#!/bin/sh  
# Set TARGET to the target machine (produce assembly files for this
# machine). The choices are:
#            i386-linux - Intel 80386 with 80387 coprocessor running Linux
#            vax  - Digital Equipment Corp's Vax
#            m68k - Motorola 68020 (and equivalents) with 68881 coprocessor
#            m88k - Motorola 88k processor
#            sparc-solaris - Sun sparc processor Solaris OS
#            sparc-sunos - Sun sparc processor SunOS OS
#            mips-irix - Mips R3000 chipset running IRIX
#            alpha-linux - DEC alpha architecture running Linux
#            arm-linux - ARM architecture running Linux
#            thumb-linux - ARM architecture subset Thumb running Linux
#TARGET=mips-irix

# Set HOST to the host machine (machine that you want the executables
# for the compiler to run on). The choices are:
#            i386-linux - Intel 80386 with 80387 coprocessor running Linux
#            vax  - Digital Equipment Corp's Vax
#            m68k - Motorola 68020 (and equivalents) with 68881 coprocessor
#            m88k - Motorola 88k processor
#            sparc-solaris - Sun sparc processor Solaris OS
#            sparc-sunos - Sun sparc processor SunOS OS
#            mips - Mips R3000 chipset running IRIX
#            alpha-linux - DEC alpha architecture running Linux
#            arm-linux - ARM architecture running Linux
#            thumb-linux - ARM architecture subset Thumb running Linux
# See the driver program for adding new machines.
#HOST=mips-irix

# Set EASE to true if you want to build the Ease verion of
# vpo; set it to nothing if you do not want Ease
EASE=
#USEEASE=TRUE
# Set USEVET to true if you want to build VET
USEVET=TRUE
# Set USEVISTA to true if you want to build VISTA
USEVISTA=TRUE
# Set USEDEOPT to true if you want to use de-optimizations
USEDEOPT=TRUE

# Set BINDIR to the directory that should contain the vpcc driver program
#BINDIR=/home/vpo-vpcc/src/vpcc-setup/bin-mips-gcc

# Set EXEROOT to the base directory that should contain the executables for
# all machines. Under this directory subdirectories for each machine will
# be created. These machine directories will contain executables that
# will run on that machine (HMACHINE).  There may be several instances
# of each executable, i.e., one for each target machine.
#EXEROOT=/home/vpo-vpcc/src/vpcc-setup/bin-mips-gcc/lib


# Set SOURCE to the directory that is the root of the directory tree
# that contains the compiler.
#SOURCE=/home/vpo-vpcc/src/vpcc-setup

# Set CC to the C compiler you wish to use to build the software
CC=gcc

# Set CCFLAGS to the C compiler flags you wish to use in building
# the software
#CFLAGS="-pedantic -g"
#CFLAGS="-ansi -pedantic -g"
CFLAGS="-g"
#CFLAGS="-O2"
CFLAGS="$CFLAGS "

if [ $USEVET"X" = "TRUEX" ]; then
	echo Using Vet
	CFLAGS="$CFLAGS -DVET"
elif [  $USEVET"X" = "FALSEX" ]; then
	echo Not Using Vet
	CFLAGS="$CFLAGS"
else
	echo "Warning!  USEVET not set, assuming VET is not needed"
	echo "export USEVET=TRUE	if/when VET is needed."
fi


# nasty hack.  all x86's should be i386's
if [ ${MACH} = "x86" ]; then
	MACH=i386
	export MACH
fi


LIBTARGET=lib.${MACH}.a

TARGET_OPTION=normal

if [ $USEEASE"X" = "TRUEX" ]; then
	echo "Enabling MEASURE flag (for EASE)"
	CFLAGS="$CFLAGS -DMEASURE"
	TARGET_OPTION=ease
	LIBTARGET=lib.${MACH}-ease.a
else
	echo "Note: export USEVISTA=TRUE if/when VISTA is needed."
fi

if [ $USEVISTA"X" = "TRUEX" ]; then
    echo "Enabling INTERACT/VET/MEASURE flags (for VISTA)"
    CFLAGS="$CFLAGS -DINTERACT -DMEASURE"
    if [ $MACH"X" = "sparcX" -a $USEDEOPT"X" = "TRUEX" ]; then
	echo "Enabling De-optimization flags (for VISTA)"
	CFLAGS="$CFLAGS -DDEOPT"
    fi
    if [ $MACH"X" = "armX" -a $USEDEOPT"X" = "TRUEX" ]; then
	echo "Enabling De-optimization flags (for VISTA)"
	CFLAGS="$CFLAGS -DDEOPT"
    fi
    TARGET_OPTION=vista
    LIBTARGET=lib.${MACH}-vista.a
else
    echo "Note: export USEVISTA=TRUE  if/when VISTA  is needed."
    echo "Currently VISTA is not enabled. "
fi

# Currently VISTA doesnt support cross-jumping. So turn on cross jumping
# only if TARGET_OPTION is normal
if [ $TARGET_OPTION"X" = "normalX" ] ; then
	CFLAGS="$CFLAGS -DCROSS"
fi
 
#### Do not make unnecessary changes below this line ####
EXEDIR=${EXEROOT}/${HOST}

# Make the auxiliary tools
echo "Building the auxiliary tools"
cd ${SOURCE}/vpo/tools/rtlprep ; make -f Makefile-${HOST} CC=${CC} CFLAGS="${CFLAGS}"
cd ${SOURCE}/vpo/tools/yacc; make CC=${CC} PARSER=${EXEDIR} CFLAGS="${CFLAGS}"
cd ${SOURCE}/vpo/tools/yyfast; make CC=${CC} CFLAGS="${CFLAGS}"
cd ${SOURCE}/vpo/tools/regtool; make CC=${CC} CFLAGS="${CFLAGS}"
cd ${SOURCE}/vpo/tools/decode; make CC=${CC} CFLAGS="${CFLAGS}"
cd ${SOURCE}/vpo/tools/encode; make CC=${CC} CFLAGS="${CFLAGS}"
cd ${SOURCE}/vpo/tools/pdtool; make CC=${CC} CFLAGS="${CFLAGS}"
cd ${SOURCE}/vpo/tools/vpoiso; make CC=${CC} CFLAGS="${CFLAGS}" vpoiso

# Install the tools
echo ; echo "Installing the auxiliary tools"
if test ! -d ${EXEDIR} ; then mkdir -p ${EXEDIR} ; fi
cp ${SOURCE}/vpo/tools/yacc/ouryacc ${EXEDIR}/ouryacc
cp ${SOURCE}/vpo/tools/yyfast/yyfast ${EXEDIR}/yyfast
cp ${SOURCE}/vpo/tools/yacc/yaccpar ${EXEDIR}/yaccpar
cp ${SOURCE}/vpo/tools/regtool/regtool ${EXEDIR}/regtool
cp ${SOURCE}/vpo/tools/vpoiso/vpoiso ${EXEDIR}/vpoiso



# also put a copy of encode/decode/vpoiso in the install dir for the
# driver scripts
cp ${SOURCE}/vpo/tools/encode/encode ${SOURCE}/../driver/install/${MACH}/dec-encode
cp ${SOURCE}/vpo/tools/decode/decode ${SOURCE}/../driver/install/${MACH}/cex-decode
cp ${SOURCE}/vpo/tools/vpoiso/vpoiso ${SOURCE}/../driver/install/${MACH}/vpoiso


case ${TARGET} in

   sparc-solaris)

	# Build the normal VPO library 
	# First build the headers (regs.h/y.tab.h)
	echo "Building headers for ${TARGET}"
	cd ${SOURCE}/vpo/${MACH}
        make -f Makefile CC=${CC} OTHER_FLAGS="${CFLAGS} -DMUSTALIGN" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool hdrs

	# The program needs to be linked with the ease library.
	if [ $TARGET_OPTION"X" = "easeX" ] || [ $TARGET_OPTION"X" = "vistaX" ] ; then
	 	echo "Building ease lib for ${TARGET}"
         	cd ${SOURCE}/vpo/${MACH}/ease
		make -f Makefile CC=${CC} ease.lib.a
	fi
        echo ; echo "Building ${TARGET_OPTION} vpo lib for ${TARGET}"
        cd ${SOURCE}/vpo/lib
        make -f Makefile-${HOST} EXTRADFLAGS="" CC=${CC} OTHER_FLAGS="${CFLAGS}"  ${LIBTARGET}

        # Build the normal optimizer
        echo ; echo "Building ${TARGET_OPTION} vpo.out for ${TARGET}"
        cd ${SOURCE}/vpo/${MACH}
        make -f Makefile EXTRADFLAGS="" CC=${CC} OTHER_FLAGS="${CFLAGS}" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc LIBS="../lib/${LIBTARGET}"
      ;;

    tms320c54x-solaris)
	# Build the normal VPO library 
	# First build the headers (regs.h/y.tab.h)
	echo "Building headers for ${TARGET}"
        cd ${SOURCE}/vpo/${MACH}
  	CFLAGS="$CFLAGS -DDISPOSE"
        make -f Makefile CC=${CC} OTHER_FLAGS="${CFLAGS} -DMUSTALIGN" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool hdrs

	if [ $TARGET_OPTION"X" = "easeX" -o $TARGET_OPTION"X" = "vistaX" ] ; then

		if [ "X"$TIDIR = "X" ]; then
			echo Please set the TIDIR environment variable to the ti-toolset directory
			exit 1
		fi
		echo "Building ${TARGET_OPTION} lib for ${TARGET}"
		EASELIB=ease/ease.lib.a
         	cd ${SOURCE}/vpo/${MACH}/ease
		make -f Makefile ease.lib.a
	fi

         echo ; echo "Building ${TARGET_OPTION} vpo lib for ${TARGET}"
         cd ${SOURCE}/vpo/lib
         make -f Makefile-${HOST} EXTRADFLAGS="" CC=${CC} OTHER_FLAGS="${CFLAGS}"  ${LIBTARGET}

         # Build the normal optimizer
         echo ; echo "Building ${TARGET_OPTION} vpo.out for ${TARGET}"
         cd ${SOURCE}/vpo/${MACH}

         make -f Makefile EXTRADFLAGS="" CC=${CC} OTHER_FLAGS="${CFLAGS}" LIBS="../lib/${LIBTARGET}" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool
      ;;

   mips-irix)
      MACH=mips
      if test ${EASE} ; then

         cd ${SOURCE}/vpo/${MACH}
         make -f Makefile-mips-irix CC=${CC} CFLAGS="${CFLAGS}" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool hdrs
         # Now build the EASE version of the VPO library
         echo ; echo "Building EASE vpo lib for ${TARGET}"
         cd ${SOURCE}/vpo/lib
         make -f Makefile lib.${MACH}-ease.a CC=${CC} CFLAGS="${CFLAGS}"

         # Build the ease report generator and ease libraries
         echo ; echo "Building the ease libraries"
         cd ${SOURCE}/vpo/${MACH}/ease
         make
         make ease.lib.a

         # Build the EASE version of VPO
         echo ; echo "Build Ease enabled vpo for ${TARGET}"
	 cd ${SOURCE}/vpo/${MACH}
         make -f Makefile vpo-ease.out EXTRADFLAGS=-DMEASURE CC=${CC} CFLAGS="${CFLAGS}" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool
     else
         cd ${SOURCE}/vpo/${MACH}
         make -f Makefile-mips-irix CC=${CC} CFLAGS="${CFLAGS}" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool hdrs
         # Build the normal VPO library 
         echo ; echo "Building normal vpo lib for ${TARGET}"
         cd ${SOURCE}/vpo/lib
         make -f Makefile-${HOST} EXTRADFLAGS="" CC=${CC} CFLAGS="${CFLAGS} -DDISPOSE -DMUSTALIGN -g"

         # Build the normal optimizer
         echo ; echo "Building normal vpo for ${TARGET}"
         cd ${SOURCE}/vpo/${MACH}
         make -f Makefile-mips-irix CC=${CC} CFLAGS="${CFLAGS}" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool
      fi
      ;;
   arm-linux)

            # Build the normal VPO library
            # First build the headers (regs.h/y.tab.h)
	    echo "Building headers for ${TARGET}"
	    cd ${SOURCE}/vpo/${MACH}
	    CFLAGS="$CFLAGS -ansi"
	    make -f Makefile CC=${CC} OTHER_FLAGS="${CFLAGS} -DMUSTALIGN" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool hdrs

	    if [ $TARGET_OPTION"X" = "easeX" -o $TARGET_OPTION"X" = "vistaX" ] ; then
                echo "Building ${TARGET_OPTION} lib for ${TARGET}"
                cd ${SOURCE}/vpo/${MACH}/ease
                make -f Makefile 
	    fi
	    
	    echo ; echo "Building ${TARGET_OPTION} vpo lib for ${TARGET}"
	    cd ${SOURCE}/vpo/lib
	    if [ $USEVISTA"X" = "TRUEX" ] ; then
	       make -f Makefile-${HOST} EXTRAFLAGS="" CC="${CC} -I../${MACH}" CFLAGS="${CFLAGS} -DMUSTALIGN -DHARDFLOAT -DALLOW_BACKWARDS_PRED -DCM_FIX_RTL -DPRED -DDISPOSE -DMUSTALIGN -DSCHED -DMAXRTS=12 -DEARLY_FRAME_SETUP -DNEEDS_MEM_TYPE -DCROSS -DEST_CYCLE " lib.${MACH}-vista.a
	    else
	       make -f Makefile-${HOST} EXTRADFLAGS="" CC=${CC} CFLAGS="${CFLAGS} -DMUSTALIGN "
	    fi
	    
         # Build the normal optimizer
	    echo ; echo "Building ${TARGET_OPTION} vpo.out for ${TARGET}"
	    cd ${SOURCE}/vpo/${MACH}
	    if [ $USEVISTA"X" = "TRUEX" ] ; then
	       make -f Makefile EXTRAFLAGS="" CC=${CC} CFLAGS="${CFLAGS}" LIBS="../lib/${LIBTARGET}" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool vpo.out
	    else
	       make -f Makefile EXTRADFLAGS="" CC=${CC} CFLAGS="${CFLAGS}" LIBS="../lib/${LIBTARGET}" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool
	    fi
	;;


   thumb-linux)
      CFLAGS="-ansi -g $CFLAGS"
      MACH=thumb
      if test ${EASE} ; then
         # Now build the EASE version of the VPO library
         echo ; echo "Building EASE vpo lib for ${TARGET}"
         cd ${SOURCE}/vpo/lib
         make -f Makefile lib.${MACH}-ease.a CC=${CC} CFLAGS="${CFLAGS} "

         # Build the ease report generator and ease libraries
         echo ; echo "Building the ease libraries"
         cd ${SOURCE}/vpo/${MACH}/ease
         make
         make ease.lib.a

         # Build the EASE version of VPO
         echo ; echo "Build Ease enabled vpo for ${TARGET}"
         cd ${SOURCE}/vpo/${MACH}
         make -f Makefile vpo-ease.out EXTRADFLAGS=-DMEASURE CC=${CC} CFLAGS="${CFLAGS} " YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool
     else
	 # build regs.h 
         cd ${SOURCE}/vpo/${MACH}
         make -f Makefile CC=${CC} CFLAGS="${CFLAGS} -DMUSTALIGN " YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool hdrs
         # Build the normal VPO library
         echo ; echo "Building normal vpo lib for ${TARGET}"
         cd ${SOURCE}/vpo/lib
         make -f Makefile-${HOST} EXTRADFLAGS="" CC=${CC} CFLAGS="${CFLAGS}  -DMUSTALIGN"

         # Build the normal optimizer
         echo ; echo "Building normal vpo for ${TARGET}"
         cd ${SOURCE}/vpo/${MACH}
         make -f Makefile CC=${CC} CFLAGS="${CFLAGS} -DMUSTALIGN " YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool
      fi
      ;;



   alpha-linux)
      MACH=alpha
      if test ${EASE} ; then
         # Now build the EASE version of the vpo library
         # Not available yet for Alpha
         # echo ; echo "Building EASE vpo lib for ${TARGET}"
         # cd ${SOURCE}/vpo/lib
         # make -f Makefile-${MACH}-linux lib.${MACH}-ease.a CC=${CC} CFLAGS="${CFLAGS}"

         # Build the ease report generator and ease libraries
         echo ; echo "Building the ease libraries"
         cd ${SOURCE}/vpo/${MACH}/ease
         make
         make ease.lib.a

         # Build the EASE version of the Alpha Linux optimizer
         echo ; echo "Build Ease enabled vpo for ${TARGET}"
         make -f Makefile vpo-ease.out EXTRADFLAGS=-DMEASURE CC=${CC} CFLAGS="${CFLAGS}" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool
     else
         # Build the normal vpo library for alpha running Linux
         echo ; echo "Building normal vpo lib for ${TARGET}"
         cd ${SOURCE}/vpo/lib
         make -f Makefile CC=${CC} CFLAGS="${CFLAGS}"

         # Build the normal Alpha Linux optimizer
         echo ; echo "Building normal vpo for ${TARGET}"
         cd ${SOURCE}/vpo/${MACH}
         make -f Makefile CC=${CC} CFLAGS="${CFLAGS}" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool
      fi
      ;;

   i386-linux)
      MACH=i386
      if test ${EASE} ; then
         # Now build the EASE version of the vpo library
         echo ; echo "Building EASE vpo lib for ${TARGET}"
         cd ${SOURCE}/vpo/${MACH}
         make -f Makefile-i386-linux EXTRADFLAGS="" CC=${CC} CFLAGS="${CFLAGS} -DMUSTALIGN" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool hdrs
         cd ${SOURCE}/vpo/lib
         make -f Makefile lib.${MACH}-ease.a CC=${CC} CFLAGS="${CFLAGS}"

         # Build the ease report generator and ease libraries
         echo ; echo "Building the ease libraries"
         cd ${SOURCE}/vpo/${MACH}/ease
         make
         make ease.lib.a

         # Build the EASE version of the i386 Linux optimizer
         echo ; echo "Build Ease enabled vpo for ${TARGET}"
	 cd ${SOURCE}/vpo/${MACH}
	 make -f Makefile-i386-linux vpo-ease.out EXTRADFLAGS=" \
-DMEASURE" CC=${CC} CFLAGS="${CFLAGS}" YYFAST=${EXEDIR}/yyfast \
YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool
     else
         # Build the normal vpo library for i386 Linux
         echo ; echo "Building normal vpo lib for ${TARGET}"
         cd ${SOURCE}/vpo/${MACH}
         make -f Makefile-i386-linux EXTRADFLAGS="" CC=${CC} CFLAGS="${CFLAGS} -DMUSTALIGN" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool hdrs
         cd ${SOURCE}/vpo/lib
         make -f Makefile-i386-linux EXTRADFLAGS="" CC=${CC} CFLAGS="${CFLAGS}"

         # Build the normal i386 Linux optimizer
         echo ; echo "Building normal vpo for ${TARGET}"
         cd ${SOURCE}/vpo/${MACH}
         make -f Makefile-i386-linux EXTRADFLAGS="" CC=${CC} CFLAGS="${CFLAGS}" YYFAST=${EXEDIR}/yyfast YACC=${EXEDIR}/ouryacc REGTOOL=${EXEDIR}/regtool
      fi
      ;;

   *)
      echo $1: unknown target processor >&2
      exit 1;
      ;;
esac



