#!/bin/sh
# Set TARGET to the target machine (produce assembly files for this
# machine). The choices are:
#            i386-linux - Intel 80386 with 80387 coprocessor running Linux
#            vax  - Digital Equipment Corp's Vax
#            m68k - Motorola 68020 (and equivalents) with 68881 coprocessor
#            m88k - Motorola 88k processor
#            sparc-solaris - Sun sparc processor Solaris OS
#            sparc-sunos - Sun sparc processor SunOS OS
#            mips-irix - Mips R3000 chipset running IRIX
#            alpha-linux - DEC alpha architecture running Linux
#TARGET=i386-linux

# Set HOST to the host machine (machine that you want the executables
# for the compiler to run on). The choices are:
#            i386-linux - Intel 80386 with 80387 coprocessor running Linux
#            vax  - Digital Equipment Corp's Vax
#            m68k - Motorola 68020 (and equivalents) with 68881 coprocessor
#            m88k - Motorola 88k processor
#            sparc-solaris - Sun sparc processor Solaris OS
#            sparc-sunos - Sun sparc processor SunOS OS
#            mips-irix - Mips R3000 chipset running IRIX
#            alpha-linux - DEC alpha architecture running Linux
# See the driver program for adding new machines.

#HOST=i386-linux

# Set BINDIR to the directory that should contain the vpcc driver program
#BINDIR=/home/vpo-vpcc/src/vpcc-setup/bin-i386-gcc

# Set EXEROOT to the base directory that should contain the executables for
# all machines. Under this directory subdirectories for each machine will
# be created. These machine directories will contain executables that
# will run on that machine (HMACHINE).  There may be several instances
# of each executable, i.e., one for each target machine.
#EXEROOT=/home/vpo-vpcc/tmp/vpcc-setup/bin-i386-gcc/lib

# Set SOURCE to the directory that is the root of the directory tree
# that contains the compiler.
#SOURCE=/home/vpo-vpcc/src/vpcc-setup

# Set CC to the C compiler you wish to use to build the software
CC=gcc

# Set CCFLAGS to the C compiler flags you wish to use in building
# the software
CFLAGS="-ansi -pedantic"
 
#### Do not make unnecessary changes below this line ####
EXEDIR=${EXEROOT}/${HOST}

# Clean out the auxiliary tools
echo "Cleaning out the auxiliary tools directories" ; echo
cd ${SOURCE}/vpo/tools/rtlprep ; make -f Makefile-${TARGET} clean
cd ${SOURCE}/vpo/tools/yacc; make clean
cd ${SOURCE}/vpo/tools/yyfast; make clean
cd ${SOURCE}/vpo/tools/regtool; make clean
cd ${SOURCE}/vpo/tools/decode; make clean
cd ${SOURCE}/vpo/tools/encode; make clean
cd ${SOURCE}/vpo/tools/pdtool; make clean
cd ${SOURCE}/vpo/tools/vpoiso; make clean

case ${TARGET} in

   sparc-solaris)
        MACH=sparc
	if [ $USEEASE"X" = "TRUEX" ]; then
	    echo ; echo "Cleaning out EASE VPO lib for ${TARGET}"	
	    cd ${SOURCE}/vpo/lib ; make -f Makefile-${TARGET} easeclean
	elif [ $USEVISTA"X" = "TRUEX" ]; then
	    echo ; echo "Cleaning out VISTA VPO lib for ${TARGET}"
            echo ${SOURCE}/vpo/lib
	    cd ${SOURCE}/vpo/lib ; make -f Makefile-${TARGET} vistaclean
	else
            # Clean the vpo library directory
	    echo ; echo "Cleaning out the normal vpo lib for ${TARGET}"
	    cd ${SOURCE}/vpo/lib ; make -f Makefile-${TARGET} clean
            # Clean the ease report generator and ease libraries directory
            # if they exist
	    if test ! -d ${SOURCE}/vpo/${MACH}/ease ; then
		echo ; echo "Clean out the ease library directory for ${TARGET}"
		cd ${SOURCE}/vpo/${MACH}/ease ; make clean
	    fi
	fi
	    
      # Clean out the normal optimizer
      echo ; echo "Clean out the vpo directory for ${TARGET}"
      cd ${SOURCE}/vpo/${MACH} ; make -f Makefile clean

      ;;


   mips-irix)
      MACH=mips
      # Clean the vpo library directory
      echo ; echo "Cleaning out the normal vpo lib for ${TARGET}"
      cd ${SOURCE}/vpo/lib ; make -f Makefile-${TARGET} clean

      # Clean the ease report generator and ease libraries directory
      # if they exist
      if test ! -d ${SOURCE}/vpo/${MACH}/ease ; then
         echo ; echo "Clean out the ease library directory for ${TARGET}"
         cd ${SOURCE}/vpo/${MACH}/ease ; make clean
      fi

      # Clean out the normal optimizer
      echo ; echo "Clean out the vpo directory for ${TARGET}"
      cd ${SOURCE}/vpo/${MACH} ; make -f Makefile clean

      ;;

   alpha-linux)
      MACH=alpha
      # Clean the vpo library directory
      echo ; echo "Cleaning out the normal vpo lib for ${TARGET}"
      cd ${SOURCE}/vpo/lib ; make -f Makefile-${TARGET} clean

      # Clean out the ease report generator and ease libraries directory
      # if they exist
      if test ! -d ${SOURCE}/vpo/${MACH}/ease ; then
         echo ; echo "Clean out the ease library directory for ${TARGET}"
         cd ${SOURCE}/vpo/${MACH}/ease ; make clean
      fi

      # Clean out the optimizer directory
      echo ; echo "Clean out the vpo directory for ${TARGET}"
      cd ${SOURCE}/vpo/${MACH} ; make -f Makefile-${TARGET} clean

      ;;

   tms320c54x-solaris)
      MACH=tms320c54x
      # Clean out the vpo library
      echo ; echo "Cleaning out the normal vpo lib for ${TARGET}"
      cd ${SOURCE}/vpo/lib ; make -f Makefile-${TARGET} clean

      # Clean out the ease report generator and ease libraries directory
      # if they exist
      if test ! -d ${SOURCE}/vpo/${MACH}/ease ; then
         echo ; echo "Clean out the ease library directory for ${TARGET}"
         cd ${SOURCE}/vpo/${MACH}/ease ; make clean
      fi

      # Clean the normal optimizer directory
      echo ; echo "Clean out the vpo directory for ${TARGET}"
      cd ${SOURCE}/vpo/${MACH} ; make -f Makefile clean

      ;;


   mips-irix)
      MACH=mips
      # Clean out the vpo library
      echo ; echo "Cleaning out the normal vpo lib for ${TARGET}"
      cd ${SOURCE}/vpo/lib ; make -f Makefile-${TARGET} clean

      # Clean out the ease report generator and ease libraries directory
      # if they exist
      if test ! -d ${SOURCE}/vpo/${MACH}/ease ; then
         echo ; echo "Clean out the ease library directory for ${TARGET}"
         cd ${SOURCE}/vpo/${MACH}/ease ; make clean
      fi

      # Clean the normal optimizer directory
      echo ; echo "Clean out the vpo directory for ${TARGET}"
      cd ${SOURCE}/vpo/${MACH} ; make -f Makefile-${TARGET} clean

      ;;

   arm-linux)
      MACH=arm
      # Clean out the vpo library
      echo ; echo "Cleaning out the normal vpo lib for ${TARGET}"
      cd ${SOURCE}/vpo/lib ; make -f Makefile-${TARGET} clean


      # Clean the normal optimizer directory
      echo ; echo "Clean out the vpo directory for ${TARGET}"
      cd ${SOURCE}/vpo/${MACH} ; make -f Makefile clean

      ;;

   thumb-linux)
      MACH=thumb
      # Clean out the vpo library
      echo ; echo "Cleaning out the normal vpo lib for ${TARGET}"
      cd ${SOURCE}/vpo/lib ; make -f Makefile-${TARGET} clean


      # Clean the normal optimizer directory
      echo ; echo "Clean out the vpo directory for ${TARGET}"
      cd ${SOURCE}/vpo/${MACH} ; make -f Makefile clean

	;;



   i386-linux)
      MACH=i386
      # Clean the vpo library directory
      echo ; echo "Cleaning out the normal vpo lib for ${TARGET}"
      cd ${SOURCE}/vpo/lib ; make -f Makefile-${TARGET} clean

      # Clean out the ease report generator and ease libraries
      # if they exist
      if test ! -d ${SOURCE}/vpo/${MACH}/ease ; then
         echo ; echo "Clean out the ease library directory for ${TARGET}"
         cd ${SOURCE}/vpo/${MACH}/ease ; make clean
      fi

      # Clean the optimizer directory
      echo ; echo "Clean out the vpo directory for ${TARGET}"
      cd ${SOURCE}/vpo/${MACH} ; make -f Makefile-${TARGET} clean

      ;;

   *)
      echo $1: unknown target processor >&2
      exit 1;
      ;;
esac



