/*
 * Machine dependent deoptimization data and functions
 */

/* 
 * $Author: jdh8d $
 * $Date: 2004/01/21 18:39:44 $
 * $Revision: 1.1 $
 */

#include "vpo.h"
#include "deopt.h"  /* Machine dependent */

#ifdef INTERACT
#ifdef DEOPT

rlist no_replace_regs;

int next_ireg,
    next_freg;



/*
 * init_deopt - Initialize machine dependent data for deoptimizations
 */
void init_deopt(rlist *rl)
{
    char buf[REGSTRING];

    next_ireg = 16;
    next_freg = 16;

    *rl = rinit();

    /* FP*/
    makereg(buf, RTREG, 11);
    rinsert(buf, rl);

    /*stack pointer*/
    makereg(buf, RTREG, 13);
    rinsert(buf, rl);
    
    /*link register*/
    makereg(buf, RTREG, 14);
    rinsert(buf, rl);

    /* program counter*/
    makereg(buf, RTREG, 15);
    rinsert(buf, rl);
}


/*
 * next_pseudo - construct and return the next pseudo reg for this register
 *               reference. Return the new register number
 */
int next_pseudo(char o_reg[], char n_reg[], int rnum)
{
  int type;

  type = regtype(o_reg);

  if(!rnum){
    
    switch(type){

    case RTREG:
      rnum = next_ireg++;
      break;
    case FTREG:
      rnum = next_freg++;
      break;
    case DTREG:
      if (next_freg % 2)
	next_freg += 3;
      else
	next_freg += 2;
      rnum = next_freg - 2;
      break;
    case CTREG:
      rnum = regnum(o_reg);
      break;
    }
  }

  makereg(n_reg, type, rnum);
  return rnum;
}


/*
 * new_pseudoreg - Return the number for the next available pseudoregister
 */
int new_pseudoreg(int type)
{
    switch(type)
    {
        case RTREG:
            regallocptr++;
            return next_ireg++;
            break;
        case FTREG:
            regallocptr++;
            return next_freg++;
            break;
        case DTREG:
            regallocptr += 2;
            if (next_freg % 2)
                next_freg += 3;
            else
                next_freg += 2;
            return (next_freg-2);
            break;
        default:
            return -1;
    }
}

int regsize(int type)
{
    switch(type)
    {
        case DTREG:
            return 4;
        case RTREG:
        case FTREG:
        default:
            return 4;
    }
}

char getmemchar(int type)
{
    switch(type)
    {
        case RTREG:
            return 'R';
        case FTREG:
            return 'F';
        case DTREG:
            return 'D';
        default:
            fprintf(stderr,
                    "Error (getmemchar) - Bad register type %d\n", type);
            exit(1);
    }
}


#endif  /* DEOPT */
#endif  /* INTERACT */

