/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/arm/ease/ease_d.c,v $
 $Date: 2003/11/09 14:19:51 $
 $Revision: 1.1 $
 $Name:  $

 Log of changes:

 $Log: ease_d.c,v $
 Revision 1.1  2003/11/09 14:19:51  jwd
 Initial version from SPARC--incomplete for arm

 Revision 1.2  1998/04/07 20:06:12  jwd
 First ANSI C version.


*************************************************************************/

#include <stdio.h>
#include "ease.h"
#include "ease_d.h"
#include "ease_d2.h"

/*
 * The following functions support cache simulations
 */

/*
 * Structure that represents a cache line
 */
struct c_line {
   unsigned int tag;
   unsigned long lru;
};

/*
 * This array represents the maximum cache possible
 */
static struct c_line cache[MAXNUMSETS][MAXSETSIZE];

int __ease_cline_shift;	/* Log2 of cache line size in bytes */
int __ease_contextflag;	/* Number of context switches made  */
unsigned long __ease_c_switch = 0;	/* Time since last context switch   */
unsigned long __ease_c_hits = 0;	/* Total number of cache hits       */
unsigned long __ease_c_misses = 0;	/* Total number of cache misses     */

static unsigned long context_switch = 0;	/* Total number of context
						   switches */
static unsigned long lru_count;	/* Used to find the LRU cache line  */
static unsigned numsets;/* Number of sets in the cache      */
static unsigned setsize;/* Number of lines per set in cache */
static unsigned int linesize;	/* Size of each line in bytes       */
static char title[100];	/* Name of this simulation          */

/*
 * ease_init_cacher - initialize cache trace simulation
 */
void __ease_init_cacher(void) {

   char *configname;
   char *getenv(const char *);
   void read_config(), c_init();

   if (!(configname = getenv("TRACE_CONFIG")))
      configname = TRACE_IN;
   read_config(configname);

   c_init();
}

/*
 * ease_read - process a cache read
 */
void __ease_read(unsigned int data_size, long unsigned int address) {

   register unsigned long index;
   register unsigned int ctag;
   register struct c_line *cptr;
   register unsigned int i, j;
   unsigned long c_address;
   unsigned long maxlru;
   unsigned int numline;
   int pick;
   void c_init();

   /*
      Obtain the starting cache address and number of lines to read
   */
   c_address = address >> __ease_cline_shift;
   numline = ((address + data_size - 1) >> __ease_cline_shift) - c_address + 1;

   /*
      Obtain the starting cache index and cache tag line numbers
   */
   index = c_address % numsets;
   ctag = c_address / numsets;

   /*
      Determine if each line is in cache and update the appropriate information
   */
   for (i = 0; i < numline; i++) {

      /*
         Look for the item in the cache.  If the item is found, update its lru
         field and note that a cache hit has taken place.
      */
      for (j = 0; j < setsize; j++) {
	 cptr = &cache[index][j];
	 if (cptr->tag == ctag) {
	    cptr->lru = lru_count++;
	    __ease_c_switch += HITTIME;
	    __ease_c_hits++;
	    goto next;
	 }
      }

      /*
         If the item is not found, then take the penalty of a cache miss and
         load the item into the cache, bumping out the lru line, if no open
         slots remain.
      */
      __ease_c_switch += MISSTIME;
      __ease_c_misses++;
      maxlru = lru_count;
      pick = 0;
      for (j = 0; j < setsize; j++) {
	 cptr = &cache[index][j];
	 if (cptr->tag == (unsigned int) -1) {
	    pick = j;
	    break;
	 }
	 if (cptr->lru < maxlru) {
	    maxlru = cptr->lru;
	    pick = j;
	 }
      }
      cache[index][pick].tag = ctag;
      cache[index][pick].lru = lru_count++;

      /*
         Advance to the next index
      */
next:
      if (++index == numsets) {
	 index = 0;
	 ctag++;
      }
   }

   /*
      If it is time to do a contex switch, re-initialize the cache
   */
   if (__ease_c_switch > SWITCHTIME) {
      if (__ease_contextflag) {
	 c_init();
	 context_switch++;
      }
      __ease_c_switch = 0;
   }
}

/*
 * ease_write - process a cache write operation
 */
void __ease_write(unsigned int data_size, long unsigned int address) {

   register unsigned long index;
   register unsigned int ctag;
   register struct c_line *cptr;
   register unsigned int i, j;
   unsigned long c_address;
   unsigned int numline;
   void c_init();

   /*
      Obtain the starting cache address and number of lines to write
   */
   c_address = address >> __ease_cline_shift;
   numline = ((address + data_size - 1) >> __ease_cline_shift) - c_address + 1;

   /*
      Obtain the starting cache index and cache tag line numbers
   */
   index = c_address % numsets;
   ctag = c_address / numsets;

   /*
      Determine if each line is in cache and update the appropriate information
   */
   for (i = 0; i < numline; i++) {

      /*
         Look for the item in the cache.  If the item is found, update its lru
         field and note that a cache hit has taken place.
      */
      for (j = 0; j < setsize; j++) {
	 cptr = &cache[index][j];
	 if (cptr->tag == ctag) {
	    cptr->lru = lru_count++;
	    __ease_c_switch += HITTIME;
	    __ease_c_hits++;
	    goto next;
	 }
      }

      /*
         If the item was not in the cache, then we must take the penalty of a
         cache miss.
      */
      __ease_c_switch += MISSTIME;
      __ease_c_misses++;

      /*
         Advance to the next index
      */
next:
      if (++index == numsets) {
	 index = 0;
	 ctag++;
      }
   }

   /*
      If it is time to do a contex switch, re-initialize the cache
   */
   if (__ease_c_switch > SWITCHTIME) {
      if (__ease_contextflag) {
	 c_init();
	 context_switch++;
      }
      __ease_c_switch = 0;
   }
}

/*
 * ease_end_cacher - finish cache simulation by printing out results
 */
void __ease_end_cacher(char *my_name, char *suffix) {

   FILE *output_file;
   float hit_ratio;
   unsigned long accesses;
   char outname[100];
   FILE *fopen(const char *, const char *);
   int fclose(FILE *);
   void error();

   /*
      Open the cache results output file
   */
   (void) sprintf(outname, "%s.%s", my_name, suffix);
   if (!(output_file = fopen(outname, "w")))
      error("cannot open cache simulation output file");

   (void) fprintf(output_file, "%s", title);
   (void) fprintf(output_file, "Cache contains %u ", numsets);
   (void) fprintf(output_file, "%u-way set associative entries ", setsize);
   (void) fprintf(output_file, "of line size %u bytes\n", linesize);
   (void) fprintf(output_file, "Instruction buffer disabled\n");
   (void) fprintf(output_file, "Prefetch disabled\n");

   if (accesses = __ease_c_hits + __ease_c_misses)
      hit_ratio = ((float) __ease_c_hits) / (float) accesses;
   else
      hit_ratio = 0.0;

   (void) fprintf(output_file, "Total hits      : %lu\n", __ease_c_hits);
   (void) fprintf(output_file, "Total misses    : %lu\n", __ease_c_misses);
   (void) fprintf(output_file, "Total accesses  : %lu\n", accesses);
   (void) fprintf(output_file, "Hit ratio       : %8.6f\n", hit_ratio);
   (void) fprintf(output_file, "Miss ratio      : %8.6f\n", 1.0 - hit_ratio);
   (void) fprintf(output_file, "Context switches: %lu\n\n", context_switch);
   (void) fclose(output_file);
}

/*
 * read_config - read cache configuration information
 */
static void read_config(char *configfn) {

   unsigned int linesz;
   char answer[100];
   FILE *cfile;
   FILE *fopen(const char *, const char *);
   int fclose(FILE *);
   void error();

   if (!(cfile = fopen(configfn, "r")))
      error("error opening configuration file");

   if (!fgets(title, 100, cfile))
      error("error reading title in config file");

   if (fscanf(cfile, "Number of sets: %u\n", &numsets) != 1)
      error("error reading numsets in config file");
   if (numsets > MAXNUMSETS)
      error("number of sets requested exceeds MAXNUMSETS");

   if (fscanf(cfile, "Set size: %u\n", &setsize) != 1)
      error("error reading setsize in config file");
   if (setsize > MAXSETSIZE)
      error("set size requested exceeds MAXSETSIZE");

   if (fscanf(cfile, "Line size: %u\n", &linesize) != 1)
      error("error reading linesize in config file");
   if ((linesz = linesize) & (linesize - 1))
      error("line size must be a power of two");
   for (__ease_cline_shift = -1; linesz; linesz >>= 1)
      __ease_cline_shift++;

   if (fscanf(cfile, "Instruction buffer: %s\n", answer) != 1)
      error("no instruction buffer flag found in config file");
   if (*answer == 'y')
      error("instruction buffer cannot be enabled with this version\n");

   if (fscanf(cfile, "Prefetch: %s\n", answer) != 1)
      error("no prefetch flag found in config file");
   if (*answer == 'y')
      error("prefetch cannot be enabled with this version\n");

   if (fscanf(cfile, "Context switching: %s\n", answer) != 1)
      error("no context switch flag found in config file");
   __ease_contextflag = (*answer == 'y') ? 1 : 0;

   (void) fclose(cfile);
}

/*
 * c_init - initialize cache
 */
static void c_init(void) {

   register int ind1, ind2;

   for (ind1 = 0; ind1 < numsets; ind1++)
      for (ind2 = 0; ind2 < setsize; ind2++)
	 cache[ind1][ind2].tag = (unsigned int) -1;
   lru_count = 0;
}

/*
 * The following functions support the main tasks
 */

/*
 * error - print error message and exit
 */
static void error(char *s) {

   int fflush(FILE *);
   void exit(int);

   (void) fprintf(stderr, "ease: %s\n", s);
   (void) fflush(stderr);
   exit(1);
}
