/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/arm/ease/ease_di.c,v $
 $Date: 2003/11/09 14:19:51 $
 $Revision: 1.1 $
 $Name:  $

 Log of changes:

 $Log: ease_di.c,v $
 Revision 1.1  2003/11/09 14:19:51  jwd
 Initial version from SPARC--incomplete for arm

 Revision 1.2  1998/04/07 20:06:30  jwd
 First ANSI C version.


*************************************************************************/

#include <stdio.h>
#include "ease.h"
#include "ease_d.h"
#include "ease_di.h"
#include "ease_di2.h"

struct ht_entry {
   int *ptr;
   int *info;
};

static struct ht_entry hash_table[MAXBLOCKS];
static int info_table[MAXBLOCKS * 5];

extern int __ease_contextflag;
extern int __ease_cline_shift;
extern unsigned long __ease_c_switch;
extern unsigned long __ease_c_hits;
extern unsigned long __ease_c_misses;

/*
 * The following functions support unified cache simulations
 */

/*
 * ease_unified_init - called to initialize address values for all blocks
 */
void __ease_unified_init(void) {

   register int i;
   int *__ease_main_ptr();
   void module_init();

   /*
      Initialize the block information pointer hash table
   */
   for (i = 0; i < MAXBLOCKS; hash_table[i++].ptr = (int *) NULL);

   /*
      Initialize simulation information one module at a time
   */
   module_init(__ease_main_ptr());
}

/*
 * ease_unified_inst - fetch instructions for unified cache simulation
 */
void __ease_unified_inst(register int *bptr) {

   static int *expected_ptr = (int *) &hash_table[0];
   register int insts, offset, i, *ptr;
   void __ease_read(), error();

   /*
      Ensure that this pointer item is the expected one, otherwise, we must
      look for the information for this block's pointer
   */
   if (bptr != (int *) *expected_ptr) {
      register int hash;

      hash = (((int) bptr) >> 2) & (MAXBLOCKS - 1);
      for (;;) {
	 if (bptr == hash_table[hash].ptr) {
	    ptr = hash_table[hash].info;
	    break;
	 }
	 if (!hash_table[hash].ptr)
	    error("function block not in hash table");
	 hash = (hash + 1) & (MAXBLOCKS - 1);
      }
   }
   else
      ptr = expected_ptr;

   /*
      Determine the number of instructions in this block
   */
   insts = *bptr++;

   /*
      If executing this block could possibly cause a context switch, then make
      each call to the cache simulator individually
   */
   if (__ease_contextflag && (__ease_c_switch + *(ptr + 3)) > SWITCHTIME) {
      offset = *(ptr + 1);

#ifdef SAMESIZE
      for (i = 0; i < insts; i++) {
	 __ease_read(SAMESIZE, (unsigned long) offset);
	 offset += SAMESIZE;
      }
#else
      for (i = 0; i < insts; i++) {
	 int size;

	 __ease_read((unsigned int) (size = *bptr++), (unsigned long) offset);
	 offset += size;
      }
#endif
   }

   /*
      Otherwise we try a somewhat faster approach, where we call the cache
      routine only once and make up for the extra references caused by the
      instructions that overlap cache line boundaries with some fancy
      calculations.
   */
   else {
      insts = __ease_c_hits + __ease_c_misses;
      __ease_read((unsigned int) *(ptr + 4), (unsigned long) *(ptr + 1));
      __ease_c_hits +=
       (offset = (*(ptr + 2) - ((__ease_c_hits + __ease_c_misses) - insts)));
      __ease_c_switch += offset * HITTIME;
   }

   /*
      Save the pointer to the next block as the most likely candidate for the
      next call.
   */
   expected_ptr = ptr + 5;
}

/*
 * ease_uread - process a memory read access for unified cache simulation
 */
void __ease_uread(unsigned int data_size, long unsigned int address, register int *ptr) {

   void __ease_unified_inst(register int *bptr), __ease_read();

   /*
      If there are instructions to simulate at this time, do them before the
      memory read access.
   */
   if (ptr)
      __ease_unified_inst(ptr);

   /*
      Now we perform the requested read access
   */
   __ease_read(data_size, address);
}

/*
 * ease_uwrite - process a memory write for unified cache simulation
 */
void __ease_uwrite(unsigned int data_size, long unsigned int address, register int *ptr) { 

   void __ease_unified_inst(register int *bptr), __ease_write();

   /*
      If there are instructions to simulate at this time, do them before the
      memory write access.
   */
   if (ptr)
      __ease_unified_inst(ptr);

   /*
      Now we perform the requested write access
   */
   __ease_write(data_size, address);
}

/*
 * module_init - initialize a module for instruction cache simulation
 */
static void module_init(register int *ptr) {

   int blocks, i, hash, *mptr, *nptr, *(*fptr) ();
   void function_init();

   /*
      Obtain a pointer to the first function in the module
   */
   mptr = (int *) *(ptr + 1);

   /*
      If this module has already been initialized, then return
   */
   hash = (((int) (mptr + 3)) >> 2) & (MAXBLOCKS - 1);
   for (;;) {
      if (hash_table[hash].ptr == (mptr + 3))
	 return;
      if (!hash_table[hash].ptr)
	 break;
      hash = (hash + 1) & (MAXBLOCKS - 1);
   }

   /*
      Make a pass through each function if the module and initialize them
   */
   while (*mptr) {

      /*
         Initialize the information for this function
      */
      function_init(mptr + 2);

      /*
         Advance to the next function in this module
      */
      mptr = (int *) *mptr;
   }

   /*
      Now start back at the first function in the module and ensure that each
      item in the call graph is also visited.
   */
   mptr = (int *) *(ptr + 1);
   while (*mptr) {

      /*
         Determine the number of blocks in this function and skip past each
         one.  We want to get to the call graph data information at the end of
         the function information block.
      */
      ptr = mptr + 2;
      blocks = *ptr++;

#ifdef SAMESIZE
      ptr += blocks;
#else
      for (i = 0; i < blocks; i++)
	 ptr += (*ptr) + 1;
#endif

      /*
         Determine the number of items in the call graph and call each one to
         ensure that their information has been initialized.
      */
      blocks = *ptr++;
      for (i = 0; i < blocks; i++) {
	 fptr = (int *(*) ()) *ptr++;
	 if (nptr = fptr())
	    module_init(nptr);
      }

      /*
         Advance to the next function in this module
      */
      mptr = (int *) *mptr;
   }
}

/*
 * function_init - initialize a function for instruction cache simulation
 */
static void function_init(register int *ptr) {

   static int offset = 0;
   static int numblocks = 0;
   static int *bptr = &info_table[0];
   register j, insts, i, blocks, address;
   int hash, refs, lines;

#ifndef SAMESIZE
   int size;
#endif

   void error();

   /*
      Get the number of blocks in this function
   */
   blocks = *ptr++;

   /*
      For each block in the function
   */
   for (i = 0; i < blocks; i++) {

      /*
         Ensure that the maximum number of blocks is not exceeded and insert
         the item in the hash table for quick reference later on.
      */
      numblocks++;
      if (numblocks >= MAXBLOCKS - 1)
	 error("too many basic blocks");
      hash = (((int) ptr) >> 2) & (MAXBLOCKS - 1);
      for (;;) {
	 if (!hash_table[hash].ptr) {
	    hash_table[hash].ptr = ptr;
	    hash_table[hash].info = bptr;
	    break;
	 }
	 hash = (hash + 1) & (MAXBLOCKS - 1);
      }
      *bptr++ = (int) ptr;

      /*
         Get the number of instructions in the block
      */
      insts = *ptr++;

      /*
         Initialize the field that indicates the starting address of this basic
         block.
      */
      *bptr++ = address = offset;

      /*
         Initially, assume that the number of references will be equal to the
         number of instructions.
      */
      refs = insts;

      /*
         For each instruction, determine its size.  Then, looking at the
         starting and ending address of the instruction, determine if the
         instruction crosses over a cache line.  If it does, then the hardware
         will have to perform two memory references to fetch it. If this is the
         case, increment the number of references that are made by this block
         by one.  Calculate the address of the next instruction by adding the
         size to the starting address.
      */

#ifdef SAMESIZE
      for (j = 0; j < insts; j++) {
	 if ((address >> __ease_cline_shift) !=
	     ((address + SAMESIZE - 1) >> __ease_cline_shift))
	    refs++;
	 address += SAMESIZE;
      }
#else
      for (j = 0; j < insts; j++) {
	 size = *ptr++;
	 if ((address >> __ease_cline_shift) !=
	     ((address + size - 1) >> __ease_cline_shift))
	    refs++;
	 address += size;
      }
#endif

      /*
         Store the total number of memory references that will be made when
         fetching all of the instructions in this block.
      */
      *bptr++ = refs;

      /*
         Now we are going to determine the worst-case scenario time that it
         would take to fetch this block.  This time is calculated by assuming
         that every new cache line referenced will result in a cache miss.
         Consider the following figure that indicates the starting address of
         the first instruction in the block (offset), the ending address
         (address) and the boundary of the cache lines between them.
      
      offset                           address |
         | |<- a1 ->|                                 |<- a2 ->| |        |
         |        | \/       \/                                \/       \/
         -------------------------------------------------------------- | cache
         line #1  | cache line #2   | cache line #3  |
         --------------------------------------------------------------
      
      Here, assume that the address of the first instruction in the block
         starts at "offset", which is somewhere in the middle of cache line #1.
         Assume also that the address of the instruction following the last
         instruction in the block falls in the middle of cache line #3.  First,
         we obtain the difference between the "offset" and "address" points.
      */
      lines = address - offset;

      /*
         Now we calculate the value of "a1", which is the distance from the
         start of cache line #1 to "offset".  If offset happens to start at the
         very beginning of a cache line, this value will be zero. We add this
         value to our previous difference to obtain the number of bytes between
         the start of cache line #1 and "address".
      */
      lines += (offset & ((1 << __ease_cline_shift) - 1));

      /*
         Now we calculate the value of "a2", which is the distance from
         "address" to the start of the cache line after cache line #3. Note
         that this value may also turn out to be zero.  We add this to our
         total so far to obtain the total number of bytes from the start of
         cache line #1 to the start of the cache line which follows cache line
         #3.
      */
      if (address & ((1 << __ease_cline_shift) - 1))
	 lines += ((1 << __ease_cline_shift) -
		   (address & ((1 << __ease_cline_shift) - 1)));

      /*
         Now we determine how many lines our block touches by dividing the
         number of bytes that we have calculated by the cache line size.
      */
      lines >>= __ease_cline_shift;

      /*
         Now we can calcutae the worst-case time to fetch this block.  If we
         assume that none of the cache lines will be present when we start the
         block, it will take MISSTIME times the number of cache lines to fetch
         these.  Then, the rest of the total memory references made by this
         block will take HITTIME cycles each.  Thus, we have our worst-time
         calculation.
      */
      *bptr++ = lines * MISSTIME + (refs - lines) * HITTIME;

      /*
         The size of this block is the starting offset minus the ending offset.
         Update the offset value to allow the next block to be placed following
         this one.
      */
      *bptr++ = address - offset;
      offset = address;
   }
}

/*
 * error - print error message and exit
 */
static void error(char *s) {

   int fflush(FILE *);
   void exit(int);

   (void) fprintf(stderr, "ease: %s\n", s);
   (void) fflush(stderr);
   exit(1);
}
