/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/arm/ease/ease_f.c,v $
 $Date: 2003/11/09 14:19:51 $
 $Revision: 1.1 $
 $Name:  $

 Log of changes:

 $Log: ease_f.c,v $
 Revision 1.1  2003/11/09 14:19:51  jwd
 Initial version from SPARC--incomplete for arm

 Revision 1.4  2003/08/21 20:30:30  jdh8d
 changes to ease for tms

 Revision 1.3  2003/08/20 18:13:00  jdh8d
 changes to get genreport, etc working for tms

 Revision 1.2  1998/04/07 20:07:39  jwd
 First ANSI C version.


*************************************************************************/

#include <stdio.h>
#include "../instdef.h"
#include "../recdef.h"
#include "ease.h"
#include "ease_f.h"

/*
 * This file contains the routines that dump out the execution frequency
 * counter information for a program.
 */

int __ease_jrelnottaken[NUMJREL];

/*
 * ease_dump_info - called to write dynamic information list to dump file
 */
void __ease_dump_info(char *my_name) {

   FILE *fp;
   int i;
   char filename[100];
   FILE *fopen(const char *, const char *);
   int (*__ease_main_ptr())[], fclose(FILE *);
   void emit_info(), error();

   /*
      Open the freqency count information output file
   */
   (void) sprintf(filename, "%s.cnt", my_name);
   if (!(fp = fopen(filename, "w")))
      error("cannot open instruction frequency output file");

   /*
      Start file by dumping out relative branch not taken counters
   */
   for (i = 0; i < NUMJREL; i++)
      (void) fprintf(fp, "%d ", __ease_jrelnottaken[i]);
   (void) fprintf(fp, "\n");

   /*
      Emit information for one function at a time
   */
   emit_info(fp, __ease_main_ptr());

   /*
      Close the frequency count information output file
   */
   (void) fclose(fp);
}

/*
 * emit_info - recursively dump out the dynamic execution information
 */
static void emit_info(FILE * fp, int (*ptr)[]) {

   int (*(*fptr) ())[], (*mptr)[], (*nptr)[];
   int i, j, k, equiv_classes, instructions;
   char *fn, funcname[256];

   /*
      If we have already visited this function, then we don't need to dump out
      its contents again.
   */
   if ((*ptr)[0])
      return;

   /*
      Obtain the address of the first function in the module.  Check it's
      visited flag.  If it is set, then we are already at work on this module,
      so we can return.
   */
   mptr = (int (*)[]) (*ptr)[2];
   if (*((*mptr) - 1))
      return;

   while ((*mptr)[0]) {

      /*
         Indicate that this information block has been dumped, this avoids
         duplicates and prevents chains in the call graph from causing infinite
         loops.
      */
      *((*mptr) - 1) = 1;

      /*
         Emit start of function marker and number of equivalence classes
         information.
      */
      (void) fprintf(fp, "-1 %d\n", equiv_classes = (*mptr)[2]);

      /*
         For each equivalence class, emit size and execution counts.  Size is
         not needed if all of the instructions on the target machine are of the
         same size.
      */

#ifdef SAMESIZE
      k = 3;
      for (i = 0; i < equiv_classes; i++)
	 (void) fprintf(fp, "%d\n", (*mptr)[k++]);
#else
      j = 3;
      k = equiv_classes + 3;
      for (i = 0; i < equiv_classes; i++)
	 (void) fprintf(fp, "%d %d\n", (*mptr)[k++], (*mptr)[j++]);
#endif

      /*
         Extract the current function name and write it out
      */
      fn = funcname;
      for (;;) {
	j = (*mptr)[k++];
	*fn++ = j;
	if (!j)
 	  break;
      }
      (void) fprintf(fp, "%s\n", funcname);

      /*
         Emit the remainder of the function information
      */
      do {
	 (void) fprintf(fp, "%d\n", (*mptr)[k]);
      } while ((*mptr)[k++] != -3);

      /*
         Emit the loop execution count information
      */
      while ((*mptr)[k] != -4) {
	 (void) fprintf(fp, "%d\n", (*mptr)[k++]);
	 (void) fprintf(fp, "%d\n", (*mptr)[k++]);
	 fn = funcname;
	 for (;;) {
	    j = (*mptr)[k++];
	    *fn++ = j;
	    if (!j)
	       break;
	 }
	 (void) fprintf(fp, "%s\n", funcname);
      }
      (void) fprintf(fp, "%d ", (*mptr)[k++]);

      /*
         Output instruction information
      */
      (void) fprintf(fp, "%d\n", instructions = (*mptr)[k++]);
      for (i = 0; i < instructions; i++) {
	 for (j = 1; j < INSTINFONUM; j++)
	    (void) fprintf(fp, "%d ", (*mptr)[k++]);
	 (void) fprintf(fp, "%d\n", (*mptr)[k++]);
      }

      /*
         Now we call the functions that this function referenced
      */
      j = (*mptr)[k++];
      for (i = 0; i < j; i++) {
	 fptr = (int (*(*) ())[]) (*mptr)[k++];
	 if (nptr = fptr())
	    emit_info(fp, nptr);
      }

      mptr = (int (*)[]) (*mptr)[0];
   }
}

/*
 * The following function support the main tasks
 */

/*
 * error - print error message and exit
 */
static void error(char *s) {

   int fflush(FILE *);
   void exit(int);

   (void) fprintf(stderr, "ease: %s\n", s);
   (void) fflush(stderr);
   exit(1);
}
