/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/arm/ease/ease_i.c,v $
 $Date: 2003/11/09 14:19:51 $
 $Revision: 1.1 $
 $Name:  $

 Log of changes:

 $Log: ease_i.c,v $
 Revision 1.1  2003/11/09 14:19:51  jwd
 Initial version from SPARC--incomplete for arm

 Revision 1.4  2002/08/07 02:09:59  clc5q
 Update EASE code per changes from David Whalley to make the SPARC EASE
 target build and execute properly.

 Revision 1.3  2000/09/19 21:46:08  jwd
 The version checked in was from a bogus repository.

 Revision 1.2  1998/04/07 20:07:51  jwd
 First ANSI C version.


*************************************************************************/

#include <stdio.h>

#include "ease.h"
#include "ease_d.h"
#include "ease_i.h"

extern int __ease_contextflag;
extern int __ease_cline_shift;
extern unsigned long __ease_c_switch;
extern unsigned long __ease_c_hits;
extern unsigned long __ease_c_misses;

double __scratch_dump[128];
  /* for saving scratch regs before a cache simulation call. */


/*
 * The following functions support instruction cache simulations
 */

/*
 * ease_trace_init - called to initialize address values for all blocks
 */
void __ease_trace_init(void) {

   int *__ease_main_ptr();
   void module_init();

   /*
      Initialize simulation information one module at a time
   */
   module_init(__ease_main_ptr());
}

/*
 * ease_trace_inst - called to indicate that a block is executing
 */
void __ease_trace_inst(register int *ptr) {
   register int insts, offset, i;
   void __ease_read();

   /*
      If executing this block could possibly cause a context switch, then make
      each call to the cache simulator individually
   */
   if (__ease_contextflag && (__ease_c_switch + *(ptr + 3)) > SWITCHTIME) {
      insts = *ptr;
      offset = *(ptr + 1);

#ifdef SAMESIZE
      for (i = 0; i < insts; i++) {
	 __ease_read(SAMESIZE, (unsigned long) offset);
	 offset += SAMESIZE;
      }
#else
      ptr += 5;
      for (i = 0; i < insts; i++) {
	 int size;

	 __ease_read((unsigned int) (size = *ptr++), (unsigned long) offset);
	 offset += size;
      }
#endif
   }

   /*
      Otherwise we try a somewhat faster approach, where we call the cache
      routine only once and make up for the extra references caused by the
      instructions that overlap cache line boundaries with some fancy
      calculations.
   */
   else {
      insts = __ease_c_hits + __ease_c_misses;
      __ease_read((unsigned int) *(ptr + 4), (unsigned long) *(ptr + 1));
      __ease_c_hits +=
       (offset = (*(ptr + 2) - ((__ease_c_hits + __ease_c_misses) - insts)));
      __ease_c_switch += offset * HITTIME;
   }
}

/*
 * module_init - initialize a module for instruction cache simulation
 */
static void module_init(register int *ptr) {

   int blocks, i, *mptr, *nptr, *(*fptr) ();
   void function_init();

   /*
      If this function has already been initialized, then return
   */
   if (*ptr)
      return;

   /*
      Obtain a pointer to the first function in the module
   */
   mptr = (int *) *(ptr + 2);

   /*
      Make a pass through each function if the module and initialize them
   */
   while (*mptr) {

      /*
         Indicate that function has been initialized
      */
      *(mptr - 1) = 1;
      function_init(mptr + 2);

      /*
         Advance to the next function in this module
      */
      mptr = (int *) *mptr;
   }

   /*
      Now start back at the first function in the module and ensure that each
      item in the call graph is also visited.
   */
   mptr = (int *) *(ptr + 2);
   while (*mptr) {

      /*
         Determine the number of blocks in this function and skip past each
         one.  We want to get to the call graph data information at the end of
         the function information block.
      */
      ptr = mptr + 2;
      blocks = *ptr++;

#ifdef SAMESIZE
      ptr += 5 * blocks;
#else
      for (i = 0; i < blocks; i++)
	 ptr += 5 + *ptr;
#endif

      /*
         Determine the number of items in the call graph and call each one to
         ensure that their information has been initialized.
      */
      blocks = *ptr++;
      for (i = 0; i < blocks; i++) {
	 fptr = (int *(*) ()) *ptr++;
	 if (nptr = fptr())
	    module_init(nptr);
      }

      /*
         Advance to the next function in this module
      */
      mptr = (int *) *mptr;
   }
}

/*
 * function_init - initialize a function for instruction cache simulation
 */
static void function_init(register int *ptr) {

   static int offset = 0;
   register j, insts, i, blocks, address;
   int *sptr, refs, lines;

#ifndef SAMESIZE
   int size;
#endif

   /*
      Get the number of blocks in this function
   */
   blocks = *ptr++;

   /*
      For each block in the function
   */
   for (i = 0; i < blocks; i++) {

      /*
         Get the number of instructions in the block
      */
      insts = *ptr;

      /*
         Initialize the field that indicates the starting address of this basic
         block.
      */
      *(ptr + 1) = address = offset;

      /*
         Initially, assume that the number of references will be equal to the
         number of instructions.
      */
      refs = insts;

      /*
         Point to where the size table for each instruction in this block
         starts or to the next block's information if we know that all of the
         instructions are of the same size.
      */
      sptr = ptr + 5;

      /*
         For each instruction, determine its size.  Then, looking at the
         starting and ending address of the instruction, determine if the
         instruction crosses over a cache line.  If it does, then the hardware
         will have to perform two memory references to fetch it. If this is the
         case, increment the number of references that are made by this block
         by one.  Calculate the address of the next instruction by adding the
         size to the starting address.
      */

#ifdef SAMESIZE
      for (j = 0; j < insts; j++) {
	 if ((address >> __ease_cline_shift) !=
	     ((address + SAMESIZE - 1) >> __ease_cline_shift))
	    refs++;
	 address += SAMESIZE;
      }
#else
      for (j = 0; j < insts; j++) {
	 size = *sptr++;
	 if ((address >> __ease_cline_shift) !=
	     ((address + size - 1) >> __ease_cline_shift))
	    refs++;
	 address += size;
      }
#endif

      /*
         Store the total number of memory references that will be made when
         fetching all of the instructions in this block.
      */
      *(ptr + 2) = refs;

      /*
         Now we are going to determine the worst-case scenario time that it
         would take to fetch this block.  This time is calculated by assuming
         that every new cache line referenced will result in a cache miss.
         Consider the following figure that indicates the starting address of
         the first instruction in the block (offset), the ending address
         (address) and the boundary of the cache lines between them.
      
      offset                           address |
         | |<- a1 ->|                                 |<- a2 ->| |        |
         |        | \/       \/                                \/       \/
         -------------------------------------------------------------- | cache
         line #1  | cache line #2   | cache line #3  |
         --------------------------------------------------------------
      
      Here, assume that the address of the first instruction in the block
         starts at "offset", which is somewhere in the middle of cache line #1.
         Assume also that the address of the instruction following the last
         instruction in the block falls in the middle of cache line #3.  First,
         we obtain the difference between the "offset" and "address" points.
      */
      lines = address - offset;

      /*
         Now we calculate the value of "a1", which is the distance from the
         start of cache line #1 to "offset".  If offset happens to start at the
         very beginning of a cache line, this value will be zero. We add this
         value to our previous difference to obtain the number of bytes between
         the start of cache line #1 and "address".
      */
      lines += (offset & ((1 << __ease_cline_shift) - 1));

      /*
         Now we calculate the value of "a2", which is the distance from
         "address" to the start of the cache line after cache line #3. Note
         that this value may also turn out to be zero.  We add this to our
         total so far to obtain the total number of bytes from the start of
         cache line #1 to the start of the cache line which follows cache line
         #3.
      */
      if (address & ((1 << __ease_cline_shift) - 1))
	 lines += ((1 << __ease_cline_shift) -
		   (address & ((1 << __ease_cline_shift) - 1)));

      /*
         Now we determine how many lines our block touches by dividing the
         number of bytes that we have calculated by the cache line size.
      */
      lines >>= __ease_cline_shift;

      /*
         Now we can calcutae the worst-case time to fetch this block.  If we
         assume that none of the cache lines will be present when we start the
         block, it will take MISSTIME times the number of cache lines to fetch
         these.  Then, the rest of the total memory references made by this
         block will take HITTIME cycles each.  Thus, we have our worst-time
         calculation.
      */
      *(ptr + 3) = lines * MISSTIME + (refs - lines) * HITTIME;

      /*
         The size of this block is the starting offset minus the ending offset.
         Update the offset value to allow the next block to be placed following
         this one.
      */
      *(ptr + 4) = address - offset;
      offset = address;

      /*
         Advance the pointer to the next block's information
      */
      ptr = sptr;
   }
}
