/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/arm/ease/genreport.c,v $
 $Date: 2003/11/09 14:19:52 $
 $Revision: 1.1 $
 $Name:  $

 Log of changes:

 $Log: genreport.c,v $
 Revision 1.1  2003/11/09 14:19:52  jwd
 Initial version from SPARC--incomplete for arm

 Revision 1.4  2003/08/21 20:30:30  jdh8d
 changes to ease for tms

 Revision 1.3  2002/11/25 14:38:50  jwd
 Made changes so EASE functionality can be built.  It had gotten out of sync
 with the rest of the compiler.

 Revision 1.2  1998/04/07 20:08:29  jwd
 First ANSI C version.


*************************************************************************/

/*
 * This file contains the machine-independent portion of the EASE report
 * generator.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "instout.h"
#include "genreport.h"

/*
 * The following items are used to contain function information for each
 * function in the file.
 */
struct linfo {
   char name[16];
   struct linfo *next;
   unsigned int entry;
   unsigned int body;
};

struct finfo {
   char name[255];
   struct finfo *next;
   unsigned int inst;
   unsigned int refs;
   unsigned int calls;
   struct linfo *loops;
};

struct finfo *funcinfo = (struct finfo *) NULL;

int width = 80;		/* Width assumed when printing reports           */
char *myname;		/* Point to the name that we where invoked with  */
int maxreg;		/* Total number of register classes defined      */
int maxcc;		/* Total number of condition code states defined */
int maxaddr;		/* Total number of addressing modes defined      */
int maxdata;		/* Total number of data types defined            */
int maxclass;		/* Total number of instruction classes defined   */
int maxinst;		/* Total number of instructions defined          */
int maxhmem;		/* Total number of implicit memory references    */
int maxrpart;		/* Total number of register partitions           */
unsigned int numexeccnt[MAXEQUIVS];	/* Execution count for each equivalence */

#ifndef SAMESIZE
unsigned int class_size[MAXEQUIVS];	/* Size of bytes of each equivalence    */
#endif

/*
 * main - drives the data collection and reporting process
 */
int main(int argc, char **argv) {

   unsigned int totalfileinst = 0;
   unsigned int totalfilerefs = 0;
   unsigned int totalinst = 0;
   unsigned int totalrefs = 0;
   unsigned int totalcalls = 0;
   int dynamic = TRUE;
   int profile = FALSE;
   int loopinfo = FALSE;
   int reminfo = FALSE;

#ifndef SAMESIZE
   unsigned int totalbytes = 0;
#endif

   struct finfo *fptr, *sptr;
   struct linfo *lptr;
   FILE *fp;
   int i, j, k, l, files, insttype, equiv_classes, instructions, icnt;
   unsigned int (*instdatacnt)[], (*ccstatcnt)[], (*addrdatacnt)[];
   unsigned int (*instfieldcnt)[], (*regpartcnt)[];
   unsigned int numexec, *p;
   char *d, filename[100];
   struct finfo *falloc(void);
   struct linfo *lalloc(void);
   FILE *fopen(const char *, const char *);
   int instindex(int type), dataindex(int type), addrindex(int mode), ccindex(int cc);

   unsigned int (*salloc())[];
   void header(char *mesg), error(char *mesg), error_str(char *mesg, char *item),
    error_int(char *mesg, int item), error_chr(char *mesg, char item), exit(int);
   void function_profile(void);
   void function_loopinfo(void);
   void instruction_distributions(void);

#ifndef SAMESIZE
   void instruction_size_information();
#endif

   void instruction_class_distributions(void);
   void addressing_mode_distributions(void);
   void memory_reference_type_distributions(void);
   void memory_reference_address_distributions(void);
   void registers_saved_and_restored(unsigned int (*regpartcnt)[], unsigned int calls);
   void register_usage_totals(unsigned int (*regpartcnt)[]);
   void register_usage_per_call(unsigned int (*regpartcnt)[], unsigned int calls);
   void condition_code_usage(unsigned int (*ccstatcnt)[]);
   void conditional_branches_taken(void);
   void control_flow_distributions(void);
   void data_distributions_per_instruction(unsigned int (*instdatacnt)[]);
   void data_distributions_per_mode(unsigned int (*addrdatacnt)[]);
   void field_mode_type_distributions(unsigned int (*instfieldcnt)[]);

   /*
      Parse command line options
   */
   myname = *argv++;
   if (--argc && **argv == '-') {
      for (d = *argv + 1; *d; d++) {
	 switch (*d) {

	    case 'f':
	       profile = TRUE;
	       break;

	    case 'l':
	       loopinfo = TRUE;
	       break;

	    case 'r':
	       reminfo = TRUE;
	       break;

	    case 's':
	       dynamic = FALSE;
	       break;

	    default:
	       error_chr("bad option '%c'", *d);
	 }
      }
      *argv++;
      argc--;
   }
   if (argc <= 0)
      error("no files specified");

   /*
      Look at the machine dependent structures to determine the values of some
      useful constants.
   */
   for (maxreg = 0; regout[maxreg].type; maxreg++);
   for (maxcc = 0; ccout[maxcc].type; maxcc++);
   for (maxaddr = 0; addrout[maxaddr].mode; maxaddr++);
   for (maxdata = 0; dataout[maxdata].type; maxdata++);
   for (maxclass = 0; classout[maxclass].type; maxclass++);
   for (maxinst = 0; instout[maxinst].type; maxinst++);
   for (maxhmem = 0; hmemout[maxhmem].type; maxhmem++);
   for (maxrpart = 0; rpartout[maxrpart].name; maxrpart++);

   /*
      Now we allocate space to hold some arrays whose sizes depend on the value
      of the constant we just determined from the machine dependent
      information.
   */
   instdatacnt = salloc(maxinst * maxdata);
   ccstatcnt = salloc(maxinst * maxcc);
   addrdatacnt = salloc(maxaddr * maxdata);
   instfieldcnt = salloc(maxinst * WRKFIELDS * maxaddr * maxdata);
   regpartcnt = salloc(maxreg * maxrpart);

   /*
      Now we initialize the arrays we have just allocated
   */
   p = *instdatacnt;
   for (i = 0; i < maxinst; i++)
      for (j = 0; j < maxdata; j++)
	 *p++ = 0;

   p = *ccstatcnt;
   for (i = 0; i < maxinst; i++)
      for (j = 0; j < maxcc; j++)
	 *p++ = 0;

   p = *addrdatacnt;
   for (i = 0; i < maxaddr; i++)
      for (j = 0; j < maxdata; j++)
	 *p++ = 0;

   p = *instfieldcnt;
   for (i = 0; i < maxinst; i++)
      for (j = 0; j < WRKFIELDS; j++)
	 for (k = 0; k < maxaddr; k++)
	    for (l = 0; l < maxdata; l++)
	       *p++ = 0;

   p = *regpartcnt;
   for (i = 0; i < maxreg; i++)
      for (j = 0; j < maxrpart; j++)
	 *p++ = 0;

   /*
      The first page contains the breakdown of instructions executed and memory
      references generated by each file.
   */
   if (dynamic)
      header("Dynamic Statistics");
   else
      header("Static Statistics");
   printf("           file          \t\tinstructions\tmemory references\n");
   printf("-------------------------\t\t------------\t-----------------\n");

   /*
      Read information from each file
   */
   for (files = 0; argc--; files++) {

      /*
         Get the next filename from the argument list.  If the file ends in
         ".cnt", then we don't need to add ".cnt" to it before we look for it.
      */
      if (strlen(*argv) < 5 || strcmp(*argv + strlen(*argv) - 4, ".cnt")) {
	 (void) sprintf(filename, "%s.cnt", *argv);
	 if (!(fp = fopen(filename, "r")))
	    error_str("cannot open %s.cnt", *argv);
	 (void) strcpy(filename, *argv++);
      }
      else {
	 if (!(fp = fopen(*argv, "r")))
	    error_str("cannot open %s", *argv);
	 (void) strcpy(filename, *argv++);
	 filename[strlen(filename) - 4] = '\0';
      }

      /*
         Read the "not taken" counts for each type of branch
      */
      for (i = 0; i < NUMJREL; i++) {
	 (void) fscanf(fp, "%d", &k);
	 if (dynamic)
	    instout[instindex(i + FIRSTJREL)].ntaken += k;
      }

      /*
         Look through the .cnt file, one function at a time
      */
      totalfileinst = totalfilerefs = 0;
      while (fscanf(fp, "%d", &i) != EOF) {

	 /*
	    Create and initialize an information structure for this function
	 */
	 fptr = falloc();
	 fptr->refs = (unsigned long) 0L;
	 fptr->inst = (unsigned long) 0L;
	 fptr->loops = (struct linfo *) NULL;

	 /*
	    Look for the start of a new function marker.  If this item is not
	    correct, then something has gone wrong.
	 */
	 if (i != -1)
	    error("input file corrupt");

	 /*
	    The first byte tells us the number of the equivalence classes in
	    this function.
	 */
	 (void) fscanf(fp, "%d", &equiv_classes);
	 if (equiv_classes >= MAXEQUIVS)
	    error("equivalence class limit exceeded");

	 /*
	    Get equivalence class execution times and sizes for each class.
	    There should be exactly "equiv_classes" pairs of these.  If all
	    instructions are of the same size, then there is no need to obtain
	    size information, since the report that we use it for is rather
	    useless for this kind of machine.
	 */

#ifdef SAMESIZE
	 for (i = 0; i < equiv_classes; i++) {
	    (void) fscanf(fp, "%d", &j);
	    if (dynamic)
	       numexeccnt[i] = j;
	 }
#else
	 for (i = 0; i < equiv_classes; i++) {
	    (void) fscanf(fp, "%d %d", &j, &k);
	    if (dynamic) {
	       numexeccnt[i] = k;
	       class_size[i] = j;
	    }
	    else
	       totalbytes += j;
	 }
#endif

	 /*
	    Get the name of the current function
	 */
	 (void) fscanf(fp, "%s\n", fptr->name);

	 /*
	    The first equivalence class is always that of the entry block,
	    which allows us to determine how many times this function was
	    called.
	 */
	 numexec = (dynamic) ? numexeccnt[0] : 1;
	 totalcalls += numexec;
	 fptr->calls = numexec;

	 /*
	    For each register class, get the number of register of each
	    partition used by the function.
	 */
	 for (i = 0; i < maxreg; i++) {
	    for (j = 0; j < maxrpart; j++) {
	       (void) fscanf(fp, "%d", &k);
	       (*regpartcnt)[i * maxrpart + j] += k * numexec;

	       /*
	          If the register partition is marked as autosave, then we
	          generate the requested number of memory references for each
	          call made to this function.
	       */
	       if (rpartout[j].autosave) {
		  totalfilerefs += k * numexec * rpartout[j].autosave;
		  fptr->refs += k * numexec * rpartout[j].autosave;
		  dataout[dataindex(regout[i].type)].refs +=
		   k * numexec * rpartout[j].autosave;
	       }
	    }
	 }

	 /*
	    The next sequence of lines are used to indicate that the execution
	    counts of some basic blocks can be inferred from the counts of
	    other blocks.  Specifically, the count for a given block is the
	    difference in the counts bewteen two others.
	 */
	 (void) fscanf(fp, "%d", &i);
	 while (i != -2) {
	    (void) fscanf(fp, "%d %d", &j, &k);

	    if (i >= equiv_classes || j >= equiv_classes || k >= equiv_classes)
	       error("bad equivalence class number");

	    numexeccnt[i] = numexeccnt[j] - numexeccnt[k];
	    (void) fscanf(fp, "%d", &i);
	 }

#ifndef SAMESIZE

	 /*
	    Now that we have determined the number of times that all of the
	    equivalence classes have been executed, we can determine the total
	    number of bytes worth of instructions that have been fetched.
	 */
	 if (dynamic)
	    for (i = 0; i < equiv_classes; i++)
	       totalbytes += numexeccnt[i] * class_size[i];
#endif

	 /*
	    The next sequence of lines are used to indicate that the number of
	    times that a branch is not taken can be inferred from the execution
	    counts of some execution classes.  There are currently three
	    different ways of calculating these.
	 */
	 (void) fscanf(fp, "%d", &i);
	 while (i != -3) {

	    if (i >= equiv_classes)
	       error("bad equivalence class number");

	    (void) fscanf(fp, "%d", &j);
	    switch (j) {

	       case 1:
		  (void) fscanf(fp, "%d", &j);
		  instout[instindex(j)].ntaken += numexeccnt[i];
		  break;

	       case 2:
		  (void) fscanf(fp, " %d %d", &j, &k);
		  if (j >= equiv_classes)
		     error("bad equivalence class number");
		  instout[instindex(k)].ntaken += numexeccnt[j] - numexeccnt[i];
		  break;

	       case 3:
		  (void) fscanf(fp, " %d %d %d", &j, &k, &l);
		  if (j >= MAXEQUIVS || k >= MAXEQUIVS)
		     error("bad equivalence class number");
		  instout[instindex(l)].ntaken +=
		   numexeccnt[i] - (numexeccnt[j] - numexeccnt[k]);
		  break;

	       default:
		  error_int("bad branch code '%d'", j);
	    }

	    (void) fscanf(fp, "%d", &i);
	 }

	 /*
	    Get the loop information, if any, for this function
	 */
	 (void) fscanf(fp, "%d", &i);
	 while (i != -4) {

	    if (i >= equiv_classes)
	       error("bad equivalence class number");

	    (void) fscanf(fp, "%d", &j);

	    if (j >= equiv_classes)
	       error("bad equivalence class number");

	    lptr = lalloc();
	    lptr->next = fptr->loops;
	    fptr->loops = lptr;
	    lptr->entry = numexeccnt[i];
	    lptr->body = numexeccnt[j];
	    (void) fscanf(fp, "%s", lptr->name);

		/* see if there's another loop, or if we exit the loops */
	    (void) fscanf(fp, "%d", &i);
	 }


	 /*
	    Determine the number of instruction information lines that follow
	 */
	 (void) fscanf(fp, "%d", &instructions);

	 for (icnt = 0; icnt < instructions; icnt++) {

	    /*
	       Look up the number of times that this instruction was executed
	    */
	    (void) fscanf(fp, "%d", &i);
	    if (i >= equiv_classes)
	       error("bad equivalence class number");
	    numexec = (dynamic) ? numexeccnt[i] : 1;
	    totalfileinst += numexec;
	    fptr->inst += numexec;

	    /*
	       Obtain specific information about this instruction
	    */
	    (void) fscanf(fp, "%d", &insttype);
	    insttype = instindex(insttype);
	    instout[insttype].count += numexec;

	    /*
	       Increment counts for each class that this instruction belongs to
	    */
	    for (k = 0; k < maxclass; k++)
	       if (instout[insttype].class & classout[k].type)
		  classout[k].count += numexec;

	    /*
	       Obtain the base data type of this instruction
	    */
	    (void) fscanf(fp, "%d", &j);
	    if (j)
	       (*instdatacnt)[insttype * maxdata + dataindex(j)] += numexec;

	    /*
	       If this machine has any condition code status types, read it
	    */
	    if (maxcc) {
	       (void) fscanf(fp, "%d", &k);
	       if (k)
		  (*ccstatcnt)[insttype * maxcc + ccindex(k)] += numexec;
	    }

	    /*
	       Obtain information regarding the mode and type of each field of
	       this instruction.
	    */
	    for (k = 0; k < MAXFIELDS; k++) {
	       (void) fscanf(fp, "%d %d", &l, &j);
	       if (l) {
		  if (l & MEMREF) {
		     totalfilerefs += numexec;
		     fptr->refs += numexec;
		     if (j)
			dataout[dataindex(j)].refs += numexec;
		     else
			error("memory reference without a type");
		     if (l & ~MEMREF)
			addrout[addrindex(l)].refs += numexec;
		     else
			error("memory reference without a mode");
		  }
		  if (j)
		     (*addrdatacnt)[addrindex(l) * maxdata + dataindex(j)] +=
		      numexec;
		  addrout[addrindex(l)].count += numexec;
		  (*instfieldcnt)[insttype * WRKFIELDS * maxaddr * maxdata +
			      k * maxaddr * maxdata + addrindex(l) * maxdata +
				  dataindex(j)] += numexec;
	       }
	    }

	    /*
	       Now examine the hidden memory reference information to see if
	       there are any implicit memory references generated by this
	       instruction. If they are, then update the information to make
	       them known.
	    */
	    for (i = 0; i < maxhmem; i++) {
	       if (instindex(hmemout[i].type) == insttype) {
		  totalfilerefs += numexec * hmemout[i].amount;
		  fptr->refs += numexec * hmemout[i].amount;
		  dataout[dataindex(hmemout[i].datatype)].refs +=
		   numexec * hmemout[i].amount;
		  addrout[addrindex(hmemout[i].mode)].imrefs +=
		   numexec * hmemout[i].amount;
		  (*addrdatacnt)[addrindex(hmemout[i].mode) * maxdata +
				 dataindex(hmemout[i].datatype)] +=
		   numexec * hmemout[i].amount;
		  addrout[addrindex(hmemout[i].mode)].imcount +=
		   numexec * hmemout[i].amount;
		  (*instfieldcnt)[insttype * WRKFIELDS * maxaddr * maxdata +
				  MAXFIELDS * maxaddr * maxdata +
				  addrindex(hmemout[i].mode) * maxdata +
				  dataindex(hmemout[i].datatype)] +=
		   numexec * hmemout[i].amount;
	       }
	    }
	 }

	 /*
	    Insert the function information structure into the list of function
	    information structures ordered by the number of instructions it
	    executed.
	 */
	 if (!funcinfo || funcinfo->inst < fptr->inst) {
	    fptr->next = funcinfo;
	    funcinfo = fptr;
	 }
	 else {
	    sptr = funcinfo;
	    for (;;) {
	       if (!sptr->next || sptr->next->inst < fptr->inst) {
		  fptr->next = sptr->next;
		  sptr->next = fptr;
		  break;
	       }
	       sptr = sptr->next;
	    }
	 }
      }

      /*
         Close the .cnt file
      */
      (void) fclose(fp);

      printf("%3d. %-20.20s\t\t%12u\t%17u\n",
	     files + 1, filename, totalfileinst, totalfilerefs);
      totalinst += totalfileinst;
      totalrefs += totalfilerefs;
   }

   printf("-------------------------\t\t------------\t-----------------\n");
   printf("     totals              \t\t%12u\t%17u\n", totalinst, totalrefs);

   /*
      Generate the various reports available
   */
   if (profile)
      function_profile();
   if (loopinfo)
      function_loopinfo();
   if (!reminfo)
      exit(0);
   instruction_distributions();

#ifndef SAMESIZE
   instruction_size_information(totalinst, totalbytes);
#endif

   instruction_class_distributions();
   addressing_mode_distributions();
   memory_reference_type_distributions();
   memory_reference_address_distributions();
   registers_saved_and_restored(regpartcnt, totalcalls);
   register_usage_totals(regpartcnt);
   register_usage_per_call(regpartcnt, totalcalls);

   if (maxcc)
      condition_code_usage(ccstatcnt);

   if (dynamic) {
      conditional_branches_taken();
      control_flow_distributions();
   }

   data_distributions_per_instruction(instdatacnt);
   data_distributions_per_mode(addrdatacnt);
   field_mode_type_distributions(instfieldcnt);

   exit(0);
}

/*
 * function_profile - print a profile of the different functions
 */
void function_profile(void) {

   struct finfo *fptr;
   unsigned long inst, refs, calls;
   void header(char *mesg);

   /*
      Do nothing if there is no function information to display
   */
   if (!(fptr = funcinfo))
      return;

   inst = refs = calls = (unsigned long) 0L;
   header("Function Profile");
   printf("function name               instructions      references     ");
   printf("   calls\n");
   printf("-------------------------  ---------------  ---------------  ");
   printf("------------\n");
   while (fptr) {
      printf("%25.25s  %15u  %15u  %12u\n",
	     fptr->name, fptr->inst, fptr->refs, fptr->calls);
      inst += fptr->inst;
      refs += fptr->refs;
      calls += fptr->calls;
      fptr = fptr->next;
   }
   printf("-------------------------  ---------------  ---------------  ");
   printf("------------\n");
   printf("%25.25s  %15u  %15u  %12u\n", "totals", inst, refs, calls);
}

/*
 * function_loopinfo - print information about the various loops
 */
void function_loopinfo(void) {

   struct finfo *fptr;
   struct linfo *lptr;
   void header(char *mesg);

   /*
      Do nothing if there is no function information to display
   */
   if (!(fptr = funcinfo))
      return;

   header("Loop Profile");
   printf("function name                 loop     iterations/entry  ");
   printf(" invoked/call\n");
   printf("-------------------------  ----------  ----------------  ");
   printf("---------------\n");
   while (fptr) {
      if (fptr->calls) {
	 for (lptr = fptr->loops; lptr; lptr = lptr->next) {
	    printf("%25.25s  %10.10s  ", fptr->name, lptr->name);
	    if (lptr->entry)
	       printf("%16.2f  ", ((double) lptr->body) / ((double) lptr->entry));
	    else
	       printf("   Never invoked  ");
	    printf("%15.2f\n", ((double) lptr->entry) / ((double) fptr->calls));
	 }
      }
      fptr = fptr->next;
   }
   printf("-------------------------  ----------  ----------------  ");
   printf("---------------\n");
}

/*
 * instruction_distributions - print instruction distributions report
 */
void instruction_distributions(void) {

   unsigned int i, all;
   void header(char *mesg);

   /*
      Determine the total number of instructions executed
   */
   all = 0;
   for (i = 0; i < maxinst; i++)
      all += instout[i].count;
   if (!all)
      return;

   header("Instruction Distributions");
   printf("instruction type                \tnumber executed\t\tper cent\n");
   printf("--------------------------------\t---------------\t\t--------\n");
   for (i = 0; i < maxinst; i++) {
      if (instout[i].count)
	 printf("%32.32s\t%15u\t\t%8.2f\n", instout[i].name,
		instout[i].count, instout[i].count / (float) all * 100.0);
   }
   printf("--------------------------------\t---------------\t\t--------\n");
   printf("                           total\t%15u\n", all);
}

#ifndef SAMESIZE
/*
 * instruction_size_information - report information about instruction sizes
 */
void instruction_size_information(unsigned int insts, unsigned int bytes) {

   void header();

   if (insts) {
      header("Instruction Size Information");
      printf("  total number of instructions executed: %14u\n", insts);
      printf("   total bytes for instructions fetched: %14u\n", bytes);
      printf("average number of bytes per instruction: %14.4f\n",
	     bytes / (float) insts);
   }
}

#endif

/*
 * instruction_class_distributions - print class distributions report
 */
void instruction_class_distributions(void) {

   int i, j;
   unsigned int all, classed;
   void header(char *mesg), subheader(char *mesg);

   all = 0;
   classed = 0;
   for (i = 0; i < maxinst; i++) {
      all += instout[i].count;
      for (j = 0; j < maxclass; j++) {
	 if (instout[i].class & classout[j].type) {
	    classed += instout[i].count;
	    break;
	 }
      }
   }
   if (!classed)
      return;

   header("Instruction Class Distributions");

   for (i = 0; i < maxclass; i++) {
      if (classout[i].count) {
	 subheader(classout[i].name);
	 printf("instruction type                \tnumber executed\t\t");
	 printf("per cent\n");
	 printf("--------------------------------\t---------------\t\t");
	 printf("--------\n");
	 for (j = 0; j < maxinst; j++) {
	    if (instout[j].count && instout[j].class & classout[i].type)
	       printf("%32.32s\t%15u\t\t%8.2f\n",
		      instout[j].name, instout[j].count,
		      instout[j].count / (float) classout[i].count * 100.0);
	 }
	 printf("--------------------------------\t---------------\t\t");
	 printf("--------\n");
	 printf(" number of instructions in class\t%15u\n\n", classout[i].count);
      }
   }

   subheader("Class Distributions");
   printf("instruction class               \tnumber executed\t\tper cent\n");
   printf("--------------------------------\t---------------\t\t--------\n");
   for (i = 0; i < maxclass; i++)
      printf("%32.32s\t%15u\t\t%8.2f\n", classout[i].name, classout[i].count,
	     classout[i].count / (float) all * 100.0);
   if (classed < all)
      printf("          all other instructions\t%15u\t\t%8.2f\n",
	     all - classed, (all - classed) / (float) all * 100.0);
   printf("--------------------------------\t---------------\t\t--------\n");
   printf(" number of instructions executed\t%15u\n\n", all);
   printf("  Note: instructions may belong to more than one class\n");
}

/*
 * addressing_mode_distributions - print addressing mode distributions report
 */
void addressing_mode_distributions(void) {

   unsigned int all;
   int i;
   char name[100];
   void header(char *mesg);

   /*
      Obtain the total addressing mode count
   */
   all = 0;
   for (i = 0; i < maxaddr; i++)
      all += addrout[i].count + addrout[i].imcount;
   if (!all)
      return;

   header("Addressing Mode Distributions");
   printf("addressing mode                 \tnumber used   \t\tper cent\n");
   printf("--------------------------------\t--------------\t\t--------\n");
   for (i = 0; i < maxaddr; i++) {
      if (addrout[i].count)
	 printf("%32.32s\t%15u\t\t%8.2f\n", addrout[i].name,
		addrout[i].count, addrout[i].count / (float) all * 100.0);
      if (addrout[i].imcount) {
	 (void) sprintf(name, "%s (implicit)", addrout[i].name);
	 printf("%32.32s\t%15u\t\t%8.2f\n", name,
		addrout[i].imcount, addrout[i].imcount / (float) all * 100.0);
      }
   }
   printf("--------------------------------\t--------------\t\t--------\n");
   printf("     total addressing modes used\t%15u\n", all);
}

/*
 * memory_reference_type_distributions - report of memory reference sizes
 */
void memory_reference_type_distributions(void) {

   int i;
   unsigned int refs, bytes;
   void header(char *mesg);

   refs = bytes = 0;
   for (i = 0; i < maxdata; i++) {
      refs += dataout[i].refs;
      bytes += dataout[i].refs * dataout[i].size;
   }
   if (!bytes)
      return;

   header("Memory Reference Type Distributions");
   printf("memory reference type     number     per cent     bytes      ");
   printf("per cent\n");
   printf("---------------------  ------------  --------  ------------  ");
   printf("--------\n");
   for (i = 0; i < maxdata; i++) {
      if (dataout[i].refs)
	 printf("%21.21s  %12u  %8.2f  %12u  %8.2f\n", dataout[i].name,
		dataout[i].refs, dataout[i].refs / (float) refs * 100.0,
		dataout[i].refs * dataout[i].size,
		(dataout[i].refs * dataout[i].size) / (float) bytes * 100.0);
   }
   printf("---------------------  ------------  --------  ------------  ");
   printf("--------\n");
   printf("               totals  %12u            %12u\n", refs, bytes);
}

/*
 * memory_reference_address_distributions - report modes used by references
 */
void memory_reference_address_distributions(void) {

   unsigned int all, refs;
   int i;
   char name[100];
   void header(char *mesg);

   all = refs = 0;
   for (i = 0; i < maxaddr; i++) {
      all += addrout[i].count + addrout[i].imcount;
      refs += addrout[i].refs + addrout[i].imrefs;
   }
   if (!all)
      return;

   header("Memory Reference Addressing Mode Distributions");
   printf("addressing mode            with reference  without reference  ");
   printf("per cent with\n");
   printf("-------------------------  --------------  -----------------  ");
   printf("-------------\n");
   for (i = 0; i < maxaddr; i++) {
      if (addrout[i].count)
	 printf("%25.25s  %14u  %17u  %13.2f\n", addrout[i].name,
		addrout[i].refs, addrout[i].count - addrout[i].refs,
		addrout[i].refs / (float) addrout[i].count * 100.0);
      if (addrout[i].imcount) {
	 (void) sprintf(name, "%s (implicit)", addrout[i].name);
	 printf("%25.25s  %14u  %17u  %13.2f\n", name,
		addrout[i].imrefs, addrout[i].imcount - addrout[i].imrefs,
		addrout[i].imrefs / (float) addrout[i].imcount * 100.0);
      }
   }
   printf("-------------------------  --------------  -----------------  ");
   printf("-------------\n");
   printf("                   totals  %14u  %17u  %13.2f\n",
	  refs, all - refs, refs / (float) all * 100.0);
}

/*
 * registers_saved_and_restored - reports registers saved and restored
 */
void registers_saved_and_restored(unsigned int (*regpartcnt)[], unsigned int calls) {

   int i, j;
   unsigned int all, total, first;
   void header(char *mesg);

   all = 0;
   for (i = 0; i < maxreg; i++)
      for (j = 0; j < maxrpart; j++)
	 if (!rpartout[j].volatil)
	    all += (*regpartcnt)[i * maxrpart + j];
   if (!all)
      return;

   header("Registers Saved and Restored");
   printf("register type          register partition  number saved  ");
   printf("per call  per cent\n");
   printf("---------------------  ------------------  ------------  ");
   printf("--------  --------\n");
   for (i = 0; i < maxreg; i++) {

      /*
         Determine if any registers are saved and restored at all for this
         register class.
      */
      total = 0;
      for (j = 0; j < maxrpart; j++)
	 if (!rpartout[j].volatil)
	    total += (*regpartcnt)[i * maxrpart + j];

      if (total) {
	 first = TRUE;
	 for (j = 0; j < maxrpart; j++) {
	    if (!rpartout[j].volatil) {
	       printf("%-21.21s  %-18.18s  %12u  %8.4f  %8.2f\n",
		      (first) ? regout[i].name : "", rpartout[j].name,
		      (*regpartcnt)[i * maxrpart + j],
		      (*regpartcnt)[i * maxrpart + j] / (float) calls,
		      (*regpartcnt)[i * maxrpart + j] / (float) all * 100.0);
	       first = FALSE;
	    }
	 }
      }
   }
   printf("---------------------  ------------------  ------------  ");
   printf("--------  --------\n");
   printf("                                   totals  %12u  %8.4f\n",
	  all, all / (float) calls);
}

/*
 * register_usage_totals - print the register usage totals report
 */
void register_usage_totals(unsigned int (*regpartcnt)[]) {

   int i, j;
   unsigned int all, total;
   char name[100];
   void header(char *mesg), subheader(char *mesg);

   /*
      Determine if any registers are used at all
   */
   all = 0;
   for (i = 0; i < maxreg; i++)
      for (j = 0; j < maxrpart; j++)
	 all += (*regpartcnt)[i * maxrpart + j];
   if (!all)
      return;

   header("Register Usage Totals");
   for (i = 0; i < maxreg; i++) {

      /*
         Determine if any registers of this class are used
      */
      total = 0;
      for (j = 0; j < maxrpart; j++)
	 total += (*regpartcnt)[i * maxrpart + j];

      if (total) {
	 (void) sprintf(name, "%s usage totals", regout[i].name);
	 subheader(name);
	 printf("register partition            \t\tnumber used\t\tper cent\n");
	 printf("------------------------------\t\t-----------\t\t--------\n");
	 for (j = 0; j < maxrpart; j++) {
	    if ((*regpartcnt)[i * maxrpart + j]) {
	       printf("%30.30s\t\t%11u\t\t%8.2f\n", rpartout[j].name,
		      (*regpartcnt)[i * maxrpart + j],
		      (*regpartcnt)[i * maxrpart + j] / (float) total * 100.0);
	    }
	 }
	 printf("------------------------------\t\t-----------\t\t--------\n");
	 printf("                         total\t\t%11u\n\n", total);
      }
   }

   subheader("all register usage totals");
   printf("register partition            \t\tnumber used\t\tper cent\n");
   printf("------------------------------\t\t-----------\t\t--------\n");
   for (i = 0; i < maxrpart; i++) {

      /*
         Determine if any registers in this partition are used
      */
      total = 0;
      for (j = 0; j < maxreg; j++)
	 total += (*regpartcnt)[j * maxrpart + i];

      if (total)
	 printf("%30.30s\t\t%11u\t\t%8.2f\n", rpartout[i].name,
		total, total / (float) all * 100.0);
   }
   printf("------------------------------\t\t-----------\t\t--------\n");
   printf("                         total\t\t%11u\n", all);
}

/*
 * register_usage_per_call - report register usage averages per call
 */
void register_usage_per_call(unsigned int (*regpartcnt)[], unsigned int calls) {

   int i, j;
   unsigned int all, total;
   char name[100];
   void header(char *mesg), subheader(char *mesg);

   /*
      Determine if any registers are used at all
   */
   all = 0;
   for (i = 0; i < maxreg; i++)
      for (j = 0; j < maxrpart; j++)
	 all += (*regpartcnt)[i * maxrpart + j];
   if (!all)
      return;

   header("Register Usage per Call");
   for (i = 0; i < maxreg; i++) {

      /*
         Determine if any registers of this class are used
      */
      total = 0;
      for (j = 0; j < maxrpart; j++)
	 total += (*regpartcnt)[i * maxrpart + j];

      if (total) {
	 (void) sprintf(name, "%s usage per call", regout[i].name);
	 subheader(name);
	 printf("register partition              \t\t\taverage per call\n");
	 printf("--------------------------------\t\t\t----------------\n");
	 for (j = 0; j < maxrpart; j++) {
	    if ((*regpartcnt)[i * maxrpart + j]) {
	       printf("%32.32s\t\t\t%16.4f\n", rpartout[j].name,
		      (*regpartcnt)[i * maxrpart + j] / (float) calls);
	    }
	 }
	 printf("--------------------------------\t\t\t----------------\n");
	 printf("                           total\t\t\t%16.4f\n\n",
		total / (float) calls);
      }
   }

   subheader("all register usage per call");
   printf("register partition              \t\t\taverage per call\n");
   printf("--------------------------------\t\t\t----------------\n");
   for (i = 0; i < maxrpart; i++) {

      /*
         Determine if any registers in this partition are used
      */
      total = 0;
      for (j = 0; j < maxreg; j++)
	 total += (*regpartcnt)[j * maxrpart + i];

      if (total)
	 printf("%32.32s\t\t\t%16.4f\n",
		rpartout[i].name, total / (float) calls);
   }
   printf("--------------------------------\t\t\t----------------\n");
   printf("                           total\t\t\t%16.4f\n",
	  all / (float) calls);
}

/*
 * condition_code_usage - report on codition code usage
 */
void condition_code_usage(unsigned int (*ccstatcnt)[]) {

   int i, j, first;
   unsigned int all, *p;
   void header(char *mesg);

   /*
      Determine if this report is even worth printing
   */
   all = 0;
   p = *ccstatcnt;
   for (i = 0; i < maxinst; i++)
      for (j = 0; j < maxcc; j++)
	 all += *p++;
   if (!all)
      return;

   header("Condition Code Usage");
   printf("instruction type\tcondition code status\tnumber used\tper cent\n");
   printf("----------------\t---------------------\t-----------\t--------\n");
   for (i = 0; i < maxinst; i++) {

      /*
         Determine if this instruction does anything with the codition codes
      */
      all = 0;
      p = &(*ccstatcnt)[i * maxcc];
      for (j = 0; j < maxcc; j++)
	 all += *p++;

      if (all) {
	 first = TRUE;
	 p = &(*ccstatcnt)[i * maxcc];
	 for (j = 0; j < maxcc; j++) {
	    if (*p) {
	       printf("%-16.16s\t%-21.21s\t%11u\t%8.2f\n",
		      (first) ? instout[i].name : "",
		    ccout[j].name, *p, *p / (float) instout[i].count * 100.0);
	       first = FALSE;
	    }
	    p++;
	 }
	 if (all < instout[i].count)
	    printf("%-16.16s\t%-21.21s\t%11u\t%8.2f\n", "", "all others",
		   instout[i].count - all,
		 (instout[i].count - all) / (float) instout[i].count * 100.0);
	 printf("\n");
      }
   }
}

/*
 * conditional_branches_taken - report on the direction of branches
 */
void conditional_branches_taken(void) {

   int i;
   unsigned int count, ntaken;
   void header(char *mesg);

   /*
      Determine if the report is even worth printing
   */
   count = ntaken = 0;
   for (i = 0; i < maxinst; i++) {
      if (instout[i].cflow & COND_BRANCH && instout[i].count) {
	 count += instout[i].count;
	 ntaken += instout[i].ntaken;
      }
   }
   if (!count)
      return;

   header("Conditional Branches Taken");
   printf("instruction type   \tbranch taken\tnot taken   \tper cent taken\n");
   printf("-------------------\t------------\t------------\t--------------\n");
   for (i = 0; i < maxinst; i++) {
      if (instout[i].cflow & COND_BRANCH && instout[i].count) {
	 printf("%19.19s\t%12u\t%12u\t%14.2f\n", instout[i].name,
		instout[i].count - instout[i].ntaken, instout[i].ntaken,
		(instout[i].count - instout[i].ntaken) /
		(float) instout[i].count * 100.0);
      }
   }
   printf("-------------------\t------------\t------------\t--------------\n");
   printf("             totals\t%12u\t%12u\t%14.2f\n",
	  count - ntaken, ntaken, (count - ntaken) / (float) count * 100.0);
}

/*
 * control_flow_distributions - report distribution of control flow changes
 */
void control_flow_distributions(void) {

   int i, j;
   unsigned int insts, all;

   insts = all = 0;
   for (i = 0; i < maxinst; i++) {
      if (instout[i].cflow) {
	 all += instout[i].count - instout[i].ntaken;
	 for (j = 0; cfout[j].type; j++)
	    if (cfout[j].type & instout[i].cflow)
	       cfout[j].count += instout[i].count - instout[i].ntaken;
      }
      insts += instout[i].count;
   }
   if (!all)
      return;

   header("Control Flow Distributions");
   printf("type of control flow instruction\tnumber executed\t\tper cent\n");
   printf("--------------------------------\t---------------\t\t--------\n");
   for (i = 0; cfout[i].type; i++) {
      printf("%32.32s\t%15u\t\t%8.2f\n", cfout[i].name,
	     cfout[i].count, cfout[i].count / (float) all * 100.0);
   }
   printf("--------------------------------\t---------------\t\t--------\n");
   printf("      total control flow changes\t%15u\t\t%8.2f\n\n",
	  all, all / (float) insts * 100.0);
   printf(" total of all other instructions\t%15u\t\t%8.2f\n\n",
	  insts - all, (insts - all) / (float) insts * 100.0);
   printf("average number of instructions executed between ");
   printf("control flow changes: %5.2f\n", (insts - all) / (float) all);
}

/*
 * data_distributions_per_instruction - report data types used by instructions
 */
void data_distributions_per_instruction(unsigned int (*instdatacnt)[]) {

   int i, j, first;
   unsigned int all, *p;
   void header(char *mesg);

   /*
      Determine if this report is worth printing
   */
   all = 0;
   p = *instdatacnt;
   for (i = 0; i < maxinst; i++)
      for (j = 0; j < maxdata; j++)
	 all += *p++;
   if (!all)
      return;

   header("Data Type Distribution per Instruction");
   printf("instruction type   \tdata type         \tnumber used \tper cent\n");
   printf("-------------------\t------------------\t------------\t--------\n");
   for (i = 0; i < maxinst; i++) {

      /*
         Determine the total number of data references made by this type of
         instruction.
      */
      all = 0;
      p = &(*instdatacnt)[i * maxdata];
      for (j = 0; j < maxdata; j++)
	 all += *p++;

      if (all) {
	 p = &(*instdatacnt)[i * maxdata];
	 first = TRUE;
	 for (j = 0; j < maxdata; j++) {
	    if (*p) {
	       printf("%-19.19s\t%-18.18s\t%12u\t%8.2f\n",
		      (first) ? instout[i].name : "",
		      dataout[j].name, *p, *p / (float) all * 100.0);
	       first = FALSE;
	    }
	    p++;
	 }
	 printf("\n");
      }
   }
}

/*
 * data_distributions_per_mode - report on data types used by addressing modes
 */
void data_distributions_per_mode(unsigned int (*addrdatacnt)[]) {

   int i, j, first;
   unsigned int all, *p;
   void header(char *mesg);

   /*
      Determine if this report is even worth printing
   */
   all = 0;
   p = *addrdatacnt;
   for (i = 0; i < maxaddr; i++)
      for (j = 0; j < maxdata; j++)
	 all += *p++;
   if (!all)
      return;

   header("Data Type Distributions per Addressing Mode");
   printf("addressing mode    \tdata type         \tnumber used \tper cent\n");
   printf("-------------------\t------------------\t------------\t--------\n");
   for (i = 0; i < maxaddr; i++) {

      /*
         Determine if there is anything to print for this addressing mode
      */
      all = 0;
      p = &(*addrdatacnt)[i * maxdata];
      for (j = 0; j < maxdata; j++)
	 all += *p++;

      if (all) {
	 first = TRUE;
	 p = &(*addrdatacnt)[i * maxdata];
	 for (j = 0; j < maxdata; j++) {
	    if (*p) {
	       printf("%-19.19s\t%-18.18s\t%12u\t%8.2f\n",
		      (first) ? addrout[i].name : "",
		      dataout[j].name, *p, *p / (float) all * 100.0);
	       first = FALSE;
	    }
	    p++;
	 }
	 printf("\n");
      }
   }
}

/*
 * field_mode_type_distributions - report on the modes and types used by fields
 */
void field_mode_type_distributions(unsigned int (*instfieldcnt)[]) {

   int i, j, k, l, first, firstf, firsta;
   unsigned int all, *p;
   void header(char *mesg);

   /*
      Determine if this report is worth printing
   */
   all = 0;
   p = *instfieldcnt;
   for (i = 0; i < maxinst; i++)
      for (j = 0; j < WRKFIELDS; j++)
	 for (k = 0; k < maxaddr; k++)
	    for (l = 0; l < maxdata; l++)
	       all += *p++;
   if (!all)
      return;

   header("Instruction Field Mode and Type Distributions");
   printf("instruction type  field name   addressing mode  ");
   printf("data type number used per cent\n");
   printf("---------------- ------------ ----------------- ");
   printf("--------- ----------- --------\n");
   for (i = 0; i < maxinst; i++) {
      first = TRUE;
      for (j = 0; j < WRKFIELDS; j++) {

	 /*
	    Count all of the modes and data types that this instruction uses
	 */
	 all = 0;
	 p = &(*instfieldcnt)[i * WRKFIELDS * maxaddr * maxdata +
			      j * maxaddr * maxdata];
	 for (k = 0; k < maxaddr; k++)
	    for (l = 0; l < maxdata; l++)
	       all += *p++;

	 if (all) {
	    firstf = TRUE;
	    p = &(*instfieldcnt)[i * WRKFIELDS * maxaddr * maxdata +
				 j * maxaddr * maxdata];
	    for (k = 0; k < maxaddr; k++) {
	       firsta = TRUE;
	       for (l = 0; l < maxdata; l++) {
		  if (*p) {
		     printf("%-16.16s %-12.12s %-17.17s %-9.9s %11u %8.2f\n",
			    (first) ? instout[i].name : "",
			    (firstf) ? ((j == MAXFIELDS) ? "implicit" :
					instout[i].fieldname[j]) : "",
			    (firsta) ? addrout[k].name : "",
			    dataout[l].name, *p, *p / (float) all * 100.0);
		     first = firstf = firsta = FALSE;
		  }
		  p++;
	       }
	    }
	 }
      }
      if (!first)
	 printf("\n");
   }
}

/*
 * header - print a header on a report
 */
void header(char *mesg) {

   static int first = TRUE;
   void subheader(char *mesg);

   if (!first)
      (void) printf("\f");
   else
      first = FALSE;
   subheader(mesg);
   (void) printf("\n");
}

/*
 * subheader - print a sub-header on a report
 */
void subheader(char *mesg) {

   int tab, i;

   (void) printf("\n");
   if (strlen(mesg) > width)
      (void) printf("%s\n", mesg);
   else {
      tab = (width - strlen(mesg)) >> 1;
      for (i = 0; i < tab; i++)
	 (void) printf(" ");
      (void) printf("%s\n", mesg);
   }
   (void) printf("\n");
}

/*
 * instindex - return the index for a particular instruction type
 */
int instindex(int type) {

   register int i, t;
   void error_int(char *mesg, int item);

   for (i = 0; t = instout[i].type; i++)
      if (t == type)
	 return (i);

   error_int("bad instruction type '%d'", type);
   return (0);
}

/*
 * addrindex - return the index for a particular addressing mode
 */
int addrindex(int mode) {

   register int i, m;
   void error_int(char *mesg, int item);

   mode = mode & ~MEMREF;
   for (i = 0; m = addrout[i].mode; i++)
      if (m == mode)
	 return (i);

   error_int("bad mode type '%d'", mode);
   return (0);
}

/*
 * dataindex - return index for a particular data type
 */
int dataindex(int type) {

   register int i, t;
   void error_int(char *mesg, int item);

   for (i = 0; t = dataout[i].type; i++)
      if (t == type)
	 return (i);

   error_int("bad data type '%d'", type);
   return (0);
}

/*
 * ccindex - return the index for a particular codition code status
 */
int ccindex(int cc) {

   register int i, t;
   void error_int(char *mesg, int item);

   for (i = 0; t = ccout[i].type; i++)
      if (t == cc)
	 return (i);

   error_int("bad condition code status '%d'", cc);
   return (0);
}

/*
 * salloc - allocate memory while watching out for an out of memory error
 */
unsigned int (*salloc(int size))[] {

   char *ptr;
   void error(char *mesg);

   if (!(ptr = malloc(((unsigned int) size) * sizeof(unsigned int))))
      error("out of memory");
   return ((unsigned int (*)[]) ptr);
}

/*
 * falloc - allocate memory while watching out for an out of memory error
 */
struct finfo *falloc(void) {

   char *ptr;
   void error(char *mesg);

   if (!(ptr = malloc(sizeof(struct finfo))))
      error("out of memory");
   return ((struct finfo *) ptr);
}

/*
 * lalloc - allocate memory while watching out for an out of memory error
 */
struct linfo *lalloc(void)
{
   char *ptr;
   void error(char *mesg);

   if (!(ptr = malloc(sizeof(struct linfo))))
      error("out of memory");
   return ((struct linfo *) ptr);
}

/*
 * error - display an error message and exit with a non-zero error code
 */
void error(char *mesg) {

   void exit(int);

   (void) fprintf(stderr, "%s: %s\n", myname, mesg);
   exit(1);
}

/*
 * error_str - display an error message and exit with a non-zero error code
 */
void error_str(char *mesg, char *item) {

   void exit(int);

   (void) fprintf(stderr, "%s: ", myname, mesg);
   (void) fprintf(stderr, mesg, item);
   (void) fprintf(stderr, "\n");
   exit(1);
}

/*
 * error_int - display an error message and exit with a non-zero error code
 */
void error_int(char *mesg, int item) {

   void exit(int);

   (void) fprintf(stderr, "%s: ", myname, mesg);
   (void) fprintf(stderr, mesg, item);
   (void) fprintf(stderr, "\n");
   exit(1);
}

/*
 * error_chr - display an error message and exit with a non-zero error code
 */
void error_chr(char *mesg, char item) {

   void exit(int);

   (void) fprintf(stderr, "%s: ", myname, mesg);
   (void) fprintf(stderr, mesg, item);
   (void) fprintf(stderr, "\n");
   exit(1);
}
