/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/arm/ease/instout.h,v $
 $Date: 2003/11/09 14:19:52 $
 $Revision: 1.1 $
 $Name:  $

 Log of changes:

 $Log: instout.h,v $
 Revision 1.1  2003/11/09 14:19:52  jwd
 Initial version from SPARC--incomplete for arm

 Revision 1.2  2003/08/20 18:08:50  jdh8d
 changes to get genreport, etc working for tms

 Revision 1.1  2003/08/19 18:10:23  jg9h
 *** empty log message ***

 Revision 1.3  2002/08/07 02:09:52  clc5q
 Update EASE code per changes from David Whalley to make the SPARC EASE
 target build and execute properly.

 Revision 1.2  1998/04/07 20:08:42  jwd
 First ANSI C version.


*************************************************************************/

/*
 * This file contains the machine dependent information that tailors the
 * report generator to a specific machine.
 */

#define MAXEQUIVS 50000
#define END 0
#define NONE 0
#define FALSE 0
#define TRUE 1

#define MEASURE
#include "addrdef.h"
#include "instdef.h"
#include "recdef.h"

#define WRKFIELDS (MAXFIELDS + 1)

/*
 * If all of the instructions on this machine are of the same size, then
 * declare the SAMESIZE constant.
 */
#define SAMESIZE 4


/*
 * Declare the different data types on the target machine.  The first field
 * indicates which data type is being defined.  These data types are declared
 * in the addrdef.h file for this target machine.  The second field indicates
 * the number of bytes that a memory item of the type being defined occupies
 * in memory.  The third field is the name of the data type as it should be
 * printed on reports.  None of the remaining fields should be explicitly
 * initialized.  The END constant must be used to mark the end of the data.
 */
struct {
    int type;
    int size;
    char *name;
    unsigned int refs;
    } dataout[] =

   { { BYTE,    1, "byte"   },
     { WORD,    2, "word"   },
     { LONG,    4, "long"   },
     { FLOAT,   4, "float"  },
     { DOUBLE,  8, "double" },
     { ANY,     4, "any"    },
     { END                  } };


/*
 * Declare the different register classes on the target machine.  The first
 * field denotes the basic data type that the register class falls into. The
 * second field contains the name of the register class as it will appear in
 * reports.  The END contant must be used to mark the end of the information.
 */
struct {
    int type;
    char *name;
    } regout[] =

   { 
     	{ LONG,    "general-purpose"  },
     	{ DOUBLE,    "floating-point" },
     { END                          } };


/*
 * Declare the register partitions.  These must agree with the number of
 * 'registers used numbers' emitted by the dumpreg() function in linstinfo.c
 * for the target machine.  The first field defines the name of the register
 * partition that is being declared.  Every register class is assumed to have
 * one of each partiton, although in some cases the partition might be empty
 * for a particular register class.  The second field indicates if the parti-
 * tion is volatile or not.  A volatile partition is one whose register values
 * may change across a function call.  A non-volatile partition is one whose
 * register values are maintained (saved and restored) across a function call.
 * The third field indicates that the registers in the partition are
 * automatically saved and restored by the calling mechanism.  This field
 * should be zero if the registers are saved and restored explicitly.  If
 * this field is non-zero, then the value given will be multiplied by the
 * number of registers in the partition used by each function and the number
 * of times that the function is called.  The resulting value will be added
 * to the number of memory references made by the program.  The type of the
 * memory reference is given by the first field of the register class
 * declaration.  The END constant must be used to mark the end of the
 * information.
 */
struct {
    char *name;
    int volatil;
    int autosave;
    } rpartout[] =

   { { "scratch",      TRUE,   0  },
     { "non-scratch",  FALSE,  0  },
     { END                        } };


/*
 * Declare the different condition code status items for this machine.  The
 * first field indicates which condition code status is being declared.  The
 * items in this field are declared in the instdef.h file for the target
 * machine.  The second field declares the name of the status as it will
 * appear on reports.  The END constant must be used to mark the end of the
 * information.
 */
struct {
    int type;
    char *name;
    } ccout[] =
   { { CCSET,  "condition codes set"    },
     { CCUSED, "condition codes used"   },
     { END                              } };


/*
 * Declare the different addressing modes available on the target machine.  The
 * first field indicates which addressing mode is being defined.  The items in
 * this field are declared in the addrdef.h file for the target machine.  The
 * second field gives the name of the addressing mode as it will appear on the
 * various reports.  The remaining fields must not be explicitly initialized.
 * The END constant must be used to mark the end of the information.
 */
struct {
    int mode;
    char *name;
    unsigned int count;
    unsigned int refs;
    unsigned int imcount;
    unsigned int imrefs;
    } addrout[] =

   { { EXPR,       "expression"   },
     { RREGISTER,  "long reg"     },
     { FREGISTER,  "float reg"    },
     { DREGISTER,  "double reg"   },
     { ABSOLUTE,   "absolute"     },
     { DISP,       "displacement" },
     { REGDFR,     "defered"      },
     { IMMED,      "immediate"    },
     { LABEL,      "label"        },
     { END                       } };


/*
 * Instruction class definitions.  These are usually defined as powers of
 * two to permit a single instruction to belong to more than one class.
 */
#define DATAMOVE (1 << 0)
#define ARITHLOG (1 << 1)
#define CONTROL  (1 << 2)
#define CALLSEQ  (1 << 3)
#define SLOTFILL (1 << 4)

/*
 * Define the instruction classes.  The first field indicates which instruction
 * class is being defined.  The second field denotes the name of the class as
 * it will appear on the various reports.  The remaining fields must not be
 * explicitly initialized.  The END constant must be used to mark the end of
 * the information.
 */
struct {
    int type;
    char *name;
    unsigned int count;
    } classout[] =

   { { ARITHLOG,    "arithmetic/logical" },
     { CALLSEQ,     "call sequence"      },
     { CONTROL,     "control"            },
     { DATAMOVE,    "data movement"      },
     { SLOTFILL,    "delay slot fill"    },
     { END                               } };


/*
 * Define the control flow classes.  These are defined and assigned to simplify
 * the tailoring of the control flow instruction distributions report.  The
 * class COND_BRANCH must exist, all others are optional.
 */
#define COND_BRANCH (1 << 0)
#define UNCOND_JUMP (1 << 1)
#define FUNC_CALL   (1 << 2)

/*
 * The first field indicates which control-flow class is being defined.  The
 * second field defines the name of the control flow class as it will appear
 * in reports. The remaining fields must not be implicitly initilaized.  The
 * END constant must be used to mark the end of the information.
 */
struct {
    int type;
    char *name;
    unsigned int count;
    } cfout[] =

   { { COND_BRANCH,   "conditional branches taken"  },
     { UNCOND_JUMP,   "unconditional jumps"         },
     { FUNC_CALL,     "function calls"              },
     { END                                          } };


/*
 * Instruction semantic record field definitions.  These aren't absolutely
 * necessary, but they do simplify the declaration of the instruction types.
 */
#define SR_BINST  { "left", "right", "destination" }
#define SR_CMP    { "left", "right" }
#define SR_MOV    { "source", "destination" }
#define SR_LOAD   { "memory", "destination" }
#define SR_STORE  { "source", "memory" }
#define SR_CALL   { "address" }
#define SR_JMP    { "label" }
#define SR_JMPL   { "label", "destination" }
#define SR_NOP


/*
 * Declare the instruction types.  The first field indicates which instruction
 * is being declared.  The items in this field are declared in the instdef.h
 * file for the target machine.  The second field indicates which instruction
 * class(es) the instruction belongs to.  If the instruction classes have been
 * defined to be powers of two, then a particular instruction may be placed
 * in more than one instruction class by using the "|" operator.  The third
 * field defines the name of the instruction as it will appear in the various
 * reports.  The fourth field indicates which control-flow class, if any, the
 * instruction belongs to.  The fifth field names the address fields of each
 * instruction.  These should correspond to the declaration of the sem_rec
 * structure in the recdef.h file for the target machine.  The number of
 * fields named must not exceed the value of the constant MAXFIELDS as declared
 * in the same recdef.h file.  The END constant must be used to mark the end
 * of the information.
 */
struct {
    int type;
    int class;
    char *name;
    int cflow;
    char *fieldname[MAXFIELDS];
    unsigned int count;
    unsigned int ntaken;
    } instout[] =

   { 
     /*	{ADDI   ,      ARITHLOG,    "add",      NONE,            SR_BINST  },
	{ANDI   ,      ARITHLOG,    "and",      NONE,            SR_BINST  },
	{DIVI   ,      ARITHLOG,    "div",      NONE,            SR_BINST  },
	{DIVUI  ,      ARITHLOG,    "divu",     NONE,            SR_BINST  },
	{MULI   ,      ARITHLOG,    "mul",      NONE,            SR_BINST  },
	{MACI   ,      ARITHLOG,    "mac",      NONE,            SR_BINST  },
	{ORI    ,      ARITHLOG,    "or",       NONE,            SR_BINST  },
	{REMI   ,      ARITHLOG,    "rem",      NONE,            SR_BINST  },
	{REMUI  ,      ARITHLOG,    "remu",     NONE,            SR_BINST  },
	{SLLI   ,      ARITHLOG,    "sll",      NONE,            SR_BINST  },
	{SRAI   ,      ARITHLOG,    "sra",      NONE,            SR_BINST  },
	{SRLI   ,      ARITHLOG,    "srl",      NONE,            SR_BINST  },
	{SUBI   ,      ARITHLOG,    "sub",      NONE,            SR_BINST  },
	{XORI   ,      ARITHLOG,    "xor",      NONE,            SR_BINST  },
	{NEGI   ,      ARITHLOG,    "neg",      NONE,            SR_BINST  },
	{NOTI   ,      ARITHLOG,    "not",      NONE,            SR_BINST  },
	{RSBI   ,      ARITHLOG,    "rsb",      NONE,            SR_BINST  },
	{LDI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{STI    ,      ARITHLOG,    "st",       NONE,            SR_STORE  },
	{CMP    ,      ARITHLOG,    "cmp",      NONE,            SR_CMP  },
	{BRELI  ,      ARITHLOG,    "brel",     NONE,            SR_JMP  },
	{BAI    ,      ARITHLOG,    "ba",       NONE,            SR_JMP  },
	{CALLI  ,      ARITHLOG,    "call",     NONE,            SR_CALL  },
	{RETI   ,      ARITHLOG,    "ret",      NONE,            SR_BINST  },
	{NOPI   ,      ARITHLOG,    "nop",      NONE,            SR_NOP  },
	{MOVI   ,      ARITHLOG,    "mov",      NONE,            SR_MOV  },
	{CVTI   ,      ARITHLOG,    "cvt",      NONE,            SR_BINST  },
	{SLLP   ,      ARITHLOG,    "srl",      NONE,            SR_BINST  },
	{SRAP   ,      ARITHLOG,    "sra",      NONE,            SR_BINST  },
	{SRLP   ,      ARITHLOG,    "srl",      NONE,            SR_BINST  },
	{MULP   ,      ARITHLOG,    "mul",      NONE,            SR_BINST  },
	{DIVP   ,      ARITHLOG,    "div",      NONE,            SR_BINST  },
	{DIVUP  ,      ARITHLOG,    "divu",     NONE,            SR_BINST  },
	{REMP   ,      ARITHLOG,    "rem",      NONE,            SR_BINST  },
	{REMUP  ,      ARITHLOG,    "remu",     NONE,            SR_BINST  },
	{CVTFI  ,      ARITHLOG,    "cvtfi",    NONE,            SR_BINST  },
	{CVTFL  ,      ARITHLOG,    "cvtfl",    NONE,            SR_BINST  },
	{CVTFU  ,      ARITHLOG,    "cvtfu",    NONE,            SR_BINST  },
	{CVTFUL ,      ARITHLOG,    "cvtul",    NONE,            SR_BINST  },
	{CVTIF  ,      ARITHLOG,    "cvtif",    NONE,            SR_BINST  },
	{CVTLF  ,      ARITHLOG,    "cvtlf",    NONE,            SR_BINST  },
	{CVTUF  ,      ARITHLOG,    "cvtuf",    NONE,            SR_BINST  },
	{CVTULF ,      ARITHLOG,    "cvtulf",   NONE,            SR_BINST  },
     */
	{ADDI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{ANDI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{DIVI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{DIVUI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{MULI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{ORI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{REMI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{REMUI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{SLLI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{SRAI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{SRLI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{SUBI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{XORI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{NEGI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{NOTI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{RSBI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{LDI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{STI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{CMP    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{BRELI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{BAI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{CALLI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{RETI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{NOPI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{MOVI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{CVTI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{MLAI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{MVNI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{LASTI    ,      ARITHLOG,    "ld",       NONE,            SR_LOAD  },
	{ END                                                            } };

/*
 * Define the hidden memory references.  These are provided to handle instances
 * where a particular instruction generates a known number of memory references
 * implicitly.  Since these references cannot be assigned to any particular
 * field, it is better that these items be handled through this separate
 * mechanism.  Also, it is possible to ascribe more than one type of memory
 * access to the same instruction using this mechanism.  The first field
 * indicates which instruction type contains the implicit memory reference.
 * The second field defines the addressing mode implicitly used to reference
 * memory.  The third field states the data type of the reference.  The fourth
 * field indicates how many such references are generated for each single
 * instruction executed.
 */
struct {
    int type;
    int mode;
    int datatype;
    unsigned int amount;
    } hmemout[] =

   { 
     { END                           } };
