
/************************************************************************/
/*                                                                      */
/* Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   */
/*     The Rector and Visitors of the University of Virginia            */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from the University of Virginia.                                */
/* This code is distributed WITHOUT ANY WARRANTY.  No claims are        */
/*      made as to whether it serves any particular purpose or even     */
/*      works at all.                                                   */
/*                                                                      */
/************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/arm/inst.c,v $
 $Date: 2004/01/21 18:39:45 $
 $Revision: 1.79 $
 $Name:  $

 Log of changes:

*************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>

#include "../lib/vpo.h"
#include "../lib/assign.h"
#include "../lib/combine.h"
#include "../lib/misc.h"

#include "recdef.h"
#include "addrdef.h"
#include "instdef.h"
#include "inst.h"
#include "regs.h"
#include "regs2.h"

#ifdef INTERACT

/***********************************************************

  The following definitions are preprocessor trickery to 
  selectively get output to go to a string instead of the
  standard output. A set of machine-independent routines 
  replace the standard output routines to get the desired
  effect. A replacement yyparse is also required. 

************************************************************/

#include "../lib/misc.h"
#ifdef putchar
#undef putchar
#endif

#define putchar(c) fputc((c), stdout)
#define printf sfprintf
#define fputc sfputc
#define puts sputs
#define fputs sfputs

#endif

#ifdef MEASURE
#include "linstinfo.h"

#define ST(a, b, c, d, e, f) stinstinfo(a, b, c, d, e, f)

static struct addr_entry __lab = { LONG, LABEL };
static struct addr_entry __rg0 = { LONG, RREGISTER };

#define LAB &(__lab)
#define RG0 &(__rg0)
#else
#define ST(a, b, c, d, e, f)
#endif

int minargrefs = 0;      /* Minimum number of argument references    */
                         /* required to make register replacement    */
                         /* economical                               */
int minlocrefs = 0;      /* Minimum number of local variable refs    */
                         /* required to make register replacement    */
                         /* economical                               */

int condret = 0;         /* indicates if RETURN can be the target    */
                         /* of a conditional jump, i.e., conditional */
                         /* returns are supported by the machine     */

int dassem = 0;          /* Assembly output flag                     */

char *labprt = ".%s:\n";    /* Used to print a label for the assembler  */
char *acomment = " @ %s\n";  /* Used to print a comment to the assembler */


int str_ptr, i_ptr;
int valid_integer;
int valid_negated;
int valid_flipped;
int valid_load;

static int pre_mod, post_mod, used_right_mem, used_right_mod, used_negated;

static struct addr_entry itable[ITABSIZE];

char *r_regs[16] = { "r0", "r1", "r2", "r3",
                            "r4", "r5", "r6", "r7",
                            "r8", "r9", "r10", "fp",
                            "r12", "sp", "lr", "pc"};
char *f_regs[8] = { "f0", "f1", "f2", "f3",
                            "f4",  "f5", "f6", "f7"};

static char *i2op[14] = { "", "\tadd", "\tand", "", "",
                          "\tmul", "\torr", "", "",
                          "", "", "", "\tsub", "\teor" };

static char *f2op[14]= { "", "\tad", "", "\tdv", "", 
                          "\tmu", "", "", "",
                          "", "", "", "\tsu", "" };

static char *i2expr[14] = { "", " + ", " & ", " / ", " / ", " * ", " | ", " % ",
                            " % ", " << ", " >> ", " >> ", " - ", " ^ " };


static char *sanyo_i2expr[14] = { "", " + ", " :AND: ", " / ", " / ", " * ", " :OR: ", " :MOD: ",
                            " :MOD: ", " :SHL: ", " :SHR: ", " :SHR: ", " - ", " :EOR: " };


#ifdef SOFTFLOAT 

/* the floating point operands that are valid */
static char *soft_f2op[14]= { "",            /* not used */
                         "\tbl\t__addsf3\n",    /* add */
                         "",            /* and */
                         "\tbl\t__divsf3\n",    /* divide */
                         "",            /* divide unsigned */
                         "\tbl\t__mulsf3\n",    /* multiply */
                         "",            /* or */
                         "",            /* remainder */
                         "",            /* unsigned remainder */
                         "",            /* logical shift left */
                         "",            /* arith shift right */
                         "",            /* logical shift right */
                         "\tbl\t__subsf3\n",    /* subtract */
                         "" };          /* exclusive or */

/* the double-proc operands that are valid */
static char *soft_d2op[14]= { "",            /* not used */
                         "\tbl      __adddf3\n",        /* add */
                         "",            /* and */
                         "\tbl      __divdf3\n",        /* divide */
                         "",            /* divide unsigned */
                         "\tbl      __muldf3\n",        /* multiply */
                         "",            /* or */
                         "",            /* remainder */
                         "",            /* unsigned remainder */
                         "",            /* logical shift left */
                         "",            /* arith shift right */
                         "",            /* logical shift right */
                         "\tbl      __subdf3\n",        /* subtract */
                         "" };          /* exclusive or */

static char *d_regs[14] =  {
		"{r0, r1}", 	/* d[0] */
		"{r1, r2}", 	/* d[1] */
		"{r2, r3}", 	/* d[2] */
		"{r3, r4}", 	/* d[3] */
		"{r4, r5}", 	/* d[4] */
		"{r5, r6}", 	/* d[5] */
		"{r6, r7}", 	/* d[6] */
		"{r7, r8}", 	/* d[7] */
		};

static char *dreg_half1[14]= {
		"r0", 	/* d[0] */
		"r1", 	/* d[1] */
		"r2", 	/* d[2] */
		"r3", 	/* d[3] */
		"r4", 	/* d[4] */
		"r5", 	/* d[5] */
		"r6", 	/* d[6] */
		"r7", 	/* d[7] */
		};
static char *dreg_half2[14]= {
		"r1", 	/* d[0] */
		"r2", 	/* d[1] */
		"r3", 	/* d[2] */
		"r4", 	/* d[3] */
		"r5", 	/* d[4] */
		"r6", 	/* d[5] */
		"r7", 	/* d[6] */
		"r8", 	/* d[7] */
	};

#endif

#ifdef MEASURE
#include "../lib/measure.h"
#endif

int 
contains_reg(struct addr_entry *t);
void 
check_prepost_mod(struct sem_rec *t, struct sem_rec* bin_st,
                  struct addr_entry* mem);
void 
valid_int(unsigned int b);
void 
set_fp_se(int);
void 
set_cc_se();
int 
contains_sr_expr(struct addr_entry *t);
void 
compare_preds(char* a, char* b);
int 
count_locals(struct addr_entry* t);
void 
check_local_count(char* insn_type, ...);


/*
 * addr - build record for addressing mode
 */
struct addr_entry *addr(int type, int mode, struct addr_entry *p1,
 struct addr_entry *p2, struct addr_entry *p3) 
{

    register struct addr_entry *t;
    int x;

    /*
     * Allocate an initialize an address record for this item
     */
    t = &itable[i_ptr++];

    /* we've ran out of recoverable itable entries.  now we start
	leaking them */
    if(i_ptr>=ITABSIZE)
    {
	t=(struct addr_entry *)CALLOC(sizeof(struct addr_entry));
    }
    t->type = type;
    t->mode = mode;
    t->arg1.ptr = p1;	/* p1 is an int casted to a ptr for exprs */
    t->arg2.ptr = p2;
    t->arg3.ptr = p3;

    switch (mode) 
    {

      case RREGISTER:
      case FREGISTER:
      case DREGISTER:
        t->cost = 1;
        break;

      case FCONST:
        t->cost = 2;
        break;

      case ICONST:
      {
        int a=atoi((char*)p1);
        if (!a)
          t->cost = 0;
        else
          t->cost = 1;
        break;
      }
      case LIDENT:
        ikind |= LOCREF;

      case GIDENT:
      case SPILLOC:
      case LABEL:
        t->cost = 2;
        break;

      case EXPR:
        if (isreloc(p2) && isreloc(p3))
          erflag = TRUE;
        if ((int) p1 == ANDI && isreloc(p2))
          erflag = TRUE;
        t->cost = 1;
        break;

      case IMMED:
        t->cost = p1->cost;
        break;

      case ABSOLUTE:
        /* check to make sure any constant is not greater than 65K */
        if (ConstantTooBig(p1)) 
           erflag = 1;
        t->cost = 5;
        break;

      case REGDFR:
      case DISP:
        t->cost = 6;
        break;

      case TEMPADDR:
        t->cost = 3;
        break;

      default:
        fprintf(stderr, "vpo (addr): bad addressing mode %d\n", mode);
        exit(1);
      }

    return(t);
}

struct addr_entry *sr_addr(int type, int mode, struct addr_entry *p1,
 struct addr_entry *p2, struct addr_entry *p3) 
{

    register struct addr_entry *t;
    int x;

    /*
     * Allocate an initialize an address record for this item
     */
    t = &itable[i_ptr++];

    /* we've ran out of recoverable itable entries.  now we start
	leaking them */
    if(i_ptr>=ITABSIZE)
    {
	t=(struct addr_entry *)CALLOC(sizeof(struct addr_entry));
    }
    t->type = type;
    t->mode = mode;
    t->arg1.ptr = p1;	/* p1 is an int casted to a ptr for exprs */
    t->arg2.ptr = p2;
    t->arg3.ptr = p3;

    switch (mode) 
    {
      case SR_EXPR:
	if((int)p1 == MULI)
	{
		int a=ConstantValue(p3);
	
		/* err if mul by a something that's not a power of 2. */
		if(((a) & (a-1))!=0)
			erflag=1;
		else if(a<0)
			erflag=1;
		else if(log2(a)>31)
			erflag=1;
	}
	else
	{
		if(!contains_reg(p3))	
		{
			int a=ConstantValue(p3);
			if(a<=0 || a>=32)
				erflag=1;
		}
	}
        t->cost = 2;
        break;

      default:
        fprintf(stderr, "vpo (addr): bad addressing mode %d\n", mode);
        exit(1);
      }

    return(t);
}

/*
 * icompare - compare two addr_entry type records for equality
 */
int icompare(struct addr_entry *a1, struct addr_entry *a2) 
{

    if (!a1)
      return(!a2);
    if (!a2)
      return(FALSE);
    if (a1->type != a2->type)
      return(FALSE);
    if (a1->mode == a2->mode) {
      switch (a1->mode) {

        case SR_EXPR:
        case EXPR:
          return(((char) (int)a1->arg1.str) == ((char) (int)a2->arg1.str) &&
                 icompare(a1->arg2.ptr, a2->arg2.ptr) &&
                 icompare(a1->arg3.ptr, a2->arg3.ptr));

        case DISP:
          return(icompare(a1->arg1.ptr, a2->arg1.ptr) &&
                 icompare(a1->arg2.ptr, a2->arg2.ptr));

        case SPILLOC:
        case IMMED:
        case ABSOLUTE:
        case REGDFR:
          return(icompare(a1->arg1.ptr, a2->arg1.ptr));

        case LABEL:
        case FCONST:
        case ICONST:
          return(((int) a1->arg2.ptr) == ((int) a2->arg2.ptr) &&
                 !strncmp(a1->arg1.str, a2->arg1.str, (int) a1->arg2.ptr));

        case RREGISTER:
        case FREGISTER:
        case DREGISTER:
        case GIDENT:
        case LIDENT:
          return(*a1->arg1.str == *a2->arg1.str &&
                 *(a1->arg1.str + 1) == *(a2->arg1.str + 1));

        default:
          fprintf(stderr, "vpo (icompare): unknown mode %d\n", a1->mode);
          exit(1);
        }
      }
    return(FALSE);
}

/*
 * dump_addr - emit a address expression in assembly format
 */
void dump_addr2(struct addr_entry *t, int start_const) 
{
   register char *s;
   register int x;

   switch (t->mode) {

      case EXPR:
         switch ((int) t->arg1.str) {

            case NEGI:
		if(!contains_reg(t) && start_const) 
			putchar('#');
               putchar('-');
	       dump_addr2(t->arg2.ptr, FALSE);
               break;

            case NOTI:
                if(start_const) putchar('#');
                if (debug_opt_i("SANYOASM"))
                        fputs(" :NOT: ",stdout);
                else    
                        putchar('~');
                dump_addr2(t->arg2.ptr, FALSE);
               break;


            case ADDI:

            if (isreloc(t->arg3.ptr)) 
            {
               dump_addr2(t->arg3.ptr, FALSE);
               putchar(' ');
               if (EXPRMODE(t->arg2.ptr) &&
                  ((int) t->arg2.ptr->arg1.ptr) == NEGI) 
               {
                  putchar('-');
                  putchar(' ');
                  dump_addr2(t->arg2.ptr->arg2.ptr, FALSE);
               }
               else 
               {
                 putchar('+');
                 putchar(' ');
                 dump_addr2(t->arg2.ptr, FALSE);
               }
            }
            else 
            {
	       if(start_const) putchar('#');
               dump_addr2(t->arg2.ptr, FALSE);
               putchar(' ');
               if (EXPRMODE(t->arg3.ptr) &&
                ((int) t->arg3.ptr->arg1.ptr) == NEGI) 
               {
                  putchar('-');
                  putchar(' ');
                  dump_addr2(t->arg3.ptr->arg2.ptr, FALSE);
               }
               else 
               {
                  putchar('+');
                  putchar(' ');
                  dump_addr2(t->arg3.ptr, FALSE);
               }
            }
            break;

          default:
             	if ( !contains_reg(t->arg3.ptr) && !contains_reg(t->arg2.ptr) && start_const)  
			putchar('#');
             	dump_addr2(t->arg2.ptr, FALSE);
		if (debug_opt_i("SANYOASM")) 
             		fputs(sanyo_i2expr[(int) t->arg1.str], stdout);
		else
             		fputs(i2expr[(int) t->arg1.str], stdout);
             	dump_addr2(t->arg3.ptr, FALSE);
             break;
        }
        break;

      case SR_EXPR:
	{
                dump_addr2(t->arg2.ptr, TRUE);
		if(ConstantValue(t->arg3.ptr)!=0 || contains_reg(t->arg3.ptr))
		{
         	  switch ((int) t->arg1.str) 
		  {
			case SLLI:
				printf( ", lsl ");
                		dump_addr2(t->arg3.ptr, TRUE);
				break;
			case SRAI:
				printf( ", asr ");
                		dump_addr2(t->arg3.ptr, TRUE);
				break;
			case SRLI:
				printf( ", lsr ");
                		dump_addr2(t->arg3.ptr, TRUE);
				break;
			case MULI:
			{
                		int log2(int a);
				int a;
                		fputs(", lsl #", stdout);
                		a=ConstantValue(t->arg3.ptr);
                		printf( "%d", log2(a));
				break;
			}
		  }
		}
		break;
	}
      case GIDENT:
        fputs(globalname(t->arg1.str), stdout);
	if(debug_opt("PIC"))
		printf( "(PLT) ");
        break;

      case LIDENT:
	if(start_const) putchar('#');
	if (debug_opt_i("SANYOASM")) {
	  printf("%d",(localrecord(t->arg1.str))->offset);
	} else
	  fputs(localname(t->arg1.str), stdout);
        break;

      case LABEL:
	 /* We want labels to have a dot in front */
	if (debug_opt_i("SANYOASM"))
	  putchar('|');
        putchar('.');
        s = t->arg1.str;

        for (x = (int) t->arg2.ptr; x--; s++)
          putchar(*s);
	if (debug_opt_i("SANYOASM"))
	  putchar('|');
        break;

      case ICONST:
	if(start_const) putchar('#');
        s = t->arg1.str;
        for (x = (int) t->arg2.ptr; x--; s++)
		putchar(*s);
	break;
      case FCONST:
	if(start_const) putchar('#');
        s = t->arg1.str;

        for (x = (int) t->arg2.ptr; x--; s++)
          putchar(*s);

        break;

      case SPILLOC:
	if(start_const) putchar('#');
        printf("%d", spilloff + atoi(t->arg1.ptr->arg1.str));
        break;


/* if we're doing soft float, dump floats out as r-regs */
#ifdef SOFTFLOAT
      case FREGISTER:
#endif
      case RREGISTER:
      {
	static int next_reg=0;
	if(regnum(t->arg1.str)>15) fputs(r_regs[regmap[next_reg++%14]], stdout);
        else fputs(r_regs[regmap[regnum(t->arg1.str)]], stdout);
          break;

      }
/* if we're doing hard float, dump floats out as f-regs */
#ifdef HARDFLOAT
      case FREGISTER:
#endif
      case DREGISTER:
      {
	static int next_reg=0;
#ifdef HARDFLOAT
	if(regnum(t->arg1.str)>15) fputs(f_regs[next_reg++%7], stdout);
        else  fputs(f_regs[regnum(t->arg1.str)], stdout);
#else
	if(regnum(t->arg1.str)) fputs(d_regs[regnum(t->arg1.str)], stdout);
        else  fputs(d_regs[regnum(t->arg1.str)], stdout);
#endif
      }
        break;

      case IMMED:
	if(start_const) putchar('#');
        if (CONMODE(t->arg1.ptr) && !atoi(t->arg1.ptr->arg1.str))
          fputs("0", stdout);
        else
          dump_addr2(t->arg1.ptr, FALSE);
        break;

      case ABSOLUTE:
        dump_addr2(t->arg1.ptr, FALSE);
        break;

      case DISP:
        putchar('[');
        dump_addr2(t->arg1.ptr,TRUE);
        putchar(',');
        dump_addr2(t->arg2.ptr,TRUE);
        putchar(']');
        break;

      case REGDFR:
        putchar('[');
        dump_addr2(t->arg1.ptr, TRUE);
        putchar(']');
        break;

      default:
        fprintf(stderr, "vpo (dump_addr): unknown addressing mode %d\n",
                               t->mode);
        exit(1);
      }
}
void dump_addr(struct addr_entry *t) 
{
	dump_addr2(t,TRUE);
}

/*
 * isreloc - is there a relocatable symbol in the expression
 */
int isreloc(struct addr_entry *p) 
{

    while (p) {
      switch (p->mode) {

        case GIDENT:
        case LABEL:
          return(TRUE);

        case EXPR:
        case DISP:
          if (isreloc(p->arg2.ptr))
            return(TRUE);
          p = p->arg3.ptr;
          break;

	case IMMED:
	case ABSOLUTE:
	  p = p->arg1.ptr;
	  break;

        default:
          return(FALSE);
        }
      }
    return(FALSE);
}

/*
 * initexpr - setup addr_entry to hold one register and an expression
 */
struct addr_entry *initexpr(struct addr_entry *reg, int neg, struct addr_entry *exp) 
{
    struct addr_entry *texp;

    texp = (neg) ? addr(ANY, EXPR, AE NEGI, exp, AN) : exp;
    return(addr(ANY, TEMPADDR, reg, texp, AN));
}

/*
 * addexpr - build expression onto the second argument
 */
struct addr_entry *addexpr(struct addr_entry *rexp, int op, struct addr_entry *exp) 
{

    rexp->arg2.ptr = addr(ANY, EXPR, AE op, rexp->arg2.ptr, exp);
    return(rexp);
}

/*
 * displacement - change the address item to a displacement mode address
 */
struct addr_entry *displacement(struct addr_entry *a, int type, int mode) 
{
    a->type = ANY;
    a->mode = DISP;
    a->cost = 6;

    return (a);
}

/*
 * memref - add type to memory reference item
 */
struct addr_entry *memref(int type, struct addr_entry *p, int max_disp) 
{
    unsigned int a;
    int b;
    p->type = type;

    if(p->arg2.ptr)
    	a=ConstantValue(p->arg2.ptr);
    else
	return (p);
    b=*(int*)&a;

#ifdef SOFTFLOAT
    if(type==DOUBLE && b!=0)
	erflag=1;

    /* adjust a float load/store to be the int load/store offset */
    if(type==FLOAT)
	max_disp=4095;
#endif

    if(max_disp < b || b < -max_disp)
	erflag=1;
    return(p);
}

/*
 * branch - check semantics of compare and branch
 */
void branch(struct sem_rec *t) 
{

    assert( NULL != t);
    ikind|=BRELI | DBOTH;
    if (dassem) 
    {
      	putchar('\t');
      	putchar('b');
      	fputs(t->pred, stdout);
      	putchar('\t');
      	dump_addr(t->sem.branch.lbl);
      	putchar('\n');
	ST(BRELI, LONG, CCUSED, AN, AN, AN);
    }
    else
      	icost = 10;
}

void mla(struct sem_rec *t, struct sem_rec* sets_cc)
{
    ikind=MULI;
    icost = 50;
    if(icompare(t->sem.mla.dst, t->sem.mla.left))
    {
	struct addr_entry* tmp;
	tmp=t->sem.mla.left;
	t->sem.mla.left=t->sem.mla.right;
	t->sem.mla.right=tmp;
    }
    if (dassem)
    {
       	fputs("\tmla",stdout);
	if(t->pred!=0)  fputs(t->pred, stdout);
	if(sets_cc)	fputs("s", stdout);
       	putchar('\t');
      	dump_addr(t->sem.mla.dst);
	putchar(',');
      	dump_addr(t->sem.mla.left);
	putchar(',');
      	dump_addr(t->sem.mla.right);
	putchar(',');
      	dump_addr(t->sem.mla.inc);
      	putchar('\n');
	ST(MLAI, LONG, 0, AN, AN, AN);
    }
    else
    {

	/* we flip the left/right earlier.. if they're still equal, the left==right
		and we have to reject */
    	if(icompare(t->sem.mla.dst, t->sem.mla.left))
		erflag=1;

	if(
		t->sem.mla.dst->mode!=RREGISTER ||
		t->sem.mla.left->mode!=RREGISTER ||
		t->sem.mla.right->mode!=RREGISTER ||
		t->sem.mla.inc->mode!=RREGISTER)
	{
		erflag=1;
		reason="Non r-reg in mul-accum";

	}

	if(sets_cc)
	{
		if(sets_cc->sem.binst.op!=MLAI) erflag=1;
		else
		{
			if(
		   	!icompare(t->sem.mla.left, sets_cc->sem.mla.left) ||
		   	!icompare(t->sem.mla.right, sets_cc->sem.mla.right) ||
		   	!icompare(t->sem.mla.inc, sets_cc->sem.mla.inc) 
		  	)
				erflag=1;
		}
		compare_preds(t->pred,sets_cc->pred);
	}
        icost = 100;

    }


}


void cmp(struct sem_rec *t)
{

	ikind|=CMP;
  if(dassem)
  {
	fputs("\tcmp", stdout);
        if(t->pred) fputs(t->pred, stdout);
	fputs("\t", stdout);
	dump_addr(t->sem.cmp.left);
	fputs(", ", stdout);
	dump_addr(t->sem.cmp.right);
        putchar('\n');
	ST(CMP, LONG, 0, AN, AN, AN);
  }
  else
  {
      	if (!COMPATIBLE(t->sem.cmp.left, t->sem.cmp.right))
         	erflag = 1;
	if(CONMODE(t->sem.cmp.right) || EXPRMODE(t->sem.cmp.right)
         || IMMEDMODE(t->sem.cmp.right))
        {
		int found_mode=0;
     		ConstantTooBig(t->sem.cmp.right);	
		if(!valid_integer)
			erflag=1;	
	}

	/* variadic func, the 0 terminates */
	check_local_count("cmp", t->sem.cmp.dst,t->sem.cmp.left, t->sem.cmp.right,NULL);
 	icost = 1 + t->sem.cmp.left->cost + t->sem.cmp.right->cost;
  }

}

/*
 * floating poitn branch
 */
void fbranch(struct sem_rec *t)
{
  ikind|=BRELI;
#ifdef INTERACT
  ikind |= FPI;
#endif  
    if (dassem) 
    {
      fputs("\tb", stdout);
      fputs(t->pred, stdout);
      putchar('\t');
      dump_addr(t->sem.branch.lbl);
      putchar('\n');
      ST(BRELI, LONG, CCUSED, AN, AN, AN);
    }
    else
      icost = 12;
}

void cc_semantic_check(struct sem_rec *t, struct sem_rec *sets_cc)
{

	/* if there's a set of the CC, make note that this might be a compare */
	ikind|=CMP;


	/* the op must be the same */
	if(t->sem.binst.op!=sets_cc->sem.binst.op)  erflag=1;

	/* the left and right operands must be the same */
	if(!icompare(t->sem.binst.right, sets_cc->sem.binst.right)) erflag=1;
	if(!icompare(t->sem.binst.left, sets_cc->sem.binst.left)) erflag=1;

	/* the destination of the CC operand must be zero, hopefully
	 * the machine description (md.y) will enforce this, but i'm checking
	 * just in case 
	 */
	if(sets_cc->sem.binst.dst!=0)	erflag=1;

	/* the predicates must be the same (yes, we allow predicated instructions to set the CC) */


	compare_preds(t->pred,sets_cc->pred);

	return;

}

void compare_preds(char* a, char*b)
{
	if(a||b)
	{
		if(a&&b)
		{
			if(strcmp(a,b)!=0)
				erflag=1;	
		}
		else
			erflag=1;

	}
}

void compares(struct sem_rec *sets_cc)
{
	/* operate and compare instructions, discard results */
	ikind|=CMP|ADDI;
	if(sets_cc->sem.binst.dst!=0)	erflag=1;
	if(dassem)
	{	
		switch(sets_cc->sem.binst.op)
		{
			case ADDI:
			{
				fputs("\tcmn", stdout);
				break;
			}
			case SUBI:
			{
				fputs("\tcmp", stdout);
				break;
			}
			case ANDI:
			{
				fputs("\ttst", stdout);
				break;
			}
			case XORI:
			{
				fputs("\tteq", stdout);
				break;
			}
			default:
				printf("unknown compare? :(\n");
				break;
		}
		if(sets_cc->pred) fputs(sets_cc->pred, stdout);
		fputs("\t", stdout);
                dump_addr(sets_cc->sem.binst.left);
                putchar(',');
                dump_addr(sets_cc->sem.binst.right);
                putchar('\n');
      
		ST(CMP, LONG, 0, AN, AN, AN);
	}
	else
	{	
		switch(sets_cc->sem.binst.op)
		{
			case ADDI:
			case SUBI:
			case ANDI:
			case XORI:
			{
        			/* now, lets check the constant */
      				if(!erflag && (IMMEDMODE(sets_cc->sem.binst.right) || CONMODE(sets_cc->sem.binst.right) ||
					EXPRMODE(sets_cc->sem.binst.right)))
      				{
        				/* we hunt for a "mode" to load this constant with */
        				int found_mode=0;
        				ConstantTooBig(sets_cc->sem.binst.right->arg1.ptr);

        				/* if it's valid on it's own, then we can just use it */
        				if(valid_integer)
        				{
                				/* do nothing */
        				} 
					else if (valid_negated)
        				{
                				/* is this an instruction we can use the opposite sign? */
                				if(sets_cc->sem.binst.op == ADDI || sets_cc->sem.binst.op == SUBI)
                        				found_mode=1;
						else	
							erflag=1;
        				}
					else
                				erflag = 1;
      				}
				break;
			}
			default:
				erflag=1;
				break;
		}

	}
      	icost = 10 + (2*sets_cc->sem.binst.left->cost) + sets_cc->sem.binst.right->cost;
}

/*
 * bin_emit - check semantics of binary instruction
 */
void bin_emit(struct sem_rec *t, struct sem_rec *sets_cc) 
{
    ikind|=ADDI;
#ifdef INTERACT
    if (REALTYPE(t->sem.binst.dst))
      ikind |= FPI;
    ikind |= t->sem.binst.op;
#endif    
    /* variadic func, the 0 terminates */
    check_local_count("binary", t->sem.binst.dst,t->sem.binst.left, t->sem.binst.right,NULL);
    if(t->sem.binst.op==MLAI)
    {
		mla(t,sets_cc);
		return;
    }

    if(sets_cc) cc_semantic_check(t, sets_cc);

    if (dassem) 
    {
#ifdef SOFTFLOAT
	if(t->sem.binst.dst->type==FLOAT)
	{
		printf( "%s\n", soft_f2op[t->sem.binst.op]);
		return;
	}
	else if(t->sem.binst.dst->type==DOUBLE)
	{
		printf( "%s\n", soft_d2op[t->sem.binst.op]);
		return;
	}
#endif

      if(t->sem.binst.op==RSBI)
      {
		fputs("\trsb", stdout);
        	if(t->pred) 	fputs(t->pred, stdout);
		if(sets_cc)	fputs("s", stdout);
		fputs("\t", stdout);
        	dump_addr(t->sem.binst.dst);
        	putchar(',');
        	dump_addr(t->sem.binst.left);
        	putchar(',');
        	dump_addr(t->sem.binst.right);
        	putchar('\n');
      }
      else if(t->sem.binst.op==SLLI)
      {
		fputs("\tmov", stdout);
        	if(t->pred) fputs(t->pred, stdout);
		if(sets_cc) fputs("s", stdout);
		fputs("\t", stdout);
        	dump_addr(t->sem.binst.dst);
        	putchar(',');
        	dump_addr(t->sem.binst.left);
		if(ConstantValue(t->sem.binst.right)!=0 || contains_reg(t->sem.binst.right))
		{
        		putchar(',');
			fputs(" lsl ", stdout);
        		dump_addr(t->sem.binst.right);
		}
        	putchar('\n');
      }
      else if(t->sem.binst.op==SRLI)
      {
                fputs("\tmov", stdout);
        	if(t->pred) fputs(t->pred, stdout);
		if(sets_cc) fputs("s", stdout);
		fputs("\t", stdout);
                dump_addr(t->sem.binst.dst);
                putchar(',');
                dump_addr(t->sem.binst.left);
		if(ConstantValue(t->sem.binst.right)!=0 || contains_reg(t->sem.binst.right))
		{
                	putchar(',');
                	fputs(" lsr ", stdout);
                	dump_addr(t->sem.binst.right);
		}
        	putchar('\n');
      }
      else if(t->sem.binst.op==SRAI)
      {
                fputs("\tmov", stdout);
        	if(t->pred) fputs(t->pred, stdout);
		if(sets_cc) fputs("s", stdout);
		fputs("\t", stdout);
                dump_addr(t->sem.binst.dst);
                putchar(',');
                dump_addr(t->sem.binst.left);
		if(ConstantValue(t->sem.binst.right)!=0 || contains_reg(t->sem.binst.right))
		{
                	putchar(',');
                	fputs(" asr ", stdout);
                	dump_addr(t->sem.binst.right);
		}
        	putchar('\n');
      }
      else if(t->sem.binst.op==MULI && t->sem.binst.dst->type==LONG)
      {
	 if(t->sem.binst.right->mode!=RREGISTER)
	 {
		unsigned int a;
		int log2(int a);
      	  	fputs("\tmov", stdout);
        	if(t->pred) fputs(t->pred, stdout);
		if(sets_cc) fputs("s", stdout);
		fputs("\t", stdout);
          	putchar('\t');
          	dump_addr(t->sem.binst.dst);
          	putchar(',');
          	dump_addr(t->sem.binst.left);
		fputs(", lsl #", stdout);
               	a=ConstantValue(t->sem.binst.right->arg1.ptr);
		printf( "%d\n", log2(a));
		ST(ADDI, LONG, 0, AN, AN, AN);
		return;
	 }
     	 fputs(i2op[t->sem.binst.op], stdout);
         if(t->pred) fputs(t->pred, stdout);
	 if(sets_cc) fputs("s", stdout);
	 fputs("\t", stdout);
         putchar('\t');
         dump_addr(t->sem.binst.dst);
         putchar(',');


  	 /* if dst and left are the same, flip the operands around */
  	 if(strncmp(t->sem.binst.dst->arg1.str,t->sem.binst.left->arg1.str,2)==0)
  	 {
      		dump_addr(t->sem.binst.right);
      		putchar(',');
        	dump_addr(t->sem.binst.left);
  	 }
  	 else
  	 {
        	dump_addr(t->sem.binst.left);
      		putchar(',');
     		dump_addr(t->sem.binst.right);
  	 }
        putchar('\n');
      }
      else
      {
        if(!REALTYPE(t->sem.binst.dst))
      	  fputs(i2op[t->sem.binst.op], stdout);
        else
      	  fputs(f2op[t->sem.binst.op], stdout);
        if (REALTYPE(t->sem.binst.dst))
        {
	  if(t->sem.binst.dst->type==FLOAT)
		  printf("fs");
	  else
		  printf("fd");
        }
        if(t->pred) fputs(t->pred, stdout);
	if(sets_cc) fputs("s", stdout);
	fputs("\t", stdout);
        dump_addr(t->sem.binst.dst);
        putchar(',');
        dump_addr(t->sem.binst.left);
        putchar(',');
        dump_addr(t->sem.binst.right);
        putchar('\n');
      }

/* FIXME, use the actual type here, instead of just calling all binaries an add */
      ST(ADDI, LONG, 0, AN, AN, AN);
    }
    else 
    {
      /* check the constant is <=31 for shifts */
      if(t->sem.binst.op==SLLI || t->sem.binst.op==SRLI ||
		t->sem.binst.op==SRAI)
      {
		if(IMMEDMODE(t->sem.binst.right)||EXPRMODE(t->sem.binst.right))
		{
               		unsigned int a=ConstantValue(t->sem.binst.right);
			int b=*(int*)&a;
			if(b>=32 || b<=-31)
				erflag=1; 
		}
		if(contains_sr_expr(t->sem.binst.right))
			erflag=1;
      }
      /* multiples can't use pseudo regs, otherwise the combiner gives
       * us shit.  Constants aren't allowed in multiples, and 
       * the left source and the dst of int multiples can't be the same
       */
      if(t->sem.binst.op==MULI)
      {
	if(!REGMODE(t->sem.binst.right))
	{
		unsigned int b=ConstantValue(t->sem.binst.right);
		int a=*(int*)&b;
		if((a&(a-1))!=0 || a<=0)
			/* hey, this constant isn't a power of two */
			erflag=1;
	}
	if(contains_sr_expr(t->sem.binst.right))
	{
		erflag=1;
		return;
	}
       
	/* integer multiplys can't have the dst and the left source the same */
	/* although this is true, we can rearrange this and use the right source 
		instead */
	if(     strncmp(t->sem.binst.dst->arg1.str,t->sem.binst.left->arg1.str,2)==0&&
		strncmp(t->sem.binst.dst->arg1.str,t->sem.binst.right->arg1.str,2)==0&&
			t->sem.binst.dst->type==LONG)
		erflag=1;

	/* lets not allow pseudo-regs in a multiply.. only combine after we get
		real registers */
	if(regnum(t->sem.binst.dst->arg1.str)>15)
		erflag=1;
	if(regnum(t->sem.binst.left->arg1.str)>15)
		erflag=1;
	if(REGMODE(t->sem.binst.right) &&regnum(t->sem.binst.right->arg1.str)>15)
		erflag=1;
      }

	/* double check that it's a valid operation because float and int
		aren't the same */
      if((t->sem.binst.dst->type==FLOAT || t->sem.binst.dst->type==DOUBLE)
		&& t->sem.binst.op<14 && *f2op[t->sem.binst.op]!='\t' )
		erflag=1;
      if((t->sem.binst.dst->type==WORD) && t->sem.binst.op<14  &&
		*i2op[t->sem.binst.op]!='\t')
		erflag=1;

      if(t->sem.binst.op==RSBI && t->sem.binst.dst->type!=LONG)
		erflag=1;
	
      if (isreloc(t->sem.binst.left) || isreloc(t->sem.binst.right))
        erflag = 1;

	/* now, lets check the constant */
      if(!erflag && IMMEDMODE(t->sem.binst.right))
      {
	/* we hunt for a "mode" to load this constant with */
	int found_mode=0;
     	ConstantTooBig(t->sem.binst.right->arg1.ptr);	

	/* if it's valid on it's own, then we can just use it */
	if(valid_integer)
	{	
		/* do nothing */
		found_mode=1;
	}

	/* add and substract instructions support "negated" constants
		by using add instead of subtract, the assembler deals
		with this for us, so we don't have to emit anything special */ 
	if (valid_negated)
	{
		/* is this an instruction we can use the opposite sign? */
		if(t->sem.binst.op == ADDI || t->sem.binst.op == SUBI)
			found_mode=1;
	}

	/* are there any that support a bit-flip of the operand? */
	if (valid_flipped)
	{
		/* any binary instructions we can "flip" the consant? */
		/* can use BIC (and-not) */
		if(t->sem.binst.op == ANDI )
			found_mode=1;
	}

	/* if we can't find a way to support it, give up */
	if(!found_mode)
        	erflag = 1;
      }
      icost = 10 + t->sem.binst.dst->cost + t->sem.binst.left->cost +
              t->sem.binst.right->cost;

#ifdef SOFTFLOAT
        if(t->sem.binst.dst->type==FLOAT)
                set_fp_se(TRUE);
        if(t->sem.binst.dst->type==DOUBLE)
                set_fp_se(FALSE);
#endif

    }
}

/*
 * fcmp - check semantics of float compare instruction
 */
void fcmp(struct sem_rec *t) 
{
	ikind|=CMP;
#ifdef INTERACT
	ikind |= FPI;
#endif	
    if (dassem) 
    {
#ifdef SOFTFLOAT
        if(t->sem.fcmp.left->type==FLOAT)
                printf( "\tbl\t__%ssf2\n", t->sem.fcmp.op);
        else
                printf( "\tbl\t__%sdf2\n", t->sem.fcmp.op);
#else

      fputs("\tcmfe\t", stdout);
      dump_addr(t->sem.fcmp.left);
      putchar(',');
      dump_addr(t->sem.fcmp.right);
      putchar('\n');
#endif
      ST(CMP, LONG, 0, AN, AN, AN);
    }
    else 
    {
#ifdef SOFTFLOAT
      	  if(t->sem.fcmp.op==0)
		erflag=1;
	  if(t->sem.fcmp.dst==0)
	  {
		erflag=1;
		return;
	  }
          if (!COMPATIBLE(t->sem.fcmp.left, t->sem.fcmp.right))
        	erflag = 1;
               if(strcmp(t->sem.fcmp.op, "eq")==0)      {}
          else if(strcmp(t->sem.fcmp.op, "ne")==0)      {}
          else if(strcmp(t->sem.fcmp.op, "ge")==0)      {}
          else if(strcmp(t->sem.fcmp.op, "le")==0)      {}
          else if(strcmp(t->sem.fcmp.op, "lt")==0)      {}
          else if(strcmp(t->sem.fcmp.op, "gt")==0)      {}
          else erflag=1;

        if(t->sem.fcmp.dst->type!=LONG)
                erflag=1;
        if(t->sem.fcmp.left->type==LONG)
                erflag=1;
        if(t->sem.fcmp.right->type==LONG)
                erflag=1;
        if(regnum(t->sem.fcmp.dst->arg1.str)!=0)
                erflag=1;
        if(regnum(t->sem.fcmp.left->arg1.str)!=0)
                erflag=1;
        if(regnum(t->sem.fcmp.right->arg1.str)==1 && t->sem.fcmp.right->type==FLOAT)
        {}
        else if(regnum(t->sem.fcmp.right->arg1.str)==2 && t->sem.fcmp.right->type==DOUBLE)
        {}
        else
                erflag=1;
#else
      if (!COMPATIBLE(t->sem.fcmp.left, t->sem.fcmp.right))
        erflag = 1;
      if(t->sem.fcmp.dst!=0)
	erflag=1;
#endif
      if(t->pred)
        erflag = 1;
      icost = 10; 

      check_local_count("fcmp",t->sem.cmp.left, t->sem.cmp.right,NULL);
      set_fp_se(TRUE);

    }
}

int bx_ok()
{
        /* walk the function looking for metadata.
         * when we find it, see if it specifies an irq function
         */

        struct bblock* cblk;
        struct list* nptr;
	extern struct bblock* top;

	if(debug_opt("bx_ok"))
		return TRUE;
        for (cblk = top; cblk; cblk = cblk->down)
        {
                for (nptr = cblk->rtls; nptr; nptr = nptr->next)
                {
                        if (nptr->type&METADATALINE)
                        {
                                if(strcmp(nptr->val, "bx_ok")==0)
                                        return TRUE;
                        }
                }
        }
        return FALSE;
}

/*
 * call - check semantics of branch and link (bl)
 */
void call(struct sem_rec *t) 
{
    int n;
    register char *a;
	ikind|=CALLI;

    if (dassem) 
    {
      	if (ABSMODE(t->sem.bl.target))
	{	
		struct addr_entry* target;
		char alloca_fn_name[10];

		if (debug_opt_i("SANYOASM")) 
		  strcpy(alloca_fn_name,"_alloca");
		else
		  strcpy(alloca_fn_name,"alloca");
		
		target=t->sem.bl.target;
		if(target->mode==ABSOLUTE)
			target=target->arg1.ptr;
		if(target->mode==GIDENT && strcmp(globalname(target->arg1.str),
			 alloca_fn_name)==0)
		{
                        extern int arg_build_size;
                        assert(!t->pred);
			if (!debug_opt_i("SANYOASM")) 
			  fputs(" @ vpo found call to alloca\n", stdout);
                        fputs("\tmvn\tr1, #3\n", stdout);
                        fputs("\tadd\tr0, r0, #3\n", stdout);
                        fputs("\tand\tr0, r0, r1\n", stdout);
                        fputs("\tsub\tsp, sp, r0\n", stdout);
                        printf( "\tadd\tr0,sp,#%d", arg_build_size);
		}
		else
		{
			fputs("\tbl", stdout);
			if(t->pred)  fputs(t->pred,stdout);	
			fputs("\t", stdout);
        		dump_addr(t->sem.bl.target);
		}
	}
      	else
	{
		fputs("\tmov",stdout);
		if(t->pred)  fputs(t->pred,stdout);	
		fputs("\tlr, pc\n",stdout);
		if(is_interwork_func() || bx_ok())
			fputs("\tbx",stdout);
		else
			fputs("\tmov",stdout);
		if(t->pred)  fputs(t->pred,stdout);	
		fputs("\t", stdout);
		if(!is_interwork_func() && !bx_ok())
			fputs("\tpc, ",stdout);
       	 	dump_addr(t->sem.bl.target->arg1.ptr);
	}
      	putchar('\n');
        ST(CALLI, LONG, 0, AN, AN, AN);
    }
    else 
    {
      if (!(t->sem.bl.target->mode & (ABSOLUTE | REGDFR)))
        erflag = 1;

      if (side_effect) 
      {
	  char a[REGSTRING+3], *b, *c;
	  char reg[REGSTRING];
	  char *se=side_effect;
	  /* pointer to this rtl */
	  extern struct list* curptr;
	  struct list* next;

	  assert(curptr);
	  next=curptr->next;
	  if(next && (next->type&PARMLINE) != PARMLINE)
			next=0;

	  	
	  if(next)
	  {
	  	c=next->val;
	  	while(c=gtrg(c))
	  	{
			copyloc(reg,unskiploc(c));
			*skiploc(reg)=0;
	  		sprintf(a,"=%s;",reg);
	  		b=a;
          		while (*side_effect = *b++)
            			side_effect++;
			
	  	}
	  }
      }
      icost = 50 + t->sem.bl.target->cost;
    }
}

/*
 * jump - check semantics of jmp instruction
 */
void jump(struct sem_rec *t) 
{
	ikind|= BAI | DBOTH;
    if (dassem) 
    {
	assert(!t->pred);
	if(t->sem.jmp.target->mode==RREGISTER)
	{
      		fputs("\tmov\tpc,", stdout);
      		dump_addr(t->sem.jmp.target);
      		putchar('\n');
	}
	else
	{
      		fputs("\tb\t", stdout);
      		dump_addr(t->sem.jmp.target);
      		putchar('\n');
	}
        ST(BAI, LONG, 0, AN, AN, AN);
    }
    else
    {
	char *a,*b,*c;
	assert(!t->pred);
      	icost = 12;
	/* jumps to globals are really return instructions converted by fixentry */
	if(t->sem.jmp.target->mode==GIDENT)
	{
		if(side_effect)
      		{
		  char a[REGSTRING+3], *b;
		  char reg[REGSTRING];
		  char *se=side_effect;
		  /* pointer to this rtl */
		  extern struct list* curptr;
		  struct list* next;
	
		  assert(curptr);
		  next=curptr->prev;
		  while(next)	
		  {
		  	if((next->type&PARMLINE) == PARMLINE)
				break;
			if(cnts(next))
			{
				next=NULL; 	
				break;
			}
			next=next->prev;
		  }
	  	  if(next)
	  	  {
		  	c=next->val;
		  	while(c=gtrg(c))
		  	{
				copyloc(reg,unskiploc(c));
				*skiploc(reg)=0;
		  		sprintf(a,"=%s;",reg);
		  		b=a;
       		   		while (*side_effect = *b++)
       		     			side_effect++;
				
		  	}
		  }
      		}
	}
    }
	
}

/*
 * l - check semantics of load operation
 */
void l(struct sem_rec *t) 
{
    int x;
#ifdef MEASURE
    int itype;
#endif
    ikind|=LDI;
#ifdef INTERACT
    if (REALTYPE(t->sem.l.dst))
      ikind |= FPI;
#endif    

    if (dassem) 
    {
      /* we check here so we can find out if it's pre_mod or post_mod easily */
      pre_mod=post_mod=used_right_mem=used_right_mod=0;
      if(t->sem.l.incr)
	check_prepost_mod(t, t->sem.l.incr, t->sem.l.mem); 
      if (REALTYPE(t->sem.l.dst))
      {
#ifdef HARDFLOAT
	if(t->sem.l.dst->type==FLOAT)
      		fputs("\tldfs", stdout);
	else
      		fputs("\tldfd", stdout);
#else
	if(t->sem.l.dst->type==FLOAT)
      		fputs("\tldr", stdout);
	else
	{
      		fputs("\tldmia ", stdout);
        	dump_addr(t->sem.l.mem->arg1.ptr);
      		printf( ", ");
        	dump_addr(t->sem.l.dst);
      		printf( "\n");
		ST(LDI, 0, 0, AN, AN, AN);
		return;
	}
#endif
      }
      else
      {
      	fputs("\tldr", stdout);
        if(t->pred)  fputs(t->pred,stdout);	
	if(t->sem.l.unsign && CONMODE(t->sem.l.unsign))
	{	
		/* do nothing, ldr is unsigned on its own */
	}
	else if(t->sem.l.unsign)
	{
		/* else, it's signed, emit the S.*/
		if(t->sem.l.mem->type!=LONG)
			/* but only if it's less than a 32 bit load */
			fputs("s", stdout);
	}
	if(t->sem.l.mem->type==BYTE)
		fputs("b", stdout);
	else if(t->sem.l.mem->type==WORD)
		fputs("h", stdout);
      }
      putchar('\t');
      dump_addr(t->sem.l.dst);
      putchar(',');
      if(!used_right_mem)
        dump_addr(t->sem.l.mem);
      else
      {
        printf("[");
        dump_addr(t->sem.l.mem->arg2.ptr);
        printf(",");
        dump_addr(t->sem.l.mem->arg1.ptr);
        printf("]");
      }
      if(pre_mod)
      { 
        /* the addr expression is correct in the mem ref, we simply */
        /* have to emit an ! to allow a write-back                  */
        printf("!");
      }
      else if(post_mod)
      {
        /* the memory address looks OK, but we have to emit info about how to update */
        printf(", ");
        if(used_right_mod)
	{
		if(contains_reg(t->sem.l.incr->sem.binst.right) && used_negated)		
			printf("-");
		else if(!contains_reg(t->sem.l.incr->sem.binst.right) && used_negated)		
			printf("#-");
                dump_addr2(t->sem.l.incr->sem.binst.right,!used_negated);
	}
        else
	{
		if(contains_reg(t->sem.l.incr->sem.binst.left) && used_negated)		
			printf("-");
		else if(!contains_reg(t->sem.l.incr->sem.binst.left) && used_negated)		
			printf("#-");
                dump_addr2(t->sem.l.incr->sem.binst.left,!used_negated);
	}
      } 
      printf("\n");

#ifdef MEASURE
      t->sem.l.mem->mode |= MEMREF;
      ST(LDI, t->sem.l.mem->type, 0, t->sem.l.mem, t->sem.l.dst, AN);
      t->sem.l.mem->mode &= ~MEMREF;
#endif

    }
    else 
    {
      if (t->sem.l.unsign) 
      {

        /*
         * A load signed byte or halfword also performs a sign extension,
         * while a load unsigned byte or halfword also performs a zero-fill.
         */


	/* it's a constand mode if it's &255 or &65535 */
        if (!CONMODE(t->sem.l.unsign)) {


	  /* check signed variants */
          x = atoi(t->sem.l.unsign->arg1.ptr->arg1.str);
          if (!(LONGTYPE(t->sem.l.dst) && WORDTYPE(t->sem.l.mem) && x == 16) &&
              !(LWTYPE(t->sem.l.dst) && BYTETYPE(t->sem.l.mem) && x == 24))
            erflag = 1;

	
 	  /* signed halfs and bytes don't have sr-expr variants */
	  /* don't forget to check the increment portion for an SR expr */
	  if(contains_sr_expr(t->sem.l.mem) || (t->sem.l.incr && contains_sr_expr(t->sem.l.incr->sem.binst.right)))
		erflag=1;
        }
        else 
        {
	  /* check unsigned variants */
          x = atoi(t->sem.l.unsign->arg1.str);
          if (!(LWTYPE(t->sem.l.dst) && WORDTYPE(t->sem.l.mem) && x == 65535) &&
              !(!REALTYPE(t->sem.l.dst) && BYTETYPE(t->sem.l.mem) && x == 255))
            erflag = 1;


 	  /* unsigned halfs don't have sr-expr variants */
	  if(WORDTYPE(t->sem.l.mem))
	  {
	  	if(contains_sr_expr(t->sem.l.mem) || (t->sem.l.incr && contains_sr_expr(t->sem.l.incr->sem.binst.right)))
			erflag=1;
	  }
        }

      }
      else if (!COMPATIBLE(t->sem.l.mem, t->sem.l.dst)) 
      {
	/* if the dst is float, the mem type better be float */
	if( t->sem.l.dst->type==FLOAT && t->sem.l.mem->type != FLOAT)
		erflag=1;
	/* if the dst is double, the mem type better be double */
	else if( t->sem.l.dst->type==DOUBLE && t->sem.l.mem->type != DOUBLE)
		erflag=1;
	/* otherwise, the dst is an r-reg, the mem better not be double/float*/
	else if(t->sem.l.dst->type==LONG && 
		(t->sem.l.mem->type == FLOAT || t->sem.l.mem->type == DOUBLE))
		erflag=1;

      }

      /* must specify signed or unsigned for byte/word types */
      if(!t->sem.l.unsign && 
		(t->sem.l.mem->type==BYTE || t->sem.l.mem->type==WORD))
	erflag=1;

      /* no such thing as sign/unsigned  long load/store */
      /* float/doubles are always signed */
      /* thus, can't specify signed/unsigned type */
      if(t->sem.l.unsign && (t->sem.l.mem->type==LONG ||
		t->sem.l.mem->type==FLOAT || t->sem.l.mem->type==DOUBLE))
	erflag=1;

#ifdef HARDFLOAT
	if(t->sem.s.mem->type==FLOAT || t->sem.s.mem->type==DOUBLE)
	if(t->sem.l.mem->arg2.ptr && !contains_reg(t->sem.l.mem->arg2.ptr)
 	   && (ConstantValue(t->sem.l.mem->arg2.ptr)&0x3)!=0)
		erflag=1;
#endif

      if(t->sem.l.incr)
	check_prepost_mod(t, t->sem.l.incr, t->sem.l.mem); 

      icost = 20 + t->sem.l.mem->cost + t->sem.l.dst->cost;
    }
    check_local_count("load",t->sem.l.mem, t->sem.l.dst,NULL);
}

/*
 * move - check semantics of mov instruction
 */
void move(struct sem_rec *t, struct sem_rec* sets_cc) 
{
  ikind|=MOVI;
#ifdef INTERACT
  if(REALTYPE(t->sem.mov.src))
     ikind |= FPI;
#endif

    if (dassem)  
    {
#ifdef HARDFLOAT
      if (REALTYPE(t->sem.mov.src))
      {
#if 0
/* this doesn't work because apparently adding 0 to 
	a floating point number doesn't guarentee the
	answer is identical.  gcc's ref run fails in
	real_value_truncate when trying to convert a 1.6
	from a double to a float. */

	if(t->sem.mov.src->type==t->sem.mov.dst->type && 
		t->sem.mov.dst->type==FLOAT)
		fputs("\tadfs\t", stdout);
	else if(t->sem.mov.src->type==t->sem.mov.dst->type ) 
		fputs("\tadfd\t", stdout);
	else
	{
		fprintf(stderr, "Cannot assemble move rtl :( \n");
		return;
	}

        dump_addr(t->sem.mov.dst);
        putchar(',');
        dump_addr(t->sem.mov.src);
	fputs("\t, #0",stdout);
        putchar('\n');
#endif
	if(t->sem.mov.src->type==t->sem.mov.dst->type && 
		t->sem.mov.dst->type==FLOAT)
		fputs("\tmvfs\t", stdout);
	else if(t->sem.mov.src->type==t->sem.mov.dst->type ) 
		fputs("\tmvfd\t", stdout);
	else
	{
		fprintf(stderr, "Cannot assemble move rtl :( \n");
		return;
	}

        dump_addr(t->sem.mov.dst);
        putchar(',');
        dump_addr(t->sem.mov.src);
        putchar('\n');
      }
      else
      {
        printf("\tmov");
        if(t->pred)  fputs(t->pred,stdout);	
        if(sets_cc) fputs("s",stdout);
        fputs("\t", stdout);
        dump_addr(t->sem.mov.dst);
        putchar(',');
        dump_addr(t->sem.mov.src);
        putchar('\n');
      }
#else
	if(t->sem.mov.dst->type==DOUBLE)
	{
		if(regnum(t->sem.mov.dst->arg1.str)==regnum(t->sem.mov.src->arg1.str)+1) 
		{
			printf("\tmov\t%s,%s\n",
				dreg_half2[regnum(t->sem.mov.dst->arg1.str)],
				dreg_half2[regnum(t->sem.mov.src->arg1.str)]
				);
			printf("\tmov\t%s,%s\n",
				dreg_half1[regnum(t->sem.mov.dst->arg1.str)],
				dreg_half1[regnum(t->sem.mov.src->arg1.str)]
				);
		}
		else
		{
			printf("\tmov\t%s,%s\n",
				dreg_half1[regnum(t->sem.mov.dst->arg1.str)],
				dreg_half1[regnum(t->sem.mov.src->arg1.str)]
				);
			printf("\tmov\t%s,%s\n",
				dreg_half2[regnum(t->sem.mov.dst->arg1.str)],
				dreg_half2[regnum(t->sem.mov.src->arg1.str)]
				);
		}
	
		return;
	}
	else{
	  printf("\tmov");
	  if(t->pred)  fputs(t->pred,stdout);	
	  if(sets_cc) fputs("s",stdout);
	  fputs("\t", stdout);
	  dump_addr(t->sem.mov.dst);
	  putchar(',');
	  dump_addr(t->sem.mov.src);
	  putchar('\n');
	}
#endif
        ST(MOVI, LONG, 0, AN, AN, AN);
    }
    else 
    {
      	int expr=EXPRMODE(t->sem.mov.src);
      	int iconst=CONMODE(t->sem.mov.src);
      	int immed_mode=IMMEDMODE(t->sem.mov.src);
      	if (!COMPATIBLE(t->sem.mov.dst, t->sem.mov.src)) 
      	{
        	if (!(LONGTYPE(t->sem.mov.src) && BWTYPE(t->sem.mov.dst)) ||
            		!(WORDTYPE(t->sem.mov.src) && BYTETYPE(t->sem.mov.dst)))
          	erflag = 1;
      	}
      	if((iconst || expr || immed_mode) && t->sem.mov.dst->type==LONG)
      	{
		int a=ConstantTooBig(t->sem.mov.src);
		if(!valid_integer && !valid_flipped)
			erflag=1;
      	}
      	else if (immed_mode || iconst || expr)
		erflag=1;

      	if (icompare(t->sem.mov.dst, t->sem.mov.src))
        	icost = 0;
      	else
        	icost = 3 + t->sem.mov.dst->cost + t->sem.mov.src->cost;

        if(sets_cc)
        {
                if(CONMODE(sets_cc->sem.cmp.right) || EXPRMODE(sets_cc->sem.cmp.right) || 
			IMMEDMODE(sets_cc->sem.cmp.right))
                {
                        int value=ConstantValue(sets_cc->sem.cmp.right);
                        if(value!=0) erflag=1;
                }
                else
                        erflag=1;
		compare_preds(t->pred,sets_cc->pred);

                if(!icompare(t->sem.mov.src,sets_cc->sem.cmp.left))      erflag=1;
		if(sets_cc->sem.cmp.type=='?')				 erflag=1;



		/* make sure the cost is non-zero if we're setting the CC */
		icost++;
        }

    }
    check_local_count("move",t->sem.mov.dst, t->sem.mov.src,NULL);
}


void fixz(struct sem_rec *t)
{
#ifdef INTERACT
     ikind |= FPI;
#endif
    if (dassem)
    {
#ifdef HARDFLOAT
        fputs("\tfixz\t", stdout);
      	dump_addr(t->sem.mov.dst);
      	putchar(',');
      	dump_addr(t->sem.mov.src);
      	putchar('\n');
        ST(CVTI, LONG, 0, AN, AN, AN);
#else
        if(t->sem.mov.src->type==DOUBLE)
                fputs("\tbl\t__fixdfsi\n", stdout);
        else
                fputs("\tbl\t__fixsfsi\n", stdout);
        ST(CALLI, LONG, 0, AN, AN, AN);
#endif
    }
    else
    {
	if(t->sem.mov.dst->type!=LONG)
		erflag=1;
	if(t->sem.mov.src->type!=DOUBLE && t->sem.mov.src->type!=FLOAT)
		erflag=1;
        icost = 3 + t->sem.mov.dst->cost + t->sem.mov.src->cost;
    }
}
void mvfs(struct sem_rec *t)
{
#ifdef INTERACT
     ikind |= FPI;
#endif
    if (dassem)
    {
#ifdef HARDFLOAT
        fputs("\tmvfs\t", stdout);
      	dump_addr(t->sem.mov.dst);
      	putchar(',');
      	dump_addr(t->sem.mov.src);
	putchar('\n');
        ST(CVTI, LONG, 0, AN, AN, AN);
#else
        fputs("\tbl\t__truncdfsf2\n", stdout);
        ST(CALLI, LONG, 0, AN, AN, AN);
#endif
    }
    else
    {
	if(t->sem.mov.dst->type!=FLOAT)
		erflag=1;
	if(t->sem.mov.src->type!=DOUBLE)
		erflag=1;
        icost = 3 + t->sem.mov.dst->cost + t->sem.mov.src->cost;
    }
}

void flt(struct sem_rec *t)
{
#ifdef INTERACT
     ikind |= FPI;
#endif
    if (dassem)
    {
#ifdef HARDFLOAT
        fputs("\tflt", stdout);
	if(t->sem.mov.dst->type==FLOAT)
		fputs("s\t",stdout);
	else
		fputs("d\t",stdout);
      	dump_addr(t->sem.mov.dst);
      	putchar(',');
      	dump_addr(t->sem.mov.src);
      	putchar('\n');
        ST(CVTI, LONG, 0, AN, AN, AN);
#else
        if(t->sem.mov.dst->type==FLOAT)
                fputs("\tbl\t__floatsisf\n", stdout);
        else
                fputs("\tbl\t__floatsidf\n", stdout);
        ST(CALLI, LONG, 0, AN, AN, AN);
#endif
    }
    else
    {
	if(t->sem.mov.dst->type!=FLOAT && t->sem.mov.dst->type!=DOUBLE)
		erflag=1;
	if(t->sem.mov.src->type==FLOAT || t->sem.mov.src->type==DOUBLE)
		erflag=1;
        icost = 3 + t->sem.mov.dst->cost + t->sem.mov.src->cost;
    }
}

void mvfd(struct sem_rec *t)
{
#ifdef INTERACT
     ikind |= FPI;
#endif
    if (dassem)
    { 
#ifdef HARDFLOAT
        fputs("\tmvfd\t", stdout);
        dump_addr(t->sem.mov.dst);
        putchar(',');
        dump_addr(t->sem.mov.src);
	putchar('\n');
        ST(CVTI, LONG, 0, AN, AN, AN);
#else
        fputs("\tbl\t__extendsfdf2\n", stdout);
        ST(CALLI, LONG, 0, AN, AN, AN);
#endif
    }
    else
    {
        if(t->sem.mov.dst->type!=DOUBLE)
                erflag=1;
        if(t->sem.mov.src->type!=FLOAT)
                erflag=1;
        icost = 3 + t->sem.mov.dst->cost + t->sem.mov.src->cost;
    }


}

void cvt(struct sem_rec *t)
{
    ikind|=CVTI;
	if(REALTYPE(t->sem.mov.dst) && LONGTYPE(t->sem.mov.src))
		flt(t);
	else if(LONGTYPE(t->sem.mov.dst) && REALTYPE(t->sem.mov.src))
		fixz(t);
	else if(FLOATTYPE(t->sem.mov.dst) && DOUBLETYPE(t->sem.mov.src))
		mvfs(t);
	else if(DOUBLETYPE(t->sem.mov.dst) && FLOATTYPE(t->sem.mov.src))
		mvfd(t);
	else
		erflag=1;

	/* no conversion operations can be predicated */
	if(t->pred)
		erflag=1;

#ifdef SOFTFLOAT
        if(regnum(t->sem.mov.dst->arg1.str)!=0 || 
		regnum(t->sem.mov.src->arg1.str)!=0)
                erflag=1;

        if(t->sem.binst.dst->type==FLOAT || t->sem.binst.dst->type==LONG)
                set_fp_se(TRUE);
        if(t->sem.binst.dst->type==DOUBLE)
                set_fp_se(FALSE);
#endif
        icost = 3 + t->sem.mov.dst->cost + t->sem.mov.src->cost;

}

void cvtu(struct sem_rec *t)
{
#ifdef SOFTFLOAT
        ikind|=CVTI;
        if (dassem)
        {
                if(FLOATTYPE(t->sem.mov.src))
                        fputs("\tbl\t__fixunssfsi\n", stdout);

                else if(DOUBLETYPE(t->sem.mov.src))
                        fputs("\tbl\t__fixunsdfsi\n", stdout);
        	ST(CALLI, LONG, 0, AN, AN, AN);
        }
        if(!LONGTYPE(t->sem.mov.dst))
                erflag=1;
        if(LONGTYPE(t->sem.mov.src))
                erflag=1;
        if(regnum(t->sem.mov.dst->arg1.str)!=0 || regnum(t->sem.mov.src->arg1.str)!=0)
                erflag=1;
        if(t->sem.binst.dst->type==FLOAT || t->sem.binst.dst->type==LONG)
                set_fp_se(TRUE);
        if(t->sem.binst.dst->type==DOUBLE)
                set_fp_se(FALSE);
#else
	erflag=1;
#endif
        icost = 3 + t->sem.mov.dst->cost + t->sem.mov.src->cost;
}



/*
 * adr - check semantics of adr instruction
 */
void adr(struct sem_rec *t) 
{
    ikind|=MOVI;
    assert (!REALTYPE(t->sem.mov.src));
    if (dassem)
    {
      /* no floating point addresses */
      printf("\tadr");
      if(t->pred) fputs(t->pred, stdout);
      fputs("\t", stdout);
      dump_addr(t->sem.mov.dst);
      putchar(',');
      dump_addr(t->sem.mov.src);
      putchar('\n');
      ST(ADDI, LONG, 0, AN, AN, AN);
    }
    else
    {
      if (!COMPATIBLE(t->sem.mov.dst, t->sem.mov.src))
      {
        if (!(LONGTYPE(t->sem.mov.src) && BWTYPE(t->sem.mov.dst)) &&
            !(WORDTYPE(t->sem.mov.src) && BYTETYPE(t->sem.mov.dst)))
          erflag = 1;
        }
    }
    icost = 3 + t->sem.mov.dst->cost + t->sem.mov.src->cost;
}


void nop() 
{
    	icost = 0;
	ikind|=NOPI;
     	ST(NOPI, LONG, 0, AN, AN, AN);
}

/*
 * retrn - check semantics of return
 */
void retrn() 
{
    char *a;
    char *b;

	ikind|=RETI;
    if (dassem) 
    {
	/* 
         * we actually don't need to "return" because the ldmea
         * that fixentry adds is OK.  we still keep this so that
         * the optimizer knows where the return locatiosn are 
	 */
   	ST(RETI, LONG, 0, AN, AN, AN);
    }
    else 
    {
      icost = 10;
      if ((b=side_effect) && sret)
        for (a = sret; *b = *a++; b++);
    }
}

/*
 * neg - check semantics of negate operation
 */
void neg(struct sem_rec *t, struct sem_rec* sets_cc) 
{
	ikind|=NEGI;
    if (dassem) 
    {
      if (REALTYPE(t->sem.neg.src))
      {
	fprintf(stderr, "No floating point neg instr\n");
        printf("\tneg.%c\t", TYPECHAR(t->sem.neg.src));
      }
      else
      {
        fputs("\trsb", stdout);
        if(t->pred) fputs(t->pred, stdout);
        if(sets_cc) fputs("s",stdout);
        fputs("\t", stdout);
      }
      dump_addr(t->sem.neg.dst);
      putchar(',');
      dump_addr(t->sem.neg.src);
      fputs(", #0\n", stdout);
      ST(NEGI, LONG, 0, AN, AN, AN);
    }
    else 
    {
      	if (!COMPATIBLE(t->sem.mov.src, t->sem.mov.dst))
        	erflag = 1;
      	if(t->sem.mov.src->type!=LONG)
		erflag=1;
       	if(sets_cc)
       	{
               	if(!icompare(t->sem.mov.src,sets_cc->sem.mov.src))      erflag=1;
		compare_preds(t->pred,sets_cc->pred);
       	}
      	icost = 2 + t->sem.mov.src->cost;
    }
}

/*
 * mvn - check semantics of negate operation
 */
void mvn(struct sem_rec *t, struct sem_rec* sets_cc) 
{

	ikind|=MOVI;
    if (dassem)
    {
      	fputs("\tmvn", stdout);
      	if(t->pred) fputs(t->pred, stdout);
      	if(sets_cc) fputs("s", stdout);
      	fputs("\t", stdout);
      	dump_addr(t->sem.neg.dst);
      	putchar(',');
      	dump_addr(t->sem.neg.src);
      	fputs("\n", stdout);
        ST(MVNI, LONG, 0, AN, AN, AN);
    }
    else
    {
      	if (!COMPATIBLE(t->sem.mov.src, t->sem.mov.dst))
        	erflag = 1;
      	if(t->sem.mov.src->type!=LONG)
        	erflag=1;
      	icost = 2 + t->sem.mov.src->cost;
       	if(sets_cc)
       	{
               	if(!icompare(t->sem.mov.src,sets_cc->sem.mov.src))      erflag=1;
		compare_preds(t->pred,sets_cc->pred);
       	}
    }
}


/*
 * not - check semantics of move complement operation
 */
void not(struct sem_rec *t) 
{

	ikind|=NOTI;
    if (dassem) 
    {
      fputs("\tnot", stdout);
      if(t->pred) fputs(t->pred, stdout);
      fputs("\t", stdout);
      dump_addr(t->sem.not.dst);
      putchar(',');
      dump_addr(t->sem.not.src); 
      putchar('\n');
      ST(NOTI, t->sem.neg.dst->type, 0, t->sem.neg.dst, t->sem.neg.src, AN);
    }
    else 
    {
      if (!COMPATIBLE(t->sem.not.src, t->sem.not.dst) ||
          REALTYPE(t->sem.not.src))
        erflag = 1;
      icost = 2 + t->sem.not.src->cost;
    }
}

/*
 * s - check semantics of store operation
 */
void s(struct sem_rec *t) 
{
  ikind|=STI;
#ifdef INTERACT
  if (REALTYPE(t->sem.s.src)) 
    ikind |= FPI;
#endif

    if (dassem) 
    {
      /* we check here so we can find out if it's pre_mod or post_mod easily */
      pre_mod=post_mod=used_right_mem=used_right_mod=0;
      if(t->sem.s.incr)
	check_prepost_mod(t, t->sem.s.incr, t->sem.s.mem); 
      if (REALTYPE(t->sem.s.src)) 
      {
#ifdef HARDFLOAT
	if(t->sem.s.src->type==FLOAT)
      		fputs("\tstfs\t", stdout);
	else
      		fputs("\tstfd\t", stdout);
#else
        if(t->sem.l.dst->type==FLOAT)
                fputs("\tstr\t", stdout);
        else
        {
                fputs("\tstmia ", stdout);
                dump_addr(t->sem.s.mem->arg1.ptr);
                printf( ", ");
                dump_addr(t->sem.s.src);
                printf( "\n");
		ST(STI, 0, 0, AN, AN, AN);
                return;
        }
#endif
      }
      else 
      {
        printf("\tstr");
        if(t->pred) fputs(t->pred, stdout);
        if(t->sem.s.mem->type==BYTE)
                fputs("b", stdout);
        else if(t->sem.s.mem->type==WORD)
                fputs("h", stdout);
	fputs("\t", stdout);
      }

      dump_addr(t->sem.s.src);
      putchar(',');
      if(!used_right_mem)
      	dump_addr(t->sem.s.mem);
      else
      {
	printf("[");
	dump_addr(t->sem.s.mem->arg2.ptr);
	printf(",");
	dump_addr(t->sem.s.mem->arg1.ptr);
	printf("]");
      }
      if(pre_mod)	
      {
	/* the addr expression is correct in the mem ref, we simply */
	/* have to emit an ! to allow a write-back		    */
	printf("!");
      }
      else if(post_mod)
      {
        /* the memory address looks OK, but we have to emit info about how to update */
        printf(", ");
        if(used_right_mod)
	{
		if(contains_reg(t->sem.s.incr->sem.binst.right) && used_negated)		
			printf("-");
		else if(!contains_reg(t->sem.s.incr->sem.binst.right) && used_negated)		
			printf("#-");
                dump_addr2(t->sem.s.incr->sem.binst.right,!used_negated);
	}
        else
	{
		if(contains_reg(t->sem.s.incr->sem.binst.left) && used_negated)		
			printf("-");
		else if(!contains_reg(t->sem.s.incr->sem.binst.left) && used_negated)		
			printf("#-");
                dump_addr2(t->sem.s.incr->sem.binst.left,!used_negated);
	}
      } 
	
      putchar('\n');
#ifdef MEASURE
      t->sem.s.mem->mode |= MEMREF;
      ST(STI, t->sem.s.mem->type, 0, t->sem.s.mem, t->sem.s.src, AN);
      t->sem.s.mem->mode &= ~MEMREF;
#endif
    }
    else 
    {
      if (!COMPATIBLE(t->sem.s.mem, t->sem.s.src)) 
      {
        /* if the dst is float, the mem type better be float */
        if( t->sem.s.src->type==FLOAT && t->sem.s.mem->type != FLOAT)
                erflag=1;
        /* if the dst is double, the mem type better be double */
        else if( t->sem.s.src->type==DOUBLE && t->sem.s.mem->type != DOUBLE)
                erflag=1;
        /* otherwise, the dst is an r-reg, the mem better not be double/float*/
	else if(t->sem.s.src->type==LONG && 
		(t->sem.s.mem->type == FLOAT || t->sem.s.mem->type == DOUBLE))
                erflag=1;
      }



	/* check for shift expressions in strh, stfs, and stfd */
	if(contains_sr_expr(t->sem.s.mem) || (t->sem.s.incr && contains_sr_expr(t->sem.s.incr->sem.binst.right)))
	{

		if(t->sem.s.mem->type==LONG)
		{	/* do nothing */ ;	}
		else if(t->sem.s.mem->type==BYTE )
		{	/* do nothing */ ;	}
		else
		{
			reason="Shift expressions not allowed in strh, stfs and stfd";
			erflag=1;
		}
	}

      	if(t->sem.s.incr)
		check_prepost_mod(t, t->sem.s.incr, t->sem.s.mem); 
#ifdef HARDFLOAT

	if(t->sem.s.mem->type==FLOAT || t->sem.s.mem->type==DOUBLE)
		if(t->sem.s.mem->arg2.ptr && !contains_reg(t->sem.s.mem->arg2.ptr)
 	   	   && (ConstantValue(t->sem.s.mem->arg2.ptr)&0x3)!=0)
			erflag=1;
#endif
      icost = 20 + t->sem.s.mem->cost + t->sem.s.src->cost;
    }
    check_local_count("move",t->sem.s.src, t->sem.s.mem,NULL);
}

/*
 * ConstantTooBig - make sure constant doesn't exceed 16-bit signed constant
 */
unsigned int ConstantValue(struct addr_entry *t);
int ConstantTooBig(struct addr_entry *t)
{
	unsigned int constValue=ConstantValue(t);


	/* ask the int solver if this const is valid */
	valid_int(constValue);


	/* hey, it just might be OK.  Lets give it a try */  
	return !(valid_integer || valid_negated || valid_flipped || valid_load);
}



unsigned int ConstantValue(struct addr_entry *t) 
{
   char *s;
   unsigned int a=65535; /* initially a bad value */

   switch (t->mode) 
   {
      case SR_EXPR:
		a=0;
		break;
      case EXPR:
      {
	    unsigned int b,c; 
	    int operation=(int)t->arg1.ptr;

            if(operation== NEGI)
	    {
		int d;
	        b= ConstantValue(t->arg2.ptr); 
                d=-*(int*)&b;
		a=*(unsigned int*)&d;
            }
            else if(operation== NOTI)
	    {
	        b= ConstantValue(t->arg2.ptr); 
                a=~b;
	    }
            else if(operation== ADDI)
	    {
	        b= ConstantValue(t->arg2.ptr); 
		c= ConstantValue(t->arg3.ptr); 
                a=b+c;
	    }
            else if(operation== SUBI)
	    {
	        b= ConstantValue(t->arg2.ptr); 
		c= ConstantValue(t->arg3.ptr); 
                a=b-c;
	    }
            else if(operation== REMI || operation==REMUI)
	    {
	        b= ConstantValue(t->arg2.ptr); 
		c= ConstantValue(t->arg3.ptr); 
		if(c!=0)
                	a=b%c;
	    }
            else if(operation== DIVI ||  operation== DIVUI)
	    {
	        b= ConstantValue(t->arg2.ptr); 
		c= ConstantValue(t->arg3.ptr); 
		if(c!=0)
            	    a=b/c;
	    }
            else if(operation== MULI)
	    {
	        b= ConstantValue(t->arg2.ptr); 
		c= ConstantValue(t->arg3.ptr); 
                a=b*c; 
	    }
	    else if(operation== SLLI)
	    {
	        b= ConstantValue(t->arg2.ptr); 
		c= ConstantValue(t->arg3.ptr); 
	     	a=b<<c;	
	    }
	    else if(operation== SRLI)
	    {
	        b= ConstantValue(t->arg2.ptr); 
		c= ConstantValue(t->arg3.ptr); 
		a=(*(unsigned int*)&b)>>c;
	    }
	    else if(operation== SRAI)
	    { 
	        b= ConstantValue(t->arg2.ptr); 
		c= ConstantValue(t->arg3.ptr); 
		a=(*(int*)&b)>>c;
	    }
	    else if(operation== XORI)
	    { 
	        b= ConstantValue(t->arg2.ptr); 
		c= ConstantValue(t->arg3.ptr); 
		a=b^c;
	    }
	    else if(operation== ORI)
	    { 
	        b= ConstantValue(t->arg2.ptr); 
		c= ConstantValue(t->arg3.ptr); 
		a=b|c;
	    }
	    else if(operation== ANDI)
	    { 
	        b= ConstantValue(t->arg2.ptr); 
		c= ConstantValue(t->arg3.ptr); 
		a=b&c;
	    }
	    else
		fprintf(stderr, "Unkonwn expr mode\n");

	break;
      }
      case LIDENT:
	a=localrecord(t->arg1.str)->offset;
	break;
      case SPILLOC:
	 a=atoi(t->arg1.str+2);
         break;
      case GIDENT:
      case LABEL:
      case RREGISTER:
	 a=0;
         break;


      case IMMED:
	a=ConstantValue(t->arg1.ptr);
	break;

      case ICONST:
	/* arm can have an 8 bit constant rotated right by
		 -14, -12, ..., -2, 0, 2, 4 ... 16 
		first pass:  just the 8 big constant not shifted 
	*/
         a=strtoul(t->arg1.str,0,10);
	break;

      case FCONST:
         break;

      default:
        fprintf(stderr, "vpo (ConstantTooBig): unknown addr mode %d\n",
		t->mode);
        exit(1);
   }
   return a;
}


void valid_int(unsigned int b)
{
        int a;
        unsigned int c;
        a=*(int*)&b;
        a=-a;
        c=*(unsigned int*)&a;

        valid_integer=valid_int_help(b);
        valid_negated=valid_int_help(c);
        valid_flipped=valid_int_help(~b);

	if (!(valid_integer || valid_negated || valid_flipped))
	  valid_integer = rotated_int_valid(b);

	if(-4095<= a && a <= 4095)
		valid_load=1;
	else
		valid_load=2;


}

int valid_int_help(unsigned int a)
{
        unsigned int b=1;
        unsigned int c;
        int first_bit_set=-1;
        int last_bit_set=-1;
        int i=0;

        if(a==0)
                return 1;

        if(a<0)
                return 0;

        c=*(unsigned int*)&a;

        while(b)
        {
                if(c&b)
                {
                        last_bit_set=i;
                        if(first_bit_set==-1)
                                first_bit_set=i;
                }
                b<<=1;
                i++;
        }

        if(( last_bit_set - (first_bit_set&(~1))+1)<=8)
                return 1;

        return 0;
}

/* Return true if the integer is valid after rotate by 16 */
int rotated_int_valid(unsigned int a)
{
  unsigned int x = a & 0xffff,y= a>>16;
  x = (a << 16) | y;
  return valid_int_help(x);
}

int contains_reg(struct addr_entry *t)
{
   char *s;
   unsigned int a=65535; /* initially a bad value */

   switch (t->mode) 
   {
      case SR_EXPR:
		return TRUE;
      case EXPR:
      {
	    unsigned int b,c; 
	    int operation=(int)t->arg1.ptr;

            if(operation== NEGI || operation== NOTI)
	    {
	        return contains_reg(t->arg2.ptr); 
            }
            else if(operation== ADDI || operation== SUBI || 
            		operation== REMI || operation==REMUI ||
            		operation== DIVI ||  operation== DIVUI ||
            		operation== MULI || operation== SLLI ||
	    		operation== SRLI || operation== SRAI ||
	    		operation== XORI || operation== ORI ||
	    		operation== ANDI)
	    {
	        return contains_reg(t->arg2.ptr)|| contains_reg(t->arg3.ptr); 
	    }
	    else
	    {
		fprintf(stderr, "Unkonwn expr mode\n");
		exit(1);
	    }
	break;
      }
      case GIDENT:
      case LIDENT:
      case LABEL:
      case SPILLOC:
      case ICONST:
      case FCONST:
	return FALSE;
         break;
      case RREGISTER:
      case FREGISTER:
      case DREGISTER:
	return TRUE;


      case IMMED:
	return contains_reg(t->arg1.ptr);
	break;

      default:
        fprintf(stderr, "vpo (contains_reg): unknown addr mode %d\n",
		t->mode);
        exit(1);
   }
}

int contains_sr_expr(struct addr_entry *t)
{
   char *s;
   unsigned int a=65535; /* initially a bad value */

   if(!t)
	return FALSE;

   switch (t->mode) 
   {
      case SR_EXPR:
		return TRUE;
      case EXPR:
      {
	    unsigned int b,c; 
	    int operation=(int)t->arg1.ptr;

            if(operation== NEGI || operation== NOTI)
	    {
	        return contains_sr_expr(t->arg2.ptr); 
            }
            else if(operation== ADDI || operation== SUBI || 
            		operation== REMI || operation==REMUI ||
            		operation== DIVI ||  operation== DIVUI ||
            		operation== MULI || operation== SLLI ||
	    		operation== SRLI || operation== SRAI ||
	    		operation== XORI || operation== ORI ||
	    		operation== ANDI)
	    {
	        return contains_sr_expr(t->arg2.ptr)|| contains_sr_expr(t->arg3.ptr); 
	    }
	    else
	    {
		fprintf(stderr, "Unkonwn expr mode\n");
		exit(1);
	    }
	break;
      }
      case GIDENT:
      case LIDENT:
      case LABEL:
      case SPILLOC:
      case ICONST:
      case FCONST:
      case RREGISTER:
      case FREGISTER:
      case DREGISTER:
      case ABSOLUTE:
	return FALSE;
      case DISP:
          return(contains_sr_expr(t->arg1.ptr) ||
                 contains_sr_expr(t->arg2.ptr));
      case REGDFR:
          return(contains_sr_expr(t->arg1.ptr) );


      case IMMED:
	return contains_sr_expr(t->arg1.ptr);
	break;

      default:
        fprintf(stderr, "vpo (contains_reg): unknown addr mode %d\n",
		t->mode);
        exit(1);
   }
}

void checklabel(struct addr_entry* t)
{
	/* since each instruction count 4 bytes, the differnce between
	 * where the label is and where the instruction that references
	 * it should be a multiple of 4.  Thus, the lower two bits of
	 * that constant will be 0.  The offset from the pc can then be
	 * as high as 1024 before we see a problem.  1024/4=255 so we 	
	 * have a potential for 255 instructions.  
	 *
	 * The current fixentry can insert lots of instructions, 2 to
	 * fix the stackpointer, 4 to move arguments, 2 to store registers
	 * and more if the offsets are large.  Lots more.  We'll assume
	 * it inserts 2+4+2=8 instructions to be safe.
	 *
	 * return locations likewise get new instructions.  1 to fix the
	 * stack pointer, 2 to restore registers.	
	 *
	 * We could try to account for large stack offsets, but that
	 * could be tough, and likely unprofitable.  Watch for future bugs
	 * that move a label because of large stack frames.
	 */ 
	#define PRELOGUE_INSTRS 8
	#define EPILOGUE_INSTRS 3
	#define MAX_INSTRS	255
	extern struct list* curptr;
	struct list* ptr;
	struct bblock* cblk;
	extern struct bblock* top;
	int counter=PRELOGUE_INSTRS;
	int count_non_rtls=FALSE;
	int rets=0;
	int in_data=TRUE;

	/* OK, we already know that t represents a label in a move/load
	 * instruction.   What we need to do is look from the beginning
	 * of this function until we find curptr.  If curptr
	 * is within a reasonable number of instructions from
	 * the beginning, all is well.  Otherwise, we set an error flag. 
	 */

        for (cblk = top; cblk; cblk = cblk->down)
        {
                for (ptr = cblk->rtls; ptr; ptr = ptr->next)
                {
			/* stop when we find it */
			if(curptr==ptr || counter>MAX_INSTRS)
				goto check;

			/* we won't want to count asmlines, etc
			 * until we reach the start of the function.
			 * we start counting when we it becomes
			 * possible to have a real instruction
			 */ 

			/* a non-empty rtl line */
			if((ptr->type&RTLINE) == RTLINE && ptr->val)
			{
				in_data=FALSE;
				count_non_rtls=TRUE;
				counter++;
			}
			if(((ptr->type&RTLINE) == RTLINE) && ptr->val &&
				strcmp(ptr->val, "PC=RT;")==0)
			  counter+=EPILOGUE_INSTRS;
			if(((ptr->type&ASMLINE) == ASMLINE) && ptr->val &&
			  strcmp(ptr->val, ".data")==0)
				in_data=TRUE;
			if(((ptr->type&ASMLINE) == ASMLINE) && ptr->val &&
			  strcmp(ptr->val, ".data")==0)
				in_data=FALSE;
		}
	}

/*
	fprintf(stderr, "Warning! Label lookup failed.  :< accepting label\n");
here, we used to accept this, now we reject on general princple, it can't be
valid
*/
	erflag=1;
	return;

check:
	if(counter>MAX_INSTRS)
	{
		erflag=1;
	}
}

int is_pre_modified(struct sem_rec *t, struct addr_entry* mem)
{
      	if(!t->sem.l.incr)
		return FALSE;

	if( mem->mode == REGDFR )
		/* we have just a register thus it's a post increment */
		return FALSE;
	else if( mem->mode == DISP)
		/* we have just a register thus it's a pre increment */
		return TRUE;
	else if( mem->mode == ABSOLUTE)
		/* this won't work either */
		return FALSE;
	else
		fprintf(stderr, "Error!  unknown addressing mode\n"); 
		assert(0);
	
	return FALSE;
}
int is_post_modified(struct sem_rec *t, struct addr_entry* mem)
{
      	if(!t->sem.l.incr)
		return FALSE;


	if( mem->mode == REGDFR )
		/* we have just a register thus it's a post increment */
		return TRUE;
	else if( mem->mode == DISP)
	{
		/* we have just a register thus it's a pre increment 
		 * but, it might be a reg+0 or reg-0 which we can 
		 * potentially treat as a post-mod 
		 * instead, this is less restrictive 
		 */

		/* arg1 should be the pointer to the register */
		struct addr_entry* arg1=mem->arg1.ptr;
	
		/* arg2 should be a pointer to +/- reg or const expr 
		 * if it doesn't contain a register and the expr is equal
		 * to zero then we can do the post increment 
	 	 */
		struct addr_entry* arg2=mem->arg2.ptr;
		if(!contains_reg(arg2) && ConstantValue(arg2)==0)
			return TRUE;
		return FALSE;
	}
	else if( mem->mode == ABSOLUTE)
		/* this won't work either */
		return FALSE;
	else
		fprintf(stderr, "Error!  unknown addressing mode\n"); 
		assert(0);
	
	return FALSE;
}


#define set_error_and_ret { erflag=1; return; }

void 
check_prepost_mod(struct sem_rec *t, struct sem_rec* bin_instr,
                  struct addr_entry* mem)
{

	if(bin_instr->sem.binst.dst==0 ||
	   bin_instr->sem.binst.left==0 ||
	   bin_instr->sem.binst.right==0 )
	{
		/* ah, this may be a move instead of a binary instruction */
		/* it's only valid if the dst it the same as the src */
		/* and, if so, we don't treat it as a pre/post mod at all */
		if(bin_instr->sem.mov.src->mode==RREGISTER && 
			strncmp(bin_instr->sem.mov.dst->arg1.str,
			bin_instr->sem.mov.src->arg1.str,2)==0)
			return;
		erflag=1;
		return;
	}
#ifdef SOFTFLOAT
        if(mem->type==DOUBLE || mem->type==FLOAT)
	{
		erflag=1;
		return;
	}
#endif
	if(t->pred && bin_instr->pred)
	{
		/* assure predicates are equal */
		if(strcmp(t->pred, bin_instr->pred)!=0)
		{
			erflag=1;
			return;
		}
			
	}
	else if(t->pred || bin_instr->pred)
	{
		erflag=1;
		return;
	}
	
	
	used_right_mem=used_right_mod=FALSE;
	pre_mod=is_pre_modified(t,mem);
	post_mod=is_post_modified(t,mem);
	used_negated=FALSE;

	if(!pre_mod && !post_mod)
		set_error_and_ret;


	#define try_next	  { post_mod=0; goto next;	}
	if(post_mod)
	{
		/* now, we hope we know this is a load/store of the form
 		 *
 		 *	r[x]=R[r[y]];<binary instruction>
		 *
		 * now we try to check that r[y] is the defined operation in
		 * the binary instruction 
		 */
		struct addr_entry* arg1=mem->arg1.ptr;
		char* base_reg;
		int constant_value;

		assert(arg1->mode==RREGISTER);
		assert(arg1->arg1.str);
	
		base_reg=arg1->arg1.str;

		/* this has to be a r[y]=r[]+/- */
		/* the right side doesn't have to be a register */
		if(bin_instr->sem.binst.dst->mode!=RREGISTER)
			try_next;
		if(bin_instr->sem.binst.left->mode!=RREGISTER)
			try_next;

		/* the binary instruction has to set the same reg as the base */
		if(strncmp(bin_instr->sem.binst.dst->arg1.str, base_reg,2)!=0)
			try_next;


		/* if it's an add, we check that it's OK */
		if(bin_instr->sem.binst.op==ADDI)
		{
			/* when we add, one of the sources has to match the dest */
			if(bin_instr->sem.binst.left->mode==RREGISTER &&
				strncmp(bin_instr->sem.binst.left->arg1.str, base_reg,2)==0)
			{
				/* the left one matches */

				/* the modifier is on the right */
				used_right_mod=TRUE;
			}
			else if(bin_instr->sem.binst.right->mode==RREGISTER &&
				strncmp(bin_instr->sem.binst.right->arg1.str, base_reg,2)==0)
			{
				/* the right one matches */
			}
			else
				try_next;
			
		}
		/* if it's a sub, we check that it's OK */
		else if(bin_instr->sem.binst.op==SUBI)
		{
			/* when we sub, the left has to match the dest */
			if(bin_instr->sem.binst.left->mode==RREGISTER &&
				strncmp(bin_instr->sem.binst.left->arg1.str, base_reg,2)==0)
			{
				/* the left one matches */
				/* the modifier is on the right */
				used_right_mod=TRUE;
				used_negated=TRUE;
			}
			else
				try_next;
			
		}
		/* we can't support any other instructions, give up */
		else
			try_next;




/* 
 *
 *Does this mean we can support pre/post-increments/decrements with a shift expression?
 *
 */




		

		/* if we get here, we know it's an add/sub and that the left item is a
			register.  We also know that the base_reg used in
			the load is both used and set in this instruction */ 
		/* now, if the right side is a register, or an integer in the valid range, 
			we're cool :) */


		if(bin_instr->sem.binst.right->mode==RREGISTER)
			goto looks_ok;
		
		constant_value=ConstantValue(bin_instr->sem.binst.right);
	
		if(!check_const_range(mem,constant_value))
			try_next;	
		goto looks_ok;
	}

/* if we get here, post_modification wasn't successful.  Lets try pre-mod */
next:
#undef try_next
#define try_next	  { pre_mod=0; goto next2;	}
	if(pre_mod)
	{
		/* now, we hope we know this is a load/store of the form
 		 *
 		 *	r[x]=R[r[y]+-<reg or expr>];<binary instruction>
		 *
		 * now we try to check that r[y] is the defined operation in
		 * the binary instruction 
		 */
		struct addr_entry* arg1=mem->arg1.ptr;
		struct addr_entry* arg2=mem->arg2.ptr;
		char* base_reg1=0;
		char* base_reg2=0;
		char* base_reg=0;
		int constant_value;

		assert(arg1->mode==RREGISTER);
		assert(arg1->arg1.str);
	
		base_reg1=arg1->arg1.str;
		if(arg2->mode==RREGISTER)
		{
			base_reg2=arg2->arg1.str;
		}

		/* OK, we may have 2 canidates for "base" registers */
		/* let's start parsing the binary op and see if we can */
		/* figure out which one is the one we want to use as base */

		/* this has to be a r[y]=r[]+/- */
		/* the right side doesn't have to be a register */
		if(bin_instr->sem.binst.dst->mode!=RREGISTER)
			try_next;
		if(bin_instr->sem.binst.left->mode!=RREGISTER)
			try_next;

		/* the binary instruction has to set the same reg as the base */
		if(strncmp(bin_instr->sem.binst.dst->arg1.str, base_reg1,2)==0)
			base_reg=base_reg1;
		else if(base_reg2 && strncmp(bin_instr->sem.binst.dst->arg1.str, base_reg2,2)==0)
		{
			base_reg=base_reg2;
			used_right_mem=TRUE;
		}
		else
			/* the thing defined isn't base_reg1 or 2, this can't be a valid pre-mod */
			try_next;

		/* if it's an add, we check that it's OK */
		if(bin_instr->sem.binst.op==ADDI)
		{
			/* when we add, one of the sources has to match the dest */
			if(bin_instr->sem.binst.left->mode==RREGISTER &&
				strncmp(bin_instr->sem.binst.left->arg1.str, base_reg,2)==0)
			{
				/* the left one matches */
				/* now we have to compare the right argument of the bin_op
				/* to the "other" argument of the of the address expression */
				struct addr_entry *other, *right;
				if(base_reg==base_reg2)	 other=mem->arg1.ptr;
				else			 other=mem->arg2.ptr;

				right=bin_instr->sem.binst.right;
				/* the binary instruction may have an 'immed' node above it */
				if(right->mode==IMMED)
					right=right->arg1.ptr;
                                if(!icompare(right,other))
					try_next;

				/* hey! these appear to match! */

			}
			else if(bin_instr->sem.binst.right->mode==RREGISTER &&
				strncmp(bin_instr->sem.binst.right->arg1.str, base_reg,2)==0)
			{
				/* the right one matches */
				/* now we have to compare the left argument of the bin_op
				/* to the "other" argument of the of the address expression */
				struct addr_entry *other, *left;
				if(base_reg==base_reg2)	 other=mem->arg1.ptr;
				else			 other=mem->arg2.ptr;

				used_right_mod=TRUE;

				left=bin_instr->sem.binst.left;
				/* the binary instruction may have an 'immed' node above it */
				if(left->mode==IMMED)
					left=left->arg1.ptr;
				if(!icompare(left, other))
					try_next;

				/* hey! these appear to match! */

			}
			else
				try_next;
			
		}
		/* if it's a sub, we check that it's OK */
		else if(bin_instr->sem.binst.op==SUBI)
		{
			/* when we sub, the left has to match the dest */
                        if(bin_instr->sem.binst.left->mode==RREGISTER &&
                                strncmp(bin_instr->sem.binst.left->arg1.str, base_reg,2)==0)
                        {
				struct addr_entry *right;
                                /* the left one matches */
				/* but we can't allow: r[x]=R[r[z]-r[y]]; r[y]=r[y]-10 */
				/* so we have to make sure we used base_reg1 */
                                if(base_reg!=base_reg1) 
					try_next;
				/* otherwise we compare the 2nd half of the memory to */
				/* the 2nd half of the binary operator */

				right=bin_instr->sem.binst.right;
				/* the binary instruction may have an 'immed' node above it */
				if(right->mode==IMMED)
					right=right->arg1.ptr;

				/* the memory op is going to have a "neg" node about it if it's 						negative.  We should zap this before icompare */	
				if((int)(mem->arg2.ptr->arg1.str)!=NEGI)
					try_next;
                                if(!icompare(right, mem->arg2.ptr->arg2.ptr))
                                        try_next;

                                /* hey! these appear to match! */
				used_negated=TRUE;
                        }

			else
				try_next;
			
		}
		/* we can't support any other instructions, give up */
		else
			try_next;
		

		/* if we get here, we know it's an add/sub and that the left item is a
			register.  We also know that the base_reg used in
			the load is both used and set in this instruction 

		   we also know that the remaining "parts" match perfectly, thus we have
			exactly the setup we need for a pre-incr load

		*/ 


		goto looks_ok;
	}

next2:
		;

looks_ok:

	if((pre_mod || post_mod) &&  (mem->type==FLOAT || mem->type==DOUBLE))
	{

		/* if we have a floating point pre/post modification */
		/* we have to have a constant smoewhere, if they all */
		/* have regs, it can't be a pre/post modification    */
		if(contains_reg(bin_instr->sem.binst.left) && contains_reg(bin_instr->sem.binst.right))
		{
			post_mod=pre_mod=0;
		}
	}
	

	if(!pre_mod && !post_mod)
	{
		used_right_mem=used_right_mod=FALSE;
		erflag=1;
	}
	return;
}

int check_const_range(struct addr_entry* mem,int constant_value)		
{
	if(mem->type==LONG && -4095<constant_value && constant_value<4095)
		return TRUE;
	if(mem->type==WORD && -255<constant_value && constant_value<255)
		return TRUE;
	if(mem->type==BYTE && -255<constant_value && constant_value<255)
		return TRUE;
	if(mem->type==FLOAT && -1023<constant_value && constant_value<1023)
		return TRUE;
	if(mem->type==DOUBLE && -1023<constant_value && constant_value<1023)
		return TRUE;

	return FALSE;
}


int log2(int a)
{
	int b=-1;	

	while(a>0)
	{
		a>>=1;
		b++;
	}
	return b;
}



void set_fp_se(int kill_r1)
{
#ifdef SOFTFLOAT
        int i;
        char reg[REGSTRING+3], a[REGSTRING+3];
        char *b;
        if(!side_effect)
                return;
        if(kill_r1)
                i=1;
        else
                i=2;
	/* kill 1-4 or 2-4 */
        for(;i<4;i++)
        {
          makereg(reg,RTREG,i);
          sprintf(a,"%s=;",reg);
          b=a;
          while (*side_effect = *b++)
                side_effect++;
        }

	/* and also kill 14 */
        makereg(reg,RTREG,14);
        sprintf(a,"%s=;",reg);
        b=a;
        while (*side_effect = *b++)
              side_effect++;

	/* and also kill 12 */
        makereg(reg,RTREG,12);
        sprintf(a,"%s=;",reg);
        b=a;
        while (*side_effect = *b++)
              side_effect++;
        makereg(reg,CTREG,0);
        sprintf(a,"%s=;",reg);
        b=a;
        while (*side_effect = *b++)
              side_effect++;
#endif
}

void check_cc_semantics(struct sem_rec* rhs,char comparison_type)
{
	switch(comparison_type)
	{
		case '?':
			/* Add and subract potentially overflow, so they aren't OK */
			if(rhs->sem.binst.op==ADDI) erflag=1;
			if(rhs->sem.binst.op==SUBI) erflag=1;
			if(rhs->sem.binst.op==SRAI) erflag=1;
			if(rhs->sem.binst.op==SLLI) erflag=1;
			if(rhs->sem.binst.op==SRLI) erflag=1;
			/* otherwise, things're cool */
			break;

		case ':':
			/* everything is OK under == */
			break;

		case '!':
			/* everything is OK under != */
			break;
	
		default:
			/* Unknown comparison type */
			break;
	}
	

}

void check_local_count(char* insn_type, ...)
{
	struct addr_entry *t;
	int count=0;
	va_list va_ptr;
	va_start(va_ptr,insn_type);

	while(1)
	{
		
		t=(struct addr_entry *)va_arg(va_ptr,int);
		if(!t) break;
		count+=count_locals(t);
	}
	if(count>1)
		erflag=1;
}


int count_locals(struct addr_entry* t)
{
   char *s;
   unsigned int a=65535; /* initially a bad value */
   struct addr_entry* p1,*p2,*p3;
   p1=(struct addr_entry*)t->arg1.ptr ;
   p2=(struct addr_entry*)t->arg2.ptr ;
   p3=(struct addr_entry*)t->arg3.ptr ;

   switch (t->mode) 
   {
      case SR_EXPR:
	        return count_locals(p3);
      case EXPR:
      {
	    unsigned int b,c; 
	    int operation=(int)t->arg1.ptr;

            if(operation== NEGI || operation== NOTI)
	    {
	        return count_locals(p2);
            }
            else if(operation== ADDI || operation== SUBI || 
            		operation== REMI || operation==REMUI ||
            		operation== DIVI ||  operation== DIVUI ||
            		operation== MULI || operation== SLLI ||
	    		operation== SRLI || operation== SRAI ||
	    		operation== XORI || operation== ORI ||
	    		operation== ANDI)
	    {
	        return count_locals(t->arg2.ptr) + count_locals(t->arg3.ptr); 
	    }
	    else
	    {
		fprintf(stderr, "Unkonwn expr mode\n");
		exit(1);
	    }
	break;
      }
      case LIDENT:
	        return 1;
      case GIDENT:
      case LABEL:
      case SPILLOC:
      case ICONST:
      case FCONST:
      case RREGISTER:
      case FREGISTER:
      case DREGISTER:
      case ABSOLUTE:
	return 0;
      case DISP:
          return(count_locals(t->arg1.ptr) || count_locals(t->arg2.ptr));
      case REGDFR:
          return(count_locals(t->arg1.ptr));


      case IMMED:
	return count_locals(t->arg1.ptr);
	break;

      default:
        fprintf(stderr, "vpo (contains_reg): unknown addr mode %d\n",
		t->mode);
        exit(1);
   }
}

