
/*
 * Define constants for instruction definitions
 */
#define ADDI    1   /* Note:  The first thirteen items are also used in an */
#define ANDI    2   /*        array to convert instructions to appropriate */
#define DIVI    3   /*        assembler output.  They must not be changed. */
#define DIVUI   4
#define MULI    5
#define ORI     6
#define REMI    7
#define REMUI   8
#define SLLI    9
#define SRAI   10
#define SRLI   11
#define SUBI   12
#define XORI   13
#define NEGI   14
#define NOTI   15
#define RSBI   16
#define LDI    17
#define STI    18
#define CMP    19
#define BRELI  20
#define BAI    21
#define CALLI  22
#define RETI   23
#define NOPI   24
#define MOVI   25
#define CVTI   26
#define MLAI   27
#define MVNI   28
#define LASTI  29


#define FIRSTJREL BRELI
#define LASTJREL BRELI
#define NUMJREL (LASTJREL - FIRSTJREL + 1)

#define CCSET  1
#define CCUSED 2


/*
 * Machine-dependent items used in the "kind" field of a "list" structure.
 * Bits 10 through 19 are reserved for our use.
 */
#define FPI       (1 << 10)  /* Floating-point instruction                */
#define SETCC     (1 << 11)  /* Instruction sets the condition codes      */
#define SETPC	  (1 << 12)  /* sets the pc, for predicated branches      */
#define USEPC	  (1 << 13)  /* all instrs use the PC, so we can't move thinsg
				past branches */



/*
 * These items are used to indicate which items are "done" and which items
 * "block" a given RTL.
 */
#define MEMREAD   (1 << 0)   /* Read from memory                          */
#define MEMWRITE  (1 << 1)   /* Writes to memory                          */
#define CCREAD    (1 << 2)   /* Read from integer condition code register */
#define CCWRITE   (1 << 3)   /* Write integer condition code register     */
#define USES_SCRATCH 	  (1 << 4)   /* to stop fp ops from moving past scratch regs */

/*
 * This array is used to define the static advantage of issuing each kind
 * of instruction.
 */

#ifdef SCHEDULE
int _av[LASTI] = {  0,  /* None                                     */
                   32,  /* ADDI                                     */
                   32,  /* ANDI                                     */
                   32,  /* DIVI                                     */
                   32,  /* DIVUI                                    */
                   32,  /* MULI       				    */
                   32,  /* ORI                                      */
                   32,  /* REMI                                     */
                   32,  /* REMUI      				    */
                   32,  /* SLLI                                     */
                   32,  /* SRAI                                     */
                   32,  /* SRLI                                     */
                   32,  /* SUBI                                     */
                   32,  /* XORI                                     */
                   32,  /* NEGI                                     */
                   32,  /* NOTI                                     */
                   32,  /* RSBI                                     */
                   38,  /* LDI    Loads scheduled early             */
                   32,  /* STI    Stores schedule late              */
                   64,  /* CMP    compares way early                */
                   0,   /* BRELI                                    */
                   0,   /* BAI                                      */
                   32,  /* CALLI                                    */
                   32,  /* RETI                                     */
                   32,  /* MOVI                                     */
                   32,  /* NOPI                                     */
                   32,  /* CVTI                                     */
		   32,  /* MLA					    */
		}; 

#else
extern int _av[LASTI];
#endif

/*
 * Macro declarations required by the machine-dependent portion of vpo
 */
#define LONER(i) (FALSE)
#define SAMECLASS(x, y) TRUE
#define NOCONFLICT(x) (TRUE)
#define ADVANTAGE(kind) _av[(kind) & BASEINST]

/*
 * Other useful macro declarations
 */
#define ISFP(i) (i & FPI)

#ifdef INTERACT

#define FPADD      0
#define FPMUL      1
#define FPDIV      2
#define FPUNKNOWN  3
#define INTMUL     4
#define INTDIV     5
#define STORE      6
#define LOAD       7
#define INTALU     8
#define INS_TYPES 10

#define FU_IALU    0
#define FU_IMULT   1
#define FU_MEMPORT 2
#define FU_FPALU   3
#define FU_FPMULT  4
#define FU_TYPES   5

#endif /*INTERACT*/

