
/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/arm/isem.c,v $
 $Date: 2002/07/16 19:44:03 $
 $Revision: 1.11 $
 $Name:  $

 Log of changes:

 $Log: isem.c,v $
 Revision 1.11  2002/07/16 19:44:03  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.10  2001/09/05 18:44:16  jdh8d
 *** empty log message ***

 Revision 1.9  2001/08/31 15:34:05  jdh8d
 *** empty log message ***

 Revision 1.8  2001/08/21 00:11:03  jdh8d
 *** empty log message ***

 Revision 1.7  2001/05/21 18:42:56  jdh8d
 *** empty log message ***

 Revision 1.6  2001/05/16 18:53:32  jdh8d
 Changes to support "soft floating point"

 Revision 1.5  2001/04/19 15:11:00  jdh8d
 Initial implentation of "if conversion" for basic predicated
 execution.

 Revision 1.4  2001/04/06 18:38:01  jdh8d
 Disallowed shifts by 68 (or more than 32, really).
 added a typecast to get rid of a warning.

 Revision 1.3  2001/04/04 19:40:17  jdh8d

 changes to support pre/post modification of an address register.

 in simp.c we added support to change:

 	r[X]=R[..];r[Y]=r[Y]; to simply remove the 2nd effect.

 This came from

 	r[X]=R[..];r[Y]=r[Y]+r[Z]; to simply remove the 2nd effect.

 but r[Z] was later recognized to be zero.  It works if the useless
 effect comes first as well.

 Revision 1.2  2001/03/09 04:42:25  jdh8d
 *** empty log message ***

 Revision 1.1  2001/02/01 22:55:19  jdh8d

 Initial arm version. Much stuff is still wrong.  Mostly this is a copy
 of the mips code but setup as an "arm" target.  Next I'll start converting
 functions so that they work with ARM instead of mips.  inst.c here I come.

 Revision 1.2  2000/06/10 18:44:54  jwd
 Converted over to ANSI C.


*************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "../lib/vpo.h"
#include "../lib/combine.h"

#include "addrdef.h"
#include "recdef.h"
#include "instdef.h"
#include "inst.h"
#include "isem.h"


int sem_ptr;                            /* Semantic record allocation pointer */

static struct sem_rec sem_table[SEMTABLESIZE]; /* Semantic record table       */

/*
 * record - get a semantic record
 */
struct sem_rec *record(struct addr_entry *d1, struct addr_entry *d2,
 struct addr_entry *d3) {
    register struct sem_rec *t;

    t = &sem_table[sem_ptr++];
    t->sem.fields[0] = d1;
    t->sem.fields[1] = d2;
    t->sem.fields[2] = d3;
    t->pred=0;
    return(t);
}

/*
 * cmp_record - get a semantic record
 */
struct sem_rec *cmp_record(struct addr_entry *d1, struct addr_entry *d2, struct addr_entry *d3, char d4) 
{
    register struct sem_rec *t;

    t = &sem_table[sem_ptr++];
    t->sem.fields[0] = d1;
    t->sem.fields[1] = d2;
    t->sem.fields[2] = d3;
    t->sem.cmp.type=d4;
    t->pred=0;
    return(t);
}



/*
 * pre_record - get a semantic record for a pre/post inc/dec load/store
 */
struct sem_rec *pre_record(struct addr_entry *d1, struct addr_entry *d2,
 struct addr_entry *d3, struct sem_rec *rec) 
{
    register struct sem_rec *t;

    t = &sem_table[sem_ptr++];
    t->sem.fields[0] = d1;
    t->sem.fields[1] = d2;
    t->sem.fields[2] = d3;
    t->sem.fields[3] = (struct addr_entry*)rec;
    return(t);
}

/*
 * brecord - get a semantic record for a binary operation
 */
struct sem_rec *brecord(int op, struct addr_entry *d1, struct addr_entry *d2) 
{
    register struct sem_rec *t;

    t = &sem_table[sem_ptr++];
    t->sem.binst.left = d1;
    t->sem.binst.right = d2;
    t->sem.binst.dst = AN;
    t->sem.binst.op = op;
    return(t);
}

/*
 * brecord - get a semantic record for a binary operation
 */
struct sem_rec *mla_record( struct sem_rec *s, struct addr_entry* inc)
{
    	s->sem.mla.inc = inc;

	/* we check that the old OP was MUL, otherwise it's an error */
	if(s->sem.binst.op!=MULI)
	{ 
		erflag=1;
		reason="accumlate without multiply";
	}
	
    	s->sem.binst.op = MLAI;
    	return(s);
}



/*
 * binary_op - check semantics of binary operation
 */
struct sem_rec *binary_op(struct addr_entry *dst, struct sem_rec *s) 
{

	/* handle the case were the dest is the CC */
	if(dst==0)
	{
		if(s->sem.binst.left->type==FLOAT || 
		   s->sem.binst.left->type==DOUBLE || 
		   s->sem.binst.right->type==FLOAT || 
		   s->sem.binst.right->type==DOUBLE )
			erflag=1;
	}
	else
	{ 
		/* 
		 * Ensure that the sources are compatible with the destination 
		 */
    		if (!COMPATIBLE(s->sem.binst.left, s->sem.binst.right) ||
       		 !COMPATIBLE(dst, s->sem.binst.right))
      			erflag = 1;
    		else if (!COMPATIBLE(dst, s->sem.binst.left))
      			erflag = 1;

#ifdef SOFTFLOAT
        		/* floats msut be of the form f[0]=f[0]+f[1] */
       		if(dst->type==FLOAT && (regnum(dst->arg1.str)!=0 ||
               		regnum(s->sem.binst.left->arg1.str)!=0 ||
               		regnum(s->sem.binst.right->arg1.str)!=1))
                       	erflag=1;

       		/* doubles msut be of the form d[0]=d[0]+d[2] */
       		if(dst->type==DOUBLE && (regnum(dst->arg1.str)!=0 ||
               		regnum(s->sem.binst.left->arg1.str)!=0 ||
               		regnum(s->sem.binst.right->arg1.str)!=2))
                    	erflag=1;
#endif
	}

    s->sem.binst.dst = dst;
    return(s);
}

/*
 * same - ensure that two items are the same
 */
struct addr_entry *same(struct addr_entry *i1, struct addr_entry *i2) 
{

    if (!icompare(i1, i2))
      erflag = 1;
    return(addr(ANY, TEMPADDR, i1, i2, AN));
}

/* predicate is the char string that tells what type
	of predicate is used.  second_src is the 2nd "source"
	it should match the destination
*/
struct sem_rec*  add_predicate(struct sem_rec* sem, char* predicate, struct addr_entry* second_src)
{
	sem->pred=predicate;

	/* we'll try to make sure the destination is the same
		as the 2nd source here .. maybe.. if so, we can leave
		the semantic routines alone.. yay! 
	*/

	/* if we have both a destination and a 2nd predicate, ensure they are the same */
        /* otherwise we have an error */
	if(sem->sem.fields[0]!=0 && second_src!=0 && predicate)
	{
		if(!icompare(sem->sem.fields[0], second_src))
      			erflag = 1;
	}
	/* other wise, if one of them is set (but not both) this is an error */
	else if(predicate && (sem->sem.fields[0]!=0 || second_src!=0))
		erflag=1;
	else if(second_src && !predicate)
		erflag=1;

	/* check the destination item */
	/* if it's not a word item */
	if(second_src && (second_src->type == FLOAT || second_src->type == DOUBLE)) 
	{
		erflag=1;
	}
	return sem;
}

struct sem_rec *fcmp_record(struct addr_entry *d1, struct addr_entry *d2,
        struct addr_entry *d3, char* d4)
{
    register struct sem_rec *t;

    t = &sem_table[sem_ptr++];
    t->sem.fcmp.dst = d1;
    t->sem.fcmp.left = d2;
    t->sem.fcmp.right = d3;
    t->sem.fcmp.op = d4;
    return(t);


}

