
/************************************************************************
 *                                                                      *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   *
 *     Zephyr Software, LLC                                             *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from Zephyr Software, LLC                                       *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/arm/linstinfo.c,v $
 $Date: 2004/01/21 18:39:46 $
 $Revision: 1.6 $
 $Name:  $

 Log of changes:

 $Log: linstinfo.c,v $
 Revision 1.6  2004/01/21 18:39:46  jdh8d
 Latest Vista additions/changes.  Don't yell at me if it's broken/bad.. prasad wanted these changes.  I've reviewed them, and tested,  minimally.. things appear OK (for the batch compiler)

 Revision 1.5  2003/11/10 18:08:54  jdh8d
 lots of ease fixes for arm.

 Revision 1.4  2003/11/09 17:17:19  jdh8d
 changes to get EASE going for the arm

 Revision 1.3  2003/11/09 14:36:47  jwd
 Fixed RCS information at beginning


*************************************************************************/

/*------------------------------------------------
  History of Changes
  ------------------
  1/13/98: Function prototypes made ansi-c
------------------------------------------------*/

#if defined(MEASURE) || defined(INTERACT)

#include <stdio.h>
#include <assert.h>

#include "../lib/vpo.h"
#include "../lib/rtlspec.h"
#include "../lib/assign.h"
#include "../lib/combine.h"
#include "../lib/misc.h"

#include "addrdef.h"
#include "recdef.h"
#include "instdef.h"
#include "regs.h"
#include "regs2.h"
#include "rtl.h"
#include "linstinfo.h"

char *startblklabel = (char *) NULL;	/* No labels needed to obtain sizes */
char *endblklabel = (char *) NULL;

char *main_fn_name="main";

#include "../lib/measure.h"

static char *EasePrefix = "__";

void	calc_funcs();

static char scratch[MAXSTRING], fscratch[MAXSTRING];
static int cc = 0;
static int info_length = 0;
int funcnum1 = 0;
static int ofs;

/*
 * register_usage_information - add register use information
 */
void register_usage_information() {

   regarray tr;
   int items, index, scratch, nonscratch;
   char reg[REGSTRING];

   /*
      Note that the register items that we obtain from the regstate vector of
      the entry block correspond to the registers as they currently appear in
      the RTLs.  In order to ascertain which register will actually be used
      when the RTLs are dumped out in assembly language, we must go through the
      map() function.
   */

   tr = (regarray) top->regstate;

   scratch = nonscratch = 0;
   items = availreg(RTREG);
   for (index = firstreg(RTREG); items--; index += skipreg(RTREG)) {
      if (tr->reg[index].status != NOTUSED) {
	 makereg(reg, RTREG, indextoreg(RTREG, index));
	 makereg(reg, RTREG, map(reg));
	 if (scratchreg(reg))
	    scratch++;
	 else
	    nonscratch++;
      }
   }
   *info_ptr++ = scratch;
   *info_ptr++ = nonscratch;

   scratch = nonscratch = 0;
   items = availreg(DTREG);
   for (index = firstreg(DTREG); items--; index += skipreg(DTREG)) {
      if (tr->reg[index].status != NOTUSED) {
	 makereg(reg, DTREG, indextoreg(DTREG, index));
	 makereg(reg, DTREG, map(reg));
	 if (scratchreg(reg))
	    scratch++;
	 else
	    nonscratch++;
      }
   }
   *info_ptr++ = scratch;
   *info_ptr++ = nonscratch;
}

/*
 * stinstinfo - place instruction information in the information array
 */
void stinstinfo(int inst, int type, int cc, struct addr_entry * p1, struct addr_entry * p2, struct addr_entry * p3) {

   if (swef && curptr->onum) {
      *info_ptr++ = curptr->daddy->equivnum;
      *info_ptr++ = inst;
      *info_ptr++ = type;
      *info_ptr++ = cc;
      *info_ptr++ = (p1) ? ((ACONMODE(p1)) ? IMMED : p1->mode) : 0;
      *info_ptr++ = (p1) ? p1->type : 0;
      *info_ptr++ = (p2) ? ((ACONMODE(p2)) ? IMMED : p2->mode) : 0;
      *info_ptr++ = (p2) ? p2->type : 0;
      *info_ptr++ = (p3) ? ((ACONMODE(p3)) ? IMMED : p3->mode) : 0;
      *info_ptr++ = (p3) ? p3->type : 0;
      *info_ptr++ = 0;
      *info_ptr++ = 0;
      *info_ptr++ = 0;
      *info_ptr++ = 0;
      info_length++;


	/* this needs to emit MAXFIELD number of pairs.  If this assert fails, we need
	 * to patch up the code above, and probably rebuild genreport too
	 */ 
	assert(MAXFIELDS==5);
   }
}

/*
 * reset_info_length - reset the info_length variable
 */
void reset_info_length()
{
  info_length = 0;
}

/*
 * is_end_ins - check if to add the special nop instruction to indicate to
 *              the simulator to stop getting cycle counts. For the ARM, this
 *              instruction needs to be added 2 insts before PC=RT to take
 *              care of the ldmfd instruction
 */
int is_end_ins(struct list *rtl)
{
  if(callsexit(rtl))
    return 1;

  return 0;
}

/*
 * is_add_before_ret - add nop instruction two instructions before PC=RT to
 *                     take care of ldmfd inst. generated by fix entry
 *                     exit. Only needed for the ARM
 */
int is_add_before_ret(struct list *rtl)
{

  /*assuming that there are no comment instructions between the ldfmd and
    return instruction. Should hold true since these instructions are all
    generated by VPO itself*/
  if(rtl && rtl->next && rtl->next->next && rtl->next->next->val &&
     is_return(rtl->next->next->val))
    return 1;

  return 0;
}

/*
 * get_label_number - generate a new label number
 */
int get_label_number() {
   static int label_number = 0;
   return label_number++;
}

/*
 * increment_class_counter - insert instruction to increment a frequency counter
 */
int increment_class_counter(struct bblock * cblk, struct list * ptr, int incr, int execnum) {

   struct list *tptr;
   int ofs;
   int word_label, branch_label;
   char new[MAXSTRING];
   
   /*
      This function must insert code at the point indicated to add the amount
      indicated by "incr" to the appropriate equivalence class counter.  If the
      value of "incr" is other than one, this function has the option of not
      inserting appropriate code provided that it returns the value FALSE.
      Whenever the code is properly inserted, the value TRUE should be
      returned.
   */

   /* ask Jason what the arm allows */
   if (incr > 250)
      return (FALSE);

   if (cnts(ptr) && is_return(ptr->val) && ptr->prev && ptr->prev->prev &&
       (ptr->prev->type & ASMLINE) && (ptr->prev->prev->type & ASMLINE))
     ptr = ptr->prev;
   
   ofs = (execnum + 4) << 2;

   word_label = get_label_number();
   branch_label = get_label_number();

   sprintf(new, "\tstmfd sp!, {r12,r14}");
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(), ASMLINE);

   (void) sprintf(new, "\tldr\tr14,.Lease00%d", word_label );
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(), ASMLINE);

   (void) sprintf(new, "\tldr\tr12,[r14]");
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(), ASMLINE);

   (void) sprintf(new, "\tadd\tr12,r12,#%d", incr);
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(), ASMLINE);
   
   (void) sprintf(new, "\tstr\tr12,[r14]"); 
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(), ASMLINE);

   sprintf(new, "\tldmfd sp!, {r12,r14}");
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(), ASMLINE);

   (void) sprintf(new, "\tb\t.Lease00%d", branch_label );
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(), ASMLINE);

   (void) sprintf(new, ".Lease00%d:\t.word\t %sease_%s_info+%d", word_label, EasePrefix,
    fn, ofs);
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(), ASMLINE);

   (void) sprintf(new, ".Lease00%d:", branch_label);
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(), ASMLINE);

   return (TRUE);
}

/*
 * increment_class_counter_by_expression - add an expression to a class counter
 */
int increment_class_counter_by_expression(struct bblock * cblk, struct list * ptr, char *expr, int execnum) {
/*ARGSUSED*/

   /*
      This function always has the option of not instrumenting the code as
      requested provided that it returns the value FALSE.  Whenever appropriate
      code is inserted, the value TRUE must be returned.
   */

   return (FALSE);
}

/*
 * increment_if_branch_not_taken - insert a counter to measure branch path
 */
void increment_if_branch_not_taken(struct bblock * cblk, int type) {

   struct list *tptr;
   char new[MAXSTRING];


   /*
      Insert the increment instructions
   */
   tptr = add_rtl_with_type(cblk, (struct list *) NULL, "\tstd\t%o4,[%sp-8]", rinit(),ASMLINE);
   (void) sprintf(new, "\tsethi\t%%hi(%sease_jrelnottaken+%d),%%o5",
		  EasePrefix, (type - FIRSTJREL) << 2);
   tptr = add_rtl_with_type(cblk, (struct list *) NULL, new, rinit(),ASMLINE);
   (void) sprintf(new, "\tld\t[%%o5+%%lo(%sease_jrelnottaken+%d)],%%o4",
		  EasePrefix, (type - FIRSTJREL) << 2);
   tptr = add_rtl_with_type(cblk, (struct list *) NULL, new, rinit(),ASMLINE);
   tptr = add_rtl_with_type(cblk, (struct list *) NULL, "\tadd\t%o4,1,%o4", rinit(),ASMLINE);
   (void) sprintf(new, "\tst\t%%o4,[%%o5+%%lo(%sease_jrelnottaken+%d)]",
		  EasePrefix, (type - FIRSTJREL) << 2);
   tptr = add_rtl_with_type(cblk, (struct list *) NULL, new, rinit(),ASMLINE);
   tptr = add_rtl_with_type(cblk, (struct list *) NULL, "\tldd\t[%sp-8],%o4", rinit(),ASMLINE);
}

/*
 * initial_call - insert call to initialize the measurement process
 */
void initial_call(struct bblock * cblk, struct list * ptr, char *suffix) {


   struct list *tptr;
   char new[MAXSTRING];


   /*
      Insert the call to initialize measurements
   */
   tptr = add_rtl_with_type(cblk, ptr, "\tstmfd\tsp!,{r0-r3}", rinit(), ASMLINE);
   (void) sprintf(new, "\tbl\t%sease_start_%s", EasePrefix, suffix);
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(), ASMLINE);
   tptr = add_rtl_with_type(cblk, ptr, "\tldmfd\tsp!,{r0-r3}", rinit(), ASMLINE);
}

/*
 * final_call - insert call to terminate the measurement process
 */
void final_call(struct bblock * cblk, struct list * ptr, char *suffix) {


   struct list *tptr;
   char new[MAXSTRING];


   savescratch(cblk, ptr->prev);

   (void) sprintf(new, "\tbl\t%sease_end_%s", EasePrefix, suffix);
   tptr = add_rtl_with_type(cblk, ptr->prev, new, rinit(),ASMLINE);

   restorescratch(cblk, ptr->prev);
}


/*
 * freq_info - emit the declarations needed to perform execution measurements
 */
#ifdef INTERACT
void freq_info(FILE *fp) 
#else
void freq_info() 
#endif
{
#ifndef INTERACT
	FILE *fp=stdout;
#endif

   int *ptr, i, j;
 
   ptr = func_instinfo;

   /*
      Output a declaration for a pointer to the next function block
   */
   (void) fprintf(fp,".data\n");
   (void) fprintf(fp,".align\t2\n");
   (void) fprintf(fp,"%sease_%s_info:\n", EasePrefix, fn);
   (void) fprintf(fp,".word 0\n");


#ifdef INTERACT
	link_module(fp);
#else
	link_module();
#endif

   /*
      Output a declaration for the number of equivalence classes in this
      function.
   */
   (void) fprintf(fp,".word %d\n", equiv_classes);

   /*
      Output a declaration for the execution counters for each equivalence
      class.
   */
   if (equiv_classes) {
      (void) fprintf(fp,".word ");
      for (j = 1; j <= equiv_classes; j++) {
	 if (!(j & 15) || j == equiv_classes) {
	    (void) fprintf(fp,"0\n");
	    if (j < equiv_classes)
	       (void) fprintf(fp,".word ");
	 }
	 else
	    (void) fprintf(fp,"0, ");
      }
   }

   /*
      Output a global declaration for the block size array.  This does not need
      to be done on machines whose instructions are all the same size.
   */

   /*
      Output function-specific information
   */
   j = 0;
   (void) fprintf(fp,".word ");
   while (*ptr != -4) {
      if (j++ == 12) {
	 (void) fprintf(fp,"%d\n", *ptr++);
	 (void) fprintf(fp,".word ");
	 j = 0;
      }
      else
	 (void) fprintf(fp,"%d, ", *ptr++);
   }
   (void) fprintf(fp,"%d\n", *ptr++);



   /*
      Output instruction information
   */
   (void) fprintf(fp,".word %d\n", info_length);
   for (i = 0; i < info_length; i++) {
      (void) fprintf(fp,".word ");
      for (j = 1; j < 4 + MAXFIELDS * 2; j++)
	 (void) fprintf(fp,"%d, ", *ptr++);
      (void) fprintf(fp,"%d\n", *ptr++);
   }
   info_length = 0;

   /*
      Output call graph information to reach other modules
   */
   call_graph(fp);
}

/*
 * trace_call - insert call to indicate that a basic block has executed
 */
void trace_call(struct bblock * cblk, struct list * ptr, int blks, int insts) {

   struct list *tptr;
   char new[MAXSTRING];
   int branch_label, word_label;
 
   /*
      Since we are going to insert a call to the basic block trace function, we
      need to save all of the scratch registers that are currently live.
   */
   savescratch(cblk, ptr);

   /*
      Insert the code to make a call to the instruction trace function.  The
      argument passed to the function is the information structure of the basic
      block that we are going to execute adjusted by the starting block offset
      of this file.
   */
   (void) sprintf(new, "\tldr\tr14,.Lease00%d", word_label = get_label_number());
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(), ASMLINE);


   tptr = add_rtl_with_type(cblk, ptr, new, rinit(),ASMLINE);
   (void) sprintf(new, "\tbl\t%sease_trace_inst", EasePrefix);
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(),ASMLINE);

   (void) sprintf(new, "\tb\t.Lease00%d", branch_label = get_label_number());
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(), ASMLINE);

   (void) sprintf(new, ".Lease00%d:\t.word\t %sease_%s_info+%d", word_label, EasePrefix,
    fn, 16 + blks * 20);
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(), ASMLINE);

   (void) sprintf(new, ".Lease00%d:", branch_label);
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(), ASMLINE);

   /*
      It is safe to restore the scratch registers that were saved
   */
   restorescratch(cblk, ptr);
}

/*
 * unify_call - insert call to indicate that a new block has been entered
 */
void unify_call(struct bblock * cblk, struct list * ptr, int blk) {

   struct list *tptr;
   char new[MAXSTRING];
 
   /*
      Since this call can be placed at almost any point in the function that we
      are instrumenting, we need to save any scratch registers that might be
      affected by the call that we are going to make.
   */
   savescratch(cblk, ptr);

   /*
      Insert the code to make a call to the function that signals that we have
      entered or re-entered the function.
   */
   (void) sprintf(new, "\tsethi\t%%hi(%sease_%s_unify_%d),%%o0", 
    EasePrefix, fn, blk);
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(),ASMLINE);
   (void) sprintf(new, "\tcall\t%sease_unified_inst,1", EasePrefix);
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(),ASMLINE);
   (void) sprintf(new, "\tadd\t%%o0,%%lo(%sease_%s_unify_%d),%%o0", EasePrefix,
    fn, blk);
   tptr = add_rtl_with_type(cblk, ptr, new, rinit(),ASMLINE);

   /*
      It is safe to restore the scratch registers that were saved
   */
   restorescratch(cblk, ptr);
}

/*
 * inst_labels - insert labels around instructions to determine their size
 */
struct list *inst_labels(struct list * ptr) {

   /*
      This function should only return a pointer to the next RTL line if all of
      the instructions are of the same size.
   */
   return (ptr->next);
}

/*
 * trace_memory_references - check the RTL for data memory references
 */
int trace_memory_references(struct list * ptr, int blk) {

   int anymemsofar;
   char *rtl, *p, dst[MAXSTRING], src[MAXSTRING];

   /*
      Keep track of the number of memory references in this RTL
   */
   anymemsofar = 0;
   rtl = ptr->val;
   (void) gtds(ptr->val, dst, src);

   /*
      Determine if there data memory references in the source.  On this target
      machine, it is only necessary to look for one.
   */
   p = src;
   while (p = gtmem(p))
      anymemsofar = tracememaddr(ptr, blk, bpos, anymemsofar, FALSE);

   /*
      Determine if there is a data memory reference in a destination (have to
      examine all since many instructions could set the condition codes).
   */
   while (rtl = gtds(rtl, dst, src)) {
      if (ismem(dst))
	 anymemsofar = tracememaddr(ptr, blk, dst, anymemsofar, TRUE);
   }

   /*
      Handle any special cases
   */
   anymemsofar = traceothermemaddr(ptr, blk, anymemsofar);

   return (anymemsofar);
}

/*
 * inst_info - emit declarations required to run an instruction cache simulation
 */
#ifdef INTERACT
void inst_info(FILE *fp) 
#else
void inst_info() 
#endif
{
#ifndef INTERACT
	FILE *fp=stdout;
#endif

   int i;

     /*
      Output a declaration for a pointer to the next function block
   */
   (void) fprintf(fp,".data\n");
   (void) fprintf(fp,".align\t2\n");
   (void) fprintf(fp,"%sease_%s_info:\n", EasePrefix, fn);
   (void) fprintf(fp,".word 0\n");
   link_module(fp);

   /*
      Output a declaration of the number of small blocks in this function
   */
   (void) fprintf(fp,".word %d\n", totalblks);

   /*
      For each small block
   */
   for (i = 0; i < totalblks; i++) {

      /*
         Emit a structure that contains the number of instructions in the block
         as the first item.  The rest of the items will be filled later by the
         measurement library functions.
      */
      (void) fprintf(fp,".word %d, 0, 0, 0, 0\n", numinsts[i]);

      /*
         On machines where the sizes of instructions vary, emit a list of items
         that specify the size of each instruction.
      */
   }

   /*
      Output call graph information to reach other modules
   */
   call_graph(fp);
}

/*
 * unified_info - emit declarations required to run a unified cache sumilation
 */
#ifdef INTERACT
void unified_info(FILE *fp) 
#else
void unified_info()
#endif
{
#ifndef INTERACT
	FILE *fp=stdout;
#endif

   int i;
   

   /*
      Since we want to avoid perturbing the data addresses as much as possible
      in a unified cache simulation, we will place these declarations in the
      text segment, since the library calls that use these declarations will
      not attempt to write into them.
   */

   /*
      Output a declaration stating the number of small blocks in this function
   */
   (void) fprintf(fp,".text\n");
   (void) fprintf(fp,".align\t2\n");
   (void) fprintf(fp,"%sease_%s_info:\n", EasePrefix, fn);
   link_module(fp);

   /*
      Output a declaration of the number of small blocks in this function
   */
   (void) fprintf(fp,".word %d\n", totalblks);

   /*
      For each small block
   */
   for (i = 0; i < totalblks; i++) {

      /*
         Emit a marker that indicates which block in the function we are going
         to output information for.  This item should be negative because it
         indicates the end of the previous block as well.
      */
      (void) fprintf(fp,"%sease_%s_unify_%d:\n", EasePrefix, fn, i);

      /*
         Emit a declaration stating the number of instructions in this basic
         block.  If the size of every instruction in the target machine is the
         same, this is the only item required for the block.
      */
      (void) fprintf(fp,".word %d\n", numinsts[i]);

      /*
         On machines where the sizes of instructions vary, emit a list of items
         that specify the size of each instruction.
      */
   }

   /*
      Output call graph information to reach other modules
   */
   call_graph(fp);

   /*
      Make sure to go back to the data segment in case anything that follows
      expects to be there.
   */
   (void) fprintf(fp,".data\n");
}

/*
 * end_info - called at end to terminate module link
 */
#ifdef INTERACT
void end_info(FILE *fp) 
#else
void end_info() 
#endif
{
#ifndef INTERACT
	FILE *fp=stdout;
#endif

   if (swef || swei) {
      	fprintf(fp,".data\n");
      	fprintf(fp,".align\t2\n");
      	fprintf(fp,"%sease_link_%d:\n", EasePrefix, funcnum1);
      	fprintf(fp,".word\t0\n");
   	fprintf(fp, ".data\n");
   	fprintf(fp, ".align 2\n");
   	fprintf(fp, "tmp_space:\t.word\t0\n");
      	if (ext_num)
	 	(void) fprintf(stderr, "vpo (end_info): loose call graph pointers\n");
   }
}

/*
 * callsexit - determine if the RTL calls the exit function
 */
int callsexit(struct list *rtl_ptr) {

  char *s=rtl_ptr->val;
   /*
      If the RTL does not contain a global identifier in the fifth character
      position, then it cannot be a call to the exit routine.
   */
   if (!isglob(s+3))
      return (FALSE);

   return (!strcmp(globalname(s + 3), "exit") ||
	   !strcmp(globalname(s + 3), "_exit"));
}

/*
 * is_anyrjmp - is RTL any type of relational jump
 */
int is_anyrjmp(register char *s) {

   return (*s == 'P' && *(s + 1) == 'C' && *(s + 4) == 'C');
}

/*
 * tracememaddr - if needed then insert a call to trace a data memory address
 */
int tracememaddr(struct list * ptr, int blk, char *start, int anymemsofar, int write) {

   struct list *tptr;
   int depth, found;
   char *p, *s, c;
   char new[MAXSTRING], reg_o1[REGSTRING], reg[REGSTRING];

   /*
      If the item is in a delay slot, back up so that we place the call before
      the instruction that creates it.  If the item that created the delay slot
      is a relative branch, then we must back up until we can place the call
      before the instruction that sets the condition codes.  If we just can't
      move back, then place the call before the relative branch and take a hit
      on the extra effort required to save and restore the condition codes.
   */
   if (ptr->kind & (TAKEN | NOTAKEN)) {
      tptr = ptr = ptr->prev;
      if (cnts(ptr) && is_rjmp(ptr->val)) {
	 for (;;) {
	    if (!(ptr = ptr->prev)) {
	       ptr = tptr;
	       break;
	    }
	    if (cnts(ptr) && ptr->onum && set_cc(ptr->val))
	       break;
	 }
      }
   }

   /*
      If no memory references have been located so far then save the scratch
      registers used.
   */
   savescratch(ptr->daddy, ptr);

   /*
      Start the RTL that will calculate the address referenced
   */
   makereg(reg, RTREG, map(reg_o1));
   makereg(new, RTREG, map(reg));
 	s=skiploc(new);	
   *s++ = '=';

   /*
      We wish to obtain the actual address referenced.  We have to be careful
      with expressions since the machine description accepts address
      expressions that cannot be simply assigned to a register.
   */
   depth = 1;
   found = FALSE;
   for (p = start + 2; c = *p; p++) {
      if (c == '[')
	 depth++;
      else if (c == ']') {
	 if (!--depth)
	    break;
      }
      else if (isRegister(p)) {
	 copyloc(s,p);
	 s=skiploc(s);	 
	 *s++ = '+';
      }
      else if (!isoper(c))
	 found = TRUE;
   }
   if (found) {
      depth = 1;
      for (p = start + 2; c = *p; p++) {
	 if (c == '[')
	    depth++;
	 else if (c == ']') {
	    if (!--depth)
	       break;
	 }
	 else if (isRegister(p)) {
		 p=skiploc(p)-1;
	    if (*(p+1) == '+') {
	       ++p;
			 continue;
	    }
	    c = '0';
	 }
	 *s++ = c;
      }
   }
   else
      s--;

   /*
      Finish the RTL string and insert it into the function
   */
   *s++ = ';';
   *s = '\0';
   tptr = add_rtl(ptr->daddy, ptr, new, rinit());

   /*
      If we are generating code for a unified cache simulation, we must emit
      code to pass a pointer to this block's information block. Since there are
      no instructions that access memory more than once, we will always have an
      information block pointer to pass.
   */
   if (swei && swed) {
      (void) sprintf(new, "\tsethi\t%%hi(%sease_%s_unify_%d),%%o2",
       EasePrefix, fn, blk);
      tptr = add_rtl_with_type(ptr->daddy, ptr, new, rinit(),ASMLINE);
      (void) sprintf(new, "\tadd\t%%o2,%%lo(%sease_%s_unify_%d),%%o2",
		     EasePrefix, fn, blk);
      tptr = add_rtl_with_type(ptr->daddy, ptr, new, rinit(),ASMLINE);
   }

   /*
      Pass the arguments and invoke the simulator directly
   */
   if (write)
      if (swei) {
         (void) sprintf(new, "\tcall\t%sease_uwrite,2", EasePrefix);
	 tptr = add_rtl(ptr->daddy, ptr, new, rinit());
      }
      else {
         (void) sprintf(new, "\tcall\t%sease_write,2", EasePrefix);
	 tptr = add_rtl(ptr->daddy, ptr, new, rinit());
     }
   else if (swei) {
         (void) sprintf(new, "\tcall\t%sease_uread,2", EasePrefix);
         tptr = add_rtl(ptr->daddy, ptr, new, rinit());
   }
   else {
      (void) sprintf(new, "\tcall\t%sease_read,2", EasePrefix);
      tptr = add_rtl(ptr->daddy, ptr, new, rinit());
   }
   tptr->type = ASMLINE;
   (void) sprintf(new, "\tmov\t%d,%%o0", memsize(*start));
   tptr = add_rtl(ptr->daddy, ptr, new, rinit());
   tptr->type = ASMLINE;

   /*
      Restore the scratch register values if they were any memory references in
      the RTL.
   */
   restorescratch(ptr->daddy, ptr);

   return (anymemsofar + 1);
}

/*
 * traceothermemaddr - check for implicit memory references
 */
int traceothermemaddr(struct list * ptr, int blk, int anymemsofar) {

  /* copy from sparc later */
}

/*
 * savescratch - save the scratch register values live at this point
 */
void savescratch(struct bblock * cblk, struct list * ptr) {

   	static int ccnum = 0;
   	struct list *tptr;
   	char *a, new[MAXSTRING];
	
   	/*
      	 * Initialize the "scratch" and "fscratch" strings to contain the mask of
      	 * all of the scratch registers live at this point in the function.
   	 */
	
   	live_scratch(cblk, ptr);
   	cc = save_cc(cblk, ptr);

	/* we need to save r0-r3,r12,r14 and f0-f3 for this, also the CC 
	 * if it's live 
	 */
	add_rtl_with_type(cblk,ptr,"stmfd	sp!, {r0-r3,r12,r14}", rinit(), ASMLINE);
#ifdef HARDLFOAT
	/* here we need to save f0-f3 */
	add_rtl_with_type(cblk,ptr, "sfmfd	sp!, f0,4,[sp]!",  rinit(), ASMLINE);

#endif
	/* here we need to save the CC if it's live */

}

/*
 * restorescratch - restore the scratch registers live at this point
 */
void restorescratch(struct bblock * cblk, struct list * ptr) {

   	struct list *tptr;
   	char *a, new[MAXSTRING];

        /* we need to restore r0-r3,r12,r14 and f0-f3 for this, also the CC
         * if it's live
         */
	add_rtl_with_type(cblk,ptr,"ldmfd	sp!, {r0-r3,r12,r14}", rinit(), ASMLINE);
#ifdef HARDLFOAT
	/* here we need to save f0-f3 */
	add_rtl_with_type(cblk,ptr, "lfmfd	sp!, f0,4,[sp]!",  rinit(), ASMLINE);
         
#endif  
        /* here we need to save the CC if it's live */


}

/*
 * live_scratch - determine which scratch registers are currently live
 */
void live_scratch(struct bblock * cblk, struct list * ptr) {

   int index, items;
   char *a, *b, reg[MAXSTRING], mapreg[MAXSTRING], ireg[REGSTRING];
  
   /*
      Obtain a list of all of the data registers that are not live at the point
      of the function indicated.  Then go through the list of data registers.
      Any scratch data register that is not available must be in use, so it
      should be saved and restored at this point.
   */
   b = scratch;
   (void) get_freereg(cblk, ptr, 0, RTREG, reg, mapreg);
   items = availreg(RTREG);
   for (index = firstreg(RTREG); items--; index += skipreg(RTREG)) {
      makereg(ireg, RTREG, indextoreg(RTREG, index));
		for (a = reg; *a; a = skiploc(a)+1) {
	 if (cmploc(a,ireg)==0)
	    break;
      }
      if (!*a) {
	 makereg(ireg, RTREG, map(ireg));
	 if (scratchreg(ireg)) {
	    copyloc(b,ireg);
		 *skiploc(b)='\0';
		 b=skiploc(b)+1;
	 }
      }
   }
   *b = '\0';


   /*
      Do the same thing with the floating-point registers
   */
   b = fscratch;

	/* only need to do this for hard float though, for softfloat we only 
	 * really have the integer registers */
#ifdef HARDFLOAT
   (void) get_freereg(cblk, ptr, 0, DTREG, reg, mapreg);
   items = availreg(DTREG);
   for (index = firstreg(DTREG); items--; index += skipreg(DTREG)) {
      makereg(ireg, DTREG, indextoreg(DTREG, index));
	 	for (a = reg; *a; a = skiploc(a)+1) {
	 if (cmploc(a,ireg)==0)
	    break;
      }
      if (!*a) {
	 makereg(ireg, DTREG, map(ireg));
	 if (scratchreg(ireg)) {
	    copyloc(b,ireg);
		 *skiploc(b)='\0';
		 b=skiploc(b)+1;
	 }
      }
   }
#endif
   *b = '\0';
}

/*
 * save_cc - determine if the condition codes are live at this point
 */
int 
save_cc(struct bblock * sblk, struct list * ptr) {

   struct bblock *cblk = sblk;

   do {
      while (ptr) {
         if (cnts(ptr)) {
	         if (is_call(ptr->val) || set_cc(ptr->val))
	            return 0;
            if (is_rjmp(ptr->val))
	            return ((*(ptr->val + 3) == 'I') ? 1 : 2);
         }
         ptr = ptr->next;
      } /* end while (ptr) */

      assert(NULL != cblk);
      if (cblk->left && !cblk->right) {
         cblk = cblk->left;
         ptr = cblk->rtls;
      }
      else
         cblk = (struct bblock *) NULL;
	} while (cblk);

   return 0;
}

/*
 * memsize - returns number of bytes that a memory reference fetches
 */
int memsize(char m) {

   switch (m) {

	 case 'B':
	 return (1);

      case 'W':
	 return (2);

      case 'F':
      case 'R':
	 return (4);

      case 'D':
	 return (8);

      default:
	 (void) fprintf(stderr, "vpo (memsize): invalid reference %c\n", m);
	 exit(1);
   }

#ifdef lint
   return (0);
#endif
}

/*
 * link_module - link functions within this module
 */
#ifdef INTERACT
void link_module(FILE *fp)
#else
void link_module()
#endif
{
#ifndef INTERACT
	FILE *fp=stdout;
#endif
   (void) fprintf(fp,"%sease_link_%d:\n", EasePrefix, funcnum1);
   (void) fprintf(fp,".word %sease_link_%d\n", EasePrefix, ++funcnum1);
   (void) fprintf(fp,".word %sease_link_0\n", EasePrefix);
}

/*
 * call_graph - emit a structure that connects functions at link time
 */
#ifdef INTERACT
void call_graph(FILE *fp)
#else
void call_graph()
#endif
{
#ifndef INTERACT
	FILE *fp=stdout;
#endif
   struct callfunc *fptr;
   struct list *ptr;
   char new[MAXSTRING];
   int word_label;

   /*
      Output external functions referenced from this function
   */
	calc_funcs();
   (void) fprintf(fp,".word %d\n", ext_num);
   for (fptr = ext_funcs; fptr; fptr = fptr->next)
      (void) fprintf(fp,".word %sease_%s_ptr\n", EasePrefix, fptr->name);

   /*
      Output a function that will return a pointer to this function's
      information block.  This function must be static if the current function
      is static and global otherwise.
   */
   (void) fprintf(fp,".text\n");
   (void) sprintf(new, ".globl\t%s", fn);

#if 0  /* always emit .globl directive */
   for (ptr = top->rtls; ptr; ptr = ptr->next) {
      if ((ptr->type & ASMLINE) && ptr->val && !strcmp(ptr->val, new)) {
#endif
	 (void) fprintf(fp,".globl\t%sease_%s_ptr\n", EasePrefix, fn);
#if 0
	 break;
      }
   }
#endif

   fprintf(fp,"%sease_%s_ptr:\n", EasePrefix, fn);
   fprintf(fp,"\tldr\tr0,.Lease00%d\n", word_label = get_label_number());
   fprintf(fp,"\tmov pc, lr\n");
   fprintf(fp, ".Lease00%d:\t.word\t%sease_%s_info\n", word_label, EasePrefix, fn);

}

void	calc_funcs()
{
	struct bblock* cblk;
	struct list* rtlptr;

	for(cblk=top;cblk;cblk=cblk->down)
	{
		for(rtlptr=cblk->rtls; rtlptr; rtlptr=rtlptr->next)
		{
			/* check for ST=<global>.. so, the +3 is correct here */
			if(cnts(rtlptr) && is_call(rtlptr->val) && isglob(rtlptr->val+3))
			{
				/* if we're calling a global function */
				add_func(globalname(rtlptr->val+3));

			}

		}

	}
}
#endif  /* MEASURE */




