/*
 * This machine description describes the StrongARM chip
 *
 * It contains 0 shift/reduce and 0 reduce/reduce conflicts
 */

%{
#ifdef INTERACT
#define yyparse stdout_yyparse
#endif

#include <stdio.h>
#include "recdef.h"
#include "addrdef.h"
#include "instdef.h"
#include "inst.h"
#include "vpo.h"
#include "isem.h"
#include "regs.h"

extern int yylength;
extern int erflag;

void makereg(char *reg, int rt, int rnum);

static char* check_codes(char* reg)
{
	if(getRegNum(reg)!=0)
		erflag=1;


	return 0;
}

static char* getPCStr()
{
	static char pcreg[REGSTRING];
        makereg(pcreg,RTREG,15);
        return pcreg;
	
}


%}

%union {
   char *str_ptr;
   struct sem_rec *rec_ptr;
   struct addr_entry *addr_ptr;
   char compare_const;
   };

/*
 * The GLOBAL token always represents a global identifier
 */
%token <str_ptr> GLOBAL
%token ERROR

/*
 * The LOCAL token always represents a local identifier
 */
%token <str_ptr> LOCAL

/*
 * The LBL token always represents a label
 */
%token <str_ptr> LBL

/*
 * The CON token always represents an integer constant above 10.
 * The single character constants ('0'..'9') represent themselves.
 */
%token <str_ptr> CON

/*
 * The FCON token always represents an floating point constant
 */
%token <str_ptr> FCON

/*
 * The following upper-case letter constants are defined by the RTL standard
 *
 *     PC - represents the program counter
 *     RT - represents the return label
 *     ST - represents a call instruction
 */
%token PC RT ST 

/*
 * The REGx tokens always represent register encodings of type x (hexadecimal)
 * REG0 is an integer reg, REG1 is an float reg, REG2 is a double REG 
 */
%token REG0 REG1 REG2 REG3

/*
 * The xMEM tokens always represent the "x[" memory reference sequence
 * char, double, float, int, short 
 */
%token BMEM DMEM FMEM RMEM WMEM 


/*
 * Two letter combinations always represent those two letters 
 *
 * SL - Spill location 
 * CV - convert, from single to double or double to single
 */
%token  SL CV YZ ZY YI ZI IY IZ UY YU UZ ZU

/* an "invariant" load.. no need to worry about checking
for aliases with other loads/stores */
%token LA 



%left '='
%left '+' '-'
%left '|' '&' '^'
%left '"' '{' '}'
%left '*' '/' '\\' '%' '#'
%nonassoc '~' ','


%type <rec_ptr> branch,jump
%type <rec_ptr> cc_binary_op, binary_op, move, neg, l, s, ret, call
%type <rec_ptr> rhs, adr, cmp, fcmp, cvt, mvn, cvtu, negcc, mvncc

%type <addr_ptr> shifted_reg, src, mem, dst, faddr, addr, expr, expr, addr, adraddr
%type <addr_ptr> fregexp,regexp, label, con, immed, reg, freg, rreg, dreg
%type <addr_ptr> regpreg, negcon, labelexp, opt_pred_end, opt_pred_mem
%type <str_ptr> rel predicate opt_pred_equal opt_pred_cc pred_div CC 

%type <compare_const> compare_const


%%

inst 	: 
		branch			{ branch($1);		}
	| 	jump			{ jump($1);		}
	|	binary_op		{ bin_emit($1,NULL); 	}
	|	binary_op cc_binary_op	{ bin_emit($1,$2);   	}
	|	cc_binary_op binary_op	{ bin_emit($2,$1);   	}
	|	cc_binary_op		{ compares($1);		}
	|	move cmp		{ move($1, $2);	 	} 
	|	cmp move		{ move($2, $1);	 	}	 
	|	move			{ move($1,NULL);	} 
	|	adr			{ adr($1); 		} 
	|	neg negcc		{ neg($1,$2); 		}
	|	negcc neg		{ neg($2,$1); 		}
	|	neg			{ neg($1,NULL);		}
	|	mvn mvncc		{ mvn($1,$2); 		}
	|	mvncc mvn		{ mvn($2,$1); 		}
	|	mvn			{ mvn($1,NULL);		}
	| 	l			{ l($1);		}
	| 	s			{ s($1);		}
 	|       ret             	{ retrn(); 		}
 	|       call            	{ call($1); 		}
	|	cmp			{ cmp($1);		}
	|	fcmp			{ fcmp($1);		}
	| 	cvt			{ cvt($1);		}
	| 	cvtu			{ cvtu($1);		}
	;


cvt 	:	dst opt_pred_equal  cv_op '[' dst ']' opt_pred_end 
		{ $$ = add_predicate(record($1, $5, AN), $2, $7);}
	;
cv_op	:	YZ | ZY | CV | YI | ZI | IY | IZ ;

cvtu 	:	dst opt_pred_equal  cvu_op '[' dst ']' opt_pred_end 
		{ $$ = add_predicate(record($1, $5, AN), $2, $7);}
	;

cvu_op	:	YU | ZU | UY | UZ 
	;

branch  : 	PC '=' predicate label ';' 	{ $$ = add_predicate(record(AN, $4, AN), $3, AN);}
	| 	PC '=' predicate rreg  ';' 	{ $$ = add_predicate(record(AN, $4, AN), $3, AN);}
	;

jump	: 	PC '=' label ';'        	{ $$ = record(AN, $3, AN); }
        |       PC '=' rreg ';'         	{ $$ = record(AN, $3, AN); }
        |       PC '=' GLOBAL ';'         	{ $$ = record(AN, addr(ANY,GIDENT,AE $3,AN, AN), AN); }
	;


cmp	
	: 	CC opt_pred_equal rreg compare_const rreg opt_pred_cc   
		{ $$=add_predicate(cmp_record(AN, $3, $5,$4),$2, AN); }
	| 	CC opt_pred_equal rreg compare_const negcon opt_pred_cc 
		{ $$=add_predicate(cmp_record(AN, $3, $5,$4), $2, AN);}
	| 	CC opt_pred_equal rreg '?' rreg opt_pred_cc   
		{ $$=add_predicate(cmp_record(AN, $3, $5,'?'),$2, AN); }
	| 	CC opt_pred_equal rreg '?' negcon opt_pred_cc 
		{ $$=add_predicate(cmp_record(AN, $3, $5,'?'), $2, AN);}
	;

fcmp	: 	CC opt_pred_equal freg '?' freg opt_pred_cc 
		{$$=add_predicate(fcmp_record(AN,$3,$5,(char*)AN),$2,AN);}
	| 	CC opt_pred_equal dreg '?' dreg opt_pred_cc 
		{$$=add_predicate(fcmp_record(AN,$3,$5, (char*)AN),$2,AN);}
	| 	dst opt_pred_equal dst rel dst opt_pred_end 
		{$$=add_predicate(fcmp_record($1,$3,$5,$4),$2,AN);}
	;


binary_op: dst   opt_pred_equal rhs opt_pred_end { $$ = add_predicate(binary_op($1, $3),$2,$4);}
	;
/*

yyfast breaks when I add this to binary_op, dunno why
{ $$ = add_predicate(binary_op(addr(LONG, RREGISTER, AE getPCStr(), AN, AN), $3),NULL,$4);}

*/



/* 
 * basically we allow the CC to be set in any way, we let 
 * bin_emit do the checking about whether
 * this op matches that op, etc. 
 */ 
cc_binary_op
	:  	CC opt_pred_equal rhs compare_const '0' opt_pred_cc 
	{ $$ = add_predicate(binary_op(0, $3),$2,0); check_cc_semantics($3,$4);}
	;

compare_const
	:	':'	{ $$=':';}
	|	'!'	{ $$='!';}

move	:	dst opt_pred_equal expr opt_pred_end 
		{ $$ = add_predicate(record($1, $3, AN), $2, $4);}
	|	dst opt_pred_equal dst opt_pred_end 
		{ $$ = add_predicate(record($1, $3, AN), $2, $4);}
	;

adr	: 	dst opt_pred_equal adraddr opt_pred_end 
		{ $$ = add_predicate(record($1, $3, AN), $2, $4);}
	;

mvn	: 	dst opt_pred_equal '~' dst opt_pred_end 	
		{ $$ = add_predicate(record($1, $4, AN), $2, $5);}
	;
mvncc	: 	CC opt_pred_equal '~' dst compare_const '0' opt_pred_cc 	
		{ $$ = add_predicate(record(AN, $4, AN), $2, AN);}
	;

neg	: 	dst opt_pred_equal '-' dst opt_pred_end	{ $$ = add_predicate(record($1, $4, AN), $2, $5);}
	;
negcc	: 	CC opt_pred_equal '-' dst compare_const '0' opt_pred_cc
		{ $$ = add_predicate(record(AN, $4, AN), $2, AN);}
	;


l 	/* 
	 * these do not have a pre/post modification 
	 */
	:       dst opt_pred_equal mem opt_pred_end             
	{ $$ = add_predicate(pre_record($1, $3, AN, (struct sem_rec*)AN), $2, $4);}
        |       dst opt_pred_equal mem '&' con opt_pred_end     
	{ $$ = add_predicate(pre_record($1, $3, $5, (struct sem_rec*)AN), $2, $6);}
        |       dst opt_pred_equal '(' mem '{' con ')' '}' con opt_pred_end 
	{ $$ = add_predicate(pre_record($1, $4, same($6, $9), (struct sem_rec*)AN), $2, $10);}

	/* this have the dst as the PC.  We only need to support integer loads though */
	|       PC opt_pred_equal mem opt_pred_end             
	{ $$ = add_predicate(pre_record(addr(LONG, RREGISTER, AE getPCStr() , AN, AN), $3, AN, (struct sem_rec*)AN), $2, $4);}



	/* 
	 * these have the pre/post modification after the load effect 
	 */
	|       dst opt_pred_equal mem opt_pred_end binary_op   
	{ $$ = add_predicate(pre_record($1, $3, AN, $5), $2, $4);}
        |       dst opt_pred_equal mem '&' con opt_pred_end binary_op 
	{ $$ = add_predicate(pre_record($1, $3, $5, $7), $2, $6);}
        |       dst opt_pred_equal '(' mem '{' con ')' '}' con opt_pred_end binary_op 
	{ $$ = add_predicate(pre_record($1, $4, same($6, $9), $11), $2, $10);}

	/* 
	 * these have the pre/post modification before the load effect 
	 */
	|       binary_op dst opt_pred_equal mem opt_pred_end             
	{ $$ = add_predicate(pre_record($2, $4, AN, $1), $3, $5);}
        |       binary_op dst opt_pred_equal mem '&' con opt_pred_end     
	{ $$ = add_predicate(pre_record($2, $4, $6, $1), $3, $7);}
        |       binary_op dst opt_pred_equal '(' mem '{' con ')' '}' con opt_pred_end 
	{ $$ = add_predicate(pre_record($2, $5, same($7, $10), $1), $3, $11);}

	/* 
	 * this may have an empty "move" effect 
	 */
	|       dst opt_pred_equal mem opt_pred_end move   
	{ $$ = add_predicate(pre_record($1, $3, AN, $5), $2, $4);}
        |       dst opt_pred_equal mem '&' con opt_pred_end move 
	{ $$ = add_predicate(pre_record($1, $3, $5, $7), $2, $6);}
        |       dst opt_pred_equal '(' mem '{' con ')' '}' con opt_pred_end move 
	{ $$ = add_predicate(pre_record($1, $4, same($6, $9), $11), $2, $10);}
	|       move dst opt_pred_equal mem opt_pred_end             
	{ $$ = add_predicate(pre_record($2, $4, AN, $1), $3, $5);}
        |       move dst opt_pred_equal mem '&' con opt_pred_end     
	{ $$ = add_predicate(pre_record($2, $4, $6, $1), $3, $7);}
        |       move dst opt_pred_equal '(' mem '{' con ')' '}' con opt_pred_end 
	{ $$ = add_predicate(pre_record($2, $5, same($7, $10), $1), $3, $11);}
	;

s	:	mem opt_pred_equal dst opt_pred_mem             
	{ $$ = add_predicate(pre_record($1, $3, AN,(struct sem_rec*)AN), $2, $4);}
        |       mem opt_pred_equal dst opt_pred_mem  binary_op  
	{ $$ = add_predicate(pre_record($1, $3, AN,$5), $2, $4);}
        |       mem opt_pred_equal dst opt_pred_mem  move       
	{ $$ = add_predicate(pre_record($1, $3, AN,$5), $2, $4);}
        |       binary_op mem opt_pred_equal dst opt_pred_mem   
	{ $$ = add_predicate(pre_record($2, $4, AN,$1), $3, $5);}
	|	move      mem opt_pred_equal dst opt_pred_mem   
	{ $$ = add_predicate(pre_record($2, $4, AN,$1), $3, $5);}
	;



call	: ST opt_pred_equal rreg ';'   
	{$$=add_predicate(record(AN, addr(ANY,REGDFR, $3,AN, AN), AN), $2, AN);}
	/* oh, you so don't want to mess with this line */
	| ST opt_pred_equal GLOBAL ';' 
	{$$=add_predicate(record(AN, addr(ANY,ABSOLUTE,addr(ANY,GIDENT,AE $3,AN, AN),AN,AN), AN),$2,AN);}
	;

ret	:       PC '=' RT ';' { $$=0; }
	;


rhs     :       dst '+' src             { $$ = brecord(ADDI, $1, $3); }
	|       dst '-' src             { $$ = brecord(SUBI, $1, $3); }
        |       dst '^' src             { $$ = brecord(XORI, $1, $3); }
        |       dst '&' src             { $$ = brecord(ANDI, $1, $3); }
        |       dst '|' src             { $$ = brecord(ORI, $1, $3);  }
	|       dst '/' src             { $$ = brecord(DIVI, $1, $3); }
        |       dst '{' src             { $$ = brecord(SLLI, $1, $3); }
        |       dst '}' src             { $$ = brecord(SRAI, $1, $3); }
        |       dst '"' src             { $$ = brecord(SRLI, $1, $3); }
	|       expr '-' dst            { $$ = brecord(RSBI, $3, addr(ANY, IMMED, $1, AN, AN)); }   
	|	'(' rhs ')' '+' dst 		{ $$=  mla_record($2, $5); 	}
	|	dst '+' '(' rreg '*' rreg ')'	{ $$=  mla_record(brecord(MULI,$4, $6), $1); }
        |       dst '*' src             { $$ = brecord(MULI, $1, $3); }
	|       '-' dst '+' src         { $$ = brecord(RSBI, $2, $4); }  
 	| 	'(' rhs ')'			{ $$=$2; }
	;

src     :       dst
        |       immed
	|	shifted_reg
        ;

shifted_reg :	'(' rreg '{' rreg ')'	{ $$ = sr_addr(ANY, SR_EXPR, AE SLLI, $2, $4); }
	|	'(' rreg '}' rreg ')'	{ $$ = sr_addr(ANY, SR_EXPR, AE SRAI, $2, $4); }
	|	'(' rreg '"' rreg ')'	{ $$ = sr_addr(ANY, SR_EXPR, AE SRLI, $2, $4); }
	|	'(' rreg '{' immed ')'	{ $$ = sr_addr(ANY, SR_EXPR, AE SLLI, $2, $4); }
	|	'(' rreg '}' immed ')'	{ $$ = sr_addr(ANY, SR_EXPR, AE SRAI, $2, $4); }
	|	'(' rreg '"' immed ')'	{ $$ = sr_addr(ANY, SR_EXPR, AE SRLI, $2, $4); }
	|	'(' rreg '*' immed ')'	{ $$ = sr_addr(ANY, SR_EXPR, AE MULI, $2, $4); }
	;

dst     :       REG0    { $$ =addr(LONG, RREGISTER, AE $<str_ptr>1, AN, AN); }
/*	| 	PC      { $$ =addr(LONG, RREGISTER, AE getPCStr() , AN, AN); }*/
        |       REG1    { $$ =addr(FLOAT, FREGISTER, AE $<str_ptr>1, AN, AN); }
        |       REG2    { $$ =addr(DOUBLE, DREGISTER, AE $<str_ptr>1, AN, AN); }
	;

rreg	:	REG0    { $$ =addr(LONG, RREGISTER, AE $<str_ptr>1, AN, AN); }
	;

freg	:	REG1    { $$ =addr(FLOAT, FREGISTER, AE $<str_ptr>1, AN, AN); }
	;

dreg	:	REG2    { $$ =addr(DOUBLE, DREGISTER, AE $<str_ptr>1, AN, AN); }
	;

immed   :       expr            	{ $$ = addr(ANY, IMMED, $1, AN, AN); }
	;


mem     :       BMEM addr ']'   { $$ = memref(BYTE, $2,255); }
        |       WMEM addr ']'   { $$ = memref(WORD, $2,255); }
        |       RMEM addr ']'   { $$ = memref(LONG, $2,4095); }
        |       FMEM faddr ']'  { $$ = memref(FLOAT, $2,1023); }
        |       DMEM faddr ']'  { $$ = memref(DOUBLE, $2,1023); }
        |       LA '[' addr ']' { $$ = memref(LONG, $3,4095); }
        ;

addr    :       reg             { $$ = addr(ANY, REGDFR, $1, AN, AN); }
	|	label		{ $$ = addr(ANY, ABSOLUTE, $1, AN, AN); checklabel($1);}
        |       '(' reg ')'     { $$ = addr(ANY, REGDFR, $2, AN, AN); }
	|	'(' label ')'	{ $$ = addr(ANY, ABSOLUTE, $2, AN, AN); checklabel($2);}
        |       regexp          { $$ = displacement($1,0,0); }
        |       regpreg         { $$ = displacement($1,0,0); }
	|	labelexp	{ $$ = addr(ANY, ABSOLUTE, $1, AN, AN); }
        ;

regpreg :       reg '+' reg             { $$ = initexpr($1, 0, $3); }
 	|       reg '-' reg             { $$ = initexpr($1, 1, $3); }
	|	reg '+' shifted_reg	{ $$ = initexpr($1, 0, $3); }
	|	reg '-' shifted_reg	{ $$ = initexpr($1, 1, $3); }
	;

labelexp:       label '+' expr            { $$ = addr(ANY, EXPR, AE ADDI, $1, $3); checklabel($1);}
        |       label '-' expr            { $$ = addr(ANY, EXPR, AE SUBI, $1, $3); checklabel($1);}
        |       expr '+' label            { $$ = addr(ANY, EXPR, AE ADDI, $3, $1); checklabel($3);}
	|       labelexp '+' expr         { $$ = addr(ANY, EXPR, AE ADDI, $1, $3); }
        |       labelexp '-' expr         { $$ = addr(ANY, EXPR, AE SUBI, $1, $3); }
        |       '(' labelexp ')'          { $$ = $2; }
	;
regexp  :       reg '+' expr            { $$ = initexpr($1, 0, $3); }
        |       reg '-' expr            { $$ = initexpr($1, 1, $3); }
        |       expr '+' reg            { $$ = initexpr($3, 0, $1); }
	|       regexp '+' expr         { $$ = addexpr($1, ADDI, $3); }
        |       regexp '-' expr         { $$ = addexpr($1, SUBI, $3); }
        |       '(' regexp ')'          { $$ = $2; }
        ;

faddr   :       expr            { $$ = addr(ANY, ABSOLUTE, $1, AN, AN); }
        |       reg             { $$ = addr(ANY, REGDFR, $1, AN, AN); }
        |       fregexp         { $$ = displacement($1,0,0); }
        ;


fregexp :       reg '+' expr            { $$ = initexpr($1, 0, $3); }
        |       reg '-' expr            { $$ = initexpr($1, 1, $3); }
        |       expr '+' reg            { $$ = initexpr($3, 0, $1); }
        |       regexp '+' expr         { $$ = addexpr($1, ADDI, $3); }
        |       regexp '-' expr         { $$ = addexpr($1, SUBI, $3); }
        |       '(' regexp ')'          { $$ = $2; }
        ;


adraddr :       GLOBAL          	{ $$ = addr(ANY,GIDENT,AE $1, AN, AN); }
        |       label			{ $$=$1; checklabel($1); 	       }
	;

reg     :       REG0    { $$ = addr(LONG, RREGISTER, AE $<str_ptr>1, AN, AN); }
	;

CC	:	REG3	{ $$=check_codes($<str_ptr>1); }
	;

expr    :       expr '+' expr   { $$ = addr(ANY, EXPR, AE ADDI, $1, $3); }
        |       expr '-' expr   { $$ = addr(ANY, EXPR, AE SUBI, $1, $3); }
        |       expr '*' expr   { $$ = addr(ANY, EXPR, AE MULI, $1, $3); }
        |       expr '/' expr   { $$ = addr(ANY, EXPR, AE DIVI, $1, $3); }
        |       expr '%' expr   { $$ = addr(ANY, EXPR, AE REMI, $1, $3); }
        |       expr '{' expr   { $$ = addr(ANY, EXPR, AE SLLI, $1, $3); }
        |       expr '"' expr   { $$ = addr(ANY, EXPR, AE SRLI, $1, $3); }
        |       expr '^' expr   { $$ = addr(ANY, EXPR, AE XORI, $1, $3); }
        |       expr '&' expr   { $$ = addr(ANY, EXPR, AE ANDI, $1, $3); }
        |       expr '|' expr   { $$ = addr(ANY, EXPR, AE ORI, $1, $3); }
        |       '-' expr %prec '~' { $$ = addr($2->type, EXPR, AE NEGI, $2, AN); }
        |       '~' expr        { $$ = addr(ANY, EXPR, AE NOTI, $2, AN); }
        |       SL '[' con ']'  { $$ = addr(ANY, SPILLOC, $3, AN, AN); }
        |       con
        |       FCON    { $$ = addr(FPANY, FCONST, AE $1, AE yylength, AN); }
        | 	'(' expr ')'    { $$ = $2; }
        |       LOCAL          	{ $$ = addr(ANY,LIDENT,AE $1, AN, AN); }
        ;

label   :       LBL     { $$ = addr(LONG, LABEL, AE $1, AE yylength, AN); }
        ;

negcon  : 	'-' con { $$ = addr($2->type, EXPR, AE NEGI, $2, AN); }
	| 	con     { $$ = $1; }
	;

con     :       CON     { $$ = addr(ANY, ICONST, AE $1, AE yylength, AN); }
        |       '0'     { $$ = addr(ANY, ICONST, AE "0", AE 1, AN); }
        |       '1'     { $$ = addr(ANY, ICONST, AE "1", AE 1, AN); }
        |       '2'     { $$ = addr(ANY, ICONST, AE "2", AE 1, AN); }
        |       '3'     { $$ = addr(ANY, ICONST, AE "3", AE 1, AN); }
        |       '4'     { $$ = addr(ANY, ICONST, AE "4", AE 1, AN); }
        |       '5'     { $$ = addr(ANY, ICONST, AE "5", AE 1, AN); }
        |       '6'     { $$ = addr(ANY, ICONST, AE "6", AE 1, AN); }
        |       '7'     { $$ = addr(ANY, ICONST, AE "7", AE 1, AN); }
        |       '8'     { $$ = addr(ANY, ICONST, AE "8", AE 1, AN); }
        |       '9'     { $$ = addr(ANY, ICONST, AE "9", AE 1, AN); }
        ;

rel     :       ':'     { $$ = "eq"; }
        |       '!'     { $$ = "ne"; }
        |       '`'     { $$ = "ge"; }
        |       '\''    { $$ = "le"; }
        |       '<'     { $$ = "lt"; }
        |       '>'     { $$ = "gt"; }
        |       'g'     { $$ = "cs"; }
        |       's'     { $$ = "ls"; }
        |       'l'     { $$ = "cc"; }
        |       'h'     { $$ = "hi"; }
        ;


predicate:	CC rel '0' ',' 		{ $$ = $2; }
	;

opt_pred_equal:	'=' predicate   	{ $$ = $2; }
	|	'='			{ $$ = (char*)AN; }
	;

opt_pred_end:	pred_div dst ';'  	{ $$ = $2; }
	|	';' 			{ $$ = AN; }
	;

opt_pred_mem:	pred_div mem ';'  	{ $$ = $2; }
	|	';' 			{ $$ = AN; }
	;

opt_pred_cc:	pred_div CC ';'  	{ $$ = (char*)AN; }
	|	';' 			{ $$ = (char*)AN; }
	;

pred_div:	','			{ $$=(char*)AN; }
	;



%%

#include <stdio.h>

yyerror(char* str)
{
        fprintf(stderr, "parse error: %s\n", str);
}
