#ifndef _PLACEMENT_H
#define _PLACEMENT_H 1

/************************************************************************
 *                                                                      *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   *
 *     Zephyr Software, LLC                                             *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from Zephyr Software, LLC                                       *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/

/* How far from a base register value can an offset be and still be addressed
 * using an immediate displacement field within the opcode.
 */
#define DISPLACEMENT_ADDR_MODE_LIMIT 4095

/* Definitions of cache hierarchy for UltraSparc-I and UltraSparc-II.
 */
#define L1_DATA_CACHE_SIZE       16384
#define L1_DATA_CACHE_SUBBLOCK_SIZE 16
#define INTS_PER_SUBBLOCK (L1_DATA_CACHE_SUBBLOCK_SIZE / sizeof(int))
#define L1_DATA_CACHE_LINE_SIZE     32
#define L1_DATA_CACHE_ASSOCIATIVITY  1
#define L1_DATA_CACHE_WAY_SIZE (L1_DATA_CACHE_SIZE / L1_DATA_CACHE_ASSOCIATIVITY)
#define L1_DATA_CACHE_NO_WRITE_ALLOCATE  TRUE /* really TRUE; measuring. */

#define L2_DATA_CACHE_SIZE      262144
#define L2_DATA_CACHE_LINE_SIZE     64
#define L2_DATA_CACHE_ASSOCIATIVITY  1
#define L2_DATA_CACHE_WAY_SIZE (L2_DATA_CACHE_SIZE / L2_DATA_CACHE_ASSOCIATIVITY)
#define L2_UNIFIED_CACHE          TRUE

/* Definitions used for packing cache lines, partitioning data sets using
 * Chaco, and the static profiling needed as input to placement optimizations.
 */
#define CACHE_LINE_PACKING_LOGARITHM 6 /* 2**N = size of packed cache lines. */
#define CACHE_LINE_PACKING_QUANTUM 64  /* size of packed cache line. */
#define CACHE_LINES_PER_PARTITION \
 (L1_DATA_CACHE_WAY_SIZE / CACHE_LINE_PACKING_QUANTUM)
#define DATA_PROF_SIZE_THRESHOLD (1 << (1 + CACHE_LINE_PACKING_LOGARITHM))

/* Constants that tell which locals are addressed from the frame pointer,
 * the stack pointer, etc.  These go in the addr_mode field of the locuse
 * record.
 */
#define NO_ADDR_MODE        -1
#define FP_RELATIVE          0
#define SP_RELATIVE          1
#define MUST_BE_FP_RELATIVE  2
#define LAST_SP_RELATIVE     3
#define NOT_LAST_SP_RELATIVE 4

/* These flags enable the coalescing and placement optimizations. */
#define COALESCE_INTS        FALSE
#define COALESCE_FLOATS      FALSE
#define DATA_PLACEMENT             FALSE

#define CLUSTERING_DATA_PLACEMENT  FALSE  /* WARNING: not ready for use. */

/* Performs static profiling of local data item use below some size limit. */
int   BuildBlockDataProfile(int);

/* Rearranges the stack frame to reduce cache misses and addressing costs. */
int   PlaceStackData(int, int);

#endif /* _PLACEMENT_H */
