

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "../lib/vpo.h"
#include "../lib/assign.h"
#include "../lib/misc.h"

#include "recdef.h"
#include "addrdef.h"
#include "instdef.h"
#include "inst.h"
#include "predicates.h"
#include "regs.h"
#include "regs2.h"

/* No more than 3 RTLs in a predicated block. */
/* A branch takes 3 cycles to complete, so, never go past this. */
int max_pred_cost=3;

/* This character is used to seperate the predicate from the re-use of the
   destination. */
static char pred_div=',';

/*
 * This routine checks to see if the RTL can be predicated 
 * on the condition given.  It returns true if possible, false otherwise 
 */
int can_predicate_rtl(struct list* rtl, int * cc_set, char* desired_condition)
{
	char *a;

	/* on the arm, any non-rtl line is OK to predicate */
	if((rtl->type & RTLINE) != RTLINE)
		return TRUE;


	/* or, any rtl with no value.. this is just a list of dead regs */
	if(rtl->val == 0)
		return TRUE;


	a=rtl->val;

	while(a=gtrg(a))
	{
		/* if it has a floating point reg, we can't predicate..sorry charlie */
		if(regtype(unskiploc(a))==FTREG || regtype(unskiploc(a))==DTREG)
			return FALSE; 
	}


	/* OK, now check for other predicates already in this rtl */
	a=rtl->val;
	while(*a++ != '=');

	if(cmploc(a,cc_reg)==0)
	{
		/* uh oh.. we already have a predicate if there is an C after the equal sign */
		/* but!  maybe we'll get lucky and the predicate matches the desired predicate */
		if(!fmatch(a,desired_condition))
			/* bummer, the condition is different, return now */
			return FALSE;
	}

	/* if we get here, we know that this rtl either doesn't have a predicate, or the
		predicate matches the condition.  If the CC is set but this rtl doesn't have
		a predicate, we can't predicate it.. the predicates conflict (none and the
		one we'd like to add */
	if(*cc_set && cmploc(a,cc_reg)!=0)
	{
		return FALSE;	
	}



	/* decide if the CC is set */
	if(*rtl->val=='S' && *(rtl->val+1)=='T')
		*cc_set=TRUE;
	if(cmploc(cc_reg, rtl->val)==0)
		*cc_set=TRUE;


	/* for now, we won't predicate returns.. that's complicated and we
		have to fixup fixentry */
	if(strcmp(rtl->val, "PC=RT;")==0)
		return FALSE;


	/* we either didn't find a float reg, didn't find a predicate that doesn't match, 
	so we can go ahead and predicate this rtl.  However, not here. */
	return TRUE;
}


void predicate_rtl(struct list* rtl, char* condition)
{
	/* OK, we've decided it's safe to predicate this rtl.. and that
		it's profitable as well.  Here's the code to insert the predication */


	char new[MAXSTRING], new2[MAXSTRING];
	char dst[MAXSTRING];
        char *a, *b;

        /* on the arm, any non-rtl doesn't need any modifications to predicate */
        if((rtl->type&RTLINE) != RTLINE)
                return;


        /* or, any rtl with no value.. this is just a list of dead regs */
        if(rtl->val == 0)
                return;


	/* since the new rtl will be bigger, we work on a copy for now */
	strcpy(new,rtl->val);

	a=new;
	while(*a)
	{
		b=dst;
		/* copy the destination into dst */
        	while((*b++ = *a++)  != '=');
	
	
		/* oops, we don't need the equals */
		b--;
		/* don't forget to null terminate the dest. */
		*b=0;
	
	
		/* ooh, we munched up the destination.. and found that the rtl already has 
			a source side predicate though .. it must have matched in 
			can_predicate_rtl.. we just return because this needs no work */
		if(cmploc(a,cc_reg)==0)
        	{
			break;
        	}


		/* alright, it's safe to add the predicate. */
		sprintf(new2, "=%s,", condition);
		a=sub(a-1,a, new2);

		if(strncmp(rtl->val, "ST=",3)==0)
		{
			/* aah, branches and calls don't need a repeat of the destination */
			/* However, we do need a P line that specifies that all scratch regs */
			/* are potentially used */
			char p_line[MAXSTRING]={0,};
			char *p_line_ptr=p_line;
			int i;
			for(i=0;i<4;i++)
			{
				makereg(p_line_ptr,RTREG,i);	
				p_line_ptr=skiploc(p_line_ptr);
				makereg(p_line_ptr,FTREG,i);	
				p_line_ptr=skiploc(p_line_ptr);
				makereg(p_line_ptr,DTREG,i);	
				p_line_ptr=skiploc(p_line_ptr);
			}
			makereg(p_line_ptr,RTREG,12);	
			p_line_ptr=skiploc(p_line_ptr);
			*p_line_ptr=0;
			add_rtl_with_type(rtl->daddy,rtl,p_line,rinit(), PARMLINE);
			break;
		}
		if(strncmp(rtl->val, "PC=",3)==0)
		{
			/* aah, branches don't need a repeat of the
				destination, so we just exit now rtl.  */
			break;
		}
		/* now, hunt down the end of the effect.*/
       		while(*(a++) != ';');

		/* and insert the end of effect predication divider and 
			add the destination as another source */
		/* maybe we can handle stores differently and only see the address as re-used
			instead of the memory itself.  have to think about this */
		sprintf(new2, "%c%s;", pred_div, dst);
		a=sub(a-1, a, new2);

		/* ok, go to the next effect if there is one */
	}

	if(debug_opt("dump_rtl_after_predicate_rtl"))
	{
		fprintf	(stderr, "Rtl: ");
		show_rtl(stderr, rtl->val);
		fprintf (stderr, "under predicate: ");
		show_rtl(stderr, condition); 
		fprintf (stderr, "Converted to:");
		show_rtl(stderr, new);
		fprintf (stderr, "\n");
	}


	/* now, put the rtl over the old one */
	FREE(rtl->val);
	rtl->val=strcpy(CALLOC(strlen(new)+1), new);
#ifdef INTERACT
	mod_rtl(rtl, NULL, new);
#endif	
}


/* determine the cost of predicating this rtl */
int predication_cost(struct list* rtl, char* condition)
{
        /* on the arm, any non-rtl doesn't have a cost to predicate*/
        if((rtl->type & RTLINE) != RTLINE)
                return 0;


        /* or, any rtl with no value.. this is just a list of dead regs */
        if(rtl->val == 0)
                return 0;

	/* multiplies take 3 cycles to complete */
	/* so, if we want to predicate one, dont' predicate anything else */
	if(fmatch(rtl->val, "*") && !is_const(epos))
		return 3;
	
	return 1;
}


/* simply copy the condition from the conditional branch */
int get_predicate_condition(struct bblock* block, char* condition)
{
	struct list* rtl=block->rtlend;
	char* a=rtl->val+3;
	char* b=condition;

	if((rtl->type&RTLINE)!=RTLINE || rtl->val==0)
		return FALSE;
	if(!fmatch(rtl->val,","))
		return FALSE;

	while( (*b++=*a++) != ',');

	/* oops, we dont' need that comma */
	b--;
	*b=0;

	/* now, the branch is taken when this predicate is true.  We need to reverse it
		so that it can be applied to the rtls along the not-taken path */


	a=skiploc(condition);

	/* some code stolen from reverse */
   	switch (*a)
   	{

      		case ':':
         		*a = '!';
         		break;

      		case '!':
         		*a = ':';
         		break;
		
      		case '>':
         		*a = '\'';
         		break;
		
      		case '<':
         		*a = '`';
         		break;
		
      		case '`':
         		*a = '<';
         		break;
      		case '\'':
         		*a = '>';
         		break;
		
      		case 'h':
         		*a = 's';
         		break;
		
      		case 'l':
         		*a = 'g';
         		break;
		
      		case 'g':
         		*a = 'l';
         		break;
		
      		case 's':
         		*a = 'h';
         		break;
		
      		default:
			return FALSE;
   	}

	/* OK, we've reversed it */
	return TRUE;


}


/* determine if two rtls branch on the same conditions to the same place, for arm, that means
	they are identical */
int same_condition(struct list* rtl1, struct list* rtl2)
{
	if((rtl1->type&RTLINE)==RTLINE && (rtl2->type&RTLINE)==RTLINE && 
		rtl1->val && rtl2->val)
	{
		/* return true if they are identical, false otherwise */
		return strcmp(rtl1->val, rtl2->val)==0;
	}

	/* if they aren't rtlines, give up */
	return FALSE;
}


void reverse_pred_condition(char *condition)
{
  char *a = condition;

  a=skiploc(condition);
  
  /* some code stolen from reverse */
  switch (*a)
    {
      
    case ':':
      *a = '!';
      break;
      
    case '!':
      *a = ':';
      break;
      
    case '>':
      *a = '\'';
      break;
      
    case '<':
      *a = '`';
      break;
      
    case '`':
      *a = '<';
      break;
    case '\'':
      *a = '>';
      break;
      
    case 'h':
      *a = 's';
      break;
      
    case 'l':
      *a = 'g';
      break;
      
    case 'g':
      *a = 'l';
      break;
      
    case 's':
      *a = 'h';
      break;
      
    default:
      break;
    }


}
