#ifndef RECDEF_H
#define RECDEF_H

#include "addrdef.h"

#define SEMTABLESIZE 128

/* MLA has 5.. dst, left, right, inc, and op, although op is redundant */
/* it needs to match binst's op */
#define MAXFIELDS 5

struct sem_rec 
{
    union 
    {
      struct addr_entry *fields[MAXFIELDS];

      /* no dest for a cond_branch, but we leave it as a placeholder */
      struct { struct addr_entry *dst, *lbl; char *rel; }         cond_branch;

      /* no dests for branches.. but we leave it as a placeholder */
      struct { struct addr_entry *dst, *target; }                 bl;
      struct { struct addr_entry *dst, *target; }                 jmp;
      struct { struct addr_entry *dst, *lbl; }                    branch;
      struct { struct addr_entry *dst, *left, *right; char type;} cmp;
      struct { struct addr_entry *dst, *left, *right; char* op;}  fcmp;

      /* these things really do have dsts, add_predicate will make sure the
      second source matches the dest */
      struct { struct addr_entry *dst, *left, *right; int op; }   binst;
      struct { struct addr_entry *dst, *left, *right; int op;
               struct addr_entry *inc;  }                         mla;
      struct { struct addr_entry *dst, *mem, *unsign; 
               struct sem_rec* incr;    }                         l;
      struct { struct addr_entry *dst, *src; }                    cvt;
      struct { struct addr_entry *dst, *src; }                    mov;
      struct { struct addr_entry *dst, *src; }                    neg;
      struct { struct addr_entry *dst, *src; }                    not;


      /* do not remove "dummy" because pre_record assumes incr is
         the 4th field */
      struct { struct addr_entry *mem, *src, *dummy;
               struct sem_rec* incr; }                            s;
    } sem;

    /* Potentially any instruction can have a predicate. We record it here. *
     * The grammar records a 2 character string for each possible predicate *
     * or sets this to null if no predicate exists.
     */
    char *pred;
};
#endif
