
/************************************************************************/
/*                                                                      */
/* Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   */
/*     The Rector and Visitors of the University of Virginia            */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from the University of Virginia.                                */
/* This code is distributed WITHOUT ANY WARRANTY.  No claims are        */
/*      made as to whether it serves any particular purpose or even     */
/*      works at all.                                                   */
/*                                                                      */
/************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/arm/sched.c,v $
 $Date: 2002/07/16 19:44:04 $
 $Revision: 1.4 $
 $Name:  $

 Log of changes:

 $Log: sched.c,v $
 Revision 1.4  2002/07/16 19:44:04  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.3  2001/09/24 20:35:36  jdh8d
 *** empty log message ***

 Revision 1.2  2001/05/21 18:42:56  jdh8d
 *** empty log message ***

 Revision 1.1  2001/04/26 19:00:16  jdh8d

 initial version of machine dependend scheduling for arm.

 Revision 1.5  2000/10/10 19:48:08  jwd
 Cleaned up the code to make it clearer what hardware registers are being
 used by constructing the register strings with makeregs.

 Revision 1.4  1998/04/29 19:37:19  jwd
 Added header file string.h.
 Added void return type to function setup_noop().

 Revision 1.3  1998/04/07 20:03:46  jwd
 First ANSI C version.

*************************************************************************/

/*------------------------------------------------
   History of Changes
   ------------------
    1/19/98: Modified for use of ANSI C header files

------------------------------------------------*/

#include <stdio.h>
#include <string.h>

#include "../lib/vpo.h"
#include "../lib/assign.h"
#include "../lib/misc.h"
#include "../lib/schedule.h"

#include "instdef.h"
#include "regs.h"
#include "regs2.h"
#include "rtl.h"
#include "sched.h"


/*
 * md_sets_uses - handle machine-dependent set and use information
 */
int md_sets_uses(struct list * ptr) 
{
   register int i;
   char reg[REGSTRING];
   char* a;

   /*
      A load instruction reads from memory and cannot move past any instruction
      that does a memory write.  This means that reads may slide past each
      other.
   */

   ptr->does |= USEPC;
   if(is_pc(ptr->val))
   	ptr->blocks |= USEPC|SETPC;

/* note, we dont' need to worry so much about invariant loads */
   if (ISBASE(ptr, LDI) && !fmatch(ptr->val, "LA")) {
      ptr->does |= MEMREAD;
      ptr->blocks |= MEMWRITE;
   }

#ifdef SOFTFLOAT
	a=ptr->val;
	if(!gtmem(a) || (ptr->kind&CVTI))
	{
	/* non memory floating ops block the scratch and CC regs */
	   while(a=gtrg(a))
	   {
		if(regtype(a-2)!=RTREG)
		{
			ptr->blocks |= USES_SCRATCH;
	 		ptr->blocks |= CCREAD | CCWRITE;
	 		ptr->does |= CCWRITE;
			ptr->does|=USES_SCRATCH;
			break;
		}
	   }
	}
	a=ptr->val;
	while(a=gtrg(a))
	{
		if(scratchreg(a-2))
			ptr->does|=USES_SCRATCH;
		
	}
#endif

   /*
      A store instruction writes to memory and cannot move past any instruction
      that either reads or writes to memory.
   */
   if (ISBASE(ptr, STI)) {
      ptr->does |= MEMWRITE;
      ptr->blocks |= MEMREAD | MEMWRITE;
   }

   /*
      Depending on their kind (float or integer) a compare instructions and
      some arithmetic instructions set the condition codes.
   */
   if (ptr->kind & SETCC) {
	 ptr->does |= CCWRITE;
	 ptr->blocks |= CCREAD | CCWRITE;
   }

   /*
	any instruction may use the CC.  we loop through looking for 
	sets and uses, just to be sure we don't miss any
   */
   epos=ptr->val;
   while (fmatch(epos, cc_reg))
   {
	if(*epos!='=') 
	{
	  ptr->does |= CCREAD;
	  ptr->blocks |= CCWRITE;
	}
	else
	{
	 ptr->does |= CCWRITE;
	 ptr->blocks |= CCREAD | CCWRITE;
	}
   }

   /*
      A call instruction potentially sets and calls a lot of things.  We must
      enumerate them here.
   */
   if (ISBASE(ptr, CALLI)) 
   {

      /*
         Sets all types of scratch variables and the link reg
      */
      makereg(reg, RTREG, 0);
      touchreg(ptr->sets, reg);
      makereg(reg, RTREG, 1);
      touchreg(ptr->sets, reg);
      makereg(reg, RTREG, 2);
      touchreg(ptr->sets, reg);
      makereg(reg, RTREG, 3);
      touchreg(ptr->sets, reg);
      makereg(reg, RTREG, 12);
      touchreg(ptr->sets, reg);
      makereg(reg, RTREG, 14);
      touchreg(ptr->sets, reg);

      /*
         Sets integer condition codes
      */
      ptr->does |= CCWRITE;
      ptr->blocks |= CCREAD | CCWRITE;

      /*
         potentially read and write memory and mess with floating-point stuff
      */
      ptr->does |= MEMREAD | MEMWRITE;
      ptr->blocks |= MEMREAD | MEMWRITE;

      /*
	    Sets all scratch floating-point registers
      */
      makereg(reg, DTREG, 0);
      touchreg(ptr->sets, reg);
      makereg(reg, DTREG, 1);
      touchreg(ptr->sets, reg);
      makereg(reg, DTREG, 2);
      touchreg(ptr->sets, reg);
      makereg(reg, DTREG, 3);
      touchreg(ptr->sets, reg);
   }

   /*
      A return value of FALSE indicates that the standard sets and uses
      information should be gathered for this RTL.
   */
   return (FALSE);
}

/*
 * md_uses - handle use information for delay slot fills
 */
int md_uses(struct list * ptr, int *does, int *blocks, vector uses, vector sets2) 
{

   char *a;
   char reg[REGSTRING];

   /*
        any instruction may use the CC.  we loop through looking for
        uses, just to be sure we don't miss any
   */
   epos=ptr->val;
   while (fmatch(epos, cc_reg))
   {
        if(*epos!='=')
        {
             *does |= CCREAD;
             *blocks |= CCWRITE;
        }
        else
        {
             *does |= CCWRITE;
             *blocks |= CCREAD | CCWRITE;
        }
   }
   *does |= USEPC;
   if(is_pc(ptr->val))
   	*blocks |= USEPC|SETPC;

#ifdef SOFTFLOAT
        a=ptr->val;
	if(!gtmem(a) || (ptr->kind&CVTI))
        {
        /* non memory floating ops block the scratch and CC regs */
           while(a=gtrg(a))
           {
                if(regtype(a-2)!=RTREG)
                {
                        ptr->blocks |= USES_SCRATCH;
                        ptr->blocks |= CCREAD | CCWRITE;
                        ptr->does |= CCWRITE;
                        ptr->does|=USES_SCRATCH;
                        break;
                }
           }
        }
        a=ptr->val;
        while(a=gtrg(a))
        {
                if(scratchreg(a-2))
                        ptr->does|=USES_SCRATCH;

        }
#endif


   /*
      A call instruction places its return value into r0 immediately
   */
   if (ISBASE(ptr, CALLI))
   {
      makereg(reg, RTREG, 0);
      touchreg(sets2, reg);
      *does |= CCWRITE;
      *blocks |= CCREAD | CCWRITE;
   }

   /*
      Only use the items in the actual RTL, don't bother with the side effects
   */
   a = ptr->val;
   while (a = gtrg(a)) 
   {
      if (*a != '=')
	 touchreg(uses, a - 2);
   }

   /*
      Don't do the standard uses analysis on this RTL
   */
   return (TRUE);
}

/*
 * sched_adv - return the dynamic advantage of issuing this RTL at this time
 */
int sched_adv(register int adv, struct list * ptr, struct list * tail) 
{

   register struct list *prev;
 
   /*
      This function is extremely machine-dependent.  The idea is to give the
      instruction scheduler the perceived advantage of issuing an instruction
      based on the fact that the RTL pointed to by "tail" indicates which RTL
      was issued last.  This function can be used to encourage the instruction
      scheduler to issue particular instruction pair as far from each other as
      possible.
   */

   /*
      See if we can determine the instruction issued before the last
      instruction.
   */
   prev = prevexec((struct bblock *) NULL, tail->prev, ptr);

   /*
      Space out floating-point instructions
   */
   if (ISFP(ptr->kind)) {
      if (ISFP(tail->kind))
	 adv >>= 2;
      if (prev && ISFP(prev->kind))
	 adv >>= 1;
   }

   /*
      Avoid consecutive memory ops
   */
   if (ISBASE(ptr, STI) || ISBASE(ptr, LDI))
   {
      if (ISBASE(tail, STI) || ISBASE(tail, LDI))
	 adv >>= 2;
      if (prev && (ISBASE(prev, STI) || ISBASE(prev, LDI)))
	 adv >>= 1;
   }

   /*
      Make consecutive instructions independent
   */
   if (REGCOM(tail->sets, ptr->uses))
      adv >>= 2;
   if (prev && (REGCOM(prev->sets, ptr->uses)))
      adv >>= 1;

   return (adv);
}

/*
 * modify - modify RTL to make it issuable, if possible
 */
int modify(struct bblock * cblk, struct list * ptr, struct list * tail) {
/*ARGSUSED*/


   /*
      There is nothing that we can do on the SPARC (as far as I know) to make
      an instruction more or less issuable.
   */
   return (FALSE);
}

/*
 * noconflict - indicate whether the given RTL can be issued or not
 */
int noconflict(struct list * pptr, struct list * cptr) {
 /* pointer to previous RTL */
 /* pointer to current RTL */

   /*
	any instructions can safely (if not effeciently) 
	be executed immediately after any other 
   */
   return TRUE;
}

/*
 * create_noop - create the components that make up a no-op RTL
 */
void create_noop(char *rtl) {

    (void) strcpy(rtl, ";");
}

/*
 * setup_noop - set the kind and cost fields of a noop instruction
 */
void setup_noop(struct list * ptr) {

   ptr->kind = NOPI; 
   ptr->cost = 1;
   ptr->sets = regnew();
   ptr->uses = regnew();
}

#ifdef INTERACT
static int _unit_lat[INS_TYPES] = {1, 1, 12, 1, 1, 19, 1, 1, 1, 1};
static int _ins_lat[INS_TYPES] = {2, 4, 12, 2, 3, 20, 1, 1, 1, 1};
static int fu_units[FU_TYPES];
static int res_avail[MAXREGS];

/*
 * fu_unit_init - initialize the functional units
 */
void fu_unit_init()
{
  int i;
  for(i=0 ; i<FU_TYPES ; i++)
    fu_units[i] = 0;
  for(i=0 ; i<MAXREGS ; i++)
    res_avail[i] = 0;
}

/*
 * exec_ins - return NULL if all instructions have finished execution
 */
struct list *exec_ins(int cycle)
{
  int i;

  for(i=0 ; i<FU_TYPES ; i++){
    if(fu_units[i] > cycle)
      return TRUE;
  }

  return FALSE;
}


/*
 * issue_ins - an instruction can be issued if all its source operands are
 *             available and a functional unit is available
 */
int issue_ins(struct list *ptr, double *sup_cycle, int cycle)
{
  char *rtlin;
  int rnum, rtype, dst, src1, src2, src3, ins_type, fu_type;
  int flag1, flag2, flag3;

  /*find the source and destination registers*/
  dst = src1 = src2 = src3 = 0;
  flag1 = flag2 = flag3 = 0;
  rtlin = ptr->val;
  
  while(rtlin && *rtlin){
    if(isRegister(rtlin)){
      rnum = getRegNum(rtlin);
      rtlin = skiploc(rtlin);

      if((*rtlin == '=') || (*rtlin==']' && *(rtlin+1) && *(rtlin+1)=='=')){
	dst = rnum;
	rtype = getRegType(unskiploc(rtlin));
      }
      else{
	if(!flag1){
	  src1 = rnum;
	  flag1 = TRUE;
	}
	else if(!flag2){
	  src2 = rnum;
	  flag2 = TRUE;
	}
	else{
	  src3 = rnum;
	  flag3 = TRUE;
	}
      }
    }
    else
      rtlin++;
  }

  /*find the instruction class*/
  if(ISFP(ptr->kind) ||  rtype==FTREG || rtype==DTREG){
    ptr->kind &= 0xFF;
    if((ptr->kind == ADDI) || (ptr->kind == CMP) || (ptr->kind == CVTI)){
      ins_type = FPADD;
      fu_type = FU_FPALU;
    }
    else if(ptr->kind == MULI){
      ins_type = FPMUL;
      fu_type = FU_FPMULT;
    }
    else if((ptr->kind == DIVI) || (ptr->kind == DIVUI)){
      ins_type = FPDIV;
      fu_type = FU_FPMULT;
    }
    else if(ptr->kind == STI){
      ins_type = STORE;
      fu_type = FU_MEMPORT;
    }
    else if(ptr->kind == LDI){
      ins_type = LOAD;
      fu_type = FU_MEMPORT;
    }
    else{
      ins_type = FPUNKNOWN;
      fu_type = FU_FPALU;
    }
  }
  else{
    ptr->kind &= 0xFF;
    if(ptr->kind == MULI){
      ins_type = INTMUL;
      fu_type = FU_IMULT;
    }
    else if((ptr->kind == DIVI) || (ptr->kind == DIVUI)){
      ins_type = INTDIV;
      fu_type = FU_IMULT;
    }
    else if(ptr->kind == STI){
      ins_type = STORE;
      fu_type = FU_MEMPORT;
    }
    else if(ptr->kind == LDI){
      ins_type = LOAD;
      fu_type = FU_MEMPORT;
    }
    else{
      ins_type = INTALU;
      fu_type = FU_IALU;
    }
  }

  /*determine when this instruction can be issued based on when it's
    source operands would be available*/
  if(flag1 && res_avail[src1] > cycle)
    return FALSE;
  if(flag2 && res_avail[src2] > cycle)
    return FALSE;
  if(flag3 && res_avail[src3] > cycle)
    return FALSE;

  /*check for output dependence*/
  if(res_avail[dst] > cycle)
    return FALSE;
  
  /*issue if the functional unit is available*/
  if(fu_units[fu_type] <= cycle){

    fu_units[fu_type] = cycle + _unit_lat[ins_type];
    res_avail[dst] = cycle + _ins_lat[ins_type];
    
    /*if this instruction is a load/branch/jump, then add the estimated
      misprediction penalty in terms of cycles to sup_cycle*/
    /*if(ins_type == LOAD)
      *sup_cycle += 0.5;
    if(is_pc(ptr->val))
    *sup_cycle += 0.01;*/
    
    return TRUE;
  }
    
  return FALSE;
}
 

/*
 * ins_latency - determine when this instruction can be issued based on
 *               instruction dependencies and resource conflicts. Return the
 *               additional number of cycles from the curr_cycle when the
 *               result would be available
 */
int ins_latency(struct list *ptr, int cycle, double *sup_cycle,
		int reg_avail[], int unit_avail[])
{

  char *rtlin;
  int rnum, rtype, dst, src1, src2, ins_type;

  /*find the source and destination registers*/
  dst = src1 = src2 = 0;
  rtlin = ptr->val;
  
  while(rtlin && *rtlin){
    if(isRegister(rtlin)){
      rnum = getRegNum(rtlin);
      rtlin = skiploc(rtlin);

      if(*rtlin == '='){
	dst = rnum;
	rtype = getRegType(unskiploc(rtlin));
      }
      else{
	if(!src1)
	  src1 = rnum;
	else
	  src2 = rnum;
      }
    }
    else
      rtlin++;
  }

  /*find the instruction class*/
  if(ISFP(ptr->kind) ||  rtype==FTREG || rtype==DTREG){
    ptr->kind &= 0xFF;
    if((ptr->kind == ADDI) || (ptr->kind == CMP) || (ptr->kind == CVTI))
      ins_type = FPADD;
    else if(ptr->kind == MULI)
      ins_type = FPMUL;
    else if((ptr->kind == DIVI) || (ptr->kind == DIVUI))
      ins_type = FPDIV;
    else if(ptr->kind == STI)
      ins_type = STORE;
    else if(ptr->kind == LDI)
      ins_type = LOAD;
    else
      ins_type = FPUNKNOWN;
  }
  else{
    ptr->kind &= 0xFF;
    if(ptr->kind == MULI)
      ins_type = INTMUL;
    else if((ptr->kind == DIVI) || (ptr->kind == DIVUI))
      ins_type = INTDIV;
    else if(ptr->kind == STI)
      ins_type = STORE;
    else if(ptr->kind == LDI)
      ins_type = LOAD;
    else
      ins_type = INTALU;
  }

  /*determine when this instruction can be issued based on when it's
    source operands would be available*/
  if(src1 && reg_avail[src1] > cycle)
    cycle = reg_avail[src1];
  if(src2 && reg_avail[src2] > cycle)
    cycle = reg_avail[src2];
  
  /*account for resource conflict*/
  if(unit_avail[ins_type] > cycle)
    cycle = unit_avail[ins_type];
  
  /*the divide engines are not pipelined*/
  unit_avail[ins_type] = cycle + _unit_lat[ins_type];

  /*add the number of cycles required for this instruction execution to
    determine when the destination register would be available*/
  cycle += _ins_lat[ins_type];
  reg_avail[dst] = cycle;

  /*if this instruction is a load/branch/jump, then add the estimated
    misprediction penalty in terms of cycles to sup_cycle*/
  if(ins_type == LOAD)
    *sup_cycle += 0.1;
  if(is_pc(ptr->val))
    *sup_cycle += 0.01;

  return cycle;
}
#endif /*INTERACT*/
