static char rcsid[] = "$Id: arena.c,v 1.4 2003/12/16 22:00:43 clc5q Exp $";
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include "vpo.h"
#include "arena.h"

#define THRESHOLD 10
struct Arena_T {
	Arena_T prev;
	char *avail;
	char *limit;
};
union align {
#ifdef MAXALIGN
	char pad[MAXALIGN];
#else
	int i;
	long l;
	long *lp;
	void *p;
	void (*fp)(void);
	float f;
	double d;
	long double ld;
#endif
};
union header {
	struct Arena_T b;
	union align a;
};
static Arena_T freechunks=NULL;
static int nfree;
Arena_T Arena_new(void) {
	Arena_T arena = malloc(sizeof (*arena));
	if (arena == NULL)
	{
		fprintf(stderr,"Null arena badness\n");
		exit(1);
	}
	arena->prev = NULL;
	arena->limit = arena->avail = NULL;
	return arena;
}
void Arena_dispose(Arena_T *ap) {
	assert(ap && *ap);
	Arena_free(*ap);
	free(*ap);
	*ap = NULL;
}
void *Arena_alloc(Arena_T arena, long nbytes)
{
	assert(arena);
	assert(nbytes > 0);
	nbytes = ((nbytes + sizeof (union align) - 1)/
		(sizeof (union align)))*(sizeof (union align));
	while (nbytes > arena->limit - arena->avail) {
		Arena_T ptr;
		char *limit;
		if ((ptr = freechunks) != NULL) {
			freechunks = freechunks->prev;
			nfree--;
			limit = ptr->limit;
		} else {
			long m = sizeof (union header) + nbytes + 10*1024;
			ptr = malloc(m);
			if (ptr == NULL)
			{
				fprintf(stderr,"virtual memory exhausted\n");
				exit(1);
			}
			limit = (char *)ptr + m;
		}
		*ptr = *arena;
		arena->avail = (char *)((union header *)ptr + 1);
		arena->limit = limit;
		arena->prev  = ptr;
	}
	arena->avail += nbytes;
	return arena->avail - nbytes;
}
void *Arena_calloc(Arena_T arena, long count, long nbytes)
{
	void *ptr;
	assert(count > 0);
	ptr = Arena_alloc(arena, count*nbytes);
	memset(ptr, '\0', count*nbytes);
	return ptr;
}

void Arena_free(Arena_T arena) {
   /* We have bugs in our code that are hard to debug but
    * go away when we stop freeing our temporary arenas.
    * Until we have time to debug them, we will just return.
    * If certain optimizations need more memory, they can
    * disable this premature return, but not all test cases
    * and benchmarks will pass.
    */
   if (!swq && !(swz && swpg && swpw))  /* Global data placement. */
      return;

        assert(arena);
	while (arena->prev) {
		struct Arena_T tmp = *arena->prev;
		if (nfree < THRESHOLD) {
			arena->prev->prev = freechunks;
			freechunks = arena->prev;
			nfree++;
			freechunks->limit = arena->limit;
		} else
			free(arena->prev);
		*arena = tmp;
	}
	assert(arena->limit == NULL);
	assert(arena->avail == NULL);
}

void debug_Arena()
{
  Arena_T ptr;

  for(ptr=freechunks ; ptr ; ptr=ptr->prev)
    fprintf(stderr, "%x ", ptr);
  fprintf(stderr, "\n");
}
