/*trans.c: functions for translating an assembly instruction into an encoded
  RTL. The main function is Translate(source, result), which translates
  source from assembly into encoded RTL and store the result in result.
  
  Author: Baosheng Cai
  Date: 10/04/99 */
#ifdef INTERACT

#define _C_POSIX_SOURCE 2
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>

#include "assem_trans.h"
/*test register map*/
#define MAXLINE 50

void send_error(char *err_msg);
char *find_global(char *global_name);
char *find_local(char *global_name);

/*Translate: the function to be called.
  source: the source assembly line to be translated
  cexresult: the cex format result */
int Translate(char *source, char *cexresult)
{
   STATEMENT sta1;

   /*Initialize statements, and store the information to stat1 structure*/
   InitStatement(&sta1, source);
   /*Translate the statement to cex format*/
   Trans(cexresult, &sta1);
   /*if the translation failed, it prints out error message*/
   
   if (GloError){
      fprintf(stderr, "%s\n", ErrMsg);
      return GloError;
   }
   
   return (0);  
}

/*Initialize the STATEMENT structure using the information in src */
void InitStatement(STATEMENT *sta1, char *src)
{
   char op[MAXOP_LEN];
   int i;
    /*get the operation string->op*/
   GetOp(op, sta1->oprnds, src);
   /*assume that bl,a=bl*/
   CutFix(op);
   /*compare the operator string to findout which operator it is*/
   for (i=0; i<100; i++){
      if (strncmp(op, operators[i], 10)==0)
         break;      
   }
   if (i>=100)
      ERR_RET("No Such instruction.\n");
   sta1->op_tor=i;
   sta1->n_oprnd=GetOprndNo(i);    
}

/*deal with instructions like bl,a, assume that they are
 the same as bl */
void CutFix(char *op)
{
   char *p;
   for (p=op;*p&&(*p!=','); p++);
   if (*p==','){
      if (strcmp(p,",a")==0)
         *p='\0';
   }
}


/*sta1: the statement to be translated
 result: translated result, in cex format*/
/*sta1: the statement containing source instruction*/
void Trans(char *result, STATEMENT *sta1)
{
   char s_oprnd[4][MAXOP_LEN];
   int i, n_read;
   char errmsg[100];
   /*read register name from input string*/
   /*get the number of operands*/
   if (sta1->oprnds[0]!='\0')
      n_read=sscanf(sta1->oprnds, "%s %s %s %s",s_oprnd[0],
                   s_oprnd[1], s_oprnd[2], s_oprnd[3]);
   else
      n_read=0;
   if (n_read!=sta1->n_oprnd){
      sprintf(errmsg,"%d operands required.\n", sta1->n_oprnd);
      ERR_RET(errmsg);
   }
   for (i=0;   i<sta1->n_oprnd; i++){
      /*Parse each operand*/
      Parse(s_oprnd[i], sta1->operand+i);         
   }
   
#ifdef DEBUG
   /*the SynAnalysis may not be needed. But it is useful in debugging*/
   SynAnalysis(sta1);
#endif
   /*Convert register type into rtl type*/
   /*from r,sw ... into 1, 2, 3... */
   CexGen(result,sta1); 
}

#ifdef DEBUG
/*Analysis the syntax of this statement before convertion*/
void SynAnalysis(STATEMENT *sta1)
{
   char errmsg[100];
   switch(sta1->op_tor){
   case ADD:
   case ADDCC:
   case SUB:
   case SUBCC:
   case AND:
   case ANDCC:
   case OR:
   case ORCC:
   case ORN:
   case ORNCC:
   case XOR:
   case XORCC:
   case XNOR:
   case XNORCC:
   case RESTORE:
   case SAVE:
   case SLL:
   case SRL:
   case SRA:
      if (sta1->operand[0].operand_type!=REG){
         sprintf(errmsg, "The first operand must be register.\n");
         ERR_RET(errmsg);
      }       
      if (sta1->operand[2].operand_type!=REG){
         sprintf(errmsg, "The third operand must be register.\n");
         ERR_RET(errmsg);
      }
      break;
   case BNE:
   case BE:
   case BG:
   case BLE:
   case BGE:
   case BL:
   case BGU:
   case BLEU:
   case BA:
   case FBG:
   case FBL:
   case FBNE:
   case FBE:
   case FBGE:
   case FBLE:
      if (sta1->operand[0].operand_type!=LABEL){
         sprintf(errmsg, "Operand must be a label.\n");
         ERR_RET(errmsg);
      }
      break;
   case LDSB:
   case LDSH:
   case LDUB:
   case LDUH:
   case LD:
   case LDD:      
      if ((sta1->operand[0].operand_type!=EXPR_ADDR) &&\
          (sta1->operand[0].operand_type!=REG_ADDR)   ) {
         sprintf(errmsg, "The first operand must be address.\n");
         ERR_RET(errmsg);
      }
      if (sta1->operand[1].operand_type!=REG) {
         sprintf(errmsg, "The second operand must be a register.\n");
         ERR_RET(errmsg);
      }
      
      break;
   case STB:
   case STH:
   case ST:
   case STD:
      if ((sta1->operand[1].operand_type!=EXPR_ADDR) &&\
          (sta1->operand[1].operand_type!=REG_ADDR) ){
         sprintf(errmsg, "The second operand must be address.\n");
         ERR_RET(errmsg);
      }   
      if (sta1->operand[0].operand_type!=REG) {
         sprintf(errmsg, "The first operand must be a register.\n");
         ERR_RET(errmsg);
      }
      
      break;
   case FITOS:
   case FSTOI:
   case FDTOI:
   case FDTOS:
   case FMOVS:
   case FNEGS:
   case FCMPS:
   case FCMPD:
   case FITOD:
   case FSTOD:
      if ((sta1->operand[1].operand_type!=REG) ||\
          (sta1->operand[1].register_type!=F)){
         sprintf(errmsg, "The second operand must be a floating point register.\n");
         ERR_RET(errmsg);
      }
      
      if ((sta1->operand[0].operand_type!=REG) ||\
          (sta1->operand[0].register_type!=F)){
         sprintf(errmsg, "The first operand must be a floating point register.\n");
         ERR_RET(errmsg);
      }
      break;
   case FADDD:
   case FMULD:
   case FSUBD:
   case FDIVD:
   case FADDS:
   case FSUBS:
   case FMULS:
   case FDIVS:
      if ((sta1->operand[1].operand_type!=REG) ||\
          (sta1->operand[1].register_type!=F)){
         sprintf(errmsg, "The second operand must be a floating point register.\n");
         ERR_RET(errmsg);
      }
         
      if ((sta1->operand[0].operand_type!=REG) ||\
          (sta1->operand[0].register_type!=F)){
         sprintf(errmsg, "The first operand must be a floating point register.\n");
         ERR_RET(errmsg);
      }
         
      if ((sta1->operand[2].operand_type!=REG) ||\
          (sta1->operand[2].register_type!=F)   ){
         sprintf(errmsg, "The Third operand must be a floating point register.\n");
         ERR_RET(errmsg);
      }
      break;
      
   case CMP:
      if (sta1->operand[0].operand_type!=REG){
         sprintf(errmsg, "The first operand must be a register.\n");
         ERR_RET(errmsg);
      }
         
      if ((sta1->operand[1].operand_type!=REG) && \
          (sta1->operand[1].operand_type!=NUM)){
         sprintf(errmsg, "The second operand must be a register or number.\n");
         ERR_RET(errmsg);
      }
      
      break;
   case MOV:
      if (sta1->operand[1].operand_type!=REG){
         sprintf(errmsg, "The second operand must be a register.\n");
         ERR_RET(errmsg);
      }         
    if ((sta1->operand[0].operand_type!=REG) && \
         (sta1->operand[0].operand_type!=NUM)){
       sprintf(errmsg, "The first operand must be a register or number.\n");
       ERR_RET(errmsg);
      }
      break;
   case SETHI:
      if (sta1->operand[0].operand_type!=HI_VALUE){
         sprintf(errmsg, "The first operand must be %%hi(value).\n");
         ERR_RET(errmsg);
      }
      if (sta1->operand[1].operand_type!=REG){
         sprintf(errmsg, "The second operand must be a register.\n");
         ERR_RET(errmsg);
      }
      
      break;
      
   case RET:
   case RETL:
   case NOP:
      break;
   case CALL:
      if ((sta1->operand[0].operand_type!=LABEL) && \
          (sta1->operand[0].operand_type!=REG)    &&
          (sta1->operand[0].operand_type!=NUM)){
         sprintf(errmsg, "The first operand must be a register or a number or a label.\n");
         ERR_RET(errmsg);
      }
      break;
   default:
      ERR_RET("No such instruction.\n");
      break;
   }
   /*another switch: some of the instructions need even registers.*/
   switch(sta1->op_tor){
   case LDD:
   case FITOD:
   case FSTOD:
      if ((sta1->operand[1].operand_no % 2)!=0){
         sprintf(errmsg, "The second register number must be even.\n");
         ERR_RET(errmsg);
      }
      break;
   case STD:
   case FDTOI:
   case FDTOS:
      if ((sta1->operand[0].operand_no % 2)!=0){
         sprintf(errmsg, "The first register number must be even.\n");
         ERR_RET(errmsg);
      }
      break;
   case FADDD:
   case FSUBD:
   case FMULD:
   case FDIVD:      
      /*The register number must be even*/
      if ((sta1->operand[0].operand_no % 2)!=0){
         sprintf(errmsg, "The first register number must be even.\n");
         ERR_RET(errmsg);
      }
      /*The register number must be even*/
      if ((sta1->operand[1].operand_no % 2)!=0){
         sprintf(errmsg, "The second register number must be even.\n");
         ERR_RET(errmsg);
      }
       
      /*The register number must be even*/
      if ((sta1->operand[0].operand_no % 2)!=0){
         sprintf(errmsg, "The third register number must be even.\n");
         ERR_RET(errmsg);
      }
    
      break;
      break;
   case FCMPD:
      /*The register number must be even*/
      if ((sta1->operand[0].operand_no % 2)!=0){
         sprintf(errmsg, "The first register number must be even.\n");
         ERR_RET(errmsg);
      }
      /*The register number must be even*/
      if ((sta1->operand[1].operand_no % 2)!=0){
         sprintf(errmsg, "The second register number must be even.\n");
         ERR_RET(errmsg);
      }
      break; 
   default:
      break;    
   }   
}
#endif   /*end if DEBUG*/

/*Generate the cex result*/
/*input: statem
   oupput: result*/
void CexGen( char *result, STATEMENT *statem)
{
   /*first assume that the operate type is add*/
   int i;
   switch (statem->op_tor) {
   case ADD:
      sprintf(result, "%s=%s+%s;", statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case ADDCC:
      sprintf(result, "IC=%s+%s?0;%s=%s+%s;",statem->operand[0].s_oprnd, \
              statem->operand[1].s_oprnd,statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case SUB:
      sprintf(result, "%s=%s-%s;", statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
    
      break;
   case SUBCC:
      sprintf(result, "IC=%s-%s?0;%s=%s-%s;",statem->operand[0].s_oprnd, \
              statem->operand[1].s_oprnd,statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case AND:
      sprintf(result, "%s=%s&%s;", statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
    
      break;
   case ANDCC:
      sprintf(result, "IC=%s&%s?0;%s=%s&%s;",statem->operand[0].s_oprnd, \
              statem->operand[1].s_oprnd,statem->operand[2].s_oprnd,\
              statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case OR:
      sprintf(result, "%s=%s|%s;", statem->operand[2].s_oprnd,\
              statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      
      break;
   case ORCC:
      sprintf(result, "IC=%s|%s?0;%s=%s|%s;",statem->operand[0].s_oprnd, \
              statem->operand[1].s_oprnd,statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case ORN:
      sprintf(result, "%s=%so%s;", statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case ORNCC:
      sprintf(result, "IC=%so%s?0;%s=%so%s;",statem->operand[0].s_oprnd, \
              statem->operand[1].s_oprnd,statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case XOR:
      sprintf(result, "%s=%s^%s;", statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
    
      break;
   case XORCC:
      sprintf(result, "IC=%s^%s?0;%s=%s^%s;",statem->operand[0].s_oprnd, \
              statem->operand[1].s_oprnd,statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case XNOR:
      sprintf(result, "%s=%sx%s;", statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      
      break;
   case XNORCC:
      sprintf(result, "IC=%sx%s?0;%s=%sx%s;",statem->operand[0].s_oprnd, \
              statem->operand[1].s_oprnd,statem->operand[2].s_oprnd,\
              statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case SLL:
      sprintf(result, "%s=%s{%s;", statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case SRL:
      sprintf(result, "%s=%s\"%s;", statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case SRA:
      sprintf(result, "%s=%s}%s;", statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case BNE:
      sprintf(result, "PC=IC!0,%s;", statem->operand[0].s_oprnd);
      break;
   case BE:
      sprintf(result, "PC=IC:0,%s;", statem->operand[0].s_oprnd);
      break;
   case BG:
      sprintf(result, "PC=IC>0,%s;", statem->operand[0].s_oprnd);
      break;
   case BLE:
      sprintf(result, "PC=IC'0,%s;", statem->operand[0].s_oprnd);
      break;
   case BGE:
      sprintf(result, "PC=IC`0,%s;", statem->operand[0].s_oprnd);
      break;
   case BL:
      sprintf(result, "PC=IC<0,%s;", statem->operand[0].s_oprnd);
      break;
   case BGU:
      sprintf(result, "PC=ICh0,%s;", statem->operand[0].s_oprnd);
      break;
   case BLEU:
      sprintf(result, "PC=ICs0,%s;", statem->operand[0].s_oprnd);
      break;
   case BA:
      sprintf(result, "PC=%s;", statem->operand[0].s_oprnd);
      break;
   case FBG:
      sprintf(result, "PC=FC>0,%s;", statem->operand[0].s_oprnd);
      break;
   case FBL:
      sprintf(result, "PC=FC<0,%s;", statem->operand[0].s_oprnd);
      break;
   case FBNE:
      sprintf(result, "PC=FC!0,%s;", statem->operand[0].s_oprnd);
      break;
   case FBE:
      sprintf(result, "PC=FC:0,%s;", statem->operand[0].s_oprnd);
      break;
   case FBGE:
      sprintf(result, "PC=FC`0,%s;", statem->operand[0].s_oprnd);
      break;
   case FBLE:
      sprintf(result, "PC=FC'0,%s;", statem->operand[0].s_oprnd);
      break;
   case LDSB:
      sprintf(result, "%s=(B[%s]{24)}24;", statem->operand[1].s_oprnd, \
              statem->operand[0].s_oprnd);
      break;
   case LDSH:
      sprintf(result, "%s=(W[%s]{16)}16;",statem->operand[1].s_oprnd, \
              statem->operand[0].s_oprnd);
      break;
   case LDUB:
      sprintf(result, "%s=B[%s]&255;", statem->operand[1].s_oprnd, \
                 statem->operand[0].s_oprnd);
      break;
   case LDUH:/*don't know if it is right (&65535). */
      sprintf(result, "%s=W[%s]&65535;",statem->operand[1].s_oprnd, \
                 statem->operand[0].s_oprnd);
      break;
   case LD:
      sprintf(result, "%s=R[%s];",statem->operand[1].s_oprnd, \
              statem->operand[0].s_oprnd);
      break;
   case LDD:
      ConvOprndD(&(statem->operand[1]));
      sprintf(result, "%s=D[%s];", statem->operand[1].s_oprnd, \
              statem->operand[0].s_oprnd);
      break;
   case STB:
      sprintf(result, "B[%s]=%s;",statem->operand[1].s_oprnd,\
              statem->operand[0].s_oprnd);
      break;
   case STH:
      sprintf(result, "W[%s]=%s;",statem->operand[1].s_oprnd,\
            statem->operand[0].s_oprnd);
      break;
   case ST:
      sprintf(result, "R[%s]=%s;",statem->operand[1].s_oprnd,\
              statem->operand[0].s_oprnd);
      break;
   case STD:
      ConvOprndD(&(statem->operand[0]));
      sprintf(result, "D[%s]=%s;",statem->operand[1].s_oprnd,\
              statem->operand[0].s_oprnd);
      break;
   case FADDS:
      sprintf(result, "%s=%s+%s;", statem->operand[2].s_oprnd,\
              statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case FADDD:
      for (i=0; i<3; i++)
         ConvOprndD(&(statem->operand[i]));
      sprintf(result, "%s=%s+%s;", statem->operand[2].s_oprnd,\
              statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
       break;
   case FSUBS:
      sprintf(result, "%s=%s-%s;", statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case FSUBD:
      for (i=0; i<3; i++)
         ConvOprndD(&(statem->operand[i]));
      sprintf(result, "%s=%s-%s;", statem->operand[2].s_oprnd,\
               statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case FMULS:
      sprintf(result, "%s=%s*%s;", statem->operand[2].s_oprnd,\
              statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case FMULD:
      for (i=0; i<3; i++)
         ConvOprndD(&(statem->operand[i]));
      sprintf(result, "%s=%s*%s;", statem->operand[2].s_oprnd,\
              statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case FDIVS:
      sprintf(result, "%s=%s/%s;", statem->operand[2].s_oprnd,\
              statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case FDIVD:
      for (i=0; i<3; i++)
         ConvOprndD(&(statem->operand[i]));
      sprintf(result, "%s=%s/%s;", statem->operand[2].s_oprnd,\
              statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case FITOS:
      sprintf(result, "%s=FI[%s];",   \
              statem->operand[1].s_oprnd, statem->operand[0].s_oprnd);
      break;
   case FITOD:
      ConvOprndD(&(statem->operand[1]));
      sprintf(result, "%s=FI[%s];",   \
              statem->operand[1].s_oprnd, statem->operand[0].s_oprnd);
      break;
   case FSTOI:
      sprintf(result, "%s=RZ[%s];",   \
              statem->operand[1].s_oprnd, statem->operand[0].s_oprnd);
      break;
   case FDTOI:
      ConvOprndD(&(statem->operand[0]));
      sprintf(result, "%s=RZ[%s];",   \
              statem->operand[1].s_oprnd, statem->operand[0].s_oprnd);
      break;
   case FSTOD:
      ConvOprndD(&(statem->operand[1]));
      sprintf(result, "%s=DC[%s];",   \
              statem->operand[1].s_oprnd, statem->operand[0].s_oprnd);
      break;
   case FDTOS:
      ConvOprndD(&(statem->operand[0]));
      sprintf(result, "%s=SC[%s];",   \
              statem->operand[1].s_oprnd, statem->operand[0].s_oprnd);
      break;
   case FMOVS:
      sprintf(result, "%s=%s;",   \
              statem->operand[1].s_oprnd, statem->operand[0].s_oprnd);
      break;
   case FNEGS:
      sprintf(result, "%s=-%s;",   \
              statem->operand[1].s_oprnd, statem->operand[0].s_oprnd);
      break;
   case FCMPS:
      sprintf(result, "FC=%s?%s;",   \
              statem->operand[1].s_oprnd, statem->operand[0].s_oprnd);
      break;
   case FCMPD:
      ConvOprndD(&(statem->operand[0]));
      ConvOprndD(&(statem->operand[1]));      
      sprintf(result, "FC=%s?%s;",   \
              statem->operand[1].s_oprnd, statem->operand[0].s_oprnd);
      break;
   case CMP:
      sprintf(result, "IC=%s?%s;",   \
              statem->operand[1].s_oprnd, statem->operand[0].s_oprnd);
      break;
   case MOV:
      sprintf(result, "%s=%s;",   \
              statem->operand[1].s_oprnd, statem->operand[0].s_oprnd);
      break;
   case SETHI:/*to be corrected*/
      sprintf(result, "%s=%s;",   \
              statem->operand[1].s_oprnd, statem->operand[0].s_oprnd);
      break;
      /*restore and save are not quite right*/
   case RESTORE:
      /*deal with the situation when it is "restore ?, %g0, ?"*/
      ResSaveSpec(statem);
      if (*(statem->operand[1].s_oprnd)=='\0')
         sprintf(result, "%s=RS[%s];", statem->operand[2].s_oprnd,\
                 statem->operand[0].s_oprnd);
      
      else if (*(statem->operand[0].s_oprnd)=='\0')
         sprintf(result, "%s=RS[%s];", statem->operand[2].s_oprnd,\
                 statem->operand[1].s_oprnd);      
      else
         sprintf(result, "%s=RS[%s+%s];", statem->operand[2].s_oprnd,\
                 statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
    
      break;
   case SAVE:
      ResSaveSpec(statem);
      if (*(statem->operand[1].s_oprnd)=='\0')
         sprintf(result, "%s=SV[%s];", statem->operand[2].s_oprnd,\
                 statem->operand[0].s_oprnd);
      
      else if (*(statem->operand[0].s_oprnd)=='\0')
         sprintf(result, "%s=SV[%s];", statem->operand[2].s_oprnd,\
                 statem->operand[1].s_oprnd);      
      else
         sprintf(result, "%s=SV[%s+%s];", statem->operand[2].s_oprnd,\
                 statem->operand[0].s_oprnd, statem->operand[1].s_oprnd);
      break;
   case RET:
      sprintf(result, "PC=RT;");
      break;
   case RETL:
      sprintf(result, "PC=RT;");
      break;
   case CALL:
      if (statem->operand[0].operand_type==REG)
         sprintf(result, "ST=%s;", statem->operand[0].s_oprnd);
      else
         sprintf(result, "ST=HI[%s]+LO[%s];", \
                 statem->operand[0].s_oprnd, statem->operand[0].s_oprnd);
         
      break;
   case NOP:
      sprintf(result, "NL=NL;");
      break;
    /*to be continued*/
   default:
      printf("not finished\n");
      break;
   }
}

/*separate the statement to 2 part:the operator and operand part
 input: string src
 ouput: string op, the operator part
            string oprnd, the operand part*/
void GetOp(char *op,char *oprnd, char *src)
{
   int i, j, opstop;
   char c;
   for (i=0;i<MAXOP_LEN;i++){
      if (src[i]==' '||src[i]=='\t'||src[i]=='\0'){
         break;
      }
      *(op+i)=*(src+i);
   }
   *(op+i)='\0';
   opstop=i;
   /*get the operands string*/
   oprnd[0]='\0';
   for (i=opstop, j=0; *(src+i)&&i<(MAXOPRND_LEN-opstop); i++){
      c=*(src+i);
      if ((c!=' ')&&(c!='\t')){
         *(oprnd+j)=c;
         j++;
      }/*end if*/
   }/*end for*/
   *(oprnd+j)='\0';
   /*replace all the , s to space*/
   for (i=0; *(oprnd+i)&&i<MAXOPRND_LEN; i++){
      if (*(oprnd+i)==',')
         *(oprnd+i)=' ';
   }
   /*if the op is restore or save, then we pre-add
    operands to them*/
   if (((strcmp (op, "restore")==0 ) || \
        (strcmp (op, "save")==0)) &&(oprnd[0]=='\0'))
      sprintf(oprnd, "%s", "%g0 %g0 %g0");
   
}
/*substitue the () with space*/
void SubtoSpace(char *oprnd)
{
   int i;
   /*replace all the () s to space*/
   for (i=0; *(oprnd+i)&&i<MAXOPRND_LEN; i++){
      if ((*(oprnd+i)=='(')||(*(oprnd+i)==')'))
         *(oprnd+i)=' ';
   }   
}
/*"It" means "(" or ')'*/
void IgnoreIt(char *oprnd)
{
   char tmpstr[20];
   SubtoSpace(oprnd);
   sscanf(oprnd, " %s ", tmpstr);
   strcpy(oprnd, tmpstr);
}
/*Assert if character c is a separator or not*/
int IsSeparator(char c)
{
   /*I don't count the % as a separtor*/
   if ((c=='+')||(c=='-')|| \
       (c=='*')||(c=='/')){
      return (1);
   }
   return(0);
   
}

/*srcstr will be changed to the current position*/
char* GetSepStr(char *srcstr, char *rslstr, char *c)
{
   int i;
   for (i=0; i<MAXOPRND_LEN; i++, srcstr++){
      if ((!*srcstr)||IsSeparator(*srcstr)){
         *(rslstr+i)='\0';
         /*str points to the next char of '+'*/
         *c=(*srcstr);
         srcstr++;
         break;
      }
      *(rslstr+i)=*srcstr;
   }
   return(srcstr);
   
}
/*examine if a string has a separator or not.*/
int HaveSeparator(char *str)
{
   char *ptr=str;
   int n=0;
   for ( ; *ptr; ptr++){
      if (IsSeparator(*ptr)){
         n++;
      }
   }
   return (n);
}
/*separate the situations where there is a separator
 and where there is not*/
OPRND_TYPE ParseExpression(char *s_oprnd, OPRND *oprnd)
{
   int sepno, i;
   char tmpc;
   char *tmpptr=s_oprnd;
   /*scan if there is isolators, then divide them. 
      and then call parse again.*/
   sepno=HaveSeparator(s_oprnd);
   if (sepno>0){
      char tmpsOprnd[MAXOPRND_LEN];
      OPRND   oprnd_tmp;
      /*set the original string length to zero,
       in order to strcat*/
      oprnd_tmp.s_oprnd[0]='\0';
      oprnd->s_oprnd[0]='\0';
      for (i=0;i<sepno+1; i++){      
         tmpptr=GetSepStr(tmpptr, tmpsOprnd, &tmpc);
         Parse(tmpsOprnd, &oprnd_tmp);         
         strcat(oprnd->s_oprnd, oprnd_tmp.s_oprnd);
         strcat (oprnd->s_oprnd, &tmpc);
       
       }
      oprnd->operand_type=EXPRESSION;
      return (oprnd->operand_type);
   }
   else{
      return (Parse(s_oprnd, oprnd));
   }
}

/*Parse: used when met with %hi or %lo,
 at most one relocatable symbol is allowed.
no registers*/

OPRND_TYPE ParseValue(char *s_oprnd, OPRND *oprnd)
{
   int sepno, i;
   char tmpc;
   char *tmpptr=s_oprnd;
   int nVar=0;
   /*scan if there is isolators, then divide them. 
      and then call parse again.*/
   sepno=HaveSeparator(s_oprnd);
   if (sepno>0){
      char tmpsOprnd[MAXOPRND_LEN];
      OPRND   oprnd_tmp;
      /*set the original string length to zero,
       in order to strcat*/
      oprnd_tmp.s_oprnd[0]='\0';
      oprnd->s_oprnd[0]='\0';
      for (i=0;i<sepno+1; i++){      
         tmpptr=GetSepStr(tmpptr, tmpsOprnd, &tmpc);
         Parse(tmpsOprnd, &oprnd_tmp);
         if (oprnd_tmp.operand_type==REG)
            ERR_RET("No registers in value\n");
         if ((oprnd_tmp.operand_type==LABEL) ||
             (oprnd_tmp.operand_type==GLOVAR)||
             (oprnd_tmp.operand_type==LOCVAR))
            nVar++;
         strcat(oprnd->s_oprnd, oprnd_tmp.s_oprnd);
         strcat (oprnd->s_oprnd, &tmpc);       
      }

      if (nVar>1)
         ERR_RET("More than one relocatable symbol.\n");         
      oprnd->operand_type=VALUE;
      return (oprnd->operand_type);
   }
   else{
      return (Parse(s_oprnd, oprnd));
   }
}

/*Parse the operands without separators
 input: string s_oprnd,
 output: oprnd*/
OPRND_TYPE Parse(char *s_oprnd, OPRND *oprnd)
{
   char s_type[3], c;
   int ntype, no;
   OPRND sub_oprnd;
   char sub_str[MAXOPRND_LEN];
   /*code pointer used for a local or global variable.*/
   char *var_code;
    
   /*if the operand is immediate number*/
   if (IsNumber(s_oprnd)){
      oprnd->operand_type=NUM;
      strcpy(oprnd->s_oprnd, s_oprnd);
   }
   else if (*s_oprnd=='('){
      /*Ignore the ()s */
      IgnoreIt(s_oprnd);
      Parse(s_oprnd, oprnd);
   }
   else if (strncmp(s_oprnd, "%lo", 3)==0){
      SubtoSpace(s_oprnd);
      sscanf(s_oprnd, "%%lo %s",sub_str);
      ParseValue (sub_str, &sub_oprnd);
      sprintf(oprnd->s_oprnd, "LO[%s]", sub_oprnd.s_oprnd);
      oprnd->operand_type=LO_VALUE;      
   }      
   else if (strncmp(s_oprnd, "%hi", 3)==0){
      SubtoSpace(s_oprnd);
      sscanf(s_oprnd, "%%hi %s",sub_str);
      ParseValue(sub_str, &sub_oprnd);
      sprintf(oprnd->s_oprnd, "HI[%s]", sub_oprnd.s_oprnd);
      oprnd->operand_type=HI_VALUE;      
   }      
   else{
      switch (*s_oprnd){
      case '%':
         c=*(s_oprnd+2);
         /*if the register is like %r1,r2, r3, o1,... */
         if (isdigit((int)c)){
            char err_detect='\0';
            sscanf(s_oprnd, "%%%c%d%c",s_type, &(oprnd->operand_no), &err_detect);
            if (err_detect){
               ERR_RET("Syntax error.\n");
            }
            s_type[1]='\0';
         }
         /*if the register is like %sp, %fp, ...*/
         else{
            sscanf(s_oprnd, "%%%s", s_type);
         }
         /*convert the type in assembly (r, sp, fp...) into
           rtl type, like 1, 2, 3, ...*/
         /*TypeConv(&(oprnd->register_type), s_type);*/
         TypeConv(oprnd, s_type);
         ntype=oprnd->register_type;
         no=oprnd->operand_no;
         TREG(ntype, no);
         sprintf(oprnd->s_oprnd, "%c%c", ntype, no);
         oprnd->operand_type=REG;
         break;
      case '[':
         ExtractAddr(s_oprnd, sub_str);
         if (ParseExpression(sub_str, &sub_oprnd)==REG){
            strcpy(oprnd->s_oprnd, sub_oprnd.s_oprnd);
            oprnd->operand_type=REG_ADDR;
         }
         else{
            strcpy(oprnd->s_oprnd, sub_oprnd.s_oprnd);
            oprnd->operand_type=EXPR_ADDR;
         }
         break;
      default:
         if (var_code=find_local(s_oprnd)){
            oprnd->operand_type=LOCVAR;
            strcpy(oprnd->s_oprnd, var_code);
         }
         else if (var_code = find_global(s_oprnd)){
            oprnd->operand_type=GLOVAR;
            strcpy(oprnd->s_oprnd, var_code);
         }
         else {
            oprnd->operand_type=LABEL;
            strcpy(oprnd->s_oprnd, s_oprnd);
         }
         break;
      }
   }
   return(oprnd->operand_type);
 
}
/*get dst from the form: src="[dst]"*/
void ExtractAddr(char *src, char *dst)
{
   int i;
   char *ptr;
   ptr=src+1;
   for (i=0; *ptr && (*ptr)!=']'; ptr++, i++){
      *(dst+i)=*ptr;
   }
   *(dst+i)='\0';
}

/*convert the type in assembly (r, sp, fp...) into
 rtl type, like 1, 2, 3, ...*/
void TypeConv(OPRND *oprnd, char *reg_typec)
{
   char errmsg[100];
   char allreg_type[10][3]={"r", "g","o","l","i","sp",\
                              "fp", "f","lo", "hi"};
   enum type_index{r,g,o,l,i,sp,fp,f,lo,hi} indx;
   int j;
   /*to determine which is the register type*/
   for (j=0;j<10; j++){
      if (strcmp(allreg_type[j], reg_typec)==0)
         break;
   }
   if (j>=10)
      ERR_RET("syntax error\n");
   indx=j;
   switch (indx){
   case r:
      oprnd->register_type=R;
      break;
   case g:
      oprnd->register_type=R;
      break;
   case o:
      oprnd->register_type=R;
      oprnd->operand_no+=8;
      break;
   case l:
      oprnd->register_type=R;
      oprnd->operand_no+=16;
      break;
   case i:
      oprnd->register_type=R;
      oprnd->operand_no+=24;
      break;
   case sp:
      oprnd->register_type=R;
      oprnd->operand_no=14;
      break;
   case fp:
      oprnd->register_type=R;
      oprnd->operand_no=30;
      break;   
   case f:
      oprnd->register_type=F;
      break;   
   case lo:/*to be continued*/
      oprnd->register_type=LO;
      break;   
   case hi:
      oprnd->register_type=HI;
      break;   
   default:
      sprintf(errmsg, "%s: no such register", oprnd->s_oprnd);
      ERR_RET (errmsg);
      break;      
   }
}

/*Get the number of operands corresponding to the operator*/
int GetOprndNo(int op)
{
   switch (op) {
   case ADD:
   case ADDCC:
   case SUB:
   case SUBCC:
   case AND:
   case ANDCC:
   case OR:
   case ORCC:
   case ORN:
   case ORNCC:
   case XOR:
   case XORCC:
   case XNOR:
   case XNORCC:
   case SLL:
   case SRL:
   case SRA:
   case FADDS:
   case FADDD:
   case FSUBS:
   case FSUBD:
   case FMULS:
   case FMULD:
   case FDIVS:
   case FDIVD:
   case RESTORE:
   case SAVE:
      return (3);
      
   case BNE:
   case BE:
   case BG:
   case BLE:
   case BGE:
   case BL:
   case BGU:
   case BLEU:
   case BA:
   case FBG:
   case FBL:
   case FBNE:
   case FBE:
   case FBGE:
   case FBLE:
   case CALL:
      return(1);
      
   case RET:
   case RETL:
   case NOP:
      return (0);

   default:
      return(2);
   }
   
}

#ifdef DEBUG
void Show(FILE * fp, register char *rtl)
{

    register char c;
    char regprt[]="bwrfd       ";
    while (c = *rtl)
       if (isrg(rtl)) {
          (void) fprintf(fp, "%c[%d]", regprt[regtype(rtl)], regnum(rtl));
          rtl += 2;
       }
    /*the follwoing is about local and global names,
      not sure to be right*/
       else if ((c & 0xe0) == 0xa0) {
          (void)fprintf(fp, "%s", locnames[locnum(rtl)]);
          rtl += 2;
       }
       else if ((c & 0xe0) == 0x80) {
          (void)fprintf(fp, "%s", glonames[glonum(rtl)]);
          rtl += 2;
       }
       else {
          (void) putc(c, fp);
          rtl++;
       }
}
#endif

void ConvOprndD(OPRND *oprnd)
{
   int ntype, no;
                 
   ntype=D;
   no=oprnd->operand_no;
   
   TREG(ntype, no);
   sprintf(oprnd->s_oprnd, "%c%c", ntype, no);
}

/*deal with restore and save separately*/
void ResSaveSpec(STATEMENT *statem)
{
   int i;
   
   if ((statem->op_tor==RESTORE) ||
       (statem->op_tor==SAVE)){

      for (i=0; i<2; i++){
         /*deal with the situation when it is "restore ?, %g0, ?"*/
         if ( (statem->operand[i].operand_type==REG) &&
              (statem->operand[i].register_type==R) &&
              (statem->operand[i].operand_no==0)){
            *(statem->operand[i].s_oprnd)='\0';
         }/*end if*/
      }/*end for*/

      if ( (statem->operand[2].operand_type==REG) &&
           (statem->operand[2].register_type==R) &&
           (statem->operand[2].operand_no==0)){
         sprintf(statem->operand[2].s_oprnd,"%s", "NL");
      }
      
   }/*end if*/   
}

/*determine if the string is a number or not*/
int IsNumber(char *oprnd)
{
   char *p;
   p=oprnd;
   if (!isdigit((int)(*p))&&(*p!='-')){
      return 0;
   }
   else{
      p++;
      while(*p){
         if (!isdigit((int)(*p)))
            return 0;
         p++;
      }
   }
   return 1; 
}

#endif
