/************************************************************************
 *                                                                      *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   *
 *     The Rector and Visitors of the University of Virginia            *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from the University of Virginia.                                *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/assign.h,v $
 $Date: 2003/11/18 01:52:41 $
 $Revision: 1.10 $
 $Name:  $

 Log of changes:

 $Log: assign.h,v $
 Revision 1.10  2003/11/18 01:52:41  clc5q
 Added gtglobal() to get global reference in RTL.

 Revision 1.9  2003/08/01 02:16:00  jg9h
 Added a few helper functions.

 Revision 1.8  2003/06/17 13:58:53  clc5q
 Make the function assign() a static private function.  It is only called
    from within this module, and it has a name conflict with the assign()
    function in the Chaco graph partitioning library.

 Revision 1.7  2003/03/12 15:35:08  jdh8d
 *** empty log message ***

 Revision 1.6  2002/10/10 02:02:29  jg9h
 Major change to the local register assigner.
 It no longer bails out on seeing that only conflicting registers
 are available for assignment. (This is important in processors like
 the tms320c54x where it is easy to run out of registers).
 It temporarily uses a conflicting register for the rtl, but frees
 that register right after the rtl.
 Added gt_any_loc().

 Revision 1.5  2002/08/21 20:17:24  clc5q
 Add function gtloc() to get the next local variable in an RTL.

 Revision 1.4  2002/07/11 15:19:42  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.3  2001/04/10 16:40:44  rv3s
 *** empty log message ***

 Revision 1.2  1998/04/28 19:18:00  jwd
 Added void to some prototypes.  Deleted some commented out
 code.
 Revision 1.1  1998/04/03 16:49:05  jwd
 Initial revision

*************************************************************************/
#ifndef ASSIGN_H
#define ASSIGN_H

#include "vpo.h"

#ifdef MEASURE
int
get_freereg(struct bblock *, struct list *, int, int,char *, char *);
#endif

#ifdef FIFOS
char * gtfrg(char *);
#endif
char* gtstackrg(char *);

extern int hascs;              /* has caller save registers */
extern int mrveclen;
extern int maxregs;
extern char *memprt;
extern char *trash_cc;         /* String needed to trash condition codes */
extern int isused[MAXREGS];    /* TRUE if there is a use line for this reg. */
extern int usedtemps[MAXTEMP]; /* TRUE if spill area is used by function */

void
registers_used(void);

int
local_register_assignment(int);

int
assign_basic_block(struct bblock *);

void
dodeads(struct bblock *, rlist *, int);

void
solveconflicts(struct bblock *);

void
loopconflicts(struct bblock *, struct bblock *);

struct bblock *
add_fix_code(struct bblock *, struct bblock *, struct bblock *, char *,
             rlist *);

void
save_entry(struct bblock *);

void
initassign(regvec *);

void
unassociate(struct bblock *);

void
savregs(struct bblock *, struct list *, int *);

void
assoc(struct bblock *, struct list *, char *, int *, int);

struct list *
reserve(struct bblock *, struct list *, int *);

int
gtreg(struct bblock *, struct list *, struct setlist *, int, int *, int);

int
findreg(struct bblock *, struct list *, struct setlist *, int, int *, int, int);

long
distance(struct list *, char *);

int
fitreg(struct bblock *, struct list *, struct setlist *, int, int *, int *,
       int);

void
spillreg(struct bblock *, struct list *, int, int);

void
loadreg(struct bblock *, struct list *, struct setlist *, int, int, int);

int
fittemp(struct bblock *, int, int *);

void
conflicts(struct list *, char *, int *, int, int *);

void
doconflicts(struct bblock *, unsigned int, int *);

void
markreg(struct bblock *, struct setlist *, int, int);

void     
touchindex(vector regs, int index);

void     /* actually vector */
touchreg(rlist, char *);

void
releasereg(rlist, char *);

void
marktemp(struct bblock *, struct setlist *, int, int);

void
freereg(struct bblock *, struct setlist *);

void
freetemp(struct bblock *, struct setlist *);

void
freetemp_index(struct bblock * cblk, struct setlist * sptr, int index);

void
conflictreg(struct bblock *, struct setlist *);

void
conflicttemp(struct bblock *, struct setlist *);

int
find_unused_register(struct bblock *, int, int *);

char *gtrg(char *);

char *gtrgstr(char *, char *);

char *
gthrg(char *);

void
makerg(char *, int, int);

void
makereg(char *, int, int);

int
memtype(char *);

char *
gtmem(char *);

int
ismem(char *);

int
is_a_load(char*);

int
is_a_store(char*);

char *
gt_any_loc(char *);

char *gtloc(char *);
char *gtglobal(char *);

int
canreplace(int, rlist, char *, int, int);

int
cancolor(int, rlist, char *, int);

rlist
regnew(void);

void
regcopy(rlist, rlist);

void
inithardregs(void);

void
markinvariant(void);

char *
aliases(char *);

int
regtoindexf(int, int);

int
allocregf(int);

int
contregf(int);

int
is_cc(char *);

int
set_cc(char *);

int
checktype(int, int);
void set_usedtemp(int index);


int 
find_spill_location(int type, int *index);

struct llist *
spill_reg_to_loc(struct bblock *cblk, struct list *ptr, char* reg, int type, int tindex);

struct llist *
load_reg_from_loc(struct bblock *cblk, struct list *ptr, char *reg, int type, int tindex);




#endif

