
/************************************************************************
 *                                                                      *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   *
 *     The Rector and Visitors of the University of Virginia            *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from the University of Virginia.                                *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/controlflow.c,v $
 $Date: 2004/01/29 01:13:59 $
 $Revision: 1.68 $
 $Name:  $

 Log of changes:

 $Log: controlflow.c,v $
 Revision 1.68  2004/01/29 01:13:59  clc5q
 Use new blockCount field in struct loopl.

 Revision 1.67  2004/01/21 18:39:51  jdh8d
 Latest Vista additions/changes.  Don't yell at me if it's broken/bad.. prasad wanted these changes.  I've reviewed them, and tested,  minimally.. things appear OK (for the batch compiler)

 Revision 1.66  2003/11/18 01:56:36  clc5q
 Move DEFAULT_LOOP_ITERATIONS from controlflow.c to controlflow.h for use
     in local and global profiling.
 Call iterations() for any loop without a valid iterations count in
     estimate_execution_frequency().

 Revision 1.65  2003/09/15 14:47:28  clc5q
 Ran indent tool over file and fixed up a few cosmetic details by hand.

 Revision 1.64  2003/07/29 21:11:17  jg9h
 Added regs_in field to loopl structure.

 Revision 1.63  2003/07/17 04:33:42  jg9h
 VISTA update

 Revision 1.62  2003/05/01 00:55:00  jg9h
 Added a parameter to comparison()/md_replace_induction_var()/md_can_replace_induction_var()

 Revision 1.61  2003/04/24 04:55:32  jg9h
 Changed rtl.c:comparison() to take a struct list rather than char*

 Revision 1.60  2003/04/23 17:55:14  clc5q
 Add new functions loop_block_count(), innermost_containing_loop(), and
      inner_loop_iterations().
 Modify estimate_execution_frequency() to use the actual number of loop
      iterations where available, rather than assuming 10 iterations
      per loop.

 Revision 1.59  2003/04/09 20:54:54  clc5q
 Link each ignode and basic block to the loops that contain it.

 Revision 1.58  2003/04/03 20:06:09  jg9h
 Merged latest in VISTA with latest in vpo.

 Revision 1.57  2003/03/18 17:18:33  jdh8d
 *** empty log message ***

 Revision 1.56  2003/03/17 17:04:07  jdh8d
 fixed my previous blunder

 Revision 1.55  2003/03/17 14:12:49  jdh8d
 *** empty log message ***

 Revision 1.54  2003/03/14 01:34:20  jdh8d
 i386 bug fixes

 Revision 1.53  2003/02/06 04:46:27  jg9h
 Merged VISTA into mainline vpo under the INTERACT preprocessor flag

 Revision 1.52  2002/07/16 21:45:56  clc5q
 Export CHAIN() macro for thumb target.

 Revision 1.51  2002/07/11 15:22:30  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.50  2002/06/05 20:08:53  jdh8d
 Changes to support the stache, including routines to build a callgraph.

 Some cosmetic changes.

 Minor bug fixes.

 Changed build scripts to build the mips compiler for SimpleScalar if
 SSAS is set to "-DSS_AS"

 Revision 1.49  2002/04/22 16:00:32  jwd
 Updated a comment so that it was a little clearer about what was happening.

 Revision 1.48  2002/04/05 01:22:55  jwd
 Based on class discussions, we cleaned up some comments and reformatted some
 code.

 Revision 1.47  2002/04/02 21:15:55  rv3s
 Undo earlier changes to remv_unreachable_code(). An unreachable block is removed in its entirety. It is the responsiblity of machine dependent optimizations to handle any asm lines in the block

 Revision 1.46  2002/03/29 19:49:38  rv3s
 In remv_unreachable_code(), if a block is unreachable but has asm lines, all
 block info and data structures must be freed and reset so that subsequent
 optimizations do not process any old data from this unreachable block. The
 routine clear_block_info() nulls out the block data structures, but retains
 the block in the list. The block status should be changed to FALLTHRU.

 build_dominance_frontier assumes that only reachable basic blocks are present
 in the basic blocks list. Hence it assumes every block has an immediate
 dominator (which is not true for unreachable blocks). Added a check to see if
 cblk->idom is non-null in build_dominance_frontier().

 Revision 1.45  2002/03/20 18:02:03  jdh8d
 *** empty log message ***

 Revision 1.44  2002/03/19 01:28:58  jdh8d
 *** empty log message ***

 Revision 1.43  2002/03/18 22:21:27  jdh8d
 *** empty log message ***

 Revision 1.42  2002/03/18 15:49:39  jdh8d
 *** empty log message ***

 Revision 1.41  2002/03/15 20:40:22  jwd
 Cleaned up and elaborated a few comments.

 Revision 1.40  2002/03/14 00:11:51  rv3s
 remv_unreachable_code() now removes a basic block only if does not have any
 assembly (asm) lines. If there are any asm lines the block, the non-asm rtls are
 removed, but the block is not removed. The function
 remove_non_asm_lines_in_block() removes the non-asm rtls.

 Revision 1.39  2002/02/12 17:07:27  rv3s
 Fixed infinite looping due to flipping of minimize_loop_jumps and reverse_jumps
 . Each of the two routines were undoing the transformation of the other. The flipping was happening between two blocks, one inside and one outside the loop. Added a check in reverse_jumps not to perform the reverse if both the blocks (target and branch) are not inside the same loop and if the target is inside a loop

 Revision 1.38  2002/01/20 13:58:31  jdh8d
 *** empty log message ***

 Revision 1.37  2002/01/15 15:38:21  jdh8d
 *** empty log message ***

 Revision 1.36  2002/01/07 22:00:10  jdh8d
 *** empty log message ***

 Revision 1.35  2001/12/30 23:47:12  jdh8d
 *** empty log message ***

 Revision 1.34  2001/12/29 13:40:02  jdh8d
 Minimize registers and a ton of fixes

 Revision 1.33  2001/12/19 17:07:20  jdh8d
 *** empty log message ***

 Revision 1.32  2001/11/29 21:17:12  jdh8d
 *** empty log message ***

 Revision 1.31  2001/11/27 22:01:50  rv3s
 *** empty log message ***

 Revision 1.30  2001/11/27 22:00:46  rv3s
 is_chain() routine added to check if a block is an empty block or one RTL block with a jump rtl

 Revision 1.29  2001/11/23 21:50:28  jdh8d
 *** empty log message ***

 Revision 1.28  2001/11/23 17:37:53  jdh8d
 *** empty log message ***

 Revision 1.27  2001/11/16 17:23:11  jdh8d
 *** empty log message ***

 Revision 1.26  2001/11/16 17:14:57  rv3s
 set_rtl_sides called after change_target changes a branch instruction. This is to ensure that the SE is set correctly, especially if the new rtl is a return instruction

 Revision 1.25  2001/11/16 15:06:43  rv3s
 New routine only_one_useful_rtl added to ensure there is only one RTLINE in
 the rtl. Used by remv_branch_chains

 Revision 1.24  2001/11/09 18:34:25  jdh8d
 *** empty log message ***

 Revision 1.23  2001/10/10 15:58:09  jdh8d
 *** empty log message ***

 Revision 1.22  2001/10/10 01:26:19  jdh8d
 *** empty log message ***

 Revision 1.21  2001/10/09 17:15:42  jdh8d
 *** empty log message ***

 Revision 1.20  2001/10/08 22:32:46  jdh8d
 *** empty log message ***

 Revision 1.19  2001/09/24 14:08:59  jdh8d
 *** empty log message ***

 Revision 1.18  2001/05/25 00:36:05  jdh8d
 *** empty log message ***

 Revision 1.17  2001/05/16 18:53:40  jdh8d
 Changes to support "soft floating point"

 Revision 1.16  2001/04/25 21:18:49  jdh8d
 *** empty log message ***

 Revision 1.15  2001/04/25 19:40:44  jdh8d

 changed controlflow to check for "useless" blocks
 by checking that there are no "useful" rtls.  added
 an optimization to view.h so we can vpoiso some stuff,
 and made vpo.c call some of the label optimizations more
 frequently.

 Revision 1.14  2001/04/19 15:11:03  jdh8d
 Initial implentation of "if conversion" for basic predicated
 execution.

 Revision 1.13  2001/04/12 22:14:51  jdh8d

 fixed bug.

 Revision 1.12  2001/03/07 15:50:16  jdh8d


 eod fix to observe t* lines

 Revision 1.11  2000/06/09 19:07:42  jwd
 Changed calls to VET to indicate whether a phase was mandatory or optional.

 Revision 1.10  1999/03/05 17:15:32  jwd
 Fixed a problem with remv_branch_chains(). It would follow
 a NULL pointer when there was a branch to a block that
 had no code and it was the last block in the function.
 Revision 1.9  1999/02/11 15:26:04  jwd
 Fixed a minor glitch with the VET. Some code needed braces
 around it.
 Revision 1.8  1999/01/06 15:56:50  jwd
 Included assign.h to get some needed prototypes.
 Revision 1.7  1998/11/19 15:15:35  jwd
 Added VET code.
 Revision 1.5  1998/11/16 19:37:54  jwd
 Added new optimization to fix control flow that is generated by
 the SUIF code generator. The optimization gets rid of extra
 tests.
 Revision 1.4  1998/04/28 20:08:03  jwd
 Included some necessary header files.
 Change define constant FT to FALLTHRU to avoid a name
 conflict.
 Revision 1.3  1998/04/03 18:18:04  jwd
 First ANSI C version.

*************************************************************************/

/* ----History of Modifications------------------

   1/9/98: Modified to be Ansi-C compatible with header prototyping.
         Implemented the use of header files created last summer.
 */

#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "vpo.h"
#include "vpo2.h"
#include "assign.h"
#include "vectors.h"
#include "controlflow.h"
#include "misc.h"
#include "links.h"

#include "rtl.h"   /* machine dependent */
#include "inst.h"  /* machine dependent */

#include "view.h"

/*
 * This file contains functions that analyze and optimize the control
 * flow of procedures.
 */

int   labnum = 0;               /* counter used to make new labels  */

/*
 * The following items maintain the loop information structure lists
 */
struct loopl *loops = (struct loopl *) NULL;    /* Start of loop list     */
struct loopl *loope = (struct loopl *) NULL;    /* End of loop list       */

#ifdef FIX

/*
 *  fix_control_flow - change code that computes a boolean value and
 *  then only uses that value to jump, to code that just jumps
 */

int fix_control_flow(int invocation)
{
   struct bblock *cblk;
   struct bblock *BBlockB2, *BBlockB3;
   struct list *nptr;
   struct list *DefRtlPtr;
   struct list *SetRtlPtr;
   struct list *JmpRtlPtr;
   struct list *B1JmpRtlPtr;
   struct list *TestRtlPtr;
   char  ValueRegister[REGSTRING];
   char  SetRegister[REGSTRING];
   char *p, *r, *ptr1, *ptr2, expr[MAXSTRING], lab[MAXSTRING];
   int   change = FALSE;

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_fix_control_flow", invocation))
      dump_function_i("before fix_control_flow", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS);
   /*-
    * Search through each basic block for blocks that begin with a
    * conditional jump (call this block B1).  If the block (call it B2)
    * that falls through to basic block B1 sets a boolean value that is
    * tested by the first RTL to B1, delete the block B1 and the change
    * the code in B2 to jump directly to the appropriate label.  Also
    * delete the RTLs that set the boolean value.  Here's an example
    * r[32]=10
    * r[33]=r[30]+LOC[0]
    * r[34]=R[r[33]] r[33]
    * r[35]=0                   !this is basic block b3
    * IC=r[32]?r[34] r[32]r[34]
    * PC=IC`0,L3
    * r[35]=r[35]+1             !this is basic block b2
    * L3
    * IC=r[35]?0 r[35]          !this is basic block b1
    * PC=IC:0,L1
    * <Block B3>
    * L1
    * <Block B4>
    * 
    * becomes
    * 
    * r[32]=10
    * r[33]=r[30]+LOC[0]
    * r[34]=R[r[33]] r[33]
    * IC=r[32]?r[34] r[32]r[34]
    * PC=IC`0,L1
    * <Block B3>
    * L1
    * <Block B4>
    * 
    */
#ifdef INTERACT
   begin_phase(FIX_CONTROL_FLOW_TRANS);
#endif
   for (cblk = top; cblk; cblk = cblk->down) {
      nptr = cblk->rtls;
      if (nptr && (nptr->type & RTLINE) && nptr->val != NULL
          && is_cc(nptr->val)) {

         /* Make sure the next instruction is a branch */
         B1JmpRtlPtr = nptr->next;
         if (!B1JmpRtlPtr || !is_rjmp(B1JmpRtlPtr->val))
            continue;

         /* There must be exactly two preds from b3 and b2 */
         if (!cblk->preds || !cblk->preds->next || cblk->preds->next->next)
            continue;

         /* Determine which pred is b2; the fall-thru is always on the left. */
         if (cblk->preds->ptr->left == cblk)
            BBlockB2 = cblk->preds->ptr;
         else if (cblk->preds->next->ptr->left == cblk)
            BBlockB2 = cblk->preds->next->ptr;
         else
            continue;

         /* SetRtlPtr is the 1st rtl in b2 */
         SetRtlPtr = BBlockB2->rtls;
         if (!SetRtlPtr || !(SetRtlPtr->type & RTLINE))
            continue;

         /* Need to check that ValueRegister is on nptr's deadlist */
         gtrgstr(nptr->val, ValueRegister);
         if (!rexists(ValueRegister, nptr->deads))
            continue;

         /* check that the expression is IC=ValueRegister?0; */
         if (!comparison(nptr->val, nptr, ValueRegister, expr, &ptr1, &ptr2)
             || (*expr != '0' && *expr != '1') || *(expr + 1) != 0)
            continue;

         /* Check that previous RTL along the fall-thru sets ValueRegister. */
         if (!isset(SetRtlPtr->val, ValueRegister)
             || !SetToTrue(SetRtlPtr->val))
            continue;

         if (!SetRtlPtr->daddy->preds
             || !(JmpRtlPtr = SetRtlPtr->daddy->preds->ptr->rtlend))
            continue;

         BBlockB3 = JmpRtlPtr->daddy;
         if (!is_rjmp(JmpRtlPtr->val))
            continue;
         if (!(TestRtlPtr = prev_rtl(JmpRtlPtr)))
            continue;
         if (!is_cc(TestRtlPtr->val))
            continue;
         if (!(DefRtlPtr = prev_rtl(TestRtlPtr)))
            continue;
         gtrgstr(SetRtlPtr->val, SetRegister);
         if (strcmp(SetRegister, ValueRegister) != 0)
            continue;
         if (!SetToFalse(DefRtlPtr->val))
            continue;

         /* Change the target of the b3 branch & fix the control flow */

         /*-
          * Check whether the branch needs to be negated.  There are four
          *  cases to consider:
          *   case 1:  NEGATE
          *     IC=reg?0;
          *     PC=IC!0,L0;
          *   case 2:  NEGATE
          *     IC=reg?1;
          *     PC=IC:0,L0;
          *   case 3:  DON'T NEGATE
          *     IC=reg?0;
          *     PC=IC:0,L0;
          *   case 4:  DON'T NEGATE
          *     IC=reg?1;
          *     PC=IC!0,L0;
          */
         if ((*expr == '0' && IsNotEqualRelJmp(B1JmpRtlPtr->val))
             || (*expr == '1' && IsEqualRelJmp(B1JmpRtlPtr->val))) {
#ifdef INTERACT
            begin_trans(VET_OPT_TRANS);
#endif
            /* extract the new target label from the cblk branch */
            for (p = nptr->next->val; *p && *p != ','; p++) ;
            if (!*p)
               continue;
            p++;
            for (r = lab; *p && *p != ';'; p++)
               *r++ = *p;
            *r = '\0';

            /* Negate the branch condition using reverse() from md */
            reverse(JmpRtlPtr, lab);
         }
         else if ((*expr == '0' && IsEqualRelJmp(B1JmpRtlPtr->val))
                  || (*expr == '1' && IsNotEqualRelJmp(B1JmpRtlPtr->val))) {
#ifdef INTERACT
            begin_trans(VET_OPT_TRANS);
#endif
            /* keep the original branch condition, but change the target */
            change_target(BBlockB3, BBlockB3->right, cblk->right);
         }
         else
            continue;

         /* Update control flow -- set the predecessor of the target block
          * of B1's branch to B3, set B3's right child to the new target,
          * remove B3 from B1's predecessor list (it has only B2 as its
          * pred after the opt), and remove B1's right child since B1's
          * branch is eliminated.
          */
         change_predecessor(cblk, BBlockB3, cblk->right);
         BBlockB3->right = cblk->right;
         remv_predecessor(BBlockB3, cblk);
         cblk->right = (struct bblock *) NULL;

         /* Delete the unnecessary RTL's so their blocks will be deleted */
         remv_links(SetRtlPtr);
         remv_rtl(SetRtlPtr);
         remv_links(nptr->next);
         remv_rtl(nptr->next);
         remv_links(nptr);
         remv_rtl(nptr);
         remv_links(DefRtlPtr);
         remv_rtl(DefRtlPtr);

         change = TRUE;
#ifdef INTERACT
         end_trans();
#endif
      }
   }
#ifdef INTERACT
   end_phase(FIX_CONTROL_FLOW_TRANS);
#endif

   if (change) {
      (void) remv_useless_basic_blocks(100);
   }


   if (swd && debug_opt_i("dump_after_fix_control_flow", invocation))
      dump_function_i("after fix_control_flow", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS);

   return change;
} /* end of fix_control_flow() */

#endif  /* FIX */

/*
 * remv_branch_chains - remove branch chains from the current function
 */
int remv_branch_chains(int invocation)
{
   struct bblock *cblk, *nblk;
   struct blist *bptr;
   struct bblock *ntarget, *target, *ptarget;
   struct blist *plast, *rptr;
   struct list *rtl, *dptr;
   int   change, err;
#ifdef INTERACT
   int   inscope;
   extern int limited_scope;
#endif

#ifdef VET
#ifdef INTERACT
   begin_phase(BRANCH_CHAINING_TRANS);
#else
   begin_phase(BRANCHCHAIN_PHASE);
#endif
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_remv_branch_chains", invocation))
      dump_function_i("before remv_branch_chains()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS);

   /*
    * Search through every basic block in the function for blocks that are part
    * of a branch chain.  A basic block is part of a branch chain if it
    * contains a label and is either empty, so that it merely falls into the
    * block below it, or if it contains only an unconditional jump to another
    * basic block.
    */
   change = err = FALSE;
   for (cblk = top; cblk; cblk = cblk->down) {

      /*
       * To emliminate branch chains, we start with a branch and follow it to
       * its logical conclusion.  If the final destination is someplace other
       * than where the branch jumps to, we can improve the code by changing
       * the branch label to go to the final destination.  Here, we check to
       * see if this basic block has a label but performs no useful function.
       * If this is the case, then we want to change all of the items that
       * transfer control to this block to transfer control to the final target
       * that we will reach if we follow the control flow chain from this point
       * forward.
       */
      if (cblk->label && CHAIN(cblk)) {

         /*
          * Clear the DONE bits from every block to allow us to detect loops in
          * branch chains.
          */
         for (nblk = top; nblk; nblk = nblk->next)
            nblk->status &= ~DONE;

         /*
          * This is a basic block with a label containing a single jump to some
          * other basic block or no RTLs at all.  This block is useless after
          * all references to it are replaced with references to its target or
          * the to the basic block that it falls into.  First, we need to
          * determine the final target of this branch chain.  We will keep
          * track of both the head of the chain and the previous item in the
          * chain in cases where we cannot transfer control to the final target
          * directly.
          */
         ptarget = (struct bblock *) NULL;
         target = cblk->left;
         if (target == ret_block && debug_opt("size"))
            continue;
#ifdef INTERACT
         if (swq)
            inscope = cblk->inscope;
#endif
         while (target && CHAIN(target) && target != ret_block) {
#ifdef INTERACT
            if (swq)
               inscope |= target->inscope;
#endif
            target->status |= DONE;
            ptarget = target;
            if (target->left && target->left->status & DONE) {
               if (!sww && !err) {
                  (void) fprintf(stderr,
                                 "vpo (remv_branch_chains): infinite loop");
                  if (fn)
                     (void) fprintf(stderr, " in %s()\n", fn);
                  else
                     (void) fprintf(stderr, "\n");
                  err = TRUE;
               }
               target = NULL;
               break;
            }

            target = target->left;
         }

         /*
          * If we have located a branch chain, attempt to change the target of
          * all of the items that branch to this basic block to go directly to
          * the end of the chain.
          */
         /* Oops, be careful that there is a ret_block is set.  
          * some functions may not have a ret block, so we can optimize them
          * better.   More subtle, if ret_block is null, AND target is null
          * (for a reason) then target was becoming ptarget at the wrong time
          * and making un-reasonable optimizations.
          */
         if (ret_block && target == ret_block && debug_opt("size"))
            target = ptarget;
#ifdef INTERACT
         if (target && target->label && !(swq && limited_scope && !inscope))
#else
         if (target && target->label)
#endif
         {
#ifdef VET
            begin_trans(VET_OPT_TRANS);
#endif

            /*
             * Look at all of the predecessors of the basic block that starts
             * the branch chain.  We wish to change all of them to transfer
             * control directly to the final target of the chain.
             */
            plast = (struct blist *) NULL;
            bptr = cblk->preds;
            while (bptr) {

               /*
                * If the predecessor ends in an RTL that references the start
                * of the branch chain, then we can modify it to transfer
                * control to the end of the chain, provided that we do not try
                * to move a return into a switch lookup table.
                */
               if ((rtl = find_control_rtl(bptr->ptr, &dptr, FALSE))
                   && !(condret && target == ret_block
                        && rtl->type & CASELINE)
                   && fmatch(rtl->val, cblk->label) && (*epos > '9'
                                                        || *epos < '0')) {

                  /*
                   * If this is a conditional branch, then we must determine if
                   * the target machine permits us to return conditionally, if
                   * not, then branch to the location that performs the return.
                   */
#if 0
                  /*
                   * For size optimizations, we don't want to remove a chain
                   * to the return block.  The return block is likely large,
                   * and duplicating the PC=RT; is expensive.
                   * We fix this by stopping the chain one early.. see above 
		             * if(debug_opt("size") && target==ret_block)
                   */
                  if (FALSE) {
                     ntarget = NULL;
                  }
                  else
#endif /* 0 */
                  if (bptr->ptr->status & JUMP || target != ret_block
                           || condret) {
                     ntarget = target;
                  }
                  else {
                     ntarget = ptarget;
                  }

                  /*
                   * If we can still eliminate the branch chain, then alter the
                   * label and change the predecessor information to show that
                   * the basic block that used to be our predecessor is now the
                   * target block's predecessor.
                   */
                  if (ntarget) {
                     if (!change_target(bptr->ptr, cblk, ntarget)) {
                        (void) fprintf(stderr,
                        "vpo (remv_branch_chains): failed to change target\n");
                        exit(EXIT_FAILURE);
                     }
                     rptr = bptr;
                     bptr = bptr->next;
                     rptr->next = ntarget->preds;
                     ntarget->preds = rptr;
                     change = TRUE;
                     continue;
                  }
               }

               /*
                * Otherwise, we cannot change the target of the predecessor to
                * transfer control down the chain, so we keep the predecessor
                * in the predecessor list of this basic block.
                */
               if (plast)
                  plast->next = bptr;
               else
                  cblk->preds = bptr;
               plast = bptr;
               bptr = bptr->next;
            }

            /*
             * If this block retains any of its predecessors, then fix up the
             * predecessor list appropriately.
             */
            if (plast)
               plast->next = (struct blist *) NULL;
            else{
	      cblk->preds = (struct blist *) NULL;
	      /*mark_blk_unreachable(cblk);*/
	      remove_block(cblk);
	    }
#ifdef VET
            end_trans();
#endif
         }
      }
   }

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_remv_branch_chains", invocation))
      dump_function_i("after remv_branch_chains()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS);

#ifdef VET
#ifdef INTERACT
   end_phase(BRANCH_CHAINING_TRANS);
#else
   end_phase(BRANCHCHAIN_PHASE);
#endif
#endif
   /*
    * Return indicating if any changes have been made to the control flow graph
    * of the function.
    */
   return change;
} /* end of remv_branch_chains() */


/*
 * mark_block_unreachable - if any block is left unreachable after branch
 *                          chaining, then we do not want to remove it as we
 *                          want remv_unreachabl_code to still show results.
 *                          So instead we remove its links but still keep it
 *                          in the control flow
 */
int mark_blk_unreachable(struct bblock *blk)
{
  int count;
  struct list *rtl;
  
  /*this block should only have one useful RTL which is the unconditional
    jump*/
  count = 0;
  for(rtl=blk->rtls ; rtl ; rtl=rtl->next){
    if(cnts(rtl))
      count++;
  }

  if(count < 2){
    
    /*mark the unconditional jump as a comment*/
    if(blk->rtlend && cnts(blk->rtlend) && is_ucjmp(blk->rtlend->val)){
      mark_comment(blk->rtlend);
    }
    
    /*if this block has any children then remove links to them*/
    if(blk->left)
      update_bb_ptrs(MOD_LEFT, blk, NULL);

    if(blk->right)
      update_bb_ptrs(MOD_RIGHT, blk, NULL);
  }
}

static int has_useful_rtls(struct bblock *cblk)
{
   struct list *rtl;
   struct loopl *looplist;

   for (rtl = cblk->rtls; rtl; rtl = rtl->next) {
      if ((rtl->type & RTLINE) == RTLINE && rtl->val)
         return TRUE;
      if (rtl->type & ASMLINE)
         return TRUE;
   }

   /* If this block is a preheader to a loop, we want
    * to keep it around.. even if it is seemingly useless.
    */
   for (looplist = loops; looplist; looplist = looplist->next) {
      if (cblk == looplist->preheader)
         return TRUE;
   }
   return FALSE;
}

/*
 * remv_useless_basic_blocks - remove useless basic blocks from the function
 */
int remv_useless_basic_blocks(int invocation)
{
   struct bblock *cblk;
   struct bblock *dblk;
   int   change;
#ifdef INTERACT
   extern int limited_scope;
#endif

#ifdef VET
#ifdef INTERACT
   begin_phase(ELIM_EMPTY_BLOCKS_TRANS);
#else
   begin_phase(DEADCODEELIM_PHASE);
#endif
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd
       && debug_opt_i("dump_before_remv_useless_basic_blocks", invocation))
      dump_function_i("before remv_useless_basic_blocks()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS);

   /*
    * No changes have been made to the control flow graph yet
    */
   change = FALSE;

   /*
    * Search through the function for useless basic blocks.  Skip the entry
    * block, since that is always useful.
    */
   for (cblk = top->down; cblk; cblk = cblk->down) {
#ifdef INTERACT
      if (swq && limited_scope && !cblk->inscope)
         continue;
#endif

#ifdef VET
      begin_trans(VET_OPT_TRANS);
#endif
      /*
       * A basic block that has no RTLs and no predecessors is useless
       */
      if (!has_useful_rtls(cblk) && !cblk->preds) {

         /*
          * Uncouple the basic block from the output order list
          */
         if (cblk->up->down = cblk->down)
            cblk->down->up = cblk->up;
#ifdef VET
         mod_bb_ptrs(MOD_DOWN, cblk->up, cblk->up->down);
#endif

         /*
          * Uncouple the basic block from the depth-first-search order list
          */
         if (cblk->prev_dfs)
            cblk->prev_dfs->next_dfs = cblk->next_dfs;
         if (cblk->next_dfs)
            cblk->next_dfs->prev_dfs = cblk->prev_dfs;
         else
            bottom = cblk->prev_dfs;

         /*
          * If the basic block has a left child, ensure that the reference to
          * this basic block in the child's predecessor list is removed.  Also,
          * if this basic block used to fall into the basic block below it,
          * then it does not do so any more.
          */
         if (cblk->left) {
            remv_predecessor(cblk, cblk->left);
            cblk->left->status &= ~FALLTHRU;
         }

         /*
          * Remove this basic block from the function and return it to the heap
          */
         dblk = cblk;
         cblk = cblk->up;

         remv_basic_block(dblk);
         change = TRUE;
      }

      /*
       * A basic block that contains no RTLs, and is fallen into by its only
       * predecessor is useless.
       */
      else if (!has_useful_rtls(cblk) && (cblk->status & FALLTHRU)
               && !cblk->preds->next) {

         /*
          * Uncouple the basic block from the output order list
          */
         if (cblk->up->down = cblk->down)
            cblk->down->up = cblk->up;
#ifdef VET
         mod_bb_ptrs(MOD_DOWN, cblk->up, cblk->up->down);
#endif

         /*
          * Uncouple the basic block from the depth-first-search order list
          */
         if (cblk->prev_dfs)
            cblk->prev_dfs->next_dfs = cblk->next_dfs;
         if (cblk->next_dfs)
            cblk->next_dfs->prev_dfs = cblk->prev_dfs;
         else
            bottom = cblk->prev_dfs;

         /*
          * If the basic block has a left child, ensure that the reference to
          * this basic block in the child's predecessor list is removed.
          */
         if (cblk->preds->ptr->left = cblk->left)
            change_predecessor(cblk, cblk->preds->ptr, cblk->left);
#ifdef VET
         mod_bb_ptrs(MOD_LEFT, cblk->preds->ptr, cblk->left);
#endif

         /*
          * Remove this basic block from the function and return it to the heap
          */
         dblk = cblk;
         cblk = cblk->up;

         remv_basic_block(dblk);
         change = TRUE;
      }
#ifdef VET
      end_trans();
#endif
   }
#ifdef VET
#ifdef INTERACT
   end_phase(ELIM_EMPTY_BLOCKS_TRANS);
#else
   end_phase(DEADCODEELIM_PHASE);
#endif
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_remv_useless_basic_blocks", invocation))
      dump_function_i("after remv_useless_basic_blocks()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS);

   /*
    * Return indicating if any changes have been made to the control flow graph
    * of the function.
    */
   return change;
} /* end of remv_useless_basic_blocks() */

/*
 * remv_useless_jumps - remove useless jumps from the function
 */
int remv_useless_jumps(int invocation)
{
   struct bblock *cblk, *target;
   struct list *ptr;
   struct list *dptr;
   int   change;
#ifdef INTERACT
   extern int limited_scope;
#endif

#ifdef VET
#ifdef INTERACT
   begin_phase(USELESS_JUMP_ELIM_TRANS);
#else
   begin_phase(USELESSJUMPS_PHASE);
#endif
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_remv_useless_jumps", invocation))
      dump_function_i("before remv_useless_jumps()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_DOM |
                      SHOW_SUCCS | SHOW_PREDS);

   change = FALSE;

   /*
    * The first pass eliminates very simple and obvious useless jumps. These
    * consist of either an unconditional branch directly to the basic block
    * which follows the branch or a conditional branch which reaches the same
    * location regardless of which path is taken.
    */
   for (cblk = top; cblk; cblk = cblk->down) {

#ifdef INTERACT
      if (swq && limited_scope && !cblk->inscope)
         continue;
#endif
      /*
       * Look for a basic block ending in an unconditional jump that transfers
       * control directly to the basic block which follows it.  If any such
       * block is found, delete the jump.
       */
#ifdef VET
            begin_trans(VET_OPT_TRANS);
#endif
      if (cblk->left && !cblk->right && cblk->status & JUMP
          && cblk->left == cblk->down) {

         /*
          * If the unconditional branch is found, then remove it from the basic
          * block and indicate that the block no longer contains an
          * unconditional branch instruction and that we now fall into the
          * basic block below us.  If there was anything in the branch's delay
          * slot, then note that it is no longer a delayed instruction.
          */
         if ((ptr = find_control_rtl(cblk, &dptr, TRUE)) && cnts(ptr)) {
            remv_links(ptr);
            remv_rtl(ptr);
            cblk->status &= ~JUMP;

            cblk->down->status |= FALLTHRU;
            if (dptr) {
               dptr->kind &= ~(TAKEN | NOTAKEN);

#ifdef INTERACT
               mod_kind(dptr);
#endif

            }
            change = TRUE;
         }
      }
#ifdef VET
      end_trans();
#endif      

      /*
       * Look for a basic block that contains a conditional branch and both
       * children are the same basic block.  If any such block is found, delete
       * the conditional branch.
       */
#ifdef VET
            begin_trans(VET_OPT_TRANS);
#endif
      if (cblk->left && cblk->left == cblk->right) {
         if ((ptr = find_control_rtl(cblk, &dptr, TRUE)) && cnts(ptr)) {

            remv_links(ptr);
            remv_rtl(ptr);
            cblk->right = (struct bblock *) NULL;
#ifdef VET
            mod_bb_ptrs(MOD_RIGHT, cblk, (struct bblock *) NULL);
#endif
            if (dptr) {
               dptr->kind &= ~(TAKEN | NOTAKEN);
#ifdef INTERACT
               mod_kind(dptr);
#endif
            }

            change = TRUE;
         }
      }
#ifdef VET
      end_trans();
#endif      
   }

   /*
    * The second pass eliminates useless branches that are a little bit harder
    * to spot because they actually branch through basic blocks that contain no
    * items.  Such basic blocks are sometimes left behind when optimizations
    * like CSE and code motion operate on a function.
    */
   for (cblk = top; cblk; cblk = cblk->down) {

      /* This arm is handling an unconditional branch to a block that can
       * be reached without executing the branch. So the branch will be removed
       */
      if (cblk->left && !cblk->right && cblk->status & JUMP
          && cblk->left != cblk->down) {
         if ((ptr = find_control_rtl(cblk, &dptr, TRUE)) && cnts(ptr)) {
#ifdef VET
            begin_trans(VET_OPT_TRANS);
#endif

            /* oops, cblk->down can't have rtls either */
            if (cblk->down && !cblk->down->rtls)
               target = cblk->down;
            else
               target = 0;
            while (target) {
               if (cblk->left == target) {
                  remv_predecessor(cblk, target);
                  remv_links(ptr);
                  remv_rtl(ptr);
                  add_predecessor(cblk, cblk->down);
                  cblk->left = cblk->down;
#ifdef VET
                  mod_bb_ptrs(MOD_LEFT, cblk, cblk->left);
#endif
                  cblk->status &= ~JUMP;
                  cblk->down->status |= FALLTHRU;
                  if (dptr) {
                     dptr->kind &= ~(TAKEN | NOTAKEN);
#ifdef INTERACT
                     mod_kind(dptr);
#endif
                  }
                  change = TRUE;
                  break;
               }
               if (!target->rtls)
                  target = target->down;
               else
                  break;
            }
#ifdef VET
            end_trans();
#endif
         }
      }

      /* This arm is handling a conditional branch to a block that can
       * be reached without executing the branch. So the branch will be removed
       */
      if (cblk->left && cblk->right) {
         if ((ptr = find_control_rtl(cblk, &dptr, TRUE)) && cnts(ptr)) {
#ifdef VET
            begin_trans(VET_OPT_TRANS);
#endif
            target = cblk->down;
            while (target) {
               if (cblk->right == target) {
                  remv_predecessor(cblk, target);
                  remv_links(ptr);
                  remv_rtl(ptr);
                  cblk->right = (struct bblock *) NULL;
#ifdef VET
                  mod_bb_ptrs(MOD_RIGHT, cblk, cblk->right);
#endif
                  if (dptr) {
                     dptr->kind &= ~(TAKEN | NOTAKEN);
#ifdef INTERACT
                     mod_kind(dptr);
#endif
                  }
                  change = TRUE;
                  break;
               }
               if (!target->rtls)
                  target = target->down;
               else
                  break;
            }
#ifdef VET
            end_trans();
#endif
         }
      }
   }
#ifdef VET
#ifdef INTERACT
   end_phase(USELESS_JUMP_ELIM_TRANS);
#else
   end_phase(USELESSJUMPS_PHASE);
#endif
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_remv_useless_jumps", invocation))
      dump_function_i("after remv_useless_jumps()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_DOM |
                      SHOW_SUCCS | SHOW_PREDS);

   /*
    * Return indicating if any changes have been made to the control flow graph
    * of the function.
    */
   return change;
} /* end of remv_useless_jumps() */

/*
 * remv_unreachable_code - eliminate unreachable (dead) code
 */
int remv_unreachable_code(int invocation)
{
   struct bblock *cblk, *end;
   struct blist *ptr;
   struct loopl *loop;
   int   change;
   int   ok_to_delete;
   extern int limited_dead_code_elim;

#ifdef VET
#ifdef INTERACT
   begin_phase(DEAD_CODE_ELIM_TRANS);
#else
   begin_phase(DEADCODEELIM_PHASE);
#endif
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_remv_unreachable_code", invocation))
      dump_function_i("before remv_unreachable_code()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS);

   /*
    * Reset the DONE bit on all basic blocks
    */
   for (cblk = top; cblk; cblk = cblk->next)
      cblk->status &= ~DONE;

   /*
    * Mark only those blocks that are reachable from the entry point of the
    * function.  All unmarked items are then dead code.
    */
   mark_reachable_code(top);

   /*
    * The pointer, "end", will point to the last basic block in the basic block
    * list when the loop finishes.  Note that since the "top" block is the loop
    * entry point, it will always be set to DONE.  Blocks not marked DONE
    * cannot be reached from the entry point, therefore, they are dead code.
    * We avoid deleting the return block even if it is unreachable to prevent
    * the "ret_block" item from pointing to a structure that may be reclaimed
    * and used for something else in the future.  Since all of the predecessors
    * to the return block will also be unreachable in this case, the return
    * block will end up with no predecessor items in its predecessor list.
    */
   change = FALSE;
   end = (struct bblock *) NULL;
   for (cblk = top; cblk; cblk = end) {
      if (cblk->status & DONE || cblk == ret_block) {
         end = cblk->next;
         if (end)
            last = end;
      }
      else {

         /* Remove the basic block only if does not have asm lines */
         end = cblk->next;

#ifdef VET
         begin_trans(VET_OPT_TRANS);
#endif

	 remove_block(cblk);
	 
#ifdef VET
         end_trans();
#endif

         change = TRUE;
      }
   }
   for (cblk = top; cblk; cblk = cblk->next)
      last = cblk;
   /*
    * The last item must mark the end of the list
    */

   if (last)
      last->next = (struct bblock *) NULL;
#ifdef VET
#ifdef INTERACT
   end_phase(DEAD_CODE_ELIM_TRANS);
#else
   end_phase(DEADCODEELIM_PHASE);
#endif
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_remv_unreachable_code", invocation))
      dump_function_i("after remv_unreachable_code()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS);

   /*
    * Return indicating if any changes have been made to the control flow graph
    * of the function.
    */
   return change;
} /* end of remv_unreachable_code() */


/*
 * remove_block - remove a basic block from the control flow
 */
void remove_block(struct bblock *cblk)
{
  struct blist *ptr;
  extern int limited_dead_code_elim;
  
  /*in limited unreachable code elimination we only remove the
    left/right and predecessor indicators from the nodes*/
  if(!limited_dead_code_elim){
    
    /*
     * 1     Uncouple the basic block from the output order list
     */
    if (cblk->up){
      cblk->up->down = cblk->down;
#ifdef VET
      mod_bb_ptrs(MOD_DOWN, cblk->up, cblk->down);
#endif
    }
    if (cblk->down)
      cblk->down->up = cblk->up;
	   
    /*
     * Uncouple the basic block from the depth-first-search order list
     */
    if (cblk->prev_dfs)
      cblk->prev_dfs->next_dfs = cblk->next_dfs;
    if (cblk->next_dfs)
      cblk->next_dfs->prev_dfs = cblk->prev_dfs;
    else
      bottom = cblk->prev_dfs;
  }

  /*
   * If the basic block has a left child, ensure that the reference to
   * this basic block in the child's predecessor list is removed.  Also,
   * if this basic block used to fall into the basic block below it,
   * then it does not do so any more.
   */
  if (cblk->left) {
    remv_predecessor(cblk, cblk->left);
    if (!(cblk->status & JUMP))
      cblk->left->status &= ~FALLTHRU;
  }

  /*
   * Do the same thing with the right child, if any
   */
  if (cblk->right)
    remv_predecessor(cblk, cblk->right);
  
  /*
   * It is intuitively obvious that since this basic block cannot be
   * reached from the entry point that, by definition, none of its
   * predecessors can be reached from the entry point either.  Since
   * this is the case, we go ahead and make these radical changes to the
   * control flow graph which must be made to ensure that we do not
   * attempt to go into the predecessor items of the basic block that we
   * are about to free up in the future through the predecessor's child
   * links.  Just look at the code that preceedes this comment and
   * imagine what might happen if we don't remove the child pointers to
   * this basic block.
   */
  for (ptr = cblk->preds; ptr; ptr = ptr->next) {
    if (ptr->ptr->left == cblk) {
      ptr->ptr->left = (struct bblock *) NULL;
#ifdef VET
      mod_bb_ptrs(MOD_LEFT, ptr->ptr, (struct bblock *) NULL);
#endif
    }
    if (ptr->ptr->right == cblk) {
      ptr->ptr->right = (struct bblock *) NULL;
#ifdef VET
      mod_bb_ptrs(MOD_RIGHT, ptr->ptr, (struct bblock *) NULL);
#endif
    }
  }
  
  
  /*
   * Return the basic block to the heap
   */
  if(!limited_dead_code_elim)
    remv_basic_block(cblk);
  else
    mark_blk_unreachable(cblk);  
}



/*
 * mark_reachable_code - recursively mark all reachable blocks as DONE
 */
void mark_reachable_code(register struct bblock *cblk)
{
   while (cblk && !(cblk->status & DONE)) {
      cblk->status |= DONE;
      if (cblk->right) {
         mark_reachable_code(cblk->left);
         cblk = cblk->right;
      }
      else
         cblk = cblk->left;
   }
}

/*
 * might_remove - determine if cblk is save to remove by reversing the
 * condition of the previous block
 */
static int might_remove(struct bblock *cblk)
{
   struct list *rtlptr = cblk->rtls;

   for (rtlptr = cblk->rtls; rtlptr; rtlptr = rtlptr->next) {
      if (cnts(rtlptr) && is_pc(rtlptr->val))
         return TRUE;
      if (cnts(rtlptr))
         return FALSE;

      /* don't remove blocks with asmlines */
      if (rtlptr->type & ASMLINE)
         return FALSE;
   }
   /* this is an empty block if we get this far */
   return TRUE;
}


/*
 * reverse_jumps - optimize jump sequences by reversing the condition
 */
int reverse_jumps(int invocation)
{
   struct bblock *cblk, *target, *branch, *junk;
   struct list *jump;
   int   change;
#ifdef INTERACT
   char  oldrtl[MAXSTRING];
   extern int limited_scope;
#endif

#ifdef VET
#ifdef INTERACT
   begin_phase(REVERSE_BRANCHES_TRANS);
#else
   begin_phase(JUMPS_PHASE);
#endif
#endif

   /*-
    * This particular optimization involves changing a sequence of branches
    * such as:
    * 
    * PC = NZ > 0, L50; PC = L45; L50   ...
    * 
    * to:
    * 
    * PC = NZ <= 0, L45; L50   ...
    */

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_reverse_jumps", invocation))
      dump_function_i("before reverse_jumps()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS | SHOW_DFS);

   change = FALSE;
   target = branch = junk = (struct bblock *) NULL;
   jump = (struct list *) NULL;
   for (cblk = top; cblk; cblk = cblk->down) {

      /*
       * If we have located a sequence of branches that might potentially be
       * replaced with a single branch, check to see if this basic block is the
       * target of the conditional branch.
       */
      if (branch) {
         if (cblk == target) {

            /* If the target block is in a loop and the branch block is not
             * inside the same loop, then its not a good idea to reverse the
             * jumps for 2 reasons:
             * - minimize_loop_jumps would flip it back, causing flip-flop
             *    between reverse_jumps and minimize_loop_jumps.
             * - a jump into a  block inside the loop is more likely.
             */
            if (!(inside_some_loop(target)
		  && in_different_loops(target, branch))){
#ifdef VET
	      begin_trans(VET_OPT_TRANS);
#endif
	      /*
	       * If we have located such a sequence, reverse the conditional
	       * branch and get rid of the unconditional branch.
	       */
#ifdef INTERACT
	      if (swq) {
		remv_rtl(junk->rtls);
		strcpy(oldrtl, jump->val);
	      }
#endif
	      reverse(jump, branch->label);
#ifdef VET
#ifdef INTERACT
	      mod_rtl(jump, oldrtl, jump->val);
#else
	      mod_rtl(jump);
#endif
#endif
	      combmark(jump);
	      change = TRUE;
	      jump->daddy->right = branch;
	      jump->daddy->left = cblk;
	      /*
	       * What used to be the predecessor can be replaced with the new
	       * predecessor directly in the list.
	       */
#ifdef INTERACT
	      if (swq)
		change_predecessor(junk, jump->daddy, branch);
#endif
#ifdef VET
	      mod_bb_ptrs(MOD_RIGHT, jump->daddy, branch);
	      mod_bb_ptrs(MOD_LEFT, jump->daddy, cblk);
#endif
	      cblk->status |= FALLTHRU;
#ifdef INTERACT
	      if (!swq)
#endif
		{
		  change_predecessor(junk, jump->daddy, branch);
		}

	      /*
	       * Uncouple this basic block from the depth-first-search order list
	       */
	      if (junk->prev_dfs)
		junk->prev_dfs->next_dfs = junk->next_dfs;
	      if (junk->next_dfs)
		junk->next_dfs->prev_dfs = junk->prev_dfs;
	      else
		bottom = junk->prev_dfs;

	      /*
	       * Uncouple this basic block from the output order list
	       */
	      jump->daddy->down = cblk;
#ifdef VET
	      mod_bb_ptrs(MOD_DOWN, jump->daddy, cblk);
#endif
	      cblk->up = jump->daddy;
	      
	      /*
	       * Remove the basic block from the next list and return it to the
	       * heap.
	       */

	      remv_basic_block(junk);
#ifdef VET
	      end_trans();
#endif
	    }
	 }

         target = branch = junk = (struct bblock *) NULL;
      }

      /*
       * If the previous basic block ended in a conditional branch and this
       * basic block contains an unconditional branch, then set the item
       * "branch" to point to the basic block which the unconditional branch
       * transfers control to.
       */
      else if (target) {
         if (!cblk->right && cblk->left && cblk->preds && !cblk->preds->next) {

            /*
             * If this is an empty basic block, then it is completely useless,
             * and we can just get rid of it at this time.
             */
            if (!cblk->rtls) {

#ifdef VET
               begin_trans(VET_OPT_TRANS);
#endif
               /*
                * This block is no longer the predecessor of the block that
                * follows it.  The block immediately above us is.
                */
               change_predecessor(cblk, cblk->up, cblk->down);

               /*
                * Uncouple this basic block from the depth-first-search order
                * list
                */
               if (cblk->prev_dfs)
                  cblk->prev_dfs->next_dfs = cblk->next_dfs;
               if (cblk->next_dfs)
                  cblk->next_dfs->prev_dfs = cblk->prev_dfs;
               else
                  bottom = cblk->prev_dfs;

               /*
                * Uncouple this basic block from the output order list
                */
               cblk->up->left = cblk->up->down = cblk->down;
#ifdef VET
               mod_bb_ptrs(MOD_LEFT, cblk->up, cblk->down);
               mod_bb_ptrs(MOD_DOWN, cblk->up, cblk->down);
#endif
               cblk->down->up = cblk->up;

               /*
                * Remove the basic block from the next list and return it to
                * the heap.
                */
               junk = cblk;
               cblk = cblk->up;

               remv_basic_block(junk);
#ifdef VET
               end_trans();
#endif
               continue;
            }

            /*
             * Otherwise, if the basic block has an unconditional branch, then
             * it may be a part of a sequence that can be eliminated by
             * reversing the conditional branch that preceedes it.
             */
            else if (cblk->left->label && cblk->left != ret_block
                     && might_remove(cblk)) {
               branch = cblk->left;
               junk = cblk;
               continue;
            }
         }
         target = (struct bblock *) NULL;
      }

      /*
       * This section locates a basic block that contains a conditional branch
       * as its last RTL.  The item "target" is then set to point to the basic
       * block that the conditional branch transfers control to when the branch
       * is taken.  The item "jump" points to the RTL containing the
       * conditional branch.
       */
      if (cblk->left && cblk->right && cblk->rtlend && cnts(cblk->rtlend)
          && is_rjmp(cblk->rtlend->val)) {
#ifdef INTERACT
         if (!
             (swq && limited_scope && !cblk->inscope && !cblk->left->inscope))
#endif
         {
            target = cblk->right;
            jump = cblk->rtlend;
         }
      }
   }
#ifdef VET
#ifdef INTERACT
   end_phase(REVERSE_BRANCHES_TRANS);
#else
   end_phase(JUMPS_PHASE);
#endif
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_reverse_jumps", invocation))
      dump_function_i("after reverse_jumps()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS | SHOW_DFS);

   /*
    * Return indicating if any changes have been made to the control flow graph
    * of the function.
    */
   return change;
} /* end of reverse_jumps() */

/*
 * minimize_jumps - minimize the number of jumps by simple code reordering
 */
int minimize_jumps(int invocation)
{
   struct bblock *cblk, *start, *end;
   struct bblock *last;
   int   change;
#ifdef INTERACT
   extern int limited_scope;
#endif

#ifdef VET
#ifdef INTERACT
   begin_phase(BLOCK_REORDERING_TRANS);
#else
   begin_phase(REORDERBLOCKS_PHASE);
#endif
#endif
   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_minimize_jumps", invocation))
      dump_function_i("before minimize_jumps()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS);

   /*
    * Reset the DONE bit on all basic blocks so that we know which ones are
    * properly located, and which ones can be moved.
    */
   for (cblk = top; cblk; cblk = cblk->next)
      cblk->status &= ~DONE;

   change = FALSE;

   /*
    * Start at the entry block, which is always placed first in the output
    * order and scan the basic blocks in the current output order until we
    * locate an opportunity to get rid of an unconditional branch by
    * rearranging the output order of the basic blocks.  Note that we can do
    * this without having to update either the depth-first-search or the
    * dominator information, since we do not actually affect the flow of the
    * control graph in any way, we are just choosing a more compact way of
    * representing it.
    */
   for (cblk = top; cblk; cblk = cblk->down) {

      /*
       * Mark this basic block done so that it, along with all of the blocks
       * that have already been ordered, will not be moved again.
       */
      cblk->status |= DONE;

#ifdef INTERACT
      if (swq && limited_scope && !cblk->inscope)
         continue;
#endif

      /*
       * If this basic block has an unconditional jump to a basic block that
       * has not been placed in the output list yet, and it is not fallen
       * through to by anyone else, then we can remove the jump by placing the
       * basic block directly after this one.  We remove the jump here and set
       * the "go" flag to indicate that a change has been made in the
       * control-flow graph.
       */
      if (cblk->status & JUMP && (start = cblk->left) != ret_block
          && !(start->status & (FALLTHRU | DONE))) {
#ifdef VET
         begin_trans(VET_REQUIRED_TRANS);
#endif
         remv_links(cblk->rtlend);
         remv_rtl(cblk->rtlend);
         cblk->status &= ~JUMP;
         start->status |= FALLTHRU;
         change = TRUE;

         /*
          * If this just happened to be a useless jump, then there is no need
          * to move any basic blocks around.
          */
         if (start != cblk->down) {

            /*
             * Determine which basic blocks can be moved up under the current
             * basic block without interrupting any fall through points.
             */
            last = start;
            for (end = start->down; end && end->status & FALLTHRU;
                 end = end->down)
               last = end;

            /*
             * Now move the set of contiguous blocks below the current block
             * while leaving the original order as undisturbed as possible.
             */
            if (last->down = cblk->down)
               last->down->up = last;
#ifdef VET
            mod_bb_ptrs(MOD_DOWN, last, cblk->down);
#endif
            cblk->down = start;
#ifdef VET
            mod_bb_ptrs(MOD_DOWN, cblk, start);
#endif
            if (start->up->down = end)
               end->up = start->up;
#ifdef VET
            mod_bb_ptrs(MOD_DOWN, start->up, end);
#endif
            start->up = cblk;
         }
#ifdef VET
         end_trans();
#endif
      }
   }
#ifdef VET
#ifdef INTERACT
   end_phase(BLOCK_REORDERING_TRANS);
#else
   end_phase(REORDERBLOCKS_PHASE);
#endif
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_minimize_jumps", invocation))
      dump_function_i("after minimize_jumps()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS);

   /*
    * Return indicating if any changes have been made to the control flow graph
    * of the function.
    */
   return change;
} /* end of minimize_jumps() */

/*
 * merge_basic_blocks - merge those basic blocks that can be merged into one
 */
int merge_basic_blocks(int invocation)
{
   struct bblock *cblk, *fall;
   struct list *ptr;
   int   change;
#ifdef INTERACT
   extern int limited_scope;
#endif


#ifdef VET
#ifdef INTERACT
   begin_phase(MERGE_BLOCKS_TRANS);
#else
   begin_phase(MERGEBLOCKS_PHASE);
#endif
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_merge_basic_blocks", invocation))
      dump_function_i("before merge_basic_blocks()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS);

   /*
    * No changes have been made to the control flow graph yet
    */
   change = FALSE;

   /*
    * Search through the function for basic blocks that can be merged
    */
   cblk = top;
   while (cblk) {

      /*
       * If this basic block has no conditional branch and falls into the
       * following basic block, which contains only this basic block as its
       * predecessor, then the two basic blocks can be merged into one.
       */
      if (!cblk->right && (fall = cblk->left) && !(cblk->status & JUMP)
          && !fall->preds->next
#ifdef INTERACT
          && !(swq && limited_scope && !cblk->inscope && !fall->inscope)
#endif
       ) {

#ifdef VET
         begin_trans(VET_OPT_TRANS);
#endif
         /*
          * Add all of the RTLs in the "fall" block to the end of this basic
          * block.  Mark the new RTLs to indicate that they now belong to a new
          * basic block.  Remove these RTLs from the "fall" block to ensure
          * that they will not be returned to the heap along with the rest of
          * the basic block.
          */
         if (fall->rtls) {
            if (fall->rtls->prev = cblk->rtlend)
               cblk->rtlend->next = fall->rtls;
            else
               cblk->rtls = fall->rtls;
            cblk->rtlend = fall->rtlend;
            for (ptr = fall->rtls; ptr; ptr = ptr->next) {
               ptr->daddy = cblk;
#ifdef VET
#ifndef INTERACT
               move_rtl(ptr, fall, (struct list *) NULL, cblk);
#endif
#endif
            }
#ifdef INTERACT
            /*An alternative way is to call merge_bb instead of move_rtl
             * Reduced traffic */
            move_rtls(fall, cblk, fall->rtls, fall->rtlend);
#endif

            fall->rtls = fall->rtlend = (struct list *) NULL;

         }

         /*
          * If the "fall" block ended in an unconditional branch, then the
          * merged block does also.
          */
         if (fall->status & JUMP)
            cblk->status |= JUMP;

         /*
          * Now we need to fix the control flow up and get rid of the "fall"
          * block.
          */
         if (cblk->left = fall->left)
            change_predecessor(fall, cblk, fall->left);
#ifdef VET
         mod_bb_ptrs(MOD_LEFT, cblk, cblk->left);
#endif
         if (cblk->right = fall->right)
            change_predecessor(fall, cblk, fall->right);
#ifdef VET
         mod_bb_ptrs(MOD_RIGHT, cblk, cblk->right);
#endif

         /*
          * Uncouple the "fall" basic block from the output order list
          */
         if (cblk->down = fall->down)
            fall->down->up = cblk;
#ifdef VET
         mod_bb_ptrs(MOD_DOWN, cblk, cblk->down);
#endif
         /*
          * Uncouple the basic block from the depth-first-search order list
          */
         if (fall->prev_dfs)
            fall->prev_dfs->next_dfs = fall->next_dfs;
         if (fall->next_dfs)
            fall->next_dfs->prev_dfs = fall->prev_dfs;
         else
            bottom = fall->prev_dfs;

         /*
          * Return the "fall" basic block, which is no longer needed, to the
          * heap.
          */

         remv_basic_block(fall);

         change = TRUE;
#ifdef VET
         end_trans();
#endif
      }

      /*
       * Otherwise, go on to the next block to look for pairs of basic blocks
       * that can be merged.
       */
      else
         cblk = cblk->down;
   }

#ifdef VET
#ifdef INTERACT
   end_phase(MERGE_BLOCKS_TRANS);
#else
   end_phase(MERGEBLOCKS_PHASE);
#endif
#endif
   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_merge_basic_blocks", invocation))
      dump_function_i("after merge_basic_blocks()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS);

   /*
    * Return indicating if any changes have been made to the control flow graph
    * of the function.
    */
   return change;
} /* end of merge_basic_blocks() */

/*
 * depth_first_search - obtain a depth-first-search list of the data-flow graph
 */
void depth_first_search(int invocation)
{
   struct bblock *cblk;
   int   number_dfs;

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_depth_first_search", invocation))
      dump_function_i("before depth_first_search()", invocation,
                      SHOW_LABELS | SHOW_SUCCS | SHOW_DFS | SHOW_PREDS);

   /*
    * Reset the DONE bit on all basic blocks
    */
#ifdef INTERACT
   for (cblk = top; cblk; cblk = cblk->next)
#else
   for (cblk = top; cblk; cblk = cblk->down)
#endif
      cblk->status &= ~DONE;

   /*
    * Start a recursive depth-first-search.  When the search is finished, the
    * next_dfs/prev_dfs pointers in the basic blocks will contain the
    * depth-first-search order of the basic blocks.  The entry block is always
    * the first item in the list and the "bottom" pointer indicates the last
    * item in the list.
    */
   bottom = (struct bblock *) NULL;
   number_dfs = 1;
   dfs(top, (struct bblock *) NULL, &number_dfs);
   bottom->next_dfs = (struct bblock *) NULL;

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_depth_first_search", invocation))
      dump_function_i("after depth_first_search()", invocation,
                      SHOW_LABELS | SHOW_SUCCS | SHOW_DFS | SHOW_PREDS);
}

/*
 * dfs - recursive depth-first-search
 */
void dfs(struct bblock *cblk, struct bblock *parent_dfs,
         int *number_dfs)
{
   /*
    * Mark this basic block visited, remember its parent, give it the next
    * depth-first-search number and add it to the depth-first-search-list.
    */
   cblk->status |= DONE;
   cblk->parent_dfs = parent_dfs;
   cblk->bnum = (*number_dfs)++;
   if (bottom)
      bottom->next_dfs = cblk;
   cblk->prev_dfs = bottom;
   bottom = cblk;

   /*
    * Recursively visit the successors of this basic block in a
    * depth-first-search order.
    */
   if (cblk->left && !(cblk->left->status & DONE))
      dfs(cblk->left, cblk, number_dfs);
   if (cblk->right && !(cblk->right->status & DONE))
      dfs(cblk->right, cblk, number_dfs);
}

/*
 * add_fix_block - insert a basic block for fix code between two basic blocks
 */
struct bblock *add_fix_block(struct bblock *from, struct bblock *to)
{
   struct bblock *cblk, *nblk;
   char  jmp[MAXSTRING];

   /*
    * If the basic block that we are coming from has an unconditional jump or
    * falls through into the "to" basic block, then we can insert the transfer
    * RTLs at the end of the from block without affecting any other control
    * flow paths.
    */
   if (from->status & JUMP || !from->right)
      cblk = from;

   /*
    * Otherwise we have to create a new block to contain the transfer RTLs and
    * modify the control flow appropriately.
    */
   else if (from->left == to) {
      struct bblock *old_to_up = to->up;

      /*-
       * if we're falling through from "from" to "to" handle it special 
       *
       * WAIT!  Neither of these may be possible!  
       * if from->left==to, then then we can't modify
       * the branch properly.. but we also dont' want
       * to add a new jump to this block 
       */

      cblk = new_labeled_basic_block();
      /*
       *  Link the new basic block into the control-flow graph
       */
      cblk->left = cblk->down = to;
      cblk->up = to->up;
      cblk->right = NULL;
      to->up = cblk;
      from->down = cblk;

      if (cblk->prev_dfs = to->prev_dfs)
         cblk->prev_dfs->next_dfs = cblk;
      to->prev_dfs = cblk;

      to->status |= FALLTHRU;
      cblk->status |= FALLTHRU;

      add_predecessor(from, cblk);
      change_predecessor(from, cblk, to);


      from->left = cblk;
#ifdef VET
      mod_bb_ptrs(MOD_LEFT, cblk, to);
      mod_bb_ptrs(MOD_DOWN, cblk, to);
      mod_bb_ptrs(MOD_DOWN, old_to_up, cblk);
      mod_bb_ptrs(MOD_LEFT, from, cblk);
#endif

      if (from->left == to) {
         from->left = cblk;
#ifdef VET
         mod_bb_ptrs(MOD_LEFT, from, cblk);
#endif
      }
      else if (from->right == to) {
         from->right = cblk;
#ifdef VET
         mod_bb_ptrs(MOD_RIGHT, from, cblk);
#endif
      }


   }
   else if (to->label) {

      /*
       * Check to see if the user would like to see the state of the function
       * at this point in time.
       */
      if (swd && debug_opt_l("dump_before_add_fix_block", to->label))
         dump_function_l("before add_fix_block()", to->label,
                         SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                         SHOW_PREDS | SHOW_DFS);

      /*
       * Create a new, labeled basic block
       */
      cblk = new_labeled_basic_block();

      /*
       * Either modify an existing branch label or insert a new branch RTL to
       * transfer control to the appropriate RTL.
       * 
       */
      if (!change_target(from, to, cblk)) {
         (void) sprintf(jmp, "%s%s;", pc_equ, cblk->label);
         (void) add_rtl(from, (struct list *) NULL, jmp, rinit());
         from->status |= JUMP;
         from->left = cblk;
#ifdef VET
         mod_bb_ptrs(MOD_LEFT, from, from->left);
#endif
      }

      /*
       * If some other block already falls into the block that we have to
       * transfer control to, then we have to insert a branch in the basic
       * block that fell through to this one to avoid the fixup code that we
       * are about to introduce.  If the previous block contained a conditional
       * branch, then we have to build a whole new basic block to insert the
       * new jump RTL in.
       */
      if (to->status & FALLTHRU) {
         (void) sprintf(jmp, "%s%s;", pc_equ, to->label);
         if (to->up->right) {
            nblk = new_basic_block((char *) NULL);
            nblk->status |= FALLTHRU;
            change_predecessor(to->up, nblk, to);
            nblk->left = to->up->left;
#ifdef VET
            mod_bb_ptrs(MOD_LEFT, nblk, nblk->left);
#endif
            nblk->up = to->up;
            nblk->down = to;
#ifdef VET
            mod_bb_ptrs(MOD_DOWN, nblk, to);
#endif
            if (nblk->prev_dfs = to->prev_dfs)
               nblk->prev_dfs->next_dfs = nblk;
            nblk->next_dfs = to;
            to->up->left = to->up->down = nblk;
#ifdef VET
            mod_bb_ptrs(MOD_LEFT, to->up, nblk);
            mod_bb_ptrs(MOD_DOWN, to->up, nblk);
#endif
            add_predecessor(to->up, nblk);
            to->prev_dfs = to->up = nblk;
         }
         (void) add_rtl(to->up, (struct list *) NULL, jmp, rinit());
         to->up->status |= JUMP;
      }

      /*
       * Link the new basic block into the control-flow graph
       */
      {
         struct bblock *old_to_up;
         cblk->left = cblk->down = cblk->next_dfs = to;
         old_to_up = to->up;
         to->up = cblk;
#ifdef VET
         mod_bb_ptrs(MOD_LEFT, cblk, to);
         mod_bb_ptrs(MOD_DOWN, cblk, to);
#endif
         cblk->up = old_to_up;
         if (cblk->prev_dfs = to->prev_dfs)
            cblk->prev_dfs->next_dfs = cblk;
         to->prev_dfs = old_to_up->down = cblk;
#ifdef VET
         mod_bb_ptrs(MOD_DOWN, old_to_up, cblk);
#endif
      }
      to->status |= FALLTHRU;
      add_predecessor(from, cblk);
      change_predecessor(from, cblk, to);

      /*
       * Check to see if the user would like to see the state of the function
       * at this point in time.
       */
      if (swd && debug_opt_l("dump_after_add_fix_block", to->label))
         dump_function_l("after add_fix_block()", to->label,
                         SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                         SHOW_PREDS | SHOW_DFS);
   }

   /*
    * There is no reason to believe that a valid conflict should ever show up
    * at a block that has no label (since there must be a merge in the control
    * flow to create a conflict and there is no way to merge without a label),
    * so if we cannot find a label here, then we should issue an error message.
    */
   else {
      (void) fprintf(stderr,
                     "vpo (add_fix_block): conflict at an unlabled block\n");
      exit(EXIT_FAILURE);
   }

   /*
    * Return a pointer to the basic block that the fix code should be added to
    * the end of.
    */
   return cblk;
} /* end of add_fix_block() */

/*
 * build_dominator_tree - build the dominator tree for the function
 */
void build_dominator_tree(int invocation)
{
   struct bblock *cblk, *ublk, *vblk;
   struct blist *bptr;
#ifdef INTERACT
   void  check_cf(int inform);
#endif
   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_build_dominator_tree", invocation))
      dump_function_i("before build_dominator_tree()", invocation,
                      SHOW_LABELS | SHOW_BNUM | SHOW_DFS | SHOW_PARENT |
                      SHOW_SUCCS | SHOW_PREDS);

   /*
    * Initialize the semi-dominator pointer of each basic block and clear out
    * the non-pervasive status bits to indicate that the basic block has not
    * yet been processed.
    */
#ifdef INTERACT
   /*Added by Cai, Baosheng, 10/19/00 for debugging purposes */

#if 0
   /* cannot do this because check_cf causes vpo to crash */
   if (swq)
     check_cf(1);
#endif
#endif

   for (cblk = top; cblk; cblk = cblk->next) {
      cblk->status &= CLEAR;
      cblk->sdom = cblk;
#ifdef INTERACT
      cblk->idom = NULL;
#endif      
   }

   /*
    * This is an implementation of the Lengauer/Tarjan algorithm for finding
    * dominators in a flow graph.  Here we are performing steps two and three
    * of the algorithm.  We examine each basic block in reverse
    * depth-first-search order, skipping the entry block of the function.
    */
   for (cblk = bottom; cblk != top; cblk = cblk->prev_dfs) {

      /*
       * Determine the semidominator for this basic block.  The semidominator
       * is the basic block with the least depth-first-search number so that
       * there exists a path from the current basic block to the semidominator
       * where every basic block in the path (except for the start and the end
       * block) has a depth-first-search number greater than that of the
       * current basic block.
       */
      for (bptr = cblk->preds; bptr; bptr = bptr->next) {
         ublk = bdt_eval(bptr->ptr);
         if (ublk->sdom->bnum < cblk->sdom->bnum)
            cblk->sdom = ublk->sdom;
      }

      /*
       * Indicate that this basic block has been processed, so that it forms a
       * tree with its depth-first-search parent as its ancestor and indicate
       * that this basic block needs to be examined when it is time to look at
       * the blocks that have this semidominator.
       */
      cblk->status |= DONE | UNDONE;

      /*
       * Now locate and process every basic block that has the depth-first-
       * search parent of this basic block as its semidominator and has not yet
       * been processed by this step.  Unmark the basic blocks found so that we
       * do not examine them again in the future.
       */
      for (vblk = top; vblk; vblk = vblk->next) {
         if (vblk->status & UNDONE && vblk->sdom == cblk->parent_dfs) {
            vblk->status &= ~UNDONE;
            ublk = bdt_eval(vblk);
            vblk->idom =
             (ublk->sdom->bnum < vblk->sdom->bnum) ? ublk : cblk->parent_dfs;
         }
      }
   }

   /*
    * Now we complete the algorithm with step four, which fills in the
    * immediate dominators not computed in step three.
    */
   for (cblk = top->next_dfs; cblk; cblk = cblk->next_dfs) {
      if (cblk->idom && (cblk->idom != cblk->sdom))
#ifdef INTERACT
         if ((!cblk->idom) && swq)
            fprintf(stderr, "build_dominator_tree(): cblk->idom is NULL\n");
         else
#endif
            cblk->idom = cblk->idom->idom;
   }
   top->idom = (struct bblock *) NULL;

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_build_dominator_tree", invocation))
      dump_function_i("after build_dominator_tree()", invocation,
                      SHOW_LABELS | SHOW_BNUM | SHOW_DFS | SHOW_PARENT |
                      SHOW_DOM | SHOW_SUCCS | SHOW_PREDS);
} /* end of build_dominator_tree() */

/*
 * bdt_eval - return any minimum block in the path to the root of this tree
 */
struct bblock *bdt_eval(struct bblock *cblk)
{
   struct bblock *eblk;

   /*
    * If this basic block is the root of a tree, then return it
    */
   if (!(cblk->status & DONE))
      return cblk;

   /*
    * Otherwise, locate a basic block with the minimum depth-first-search
    * numbered semidominator item in the path from this item to the root of the
    * tree (other than the root itself) and return that basic block.
    */
   eblk = cblk;
   while ((cblk = cblk->parent_dfs) && cblk->status & DONE) {
      if (cblk->sdom->bnum < eblk->sdom->bnum)
         eblk = cblk;
   }
   return eblk;
}

/*
 * find_dominance_frontiers - determine the dominatance frontier of each block
 */
void find_dominance_frontiers(int invocation)
{
   struct bblock *cblk;
   struct blist *ptr;

   /*
    * Note that in order to determine the correct dominance frontier for each
    * basic block that the reverse depth-first-search order and the dominator
    * tree for the function must be up-to-date.
    */

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_find_dominance_frontiers", invocation))
      dump_function_i("before find_dominance_frontiers()", invocation,
                      SHOW_LABELS | SHOW_DFS | SHOW_DOM | SHOW_SUCCS |
                      SHOW_PREDS);

   /*
    * Visit each basic block in reverse depth-first-search order and determine
    * the dominance frontier of each basic block.
    */
   for (cblk = bottom; cblk; cblk = cblk->prev_dfs) {
      cblk->dom_front = NULL;
   }

   for (cblk = bottom; cblk; cblk = cblk->prev_dfs) {

      /*
       * First calculate the local dominance frontier set for each basic block.
       * This is done by looking at the successors of this basic block and, if
       * this basic block is not the immediate dominator of any of them, then
       * any such block belongs in the dominance frontier of this basic block.
       */
      if (cblk->left) {
         if (cblk->left->idom != cblk)
            add_frontier_basic_block(cblk, cblk->left);
      }
      if (cblk->right) {
         if (cblk->right->idom != cblk)
            add_frontier_basic_block(cblk, cblk->right);
      }

      /*
       * Now pass the dominance frontier items that belong to this basic block
       * upward to the basic block that immediately dominates this basic block.
       */
      if (cblk->idom) {
         for (ptr = cblk->dom_front; ptr; ptr = ptr->next)
            if (ptr->ptr->idom != cblk->idom)
               add_frontier_basic_block(cblk->idom, ptr->ptr);
      }
   }

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_find_dominance_frontiers", invocation))
      dump_function_i("after find_dominance_frontiers()", invocation,
                      SHOW_LABELS | SHOW_DFS | SHOW_DOM | SHOW_FRONTIER |
                      SHOW_SUCCS | SHOW_PREDS);
}

/*
 * find_loops - locate the loops in the function
 */
int find_loops(int invocation)
{
   struct bblock *cblk;
   int   found;

   /*
    * This function assumes that the dominator tree for the current function is
    * up-to-date.  This can be done by either calling build_dominator_tree()
    * before this function is invoked, or by ensuring that all of the changes
    * made to the control flow graph since the last time that the dominator
    * tree was built have updated the immediate dominator pointers correctly.
    */

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_find_loops", invocation))
      dump_function_i("before find_loops()", invocation,
                      SHOW_LABELS | SHOW_DOM | SHOW_SUCCS | SHOW_PREDS);

   /*
    * Find the natural loops in the function by locating the back edges in the
    * control flow graph.  A back edge is an edge that transfers control to a
    * basic block that dominates the basic block that the edge comes from.  We
    * consider two natural loops to belong to the same loop if they have the
    * same basic block as their head.
    */
   found = FALSE;
   for (cblk = top; cblk; cblk = cblk->down) {
      if (cblk->right && dominates(cblk->right, cblk)) {
         add_natural_loop(cblk, cblk->right);
         found = TRUE;
      }
      if (cblk->left && dominates(cblk->left, cblk)) {
         add_natural_loop(cblk, cblk->left);
         found = TRUE;
      }
   }

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_find_loops", invocation))
      dump_function_i("after find_loops()", invocation,
                      SHOW_LABELS | SHOW_FREQ | SHOW_LOOP | SHOW_DOM |
                      SHOW_SUCCS | SHOW_PREDS);

   /*
    * Indicate if any loops have been found
    */
   return found;
}

/*
 * add_natural_loop - add a natural loop to the list of loop structures
 */
void add_natural_loop(struct bblock *tail, struct bblock *head)
{
   /*
    * The back edge tells us that the head block is the entry point of a
    * natural loop.  If the head block has not been marked as the head of any
    * other loop (more than one natural loop may make up a loop), then we have
    * found a new loop and we need to allocate a new loop information structure
    * for it.
    */
   if (!head->loop) {
      head->freq++;
      head->loop =
       (struct loopl *) SALLOCTYPE(sizeof(struct loopl), LOOPLTYPE);
      head->loop->next = loops;
      loops = head->loop;
      head->loop->ptr = head;
      head->loop->preheader = (struct bblock *) NULL;
      head->loop->blocks = head->loop->exits = (struct blist *) NULL;
      head->loop->regs = vinit();
      head->loop->regs_in = vinit();
      head->loop->invareg = rinit();
      head->loop->stored = rinit();
      head->loop->loaded = rinit();
      head->loop->valid = FALSE;
      head->loop->invaread = FALSE;
      head->loop->control = NULL;
      head->loop->iterations[0] = '\0';
      head->loop->blockCount = 0;
   }

   /*
    * Add the head block to the list of blocks in the loop.  Now find all of
    * the other basic blocks that make up the natural loop by starting at the
    * tail (which is dominated by the head block) and recursively adding the
    * predecessors of each basic block in the loop to the loop until we reach
    * the head block.
    */
   (void) add_loop_block(head->loop, head);
   find_loop_blocks(head->loop, tail);
}

/*
 * find_loop_blocks - find all of the basic blocks that make up a natural loop
 */
void find_loop_blocks(struct loopl *loop, struct bblock *cblk)
{
   struct blist *ptr;

   /*
    * If this basic block is not already part of the loop, then add it and add
    * all of its predecessors to the loop as well.
    */
   if (add_loop_block(loop, cblk)) {
      cblk->freq++;
      for (ptr = cblk->preds; ptr; ptr = ptr->next)
         find_loop_blocks(loop, ptr->ptr);
   }
}

/*
 * add_loop_block - add a basic block to the list of blocks in a loop
 */
int add_loop_block(struct loopl *loop, struct bblock *cblk)
{
   struct blist *cptr, *pptr;

   /*
    * Check the loop block list to determine if the block has already been
    * added to the list.
    */
   pptr = (struct blist *) NULL;
   for (cptr = loop->blocks; cptr; cptr = cptr->next) {
      if (cptr->ptr == cblk)
         return (FALSE);
      pptr = cptr;
   }

   /*
    * If the block has not yet been added to the list, create a new list
    * structure item and append it to the end of the list.
    */
   cptr = (struct blist *) SALLOCTYPE(sizeof(struct blist), BLISTTYPE);
   cptr->ptr = cblk;
   cptr->next = (struct blist *) NULL;
   if (pptr)
      pptr->next = cptr;
   else
      loop->blocks = cptr;

   /* Add loop to the list of loops in which cblk is contained */
   add_loop_to_block(loop, cblk);

   /* Increment the count of blocks in the loop. */
   loop->blockCount += 1;
   return TRUE;
}

/*
 * block_in_loop - return TRUE if a basic block is part of a loop
 */
int block_in_loop(struct loopl *loop, struct bblock *cblk)
{
   struct listl *lptr;

   for (lptr = cblk->looplist; lptr; lptr = lptr->next) {
      if (lptr->ptr == loop)
         return TRUE;
   }
   return FALSE;
}

/*
 * any_block_in_loop - return TRUE if any block in a loop is also in a vector
 */
int any_block_in_loop(struct loopl *loop, dvector vec)
{
   struct blist *cptr;

   for (cptr = loop->blocks; cptr; cptr = cptr->next) {
      if (din(vec, cptr->ptr->bnum))
         return TRUE;
   }
   return FALSE;
}

/*
 * loop_block_count - return number of basic blocks within the loop.
 */
int loop_block_count(struct loopl *loop)
{
#if 0
   struct blist *cptr;
   int   count;

   assert(NULL != loop);

   for (cptr = loop->blocks; cptr; cptr = cptr->next)
      ++count;

   return count;
#else
   return loop->blockCount;
#endif
}

/*
 * innermost_containing_loop -  return innermost loop that contains both
 * of the given basic blocks.  Return NULL if the blocks do not share a loop.
 */
struct loopl *innermost_containing_loop(struct bblock *blk1,
                                        struct bblock *blk2)
{
   int   MinBlocks = INT_MAX - 5;
   int   CurrBlocks;
   struct loopl *ret_value = NULL;
   struct listl *lptr;

   if ((NULL == blk1) || (NULL == blk2))
      return ret_value;

   /* For each loop that contains blk1, see if it contains blk2.  If so,
    * see if that loop contains fewer blocks than our previous minimum.
    * Keep track of the minimum-block-count common loop and return it.
    * NOTE that irreducible flow graphs may produce strange results.
    */
   for (lptr = blk1->looplist; lptr; lptr = lptr->next) {
      if (block_in_loop(lptr->ptr, blk2)) {
         CurrBlocks = loop_block_count(lptr->ptr);
         if (CurrBlocks < MinBlocks) {
            MinBlocks = CurrBlocks;
            ret_value = lptr->ptr;
         }
      }
   }
   return ret_value;
}

/*
 * inner_loop_iterations - return number of times a basic block is executed.
 */
int inner_loop_iterations(struct bblock *cblk)
{
   struct listl *lptr;
   int   iters = 1;

   for (lptr = cblk->looplist; lptr; lptr = lptr->next) {
      assert(NULL != lptr->ptr);
      if (lptr->ptr->valid && isdigit(lptr->ptr->iterations[0]))
         iters *= atoi(lptr->ptr->iterations);
      else
         iters *= DEFAULT_LOOP_ITERATIONS;
   }
   assert(1 <= iters);
   return iters;
}

/*
 * minimize_loop_jumps - remove useless unconditional branch at end of loop
 */
int minimize_loop_jumps(int invocation)
{
   struct loopl *loop;
   struct blist *bptr;
   struct loopl *lptr;
   struct bblock *cblk;
   struct bblock *branch, *fix, *nptr;
   struct list *jmp;
   int   change, deleted, others;
   char  rtl[MAXSTRING], *a;
   int   failed;
#ifdef INTERACT
   char  oldrtl[MAXSTRING];
#endif

#ifdef VET
#ifdef INTERACT
   begin_phase(MINIMIZE_LOOP_JUMPS_TRANS);
#else
   begin_phase(OPTLOOPS_PHASE);
#endif
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_minimize_loop_jumps", invocation))
      dump_function_i("before minimize_loop_jumps()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_FREQ | SHOW_LOOP |
                      SHOW_SUCCS | SHOW_PREDS | SHOW_DFS);

   change = deleted = FALSE;

   /*-
    * Examine all loops.  What we are going to do in this section of code is
    * locate loops that look something like this:
    * 
    * ... 
    * L50:  
    * ...                     --top of loop (entry) 
    * ... PC = NZ >= 0,L52
    * L51:                              --conditional branch exits loop 
    * ... PC = L50;             --unconditional branch to top L50:  
    * L52:      ...
    * 
    * So we can change it to something like this:
    * 
    * ... 
    * PC = L50;               --Outside of loop--cheap! 
    * L00:      ...                     --New top of loop 
    * L50:  
    * ...                     --Old top of loop 
    * ... 
    * PC = NZ < 0, L00;       --Note that condition is reversed L51:  ...
    * L52:
    * 
    * Now we have one less unconditional branch executed each time though the
    * loop at the expense of one unconditional branch executed only once when
    * entering the loop.  On the average, it is a big win.
    */
   for (loop = loops; loop; loop = loop->next) {
      cblk = loop->ptr;

      /*
       * Locate a basic block that is part of this loop, has an unconditional
       * branch back to the top of the loop, is not the entry block to the loop
       * itself, and whose immediate predecessor has a conditional jump. Also,
       * check to be sure that we don't affect the structure of a loop with
       * other embedded loops in a way that has a pejorative impact on the
       * control flow of the inner loop.
       */
      for (bptr = cblk->preds; bptr; bptr = bptr->next)
         if (block_in_loop(loop, branch = bptr->ptr) && branch->status & JUMP
             && branch != cblk && (fix = branch->up)
             && (!(branch->status & FALLTHRU)
                 || (fix->rtlend && cnts(fix->rtlend)
                     && is_rjmp(fix->rtlend->val)
                     && branch->freq >= fix->freq)))
            break;

      /*
       * If this loop doesn't fit the profile, continue on to the next one
       */
      if (!bptr)
         continue;

      /*
       * If fix (the block we're going to reverse) has a target that is also
       * in the loop, the assumption that fix's branch is not-taken is not very
       * sound.  We came across an infinite loop because of this, so we'll 
       * avoid the "un-reverse" of fix if the target is also in the loop.
       */
      if (fix->right && block_in_loop(loop, fix->right))
         continue;

      /*
       * another fix to check if the branch we think we are going to reverse
       * is a conditional branch*/
      if(!(fix->rtlend && cnts(fix->rtlend) && is_rjmp(fix->rtlend->val)))
	continue;

      /*
       * We set change to TRUE to indicate that the flow of the function will
       * be seriously changed so that control flow analysis will need to be
       * re-calculated.
       */
      change = TRUE;

#ifdef VET
      begin_trans(VET_OPT_TRANS);
#endif
      /*
       * Unlink the unconditional branch basic block from the output list
       */
      fix->down = branch->down;
      if (fix->down != NULL)
         branch->down->up = fix;
#ifdef VET
      mod_bb_ptrs(MOD_DOWN, fix, branch->down);
#endif

      /*
       * Set "jmp" to point to the RTL containing the unconditional branch. If
       * there are other RTLs in the same block with the unconditional branch,
       * then they need to be moved up to the top of the loop.
       */
      if ((jmp = branch->rtlend)->prev) {

         /*
          * Remove the unconditional branch from the rest of the basic block
          */
         jmp->prev->next = (struct list *) NULL;

         /*
          * See if we fall through into the top of the loop
          */
         if (cblk->status & FALLTHRU) {

            struct bblock *old_jmp_daddy;
            /*
             * If we do fall through, but the RTL at the end of the lexically
             * previous block is a conditional branch, then we cannot just add
             * another branch to it, so we just create a new basic block and
             * put it in place so that we can add some code to it later on.
             */
            if (cblk->up->right) {

               struct bblock *old_cblk_up;
               /*
                * Create the new basic block and place it in the function
                * imediately before the entry block of the loop.
                */
               nptr = new_basic_block((char *) NULL);
               add_predecessor(nptr->up = cblk->up, nptr);
               nptr->down = nptr->next_dfs = nptr->left = cblk;
               old_cblk_up = cblk->up;
               cblk->up = nptr;
#ifdef VET
               mod_bb_ptrs(MOD_LEFT, nptr, cblk);
               mod_bb_ptrs(MOD_DOWN, nptr, cblk);
#endif
               old_cblk_up->down = old_cblk_up->left = nptr;
#ifdef VET
               mod_bb_ptrs(MOD_LEFT, old_cblk_up, nptr);
               mod_bb_ptrs(MOD_DOWN, old_cblk_up, nptr);
#endif
               /* update the dfs links so the new block is incorporated */
               if (nptr->prev_dfs = cblk->prev_dfs)
                  nptr->prev_dfs->next_dfs = nptr;
               cblk->prev_dfs = nptr;

               nptr->status |= FALLTHRU;
               /* locate cblk->up in cblk predecessor list */
               for (bptr = cblk->preds; bptr && bptr->ptr != nptr->up;
                    bptr = bptr->next) ;
               /* Change this predecessor entry to be the block just added. */
               if (bptr)
                  bptr->ptr = nptr;
               else {
                  (void) fprintf(stderr,
             "vpo (minimize_loop_jumps): cannot locate predecessor pointer\n");
                  exit(EXIT_FAILURE);
               }
            }

            /*
             * Mark the new last RTL for the block containing the unconditional
             * branch that we have removed.
             */
            branch->rtlend = jmp->prev;

            /*
             * Place the unconditional branch immediately before the top of the
             * loop, and change the status of the basic block that receives it
             * to indicate that it has an unconditional branch at its end.
             */
            if (jmp->prev = cblk->up->rtlend)
               jmp->prev->next = jmp;
            else
               cblk->up->rtls = jmp;
            cblk->up->rtlend = jmp;

            old_jmp_daddy = jmp->daddy;
            jmp->daddy = cblk->up;

#ifdef VET
            move_rtl(jmp, old_jmp_daddy, jmp->next, cblk->up);
#endif
            cblk->up->status |= JUMP;
         }
         else {

            /*
             * If the unconditional branch is in a basic block all by itself,
             * then we can just delete it from the block.  We also decrement
             * the number of items branching to the top of the loop when we do
             * so.
             */
            remv_links(jmp);
            remv_rtl(jmp);
         }

         /*
          * The basic block that we have removed the unconditional block for
          * must reflect that change in its status.
          */
         branch->status &= ~JUMP;

         /*
          * Now move the rest of the RTLs that came after the conditional
          * branch but before the unconditional branch to the top of the loop.
          * Since they are all in a single basic block, we can just move the
          * block up to the top of the loop.
          */
         if (branch->up = cblk->up) {
            cblk->up->down = branch;
#ifdef VET
            mod_bb_ptrs(MOD_DOWN, cblk->up, branch);
#endif
         }
         else
            top = branch;
         cblk->up = branch;
         branch->down = cblk;
#ifdef VET
         mod_bb_ptrs(MOD_DOWN, branch, cblk);
#endif

         /*
          * Since the block we moved up to the top has no jumps, it falls into
          * the block that used to be the top of the loop, so set that block's
          * status to indicate that it is fallen into.
          */
         cblk->status |= FALLTHRU;
      }

      /*
       * If the unconditional branch is in a block all by itself, we can just
       * get rid of this basic block since we don't need it anymore.  In this
       * case, alter the predecessor list of the entry block of the loop and
       * put the conditional branch block in place of the unconditional branch
       * block.
       */
      else {
         bptr->ptr = fix;
         if (branch->next_dfs)
            branch->next_dfs->prev_dfs = branch->prev_dfs;
         else
            bottom = branch->prev_dfs;
         if (branch->prev_dfs)
            branch->prev_dfs->next_dfs = branch->next_dfs;

         remv_basic_block(branch);
         branch = cblk;
         deleted = TRUE;
      }

      if (deleted || branch->status & FALLTHRU) {

         /*
          * If the new top of the loop does not have a label, we give it one
          */
         label_basic_block(branch);

         /*
          * Reverse the conditional branch at the end of the loop to jump back
          * to the top of the loop.
          */
#ifdef INTERACT
         if (swq)
            strcpy(oldrtl, fix->rtlend->val);
#endif
         reverse(fix->rtlend, branch->label);
#ifdef VET
#ifdef INTERACT
         mod_rtl(fix->rtlend, oldrtl, fix->rtlend->val);
#else
         mod_rtl(fix->rtlend);
#endif
#endif
         fix->left = fix->right;
#ifdef VET
         mod_bb_ptrs(MOD_LEFT, fix, fix->right);
#endif
         fix->right = branch;
#ifdef VET
         mod_bb_ptrs(MOD_RIGHT, fix, branch);
#endif

         /*
          * Unless we completely deleted the block that used to be at the end
          * of the loop, it is now at the top and we no longer fall through to
          * it.
          */
         if (deleted)
            deleted = FALSE;
         else
            branch->status &= ~FALLTHRU;

         /*
          * If the basic block whose condition we just reversed now falls
          * through out of the loop to the block that sits right under it, then
          * set the status of the block right under it to indicate that fact.
          */
         if (fix->down == fix->left)
            fix->down->status |= FALLTHRU;
         else {

            /*
             * If we are supposed to exit out of the loop to a basic block
             * other than the one that is right below us, then we have to add a
             * new basic block to the loop with an unconditional branch to get
             * us there.
             */
            (void) sprintf(rtl, "%s%s;", pc_equ, fix->left->label);
            branch = new_basic_block((char *) NULL);
            for (bptr = fix->left->preds; bptr && bptr->ptr != fix;
                 bptr = bptr->next) ;
            if (bptr)
               bptr->ptr = branch;
            else {
               (void) fprintf(stderr,
             "vpo (minimize_loop_jumps): cannot locate predecessor pointer\n");
               exit(EXIT_FAILURE);
            }
            if (branch->down = fix->down)
               fix->down->up = branch;
#ifdef VET
            mod_bb_ptrs(MOD_DOWN, branch, fix->down);
#endif
            if (branch->next_dfs = fix->next_dfs)
               branch->next_dfs->prev_dfs = branch;
            else
               bottom = branch;
            branch->up = branch->prev_dfs = fix;
            fix->down = fix->next_dfs = branch;
#ifdef VET
            mod_bb_ptrs(MOD_DOWN, fix, branch);
#endif
            branch->status |= (FALLTHRU | JUMP);
            (void) add_rtl(branch, (struct list *) NULL, rtl, rinit());
            branch->left = fix->left;
#ifdef VET
            mod_bb_ptrs(MOD_LEFT, branch, fix->left);
#endif
            fix->left = branch;
#ifdef VET
            mod_bb_ptrs(MOD_LEFT, fix, branch);
#endif
            add_predecessor(fix, branch);

            /*
             * Determine how many loops this new basic block is a part of. We
             * must do this in order to ensure that the remaining loop jump
             * minimization optimizations operate properly.
             */
            for (lptr = loops; lptr; lptr = lptr->next) {
               if (block_in_loop(lptr, fix)
                   && block_in_loop(lptr, branch->left)) {
                  branch->freq++;
                  (void) add_loop_block(lptr, branch);
               }
            }
         }
      }
#ifdef VET
      end_trans();
#endif
   }

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_between_minimize_loop_jumps", invocation))
      dump_function_i("between minimize_loop_jumps()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_FREQ | SHOW_LOOP |
                      SHOW_SUCCS | SHOW_PREDS | SHOW_DFS);

   /*-
    * Examine all loops.  This time we are looking for a loop that looks
    * something like this:
    * 
    * ...                       --code before loop 
    * PC = L50;                 --jump into loop 
    * L51:  
    * ... 
    * ... 
    * L50:      ...             --entry point of loop 
    * NZ = r[5] ? 25;           --check exit condition 
    * PC = NZ < 0, L51;         --loop again 
    * L52:    ...               --exit from loop
    * 
    * And transform it into:
    * 
    * ...                       --old code before loop 
    * ...                       --copied from old entry point of loop 
    * NZ = r[5] ? 25; 
    * PC = NZ >= 0, L52;        --note the reverse check L51:  
    * ...                       --now we just fall into the loop 
    * ... 
    * L50:      ...             --with luck, the label can go away 
    * NZ = r[5] ? 25;           --this is still here 
    * PC = NZ < 0, L51;         --and so is this 
    * L52:  
    * ...                       --exit from loop
    * 
    * At first glance, it only seems like we saved ourselves a single
    * unconditional branch.  Seems like a lot of work to go through for just
    * one branch.  If the loop is part of another loop, however, that one
    * branch may have been executed quite a few times, so we win. There's more
    * to it, though.  The CSE phase may be able to determine the value of r[5]
    * before entering the loop, thus with some manipulations it could determine
    * that the initial comparison is not required and that the conditional
    * branch may never be taken, so that it can be removed. This was not
    * possible with the previous layout of the loop.  So this new
    * transformation can help yield even better code.  Finally, note that the
    * label at the old entry point of the loop (L50) is almost always
    * referenced only by the initial unconditional branch into it.  This means
    * that it can go away, thus allowing us to merge the two basic blocks into
    * a single one.  On machines with instruction scheduling, this may permit
    * us to generate far better code than was possible with the previous
    * layout.  So this one innocent transformation is more desirable than a
    * first glance might suggest (and that's why we do it).
    */
   for (loop = loops; loop; loop = loop->next) {
      cblk = loop->ptr;

      /*
       * Look for an entry block that end with a conditional branch to a basic
       * block that is also in the loop and falls through to a block that is
       * outside of the loop.  Ensure that the conditional branch is something
       * that can be reversed.
       */
      if (!(branch = cblk->right) || !block_in_loop(loop, branch)
          || block_in_loop(loop, cblk->down) || !(fix = branch->up)
          || block_in_loop(loop, fix) || !(fix->status & JUMP)
          || fix->left != cblk || !cblk->rtlend || !cnts(cblk->rtlend)
          || !is_rjmp(cblk->rtlend->val))
         continue;

      /*
       * fix: if the jump before the loop to the loop header is a useless
       * jump then we do not need to do anything here, it should really be
       * handled by useless jump elimination
       */
      if(fix->down == branch)
	continue;

      /*
       * Now we check to see if there are any basic blocks that are not in the
       * loop but entered the loop through the original entry block.  If there
       * are, then we must change these predecessors to transfer control to the
       * new entry point through the check code that we will introduce in the
       * new loop preheader in order to avoid forming an irreducible flow
       * graph.
       */
      others = FALSE;
      failed = FALSE;
      for (bptr = cblk->preds; bptr; bptr = bptr->next) {
         if (!block_in_loop(loop, bptr->ptr)) {
            if (cblk == fix)
               continue;
            others = TRUE;

            /*
             * Later, this optimization will do:
             *    change_target(bptr->ptr, cblk, fix))
             * where bptr->ptr is a pred of cblk.  This isn't always possible,
             * so we check that it is.
             * If bptr->ptr has only one successor, we know we can deal with
             * it.  However, sometimes if it has two, we're hosed.
             */
            if (bptr->ptr->left && bptr->ptr->right) {
               /* Namely, if bptr->ptr->left == cblk */
               if (bptr->ptr->left == cblk) {
                  failed = TRUE;
               }
            }
         }
      }
      if (failed)
         continue;


#ifdef VET
      begin_trans(VET_OPT_TRANS);
#endif
      /*
       * Get rid of the unconditional branch to the entry point of the loop so
       * that we can now just fall into the top of the loop.
       */
      remv_links(fix->rtlend);
      remv_rtl(fix->rtlend);
      fix->status &= ~JUMP;

      /*
       * Remove the predecessor link of the old entry point of the loop
       */
      remv_predecessor(fix, cblk);

      /*
       * If we will be needing to transfer control to the copy of the exit code
       * that we are going to make then we must ensure that we have a preheader
       * block that has no other code.  In this case, the basic block pointed
       * to by "fix" is the new preheader block.  If we find no cases where we
       * will have to transfer control to the preheader, then it doesn't matter
       * if the current "fix" block has code.  If we do find instances where
       * the new code needs to be branched to, then we can use the current
       * "fix" block only if it is empty, otherwise, we must create a new basic
       * block and place it immediately below the current "fix" block.
       */
      if (others) {

         /*
          * We can use this basic block only if it is empty
          */
         if (fix->rtls) {

            /*
             * Create a new basic block with a label and insert it immediately
             * below the "fix" block.  Once the new preheader is inserted, we
             * make it the new "fix" block.
             */
            branch->up = nptr = new_labeled_basic_block();
            nptr->up = fix;
            if (nptr->prev_dfs = branch->prev_dfs)
               nptr->prev_dfs->next_dfs = nptr;
            nptr->next_dfs = nptr->down = branch;
#ifdef VET
            mod_bb_ptrs(MOD_DOWN, nptr, branch);
#endif
            branch->prev_dfs = fix->down = fix->left = nptr;
#ifdef VET
            mod_bb_ptrs(MOD_DOWN, fix, nptr);
            mod_bb_ptrs(MOD_LEFT, fix, nptr);
#endif
            nptr->status |= FALLTHRU;
            add_predecessor(fix, nptr);
            fix = nptr;
         }

         /*
          * Ensure that the preheader block has a label so that we can transfer
          * control to it.
          */
         else
            label_basic_block(fix);
      }

      /*
       * Change the control flow to indicate that the code before the loop now
       * falls right through into the top of the loop.
       */
      fix->left = branch;
#ifdef VET
      mod_bb_ptrs(MOD_LEFT, fix, branch);
#endif
      add_predecessor(fix, branch);
      branch->status |= FALLTHRU;

      /*
       * Indicate that the entry block of the loop has now changed
       */
      cblk->loop = (struct loopl *) NULL;
      branch->loop = loop;
      loop->ptr = branch;
      add_loop_to_block(loop, branch);
      change = TRUE;

      /*
       * The following code places a command to the CSE phase to trash (forget)
       * the value of the condition codes upon entry to the new loop.  This is
       * required because the loop transformation we are doing might cause CSE
       * to attempt to remove an RTL which sets the condition codes at the
       * start of the loop, but the code motion phase may then come around and
       * move code into the loop preheader which would, on most machines,
       * invalidate the value of the condition codes.
       */
      if (swm && swc && trash_cc) {
         a = trash_cc;
         while (*a) {
#ifdef INTERACT
            jmp =
             add_rtl_with_type(branch, branch->rtls, a, rinit(), TRASHLINE);
            mod_type(jmp);
#else
            jmp = add_rtl(branch, branch->rtls, a, rinit());
            jmp->type = TRASHLINE;
#endif
            while (*a++) ;
         }
      }

      /*
       * Copy the code from the entry point of the loop into the preheader.
       * Note that items of type DELAYLINE are never duplicated, since they are
       * strictly declarations that would cause the assembler to complain if
       * issued more than once.
       */
      for (jmp = cblk->rtls; jmp; jmp = jmp->next) {
         if (!(jmp->type & DELAYLINE))
            (void) copy_rtl(fix, (struct list *) NULL, jmp);
      }

      /*
       * Make sure that the block after the loop has a label and reverse the
       * conditional branch copied to the block before the loop to branch to
       * it.
       */
      label_basic_block(cblk->down);
#ifdef INTERACT
      if (swq)
         strcpy(oldrtl, fix->rtlend->val);
#endif
      reverse(fix->rtlend, cblk->down->label);
#ifdef VET
#ifdef INTERACT
      mod_rtl(fix->rtlend, oldrtl, fix->rtlend->val);
#else
      mod_rtl(fix->rtlend);
#endif
#endif
      fix->right = cblk->down;
#ifdef VET
      mod_bb_ptrs(MOD_RIGHT, fix, cblk->down);
#endif
      add_predecessor(fix, cblk->down);

      /*
       * Now look at all of the other predecessors to the old entry block of
       * the loop.  If we find any that come from outside of the loop, then we
       * will modify them to branch to the new entry block above the loop with
       * the reversed conditional.
       */
      if (others) {
         bptr = cblk->preds;
         while (bptr) {
            struct bblock *temp_ptr = 0;
            if (!block_in_loop(loop, bptr->ptr)) {
               if (!change_target(bptr->ptr, cblk, fix)) {

                  /* It might be the case that bptr->ptr falls through to "fix"
                   * thus we'll have to add an unconditional branch to the end
                   * in order to get rid of the "cblk" block
                   */
                  char  new_rtl[MAXSTRING];
                  if (bptr->ptr->down != fix) {
                     sprintf(new_rtl, "PC=%s;", fix->label);
                     add_rtl(bptr->ptr, NULL, new_rtl, rinit());
                     bptr->ptr->status |= JUMP;
                     bptr->ptr->left = fix;
                     if (bptr->ptr->down)
                        bptr->ptr->down->status &= ~FALLTHRU;
                  }
               }
               temp_ptr = bptr->ptr;
               remv_predecessor(temp_ptr, cblk);
               add_predecessor(temp_ptr, fix);
               bptr = cblk->preds;
            }
            else
               bptr = bptr->next;
         }
      }
#ifdef VET
      end_trans();
#endif
   }

#ifdef VET
#ifdef INTERACT
   end_phase(MINIMIZE_LOOP_JUMPS_TRANS);
#else
   end_phase(OPTLOOPS_PHASE);
#endif
#endif
   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_minimize_loop_jumps", invocation))
      dump_function_i("after minimize_loop_jumps()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_FREQ | SHOW_LOOP |
                      SHOW_SUCCS | SHOW_PREDS | SHOW_DFS);

   /*
    * Return indicating whether a change has been made to the function that
    * required the re-calculation of the control-flow information.
    */
   return change;
} /* end of minimize_loop_jumps() */

/*
 * find_loop_exits - examine loops to locate the exit blocks
 */
void find_loop_exits()
{
   struct bblock *cblk;
   struct loopl *loop;
   struct blist *bptr;

   /*
    * Look through all of the basic blocks in each loop in the function
    */
   for (loop = loops; loop; loop = loop->next) {
      for (bptr = loop->blocks; bptr; bptr = bptr->next) {
         cblk = bptr->ptr;

         /*
          * If any of the children of this block is outside of the loop that
          * this block is in, then that child belongs in the exit list of this
          * loop.  Do not add the same child more than once to the list.
          */
         if (cblk->left && !block_in_loop(loop, cblk->left))
            (void) add_exit_block(loop, cblk->left);
         if (cblk->right && !block_in_loop(loop, cblk->right))
            (void) add_exit_block(loop, cblk->right);
      }
   }
}

/*
 * add_exit_block - add a basic block to the list of exit blocks in a loop
 */
int add_exit_block(struct loopl *loop, struct bblock *cblk)
{
   struct blist *cptr, *pptr;

   /*
    * Check the loop block list to determine if the block has already been
    * added to the list.
    */
   pptr = (struct blist *) NULL;
   for (cptr = loop->exits; cptr; cptr = cptr->next) {
      if (cptr->ptr == cblk)
         return (FALSE);
      pptr = cptr;
   }

   /*
    * If the block has not yet been added to the list, create a new list
    * structure item and append it to the end of the list.
    */
   cptr = (struct blist *) SALLOCTYPE(sizeof(struct blist), BLISTTYPE);
   cptr->ptr = cblk;
   cptr->next = (struct blist *) NULL;
   if (pptr)
      pptr->next = cptr;
   else
      loop->exits = cptr;
   return TRUE;
}

/*
 * find_loop_preheaders - Find preheaders for all loops, if they exist. If
 *                        a loop does not have a natural preheader, then
 *                        return null for the preheader of that loop.
 */
void find_loop_preheaders()
{
  struct loopl *loop;
  extern struct loopl *loops;

  for(loop=loops ; loop ; loop=loop->next)
    loop->preheader = natural_preheader(loop->ptr);
}

/*
 * add_loop_to_block - Add loop to the the beginning of list of loops that 
 * contains the basic block
 */
void add_loop_to_block(struct loopl *loop, struct bblock *cblk)
{
   struct listl *lptr;

   lptr = (struct listl *) SALLOCTYPE(sizeof(struct listl), LISTLTYPE);
   lptr->ptr = loop;
   lptr->next = (struct listl *) NULL;
   if (cblk->looplist) {
      lptr->next = cblk->looplist;
   }
   cblk->looplist = lptr;
}

/*
 * order_loops - order loop information with deepest loops first
 */
void order_loops()
{
   struct loopl *loop, *last, *ins, *l;

   /*
    * Order the loop information structures so that the outer loops are placed
    * at the front of the list and inner loops towards the end.
    */
   if (loop = loops) {
      for (ins = loops->next; ins; ins = ins->next) {
         last = (struct loopl *) NULL;
         for (l = loops; l != ins; l = l->next) {
            if (l->ptr->freq < ins->ptr->freq) {
               loop->next = ins->next;
               ins->next = l;
               if (last)
                  last->next = ins;
               else
                  loops = ins;
               ins = loop;
               break;
            }
            else
               last = l;
         }
         loop = ins;
      }
   }

   /*
    * Using the list that groups loops from the outer loops to the inner loops,
    * set up a reverse list and let the "loope" variable point to the start of
    * this list.
    */
   last = (struct loopl *) NULL;
   for (loop = loops; loop; loop = loop->next) {
      loop->prev = last;
      last = loop;
   }
   loope = last;
}

/*
 * clear_loops - release all loop information structures
 */
void clear_loops()
{
   struct loopl *loop, *dloop;
   struct blist *bloop, *rloop;
   struct bblock *cblk;
   struct listl *lptr, *rptr;

   /*
    * Clear all loop information from the basic blocks in this function
    */
   for (cblk = top; cblk; cblk = cblk->next) {
      cblk->loop = (struct loopl *) NULL;
      cblk->freq = 0;
      /*
       * Release the  list of loops 
       */
      lptr = cblk->looplist;
      while (lptr) {
         rptr = lptr;
         lptr = lptr->next;
         FREETYPE(rptr, LISTLTYPE);
      }
      cblk->looplist = (struct listl *) NULL;

   }

   /*
    * Free all loop structure information
    */
   loop = loops;
   while (loop) {
      rfree(loop->invareg);
      rfree(loop->stored);
      rfree(loop->loaded);
      mrfree(loop->regs);

      /*
       * Release the basic blocks in loop list
       */
      bloop = loop->blocks;
      while (bloop) {
	rloop = bloop;
	bloop = bloop->next;
	FREETYPE(rloop, BLISTTYPE);
      }

      /*
       * Release the exit basic blocks of the loop list
       */
      bloop = loop->exits;
      while (bloop) {
         rloop = bloop;
         bloop = bloop->next;
         FREETYPE(rloop, BLISTTYPE);
      }

      /*
       * Return the loop structure to the heap
       */
      dloop = loop;
      loop = loop->next;
      FREETYPE(dloop, LOOPLTYPE);
   }

   loops = loope = (struct loopl *) NULL;
}

/*
 * estimate_execution_frequency - estimate how often a basic block is executed
 * and put the result in the "estimate" field in the block.
 * Returns TRUE if the control flow graph is irreducible.
 */
int estimate_execution_frequency(int invocation)
{
   struct bblock *cblk;
   struct blist *bptr;
   struct bblock *sblk;
   int   change, miss, go, succs;


   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd
       && debug_opt_i("dump_before_estimate_execution_frequency", invocation))
      dump_function_i("before estimate_execution_frequency()", invocation,
                      SHOW_LABELS | SHOW_FREQ | SHOW_LOOP | SHOW_SUCCS |
                      SHOW_PREDS);

   /*
    * Reset the DONE bit on all basic blocks and initialize all estimates to
    * zero
    */
   for (cblk = top; cblk; cblk = cblk->next) {
      cblk->status &= ~DONE;
      cblk->estimate = 0.0;
   }

   /*
    * Start by initializing the only block that we are sure about, the entry
    * basic block, to one.
    */
   top->estimate = 1.0;

   /*
    * Iterate through the basic blocks, updating our estimate values, until no
    * changes remain.
    */
   do {
      change = miss = FALSE;

      /*
       * Visit every basic block whose execution frequency estimate has not yet
       * been determined.
       */
      for (cblk = top; cblk; cblk = cblk->down) {
         if (!(cblk->status & DONE)) {

            /*
             * Ensure that we have obtained an execution estimate from every
             * predecessor for this basic block before we can determine that
             * the estimate for this basic block is complete.  Ignore back
             * edges at this stage.
             */
            go = TRUE;
            for (bptr = cblk->preds; bptr; bptr = bptr->next) {
               if (!dominates(cblk, bptr->ptr))
                  if (!(bptr->ptr->status & DONE))
                     go = FALSE;
            }

            /*
             * If we have obtained an execution freqency estimate for all of
             * the predecessors of this basic block, then we indicate that this
             * block's execution estimate has been appropriately calculated,
             * and we can then propagate the effects of this block's
             * execution estimate to all of its successors.
             */
            if (go) {
               cblk->status |= DONE;

               /*
                * Now we determine how many children this basic block has.  We
                * must make a special case for switch lookup tables.  Note that
                * these lookup tables are made up of a chain of basic blocks,
                * but that for the purposes of execution estimation we will
                * treat such a sequence as if it were one single basic block.
                */
               succs = 0;
               if (cblk->rtlend && (cblk->rtlend->type & CASELINE)) {
                  if (!dominates(cblk->right, cblk))
                     succs++;
                  for (sblk = cblk->left;
                       sblk->rtlend && (sblk->rtlend->type & CASELINE);
                       sblk = sblk->left) {
                     sblk->status |= DONE;
                     sblk->estimate = cblk->estimate;
                     if (!dominates(sblk->right, cblk))
                        succs++;
                  }
               }
               else {
                  if (cblk->left && !dominates(cblk->left, cblk))
                     succs++;
                  if (cblk->right && !dominates(cblk->right, cblk))
                     succs++;
               }

               /*
                * Now we take the estimated execution frequency for this basic
                * block, and divide it up amongst all of the successors that do
                * not dominate us (again, avoid those back edges).
                */
               if (cblk->rtlend && (cblk->rtlend->type & CASELINE)) {
                  if (!dominates(cblk->right, cblk))
                     cblk->right->estimate += (cblk->estimate / succs);
                  for (sblk = cblk->left;
                       sblk->rtlend && (sblk->rtlend->type & CASELINE);
                       sblk = sblk->left) {
                     if (!dominates(sblk->right, cblk))
                        sblk->right->estimate += (cblk->estimate / succs);
                  }
               }
               else {
                  if (cblk->left && !dominates(cblk->left, cblk))
                     cblk->left->estimate += (cblk->estimate / succs);
                  if (cblk->right && !dominates(cblk->right, cblk))
                     cblk->right->estimate += (cblk->estimate / succs);
               }

               /*
                * Since we have made some additional progress, perform at least
                * another iteration through this loop.
                */
               change = TRUE;
            }
            else
               miss = TRUE;
         }
      }

   } while (change);

   if (miss && !sww) {
#ifndef INTERACT   
      (void) fprintf(stderr,
                "vpo (estimate_execution_frequency): irreducible flow graph");
      if (fn)
         (void) fprintf(stderr, " in %s()\n", fn);
      else
         (void) fprintf(stderr, "\n");
#endif
   }
   
   /*
    * Take the static estimates and multiply them by the cumulative execution
    * count of the innermost loop containing each block
    * to obtain the actual execution frequency estimates used.
    */
   else {
      /* Update any iterations fields that are not yet valid. */
      struct loopl *loop;
      for (loop = loops; NULL != loop; loop = loop->next) {
         if (!loop->valid) {
            (void) iterations(loop, &loop->control, loop->iterations, 0, 0, 0, 0);
            loop->valid = TRUE;
         }
      }
      for (cblk = top; cblk; cblk = cblk->next) {
         if (cblk->freq)
            cblk->estimate *= inner_loop_iterations(cblk);
      }
   }

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd
       && debug_opt_i("dump_after_estimate_execution_frequency", invocation))
      dump_function_i("after estimate_execution_frequency()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_FREQ | SHOW_ESTIMATE |
                      SHOW_LOOP | SHOW_SUCCS | SHOW_PREDS);

   /*
    * Indicate whether or not the control flow graph for this function is
    * irreducible.
    */
   return miss;
} /* end of estimate_execution_frequency() */

/*
 * preheader - return pointer to preheader for loop, make one if need be
 */
struct bblock *preheader(struct bblock *cblk)
{
   struct blist *bptr;
   struct bblock *hblk, *pblk, *ltop, *lbottom;
   struct loopl *loop, *list[MAXLOOPS];
   struct blist *prev, *nptr;
   int   lind, i;
   char  jmp[MAXSTRING];

   /*
    * Check to see if the loop already has an appropriate preheader basic block
    * (many do) and return a pointer to it if we find one.
    */
   if (hblk = natural_preheader(cblk))
      return (hblk);

   /*
    * If we have not located a valid preheader basic block, then we perform a
    * simple sanity check to ensure that the basic block that we think is the
    * entry point of the loop has a valid label (otherwise, how could anyone
    * branch back to it, thus making the loop).
    */
   if (!cblk->label) {
      (void) fprintf(stderr, "vpo (preheader): loop header has no label\n");
      exit(1);
   }

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_l("dump_before_preheader", cblk->label))
      dump_function_l("before preheader()", cblk->label,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS);

   /*
    * Since we are definitely going to need a new basic block to form an an
    * appropriate loop preheader basic block, create one with a brand-new
    * label.
    */
   hblk = new_labeled_basic_block();

   /*
    * Now we need to locate an appropriate location to insert the new preheader
    * basic block.  The best location would be immediately before the loop
    * entry block so that we can simply fall through into it.  If the block
    * already before the loop entry already falls into the loop entry point and
    * it is part of the loop itself, then chances are that it is a highly
    * travelled path and we would probably affect the execution time of the
    * loop detrimentally if we where to place the preheader there.
    */
   pblk = cblk->up;
   if (cblk->status & FALLTHRU && block_in_loop(cblk->loop, pblk)) {

      /*
       * Failing to place the preheader right before the loop entry point, we
       * then search for a basic block that is not part of the loop yet jumps
       * directly to the loop entry point.  If we place the preheader basic
       * block immediately after this basic block, we can get rid of the branch
       * since we now wish to execute the loop preheader block before
       * transfering control to the loop entry point.
       */
      for (bptr = cblk->preds; bptr; bptr = bptr->next)
         if (!block_in_loop(cblk->loop, bptr->ptr) 
             && bptr->ptr->left == cblk) {
            pblk = bptr->ptr;
            break;
         }

      /*
       * If all attempts to find a location for the preheader basic block so
       * far fail, then we are forced to place the preheader basic block
       * anywhere.  If this is the case, then we attempt to place it after a
       * basic block that ends in an unconditional branch.  By doing so, we
       * minimize the impact of the new basic block placement on the control
       * flow and execution time of the function.
       */
      /* find the top and bottom of the loop in the control flow of the
       * function. It will be worthwhile to start the scan for such a block
       * from a point preceeding the loop in the control flow, to make some
       * loop structures consistent (added by Prasad)*/
      if (!bptr){
	ltop = lbottom = (struct bblock *)NULL;
	findLoopTopBottom(cblk->loop, &ltop, &lbottom);
	if(ltop && ltop->up)
	  pblk = ltop->up;
	else
	  pblk = top;
         for (; pblk->down && !(pblk->status & JUMP); pblk = pblk->down)
	   ;
      }
   }

   /*
    * Having found a location for the new preheader basic block, we link it in
    * to the existing control flow structure accordingly.
    */
   hblk->up = hblk->prev_dfs = pblk;
   if (hblk->down = pblk->down)
      pblk->down->up = hblk;
#ifdef VET
   mod_bb_ptrs(MOD_DOWN, hblk, pblk->down);
#endif
   if (hblk->next_dfs = pblk->next_dfs)
      hblk->next_dfs->prev_dfs = hblk;
   else
      bottom = hblk;
   pblk->down = pblk->next_dfs = hblk;
   hblk->left = cblk;
#ifdef VET
   mod_bb_ptrs(MOD_DOWN, pblk, hblk);
   mod_bb_ptrs(MOD_LEFT, hblk, cblk);
#endif

   /*
    * Now update the dominator tree by making the new preheader block the
    * immediate dominator of the loop entry block.
    */
   hblk->idom = cblk->idom;
   cblk->idom = hblk;

   /*
    * Determine which loops enclose the loop that we are building a preheader
    * basic block for.
    */
   lind = 0;
   for (loop = loops; loop; loop = loop->next) {
      if (block_in_loop(loop, cblk)) {
         if (lind == MAXLOOPS)
            (void) fprintf(stderr,
                           "vpo (preheader): loop nesting level exceeded\n");
         list[lind++] = loop;
      }
   }

   /*
    * Check to see if the block above the new preheader basic block falls
    * through into the preheader and update the status if it does.
    */
   if (!(pblk->status & JUMP))
      hblk->status |= FALLTHRU;

   /*
    * Check to see if this block makes an uncoditional jump to the loop entry
    * point (but its not part of the actual loop), in which case we can build
    * the preheader directly below it and just fall through.
    */
   if (pblk->status & JUMP && pblk->left == cblk
       && !block_in_loop(cblk->loop, pblk)) {
      hblk->status |= FALLTHRU;
      pblk->status &= ~JUMP;
      remv_links(pblk->rtlend);
      remv_rtl(pblk->rtlend);
   }

   /*
    * If we managed to place the preheader basic block immediately before the
    * loop entry, then we note the fact that we have a fall through condition,
    * otherwise, we must insert an unconditional branch RTL at the end of the
    * preheader basic block to tarnsfer control directly to the loop entry
    * point.
    */
   if (hblk->down == cblk)
      cblk->status |= FALLTHRU;
   else {
      (void) sprintf(jmp, "%s%s;", pc_equ, cblk->label);
      (void) add_rtl(hblk, (struct list *) NULL, jmp, rinit());
      hblk->status |= JUMP;
   }

   /*
    * Now we make any changes to the code that are required to execute the new
    * preheader basic block before entering the loop entry point.
    */
   prev = (struct blist *) NULL;
   bptr = cblk->preds;
   while (bptr) {

      /*
       * Look at the items that branch to the loop header from outside of the
       * loop.
       */
      if (!block_in_loop(cblk->loop, bptr->ptr)) {
         for (i = 0; i < lind; i++) {
            if (list[i])
               if (!block_in_loop(list[i], bptr->ptr))
                  list[i] = (struct loopl *) NULL;
         }
         if (!change_target(bptr->ptr, cblk, hblk)) {
            if (bptr->ptr->left == cblk) {
               bptr->ptr->left = hblk;
#ifdef VET
               mod_bb_ptrs(MOD_LEFT, bptr->ptr, hblk);
#endif
            }
            else {
               (void) fprintf(stderr,
                              "vpo (preheader): cannot locate jump to %s\n",
                              cblk->label);
               exit(EXIT_FAILURE);
            }
         }
         if (prev)
            prev->next = nptr = bptr->next;
         else
            cblk->preds = nptr = bptr->next;
         bptr->next = hblk->preds;
         hblk->preds = bptr;
         bptr = nptr;
      }

      /*
       * Leave items that branch to the loop header from inside the loop alone
       */
      else {
         prev = bptr;
         bptr = bptr->next;
      }
   }
   change_loop_exits(cblk, hblk);
   add_predecessor(hblk, cblk);

   /*
    * Since the new preheader basic block may now be a part of other loops
    * (since loops nest), then mark the preheader basic block as a member of
    * all the loops that it belongs to (it does not belong to the loop that it
    * is a preheader for, since, by definition, the preheader is only executed
    * once before entering a loop).
    */
   for (i = 0; i < lind; i++) {
      if (list[i])
         (void) add_loop_block(list[i], hblk);
   }

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_l("dump_after_preheader", cblk->label))
      dump_function_l("after preheader()", cblk->label,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS);

   /*
    * Return the pointer to the newly formed preheader basic block
    */
   return hblk;
} /* end of preheader() */

/*
 * natural_preheader - determine if a basic block has a natural preheader
 */
struct bblock *natural_preheader(struct bblock *cblk)
{
   struct bblock *hblk;
   struct blist *bptr;

   /*
    * A basic block is the natural preheader of a loop only if it is the only
    * basic block outside of the loop that transfers control to the entry block
    * of the loop and it either jumps unconditionally or falls directly into
    * the basic block.
    */
   hblk = (struct bblock *) NULL;
   for (bptr = cblk->preds; bptr; bptr = bptr->next) {
      if (!block_in_loop(cblk->loop, bptr->ptr)) {
         if (!hblk && (hblk = bptr->ptr)->left == cblk && !hblk->right)
            continue;
         else
            return ((struct bblock *) NULL);
      }
   }
   return hblk;
}

/*
 * change_loop_exits - update exit lists in current loop structures
 */
void change_loop_exits(struct bblock *old, struct bblock *new)
{
   struct blist *bptr;
   struct loopl *loop;
   int   found = FALSE;

   for (loop = loops; loop; loop = loop->next) {
      for (bptr = loop->exits; bptr; bptr = bptr->next) {
         if (bptr->ptr == old) {
            bptr->ptr = new;
            found = TRUE;
         }
      }
   }
   if (found)
      rcpy(&new->ulist, old->ulist);
}

/*
 * loop_exit - create a clean exit block from a loop
 */
struct bblock *loop_exit(struct loopl *loop, struct bblock *cblk)
{
   struct blist *bptr;
   struct bblock *hblk;
   struct blist *nptr, *prev;
   char  jmp[MAXSTRING];

   /*
    * If all of the predecessors of this basic block are in the loop that we
    * need an exit point out of, then this block is a natural exit block.
    */
   for (bptr = cblk->preds; bptr && block_in_loop(loop, bptr->ptr);
        bptr = bptr->next) ;
   if (!bptr)
      return cblk;

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_l("dump_before_loop_exit", cblk->label))
      dump_function_l("before loop_exit()", cblk->label,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_DOM |
                      SHOW_SUCCS | SHOW_PREDS);

   /*
    * Since the exit block is not a natural exit block, we have to make one.
    * First, allocate a new basic block and give it a unique label.
    */
   hblk = new_labeled_basic_block();
   /*
    * Now look at the predecessors of the old exit block.  All those that are
    * part of the loop that we are exiting must now go through the new basic
    * block instead.  All others can keep going just the way they were.
    */
   prev = (struct blist *) NULL;
   bptr = cblk->preds;
   while (bptr) {

      /*
       * Look at the items that branch to the exit basic block from within the
       * loop and change their target to branch to the new basic block.
       */
      if (block_in_loop(loop, bptr->ptr)) {
         nptr = bptr->next;
         if (prev)
            prev->next = nptr;
         else
            cblk->preds = nptr;
         bptr->next = hblk->preds;
         hblk->preds = bptr;
         if (!change_target(bptr->ptr, cblk, hblk)) {
            if (cblk->status & FALLTHRU && cblk->up == bptr->ptr) {
               if (bptr->ptr->left == cblk)
                  bptr->ptr->left = hblk;
               else {
                  assert(bptr->ptr->right == cblk);
                  bptr->ptr->right = hblk;
               }
#ifdef VET
               mod_bb_ptrs(MOD_LEFT, bptr->ptr, hblk);
#endif
            }
            else
               (void) fprintf(stderr,
                              "vpo (loop_exit): cannot locate label %s\n",
                              cblk->label);
         }
         bptr = nptr;
      }

      /*
       * Skip any items that transfer control to the exit basic block, but are
       * not inside the loop.
       */
      else {
         prev = bptr;
         bptr = bptr->next;
      }
   }

   /*
    * Now tie in the new basic block to the rest of the procedure so that
    * branches are minimized.
    */
   if (cblk->status & FALLTHRU && !block_in_loop(loop, cblk->up)) {
      (void) sprintf(jmp, "%s%s;", pc_equ, cblk->label);
      (void) add_rtl(cblk->up, (struct list *) NULL, jmp, rinit());
      cblk->up->status |= JUMP;
   }
   else
      cblk->status |= FALLTHRU;
   if (hblk->up = cblk->up) {
      cblk->up->down = hblk;
#ifdef VET
      mod_bb_ptrs(MOD_DOWN, cblk->up, hblk);
#endif
   }
   if (hblk->prev_dfs = cblk->prev_dfs)
      hblk->prev_dfs->next_dfs = hblk;
   cblk->up = cblk->prev_dfs = hblk;
   hblk->down = hblk->next_dfs = hblk->left = cblk;
#ifdef VET
   mod_bb_ptrs(MOD_DOWN, hblk, cblk);
   mod_bb_ptrs(MOD_LEFT, hblk, cblk);
#endif
   add_predecessor(hblk, cblk);

   /*
    * Indicate that this new exit block is now the immediate dominator of the
    * basic block that it transfers control to.
    */
   hblk->idom = cblk->idom;
   cblk->idom = hblk;

   /*
    * Fix up the loop's exits list to reflect the changes made in the loop exit
    * blocks.
    */
   for (bptr = loop->exits; bptr; bptr = bptr->next) {
      if (bptr->ptr == cblk)
         bptr->ptr = hblk;
   }

   if (hblk->up->left == hblk || hblk->up->right == hblk)
      hblk->status |= FALLTHRU;

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_l("dump_after_loop_exit", cblk->label))
      dump_function_l("after loop_exit()", cblk->label,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_DOM |
                      SHOW_SUCCS | SHOW_PREDS);

   return hblk;
}  /* end of loop_exit() */

/*
 * label_basic_block - make a new label for the basic block if it has none
 */
void label_basic_block(struct bblock *cblk)
{
   char  jmp[MAXSTRING];

   if (!cblk->label) {
      (void) sprintf(jmp, lab_mask, labnum++);
      update_label(cblk, strcpy(CALLOC(strlen(jmp) + 1), jmp));
   }
}

/*
 * new_labeled_basic_block - return a new basic block with a newly-made label
 */
struct bblock *new_labeled_basic_block()
{
   char *lptr, lab[MAXSTRING];

   (void) sprintf(lab, lab_mask, labnum++);
   lptr = strcpy(CALLOC(strlen(lab) + 1), lab);
   return new_basic_block(lptr);
}

/*
 * change_target - attempt to alter the target of a branch RTL
 */
int change_target(struct bblock *cblk, struct bblock *from, struct bblock *to)
{
   struct list *rtl, *dptr;
   char  jmp[MAXSTRING];

   /*
    * Locate the RTL that should contain the desired label and then make the
    * change required to the RTL to indicate the new target of the branch.
    * Return TRUE if the change is made appropriately.
    */
   if (rtl = find_control_rtl(cblk, &dptr, FALSE)) {
      (void) strcpy(jmp, rtl->val);
      if (fmatch(jmp, from->label) && (*epos > '9' || *epos < '0')) {
         (void) sub(bpos, epos, to->label);
         if (cblk->right == from) {
            cblk->right = to;
#ifdef VET
            mod_bb_ptrs(MOD_RIGHT, cblk, to);
#endif
         }
         else {
            cblk->left = to;
#ifdef VET
            mod_bb_ptrs(MOD_LEFT, cblk, to);
#endif
         }
         update_rtl(rtl, jmp);
         set_rtl_sides(rtl);
         return TRUE;
      }
   }

   /*
    * Return FALSE only if we could not locate the RTL that contains the
    * correct label.
    */
   return FALSE;
}

/*
 * find_control_rtl - attempt to find the control RTL in a basic block
 */
struct list *find_control_rtl(struct bblock *cblk, struct list **dptr,
                              int for_removal)
{
   struct list *ptr;

   /*
    * Start at the last item in the basic block and search upward for the first
    * RTL.  In cases where branch instructions have delay slots, the first RTL
    * that we locate might be a delayed instruction.  In such cases, we set
    * "dptr" to point to any such instruction and continue our search for the
    * branch instruction.
    */
   *dptr = (struct list *) NULL;
   for (ptr = cblk->rtlend; ptr; ptr = ptr->prev) {

      /*
       * A switch statament lookup table item is responsible for the control
       * flow of this basic block.
       */
      if (ptr->type & CASELINE)
         return ptr;

      /*
       * Ignore any non-RTL items
       */
      if (!cnts(ptr))
         continue;

      /*
       * If the item is a delayed instruction, then remember it.  If the
       * "for_removal" argument is set, then we can skip this delayed item only
       * if it would be safe to take it from its delay slot by removing the
       * item that forms the slot.
       */
      if (((ptr->kind & (TAKEN | NOTAKEN)) == (TAKEN | NOTAKEN))
          || (!for_removal && (ptr->kind & (TAKEN | NOTAKEN)))) {

         if (*dptr) {
            (void) fprintf(stderr,
                           "vpo (find_control_rtl): multiple delay items\n");
            exit(EXIT_FAILURE);
         }

         *dptr = ptr;
         continue;
      }

      /*
       * If the RTL updates the program counter, then it is a control
       * instruction, otherwise, if the last non-delayed item in the basic
       * block is not a control instruction, then we can be sure that there is
       * no such instruction in this basic block.
       */
      return ((is_pc(ptr->val)) ? ptr : (struct list *) NULL);
   }

   /*
    * We did not locate any control instruction in this basic block
    */
   return ((struct list *) NULL);
} /* end of find_control_rtl() */

#ifdef MEASURE

/*
 * reveal_exit_points - determine the impact of "exit()" calls to the function
 */
void reveal_exit_points(int invocation)
{
   struct bblock *cblk;
   struct list *ptr;
   int first;

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_reveal_exit_points", invocation))
      dump_function_i("before reveal_exit_points()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS | SHOW_DFS);

   /*
    * Consider each basic block in the current function
    */
   for (cblk = top; cblk; cblk = cblk->down) {

      /*
       * Cut off blocks that have a call to a function that terminates the
       * execution of the program so that accurate measurements can be obtained
       * since some execution counts are inferred from other basic blocks.
       */
      for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
         if (cnts(ptr)) {
            if (is_call(ptr->val) && callsexit(ptr)) {

               /*
                * If this basic block falls into another basic block, then
                * update the control flow information to show that the basic
                * block is not fallen into any more.
                */
               if (cblk->left) {
                  remv_predecessor(cblk, cblk->left);
                  if (!(cblk->status & JUMP))
                     cblk->left->status &= ~FALLTHRU;
                  else
                     cblk->status &= ~JUMP;
                  cblk->left = (struct bblock *) NULL;
               }

               if (cblk->right) {
                  remv_predecessor(cblk, cblk->right);
                  cblk->right = (struct bblock *) NULL;
               }

               /*
                * Any RTLs that remain in the basic block are never executed,
                * so we go through and remove them.  Be careful not to remove
                * any instruction that may be in the delay slot of the call.
                */
               first = TRUE;
               for (ptr = ptr->next; ptr; ptr = ptr->next) {
                  if (cnts(ptr)) {
                     if (!first || !(ptr->kind & (TAKEN | NOTAKEN))) {
                        remv_links(ptr);
                        remv_rtl(ptr);
                     }
                     first = FALSE;
                  }
               }

               break;
            }
         }
      }
   }

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_reveal_exit_points", invocation))
      dump_function_i("after reveal_exit_points()", invocation,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP | SHOW_SUCCS |
                      SHOW_PREDS | SHOW_DFS);
} /* end of reveal_exit_points() */

/*
 * equivalence_classes - which blocks are in the same equivalence class
 */
void equivalence_classes()
{
   struct bblock *cblk, *fblk;
   dvector dptr;
   int   change;

   /*
    * Number the basic blocks and calculate the length of the vectors required
    * to reference them.
    */
   (void) number_basic_blocks();

   /*
    * Throw away the current dominator information so that we can determine the
    * always successor information using the dominator vectors.
    */
   clear_dominator_vectors();

   /*
    * Initialize all blocks except those that exit from the function as having
    * all blocks always succeding them.  Mark exit blocks as always being
    * succeeded only by themselves.
    */
   for (cblk = top; cblk; cblk = cblk->next) {
      if (!cblk->left && !cblk->right)
         dinsert(&cblk->dom, cblk->bnum);
      else
         cblk->dom = dall();
   }

   /*
    * Iterate to determine the always successor information for each basic
    * block.  The always successors are those blocks that will always be
    * executed later if this block is executed now.
    */
   dptr = dall();
   do {
      change = FALSE;
      for (cblk = bottom; cblk; cblk = cblk->prev_dfs) {
         dcpy(&dptr, cblk->dom);
         if (cblk->left)
            dinter(cblk->dom, cblk->left->dom);
         if (cblk->right)
            dinter(cblk->dom, cblk->right->dom);
         dinsert(&cblk->dom, cblk->bnum);
         if (!dcmp(cblk->dom, dptr))
            change = TRUE;
      }
   } while (change);

   /*
    * For each block, initialize the equivalence class vector to contain the
    * set of blocks that are we are always successors of.
    */
   for (cblk = top; cblk; cblk = cblk->down) {
      for (fblk = top; fblk; fblk = fblk->down) {
         if (din(fblk->dom, cblk->bnum))
            dinsert(&cblk->equivclass, fblk->bnum);
      }
   }

   /*
    * Calculate the dominator information for the basic blocks in this
    * function.
    */
   clear_dominator_vectors();
   find_dominators();

   /*
    * Determine the equivalence classes by determining which of the blocks that
    * this block is always a successor of dominate us and are not in a
    * different loop.  Note that this results in equivalence classes for each
    * basic block that are made up only of blocks that dominate us.
    */
   for (cblk = top; cblk; cblk = cblk->down) {
      dinter(cblk->equivclass, cblk->dom);
      for (fblk = top; fblk; fblk = fblk->down) {
         if (fblk != cblk && din(cblk->equivclass, fblk->bnum))
            if (in_different_loops(fblk, cblk))
               ddelete(&cblk->equivclass, fblk->bnum);
      }
   }

   /*
    * Since we only have predecessor blocks in the same equivalence class, we
    * need to indicate all other blocks in the equivalence class.
    */
   for (cblk = top; cblk; cblk = cblk->down) {
      for (fblk = top; fblk; fblk = fblk->down) {
         if (cblk != fblk) {
            if (din(cblk->equivclass, fblk->bnum))
               dinsert(&fblk->equivclass, cblk->bnum);
            else if (din(fblk->equivclass, cblk->bnum))
               dinsert(&cblk->equivclass, fblk->bnum);
         }
      }
   }
}

#endif /* MEASURE */

#if defined(MEASURE) || defined(INTERACT)

/*
 * find_dominators - determine which blocks dominate each basic block
 */
void find_dominators()
{
   struct bblock *cblk;
   dvector dptr;
   int   change;

   /*
    * Initialize the dominator vectors so that initially the entry basic block
    * dominates only itself and all other blocks are dominated by all of the
    * other blocks.
    */
   dinsert(&top->dom, top->bnum);
   for (cblk = top->next; cblk; cblk = cblk->next)
      cblk->dom = dall();

   /*
    * Iterate over the basic blocks until the dominator information does not
    * change.
    */
   dptr = dall();
   do {
      change = FALSE;
      for (cblk = top; cblk; cblk = cblk->down) {
         if (cblk->left) {
            dcpy(&dptr, cblk->left->dom);
            dinter(cblk->left->dom, cblk->dom);
            dinsert(&cblk->left->dom, cblk->left->bnum);
            if (!dcmp(cblk->left->dom, dptr))
               change = TRUE;
         }
         if (cblk->right) {
            dcpy(&dptr, cblk->right->dom);
            dinter(cblk->right->dom, cblk->dom);
            dinsert(&cblk->right->dom, cblk->right->bnum);
            if (!dcmp(cblk->right->dom, dptr))
               change = TRUE;
         }
      }
   } while (change);

   dfree(dptr);
} /* end of find_dominators() */
#endif /* MEASURE || INTERACT */

#ifdef MEASURE

/*
 * clear_dominator_vectors - clear dominator information
 */
void clear_dominator_vectors()
{
   struct bblock *cblk;

   for (cblk = top; cblk; cblk = cblk->next) {
      dfree(cblk->dom);
      cblk->dom = dinit();
   }
}

#endif


/*
 * in_different_loops - determines if two blocks are in any different loops
 */
int in_different_loops(struct bblock *ptr1, struct bblock *ptr2)
{
   if (find_loop(ptr1, ptr2))
      return TRUE;
   return (find_loop(ptr2, ptr1));
}


/*
 * find_loop - look for path from ptr1 back to ptr1 that avoids ptr2
 */
int find_loop(struct bblock *ptr1, struct bblock *ptr2)
{
   struct bblock *cblk;

   /*
    * Reset DONE bits from each basic block
    */
   for (cblk = top; cblk; cblk = cblk->down)
      cblk->status &= ~DONE;

   if (ptr1->left && find_cycle(ptr1->left, ptr1, ptr2))
      return TRUE;
   if (ptr1->right && find_cycle(ptr1->right, ptr1, ptr2))
      return TRUE;
   return FALSE;
}

/*
 * find_cycle - look for a path to ptr1 that avoids ptr2
 */
int find_cycle(struct bblock *ptr, struct bblock *ptr1, struct bblock *ptr2)
{
   if (ptr == ptr1)
      return TRUE;

   if (ptr == ptr2 || ptr->status & DONE)
      return FALSE;

   ptr->status |= DONE;

   if (ptr->left && find_cycle(ptr->left, ptr1, ptr2))
      return TRUE;

   if (ptr->right && find_cycle(ptr->right, ptr1, ptr2))
      return TRUE;

   return FALSE;
}

/*
 * rtl_dominates - check to see if one RTL dominates another
 */
int rtl_dominates(struct list *rtl1, struct list *rtl2)
{
   struct list *ptr;

   /*
    * First, check to see if the two RTLs  are in the same block and if they
    * are just look at the basic block to see if rtl1 comes before rtl2.
    */
   if (rtl1->daddy == rtl2->daddy) {
      for (ptr = rtl2; ptr; ptr = ptr->prev)
         if (ptr == rtl1)
            return TRUE;
      return FALSE;
   }

   /*
    * Otherwise, check to see if the basic block containing the first RTL
    * dominates the basic block containing the second RTL.
    */
   return dominates(rtl1->daddy, rtl2->daddy);
}

/*
 * dominates - check to see if one basic block dominates another
 */
int dominates(struct bblock *blk1, register struct bblock *blk2)
{
   /*
    * Check to see if the first block is a (not necessarily proper) ancestor of
    * the second block in the dominator tree.
    */
   while (blk2) {
      if (blk1 == blk2)
         return TRUE;
      blk2 = blk2->idom;
   }
   return FALSE;
}

/* Return true if there is only 1 useful RTL.
 * Only comment lines are useless RTLs.
 */
int only_one_useful_rtl(struct bblock *cblk)
{
   struct list *ptr;
   int   nurtls = 0;
   for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
      if (cnts(ptr))
         nurtls++;
      if (nurtls > 1)
         return FALSE;
   }

   return (nurtls == 1);
}

int is_chain(struct bblock *cblk)
{
#ifdef INTERACT
  extern int limited_branch_chaining;
#endif  
   struct list *ptr;
   struct loopl *loop;
   int   nurtls = 0;
   extern struct loopl *loops;
   
   for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
      if (cnts(ptr))
         nurtls++;
      if (ptr->type & ASMLINE)
         return FALSE;
      if (ptr->type & SETLINE)
         return FALSE;
      if (nurtls > 1)
         return FALSE;
   }

   if (nurtls == 0)
      return TRUE;
#ifdef INTERACT
   /*I wish to be able to remove only those blocks which are empty, but are
     caught during branch chaining. Adding code here seems to be the easiest
     fix */
   if(limited_branch_chaining)
     return FALSE;

   /*preheader blocks cannot form parts of branch chains since these may have
     been explicitly created during loop optimizations. Removing them here
     may cause the compiler to go into an infinite loop*/
   for(loop=loops ; loop ; loop=loop->next){
     if(cblk == loop->preheader)
       return FALSE;
   }
#endif   
   
   return ((cblk->status & JUMP) && (nurtls == 1));
}


int inside_some_loop(struct bblock *blk)
{
   struct loopl *loop;
   for (loop = loops; loop; loop = loop->next)
      if (block_in_loop(loop, blk))
         return TRUE;

   return FALSE;
}

void test_next_list()
{
   struct bblock *cblk;
   for (cblk = top; cblk; cblk = cblk->down)
      cblk->status |= DONE;
   for (cblk = top; cblk; cblk = cblk->next)
      cblk->status &= ~DONE;
   for (cblk = top; cblk; cblk = cblk->down) {
      if (cblk->status & DONE) {
         printf("Block %x with label %s not on 'next' list\n", cblk,
                cblk->label);
      }
   }
   fflush(stdout);
}


/*
 * check_loop_exits - Return FALSE is any loop exit of the arugument loop
 * ends with an assembly line, and has a left child follow through block and 
 * a right child.
 */
int check_loop_exits(struct loopl *loop)
{
   struct blist *blklist;
   for (blklist = loop->exits; blklist; blklist = blklist->next) {
      struct bblock *the_exit = blklist->ptr;
      if (the_exit->rtlend && (the_exit->rtlend->type & ASMLINE)
          && the_exit->left == the_exit->down && the_exit->right)
         return FALSE;
   }
   return TRUE;
}
