/************************************************************************
 *                                                                      *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   *
 *     The Rector and Visitors of the University of Virginia            *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from the University of Virginia.                                *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/controlflow.h,v $
 $Date: 2003/11/18 01:55:51 $
 $Revision: 1.11 $
 $Name:  $

 Log of changes:

 $Log: controlflow.h,v $
 Revision 1.11  2003/11/18 01:55:51  clc5q
 Move DEFAULT_LOOP_ITERATIONS from controlflow.c to controlflow.h for use
     in local and global profiling.

 Revision 1.10  2003/04/23 17:50:28  clc5q
 Add prototype for new function inner_loop_iterations().

 Revision 1.9  2003/04/09 20:54:54  clc5q
 Link each ignode and basic block to the loops that contain it.

 Revision 1.8  2003/03/18 17:18:34  jdh8d
 *** empty log message ***

 Revision 1.7  2002/07/16 21:45:56  clc5q
 Export CHAIN() macro for thumb target.

 Revision 1.6  2002/07/11 15:20:48  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.5  2002/06/05 20:08:53  jdh8d
 Changes to support the stache, including routines to build a callgraph.

 Some cosmetic changes.

 Minor bug fixes.

 Changed build scripts to build the mips compiler for SimpleScalar if
 SSAS is set to "-DSS_AS"

 Revision 1.4  1999/01/06 15:34:51  jwd
 Added prototype for fix_control_flow. This fixes control flow for
 boolean expressions.
 Revision 1.3  1998/06/26 20:55:45  jwd
 Added prototypes that are used by the ease code.
 This were not added when the code was converted to ANSI.
 Revision 1.2  1998/04/28 20:09:13  jwd
 Added void to some function prototypes (not sure why they
 were missing).
 
 Also deleted some commented out code.
 Revision 1.1  1998/04/03 16:53:32  jwd
 Initial revision

*************************************************************************/

#ifndef CONTROLFLOW_H
#define CONTROLFLOW_H


/*
 * CHAIN macro evaluates to TRUE if the basic block either has no RTLs, thus
 * being part of a branch chain to its predecessor or if it contains only an
 * unconditional jump.
 */
 
 /*#define CHAIN(cblk) (!(cblk)->rtls || \
   ((cblk)->status & JUMP && (cblk)->rtls == (cblk)->rtlend))*/
 
 /*#define CHAIN(cblk) (!(cblk)->rtls || \
   ((cblk)->status & JUMP && only_one_useful_rtl((cblk))))*/

extern int is_chain(struct bblock *);

#define CHAIN(cblk) is_chain((cblk))

/* We assume each loop executes this many times, unless we know otherwise. */
#define DEFAULT_LOOP_ITERATIONS 10

#ifdef MEASURE
void
reveal_exit_points(int invocation);
#endif

extern struct loopl *loops; /* start of loop list */
extern struct loopl *loope; /* end of loop list */
extern int labnum;          /* counter used to make new labels  */

int fix_control_flow(int); 
int remv_branch_chains(int);
int remv_useless_basic_blocks(int);
int remv_useless_jumps(int);
int remv_unreachable_code(int);
void mark_reachable_code(struct bblock *);
int reverse_jumps(int);
int minimize_jumps(int);
int merge_basic_blocks(int);
void depth_first_search(int);
void dfs(struct bblock *, struct bblock *, int *);
struct bblock *add_fix_block(struct bblock *, struct bblock *);
void build_dominator_tree(int);
struct bblock *bdt_eval(struct bblock *);
void find_dominance_frontiers(int);
int find_loops(int);
void add_natural_loop(struct bblock *, struct bblock *);
void find_loop_blocks(struct loopl *, struct bblock *);
int add_loop_block(struct loopl *, struct bblock *);
int block_in_loop(struct loopl *, struct bblock *);
void add_loop_to_block(struct loopl *, struct bblock *); 
int any_block_in_loop(struct loopl *, rlist);
int inner_loop_iterations(struct bblock *);
int minimize_loop_jumps(int);
void find_loop_exits(void);
int add_exit_block(struct loopl *, struct bblock *);
void order_loops(void);
void clear_loops(void);
int estimate_execution_frequency(int);
struct bblock *preheader(struct bblock *);
struct bblock *natural_preheader(struct bblock *);
void change_loop_exits(struct bblock *, struct bblock *);
struct bblock *loop_exit(struct loopl *, struct bblock *);
void label_basic_block(struct bblock *);
struct bblock *new_labeled_basic_block(void);
int change_target(struct bblock *, struct bblock *, struct bblock *);
struct list *find_control_rtl(struct bblock *, struct list **, int);
int rtl_dominates(struct list *, struct list *);
int dominates(struct bblock *, struct bblock *);
int check_loop_exits(struct loopl *);
void remove_block(struct bblock *cblk);
#ifdef MEASURE
void find_dominators();
void clear_dominator_vectors();
#endif
#endif
