
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "vpo.h"
#include "assign.h"
#include "cdmotion.h"
#include "cse.h"
#include "combine.h"
#include "controlflow.h"
#include "database.h"
#include "dataflow.h"
#include "eod.h"
#include "locals.h"
#include "links.h"
#include "misc.h"
#include "vectors.h"
#include "vpo2.h"
#include "view.h"

#include "regs2.h"   /* Machine Dependent */
#include "rtl.h"     /* Machine Dependent */
#include "y.tab2.h"  /* Tools Dependent */


int 
find_common_instructions (struct bblock* cblk, struct list* jmp_rtl,
                          char used_regs[], char set_regs[]);

int 
cross_jump(int invocation);

int 
cross_jumping(int invocation)
{
 	int change=FALSE;
	if(debug_opt("no_cross_jump"))
		return change;
#ifndef CROSS
        return 0;

#else
	
	change=unification(invocation);
	change|=cross_jump(invocation);
	return change;
#endif
}
int unification(int invocation)
{

        int change=FALSE;
#ifdef CROSS


        struct bblock* cblk, *left_block, *right_block;
        struct list* rtlptr, *jmp_rtl=NULL;
	char set_regs[MAXSTRING], used_regs[MAXSTRING];
	set_regs[0]=used_regs[0]=0;

#ifdef VET
begin_phase(CROSSJUMPING_PHASE);
#endif

        for(cblk=top; cblk; cblk=cblk->next)
        {
		/* 
		 * OK, if we have 2 fall-throughs we want to look for
		 * instructions that can move to this block without 
		 * causing any problems.  
		 */
		if(cblk->left && cblk->right && cblk->left!=cblk->right)
		{
			left_block=cblk->left;
			right_block=cblk->right;

			/* the left/right block can only have 1 predecessor */
			if(left_block->preds->ptr!=cblk || left_block->preds->next!=0)
				continue;
			if(right_block->preds->ptr!=cblk || right_block->preds->next!=0)
				continue;

			jmp_rtl=find_unification_point(cblk,used_regs,set_regs);
			/* we need to be able to find the jmp in this block before attemping to move
				rtls in here */
			if(jmp_rtl==0)
				continue;

			change|=find_common_instructions(cblk, jmp_rtl, used_regs, set_regs);
		}
	}

#ifdef VET
end_phase(CROSSJUMPING_PHASE);
#endif
#endif

	return change;
}

int compare_rtls(struct list ** left_rtl, struct list ** right_rtl, char used_regs[], char set_regs[])
{
	while(*left_rtl && (*left_rtl)->type & COMMENT)
		*left_rtl=(*left_rtl)->next;
	while(*right_rtl && (*right_rtl)->type & COMMENT)
		*right_rtl=(*right_rtl)->next;

	if(!*left_rtl || !*right_rtl)
		return FALSE;

	if(BASETYPE((*left_rtl)->type) != BASETYPE((*right_rtl)->type))
		return FALSE;

	if((*left_rtl)->type & ASMLINE)
		return FALSE;


	/* if only one has a value, we can't optimize now */
	if((*left_rtl)->val!=0 && (*right_rtl)->val==0)
		return FALSE;
	if((*left_rtl)->val==0 && (*right_rtl)->val!=0)
		return FALSE;

	/* if they are a branch.. well, that's not what we want */
	if((*left_rtl)->val!=0 && is_pc((*left_rtl)->val))
		return FALSE;

	/* if neither do, we're cool */
	if((*left_rtl)->val==0 && (*right_rtl)->val==0)
		return TRUE;

	if(strcmp((*left_rtl)->val, (*right_rtl)->val)==0)
	{
		char *a=(*left_rtl)->val;
		while(a=gtrg(a))
		{
			char pattern[REGSTRING];
			copyloc(pattern,unskiploc(a));
			*skiploc(pattern)=0;
			if(fmatch(set_regs,pattern))
				return FALSE;
			if(*a=='=')
				if(fmatch(used_regs,pattern))
					return FALSE;
			
		}
		a=(*left_rtl)->se;
		while(a && (a=gtrg(a)))
		{
			char pattern[REGSTRING];
			copyloc(pattern,unskiploc(a));
			*skiploc(pattern)=0;
			if(fmatch(set_regs,pattern))
				return FALSE;
			if(*a=='=')
				if(fmatch(used_regs,pattern))
					return FALSE;
		}


		if(is_call((*left_rtl)->val))
		{
			char* a=used_regs;
			while(a && *a)
			{
				if(scratchreg(a))
					return FALSE;
				a=skiploc(a);
			}
		}
		return TRUE;
	}

	return FALSE;
}

int find_common_instructions ( struct bblock* cblk, struct list* jmp_rtl, char used_regs[], char set_regs[])
{
	char reg[REGSTRING];
	struct list* left_rtl=cblk->left->rtls;
	struct list* right_rtl=cblk->right->rtls;
	struct list* next_left_rtl=0;
	struct list* next_right_rtl=0;
	int change=FALSE;

	while(compare_rtls(&left_rtl, &right_rtl, used_regs, set_regs))
	{
		/* we've determined that left/right are the same, so
		 * we're going to move them.  The left gets removed
		 * from it's location and inserted before jmp_rtl.
		 */

		/* memorize the next RTLs to look at */
		next_left_rtl=left_rtl->next;
		next_right_rtl=right_rtl->next;

#ifdef VET
begin_trans(VET_OPT_TRANS);
#endif
		change=TRUE;
		/* unlink the left rtl */
		if(left_rtl->prev)
			left_rtl->prev->next=left_rtl->next;
		else
			cblk->left->rtls=left_rtl->next;

		if(left_rtl->next)
			left_rtl->next->prev=left_rtl->prev;
		else
			cblk->left->rtlend=left_rtl->prev;

		/* remove the right rtl */
		remv_links_to(right_rtl);
		remv_links(right_rtl);
		remv_rtl(right_rtl);
		/* remove the links to the left one, they may not be safe */
		remv_links_to(left_rtl);
		remv_links(left_rtl);

		/* put the left rtl before jmp_rtl */
		left_rtl->prev=jmp_rtl->prev;
		left_rtl->next=jmp_rtl;
		if(jmp_rtl->prev)
			jmp_rtl->prev->next=left_rtl;
		else
			jmp_rtl->daddy->rtls=left_rtl;

		jmp_rtl->prev=left_rtl;
		left_rtl->daddy=jmp_rtl->daddy;
		if(cnts(left_rtl)) combmark(left_rtl);


#ifdef VET
end_trans();
#endif
		/* advance to the next rtl */
		left_rtl=next_left_rtl;
		right_rtl=next_right_rtl;
		
	};
	return change;
}


int cross_jump(int invocation)

{

        int change=FALSE;
#ifdef CROSS


        struct bblock* cblk, *left_block, *right_block;
        struct list* rtlptr, *jmp_rtl=NULL;
	char set_regs[MAXSTRING], used_regs[MAXSTRING];
	set_regs[0]=used_regs[0]=0;

#ifdef VET
begin_phase(CROSSJUMPING_PHASE);
#endif

        for(cblk=top; cblk; cblk=cblk->next)
        {
		/* 
		 * OK, if we have 2 fall-throughs we want to look for
		 * instructions that can move to this block without 
		 * causing any problems.  
		 */
		struct blist* preds;
		struct list* rtl=NULL;
		int rtls_found=0;
		struct list* rtls_to_remove[MAXRTS];
		char *set_reg_ptr, *used_reg_ptr;
		int i;

trynextrtl:
		rtls_found=0;
	
		/* we need at least 2 preds for this to be useful */
		if(cblk->preds==0 || cblk->preds->next==0)
			continue;
#ifndef SIZE_OPTS
		if(cblk==ret_block)
			continue;
#endif
		for(preds=cblk->preds; preds; preds=preds->next)
		{
		        /* if predecessor has more than one successor, then
                         * we can't move RTLs into this successor (cblk) 
			 */
			if(preds->ptr->left && preds->ptr->right)
				goto trynextblock;
			if(!preds->ptr->left && !preds->ptr->right)
				goto trynextblock;

			rtl=preds->ptr->rtlend;
			set_regs[0]=used_regs[0]=0;
			set_reg_ptr=set_regs;
			used_reg_ptr=used_regs;

			/* back up over any comments, and over the jump, if any */
			while(rtl && rtl->type & COMMENT)
			  rtl=rtl->prev;
			if(cnts(rtl) && is_pc(rtl->val))
			{
				char *a=rtl->val;
				while(a && (a=gtrg(a)))
				{
					if(*a=='=')
					{
						copyloc(used_reg_ptr,unskiploc(a));
						*(used_reg_ptr=skiploc(used_reg_ptr))=0;
					}
					else
					{
						copyloc(set_reg_ptr,unskiploc(a));
						*(set_reg_ptr=skiploc(set_reg_ptr))=0;
					}
				}
				a=rtl->se;
				while(a && (a=gtrg(a)))
				{
					if(*a=='=')
					{
						copyloc(used_reg_ptr,unskiploc(a));
						*(used_reg_ptr=skiploc(used_reg_ptr))=0;
					}
					else
					{
						copyloc(set_reg_ptr,unskiploc(a));
						*(set_reg_ptr=skiploc(set_reg_ptr))=0;
					}
				}
				rtl=rtl->prev;
			}
			while(rtl && rtl->type & COMMENT)
			  rtl=rtl->prev;

			if(cnts(rtl))
			{
				char *a=rtl->val;
				while(a=gtrg(a))
				{
					char reg[REGSTRING];
					copyloc(reg,unskiploc(a));
					*skiploc(reg)=0;
					if(fmatch(set_regs,a))
						goto trynextblock;
					if(*a=='=' && fmatch(used_regs,a))
						goto trynextblock;
				}
			}
			if(!rtl) goto trynextblock;
			rtls_to_remove[rtls_found++]=rtl;

			/* too many predecessors to worry about */
			if(rtls_found>MAXRTS) 
				goto trynextblock;
		}
		/* lets not merge asm lines, we don't want to mess with 'em. */
		if(rtls_to_remove[0]->type & ASMLINE)
			goto trynextblock;

		rtl=rtls_to_remove[0];
		for(i=1;i<rtls_found;i++)
		{
		  if(BASETYPE(rtls_to_remove[i]->type)!=BASETYPE(rtl->type))
		    goto trynextblock;
		  if(rtls_to_remove[i]->val==0 && rtl->val!=0)
		    goto trynextblock;
		  if(rtls_to_remove[i]->val!=0 && rtl->val==0)
		    goto trynextblock;
		  if(rtl->val!=0 && strcmp(rtls_to_remove[i]->val,rtl->val)!=0)
		    goto trynextblock;
		}
#ifdef VET
begin_trans(VET_OPT_TRANS);
#endif
		for(i=0;i<rtls_found;i++)
		{
			if(rtls_to_remove[i]->next)	rtls_to_remove[i]->next->prev=rtls_to_remove[i]->prev;
			else				rtls_to_remove[i]->daddy->rtlend=rtls_to_remove[i]->prev;
			if(rtls_to_remove[i]->prev)	rtls_to_remove[i]->prev->next=rtls_to_remove[i]->next;
			else				rtls_to_remove[i]->daddy->rtls=rtls_to_remove[i]->next;
			remv_links_to(rtls_to_remove[i]);
			remv_links(rtls_to_remove[i]);

			if(i!=0)
			{
				remv_rtl(rtls_to_remove[i]);
			}
		}
		rtl->next=cblk->rtls;
		if(rtl->next) rtl->next->prev=rtl;
		else	cblk->rtlend=rtl;
		cblk->rtls=rtl;
		rtl->prev=NULL;
		rtl->daddy=cblk;
		cblk->rtls=rtl;
		if(cnts(rtl)) combmark(rtl);

		/* the optimization also moves empty RTLs, which sets the
		   change variable although no 'real' change is made to the
		   program code. Only set 'change' if the RTL moved is an
		   actual instruction*/
		if(cnts(rtl))
		   change=TRUE;

#ifdef VET
end_trans();
#endif
		goto trynextrtl;
		


		trynextblock:
		/* at this point we try the next block, something has gone wrong */
			continue;
	}

#ifdef VET
end_phase(CROSSJUMPING_PHASE);
#endif
#endif

	return change;
}
