
/************************************************************************/
/*                                                                      */
/* Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   */
/*     The Rector and Visitors of the University of Virginia            */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from the University of Virginia.                                */
/* This code is distributed WITHOUT ANY WARRANTY.  No claims are        */
/*      made as to whether it serves any particular purpose or even     */
/*      works at all.                                                   */
/*                                                                      */
/************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/database.c,v $
 $Date: 2003/03/05 15:45:58 $
 $Revision: 1.14 $
 $Name:  $

 Log of changes:

 $Log: database.c,v $
 Revision 1.14  2003/03/05 15:45:58  clc5q
 Add assertions and diagnostic output.

 Revision 1.13  2003/01/24 20:18:24  clc5q
 Move maybe_same() from each target's rtl.c module up into combine.c.
    Make calls into rtl.c for machine-dependent information or processing.

 Revision 1.12  2002/07/11 15:22:31  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.11  2002/06/05 20:08:53  jdh8d
 Changes to support the stache, including routines to build a callgraph.

 Some cosmetic changes.

 Minor bug fixes.

 Changed build scripts to build the mips compiler for SimpleScalar if
 SSAS is set to "-DSS_AS"

 Revision 1.10  2001/07/23 18:30:51  rv3s
 Added a routine clrmem() to clear out only memory items from the equivalence
 list

 Revision 1.9  2001/05/28 20:30:20  jdh8d
 *** empty log message ***

 Revision 1.8  2001/05/23 20:28:20  jdh8d


 Re-worked all of the backend to support variable
 sized registers.  IT IS  NO LONGER SAFE TO ASSUME
 THAT REGISTERS ARE TWO BYTES.  Mips currently has
 4-byte registers, others will follow (or be destroyed!)

 Revision 1.7  2001/04/09 16:05:04  rv3s
 Check if the item length in e_insert is greater than the limit. If so, do not insert the item and return NULL

 Revision 1.6  2001/03/07 15:50:16  jdh8d


 eod fix to observe t* lines

 Revision 1.5  2000/06/09 19:09:56  jwd
 Changed class to VET to indicate which phases are mandatory and
 which are optional.

 Revision 1.4  1999/01/06 15:35:51  jwd
 Added VET directives.
 Revision 1.3  1998/04/03 18:19:04  jwd
 First ANSI C version.

*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "vpo.h"
#include "assign.h"
#include "combine.h"
#include "misc.h"
#include "vectors.h"
#include "database.h"

#include "rtl.h"    /*Machine Dependent */

#ifdef VET
#include "view.h"
#endif

/*
 * This file contains functions that perform common housekeeping and
 * transaction chores on some important data structures.
 */


/*------------History of Modifications--------------------------
  1/9/98:  Took ansi-c prototyping work of the summer and applied it, 
            removing local scope prototypes from the functions


--------------------------------------------------------*/


struct setlist sets[MAXRVEC];	/* register information table   */
struct setlist *setsused = NULL;	/* pointer to sets used         */

/*
 * The first group of functions deal with structure that keep track of
 * register and local variable information.
 */

/*
 * s_add - insert item into set list if not already there
 */
struct setlist *s_add(char *item) {

   struct setlist *sptr;
   int i;
   int new_location;
   char reg[REGSTRING];


   new_location=USP(item);
   i = vecmap[new_location];
   if (i < 0) {
      makereg(reg, regtype(item), regnum(item));
      (void) fprintf(stderr, "vpo (s_add): unallocated item, '");
      show_rtl(stderr, reg);
      (void) fprintf(stderr, "'\n");
      if (NULL != fn)
         (void) fprintf(stderr, " in function: %s\n", fn);
      assert(("Unallocated item", i >= 0));
      exit(EXIT_FAILURE);
   }
   sptr = &sets[i];
   if (!sptr->valid) {
      sptr->next = setsused;
      sptr->valid = setsused = sptr;
      sptr->sets = sptr->dead = sptr->ldead = (struct list *) NULL;
      sptr->regstatus = UNASSIGNED;
      sptr->assign = 0;
      sptr->uses = 0;
      sptr->rnum = 0;
      sptr->triple = (struct itriple *) NULL;
      sptr->setin = (struct blist *) NULL;
   }
   return (sptr);
}

/*
 * remember_deaths - remember where registers died
 */
void remember_deaths(struct list * ptr) {

   struct setlist *sptr;
   unsigned int c, j;
   unsigned int *rptr, *end;

   end = (rptr = (unsigned int *) ptr->deads) + rveclen;
   sptr = sets;
   while (rptr != end) {
      if (c = *rptr++) {
	 for (j = 1; j; j <<= 1) {
	    if (c & j) {
	       if (!sptr->valid) {
		  sptr->next = setsused;
		  sptr->valid = setsused = sptr;
		  sptr->sets = sptr->dead = (struct list *) NULL;
		  sptr->regstatus = UNASSIGNED;
		  sptr->assign = 0;
		  sptr->uses = 0;
		  sptr->triple = (struct itriple *) NULL;
	       }
	       sptr->ldead = ptr;
	       sptr->rnum = ptr->rnum;
	    }
	    sptr++;
	 }
      }
      else
	 sptr += (bitsperint + 1);
   }
}

/*
 * clrsets - initializes the sets database
 */
void clrsets() {

   register struct setlist *ptr;

   for (ptr = setsused; ptr; ptr = ptr->next)
      ptr->valid = (struct setlist *) NULL;
   setsused = (struct setlist *) NULL;
}

/*
 * clrsetdeads - clear all set and dead links in set list
 */
void clrsetdeads() {

   struct setlist *ptr;

   for (ptr = setsused; ptr; ptr = ptr->next)
      ptr->sets = ptr->dead = (struct list *) NULL;
}

/*
 * The following set of functions manipulate equivalence tables for the
 * common subexpression eliminator
 */

/*
 * e_insert - insert an item in an equivalence table
 */
struct eqvlist *e_insert(struct eqvtable * list, char *item) {

   struct eqvlist *eptr;
   int h;
   char *a;

   /* New check */
   /* Dont add the item if its too big */
   if (strlen(item) > ITEM_SIZE_LIMIT)
     return 0;

   /* end check */

   /*
      Don't add an item that is already in the equivalence table
   */
   if (!(eptr = e_lookup(list, item))) {

      /*
         Allocate space for a new equivalence list structure
      */
      eptr = (struct eqvlist *) SALLOCTYPE(sizeof(struct eqvlist), EQVLISTTYPE);

      /*
         Insert the item into its appropriate hash bucket for this table used
         to look up the item quickly.
      */
      if (eptr->next = list->eqiv[h = hash(item)])
	 eptr->next->prev = eptr;
      list->eqiv[h] = eptr;
      eptr->prev = (struct eqvlist *) NULL;

      /*
         Insert the item into the definition list which just keeps track of all
         of the items defined in this equivalence table.
      */
      if (eptr->ndef = list->used)
	 eptr->ndef->pdef = eptr;
      list->used = eptr;
      eptr->pdef = (struct eqvlist *) NULL;

      /*
         Initialize the equivalence list to indicate that this item is as yet
         not equivalent to anything else.
      */
      eptr->nequiv = eptr->pequiv = (struct eqvlist *) NULL;

      /*
         Initialize the remaining fields of this new structure
      */
      eptr->item = strcpy(CALLOC(strlen(item) + 1), item);
      eptr->sets = eptr->dead = (struct list *) NULL;
      eptr->uses = 0;
      eptr->reg = 0;
      eptr->memtype = NOMEM;
      if (!(eptr->reg = isareg(item)))
	 for (a = item; a = gtmem(a); eptr->memtype |= get_memtype(bpos))
	    a = bpos + 1;
   }

   return (eptr);
}

/*
 * e_insert_reg - insert a register in an equivalence table
 */
struct eqvlist *e_insert_reg(struct eqvtable * list, char *item) {

   struct eqvlist *eptr;
   int h;
   if(!isRegister(item))
   {
	fprintf(stderr, "Bad e_insert_reg found :(\n");
   }	
   
   /*
      Don't add an item that is already in the equivalence table
   */
   if (!(eptr = e_lookup_reg(list, item))) {

      /*
         Allocate space for a new equivalence list structure
      */
      eptr = (struct eqvlist *) SALLOCTYPE(sizeof(struct eqvlist), EQVLISTTYPE);

      /*
         Insert the item into its appropriate hash bucket for this table used
         to look up the item quickly.
      */
      if (eptr->next = list->eqiv[h = hash(item)])
	 eptr->next->prev = eptr;
      list->eqiv[h] = eptr;
      eptr->prev = (struct eqvlist *) NULL;

      /*
         Insert the item into the definition list which just keeps track of all
         of the items defined in this equivalence table.
      */
      if (eptr->ndef = list->used)
	 eptr->ndef->pdef = eptr;
      list->used = eptr;
      eptr->pdef = (struct eqvlist *) NULL;

      /*
         Initialize the equivalence list to indicate that this item is as yet
         not equivalent to anything else.
      */
      eptr->nequiv = eptr->pequiv = (struct eqvlist *) NULL;

      /*
         Initialize the remaining fields of this new structure
      */
      eptr->item = CALLOC(REGSTRING);
      copyloc(eptr->item, item);
      *(skiploc(eptr->item)) = '\0';
      eptr->sets = eptr->dead = (struct list *) NULL;
      eptr->uses = 0;
      eptr->reg = TRUE;
      eptr->memtype = NOMEM;
   }

   return (eptr);
}

/*
 * dstroy - remove equivalences of all items referencing the given expression
 */
void dstroy(struct eqvtable * list, char *exp) {

   char c, *lin, d, *a, *b;
   struct eqvlist *ptr, *dptr;

   /*
      Look through all of the items defined in this equivalence table
   */
   ptr = list->used;
   c = *exp++;
nexti:
   if (!ptr)
      return;

   /*
      Locate all of those items that contain "exp" as part of the item
   */
   lin = ptr->item;
   while (d = *lin++) {
      if (d == c) {
	 b = lin;
	 a = exp;
	 if (!(d = *a++))
	    goto found;
	 while (d == *b++) {
	    if (!(d = *a++)) {
	 found:

	       /*
	          Remove the item from the list of all its equivalents
	       */
	       if (ptr->pequiv)
		  ptr->pequiv->nequiv = ptr->nequiv;
	       if (ptr->nequiv)
		  ptr->nequiv->pequiv = ptr->pequiv;

	       /*
	          Remove the item from its hash bucket list
	       */
	       if (ptr->prev)
		  ptr->prev->next = ptr->next;
	       else
		  list->eqiv[hash(ptr->item)] = ptr->next;
	       if (ptr->next)
		  ptr->next->prev = ptr->prev;

	       /*
	          Remove the item from the list of items in this equivalence
	          table
	       */
	       if (ptr->pdef)
		  ptr->pdef->ndef = ptr->ndef;
	       else
		  list->used = ptr->ndef;
	       if (ptr->ndef)
		  ptr->ndef->pdef = ptr->pdef;

	       /*
	          Release the item and all other elements associated with it
	       */
	       dptr = ptr;
	       ptr = ptr->ndef;
	       FREE(dptr->item);
	       FREETYPE(dptr, EQVLISTTYPE);
	       goto nexti;
	    }
	 }
      }
   }
   ptr = ptr->ndef;
   goto nexti;
}

/*
 * mem_aliases - remove any potential alias memory locations
 */
void mem_aliases(struct eqvtable * list, char *memref) {

   struct eqvlist *ptr, *next;
   int memtype;
   char *a;
   
   /*
      Determine roughly what kind of memory items the memory reference given
      might potentially touch.
   */
   if (ptr = e_lookup(list, memref))
      memtype = ptr->memtype;
   else
      memtype = get_memtype(memref);

   /*
      Examine each item in the equivalence list
   */
   for (ptr = list->used; ptr; ptr = next) {
      next = ptr->ndef;

     /*
      *	first check to see if ptr->item has a memory type yet
      */

      /*
         If the item contains a memory reference that might potentially
         conflict with the item that we are storing to, check it more closely.
         If we cannot be sure that the items are not referencing the same
         memory location, then we have to play it safe and remove the items
         from the equivalence table.
      */
      if (ptr->memtype & memtype) {
	 a = ptr->item;
	 while (a = gtmem(a)) {
	    if (maybe_same(memref, bpos)) {
      	       if(swd && debug_opt_i("debug_mem_aliases"))
	       {
			fprintf(stderr, "Dumping");
			show_rtl(stderr, ptr->item);
			fprintf(stderr, " and ");
			show_rtl(stderr, bpos);
			fprintf(stderr, " types are the same\n");

	       }
	       e_delete(list, ptr->item);
	       break;
	    }
      	    else if(swd && debug_opt_i("debug_mem_aliases"))
	    {
		fprintf(stderr, "Keeping ");
		show_rtl(stderr, ptr->item);
		fprintf(stderr, " and ");
		show_rtl(stderr, bpos);
		fprintf(stderr, " because maybe_same says so\n");
 	    }
	 }
      }
      else if(swd && debug_opt_i("debug_mem_aliases"))
      {
	fprintf(stderr, "Keeping ");
	show_rtl(stderr, ptr->item);
	fprintf(stderr, " and ");
	show_rtl(stderr, bpos);
	fprintf(stderr, " because mem-types are different\t");
	fprintf(stderr, "ptr->memtype=%d, memtype=%d\n", ptr->memtype, memtype);

	
      }
   }
}

/*
 * e_delete - remove an item from the equivalence list
 */
void e_delete(struct eqvtable * list, char *item) {

   struct eqvlist *ptr;
   

   /*
      Search the current equivalence table for the requested item
   */
   if (ptr = e_lookup(list, item)) {

      /*
         Remove the item from the equivalence chain
      */
      if (ptr->pequiv)
	 ptr->pequiv->nequiv = ptr->nequiv;
      if (ptr->nequiv)
	 ptr->nequiv->pequiv = ptr->pequiv;

      /*
         Remove the item from the hash bucket list
      */
      if (ptr->prev)
	 ptr->prev->next = ptr->next;
      else
	 list->eqiv[hash(item)] = ptr->next;
      if (ptr->next)
	 ptr->next->prev = ptr->prev;

      /*
         Remove the item from the list of item defined in this table
      */
      if (ptr->pdef)
	 ptr->pdef->ndef = ptr->ndef;
      else
	 list->used = ptr->ndef;
      if (ptr->ndef)
	 ptr->ndef->pdef = ptr->pdef;

      /*
         Release the item and all elements associated with it
      */
      FREE(ptr->item);
      FREETYPE(ptr, EQVLISTTYPE);
   }
}

/*
 * e_lookup - search for an item in an equivalence list and return its pointer
 */
struct eqvlist *e_lookup(struct eqvtable * list, char *item) {

   struct eqvlist *ptr;
   

   /*
      Look for the entry in the appropriate hash list
   */
   for (ptr = list->eqiv[hash(item)]; ptr; ptr = ptr->next)
      if (!strcmp(item, ptr->item))
	 return (ptr);

   /*
      Return NULL pointer if no entry is found
   */
   return ((struct eqvlist *) NULL);
}

/*
 * e_lookup_reg - search for a register in an equivalence list
 */
struct eqvlist *e_lookup_reg(struct eqvtable * list, char *item) {

   struct eqvlist *ptr;
   char c;

   /*
      Look for the entry in the appropriate hash list.  Note that registers are
      only two characters long.  Also note that the first character of all
      register has the same two most-significant-bits.  Since the hash function
      ensures that characters starting with the same least- significant-six
      bits (at least) fall into the same hash bucket, then there is no need to
      compare for the first character, just the second one is enough to tell us
      if we have the right item or not.

        now, however, registers are variable sized depending on the target
        (currently 4 for mips, 2 for everything else)
        so we need to use the compareloc macro to compare properly

   */
   for (ptr = list->eqiv[hash(item)]; ptr; ptr = ptr->next)
   {
        if(cmploc(ptr->item, item)==0 && *skiploc(ptr->item)==0)
                return (ptr);
   }

   /*
      Return NULL pointer if no entry is found
   */
   return ((struct eqvlist *) NULL);
}

/*
 * clrequiv - clear an equivalence table
 */
void clrequiv(struct eqvtable * list) 
{

   struct eqvlist *ptr, *dptr;
   int loop;

   ptr = list->used;
   while (ptr) {
      dptr = ptr;
      ptr = ptr->ndef;
      FREE(dptr->item);
      FREETYPE(dptr, EQVLISTTYPE);
   }
   list->used = (struct eqvlist *) NULL;
   for (loop = 0; loop < ESIZE; list->eqiv[loop++] = (struct eqvlist *) NULL);
}

/*
 * newequiv - allocates a new equivalence list and clears it
 */
struct eqvtable *newequiv() {

   struct eqvtable *list;
   int loop;

   list = (struct eqvtable *) SALLOCTYPE(sizeof(struct eqvtable), EQVTABLETYPE);
   for (loop = 0; loop < ESIZE; list->eqiv[loop++] = (struct eqvlist *) NULL);
   list->used = (struct eqvlist *) NULL;
   list->uses = 1;
   return (list);
}

/*
 * e_clrsets - clear all sets pointers to prevent definition deletion
 */
void e_clrsets(struct eqvtable * list)
{
   struct eqvlist *ptr;

   for (ptr = list->used; ptr; ptr = ptr->ndef)
      ptr->sets = (struct list *) NULL;
   list->uses = 2;
}

/*
 * e_copy - make a copy of an equivalence list
 */
void e_copy(struct eqvtable * list1, struct eqvtable * list2) 
{

   struct eqvlist *ptr, *new, *eptr;
   

   /*
      Loop through all items defined in list2
   */
   for (ptr = list2->used; ptr; ptr = ptr->ndef) {

      /*
         If the item starts an equivalence list with more than one definition,
         then copy it over.  There is no reason to move items than have no
         equivalences, since they will not play a part in finding common
         subexpressions.
      */
      if (!ptr->pequiv && ptr->nequiv) {

	 /*
	    Create a new item with no equivalence links.  Copy over the value
	    of the "dead" field that tells us where the item was last used.
	 */
	 if (ptr->reg)
	    new = e_insert_reg(list1, ptr->item);
	 else
	    new = e_insert(list1, ptr->item);
	 new->dead = ptr->dead;

	 /*
	    Now locate each of the equivalent items and make copies of them in
	    the new equivalence table.
	 */
	 for (eptr = ptr->nequiv; eptr; eptr = eptr->nequiv) {
	    if (eptr->reg)
	       new->nequiv = e_insert_reg(list1, eptr->item);
	    else
	       new->nequiv = e_insert(list1, eptr->item);
	    new->nequiv->pequiv = new;
	    new = new->nequiv;
	    new->dead = eptr->dead;
	 }
      }
   }
}

/*
 * e_merge - merge two equivalence lists keeping only "safe" expressions
 */
void e_merge(struct eqvtable * list1, struct eqvtable * list2) 
{

   struct eqvlist *ptr1, *ptr2, *tail, *pe1, *next, *pe2, *px2,
	*myptr1;
   int heads;

   /*
      Reset the "seen" flag on all of the items in the accumulating equivalence
      table and count the number separate equivalence sets that there are.
   */
   heads = 0;
   for (ptr1 = list1->used; ptr1; ptr1 = ptr1->ndef) {
      ptr1->seen = FALSE;
      if (!ptr1->pequiv)
	 heads++;
   }

   /*
      Loop through the equivalence table until all of the equivalence sets in
      the accumulating table have been handled.
   */
   while (heads) {

      /*
         Make a pass through all of the items in the accumulating equivalence
         table.
      */
      for (ptr1 = list1->used; ptr1; ptr1 = next) 
      {

	 /*
	    Remember the next item in the table.  This is used because there
	    will be instances where the current item will be deleted from the
	    accumalting equivalence table.
	 */
	 next = ptr1->ndef;

	 /*
	    If this item is the head of an equivalence set that we haven't
	    handeled yet, then examine it more closely.
	 */
	 if (!(ptr1->pequiv || ptr1->seen)) {

	    /*
	       Look for the item in the updating equivalence list
	    */
	    if (ptr2 = e_lookup(list2, ptr1->item)) {

	       /*
	          If the item has equivalences in the accumulating equivalence
	          list, then mark it as seen.
	       */
	       if (pe1 = ptr1->nequiv)
		  ptr1->seen = TRUE;

	       /*
	          Merge the death points of the accumulating list and the
	          updating list for this item.
	       */
	       merge_deaths(ptr1, ptr2);

	       /*
	          Now back to the start of the equivalance set in the updating
	          list so that we can determine which items are shared by the
	          two sets.
	       */
	       while (ptr2->pequiv)
		  ptr2 = ptr2->pequiv;

	       /*
	          Examine the items that are equivalent to this item in the
	          accumulating equivalence table.
	       */
	       while (pe1) {

		  /*
		     Look up the equivalent item in the updating equivalence
		     table. If we find it, then look through if the items are
		     equivalent in both of the equivalence tables, then mark
		     the item as having been seen and merge their death points.
		  */
		  if (pe2 = e_lookup(list2, pe1->item)) {
		     px2 = ptr2;
		     do {
			if (px2 == pe2) {
			   pe1->seen = TRUE;
			   merge_deaths(pe1, pe2);
			   break;
			}
		     } while (px2 = px2->nequiv);
		  }

		  /*
		     Continue through all of the equivalent items in this set
		  */
		  pe1 = pe1->nequiv;
	       }
	    }

	    /*
	       Initialize a new list
	    */
	    tail = (struct eqvlist *) NULL;

	    /*
	       Go through all of the items that are equivalent to this item in
	       the accumulating equivalence table.  Place all items in this set
	       of equivalences that do not also exist in an equivalence set in
	       the updating equivalence table into a new list.
	    */
	    for (pe1 = ptr1; pe1; pe1 = px2) {
	       px2 = pe1->nequiv;
	       if (!pe1->seen) {
		  if (pe1->pequiv)
		     pe1->pequiv->nequiv = pe1->nequiv;
		  if (pe1->nequiv)
		     pe1->nequiv->pequiv = pe1->pequiv;
		  pe1->nequiv = (struct eqvlist *) NULL;
		  if (pe1->pequiv = tail) {
		     tail->nequiv = pe1;
		     tail = pe1;
		  }
		  else {
		     tail = pe1;
		     heads++;
		     if (pe1 == ptr1) {
			if (px2)
			   heads++;
			break;
		     }
		  }
	       }
	    }

	    /*
	       If we have created a new equivalence set at this point, we check
	       to see if it is a solitary item.  If it is, then we can go ahead
	       and remove the item from the accumulating equivalence table.  We
	       decrement the number of sets that we have to work with and
	       update the next pointer if we happen to be deleting the next
	       item that we were going to visit.
	    */
	    if (tail && !tail->pequiv) {
	       heads--;
	       if (next == tail)
		  next = next->ndef;
	       e_delete(list1, tail->item);
	    }

	    /*
	       Now there is one less set to handle.  We break out of the loop
	       if we have handled all of the sets (even the new ones we
	       create).
	    */
	    if (!heads--)
	       break;
	 }
      }
   }
}

/*
 * merge_deaths - merge deaths of two equivalence lists
 */
void merge_deaths(struct eqvlist * cptr, struct eqvlist * dptr) 
{



   if (dptr->dead)
   {
      if (cptr->dead) {
#ifdef VET
        begin_trans(VET_OPT_TRANS);
#endif
	 if (dptr->dead->onum < cptr->dead->onum)
	    rdelete(dptr->item, &dptr->dead->deads);
	 else if (dptr->dead->onum > cptr->dead->onum) {
	    rdelete(dptr->item, &cptr->dead->deads);
	    cptr->dead = dptr->dead;
	 }
#ifdef VET
        end_trans();
#endif
      }
      else
      {
	 cptr->dead = dptr->dead;
      }
   }
}

/*
 * e_invar - remove any items that are not loops invariant from the list
 */
void e_invar(struct eqvtable * list, struct loopl * loop) 
{

   char *a;
   struct eqvlist *ptr, *next;


   /*
      Examine each item in the equivalence table independently and determine if
      it contains any non-invariant item.  Remove any such items.
   */
   for (ptr = list->used; ptr; ptr = next) {
      next = ptr->ndef;

      /*
         If memory reads are not loop invaraint, then we delete the item if it
         references memory.
      */
      if (!loop->invaread) {
	 if (ptr->memtype) {
	    e_delete(list, ptr->item);
	    continue;
	 }
      }

      /*
         Examine the registers used in this item, and remove any item that
         contains any register that is not marked loop invariant.
      */

	if(isMangledReg(ptr->item,"*") || isMangledReg(ptr->item,"*?*"))
		e_delete(list,ptr->item);
	else
	{
      		a = ptr->item;
      		while (a = gtrg(a)) {
	 		if (!loop->invareg || !rexists(unskiploc(a), loop->invareg))
	    		break;
      		}
      		if (a)
	 		e_delete(list, ptr->item);
	}
   }
}

/* clrmem -- remove all items using memory values */
void clrmem(struct eqvtable *list) 
{
  char *a;
  struct eqvlist *ptr, *next;

  /*
      Examine each item in the equivalence list
  */
  for (ptr = list->used; ptr; ptr = next) {
    next = ptr->ndef;
    
    /*
      If the item references memory other than a spill location, then remove
      it from the equivalence table.
    */
    if (ptr->memtype && ptr->memtype != SPILLMEM) {
      e_delete(list, ptr->item);
      continue;
    }
  }



}
/*
 * clrscratchmem - remove all items using scratch registers or memory values
 */
void clrscratchmem(struct eqvtable * list) 
{

   char *a;
   struct eqvlist *ptr, *next;

    /*
      Examine each item in the equivalence list
   */
   for (ptr = list->used; ptr; ptr = next) {
      next = ptr->ndef;

      /*
         If the item references memory other than a spill location, then remove
         it from the equivalence table.
      */
      if (ptr->memtype && ptr->memtype != SPILLMEM) {
	 e_delete(list, ptr->item);
	 continue;
      }

      /*
         If the item references any scratch register items, then remove it from
         the equivalence table.
      */
      a = ptr->item;
      while (a = gtrg(a)) {
	 if (scratchreg(unskiploc(a)))
	    break;
      }
      if (a) {
	 e_delete(list, ptr->item);
	 continue;
      }
   }
}

/*
 * clrscratch - remove all items using scratch registers
 */
void clrscratch(struct eqvtable * list) 
{

   char *a;
   struct eqvlist *ptr, *next;

   /*
      Examine each item in the equivalence list
   */
   for (ptr = list->used; ptr; ptr = next) {
      next = ptr->ndef;

      /*
         If the item references any scratch register items, then remove it from
         the equivalence table.
      */
      a = ptr->item;
      while (a = gtrg(a)) {
	 if (scratchreg(unskiploc(a)))
	    break;
      }
      if (a) {
	 e_delete(list, ptr->item);
	 continue;
      }
   }
}

























