/* Function de_optimize.c
   Contains functions to take in an optimized function and reverse the
   changes to remove the optimizations
*/

/* 
 * $Author: jdh8d $
 * $Date: 2004/01/21 18:39:54 $
 * $Revision: 1.2 $
 */

#include <stdio.h>
#include "assign.h"
#include "controlflow.h"
#include "locals.h"
#include "misc.h"
#include "vpo.h"
#include "vpo2.h"
#include "view.h"
#include "de_optimize.h"
#include "deopt.h"  /* Machine dependent */

#ifdef INTERACT
#ifdef DEOPT

#ifndef locnum
#define locnum(s)   (((*s & 0x1f) << 7) + (*(s+1) & 0x7f))
#endif

struct rig_node *rig_top,       /* Start of the Register Interference Graph */
                *rig_end;       /* End of the Register Interference Graph */
int rig_nums = 0,               /* Total number of RIG nodes */
    old_regallocptr;            /* Original regallocptr value (for resizing) */

static int verify_outs(struct bblock *cblk);

int de_optimize_func()
{
    int changes = FALSE;
    changes |= deopt_loop_inv_cdmotion();
    return changes;
}


/*
 * coalesce_fp_moves - Combine 2 single moves into a double move
 */
int coalesce_fp_moves()
{
    int change,                 /* Boolean flag to signify modification */
        len,                    /* String length of match */
        removed_rtl;            /* Whether rtl was removed or not */
    struct bblock *cblk;        /* The current basic block we are in */
    struct list *rtl,           /* The rtl we are working from */
        *trtl,                  /* Testing rtls for matches */
        *next;                  /* The next RTL to start testing with */
    char *c,                    /* Start of RTL information */
         match[60],             /* Buffer to construct match string */
         new_text[60],          /* Buffer for new coalesced RTL */
         lreg[REGLOCLENGTH+1],  /* Left fp register (assigned to) */
         rreg[REGLOCLENGTH+1];  /* Right fp register (assigned from) */

    begin_phase(COALESCE_FP_MOVES_TRANS);

    change = FALSE;
    removed_rtl = 0;

    /* Loop over each basic block starting from top */
    for (cblk = top; cblk; cblk = cblk->down)
    {
        /* Loop over all RTLs in the basic block */
        for (rtl = cblk->rtls; rtl; rtl = next)
        {
            next = rtl->next;
            /* Only look at actual RTLs and not other data */
            if (!cnts(rtl))
                continue;
            if (rtl->val
                && isreg(rtl->val)
                && regtype(rtl->val) == FTREG
                && !(regnum(rtl->val) & 1)   /* Only even regs */
                && *skiploc(rtl->val) == '='
                && isreg(skiploc(rtl->val)+1)
                && regtype(skiploc(rtl->val)+1) == FTREG
                && !(regnum(skiploc(rtl->val)+1) & 1)    /* Only even regs */
                && *skiploc(skiploc(rtl->val)+1) == ';'     /* End of rtl */
               )
            {
                /* Have to keep separate calls to regnum due to static */
                sprintf(lreg, "%s", getReg(FTREG, 1+regnum(rtl->val)));
                sprintf(rreg, "%s", 
                        getReg(FTREG, 1+regnum(skiploc(rtl->val)+1)));
                sprintf(match, "%s=%s;", lreg, rreg);
                len = 1+strlen(match);
                for (trtl = rtl->next; trtl; trtl = trtl->next)
                {
                    if (!(rtl->type & RTLINE))
                        continue;
                    if (!strncmp(match, trtl->val, len))
                    {
                        change = TRUE;

                        begin_trans(VET_OPT_TRANS);

                        sprintf(lreg, "%s", getReg(DTREG, regnum(rtl->val)));
                        sprintf(rreg, "%s", 
                                getReg(DTREG, regnum(skiploc(rtl->val)+1)));
                        sprintf(new_text, "%s=%s;", lreg, rreg);

                        add_rtl_with_type(cblk, trtl->next, new_text, rinit(), RTLINE);
                        remv_rtl(rtl);
                        remv_rtl(trtl);

                        end_trans();
                        break;
                    }
                }
            }
        }
    }

    end_phase(COALESCE_FP_MOVES_TRANS);


    return change;
}


/*
 * deopt_loop_inv_cdmotion - Move loop invariant code back into the loop
 */
int deopt_loop_inv_cdmotion()
{
    int i = 0,                  /* Count of loops */
        replace,                /* Whether or not RTL is reinsertable */
        writes,                 /* Whether or not loop has writes/calls */
        store_seen,             /* Are there stores in the loop? */
        call_seen,              /* Are there calls in the loop? */
        rename_type,            /* 0 none, 1 use/set, 2 use/dead */
        change = FALSE;         /* Change made to RTL's? */
    struct loopl *cur_loop;     /* The current loop we are examining */
    struct blist *bl;           /* List of blocks in the cur_loop */
    struct list *rtl,           /* RTL to reinsert */
                *trtl,          /* Potential RTL to examine  */
                *brtl;          /* Backwards RTL to examine (for renaming) */
    rlist rset,                 /* Regs set */
          rused,                /* Regs used */
          test_set,             /* Registers set in backtrack */
          live_regs,            /* Regs that are live in the loop */
          old_deads;            /* Old dead registers */
    vector vlive;               /* Live registers as a vector */
    char *j,                    /* Generic string pointer */
         *reg_to_rename,        /* Register candidate for renaming */
         old_rtl[MAXSTRING],    /* Old rtl text for mod_rtl */
         new_reg[MAXSTRING],    /* New register to use for renaming */
         rset_str[MAXSTRING],   /* String of registers set */
         rused_str[MAXSTRING];  /* String of registers used */

    rset = rinit();
    rused = rinit();
    test_set = rinit();
    live_regs = rinit();
    old_deads = rinit();

    /* Simple Loop from outermost (end of list) to innermost (start) */
    for (cur_loop = loope; cur_loop; cur_loop = cur_loop->prev)
    {
        i++;

        /* Only handle loops with a natural preheader */
        if ( ( cur_loop->preheader = natural_preheader(cur_loop->ptr)) == NULL)
            continue;

        call_seen = FALSE;

        /* Calculate dead registers so that we can move more stuff in */
        remv_dead_variables(1, TRUE, FALSE);
        begin_phase(DEOPT_LOOP_INV_CDMOTION_TRANS);

        /* Do the loop invariant analysis */
        writes = 0;
        store_seen = FALSE;
        loop_invariant_init(cur_loop);
        loop_invariant_analysis(cur_loop, TRUE, TRUE, &writes);

        /* If the loop reads are not invariant, check for store instructions */
        if (writes)
            store_seen = has_stores(cur_loop);

        /* Iterate in reverse through preheader */
        for (rtl = cur_loop->preheader->rtlend; rtl; rtl = rtl->prev)
        {
            if (!cnts(rtl))
                continue;

            regs_set(rtl, &rset);
            replace = TRUE;

            if (is_call(rtl->val))
            {
                /* Function calls cannot be invariant and also  */
                /* cause other rtls to no longer be invariant */
                call_seen = TRUE;
                replace = FALSE;
            }
            else if (is_a_load(rtl->val) && !cur_loop->invaread)
            {
                /* invaread detects stores and calls so now */
                /* we check to see if it is merely a local */
                /* with no stores or indirect references */
                if (!store_seen && gtloc(rtl->val) && !has_indref(cur_loop))
                    ;
                else
                    replace = FALSE;
            }
            else if (is_a_store(rtl->val))
            {
                /* Can't manipulate store instructions */
                replace = FALSE;
                store_seen = TRUE;
            }

            if (replace && !invariant_rtl(rtl, cur_loop->invareg))
            {
                /* Register Renaming */
                replace = FALSE; /* Flag to make sure renaming happens */
                /* First make sure that all sets are invariant */
                reget(rset, rset_str);
                for (j = rset_str; *j; j = skiploc(j) + 1)
                {
                    if (!rexists(j, cur_loop->invareg))
                        goto DEL_REGS;
                }

                regs_used(rtl, &rused);
                reget(rused, rused_str);
                reg_to_rename = NULL;
                rename_type = 0;

                for (j = rused_str; *j; j = skiploc(j) + 1)
                {
                    if (rexists(j, rset))
                    {
#ifdef SH_DEBUG
                        fprintf(stderr, "Found a set/reset\n");
#endif
                        if (!reg_to_rename)
                        {
                            reg_to_rename = j;
                            rename_type = 1;
                        }
                        else
                        {
                            rename_type = 0;
                            break;
                        }
                    }
                    else if (!rexists(j, cur_loop->invareg) &&
                             rexists(j, rtl->deads))
                    {
#ifdef SH_DEBUG
                        fprintf(stderr, "Found a dead set\n");
#endif
                        if (!reg_to_rename)
                        {
                            reg_to_rename = j;
                            rename_type = 2;
                        }
                        else
                        {
                            rename_type = 0;
                            break;
                        }
                    }
                }
                /* Only handle single register renames */
                if (rename_type != 0)
                {
                    live_regs = NULL;
#ifdef SH_DEBUG
                    fprintf(stderr, "Calculating live regs: ");
#endif
                    live_reg_calc(cur_loop, &live_regs);
#ifdef SH_DEBUG
                    rshow(live_regs);
                    fprintf(stderr, "\n");
#endif

                    vlive = regnew();
                    rlist_to_vector(live_regs, &vlive);
#ifdef SH_DEBUG
                    fprintf(stderr, "vlive: ");
                    vshow(vlive);
                    fprintf(stderr, "\n");
#endif

                    /* See if a new register can be allocated */
                    if (cancolor(RTREG, vlive, new_reg, 1))
                    {
#ifdef SH_DEBUG
                        fprintf(stderr, "Can find a rename reg: ");
                        show(new_reg);
#endif

                        /* Search for the previous set of the register */
                        /* we are looking to rename */
                        for (brtl = rtl->prev; brtl; brtl = brtl->prev)
                        {
                            if (!cnts(brtl))
                                continue;
#ifdef SH_DEBUG
                            fprintf(stderr, "back: ");
                            show(brtl->val);
#endif
                            regs_set(brtl, &test_set);

                            if (rexists(reg_to_rename, test_set))
                            {
#ifdef SH_DEBUG
                                fprintf(stderr, "Renaming "); /* DEBUG */
                                show(reg_to_rename);
                                fprintf(stderr, " --> ");
                                show(new_reg);
#endif

                                begin_trans(VET_OPT_TRANS);
                                change = TRUE;
                                replace = TRUE;
                                /* Do the renaming for the most recent */
                                /* set of reg_to_rename */
                                for (j = brtl->val; *j; j = skiploc(j)+1)
                                {
                                    if (j[0] == reg_to_rename[0] 
                                        && j[1] == reg_to_rename[1])
                                    {
                                        strcpy(old_rtl, brtl->val);
                                        j[0] = new_reg[0];
                                        j[1] = new_reg[1];
                                        mod_rtl(brtl, old_rtl, brtl->val);
                                        break;
                                    }
                                }
                                /* Do the renaming for the uses, */
                                /* noting that we may have multiple */
                                /* uses in a single line */
                                for (j = rtl->val; *j; j = skiploc(j)+1)
                                {
                                    if (j[0] == reg_to_rename[0]
                                        && j[1] == reg_to_rename[1])
                                    {
                                        strcpy(old_rtl, rtl->val);
                                        j[0] = new_reg[0];
                                        j[1] = new_reg[1];
                                        rcpy(&old_deads, brtl->deads);
                                        rdelete(reg_to_rename, &rtl->deads);
                                        mod_deads(rtl, old_deads, rtl->deads);
                                        mod_rtl(rtl, old_rtl, rtl->val);
                                    }
                                }
                                end_trans();
                                break;
                            }
                        }
                    }

                    mrfree(vlive);
                }
            }

            /* Check if we are replacing and if rtl regs are invariant */
            if (replace && invariant_rtl(rtl, cur_loop->invareg))
            {
                /* Traverse all basic blocks in the body of the loop */
                for (bl = cur_loop->blocks; bl; bl = bl->next)
                {
                    /* Look in each rtl of the basic block */
                    for (trtl = bl->ptr->rtls; trtl; trtl = trtl->next)
                    {
                        if (!cnts(trtl))
                            continue;
#ifdef SH_DEBUG
                        fprintf(stderr, "body %d ", i); /* DEBUG */
                        show(trtl->val);
#endif

                        regs_used(trtl, &rused);
                        /* If we use a set register, copy the set rtl */
                        if (rcommon(rused, rset))
                        {
#ifdef SH_DEBUG
                            fprintf(stderr, "Adding rtl\n"); /* DEBUG */
#endif
                            begin_trans(VET_OPT_TRANS);
                            copy_rtl(bl->ptr, trtl, rtl);
                            end_trans();
                            change = TRUE;
                        }
                    }
                }
            }
DEL_REGS:
            /* Remove any set register from the invariant register list */
            rminus(&cur_loop->invareg, rset);
        }
        end_phase(DEOPT_LOOP_INV_CDMOTION_TRANS);
    }
#ifdef SH_DEBUG
    fprintf(stderr, "%d loops processed\n", i); /* DEBUG */
#endif

    rfree(rset);
    rfree(rused);
    rfree(test_set);
    rfree(live_regs);
    rfree(old_deads);
    return change;
}


/*
 * has_calls - Detect presence of function calls in a loop
 */
int has_calls(struct loopl *cur_loop)
{
    struct blist *bl;   /* List of blocks in the loop */
    struct list *rtl;   /* RTL for iteration */

    for (bl = cur_loop->blocks; bl; bl = bl->next)
    {
        for (rtl = bl->ptr->rtls; rtl; rtl = rtl->next)
        {
            if (cnts(rtl) && is_call(rtl->val))
                return TRUE;
        }
    }
    return FALSE;
}


/*
 * has_stores - Detect presence of stores in a loop
 */
int has_stores(struct loopl *cur_loop)
{
    struct blist *bl;   /* List of blocks in the loop */
    struct list *rtl;   /* RTL for iteration */

    for (bl = cur_loop->blocks; bl; bl = bl->next)
    {
        for (rtl = bl->ptr->rtls; rtl; rtl = rtl->next)
        {
            if (cnts(rtl) && ismem(rtl->val))
                return TRUE;
        }
    }
    return FALSE;
}


/*
 * has_indref - Detect presence of indirect references in a loop
 */
int has_indref(struct loopl *cur_loop)
{
    struct blist *bl;       /* List of blocks in the loop */
    struct list *rtl;       /* RTL for iteration */
    char *c,                /* Start of actual RTL value */
         code[REGSTRING];   /* Code for local reference */
    int type,               /* Type of local reference */
        ind,                /* Is local reference indirect? */
        set;                /* Is local reference being set? */

    for (bl = cur_loop->blocks; bl; bl = bl->next)
    {
        for (rtl = bl->ptr->rtls; rtl; rtl = rtl->next)
        {
            if (c = cnts(rtl))
            {
                while (locget(&c, code, &type, &ind, &set, rtl))
                {
                    if (ind)
                        return TRUE;
                }
            }
        }
    }
    return FALSE;
}


/*
 * invariant_rtl - Check that rtl only uses invariant registers
 */
int invariant_rtl(struct list *rtl, rlist invareg)
{
    int invar = TRUE;           /* Is RTL invariant? */
    rlist rset,                 /* Regs set by RTL */
          rused;                /* Regs used by RTL */
    char *i,                    /* Iterate through rset_str */
         *j,                    /* Iterate through rused_str */
         rset_str[MAXSTRING],   /* String of registers set */
         rused_str[MAXSTRING];  /* String of registers used */

    if (invareg == NULL)
        return FALSE;

    rset = rinit();
    rused = rinit();

    regs_set(rtl, &rset);
    reget(rset, rset_str);

    /* Make sure registers being set are invariant for the loop */
    for (i = rset_str; *i; i = skiploc(i) + 1)
    {
        if (rexists(i, invareg))
        {
            regs_used(rtl, &rused);
            reget(rused, rused_str);
            for (j = rused_str; *j; j = skiploc(j) + 1)
            {
                /* Verify that each register that is used is invariant */
                if (!rexists(j, invareg))
                {
                    invar = FALSE;
                    break;
                }
            }
            rfree(rused);
        }
        else
            invar = FALSE;

        rfree(rset);
    }
    return invar;
}


/*
 * live_reg_calc - Calculate the live registers for a loop
 */
void live_reg_calc(struct loopl *cur_loop, rlist *live_regs)
{
    struct blist *bl;   /* List of blocks in a loop */
    struct list *rtl;   /* RTL for iteration */
    rlist rtmp=rinit(); /* Temp for holding regs used or set */

    rfree(*live_regs);
    *live_regs = rinit();

    /* Loop through all the blocks and collect the registers  */
    /* that are set or used by any RTLs */
    for (bl = cur_loop->blocks; bl; bl = bl->next)
    {
        for (rtl = bl->ptr->rtls; rtl; rtl = rtl->next)
        {
            regs_used(rtl, &rtmp);
            runion(live_regs, rtmp);
            rfree(rtmp);
            regs_set(rtl, &rtmp);
            runion(live_regs, rtmp);
            rfree(rtmp);
        }
    }
}


/*
 * gtglob - Get the next global variable
 */
char *gtglob(register char *s)
{
    if (!s) return NULL;
    for (;;)
    {
        while (*s)
        {
            if (isGlobal(s))
                return (s + GLOBALLOCLENGTH);
            s = skiploc(s);
        }
        return ((char *) NULL);
    }
}


/*
 * rlist_to_vector - Convert an rlist into a vector
 */
void rlist_to_vector(rlist rl, vector *vec)
{
    char rl_str[MAXSTRING],
         *c;

    reget(rl, rl_str);

    for (c = rl_str; *c; c = skiploc(c) + 1)
    {
        touchreg(*vec, c);
    }
}


/*
 * deopt_reg_alloc - Deoptimize register allocation
 */
int deopt_reg_alloc()
{
    int i,
        changes = FALSE,
        reg_type,
        pseudosize,
        merge_use_set_ranges,
        cur_regallocptr,
        hi_locnum = -1;
    struct bblock *cblk;
    struct list *rtl;
    struct rig_node *rn,
                    *sib;
    struct ruse_node *rund;
    char *c,
         *s,
         *a;
    extern int valid_analysis;
    struct setlist *sptr;
    char new_rtl[MAXSTRING],
         newregname[REGSTRING],
         newloccode[REGSTRING],
         newlocname[MAXSTRING],
         new_decl[MAXSTRING];
    rlist old_deads;
    struct locuse *lu;

    /*
     * Deoptimizing Register Allocation
     * --------------------------------
     * 1) Build the list of registers that cannot be replaced for the
     *    current architecture.
     * 2) Construct the RIG itself, performing the necessary merging
     *    of RIG nodes on completion
     * 3) Replace all candidate middle registers with pseudoregisters
     * 4) Replace spanning registers with stores/loads as required
     * 5) Destroy the RIG appropriately
     * 6) Deallocate the non-replaceable register list
     */

    begin_phase(DEOPT_REG_ALLOC_TRANS);
    init_deopt(&no_replace_regs);

    merge_use_set_ranges = 1;
    build_rig(merge_use_set_ranges);
#ifdef SH_DEBUG
    fprintf(stderr, "Flagged for merging use/set ranges (Steve)\n");
    display_rig2();
#endif

    if (!merge_use_set_ranges)
    {
        /* For Prasad's version, we must recalculate the end of the */
        /* hardware registers, locals and other declarations. This */
        /* method will eliminate the excess pseudoregisters. */
        for (i = hregnum; i < regallocptr; i++)
        {
            if (isreg(sets[i].item) && pseudoreg(sets[i].item))
                break;
        }
        regallocptr = i;
    }

    old_regallocptr = regallocptr;

    /* Calculate the current maximum local var number */
    for (lu = locs; lu; lu = lu->next)
        if (locnum(lu->code) > hi_locnum)
            hi_locnum = locnum(lu->code);

    for (rn = rig_top; rn; rn = rn->next)
        rn->done = FALSE;

    for (rn = rig_top; rn; rn = rn->next)
    {
        if (!rn->can_replace || rn->done)
            continue;

        rn->done = TRUE;
        /* Middle register if it has no siblings and is replaceable */
        /* (e.g. live range is contained in a single basic block) */
        if (rn->psib == rn)
        {
            /* Check that block is truly replaceable */
            if ((sptr = s_lookup(rn->reg_set)) && vin(rn->non_alloc,
                        regtoindex(regtype(rn->reg_set),
                            regnum(rn->reg_set))))
                continue;

            reg_type = regtype(rn->reg_set);
            makereg(rn->pseudo, reg_type, new_pseudoreg(reg_type));

        }
        else
        {
            /* Register spans multiple blocks, so we must break it down */
            hi_locnum++;
            newloccode[0] = 0xa0 | (hi_locnum>>7);
            newloccode[1] = 0x80 | (hi_locnum & 0x7f);
            newloccode[2] = '\0';
            snprintf(newlocname, REGSTRING, "__%s_%d", fn, hi_locnum);
            snprintf(new_decl, MAXSTRING, "d%s\t%c%c\t2\t%d\t%d", 
                    newlocname, newloccode[0], newloccode[1], 
                    regtype(rn->reg_set), regsize(regtype(rn->reg_set)));
            fprintf(stderr, "decl: %s\n", new_decl);

            declaration(copy_string(new_decl));

            reg_type = regtype(rn->reg_set);
            makereg(rn->pseudo, reg_type, new_pseudoreg(reg_type));
            rn->memloc = localrecord(newloccode);
            rn->memloc->atype = reg_type;
            rn->memloc->memChar = getmemchar(reg_type);

            for (sib = rn->nsib; sib != rn; sib = sib->nsib)
            {
                sib->done = TRUE;
                copyloc(sib->pseudo, rn->pseudo);
                *(skiploc(sib->pseudo)) = 0;
                sib->memloc = rn->memloc;
            }
        }
    }
    rveclen = regallocptr >> intscale;
    if (regallocptr & bitsperint)
        rveclen++;

    for (cblk = top; cblk; cblk = cblk->next)
    {
        rfree(cblk->dlist);
        rfree(cblk->ulist);
        rfree(cblk->olist);

        for (rtl = cblk->rtls; rtl; rtl = rtl->next)
            if (rtl->deads)
                rtl->deads = resize_rlist(rtl->deads);
    }

    cur_regallocptr = old_regallocptr;

    for (rn = rig_top; rn; rn = rn->next)
        rn->done = FALSE;

    for (rn = rig_top; rn; rn = rn->next)
    {
        /*if (rn->done) */
        /*    continue; */

        if (rn->can_replace && (s = rn->pseudo) && *s && regtovec(s) < 0)
        {
            if (cur_regallocptr == MAXRVEC)
            {
                fprintf(stderr, 
                        "deopt_reg_alloc - Too many registers for vector\n");
                exit(1);
            }

            /* No checking for afifo here, since all pseudoregs */
            sets[cur_regallocptr].status &= ~FIFOREG;
            a = sets[cur_regallocptr].item;
            copyloc(a,s);
            a = skiploc(a);
            *a = '\0';
#ifdef SH_DEBUG
            fprintf(stderr, "added: ");
            show(sets[cur_regallocptr].item);
#endif
            sets[cur_regallocptr].num = cur_regallocptr;
            vecmap[regname(cur_regallocptr)] = cur_regallocptr;
            cur_regallocptr++;
        }
    }

    i = old_regallocptr;
    for (pseudos = rinit(); i < cur_regallocptr; i++)
        rins((unsigned int) i, &pseudos);
#ifdef SH_DEBUG
    rshow(pseudos);
#endif

    old_deads = rinit();
    for (rn = rig_top; rn; rn = rn->next)
    {
        if (!rn->can_replace || rn->done)
            continue;
        rn->done = TRUE;

        if (!rn->pseudo[0])
        {
            fprintf(stderr, "No pseudo allocated ???\n");
            continue;
        }

        if (!rn->memloc)
        {
            /* Handle middle replacements */
            for (rund = rn->sets; rund; rund = rund->next)
                replace_reg(rund->rtl, rn->reg_set, rn->pseudo);
            for (rund = rn->uses; rund; rund = rund->next)
                replace_reg(rund->rtl, rn->reg_set, rn->pseudo);

            if (rn->finish)
            {
                begin_trans(VET_OPT_TRANS);
                rcpy(&old_deads, rn->finish->deads);
                rdelete(rn->reg_set, &rn->finish->deads);
                rinsert(rn->pseudo, &rn->finish->deads);
                mod_deads(rn->finish, old_deads, rn->finish->deads);
                end_trans();
            }
        }
        else
        {
            /* Handle multi-range replacements */
            sib = rn;
            do
            {
                /* Replace uses first in case of use/set (that way the */
                /* old value is not accidentally used) */
                /* TODO: maybe use rn->pseudo and such for better code */
                /* TODO: Figure out why code for mem is not generated */

                fprintf(stderr, "after s_lookup: %p\n", sptr);

                sib->done = TRUE;
                for (rund = sib->uses; rund; rund = rund->next)
                {
                    begin_trans(VET_OPT_TRANS);
                    load(sib->memloc->type, FALSE, sib->memloc->code,
                         rund->rtl->daddy, rund->rtl, sib->pseudo, FALSE);
                    end_trans();
                    replace_reg(rund->rtl, sib->reg_set, sib->pseudo);
                }
                for (rund = sib->sets; rund; rund = rund->next)
                {
                    begin_trans(VET_OPT_TRANS);
                    store(sib->memloc->type, FALSE, sib->memloc->code,
                         rund->rtl->daddy, rund->rtl->next, sib->pseudo, FALSE);
                    end_trans();
                    replace_reg(rund->rtl, sib->reg_set, sib->pseudo);
                    begin_trans(VET_OPT_TRANS);
                    rcpy(&old_deads, rund->rtl->next->deads);
                    rinsert(sib->pseudo, &rund->rtl->next->deads);
                    mod_deads(rund->rtl->next, old_deads, rund->rtl->next->deads);
                    end_trans();
                }

                if (sib->finish)
                {
                    begin_trans(VET_OPT_TRANS);
                    rcpy(&old_deads, sib->finish->deads);
                    rdelete(sib->reg_set, &sib->finish->deads);
                    rinsert(sib->pseudo, &sib->finish->deads);
                    mod_deads(sib->finish, old_deads, sib->finish->deads);
                    end_trans();
                }
                sib = sib->nsib;
            } while (sib != rn);
        }
    }
    fprintf(stderr, "Finished replacements\n");

    destroy_rig();
    rfree(no_replace_regs);
    end_phase(DEOPT_REG_ALLOC_TRANS);

    /*if (!(~valid_analysis & DEPTH_FIRST_SEARCH_ANALYSIS)) */
        depth_first_search(1);
    /*if (!(~valid_analysis & LIVE_VAR_ANALYSIS)) */
        live_variable_analysis(1, trans_done & REG_ASSIGNMENT_BIT, FALSE);

    local_register_assignment(1);
    number_rtls();

    return changes;
}


/*
 * replace_reg - Replace oldreg with newreg in RTL
 */
void replace_reg(struct list *rtl, char *old_reg, char *new_reg)
{
    char new_rtl[MAXSTRING],
         *c;

    strcpy(new_rtl, rtl->val);
    for (c = new_rtl; *c; c = skiploc(c))
        if (!cmploc(c, old_reg))
        {
            copyloc(c, new_reg);
        }
    begin_trans(VET_OPT_TRANS);
    update_rtl(rtl, new_rtl);
    end_trans();
}


/*
 * resize_rlist - Resize a given rlist to newsize maintaining data
 */
rlist resize_rlist(rlist oldrl)
{
    rlist newrl;
    unsigned int *p1, *p2, *end, old_rveclen;

    old_rveclen = old_regallocptr >> intscale;
    if (old_regallocptr & bitsperint)
        old_rveclen++;

    if (p2 = (unsigned int *) oldrl)
    {
        newrl = rinit();
        newrl = (rlist) (p1 = RALLOC);
        for (end = p1 + old_rveclen; p1 != end; *p1++ = *p2++);
        for (end = newrl + rveclen; p1 != end; *p1++ = 0);
    }
    else
        newrl = rinit();

    rfree(oldrl);
    return newrl;
}


/*
 * build_rig - Build an IG for registers, not locals
 *     merge_use_set_ranges - merge use and set in same rtl into 1 live range?
 */
void build_rig(int merge_use_set_ranges)
{
    struct bblock *cblk;        /* Basic Block pointer for traversal */
    struct list *rtlptr;        /* RTL pointer for traversal */
    char *rtlin,                /* RTL value pointer for traversal */
         reg[BIGSTRING];        /* Register names */
    struct setlist *sptr;       /* Setlist pointer */
    struct rig_node *rn;        /* Node for constructing RIG */
    rlist rl;                   /* rlist of registers */

    rig_top = NULL;
    rig_end = NULL;
    rig_nums = 0;
    rl = rinit();

    /* Traverse all basic blocks */
    for (cblk = top; cblk; cblk = cblk->down)
    {
        cblk->blkins = NULL;

        /* Get all the registers that are live on entry to the block */
        reget(cblk->ulist, reg);
        for (rtlin = reg; *rtlin; rtlin = skiplocOr2(rtlin) + 1)
        {
            if ((sptr = s_lookup(rtlin)) && isreg(rtlin))
            {
                rn = new_rig_node();
                /*regtoindex(RTREG,sptr->item); */
                /*regnum(rtlin); */
                strcpy(rn->reg_set, rtlin);
                rn->block = cblk;
                rinsert(rn->reg_set, &rl);
                if (rcommon(rl, no_replace_regs))
                    rn->can_replace = 0;
                rdelete(rn->reg_set, &rl);
                if (rn->blkins = rn->blkouts = cblk->blkins)
                    cblk->blkins->blkprev = rn;
                cblk->blkins = rn;
            }
        }

        /* Now initialize the outgoing regs as the incoming regs */
        cblk->blkouts = cblk->blkins;

        /* Traverse the block's RTLs for more live ranges */
        for (rtlptr = cblk->rtls; rtlptr; rtlptr = rtlptr->next)
        {

            /* Collect used regs first, since we may not be  */
            /* doing merge_use_set_ranges */

            /* Adjust registers used in the rtl itself */
            regs_used(rtlptr, &rl);
            reget(rl, reg);

            if (rtlptr->type == PARMLINE)
            {
                /* Handle parmlines by noting it in RIG nodes */
                for (rtlin = reg; *rtlin; rtlin = skiplocOr2(rtlin) + 1)
                {
                    for (rn = cblk->blkouts; rn; rn = rn->blkouts)
                    {
                        touchreg(rn->non_alloc, rtlin);
                        if (!strcmp(rn->reg_set, rtlin))
                            rn->can_replace = FALSE;
                    }
                }
            }
            else
            {
                for (rtlin = reg; *rtlin; rtlin = skiplocOr2(rtlin) + 1)
                {
                    rig_add_use(rtlptr, rtlin);
                }
            }

            /* Adjust registers set in the rtl itself */
            regs_set(rtlptr, &rl);
            reget(rl, reg);
            if (rtlptr->type == RESLINE)
            {
#ifdef SH_DEBUG
                fprintf(stderr, "RESLINE !!!\n");
#endif
                rn = rig_add_set(rtlptr, rtlin, merge_use_set_ranges);
                rn->can_replace = FALSE;
                /*for (rn = cblk->blkouts; rn; rn = rn->blkouts)
                {
                    if (!strcmp(rn->reg_set, rtlin))
                    {
                    }
                }*/
            }
            else if (!merge_use_set_ranges && rtlptr->type == SETLINE)
            {
                /* Handle setlines when not merging ranges */
                for (rtlin = reg; *rtlin; rtlin = skiplocOr2(rtlin) + 1)
                {
                    for (rn = cblk->blkouts; rn; rn = rn->blkouts)
                    {
                        touchreg(rn->non_alloc, rtlin);
                        if (!strcmp(rn->reg_set, rtlin))
                            rn->can_replace = FALSE;
                    }
                }
            }
            else if (rtlptr->type != SETLINE)
            {
                for (rtlin = reg; *rtlin; rtlin = skiplocOr2(rtlin) + 1)
                {
                    rig_add_set(rtlptr, rtlin, merge_use_set_ranges);
                }
            }

            /* Work through the dead list to calculate end of live ranges */
            reget(rtlptr->deads, reg);
            for (rtlin = reg; *rtlin; rtlin = skiplocOr2(rtlin) + 1)
            {
                for (rn = cblk->blkouts; rn; rn = rn->blkouts)
                {
                    if (!rn->finish && !strcmp(rn->reg_set, rtlin))
                    {
                        rn->finish = rtlptr;
                        if (rn->blkouts)
                            rn->blkouts->blkprev = rn->blkprev;
                        if (rn->blkprev)
                            rn->blkprev->blkouts = rn->blkouts;
                        else
                            cblk->blkouts = rn->blkouts;
                        rn->blkouts = NULL;
                    }
                }
            }
        }
        /*if (!verify_outs(cblk))
        {
            fprintf(stderr, "Error: verify_outs\n");
            fprintf(stderr, "olist: ");
            rshow(cblk->olist);
            fprintf(stderr, "\nblkouts: ");
            for (rn = cblk->blkouts; rn; rn = rn->blkouts)
            {
                show_rtl(stderr, rn->reg_set);
            }
            fprintf(stderr, "\n");
        }*/
    }
    rfree(rl);

    merge_rig_nodes();
}


/*
 * rig_add_use - Add a used register to the appropriate RIG node
 */
void rig_add_use (struct list *rtl, char *ruse)
{
    struct rig_node *rn;
    struct ruse_node *rund;

    /* Add a register use to the appropriate RIG node */
    for (rn = rtl->daddy->blkouts; rn; rn = rn->blkouts)
    {
        if (!rn->finish && !strcmp(rn->reg_set, ruse))
        {
            rund = SALLOC(sizeof(struct ruse_node));
            rund->rtl = rtl;
            rund->next = rn->uses;
            rn->uses = rund;
            return;
        }
    }
#ifdef SH_DEBUG
    fprintf(stderr, "Error (rig_add_use): Could not find a use of "
            "%d in block %d, rtl %d\n", regnum(ruse), 
            rtl->daddy->block_id, rtl->rtl_id);
#endif
    /*exit(1); */

    /* This happens when we are either looking at a parameter or a */
    /* return value from a called function, so we need to add a new */
    /* node appropriately */

    rn = new_rig_node();
    strcpy(rn->reg_set, ruse);
    rn->block = rtl->daddy;
    if (rn->blkouts = rn->block->blkouts)
        rn->block->blkouts->blkprev = rn;
    rn->block->blkouts = rn;

    rn->can_replace = FALSE;
    rund = SALLOC(sizeof(struct ruse_node));
    rund->rtl = rtl;
    rund->next = rn->uses;
    rn->uses = rund;
}


/*
 * rig_add_set - Add a set register as a new RIG node
 */
struct rig_node * rig_add_set (struct list *rtl, char *rset, int merge_use_set_ranges)
{
    struct rig_node *rn;
    struct ruse_node *rund;
    rlist rl;
    int found = 0;

    /* First look for the RIG node in the current blocks */
    for (rn = rtl->daddy->blkouts; rn; rn = rn->blkouts)
    {
        /* If it is found, then possibly this is the actual set of the  */
        /* register. Add info if necesary */
        if (!strcmp(rn->reg_set, rset) && !rn->finish)
        {
            if (merge_use_set_ranges)
            {
#ifdef SH_DEBUG
                fprintf(stderr, "Found a use/set: ");
                show(rtl->val);
#endif
                found = 1;
            }
            else
            {
                /*fprintf(stderr, "Splitting the live range: "); */
                /*show(rtl->val); */
                /*rn->finish = rtl; */
                if (rn->blkouts)
                    rn->blkouts->blkprev = rn->blkprev;
                if (rn->blkprev)
                    rn->blkprev->blkouts = rn->blkouts;
                else
                    rn->block->blkouts = rn->blkouts;
                rn->blkouts = NULL;
            }
            break;
        }
    }

    /* If it is not currently in a RIG node, then add it with the  */
    /* appropriate fields */
    if (!found)
    {
        rn = new_rig_node();
        strcpy(rn->reg_set, rset);
        rn->block = rtl->daddy;
        rn->start = rtl;
        if (rn->blkouts = rn->block->blkouts)
            rn->block->blkouts->blkprev = rn;
        rn->block->blkouts = rn;
    }

    rl = rinit();
    rinsert(rn->reg_set, &rl);

    /* If register is hardware specific (e.g. 0, SP, FP, ...) or if */
    /* the set is from a function call return, then it is not */
    /* replaceable */
    if (rcommon(rl, no_replace_regs))
        rn->can_replace = FALSE;
    else
        rn->can_replace = (rtl->type != RESLINE);
    rfree(rl);

    rund = SALLOC(sizeof(struct ruse_node));
    rund->rtl = rtl;
    rund->next = rn->sets;
    rn->sets = rund;
    return rn;
}


/*
 * merge_rig_nodes - Merge the lifespans for registers in the RIG
 */
void merge_rig_nodes()
{
    struct bblock *cblk;
    struct blist *bptr;
    struct rig_node *rn1,
                    *rn2,
                    *rn3;

    /* Starting block ins can never be replaced */
    if (top)
    {
#ifdef SH_DEBUG      
        fprintf(stderr, "TOP: %d\n", top->block_id);*/
#endif	  
        for (rn1 = top->blkins; rn1; rn1 = rn1->blkins)
            rn1->can_replace = FALSE;
    }

    if (ret_block)
    {
#ifdef SH_DEBUG      
        fprintf(stderr, "RET_BLOCK: %d\n", ret_block->block_id);
#endif	
        for (bptr = ret_block->preds; bptr; bptr = bptr->next)
        {
            for (rn1 = bptr->ptr->blkouts; rn1; rn1 = rn1->blkouts)
            {
                rn1->can_replace = FALSE;
            }
        }
    }

    /* Loop over all basic blocks */
    for (cblk = top; cblk; cblk = cblk->down)
    {
        /* Search through each of the predecessors of the cblk */
        for (bptr = cblk->preds; bptr; bptr = bptr->next)
        {
            /* Look at the cblk incoming registers */
            for (rn2 = cblk->blkins; rn2; rn2 = rn2->blkins)
            {
                /* Loop until we find a match or run out of choices */
                for (rn1 = bptr->ptr->blkouts; rn1 && strcmp(rn1->reg_set,
                     rn2->reg_set); rn1 = rn1->blkouts)
                    ;
                if (rn1)
                {
                    for (rn3 = rn1->nsib; rn3 != rn1 && rn3 != rn2; rn3 = rn3->nsib)
                        ;
                    if (rn3 != rn2)
                    {
                        /* How to handle problem of r[18]=r[18]+1 ? */
                        rn2->psib->nsib = rn1->nsib;
                        rn1->nsib->psib = rn2->psib;
                        rn1->nsib = rn2;
                        rn2->psib = rn1;
                        if (!rn2->can_replace || !rn1->can_replace)
                        {
                            for (rn3 = rn2->nsib; rn3 != rn2; rn3 = rn3->nsib)
                                rn3->can_replace = FALSE;
                        }
                        /*fprintf(stderr, "rn2: ");
                        print_rig_node(rn2);
                        fprintf(stderr, "rn3: ");
                        print_rig_node(rn3);
                        fprintf(stderr, "%p %p\n", rn2, rn3);*/
                    }
                }
            }
        }
    }
}


/*
 * destroy_rig - Clean up all data structures for the RIG
 */
void destroy_rig()
{
    struct rig_node *rn, *next;

    for (rn = rig_top; rn; rn = next)
    {
        next = rn->next;
        destroy_rig_node(rn);
    }

    rig_top = rig_end = NULL;
}


/*
 * display_rig - Debugging printer for the RIG
 */
void display_rig()
{
    struct rig_node *rn;
    struct ruse_node *rund;

    fprintf(stderr, "Register Interference Graph\n");
    /* Type, RegNumber, Block, Start, Finish, Sets of Reg, Uses of Reg */
    fprintf(stderr, "?T[ RN ]| BL | ST | FN | SETS | USES | SIBS | Vect |\n");
    rn = rig_top;
    for (rn = rig_top; rn; rn = rn->next)
    {
        print_rig_node(rn);
        fprintf(stderr, "---------------------------------\n");
    }
}


/*
 * display_rig2 - Debugging printer for the RIG (more complex)
 */
void display_rig2()
{
    struct bblock *cblk;
    struct rig_node *rn;

    for (rn = rig_top; rn; rn = rn->next)
        rn->done = 0;

    fprintf(stderr, "Register Interference Graph 2\n");
    /* Type, RegNumber, Block, Start, Finish, Sets of Reg, Uses of Reg */
    fprintf(stderr, "?T[ RN ]| BL | ST | FN | SETS | USES | SIBS | Vect |\n");
    rn = rig_top;
    for (cblk = top; cblk; cblk = cblk->down)
    {
        fprintf(stderr, "Ins\n");
        for (rn = cblk->blkins; rn; rn = rn->blkins)
        {
            rn->done = 1;
            print_rig_node(rn);
            fprintf(stderr, "---------------------------------\n");
        }
        fprintf(stderr, "Outs\n");
        for (rn = cblk->blkouts; rn; rn = rn->blkouts)
        {
            rn->done = 1;
            print_rig_node(rn);
            fprintf(stderr, "---------------------------------\n");
        }
    }

    fprintf(stderr, "Middles\n");
    for (rn = rig_top; rn; rn = rn->next)
    {
        if (!rn->done)
        {
            print_rig_node(rn);
            fprintf(stderr, "---------------------------------\n");
        }
        /*if (rn->psib == rn && rn->start && rn->finish &&
                rn->start->daddy == rn->finish->daddy)
        {
            print_rig_node(rn);
            fprintf(stderr, "---------------------------------\n");
        }*/
    }
}


/*
 * print_rig_node - Print information to the screen for a RIG node structure
 */
void print_rig_node(struct rig_node *rn)
{
    struct ruse_node *rund;
    struct rig_node *rs;

    if (!rn)
        return;

    fprintf(stderr, "%c%c[%4d]|%4d|", (rn->can_replace? ' ':'X'), regprt[regtype(rn->reg_set)], regnum(rn->reg_set), rn->block->block_id);
    if (rn->start)
        fprintf(stderr, "%4d|", rn->start->rtl_id);
    else
        fprintf(stderr, "    |");
    if (rn->finish)
        fprintf(stderr, "%4d|", rn->finish->rtl_id);
    else
        fprintf(stderr, "    |");
    for (rund = rn->sets; rund; rund = rund->next)
    {
        fprintf(stderr, "%d ", rund->rtl->rtl_id);
    }
    fprintf(stderr, "| ");
    for (rund = rn->uses; rund; rund = rund->next)
    {
        fprintf(stderr, "%d ", rund->rtl->rtl_id);
    }
    fprintf(stderr, "| ");
    for (rs = rn->nsib; rs != rn; rs = rs->nsib)
    {
        fprintf(stderr, "%d ", rs->block->block_id);
    }
    fprintf(stderr, "| ");
    vshow(rn->non_alloc);
    fprintf(stderr, " |\n");
}


/*
 * new_rig_node - Create a new node for the Register Interference Graph
 */
struct rig_node *new_rig_node()
{
    struct rig_node *rn;
    rn = SALLOC(sizeof(struct rig_node));

    if (rig_top)
    {
        rn->prev = rig_end;
        rig_end->next = rn;
        rig_end = rn;
    }
    else
    {
        rn->prev = NULL;
        rig_top = rig_end = rn;
    }
    rn->next = NULL;

    /* Generic settings */
    rn->blkprev = NULL;
    rn->blkins = NULL;
    rn->blkouts = NULL;
    rn->psib = rn->nsib = rn;

    rn->start = NULL;
    rn->finish = NULL;

    strcpy(rn->pseudo, "");
    rn->memloc = NULL;
    rn->sets = NULL;
    rn->uses = NULL;

    rn->can_replace = TRUE;
    rn->non_alloc = regnew();
    rig_nums++;
    return rn;
}


/*
 * destroy_rig_node - Cleans up a RIG node's memory allocation
 */
void destroy_rig_node(struct rig_node *rn)
{
    struct ruse_node *rund;
    if (!rn)
        return;

    for (rund = rn->sets; rund; rund = rn->sets)
    {
        rn->sets = rn->sets->next;
        FREE(rund);
    }
    for (rund = rn->uses; rund; rund = rn->uses)
    {
        rn->uses = rn->uses->next;
        FREE(rund);
    }
    if (rn->prev)
        rn->prev->next = rn->next;
    else
        rig_top = rn->next;

    if (rn->next)
        rn->next->prev = rn->prev;
    else
        rig_end = rn->prev;
    vfree(rn->non_alloc);
    FREE(rn);
    rig_nums--;
}


/*
 * verify_outs - Verify that olist and blkouts of a basic block match exactly
 */
static int verify_outs(struct bblock *cblk)
{
    struct rig_node *rn;
    char regs_str[MAXSTRING],
         *c;
    int ok,
        ctr;

    /* Using the olist search the blkouts for matches */
    /* Keep a count of the regs so that we can verify in both directions */
    reget(cblk->olist, regs_str);
    ctr = 0;
    for (c = regs_str; *c; c = skiploc(c) + 1)
    {
        if (!isreg(c))
            continue;
        ok = 0;
        ctr++;
        for (rn = cblk->blkouts; !ok && rn; rn = rn->blkouts)
            if (!strcmp(c, rn->reg_set))
                ok = 1;
        if (!ok)
            return 0;
    }
    for (rn = cblk->blkouts; rn; rn = rn->blkouts)
        ctr--;
    return (ctr == 0);
}

/**********************
 * Routines from Bill *
 **********************/


/*
 * reginvect - Determine if a register in a vector
 *
 * Parameters: reg: register we are interested in
 *             regs: vector of registers
 */
int reginvect(char *reg, vector regs)
{
   int             index,
                   rt,
                   rcont;
   extern int      intscale,
                   bitsperint;

   rt = regtype(reg);
   index = regtoindex(rt, regnum(reg));
   for (rcont = contreg(rt); rcont--; index++)
      if (regs[index >> intscale] & (((unsigned) 1) << (index & bitsperint)))
         return TRUE;
   return FALSE;
}


/*
 * vshow - Show the contents of a vector
 */
void vshow(vector vlist)
{
   int i;
   char reg_string[REGSTRING];

   for (i = 0; i < 32; i++)
   {
      makereg(reg_string, RTREG, indextoreg(RTREG, i));
      if (reginvect(reg_string, vlist))
         show_rtl(stderr, reg_string);
   }
   for (i = 0; i < 32; i++)
   {
      makereg(reg_string, FTREG, indextoreg(FTREG, i));
      if (reginvect(reg_string, vlist))
         show_rtl(stderr, reg_string);
   }
   fflush(stderr);
}


/*
 * rminus - subtracts the bits of one vector from another
 */
void rminus(rlist * result, rlist vec)
{

   register unsigned int *ptr1, *ptr2, *end;
#ifdef VET
   rlist           rold;
   void            x_send_deads(), rsave();
#endif

#ifdef VET
   rsave(&rold, *result);
#endif

   if (ptr2 = (unsigned int *) vec)
      if (ptr1 = (unsigned int *) *result)
     for (end = ptr1 + rveclen; ptr1 != end; *ptr1++ &= ~(*ptr2++));
#ifdef VET
   x_send_deads(rold, result);
#endif
}


/*
 * is_reg_affected - determines if a register is set by an RTL line
 *
 * Parameters:       ptr: An rtl value line
 *                   reg: Representation of a register
 */
int is_reg_affected(struct list *ptr, char *reg)
{
   char *rtlin;

   if (cnts(ptr))
   {
      rtlin = ptr->val;
      while (rtlin = gtrg(rtlin))
	if (*rtlin == '=' && cmploc(unskiploc(rtlin), reg)==0)
	  return TRUE;
      if (rtlin = ptr->se)
	while (rtlin = gtrg(rtlin))
	  if (*rtlin == '=' && cmploc(unskiploc(rtlin), reg)==0)
	    return TRUE;
   }
   else if ((ptr->type & RESLINE) && cmploc(ptr->val, reg)==0)
     return TRUE;

   else if (ptr->type & SETLINE)
     for (rtlin = ptr->val; *rtlin; rtlin = skiploc(rtlin))
       if (cmploc(rtlin, reg)==0)
	 return TRUE;
   return FALSE;
}

/*
 * reg_set_in_block - checks if a register is set in a block
 *
 * Parameters:        cblk: Block we would like to check
 *                    reg: Register we want to know has been set or not
 */
int reg_set_in_block(struct bblock *cblk, char *reg)
{
   struct list    *ptr;

   for (ptr = cblk->rtls; ptr; ptr = ptr->next)
      if (is_reg_affected(ptr, reg))
         return TRUE;
   return FALSE;
}

/*
 * regs_set -  determine the registers set by an RTL line
 *
 * Parameters: ptr: An RTL value line
 *             regs: place to store the regs set by this RTL 
 */
void regs_set(struct list *ptr, rlist *regs)
{
   char *rtlin;

   rfree(*regs);
   *regs = rinit();
   if (cnts(ptr))
   {
      rtlin = ptr->val;
      while (rtlin = gtrg(rtlin))
         if (*rtlin == '=')
            rinsert(unskiploc(rtlin), regs);
      if (rtlin = ptr->se)
         while (rtlin = gtrg(rtlin))
            if (*rtlin == '=')
               rinsert(unskiploc(rtlin), regs);
   }
   else if (ptr->type & RESLINE)
   {
      rinsert(ptr->val, regs);
   }
   else if (ptr->type & SETLINE)
   {
      for (rtlin = ptr->val; *rtlin; rtlin = skiploc(rtlin))
         rinsert(rtlin, regs);
   }
}

/*
 * regs_used - determine the registers used by an RTL line
 *
 * Parameters: ptr: An RTL value line
 *             regs: place to store the regs used by this RTL 
 */
void regs_used(struct list *ptr, rlist *regs)
{
   char *rtlin;

   rfree(*regs);
   *regs = rinit();
   if (!ptr)
      return;
   if (cnts(ptr))
   {
      rtlin = ptr->val;
      while (rtlin = gtrg(rtlin))
         if (*rtlin != '=')
            rinsert(unskiploc(rtlin), regs);
      if (rtlin = ptr->se)
         while (rtlin = gtrg(rtlin))
            if (*rtlin != '=')
               rinsert(unskiploc(rtlin), regs);
   }
   else if (ptr->type & PARMLINE)
   {
      for (rtlin = ptr->val; *rtlin; rtlin = skiploc(rtlin))
         rinsert(rtlin, regs);
   }
}

/*
 * regs_set_before_used - determine the registers that are set before being
 *                        used in a block
 * 
 * Parameters:            cblk: Block we are interested in
 *                        set_before_used: space to store the registers
 */
void regs_set_before_used(struct bblock *cblk, rlist *set_before_used)
{
   struct bblock  *tblk;
   struct list    *ptr;
   rlist           used, usedinrtl, setinrtl;

   for (tblk = top; tblk; tblk = tblk->down)
      tblk->status &= ~DONE;
   tblk = cblk;
   do
   {
      used = rinit();
      usedinrtl = rinit();
      setinrtl = rinit();

      rfree(*set_before_used);
      *set_before_used = rinit();
      for (ptr = cblk->rtls; ptr; ptr = ptr->next)
      {
         regs_used(ptr, &usedinrtl);
         runion(&used, usedinrtl);
         regs_set(ptr, &setinrtl);
         rminus(&setinrtl, used);
         runion(set_before_used, setinrtl);
      }
      tblk->status |= DONE;
      if (tblk->left && !tblk->right)
         tblk = tblk->left;
   } while (tblk && !(tblk->status & DONE));

   rfree(used);
   rfree(usedinrtl);
   rfree(setinrtl);
}

/*
 * assign_pseudo_regs - assign pseudo regs to all the nodes in the RIG
 */
void assign_pseudo_regs()
{
  int reg_type, reg_num;
  struct setlist *sptr;
  struct rig_node *rn, *sib, *ig_ptr;

  /*mark all nodes undone*/
  for(rn=rig_top ; rn ; rn=rn->next)
    rn->done = FALSE;

  /*for each RIG node assign pseudo regs, if not already done*/
  for(rn=rig_top ; rn ; rn=rn->next){

    if(rn->can_replace){
      if (!((sptr = s_lookup(rn->reg_set)) &&
	    vin(rn->non_alloc,regtoindex(regtype(rn->reg_set),regnum(rn->reg_set))))){
	reg_type = regtype(rn->reg_set);
	reg_num = new_pseudoreg(reg_type);
	if(reg_num != -1){
	  make_pseudo_reg(rn->pseudo, reg_type, reg_num);
	  /*rn->done = TRUE; */
	  
	  /*all siblings of this node should have the same pseudo reg*/
	  for(ig_ptr=rn->nsib ; ig_ptr!=rn ; ig_ptr=ig_ptr->nsib){
	    strcpy(ig_ptr->pseudo, rn->pseudo);
	    ig_ptr->done = TRUE;
	  }
	}
      }
    }
  }
}

/*
 * assign_pseudo_regs - assign pseudo reg to a particular node in the RIG
 */
int assign_pseudo_reg(struct rig_node *rn)
{
  int reg_type, reg_num;
  struct setlist *sptr;
  struct rig_node *sib, *ig_ptr;

  if(rn->pseudo && rn->pseudo[0]!='\0')
    return TRUE;

  if(rn->can_replace){
    if (!((sptr = s_lookup(rn->reg_set)) &&
	  vin(rn->non_alloc,regtoindex(regtype(rn->reg_set),regnum(rn->reg_set))))){
      reg_type = regtype(rn->reg_set);
      reg_num = new_pseudoreg(reg_type);
      if(reg_num == -1)
	return FALSE;
      make_pseudo_reg(rn->pseudo, reg_type, reg_num);

      /*all siblings of this node should have the same pseudo reg*/
      for(ig_ptr=rn->nsib ; ig_ptr!=rn ; ig_ptr=ig_ptr->nsib)
	strcpy(ig_ptr->pseudo, rn->pseudo);

      return TRUE;
    }
  }
  return FALSE;
}


/*
 * get_pseudo_reg - get the pseudo register corresponding to the hardware
 *                  register assigned during creating the register
 *                  interferece graph
 */
void get_pseudo_reg(struct list *rtl, char *o_reg, char *new_reg, int is_use)
{
  char *cptr;
  static struct rig_node *rig_start=NULL;
  struct rig_node *rn;
  struct ruse_node *pruse;

  *new_reg = '\0';
  
  /*We work on the assumption that since there has been no block re-ordering
    between building the RIG and calling this function the nodes in the graph
    should be in the same order as the basic blocks. The function still works
    if this is not the case. But in any case the nodes should be clustered by
    basic block numbers*/
  if(rig_start)
    rn = rig_start;
  else
    rn = rig_top;
  
  if(rn->block->block_id != rtl->daddy->block_id){

    while(rn && rn->block->block_id != rtl->daddy->block_id)
      rn = rn->next;

    if(!rn){
      rn = rig_top;
      while(rn && rn->block->block_id!=rtl->daddy->block_id)
	rn = rn->next;
    }
  }

  if(rn){

    /*We have located the cluster of nodes with same block number as the
      RTL. Now we look for the node for this RTL*/
    while(rn && rn->block->block_id == rtl->daddy->block_id){

      if((strcmp(rn->reg_set, o_reg)==0) && rn->can_replace){

	if(is_use){
	  /*check if this RTL is present in the uses list for this node*/
	  for(pruse=rn->uses ; pruse ; pruse=pruse->next){
	    if(pruse->rtl == rtl){
	      if(assign_pseudo_reg(rn))
		strncpy(new_reg, rn->pseudo, REGSTRING);
	      break;
	    }
	  }
	}
	else{
	  /*check if RTL is present in the set list for this node*/
	  for(pruse=rn->sets ; pruse ; pruse=pruse->next){
	    if(pruse->rtl == rtl){
	      if(assign_pseudo_reg(rn))
		strncpy(new_reg, rn->pseudo, REGSTRING);
	      break;
	    }
	  }
	}
      }
      rn = rn->next;
    }

    if(new_reg && isRegister(new_reg)){
      cptr = new_reg;
      cptr = skiploc(new_reg);
      *cptr = '\0';
    }
  }

  /*if the register is not mapped we assum it is a hardware register and only
    remap it, so that two different registers, mapped to the same register,
    appear the same*/
  if(new_reg[0] == '\0')
    remap_hardware_reg(new_reg, o_reg, REGSTRING);
}


void remap_hardware_reg(char new_reg[], char hreg[], int reglength)
{
  int rnum;
  
  
  if(regtype(hreg) == RTREG){
    rnum = map(hreg);
    makereg(new_reg, RTREG, rnum);
  }
}

#endif  /* DEOPT */
#endif  /* INTERACT */

