/* Header file corresponding to the file de_optimize.c
 */

/*
 * $Author: jdh8d $
 * $Date: 2004/01/21 18:39:54 $
 * $Revision: 1.2 $
 */

#ifndef DE_OPTIMIZE_H
#define DE_OPTIMIZE_H

#ifdef INTERACT
#ifdef DEOPT

/* Turn on Debugging features*/
/*#define SH_DEBUG*/

/* DATA STRUCTURES */

/* Register Use List node*/

struct ruse_node
{
  struct ruse_node *next;     /* Next use of register in an RTL*/
  struct list *rtl;           /* RTL with a use of the register*/
};

/* Register Interference Graph node*/
struct rig_node 
{
    struct rig_node *prev,
                    *next,
                    *blkprev,   /* Previous node in block*/
                    *blkins,    /* Incoming node in block*/
                    *blkouts,   /* Outgoing node in block*/
                    *psib,      /* Previous related sibling (circ. list)*/
                    *nsib;      /* Next related sibling (circ. list)*/

  int can_replace,              /* Can we replace this register?*/
                  done;
  struct list *start,           /* Beginning of live range*/
                *finish;        /* End of live range*/
  struct bblock *block;         /* Basic block for live range*/
  char reg_set[REGSTRING],      /* Register being set for node*/
         pseudo[REGSTRING];     /* Pseudoregister assigned to node*/
  struct locuse *memloc;        /* New memory location for load/stores*/
  struct ruse_node *sets,       /* Places where reg_set is being set*/
                     *uses;     /* Places where reg_set is being used*/
  vector non_alloc;             /* Non-allocatable registers in this range*/
};


/*extern struct rig_node *rig_top,   
                         *rig_end;   
*/

/* FUNCTION PROTOTYPES */
int de_optimize_func();
int coalesce_fp_moves();
int deopt_loop_inv_cdmotion();
int has_calls(struct loopl *cur_loop);
int has_stores(struct loopl *cur_loop);
int invariant_rtl(struct list *rtl, rlist invareg);
void live_reg_calc(struct loopl *cur_loop, rlist *live_regs);
char *gtglob(char *s);
void rlist_to_vector(rlist rl, vector *vec);
int deopt_reg_alloc();
void build_rig(int merge_use_set_ranges);
void merge_rig_nodes();
void destroy_rig();
void display_rig();
void display_rig2();
void print_rig_node(struct rig_node *rn);
struct rig_node *new_rig_node();
void destroy_rig_node(struct rig_node *rn);
void rig_add_use(struct list *rtl, char *ruse);
struct rig_node * rig_add_set(struct list *rtl, char *rset, int merge_use_set_ranges);
rlist resize_rlist(rlist oldrl);
void replace_reg(struct list *rtl, char *old_reg, char *new_reg);

/* Bill's functions */
int reginvect(char *reg, vector regs);
void vshow(vector vlist);
void rminus(rlist *result, rlist vec);
int is_reg_affected(struct list *ptr, char *reg);
int reg_set_in_block(struct bblock *cblk, char *reg);
void regs_set(struct list *ptr, rlist *regs);
void regs_used(struct list *ptr, rlist *regs);
void regs_set_before_used(struct bblock *cblk, rlist *set_before_used);
void remap_hardware_reg(char new_reg[], char hreg[], int reglength);

#endif  /* DEOPT */
#endif  /* INTERACT */
#endif  /* DE_OPTIMIZE_H */
