/*****************************************************
Functions to encode a human readable RTL into encoded RTL
encode_rtl(src, dst): encode src into dst.
By cai, baosheng
*******************************************************/
#ifdef INTERACT
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "vpo.h"
#include "vpo2.h"
#include "undo.h"
#include "encode_rtl.h"

#include "regs.h"

#define ERR_RET(MSG) {send_error(MSG); return;}

struct glouse *glos_head, *glos_tail;


/*src is the human readable rtl, and dst is the encoded, machine readable rtl*/
void encode_rtl(char *src, char *dst)
{

   register char c;
   int offset;
   char src_no_space[MAXSTRING];
   char *srcptr;
   char *dstptr = dst;
   /*empty the destination string*/
   *dstptr = '\0';
   /*first, ignore all the spaces in the source rtl*/
   ignore_space(src, src_no_space);
   srcptr = src_no_space;
   
   while (*srcptr)
      /*if it begins with b, w, r, f, or d, then it is a register*/
      if (is_register(srcptr)){
         
         char reg_string[REGSTRING];
         int reg_type, reg_num;
         
         offset = get_reg(srcptr, &reg_type, &reg_num);
         makereg (reg_string, reg_type, reg_num);
         strcat (dst, reg_string);
         srcptr += offset;
         dstptr = skiploc(dstptr);
      }

   /* if it begins with a ., then it is a local variable. */
      else if (is_local(srcptr)) {
         char local_name[MAXSTRING];
         char *code;
         /*increase the srcptr to jump the "local("*/
         srcptr +=6;
         offset = get_name(srcptr, local_name);
         srcptr += offset;
         if (*srcptr != ')')
            ERR_RET("encode_rtl(): Syntax error in encode_rtl():local()\n");
         code = find_local(local_name);
         if (!code)
            ERR_RET("encode_rtl(): Can't find the local name\n");
	
         strcat (dst, code);
         srcptr ++;
         dstptr = skiploc(dstptr);
      }
      else if (is_global(srcptr)){
         char global_name[MAXSTRING];
         char *code;
         /*increase the srcptr to jump the "local("*/
         srcptr +=7;
         offset = get_name(srcptr, global_name);
         srcptr += offset;
         if (*srcptr != ')')
            ERR_RET("encode_rtl(): Syntax error in encode_rtl():global()\n");
         code = find_global(global_name);
         if (!code)
            ERR_RET("encode_rtl(): Can't find the global name\n");
         strcat (dst, code);
         srcptr ++;
         dstptr = skiploc(dstptr);         
      }
    else {

         *dstptr = *srcptr;                  
         *(dstptr+1) = '\0';
         srcptr++;
         dstptr ++;
      }
   *dstptr = '\0';
   
#ifdef DEBUG
   fprintf(stderr, "encode_rtl(): The encoded rtl is: %s\n", dst);
#endif
}

/*Ignore the spaces in the source human readable rtl*/
void ignore_space(char *src, char *dst)
{
   for (;*src; src++){
      if (*src == ' '|| *src == '\t')
         continue;
      *dst = *src;
      dst++;
   }
   *dst = '\0';
}


/*is the string represent a register?*/
int is_register(char *src)
{
   if ( (*src =='b' || *src =='w' || *src =='r'
          ||*src == 'f' || *src == 'd') && *(src+1) == '[')
      return 1;
   else
      return 0;
}

/*is src a local variable? */
int is_local(char *src)
{
   if (!strncmp(src, "local(", 6))
      return 1;
   else return 0;
}
/*is src a local variable? */
int is_global(char *src)
{
   if (!strncmp(src, "global(", 7))
      return 1;
   else return 0;
}


/*get the register type and number*/
int get_reg(char *srcptr, int *reg_type, int *reg_num)
{
   int i;
   char c;
   
   if ((*reg_type = get_type(*srcptr))==-1)
      ERR_RET("get_reg(): Register type error\n");

   if (sscanf(srcptr, "%c[%d]",&c, reg_num) != 2)
      ERR_RET("get_reg(): Register number error\n");
      
   for (i = 0; ; i++, srcptr++){
      if (*srcptr == ']')
         return i+1;
   }
}

/*maybe not correct... routine should go into machine-dependent part*/
int get_type (char c)
{
   switch (c){
   case 'r':
      return RTREG;
   case 'f':
      return FTREG;
   case 'd':
      return DTREG;
   default:
      return -1;
   }
}
/*get the name till separators.*/
int get_name(char *src, char *name)
{
   int i;
   for (i =0; ; i++){
      if (*(src+i)!= '.' && *(src +i) != '_' && !isalnum(*(src+i)))
         break;
      *(name+i) = *(src +i);
   }
   *(name +i) = '\0';
   return i;   
}


/*find the code of the local variable*/
char *find_local(char *local_name)
{
   extern struct locuse *locs;
   register struct locuse *ptr;
   for (ptr = locs; ptr; ptr = ptr->next) {
      if (!strcmp(ptr ->item, local_name))
         return ptr->code;
   }
   return NULL;
}
/*find the code of the global varialbe*/
char *find_global(char *global_name)
{
   register struct glouse *ptr;
   for (ptr = glos_head; ptr; ptr = ptr->next) {
      if (!strcmp(ptr->global_name_ptr, global_name))
         return ptr->code;
   }
   return NULL;
}

/*add a node to the glocs_head link list*/
void add_glouse_node(struct glouse *gloptr)
{
   if (gloptr==NULL){
      fprintf(stderr, "Null pointer: gloptr");
      exit(-1);
   }
   
   if (glos_head==NULL){
      /*If this is the first node, then initialize
         the node head, current, and tail*/
      glos_head = gloptr;
      glos_tail = gloptr;
   }
   else{
      /*If it is not the new node, then add this node to the tail*/
      glos_tail->next = gloptr;
      glos_tail= gloptr;
   }
}


#endif
