
/************************************************************************/
/*                                                                      */
/* Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   */
/*     The Rector and Visitors of the University of Virginia            */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from the University of Virginia.                                */
/* This code is distributed WITHOUT ANY WARRANTY.  No claims are        */
/*      made as to whether it serves any particular purpose or even     */
/*      works at all.                                                   */
/*                                                                      */
/************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/eod.c,v $
 $Date: 2004/01/21 18:39:55 $
 $Revision: 1.22 $
 $Name:  $

 Log of changes:

 $Log: eod.c,v $
 Revision 1.22  2004/01/21 18:39:55  jdh8d
 Latest Vista additions/changes.  Don't yell at me if it's broken/bad.. prasad wanted these changes.  I've reviewed them, and tested,  minimally.. things appear OK (for the batch compiler)

 Revision 1.21  2003/03/12 15:35:09  jdh8d
 *** empty log message ***

 Revision 1.20  2003/02/06 04:46:28  jg9h
 Merged VISTA into mainline vpo under the INTERACT preprocessor flag

 Revision 1.19  2002/07/11 15:22:31  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.18  2002/06/05 20:08:53  jdh8d
 Changes to support the stache, including routines to build a callgraph.

 Some cosmetic changes.

 Minor bug fixes.

 Changed build scripts to build the mips compiler for SimpleScalar if
 SSAS is set to "-DSS_AS"

 Revision 1.17  2002/05/16 19:39:41  jdh8d
 some changes from Mas

 Revision 1.16  2002/04/02 21:07:52  jwd
 Added some comments.

 Revision 1.15  2001/08/31 15:34:05  jdh8d
 *** empty log message ***

 Revision 1.14  2001/08/22 22:27:59  jdh8d
 *** empty log message ***

 Revision 1.13  2001/08/15 21:18:22  jdh8d
 *** empty log message ***

 Revision 1.12  2001/07/23 18:32:34  rv3s
 Replaced skiploc() with skiplocOr2() and copyloc() with copylocOr2 for processing register lists

 Revision 1.11  2001/05/23 20:28:20  jdh8d


 Re-worked all of the backend to support variable
 sized registers.  IT IS  NO LONGER SAFE TO ASSUME
 THAT REGISTERS ARE TWO BYTES.  Mips currently has
 4-byte registers, others will follow (or be destroyed!)

 Revision 1.10  2001/05/12 18:23:07  jdh8d
 changed EOD to prevent memory operations from trading places.  added
 side effects to inst.c, and added the mips version of maybe_same

 Revision 1.9  2001/04/04 19:41:39  jdh8d

 fixed bugs and lots of them.  cse was removing a "redundant" store
 because it didn't recognize an aliased load.  Code motion was calculating
 the offset wrong if the update of the induction variable happened
 in the same RTL as the load/store using that var, etc.

 Revision 1.8  2001/03/07 15:50:16  jdh8d


 eod fix to observe t* lines

 Revision 1.7  2000/06/09 19:36:59  jwd
 Changed calls to VET to indicate which phases are mandatory and
 which are optional.
 Revision 1.6  1999/03/05 16:59:39  jwd
 Missing a declaration of moveptr needed for some VET code in
 function movedown().
 Revision 1.5  1999/01/06 16:46:09  jwd
 Added VET support.
 Revision 1.4  1998/04/28 20:17:57  jwd
 Included some necessary header files.
 Revision 1.3  1998/04/03 18:19:22  jwd
 First ANSI C version.

*************************************************************************/

/*------HISTORY OF CHANGES---------------
  1/12/98 Modified to use ANSI compatible function protyping.
  ----------------------------------------*/


#include <stdio.h>
#include <string.h>

#include "vpo.h"
#include "vpo2.h"
#include "assign.h"
#include "links.h"
#include "misc.h"
#include "eod.h"
#include "vectors.h"

#include "rtl.h"  /*Machine Dependent */

#ifdef VET
#include "view.h"
#endif

/*
 * This file contains functions that perform evaluation order determination
 * and targeting on a procedure.
 */

static struct list *ccs[128];
static struct list *reserveA[MAXRTS];


/*
 * evaluation_order_determination - perform evaluation order determination
 */
void evaluation_order_determination(int invocation) {

   struct bblock *cblk;
   struct list *ptr, *bptr;
   int i, j, x, cost, use, carry, costs[MAXLINKS];

#ifdef VET
#ifdef INTERACT
  begin_phase(EVAL_ORDER_DETER_TRANS);	
#else
    begin_phase(EVALORDERDET_PHASE);
#endif
#endif

   /*
      Check to see if the user would like to see the state of the function at
      this point in time.
   */
   if (swd &&
       debug_opt_i("dump_before_evaluation_order_determination", invocation))
      dump_function_i("before evaluation_order_determination()", invocation,
		      SHOW_NONRTLS | SHOW_SE | SHOW_COST);

   /*
      Visit each basic block in the current function
   */
   for (cblk = top; cblk; cblk = cblk->down) {
		
#ifdef INTERACT
      if (swq && limited_scope && !cblk->inscope)
         continue;
#endif
		
      /*
         Clear the reserve pointer and the condition code set array to ensure
         that no combiner links are set across basic blocks.
      */
      reserveA[0] = ccs[0] = (struct list *) NULL;

      /*
         Perform a preliminary pass through the basic block and set the "cost"
         field in each RTL to reflect the number of registers required to
         calculate each expression.
      */
#ifdef INTERACT
      begin_trans(VET_OPT_TRANS);
#endif      
      for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
	 if (cnts(ptr)) {
	    /* cost is the number of registers used by the RTL */
            ptr->cost = rusage(ptr);

	    /*
	       Look at the children of this RTL (if any) and determine the
	       number of registers required to calculate this whole expression.
	       First we create an array containing the costs of all of the
	       children sorted from the highest to the lowest.  Then we examine
	       this array to determine how many registers are actually required
	       to calculate the complete expresssion.
	    */
	    if (ptr->lnks[0]) {
	       for (i = 0; i < MAXLINKS && (bptr = ptr->lnks[i]); i++) {
		  for (x = 0; x < i; x++)
		     if (bptr->cost > costs[x])
			break;
		  if (x >= i)
		     costs[i] = bptr->cost;
		  else {
		     for (j = i; j > x; j--)
			costs[j] = costs[j - 1];
		     costs[x] = bptr->cost;
		  }
	       }
	       carry = 1;
	       use = costs[0] - 1;
	       for (j = 1; j < i; j++) {
		  if (use >= costs[j]) {
		     carry++;
		     use--;
		  }
		  else
		     carry++;
	       }
	       cost = carry + use;
	       if (cost > ptr->cost)
		  ptr->cost = cost;
	    }
	 }
      }
#ifdef INTERACT
      end_trans();
#endif      

      if (swd && cblk->label &&
	  debug_opt_i_l("dump_evaluation_order_determination",
			invocation, cblk->label))
	 dump_function_i_l("evaluation_order_determination()", invocation, cblk->label,
			   SHOW_NONRTLS | SHOW_SE | SHOW_COST);

      /*
         Now pass through the basic block from the bottom to the top and
         re-order RTLs where possible to calculate expressions that require the
         most registers first.  Here, we only consider RTLs that have at least
         two combiner links as a starting point for re-arranging the code.
      */
      /*#ifdef VET
      begin_trans(VET_OPT_TRANS);
      #endif*/
      for (ptr = cblk->rtlend; ptr; ptr = ptr->prev)
	 if (cnts(ptr) && ptr->lnks[0] && ptr->lnks[1])
	 {
		struct list* temp;
	    	temp = reorder(ptr);

		/* we may be stuck in a loop due to forward links, remove them
			and try again */
		if(temp==ptr)
			remv_forward_links_in_block(cblk);	
		else
			ptr=temp;
	 }
      /*#ifdef VET
      end_trans();
      #endif*/
   }

   /*
      Check to see if the user would like to see the state of the function at
      this point in time.
   */
   if (swd &&
       debug_opt_i("dump_after_evaluation_order_determination", invocation))
      dump_function_i("after evaluation_order_determination()", invocation,
		      SHOW_NONRTLS | SHOW_SE | SHOW_COST);
#ifdef VET
#ifdef INTERACT
   end_phase(EVAL_ORDER_DETER_TRANS);	
#else
    end_phase(EVALORDERDET_PHASE);
#endif
#endif
}

/*
 * rusage - determine the number of registers used by an RTL
 */
int rusage(struct list * ptr) {

   register int i;
   register char *rtl;
   struct list *bptr;
   int cost, number, j, ignore;
   char a, *b, c, *counted[128], reg[MAXSTRING];

   cost = number = 0;

   /*
      First, look at the registers that are used by the RTL itself.  We look at
      only pseudo-registers when considering the register usage because
      counting in hardware registers would not give results that are accurate
      in cases were reserved registers (like a frame pointer or a stack
      pointer) are used in expressions.  We do look at hardware registers to
      set up combiner links between function calls and the next use of the
      register that contains the return value.
   */
   reget(ptr->deads, reg);
   rtl = ptr->val;
   while (rtl = gtrg(rtl)) {
      if (pseudoreg(unskiploc(rtl))) {
         ignore = FALSE;
         if (*rtl == '=') {
            for (b = reg; *b; b=skiplocOr2(b) + 1) {
               if (isRegister(b) && (regtype(b) == regtype(unskiploc(rtl)))) {
                  ignore = TRUE;
                  do
                  {
                     int i;
                     if(*skiploc(b))
                     {
                        copylocOr2(b,skiplocOr2(b)+1);
                        *skiploc(b)=*skiploc(skiplocOr2(b)+1);
                        b = skiplocOr2(b)+1;
                     }
                     else
                     {
                        for(i=0;i<=skiploc(b)-b;i++)
                        {
                           b[i]=0; /* null out the last one */
                        }
                     }

                        /*
                     *b = *(b + 3);
                     *(b + 1) = *(b + 4);
                     *(b + 2) = *(b + 5);
                        */
                  } while (*b);
                  break;
               }
            }
         }
         if (!ignore) {
            for (i = 0; i < number; i++) {
                /*(*counted[i] == *(rtl - 2) && *(counted[i] + 1)==*(rtl - 1))*/
               if (cmploc(unskiploc(rtl), counted[i])==0)
                  break;
            }
            if (i == number) {
               counted[number++] = unskiploc(rtl);
               cost += allocregf(regtype(unskiploc(rtl)));
            }
         }
         if (*reserve && *rtl != '=') {
            for (i = 0; i < MAXRTS && (bptr = reserveA[i]); i++) {
               /*if(*bptr->val == *(rtl - 2) && *(bptr->val + 1) ==*(rtl - 1))*/
               if  (cmploc(bptr->val, unskiploc(rtl))==0){
                  while (!cnts(bptr))
                     bptr = bptr->prev;
                  if (bptr->blnks[0])
                     bptr = bptr->blnks[0];
		if(ptr!=bptr)
                  (void) add_link(ptr, bptr);
                  while (i < MAXRTS - 1) {
                     reserveA[i] = reserveA[i + 1];
                     i++;
                  }
                  reserveA[i] = (struct list *) NULL;
                  break;
               }
            }
         }
      }
   }

   /*
      Check to see if any of the condition code items that have outstanding
      sets are used in this RTL.  If so, set an appropriate combiner link for
      them.
   */
   for (i = 0; bptr = ccs[i]; i++) {
      rtl = ptr->val;
      a = *bptr->val;
      while (c = *rtl++) {
	 if (c == a && *rtl == *(bptr->val + 1) && *(rtl + 1) != '=') {
	    (void) add_link(ptr, bptr);
	    for (j = i; ccs[j] = ccs[j + 1]; j++);
	    i--;
	    break;
	 }
      }
   }

   /*
      If this RTL sets a condition code item, then add it to the array of
      condition code items set.
   */
   if (set_cc(ptr->val)) {
      for (i = 0; bptr = ccs[i]; i++) {
	 if (*bptr->val == *ptr->val && *(bptr->val + 1) == *(ptr->val + 1)) {
	    ccs[i] = ptr;
	    break;
	 }
      }
      if (!bptr) {
	 ccs[i++] = ptr;
	 ccs[i] = (struct list *) NULL;
      }
   }

   /*
      Now look to see if any use lines follow the RTL.  These indicate that the
      RTL may use more registers than is apparent from the RTL itself. Also,
      remember the last reserve line that we have encountered.
   */
   for (ptr = ptr->next; ptr; ptr = ptr->next) {
      if (cnts(ptr))
	 break;
      if ((ptr->type & USELINE) && ptr->val) {
	 for (i = 0; i < number; i++)
	    if (cmploc(counted[i], ptr->val)==0)
	       break;
	 if (i == number) {
	    counted[number++] = ptr->val;
	    cost += allocregf(regtype(ptr->val));
	 }
      }
      else if ((ptr->type & RESLINE) && ptr->val) {
	 for (i = 0; i < MAXRTS; i++) {
	    if (!reserveA[i]) {
	       reserveA[i] = ptr;
	       if (i < MAXRTS - 1)
		  reserveA[i + 1] = (struct list *) NULL;
	       break;
	    }
	 }
      }
   }

   return (cost);
}

/*
 * reorder - reorder subexpressions to minimize the number of register needed
 */
struct list *reorder(struct list * ptr) {

   struct list *bptr, *rptr;
   int i, j, low, lownum, lowrtlnum, sum;

   /*
      Look at the combiner links for an RTL and move those items that use the
      fewest registers as close as possible.  This ensures that subexpressions
      that need the most registers to compute are done first.
   */
   rptr = ptr;
   for (i = 0; i < MAXLINKS && (bptr = ptr->lnks[i]); i++) {
      low = i;
      lownum = bptr->cost;
      lowrtlnum = RTL_sum(bptr);
      for (j = i + 1; j < MAXLINKS && (bptr = ptr->lnks[j]); j++) {
	 if (bptr->cost < lownum) {
	    lownum = bptr->cost;
	    lowrtlnum = RTL_sum(bptr);
	    low = j;
	 }
	 /*fix added so that the order of links with the same cost does not
	   affect the final output*/
	 else if(bptr->cost == lownum && ((sum=RTL_sum(bptr)) < lowrtlnum)){
	   lowrtlnum = sum;
	   low = j;
	 }
      }
      /* bptr points to the RTL that has the lowest cost */
      bptr = ptr->lnks[low];
      if (!bptr->val || !is_call(bptr->val))
	{
	 rptr = movedown(rptr, bptr);
	}
      /* reorder the links so that the cheapest is first and so it is not processed again */
      ptr->lnks[low] = ptr->lnks[i];
      ptr->lnks[i] = bptr;
   }

#ifdef VET
    mod_links(ptr);
#endif

   return (rptr);
}

/*
 * movedown - move an RTL and its associated items closer to another RTL
 */
struct list *movedown(struct list * rptr, struct list * ptr) {

   register char c, *rtl;
   struct list *eptr, *lptr;
   int setsloc;
   char *a, locref[MAXSTRING];
#ifdef VET
    struct list *moveptr;
#endif
   /*
      Group all of the items that belong to this RTL together.  Note that we
      cannot assume that the RTL is followed by another one because even though
      we have not explicitly set up global combiner links, it is sometimes
      possible for the mdlinks() function which is called by the combiner to
      have set up a link across a basic block.
   */
   for (eptr = ptr->next; eptr; eptr = eptr->next) {
      if (cnts(eptr))
	 break;
      if((eptr->type&USELINE) && eptr->val)
        {
                /* ensure we don't move an item past a "use" line because it will
                        mess up the local register assigner */
                for(rtl=aliases(eptr->val); *rtl; rtl=skiploc(rtl)+1)
                {
                        if(ptr->val && fmatch(ptr->val, rtl))
                                return ptr;
                        if(ptr->se && fmatch(ptr->se, rtl))
                                return ptr;
                }
        }
   }
   if (!eptr)
      return (ptr);
   if (eptr == rptr)
      return (reorder(ptr));

   /*
      Determine if the RTL that we are trying to move sets a local variable. If
      so, remember the reference so that we can ensure that we do not move it
      past another RTL which may use the local variable value.
   */
   setsloc = FALSE;
   if (ismem(rtl = ptr->val)) {
      for (a = locref; (c = *rtl) != '='; *a++ = c)
      {
	 if (islocal(rtl))
	    setsloc = TRUE;
	 rtl=skiploc(rtl);
      }
      *a = '\0';
   }

   /*
      Ensure that we do not try to move a register's use past a point that
      reserves or sets the register and that we don't move a trashed item past
      a point that uses the item trashed.
   */
   for (lptr = eptr; lptr && lptr != rptr; lptr = lptr->next) {
      if (lptr->type & RESLINE) {
         for (rtl = aliases(skiploc(lptr->val) + 1); *rtl; rtl = skiploc(rtl)+1)
            if (fmatch(ptr->val, rtl))
               return (ptr);
      }

      if (lptr->type & SETLINE) {
         for (rtl = lptr->val; *rtl; rtl = skiploc(rtl)) 
	 {
	    char* a;
            c = *(skiploc(rtl));
            *(skiploc(rtl)) = '\0';
	    for(a=aliases(rtl);*a;a=skiploc(a)+1)
            	if (fmatch(ptr->val, a)) 
 	    	{
               		*(skiploc(rtl)) = c;
               		return (ptr);
            	}
            *(skiploc(rtl)) = c;
         }
         if (rtl = is_linked_to_reserve(ptr)) {
            for (a = lptr->val; *a; a=skiploc(a)){
               /*if (*a == *rtl && *(a + 1) == *(rtl + 1))*/
               if (cmploc(a,rtl)==0)
                  return (ptr);
            }
         }
      }
      if((lptr->type&USELINE) && lptr->val)
	{
		/* ensure we don't move an item past a "use" line because it will
			mess up the local register assigner */
		if(lptr->val)
		for(rtl=aliases(lptr->val); *rtl; rtl=skiploc(rtl)+1)
		{
			if(ptr->val && fmatch(ptr->val, rtl))
				return ptr;
			if(ptr->se && fmatch(ptr->se, rtl))
				return ptr;
		}
	}
      if (rtl = cnts(lptr))
      {
	char defined_reg[REGSTRING];
	char my_new[MAXSTRING];
	char* b=ptr->val;	

	if(gtstackrg(ptr->val) && gtstackrg(lptr->val))
		return ptr;
#ifdef FIFOS
	if(gtfrg(ptr->val) && gtfrg(lptr->val))
		return ptr;
#endif

	while(b=gtrg(b))
	{
		/* look for defined registers in lptr */
		if(*b!='=')	
			continue;
		if(!pseudoreg(unskiploc(b)))
		{
	 		for (a = aliases(unskiploc(b)); *a; a = skiploc(a)+1)
 				if (fmatch(rtl, a)) 
	       				return (ptr);
		}
		else
		{
			copyloc(defined_reg, unskiploc(b));
			*skiploc(defined_reg)=0;
 			if (fmatch(rtl, defined_reg)) 
       				return (ptr);
		}
	}	

	b=ptr->se;
        while(b && (b=gtrg(b)))
        {
                /* look for defined registers in lptr */
                if(*b!='=')
                        continue;
		if(!pseudoreg(unskiploc(b)))
                {
	 		for (a = aliases(unskiploc(b)); *a; a = skiploc(a)+1)
       	                 if (fmatch(rtl, a))
       	                         return (ptr);
		}
		else
		{
			copyloc(defined_reg, unskiploc(b));
			*skiploc(defined_reg)=0;
       	                 if (fmatch(rtl, defined_reg))
       	                         return (ptr);
		}
        }

	b=lptr->val;
	while(b=gtrg(b))
	{
		/* look for defined registers in lptr */
		if(*b!='=')	
			continue;
		if(!pseudoreg(unskiploc(b)))
		{
	 		for (a = aliases(unskiploc(b)); *a; a = skiploc(a)+1)
                        	if (fmatch(ptr->val, a))
                                	return (ptr);
		}
		else
		{
			copyloc(defined_reg, unskiploc(b));
			*skiploc(defined_reg)=0;
       	                 if (fmatch(ptr->val, defined_reg))
       	                         return (ptr);
		}
	}	

	b=lptr->se;
	while(b && (b=gtrg(b)))
	{
		/* look for defined registers in lptr */
		if(*b!='=')	
			continue;
		if(!pseudoreg(unskiploc(b)))
                {
	 		for (a = aliases(unskiploc(b)); *a; a = skiploc(a)+1)
                                if (fmatch(ptr->val, a))
                                        return (ptr);
                }
                else
                {
			copyloc(defined_reg, unskiploc(b));
			*skiploc(defined_reg)=0;
                         if (fmatch(ptr->val, defined_reg))
                                 return (ptr);
                }
	}	
	if (setsloc && fmatch(lptr->val, locref))
	    return (rptr);

	if(reads_mem(lptr->val) && writes_mem(ptr->val))
		return ptr;
	if(writes_mem(lptr->val) && writes_mem(ptr->val))
		return ptr;
	if(writes_mem(lptr->val) && reads_mem(ptr->val))
		return ptr;
      }

      if (lptr->type & TRASHLINE && strcmp(lptr->val, "*")) {
	 if (fmatch(ptr->val, lptr->val))
	    return (ptr);
      }

      /*
       * if we encounter a trash all line, return that this
       * is as far as we can go 
       */
      if(lptr->type & TRASHLINE && strcmp(lptr->val, "*")==0) 
      {
	/* we can't move memory ops past a t* */
	if(gtmem(ptr->val))
		return ptr;
      }
   }

   /*
      If we did not encounter the target RTL, then there must be a link back or
      across basic blocks, therefore we just return.
   */
   if (!lptr)
      return (ptr);

   /*
      Now remove the desired RTL and associated items from their previous
      position and place them in their new position.  Note that nothing is done
      to prevent items from moving across function calls, since we are certain
      that the combiner links that we are working with are limited to a single
      statement and most languages permit the compiler to arrange items within
      an expression regardless of possible side effects.
   */
#ifdef VET
    /*
     * Need to move all associated RTLs with ptr such as USELINEs.
     * Move RTLs from ptr to lptr before rptr.
     */
#ifdef INTERACT
	if (swq)
#endif
#ifdef VET
	begin_trans(VET_OPT_TRANS);
#endif
    for (moveptr = ptr; moveptr != eptr; moveptr = moveptr->next) {
	move_rtl(moveptr, moveptr->daddy, rptr, rptr->daddy);
    }
#endif

   lptr = eptr->prev;
   if (eptr->prev = ptr->prev)
      ptr->prev->next = eptr;
   else
      ptr->daddy->rtls = eptr;
   if (ptr->prev = rptr->prev)
      rptr->prev->next = ptr;
   else
      rptr->daddy->rtls = ptr;
   rptr->prev = lptr;
   lptr->next = rptr;
#ifdef VET
	end_trans();
#endif
   return (reorder(ptr));
}

/*
 * is_linked_to_reserve - check to see if an RTL is linked to a reserve line
 */
char *is_linked_to_reserve(struct list * ptr) {

   struct list *bptr;
   int i;

   for (i = 0; i < MAXLINKS && (bptr = ptr->lnks[i]); i++) {
      if (is_call(bptr->val)) {
	 while (bptr = bptr->next) {
	    if (cnts(bptr))
	       break;
	    if (bptr->type & RESLINE)
	       return (skiploc(bptr->val) + 1);
	 }
      }
   }
   return ((char *) NULL);
}
