#ifdef STACHE


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "vpo.h"
#include "callgraph.h"
#include "lig.h"
#include "misc.h"
#include "stache.h"

/* here's the start of the list */
static struct LIG_node* lig=0;

static add_LIG_interference (struct LIG_node* node1, 
	struct LIG_node* node2);
static LIG_walk_CG(struct call_graph_node *start, 
		struct call_graph_node *current);
void dump_LIG();
void color_LIG();

#define MAX(a,b) ((a)>(b) ? (a) : (b))
#define NODESIZE(lig_node) ((lig_node)->precolored ?  \
			(lig_node)->precolor_size :   \
			(lig_node)->the_local->size)

void build_LIG()
{

	/* 
		pseudo_code

		for each node, N,  in the call graph

			if(N is recursive)
				continue;

			for each local, L,  in N
			{
				create a new LIG node for L
				and add it to the LIG_node list (lig)
				(make sure L has a pointer to the
					new LIG node)
			}

		for each node, N,  in the call graph
			for each node, M,  in the walk of the call graph 
			start from N 
				for each local, L1, in N
					for each local, L2, in M
						add_interference(L1, L2);
				
			
	*/	


        struct call_graph_node *cgn;
	struct call_graph_local_info* cgli;
	/* for each node, N,  in the call graph */
        for(cgn=the_call_graph; cgn; cgn=cgn->next_func)
	{
		
		if(cgn->is_indirectly_recursive)
			continue;

		/*	for each local, L,  in N */
        	for(cgli=cgn->first_local; cgli; cgli=cgli->next)
		{
			struct LIG_node* new_LIG_node= (struct LIG_node*)
				CALLOC_PERM(sizeof(struct LIG_node));
			bzero(new_LIG_node, sizeof(struct LIG_node));

			/* link the node to the local and the local
				to the node */
			new_LIG_node->the_local=cgli;
			cgli->the_LIG_node=new_LIG_node;

			/* link it into the list of LIG_node's */
			new_LIG_node->next=lig;
			lig=new_LIG_node;
		}
	}


	/* for each node, N,  in the call graph */
        for(cgn=the_call_graph; cgn; cgn=cgn->next_func)
	{
		/*for each node, M,  in the walk of 
			the call graph start from N  */

		clear_seen();
		LIG_walk_CG(cgn,cgn);
	
	}
	
}

static int does_interfere (struct LIG_node* node1, struct LIG_node* node2)
{
	struct LIG_edge* lig_edge;

	for(lig_edge=node1->first_edge; lig_edge; lig_edge=lig_edge->next)
	{
		if(lig_edge->node==node2)
			return TRUE;
	}
	return FALSE;
}

static add_LIG_interference (struct LIG_node* node1, struct LIG_node* node2)
{
	struct LIG_edge* new_edge;
	if(node1==0 || node2==0 || node1==node2)
		return;
	if(does_interfere(node1,node2))
	{
		assert(does_interfere(node2,node1));
		return;
	}

	/* add a new edge from node1 to node 2 */
	new_edge=(struct LIG_edge*)CALLOC_PERM(sizeof(struct LIG_edge));
	bzero(new_edge, sizeof(struct LIG_edge));
	new_edge->node=node2;
	new_edge->next=node1->first_edge;
	node1->first_edge=new_edge;

	/* add a new edge from node2 to node 1 */
	new_edge=(struct LIG_edge*)CALLOC_PERM(sizeof(struct LIG_edge));
	bzero(new_edge, sizeof(struct LIG_edge));
	new_edge->node=node1;
	new_edge->next=node2->first_edge;
	node2->first_edge=new_edge;

}

static LIG_walk_CG(struct call_graph_node *start, 
		struct call_graph_node *current)
{
        struct call_graph_list_node *cgln;
	struct call_graph_local_info* loc_info, *loc_info2;

	if(current->seen || current->is_indirectly_recursive || 
		!start->has_data)
		return;	

	current->seen=TRUE;

        for(cgln=current->next_child; cgln; cgln=cgln->next_list_node)
        {
                LIG_walk_CG(start, cgln->the_func);
        }

	if(current->has_data)
	{
		/* for each local, L1, in N */
        	for(loc_info=start->first_local; loc_info; 	
				loc_info=loc_info->next)
		{
			/*for each local, L2, in M*/
        		for(loc_info2=current->first_local; loc_info2; 	
				loc_info2=loc_info2->next)
			{
	
		    	add_LIG_interference(loc_info->the_LIG_node, 
				loc_info2->the_LIG_node);
			}
		}
	}
	else 
	{
		int newNodeUsage=stache_size;
                struct LIG_node* new_LIG_node= (struct LIG_node*)
                     CALLOC_PERM(sizeof(struct LIG_node));

		if(current->library_func)
			newNodeUsage=current->total_stache_usage;

                bzero(new_LIG_node, sizeof(struct LIG_node));

                /* link the node to the local and the local to the node */
		/* but we don't haev a local, so we'll set to 0 */
                new_LIG_node->the_local=0;

                /* link it into the list of LIG_node's */
                new_LIG_node->next=lig;
                lig=new_LIG_node;

		lig->precolored=1;
		if(current->name)
			lig->precolor_name=(char*)strdup(current->name);
		else
			lig->precolor_name=0;


		if(newNodeUsage!=0)
			fprintf(stderr, "Adding precolor 0-%d cuz "
				"%s calls (possibly indirectly) %s\n", 	
				newNodeUsage, start->name, current->name);

		lig->start_location=0;
		lig->precolor_size=newNodeUsage;
        	for(loc_info=start->first_local; loc_info; 
		    loc_info=loc_info->next)
		{
                        add_LIG_interference(loc_info->the_LIG_node, lig);
		}
	}
}

void dump_LIG()
{
	struct LIG_node* lig_node;
	struct LIG_edge* lig_edge;
	int count=0;

	fprintf(stderr, "Dumping lig...\n");

	for(lig_node=lig; lig_node; lig_node=lig_node->next)
	{
		if(lig_node->precolored)
		{
		 fprintf(stderr,"\nvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv\n");
		 if(lig_node->precolor_name)
			 fprintf(stderr,"precolored %s: %d",
				lig_node->precolor_name,	
				lig_node->precolor_size);
		 else
		   	fprintf(stderr,"precolored node: %d",
				lig_node->precolor_size);
		 fprintf(stderr,"\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n");
		 continue;	
		}
		fprintf(stderr, "\nvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv\n");
		assert(lig_node && lig_node->the_local && 
			lig_node->the_local->the_LIG_node==lig_node &&
			lig_node->the_local->function_cgn);
		count=0;
		fprintf(stderr,"local %s ", lig_node->the_local->item);
		if(lig_node->stache_alloced)
		{
			fprintf(stderr, "[stache: %d] ", 
				lig_node->start_location);
		}
		fprintf(stderr, "interferes with:\n");
		for(lig_edge=lig_node->first_edge; lig_edge;
			lig_edge=lig_edge->next)
		{
			if(count++==3)
			{
				fprintf(stderr, "\n");
				count=0;
			}
			if(lig_edge->node->precolored)
			{
				fprintf(stderr,"precolored 0-%d\t", 
					NODESIZE(lig_edge->node));
			}
			else
			{
				fprintf(stderr,"(%s)%s\t", 
					lig_edge->node->the_local->
						function_cgn->name,
					lig_edge->node->the_local->item);
			}
		}
		fprintf(stderr, "\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n");
	}
	fprintf(stderr, "done dumping lig\n");
}


static void setup_neighbor_weights()
{
	struct LIG_node* lig_node;
	struct LIG_edge* lig_edge;
	for(lig_node=lig; lig_node; lig_node=lig_node->next)
	{
	  for(lig_edge=lig_node->first_edge; lig_edge; lig_edge=lig_edge->next)
	  {
		int size=NODESIZE(lig_edge->node);
		lig_node->neighbors_size+=size;
	  }
	}

}

static struct LIG_node* last_removed=0;

static void remove_LIG_node(struct LIG_node* lig_node)
{
	struct LIG_edge* lig_edge;
	/* remove our cost from the neighbors */
	for(lig_edge=lig_node->first_edge; lig_edge; lig_edge=lig_edge->next)
	{
		int size=NODESIZE(lig_node);
		lig_edge->node->neighbors_size-=size;
	}

	/* and add ourselves to the removed nodes list */
	lig_node->removed=1;
	lig_node->prev_removed=last_removed;
	last_removed=lig_node;

}

float LIG_node_benefit(struct LIG_node* lig_node)
{
	int reps=0;
	int size=0;
	int neighbors=0;

	/* best to just get the precoloreds out of the way */
	if(lig_node->precolored)
		return 100000;
	reps=lig_node->the_local->static_reps;
	size=lig_node->the_local->size;
	neighbors=lig_node->neighbors_size;

	return ((float)reps*100)/(size+neighbors);

}


static void remove_LIG_nodes()
{
	struct LIG_node* lig_node;
	struct LIG_edge* lig_edge;
	int done=FALSE;
	int removing=TRUE;
	if(!lig)
		return;
	while(!done)
	{
		/* we might be done now */

		done=TRUE;
		/* if we didn't remove anything last time through the list */
		/* we must be at a spot where there is no room left, */
		/* pick the least beneficial one and spill it */
		if(!removing)
		{
			float min_benefit=-1;
			struct LIG_node *min_benefit_node=0;
			for(lig_node=lig; lig_node; lig_node=lig_node->next)
			{
				if(!lig_node->removed && !lig_node->precolored)
				{
					float this_benefit=LIG_node_benefit(
						lig_node);
					if(this_benefit>min_benefit || 
						this_benefit==-1)
					{
						min_benefit_node=lig_node;
						min_benefit=this_benefit;
					}
						
				}
			}
			remove_LIG_node(min_benefit_node);
			min_benefit_node->should_color=0;

			/* hopefully we've broken a tie, now continue */
			/* trying to remove nodes */
		}
		removing=FALSE;
		for(lig_node=lig; lig_node; lig_node=lig_node->next)
		{
			if(lig_node->removed || lig_node->precolored)
			{
				/* don't remove precolored nodes */
				/* or nodes that've already been removed*/
				continue;
			}	
			else if(lig_node->the_local->size + 
				lig_node->neighbors_size < stache_size)
			{
				removing=TRUE;
				remove_LIG_node(lig_node);
				lig_node->should_color=1;
			}
			if(!lig_node->removed)
			/* oops, n/m.. not done yet */
				done=FALSE;
		}
	}
}

static check_for_coloring(struct LIG_node* lig_node)
{

	#define MAX_INTERFERES 10000
	struct LIG_node *node_array[MAX_INTERFERES];
	struct LIG_edge* lig_edge;
	int edge_count=0,i,j;
	int potential_start=0;
	int current_node=0;

	int new_size=lig_node->the_local->size;

	/* move relevent edges into an array */
	for(lig_edge=lig_node->first_edge; lig_edge; lig_edge=lig_edge->next)
	{
		if(lig_edge->node->removed || 
			(!lig_edge->node->stache_alloced && 
			 !lig_edge->node->precolored))
			continue;
		node_array[edge_count++]=lig_edge->node;
	}

	assert(edge_count<MAX_INTERFERES);

	/* sort the list based on stache_start */
	for(i=0;i<edge_count;i++)
	{
		for(j=i;j<edge_count;j++)
		{
			if(node_array[i]->start_location >
			   node_array[j]->start_location)
			{
				struct LIG_node* temp=node_array[i];
				node_array[i]=node_array[j];
				node_array[j]=temp;
			}
		}
	}

	while(potential_start<stache_size && current_node<edge_count)
	{
		if(node_array[current_node]->start_location
			- potential_start < new_size)
		{
			
			int size=NODESIZE(node_array[current_node])+1;
			
			potential_start=MAX(
			     (node_array[current_node]->start_location+ size), 
			      potential_start);
			potential_start+=8-(potential_start%8);
			assert(potential_start%8==0);
		}
		current_node++;
	}
	if(potential_start+new_size< stache_size)
	{
		lig_node->stache_alloced=1;
		lig_node->start_location=potential_start;
		return TRUE;
	}
	return FALSE;
	
}

static void assign_LIG_nodes()
{
	struct LIG_node* lig_node;
 	for(lig_node=last_removed; lig_node; lig_node=lig_node->prev_removed)
	{
		int did_color;
		if(lig_node->precolored)
			continue;
		did_color=check_for_coloring(lig_node);		
		if((did_color && !lig_node->should_color) || 
			(!did_color && lig_node->should_color))
		{
			if(lig_node->should_color)
				fprintf(stderr, "Couldn't color item.  boo!\n");
			else
				fprintf(stderr, "Colored extra item. yay!\n");
		}

		/* add it back to the graph */
		lig_node->removed=0;
	}
}

void color_LIG()
{
	setup_neighbor_weights();
	remove_LIG_nodes();
	assign_LIG_nodes();

	if(debug_opt("dump_lig"))
		dump_LIG();
}









#endif

