
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "vpo.h"
#include "assign.h"
#include "cdmotion.h"
#include "cse.h"
#include "combine.h"
#include "controlflow.h"
#include "database.h"
#include "dataflow.h"
#include "eod.h"
#include "locals.h"
#include "links.h"
#include "misc.h"
#include "vectors.h"
#include "cross.h"
#include "live.h"
#include "vpo2.h"
#include "regs2.h"    /*Machine Dependent*/
#include "rtl.h"      /*Machine Dependent*/
#include "y.tab2.h"  /*Tools Dependent*/
#ifdef VET
#include "view.h"
#endif

#define add_use_to_live_range(rtl,rtlist,blocklist) add_to_live_range(rtl,rtlist,blocklist,TRUE)
#define add_set_to_live_range(rtl,rtlist,blocklist) add_to_live_range(rtl,rtlist,blocklist,FALSE)

void add_to_live_range
	(
		/* the rtl to add to the list */

		struct list* rtl, 
		/* an output parameter that is the list of rtls in this regs lifetime */
		struct rtl_list** rtlist, 

		/* the list of basic blocks that the register is live in */
		struct blist** blocklist,

		/* is this a use or set of the var */
		int use
	)
{
	struct rtl_list* cur_node;
	struct blist* cur_block;

	for(cur_node=*rtlist; cur_node; cur_node=cur_node->next)	
	{
		if(cur_node->rtl==rtl)
		{
			if(use)	cur_node->used=TRUE; 
			else	cur_node->set=TRUE;
			return;
		}
	}

	cur_node=(struct rtl_list*)CALLOC_PHASE(sizeof(struct rtl_list));
	cur_node->rtl=rtl;
	cur_node->next=*rtlist;
	*rtlist=cur_node;
	cur_node->used=cur_node->set=FALSE;
	if(use)	cur_node->used=TRUE; 
	else	cur_node->set=TRUE;

	
	for(cur_block=*blocklist; cur_block; cur_block=cur_block->next)	
	{
		if(cur_block->ptr==rtl->daddy)
			return;
	}
	
	cur_block=(struct blist*)CALLOC_PHASE(sizeof(struct blist));
	cur_block->ptr=rtl->daddy;
	cur_block->next=*blocklist;
	*blocklist=cur_block;

}



/* mark all blocks in the func as done */
void clear_done_bits()
{
	struct bblock *cblk;
	for(cblk=top;cblk;cblk=cblk->down)
	{
		cblk->status &= ~USE_LR_DONE;
		cblk->status &= ~SET_LR_DONE;
	}
}


/* 
 * check_live_range - return TRUE if the full live range of the variable has no 
 * 	side-effect referneces return FALSE otherwise
 */
int check_live_range
        (
                /* the register to search on.  This should point to rtl->val somewhere */
                char *reg,

                /* an output param that's the list of live vars */
                rlist *lives,

                /* an output parameter that is the list of rtls in this regs lifetime */
                struct rtl_list** rtlist,

                /* the list of basic blocks that the register is live in */
                struct blist** blocklist

        )
{

	struct rtl_list *cur_node;
	for(cur_node=*rtlist; cur_node; cur_node=cur_node->next)
	{
		if(cnts(cur_node->rtl) && cur_node->rtl->se)
		{
			if(isaliasref(cur_node->rtl->se,reg) || isaliasset(cur_node->rtl->se,reg))
				return FALSE;
		}

	}
	return TRUE;
}




/* return TRUE if the full live range of the variable is found
 *  FALSE otherwise 
 */
int find_live_range
	(
		/* input parameter rtl -- where to start the search */
		struct list* rtl, 
		/* the register to search on.  This should point to rtl->val somewhere */
		char *reg,

		/* Is this register a use (is_use=TRUE) or a set (is_use=FALSE)*/
		int is_use,

		/* an output param that's the list of live vars */
		rlist *lives, 

		/* an output parameter that is the list of rtls in this regs lifetime */
		struct rtl_list** rtlist, 

		/* the list of basic blocks that the register is live in */
		struct blist** blocklist

	)
{
	char the_reg[MAXSTRING];
	assert(rtl);
	assert(reg);
	assert(rtlist);
	assert(blocklist);

	copyloc(the_reg,reg);
	*skiploc(the_reg)=0;

	clear_done_bits();


	/* we need to find all uses start here */
	if(is_use)
	{
		if(!extend_live_range_use(rtl->daddy,rtl,the_reg,lives,rtlist,blocklist))
			return FALSE;
		/* plus, all the sets */
		if(!extend_live_range_set(rtl->daddy,rtl,the_reg,lives,rtlist,blocklist))
			return FALSE;
	}
	else
	{
		if(rtl->next)
		{
			if(!extend_live_range_use(rtl->daddy,rtl->next,the_reg,lives,rtlist,blocklist))
				return FALSE;
			if(!extend_live_range_set(rtl->daddy,rtl,the_reg,lives,rtlist,blocklist))
				return FALSE;
		}
		else
		{
			if(!extend_live_range_set(rtl->daddy,rtl->next,the_reg,lives,rtlist,blocklist))
				return FALSE;
		}

	}
	return TRUE;

}


void update_liveness_info(struct list* rtl, rlist* lives)
{
	if(rtl->type&RTLINE)
		runion(lives, rtl->deads);
	if(rtl->type&RTLINE || rtl->type&SETLINE || rtl->type&PARMLINE || rtl->type&CSLINE || rtl->type&RESLINE)
	{
		char *a;
		a=rtl->val;
		while(a && (a=gtrg(a)))
			rinsert(unskiploc(a),lives);
		a=rtl->se;
		while(a && (a=gtrg(a)))
			rinsert(unskiploc(a),lives);
	}
}

/* 
 * look for uses of reg in cblk and successors 
 */
int extend_live_range_use
	(
		struct bblock* cblk, 
		struct list* rtl,
		char* reg,

		/* an output param */
		rlist *lives,

		/* an output parameter that is the list of rtls in this regs lifetime */
		struct rtl_list** rtlist, 

		/* the list of basic blocks that the register is live in */
		struct blist** blocklist
	)	
{
	/* we've already done this happy little block.. oh yes we have */
	if(cblk->status&USE_LR_DONE)
		return TRUE;
	if(rtl==NULL) 
	{
		/* this means we're starting at the top of a basic block
		   and that we have to go up to all preds of this block to find sets/uses 
		   there. 
		*/
		struct blist* preds=cblk->preds;
		cblk->status |= USE_LR_DONE;
		for(preds=cblk->preds; preds; preds=preds->next)
		{
			if(!  extend_live_range_set(preds->ptr,NULL, reg, lives, rtlist, blocklist))
				return FALSE;
		}
		rtl=cblk->rtls;
	}

	for(; rtl; rtl=rtl->next)
	{
		update_liveness_info(rtl,lives);
		/* if we find a use, yay for us */
		if(cnts(rtl) && isref(rtl->val,reg))
			add_use_to_live_range(rtl,rtlist,blocklist);

		/* if this rtl has an alias reference, but not an "normal" reference, we give up */
		if(cnts(rtl) && isaliasref(rtl->val,reg) && !isref(rtl->val,reg))
		  return FALSE;

		/* if this rtl has a side-effect, then analyze the side-effect too */
		if (cnts(rtl) && rtl->se) {
		    if(isref(rtl->se,reg) && !isref(rtl->val,reg))
			add_use_to_live_range(rtl,rtlist,blocklist);
		    if(isaliasref(rtl->se,reg) && !isref(rtl->se,reg))
			return FALSE;
		    if (isaliasset(rtl->se,reg))
			return FALSE;
		}
		/* 
		if(rtl->se && (isaliasref(rtl->se,reg)|| isaliasset(rtl->se,reg)))
			return FALSE;
		*/


		if(rtl->type&RESLINE || rtl->type&PARMLINE || rtl->type&CSLINE || rtl->type&SETLINE)
		{
			if(rtl->val && isaliasref(rtl->val,reg))
				return FALSE;
		}

		/* if it's dead here, we're done.. yay for us! */
		if(rexists(reg,rtl->deads))
			return TRUE;

		/* if the side effect sets/uses this reg (or an alias) give up */
		/*
		if(rtl->se && (isaliasref(rtl->se,reg)|| isaliasset(rtl->se,reg)))
			return FALSE;
		*/


		if(rtl->type&RESLINE || rtl->type&PARMLINE || rtl->type&CSLINE || rtl->type&SETLINE)
		{
			if(rtl->val && isaliasref(rtl->val,reg))
				return FALSE;
		}

		/* if it's dead here, we're done.. yay for us! */
		if(rexists(reg,rtl->deads))
			return TRUE;

		/* If it's re-set in this RTL, we're done yay for us! */
		if(cnts(rtl) && isaliasset(rtl->val,reg))
			return TRUE;
	}

	if(cblk->left && rexists(reg,cblk->left->ulist))
	{
		if(!extend_live_range_use(cblk->left,NULL,reg,lives, rtlist,blocklist))
			return FALSE;
	}
	if(cblk->right && rexists(reg,cblk->right->ulist))
	{
		if(!extend_live_range_use(cblk->right,NULL,reg,lives, rtlist,blocklist))
			return FALSE;
	}
	return TRUE;


}

void find_extra_lives(struct list* rtl, rlist* lives)
{
	rlist cur_lives=rinit();
	struct list* cur_rtl;
	char *a;

	rcpy(&cur_lives, rtl->daddy->ulist);
	for(cur_rtl=rtl->daddy->rtls;cur_rtl;cur_rtl=cur_rtl->next)
	{
		/* remove the deads */
		if(cur_rtl->type&RTLINE)
		{
			rlist empty;
			empty=rinit();
			/* empty = empty + (cur_lives-rtl_deads) */
			runiond(&empty,cur_lives,cur_rtl->deads);
			rcpy(&cur_lives,empty);
			rfree(empty);
		}

		/* stop if we've found the target */
		if(rtl==cur_rtl)
			break;

		/* add the sets */
		if(a=cnts(cur_rtl))
		{
			while(a=gtrg(a))
			{
				if(*a=='=')
					rinsert(unskiploc(a),&cur_lives);
			}
		}	
		if(a=cur_rtl->se)
		{
			while(a=gtrg(a))
			{
				/* watch for a reg set in a SE, that's not just trashed */
				if(*a=='=' && *(a+1)!=';')
					rinsert(unskiploc(a),&cur_lives);
			}
		}	
		if(cur_rtl->type&SETLINE)
		{
			a=cur_rtl->val;
			while(a=gtrg(a))
			{
				char *b;
				for(b=aliases(unskiploc(a));*b;b=skiploc(b)+1)
				{
					rdelete(b,&cur_lives);
				}
			}
		}	
		if(cur_rtl->type&RESLINE)
		{
			a=cur_rtl->val;
			while(a=gtrg(a))
			{
				rinsert(unskiploc(a),&cur_lives);
			}
		}	
	}
	runion(lives,cur_lives);
	rfree(cur_lives);
}

/* 
 * look for uses of reg in successors of cblk, then search back and try to find
 * the set-point 
 */
int extend_live_range_set
	(
		struct bblock* cblk, 
		struct list* rtl,
		char* reg,

		/* an output param */
		rlist *lives,

		/* an output parameter that is the list of rtls in this regs lifetime */
		struct rtl_list** rtlist, 

		/* the list of basic blocks that the register is live in */
		struct blist** blocklist
	)	
{
	/* we've already done this happy little block.. oh yes we have */
        struct blist* preds=cblk->preds;
	if(cblk->status&SET_LR_DONE)
		return TRUE;
	if(rtl==NULL) 
	{
		/* this means we're starting at the top of a basic block
		   and that we have to go up to all preds of this block to find sets/uses 
		   there. 
		*/
		cblk->status |= SET_LR_DONE;
		if(cblk->left && rexists(reg,cblk->left->ulist))
		{
			if(!extend_live_range_use(cblk->left,NULL,reg,lives, rtlist,blocklist))
				return FALSE;
		}
		if(cblk->right && rexists(reg,cblk->right->ulist))
		{
			if(!extend_live_range_use(cblk->right,NULL,reg,lives, rtlist,blocklist))
				return FALSE;
		}
		rtl=cblk->rtlend;
	}

	for(; rtl; rtl=rtl->prev)
	{
		
		/* if the side effect sets/uses this reg (or an alias) give up */
	    	/*
		if(rtl->se && (isaliasref(rtl->se,reg)|| isaliasset(rtl->se,reg)))
			return FALSE;
		*/
	    	
	    	
	    	if (cnts(rtl) && rtl->se && isaliasset(rtl->se,reg))
		    	return FALSE;

		/* if it's dead here, we're done.. yay for us! */
		if((rtl->type&RTLINE) && rexists(reg,rtl->deads))
			return TRUE;

		/* If it's re-set in this RTL, we're done yay for us! */
		if(cnts(rtl) && isaliasset(rtl->val,reg))
		{
			if(!isset(rtl->val,reg))
				return FALSE;
			add_set_to_live_range(rtl,rtlist,blocklist);
			find_extra_lives(rtl,lives);
			return TRUE;
		}

		/* if we find a use, yay for us */
		if(cnts(rtl) && isref(rtl->val,reg))
			add_use_to_live_range(rtl,rtlist,blocklist);
		else if (cnts(rtl) && rtl->se && isref(rtl->se,reg))
			add_use_to_live_range(rtl,rtlist,blocklist);

		if(rtl->type&RESLINE || rtl->type&PARMLINE || rtl->type&CSLINE || rtl->type&SETLINE)
		{
			if(rtl->val && isaliasref(rtl->val,reg))
				return FALSE;
		}


		if(cblk==top && rtl==top->rtls)
		{
			add_set_to_live_range(rtl,rtlist,blocklist);
			find_extra_lives(rtl,lives);
			return TRUE;
		}
		update_liveness_info(rtl,lives);
	}
        for(preds=cblk->preds; preds; preds=preds->next)
        {
        	if(!  extend_live_range_set(preds->ptr,NULL, reg, lives, rtlist, blocklist))
                	return FALSE;
        }


	return TRUE;


}

int should_minimize(char* old_reg, char* new_reg, struct list* rtl, struct rtl_list** rtlist, struct blist** blocklist, rlist *lives)
{
	char new_rtl[MAXSTRING];
	rlist new_deads=rinit();
	char *a;
	static int no_live_range=FALSE;

	strcpy(new_rtl,rtl->val);
	copyloc(new_rtl+(old_reg-rtl->val), new_reg);

	rcpy(&new_deads, rtl->deads);
	rdelete(new_reg,&new_deads);

	if(*rtlist!=0 && no_live_range)
		return FALSE;
	if(*rtlist==0)
		no_live_range=FALSE;


	/* first, check to see if the RTL is valid before we hunt down it's entire lifetime */
	if(find(rtl,new_rtl,new_deads))
	{
		rfree(new_deads);
		return FALSE;
	}
	rfree(new_deads);

	/* yay!  we're halfway there.  we've found that this RTL is safe to change */
	/* now we need to hunt down the lifetime of this variable and */


	if(*rtlist==NULL) 
	{
		if(!find_live_range ( rtl, old_reg, FALSE, lives, rtlist, blocklist))
		{
			/* we failed to find the entire live range, bail out now */
			/* also, record that we failed until next time we start anew */
			no_live_range=TRUE;
			return FALSE;
		}
		if(!check_live_range(old_reg,lives,rtlist,blocklist))
		{	
			/* we've found the live range, but it has a side-effect
			 * or other thing we don't want to deal with here.
			 * thus, we bail
			 */
			no_live_range=TRUE;
			return FALSE;
		}
	}
	for(a=aliases(new_reg);*a;a=skiploc(a)+1)
	{
		if(rexists(a,*lives))
			return FALSE;
	}
	return TRUE;



}

int only_local_changes(char *old_reg,char *new_reg,struct rtl_list *rtlist)
{
  int ornum = regnum(old_reg);
  int changed=FALSE,old_reg_set=FALSE;
  struct list *rtl;
  struct rtl_list *rl;
  struct bblock *blk;
  blk = rtlist->rtl->daddy;

  for (rl = rtlist->next;rl;rl = rl->next) {
    if (rl->rtl->daddy != blk)
      return FALSE;
  }

  /* All the rtls to be changed are in the same block. Check to see if
     new_reg needs to be in the dlist */

  /* Check to see if old_reg is still in the dlist */

  for (rtl = blk->rtls;rtl;rtl = rtl->next) {
    if (!cnts(rtl))
      continue;

    if (isset(rtl->val,old_reg) || (rtl->se && isset(rtl->se,old_reg))) {
      old_reg_set = TRUE;
      break;
    }



  }

  if (!old_reg_set)
    rdelete(old_reg,&blk->dlist);

  rinsert(new_reg,&blk->dlist);
  return TRUE;

}

int attempt_minimize(char* old_reg2, char* new_reg, struct rtl_list* rtlist, struct blist* blocklist, rlist lives)
{
	char old_reg[REGSTRING];

	char *a;
	struct rtl_list* cur_node,*last_node;
	struct blist* cur_block;
	int ornum;
	if(!can_minimize(old_reg2, new_reg, rtlist, blocklist, lives))
		return FALSE;

#ifdef VET
	begin_trans(VET_OPT_TRANS);
#endif

	/* we need to preserve old_reg2 because it might change before we're done */
	copyloc(old_reg,old_reg2);
	*skiploc(old_reg)=0;
	for(last_node = cur_node=rtlist; cur_node; last_node=cur_node,cur_node=cur_node->next)
	{
		if (cur_node->rtl->val) {
			a=cur_node->rtl->val;
			while(a=gtrg(a))
			{
				if(*a=='=' && cur_node->set && cmploc(unskiploc(a),old_reg)==0)
					copyloc(unskiploc(a),new_reg);
				if(*a!='=' && cur_node->used && cmploc(unskiploc(a),old_reg)==0)
					copyloc(unskiploc(a),new_reg);
			}
		}
		if(cur_node->rtl==top->rtls)
		{
			struct locuse* lptr;
			for(lptr=locs;lptr;lptr=lptr->next)
			{
				if(*lptr->regin && cmploc(lptr->regin,old_reg)==0)
					copyloc(lptr->regin,new_reg);
			}
		}
		if(cur_node->used)
		{
			/* fix up the dead list */
			if(rexists(old_reg,cur_node->rtl->deads))
			{
				rinsert(new_reg,&cur_node->rtl->deads);
			}
			for(a=aliases(old_reg);*a;a=skiploc(a)+1)
				rdelete(a,&cur_node->rtl->deads);

		}
		if(cur_node->set || (cur_node->rtl->val && isaliasset(cur_node->rtl->val,new_reg)))
		{
			rdelete(new_reg,&cur_node->rtl->deads);
		}

	}
#ifdef VET
	end_trans();
#endif
	/* If all the changes are to rtls in a single block, then we
		dont have to do live variable analysis */
	if (!only_local_changes(old_reg,new_reg,rtlist))
		/* the live var info is all out of wack if we've changed something, redo it for the next thing */
		live_variable_analysis(0,1,0);

	return TRUE;
}


int can_minimize(char* old_reg, char* new_reg, struct rtl_list* rtlist, struct blist* blocklist, rlist lives)
{
	char new_rtl[MAXSTRING];
	rlist new_deads=rinit();
	char *a;


	struct rtl_list* cur_node;
	struct blist* cur_block;

	for(cur_node=rtlist; cur_node; cur_node=cur_node->next)
	{
		if (!cur_node->rtl->val) continue;
		else if ( (cur_node->rtl->type & RTLINE)==0) continue;
		strcpy(new_rtl,cur_node->rtl->val);
		rcpy(&new_deads,cur_node->rtl->deads);
		a=new_rtl;
		while(a=gtrg(a))
		{
			if(*a=='=' && cur_node->set && cmploc(unskiploc(a),old_reg)==0)
				copyloc(unskiploc(a),new_reg);
			if(*a!='=' && cur_node->used && cmploc(unskiploc(a),old_reg)==0)
				copyloc(unskiploc(a),new_reg);
		}

		/* fix up the dead list */
		for(a=aliases(old_reg); *a; a=skiploc(a)+1)
		{
			if(rexists(a,cur_node->rtl->deads))
			{
				rdelete(a,&new_deads);
			}
		}
		if(!isaliasset(new_rtl,new_reg))
			rinsert(new_reg,&new_deads);
		/* first, check to see if the RTL is valid before we hunt down it's entire lifetime */
		if(find(cur_node->rtl,new_rtl,new_deads))
		{
			rfree(new_deads);
			return FALSE;
		}
		rfree(new_deads);
	}
	return TRUE;


}



int minimize_registers()
{
	struct bblock* cblk;
	struct list* rtl;
	int change=FALSE;
	rlist lives=rinit();
	struct rtl_list* rtlist=NULL;
	struct blist* blocklist=NULL;

	if(debug_opt("no_minimize_regs"))
		return FALSE;


	/* this is expensive.  we only want to do it if CSE, code motion and 
	 * register assignment have been requested.  Also, we only use it for "size"
	 * optimizations, although it may help speed as well.
	 */
	if(!swo || !swc)
		return FALSE;


	/* we need to explicitly ask for expensive speed optimizations before we do this */
	if(!debug_opt("size") && !debug_opt("speed")) 
		return FALSE;


	live_variable_analysis(0,1,0);
#ifdef VET
#ifndef INTERACT
    begin_phase(REGALLOC_PHASE);
#endif
#endif

	for(cblk=top;cblk;cblk=cblk->down)
	{
	  
		for(rtl=cblk->rtls;rtl;rtl=rtl->next)
		{
			char *a, b[REGSTRING];
			char canidates[MAXSTRING];
			if(!cnts(rtl))
				continue;
			a=rtl->val;
			while(a=gtrg(a))
			{
				int rt;
				int index;
				int items;
				/* skip uses */
				if(*a!='=')	continue;
				copyloc(b,unskiploc(a));
				*skiploc(b)=0;

				/* reset the live range info because we're looking at a new reg */
				lives=rinit();
				rtlist=NULL;
				blocklist=NULL;

				rt=regtype(b);
				if(isref(rtl->val,b))  continue;
				items=availreg(rt);
				for(index=firstreg(rt); items--; index+=skipreg(rt))		
				{
					makereg(b,rt,indextoreg(rt,index));
					if(index >= regtoindex(rt,regnum(unskiploc(a))))
						continue;
					/* if the types match, and the number is less, try to minimize it */
					if(should_minimize(unskiploc(a),b,rtl, &rtlist,&blocklist,&lives))
					{
						if(attempt_minimize(unskiploc(a),b,rtlist,blocklist,lives))
						{
							change=TRUE;
							goto try_next_rtl;
						}
					}
				}
			}

			try_next_rtl:
				rfree(lives);
				;
		}
	}
#ifdef VET
#ifndef INTERACT
	end_phase(REGALLOC_PHASE);
#endif
#endif
	return change;
}
























