/************************************************************************
 *                                                                      *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   *
 *     The Rector and Visitors of the University of Virginia            *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from the University of Virginia.                                *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/

#ifndef LIVE_H
#define LIVE_H

struct rtl_list {
	struct list* rtl;
	struct rtl_list* next;

	/* a bit to tell me whether the thing is set or not */
	unsigned int set:1;
	unsigned int used:1;
};

int
minimize_registers();

void add_use_to_live_range
	(
		/* the rtl to add to the list */

		struct list* rtl, 
		/* an output parameter that is the list of rtls in this regs lifetime */
		struct rtl_list** rtllist, 

		/* the list of basic blocks that the register is live in */
		struct blist** blocklist
	);

void add_set_to_live_range
	(
		/* the rtl to add to the list */

		struct list* rtl, 
		/* an output parameter that is the list of rtls in this regs lifetime */
		struct rtl_list** rtllist, 

		/* the list of basic blocks that the register is live in */
		struct blist** blocklist
	);

/* mark all blocks in the func as done */
void 
clear_done_bits();

/* return TRUE if the full live range of the variable is found
 *  FALSE otherwise 
 */
int find_live_range
	(
		/* input parameter RTL -- where to start the search */
		struct list* rtl, 
		/* the register to search on.  This should point to rtl->val somewhere */
		char *reg,

		/* Is this register a use (is_use=TRUE) or a set (is_use=FALSE)*/
		int is_use,

		/* an output param that's the list of live vars */
		rlist* lives, 

		/* an output parameter that is the list of RTLs in this regs lifetime */
		struct rtl_list** rtllist, 

		/* the list of basic blocks that the register is live in */
		struct blist** blocklist

	);

/* 
 * look for uses of reg in cblk and successors 
 */
int extend_live_range_use
	(
		struct bblock* cblk, 
		struct list* rtl,
		char* reg,

		rlist* lives, 

		/* an output parameter that is the list of rtls in this regs lifetime */
		struct rtl_list** rtllist, 

		/* the list of basic blocks that the register is live in */
		struct blist** blocklist
	);	

/* 
 * look for uses of reg in successors of cblk, then search back and try to find
 * the set-point 
 */
int extend_live_range_set
	(
		struct bblock* cblk, 
		struct list* rtl,
		char* reg,

		rlist* lives, 

		/* an output parameter that is the list of rtls in this regs lifetime */
		struct rtl_list** rtllist, 

		/* the list of basic blocks that the register is live in */
		struct blist** blocklist
	);	


#endif  /* LIVE_H */
