/************************************************************************
 *                                                                      *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   *
 *     The Rector and Visitors of the University of Virginia            *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from the University of Virginia.                                *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/locals.h,v $
 $Date: 2003/12/16 22:14:38 $
 $Revision: 1.11 $
 $Name:  $

 Log of changes:

 $Log: locals.h,v $
 Revision 1.11  2003/12/16 22:14:38  clc5q
 Add struct globalsym fields: PadBytes, SectionNumber, and AffinityIndex.
 Add debugging function prototypes.

 Revision 1.10  2003/11/18 02:01:58  clc5q
 Major implementation portion of global data profiling and placement.

 Revision 1.9  2003/11/03 00:34:09  clc5q
 Begin implementing data placement profiling of globals.

 Revision 1.8  2003/05/04 01:59:47  clc5q
 Export "loce", which points to the end of the locals list.

 Revision 1.7  2003/02/06 04:48:21  jg9h
 Merged VISTA into mainline vpo under the INTERACT preprocessor flag

 Revision 1.6  2002/11/27 16:28:08  jwd
 Sampoorani updated this code so that variables that heavily used
 in loops can be allocated even if no variable is available.  This
 code was originally written by Raja, but it was broken from the get
 go.

 Revision 1.5  2002/08/08 19:00:12  clc5q
 Add parameter "sort" to locals_used(). Use it as a guard for whether to
 call reorder_locals().

 Revision 1.4  2002/07/11 15:20:48  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.3  2002/06/05 20:08:53  jdh8d
 Changes to support the stache, including routines to build a callgraph.

 Some cosmetic changes.

 Minor bug fixes.

 Changed build scripts to build the mips compiler for SimpleScalar if
 SSAS is set to "-DSS_AS"

 Revision 1.2  1998/04/28 20:22:36  jwd
 Added some necessary header files.

 Deleted some commented out code.

 Revision 1.1  1998/04/03 16:53:51  jwd
 Initial revision

*************************************************************************/

#ifndef LOCALS_H
#define LOCALS_H

#ifdef CALLGRAPH
#include "callgraph.h"
#endif

/* Holds linked list of ASMLINEs that are associated with a global variable
 * and need to be emitted before we finish the input file.  The first line
 * for a global will probably be the global name followed by a colon, and
 * remaining lines will usually initialize the global.  It is also possible
 * for the only line to be an uninitialized common block allocation.
 */
struct globalAsmLines {
   char *line;
   struct globalAsmLines *next;
};

/* Type of entries in the symbol list for global variables. */
struct globalsym {
   char code[REGSTRING];   /* Binary code used in RTLs. */
   int gallocIndex;        /* Index into globals array for its name. */
   int size;               /* Size of global variable in bytes. */
   short PadBytes;           /* # bytes padding needed after global. */
   short SectionNumber;      /* Internal index for assembler section. */
   short PlacementPartition; /* Data placement partition number. */
   short AffinityIndex;      /* Index into AffinityArray[]. */
   struct globalsym *next; /* Linked list pointer to next item. */
   struct globalsym *prev; /* Linked list pointer to previous item. */
   struct globalsym *CoalescedMSNeighbor;
             /* Left (most significant) chain neighbor */
   struct globalsym *CoalescedLSNeighbor;
             /* Right (least significant) chain neighbor */
   char *sectionBuffer;    /* Contains assembler data section type directive.*/
   struct globalAsmLines *FirstLine; /* First assembly line to emit. */
   struct globalAsmLines *LastLine;  /* Last assembly line to emit. */
};

extern struct globalsym *globSymStart;
extern struct globalsym *globSymEnd;
extern int CollectingGlobals;

extern struct ignode *igptr;   /* first node in interference graph list */
extern struct ignode *iglast;  /*  last node in interference graph list */
extern struct locuse *locs;    /* list of locals used in current function */
extern struct locuse *loce;    /* last local in locs list */
extern struct loopl *loops;    /* start of loop list */

struct globalsym *findGlobalFromName(char *);
struct globalsym *findGlobalFromCode(char *);
void emitAsmForGlobals(int);
void BuildBlockGlobalProfile(int);
void PlaceGlobalData(struct call_graph_node *);
void RecordCurrentCallGraphNode(struct call_graph_node *);
FILE *OpenGlobalDebugFile();
void ShowGlobalsArray();
void CloseGlobalDebugFile();

int 
simple_global_allocation(int, int *);

int 
color_global_allocation(int);

void 
loctoreg(struct locuse *, struct list *, int, char *);

void 
locals_used(int sort);

void 
reorder_local(struct locuse *);

void 
sort_interference_graph(void);
void new_sort_ig();

int 
sets_only_cc(struct list *);

int 
count_caller_save(void);

int 
callersave(char *);

void 
move_caller_save(int);

void 
clean_locals(void);

void 
clear_locals(void);

int 
declaration(char *);

void 
makeglobal(char *, char *);


#endif
