/*
 * loop_unroll.c
 * File contains functions to assist in loop unrolling
 */

#include <stdio.h>
#include <stdlib.h>
#include "vpo.h"
#include "misc.h"
#include "view.h"
#include "controlflow.h"

#define MAX_UR_BLOCKS 50
#define MAX_CHANGE_RTL 50
#define MAX_QUEUE_BLKS 50
#define MAX_UNROLL_DEPTH 8
#define MAX_CLEANUP_LENGTH 4

#ifndef UNDO_H
void clear_all_preds();
void recalculate_all_preds();
void clear_bb_preds(struct bblock *);
int is_ucjmp(char *);
#endif
#ifndef REMJUMP_H
void change_branch_label(struct list *, char *);
struct bblock *copy_block(struct bblock *);
#endif

int do_loop_analysis(unsigned int);
struct bblock *prelogue(struct loopl *loop, int iter, int UR);
struct bblock *epilogue(struct loopl *loop, int iter, int UR);
int getURfactor(struct loopl *loop, int unroll, int iter);
int isHeader(struct bblock *blk);
int canUnrollLoop(struct loopl *loop, int unroll);
struct list *findCompareRtl(struct bblock *blk);
struct list *findJumpRtl(struct bblock *blk);
int rtlCount(struct bblock *blk);
void moveRTL(struct list *rtl, struct bblock *new, struct list *nrtl);
struct bblock *createNewBranchBlock(struct loopl *loop, int loopType);
struct bblock *findLoopCntrBlk(struct loopl *loop);
struct bblock *findMainLoopExit(struct loopl *loop);
struct bblock *singleExitBlk(struct loopl *loop);
struct bblock *insCleanup(struct loopl *loop, int iter, int UR, struct bblock *epilog,
		struct bblock *exitblk, int loopType);
void insertBlk(struct bblock *nblk, struct bblock *nextblk);
void setLoop13Links(struct loopl *loop, struct bblock *dup_map[MAX_BLOCKS],
		    struct bblock *pdown, struct bblock *blkup, struct bblock *blkdown);
void setCleanup2Links(struct loopl *loop, struct bblock *dup_map[MAX_BLOCKS],
		    struct bblock *pdown, struct bblock *blkup, struct bblock *blkdown);
void setCleanup13Links(struct loopl *loop, struct bblock *dup_map[MAX_BLOCKS],
		    struct bblock *pdown, struct bblock *blkup, struct bblock *blkdown);
void removeBlk(struct bblock *blk);
void findLoopTopBottom(struct loopl *loop, struct bblock **top, struct bblock **bottom);
int getLoopStructure(struct loopl *loop);
int unrollType13Loops(struct loopl *loop, int iter, int UR, int loopType);
int unrollType2Loops(struct loopl *loop, int iter, int UR, int loopType);
void setType12PrelogLinks(struct loopl *loop, struct bblock *preheader);
void setType3PrelogLinks(struct loopl *loop, struct bblock *preheader);
void setType13EpilogLinks(struct loopl *loop, struct bblock *epiblk,
			  struct bblock *exitblk);
void setType2EpilogLinks(struct loopl *loop, struct bblock *epiblk,
			 struct bblock *exitblk);
struct list *findLastRTL(struct bblock *blk);
struct bblock *createNewJumpBlock(struct loopl *loop);
void update_bb_ptrs(int dir, struct bblock *block, struct bblock *child);
void change_branch(struct loopl *loop, struct bblock **exitblk, int iter, int UR, int);
int loopTermination(int init, int inc, char relop, int value, int iter, int loopType);
int loop_iterations(struct loopl *loop, struct setlist **cptr, char *value,
		    char *initial, int *inc, char *cval, char *reloper);
int remvLoopVarUpdate(struct loopl *loop, struct bblock *epilog,
		      struct bblock *exitblk, int loopTpe);
void getRTLComp(char *rtlval, char *set, char *op1, char *oper, char *op2);
void consRTL(char *rtlval, char *set, char *op1, char oper, char *op2);
void addToQueue(struct bblock *Qblk[], struct bblock *cblk, int *bottom);
struct bblock *remvFromQueue(struct bblock *Qblk[], int *top);
void add_to_llist(struct list *rtl, struct llist **list, int inc);
void handleUnsignedLoops(struct loopl *loop, struct bblock *clHead,
			 struct bblock *exitblk, int iter, int UR, int loopType);
struct bblock *NEW_basic_block(char *label);
void replaceReg(char *rtlval, char *reg, char *rep, char *str);
void fix_controlflow();
void remvIncBlk(struct loopl *loop, struct bblock *exitblk);
void modifyLoopExits(struct loopl *loop);
void add_block_to_loop(struct loopl *loop, struct bblock *blk);
char *initialLoopVal(char *init);
char *loopCompVal(char *cval);
int loopInc(int *inc);
char loopRelop(char *rel);
int isInteger(char *str);


struct RTLStore{
  struct list *rtl;
  char newval[MAXSTRING];
};

struct bblock *ltop, *lbottom;
int allocRegister, multipleExitLoop, updateAfterCompare;

/*
 * getLoopStructure - We consider 3 different loop structures for loop
 *                    unrolling, as illustrated in the function
 *                    'loop_unroll'. Here we determine the current loop's
 *                    structure
 */
int getLoopStructure(struct loopl *loop)
{
  int loopType;
  struct blist *bl;
  struct bblock *pheader, *exitblk, *cblk;
  
#ifdef INTERACT
  begin_trans(VET_OPT_TRANS);
#endif
      
  /*get the block preheader*/
  pheader = (struct bblock *)preheader(loop->ptr);
  if(!pheader || !pheader->down){
    fprintf(stderr, "Unrecognized loop structure. Loop preheader is the last block\n");
    return FALSE;
  }

  /*find the loop top and bottom blocks in the control flow graph. This
    will help us attatch the entire duplicated loop body in the
    function control flow structure*/
  ltop = lbottom = (struct bblock *)NULL;
  findLoopTopBottom(loop, &ltop, &lbottom);

  /*All the blocks in any loop must be contiguous in the control flow*/
  /*for(cblk=ltop ; cblk && cblk!=lbottom; cblk=cblk->down){
    if(!block_in_loop(loop, cblk))
      return FALSE;
  }
  if(!cblk)
  return FALSE;*/

  /* This code may not handle the case when the loop bottom is the last block
     in the control flow of the function. Check for this*/
  if(!lbottom->down)
    return FALSE;

  /*get the loop block which transfers control outside of the loop*/
  exitblk = findMainLoopExit(loop);

  /*For type 1 loops the header block is the topmost block in the loop, and
    the exitblock in the last block*/
  if(loop->ptr==ltop && exitblk==lbottom)
    loopType = 1;

  /*for type 2 loops, the header and exit blocks are the same and are located
    at the top of the loop. The last block should have an unconditional jump
    to the loop header*/
  else if(loop->ptr == exitblk && exitblk == ltop &&
	  (lbottom->left==loop->ptr && lbottom->right==NULL))
    loopType = 2;

  /*for type 3 loops, the header and exit blocks are the same and are located
    at the bottom of the loop*/
  else if(loop->ptr == exitblk && exitblk == lbottom)
    loopType = 3;

  /*there is not much advantage to unrolling this type of loops*/
  else if(loop->ptr == exitblk && exitblk == ltop && lbottom->right==loop->ptr)
    return FALSE;

  else{
    /*fprintf(stderr, "Unrecognized loop structure: %s\n", loop->ptr->label);*/
    loopType = FALSE;
  }
  
  /*sanity check, verify that the header block is reached by only one block
    in the loop*/
#ifdef INTERACT
  end_trans();
#endif      
  
  return loopType;
}


/*
 * loop_unroll - function called to unroll loops. Returns the number of loops
 *               sucessfully unrolled. This function assumes that all
 *               necessary loop analysis has already been performed.
 */
int loop_unroll(int unroll)
{
  char comment[MAXSTRING];
  int numLoops = 0;
  int change, iter, UR, loopType, analysis, tchange;
  struct loopl *loop;
  struct bblock *cblk;
  struct blist *bl;
  extern struct bblock *top;
  extern char *fn;
  extern int lcount;
  extern struct loopl *loops;

#ifdef INTERACT
  begin_phase(LOOP_UNROLL_TRANS);
#endif  

  tchange = change = FALSE;
  
 /*the loops can be scanned using the 'loops' list. To allow restricting the
    scope of this optimization to basic blocks, we do it by scanning the
    basic block list*/
  for(cblk = top; cblk; cblk = cblk->down){
#if 0
    /*if (lcount >= 10)
      break; */
#endif

    /*first identify the type of the loop depending on its structure in the
      control flow graph. There can be 3 loop structure:

          Type 1             Type 2              Type 3

        L00:               L00: compare        L00: j L01
             ....               br L02              ....
             ....               ....           
    	                        ....           L02: ....
        L01: compare                                ....
             br L00        L01: j L00              
                           L02:                L01: compare  
	                                            br L02

    */

    /*set the unrolled flag of blocks to FALSE*/
    cblk->status &= ~(UNROLLED);
    
    /*check if this block is the header of some loop and if it is a candidate
      for unrolling. Currently we have certain restrictions on which loops
      can be unrolled*/
    /*the function isUnrollLoop() 
      returns  0 if loop cannot be unrolled,
      returns -1 if loop iteration count is unknown, and
      returns >0 if loop iteration count is known at compile time
    */
    if((loop=cblk->loop)!=NULL && (iter=canUnrollLoop(loop, unroll))
       && (loopType=getLoopStructure(loop))){

      /*determine the loop unroll factor*/
      UR = getURfactor(loop, unroll, iter);

      /*The iter we have is the number of times the header is executed. For
	type 2 and 3 loops, the header is executed one more time than the
	loop body. In this algorithm we need the number of times the loop is
	executed. If the loop variable update instruction is after the
	compare, then we do not need to reduce the number of iterations */
      if(loopType != 1 && !updateAfterCompare)
	iter--;

      if(UR != 1){

	/*fprintf(stderr, "Loop: %s\n", loop->ptr->label);*/
	if(loopType == 1)
	  tchange = unrollType13Loops(loop, iter, UR, loopType);
	else if(loopType == 2)
	  tchange = unrollType2Loops(loop, iter, UR, loopType);
	else if(loopType == 3)
	  tchange = unrollType13Loops(loop, iter, UR, loopType);
	
	change |= tchange;
	if(tchange)
	  {
	    /*fprintf(stderr, "Loop Unrolled in function %s\n", fn);*/
	    lcount++;
	    cblk->status |= UNROLLED;
	  }

	numLoops++;
      }
    }
  }

#ifdef INTERACT
  end_phase(LOOP_UNROLL_TRANS);
#endif

  /*need to free the triple list created during the identification of basic
    induction variables*/
  freetriples();

  /*this code is only needed to mark the status of loop header blocks*/
  analysis = DOMINATOR_ANALYSIS|LOOP_ANALYSIS;
  do_needed_analysis(analysis); 
  for(loop=loops ; loop ; loop=loop->next){
    /*traverse loop blocks to check if any block in this loop is marked as
      UNROLLED */
    for(bl=loop->blocks ; bl ; bl=bl->next){
      if(bl->ptr->status & UNROLLED){

	/*add a new UNROLLED RTL to the header basic block*/
	loop->ptr->status |= UNROLLED;
	/*strcpy(comment, "Block Unrolled\n");
	add_rtl_with_type(loop->ptr, loop->ptr->rtls, comment,
	NULL, UNROLL_BLOCK);*/
	break;
      }
    }
  }
  
  return change;
}


int unrollType13Loops(struct loopl *loop, int iter, int UR, int loopType)
{
  int i, change, numDupBlocks, flag, analysis;
  struct blist *bl;
  struct bblock *cblk, *branch, *new, *nextblk, *cleanupHead, *dup_map[MAX_BLOCKS];
  struct bblock *preheader, *exitblk, *epilog, *blkup, *blkdown, *oblk, *tblk;

#ifdef INTERACT
  begin_trans(VET_OPT_TRANS);
#endif  
  /*if needed, insert code which will determine if the unrolled loop can
    be entered, else will jump straight to clean-up*/
  if(!(preheader = prelogue(loop, iter, UR))){
#ifdef INTERACT
    end_trans();
#endif
    return FALSE;
  }

  if(preheader){

    /*set the proper links for this new preheader block*/
    if(loopType == 1){
      
      /*insert this block immediately before the loop header, which is the
	firt loop block in the control flow*/
      insertBlk(preheader, loop->ptr);
      setType12PrelogLinks(loop, preheader);
    }
    else{
      
      /*insert this block immediately after the original preheader*/
      insertBlk(preheader, loop->preheader->down);
      setType3PrelogLinks(loop, preheader);
    }
    /*set this block as the new loop preheader*/
    loop->preheader = preheader;
  }
#ifdef INTERACT
  end_trans();
  begin_trans(VET_OPT_TRANS);
#endif

  /*for type 2,3 loops there is an external jump to the loop control block.
    This block can still contain additional instructions other than the
    compare and branch. Seperate these out into a new block*/
  if(loopType == 1)
    exitblk = createNewBranchBlock(loop, 1);
  else if(loopType == 3)
    exitblk = createNewBranchBlock(loop, 3);
  
  /*if we needed to add prelogue before the loop, then add epilogue block
    after the loop*/
  epilog = (struct bblock *)NULL;
  if(preheader){

    epilog = epilogue(loop, iter, UR);

    /*insert this block immediately following the loop exit*/
    insertBlk(epilog, exitblk->down);

    /*set the correct pointers for the epilogue block*/
    setType13EpilogLinks(loop, epilog, exitblk);
  }
#ifdef INTERACT
  end_trans();
#endif  

  /*cleanup code may be needed after the main unrolled loop, if the
    unroll factor is not an exact multiple of the loop bound*/
  cleanupHead = insCleanup(loop, iter, UR, epilog, exitblk, 3);

  /*now finally!! we are ready to unroll the main body of the loop*/
	/*intialize dup_map*/

  /*initialize dup_map structure to hold new basic block pointers
    corresponding to the old blocks*/
  for(i=0 ; i<MAX_BLOCKS ; i++)
    dup_map[i] = (struct bblock *)NULL;

  /*the duplicated loop body is inserted between blkup and blkdown*/
  blkdown = exitblk;
  blkup = exitblk->up;
  
  /*duplicate the blocks UR number of times*/
  for(i=1 ; i<UR ; i++){
	  
#ifdef INTERACT
    begin_trans(VET_OPT_TRANS);
#endif
    /*duplicate the blocks and adjust the control flow of the new
      blocks*/
    for(bl=loop->blocks ; bl ; bl=bl->next){

	if(bl->ptr != exitblk){
	oblk = bl->ptr;
	new = dup_map[oblk->block_id] = copy_block(oblk);
	      
	/*for type 1 loops, insert the duplicated blocks above exitblk, if
	  the actual down block has not be duplicated as yet*/
	for(tblk=oblk->down ; tblk && tblk!=lbottom->down ; tblk=tblk->down){
	  if(dup_map[tblk->block_id])
	    break;
	}
	if(tblk==NULL || !(nextblk=dup_map[tblk->block_id]))
	  nextblk = blkdown;
	insertBlk(new, nextblk);
      }
    }

    /*set the left/right links for all the new blocks*/
    setLoop13Links(loop, dup_map, blkdown, blkup, blkup->down);

    /*re-initialize dup_map and other pointer variables for the next iteration*/
    for(bl=loop->blocks ; bl ; bl=bl->next)
      dup_map[bl->ptr->block_id] = (struct bblock *)NULL;
    blkdown = blkup->down;

#ifdef INTERACT
    end_trans();
#endif  
  }

  /*We need to change the compare/branch instructions checking for exit from
    the unrolled loop. For compile time knowm bounds we set the loop bound to
    be a multiple of the unroll factor. For execute time bounds we may need
    to change the relop or bounds of the loop as well.
    Since this block is duplicated elsewhere during loop unrolling, we can
    only perform this modification at the end of this function*/
#ifdef INTERACT
  begin_trans(VET_OPT_TRANS);
#endif
  change_branch(loop, &exitblk, iter, UR, loopType);

  /*it is possible that loop unrolling produced code where the left and right
    pointers are inconsistent. Fix this now*/
  fix_controlflow();
  
  /*remvIncBlk(loop, exitblk);*/
  recalculate_all_preds();

  /*for loops with unsigned comparison/branch this approach to loop unrolling
    might give incorrect results. There is a chance that after adjusting the
    loop control value by the unroll factor, the loop control value may
    overflow. This function inserts additional code before entering the
    unrolled loop to avoid this*/
  handleUnsignedLoops(loop, cleanupHead, exitblk, iter, UR, loopType);
  
  /*now we can attempt to remove unnecessary updates to the loop variable. 
    For that we must first replace all uses of the loop variable by
    appropirate increments. Oops! cannot do this if there are multiple exits
    from the loop*/
  if(exitblk && !multipleExitLoop){
    analysis = DEPTH_FIRST_SEARCH_ANALYSIS | DOMINATOR_ANALYSIS;
    /*do_loop_analysis(analysis);*/ /*change for Bill's code*/
    do_needed_analysis(analysis);
    remvLoopVarUpdate(loop, epilog, exitblk, loopType);
  }
#ifdef INTERACT
  end_trans();
#endif

  return TRUE;
}


int unrollType2Loops(struct loopl *loop, int iter, int UR, int loopType)
{
  int i, change, numDupBlocks, flag, analysis;
  struct blist *bl;
  struct bblock *cleanupHead;
  struct bblock *cblk, *branch, *new, *nextblk, *dup_map[MAX_BLOCKS], *oblk;
  struct bblock *preheader, *exitblk, *jumpblk, *epilog, *blkup, *blkdown, *tblk;

#ifdef INTERACT
  begin_trans(VET_OPT_TRANS);
#endif  
  /*if needed, insert code which will determine if the unrolled loop can
    be entered, else will jump straight to clean-up*/
  if(!(preheader = prelogue(loop, iter, UR))){
#ifdef INTERACT
    end_trans();
#endif    
    return FALSE;
  }
  
  if(preheader){
    
    /*insert this block immediately before the loop header, which is the
      firt loop block in the control flow*/
    insertBlk(preheader, loop->ptr);
    setType12PrelogLinks(loop, preheader);

    /*set this block as the new loop preheader*/
    loop->preheader = preheader;
  }
#ifdef INTERACT
  end_trans();
  begin_trans(VET_OPT_TRANS);
#endif

  /*for type 2 loops there is an external jump to the loop control block.
    This block can still contain instructions other than the compare and
    branch. These instructions should be seperated into a new block. At the
    same time this new block has to be copied into the loop body*/
  exitblk = createNewBranchBlock(loop, 2);
  jumpblk = createNewJumpBlock(loop);
  
  /*if we needed to add prelogue before the loop, then add epilogue block
    after the loop*/
  epilog = (struct bblock *)NULL;
  if(preheader){

    epilog = epilogue(loop, iter, UR);

    /*insert this block immediately before the exit block*/
    if(!block_in_loop(loop, exitblk->left))
      insertBlk(epilog, exitblk->left);
    else
      insertBlk(epilog, exitblk->right);

    /*set the correct pointers for the epilogue block*/
    setType2EpilogLinks(loop, epilog, exitblk);
  }
#ifdef INTERACT
  end_trans();
#endif  

  /*cleanup code may be needed after the main unrolled loop, if the
    unroll factor is not an exact multiple of the loop bound*/
  cleanupHead = insCleanup(loop, iter, UR, epilog, exitblk, 2);

  /*now finally!! we are ready to unroll the main body of the loop*/
	/*intialize dup_map*/

  /*initialize dup_map structure to hold new basic block pointers
    corresponding to the old blocks*/
  for(i=0 ; i<MAX_BLOCKS ; i++)
    dup_map[i] = (struct bblock *)NULL;

  /*the duplicated loop body is inserted between blkup and blkdown*/
  blkdown = jumpblk;
  blkup = jumpblk->up;
  
  /*duplicate the blocks UR number of times*/
  for(i=1 ; i<UR ; i++){
	  
    /*duplicate the blocks and adjust the control flow of the new
      blocks*/
    for(bl=loop->blocks ; bl ; bl=bl->next){
      if(bl->ptr!=exitblk && bl->ptr!=jumpblk){
#ifdef INTERACT
	begin_trans(VET_OPT_TRANS);
#endif
	oblk = bl->ptr;
	new = dup_map[oblk->block_id] = copy_block(oblk);
	      
	/*for type 1 loops, insert the duplicated blocks above exitblk, if
	  the actual down block has not be duplicated as yet*/
	for(tblk=oblk->down ; tblk && tblk!=lbottom->down ; tblk=tblk->down){
	  if(dup_map[tblk->block_id])
	    break;
	}
	if(tblk==NULL || !(nextblk=dup_map[tblk->block_id]))
	  nextblk = blkdown;
	insertBlk(new, nextblk);
	
#ifdef INTERACT
	end_trans();
#endif
      }
    }

#ifdef INTERACT
    begin_trans(VET_OPT_TRANS);
#endif  
    /*set the left/right links for all the new blocks*/
    setLoop13Links(loop, dup_map, blkdown, blkup, blkup->down);

    /*re-initialize dup_map and other pointer variables for the next iteration*/
    for(bl=loop->blocks ; bl ; bl=bl->next)
      dup_map[bl->ptr->block_id] = (struct bblock *)NULL;
    blkdown = blkup->down;

#ifdef INTERACT
    end_trans();
#endif  
  }

  /*for compile time known bounds we need to change the branch instruction
    of the main unrolled loop to keep it a multiple of UR without adding
    any new prelogue/epilogue instructions*/
#ifdef INTERACT
  begin_trans(VET_OPT_TRANS);
#endif
  change_branch(loop, &exitblk, iter, UR, loopType);
  /*remvIncBlk(loop, exitblk);*/

  /*it is possible that loop unrolling produced code where the left and right
    pointers are inconsistent. Fix this now*/
  fix_controlflow();
  
  recalculate_all_preds();

  /*for loops with unsigned comparison/branch this approach to loop unrolling
    might give incorrect results. There is a chance that after adjusting the
    loop control value by the unroll factor, the loop control value may
    overflow. This function inserts additional code before entering the
    unrolled loop to avoid this*/
  handleUnsignedLoops(loop, cleanupHead, exitblk, iter, UR, loopType);

  /*now we can attempt to remove unnecessary updates to the loop variable. 
    For that we must first replace all uses of the loop variable by
    appropirate increments. */
  if(exitblk && !multipleExitLoop){
    analysis = DEPTH_FIRST_SEARCH_ANALYSIS | DOMINATOR_ANALYSIS;
    do_needed_analysis(analysis);
    /*do_loop_analysis(analysis);*/ /*change for Bill's code*/
    remvLoopVarUpdate(loop, epilog, exitblk, loopType);
  }
#ifdef INTERACT
  end_trans();
#endif

  return TRUE;
}

int do_loop_analysis(unsigned int analysis)
{
   int writes;
   struct loopl *loop;
   extern struct loopl *loops;

if (analysis & DEPTH_FIRST_SEARCH_ANALYSIS ||
      analysis & DOMINATOR_ANALYSIS) {
    number_basic_blocks();
    depth_first_search(1);
  }
if (analysis & DOMINATOR_ANALYSIS) {
    build_dominator_tree(1);
    find_dominance_frontiers(1);
  }
 if (analysis & LOOP_ANALYSIS) {
    clear_loops();
    find_loops(1);
    order_loops();
    find_loop_exits();
  }

   if (analysis & LOOP_INVARIANT_ANALYSIS) 
   {
      for(loop=loops ; loop ; loop=loop->next)
      {
         writes = 0;
         loop_invariant_init(loop);
         loop_invariant_analysis(loop, TRUE, TRUE, &writes);
         loop_invariant_analysis(loop, FALSE, TRUE, &writes);
      }
   }


}

  

/*
 * setType3PrelogLinks - set the left/right links of the prelogue loop for
 *                       type 3 loops
 */
void setType3PrelogLinks(struct loopl *loop, struct bblock *preheader)
{
  struct list *rtl;
  
  /*move the jump statement from the original preheader of the loop to this
    new block*/
  rtl = findLastRTL(loop->preheader);
  moveRTL(rtl, preheader, NULL);

  /*change the links of the new preheader block to the same links as the old
    preheader*/
  update_bb_ptrs(MOD_LEFT, preheader, loop->ptr);
  if(loop->preheader->left == loop->ptr){

    /*change the links of the old preheader to point to this new block*/
    update_bb_ptrs(MOD_LEFT, loop->preheader, preheader);
  }
  else{
    update_bb_ptrs(MOD_RIGHT, loop->preheader, preheader);
  }
}


/*
 * setType12PrelogLinks - set the left/right links of the prelogue loop for
 *                        type 1 and type 2 loops
 */
void setType12PrelogLinks(struct loopl *loop, struct bblock *preheader)
{
  struct list *rtl;
  
  /*change the links of the new preheader block to the same links as the old
    preheader*/
  update_bb_ptrs(MOD_LEFT, preheader, loop->ptr);

  /*change the links of the old preheader to point to this new block*/
  if(loop->preheader->left == loop->ptr)
    update_bb_ptrs(MOD_LEFT, loop->preheader, preheader);
  else
    update_bb_ptrs(MOD_RIGHT, loop->preheader, preheader);
}


/*
 * findLastRTL - find the last valid RTL in this block
 */
struct list *findLastRTL(struct bblock *blk)
{
  struct list *rtl;

  for(rtl=blk->rtlend ; rtl ; rtl=rtl->prev)
    if(cnts(rtl))
      break;

  return(rtl);
}

/*
 * insPrelogue - insert code before the loop that actually checks if the
 *               unrolled loop is to be entered
 */
struct bblock *prelogue(struct loopl *loop, int iter, int UR)
{
  char relop;
  char *s, rtlval[MAXSTRING], reg[REGSTRING+1], bounds[REGSTRING+1], str[MAXSTRING];
  int i, maxint, prelogue, rcnt, inc, value;
  struct bblock *cblk, *header, *pheader, *new, *new2, *exitblk;
  struct list *rtl, *com_rtl, *cntr_rtl;
  struct blist *bl;

  /*find the loop preheader*/
  pheader = loop->preheader = (struct bblock *)preheader(loop->ptr);

  /*
   * For compile time known bounds we change the comparison check to check for
    a number N', such that N'<N and N'%UR=0.
    So, for(i=0 ; i<N ; i+=UR) becomes
        for(i=0 ; i<N'; i+=UR).
    The number of iterations left behind i.e. (N % UR) are added during
    cleanup code.

   * For execution time bounds, when the loop variable 'i' is not used inside
    the loop....
    The unrolled loop will be entered only if (N>=UR).
    The unrolled code will look something like this:

    N = N - (UR-1);
    for(i=0 ; i<N ; i+=UR){
       ...
       <UR iterations of the loop>
       ...
    }
    N = N + (UR-1);

    The loop condition (i != N) will have to be changed to
    (i < N), for i-increasing loops, and
    (i > N), for i-decreasing loops
  */

  /*always construct a new basic block, to keep the control flow consistent
    for compile and execution time loops. If empty, this block will be
    removed later*/
  new = (struct bblock *)NEW_basic_block(NULL);
  label_basic_block(new);

  /*for execution time loops, add the instruction N = N - (UR-1);
    to this basic block*/
  if(iter < 0){

    /*if N is changed inside loops, then we need to allocate a new register*/
    if(allocRegister){

      /*get a new register*/
      if(!cancolor(regtype(loop->control->item), loop->regs, reg, FALSE))
	return NULL;

      /*set the loop control variable to this new register*/
      loopCompVal(reg);
    }

    /*if the loop bounds check variable is a constant then we do not need to
      add an instruction as we can change the bounds check condition itself*/
    strcpy(bounds, loopCompVal(0));
    if(!isInteger(loop->iterations)){
    
      /*construct an instruction to subtract (UR-1) from the loop bound
	variable*/
      inc = loopInc(0);
      UR = UR*inc;
      if(UR < 0)
	UR = (-UR);
      strcpy(bounds, loopCompVal(0));

      relop = loopRelop(0);
      switch(relop){
	
      case CMP_LT:
      case CMP_ULT:
	sprintf(rtlval, "%s=%s-%d;", bounds, loop->iterations, UR-1);
	break;
      case CMP_LE:
      case CMP_ULE:
	sprintf(rtlval, "%s=%s-%d;", bounds, loop->iterations, UR-1);
	break;
      case CMP_GT:
      case CMP_UGT:
	sprintf(rtlval, "%s=%s+%d;", bounds, loop->iterations, UR-1);
	break;
      case CMP_GE:
      case CMP_UGE:
	sprintf(rtlval, "%s=%s+%d;", bounds, loop->iterations, UR-1);
	break;
      case CMP_NE:
	if(inc > 0)
	  sprintf(rtlval, "%s=%s-%d;", bounds, loop->iterations, UR-1);
	else
	  sprintf(rtlval, "%s=%s+%d;", bounds, loop->iterations, UR-1);
	break;
      }

      /*add this RTL to the new block*/
      rtl = add_rtl_with_type(new, NULL, rtlval, NULL, RTLINE);
      
      /*verify that the instruction created is valid*/
      if(find(rtl, rtlval, NULL)){
	fprintf(stderr, "Invalid RTL in prelogue: %s\n", rtlval);
	return NULL;
      }
    }
    else{
      /*for unsigned loops we need to verify that subtracting/adding UR-1
	from the loop control variable does not cause overflow/underflow*/
      exitblk = findMainLoopExit(loop);
      if(relop = unsignedCompare(exitblk)){
	value = atoi(bounds);

	/*look for overflows*/
	if(relop==CMP_UGT || relop==CMP_UGE){

	  maxint = 0;
	  for(i=0 ; i<sizeof(int) ; i++){
	    maxint <<= 8;
	    maxint |= 0xFF;
	  }
	  if(value > maxint-(UR-1)){
	    fprintf(stderr, "Unrolling not performed as unsigned compare might overflow during algorithm\n");
	    return NULL;
	  }
	}
	/*look for underflow*/
	else{
	  if(value < UR-1){
	    fprintf(stderr, "Unrolling not performed as unsigned compare might underflow during unroll algorithm\n");
	    return NULL;
	  }
	}
      }
    }
  }

  return(new);
}


/*
 * setTyple13EpilogLinks - set the links for the left and right pointers after
 *                         inserting the epilogue block for type 1 & 3 loops
 */
void setType13EpilogLinks(struct loopl *loop, struct bblock *epiblk,
			  struct bblock *exitblk)
{
  if(exitblk->left == epiblk->down)
    update_bb_ptrs(MOD_LEFT, exitblk, epiblk);
  else
    update_bb_ptrs(MOD_RIGHT, exitblk, epiblk);

  update_bb_ptrs(MOD_LEFT, epiblk, epiblk->down);
}

/*
 * setTyple2EpilogLinks - set the links for the left and right pointers after
 *                        inserting the epilogue block for type 2 loops
 */
void setType2EpilogLinks(struct loopl *loop, struct bblock *epiblk,
			 struct bblock *exitblk)
{
  struct bblock *cblk, *loop_exit;
  struct blist *bl, *nbl;
  struct list *jmpRTL;

  /*set the links for the new epilogue block*/
  if(!block_in_loop(loop, exitblk->right)){

    update_bb_ptrs(MOD_LEFT, epiblk, exitblk->right);
    loop_exit = exitblk->right;
  }
  else{
    update_bb_ptrs(MOD_LEFT, epiblk, exitblk->left);
    loop_exit = exitblk->left;
  }

  /*change the links exiting out of the loop to point to the new epilog*/
  for(bl=loop_exit->preds ; bl ; bl=nbl){
    nbl = bl->next;
    cblk = bl->ptr;
    if(block_in_loop(loop, cblk)){
      
      if(cblk->left == loop_exit)
	update_bb_ptrs(MOD_LEFT, cblk, epiblk);
      else
	update_bb_ptrs(MOD_RIGHT, cblk, epiblk);
    }
  }
}


 /*
 * epilogue - add a basic block which contains an instruction to add the
 *            appropriate amount to the loop control variable, as subtracted
 *            in the prelogue
 */
struct bblock *epilogue(struct loopl *loop, int iter, int UR)
{
  int lcntr, inc;
  char relop;
  char rtlval[MAXSTRING], reg1[MAXSTRING], reg2[MAXSTRING], label[MAXSTRING];
  struct bblock *new, *exitblk;
  struct bblock *cntrRTL, *rtl;
  
  /*create a new basic block for the epilogue. We create a new basic block
    for all types of loops to keep the control flow consistent*/
  new = (struct bblock *)NEW_basic_block(NULL);
  label_basic_block(new);

  /*for execution time bound loops*/
  if(iter < 0 && !allocRegister){

    if(!isInteger(loop->iterations)){
      
      /*increment the loop bound variable*/
      inc = loopInc(0);
      UR = UR * inc;
      if(UR < 0)
	UR = (-UR);
      relop = loopRelop(NULL);
      switch(relop){
	
      case CMP_LT:
      case CMP_ULT:
	sprintf(rtlval, "%s=%s+%d;", loop->iterations, loop->iterations, UR-1);
	break;
      case CMP_LE:
      case CMP_ULE:
	sprintf(rtlval, "%s=%s+%d;", loop->iterations, loop->iterations, UR-1);
	break;
      case CMP_GT:
      case CMP_UGT:
	sprintf(rtlval, "%s=%s-%d;", loop->iterations, loop->iterations, UR-1);
	break;
      case CMP_GE:
      case CMP_UGE:
	sprintf(rtlval, "%s=%s-%d;", loop->iterations, loop->iterations, UR-1);
	break;
      case CMP_NE:
	if(inc > 0)
	  sprintf(rtlval, "%s=%s+%d;", loop->iterations, loop->iterations, UR-1);
	else
	  sprintf(rtlval, "%s=%s-%d;", loop->iterations, loop->iterations, UR-1);
	break;
      }
      add_rtl_with_type(new, NULL, rtlval, NULL, RTLINE);
    }
  }
  
  return(new);
}


/* findLoopTopBottom - if the loop is thought of as a string of blocks in the
 *                     control flow graph of the function, then locate the
 *                     ends of the string
 */
void findLoopTopBottom(struct loopl *loop, struct bblock **btop, struct bblock **bbottom)
{
  int i, bcnt;
  struct bblock *blk;
  struct blist *bl;
  extern struct bblock *top;

  if(loop->preheader && loop->preheader->down &&
     block_in_loop(loop, loop->preheader->down))
    blk = loop->preheader;
  else
    blk = top;

  for(bcnt=0, blk=blk->down; blk ; blk=blk->down){

    if(block_in_loop(loop, blk)){

      if(!(*btop))
	*btop = blk;

      (*bbottom) = blk;
      bcnt++;
      if(bcnt == loop->blockCount)
	break;
    }
  }
}


/*
 * insertBlk - insert 'blk' in the control flow graph before 'nextblk'
 */
void insertBlk(struct bblock *new, struct bblock *nextblk)
{
  /*insert the 'new' block before 'nextblk'*/
  update_bb_ptrs(MOD_DOWN, new, nextblk);
  new->up = nextblk->up;
  update_bb_ptrs(MOD_DOWN, nextblk->up, new);
  nextblk->up = new;
}


/*
 * setCleanup2Links - set the left/right links for the new loop blocks added
 */
void setCleanup2Links(struct loopl *loop, struct bblock *dup_map[MAX_BLOCKS],
		      struct bblock *pdown, struct bblock *blkup,
		      struct bblock *blkdown)
{
  struct blist *bl, *nbl;
  struct list *rtl;
  struct bblock *oblk, *nblk, *tblk, *left, *right, *jumpblk;

  /*Set the left and right pointers for all duplicated loop blocks. Note that
    the branch block is not duplicated, so it has no entry in dup_map*/
  for(bl=loop->blocks ; bl ; bl=bl->next){

    /*set internal links between duplicated blocks*/
    oblk = bl->ptr;
    if(nblk = dup_map[oblk->block_id]){
      
      left = (tblk=dup_map[oblk->left->block_id]) ? tblk : oblk->left;
      update_bb_ptrs(MOD_LEFT, nblk, left);
      
      if(oblk->right){
	
	right = (tblk=dup_map[oblk->right->block_id]) ? tblk : oblk->right;
	update_bb_ptrs(MOD_RIGHT, nblk, right);
      }
    }
  }

  /*the epilog block should fall through to the cleanup code*/
  update_bb_ptrs(MOD_LEFT, blkup, blkup->down);

  /*all links leaving internal loop blocks for the epilog block should 
    now point to blkdown*/
  for(bl=blkup->preds ; bl ; bl=nbl){

    nbl = bl->next;
    oblk = bl->ptr;
    if(nblk = dup_map[oblk->block_id]){
    
      if(oblk->left == blkup)
	update_bb_ptrs(MOD_LEFT, nblk, blkdown);
      else
	update_bb_ptrs(MOD_RIGHT, nblk, blkdown);
    }
  }
}




/*
 * setCleanup13Links - set the left/right links for the new loop blocks added
 */
void setCleanup13Links(struct loopl *loop, struct bblock *dup_map[MAX_BLOCKS],
		       struct bblock *pdown, struct bblock *blkup,
		       struct bblock *blkdown)
{
  struct blist *bl, *nbl;
  struct list *rtl;
  struct bblock *oblk, *nblk, *tblk, *left, *right;

  /*Set the left and right pointers for all duplicated loop blocks. Note that
    the branch block is not duplicated, so it has no entry in dup_map*/
  for(bl=loop->blocks ; bl ; bl=bl->next){

    /*set internal links between duplicated blocks*/
    oblk = bl->ptr;
    if(nblk = dup_map[oblk->block_id]){

      if(left=dup_map[oblk->left->block_id])
	update_bb_ptrs(MOD_LEFT, nblk, left);
      else
	update_bb_ptrs(MOD_LEFT, nblk, oblk->left);

      if(oblk->right){
	if(right=dup_map[oblk->right->block_id])
	  update_bb_ptrs(MOD_RIGHT, nblk, right);
	else
	  update_bb_ptrs(MOD_RIGHT, nblk, oblk->right);
      }
    }
  }

  /*join blkup with the starting block in this code*/
  update_bb_ptrs(MOD_LEFT, blkup, blkup->down);

  /*all links leaving internal loop blocks for the compare/branch block,
    now should point to blkdown*/
  for(bl=pdown->preds ; bl ; bl=nbl){

    nbl = bl->next;
    oblk = bl->ptr;
    if(nblk = dup_map[oblk->block_id]){
    
      if(oblk->left == pdown)
	update_bb_ptrs(MOD_LEFT, nblk, blkdown);
      else
	update_bb_ptrs(MOD_RIGHT, nblk, blkdown);
    }
  }
}


/*
 * setLoop13Links - set the left/right links for the new loop blocks added
 */
void setLoop13Links(struct loopl *loop, struct bblock *dup_map[MAX_BLOCKS],
	    struct bblock *pdown, struct bblock *blkup, struct bblock *blkdown)
{
  struct blist *bl, *nbl;
  struct list *rtl;
  struct bblock *oblk, *nblk, *tblk, *left, *right;

  /*Set the left and right pointers for all duplicated loop blocks. Note that
    the branch block is not duplicated, so it has no entry in dup_map*/
  for(bl=loop->blocks ; bl ; bl=bl->next){

    /*set internal links between duplicated blocks*/
    oblk = bl->ptr;
    if(nblk = dup_map[oblk->block_id]){

      if(left=dup_map[oblk->left->block_id])
	update_bb_ptrs(MOD_LEFT, nblk, left);
      else
	update_bb_ptrs(MOD_LEFT, nblk, oblk->left);
      
      if(oblk->right){
	if(right=dup_map[oblk->right->block_id])
	  update_bb_ptrs(MOD_RIGHT, nblk, right);
	else
	  update_bb_ptrs(MOD_RIGHT, nblk, oblk->right);
      }
    }
  }

  /*all blocks previously linking to the branch block should now point to the
    start block in the unrolled code. If the branch block is also the header
    then it might also be reached by blocks outside the loop. So, we check
    that only blocks belonging to this loop have their links changed*/
  for(bl=pdown->preds ; bl ; bl=nbl){

    nbl = bl->next;
    oblk = bl->ptr;
    if(block_in_loop(loop, oblk)){
      
      nblk = dup_map[oblk->block_id];
      if(oblk->left == pdown)
	update_bb_ptrs(MOD_LEFT, oblk, blkdown);
      else
	update_bb_ptrs(MOD_RIGHT, oblk, blkdown);
    }
  }
}


/*
 * removeBlk - delete a block from the control flow and update the control
 *             flow gaph
 */
void removeBlk(struct bblock *blk)
{
  struct blist *bl, *nbl;
  struct bblock *pblk, *nblk;

  if(!blk->left){
    fprintf(stderr, "Block being deleted does not have left pointer\n");
    nblk = blk->down;
  }
  else
    nblk = blk->left;
  
  /*link all predecessors of the current blk to the following block*/
  for(bl=blk->preds ; bl ; bl=nbl){

    nbl = bl->next;
    pblk = bl->ptr;
    if(pblk->left == blk)
      update_bb_ptrs(MOD_LEFT, pblk, nblk);
    else if(pblk->right == blk)
      update_bb_ptrs(MOD_RIGHT, pblk, nblk);
  }

  /*remove the block from the control flow, assuming that the block being
    deleted is not the first or last block in the function*/
  blk->down->up = blk->up;
  update_bb_ptrs(MOD_DOWN, blk->up, blk->down);

  /*release the resources held by this block*/
  remv_basic_block(blk);

  /*remove this block from the predecessor list of its sucessors*/
  if(blk->left)
    remv_predecessor(blk, blk->left);
  if(blk->right)
    remv_predecessor(blk, blk->right);  
}


/* insCleanup - if the loop unroll factor is not an exact multiple of the
 *              loop, then insert cleanup code after the main body of the
 *              unrolled loop. Returns the cleanup header block
 */
struct bblock *insCleanup(struct loopl *loop, int iter, int UR, struct bblock *epilog,
		struct bblock *exitblk, int loopType)
{
  char rtlval[MAXSTRING];
  int i, cl_iter;
  struct blist *bl;
  struct bblock *cntrblk, *jumpblk, *topblk, *oblk, *tblk;
  struct bblock *bdown, *bup, *new, *bottom, *nextblk, *dup_map[MAX_BLOCKS];
  struct list *rtl, *nrtl, *cmpRTL, *jmpRTL;

  /*the dup_map array will hold a pointer to the duplicated block
    in the corresponding array location i.e.
    dup_map[orig_blk->block_id] = (struct bblock *)dup_block;
  */

  /*determine is we need to insert cleanup code. Cleanup is always needed for
    execution time bound loops, but might be un-needed for compile time bound
    loops if the unroll factor is an exact multiple of the loop count*/
  if(iter>0 && (cl_iter=iter%UR)==0)
    return NULL;
  
  /*initialize dup_map*/
  for(i=0 ; i<MAX_BLOCKS ; i++)
    dup_map[i] = (struct bblock *)NULL;

  /*the cleanup code is placed after the epilogue/exitblk block, immediately
    outside the main loop body*/
  /*Assuming that the epilogue block is always present, even if empty, might
    simplify things*/
  /*if(epilog)*/
  bdown = epilog->down;
  /*else
    bdown = exitblk->down;*/
  bup = epilog;
  jumpblk = epilog->up;

  /*insert cleanup code for COMPILE time bound loops*/
  if(iter > 0 && cl_iter <= MAX_CLEANUP_LENGTH){

    /*duplicate the loop body cl_iter number of times, and place it before
      after the  epilogue block*/
    for(i=0 ; i<cl_iter ; i++){
      
      /*duplicate loop blocks*/
      for(bl=loop->blocks ; bl ; bl=bl->next){

	oblk = bl->ptr;
	if(oblk != exitblk && !(loopType==2 && oblk==jumpblk)){
	  
#ifdef INTERACT
	  begin_trans(VET_OPT_TRANS);
#endif    
	  new = dup_map[oblk->block_id] = copy_block(oblk);
	  
	  /*insert the new block in the control flow before 'nextblk'*/
	  for(tblk=oblk->down ; tblk && tblk!=lbottom->down ; tblk=tblk->down){
	    if(dup_map[tblk->block_id])
	      break;
	  }
	  if(tblk==NULL || !(nextblk=dup_map[tblk->block_id]))
	    nextblk = bdown;
	  insertBlk(new, nextblk);
      
#ifdef INTERACT
	  end_trans();
#endif
	}
      }
      
#ifdef INTERACT
      begin_trans(VET_OPT_TRANS);
#endif  
      /*set the left/right links for all the new blocks, remove jumps to
	exitblk*/
      if(loopType == 2)
	setCleanup2Links(loop, dup_map, exitblk, bup, bdown);
      else
	setCleanup13Links(loop, dup_map, exitblk, bup, bdown);

#ifdef INTERACT
      end_trans();
#endif  
      
      /*re-initiate the dup_map structure*/
      for(bl=loop->blocks ; bl ; bl=bl->next)
	dup_map[bl->ptr->block_id] = (struct bblock *)NULL;
      bdown = epilog->down;
    }
  }
  /*for EXECUTION time bounds*/
  else{

    /*identify block before which code would be place. Assuming there is
      always a epilog block makes this easy. The epilog block would anyway
      always be present for this case*/
    bup = epilog;
    bdown = epilog->down;
  
    /*duplicate all loop blocks as they are, including the branch blocks*/
    for(bl=loop->blocks ; bl ; bl=bl->next){

      oblk = bl->ptr;
#ifdef INTERACT
      begin_trans(VET_OPT_TRANS);
#endif    
      new = dup_map[oblk->block_id] = copy_block(oblk);
	  
      /*insert the new block in the control flow before 'nextblk'*/
      for(tblk=oblk->down ; tblk && tblk!=lbottom->down ; tblk=tblk->down){
	if(dup_map[tblk->block_id])
	  break;
      }
      if(tblk==NULL || !(nextblk=dup_map[tblk->block_id]))
	nextblk = bdown;
      insertBlk(new, nextblk);
      
#ifdef INTERACT
      end_trans();
#endif
    }

#ifdef INTERACT
      begin_trans(VET_OPT_TRANS);
#endif  
      /*set the left/right links for all the new blocks, do not remove
	jumps to exitblk*/
      if(loopType == 2){

	setCleanup2Links(loop, dup_map, epilog, bup, bdown);
      }
      /*for looptypes 1 and 3*/ 
      else{
	
	setCleanup13Links(loop, dup_map, epilog, bup, bdown);

	/*For execution time loops, we need to check if the clean-up code
	  should be entered at run-time. This depends on the loop counter which
	  is unknown at compile time. So, in the epilogue block we put an
	  unconditional jump to the exitblk of the duplicated loop, which will
	  check for this*/
	if(epilog){
	  
	  cmpRTL = (struct list *)findCompareRtl(exitblk);
	  for(rtl=cmpRTL ; rtl ; rtl=nrtl){
	    nrtl = rtl->next;
	    if(cnts(rtl))
	      add_rtl_with_type(epilog, NULL, rtl->val, NULL, RTLINE);
	  }
	  label_basic_block(bdown);
	  reverse(epilog->rtlend, bdown->label);
	  
	  /*update epilog pointers*/
	  update_bb_ptrs(MOD_RIGHT, epilog, bdown);
	}
      }
#ifdef INTERACT
      end_trans();
#endif  
  }

  return(dup_map[loop->ptr->block_id]);
}

/*
 * change_branch - change the branch instruction in block 'exitblk' of loop
 *                 'loop' so that N<iter and N%UR=0
 */
void change_branch(struct loopl *loop, struct bblock **pexitblk,
		             int iter, int UR, int loopType)
{
  char *c, relop, str[MAXSTRING], rem[MAXSTRING], coper;
  char rtlval[MAXSTRING], reg1[REGSTRING+1], reg2[REGSTRING+1], label[MAXSTRING];
  int i, lcntr, niter, init, inc, value, tvalue, type;
  struct blist *bl, *nbl;
  struct bblock *cblk, *prelogue, *epilogue, *exitblk;
  struct list *cntrRTL, *compareRTL, *brRTL;

  exitblk = *pexitblk;
  init = atoi(initialLoopVal(0));
  inc = loopInc(0);
  value = atoi(loopCompVal(0));
  relop = loopRelop(0);
      
  /*for compile time known bounds*/
  if(iter > 0){
    
    /*determine the new loop iteration factor*/
    niter = iter / UR;
  
    /*find the comparison value to use in the unrolled loop control
      instruction, given the new iteration count, the loop init value, and the
      loop increment*/
    if(niter>1){

      tvalue = loopTermination(init, inc*UR, relop, value, niter, loopType);
      
      /*change the branch exit RTL*/
      getCompareComp(exitblk, reg1, &relop, &coper, reg2, rem, label);
      sprintf(str, "%d", tvalue);
      setCompare(exitblk, reg1, relop, coper, str, rem, label);
    }
    /*if the unrolled loop only iterates once, then it makes sense to remove
      the compare/branch block altogether*/
    else{

      /*all preds of the exitblk now point to the epilog*/
      for(bl=exitblk->preds ; bl ; bl=nbl){

	nbl = bl->next;
	cblk = bl->ptr;
	if(cblk->left==exitblk)
	  update_bb_ptrs(MOD_LEFT, cblk, exitblk->down);
	else
	  update_bb_ptrs(MOD_RIGHT, cblk, exitblk->down);
      }

      /*remove the basic block*/
      update_bb_ptrs(MOD_DOWN, exitblk->up, exitblk->down);
      exitblk->down->up = exitblk->up;
      remv_basic_block(exitblk);
      /*remove this block from the predecessor list of its sucessors*/
      if(exitblk->left)
	remv_predecessor(exitblk, exitblk->left);
      if(exitblk->right)
	remv_predecessor(exitblk, exitblk->right);
   
      *pexitblk = NULL;
    }
  }
  
  /*for execute time known bounds*/
  else{

    if(!allocRegister){

      if(!isInteger(loop->iterations)){

	/*if the branch condition is a REL_NE, then for the unrolled loop we
	  change if to either a REL_GT or REL_LT*/
	getCompareComp(exitblk, reg1, &relop, &coper, reg2, rem, label);

	if(relop==CMP_NE && inc>0){
	  if(unsignedCompare(exitblk))
	    setCompare(exitblk, reg1, CMP_ULT, coper, reg2, rem, label);
	  else	    
	    setCompare(exitblk, reg1, CMP_LT, coper, reg2, rem, label);
	}
	else if(relop==CMP_NE && inc<0){
	  if(unsignedCompare(exitblk))
	    setCompare(exitblk, reg1, CMP_UGT, coper, reg2, rem, label);
	  else
	    setCompare(exitblk, reg1, CMP_GT, coper, reg2, rem, label);
	}
	else
	  setCompare(exitblk, reg1, relop, coper, reg2, rem, label);
      }
      else{

	/*change the constant loop bound variable within the branch statement
	  itself*/
	lcntr = atoi(loopCompVal(0));
	UR = UR * inc;
	if(UR < 0)
	  UR = -(UR);
	relop = loopRelop(0);
	switch(relop){

	case CMP_LT:
	case CMP_ULT:
	  lcntr = lcntr - (UR - 1);
	  break;
	case CMP_LE:
	case CMP_ULE:
	  lcntr = lcntr - (UR - 1);
	  break;
	case CMP_GT:
	case CMP_UGT:
	  lcntr = lcntr + (UR - 1);
	  break;
	case CMP_GE:
	case CMP_UGE:
	  lcntr = lcntr + (UR - 1);
	  break;
	case CMP_NE:
	  if(inc > 0)
	    lcntr = lcntr - (UR - 1);
	  else
	    lcntr = lcntr + (UR - 1);
	  break;
	}
	
	sprintf(rtlval, "%d", lcntr);
	getCompareComp(exitblk, reg1, &relop, &coper, reg2, rem, label);
	
	if(relop=='!' && inc>0){
	  if(unsignedCompare(exitblk))
	    setCompare(exitblk, reg1, CMP_ULT, coper, rtlval, rem, label);
	  else	    
	    setCompare(exitblk, reg1, CMP_LT, coper, rtlval, rem, label);
	}
	else if(relop=='!' && inc<0){
	  if(unsignedCompare(exitblk))
	    setCompare(exitblk, reg1, CMP_UGT, coper, rtlval, rem, label);
	  else
	    setCompare(exitblk, reg1, CMP_GT, coper, rtlval, rem, label);
	}
	else
	  setCompare(exitblk, reg1, relop, coper, rtlval, rem, label);
      }
    }
    
    /*for execute time loop bounds when N is used inside the loop, we have
      allocated a new register*/
    else{
      
      getCompareComp(exitblk, reg1, &relop, &coper, reg2, rem, label);
      strcpy(str, loopCompVal(0));
      setCompare(exitblk, reg1, relop, coper, str, rem, label);
    }

    /*for type 1 loops, the condition is checked after the first iteration.
      After unrolling, since we now have more iterations inside the loop, we
      need the condition checking outside the loop. We can either duplicate
      the comparison and branch in the prologue code, or jump to the exitblk
      in the prelogue*/
    if(loopType == 1){
      prelogue = loop->preheader;
      epilogue = exitblk->down;
      compareRTL = findCompareRtl(exitblk);
      brRTL = findJumpRtl(exitblk);

      add_rtl_with_type(prelogue, NULL, compareRTL->val, NULL, RTLINE);
      add_rtl_with_type(prelogue, NULL, brRTL->val, NULL, RTLINE);

      reverse(prelogue->rtlend, epilogue->label);
      update_bb_ptrs(MOD_RIGHT, prelogue, epilogue);
    }
  }
}


/*We currently do not handle this for all cases. Prior to loop unrolling we
  try to change such an increment to before the compare, so that there
  remains no instruction between the compare/branch*/
/*
 * remvIncBlk - when there is an instruction between the compare and the
 *              branch we are forced to create a new block with the sole loop
 *              update instruction in that block so that the instruction is
 *              duplicated for each unrolled path even though the branch
 *              block is not duplicated. This leaves us with one additional
 *              instance of that block, which needs to be removed at this
 *              point
 */
void remvIncBlk(struct loopl *loop, struct bblock *exitblk)
{
  struct bblock *cblk, *rblk;
  struct blist *bl, *nbl;
  int RTLCnt;

  if(!exitblk)
    return;

  RTLCnt = rtlCount(exitblk);
  if(RTLCnt > 2){

    /*remove the additional block*/
    rblk = exitblk->up;

    /*update pointers*/
    for(bl=rblk->preds ; bl ; bl=nbl){

      nbl = bl->next;
      cblk = bl->ptr;
      if(cblk->left == rblk)
	update_bb_ptrs(MOD_LEFT, cblk, exitblk);
      else
	update_bb_ptrs(MOD_RIGHT, cblk, exitblk);
    }
    update_bb_ptrs(MOD_DOWN, rblk->up, exitblk);
    exitblk->up = rblk->up;
    remv_basic_block(rblk);
    
   /*remove this block from the predecessor list of its sucessors*/
   if(rblk->left)
     remv_predecessor(rblk, rblk->left);
   if(rblk->right)
     remv_predecessor(rblk, rblk->right);   
  } 
}


/*
 *  handleUnsignedLoops - insert additional code to handle unsigned loops.
 */
void handleUnsignedLoops(struct loopl *loop, struct bblock *clHead,
			 struct bblock *exitblk, int iter, int UR, int loopType)
{
  int i, value, inc;
  char relop, cmpVal[MAXSTRING], rtlval[MAXSTRING], str[MAXSTRING];
  struct bblock *cblk, *new, *prelog;
  struct list *rtl, *cmpRTL, *brRTL;
  struct blist *bl, *nbl;

  if(allocRegister)
    strcpy(cmpVal, loopCompVal(0));
  else
    strcpy(cmpVal, loop->iterations);
  
  /*if the iteration count is known, we should be safe*/
  if(iter<0 && (relop=unsignedCompare(exitblk)) && !isInteger(cmpVal)){

    /*construct new basic block*/
    new = (struct bblock *)NEW_basic_block(NULL);
    label_basic_block(new);
    relop = loopRelop(NULL);
    inc = loopInc(0);
    UR = UR * inc;
    if(UR < 0)
      UR = -UR;
	
    if(relop==CMP_UGT || relop==CMP_UGE){

      /*for this case loop unrolling adds UR-1 to the loop control variable in
	the prelogue. Here we are required to verify that the value of the
	control variable goes not overflow after the add, i.e. in the check
	(i > n) if 'i' is less than n, then after the add the overflow might
	cause this to be not true anymore*/
      value = 0;
      for(i=0 ; i<sizeof(int) ; i++){
	value <<= 8;
	value |= 0xFF;
      }

      /*(n+UR-1) should be less than overflow, i.e. n should be greater than
	value - UR*/
      value = value - UR;
      sprintf(str, "%d", value);

      /*construct the compare instruction, if(n > value)*/
      consCompareIns(rtlval, cmpVal, str);
      add_rtl_with_type(new, NULL, rtlval, NULL, RTLINE);

      /*construct the branch for if(n > value) jmp to the cleanup code*/
      consBranchIns(rtlval, CMP_GT, clHead->label);
      add_rtl_with_type(new, NULL, rtlval, NULL, RTLINE);
    }
    else if(relop==CMP_ULT || relop==CMP_ULE){

      /*here we subtract UR-1 from the control variable 'n' in (i<n). If the
	subtract results in an underflow then the value of n after the
	underflow will be more than the present value of n. This will cause
	the loop to iterate more than the required number of times. So, check
	for underflow here*/

      /*construct the compare value*/
      sprintf(str, "%d", UR-1);

      /*construct the compare instruction (n<UR-1)*/
      consCompareIns(rtlval, cmpVal, str);
      add_rtl_with_type(new, NULL, rtlval, NULL, RTLINE);
      consBranchIns(rtlval, CMP_LT, clHead->label);
      add_rtl_with_type(new, NULL, rtlval, NULL, RTLINE);
    }

    /*insert the block immediately before the prelogue of the loop*/
    prelog = loop->preheader;
    update_bb_ptrs(MOD_DOWN, new, prelog);
    new->up = prelog->up;
    update_bb_ptrs(MOD_DOWN, new->up, new);
    prelog->up = new;

    /*the predecessors of prelog should now point to the new block*/
    for(bl=prelog->preds ; bl ; ){
      cblk = bl->ptr;
      nbl = bl->next;
      
      if(cblk->left == prelog)
	update_bb_ptrs(MOD_LEFT, cblk, new);
      else
	update_bb_ptrs(MOD_RIGHT, cblk, new);
      bl = nbl;
    }

    /*update the left/right pointers of the new block */
    update_bb_ptrs(MOD_LEFT, new, prelog);
    update_bb_ptrs(MOD_RIGHT, new, clHead);
  }
}


/*
 * loopTermination - Find the new value for the loop termination condition
 *                   for the unrolled loop
 */
int loopTermination(int init, int inc, char relop, int value, int iter, int loopType)
{
  if(inc < 0)
    inc = (-inc);
  
  switch (relop) {
    
  case CMP_LE:
  case CMP_ULE:
    return (init + (inc*iter) - 1);
    
  case CMP_GE:
  case CMP_UGE:
    return (init - (inc*iter) +1);
    
  case CMP_LT:
  case CMP_ULT:
    return (init + (inc*iter));
    
  case CMP_GT:
  case CMP_UGT:
    return (init - (inc*iter));

  case CMP_NE:
    if(init > value)
      return (init - (inc*iter));
    else
      return (init + (inc*iter));
  }
}


/*
 * remvLoopVarUpdate - Remove the unnecessary updates to the loop variable
 */
int remvLoopVarUpdate(struct loopl *loop, struct bblock *epilog,
		       struct bblock *exitblk, int loopTpe)
{
  char *rtlin, relop, oper, op1[MAXSTRING], op2[MAXSTRING], set[MAXSTRING];
  char rtlval[MAXSTRING], label[MAXSTRING], rem[MAXSTRING], coper;
  int i, inc, totalInc, isset, rtlcnt, canReplace, type;
  struct llist *ll, *loopVarSets;
  struct bblock *cblk, *Qblk[MAX_QUEUE_BLKS];
  struct list *rtl;
  struct setlist *sptr, *lptr;
  struct RTLStore changedRTL[MAX_CHANGE_RTL];
  int top, bottom;

  /*We traverse the loop in a breadth first search. All sets to the loop
    variable have been verified to be basic induction variable like updates
    and they dominate all blocks of the loop. If there are any other uses of
    the loop variable, we will try to replace them with the appropirate
    updated loop variable value, if possible and check if the resulting RTL
    is valid for the machine. If all uses could be replaced then we can
    safely remove all but one set of the loop variable*/

  /*get the loop variable*/
  getCompareComp(exitblk, op1, &relop, &coper, op2, rem, label);
  if(strcmp(op1, (char*)loopCompVal(0)) == 0)
    lptr = s_lookup(op2);
  else
    lptr = s_lookup(op1);

  /*clear the DONE bit for all blocks*/
  for(cblk=loop->preheader ; cblk ; cblk=cblk->down)
    cblk->status &= ~DONE;
  
  canReplace = TRUE;
  totalInc = 0;
  rtlcnt = 0;
  loopVarSets = (struct llist *)NULL;
  top = bottom = 0;
  addToQueue(Qblk, loop->preheader->down, &bottom);
  while(top != bottom){

    /*get the next block in the queue*/
    cblk = remvFromQueue(Qblk, &top);
    for(rtl=cblk->rtls ; rtl ; rtl=rtl->next){

      isset = FALSE;
      for(rtlin=cnts(rtl) ; rtlin && *rtlin ; rtlin=skiploc(rtlin)){

	if(isRegister(rtlin) && (sptr=s_lookup(rtlin)) && sptr==lptr){

	  /*check if the RTL sets the loop variable*/
	  if(*(rtlin+REGLOCLENGTH) == '='){

	    /*check if this is a basic induction variable*/
	    if(!is_basic_help(rtl, rtlin, &inc)){

	      /*this might still be a non-standard induction variable*/
	      if(!is_basic(rtl, rtlin, &inc)){
		fprintf(stderr, "Set of loop variable not a basic induction variable!!\n");
		return FALSE;
	      }
	      totalInc = inc;
	    }
	    else
	      totalInc += inc;
	    add_to_llist(rtl, &loopVarSets, inc);
	    isset = TRUE;
	  }

	  /*this is a use of the loop variable, which is not also a set*/
	  else if(!isset){

	    /*determine if this RTL can be legally changed by replacing the
	      current loop register with loop register + increment*/
	    sprintf(label, "(%s+%d)", sptr->item, totalInc);
	    replaceReg(rtl->val, sptr->item, label, rtlval);
	    simp(rtlval, FALSE);
	    
	    /*getRTLComp(rtl->val, set, op1, &oper, op2);
	      if(strncmp(op1, sptr->item, REGLOCLENGTH) == 0){
	      sprintf(label, "(%s+%d)", op1, totalInc);
	      consRTL(rtlval, set, label, oper, op2);
	      }
	      else{
	      sprintf(label, "(%s+%d)", op2, totalInc);
	      consRTL(rtlval, set, op1, oper, label);
	      }*/
	    if(!find(rtl, rtlval, rtl->deads)){
	      
	      /*even if this RTL change is legal, we cannot make the change
		immediately unless we check all other RTL uses in this loop*/
	      changedRTL[rtlcnt].rtl = rtl;
	      strcpy(changedRTL[rtlcnt].newval, rtlval);
	      rtlcnt++;
	    }
	    else{
	      canReplace = FALSE;
	      break;
	    }
	  }
	}
      }
      if(!canReplace)
	break;
    }

    if(!canReplace)
      break;
    
    /*add this blocks children to the queue*/
    if(cblk->left && cblk->left!=loop->exits->ptr && cblk->left!=exitblk)
      addToQueue(Qblk, cblk->left, &bottom);
    if(cblk->right && cblk->right!=loop->exits->ptr && cblk->right!=exitblk)
      addToQueue(Qblk, cblk->right, &bottom);
  }
  
  /*if all RTL changes are valid, then go ahead and replace the RTLs. Also
    delete all loop variable updates except the last one*/
  if(canReplace){

    for(i=0 ; i<rtlcnt ; i++){
      changedRTL[i].rtl->val =
	strcpy(CALLOC(strlen(changedRTL[i].newval) + 1), changedRTL[i].newval);
#ifdef INTERACT
      mod_rtl(changedRTL[i].rtl, NULL, changedRTL[i].newval);
#endif
    }

    /*do only if there are more than 2 updates to the variable*/
    if(loopVarSets && loopVarSets->next){
      
      for(ll=loopVarSets ; ll->next ; ll=ll->next)
	remv_rtl(ll->ptr);

      /*if this loop variable update is not a basic induction variable, but a
	non-standard induction variable, then we do not change the RTL, since
	the variable to which it is set would aready have been adjusted in
	each iteration of the unrolled loop*/
      rtlin = cnts(ll->ptr);
      if(is_basic_help(ll->ptr, rtlin, &inc)){
	rtlin = ll->ptr->val;
	getRTLComp(rtlin, set, op1, &oper, op2);
	if(oper=='-' && totalInc<0)
	  totalInc = -(totalInc);
	sprintf(op2, "%d", totalInc);
	consRTL(rtlval, set, op1, oper, op2);
      
	ll->ptr->val = strcpy(CALLOC(strlen(rtlval)+1), rtlval);
#ifdef INTERACT      
	mod_rtl(ll->ptr, NULL, rtlval);
#endif
      }
    }
  }
}


/* replaceReg - replace a register in the base string with the replace
 *              string
 */
void replaceReg(char *rtlval, char *reg, char *rep, char *str)
{
  char *c, *d;
  int i, type, num;

  type = getRegType(reg);
  num = getRegNum(reg);

  i = 0;
  for(c=rtlval ; *c ; c++){
    if(isRegister(c) && type==getRegType(c) && num==getRegNum(c)){
      for(d=rep ; *d ; d++)
	str[i++] = *d;
      c = c + REGLOCLENGTH;
    }
    str[i++] = *c;
  }
  str[i] = '\0';
}


/*
 * getRTLComp - dissect the passed RTL into components and return each
 */
void getRTLComp(char *rtlval, char *set, char *op1, char *oper, char *op2)
{
  char *c;
  int i, j;

  *op1 = *op2 = *set = *oper = '\0';
  
  /*store the set register*/
  for(i=0, c=rtlval ; *c != '=' ; c++, i++)
    set[i] = *c;
  set[i] = '\0';

  c = skiploc(c);
  
  /*get the first operand*/
  if(isRegister(c)){
    strncpy(op1, c, REGLOCLENGTH);
    c = c + REGLOCLENGTH;
  }
  else if(isGlobal(c)){
    strncpy(op1, c, GLOBALLOCLENGTH);
    c = c + GLOBALLOCLENGTH;
  }
  else if(isLocal(c)){
    strncpy(op1, c, LOCALLOCLENGTH);
    c = c + LOCALLOCLENGTH;
  }
  else if(isInteger(c)){
    sprintf(op1, "%d", atoi(c));
    c = is_const(c);
  }

  /*get the operator*/
  if(*c != ';')
    *oper = *c;

  c = skiploc(c);

  /*get the final operand*/
  if(c && *c){
    if(isRegister(c))
      strncpy(op2, c, REGLOCLENGTH);
    else if(isGlobal(c))
      strncpy(op2, c, GLOBALLOCLENGTH);
    else if(isLocal(c))
      strncpy(op2, c, LOCALLOCLENGTH);
    else if(isInteger(c))
      sprintf(op2, "%d", atoi(c));
  }
}


/*
 * consRTL - construct a RTL string from its components
 */
void consRTL(char *rtlval, char *set, char *op1, char oper, char *op2)
{
  if(op2 && *op2)
    sprintf(rtlval, "%s=%s%c%s;", set, op1, oper, op2);
  else
    sprintf(rtlval, "%s=%s;", set, op1);
}


/*
 * addToQueue - add a basic block to the queue
 */
void addToQueue(struct bblock *Qblk[], struct bblock *cblk, int *bottom)
{
  if(cblk && !(cblk->status & DONE)){
    
    if(*bottom == MAX_QUEUE_BLKS)
      *bottom = 0;
    
    Qblk[*bottom] = (struct bblock *)cblk;
    (*bottom)++;
    cblk->status |= DONE;
  }
}


/*
 * remvFromQueue - remove and return a basic block from the queue
 */
struct bblock *remvFromQueue(struct bblock *Qblk[], int *top)
{
  struct bblock *cblk;
  
  if(*top == MAX_QUEUE_BLKS)
    *top = 0;

  cblk = (struct bblock *)Qblk[*top];
  (*top)++;

  return cblk;
}


/*
 * add_to_llist - add a RTL to the end of a list
 */
void add_to_llist(struct list *rtl, struct llist **list, int inc)
{
  struct llist *ll, *ptr;

  ll = (struct llist *)SALLOCTYPE(sizeof(struct llist), LLISTTYPE);
  ll->ptr = rtl;
  ll->next = NULL;
  ll->inc = inc;

  if(*list == NULL)
    *list = ll;
  else{
    for(ptr=*list ; ptr->next ; ptr=ptr->next);
    ptr->next = ll;
  }
}


/* findLoopCntrBlk - find the block which contains the jump outside the
 *                   loop. This function only works if there is a single exit
 *                   block, otherwise it will return the first instance of a
 *                   loop block which has a statement to jump outside the
 *                   loop
 */
struct bblock *findLoopCntrBlk(struct loopl *loop)
{
  struct bblock *cblk, *exitblk;
  struct blist *bl;

  /*assuming this loop has only a single basic block*/
  exitblk = loop->exits->ptr;
  for(bl=loop->blocks ; bl ; bl=bl->next){

    cblk = bl->ptr;
    if(cblk->left==exitblk || cblk->right==exitblk)
      return cblk;
  }    
}


/* createNewJumpBlock - For type 2 loops there is a jump statement at the end
 *                      of the loop which transfers control back to the loop
 *                      header. Isolate this instruction into a block by
 *                      itself and return this block
 */
struct bblock *createNewJumpBlock(struct loopl *loop)
{
  int RTLcnt;
  struct bblock *jmpblk, *new, *cblk;
  struct blist *bl;
  struct list *rtl, *nrtl, *cmpRTL;

  /*find the jump block*/
  jmpblk = NULL;
  for(bl=loop->blocks ; bl ; bl=bl->next){

    cblk = bl->ptr;
    if(cblk->left == loop->ptr && cblk->right == NULL){
      if(jmpblk){
	fprintf(stderr, "More than one blocks transfers control to the loop header for type 2 loops. This structure is not handled.\n");
	return (struct bblock *)NULL;
      }
      jmpblk = cblk;
      break;
    }
  }

  /*if the jump block contains an unconditional jump*/
  if(cblk->left == loop->ptr){
    
    RTLcnt = rtlCount(jmpblk);
    
    /*if the block only contains this one RTL then we need not do anything*/
    if(RTLcnt > 1){
      
      /*find the last rtl*/
      rtl = findLastRTL(jmpblk);
      
      /*create a new basic block and move the jump RTL to this block*/
      new = (struct bblock *)NEW_basic_block(NULL);
      label_basic_block(new);
      moveRTL(rtl, new, NULL);
      
      /*update block pointers*/
      update_bb_ptrs(MOD_LEFT, new, jmpblk->left);
      update_bb_ptrs(MOD_LEFT, jmpblk, new);
      
      /*add this block to the loop blocks*/
      add_loop_block(loop, new);
      jmpblk = new;
    }
  }

  /****** will never happen *****/
  /*if the jumpblk contains a conditional jump to the loop header*/
  else{
    RTLcnt = rtlCount(jmpblk);
    if(RTLcnt > 2){

      /*move all RTLs upto the compare RTL to a new block*/
      cmpRTL = findCompareRtl(jmpblk);
      new = (struct bblock *)NEW_basic_block(NULL);
      label_basic_block(new);
      for(rtl=jmpblk->rtls ; rtl!=cmpRTL ; rtl=nrtl){
	nrtl = rtl->next;
	moveRTL(rtl, new, (struct list*)NULL);
      }

      /*link this block in the control flow*/
      update_bb_ptrs(MOD_DOWN, new, jmpblk);
      new->up = jmpblk->up;
      update_bb_ptrs(MOD_DOWN, new->up, new);
      jmpblk->up = new;

      /*set the left/right pointers*/
      for(bl=jmpblk->preds ; bl ; bl=bl->next){
	cblk = bl->ptr;
	if(cblk->left == jmpblk)
	  update_bb_ptrs(MOD_LEFT, cblk, new);
	else
	  update_bb_ptrs(MOD_RIGHT, cblk, new);
      }
      update_bb_ptrs(MOD_LEFT, new, jmpblk);

      /*add this block to the loop list*/
      add_block_to_loop(loop, new);

      /*if jmpblk still contains >3 RTLs then quit*/
      if(rtlCount(jmpblk) > 2){
	fprintf(stderr, "More than 2 RTLs in jmpblk\n");
	return NULL;
      }
    }
  }

  return(jmpblk);
}


/*
 * createNewBranchBlock - The loop control and branch instructions are
 *                        isolated into a separate basic block. During loop
 *                        unrolling all statements in the loop, except the
 *                        branch and jump statements are duplicated.
 *                        Separating in distinct blocks helps later during
 *                        block duplication
 */
struct bblock *createNewBranchBlock(struct loopl *loop, int loopType)
{
  int RTLcnt, count;
  struct bblock *cblk, *exitblk, *new, *lbranch, *pheader, *jmpblk, *new2;
  struct list *cmpRTL, *jmpRTL, *rtl, *nrtl;
  struct blist *bl, *nbl;

  new = (struct bblock *)NULL;
  
  /*already verified that the loop contains only a single exit*/
  lbranch = exitblk = findMainLoopExit(loop);
  
  /*locate the compare/branch RTLs*/
  cmpRTL = (struct list *)findCompareRtl(exitblk);
  jmpRTL = (struct list *)findJumpRtl(exitblk);
  RTLcnt = rtlCount(exitblk);

  /*if these RTLs are already in a block by themselves, then we don't need to
    do anything*/
  if(exitblk->rtls!=cmpRTL && RTLcnt>2 && cmpRTL && jmpRTL){

    new = (struct bblock*)NEW_basic_block(NULL);
    label_basic_block(new);

    /*move all instructions following the compare into another block*/
    for(rtl=cmpRTL ; rtl ; rtl=nrtl){
      nrtl = rtl->next;
      moveRTL(rtl, new, (struct list *)NULL);
    }

    /*link the new basic block in the control flow*/
    if(exitblk->down){
      exitblk->down->up = new;
      update_bb_ptrs(MOD_DOWN, new, exitblk->down);
    }
    new->up = exitblk;
    update_bb_ptrs(MOD_DOWN, exitblk, new);
    update_bb_ptrs(MOD_LEFT, new, exitblk->left);
    update_bb_ptrs(MOD_RIGHT, new, exitblk->right);

    /*the previous exitblk now just falls through to the new block*/
    update_bb_ptrs(MOD_LEFT, exitblk, new);
    update_bb_ptrs(MOD_RIGHT, exitblk, NULL);
    
    lbranch = new;
    if(loopType != 2){
      lbottom = new;
    
      /*insert the new branch block in the block list of the loop*/
      add_block_to_loop(loop, new);
    }

    /*for type 3 loops the preheader jumps unconditionally to the exitblk.
      For this case we need to copy the additional exitblk RTLs to the
      preheader and change the branch label in the preheader to point to
      point to the new branch block*/
    if(loopType == 3){

      pheader = (struct bblock *)preheader(loop->ptr);
      if(pheader->left != loop->ptr){
	fprintf(stderr, "Case should never happen!!! Exiting...\n");
	exit(0);
      }

      /*copy instructions into preheader*/
      for(rtl=exitblk->rtls ; rtl ; rtl=rtl->next){

	if(cnts(rtl))
	  add_rtl_with_type(pheader, pheader->rtlend, rtl->val, NULL, RTLINE);
      }

      /*change preheader branch label*/
      update_bb_ptrs(MOD_LEFT, pheader, new);
    }
    /*for type 2 loops the new block has been created before the loop. This
      block should also be duplicated within the loop. As a result the jump
      condition should now jump to the block which only contains the control
      and branch*/
    else if(loopType == 2){

      /*find the jump block*/
      jmpblk = NULL;
      for(bl=loop->blocks ; bl ; bl=bl->next){
	cblk = bl->ptr;
	if(cblk->left == loop->ptr || cblk->right==loop->ptr){
	  jmpblk = cblk;
	  break;
	}
      }

      /*copy the instructions into a new block and place it before the jump
	block*/
      new2 = (struct bblock *)copy_block(exitblk);
      update_bb_ptrs(MOD_DOWN, new2, jmpblk);
      new2->up = jmpblk->up;
      update_bb_ptrs(MOD_DOWN, jmpblk->up, new2);
      jmpblk->up = new2;

      /*all predecessors of jmpblk now fall into the new block*/
      for(bl=jmpblk->preds ; bl ; bl=nbl){

	nbl = bl->next;
	cblk = bl->ptr;
	if(cblk->left == jmpblk)
	  update_bb_ptrs(MOD_LEFT, cblk, new2);
	else if(cblk->right == jmpblk)
	  update_bb_ptrs(MOD_RIGHT, cblk, new2);
	else
	  fprintf(stderr, "Preds not correctly set in loop unroll\n");
      }

      /*the new block just falls into jmpblk*/
      update_bb_ptrs(MOD_LEFT, new2, jmpblk);
      new2->right = NULL;

      /*the loop should now jump to the block containing only the compare and
	branch */
      if(jmpblk->left == loop->ptr)
	update_bb_ptrs(MOD_LEFT, jmpblk, new);
      else
	update_bb_ptrs(MOD_RIGHT, jmpblk, new);

      /*insert the new branch block in the block list of the loop*/
      add_block_to_loop(loop, new2);

      /*the loop now has a new header and the old header is no longer a part
	of the loop*/
      loop->ptr = new;
      add_loop_to_block(loop, new);
      for(bl=loop->blocks ; bl ; bl=bl->next){
	if(bl->ptr == exitblk){
	  bl->ptr = new;
	  break;
	}
      }
    }
  }

  /*if the new block contains 3 instructions at this point, then it means
    that there is an instruction between the compare and branch*/
  RTLcnt = rtlCount(new);
  if(RTLcnt > 2){

    /*create a new basic block and copy this instruction into the new block
      as well. This is done so that this instruction will be executed for
      each loop unroll. But this also creates one additional occurance of
      this instruction which would be removed later at the end of unrolling*/
    new = (struct bblock*)NEW_basic_block(NULL);
    label_basic_block(new);
    for(rtl=cmpRTL->next ; !cnts(rtl) ; rtl=rtl->next);
    add_rtl_with_type(new, NULL, rtl->val, NULL, RTLINE);

    /*add this block in the control flow*/
    new->up = lbranch->up;
    update_bb_ptrs(MOD_DOWN, new, lbranch);
    update_bb_ptrs(MOD_DOWN, lbranch->up, new);
    lbranch->up = new;
    update_bb_ptrs(MOD_LEFT, new, lbranch);

    /*update the left/right links*/
    for(bl=lbranch->preds ; bl ; bl=nbl){

      nbl = bl->next;
      cblk = bl->ptr;
      if(block_in_loop(loop, cblk)){
	if(cblk->left == lbranch)
	  update_bb_ptrs(MOD_LEFT, cblk, new);
	else
	  update_bb_ptrs(MOD_RIGHT, cblk, new);
      }
    }
  }

  /*return the final loop exiting block*/
  return(lbranch);
}


/*
 * findCompareRtl - locate and return a comparison RTL in this basic block
 */
struct list *findCompareRtl(struct bblock *blk)
{
  struct list *rtl;

  for(rtl=blk->rtlend ; rtl ; rtl=rtl->prev){

    if(cnts(rtl) && is_compare(rtl->val))
      return rtl;
  }

  return (struct list*)NULL;
}


/*
 * findJumpRtl - locate and a branch/jump instruction in this basic block
 */
struct list *findJumpRtl(struct bblock *blk)
{
  struct list *rtl;

  for(rtl=blk->rtlend ; rtl ; rtl=rtl->prev){

    if(cnts(rtl) && is_pc(rtl->val))
      return rtl;
  }

  return (struct list *)NULL;
}


/*
 * rtlCount - Count the number of RTLs in this basic block
 */
int rtlCount(struct bblock *blk)
{
  int count;
  struct list *rtl;

  count = 0;
  if(blk){
    for(rtl=blk->rtls ; rtl ; rtl=rtl->next){
      if(cnts(rtl))
	count++;
    }
  }
  return(count);
}


/* moveRTL - move an RTL from the present basic block to the 'new' basic
             block. The new RTL is inserted before 'nrtl'
*/
void moveRTL(struct list *rtl, struct bblock *dblk, struct list *nrtl)
{
  struct bblock *sblk;

  sblk = rtl->daddy;
  
  /*change the pointers in the source block to unlink the RTL*/
  if(rtl->prev)
    rtl->prev->next = rtl->next;
  else
    sblk->rtls = rtl->next;

  if(rtl->next)
    rtl->next->prev = rtl->prev;
  else
    sblk->rtlend = rtl->prev;

  /*link the RTLs in the destination block*/
  rtl->next = nrtl;
  if(nrtl){
    rtl->prev = nrtl->prev;
    nrtl->prev = rtl;
  }
  else{
    rtl->prev = dblk->rtlend;
    dblk->rtlend = rtl;
  }
  if(rtl->prev)
    rtl->prev->next = rtl;
  else
    dblk->rtls = rtl;

  /*change the RTLs block association*/
  rtl->daddy = dblk;

#ifdef INTERACT
    move_rtl(rtl, sblk, nrtl, dblk);
#endif    
}


/*
 * getURfactor - determine the Unroll Factor for this loop. The unroll factor
 *               may be specified by the user. This function is just to get
 *               things started
 */
int getURfactor(struct loopl *loop, int unroll, int iter)
{
  int UR;
  
  UR = unroll;
  if( (iter>0 && UR<=iter) || iter<0)
    return UR;
  else if(iter>0 && UR>iter)
    return iter;
  return 1;
}

/*
 * isHeader - check if this block is the header of some loop
 */
int isHeader(struct bblock *blk)
{
  struct loopl *ll;
  extern struct loopl *loops;

  for(ll=loops ; ll ; ll=ll->next){
    if(blk == ll->ptr)
      return TRUE;
  }
  return FALSE;
}


/*
 * add_block_to_loop - insert the block in the looplist
 */
void add_block_to_loop(struct loopl *loop, struct bblock *blk)
{
  struct listl *ll;
  struct blist *bl;
  
  /*insert the new block in the block list of the loop*/
  bl = (struct blist *)SALLOCTYPE(sizeof(struct blist), BLISTTYPE);
  bl->ptr = blk;
  bl->next = loop->blocks;
  loop->blocks = bl;

  /*add the loop to the block looplist*/
  add_loop_to_block(loop, blk);
}


/*
 * update_bb_ptrs - update the left/right pointer of block 'block' to
 *                  'child'. Which pointer is specified by dir
 */
 void update_bb_ptrs(int dir, struct bblock *block, struct bblock *child)
{
  if(dir == MOD_LEFT){
    
    /*remove previous predexessor*/
    if(block->left && block->left!=block->right)
      remv_predecessor(block, block->left);

    /*set ne pointers and change branch label if any*/
    block->left = child;
    if(child != NULL){
      if(block->rtlend && cnts(block->rtlend) && is_ucjmp(block->rtlend->val))
	change_branch_label(block->rtlend, child->label);

      /*update predecessor*/
      add_predecessor(block, child);
    }
    
#ifdef INTERACT
    mod_bb_ptrs(MOD_LEFT, block, child);
#endif  
  }    
  else if(dir == MOD_RIGHT){
    
    /*remove previous predexessor*/
    if(block->right && block->left!=block->right)
      remv_predecessor(block, block->right);

    block->right = child;
    if(child != NULL){
      
      if(block->rtlend && cnts(block->rtlend) && is_rjmp(block->rtlend->val))
	change_branch_label(block->rtlend, child->label);
      add_predecessor(block, child);
    }
    
#ifdef INTERACT
    mod_bb_ptrs(MOD_RIGHT, block, child);
#endif  
  }
  else if(dir == MOD_DOWN){
    block->down = child;
#ifdef INTERACT
    mod_bb_ptrs(MOD_DOWN, block, child);
#endif
  }
}


/*
 * fix_controlflow - fix any inconsistent left and right pointers
 */
void fix_controlflow()
{
  struct bblock *cblk, *tptr;
  extern struct bblock *top;

  for(cblk=top ; cblk ; cblk=cblk->down){

    if(cblk->left && cblk->right &&
       cblk->left!=cblk->down && cblk->right==cblk->down){

      /*reverse the branch condition and switch the left and right pointers*/
      reverse(cblk->rtlend, cblk->left->label);
      tptr = cblk->left;
      update_bb_ptrs(MOD_LEFT, cblk, cblk->right);
      update_bb_ptrs(MOD_RIGHT, cblk, tptr);
    }
  }
}


/*
 * initialLoopVal - If passed an argument, then set the initial value for the
 *                  loop in question, else return the stored value
 */
char *initialLoopVal(char *init)
{
  static char initial[MAXSTRING];

  if(init)
    strncpy(initial, init, strlen(init)+1);

  return (char *)initial;
}


/*
 * loopInc - If passed an argument, store that argument as the loop increment
 *           value, return the stored value
 */
int loopInc(int *inc)
{
  static int increment;
  
  if(inc)
    increment = *inc;

  /*if(increment < 0)
    increment = (-increment);*/
  return increment;
}


/* loopCompVal - If passed a non-Null argument, then set the loop comparison
 *               value for the current loop. Return the stored value. For
 *               loops where the loop bounds variable is changed within
 *               loops, we need to allocate a new register. This function
 *               then stores/returns this new register to be used for bounds
 *               checking within unrolled loops.
 */
char *loopCompVal(char *cval)
{
  static char compVal[MAXSTRING];

  if(cval)
    strncpy(compVal, cval, strlen(cval)+1);

  return (char*)compVal;
}

/*
 * loopRelop - If passed a non-null argument, set the loop relation al
 *             operator for the current loop. Return the stored relational
 *             operator for the current loop
 */
char loopRelop(char *rel)
{
  static int relop;

  if(rel)
    relop = *rel;

  return relop;
}


#if 0
/* invertBranch - we are changing the relative order of the two operands.
 *                This will also require changing the branch direction. Note
 *                that this change is different than reversing the branch.
 *                eg. '>' branch will be changed to '<' and not '<=', which
 *                would be the case when reversing branch in the traditional
 *                sense in some function like reverse branches.
 */
char invertBranch(struct list *rtl, char *label)
{
   register char *a, *b;
   char  new[MAXSTRING], *relop;

   a = rtl->val;
   b = new;
   while ((*b++ = *a++) != 'C' || (*(a - 2) != 'I' && *(a - 2) != 'F')) ;
   relop = b;
   switch (*a++) {

    case ':':
       *b = '!';
       break;

    case '!':
       *b = ':';
       break;

    case '>':
       *b = '<';
       break;

    case '<':
       *b = '>';
       break;

    case '`':
       *b = '\'';
       break;

    case '\'':
       *b = '`';
       break;

    case 'h':
       *b = 'l';
       break;

    case 'l':
       *b = 'h';
       break;

    case 'g':
       *b = 's';
       break;

    case 's':
       *b = 'g';
       break;

    default:
       (void) fprintf(stderr, "vpo (reverse): bad conditional, '%s'\n",
                      rtl->val);
       exit(EXIT_FAILURE);
   }
   b++;
   while ((*b++ = *a++) != ',') ;
   a = label;
   while (*b++ = *a++) ;
   *(b - 1) = ';';
   *b = '\0';
   FREE(rtl->val);
   rtl->val = strcpy(CALLOC(strlen(new) + 1), new);
#ifdef INTERACT
   mod_rtl(rtl, NULL, rtl->val);
#endif

   return(*relop);
}
#endif


/*
 * canUnrollLoop - Check if this loop meets the constraints to allow loop
 *                 unrolling
 */
int canUnrollLoop(struct loopl *loop, int unroll)
{
  int i, ind, loopback, inc, val, nval, count, UR, iteration;
  char relop, iter[MAXSTRING], control[MAXSTRING], exitValue[MAXSTRING];
  char init[MAXSTRING], cval[MAXSTRING], str[MAXSTRING], *rtlin, *chptr;
  char op2[MAXSTRING], reg[REGSTRING], tstr[MAXSTRING];
  struct setlist *cvar, *sptr, *indptr;
  struct loopl *cloop;
  struct blist *bl;
  struct bblock *cblk, *eblk;
  struct list *rtl, *indrtl, *cmpRTL, *crtl, *brtl;

  /*check if this is the innermost loop. Currently we only perform loop
    unrolling on the innermost loop. If any block in this loop is the header
    of some other loop, then this is not the innermost loop*/
  for(bl=loop->blocks->next ; bl ; bl=bl->next){

    if(isHeader(bl->ptr))
      return FALSE;
  }

  /*This function will analyse and locate the basic induction variables in
    the loop. If possible, this function will also calculate the number of
    loop iterations for this loop*/
  init[0] = '\0';
  cval[0] = '\0';
  inc = 0;
  if(!(LoopAnalysesForProfiling(loop, init, &inc, cval, &relop)))
    return FALSE;
  
  initialLoopVal(init);
  loopInc(&inc);
  loopCompVal(cval);
  loopRelop(&relop);
  allocRegister = FALSE;  

  /*Sometimes the compare instruction might be of the form
    IC=r[16]-1:0;r[16]=r[16]-1
    This is a single machine instruction on many architectures where it would
    be converted into a subcc (addcc, etc.). During loop unrolling we need to
    modify the branch instruction. Here this would be most easily achieved by
    seperating the increment and the compare. Instruction selection should be
    able to combine the two after loop unrolling*/
  eblk = findMainLoopExit(loop);
  cmpRTL = findCompareRtl(eblk);
  
  if((rtlin=cnts(cmpRTL)) && (rtlin=(char *)is_compare(rtlin)) &&
     (rtlin=(char*)gtrg(rtlin)) && !(*rtlin=='?' || *rtlin=='u' || *rtlin==':')){

#ifdef INTERACT
    begin_trans(VET_OPT_TRANS);
#endif      

    /*this RTL can only be some instruction like add, sub etc. with a
      constant, as we check for this before getting here. Storethe constant*/
    if(!isInteger(rtlin) || !(val=atoi(rtlin))){
#ifdef INTERACT
      end_trans();
#endif      
      return FALSE;
    }
    
    /*we have the RTL we want. Scan ahead till the end of the first RTL in
      the RTL list*/
    for(;*rtlin!=';' ; rtlin++);
    rtlin++;
    
    /*copy the effect RTL into another string*/
    if(!(rtlin || *rtlin)){
#ifdef INTERACT
      end_trans();
#endif            
      return FALSE;
    }
    
    for(i=0 ; rtlin && *rtlin ; i++, rtlin++)
      str[i] = *rtlin;
    str[i] = '\0';

    /*copy this RTL into a seperate RTL list*/
    rtl = add_rtl_with_type(eblk, cmpRTL, str, NULL, RTLINE);

    /*change the compare RTL to remove the secondary effect. Since we are now
      performing the operation before the actual compare this RTL is a simple
      compare RTL, comparing the register with zero*/

    /*Scan the RTL to find until we after the compare register*/
    if((rtlin=(char *)is_compare(cmpRTL->val)) && (rtlin=(char *)gtrg(rtlin)))
      chptr = rtlin;

    /*verify that the comparison is with zero*/
    while(!(*rtlin==':' || *rtlin=='?' || *rtlin=='u'))
      rtlin++;
    rtlin++;
    if(!isInteger(rtlin) || (atoi(rtlin)!=0)){
#ifdef INTERACT
      end_trans();
#endif      
      return FALSE;
    }
    
    /*construct new RTL having comparison with 0*/
    *chptr = '\0';
    sprintf(cmpRTL->val, "%s?0;", cmpRTL->val);
      
#ifdef INTERACT
    mod_rtl(cmpRTL, NULL, cmpRTL->val);
    end_trans();
#endif
  }
  
  /*check that there is not more that 1 instruction between the compare and
    the branch*/
  crtl = findCompareRtl(eblk);
  brtl = findJumpRtl(eblk);
  count = 0;
  updateAfterCompare = FALSE;
  for(rtl=crtl->next ; rtl!=brtl ; rtl=rtl->next){
    if(cnts(rtl)){
      indrtl = rtl;
      count++;
    }
  }

  /*try to push the compare instruction closer to the branch*/
  if(count > 0)
    count -= push_compare(eblk);
  
  if(count == 1){
    /*check if the variable being set by this instruction is the loop
      variable, i.e. this instruction can only be the basic induction
      instruction updating the loop variable*/
    chptr = indrtl->val;
    if((rtlin=(char*)is_compare(crtl->val)) && (sptr=s_lookup(rtlin)) &&
       (indptr=s_lookup(indrtl->val)) && (is_basic(indrtl, chptr, &ind)) &&
       sptr==indptr && !isRegister(cval)){

      /*currently we handle this case if the comparison is with a constant so
	that we can easily move the induction instruction above the compare*/
      val = atoi(cval);
      nval = val + ind;
      if(unsignedCompare(eblk) && ((nval>=0 && val<0) || (nval<0 && val>=0))){
	fprintf(stderr, "Found an instruction between compare/branch, with unsigned comparison and value overflowing after adjustment. Unrolling skipped...\n");
	return FALSE;
      }
#ifdef INTERACT
      begin_trans(VET_OPT_TRANS);
#endif  
      /*change the comparison RTL*/
      sprintf(op2, "%d", nval);
      consCompareIns(str, sptr->item, op2);

      /*verify that this RTL is a legal RTL on this machine*/
      strcpy(tstr, crtl->val);
      strcpy(crtl->val, str);
      if(find(crtl, str, NULL)){
	fprintf(stderr, "Invalid compare RTL: %s\n", str);
	strcpy(crtl->val, tstr);
	return FALSE;
      }
      
      FREE(crtl->val);
      crtl->val = strcpy(CALLOC(strlen(str) + 1), str);

      /*move the increment above the comparison in the same block*/
      moveRTL(indrtl, eblk, crtl);
#ifdef INTERACT
      mod_rtl(crtl, NULL, crtl->val);
      end_trans();
#endif

      /*update the loop termination condition*/
      sprintf(str, "%d", nval);
      loopCompVal(str);
      updateAfterCompare = TRUE;
    }
    else{
      fprintf(stderr, "Loop cannot be unrolled as compare-update-branch sequence not in the correct format\n");
      return FALSE;
    }
  }
  else if(count > 1){
    bshow(crtl->daddy, 0);
    fprintf(stderr, "More than 1 instruction between compare and branch in exit block\n");
    return FALSE;
  }

  /*For comparisons with constants (like, i<512), we need to check if 
    changing the constant is valid on the particular machine*/
  if(isInteger(cval) && (count=atoi(cval))){
    
    /*fprintf(stderr, "Check RTL\n");
      show(crtl->val);*/
    /*change the constant by UR-1, as will mostly be done by loop 
      unrolling to check if it is legal*/
    if(isInteger(loop->iterations) && isInteger(init))
      iteration = atoi(loop->iterations);
    else
      iteration = -1;
    UR = getURfactor(loop, unroll, iteration);
    count = count - (UR-1);
    
    /*create a new RTL to check*/
    makereg(reg, RTREG, 1);
    sprintf(op2, "%d", count);
    consCompareIns(str, reg, op2);
    /*show(str);
      fprintf(stderr, "UR=%d, count=%d\n", UR, count);*/

    /*change RTL value*/
    strcpy(tstr, crtl->val);
    strcpy(crtl->val, str);
    if(find(crtl, str, NULL)){
      fprintf(stderr, "New RTL illegal: %s\n", str);
      strcpy(crtl->val, tstr);
      return FALSE;
    }
    
    /*reset the RTL*/
    strcpy(crtl->val, tstr);
  }
  

  /*If we are able to determine the number of loop iterations, then loop
    unrolling can be performed. Furthermore, if this value is a constant,
    then the iteration count is known at compile time which can help loop
    unrolling*/
  if(loop->iterations){
    if(isInteger(loop->iterations) && isInteger(init))
      return (atoi(loop->iterations));

    /*here the iteration count is a concatenation of the loop bounds register
      and the loop initial value. Here we only need the loop bound register*/
    /*for(i=1 ; i<=REGLOCLENGTH ; i++)
      loop->iterations[i-1] = loop->iterations[i];
      loop->iterations[i-1] = '\0';*/

    /*We need to determine if the loop variables are live coming into the
      loop, if they are set within the loop, and their uses in the loop*/
    if(analyseLoop(loop))
      return -1;
    return 0;
  }

  /*not a counting loop*/
  return 0;
}


/*
 * analyseLoop - Study the loop to determine things necessary for unrolling
 *               execution time loops. Return -1 if the loop can be unrolled,
 *               0 if not.
 */
int analyseLoop(struct loopl *loop)
{
  char *rtlin, *chptr, *c;
  int i, count;
  struct bblock *cblk, *pheader, *eblk;
  struct blist *bl;
  struct list *rtl, *rtl2, *crtl, *brtl;
  struct setlist *sptr, *cptr, *indptr;

   /*the loop bounds variable should be live coming into the loop. Also check
    if there are any sets/uses of that variable within the loop*/
  sptr = s_lookup(loop->iterations);
  if(isRegister(loop->iterations) && !rexists(loop->iterations, loop->ptr->ulist)){
    
    fprintf(stderr, "Loop variable ");
    show(loop->iterations);
    fprintf(stderr, "not live coming into the loop\n");
    return 0;
  }
  
  /*We already know that the loop bounds variable is loop-invariant. So,
    find the set of the item which must be in the header of the loop for
    type 2 and 3 loops*/
  eblk = findMainLoopExit(loop);
  for(rtl=eblk->rtls ; rtl ; rtl=rtl->next){
    
    if(rtlin = cnts(rtl)){
      if(isRegister(rtlin) && (cptr=s_lookup(rtlin)) && cptr==sptr){
	
	/*move this set to the preheader of the loop*/
	pheader = (struct bblock *)preheader(loop->ptr);
	
	/*If the preheader unconditionally jumps to header of the loop,
	  then the RTL can be safely moved to the preheader*/
	if(!pheader->rtlend){
	  moveRTL(rtl, pheader, NULL);
	}
	else if((rtl2=findLastRTL(pheader)) && (!is_pc(rtl) ||
						is_ucjmp(rtl2->val))){
	  moveRTL(rtl, pheader, rtl2);
	}
	else
	  return 0;
      }
    }
  }

  /*This approach to loop unrolling for execution time bound loops do not
    require an additional register. But it assumes that the loop bounds
    variable is not used within the loop. Check to determine if this
    assumption is true, else we need to assign another register to do the
    same thing*/
  /*For the loop control variable, the only uses of this variable should be
    in the induction RTLs, to allow us to combine these RTLs after loop
    unrolling so that we don't have muiltiple copies of this RTL when it
    could be combined into 1*/
  for(bl=loop->blocks ; bl ; bl=bl->next){
    
    cblk = bl->ptr;
    for(rtl=cblk->rtls ; rtl ; rtl=rtl->next){
      
      /*here we are only looking for the uses of the loop bounds variable.
	So skip the set register*/
      if((rtlin=cnts(rtl)) && !(isRegister(rtlin) && is_compare(rtlin)) &&
	 (rtlin=skiploc(rtlin)) && *rtlin=='='){
	
	/*the control RTL has to be allocated a register. So we don't care
	  for locals*/
	while((rtlin = (char*)gtrg(rtlin))){
	  cptr = s_lookup(unskiploc(rtlin));
	  if(cptr == sptr){
	    
	    fprintf(stderr, "The loop bounds variable is used within this loop\n");
	    allocRegister = TRUE;
	  }
	}
      }
    }
  }
  
  return -1;
}


#define MAXREGTYPES 10
/*
 * push_compare - push the compare instruction closer to the branch, if
 *                possible. Return number of times compare was pushed
 *                below any RTL
 */
int push_compare(struct bblock *eblk)
{
  int i, cmpReg[MAXREGTYPES], rtlReg[MAXREGTYPES], count;
  char *c, coper, relop, rem[MAXSTRING], label[MAXSTRING];
  char reg1[REGSTRING+1], reg2[REGSTRING+1];
  struct bblock *cblk;
  struct list *rtl, *nrtl, *cmpRTL, *brRTL;

  count = 0;
  cmpRTL = findCompareRtl(eblk);
  brRTL = findJumpRtl(eblk);

  for(i=0 ; i<MAXREGTYPES ; i++)
    cmpReg[i] = 0;
  getCompareComp(eblk, reg1, &relop, &coper, reg2, rem, label);
  if(isRegister(reg1))
    cmpReg[regtype(reg1)] |= (1 << regnum(reg1));
  if(isRegister(reg2))
    cmpReg[regtype(reg2)] |= (1 << regnum(reg2));
  
  if(rem[0] == 0){
    for(rtl=cmpRTL->next ; rtl!=brRTL ; rtl=nrtl){

      nrtl = rtl->next;
      if(cnts(rtl)){
	
	/*scan the rtl to check if any of the registers used in the compare and
	  also used in this RTL. If they are then the compare cannot be pushed
	  after this RTL*/
	for(i=0 ; i<MAXREGTYPES ; i++)
	  rtlReg[i] = 0;
	for(c=rtl->val ; c && *c ; c=skiploc(c)){
	  if(isRegister(c))
	    rtlReg[regtype(c)] |= (1 << regnum(c));
	}
	for(c=rtl->se ; c && *c ; c=skiploc(c)){
	  if(isRegister(c))
	    rtlReg[regtype(c)] |= (1 << regnum(c));
	}
	
	/*check for common register reference between the compare instruction
	  and this RTL*/
	for(i=0 ; i<MAXREGTYPES ; i++){
	  if(cmpReg[i] & rtlReg[i])
	    return count;
	}

	/*push the compare instruction below this instruction*/
	moveRTL(cmpRTL, cmpRTL->daddy, rtl->next);
	count++;
      }
    }
  }
  return count;
}


/*
 * singleExitBlk - Check if all the exit blocks in the loop jump to the same
 *                 block outside the loop, or jump to the return block.
 *                 Return the main loop exit block
 */
struct bblock *singleExitBlk(struct loopl *loop)
{
  struct bblock *cblk;

  cblk = NULL;
  if(!loop->exits)
    return NULL;

  if(loop->exits->next){

    /*Performing this step only because such control flow seems to occur in the
      code by the frontend. If the second exit condition is a block with
      only an unconditional transfer of control, then this jump can be
      replaced */
    if(!(loop->exits->ptr==ret_block || loop->exits->next->ptr==ret_block))
      modifyLoopExits(loop);

    /*if we have eliminated all loop exit blocks, then we return. The
      controlflow needs to be simplified by some other optimizations*/
    if(!loop->exits)
      return NULL;
    
    /*if we still have multiple exits return*/
    if(loop->exits->next)
      return NULL;
  }

  /*loop can never have 3 exits*/
  if(loop->exits->next && loop->exits->next->next)
    return NULL;

  /*arrange the loop exits in desired order*/
  if(loop->exits->next && loop->exits->ptr == ret_block){
    loop->exits->ptr = loop->exits->next->ptr;
    loop->exits->next->ptr = ret_block;
  }

  /*find the main exit out of the loop*/
  cblk = findMainLoopExit(loop);

  return cblk;
}


/*
 * modifyLoopExits - In case of a break statement in the loop, the control
 *                   flow produced by the frontend sometimes included an
 *                   unnecessary jump statement generated by a 'break',
 *                   before it jumps out of the loop. Try to remove that
 *                   block, to allow loop unrolling to proceed
 */
void modifyLoopExits(struct loopl *loop)
{
  struct blist *bl, *nbl, *ebl, *nebl, *pebl;;
  struct bblock *cblk, *blk;

  /*loop at each loop exit block. If it only contains a single unconditional
    jump, then attempt to remove the block*/
  for(bl=loop->exits ; bl ; ){

    nbl = bl->next;
    cblk = bl->ptr;
    if(rtlCount(cblk)==1 && cnts(cblk->rtlend) && is_ucjmp(cblk->rtlend->val)){

      /*look at all its predecessors, it is important that the block which
	falls through to it should have another jump to the next block in the
	control flow*/
      for(ebl=cblk->preds ; ebl ; ebl=nebl){

	nebl = ebl->next;
	blk = ebl->ptr;
	if(blk->left==cblk){
	  if(!blk->right || blk->right!=cblk->down)

	    /*return since control flow is not as desired*/
	    return;
	}
      }

      /*the control flow seems to be what we expect. Change the predecessors
	and remove the exit block*/
      for(ebl=cblk->preds ; ebl ; ebl=nebl){

	nebl = ebl->next;
	blk = ebl->ptr;
	if(blk->right == cblk){

	  /*chain the jump to the following block*/
	  update_bb_ptrs(MOD_RIGHT, blk, cblk->left);
	}
	else{
	  reverse(blk->rtlend, cblk->left->label);
	  update_bb_ptrs(MOD_LEFT, blk, cblk->down);
	  update_bb_ptrs(MOD_RIGHT, blk, cblk->left);
	}
      }

      /*remove the exit block*/
      cblk->down->up = cblk->up;
      update_bb_ptrs(MOD_DOWN, cblk->up, cblk->down);
      remv_basic_block(cblk);

      /*remove this block from the predecessor list of its sucessors*/
      if(cblk->left)
	remv_predecessor(cblk, cblk->left);
      if(cblk->right)
	remv_predecessor(cblk, cblk->right);
   
      
      /*remove this block from the list of block exits*/
      pebl = NULL;
      for(ebl=loop->exits ; ebl ; ){

	nebl = ebl->next;
	if(ebl->ptr == cblk){

	  if(pebl)
	    pebl->next = nebl;
	  else
	    loop->exits = nebl;
	  break;
	}
	pebl = ebl;
	ebl = nebl;
      }
    }

    /*go to the next loop exit*/
    bl = nbl;
  }
}


/*
 * findMainLoopExit - find the main loop control block
 */
struct bblock *findMainLoopExit(struct loopl *loop)
{
  struct blist *bl, *nbl;
  struct bblock *cblk, *eblk;

  /*the loop exits are already assumed to be arranged such that the first
    exit block is the main exit block out of the loop*/
  eblk = loop->exits->ptr;

  /*We allow multiple loop blocks to exit to this block. The main loop exit
    though should either be the loop header or jump to the loop header. Check
    for this condition*/
  for(bl=eblk->preds ; bl ; bl=nbl){

    nbl = bl->next;
    cblk = bl->ptr;
    if(block_in_loop(loop, cblk)){
      if(cblk==loop->ptr || (cblk->left==eblk && cblk->right==loop->ptr) ||
	 (cblk->right==eblk && cblk->left==loop->ptr))
	return cblk;
    }
  }
  return NULL;
}


/*
 * loop_iterations - attempt to determine how many times loop is executed
 */
int loop_iterations(struct loopl *loop, struct setlist **cptr, char *value,
		    char *initial, int *inc, char *cval, char *reloper)
{
   register struct bblock *cblk;
   struct blist *bptr, *nbl;
   struct llist *lptr;
   struct list *ptr, *rtl, *cmpRTL;
   struct setlist *cmp1, *cmp2;
   int   val, init, item, before, after;
   char  oper, relop, reg[MAXSTRING], *a, temp[MAXSTRING], rem[MAXSTRING], 
         coper;
   char rtlval[MAXSTRING], label[MAXSTRING], op1[REGSTRING], op2[REGSTRING];
   bzero(reg, sizeof(reg));
   bzero(temp, sizeof(temp));

#ifdef INTERACT
   begin_trans(VET_OPT_TRANS);
#endif   
   *value = '\0';

   /*for loop unrolling we only handle loops with a single exit block outside
     of the loop, or exit to the return block. Note that this means there can
     be multiple loop exit blocks. This function returns the main loop exit
     block*/
   cblk = (struct bblock *)singleExitBlk(loop);

   if(!is_single_exit_loop(loop))
     multipleExitLoop = TRUE;
   else
     multipleExitLoop = FALSE;

#ifdef INTERACT
   end_trans();
#endif
   /*
    * At this point, if we have a pointer to a valid basic block, then it is
    * the single exit point from which the loop exits.  We must now examine
    * this block to determine what the control variable of the loop is and
    * if we can determine its initial value and the exit condition of the
    * loop.
    */
   if (!cblk)
      return FALSE;

   /*
    * Check that the exit condition is executed on every iteration
    * through the loop.  Otherwise, there is little we can tell
    * about the exit condition by just looking at the last test.
    */
   if (!everytime(loop, cblk))
      return FALSE;

   /*
    * Given that there is only one exit out of the loop, attempt to
    * locate the "control variable" of the loop and its exit value.
    */
   if (loop_exit_cond(loop, cblk, cblk->right&&block_in_loop(loop, cblk->right),
		      reg, value, &relop)) {

   /* Sometimes the loop variable and the control variable might be
	 reversed in the comparison. This poses some problems for this
	 routine. To make the state consistent, reverse the branch for such a
	 case */
#ifdef INTERACT
     begin_trans(VET_OPT_TRANS);
#endif
     if(isRegister(reg) && isRegister(value) &&
	(cmp1=s_lookup(reg)) && (cmp2=s_lookup(value))){
       if(!(cmp1->regstatus & BASIC) && (cmp2->regstatus & BASIC)){

	 /*reverse the order of the comparison*/
	 getCompareComp(cblk, op1, &oper, &coper, op2, rem, label);
	 setCompare(cblk, op2, oper, coper, op1, rem, label);
	 relop = invertBranch(cblk->rtlend, label);

	 /*get the exit condition variables again*/
	 loop_exit_cond(loop, cblk, cblk->right && block_in_loop(loop, cblk->right),
			reg, value, &relop);
       }	 
     }
#ifdef INTERACT
     end_trans();
#endif
     /*
       * Now check to see if the control variable is a basic induction
       * variable.  If it is, and the increment is performed every time
       * through the loop (a condition that must be true if the basic
       * block in which the increment is located in dominates every
       * predecessor block of the loop header that is also in the loop),
       * then we know how much it is incremented by on each iteration
       * through the loop.
       */
     *cptr = s_lookup(reg);
      if ((*cptr)->regstatus & BASIC) {
         before = after = 0;
         for (lptr = (*cptr)->triple->ref; lptr; lptr = lptr->next) {
            if (!everytime(loop, lptr->ptr->daddy)) {
               *value = '\0';
               return FALSE;
            }
            else if (dominates(lptr->ptr->daddy, cblk))
               before += lptr->inc;
            else
               after += lptr->inc;
         }
         after += before;
         if (!after) {
            *value = '\0';
            return FALSE;
         }
	 if(inc)
	   *inc = after;
	 if(reloper)
	   *reloper = relop;

         /*
          * Now all that remains is to look at the preheader of the loop and
          * see if we can find the value that the loop control variable is
          * initialized to.  Knowing this, we can determine exactly how many
          * times the loop is executed.
          */
         if (!(cblk = loop->preheader))
            loop->preheader = cblk = (struct bblock *)natural_preheader(loop->ptr);
         if (!cblk) {
            /* We couldn't find a pre-header. One reason why this could happen
             * is when the loop exit condition is based on a loop-invariant
             * reg. So the loop entry is protected by a conditional branch that
             * jumps over the loop if the loop condition is not met.
             * In such a case, search for a predecessor of the loop head
             * (loop->ptr) such that it is the only predecessor that is outside
             * the loop.
             */
            struct bblock *hblk, *pblk;
            struct blist *bptr;
            hblk = NULL;
            pblk = NULL;
            /* This loop tries to find two predecessors for loop->ptr, that
             * lie outside the loop. If it fails => we succeeded in our goal.
             */
            for (bptr = loop->ptr->preds; !pblk && bptr; bptr = nbl) {

	      nbl = bptr->next;
	      if (!block_in_loop(loop, bptr->ptr)) {
                  pblk = hblk;
                  hblk = bptr->ptr;
               }
            }
            if (!pblk)
               cblk = hblk;
         }
         if (cblk && (ptr = value_of(cblk->rtlend, reg, rinit(), temp))) {
            expand((struct loopl *) NULL, ptr, temp);
            simp(temp, FALSE);
	    if(initial)
	      strcpy(initial, temp);

	    /*check if the initializer is a constant*/
            if ((a = is_const(temp)) && !*a) {
	      
               init = atoi(temp) + before;

	       /*for type 1 loops, there is no need to add before to init,
		 since even though 'before' is added to the loop variable
		 before the first comparison, the loop body has already been
		 executed once at this point*/
	       if(getLoopStructure(loop) != 1)
		 sprintf(initial, "%d", init);

               /*
                * Examine the exit condition value. If the value is a constant,
                * then we can determine how many times the loop is executed.
		* We also need to check if this exit block is the one always
                * reached, else iterations cannot be determined.
                */
               if (isInteger(value) && !multipleExitLoop) {
		 if(cval)
		   strcpy(cval, value);
		 
                 val = atoi(value);
		 switch (relop) {

		 case CMP_LE:
		 case CMP_ULE:
                      if (init > val) {
                         (void) strcpy(value, "1");
                         return TRUE;
                      }
                      item = (val - init) / after;
                      if (item > -1) {
                         (void) sprintf(value, "%d", item + 2);
                         return TRUE;
                      }
                      *value = '\0';
                      return FALSE;

		 case CMP_GE:
		 case CMP_UGE:
                      if (init < val) {
                         (void) strcpy(value, "1");
                         return TRUE;
                      }
                      item = (init - val) / (-after);
                      if (item > -1) {
                         (void) sprintf(value, "%d", item + 2);
                         return TRUE;
                      }
                      *value = '\0';
                      return FALSE;

		 case CMP_LT:
		 case CMP_ULT:
                      if (init >= val) {
                         (void) strcpy(value, "1");
                         return TRUE;
                      }
                      item = (val - init - 1) / after;
                      if (item > -1) {
                         (void) sprintf(value, "%d", item + 2);
                         return TRUE;
                      }
                      *value = '\0';
                      return FALSE;

		 case CMP_GT:
		 case CMP_UGT:
                      if (init <= val) {
                         (void) strcpy(value, "1");
                         return TRUE;
                      }
                      item = (init - val - 1) / (-after);
                      if (item > -1) {
                         (void) sprintf(value, "%d", item + 2);
                         return TRUE;
                      }
                      *value = '\0';
                      return FALSE;

		 case CMP_NE:
                      if (init <= val && after > 0) {
                         item = (val - init) / after;
                         (void) sprintf(value, "%d", item + 1);
                         return TRUE;
                      }
                      else if (init >= val && after < 0) {
                         item = (init - val) / -after;
                         (void) sprintf(value, "%d", item + 1);
                         return TRUE;
                      }
                      *value = '\0';
                      return FALSE;
                  }
               }
	    }
	 }

	 /*If we get here then the loop control variable is not a constant
	   and/or noting could be said about the loop initializer. We can
	   still unroll the loop as long as the loop control variable is loop
	   invariant */
	 if (is_invariant(value, !loop->invaread, loop, NULL)) {
	   strcpy(cval, value);
	   return TRUE;
	 }
      }
   }

   /*
    *  Return the "I don't know" or "infinite" result
    */
   *value = '\0';
   return FALSE;
} /* end of iterations() */

#if 0
/*#ifndef UNDO_H*/
/* From undo.c */
/*
 * clear_all_preds - clear all of the predecessors for all of the
 *                   basic blocks
 */
void clear_all_preds()
{
   struct bblock *cblk;
   extern struct bblock *ret_block;

   /*clear all the preds for each basic block*/
   for (cblk = top; cblk; cblk = cblk->down)
      clear_bb_preds(cblk);
   clear_bb_preds(ret_block);
}

/*
 * recalculate_all_preds - recalculate all of the predecessors for all
 *                         basic blocks
 */
void recalculate_all_preds()
{
   struct bblock *cblk;
   struct blist *bl;

   clear_all_preds();
   
   for (cblk = top; cblk; cblk = cblk->down) {
     
     /*add new predecessors*/
      if (cblk->left)
         add_predecessor(cblk, cblk->left);
      if (cblk->right)
         add_predecessor(cblk, cblk->right);
   }
}

/*
 * clear_bb_preds - clear all of the predecessors for a single basic block
 */
void clear_bb_preds(struct bblock *cblk)
{
   register struct blist *bptr, *dptr;
   
   /*free the space used by predecessor nodes*/
   bptr = cblk->preds;
   while (bptr) {
      dptr = bptr;
      bptr = bptr->next;
      FREETYPE(dptr, BLISTTYPE);
   }

   /*set to NULL*/
   cblk->preds = NULL;
}

/*
 * is_ucjmp - is RTL an unconditional jump
 */
int is_ucjmp(char *s)
{
   return(*s == 'P' && *(s + 1) == 'C' && *(s + 3) == 'L');
}

#endif /*UNDO_H*/

#if 0
/*#ifndef REMJUMP_H*/
/* 
 * copy_block - make a copy of the given basic block
 */
struct bblock *copy_block(struct bblock *cblk)
{
  struct bblock *new;
  struct list *rtl, *prtl, *nrtl;

#ifdef INTERACT
  begin_trans(VET_OPT_TRANS);
#endif

  new = (struct bblock *)NEW_basic_block(NULL);
  label_basic_block(new);

#ifdef INTERACT
  end_trans();
  begin_trans(VET_OPT_TRANS);
#endif

  /*add all the RTLs of the current block in this new block*/
  prtl = NULL;
  for(rtl=cblk->rtls ; rtl ; rtl=rtl->next){
    nrtl = (struct list *)new_rtl(new, rtl->type, prtl);
    if(rtl->val){
      nrtl->val = (char *)CALLOC(strlen(rtl->val)+1);
      strcpy(nrtl->val, rtl->val);
    }
    nrtl->kind = rtl->kind;
    rcpy(&nrtl->deads, rtl->deads);
#ifdef INTERACT
    insert_rtl(nrtl);
#endif
    prtl = nrtl;
  }

#ifdef INTERACT
  end_trans();
#endif
  return (struct bblock*)new;
}

/* change_branch_label - just replace the target label in the branch
 *                       instruction with the new label. Does not affect any
 *                       basic block pointers
 */
void change_branch_label(struct list *rtl, char *label)
{
  char *pos, *val, new[MAXSTRING];
  int len;

#ifdef INTERACT
  /*begin_trans(VET_OPT_TRANS);*/
#endif

  for(len=0,pos=rtl->val ; *pos!='L' ; pos++)
    new[len++] = *pos;

  for(pos=label ; *pos ; pos++)
    new[len++] = *pos;
  new[len++] = ';';
  new[len] = '\0';

  /*copy over the new RTL*/
  FREE(rtl->val);
  rtl->val = (char *)CALLOC(strlen(new)+2);
  strcpy(rtl->val, new);

#ifdef INTERACT
  mod_rtl(rtl, NULL, rtl->val);
  /*end_trans();*/
#endif
}

#endif /*REMJUMP*/

/*
 * New_basic_block - just a wrapper for new_basic_block
 */
struct bblock *NEW_basic_block(char *label)
{
  struct bblock *cblk;

  cblk = (struct bblock *)new_basic_block(label);
  if(cblk->block_id > MAX_BLOCKS){
    fprintf(stderr, "New block number exceededed the MAX_BLOCKS used for some structures. Update MAX_BLOCKS in vpo.h\n");
  }

  return(cblk);
}


/*
 * isInteger - check to determine if the string starts with a constant
 */
int isInteger(char *str)
{
  char *c;

  c = str;
  if(*c == '-')
    c++;
  if(*c <= '9' && *c >= '0')
    return TRUE;
  return FALSE;
}

