/************************************************************************/
/*                                                                      */
/* Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   */
/*     The Rector and Visitors of the University of Virginia            */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from the University of Virginia.                                */
/* This code is distributed WITHOUT ANY WARRANTY.  No claims are        */
/*      made as to whether it serves any particular purpose or even     */
/*      works at all.                                                   */
/*                                                                      */
/************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/measure.c,v $
 $Date: 2004/01/21 18:39:57 $
 $Revision: 1.16 $
 $Name:  $

 Log of changes:

 $Log: measure.c,v $
 Revision 1.16  2004/01/21 18:39:57  jdh8d
 Latest Vista additions/changes.  Don't yell at me if it's broken/bad.. prasad wanted these changes.  I've reviewed them, and tested,  minimally.. things appear OK (for the batch compiler)

 Revision 1.15  2004/01/15 15:43:38  jdh8d
 fixed C++ style comments .. don't put them in again, YOU HEAR ME WHOEVER DID THIS??

 Revision 1.14  2003/08/21 20:30:29  jdh8d
 changes to ease for tms

 Revision 1.13  2003/08/20 17:54:44  jdh8d
 changes to get genreport, etc working for tms

 Revision 1.12  2003/07/17 04:33:42  jg9h
 VISTA update

 Revision 1.11  2003/04/24 05:01:47  jg9h
 VISTA:Use main_fn_name

 Revision 1.10  2003/04/04 04:36:47  jg9h
 Removed C++ style comments from INTERACT code

 Revision 1.9  2003/04/03 20:06:10  jg9h
 Merged latest in VISTA with latest in vpo.

 Revision 1.8  2003/02/15 00:09:31  clc5q
 Get rid of "extern int moreopts;" and other stupi "extern" statements and
   other sources of compiler warnings.

 Revision 1.7  2003/02/06 04:46:29  jg9h
 Merged VISTA into mainline vpo under the INTERACT preprocessor flag

 Revision 1.6  2002/07/11 15:22:31  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.5  2000/06/09 19:42:17  jwd
 Added a dummy static so the compiler wouldn't complain about
 an empty module.

 Revision 1.4  1998/06/26 20:58:52  jwd
 Adde some include files so the proper prototypes
 were found.
 
 Added a calll to final_call() in exit-call(). This had somehow
 been deleted.  Function final_call() generates the code
 to call the ease function to dump the information to a file.
 
 Removed a spurious prototype from branch_measure().
 
 Revision 1.3  1998/04/03 18:19:45  jwd
 First ANSI C version.

*************************************************************************/

/*------HISTORY of CHANGES --------------
   1/12/98: Modified to ANSI-C compatible function prototyping.
             Note, the header file still declares data!
----------------------------------------*/

#ifdef MEASURE

#ifdef INTERACT
#define COUNT_ALL
#endif

#define STRLENGTH 80
static int dummy = 0;  /* so module is not empty */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "vpo.h"
#include "addrdef.h"
#include "combine.h"
#include "controlflow.h"
#include "measure.h"
#include "misc.h"
#include "vectors.h"

#include "linstinfo.h"   /* Machine Dependent */
#include "rtl.h"        /* Machine Dependent */
#endif /*MEASURE*/

/*
 * These limits apply to any single function
 */
#define MAXEQUIVS 1024
#define MAXBLOCKS 4096
#define MAXINSTS 32768

#ifdef __SVR4
static char *EasePrefix = "__";
#else
static char *EasePrefix = "___";
#endif

int totalblks;		/* Total number of blocks in this function    */
int equiv_classes;	/* Number of equivalence classes in function  */
int insttype;		/* Instruction type of last item parsed       */
int *info_ptr;		/* Pointer to next free information item      */
int numblks[MAXEQUIVS];	/* Number of blocks in an equivalence class   */
int blksinec[MAXBLOCKS];/* List of blocks in an equivalence class     */
int numinsts[MAXBLOCKS];/* Number of instructions in each small block */
int func_instinfo[MAXINSTS * 8];	/* Contain execution frequency
					   information    */
extern char *main_fn_name;

#ifdef MEASURE
#ifdef INTERACT
#include "view.h"
extern struct bblock *top, *ret_block;
extern struct loopl *loops;
extern struct bblock *last, *bottom;
extern int swef, swei, swed, swep, swa;
extern char *fn, *sret, *labprt, *acomment;
extern char *startblklabel, *endblklabel;
extern char pathname[100], basename[100];
extern int varargs, LeafFunction;
extern char *ret, *linptr;
extern int funcname, dassem;
extern int sem_ptr, str_ptr, i_ptr;
extern rlist deadlist;
extern struct list *curptr;
extern int measure_phase, measure_flag, save_measure_flag;
extern int measure_init, send_messages;
extern FILE *instfp, *conf_fp;
extern char inst_filename[100], *fn;
extern struct freq_counts *initial_count;
extern unsigned int trans_done, session_over;

/* to store the global variables reflecting current function state*/
static struct bblock *store_top, *store_ret_block;
static struct bblock *store_last, *store_bottom;
static char *store_fn, *store_sret;
static int store_varargs, store_swa, store_LeafFunction;
static int store_funcnum;

/*global variable for input file*/
FILE *fp, *outfile;

#endif

/*
 * measure_freq - instrument the function to obtain execution frequency counts
 */
void measure_freq() {

   struct bblock *cblk, *tblk;
   struct list *ptr, *tptr, *rjmp;
   struct loopl *loop;
   int found, i, j, k, ismain, first, inserted;
   char *r;

   /*
      Calculate the equivalence classes in the current function
   */
   equivalence_classes();

   /*
      Associate an equivalence class number with each equivalence class
   */
   for (cblk = top; cblk; cblk = cblk->down) {
      cblk->equivnum = -2;
      for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
	 if (cnts(ptr)) {
	    cblk->equivnum = -1;
	    break;
	 }
      }
#ifdef COUNT_ALL
      cblk->equivnum = -1;
#endif	
   }
   equiv_classes = totalblks = 0;
   for (cblk = top; cblk; cblk = cblk->down) {
      if (cblk->equivnum == -1) {
	 if (equiv_classes >= MAXEQUIVS) {
	    (void) fprintf(stderr,
			"vpo (measure_freq): too many equivalence classes\n");
	    exit(1);
	 }
	 numblks[equiv_classes] = 0;
	 cblk->equivnum = equiv_classes;
	 for (tblk = cblk->down; tblk; tblk = tblk->down)
#ifdef COUNT_ALL
	    if (FALSE && tblk->equivnum == -1 &&
                din(tblk->equivclass, cblk->bnum))
#else
	    if (tblk->equivnum == -1 && din(tblk->equivclass, cblk->bnum))
#endif		 
	       tblk->equivnum = equiv_classes;
	 equiv_classes++;
      }
   }

#ifdef INTERACT
   for (i = 0, cblk = top; cblk; cblk = cblk->down) {
     cblk->equivnum = -2;
     for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
       if (cnts(ptr) && ptr->onum) {
	 cblk->equivnum = i++;
	 break;
       }
     }
   }
   equiv_classes = i;
#endif
	
   /*
      Set up to capture the instruction frequency information for this
      function.  First, place the name of the current function in the
      information so that the report generator can determine which function
      this information belongs to.
   */

	/* this is broken, it assumes a word is 4 bytes.. which isn't true */
	/* on tms, a word holds 1 byte */

	/* i'm changing this to assume that a word at least 1 byte and the func is
	 * null terminated 
	 */

   info_ptr = func_instinfo;
   r = fn;
   for (;;) {

	j=*r++;
	*info_ptr++ = j;

	if(!j)
	 break;
   }

   /*
      Add the register usage information for this function
   */
   register_usage_information();

   /*
      Insert code to increment a counter for each equivalent class within the
      function.
   */
   for (i = 0; i < equiv_classes; i++) {

      /*
         Find a place to increment the counter for this execution class
      */
      found = FALSE;
      for (cblk = top; cblk; cblk = cblk->down) {
	 if (cblk->equivnum == i) {
	    for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
	       if (cnts(ptr) && ptr->onum) {
		  found = TRUE;
		  break;
	       }
	    }
	    if (found)
	       break;
	 }
      }
      if (!cblk || !ptr) {
	 (void) fprintf(stderr, "vpo (measure_freq): no increment location\n");
	 exit(1);
      }

      /*
         If the number of times that the execution class is executed can be
         inferred from the predecessor block and another of the predecessor's
         successors (if-then-else structure), then do not insert a counter.
         Instead, write information to the information file that indicates how
         the count for the execution class can be determined.
      */
		if (
#ifdef COUNT_ALL
      FALSE && cblk->up && cblk->up->equivnum >= 0 && cblk->preds &&
#else
      cblk->up && cblk->up->equivnum >= 0 && cblk->preds &&
#endif		
	  cblk->preds->ptr == cblk->up && !cblk->preds->next &&
	  cblk->up->left == cblk && cblk->up->right &&
	  cblk->up->right->equivnum >= 0 && cblk->up->right->preds &&
	  cblk->up->right->preds->ptr == cblk->up &&
	  !cblk->up->right->preds->next) {
	 *info_ptr++ = i;
	 *info_ptr++ = cblk->up->equivnum;
	 *info_ptr++ = cblk->up->right->equivnum;
	 continue;
      }

      /*
         If equivalence class includes the header of a loop and the number of
         iterations is known, then increment the counter by the number of
         iterations in the preheader of the loop.
      */
      inserted = FALSE;
      for (loop = loops; loop; loop = loop->next) {
	 if (loop->ptr->equivnum == i) {

	    /*
	       Check to see if the loop iteration count is known and there is a
	       natural preheader for this loop.  We cannot create if there
	       isn't since it might invalidate the equivalence class
	       information that we are currently working with.
	    */
	    if (loop->valid && (loop->preheader ||
			  (loop->preheader = natural_preheader(loop->ptr)))) {

	       /*
	          If the loop iteration count is constant then attempt to move
	          the insertion of the increment by this constant out of any
	          other loops it is nested in.  Other loops are checked to
	          determine if their counts are also constant.  If so, then
	          multiply the current count by the number of iterations in the
	          outer loop.
	       */
	       if ((j = atoi(loop->iterations)) > 0) {
		  k = j;
		  tblk = loop->preheader;
		  if (tblk->equivnum == -2 && tblk->preds && tblk->preds->ptr &&
		      !tblk->preds->next && tblk->preds->ptr->left == tblk &&
		      !tblk->preds->ptr->right) {
		     tblk = tblk->preds->ptr;
		  }
		  do {
		     found = FALSE;
		     for (loop = loops; loop; loop = loop->next) {
			if (loop->ptr->equivnum == tblk->equivnum) {
			   if (loop->valid && (loop->preheader ||
			   (loop->preheader = natural_preheader(loop->ptr))) &&
			       (j = atoi(loop->iterations)) > 0) {
			      tblk = loop->preheader;
			      if (tblk->equivnum == -2 && tblk->preds &&
				  tblk->preds->ptr && !tblk->preds->next &&
				  tblk->preds->ptr->left == tblk &&
				  !tblk->preds->ptr->right)
				 tblk = tblk->preds->ptr;
			      k = k * j;
			      found = TRUE;
			   }
			   break;
			}
		     }
		  } while (found);

		  /*
		     Locate the best place at the end of the basic block to
		     insert the increment, knowing that the end of the basic
		     block may be a jump or even a jump followed by a filled
		     delay slot.  In all cases, move the increment before
		     these.
		  */
		  if (tblk->status & JUMP) {
		     for (tptr = tblk->rtlend;
			  tptr && (tptr->kind & (TAKEN | NOTAKEN));
			  tptr = tptr->prev);
		  }
		  else
		     tptr = (struct list *) NULL;

		  inserted = increment_class_counter(tblk, tptr, k, i);
	       }

	       /*
	          Otherwise, attempt to increment the counter by the variable
	          value.  As above, look out for filled delay slots and jumps.
	       */
	       else {
		  if (loop->preheader->status & JUMP) {
		     for (tptr = loop->preheader->rtlend;
			  tptr && (tptr->kind & (TAKEN | NOTAKEN));
			  tptr = tptr->prev);
		  }
		  else
		     tptr = (struct list *) NULL;
		  inserted = increment_class_counter_by_expression(
				  loop->preheader, tptr, loop->iterations, i);
	       }
	    }
	    break;
	 }
      }

      /*
         Otherwise, insert the increment in the first block of the equivalent
         class.
      */
#ifdef COUNT_ALL
      if (TRUE || !inserted) 
#else
      if (!inserted) 
#endif		
		{
	 if (!increment_class_counter(cblk, ptr, 1, i)) {
	    (void) fprintf(stderr,
			   "vpo (measure_freq): cannot instrument code\n");
	    exit(1);
	 }
      }
   }

   /*
      If this is the "main" function, insert the code to initialize the trace
   */
   if (ismain = !strcmp(fn, main_fn_name))
      enter_call();

   /*
      Insert calls to dump out instruction information (if needed) and insert
      code to increment a counter after each relational jump that was not
      taken.
   */
   *info_ptr++ = -2;
   for (cblk = top; cblk; cblk = cblk->down) {
      rjmp = (struct list *) NULL;
      first = TRUE;
      for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
	 if (cnts(ptr) && ptr->onum) {
	    if ((ismain && is_pc(ptr->val) &&
		 (cblk->left == ret_block || cblk->right == ret_block)) ||
		(is_call(ptr->val) && callsexit(ptr))) {
	       if (!first)
		  class_has_call(cblk);
	       exit_call(ptr);
	    }
	    if (is_anyrjmp(ptr->val))
	       rjmp = ptr;
	    first = FALSE;
	 }
      }

      if (rjmp)
	 branch_measure(cblk, rjmp);
   }

   /*
      Insert labels needed to calculate equivalence class instruction size
      information.
   */
   for (i = 0; i < equiv_classes; i++) {
      for (cblk = top; cblk; cblk = cblk->down) {
	 if (cblk->equivnum == i)
	    add_block_to_class(cblk);
      }
   }

   /*
      Insert information to indicate how often each loop is executed versus how
      often it is entered.
   */
   *info_ptr++ = -3;
   for (loop = loops; loop; loop = loop->next) {
      if (loop->preheader || (loop->preheader = natural_preheader(loop->ptr))) {
	 *info_ptr++ = loop->preheader->equivnum;
	 *info_ptr++ = loop->ptr->equivnum;
	 r = loop->ptr->label;
	 for (;;) {
	    j = 0;
	    j |= (int) *r++;
	    *info_ptr++ = j;
	    if(!j)
	       break;
	 }
      }
   }
   *info_ptr++ = -4;
}

/*
 * measure_cache - instrument the function to obtain cache simulation results
 */
void measure_cache() {

   register struct bblock *cblk;
   register struct list *ptr;
   struct list *sptr;
   int numinst, insts, hasmem, ismain;

   /*
      If this is the "main" function, insert the code to initialize the trace
   */
   if (ismain = !strcmp(fn, main_fn_name))
      enter_call();

   /*
      Go through each instruction in the function, inserting calls to the cache
      simulation routines as appropriate.
   */
   totalblks = insts = 0;
   for (cblk = top; cblk; cblk = cblk->down) {
      ptr = cblk->rtls;
      while (ptr) {
	 hasmem = FALSE;
	 sptr = (struct list *) NULL;

	 /*
	    For each small block (these blocks are also delimited by calls)
	 */
	 for (numinst = 0; ptr; ptr = ptr->next) {
	    if (cnts(ptr) && ptr->onum) {
	       ptr->cost = totalblks;
	       if (!numinst++) {
		  if (!swed && swei)
		     trace_call(cblk, ptr, totalblks, insts);
		  sptr = ptr;
	       }
	       insts++;
	       hasmem = (swed && trace_memory_references(ptr, totalblks));

	       /*
	          Some machines may have delay slots immediately following a
	          call instruction.  The instructions in these slots must be
	          left in their place and counted as part of the same small
	          block that the call is in.
	       */
	       if (is_call(ptr->val)) {

#ifdef FILL
		  if (cnts(ptr->next) && ptr->next->onum &&
		      (ptr->next->kind & (TAKEN | NOTAKEN))) {
		     ptr = ptr->next;
		     numinst++;
		     insts++;
		     if (swed && trace_memory_references(ptr, totalblks))
			hasmem = TRUE;
		  }
#endif

		  break;
	       }

	       /*
	          For unified cache simulations, an instruction with a memory
	          reference ends a small block.
	       */
	       if (swed && swei && hasmem)
		  break;
	    }
	 }

	 /*
	    Store the number of instructions
	 */
	 if (numinst) {
	    if (totalblks >= MAXBLOCKS) {
	       (void) fprintf(stderr, "vpo (measure_cache): too many blocks\n");
	       exit(1);
	    }
	    if (!hasmem && swei && swed)
	       unify_call(cblk, sptr, totalblks);
	    numinsts[totalblks++] = numinst;
	 }

	 if (ptr)
	    ptr = ptr->next;
      }
   }

   /*
      Insert instruction labels and the call to the information dump function
      at each possible exit point in the program.
   */
   for (cblk = top; cblk; cblk = cblk->down) {
      ptr = cblk->rtls;
      while (ptr) {
	 if (cnts(ptr) && ptr->onum) {
	    if ((ismain && is_pc(ptr->val) &&
		 (cblk->left == ret_block || cblk->right == ret_block)) ||
		(is_call(ptr->val) && callsexit(ptr)))
	       exit_call(ptr);
	    if (swei)
	       ptr = inst_labels(ptr);
	    else
	       ptr = ptr->next;
	 }
	 else
	    ptr = ptr->next;
      }
   }
}

/*
 * measure_time - instrument the function to obtain timing measurements
 */
void measure_time() {

   struct bblock *cblk;
   struct list *ptr;
   int ismain;

   /*
      If this is the "main" function, insert a call to the function that
      initializes the starting time.
   */
   if (ismain = !strcmp(fn, main_fn_name))
      enter_call();

   /*
      Insert a call to the function that prints out the total execution time at
      every possible exit point in the program.
   */
   for (cblk = top; cblk; cblk = cblk->down) {
      for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
	 if (cnts(ptr) && ptr->onum) {
	    if ((ismain && is_pc(ptr->val) &&
		 (cblk->left == ret_block || cblk->right == ret_block)) ||
		(is_call(ptr->val) && callsexit(ptr)))
	       exit_call(ptr);
	 }
      }
   }
}

/*
 * add_block_to_class - add a block to the set of blocks in an execution class
 */
void add_block_to_class(struct bblock * cblk) {

   static int blksbefore = 0;
   struct list *ptr, *tptr;
   char new[MAXSTRING];

   /*
      Locate the first RTL in the basic block
   */
   for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
      if (cnts(ptr) && ptr->onum) {
	 if (totalblks >= MAXBLOCKS) {
	    (void) fprintf(stderr, "vpo (add_block_to_class): too many blocks\n");
	    exit(1);
	 }
	 blksinec[totalblks++] = blksbefore;

	 /*
	    Indicate that there is one more basic block in the current
	    equivalence class.
	 */
	 if (numblks[cblk->equivnum] < 0)
	    numblks[cblk->equivnum]--;
	 else
	    numblks[cblk->equivnum]++;

	 /*
	    Tag the first RTL with a starting label (if needed)
	 */
	 if (startblklabel) {
	    (void) sprintf(new, startblklabel, blksbefore);
	    tptr = add_rtl(cblk, ptr, new, rinit());
	    tptr->type = ASMLINE;
#ifdef INTERACT
	    mod_type(tptr);
#endif
		 
	    /*
	       Locate the last RTL in the basic block and tag it with a end
	       label
	    */
	    for (ptr = cblk->rtlend; ptr; ptr = ptr->prev)
	       if (cnts(ptr) && ptr->onum)
		  break;
	    (void) sprintf(new, endblklabel, blksbefore++);
	    tptr = add_rtl(cblk, ptr->next, new, rinit());
	    tptr->type = ASMLINE;
#ifdef INTERACT
	    mod_type(tptr);
#endif		 
	 }

	 /*
	    We are finished with this block
	 */
	 break;
      }
   }
}

/*
 * class_has_call - denote that the list size needs to be adjusted
 */
void class_has_call(struct bblock * cblk) {


   /*
      Making the number of basic blocks in list number negative indictaes that
      the equivalence class size should be considered to have one extra
      function call.
   */
   if (numblks[cblk->equivnum] > 0)
      numblks[cblk->equivnum] = -numblks[cblk->equivnum];
}

/*
 * enter_call - insert the appropriate call to start measurements
 */
void enter_call() {

   struct bblock *cblk;
   struct list *ptr;
   char *a, suffix[MAXSTRING];
 
   /*
      Skip empty basic blocks
   */
   for (cblk = top; !cblk->rtls; cblk = cblk->down);
   if (!cblk)
      return;

   /*
     Place this call right after the call to the function name "main"
   */
   /*for(ptr=cblk->rtlend ; ptr ; ptr=ptr->prev){
     if((ptr->type & ASMLINE) && ptr->val &&
	(strcmp(ptr->val, "main:") == 0))
       break;
   }
   ptr=ptr->next;*/
   
   /*
      Place this call before the first instruction in the function
   */
   for (ptr = cblk->rtls; ptr; ptr = ptr->next)
     if (cnts(ptr) && ptr->onum)
       break;

   /*
      Determine the suffix of the name of the function to call
   */
   a = suffix;
   if (swef)
      *a++ = 'f';
   if (swed)
      *a++ = 'd';
   if (swei)
      *a++ = 'i';
   if (swep)
      *a++ = 'p';
   *a = '\0';

   initial_call(cblk, ptr, suffix);
}

/*
 * exit_call - insert the appropriate call to end measurements
 */
void exit_call(struct list * ptr) {

   char *a, suffix[MAXSTRING];

   /*
      Determine the suffix of the name of the function to call
   */
   a = suffix;
   if (swef)
      *a++ = 'f';
   if (swed)
      *a++ = 'd';
   if (swei)
      *a++ = 'i';
   if (swep)
      *a++ = 'p';
   *a = '\0';

    final_call(ptr->daddy, ptr, suffix);
}

/*
 * branch_measure - insert code to measure branch executions
 */
void branch_measure(struct bblock * cblk, struct list * ptr) {

   struct bblock *tblk;

   /*
      Determine the instruction type of the branch
   */
   insttype = 0;
   (void) find(ptr, ptr->val, ptr->deads);
   if (!insttype) {
      (void) fprintf(stderr, "vpo (branch_measure): no type for ");
      show_rtl(stderr, ptr->val);
      (void) fprintf(stderr, "\n");
      exit(1);
   }

   /*
      We can avoid incrementing a counter if the basic block fallen into has
      only one predecessor.
   */
   if (cblk->left && cblk->left->equivnum >= 0 &&
       cblk->left->preds && !cblk->left->preds->next) {
      *info_ptr++ = cblk->left->equivnum;
      *info_ptr++ = 1;
      *info_ptr++ = insttype;
      return;
   }

   /*
      We can avoid incrementing a counter if target basic block of the
      conditional branch has only one predecessor.
   */
   if (cblk->right && cblk->right->equivnum >= 0 && cblk->right->preds &&
       cblk->right->preds->ptr == cblk && !cblk->right->preds->next) {
      *info_ptr++ = cblk->right->equivnum;
      *info_ptr++ = 2;
      *info_ptr++ = cblk->equivnum;
      *info_ptr++ = insttype;
      return;
   }

   /*
      We can avoid incrementing a counter if the target basic block has only
      one other predecessor and that predecessor's only successor is the target
      basic block.
   */
   if (cblk->right && cblk->right->equivnum >= 0 && cblk->right->preds &&
       cblk->right->preds->next && !cblk->right->preds->next->next) {
      if (cblk->right->preds->ptr == cblk)
	 tblk = cblk->right->preds->next->ptr;
      else if (cblk->right->preds->next->ptr == cblk)
	 tblk = cblk->right->preds->ptr;
      else
	 tblk = (struct bblock *) NULL;
      if (tblk && tblk->equivnum >= 0 &&
	  tblk->left == cblk->right && !tblk->right) {
	 *info_ptr++ = cblk->equivnum;
	 *info_ptr++ = 3;
	 *info_ptr++ = cblk->right->equivnum;
	 *info_ptr++ = tblk->equivnum;
	 *info_ptr++ = insttype;
	 return;
      }
   }

   /*
      Otherwise, the only way to know how often a branch is not taken is to
      place a counter in the fall-through portion of the branch.
   */
   increment_if_branch_not_taken(cblk, insttype);
}


#ifdef INTERACT
/******** New Measurment Code - added by Prasad Kulkarni *******/
int get_measure(struct freq_counts **measure_cnt, int trans_cnt)
{
  int i, is_last_function, status, max_time, store_swef;
  char cmd[MAXSTRING], line[MAXSTRING], base[MAXSTRING], perf_file[MAXSTRING];
  char expected_file[MAXSTRING], actual_file[MAXSTRING];
  char confbuf[BUFSIZ];
  struct freq_counts *freq;
  FILE *temp_fp;
  extern int cex_offset, swef, viewing, execute_status;
  extern int funcnum1, function_number, valid_analysis;
  extern char instbuf[BUFSIZ];
  extern FILE *comb_log;

  /*open the configuration file containing the execution/simulation commands*/
  sprintf(cmd, "%stest.config", pathname);
  if( (conf_fp = fopen(cmd, "r")) == NULL){
    write_record_id_to_socket(CHANGE_TEST_CONFIG_ID);
    change_test_config();
    if( (conf_fp = fopen(cmd, "r")) == NULL){
      fprintf(stderr, "Unable to create/open %s\n", cmd);
    }
  }
  else
    setbuf(conf_fp, confbuf);
  
  /*Currently measurements are only supported on the sparc solaris. Check the
    machine type for compatibility*/
  if(conf_fp && (execute_status & (EXECUTE|SIMULATE))){
  
    /*--- save the current states of the assembly file and the .trans file ---*/

    /*set flag to halt reading from the socket*/
    measure_phase = TRUE;
    store_swef = swef;
    if(execute_status & EASE)
      swef = TRUE;
    
    /*copy the instrumented assembly file to a temporary location*/
    fclose(instfp);
    sprintf(cmd, "cp %s %stemp", inst_filename, inst_filename);
    system(cmd);
    instfp = fopen(inst_filename, "a");
    setbuf(instfp, instbuf);
    
    /*save the original .transout file so that it can be restored after
      measurement phase*/
    sprintf(cmd, "cp %s%s.transout %s%stemp.transout", pathname, basename,
	    pathname, basename);
    system(cmd);
    
    /*verify that the required optimizations have been performed. Need to do
      both register assignment and fix-entry exit.*/
    save_measure_flag = measure_flag;
    measure_flag = TRUE;
    apply_epilogue_trans(&valid_analysis);
    
    /*--- dump out the transformations ---*/
    dump_trans(initial_count);
    initial_count = NULL;
    
    /*--- insert the code to do measurements ---*/
    if(execute_status & EASE)
      take_measures(execute_status, 1);
    measure_flag = save_measure_flag;
    
    /*in case we are simulating the execution, then we intend to get cycle 
      counts for only the function which we are optimizing. For this function
      we generate a flag in the assembly so that the simulator can know for
      which instructions to run the cycle-accurate simulator, and for the
      rest we can run the fast functional simulator*/
    /*if((execute_status & SIMULATE) && is_curr_fn)
      dump_rtl_asm(1, instfp, 1);
      else*/
    dump_rtl_asm(1, instfp, 0);

    /*Emit any additional declarations needed to perform measurements on this
      function*/
    if (/*(execute_status & EXECUTE) &&*/ fn && (execute_status & EASE))
      freq_info(instfp);
    do_fn_clean();
    /*clear the list of transforms*/
    clear_transform_list();
    
    /*--- do the following for each of the remaining functions ---*/
    is_last_function = FALSE;
    while(!is_last_function)
      {
	/*we do not need to store these messages in the transform list*/
	save_measure_flag = measure_flag;
	measure_flag = TRUE;
	/* read next function from the .cex file */
	is_last_function = read_next_function();
	if(fn)
	  do_fn_prereqs();
	measure_flag = save_measure_flag;

	/*apply the initial trans*/
	if(fn)
	  apply_inittrans(&initial_count, trans_cnt, APPLY_FROM_FILE);
	
	save_measure_flag = measure_flag;
	measure_flag = TRUE;
	apply_epilogue_trans(&valid_analysis);
	
	dump_trans(initial_count);
	initial_count = NULL;
	if(execute_status & EASE)
	  take_measures(execute_status, 0);
	measure_flag = save_measure_flag;

	/*if((execute_status & SIMULATE) && is_curr_fn)
	  dump_rtl_asm(1, instfp, 1);
	  else*/
	dump_rtl_asm(1, instfp, 0);
	
	if (/*!(exe_stat & SIMULATE) && */fn && (execute_status & EASE))
	  freq_info(instfp);
	do_fn_clean();
	clear_transform_list();
      }
    if(execute_status & EASE)
      end_info(instfp);
    
    /*--- execute the instrumented assembly file ---*/
    
    /*open the configuration file*/

    fclose(instfp);

    /*ignore the first line which contains the execution type*/
    fgets(cmd, MAXSTRING-1, conf_fp);
    
    /*link the instrumented assembly file*/
    fscanf(conf_fp, "link command: ");
    fgets(cmd, MAXSTRING, conf_fp);
    
    if(status = system(cmd)){
      /*write link error to log file*/
      log_error("Link Error", 10);
    }
    
    if(!status){
      
      /*get the filename for the desired and actual output files*/
      fscanf(conf_fp, "desired output file: %s\n", line);
      sprintf(expected_file, "%s%s", pathname, line);
      fscanf(conf_fp, "actual output file: %s\n", line);
      sprintf(actual_file, "%s%s", pathname, line);
      /*remove the previous output file if it exists*/
      /*if(temp_fp = fopen(actual_file, "r")){
	fclose(temp_fp);*/
      sprintf(cmd, "rm %s", actual_file);
      system(cmd);
	/*}*/
      
      /*get the execute command and the maximum time allowed for execution*/
      fscanf(conf_fp, "execute command: ");
      fgets(line, MAXSTRING, conf_fp);
      fscanf(conf_fp, "maximum time: %d\n", &max_time);
      
      /*execute the program to get the counts file*/
      line[strlen(line)-1] = '\0';
      if(execute_status & EXECUTE)
	sprintf(cmd, "ulimit -t %d; %s > %s", max_time, line, actual_file);
      else
	sprintf(cmd, "%s > %s", line, actual_file);
      if(system(cmd)/*status=0*/){
	log_error("Execute Error", 13);
      }

      /*extract the name of the count file from the execution command*/
      get_perf_file(line, perf_file, execute_status);
	
      /*comapre if current output matched the expected output*/
      sprintf(cmd, "diff %s %s", actual_file, expected_file);
      status = system(cmd);
      /*status = (status>0 || status==-1) ? -1 : 0;*/
      fprintf(stderr, "After comparing outputs, status: %d\n", status);
    }
    
    /*--- restore the system state to that before this function call ---*/
    
    /*restore the trans file*/
    sprintf(cmd, "cp %s%s.transout %s%s.trans", pathname,
	    basename, pathname, basename);
    system(cmd);
    
    /*restore the transout file*/
    sprintf(cmd, "cp %s%stemp.transout %s%s.transout", pathname,
	    basename, pathname, basename);
    system(cmd);
    
    /*flag to indicate the function initialization phase is on*/
    measure_init = 1;

    swef = store_swef;
  
    /*reset the input file offset*/
    if(fseek(stdin, cex_offset, SEEK_SET) != 0)
      {
	fprintf(stderr, "not able to fseek\n");
	exit(0);
      }
    
    /*reset the output assembly file*/
    sprintf(cmd, "mv %stemp %s", inst_filename, inst_filename);
    system(cmd);
    instfp = fopen(inst_filename, "a");
    setbuf(instfp, instbuf);
    
    /*reset funcnum which is used to instrument the measurement code*/
    funcnum1 = function_number;
    
    /*read the current function again*/
    read_next_function();
    set_function_name(fn);
    do_fn_prereqs();
    /*clear the list of transforms*/
    clear_transform_list();
    apply_inittrans(&initial_count, trans_cnt, APPLY_FROM_FILE);
    
    /*--- read measurements and store in data structure ---*/

    /*read the counts from the count file*/
    if((execute_status & EASE) && (execute_status & EXECUTE)){
      if(measure_cnt){
	freq = (struct freq_counts *)get_ins_count(top, perf_file, fn, 0);
	*measure_cnt = freq;
      }
    }
    else if(execute_status & SIMULATE){
      /*get_sim_counts(perf_file);*/
    }
    
    /*close the config file*/
    fclose(conf_fp);
    
    /*reset measure initialisation flag*/
    measure_init = FALSE;
    
    /*end measurement phase*/
    measure_phase = FALSE;
  }
  /*if the machine is not a sparc*/
  else{
    int total_changes;

    /*undo changes made during checking the current sequence*/
    total_changes = get_num_changes(get_num_transforms());
    undo_changes(total_changes - trans_cnt);
    clear_all_links(top);
    set_function_name(fn);

    /*make dummy data structure containing all zero measures*/
    if(measure_cnt && *measure_cnt){
      freq = make_dummy_counts();
      *measure_cnt = freq;
    }
    status = 0;
  }
  
  /*return the status after executing the program*/
  /*status = (status == -1) ? -1 : 0;*/
  return status;
}

/* make_dummy_counts - if for some reason the count file could not be read
 *                     then create a dummy frequency structure with all
 *                     frequency counts 0
 */
struct freq_counts *make_dummy_counts()
{
  struct freq_counts *freq, *new_ptr, *end_ptr;
  struct bblock *cblk;
  extern struct bblock *top;
  
  /*initialize the frequency count structure*/
  for(cblk=top, freq=NULL ; cblk ; cblk=cblk->down){
    
    /* allocate new record for this block frequency*/
    new_ptr = (struct freq_counts*)CALLOC(sizeof(struct freq_counts));
    new_ptr->count = 0;
    new_ptr->next = NULL;
    if(freq == NULL)
      freq = new_ptr;
    else
      end_ptr->next = new_ptr;
    end_ptr = new_ptr;
    cblk->frequency = 0;
  }
  
  return freq;
}


/*function will insert the code to take measurements at execution time and
  write this code to the output file*/
void take_measures(int exe_stat, int is_curr_fn)
{
  int store_swef;
  extern int swf;

  if (top && fn)
    {
      store_swef = swef;
      swef = TRUE;
      
     /*Before calculating the equivalence classes to instrument the function,
       we remove the basic blocks in the function to reduce the size of the
       data structures that are needed to maintain the information about the
       basic blocks in the function, and determine the impact of calling the
       "exit()" function on the control flow graph of the function*/
     (void) remv_useless_basic_blocks(1);
     reveal_exit_points(1);
     
     /*perform loop analysis on the function*/
     depth_first_search(d_f_s_count++);
     build_dominator_tree(b_d_t_count++);
     find_dominance_frontiers(f_d_f_count++);
     clear_loops();
     (void) find_loops(1);
      
     /*In order to be able to determine which registers are live at any given
       point in the function in order to ensure that the instrumented code
       does not affect any of the registers used, determine the live variable
       information for the function*/
#ifdef FILL
     live_variable_analysis(1, FALSE, swf);
#else
     live_variable_analysis(1, FALSE, FALSE);
#endif
	 
     /*Also determine which registers are used in the function*/
     registers_used();
	 
     /*Number each RTL in the function so that we can tell the RTLs that
       are part of the original function from those that are added in
       order to instrument the code*/
     number_rtls();
     
     measure_freq();

     swef = store_swef;
    }
  /*measure_flag = FALSE;*/
   
  /*dump out the instrumented code to the output file*/
  reset_info_length();
}

#endif /*INTERACT*/

#endif /*MEASURE*/
