/************************************************************************
 *                                                                      *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   *
 *     The Rector and Visitors of the University of Virginia            *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from the University of Virginia.                                *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/misc.h,v $
 $Date: 2004/01/21 18:39:59 $
 $Revision: 1.18 $
 $Name:  $

 Log of changes:

 $Log: misc.h,v $
 Revision 1.18  2004/01/21 18:39:59  jdh8d
 Latest Vista additions/changes.  Don't yell at me if it's broken/bad.. prasad wanted these changes.  I've reviewed them, and tested,  minimally.. things appear OK (for the batch compiler)

 Revision 1.17  2003/09/02 14:33:14  jdh8d
 changed include from ../sparc/regs.h to regs.h to allow makefiles to specify the path to regs.h (for tms)

 Revision 1.16  2003/08/28 18:29:35  jdh8d
 changes to get tms working; rtl deleted twice in cdmotion.c

 Revision 1.15  2003/08/01 04:40:15  jg9h
 Added more rtl list manipulation functions.
 Changed calls to prev_rtl() into up_rtl().

 Revision 1.14  2003/02/06 04:48:21  jg9h
 Merged VISTA into mainline vpo under the INTERACT preprocessor flag

 Revision 1.13  2002/11/25 14:40:32  jwd
 Made changes so EASE functionality can be built.  It had gotten out of sync
 with the rest of the compiler.  This required moving some functions from
 rtlspec.c to misc.c because the EASE library uses rtlspec.c.

 Revision 1.12  2002/08/09 19:57:19  clc5q
 Add #include of vectors.h to get dveclen, rveclen, and veclen.

 Revision 1.11  2002/07/11 15:20:48  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.10  2002/06/05 20:08:54  jdh8d
 Changes to support the stache, including routines to build a callgraph.

 Some cosmetic changes.

 Minor bug fixes.

 Changed build scripts to build the mips compiler for SimpleScalar if
 SSAS is set to "-DSS_AS"

 Revision 1.9  2002/03/29 19:54:33  rv3s
 *** empty log message ***

 Revision 1.8  2002/03/25 17:17:16  rv3s
 Added prototype for nullify_rtl

 Revision 1.7  2001/10/29 13:55:26  jdh8d
 *** empty log message ***

 Revision 1.6  2001/09/06 20:32:27  jdh8d
 *** empty log message ***

 Revision 1.5  2001/08/29 16:44:32  jdh8d
 *** empty log message ***

 Revision 1.4  2001/04/13 21:49:53  jdh8d
 *** empty log message ***

 Revision 1.3  2000/06/09 19:59:18  jwd
 Added prototype for add_rtl_with_type.

 Revision 1.2  1998/04/28 20:25:28  jwd
 Added void to function prototypes (not sure why they were missing).
 
 Deleted some commented out code.
 Revision 1.1  1998/04/03 18:15:54  jwd
 Initial revision

*************************************************************************/

#ifndef MISC_H
#define MISC_H

#define WRITE_CEX 1
#define FUNC_INFO 0

#include <stdlib.h>
#include <sys/time.h>

#include "vpo.h"
#include "vectors.h"

#ifdef DISPOSE
void *myalloc_perm (long);
void *myalloc_func (long);
void *myalloc_phase(long);
void *myalloc_trans(long size);
void *myalloc_comb (long size);
void *myalloc_profile (long size);
void dispose_perm (void);
void dispose_func (void);
void dispose_phase(void);
void dispose_trans(void);
void dispose_comb(void);
void dispose_profile(void);
unsigned int *frmalloc(int);
void myfree (char *);
unsigned int *myalloc(register int);
void dispose (register char *);
#else
void do_free(void *);
char *salloc(int);
#endif

#ifdef DISPOSE
#define SALLOC(size) myalloc_func(size)
#define CALLOC(size) ((char *) myalloc_func(size))
#define SALLOC_PERM(size) myalloc_perm(size)
#define CALLOC_PERM(size) ((char *) myalloc_perm(size))
#define SALLOC_PHASE(size) myalloc_phase(size)
#define CALLOC_PHASE(size) ((char *) myalloc_phase(size))
#define SALLOCTYPE(size, type) (freeptrs[type] ? frmalloc(type) : myalloc_func((size)>4 ? (size) : 4))
#define BALLOC (struct bblock *) SALLOCTYPE(sizeof(struct bblock), BBLOCKTYPE)
#define VALLOC ((unsigned int*) SALLOCTYPE(veclen << int2char, VECTORTYPE))
#define DALLOC ((unsigned int*)SALLOC(dveclen << int2char))
#define RALLOC ((unsigned int*)SALLOCTYPE(rveclen << int2char, RLISTTYPE))
#define MRALLOC ((unsigned int*)SALLOCTYPE(mrveclen << int2char, MRVECTORTYPE))
#define FREE(ptr) ((void)(ptr=0))
/* #define FREETYPE(ptr, type)*/

/*#define FREETYPE(ptr, type) (*((char ***) ptr) = freeptrs[type], freeptrs[type] = (char **) ptr, freetype_func(ptr,type), ptr=0)*/

#define FREETYPE(ptr, type) FREE_func(ptr, type)

extern void freetype_func(void* ptr, int type);
void FREE_func(void *ptr, int type);



#define free(ptr) {myfree(ptr),ptr=0}
#define vfree(ptr) if (ptr) {FREETYPE(ptr, VECTORTYPE); ptr=0;}
#define dfree(ptr) (ptr=0)
#define rfree(ptr) if (ptr) {FREETYPE(ptr, RLISTTYPE); ptr=0;}
#define mrfree(ptr) if (ptr) {FREETYPE(ptr, MRVECTORTYPE); ptr=0;}
#define regfree(ptr)

#else   /* not DISPOSE */
#define SALLOC(size) 		 salloc(size)
#define SALLOC_PERM(size) 		 salloc(size)
#define SALLOC_PHASE(size) 		 salloc(size)
#define CALLOC(size) ((char *)   salloc(size))
#define CALLOC_PERM(size) ((char *)   salloc(size))
#define CALLOC_PHASE(size) ((char *)   salloc(size))
#define SALLOCTYPE(size, type)   salloc(size)
#define BALLOC (struct bblock *) salloc(sizeof(struct bblock))
#define VALLOC (unsigned int *)  salloc(veclen << int2char)
#define DALLOC (unsigned int *)  salloc(dveclen << int2char)
#define RALLOC (unsigned int *)  salloc(rveclen << int2char)
#define MRALLOC (unsigned int *) salloc(mrveclen << int2char)
#define FREE(ptr) 	    {do_free((char *) ptr); ptr=0;}
#define FREETYPE(ptr, type) {do_free((char *) ptr); ptr=0;}
#define vfree(ptr) 	    {do_free((char *) ptr); ptr=0;}
#define dfree(ptr) 	    {do_free((char *) ptr); ptr=0;}
#define rfree(ptr) 	    {do_free((char *) ptr); ptr=0;}
#define mrfree(ptr) 	    {do_free((char *) ptr); ptr=0;}
#define regfree(ptr) 	    {do_free((char *) ptr); ptr=0;}

#endif  /* end if DISPOSE ... else ... */


extern int ignums;  /* total interference graph nodes */
extern int nearly_exhausted_memory;
extern struct callfunc *ext_funcs;   /* list of external function calls */
extern int ext_num;	/* number of external functions    */

void allocate_reg_bit_positions(void);
void vectorize_deads(int);
void clear_call_graph(void);
void add_func(char *);
void clear_function(void);
struct bblock *new_basic_block(char *);
void remv_basic_block(struct bblock *);
void free_basic_block(struct bblock *);
void add_predecessor(struct bblock *, struct bblock *);
void change_predecessor(struct bblock *, struct bblock *, struct bblock *);
void remv_predecessor(struct bblock *, struct bblock *);
void add_frontier_basic_block(struct bblock *, struct bblock *);
struct list *add_rtl(struct bblock *, struct list *, char *, rlist);
struct list *add_rtl_with_type(struct bblock *, struct list *, char *, rlist deads, unsigned int);
struct list *add_rtl_with_type_trans(struct bblock *, struct list *, char *, rlist deads, unsigned int);
struct list *new_rtl(struct bblock *, int, struct list *);
struct list *up_rtl(register struct list * ptr);
struct list *prev_rtl(struct list *);
struct list *next_rtl(struct list *);
void show_rtl(FILE *, char *);
struct list *copy_rtl(struct bblock *, struct list *, struct list *);
void remv_rtl(struct list *);

struct llist *newref(struct list *);
int is_present_in_ref(struct llist *lptr, struct list *ptr);
int count_ref(struct llist *lptr);
struct llist *merge_refs(struct llist *lptr1, struct llist *lptr2);
void add_to_ref(struct llist **lptr, struct list *ptr);

struct itriple *newtriple(struct setlist *);
void freetriples(void);
struct memgroup *newpartition(char *);
struct memitem *newmem(struct list *, char *, char *, int);
struct ignode *newignode(struct setlist *);
void remvignode(struct ignode *);
void iconflict(struct ignode *, struct ignode *);
void dconflict(struct ignode *, struct ignode *);
void clear_interference_graph(void);
struct bblock *getlabel(char *);
void gtrtl(char **, char **);
char *gtline(void);
char *gtlab(void);
int fmatch(char *, char *);
/* these are important when calling fmatch */
extern char* bpos, *epos;

void gtstr(char *, char *, char *);
char *gtds(char *, char *, char *);
char *sub(char *, char *, char *);
char *isset(char *, char *);
char *isaliasset(char *, char *);
int isref(char *, char *);
int isaliasref(char *, char *);
char *is_const(char *);
struct list *value_of(struct list *, char *, rlist, char *);

void nullify_rtl(struct list*);

void clear_block_info(struct bblock*);

char* get_glob_name(char*);

void update_label(struct bblock *, char *);
void update_se(struct list *, char *);
void update_rtl(struct list *, char *);
void bshow(struct bblock *cblk, int links);
void rfshow(FILE *fp, rlist vec);

#ifdef INTERACT

#define MAX_ATTEMPTS 20

void get_num(char**, int*);
void get_string(char**, char*);
void swapint(int*, int*);
void swapdouble(double*, double*);
int is_sequence_valid(int seq[], int base_opt[], int n);
int is_opt_valid(int opt, unsigned int *done_trans);
void get_random_seq(int popu_str[], struct comb_struct *comb);
int is_same_seq(int str1[], int str2[], int strlength);
void update_hashtable(struct comb_struct *comb, int sort_phases[]);
int in_hashtable(int[], int flags, int strlength, unsigned long long*,
		 unsigned long long*);
void update_succ_htable(struct comb_struct *comb, int seq[], int length,
			unsigned long long, unsigned long long);
int in_succ_hashtable(int seq[], int length,
		      unsigned long long *, unsigned long long *);
void reproduction(struct comb_struct comb, int sorted_list[]);
void crossover(struct comb_struct comb);
void gen_crossover(int base1[], int base2[], int replace1[],
		   int replace2[], int cross_pt, int length);
void mutation(struct chromosome *chrom, int base_opt[], int [], int length,
	      int phases, int, int, int);
void sort_population(struct comb_struct comb, int list[]);
unsigned long long sel_seq_measures(int seq_count, int weight);
void init_hashtables(struct comb_struct *);
int check_func_instance(unsigned long long*, unsigned long long *,
			struct Func_hash **, int[], int, int, int*);
int check_rtl_instance(struct list *rtl, unsigned int *hashvalue, unsigned int *, int, int);
int get_search_flags(struct comb_struct *comb, char *fn);
unsigned int  update_tree(struct comb_struct *comb, int opt, unsigned int curr_ptree,
			  int change, int succ_num, int succ_phases[]);
int analyse_func(int seq_count);
void do_reqd_optimization(int is_reg_assign, int is_fix_entry_exit, int *valid_analysis);
void apply_epilogue_trans(int *valid_analysis);
void get_remapped_rtl(struct list *rtl, char *rtlval, int is_remap);
int is_same_func(int *v_analysis, unsigned long long *s_measure,
		 unsigned long long *d_measure, struct Func_hash **ret_index,
		 struct Func_hash **remap_f, int succ_phases[], int succ_num, int, int);
void get_time_diff(struct timeval ctime, unsigned long long *, unsigned long long *);
void get_num(char **, int *);
void get_string(char **, char *);
void print_tree_path(int node, FILE *fp);
int do_needed_analysis(unsigned int);
void sort_ind_phases(int strlength, int flags, int base_opt[],
		     int seq[], int sort_phases[], int seq_no, int remap);
int sort_ind_phases_pass(int strlength, int seq[], int sort[]);
int sort_ind_phases_pass_RA(int strlength, int seq[], int sort[]);
int get_next_intel_seq(struct comb_struct *comb, int next_seq[],
		       int phase_indices[], int sort_phases[], int *hits);
void get_next_comb_seq(struct comb_struct *comb, int s[], int phase_indices[]);
int get_next_ex_phase(struct comb_struct *comb, int next_seq[], int length,
		       int phase_indices[], int curr_pos[], int curr_ptree);
void remap_label(char *label, int is_modify);
int read_options_from_file(char options[]);
void clear_deads();
unsigned long long get_perf_num(struct comb_struct *comb, int seq[],
				unsigned long long *s_measure,
				unsigned long long *d_measure);
int is_execute(FILE *fp);
void insert_sim_flags(char str[]);
void get_perf_file(char line[], char filename[], int exe_stat);
void add_simulator_ins(struct list *rtl, FILE *fp, int type);
void add_start_cycle_ins(FILE *fp);
void add_end_cycle_ins(FILE *fp);
int resume_execution();
int find_curr_fn(char *fn, struct comb_struct *comb);
void store_parameters(char *fn, struct comb_struct *comb);
int resume_perf_num(unsigned long long *s_measure,
		    unsigned long long *d_measure);
void log_active_seq(int seq_no, int *succ_num, int succ_phases[]);
void log_best_seq_msg(double measure, long long sec, long long usec);
void log_random_numbers();
void log_htable_msg();
void log_valid_msg(int is_valid);
void map_phases_succ(struct comb_struct *comb, int next_seq[],
		     int succ_map[], int att_seq[]);
void resume_hashvalues(unsigned int *f_hash, unsigned int *CRC_reg,
		       int *rtl_count, int is_remap);
void log_hashvalues(unsigned int f_hash, unsigned int CRC_reg,
		    int rtl_count, int is_remap);
void log_function_match_line(int count, int is_remap);
void log_succ_seq_line(int scount);
void log_error(char *error, int length);
void print_dominators_all();
void print_dominators(int block_id);
int batch_comp_opt(int start_pc);
void add_phase_simplify_controlflow(int *pc);
int get_func_hash(unsigned int *CRC_hash, unsigned int *func_hash, int is_remap, int*);
void init_profile_func(int level);
void alloc_rec(unsigned int CRC, unsigned int fhash, int rcount,
	       unsigned int r_CRC, unsigned int r_fhash);
void alloc_remap_rec(unsigned int CRC, unsigned int fhash, int rcount,
	       unsigned int r_CRC, unsigned int r_fhash);
int find_rec(unsigned int CRC, unsigned int fhash, int rcount,
	     unsigned int *r_CRC, unsigned int *r_fhash);
int find_remap_rec(unsigned int CRC, unsigned int fhash, int rcount,
		   unsigned int r_CRC, unsigned int r_fhash);
void store_profile_base(char **seq, char base[]);
void store_profile_func();
void ex_dfs_init(int pc, int *analysis);
int log_profile_func(int level, int change, int *analysis, int is_undo);
void log_dfs_func(int change, int *valid_analysis, int *pc);
void end_profile_func();
struct CFList *new_cf(int id, struct CFList *prev);
struct CFList *get_func_cf(int *is_new);
struct BlockList * update_cf_blk(struct CFList *curr_cf, struct BlockList *prev,
		 int blk, int left, int right, int status, long long count);
void remap_label2(struct list *rtl, char rtlval[], char *rtlptr);
void store_as_cex(char *store_ptr[], char store_file[], int is_file);
void store_cex_line(char *line, char **str_store, FILE *fp, int *num, int is_file);
void store_rtl(char *buf, struct list *rtl, char *reg);
void reset_kind_cost();
int do_branch_opt(int *valid_analysis);
void get_next_prob_phase(int *pc, int change, int is_first);
void read_ease_counts(struct CFList *cf);
void get_batch_exec_counts(int *rcount, int *dy_count);
void read_measure_flags();
double block_weight(struct bblock *cblk);
long long get_dy_ease_counts(struct CFList *func_cf, int new_cf);
long long block_cycles(struct bblock *cblk);
int get_phase_index(char ph);
void calc_pcon(int nread);
void init_ph_paths();

extern int rtl_id;
#endif
#endif
