/************************************************************************
 *                                                                      *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   *
 *     The Rector and Visitors of the University of Virginia            *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from the University of Virginia.                                *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/


#include "vpo.h"
#include "assign.h"
#include "misc.h"
#include "pred.h"
#ifdef VET
#include "view.h"
#endif
#include "vectors.h"

#include "predicates.h"   /* Machine dependent */

static void remove_reg_death(struct list* rtl, rlist olist);
int move_and_predicate(int invocation);


int predicate_basic_blocks(int invocation)
{
	/* number the blocks so we know forward from backwards */
	struct bblock* ptr;
	int change = FALSE;
	int i=0;
	for(ptr=top; ptr; ptr=ptr->down)
		ptr->dnum=i++;
	change |= pred_to_fallthrough(invocation);
	change |= pred_to_target(invocation);
	if(!debug_opt("no_move_and_pred_rtls"))
		change |= move_and_predicate(invocation);
	return change;
	
}

int pred_to_fallthrough(int invocation)
{

#ifndef PRED
	return 0;
#endif
#ifdef PRED


	struct bblock* cblk, *block1, *block2, *block3;
	struct list* rtl;
	int rtl_count;
	char condition[MAXSTRING];
	int change=FALSE;
	int pred_cost=0;
	int cc_set=0;
	
#ifdef VET
#ifdef INTERACT
begin_phase(PREDICATION_TRANS);
#else
begin_phase(PREDICATION_PHASE);
#endif
#endif

	for(cblk=top; cblk; cblk=cblk->next)
	{

	/* watch for very simple places to predicate

		for now, we watch for:

		if()
		{
			 block 
		}

		In rtls we expect this to look like:



		 block1: ends with conditional branch

		 block2: block with no label and block 3 is lexically below block 2
                         (i.e., block2->down == block 3)
	
		 block3: target of conditional, preds don't matter


	*/


		if(cblk->left && cblk->right)
		{
			block1=cblk;
			block2=cblk->left;
			block3=cblk->right;	
		}
		else
			goto try_next_block;

		/* does block2 fallthrough block3 */
		/* block2 can't jump to block3, if we let that happen,
			and we predicate the jump, it'll fall through
			instead of jumping */
		if(block2->down!=block3)
			goto try_next_block;

		/* oops, block 2 has too many predecessors.. the condition might not
			be set from block1 if we come in from a different path */
		if(block2->preds->next)
			goto try_next_block;


		/* we can't predicate something with a label.. we might jump here
			indirectly */
		if(block2->label && block2 != ret_block && block2->preds &&
              		(block2->preds->next || !(block2->status & FALLTHRU))) 
			goto try_next_block;


		/* get the predicate.  give up on this block if we can't find it */
		if(!get_predicate_condition(block1,condition))
			goto try_next_block;

#ifdef INTERACT
		/*fprintf(stderr, "Got predicate condition:");
		show_rtl(stderr, block1->rtlend->val);
		fprintf(stderr, "\n");
		fflush(stderr);*/
#endif
		
		pred_cost=0;
		cc_set=FALSE;
		for(rtl=block2->rtls; rtl; rtl=rtl->next)
		{
			if(!can_predicate_rtl(rtl, &cc_set, condition))
				goto try_next_block;
			pred_cost+=predication_cost(rtl);
	
			if(pred_cost>=max_pred_cost)
				goto try_next_block;
		}
#ifdef VET
		begin_trans(VET_OPT_TRANS);
#endif

		/* if we've made it here, we know that each rtl in block2 can be predicated  */
		/* and that the cost of block2 is low enough.  We go ahead and predicate the */
		/* block								     */
		for(rtl=block2->rtls; rtl; rtl=rtl->next)
		{
			predicate_rtl(rtl, condition);

			/* when we predicate an rtl, we may extend the life of any 
				destination.  We have to hunt backwards and remove deaths
				so that we don't combine things */
		
			remove_reg_death(rtl, block1->olist);
#ifdef VET
#ifndef INTERACT
			mod_rtl(rtl);
#endif
#endif
		}


		/* now, we remove the conditional jump from block1, and steal the rtls
			from block2 and add them to block1 */
		remv_links_to(block1->rtlend);
		remv_links(block1->rtlend);
		remv_rtl(block1->rtlend);
		if(block1->rtlend)
			block1->rtlend->next=block2->rtls;
		else
			block1->rtls=block2->rtls;
		if(block2->rtls)
		{
			block2->rtls->prev=block1->rtlend;
			block1->rtlend=block2->rtlend;
		}
		
		for(rtl=block2->rtls; rtl; rtl=rtl->next)
			rtl->daddy=block1;
		
#ifdef INTERACT
		/*add message to move all the rtls from block2 to block1.
		  Similar to merge basic blocks message*/
		move_rtls(block2, block1, block2->rtls, block2->rtlend);
#endif

		/* no no.. the rtls aren't part of block 2 anymore */
		block2->rtlend=block2->rtls=0;


		/* block 2 is going away, the new pred is block 1 */
	      	remv_predecessor(block2, block3);

		block1->left=block3;
#ifdef INTERACT
		mod_bb_ptrs(MOD_LEFT, block1, block1->left);
#endif		
		if(block2->left!=block3)
			block1->right=block2->left;
		else
			block1->right=block2->right;
#ifdef INTERACT
		mod_bb_ptrs(MOD_RIGHT, block1, block1->right);
#endif		

		/* if block 3 didn't fall through before, it does now */
		block3->status |= FALLTHRU;

		/* the new output order is block1 followed by block3.  block2 goes away */
		block1->down=block3;
		block3->up=block1;
#ifdef INTERACT
		mod_bb_ptrs(MOD_DOWN, block1, block1->down);
#endif		


		/* if block 2 has any other preds, we have to move them	
			to block 1 */
		if(block2->left!=block3 && block2->left)
		{
			remv_predecessor(block2,block2->left);			
			add_predecessor (block1,block2->left);
		}
		if(block2->right!=block3 && block2->right)
		{
			remv_predecessor(block2,block2->right);			
			add_predecessor (block1,block2->right);
		}
		/* take block 2 out of the list of blocks, never check it again */
		remv_basic_block(block2);

		

#ifdef VET
		end_trans();
#endif
		/* note that we've made a change */
		change=TRUE;

		/* for some reason, this block can't be predicated.. lets check out the
			next one */
		try_next_block:
			continue;
	}

#ifdef VET
#ifdef INTERACT
end_phase(PREDICATION_TRANS);
#else
end_phase(PREDICATION_PHASE);
#endif
#endif
	return change;
#endif
}

static void remove_reg_death(struct list* rtl, rlist olist)
{
	struct list* ptr=0;
	char* reg=0;
	char  prev_block_dead_regs[MAXSTRING]={0,};
	char* pbdr_ptr=0;

	if(rtl->val==0 || (rtl->type&RTLINE)!=RTLINE)
		return;

	reget(olist, prev_block_dead_regs);	

	pbdr_ptr=prev_block_dead_regs;
	while(*pbdr_ptr)
	{
		if(rexists(pbdr_ptr, rtl->deads))
			rdelete(pbdr_ptr, &rtl->deads);
		
		pbdr_ptr=skiploc(pbdr_ptr)+1;
	}


	for(ptr=rtl; ptr; ptr=ptr->prev)
	{
		if((ptr->type&RTLINE)!=RTLINE)
			continue;
		reg=rtl->val;
		while(reg=gtrg(reg))
		{
			/* defined regs only */
			if(*reg!='=')
				continue;

			if(rexists(unskiploc(reg), ptr->deads))
			{
#ifdef INTERACT
			        rlist deads;
				rcpy(&deads, ptr->deads);
#endif
				rdelete(unskiploc(reg), &ptr->deads);
#ifdef VET
#ifdef INTERACT
				mod_deads(ptr, deads, ptr->deads);
#else
				mod_deads(ptr);
#endif
#endif
				return;
			}

		}
	}	
}


int pred_to_target(int invocation)
{

#ifndef PRED
	return 0;
#endif
#ifdef PRED


	struct bblock* cblk=0, *block1=0, *block2=0, *block3=0;
	struct list* rtl=0;
	int rtl_count=0;
	char condition[MAXSTRING]={0,};
	int change=FALSE;
	int pred_cost=0;
	int cc_set=FALSE;
#ifdef INTERACT
	extern int max_pred_cost;
	int cond_jump=FALSE, uncond_jump=FALSE;	
#endif

#ifdef VET
#ifdef INTERACT
begin_phase(PREDICATION_TRANS);	
#else
begin_phase(PREDICATION_PHASE);
#endif
#endif

	for(cblk=top; cblk; cblk=cblk->next)
	{

	/* watch for very simple places to predicate

		for now, we watch for:

		if()
		{
			 block 
		}

		In rtls we expect this to look like:



		 block1: ends with condtional branch

		 block2: block with no label and target of  block 3 
	
		 block3: target of conditional from block 1 and of block2, preds don't matter


	*/


		if(cblk->left && cblk->right)
		{
			block1=cblk;
			block2=cblk->left;
			block3=cblk->right;	
		}
		else
			goto try_next_block;


#ifndef ALLOW_BACKWARDS_PRED
		/* most backward branches are taken, so, don't predicate here */
		if(block3->dnum < block1->dnum)
			goto try_next_block;
#endif
			

		/* oops, block 2 has too many predecessors.. the condition might not
			be set from block1 if we come in from a different path */
		if(block2->preds->next)
			goto try_next_block;


		/* we can't predicate something with a label.. we might jump here
			indirectly */
		if(block2->label && block2 != ret_block && block2->preds &&
              		(block2->preds->next || !(block2->status & FALLTHRU))) 
			goto try_next_block;

		/* get the predicate.  give up on this block if we can't find it */
		if(!get_predicate_condition(block1,condition))
			goto try_next_block;

		/* block2 must jump to block3, not fallthrough */
                /* (fallthrough handled by pred_to_fallthrough) */
		if(block2->left==block3 && block2->right==NULL && block2->left!=block2->down)
		{
			/* block2 unconditionally jumps to block 3 */
#ifdef INTERACT
			uncond_jump=TRUE;
#endif
		}
		else if(block2->right==block3 && block2->left!=NULL) 
		{
			/* block2 conditionally jumps to block3 */
			/* make sure the predicate used to jump to block 3 matches that */
			/* used in block 1 */
#ifdef INTERACT
			cond_jump=TRUE;
			cc_set=TRUE;
#endif
			/* if block1 and 2 jump on the same condition */
			/* we can pred block2 and use block2's jump to get to block3 */
			if(!same_condition(block1->rtlend, block2->rtlend))
				goto try_next_block;
		}
		/* otherwise, block2 goes someplace else, give up on this block */
		else
			goto try_next_block;




		pred_cost=0;
		cc_set=FALSE;
		/* intentially skip the last rtl */
		for(rtl=block2->rtls; rtl!=block2->rtlend; rtl=rtl->next)
		{
			if(!can_predicate_rtl(rtl, &cc_set, condition))
				goto try_next_block;
			pred_cost+=predication_cost(rtl);
	
			if(pred_cost>=max_pred_cost)
				goto try_next_block;
		}
#ifdef VET
		begin_trans(VET_OPT_TRANS);
#endif

		/* if we've made it here, we know that each rtl in block2 can be predicated properly  */
		/* and that the cost of block2 is low enough.  We go ahead and predicate the */
		/* block								     */
		/* intentially skip the last rtl */
		for(rtl=block2->rtls; rtl!=block2->rtlend; rtl=rtl->next)
		{
			predicate_rtl(rtl, condition);

			/* when we predicate an rtl, we may extend the life of any 
				destination.  We have to hunt backwards and remove deaths
				so that we don't combine things */
		
			remove_reg_death(rtl, block1->olist);
#ifdef VET
#ifndef INTERACT
			mod_rtl(rtl);
#endif
#endif
		}
		/* remove any deaths from block2's jump rtl */
		remove_reg_death(block2->rtlend, block1->olist);

		/* now, we remove the conditional jump from block1, and steal the rtls
			from block2 and add them to block1 */
		remv_links_to(block1->rtlend);
		remv_links(block1->rtlend);
		remv_rtl(block1->rtlend);
		if(block1->rtlend)
			block1->rtlend->next=block2->rtls;
		else
			block1->rtls=block2->rtls;
		if(block2->rtls)
		{
			block2->rtls->prev=block1->rtlend;
			block1->rtlend=block2->rtlend;
		}
		for(rtl=block2->rtls; rtl; rtl=rtl->next)
			rtl->daddy=block1;
		
#ifdef INTERACT
		/*add message to move all the rtls from block2 to block1.
		  Similar to merge basic blocks message*/
		move_rtls(block2, block1, block2->rtls, block2->rtlend);
#endif

		/* no no.. the rtls aren't part of block 2 anymore */
		block2->rtlend=block2->rtls=0;


		/* block 2 is going away, the new pred is block 1 */
	      	remv_predecessor(block2, block3);
		block1->left=block2->left;
		block1->right=block2->right;
#ifdef INTERACT
		mod_bb_ptrs(MOD_LEFT, block1, block1->left);
		mod_bb_ptrs(MOD_RIGHT, block1, block1->right);
#endif		

		/* the new output order is block1 followed by block3.  block2 goes away */
		block1->down=block2->down;
#ifdef INTERACT
		mod_bb_ptrs(MOD_DOWN, block1, block1->down);
#endif		

		/* whoa bess, block2 may be the last block in the function */
		if(block2->down)
			block2->down->up=block1;


                /* if block 2 has any other preds, we have to move them
                        to block 1 */
                if(block2->left!=block3 && block2->left)
                {
                        remv_predecessor(block2,block2->left);
                        add_predecessor (block1,block2->left);
                }
                if(block2->right!=block3 && block2->right)
                {
                        remv_predecessor(block2,block2->right);
                        add_predecessor (block1,block2->right);
                }
		if(block2->status & JUMP)
			block1->status |= JUMP;

		/* take block 2 out of the list of blocks, never check it again */
		remv_basic_block(block2);
		

#ifdef VET
		end_trans();
#endif

		/* note that we've made a change */
		change=TRUE;

		/* for some reason, this block can't be predicated.. lets check out the
			next one */
		try_next_block:
			continue;
	}

#ifdef VET
#ifdef INTERACT
end_phase(PREDICATION_TRANS);	
#else
end_phase(PREDICATION_PHASE);
#endif
#endif
	return change;
#endif
}


/* 
 * Look for branch patterns like
           L0:
                CC = ...
                PC = ..., L1
                ...
           L1:
                instrn 1
                instrn 2
                ..
                instrn m
                PC = L2

 * If the number of instrns m in the target L1 is small enough and 
 * all of them can be predicated, they can be move to L0 (which is not 
 * an actual branch label, just a place marker). This would get rid of 
 * the branch to L1. The branch at L0 will be changed to PC = ....,L2
 */

int move_and_predicate(int invocation)
{
  struct bblock* cblk, *block2, *block3;
  int change = FALSE;
#ifdef PRED
#ifdef VET
#ifdef INTERACT
begin_phase(PREDICATION_TRANS);
#else
begin_phase(PREDICATION_PHASE);
#endif
#endif

  for (cblk = top; cblk; cblk = cblk->down)
    {
      if((block2=cblk->left) && (block3=cblk->right)) {
        /* Check to see if left is the fall through */
        if (block2 != cblk->down)
           continue;

           /* More than one pred for block3. Can't do this optimization */
           if (block3->preds && block3->preds->next)
              continue;


        /* Now look at left and right to see if they branch to the same target */
        {
          char condn[MAXSTRING];
          int cc_set = FALSE;
          struct list *rtl,*branch_rtl;
          int predcost = 0;

          /* Get the branch condition from cblk */
          if (!get_predicate_condition(cblk,condn))
             goto try_next_block;


          /* Reverse the condition because we want to see if it is
           * predicatable with the final condition we will use
           */
          reverse_pred_condition(condn);

          /* Check if each instrn in block3 can be predicated */
          for (rtl = block3->rtls; rtl ; rtl = rtl->next) {
            if (!can_predicate_rtl(rtl,&cc_set,condn)) {
               goto try_next_block;
            }
            predcost += predication_cost(rtl);
            if (predcost > max_pred_cost)
               goto try_next_block;
          }
          if (cc_set)
             goto try_next_block;

          /* The whole block is predicatable. Now check to see if
             it is safe to move it to cblk */
          branch_rtl = cblk->rtlend;

          for (rtl = block3->rtls; rtl; rtl = rtl->next) {
            char *a;

            /* Ignore the last one if it is a branch */
            if (rtl->next == NULL)
              if (cnts(rtl) && is_pc(rtl->val))
                break;

            /* Do not move calls; can trash condition codes */
            if (cnts(rtl) && is_call(rtl->val))
                goto try_next_block;

            if (a = cnts(rtl))
              while (a = gtrg(a)) {
                /* Cannot move a set past a use and vice-versa */
                if (isaliasset(branch_rtl->val,unskiploc(a)))
                   goto try_next_block;

                if ((*a == '=')  && (isaliasref(branch_rtl->val,unskiploc(a))))
                  goto try_next_block;
              }
            if (a = rtl->se)
              while (a = gtrg(a)) {
                /* Can move a set past a use and vice-versa */
                if (isaliasset(branch_rtl->val,unskiploc(a)))
                  goto try_next_block;

                if ((*a == '=')  && (isaliasref(branch_rtl->val,unskiploc(a))))
                  goto try_next_block;
              }
          }

          /* If we ever get here, it is safe to move the rtls from block3 (except the branch instrn) to cblk */
#ifdef VET
          begin_trans(VET_OPT_TRANS);
#endif
          for (rtl = block3->rtls;rtl; ){
            struct list *rtl1;

            if (cnts(rtl) && is_pc(rtl->val))
              break;

            predicate_rtl(rtl,condn);
            block3->rtls = rtl->next;
            rtl1 = rtl->next;
            if (rtl1)
              rtl1->prev = rtl->prev;

            /* Insert into cblk */
            rtl->next = cblk->rtlend;
            rtl->prev = cblk->rtlend->prev;
            if (cblk->rtlend->prev)
              cblk->rtlend->prev->next = rtl;
            cblk->rtlend->prev = rtl;

            if (cblk->rtlend == cblk->rtls)
              cblk->rtls = rtl;

            rtl->daddy = cblk;
#ifdef INTERACT
	    move_rtl(rtl, block3, cblk->rtlend, cblk);
#endif	    
	/* zap the dead list because it may be wrong */
	    rtl->deads=rinit();
            remv_links(rtl);
            rtl = rtl1;
            change |= TRUE;

            if (rtl == (struct list*) NULL)
              block3->rtlend = (struct list*) NULL;
          }

#ifdef VET
          end_trans();
#endif

        }

      }
    try_next_block:
      ;
    }
#ifdef VET
#ifdef INTERACT
end_phase(PREDICATION_TRANS);
#else
end_phase(PREDICATION_PHASE);
#endif
#endif
#endif /* PRED */
  return change;

}

