#ifndef READLP_H
#define READLP_H

typedef struct locality_profile locality_profile_t;
typedef struct locality_profile_range locality_profile_range_t;
typedef struct locality_profile_variable locality_profile_variable_t;

struct locality_profile_range
{
	long long start_ref;
	long long last_ref;
	int num_refs;
};

struct locality_profile_variable
{
	char *name;
	int addr;	
	int size;

	int total_refs;

	int num_ranges;
	int array_size;


	/* the partition assigned, NULL means none */
	part_list_t* assigned_part;

	locality_profile_range_t* ranges;

	/* a boolean that's true iff this variable is being compiled now */  
	int compiling;

	/* a list of rtls that correspond to the statements needed to emit this variable */
	/* note that this may be null for local variables */
	struct list* rtls;
};

struct locality_profile
{
	int array_size;
	locality_profile_variable_t* variables;
};


typedef struct my_cache_user_data my_cache_user_data_t;
struct my_cache_user_data
{
	float estimated_miss_latency;
};

typedef struct my_part_user_data my_part_user_data_t;
struct my_part_user_data
{
	float estimated_average_access_time;
	int assoc;
	int part_num;
};


int read_profile(FILE* fp, locality_profile_t **p);
int calc_tcf(locality_profile_t* p, int var_a_index, int var_b_index);
void dump_assignments(locality_profile_t* p);
void calc_locality(locality_profile_t* p, partition_description_t* pd);
void read_locality(char* lp_fn);

#ifndef STANDALONE_PARTITIONING
void drive_partitioning(char* parts_fn);
#endif

extern int total_number_of_partitions;



#endif


