/*******************************************************************
 * rename.c : Register renaming code
 ******************************************************************/

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/time.h>

#include "vpo.h"
#include "arena.h"
#include "remjump.h"
#include "view.h"
#include "misc.h"

/* stores register set-use information for each valid RTL in a basic block */
struct RENAME_INFO{
  rlist avail; /*registers available at this RTL*/
  rlist rset;  /*registers previously used in this block*/
};
struct RENAME_INFO ren_info[MAX_RTL];

/*
 * local_register_rename - perform register renaming on a basic block.
 *                         Only successful if performed after register
 *                         assignment. Requires live and dead variable
 *                         analysis to be accurate at this point.
 */
int local_register_rename()
{
  char *rtlin, *c;
  int flag, rn;
  struct bblock *cblk;
  struct list *crtl, *rtlptr;
  rlist avail_regs, regs, regs_live_entry, regs_live_exit;
  extern struct bblock *top;

  /*check if register assignment done. Return if not done*/
  flag = FALSE;
  for(cblk=top ; cblk ; cblk=cblk->down){
    for(rtl=cblk->rtls ; rtl ; rtl=rtl->next){
      if(c = cnts(rtl)){
	while(c = gtrg(c)){
	  if(pseudoreg(unskiploc(c))){
	    fprintf(stderr, "local_register_rename called before register assignment\n");
	    return 0;
	  }
	  break;
	}
	flag = TRUE;
      }
      if(flag)
	break;
    }
    if(flag)
      break;
  }

  /*we can now proceed to perform the real optimization*/
  /*With local register renaming the following registers cannot be renamed:
    1. Registers live at the start of a block, until they die
    2. The last set of registers live at the end of the block
    3. Special purpose registers such as function parameters passed in
       registers
  */

#ifdef INTERACT
  begin_phase(REG_RENAME_PHASE);
#endif

  /*initialize register vectors*/
  avail_regs = rinit();
  regs = rinit();
  regs_used = rinit();
  regs_live_entry = rinit();
  regs_live_exit = rinit();

  for(cblk=top ; cblk ; cblk=cblk->down){

    /*initialize the RENAME_INFO structure for this block, mark all possible
      registers to be available for renaming at each RTL*/
    rn = 0;
    for(crtl=top->rtls ; crtl ; crtl=crtl->next){
      if(cnts(crtl)){
      }
    }

    /*mark the registers live coming into the block as not available for
      renaming, until the registers die*/
    for(crtl=top->rtls ; crtl ; crtl=crtl->next){
      if(cnts(crtl)){
      }
    }

    /*search for calls. For each call instruction, search backwards for
      assignments to its arguments (those that are passed in registers), and
      mark those registers as not available for renaming since their last set*/
    for(crtl=top->rtls ; crtl ; crtl=crtl->next){
      if(cnts(crtl)){
	if(is_call(crtl->val)){
	}
      }
    }

    /*for registers live leaving the block, mark those as unavailable for
      renaming since their last set*/


    /*start renaming the registers. Modify the register live ranges as we
      move along*/
    for(crtl=top->rtls ; crtl ; crtl=crtl->next){
      if(cnts(crtl)){
      }
    }
  }   
  
#ifdef INTERACT
  end_phase(REG_RENAME_PHASE);
#endif  
}

