/************************************************************************
 *                                                                      *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   *
 *     The Rector and Visitors of the University of Virginia            *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from the University of Virginia.                                *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/

#ifndef _RTLSPEC_H
#define _RTLSPEC_H

/* get machine-dependent register size information */
#include "regs.h"

/* fix these */

#define MAXRVEC    8192      /* Maximum number of different registers */
#define MAPREGS    0x200000   /* Number of register map slots          */
#define GLOBALS    0x3FFFF    /* Characters reserved for global ids    */

#ifndef SAFE_ISREGISTER
#define isRegister(s) (((*(s) & 0xc0) == 0xc0) && (*((s)+1)&0x80))
#endif

/* already fixed */

#define CHARSIZE	(sizeof(char))
#define CHARBITS	(CHARSIZE*8)
#define USEABLECHARBITS (CHARBITS-1)

#ifndef MAXLOCLENGTH
#define MAXLOCLENGTH 	(2)
#endif

#ifndef REGLOCLENGTH
#define REGLOCLENGTH 	(2)
#endif

#ifndef GLOBALLOCLENGTH
#define GLOBALLOCLENGTH (2)
#endif

#ifndef LOCALLOCLENGTH
#define LOCALLOCLENGTH 	(2)
#endif

/* some macros for registers */

#define BUILDREG(t,n) 	(getReg((t), (n)))

#define MAXREGBITS 	((REGLOCLENGTH-1)*7 + 6)
#define REGTYPEBITS 	(4)

#define MAXREGNUM 	((1<<(MAXREGBITS-REGTYPEBITS))-1)
#define MAXREGTYPENUM 	((1<<(REGTYPEBITS))-1)
#define MINREGNUM 	(0)
#define MINREGTYPENUM 	(0)

#define REGISTERIDENTMASK (0x1)
#define REGISTERIDENTSIZE (1)

#define regtype(s) 	(getRegType((s)))
#define regnum(s)  	(getRegNum ((s)))
#define isreg(s) 	(isRegister((s)))
#define isrg(s) 	(isRegister((s)))
#define isareg(s)	(isRegister(s) && *skiploc(s)==0)

/* global location macros */

#define BUILDGLOB(n)		(getGlobal((n)))

#define MAXGLOBBITS	((GLOBALLOCLENGTH-1)*7 + 5)
#define MINGLOBNUM  	(0)
#define MAXGLOBNUM 	((1<<MAXGLOBBITS)-1)

#define GLOBALIDENTMASK	 	(0x0) 
#define GLOBALIDENTMASKSIZE	(2) 

#define isglob(a)	(isGlobal((a)))

/* local location macros */

#define BUILDLOCAL(n)        (getLocal((n)))

#define MAXLOCALBITS    ((LOCALLOCLENGTH-1)*7 + 5)
#define MINLOCALNUM     (0)
#define MAXLOCALNUM     ((1<<MAXGLOBBITS)-1)

#define LOCALIDENTMASK 	 	(0x1)
#define LOCALIDENTMASKSIZE 	(2)

#define islocal(a)	(isLocal((a)))

/* Use in the register implementation */

#ifndef fatal
#define fatal(arg) (fprintf(stderr, "Fatal error occurred at %d of %s: %s\n", \
		__LINE__, __FILE__, (arg)), exit(1);)
#endif

#ifdef SAFE_ISREGISTER
#define skiploc(a) (isRegister((a)) ? (a)+REGLOCLENGTH    : 		\
		   (isGlobal((a))   ? (a)+GLOBALLOCLENGTH : 		\
		   (isLocal((a))    ? (a)+LOCALLOCLENGTH  : (a)+1 )))

#define unskiploc(a) (isRegister((a)-REGLOCLENGTH) ?(a)-REGLOCLENGTH   :\
		     (isGlobal((a)-GLOBALLOCLENGTH)?(a)-GLOBALLOCLENGTH:\
		     (isLocal((a)-LOCALLOCLENGTH)  ?(a)-LOCALLOCLENGTH :(a)-1)))

#define skiplocOr2(a) (isRegister((a)) ? (a)+REGLOCLENGTH    : 		\
		   (isGlobal((a))   ? (a)+GLOBALLOCLENGTH : 		\
		   (isLocal((a))    ? (a)+LOCALLOCLENGTH  : (a)+2 )))

#else /* !SAFE_ISREGISTER */

#define skiploc(a)    ((((*(a))&0x80) && ((*((a)+1))&0x80))  ? (a)+REGLOCLENGTH : (a)+1)
#define skiplocOr2(a) ((((*(a))&0x80) && ((*((a)+1))&0x80))  ? (a)+REGLOCLENGTH : (a)+2)
#define unskiploc(a)  ((*((a)-REGLOCLENGTH))&128 ? ((a)-REGLOCLENGTH) : (a)-1)

#endif

#define isloc(a) (isRegister((a)) || isGlobal((a)) || isLocal((a)))

#define copyloc(a,b) (copyLocation((a),(b)))
#define cmploc(a,b) (compareLocation((a),(b)))

#define copylocOr2(a,b) ((!isloc((b)))   ? (void)( 	(*  (char*)(a)    =(*(char*)(b))), (*((((char*)(a))+1))=(*(((char*)(b))+1)))  ) :  (copyLocation((a),(b))))

#define cmplocOr2(a,b) (!isloc((b)) ? (!(((*(char*)(a))==(*(char*)(b))) && ((*(char*)(a)+1)==(*(char*)(b)+1)))) : compareLocation((a),(b)))


#define skipmangledreg(a,mangle) 					     \
									     \
 	((isMangledReg((a),(mangle))) ? ((a)+REGLOCLENGTH+strlen((mangle))) :\
		((char*)fprintf(stderr,"Not mangled register"))) 

#define unskipmangledreg(a,mangle)  				  \
 	(isMangledReg((a)-REGLOCLENGTH-strlen((mangle)),(mangle)) \
		? (a)-REGLOCLENGTH-strlen((mangle)) :		  \
		((char*)fprintf(stderr,"Not mangled register"))) 


void mangleReg(char* dst, char *src, char* mangler);
void demangleReg(char* dst, char* src, char* mangler);
char* isMangledReg(char* src, char*  mangler);
char* getReg(int type, int number);
#ifdef SAFE_ISREGISTER
int isRegister(char* location);
#endif
int getRegType(char* t);
int getRegNum(char* s);
char* getGlobal(int number);
int isGlobal(char* location);
int getGlobalNum(char* s);
char* getLocal(int number);
int isLocal(char* location);
int getLocalNum(char* s);
void copyLocation(char* a, char* b);
int compareLocation(char* a, char* b);
void locString(char* dst, char* src);

#endif
