
/************************************************************************/
/*                                                                      */
/* Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   */
/*     The Rector and Visitors of the University of Virginia            */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from the University of Virginia.                                */
/* This code is distributed WITHOUT ANY WARRANTY.  No claims are        */
/*      made as to whether it serves any particular purpose or even     */
/*      works at all.                                                   */
/*                                                                      */
/************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/scan.c,v $
 $Date: 2003/02/06 04:46:29 $
 $Revision: 1.11 $
 $Name:  $

 Log of changes:

 $Log: scan.c,v $
 Revision 1.11  2003/02/06 04:46:29  jg9h
 Merged VISTA into mainline vpo under the INTERACT preprocessor flag

 Revision 1.10  2002/07/17 22:14:41  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.9  2002/06/05 20:08:54  jdh8d
 Changes to support the stache, including routines to build a callgraph.

 Some cosmetic changes.

 Minor bug fixes.

 Changed build scripts to build the mips compiler for SimpleScalar if
 SSAS is set to "-DSS_AS"

 Revision 1.8  2001/10/09 19:41:43  jdh8d
 *** empty log message ***

 Revision 1.7  2001/05/28 20:30:21  jdh8d
 *** empty log message ***

 Revision 1.6  2001/05/28 15:44:48  jdh8d
 *** empty log message ***

 Revision 1.5  2001/05/23 20:28:21  jdh8d


 Re-worked all of the backend to support variable
 sized registers.  IT IS  NO LONGER SAFE TO ASSUME
 THAT REGISTERS ARE TWO BYTES.  Mips currently has
 4-byte registers, others will follow (or be destroyed!)

 Revision 1.4  1998/04/29 18:54:07  jwd
 Added vpo.h header file.
 Revision 1.3  1998/04/03 18:20:01  jwd
 First ANSI C version.

*************************************************************************/

/*------------------------------------------------
   History of changes
   1/12/98: Function prototyping modified to be more modular and ANSI-C
       compatible
--------------------------------------------------*/


#include <stdio.h>

#include "vpo.h"
#include "vpo2.h"
#include "combine.h"

/* should include this last, the defines in here may muck with other
	things, eg:  sparcs #define b 98 because b is a token interfereces
	with rtlspec.h's use of "b" as a parameter name
*/
#include "y.tab.h" /* Machine dependent?*/
#include "scan.h"
#include "scan2.h"


int yylength;		/* Indicates the length of the last label or constant
			   token */

/*
 * yylex - return next token from the input line
 */
int yylex() {

   register char c, reg, *p;
   register int token, length;
   extern int check_legal;

   switch (((int) (c = *linptr++)) & 0xff) {

      case 0x00:
	 /* end of line */
	 return (0);

      case 0x01:
      case 0x02:
      case 0x03:
      case 0x04:
      case 0x05:
      case 0x06:
      case 0x07:
      case 0x08:
      case 0x09:
      case 0x0a:
      case 0x0b:
      case 0x0c:
      case 0x0d:
      case 0x0e:
      case 0x0f:
      case 0x10:
      case 0x11:
      case 0x12:
      case 0x13:
      case 0x14:
      case 0x15:
      case 0x16:
      case 0x17:
      case 0x18:
      case 0x19:
      case 0x1a:
      case 0x1b:
      case 0x1c:
      case 0x1d:
      case 0x1e:
      case 0x1f:
	 /* control character */
	 (void) fprintf(stderr, "vpo (yylex): control character encountered\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);

      case 0x20:
      case 0x21:
      case 0x22:
      case 0x23:
      case 0x24:
      case 0x25:
      case 0x26:
      case 0x27:
      case 0x28:
      case 0x29:
      case 0x2a:
      case 0x2b:
      case 0x2c:
      case 0x2d:
      case 0x2e:
      case 0x2f:
	 /* !"#$%&'()*+,-./ character */
	 return ((int) c);

      case 0x30:
      case 0x31:
      case 0x32:
      case 0x33:
      case 0x34:
      case 0x35:
      case 0x36:
      case 0x37:
      case 0x38:
      case 0x39:
	 /* 0123456789 character */

	 /*
	    A lone character by itself is always returned as its own token
	 */
	 if (((reg = *linptr) > '9' || reg < '0') && reg != '.' && reg != 'e')
	    return ((int) c);

	 /*
	    Start by assuming that we have located an integer constant unless
	    we encounter the decimal point or an exponentiation symbol which
	    tells us otherwise.
	 */
	 p = linptr;
	 yylval.str_ptr = p - 1;
	 length = 1;
	 if (reg != '.' && reg != 'e') {
	    do {
	       length++;
	       reg = *++p;
	    } while (reg <= '9' && reg >= '0');
	 }
	 if (reg != '.' && reg != 'e') {
	    yylength = length;
	    linptr = p;
	    return (CON);
	 }

	 /*
	    If we get to this point, then we are parsing a floating point
	    constant, so we scoop up the decimal and/or exponent parts and
	    return.
	 */

#ifdef FCON
	 if (reg == '.') {
	    do {
	       length++;
	       reg = *++p;
	    } while (reg <= '9' && reg >= '0');
	 }
	 if (reg == 'e') {
	    length++;
	    reg = *++p;
	    if (reg == '+' || reg == '-') {
	       length++;
	       reg = *++p;
	    }
	    do {
	       length++;
	       reg = *++p;
	    } while (reg <= '9' && reg >= '0');
	 }
	 yylength = length;
	 linptr = p;
	 return (FCON);
#else
	 (void) fprintf(stderr, "vpo (yylex): illegal floating point token\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0x3a:
      case 0x3b:
      case 0x3c:
      case 0x3d:
      case 0x3e:
      case 0x3f:
      case 0x40:
	 /* :;<=>?@ characters */
	 return ((int) c);

      case 0x41:
      case 0x42:
      case 0x43:
      case 0x44:
      case 0x45:
      case 0x46:
      case 0x47:
      case 0x48:
      case 0x49:
      case 0x4a:
      case 0x4b:
      case 0x4d:
      case 0x4e:
      case 0x4f:
      case 0x50:
      case 0x51:
      case 0x52:
      case 0x53:
      case 0x54:
      case 0x55:
      case 0x56:
      case 0x57:
      case 0x58:
      case 0x59:
      case 0x5a:
	 /* ABCDEFGHIJKMNOPQRSTUVWXYZ character */
	 reg = *linptr++;
	 if (token = lu[((((int) c) & 0x1f) << 5) | (((int) reg) & 0x1f)])
	    return (token);
	 (void) fprintf(stderr, "vpo (yylex): %c%c is not a valid token\n",
			c, reg);
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);

      case 0x4c:
	 /* L character */
	 if ((reg = *(p = linptr++)) <= '9' && reg >= '0') {
	    yylval.str_ptr = p - 1;
	    length = 1;
	    do {
	       length++;
	    } while ((reg = *++p) <= '9' && reg >= '0');
	    yylength = length;
	    linptr = p;
	    return (LBL);
	 }
	 if (token = lu[((((int) c) & 0x1f) << 5) | (((int) reg) & 0x1f)])
	    return (token);
	 (void) fprintf(stderr, "vpo (yylex): %c%c is not a valid token\n",
			c, reg);
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);

      case 0x5b:
      case 0x5c:
      case 0x5d:
      case 0x5e:
      case 0x5f:
      case 0x60:
      case 0x61:
      case 0x62:
      case 0x63:
      case 0x64:
      case 0x65:
      case 0x66:
      case 0x67:
      case 0x68:
      case 0x69:
      case 0x6a:
      case 0x6b:
      case 0x6c:
      case 0x6d:
      case 0x6e:
      case 0x6f:
      case 0x70:
      case 0x71:
      case 0x72:
      case 0x73:
      case 0x74:
      case 0x75:
      case 0x76:
      case 0x77:
      case 0x78:
      case 0x79:
      case 0x7a:
      case 0x7b:
      case 0x7c:
      case 0x7d:
      case 0x7e:
      case 0x7f:
	 /* [\]^_`abcdefghijklmnopqrstuvwxyz{|}~<DEL> character */
	 return ((int) c);

      case 0x80:
      case 0x81:
      case 0x82:
      case 0x83:
      case 0x84:
      case 0x85:
      case 0x86:
      case 0x87:
      case 0x88:
      case 0x89:
      case 0x8a:
      case 0x8b:
      case 0x8c:
      case 0x8d:
      case 0x8e:
      case 0x8f:
      case 0x90:
      case 0x91:
      case 0x92:
      case 0x93:
      case 0x94:
      case 0x95:
      case 0x96:
      case 0x97:
      case 0x98:
      case 0x99:
      case 0x9a:
      case 0x9b:
      case 0x9c:
      case 0x9d:
      case 0x9e:
      case 0x9f:
	 /* global identifier */
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	 return (GLOBAL);

      case 0xa0:
      case 0xa1:
      case 0xa2:
      case 0xa3:
      case 0xa4:
      case 0xa5:
      case 0xa6:
      case 0xa7:
      case 0xa8:
      case 0xa9:
      case 0xaa:
      case 0xab:
      case 0xac:
      case 0xad:
      case 0xae:
      case 0xaf:
      case 0xb0:
      case 0xb1:
      case 0xb2:
      case 0xb3:
      case 0xb4:
      case 0xb5:
      case 0xb6:
      case 0xb7:
      case 0xb8:
      case 0xb9:
      case 0xba:
      case 0xbb:
      case 0xbc:
      case 0xbd:
      case 0xbe:
      case 0xbf:
	 /* local identifier */
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	 return (LOCAL);

      case 0xc0:
      case 0xc1:
      case 0xc2:
      case 0xc3:

#ifdef REG0
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REG0);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type 0 not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0xc4:
      case 0xc5:
      case 0xc6:
      case 0xc7:

#ifdef REG1
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REG1);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type 1 not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0xc8:
      case 0xc9:
      case 0xca:
      case 0xcb:

#ifdef REG2
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REG2);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type 2 not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0xcc:
      case 0xcd:
      case 0xce:
      case 0xcf:

#ifdef REG3
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REG3);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type 3 not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0xd0:
      case 0xd1:
      case 0xd2:
      case 0xd3:

#ifdef REG4
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	 return (REG4);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type 4 not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0xd4:
      case 0xd5:
      case 0xd6:
      case 0xd7:

#ifdef REG5
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REG5);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type 5 not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif


      case 0xd8:
      case 0xd9:
      case 0xda:
      case 0xdb:

#ifdef REG6
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REG6);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type 6 not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0xdc:
      case 0xdd:
      case 0xde:
      case 0xdf:

#ifdef REG7
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REG7);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type 7 not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0xe0:
      case 0xe1:
      case 0xe2:
      case 0xe3:

#ifdef REG8
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REG8);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type 8 not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0xe4:
      case 0xe5:
      case 0xe6:
      case 0xe7:

#ifdef REG9
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REG9);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type 9 not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0xe8:
      case 0xe9:
      case 0xea:
      case 0xeb:

#ifdef REGA
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REGA);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type A not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0xec:
      case 0xed:
      case 0xee:
      case 0xef:

#ifdef REGB
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REGB);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type B not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0xf0:
      case 0xf1:
      case 0xf2:
      case 0xf3:

#ifdef REGC
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REGC);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type C not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0xf4:
      case 0xf5:
      case 0xf6:
      case 0xf7:

#ifdef REGD
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REGD);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type D not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0xf8:
      case 0xf9:
      case 0xfa:
      case 0xfb:

#ifdef REGE
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REGE);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type E not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      case 0xfc:
      case 0xfd:
      case 0xfe:
      case 0xff:

#ifdef REGF
	 yylval.str_ptr = linptr - 1;
	 linptr=skiploc(linptr-1);
	/* this looks like a reg, verify that it is */
	 if(!isRegister(unskiploc(linptr))) {erflag=1;return 0;}
	 return (REGF);
#else
	 (void) fprintf(stderr, "vpo (yylex): register type F not declared\n");
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
#endif

      default:
	 (void) fprintf(stderr, "vpo (yylex): bad token %x\n",
			(int) (unsigned char) c);
	 if(check_legal)
	   return(ERROR);
#ifdef INTERACT
	 if (swq) return (ERROR);
#endif
	 exit(1);
   }

#ifdef lint
   return (0);
#endif
}
