
/************************************************************************
 *                                                                      *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   *
 *     The Rector and Visitors of the University of Virginia            *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from the University of Virginia.                                *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/schedule.c,v $
 $Date: 2003/09/15 14:47:28 $
 $Revision: 1.14 $
 $Name:  $

 Log of changes:

 $Log: schedule.c,v $
 Revision 1.14  2003/09/15 14:47:28  clc5q
 Ran indent tool over file and fixed up a few cosmetic details by hand.

 Revision 1.13  2003/03/22 03:52:08  clc5q
 Add assertions in depends() rather than permit core dumps on bad RTLs.

 Revision 1.12  2003/02/06 04:46:29  jg9h
 Merged VISTA into mainline vpo under the INTERACT preprocessor flag

 Revision 1.11  2002/07/17 22:14:41  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.10  2002/06/05 20:08:54  jdh8d
 Changes to support the stache, including routines to build a callgraph.

 Some cosmetic changes.

 Minor bug fixes.

 Changed build scripts to build the mips compiler for SimpleScalar if
 SSAS is set to "-DSS_AS"

 Revision 1.9  2001/05/23 20:28:21  jdh8d


 Re-worked all of the backend to support variable
 sized registers.  IT IS  NO LONGER SAFE TO ASSUME
 THAT REGISTERS ARE TWO BYTES.  Mips currently has
 4-byte registers, others will follow (or be destroyed!)

 Revision 1.8  2001/04/26 19:03:53  jdh8d

 Changes for first arm version of scheduling.

 Revision 1.7  2000/06/09 19:43:33  jwd
 Changed calls to VET to indicate which phases are mandatory and
 which are optional.

 Revision 1.6  1999/03/05 17:01:50  jwd
 Fixed a comment that had been messed up by reformating.
 Revision 1.5  1999/01/06 15:51:57  jwd
 Added VET directives.
 Revision 1.4  1998/04/29 19:18:54  jwd
 Changed FT to FALLTHRU. This was done to avoid a conflict
 with scan.h.
 Revision 1.3  1998/04/03 18:20:17  jwd
 First ANSI C version.

*************************************************************************/

/*------------------------------------------------
    History of Changes
    ------------------

    1/13/98: Function prototyping updated to ANSI standard and modularization
              increased through use of header files.
------------------------------------------------*/

#if SCHED || FILL
#define SCHEDULE 1

#include <stdio.h>
#include <assert.h>

#include "vpo.h"
#include "vpo2.h"
#include "rtlspec.h"
#include "assign.h"
#include "combine.h"
#include "controlflow.h"
#include "links.h"
#include "misc.h"
#include "schedule.h"

#include "instdef.h"    /* Machine dependent */
#include "regs.h"  /* Machine Dependent */
#include "sched.h" /* Machine dependent */

#ifdef FILL
#ifdef MEASURE
#include "vectors.h"
#endif
#endif

/*
 * This file contains routines which deal with the scheduling of instructions
 * and filling delay slots.
 */

#ifdef VET
#include "view.h"
static rlist rold;
#endif

#ifdef SCHED
#include <string.h>
static struct list *issued;     /* pointer to first RTL in the issued set   */
static struct list *tailissued; /* pointer to last RTL in the issued set    */

static int pready = TRUE;       /* it is okay to re-print a new ready set    */

/*
 * instruction_scheduling - perform instruction scheduling
 */
void instruction_scheduling(int invocation)
{
   register struct list *ptr, *nptr;
   register struct bblock *cblk;
   struct list *ready;

#ifdef VET
#ifdef INTERACT
   begin_phase(INST_SCHED_TRANS);
#else
   begin_phase(INSTSCHED_PHASE);
#endif
#endif

   /*
    * Ensure that the kind of all instructions has been determined.  If not,
    * call the machine description to obtain the kind.  Delete instructions
    * that have no cost, as they are likely to cause trouble. Determine the
    * sets and uses information for each RTL.
    */

#ifdef VET
   /*
    * Depend_setup calls set() which changes the side_effect
    * through yyparse(). 
    */
   begin_trans(VET_OPT_TRANS);
#endif
   depend_setup();
#ifdef VET
   end_trans();
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_instruction_scheduling", invocation))
      dump_function_i("before instruction_scheduling()", invocation,
                      SHOW_LABELS | SHOW_SUCCS | SHOW_PREDS | SHOW_NONRTLS);

   /*
    * Perform instruction scheduling on each basic block in the function
    */
   for (cblk = top; cblk; cblk = cblk->down) {

      if (swd
          && debug_opt_i_l("instruction_scheduling", invocation,
                           cblk->label)) {
         (void) fprintf(stderr, "Scheduling basic block: ");
         if (cblk->label)
            (void) fprintf(stderr, "%s\n", cblk->label);
         else
            (void) fprintf(stderr, "{NULL}\n");
         pready = TRUE;
      }

      /*
       * Initialize the issued set
       */
      issued = tailissued = (struct list *) NULL;

      /*
       * For each RTL, determine its sets, uses and memory dependencies.  Issue
       * all non-RTLs immediately in the same order that they were encountered
       * in the basic block.
       */
      for (ptr = cblk->rtls; ptr; ptr = nptr) {
         nptr = ptr->next;
         if (!cnts(ptr)
             && (!(ptr->type & ASMLINE) || ptr->type & DELAYLINE
                 || ptr->type & FUNCNAME))
            issue_non_rtl(cblk, ptr);
      }

      /*
       * If the basic block that follows this one contains only this basic
       * block as its predecessor, then go ahead and treat all of the items in
       * this and the following block as belonging to the same block. If
       * anything, this will increase the number of RTLs available for
       * scheduling, which never hurts and sometimes helps.
       */
      while (cblk->left && !cblk->right && !(cblk->status & JUMP)
             && cblk->left == cblk->down && cblk->down->preds
             && cblk->down->preds->ptr == cblk && !cblk->down->preds->next) {

         /*
          * Place all of the RTL from the previous basic block into the current
          * basic block.
          */
         for (ptr = issued; ptr; ptr = ptr->next)
            ptr->daddy = cblk->down;
         for (ptr = cblk->rtls; ptr; ptr = ptr->next)
            ptr->daddy = cblk->down;

         /*
          * Issue all non-RTLs immediately in the same order that they were
          * encountered in the basic block.
          */
         for (ptr = cblk->down->rtls; ptr; ptr = nptr) {
            nptr = ptr->next;
            if (!cnts(ptr)
                && (!(ptr->type & ASMLINE) || ptr->type & DELAYLINE))
               issue_non_rtl(cblk->down, ptr);
         }

         /*
          * Merge the RTLs from both basic blocks into the second basic block
          */
         if (cblk->rtlend) {
            if (cblk->rtlend->next = cblk->down->rtls)
               cblk->down->rtls->prev = cblk->rtlend;
            else
               cblk->down->rtlend = cblk->rtlend;
         }
         if (cblk->rtls)
            cblk->down->rtls = cblk->rtls;
         cblk->rtls = cblk->rtlend = (struct list *) NULL;
#ifdef VET
         /* Need to update basic block to remove all rtls. */
         begin_trans(VET_OPT_TRANS);
#ifdef INTERACT
         for (ptr = cblk->down->rtls; ptr; ptr = ptr->next) {
            if (ptr->old_daddy == cblk)
               move_rtl(ptr, cblk, cblk->down->old_rtls, cblk->down);
         }
#endif
         update_bb_rtls(cblk);
         end_trans();
#endif
         cblk = cblk->down;
      }

      /*
       * While there are instructions to issue, schedule instructions
       */
      while (ready = ready_set(cblk, invocation)) {
         choose_rtl(cblk, &ready, invocation);
         release_ready(cblk, ready);
      }

      /*
       * Connect the reordered RTLs back to the basic block
       */
      cblk->rtls = issued;
      cblk->rtlend = tailissued;
#ifdef VET
      /* 
       * Update basic block for reorderd RTLs.  If merged two
       * basic blocks then update second basic block.
       */
      begin_trans(VET_OPT_TRANS);
      update_bb_rtls(cblk);
      end_trans();
#endif
   }

   /*
    * Ensure that all RTLs can still be executed without conflicts
    */
   add_noops();

#ifdef VET
#ifdef INTERACT
   end_phase(INST_SCHED_TRANS);
#else
   end_phase(INSTSCHED_PHASE);
#endif
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_instruction_scheduling", invocation))
      dump_function_i("after instruction_scheduling()", invocation,
                      SHOW_LABELS | SHOW_SUCCS | SHOW_PREDS | SHOW_NONRTLS);
} /* end of instruction_scheduling() */

#endif  /* SCHED */

#ifdef FILL

/*
 * fill_delay_slots - attempt to fill any empty delay slots in the function
 */
int fill_delay_slots(int invocation)
{
   register struct list *ptr;
   register struct bblock *cblk;
   struct list *nptr;
   int   change;

#ifdef MEASURE
   register char *a;
   struct list *cptr;
   char  reg[MAXSTRING];
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_fill_delay_slots", invocation))
      dump_function_i("before fill_delay_slots()", invocation,
                      SHOW_LABELS | SHOW_SUCCS | SHOW_PREDS | SHOW_NONRTLS);

   change = FALSE;

   /*-
    * The purpose of this routine is to coordinate the filling of delay slots
    * with appropriate instructions from elsewhere in the function.  The
    * machine description is responsible for filling in the "kind" field of
    * each instruction (by setting the "ikind" global variable) with the
    * information needed in order to locate appropriate instructions to fill
    * in delay slots.  At the present time, the algorithm presented here will
    * handle only a single delay slot per instruction.  The information in
    * the "kind" field used by this algorithm requires the following bits
    * to be set appropriately on the instruction containing the delay slot:
    *
    *   DTAKEN  DNOTAKEN  DBOTH
    *    reset   reset    reset    No delay slots (normal instruction)
    *    reset   reset     set     Both paths always execute delay slot
    *    reset    set     reset    Delay slot executed if branch not taken
    *    reset    set      set     Can be executed by both paths or not taken
    *     set    reset    reset    Delay slot executed if branch is taken
    *     set    reset     set     Can be executed by both paths or if taken
    *     set     set     reset    Either taken or not taken but never both
    *     set     set      set     Any combination of taken, not taken or both
    *
    * Obviously, the last six pertain only to instructions that can change the
    * flow of control (not calls).  Instructions that do not change the flow
    * of control can only be annotated with the first or second cases.
    */

#ifdef VET
#ifdef INTERACT
   begin_phase(FILL_DELAY_SLOTS_TRANS);
#else
   begin_phase(FILLSLOTS_PHASE);
#endif
#endif
   /*
    * Ensure that dependency information is avaiable for all instructions
    */
   if (!swi) {
#ifdef VET
      /*
       * Depend_setup calls set() which changes the side_effect
       * through yyparse(). 
       */
      begin_trans(VET_OPT_TRANS);
#endif
      depend_setup();
#ifdef VET
      end_trans();
#endif
   }

   /*
    * Visit each RTL in the function looking for items that contain delay slots
    * and, if the delay slots aren't already filled, attempt to find
    * appropriate instructions to fill them with.
    */
   for (cblk = top; cblk; cblk = cblk->down) {
      for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
         if (cnts(ptr) && ptr->kind & (DTAKEN | DNOTAKEN | DBOTH)
             && !((nptr = nextexec(ptr->next, ptr))
                  && (nptr->kind & (TAKEN | NOTAKEN)))) {

#ifdef VET
            begin_trans(VET_OPT_TRANS);
#endif
            /*
             * Look for appropriate instructions to fill the delay slot with
             */
            if (ptr->kind & DBOTH && (nptr = fill_both(ptr)))
               nptr->kind |= TAKEN | NOTAKEN;
            else if (ptr->kind & DTAKEN && (nptr = fill_taken(ptr)))
               nptr->kind |= TAKEN;
            else if (ptr->kind & NOTAKEN && (nptr = fill_notaken(ptr)))
               nptr->kind |= NOTAKEN;
#ifdef INTERACT
            if (nptr)
               mod_kind(nptr);
#endif

            /*
             * If an appropriate instruction is found, it will already have
             * been removed from its previous location, with the exception that
             * we should remove its combiner links.  We must then place the new
             * instruction in the delay slot.  If no instruction is found, we
             * do not need to fill the delay slot with anything, since the
             * machine description must be smart enough to realize that the
             * slot is not filled and issue instructions accordingly.
             */
            if (nptr) {
               struct bblock *old_daddy;

#ifdef MEASURE

               /*
                * If we are possibly going to instrument the code later, we
                * need to be careful about how we move instructions so that we
                * don't move the use of a register past its death, because we
                * may need to use this information to make register allocation
                * decisions.
                */
#ifdef VET
#ifdef INTERACT
               if (swq)
#endif
                  rold = rinit();
#endif
               if (nptr->daddy == ptr->daddy) {
                  cptr = nptr->next;
                  for (;;) {
                     if (cptr->type & RTLINE) {
                        reget(cptr->deads, reg);
                        for (a = reg; *a; a = skiploc(a) + 1) {
                           if (isrg(a)) {
                              if (isref(nptr->val, a)) {
                                 rdelete(a, &cptr->deads);
#ifdef VET
#ifdef INTERACT
                                 if (swq)
#endif
                                    rsave(&rold, nptr->deads);
#endif
                                 rinsert(a, &nptr->deads);
                              }
                           }
                        }
                     }
                     if (cptr == ptr)
                        break;
                     cptr = cptr->next;
                  }
               }
#endif

               remv_links(nptr);
               if (nptr->next = ptr->next)
                  ptr->next->prev = nptr;
               else
                  cblk->rtlend = nptr;
               nptr->prev = ptr;
               ptr->next = nptr;
               old_daddy = nptr->daddy;
               nptr->daddy = cblk;
#ifdef VET
               /* rinsert call above will not mod deads since nptr was
                * unlinked from the structure */
               if (rold)
                  x_send_deads(rold, &nptr->deads);
               move_rtl(nptr, old_daddy, nptr->next, cblk);
#endif
               change = TRUE;
            }
#ifdef VET
            end_trans();
#endif
         }
      }
   }

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_fill_delay_slots", invocation))
      dump_function_i("after fill_delay_slots()", invocation,
                      SHOW_LABELS | SHOW_SUCCS | SHOW_PREDS | SHOW_NONRTLS);

#ifdef VET
#ifdef INTERACT
   end_phase(FILL_DELAY_SLOTS_TRANS);
#else
   end_phase(FILLSLOTS_PHASE);
#endif
#endif
   /*
    * Indicate if any new delay slots were filled
    */
   return change;
} /* end of fill_delay_slots() */

#endif /* FILL */

#ifdef SCHED

/*
 * ready_set - determine which items are "ready" to be issued
 */
struct list *ready_set(struct bblock *cblk, int invocation)
{
   register struct list *ptr, *nptr;
   register int cdoes, cblocks;
   register vector csets, cuses;
   struct list *ready;

   /*
    * Initialize dependency information and ready set
    */
   cdoes = 0;
   cblocks = 0;
   cuses = regnew();
   csets = regnew();
   ready = (struct list *) NULL;

   /*
    * Find the set of RTLs currently in the unissued set that have no depencies
    * with any of the other RTLs in this set and that may be immediately
    * considered as candidates to issue to the machine.
    */
   for (ptr = cblk->rtls; ptr; ptr = nptr) {
      nptr = ptr->next;

      /*
       * Assembly lines need to maintain their position relative to other RTLs
       * for various reasons, so we make a special case of them.
       */
      if (!(ptr->type & RTLINE)) {
         if (ready)
            break;
         issue_non_rtl(cblk, ptr);
         continue;
      }

      /*
       * Note that some instructions can only be placed in the ready set by
       * themselves.
       */
      if (ready && LONER(ptr->kind))
         break;

      /*
       * If there are any dependencies with any of the items that have already
       * been placed in the ready set, then don't allow this RTL to be placed
       * in the ready set.  Also, if there are any dependencies with any of the
       * preceeding RTLs that have not yet been issued or placed in the ready
       * set, then don't consider this RTL, either.
       */
      if (!(ptr->kind & READY)
          && (cdoes & ptr->blocks || ptr->does & cblocks
              || REGCOM(csets, ptr->uses) || REGCOM(cuses, ptr->sets)
              || REGCOM(csets, ptr->sets))) {

         /*
          * Update the new depend information for the next RTLs that can be
          * added to the "ready" set.
          */
         cdoes |= ptr->does;
         cblocks |= ptr->blocks;
         REGINC(csets, ptr->sets);
         REGINC(cuses, ptr->uses);

         continue;
      }

      /*
       * Update the new depend information for the next RTLs that can be added
       * to the "ready" set.
       */
      cdoes |= ptr->does;
      cblocks |= ptr->blocks;
      REGINC(csets, ptr->sets);
      REGINC(cuses, ptr->uses);

      /*
       * Move the RTL from the unissued set into the "ready" set
       */
      insert_ready(cblk, &ready, ptr);

      if (LONER(ptr->kind))
         break;
   }

   /*
    * Release sets and uses information for the basic block
    */
   mrfree(csets);
   mrfree(cuses);

   if (swd && pready
       && debug_opt_i_l("instruction_scheduling", invocation, cblk->label)) {
      int   first = TRUE;

      for (ptr = ready; ptr; ptr = ptr->next) {
         if (first) {
            (void) fprintf(stderr, "Ready set:\n");
            first = FALSE;
            pready = FALSE;
         }
         (void) fprintf(stderr, "  {%d} '", ptr->cost);
         show_rtl(stderr, ptr->val);
         (void) fprintf(stderr, "'\n");
      }
   }

   return ready;
} /* end of ready_set() */

/*
 * release_ready - release all unused items from the ready set
 */
void release_ready(struct bblock *cblk, struct list *ready)
{
   struct list *head;

   /*
    * If there are any items left in the ready set, place them at the start of
    * the non-issued set.
    */
   if (ready) {
      head = cblk->rtls;
      for (cblk->rtls = ready; ready->next; ready = ready->next) ;
      if (ready->next = head)
         head->prev = ready;
   }
}

/*
 * insert_ready - move an RTL from the unissued set to the ready set
 */
void insert_ready(struct bblock *cblk, struct list **ready, struct list *ptr)
{
   register struct list *nptr, *pptr;
   int sum1, sum2;

   /*
    * Note the fact that this item is "ready" so that we needn't check its
    * dependencies in the future.
    */
   ptr->kind |= READY;
#ifdef INTERACT
   begin_trans(VET_OPT_TRANS);
   mod_kind(ptr);
   end_trans();
#endif

   /*
    * Determine the advantage of scheduling this RTL at this time
    */
   advantage(ptr);

   /*
    * Disconnect the RTL that is to be moved from the unissued set
    */
   if (ptr->prev)
      ptr->prev->next = ptr->next;
   else
      cblk->rtls = ptr->next;
   if (ptr->next)
      ptr->next->prev = ptr->prev;
   else
      cblk->rtlend = ptr->prev;

   /*
    * Insert the RTL that is to be moved into the ready set in descending order
    * by the cost of the RTL.
    */
   if (*ready) {

      /*
       * Find the appropriate slot in the ready set for this line.  Note that
       * items of like value are arranged so that the first item in comes first
       * in the ready set.
       */
      pptr = (struct list *) NULL;
      for (nptr = *ready; nptr; nptr = nptr->next) {
#ifdef INTERACT
	/*useful while detecting identical function instances. CRC values
	  allow us to take the order of RTLs in the function into
	  consideration, but this also means that many functions which only
	  differ in the order of RTLs are detected as non-identical. This
	  should help a bit to remove more function instances as redundant*/
	if(nptr->cost==ptr->cost){
	  if((sum1=RTL_sum(ptr)) < (sum2=RTL_sum(nptr)))
	    break;
	  else if(sum1==sum2 && ptr->rtl_id<nptr->rtl_id)
	    break;
	}
#endif	  
         if (nptr->cost < ptr->cost)
            break;
         pptr = nptr;
      }

      /*
       * Connect the new RTL to the ready set
       */
      if (nptr) {
         ptr->prev = nptr->prev;
         ptr->next = nptr;
         if (nptr->prev)
            nptr->prev->next = ptr;
         else
            *ready = ptr;
         nptr->prev = ptr;
      }
      else {
         ptr->prev = pptr;
         ptr->next = (struct list *) NULL;
         pptr->next = ptr;
      }
   }

   /*
    * Inserting the first item into the ready set is trivial
    */
   else {
      *ready = ptr;
      ptr->prev = ptr->next = (struct list *) NULL;
   }
} /* end of insert_ready() */

/*
 * advantage - determines the cummulative advantage of issuing this RTL
 */
void advantage(struct list *rtl)
{
   register struct list *ptr;
   register int adv;

   /*
    * For each instruction calculate its cost plus part of the highest cost of
    * any of the instructions in this basic block that depend upon it. This
    * will allow us to determine which instruction should be scheduled first.
    */
   adv = 0;
   for (ptr = rtl->next; ptr; ptr = ptr->next) {
      if (ptr->type & RTLINE && depends(ptr, rtl))
         if (adv < ADVANTAGE(ptr->kind))
            adv = ADVANTAGE(ptr->kind);
   }
   adv >>= 1;
   adv += ADVANTAGE(rtl->kind);

   /*
    * Obtain the last issued instruction of the desired type and use it to
    * modify the advantage of this instruction based on the target machine.
    */
   if (tailissued
       && (ptr = prevexec((struct bblock *) NULL, tailissued, rtl)))
      rtl->cost = sched_adv(adv, rtl, ptr);
   else
      rtl->cost = adv;
}

/*
 * choose_rtl - choose an RTL that can safely follow the last one
 */
void choose_rtl(struct bblock *cblk, struct list **ready, int invocation)
{
   register struct list *ptr;

   /*
    * Loop until we are able to emit an RTL from the ready set
    */
   for (;;) {

      /*
       * Look through the set of ready items and issue the most appropriate
       * one.  If none can be issued because of hardware interlocks, then emit
       * a no-op instruction.
       */
      for (ptr = *ready; ptr; ptr = ptr->next) {

         /*
          * If the instruction can follow the previous one, issue it
          */
         if (canfollow(cblk, ptr, tailissued)) {
            issue_rtl(ready, ptr, invocation, cblk->label);
            return;
         }

         /*
          * If this instruction cannot follow the last executable RTL, see if
          * we can modify it so that it can follow.
          */
         if (modify(cblk, ptr, tailissued)
             && canfollow(cblk, ptr, tailissued)) {
            issue_rtl(ready, ptr, invocation, cblk->label);
            return;
         }
      }

      /*
       * If we cannot issue any instructions at this time, issue a no-op.  On
       * some architectures, we may have different classes of no-ops for
       * different execution units.  If this is the case, then the kind of
       * no-op issued might have a big impact on what can follow in the next
       * round of issues.  The mechanism here does not really allow us to
       * determine which kind of no-op would be best to issue at this
       * particular point, but what we can do is issue a no-op that will at
       * least eventually permit the first RTL in the ready set to eventually
       * issue.  This at least keeps us from looping infinitely.
       */
      issue_noop(cblk, (*ready)->kind, invocation, cblk->label);
   }
} /* end of choose_rtl() */

/*
 * issue_rtl - issue an RTL from the ready set to the issued set
 */
void issue_rtl(struct list **ready, struct list *ptr, int invocation,
               char *label)
{
   if (swd && debug_opt_i_l("instruction_scheduling", invocation, label)) {
      (void) fprintf(stderr, "Issuing: '");
      show_rtl(stderr, ptr->val);
      (void) fprintf(stderr, "'\n");
      pready = TRUE;
   }

   /*
    * Disconnect the group from its old set
    */
   if (*ready == ptr)
      *ready = ptr->next;
   else
      ptr->prev->next = ptr->next;
   if (ptr->next)
      ptr->next->prev = ptr->prev;

   /*
    * Append the group to the issued set
    */
   if (tailissued) {
      tailissued->next = ptr;
      ptr->prev = tailissued;
   }
   else {
      issued = ptr;
      ptr->prev = (struct list *) NULL;
   }
   ptr->next = (struct list *) NULL;
   tailissued = ptr;
} /* end of issue_rtl() */

/*
 * issue_noop - append a no-op to the issued set
 */
void issue_noop(struct bblock *cblk, unsigned int kind, int invocation,
                char *label)
{
   char  new[MAXSTRING];

   /*
    * Create an RTL and place it in the issued set
    */
   tailissued = new_rtl((struct bblock *) NULL, RTLINE, tailissued);
   tailissued->daddy = cblk;

   /*
    * Call the machine-dependent routine responsible for generating a no-op of
    * the appropriate kind.
    */
   create_noop(new);
   tailissued->val = strcpy(CALLOC(strlen(new) + 1), new);
   tailissued->deads = rinit();
   setup_noop(tailissued);
#ifdef VET
   insert_rtl(tailissued);
#endif

   /*
    * If we have to issue a no-op as the first RTL in a basic block, set the
    * issued pointer as required.
    */
   if (!issued)
      issued = tailissued;

   if (swd && debug_opt_i_l("instruction_scheduling", invocation, label)) {
      (void) fprintf(stderr, "Issuing: '");
      show_rtl(stderr, new);
      (void) fprintf(stderr, "'\n");
      pready = TRUE;
   }
} /* end of issue_noop() */

/*
 * issue_non_rtl - issue a non-RTL line to the issued set
 */
void issue_non_rtl(struct bblock *cblk, register struct list *ptr)
{
   /*
    * Disconect the non-RTL from its basic block
    */
   if (ptr->prev)
      ptr->prev->next = ptr->next;
   else
      cblk->rtls = ptr->next;
   if (ptr->next)
      ptr->next->prev = ptr->prev;
   else
      cblk->rtlend = ptr->prev;

   /*
    * Place the non-RTL at the end of the issued set
    */
   if (ptr->prev = tailissued)
      tailissued->next = ptr;
   else
      issued = ptr;
   ptr->next = (struct list *) NULL;
   tailissued = ptr;
}

/*
 * add_noops - ensure that RTLs in each basic block are safe to execute
 */
void add_noops()
{
   register struct list *ptr;
   register struct bblock *cblk;
   struct list *nptr;
   rlist dead;
   unsigned int kind;
   char  new[MAXSTRING];

   for (cblk = top; cblk; cblk = cblk->down) {
      for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
         if (cnts(ptr) && !canfollow(cblk, ptr, ptr->prev)) {
            create_noop(new);
            dead = rinit();
#ifdef VET
            begin_trans(VET_OPT_TRANS);
#endif
            nptr = add_rtl(cblk, ptr, new, dead);
            setup_noop(nptr);
         }
      }
   }
} /* end of add_noops() */

#endif  /* SCHED */

#ifdef FILL

/*
 * fill_both - look for instructions to fill a delay slot both ways
 */
struct list *fill_both(struct list *ptr)
{
   register struct list *rtl;
   register int cdoes, cblocks;
   register vector csets, cuses;
   int   xdoes, xblocks;
   struct list *last;

   /*
    * Initialize dependency information
    */
   xdoes = 0;
   xblocks = 0;

   cuses = regnew();
   csets = regnew();
   uses(ptr, &xdoes, &xblocks, cuses, csets);
   cdoes = xdoes;
   cblocks = xblocks;

   /*
    * Start from the instruction with the delay slot and move upwards looking
    * for an instruction that can be moved into the delay slot.
    */
   rtl = prevexec2(ptr->prev, ptr);
   for (last = ptr; rtl; rtl = prevexec2(rtl->prev, ptr)) {

      /*
       * Do not attempt to move instructions that have already been placed in
       * someone else's delay slot or that have delay slots of their own. Also,
       * any items marked as NOMOVE should be kept out of delay slots.
       */
      if (rtl->kind & (TAKEN | NOTAKEN | DTAKEN | DNOTAKEN | DBOTH | NOMOVE)) {

         /*
          * Add to the dependecy information since this instruction will be
          * part of the interval between the next instructions considered and
          * the delay slot.
          */
         cdoes |= rtl->does;
         cblocks |= rtl->blocks;
         REGINC(csets, rtl->sets);
         REGINC(cuses, rtl->uses);

         last = rtl;
         continue;
      }

      /*
       * Ensure that this instruction does not conflict with anything in the
       * interval between here and the delay slot.
       */
      if (cdoes & rtl->blocks || cblocks & rtl->does
          || REGCOM(csets, rtl->uses) || REGCOM(cuses, rtl->sets)
          || REGCOM(csets, rtl->sets)) {

         /*
          * Add to the dependecy information since this instruction will be
          * part of the interval between the next instructions considered and
          * the delay slot.
          */
         cdoes |= rtl->does;
         cblocks |= rtl->blocks;
         REGINC(csets, rtl->sets);
         REGINC(cuses, rtl->uses);

         last = rtl;
         continue;
      }

      /*
       * This instruction looks like a very likely candidate for the delay
       * slot, so let's remove it from its location and see if it causes any
       * problems when we do.
       */
      if (rtl->prev)
         rtl->prev->next = rtl->next;
      else
         rtl->daddy->rtls = rtl->next;
      if (rtl->next)
         rtl->next->prev = rtl->prev;
      else
         rtl->daddy->rtlend = rtl->prev;

      /*
       * If we cannot remove this instruction from its current location without
       * causing a hardware interlock, then we have to continue our search.
       */
      if (!canfollow(rtl->daddy, last, rtl->prev)) {

         /*
          * Re-insert the item back where it was
          */
         if (rtl->prev)
            rtl->prev->next = rtl;
         else
            rtl->daddy->rtls = rtl;
         if (rtl->next)
            rtl->next->prev = rtl;
         else
            rtl->daddy->rtlend = rtl;

         /*
          * Add to the dependecy information since this instruction will be
          * part of the interval between the next instructions considered and
          * the delay slot.
          */
         cdoes |= rtl->does;
         cblocks |= rtl->blocks;
         REGINC(csets, rtl->sets);
         REGINC(cuses, rtl->uses);

         last = rtl;
         continue;
      }

      /*
       * We are almost there.  All that is left now is to check to be sure that
       * the instruction can be safely placed in the delay slot of the original
       * instruction.
       */
      if (!canfollow(ptr->daddy, rtl, ptr)) {

         /*
          * Re-insert the item back where it was
          */
         if (rtl->prev)
            rtl->prev->next = rtl;
         else
            rtl->daddy->rtls = rtl;
         if (rtl->next)
            rtl->next->prev = rtl;
         else
            rtl->daddy->rtlend = rtl;

         /*
          * Add to the dependecy information since this instruction will be
          * part of the interval between the next instructions considered and
          * the delay slot.
          */
         cdoes |= rtl->does;
         cblocks |= rtl->blocks;
         REGINC(csets, rtl->sets);
         REGINC(cuses, rtl->uses);

         last = rtl;
         continue;
      }

      /*
       * Release sets and uses information for the basic block
       */
      mrfree(csets);
      mrfree(cuses);

      /*
       * We have found a valid instruction to fill in the delay slot, so return
       * it to the caller.
       */
      return rtl;
   }

   /*
    * Release sets and uses information for the basic block
    */
   mrfree(csets);
   mrfree(cuses);

   return ((struct list *) NULL);
} /* end of fill_both() */

/*
 * fill_taken - look for instructions to fill a delay slot from the target path
 */
struct list *fill_taken(struct list *ptr)
{
   register struct list *rtl;
   register int cdoes, cblocks;
   register vector csets, cuses;
   struct bblock *cblk, *prev, *target, *branch;
   struct list *next, *last, *nptr;
   struct blist *bptr, *fptr;
   int   paths, found;

   /*
    * Attempt to locate the first instruction executed at the target of the
    * branch.  The target basic block can be one of two cases.  First, it may
    * be reached only by the current branch, in which case we may take items
    * from it at will, without having to make any alterations to the function's
    * control flow.  The second and most likely case, however, is that the
    * branch is just one of two or more paths to the target block, in which
    * case we can move an item from the target block only if we carefully
    * re-arrange the existing control-flow graph.
    */
   paths = 0;
   prev = ptr->daddy;
   found = FALSE;
   if (!(cblk = prev->right))
      cblk = prev->left;
   while (cblk) {
      if (cblk->preds->ptr != prev || cblk->preds->next) {
         for (bptr = cblk->preds; bptr; bptr = bptr->next) {
            if (found || bptr->ptr != prev)
               paths++;
            else
               found = TRUE;
         }
      }
      if (!cblk->rtls) {
         if (cblk->right)
            return ((struct list *) NULL);
         else {
            prev = cblk;
            if (!(cblk = cblk->left))
               return ((struct list *) NULL);
            found = FALSE;
         }
      }
      else
         break;
   }

   /*
    * Initialize dependency information
    */
   cdoes = 0;
   cblocks = 0;
   cuses = regnew();
   csets = regnew();
   last = (struct list *) NULL;

   /*
    * Go to the target of the branch and attempt to locate a suitable
    * instruction for that basic block to move into the delay slot.
    */
   for (rtl = nextexec(cblk->rtls, ptr); rtl; rtl = next) {

      /*
       * Locate the following instruction, if any
       */
      next = nextexec(rtl->next, ptr);

      /*
       * Do not attempt to move instructions that have already been placed in
       * someone else's delay slot or that have delay slots of their own. Also,
       * any items marked as NOMOVE should be kept out of delay slots.
       */
      if (rtl->kind & (TAKEN | NOTAKEN | DTAKEN | DNOTAKEN | DBOTH | NOMOVE)) {

         /*
          * Add to the dependecy information since this instruction will be
          * part of the interval between the next instructions considered and
          * the delay slot.
          */
         cdoes |= rtl->does;
         cblocks |= rtl->blocks;
         REGINC(csets, rtl->sets);
         REGINC(cuses, rtl->uses);

         last = rtl;
         continue;
      }

      /*
       * Ensure that this instruction does not conflict with anything in the
       * interval between here and the delay slot.
       */
      if (cdoes & rtl->blocks || cblocks & rtl->does
          || REGCOM(csets, rtl->uses) || REGCOM(cuses, rtl->sets)
          || REGCOM(csets, rtl->sets)) {

         /*
          * Add to the dependecy information since this instruction will be
          * part of the interval between the next instructions considered and
          * the delay slot.
          */
         cdoes |= rtl->does;
         cblocks |= rtl->blocks;
         REGINC(csets, rtl->sets);
         REGINC(cuses, rtl->uses);

         last = rtl;
         continue;
      }

      /*
       * This instruction looks like a very likely candidate for the delay
       * slot, so let's remove it from its location and see if it causes any
       * problems when we do.
       */
      if (rtl->prev)
         rtl->prev->next = rtl->next;
      else
         rtl->daddy->rtls = rtl->next;
      if (rtl->next)
         rtl->next->prev = rtl->prev;
      else
         rtl->daddy->rtlend = rtl->prev;

      /*
       * If we cannot remove this instruction from its current location without
       * causing a hardware interlock, then we have to continue our search.
       */
      if (next && !canfollow(next->daddy, next, last)) {

         /*
          * Re-insert the item back where it was
          */
         if (rtl->prev)
            rtl->prev->next = rtl;
         else
            rtl->daddy->rtls = rtl;
         if (rtl->next)
            rtl->next->prev = rtl;
         else
            rtl->daddy->rtlend = rtl;

         /*
          * Add to the dependecy information since this instruction will be
          * part of the interval between the next instructions considered and
          * the delay slot.
          */
         cdoes |= rtl->does;
         cblocks |= rtl->blocks;
         REGINC(csets, rtl->sets);
         REGINC(cuses, rtl->uses);

         last = rtl;
         continue;
      }

      /*
       * We are almost there.  All that is left now is to check to be sure that
       * the instruction can be safely placed in the delay slot of the original
       * instruction.
       */
      if (!canfollow(ptr->daddy, rtl, ptr)) {

         /*
          * Re-insert the item back where it was
          */
         if (rtl->prev)
            rtl->prev->next = rtl;
         else
            rtl->daddy->rtls = rtl;
         if (rtl->next)
            rtl->next->prev = rtl;
         else
            rtl->daddy->rtlend = rtl;

         /*
          * Add to the dependecy information since this instruction will be
          * part of the interval between the next instructions considered and
          * the delay slot.
          */
         cdoes |= rtl->does;
         cblocks |= rtl->blocks;
         REGINC(csets, rtl->sets);
         REGINC(cuses, rtl->uses);

         last = rtl;
         continue;
      }

      /*
       * Now we have to consider the fact that we may have to alter the
       * control-flow path of this function in order to ensure that alternate
       * paths to the target basic block maintain their original semantics.
       */
      if (paths) {

         /*
          * This code handles the case where the instruction picked for the
          * delay slot must be placed in a basic block of its own.  The target
          * of the original branch, which will now have the delay instruction
          * in its delay slot will then jump not to this basic block, but to
          * one which follows, thus avoiding the double execution of the delay
          * slot instruction.
          */
         prev = ptr->daddy;
         if (!(cblk = prev->right))
            cblk = prev->left;
         while (cblk) {
            if (cblk->preds->ptr != prev || cblk->preds->next) {

               /*
                * If there is only one other path to the target basic block,
                * then if that other path happens to be the fall-through path
                * into the target basic block, then it might be possible to
                * just place a copy of the delay slot instruction into the
                * other path so that no actual control-flow alteration need be
                * made to the function.
                */
               if (paths == 1) {
                  for (bptr = cblk->preds; bptr; bptr = bptr->next) {
                     if (bptr->ptr != prev && !bptr->ptr->right
                         && (!bptr->ptr->rtlend
                             || !(bptr->ptr->rtlend->type & CASELINE))
                         && !(bptr->ptr->status & JUMP)) {
                        if (canfollow(bptr->ptr, rtl, bptr->ptr->rtlend)) {
                           (void) copy_rtl(bptr->ptr, (struct list *) NULL,
                                           rtl);
                           goto finished;
                        }
                     }
                  }
               }
            }

            /*
             * If we come to an empty basic block, keep following the fall-
             * through link.  There is one slight catch, however, and that is
             * the possibility that this block is empty only because we removed
             * the delay instruction from it, in which case this is a great
             * opportunity to just put a copy of the instruction back here and
             * let all other paths (other than the one that came from the delay
             * slot branch) branch to the point following this basic block.
             */
            if (!cblk->rtls) {
               if (cblk == rtl->daddy) {
                  (void) copy_rtl(cblk, (struct list *) NULL, rtl);
                  target = cblk->left;
                  label_basic_block(target);
                  branch =
                   (ptr->daddy->right) ? ptr->daddy->right : ptr->daddy->left;
                  if (change_target(ptr->daddy, branch, target))
                     remv_predecessor(ptr->daddy, branch);
                  else {
                     (void) fprintf(stderr,
                                    "vpo (fill_taken): cannot locate label\n");
                     exit(1);
                  }
                  add_predecessor(ptr->daddy, target);
                  goto finished;
               }
            }

            /*
             * This basic block has some items in it
             */
            else {
               for (next = cblk->rtls; next; next = next->next) {

                  /*
                   * If we encounter any valid RTLs along this path, then our
                   * hopes of just branching around the delayed instruction are
                   * dashed, so stop the search.
                   */
                  if (cnts(next))
                     break;

                  /*
                   * If this case is TRUE, then the delay RTL happens to be the
                   * first RTL following the path from the target branch.  If
                   * this is the case, then it may also be that the delay
                   * instruction is already in a basic block all by itself, in
                   * which case we need only to alter the branch target.
                   */
                  if (next == rtl->next || next == rtl->prev) {
                     for (nptr = next; nptr; nptr = nptr->next) {
                        if (cnts(nptr))
                           break;
                     }

                     if (!nptr && !cblk->right) {
                        (void) copy_rtl(cblk, next, rtl);
                        target = cblk->left;
                        label_basic_block(target);
                        branch =
                         (ptr->daddy->right) ? ptr->daddy->right : ptr->
                         daddy->left;
                        if (change_target(ptr->daddy, branch, target))
                           remv_predecessor(ptr->daddy, branch);
                        else {
                           (void) fprintf(stderr,
                                    "vpo (fill_taken): cannot locate label\n");
                           exit(EXIT_FAILURE);
                        }
                        add_predecessor(ptr->daddy, target);
                        goto finished;
                     }

                     break;
                  }
               }

               if (next || cblk->right)
                  break;
            }

            /*
             * Continue looking for an appropriate RTL to place in the delay
             * slot in the following basic block.
             */
            prev = cblk;
            cblk = cblk->left;
         }

         /*
          * At this point we have exhausted all of the possible simple
          * solutions, and are left only with having to build a new basic block
          * and change the control-flow structure of the function.  First, we
          * create a new basic block and place a copy of the delay slot
          * instruction in it.
          */
         target = new_labeled_basic_block();
         (void) copy_rtl(target, (struct list *) NULL, rtl);

         /*
          * Now we have to locate all of the paths to the basic block which
          * contained the delay item and force them all to branch to the new
          * basic block.
          */
         prev = ptr->daddy;
         if (!(cblk = prev->right))
            cblk = prev->left;
         while (cblk) {

            /*
             * If this basic block has predecessors other than the block that
             * contained the delay slot item, then change those predecessors to
             * branch to the block that contains the RTL that was copied.
             */
            if (cblk->preds->ptr != prev || cblk->preds->next) {
               for (bptr = cblk->preds; bptr; bptr = fptr) {
                  fptr = bptr->next;
                  if (bptr->ptr != prev) {
                     if (change_target(bptr->ptr, cblk, target)) {
                        add_predecessor(bptr->ptr, target);
                        remv_predecessor(bptr->ptr, cblk);
                     }
                  }
               }
            }

            if (!cblk->rtls) {

               /*
                * Check to see if this basic block contained the RTL that was
                * used to fill the delay slot.  If it is, then we place the new
                * basic block in front of it so that all paths other than the
                * one that contained the delay slot will execute a copy of the
                * RTL that was used to fill the slot.
                */
               if (rtl->daddy == cblk) {
                  target->left = cblk;
#ifdef VET
                  mod_bb_ptrs(MOD_LEFT, target, cblk);
#endif
                  if (target->up = cblk->up) {
                     target->up->down = target;
#ifdef VET
                     mod_bb_ptrs(MOD_DOWN, target->up, target);
#endif
                     if (target->up->left == cblk
                         && !(target->up->status & JUMP)) {
                        target->up->left = target;
#ifdef VET
                        mod_bb_ptrs(MOD_LEFT, target->up, target);
#endif
                        target->status |= FALLTHRU;
                        remv_predecessor(target->up, cblk);
                        add_predecessor(target->up, target);
                     }
                  }
                  target->down = cblk;
                  cblk->up = target;
#ifdef VET
                  mod_bb_ptrs(MOD_DOWN, target, cblk);
#endif
                  if (target->prev_dfs = cblk->prev_dfs)
                     target->prev_dfs->next_dfs = target;
                  target->next_dfs = cblk;
                  cblk->prev_dfs = target;
                  cblk->status |= FALLTHRU;
                  add_predecessor(target, cblk);
                  branch =
                   (ptr->daddy->right) ? ptr->daddy->right : ptr->daddy->left;
                  if (branch != cblk) {
                     (void) change_target(ptr->daddy, branch, cblk);
                     remv_predecessor(ptr->daddy, branch);
                     add_predecessor(ptr->daddy, cblk);
                  }
                  goto finished;
               }
            }

            else {

               /*
                * If this basic block contained the RTL that was placed in the
                * delay slot, of if it contains any executable RTL, then this
                * if the place that we must place the new basic block so that
                * all other paths have a place to transfer control to.
                */
               for (next = cblk->rtls; next; next = next->next) {
                  if (cnts(next) || next == rtl->next || next == rtl->prev) {
                     target->left = cblk;
#ifdef VET
                     mod_bb_ptrs(MOD_LEFT, target, cblk);
#endif
                     if (target->up = cblk->up) {
                        target->up->down = target;
#ifdef VET
                        mod_bb_ptrs(MOD_DOWN, target->up, target);
#endif
                        if (cblk->up->left == cblk
                            && !(cblk->up->status & JUMP)) {
                           target->up->left = target;
#ifdef VET
                           mod_bb_ptrs(MOD_LEFT, target->up, target);
#endif
                           target->status |= FALLTHRU;
                           remv_predecessor(target->up, cblk);
                           add_predecessor(target->up, target);
                        }
                     }
                     target->down = cblk;
                     cblk->up = target;
#ifdef VET
                     mod_bb_ptrs(MOD_DOWN, target, cblk);
#endif
                     if (target->prev_dfs = cblk->prev_dfs)
                        target->prev_dfs->next_dfs = target;
                     target->next_dfs = cblk;
                     cblk->prev_dfs = target;
                     cblk->status |= FALLTHRU;
                     add_predecessor(target, cblk);
                     branch =
                      (ptr->daddy->right) ? ptr->daddy->right : ptr->daddy->
                      left;
                     if (branch != cblk) {
                        (void) change_target(ptr->daddy, branch, cblk);
                        remv_predecessor(ptr->daddy, branch);
                        add_predecessor(ptr->daddy, cblk);
                     }
                     goto finished;
                  }
               }
            }

            /*
             * Continue searching
             */
            prev = cblk;
            cblk = cblk->left;
         }
      }

    finished:

      /*
       * Release sets and uses information for the basic block
       */
      mrfree(csets);
      mrfree(cuses);

      /*
       * We have found a valid instruction to fill in the delay slot, so return
       * it to the caller.
       */
      return (rtl);
   }

   /*
    * Release sets and uses information for the basic block
    */
   mrfree(csets);
   mrfree(cuses);

   return ((struct list *) NULL);
} /* end of fill_taken() */

/*
 * fill_notaken - look for instructions to fill delay slot from fall-through
 */
struct list *fill_notaken(struct list *ptr)
{
    /*ARGSUSED*/ return ((struct list *) NULL);
}

#endif  /* FILL */

/*
 * depend_setup - setup information to calculate dependencies
 */
void depend_setup()
{
   register struct list *ptr, *nptr;
   register struct bblock *cblk;
   char  se[MAXSTRING];

   /*
    * Examine each RTL in the function
    */
   for (cblk = top; cblk; cblk = cblk->down) {
      STATUS(cblk, UNDONE);
      for (ptr = cblk->rtls; ptr; ptr = nptr) {
         nptr = ptr->next;
         if (cnts(ptr)) {

            /*
             * Ensure that we have obtained the appropriate "kind" information
             * for each RTL.
             */
            if (!ptr->kind || ptr->cost == 0) {
               ikind = FALSE;
               side_effect = se;
               *se = '\0';
               icost = 0;
               if (find(ptr, ptr->val, ptr->deads)) {
                  (void) fprintf(stderr,
                                 "vpo (depend_setup): cannot assemble '");
                  show_rtl(stderr, ptr->val);
                  (void) fprintf(stderr, "'\n");
                  continue;
               }
               if (!(ptr->kind = ikind)) {
                  (void) fprintf(stderr,
                                 "vpo (depend_setup): no kind given for '");
                  show_rtl(stderr, ptr->val);
                  (void) fprintf(stderr, "'\n");
               }
               ptr->cost = icost;
#ifdef INTERACT
               mod_kind(ptr);
#endif
               set(ptr, ptr->val, se);
            }

            /*
             * Get rid of zero cost instructions, since they may cause problems
             * if they are moved into delay slots.
             */
            if (!ptr->cost) {
               remv_links(ptr);
               remv_rtl(ptr);
               continue;
            }

            /*
             * Determine dependency information for this RTL.
             */
            sets_uses(ptr);
         }
      }
   }
   side_effect = (char *) NULL;
} /* end of depend_setup() */

/*
 * depends - return TRUE if two instructions are inter-dependent
 */
int depends(struct list *cptr, struct list *ptr)
{
   /*
    * Dependencies come from two sources, one is the "does" and "blocks" sets
    * for two RTLs.  The other is the registers set and used by the RTLs.
    */
   int   return_value;

   assert(NULL != cptr);
   assert(NULL != ptr);

   return_value = (cptr->does & ptr->blocks) || (ptr->does & cptr->blocks);

   if (!return_value) {
      assert(NULL != cptr->sets);
      assert(NULL != ptr->uses);
      return_value = REGCOM(cptr->sets, ptr->uses);

      if (!return_value) {
         assert(NULL != ptr->sets);
         assert(NULL != cptr->uses);
         return_value = REGCOM(cptr->uses, ptr->sets)
          || REGCOM(cptr->sets, ptr->sets);
      }
   }
   return return_value;
} /* end of depends() */

/*
 * sets_uses - indicate which items are set and used by this RTL
 */
void sets_uses(struct list *ptr)
{
   register char *a;

   /*
    * Initialize the "sets" and "uses" vectors
    */
   ptr->sets = regnew();
   ptr->uses = regnew();

   /*
    * Set up any information required for this particular target machine. If
    * md_sets_uses() return a TRUE value, then the standard sets and uses code
    * is skipped.
    */
   if (md_sets_uses(ptr))
      return;

   /*
    * For each register in the RTL, set the appropriate vectors to indicate
    * which register items are set and used by this RTL.
    */
   a = ptr->val;
   while (a = gtrg(a)) {
      if (*a == '=')
         touchreg(ptr->sets, unskiploc(a));
      else
         touchreg(ptr->uses, unskiploc(a));
   }

   /*
    * Now do the same for the side effects of this RTL
    */
   if (a = ptr->se) {
      while (a = gtrg(a)) {
         if (*a == '=')
            touchreg(ptr->sets, unskiploc(a));
         else
            touchreg(ptr->uses, unskiploc(a));
      }
   }
} /* end of sets_uses() */

/*
 * uses - define dependency information for item moving past into delay slot
 */
void uses(struct list *ptr, int *does, int *blocks, vector uses, vector sets)
{
   register char *a;

   if (md_uses(ptr, does, blocks, uses, sets))
      return;

   /*
    * For each register in the RTL, set the appropriate vectors to indicate
    * which register items are set and used by this RTL.
    */
   a = ptr->val;
   while (a = gtrg(a)) {
      if (*a != '=')
         touchreg(uses, unskiploc(a));
   }

   /*
    * Now do the same for the side effects of this RTL
    */
   if (a = ptr->se) {
      while (a = gtrg(a)) {
         if (*a != '=')
            touchreg(uses, unskiploc(a));
      }
   }
}

/*
 * canfollow - determine if this instruction can follow the last one
 */
int canfollow(struct bblock *cblk, struct list *cptr, struct list *last)
{
   register struct blist *bptr;
   struct list *ptr;

   /*
    * Note that this function only handles cases where an instruction can
    * conflict with its previous neighbor only.  Multiple instruction conflicts
    * will be taken care of in a future version.
    */

   /*
    * Determine if it is possible to have a conflict for this instruction
    */
   if (NOCONFLICT(cptr->kind))
      return TRUE;

   /*
    * Find the last previously executed RTL of the same kind in this block
    */
   ptr = prevexec(cblk, last, cptr);

   /*
    * If a previously executed RTL was found within the same basic block, then
    * determine if there are any conflicts.
    */
   if (ptr)
      return noconflict(ptr, cptr);

   /*
    * Otherwise, for each predecessor start with the RTLs at the end of the
    * block and determine if the instruction can follow.
    */
   for (bptr = cblk->preds; bptr; bptr = bptr->next) {

      /*
       * If we have alrady visited this one, skip it
       */
      if (bptr->ptr->status & DONE)
         continue;

      /*
       * Locate the last executable item in this block
       */
      if (ptr = prevexec(bptr->ptr, bptr->ptr->rtlend, cptr)) {

         /*
          * If there is a conflict, then we cannot follow
          */
         if (!noconflict(ptr, cptr))
            return FALSE;
      }
      else {

         /*
          * No executable found at this block, back to predecessors
          */
         STATUS(bptr->ptr, DONE);
         if (!canfollow(bptr->ptr, cptr, (struct list *) NULL)) {
            STATUS(bptr->ptr, UNDONE);
            return FALSE;
         }
         STATUS(bptr->ptr, UNDONE);
      }
   }

   /*
    * If we can follow all predecessors then the instruction can follow
    */
   return TRUE;
} /* end of canfollow() */

/*
 * prevexec - returns the first previous executable RTL in this basic block
 */
struct list *prevexec(struct bblock *cblk, register struct list *ptr,
                      struct list *cptr)
{
   /*
    * Search back until an executable RTL is found within the block, the
    * current block has more than one predecessor, or there is no previous
    * block.  Note that this routine assumes that only RTLs of type RTLINE are
    * executable.
    */
   while (ptr) {

      /*
       * Attempt to locate a previous instruction in this block with the same
       * instruction class as the one we wish to find the predecessor of.
       */
      while (ptr) {
         if (cnts(ptr) && (!cptr || SAMECLASS(ptr->kind, cptr->kind)))
            return ptr;
         ptr = ptr->prev;
      }

      /*
       * If this block has only one predecessor, follow it back
       */
      if (cblk && cblk->preds && !cblk->preds->next) {
         cblk = cblk->preds->ptr;
         ptr = cblk->rtlend;
      }
   }

   return ((struct list *) NULL);
}

/*
 * prevexec2 - return the previous executable RTL in the same execution class
 */
struct list *prevexec2(register struct list *ptr, struct list *cptr)
{
   while (ptr) {
      if (cnts(ptr) && (!cptr || SAMECLASS(ptr->kind, cptr->kind)))
         return (ptr);
      ptr = prev_rtl(ptr);
   }
   return ((struct list *) NULL);
}

/*
 * nextexec - return the next executable RTL of the appropriate kind
 */
struct list *nextexec(register struct list *ptr, struct list *cptr)
{
   while (ptr) {
      if (cnts(ptr) && (!cptr || SAMECLASS(ptr->kind, cptr->kind)))
         return ptr;
      ptr = next_rtl(ptr);
   }
   return ((struct list *) NULL);
}

#endif   /* SCHED || FILL */
