
/************************************************************************
 *                                                                      *
 * Copyright (c) 2000                                                   *
 *     The Rector and Visitors of the University of Virginia            *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from the University of Virginia.                                *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/

/************************************************************************

 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/stache.c,v $
 $Date: 2003/11/03 00:32:16 $
 $Revision: 1.4 $
 $Name:  $

 Log of changes:

 $Log: stache.c,v $
 Revision 1.4  2003/11/03 00:32:16  clc5q
 Moved find_global_declaration() out to machine dependent rtl.c files.

 Revision 1.3  2002/07/16 22:01:17  clc5q
 Export global_vars and globals_found for MIPS target.

 Revision 1.2  2002/07/16 19:41:41  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.1  2002/06/05 20:08:54  jdh8d
 Changes to support the stache, including routines to build a callgraph.

 Some cosmetic changes.

 Minor bug fixes.

 Changed build scripts to build the mips compiler for SimpleScalar if
 SSAS is set to "-DSS_AS"



************************************************************************/

/* if we aren't compiling for the stache, junk this */
#ifdef STACHE

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <strings.h>

#include "vpo.h"
#include "stache.h"
#include "assign.h"
#include "callgraph.h"
#include "cdmotion.h"
#include "combine.h"
#include "controlflow.h"
#include "cse.h"
#include "database.h"
#include "dataflow.h"
#include "eod.h"
#include "lig.h"
#include "links.h"
#include "locals.h"
#include "misc.h"
#include "vectors.h"
#include "vpo2.h"

#include "inst.h"     /* Machine Dependent */
#include "isem.h"     /* Machine Dependent */
#include "regs2.h"    /* Machine Dependent */
#include "rtl.h"      /* Machine Dependent */
#include "y.tab2.h"   /* Tools Dependent */


int stache_policy = NO_STACHE;

struct global_var global_vars[MAX_GLOBALS];
int globals_found = 0;

/* a list of stacheable items */
static struct locuse* stacheables = 0;

int identify_memory_accesses() ;
static void lookup_stacheables_in_LIG();

static void promote_to_stache(struct locuse* lptr);
static void find_stacheables(void);
static void sort_stacheables(void);
static void choose_best_stacheables(void);
static add_stacheable(struct locuse* lptr);
static dump_stacheables(char* msg);
static dump_stacheables_i(FILE* stream, char* mesg, struct locuse* list);



/* the main driver function for stache allocation */
void drive_stache_allocation_phase1(void)
{
	return;

   /* don't do anything if stache allocation isn't turned on */
   if(!swpa && !swpi)
	return;
   /*
      Check to see if the user would like to see the state of the function at
      this point in time.
   */
   if (swd && debug_opt_i("dump_before_stache_alloc1", 1))
      dump_function_i("before stache_alloc()", 1,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP |
                      SHOW_SUCCS | SHOW_PREDS | SHOW_NONRTLS | SHOW_SE);


   /* make determinations about memory accesses */
   if(swpi)
   	identify_memory_accesses();



   /*
      Check to see if the user would like to see the state of the function at
      this point in time.
   */
   if (swd && debug_opt_i("dump_after_stache_alloc1", 1))
      dump_function_i("after stache_alloc()", 1,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP |
                      SHOW_SUCCS | SHOW_PREDS | SHOW_NONRTLS | SHOW_SE);



}
void drive_stache_allocation_phase2(void)
{

   /* don't do anything if stache allocation isn't turned on */
   if(!swpa && !swpi)
        return;
   /*
      Check to see if the user would like to see the state of the function at
      this point in time.
   */
   if (swd && debug_opt_i("dump_before_stache_alloc2", 1))
      dump_function_i("before stache_alloc()", 1,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP |
                      SHOW_SUCCS | SHOW_PREDS | SHOW_NONRTLS | SHOW_SE);

   /* try to pickup any new memory ops that've been added */
   if(swpi)
   	identify_memory_accesses();

   if(stache_policy==STUPID_STACHE_POLICY)
   {
        /* locate variables that're suitable for the stache */
        find_stacheables();

        /* sort those variables based on an objective function */
        sort_stacheables();

        /*make the determination about which should go into the stache*/
        /* and do the promotion */
        choose_best_stacheables();
  }
  else if(stache_policy==COLORING_STACHE_POLICY || 
	  stache_policy==BOTH_STACHE_POLICY)
  {
        /* locate variables that're suitable for the stache */
        find_stacheables();

 	lookup_stacheables_in_LIG();		
  }

  /*
      Check to see if the user would like to see the state of the function at
      this point in time.
   */
   if (swd && debug_opt_i("dump_after_stache_alloc2", 1))
      dump_function_i("after stache_alloc()", 1,
                      SHOW_LABELS | SHOW_STATUS | SHOW_UP |
                      SHOW_SUCCS | SHOW_PREDS | SHOW_NONRTLS | SHOW_SE);


}

static void lookup_stacheables_in_LIG()
{
	struct call_graph_node* cgn;
	struct locuse* lptr;
	struct call_graph_local_info* cgli;
	char* func_id;

	if(!fn)
		return;
        func_id = function_name_to_id(fn);
        if(func_id==0)
        {
                if(debug_opt("dump_stache_errors"))
                        fprintf(stderr, "unable to find function id\n");

                return;
        }
	cgn=add_func_to_call_graph(func_id);
	for(lptr=locs; lptr ; lptr=lptr->next)
	{
		for(cgli=cgn->first_local; cgli; cgli=cgli->next)
		{
			if(strcmp(cgli->code, lptr->code)==0)
			{
				struct LIG_node* lig_node;
				lig_node=cgli->the_LIG_node;
				if(lig_node && lig_node->stache_alloced)
				{
					add_stacheable(lptr);
					lptr->offset=lig_node->start_location;	
					promote_to_stache(lptr);
					lptr->stache_data->is_alloced_to_stache=TRUE;
				}
			}
		}
		
	}
}

/* locate variables that're suitable for the stache */
static void find_stacheables(void)
{
    struct locuse* lptr;
    static int invocation=0;
	

    for (lptr = locs; lptr; lptr = lptr->next) 
    {
	/* intialize just in case */
	lptr->stache_data=0;
	if(
		/* can't promote if it's volatile */
		lptr->vol

		/* lets not mess with the calling sequence for now */
		|| lptr->arg

		/*if it's already in a register, no reason to use stache space*/
		|| lptr->reg!=0

	  )
	{
   	    if (swd && debug_opt_i("dump_stacheables_decision", 1))
	    {
		fprintf(stderr, "%s is not stacheable because: vol=%d, arg=%d,"
			" reg=%d\n", lptr->item, lptr->vol, lptr->arg,
			lptr->reg);
	    }
	
	    continue;
	}

	/* aliasing is a problem for instruction paradigm     */
	/* so if an alias is possible, this one doesn't count */
	if(swpi && lptr->indref)
	{
   	    if (swd && debug_opt_i("dump_stacheables_decision", 1))
	    {
                fprintf(stderr, "%s is not stacheable because: swpi=%d, "
			"indref=%d\n", lptr->item, swpi, lptr->indref);
	
	    }
	    continue;
	}


	/* otherwise, we might be able to allocate this local to the
		stache, so we add it to the list of stacheables*/
   	if (swd && debug_opt_i("dump_stacheables_decision", 1))
		fprintf(stderr, "%s is stacheable!\n", lptr->item);

	add_stacheable(lptr);
    }

    if(swd && debug_opt("dump_stacheables"))
    {
	dump_stacheables("find_stacheables");
    }


    return;
}

/* sort those variables based on an objective function */
static void sort_stacheables(void)
{

    return;
}


/* make the determination about which of those should go into the stache */
/* and do the promotion */
static void choose_best_stacheables(void)
{
	struct bblock* cblk;
	struct list* rtlptr;
	int error=FALSE;
	struct locuse* lptr;
	int stache_used=0;
	int stache_allocation=0;
	char* func_id=0;
	struct call_graph_node* cgn=0;


	/* if it's the instruction paradigm, check that we know what
		we need to know */
        if(swpi)
   	for (cblk = top; cblk; cblk = cblk->down)
 	for(rtlptr=cblk->rtls; rtlptr; rtlptr=rtlptr->next)
	{
		if(rtlptr->type & RTLINE && rtlptr->val)
		{
			char* a=rtlptr->val;
			a=gtmem(a);
			if(a && rtlptr->variable==0)
			{
				fprintf(stderr, 
				  "****************************************\n");
				fprintf(stderr, "Mem Op to new variable in:");
				show(rtlptr->val);
				fprintf(stderr, "Function: %s\n", fn);
				fprintf(stderr, 
				  "****************************************\n");
				error=TRUE;
			}
			else if (a && debug_opt("dump_mem_ops"))
			{
				fprintf(stderr, "Mem Op to old variable in:");
				show(rtlptr->val);
				if(isLocal(rtlptr->variable))
					fprintf(stderr, "Local: %s\n", 	
					  localname(rtlptr->variable));
				else if(isGlobal(rtlptr->variable))
					fprintf(stderr, "Global: %s\n", 	
					  globalname(rtlptr->variable));
				else
					fprintf(stderr, "Unknown: %s\n", 	
					  rtlptr->variable);
			}
		}

	}
	/* lets not make guesses.. if we found a new variable, we split 
	 * if we don't know the name of this function, bail
         * if we don't find ourselves in the call graph, bail
	 */
	if(fn==0)
	{
		if(debug_opt("dump_stache_errors"))
			fprintf(stderr, "No function name\n");
		return;
	}
	if(error) 
	{
		if(debug_opt("dump_stache_errors"))
			fprintf(stderr, "Found stache error\n");
		return;
	}

	func_id = function_name_to_id(fn); 
	if(func_id==0)
	{
		if(debug_opt("dump_stache_errors"))
			fprintf(stderr, "unable to find function id\n");
		
		return;
	}
	cgn=add_func_to_call_graph(func_id);
	stache_allocation=cgn->stache_usage;

	/* bail if we aren't supposed to use the stache */
	if(stache_allocation<=0)
	{
		if(debug_opt("dump_stache_errors"))
			fprintf(stderr, "No stache allowed for %s\n", fn);
		return;
	}


	/*
	 * Look through the stachables in order and promote everything
	 * that goes.  Fixentry might fill in "gaps" with saved registers,
	 * etc so don't worry about that sort of thing here 
	 */
	
	for(lptr=locs; lptr ; lptr=lptr->next)
	{
		/* simply having the stache_data structure means
		 *	that it's possible to put this local in
		 *	the stache.
		 */
		if(!lptr->stache_data)
			continue;
		else if (debug_opt("dump_stacheables_decisions"))
		{
			fprintf(stderr, "%s is not stacheable\n",
				lptr->item);
		}

		/* enough room left for this one? */
		if(lptr->size <= stache_allocation)
		{
			stache_allocation-=lptr->size;

			/* round up to the nearest multiple of 8, just
			 * in case-- make sure we round down so
			 * we dont overallocate
			 */
			if(stache_allocation%8!=0)
			{
				stache_allocation&=7;
			}
			

			/* we set a flag and let fixentry know the 
			 * item is in stache 
			*/
			lptr->stache_data->is_alloced_to_stache=TRUE;	

			/* we look through the rtl's here and change $sp
				references to $fp references 
			   actually, this just calls "make_promotion"
				on each RTL, because promoting is machine
				dependent
			*/
			if(debug_opt("dump_stache_promotions"))
				fprintf(stderr, "%s is promoted, yay! \n",
					lptr->item);
			promote_to_stache(lptr);
			break;
		}
		else if (debug_opt("dump_stacheables_decision"))
		{
			fprintf(stderr, "Not enough room for %s in stache\n",
				lptr->item);
		}
	}

    	return;
}

static void promote_to_stache(struct locuse* lptr)
{
	struct bblock* cblk;
	struct list* rtlptr;

   	for (cblk = top; cblk; cblk = cblk->down)
 	for(rtlptr=cblk->rtls; rtlptr; rtlptr=rtlptr->next)
	{
		make_promotion(rtlptr, lptr);
	}
	

}

static add_stacheable(struct locuse* lptr)
{
    	struct stacheable* temp;
    	struct locuse* list;

	for(list=stacheables; list; list=list->stache_data->next)
		if(list==lptr)
			return;

	temp=(struct stacheable*)
		CALLOC_PERM(sizeof(struct stacheable));


	/* initialize the fields */
	temp->is_alloced_to_stache=0;
	temp->benefit=lptr->reps;

	/* add it to the locuse structure */
	lptr->stache_data=temp;


	/* and link it into the list of data */
	temp->next=stacheables;
	stacheables=lptr;

}

static dump_stacheables(char* msg)
{
	dump_stacheables_i(stderr, msg, stacheables);
}
static dump_stacheables_i(FILE* stream, char* mesg, struct locuse* list)
{
	fprintf(stream, "\n");
	if(fn)
	    fprintf(stream, "Dumping stacheables for %s: %s\n", fn, mesg);	
	else
	   fprintf(stream, "Dumping stacheables: %s\n", mesg);	
	for(; list; list=list->stache_data->next)
	{
		assert(list); 
		fprintf(stream, "-----------------------------------------\n");
		fprintf(stream, "%s: ", list->item);
		fprintf(stream, "benefit:%d ", list->stache_data->benefit);
		fprintf(stream, "size:%d ", list->size);
		fprintf(stream, "reps:%d ", list->reps);
		if(list->stache_data->is_alloced_to_stache)
			fprintf(stream, "allocated");
		else
			fprintf(stream, "not allocated");
			

		fprintf(stream, "\n");

	}	
	fprintf(stream, "-----------------------------------------\n");
	fprintf(stream, "\n");

}


static ident_invocation=0;

void cleanup_stacheables(void)
{
	struct bblock* cblk;
	struct list* rtlptr;
	struct locuse *list=stacheables, *next;

return;
	for(; list; list=next)
	{
		next=list->stache_data->next;
		FREE(list->stache_data);
	}

   	for (cblk = top; cblk; cblk = cblk->down)
		for(rtlptr=cblk->rtls; rtlptr; rtlptr=rtlptr->next)
		{
			if(rtlptr->variable!=0 && 
				strcmp(rtlptr->variable,"Unknown")!=0)
			{
				FREE(rtlptr->variable);
			}
			rtlptr->variable=0;
		}

	ident_invocation=0;
	stacheables=0;

}

/* pray */
int identify_memory_accesses() 
{
   int invocation=ident_invocation++;
   register struct bblock *cblk;
   struct list* rtlptr;
   int change;

   invocation++;

   /*
      Check to see if the user would like to see the state of the function at
      this point in time.
   */
   if (swd &&
       debug_opt_i("dump_before_identify_memory_accesses", invocation))
      dump_function_i("before identify_memory_accesses()", invocation,
                      SHOW_LABELS | SHOW_PREDS | SHOW_NONRTLS | SHOW_SE);


   if(swd && debug_opt("ident_mem_ops"))
	fprintf(stderr, "\n\n\n\n\n\n\n");;
   change = FALSE;


   /*
        Mark all of the basic blocks UNDONE and perform CSE on the function
        starting with the entry block.
   */
   for (cblk = top; cblk; cblk = cblk->down)
      STATUS(cblk, UNDONE);
   change |= annotate_memory_ops(top, invocation);

   /*
         Go back and pick up any blocks that were left in a dormant state
   */
   for (cblk = top; cblk; cblk = cblk->down)
   {
      if (cblk->status & (DRYRUN | DORMANT))
         change |= annotate_memory_ops(cblk, invocation);
   }

   /*
         Ensure that we have turned off the request for side effect information
         from the machine description.
   */
   side_effect = (char *) NULL;

   /*
         Make a final pass through the basic blocks to ensure that every block
         was properly examined and that equivalence tables are returned to the
         free memory list so that they can be re-used in the next pass.
   */
   for (cblk = top; cblk; cblk = cblk->down) 
   {
      if (!(cblk->status & DONE))
         fprintf(stderr,"vpo (Ident_mem_ops): basic block not examined\n");
      if (cblk->elist) 
      {
         clrequiv(cblk->elist);
         FREETYPE(cblk->elist, EQVTABLETYPE);
         cblk->elist = (struct eqvtable *) NULL;
      }
   }



   /* any rtl's that don't have a "variable" associated with them are
	marked as unknown */

   if(invocation==1)
   {
      for (cblk = top; cblk; cblk = cblk->down)
      {
	for(rtlptr=cblk->rtls; rtlptr; rtlptr=rtlptr->next)
	{
		if(rtlptr->variable==0) 
		{
			rtlptr->variable="Unknown";
		}
	}
      }
   }


   /*
      Check to see if the user would like to see the state of the function at
      this point in time.
   */
   if (swd &&
       debug_opt_i("dump_after_identify_memory_accesses", invocation))
      dump_function_i("after identify_memory_accesses()", invocation,
                      SHOW_LABELS | SHOW_PREDS | SHOW_NONRTLS | SHOW_SE);

}

int annotate_memory_ops(register struct bblock * cblk, int invocation) 
{
   register struct blist *bptr;
   register int go, dryrun;
   int change;

   change = FALSE;

   /*
      Determine if this basic block has been dry runned.  A dry run is an
      initial trial pass through the basic blocks that make up a loop.  The
      intent of a dry run is to determine if there are any equivalences that
      are available at the end of each loop that match those available at the
      entry point.
   */
   dryrun = (cblk->status & DRYRUN);

   /*
      Set the status of the current block to WAITING.  This will allow us to
      determine that this block is being worked on when recursive calls are
      made to this function.
   */
   STATUS(cblk, WAITING);

   /*
      Set the "go" status to permit this basic block to be examined.  Later, we
      may encounter circumstances that will prevent us from going ahead and
      examining this basic block at this time.
   */
   go = TRUE;

   /*
      If this basic block has predecessors, then we need to determine what we
      know about the equivalences at this point from them.
   */
   if (bptr = cblk->preds) 
   {
      /*
         If this basic block has more than one predecessor, then we are going
         to have to merge equivalence classes to obtain the equivalence items
         that are valid for this basic block.
      */
      if (bptr->next) 
      {

         /*
            We first examine each predecessor and try to ensure that equiv list
	    gathering  has
            been performed on each of them.  Since many functions contain
            loops, there will be cases where not all of the predecessors can
            have equiv list gathering performed on them at this time.
         */
         while (bptr) 
         {

            /*
               If we have not yet worked on the predecessor to this
               basic block, then we should go ahead and try do so in order to
               be able to start with more than just a blank equivalence list
               for this basic block.
            */
            if (bptr->ptr->status & UNDONE)
               change |= annotate_memory_ops(bptr->ptr, invocation);

            /*
               If something has prevented us from looking at the
               predecessor, then we clear the "go" flag which will prevent us
               from attempting to perform CSE on the current basic block just
               yet.
            */
            if (bptr->ptr != cblk && bptr->ptr->status & WAITING)
               go = FALSE;

            bptr = bptr->next;
         }

         /*
            If we still have a "go" indication for this basic block, we can
            continue the process of merging the equivalence lists of the
            predecessors.
         */
         if (go) 
         {

            /*
               If any predecessors indicate that they have not has CSE
               performed on them appropriately, then we cannot go ahead and do
               CSE on the current basic block.  This will happen when a loop
               that has not yet been dry runned is encountered.
            */
            for (bptr = cblk->preds; bptr; bptr = bptr->next) 
            {
               if (bptr->ptr->status & ~(DONE | DRYRUN | CLEAR) &&
                   !(dryrun && bptr->ptr == cblk)) 
               {
                  go = FALSE;
                  break;
               }
            }

            /*
               If we have valid information for all of the predecessors, then
               we can go ahead and merge their equivalence information to yield
               a list of the equivalences that are available from all of the
               possible control flow paths into this basic block.
            */
            if (go)  
            {
               dryrun = FALSE;
               bptr = cblk->preds;
               /*
                  We need to pick a place to start our merging process, if the
                  current basic block has itself as a predecessor, then if we
                  wish to start off the merge list by obtaining that
                  equivalence information first before we end up wiping it out
                  to hold the new merge list.
               */
               if (bptr->ptr != cblk) 
               {
                  /*
                     The first predecessor it not this current basic block, but
                     a quick check can be used here to determine if we will be
                     finding this block later in the predecessor list.  If so,
                     we grab the equivalence information first (if there is
                     any) so that we have it before it becomes corrupted.  If
                     we encounter it later in the merging process, we will know
                     to ignore it.
                  */
                  if (cblk->left == cblk || cblk->right == cblk) 
                  {
                     /*
                        In the case that we don't have any information for our
                        own basic block, we start with a blank equivalence
                        table.
                     */
                     if (!cblk->elist)
                        cblk->elist = newequiv();

                     /*
                        We expect to have an equivalence list to merge in for
                        every predecessor, and we complain if we do not.
                     */
                     if (!bptr->ptr->elist)
                        (void) fprintf(stderr,
                                       "vpo (blahblah): elist not available\n");
                     else
                        e_merge(cblk->elist, bptr->ptr->elist);

                     /*
                        Return the equivalence table of the predecessor that we
                        have just merged in to the free list if we determine
                        that we will not be needing it again in the future.
                     */
                     if (!(bptr->ptr->status & DRYRUN) &&
                         !--bptr->ptr->elist->uses) 
                     {
                        clrequiv(bptr->ptr->elist);
                        FREETYPE(bptr->ptr->elist, EQVTABLETYPE);
                        bptr->ptr->elist = (struct eqvtable *) NULL;
                     }
                  }

                  /*
                     The current basic block is not a predecessor of itself, so
                     we can go ahead and start a merge list without considering
                     any equivalence table information that this basic block
                     might have had.
                  */
                  else 
                  {

                     /*
                        If we had a previous equivalence list, then we go ahead
                        and return it to the free list.
                     */
                     if (cblk->elist) 
                     {
                        clrequiv(cblk->elist);
                        FREETYPE(cblk->elist, EQVTABLETYPE);
                     }

                     /*
                        If we will need to reference the predecessor's
                        equivalence information in the future, then we make a
                        copy of it in order to start our own merge list.
                     */
                     if (bptr->ptr->status & DRYRUN || --bptr->ptr->elist->uses)
                     {
                        cblk->elist = newequiv();
                        if (!bptr->ptr->elist)
                            fprintf(stderr, 
				"vpo (remv_common_subexpressions): "
				"elist not available\n");
                        else
                           e_copy(cblk->elist, bptr->ptr->elist);
                     }

                     /*
                        Otherwise, we will not need the predecessor's
                        equivalence information again, so instead of copying
                        it, we go ahead and take it so that it will now be the
                        start of our merge list.
                     */
                     else 
                     {
                        if (!(cblk->elist = bptr->ptr->elist)) 
                        {
                           fprintf(stderr,"vpo (): elist not available\n");
                           cblk->elist = newequiv();
                        }
                        bptr->ptr->elist = (struct eqvtable *) NULL;
                     }
                  }
               }
               /*
                  The merging process continues over all of the predecessors of
                  the current basic block.
               */
               while (bptr = bptr->next) 
               {

                  /*
                     Never bother to merge our own information with ourselves,
                     since we have already taken it into account at the start
                     of the merge process.  Otherwise, continue to merge in all
                     of the predecessor information to determine what, if
                     anything, is equivalent when we reach this basic block.
                  */
                  if (bptr->ptr != cblk) 
                  {

                     /*
                        We should have an equivalence structure for each
                        predecessor, but we check here just for the sake of
                        paranoia and print out an error if we don't have one.
                     */
                     if (bptr->ptr->elist)
                        e_merge(cblk->elist, bptr->ptr->elist);
                     else 
                     {
                        fprintf(stderr, "vpo (): elist not available\n");
                        e_merge(cblk->elist, newequiv());
                     }

                     /*
                        Return the equivalence table of the predecessor that we
                        have just merged in to the free list if we determine
                        that we will not be needing it again in the future.
                     */
                     if (!(bptr->ptr->status & DRYRUN) &&
                         !--bptr->ptr->elist->uses) 
                     {
                        clrequiv(bptr->ptr->elist);
                        FREETYPE(bptr->ptr->elist, EQVTABLETYPE);
                        bptr->ptr->elist = (struct eqvtable *) NULL;
                     }
                  }
               }
            }

            /*
               Otherwise, if the block was initially set up for a dry run, then
               we continue to leave it in that state.
            */
            else if (dryrun) 
            {
               STATUS(cblk, DRYRUN);
               return (change);
            }

            /*
               Otherwise, it is time to perform a real dry run from this basic
               block, so we initialize a new equivalence list and start the dry
               run at this point.
            */
            else 
            {
               go = dryrun = TRUE;

               /*
                  Check to see if we have loop information for this basic
                  block. If we do, we might be able to take advantage of loop
                  invariant information gathered by the code motion phase.
               */
               if (cblk->loop) 
               {

                  if (swd &&
                      debug_opt_i_l("ident_memops_dryrun",
                                    invocation, cblk->label))
                     (void) fprintf(stderr, "Dryrun for %s\n", cblk->label);

                  /*
                     In order to take advantage of the loop invariant items
                     that may enter this loop, we go ahead and merge in the
                     equivalences of all of the predecessors that we have
                     equivalence information for.
                  */
                  for (bptr = cblk->preds; bptr; bptr = bptr->next) 
                  {
                     if (!block_in_loop(cblk->loop, bptr->ptr) &&
                         bptr->ptr->status & DONE && bptr->ptr->elist) 
                     {
                        if (swd &&
                            debug_opt_i_l("ident_memops_dryrun",
                                          invocation, cblk->label)) 
                        {
                           if (bptr->ptr->label)
                              (void) fprintf(stderr, "%s\n", bptr->ptr->label);
                           e_show(bptr->ptr->elist);
                        }
                        if (cblk->elist)
                           e_merge(cblk->elist, bptr->ptr->elist);
                        else  
                        {
                           cblk->elist = newequiv();
                           e_copy(cblk->elist, bptr->ptr->elist);
                        }
                     }
                  }

                  /*
                     Remove any equivalences that contain non-constant items,
                     leaving only valid constant and loop-invariant items to
                     start the loop with.
                  */
                  if (cblk->elist)
                     e_invar(cblk->elist, cblk->loop);
                  else
                     cblk->elist = newequiv();

                  if (swd &&
                      debug_opt_i_l("ident_memops_dryrun", invocation, 
			cblk->label)) 
		  {
                     fprintf(stderr, "End of dryrun for %s\n", cblk->label);
                     e_show(cblk->elist);
                  }
               }

               /*
                  If we have no loop information for this basic block, start
                  the dry run with an empty equivalence table.
               */
               else
                  cblk->elist = newequiv();
            }
         }
      }

      /*
         Otherwise, this basic block has only a single predecessor, so we do
         not need to merge equivalence lists.
      */
      else 
      {

         /*
            If we have not yet performed CSE on the predecessor to this basic
            block, then we should go ahead and do so in order to be able to
            start with more than just a blank equivalence list for this basic
            block.
         */
         if (bptr->ptr->status & UNDONE)
            change |= annotate_memory_ops(bptr->ptr, invocation);

         /*
            If we were able to perform CSE on the predecessor block, then go
            ahead and set up a equivalence list so that we can perform CSE on
            this block.
         */
         if (bptr->ptr->status & (DONE | DRYRUN)) 
         {

            /*
               If the pass on the predecessor was a dry run, then the
               current basic block must also be part of a loop so that this
               will also be the dry run for this basic block.
            */
            if (dryrun = (bptr->ptr->status & DRYRUN)) 
            {

               /*
                  If the preceeding basic block has more than one child, then
                  we make a copy of its equivalence list.
               */
               if (bptr->ptr->elist->uses) 
               {
                  cblk->elist = bptr->ptr->elist;
                  bptr->ptr->elist = newequiv();
                  if (cblk->elist)
                     e_copy(bptr->ptr->elist, cblk->elist);
                  else 
                  {
                     fprintf(stderr, "vpo (): elist not available\n");
                     cblk->elist = newequiv();
                  }
               }

               /*
                  Otherwise, we can go ahead and take the equivalence list,
                  since no one else will be needing it at this point.
               */
               else 
	       {
                  if (!(cblk->elist = bptr->ptr->elist)) 
                  {
                     fprintf(stderr, "vpo (): elist not available\n");
                     cblk->elist = newequiv();
                  }
                  bptr->ptr->elist = (struct eqvtable *) NULL;
               }
            }

            /*
               Otherwise, this is not a dry run, so we proceed with a copy of
               the predecessor's equivalence list.  In this case, the
               predecessor has more than one child, and it has not given a copy
               to all of it's children yet, so we just make a copy of the
               equivalence list and leave the original one for the other child.
            */
            else if (--bptr->ptr->elist->uses) 
            {
               cblk->elist = newequiv();
               if (bptr->ptr->elist)
                  e_copy(cblk->elist, bptr->ptr->elist);
               else  
               {
                  (void) fprintf(stderr, "vpo (): elist not available\n");
                  e_copy(cblk->elist, newequiv());
               }
            }

            /*
               Otherwise, we can take the equivalence list from the preceeding
               basic block, since no one else will need to use it.
            */
            else 
            {
               if (!(cblk->elist = bptr->ptr->elist)) 
               {
                  (void) fprintf(stderr, "vpo (): elist not available\n");
                  cblk->elist = newequiv();
               }
               bptr->ptr->elist = (struct eqvtable *) NULL;
            }
         }

         /*
            Otherwise, we were not able to perform CSE on the predecessor, so
            we must put this block on hold for now.
         */
         else
            go = FALSE;

      }
   }

   /*
      Otherwise, this must be the entry block to the function, so we start with
      a clear equivalence list.
   */
   else
      cblk->elist = newequiv();

   /*
      If everything is set up to perform annotations on this basic block, then 
	do it.
      If this was just a dry run pass, then mark the block as having gone
      through its dry run, otherwise, mark it done.  Once a block is examined,
      move on to its children and call this function recursively in order to
      attempt to examine all of the basic blocks.  If this block has more than
      one child, then forget set information, so we don't try to delete sets
      that may be needed in more than one place.
   */
   if (go) 
   {
      if(cblk->loop)
        e_invar(cblk->elist, cblk->loop);
      change |= annotate_basic_blocks(cblk, invocation);
      if (dryrun)
         STATUS(cblk, DRYRUN);
      else
         STATUS(cblk, DONE);
      if (cblk->right)
         e_clrsets(cblk->elist);
      if (cblk->left && (cblk->left->status & (UNDONE | DORMANT | DRYRUN)))
         change |= annotate_memory_ops(cblk->left, invocation);
      if (cblk->right && (cblk->right->status & (UNDONE | DORMANT | DRYRUN)))
         change |= annotate_memory_ops(cblk->right, invocation);
   }

   /*
      If we do not get the go-ahead for this basic block, then place it in a
      dormant state for now.  We will be picked up again at some point in the
      future when the status of any of our predecessors changes.
   */
   else
      STATUS(cblk, DORMANT);
   /*
      Return indicating if any changes have been made to the function
   */
   return (change);
}


int annotate_basic_blocks(struct bblock * cblk, int invocation) 
{

   register struct list *ptr;
   register char *rtl, *a, *b;
   register struct eqvlist *ptr1, *ptr2;
   int change, cost, diff, i, check, kind;
   char *s, new[BIGSTRING], src[BIGSTRING], dst[BIGSTRING], se[MAXSTRING];

   change = FALSE;

   if (swd && cblk->label && debug_opt_i_l("ident_mem_ops",
                     invocation, cblk->label))
      (void) fprintf(stderr, "%s:\n", cblk->label);

   /*
      Look at each line in the basic block
   */
   for (ptr = cblk->rtls; ptr; ptr = ptr->next) 
   {
      if (ptr->type & RTLINE) 
      {
         if (rtl = ptr->val) 
         {
	    strcpy(new, ptr->val);
            deadlist = ptr->deads;
            curptr = ptr;

            /*
               Ensure that we have the most up-to-date information on the cost
               and side effects of this RTL.
            */
            if (ptr->cost && ptr->type & SIDEFFECT) 
            {
               if (ptr->se)
                  (void) strcpy(se, ptr->se);
               else
                  *se = '\0';
               cost = ptr->cost;
               kind = ptr->kind;
            }
            else 
            {
               sem_ptr = str_ptr = i_ptr = erflag = FALSE;
               linptr = rtl;
               side_effect = se;
               *se = '\0';
               icost = -1;
               ikind = 0;
               if(find(curptr, rtl, deadlist))
               {
                        bad_rtl_line_warning("annotate_basic_block", curptr);
               }
               else 
               {
                  cost = ptr->cost = icost;
                  kind = ptr->kind = ikind;
               }
            }
            if (cost < 0)  
            {
               fprintf(stderr, "vpo (cse_basic_block): icost not set for ");
               show_rtl(stderr, rtl);
               (void) fprintf(stderr, "\n");
               exit(1);
            }

            /*
                  Update the use counts on each item used by the instruction.
                  Note that items comprising the address of a memory reference
                  are used, not set.
            */
            rtl = new;
            while (rtl = gtds(rtl, dst, src)) 
            {
               if (ismem(dst))
                  upduse(cblk->elist, dst + 2, ptr);
               upduse(cblk->elist, src, ptr);
            }
            /*
                  The following code examines each of the items set in the side
                  effects of an instruction.  Any item which is also set in the
                  actual instruction is removed from the side effect.  The
                  reason this is done is to prevent the following senario:

               r[1] = R[r[1]++];       (side effect) = r[1] = r[1] + 4;

               Since the side effects are parsed after the main instruction,
                  this particular beastie would leave the equivalence table
                  indicating that r[1] had a value that was four greater than
                  its actual value.  This, of course, has the underlying
                  assumption that well-behaved machines would not increment
                  r[1] after storing a new value into it.  You might also be
                  asking yourself why anybody would be using an autoincrement
                  when the value of the register incremented will be thrown
                  away?  The answer is that there probably exists a machine
                  that can only reference memory with an autoincrement.

               Note that we also take this opportunity to update the reference
                  counts of each item used in the side effect. In this case we
                  do not check for memory references in the destination,
                  because no machine yet encountered has required a side effect
                  that sets a memory location.
            */
            if (*se) 
            {
               rtl = b = se;
               while (*(s = b)) 
               {
                  a = dst;
                  while ((*a++ = *b++) != '=')
 			;
                  *(a - 1) = '\0';
                  if (isset(new, dst))
                     while (*b++ != ';')
			;
                  else 
		  {
                     a = src;
                     while ((*a++ = *b++) != ';')
 			;
                     *(a - 1) = '\0';
                     if (*src)
                        upduse(cblk->elist, src, ptr);
                     b = s;
                     while ((*rtl++ = *b++) != ';')
			;
                  }
               }
               *rtl = '\0';
            }




/*

	look up each memory reference and print out
	the oldest "cannonical" form ofr the address.

*/

	find_locals(ptr);


            /*
                  Enter each RTL and side effect item into the CSE equivalence
                  list database for this basic block.
            */
	    strcpy(new, ptr->val);
            rtl = new;
            while (rtl = gtds(rtl, dst, src))

#ifdef FIFOS
               change |= annotate_update_equivalence_table(
			cblk->elist, dst, src, ptr, (*rtl) ? FALSE : TRUE);
#else
               change |= annotate_update_equivalence_table(
			cblk->elist, dst, src, ptr);
#endif

            rtl = se;
            while (rtl = gtds(rtl, dst, src))

#ifdef FIFOS
               change |= annotate_update_equivalence_table
			(cblk->elist, dst, src, ptr, (*rtl) ? FALSE : TRUE);
#else
               change |= annotate_update_equivalence_table
			(cblk->elist, dst, src, ptr);
#endif

	 }
         /*
            This RTL line contains only a dead register list, but we still need
            to examine it in order to determine if any registers that were set
            were not used.
         */
         else
            change |= cse_dead_register_list(cblk->elist, ptr, ptr->val);

      }
      else if (ptr->type & RESLINE) 
      {
         if (ptr->val)
            if (ptr1 = e_lookup_reg(cblk->elist, ptr->val))
               ptr1->dead = (struct list *) NULL;
      }
   }
}


#ifdef FIFOS
int annotate_update_equivalence_table(
	struct eqvtable *list, char *new, char *old, struct list *ptr, int last)
#else
int annotate_update_equivalence_table(
	struct eqvtable * list, char *new, char *old, struct list * ptr)
#endif

{
   register struct eqvlist *eptr;
   register char *a;
   struct eqvlist *nptr;
   struct list *dies;
   int change, match;
   char *p1, *p2, reg[BIGSTRING], memref[BIGSTRING];

#ifdef FIFOS
   char sav[BIGSTRING], savnew[BIGSTRING];

#endif

#ifdef AUTO
   /*
      Remove any auto-increment and auto-decrement artifacts from the RTL to
      reveal the actual value of each item in the destination.
   */
   remauto(new);
#endif

   change = FALSE;

#ifdef AUTO

   /*
      Remove auto-increment and decrement artifacts from each item in the
      source expression.
   */
   remauto(old);
#endif

   /*
      Any register which is used in the source whose re-computable value is
      known can no longer be re-computable, so go ahead and remove its value
      from the equivalence list.  Also, if the destination is a register and it
      is used in the source, set match TRUE.
   */
   match = FALSE;
   a = old;
   while (a = gtrg(a)) {
      if(cmploc(new,unskiploc(a))!=0 || !*new) {
/*

	I just want the register equivilences so we aren't going
	to get rid of these just yet.  Keep them around in case
	so we know *all* equivilences regardless of whether it's usable

         mangleReg(reg,unskiploc(a),"/");
         *(skipmangledreg(reg,"/"))=0;
         e_delete(list, reg);
*/
      }
      else
         match = TRUE;
   }

   /*
      If there is no destination item given, or if the destination item is
      something that CSE should not bother with, then return now without
      bothering to enter anything in the equivalence table.
   */
   if (!*(p2 = new) || dont_bother_with(new)) {

#ifdef FIFOS

      /*
         A fifo register value is consumed once it is used.  If there are
         values queued up for the fifo consumed, then alter their labels to
         move up the fifo values.
      */
      a = old;
      while (a = gtfrg(a)) {
         gtstr(unskiploc(a), a, reg);
         consume(list, reg);
      }
#endif

      return (change);
   }

   /*
      If the source is something that CSE should not be bothering with, then
      remove any knowledge of the destination item from the equivalence table
      and return at this point.
   */
   if (dont_bother_with(old)) {
      adstroy(list, new);
      if (ismem(new)) {
         a = new;
         while (a = gtmem(a)) {
            gtstr(p1 = bpos + 2, p2 = epos - 1, memref);
            if (oldest(list, memref))
               if (a > p2)
                  a = sub(p1, p2, memref) + 1;
               else
                  (void) sub(p1, p2, memref);
         }
         mem_aliases(list, new);
      }
      return (change);
   }

   /*
      If a register is used in a memory reference then its old re-computable
      value cannot be rescued, so go ahead and remove it from the equivalence
      list.
   */
   while (p2 = gtmem(p2)) {
      a = memref;
      gtstr(bpos + 2, epos - 1, a);
      while (a = gtrg(a)) {
/*
	again, keep thesea round for full information 

         mangleReg(reg,unskiploc(a),"/");
         *(skipmangledreg(reg,"/"))=0;
         e_delete(list, reg);
*/
      }
   }

   dies = (struct list *) NULL;

#ifdef FIFOS
   (void) strcpy(savnew, new);
   (void) strcpy(sav, old);

   /*
      We give fifo registers a different treatment when they are assigned a new
      value.  If they have an old value, then the new value does not replace
      the old one, rather, the new value is queued so that it does not become
      the value of the fifo register until all of the queued values are
      consumed.  In the equivalence list, there is no structure for keeping
      track of queues, so we simulate the effect by assigning the value at the
      head of the queue to the fifo register (let's call it f[0] for this
      example).  The value queued behind the head of the queue is noted by the
      label "f*0*[0]", and the one after that is "f*1*[0]" and so on.  So the
      queued form of register f[0] is f*n*[0], where n is a number that
      differentiates the queue order.  The following code runs through the
      existing fifo values for a fifo register and changes the string "new" to
      the appropriate label to queue up this assignment to the fifo register.
   */
   if (isreg(new) && fiforeg(new)) {
      if (eptr = e_lookup_reg(list, new)) {
         for (match = 0; eptr; match++) {
            (void) sprintf(memref, "%c*%d*%s", *new, match, new + 1);
            eptr = e_lookup(list, memref);
         }
         (void) strcpy(new, memref);
      }
   }
   else
#endif

      /*
         If the destination is a register and it is not used in the source, and
         we had an old value for the register, then save that old value because
         we might just be re-computing what we had in the register.  The old
         value is saved as the register string with an '*' inserted as the
         second character.
      */

	/* here I change it so the dest is a regsiter only.  we save old
		values whenever we can for full information */

   if (isreg(new) && (eptr = e_lookup_reg(list, new))) {
      mangleReg(memref,new,"/");
      *(skipmangledreg(memref,"/"))=0;

      if (nptr = e_insert(list, memref)) {
         nptr->sets = eptr->sets;
         nptr->uses = eptr->uses;
         nptr->dead = 0 ; /*eptr->dead;*/
         if (nptr->dead == ptr)
            dies = nptr->dead;
         if (nptr->pequiv)
            nptr->pequiv->nequiv = nptr->nequiv;
         if (nptr->nequiv) {
            nptr->nequiv->pequiv = nptr->pequiv;
            nptr->nequiv = (struct eqvlist *) NULL;
         }
         while (eptr->nequiv)
            eptr = eptr->nequiv;
         eptr->nequiv = nptr;
         nptr->pequiv = eptr;
      }
   }

   /*
      Replace memory reference of the destination with the oldest value
   */
   a = new;
   while (a = gtmem(a)) {
      gtstr(p1 = bpos + 2, p2 = epos - 1, memref);
      if (oldest(list, memref))
         if (a > p2)
            a = sub(p1, p2, memref) + 1;
         else
            (void) sub(p1, p2, memref);
   }


   /*
      If the destination item is a memory location, then examine the
      equivalence list and remove any memory locations that may be aliases.
   */
   if (ismem(new))
      mem_aliases(list, new);

   /*
      Obtain the oldest representaion of the source item
   */
   (void) oldest(list, old);

   /*
      Look to see if we already have an equivalence entry for the source item
      in the equivalence table and if the destination item is used in the
      source expression.
   */
   if ((match = fmatch(old, new)) && (eptr = e_lookup(list, old))) {

      /*
         The source and destinations may be equivalent, so we have to be a
         little careful about how we manipulate them.
      */
      if (strcmp(old, new)) {

         /*
            Disengage the destination item from the equivalence table.
            Re-insert it back in so that it now has no equivalences.
         */
         e_delete(list, new);
         nptr = e_insert(list, new);

         /*
            Attach the destination item to the equivalence set of the source
            item, and remove the source item from its equivalence set.
         */
         nptr->nequiv = eptr->nequiv;
         nptr->pequiv = eptr->pequiv;
         if (nptr->pequiv)
            nptr->pequiv->nequiv = nptr;
         if (nptr->nequiv)
            nptr->nequiv->pequiv = nptr;
         eptr->nequiv = (struct eqvlist *) NULL;
         eptr->pequiv = (struct eqvlist *) NULL;
      }
      else
         nptr = eptr;

      /*
         The following is a cheap trick.  It is intended to delete all
         equivalence list references to the destination item, which is being
         changed, except for the actual item that has just been inserted.  The
         method employed is to hide the name of the destination item so that it
         will not match the deletion criteria.
      */
      a = nptr->item;
      nptr->item = "*";
      adstroy(list, new);
      nptr->item = a;
   }
   else

      /*
         Otherwise, destroy all references that use the destination item, whose
         value is now changed.
      */
      adstroy(list, new);

   /*
      Update the destination item's information to reflect the fact that it has
      just been assigned at this point.
   */
   nptr = e_insert(list, new);
   nptr->sets = ptr;
   nptr->dead = dies;
   nptr->uses = 0;

   /*
      If the source item has no releation to the destination and it has not
      already appeared before, add it to the equivalence list and note its
      equivalence to the source.
   */
   if (!match && (eptr = e_insert(list, old))) {
      while (eptr->nequiv)
         eptr = eptr->nequiv;
      eptr->nequiv = nptr;
      nptr->pequiv = eptr;
   }

#ifdef FIFOS

   /*
      A fifo register value is consumed once it is used.  If there are values
      queued up for the fifo consumed, then alter their labels to move up the
      fifo values.
   */
   if (last) {
      a = sav;
      while (a = gtfrg(a)) {
         gtstr(unskiploc(a), a, reg);
         consume(list, reg);
      }
      p2 = savnew;
      while (p2 = gtmem(p2)) {
         a = memref;
         gtstr(bpos + 2, epos - 1, a);
         while (a = gtfrg(a)) {
            gtstr(unskiploc(a), a, reg);
            consume(list, reg);
         }
      }
   }
#endif

   /*
      Indicate if any changes were made to the function while updating the
      equivalence table.
   */
   return (change);
}



void read_globals_profile()
{
	FILE* fin;

	bzero(global_vars, sizeof(global_vars));
	fin=fopen(dataProfileName,"r");

	if(!fin)
	{
		fprintf(stderr,"Cannot open %s for reading, abandoning "
			"global profiling\n", dataProfileName);
		return;
	}

	while(!feof(fin))
	{
		char name[100];
		int count;
		float percent_count;
		int size;
		int i;
		if(fscanf(fin, "%s%d%f%d", name, &count, &percent_count, &size)
			!=4)
			break;
		if(count==0)
			continue;
		
		for(i=0;i<globals_found;i++)
		{
			if(strcmp(name, global_vars[i].name)==0)
			{
				break;
			}
		}
		if(i==globals_found)
			globals_found++;

		global_vars[i].name=(char*)strdup(name);
		global_vars[i].id[0]=0;
		global_vars[i].use_count+=count;
		global_vars[i].rel_use_count+=percent_count;
		global_vars[i].size=size;
		global_vars[i].compiling=0;
		global_vars[i].is_stached=0;
		global_vars[i].rel_benefit=0;
		if(debug_opt("dump_dataprofile"))
		{
			fprintf(stderr, "%d Found %s, size=%d,"
				" count=%d, per=%f\n",
				i, name,size,count,percent_count);
		}
	}
}

static int global_compare(const void* a, const void* b);

void assign_globals_to_stache()
{
	int i;
	static int stache_used=0;
	char type;
	char* line;

	rewind(stdin);
	if(debug_opt("dump_assign_globals_to_stache"))
		fprintf(stderr, "in assign_globals_to_stache()");
	while(!feof(stdin))
	{
		switch(type=getchar())
		{
			case EOF:
				line=".text";
				find_global_declaration(line);
				goto make_assignments;

			case '-':
			case '_':
				find_global_declaration(line=gtline());
				FREE(line);
				break;
			case '*':
				getchar();
				break;
			default:
				line=gtline();
				FREE(line); 	/* can't FREE(getline()) here */
				break;
		}
	}

make_assignments:

	rewind(stdin);

	/* assign benefits */
	for(i=0;i<globals_found;i++)
	{
		if(global_vars[i].compiling && global_vars[i].size)
			global_vars[i].rel_benefit=
				(float)global_vars[i].use_count/
				(float)global_vars[i].size;
		else
			global_vars[i].rel_benefit=-1;
		
	}

	/* sort based on benefit */
	qsort(global_vars, globals_found, sizeof(global_vars[0]), 
		global_compare);



	/* make the assignments, we actually promote in fixentry */
	i=0;
	while(stache_used<=stache_size && i<globals_found)
	{
		if(global_vars[i].size+stache_used<stache_size &&
			global_vars[i].compiling)
		{
			global_vars[i].is_stached=1;
			stache_used+=global_vars[i].size;
			if(debug_opt("dump_stacheables"))
				fprintf(stderr, "Promoting %s[%d] to stache "
					"size=%d, total_size=%d, "
					"benefit=%5.2f, uses=%d\n",
					global_vars[i].name, i,
					global_vars[i].size,
					stache_used,
					global_vars[i].rel_benefit,
					global_vars[i].use_count);
		}
		i++;
	}	

	/* we've used up stache_used bytes of stache.. we can't ever reclaim
		it, we'll just forget we had it. */
	if(!sww)
		fprintf(stderr, "globals used %d bytes of stache\n", 
			stache_used);
	stache_size-=stache_used;
	return;
}

static int global_compare(const void* a, const void* b)
{
	struct global_var* a1=(struct global_var*)a;
	struct global_var* b1=(struct global_var*)b;

	if(a1->rel_benefit > b1->rel_benefit)
		return -1;
	if(a1->rel_benefit < b1->rel_benefit)
		return 1;
	return 0;
}


#endif  /* STACHE */

