
/***********************************************************************
 * 
 * Copyright (c) 2000 Baosheng Cai.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/undo.c,v $
 * $Revision: 1.7 $
 ***********************************************************************/

/* FILE: undo.c: functions to implement undo ability.
 undo_phases: can undo several phases
 undo_transes: can undo several transformations
 undo_changes: can undo several changes
 All of them will eventually call undo_last_change, which is the core function.
 */
#ifdef INTERACT
#define __UNDO_C__

#include <stdio.h>
#include <stdlib.h>
#include "vpo.h"
#include "vpo2.h"
#include "view.h"
#include "undo.h"
#include "misc.h"
#include "hand_trans.h"
#include "measure.h"
/* Machine dependent -- needs to be in the include path */
#include "instdef.h"

#define MEASURE_POSITION '%'
#define BEFORE_INIT_TRANS 1
#define AFTER_INIT_TRANS 2
#define UNDO_STOP   'u'       /* indicates when to stop undoing
                                 transformations                   */
FILE *init_file;
int from_where;

struct transform_list *transform_head, *transform_tail;

extern int continuous_measures;
extern int measure_phase, measure_init;

void check_cf(int inform);
int check_null_block();
int determine_total_ins(struct bblock *bb_ptr);
int getRTLCount(struct bblock *bb_ptr);
struct freq_counts*  get_ins_count(struct bblock *bblk, char *file_name, char
				   *func_name, int func_offset);
void check_control_flow(int status);

#define ERR_RET(MSG) {send_error(MSG); return;}

static int preds_modified;    /* indicates if a predecessor for a
                                 block has been modified           */
int undoing_transformation = FALSE;
                              /* currently undoing transformation  */

/*
 * clear_all_preds - clear all of the predecessors for all of the
 *                   basic blocks
 */
void clear_all_preds()
{
   struct bblock *cblk;
   extern struct bblock *ret_block;

   /*clear all the preds for each basic block*/
   for (cblk = top; cblk; cblk = cblk->down)
      clear_bb_preds(cblk);
   clear_bb_preds(ret_block);
}

/*
 * clear_bb_preds - clear all of the predecessors for a single basic block
 */
void clear_bb_preds(struct bblock *cblk)
{
   register struct blist *bptr, *dptr;

   /*free the space used by predecessor nodes*/
   bptr = cblk->preds;
   while (bptr) {
      dptr = bptr;
      bptr = bptr->next;
      FREETYPE(dptr, BLISTTYPE);
   }

   /*set to NULL*/
   cblk->preds = NULL;
}

/*
 * recalculate_all_preds - recalculate all of the predecessors for all
 *                         basic blocks
 */
void recalculate_all_preds()
{
   struct bblock *cblk;
   struct blist *bl;

   clear_all_preds();
   
   for (cblk = top; cblk; cblk = cblk->down) {
     
     /*add new predecessors*/
      if (cblk->left)
         add_predecessor(cblk, cblk->left);
      if (cblk->right)
         add_predecessor(cblk, cblk->right);
   }
}

#ifdef INTERACT
/*
 * get_num_transforms - return the number of transforms applied on this
 *                      function
 */
int get_num_transforms()
{
  int init_count, count, count_flag;
  struct transform_list *ptr;

  init_count = 0;
  count = 0;
  count_flag = FALSE;
  for(ptr=transform_head ; ptr ; ptr=ptr->next){

    /*count all transformations in the list*/ 
    if(ptr->mod_flag == BEGINTRANS){
      sscanf(ptr->string, "%d", &init_count);
      count_flag = FALSE;
    }
    else if(ptr->mod_flag!=LOCAL_ALLOC && ptr->mod_flag!=MAP_REGISTER)
      count_flag = TRUE;
      
    if(ptr->mod_flag==ENDTRANS && count_flag)
      count = init_count;
 }

  /*return the number of transforms*/
  return count;
}

/*
 * get_num_changes - return the number of changes in specified number of
 *                   transforms
 */
int get_num_changes(int transforms)
{
  int init_count, change_count, trans_count, flag;
  struct transform_list *ptr;
  
  init_count = 0;
  change_count = 0;
  trans_count = 0;
  for(ptr=transform_head ; ptr && trans_count<transforms ; ptr=ptr->next){

    /*count number of changes*/
    if(ptr->mod_flag!=BEGINTRANS && ptr->mod_flag!=ENDTRANS &&
       ptr->mod_flag!=BEGINPHASE && ptr->mod_flag!=ENDPHASE &&
       ptr->mod_flag!=LOCAL_ALLOC && ptr->mod_flag!=MAP_REGISTER){
      change_count++;
      flag = TRUE;
    }

    /*count number of transforms*/
    if(ptr->mod_flag == BEGINTRANS){
      sscanf(ptr->string, "%d", &init_count);
      /*flag will be set to TRUE when we encounter a countable change in this
	transform (some changes like LOCAL_ALLOC are not counted)*/
      flag = FALSE;
    }
      
    if(ptr->mod_flag == ENDTRANS && flag)
      trans_count = init_count;
  }

  /*return the number of changes*/
  return change_count;
}
 

/*
 * undo_transes - undo a specified number of transformations
 */
void undo_transes(int num)
{
   char line[MAXSTRING]; 
   int num_transforms, num_changes, phase, save_flag;
   unsigned long long s_measure, d_measure;
   long file_offset;
   struct transform_list *ptr;
   struct freq_counts *freq;
   extern char *fn;
   extern int session_over, viewing;
   extern int cex_offset, measure_flag, measure_init;
   extern struct transform_list *transform_head, *transform_tail;

   /*set flag to disable saving chnages to list and sending them to viewer*/
   measure_init = TRUE;
   save_flag = measure_flag;
   measure_flag = TRUE;

   /*clear the current state of this function*/
   do_fn_clean();
   file_offset = ftell(stdin);

   /*reset the input file offset*/
   if(fseek(stdin, cex_offset, SEEK_SET) != 0){
     fprintf(stderr, "not able to fseek\n");
     exit(0);
   }

   /*read the current function again*/
   read_next_function();
   fseek(stdin, file_offset, SEEK_SET);
   set_function_name(fn);
   do_fn_prereqs();

   /*if num >= 0: number of transforms to discard,
        num < 0 : discard all transforms
   */
   if(num >= 0){
     
     /*get the number of transforms applied to this function*/
     num_transforms = get_num_transforms();

     /*get the number of changes in the specified number of transforms*/
     num_changes = get_num_changes(num_transforms-num);
   }
   else
     num_changes = 0;

   /*apply these changes to the function*/
   apply_inittrans(NULL, num_changes, APPLY_FROM_LIST);
   
   /*check if the last message sent is a ENDPHASE message*/
   if(transform_tail && transform_tail->mod_flag!=ENDPHASE){
     fprintf(stderr, "Last message: %c\n", transform_tail->mod_flag);
     /*find the phase number of the last phase*/
     for(ptr=transform_head ; ptr ; ptr=ptr->next){
       if(ptr->mod_flag == BEGINPHASE)
	 sscanf(ptr->string, "%d", &phase);
     }
     
     /*add the end phase message to the end of the transform list*/
     sprintf(line, "%d\n", phase);
     add_to_transform_list(ENDPHASE, line);

#ifdef MEASURE
     /*since some changes in this transformation have been discarded it is
       necessary to take measurements again*/
     freq = NULL;
     if(continuous_measures){
       get_measure(&freq, -1);
       read_perf_num(&s_measure, &d_measure);
       if(viewing && !session_over)
	 send_measures(freq, s_measure, d_measure);
     }
#endif
     add_measures_transform_list(transform_tail, freq);
   }
     
   measure_flag = save_flag;
   measure_init = FALSE;
}

/*
 * undo_changes - undo a specified number of changes
 */
void undo_changes(int num)
{
  int count = 0;
   char line[MAXSTRING]; 
   int num_transforms, num_changes, phase, save_flag;
   unsigned long long s_measure, d_measure;
   long file_offset;
   struct transform_list *ptr;
   struct freq_counts *freq;
   extern char *fn;
   extern int session_over, viewing;
   extern int cex_offset, measure_flag, measure_init;
   extern struct transform_list *transform_head, *transform_tail;
   
   /*set flag to disable saving chnages to list and sending them to viewer*/
   measure_init = TRUE;
   save_flag = measure_flag;
   measure_flag = TRUE;

   /*clear the current state of this function. Note that since the transform
     list is now in a different arena it will not be freed at this time*/
   do_fn_clean();
   file_offset = ftell(stdin);

   /*reset the input file offset*/
   if(fseek(stdin, cex_offset, SEEK_SET) != 0){
     fprintf(stderr, "not able to fseek\n");
     exit(0);
   }

   /*read the current function again*/
   read_next_function();
   fseek(stdin, file_offset, SEEK_SET);
   set_function_name(fn);
   do_fn_prereqs();
   
   /*get the number of transforms applied to this function*/
   num_transforms = get_num_transforms();

   /*get the number of changes in the specified number of transforms*/
   num_changes = get_num_changes(num_transforms);

   /*apply these changes to the function*/
   apply_inittrans(NULL, num_changes-num, APPLY_FROM_LIST);
   
   /*check if the last message sent is a ENDPHASE message*/
   if(transform_tail && transform_tail->mod_flag!=ENDPHASE){
     fprintf(stderr, "Last message: %c\n", transform_tail->mod_flag);
     /*find the phase number of the last phase*/
     for(ptr=transform_head ; ptr ; ptr=ptr->next){
       if(ptr->mod_flag == BEGINPHASE)
	 sscanf(ptr->string, "%d", &phase);
     }
     
     /*add the end phase message to the end of the transform list*/
     sprintf(line, "%d\n", phase);
     add_to_transform_list(ENDPHASE, line);

#ifdef MEASURE
     /*since some changes in this transformation have been discarded it is
       necessary to take measurements again*/
     freq = NULL;
     if(continuous_measures){
       get_measure(&freq, -1);
       read_perf_num(&s_measure, &d_measure);
       if(viewing && !session_over)
	 send_measures(freq, s_measure, d_measure);
     }
#endif
     add_measures_transform_list(transform_tail, freq);
   }
     
   measure_flag = save_flag;
   measure_init = FALSE;
}


/*
 * undo_from_list - undo changes function called during the exhaustive scan
 *                  of the optimization search space
 */
void undo_from_list(int level)
{
  int i;
  extern unsigned int trans_done;
  extern int pro_apply_base, pro_is_cex_file;
  extern int pro_cex_opt[MAX_PRO_LEVELS];
  extern char **pro_cex_file;
  extern char *list_cex_files[MAX_PRO_LEVELS][MAX_CEX_LINES];
  
  /*undo all transformations*/
#ifdef WRITE_CEX
  pro_cex_file = (char **)list_cex_files[level-1];
#else
  pro_cex_file = (char **)list_cex_files[0];
#endif /*WRITE_CEX*/
  
  pro_is_cex_file = TRUE;
  undo_transes(-1);
  pro_is_cex_file = FALSE;
  
#ifdef WRITE_CEX      
  /*update trans_done*/
  for(i=0 ; i<level ; i++){
    trans_done |= (1 << pro_cex_opt[i]);
    /*if(!(trans_done & (1<<REG_ASSIGNMENT_TRANS)) &&
       requires_RA(pro_cex_opt[i]))
       trans_done |= (1 << REG_ASSIGNMENT_TRANS);*/
  }
#endif /*WRITE_CEX*/  
}
  
#endif /*INTERACT*/

/*
 * find_last - find the last pointer in the next list
 */
struct bblock *find_last()
{
   struct bblock *cblk;

   for (cblk = top; ; cblk = cblk->next) {
      if (!cblk->next)
         break;
   }
   return cblk;
}


/*
 * dump_left_right - dump the left and right pointers of the basic blocks
 *                   to a file
 */
void dump_left_right(char *filename)
{
   struct bblock *cblk;
   char buf[BUFSIZ];
   FILE *fout;

   fout = fopen(filename, "w");
   setbuf(fout, buf);

   for (cblk = top; cblk; cblk = cblk->down)
      fprintf(fout, "%d:%d:%d;", cblk, cblk->left, cblk->right);
   fclose(fout);
}

/*
 * compare_string - compare if string1 is identical with string2.
 *                  The same functionality as strcmp, just handle
 *                  null pointers.
 */
int compare_string(char *string1, char *string2)
{
   if (string1 && string2)
      return (strcmp(string1, string2));
   else if (string1 != string2)
      return -1;
   else
      return 0;
}

/*
 * calc_status - recalculate the status of the basic blocks
 *               called when there is MOD_DOWN, MOD_LEFT, and MOD_RIGHT
 */
void calc_status()
{
   struct bblock *cblk;
   struct list *last_rtl, *prev_last_rtl;
   extern struct bblock *ret_block;
   
   for (cblk = top; cblk; cblk = cblk->down) {
      last_rtl = prev_last_rtl = NULL;

      /* find last rtl in the block */
      for (last_rtl = cblk->rtlend; last_rtl; last_rtl = last_rtl->prev)
         if (cnts(last_rtl))
            break;

      /* find previous rtl of the last one in the block */
      if (last_rtl)
         for (prev_last_rtl = last_rtl->prev; prev_last_rtl;
              prev_last_rtl = prev_last_rtl->prev)
            if (cnts(prev_last_rtl))
               break;

      /* if the cblk ends with ucjmp */
      if ((last_rtl && is_ucjmp(last_rtl->val)) ||
          (prev_last_rtl && is_ucjmp(prev_last_rtl->val)) ||
          (last_rtl && is_pc(last_rtl->val) && cblk->left == ret_block)) {
         cblk -> status |= JUMP;
         if (cblk->down)
            cblk->down->status &= ~FALLTHRU;
      }
      else {
         cblk->status &= ~JUMP;
         if (cblk->down)
            cblk->down->status |= FALLTHRU;
      }
   }
}


/* functions copied from an older verion of controlflow.c */

#define CHECKCF (1 << 10)    /* Used for checking control flow. */

/*
 * check_cf - check the control flow
 *            (only works before fixentry() and fill_slots())
 */
void check_cf(int inform)
{
    extern struct bblock *ret_block;
    struct bblock *cblk, *nblk;
    struct blist *bptr;
    struct list *jump, *ptr;
    int i, j;
    char msg[MAXSTRING];


    /* go through every block */
    (void) number_basic_blocks();

    /* check if everything in down list is in next list */
    for (cblk = top; cblk; cblk = cblk->down) {
       for (nblk = top; nblk; nblk = nblk->next)
          if (cblk == nblk)
             break;
       if (!nblk){
          sprintf(msg, "Block in down list can't be found in next list:%d\n",
                  cblk->bnum);
          blkerrmsg(cblk, msg, TRUE);
       }      
    }

    /* check if everything in next list is in down list */
    for (nblk = top; nblk; nblk = nblk->next) {
       for (cblk = top; cblk; cblk = cblk->down)
          if (nblk == cblk)
             break;
       if (!cblk && nblk != ret_block) {
          sprintf(msg, "Block in next list can't be found in down list:%d\n",
                  nblk->bnum);
          blkerrmsg(nblk, msg, TRUE);
       }      
    }    
    
    for (cblk = top; cblk; cblk = cblk->down) {

       /* check that up matches down */
       if (cblk->up && cblk->up->down != cblk) {
          sprintf(msg, "up does not match down at block %d\n",
                  cblk->up->bnum);
          blkerrmsg(cblk, msg, TRUE);
       }
       if (cblk->down && cblk->down->up != cblk) {
          sprintf(msg, "down does not match up at block %d\n",
                  cblk->down->bnum);
          blkerrmsg(cblk, msg, TRUE);
       }

       /* check if all predecessors match */
       for (bptr = cblk->preds; bptr; bptr = bptr->next)
          if (bptr->ptr->left != cblk && bptr->ptr->right != cblk) {
             fprintf(stderr, "pred %d does not have it as succ\n",
                     bptr->ptr->bnum);
             blkerrmsg(cblk, msg, TRUE);
          }

       /* find last rtl in the block */
       for (jump = cblk->rtlend; jump; jump = jump->prev)
          if (cnts(jump))
             break;

       /* if a left successor */
       if (cblk->left) {

          /* check if target has this block as a predecessor */
          if (!inblist(cblk->left->preds, cblk))
             blkerrmsg(cblk, "left pred not found", TRUE);

          /* if found such an rtl */
          if (jump) {

             /* if the rtl is an unconditional jump */
             if (is_ucjmp(jump->val)) {
                if (cblk->down && cblk->down->status & FALLTHRU &&
                    cblk->left != cblk->down) {
                   cblk->down->status &= ~FALLTHRU;
                   blkerrmsg(cblk->down, "FT after ucjmp", inform);
                }

                if (!(cblk->status & JUMP)) {
                   cblk->status |= JUMP;
                   blkerrmsg(cblk, "No JUMP at ucjmp", inform);
                }

                if (!fmatchlabel(jump->val, cblk->left->label))
                   blkerrmsg(cblk, "wrong label at ucjmp", TRUE);
             }


             /* if the rtl is a conditional branch */
             else if (is_rjmp(jump->val)) {
                if (cblk->status & JUMP) {
                   cblk->status &= ~JUMP;
                   blkerrmsg(cblk, "JUMP at rjmp", inform);
                }

                if (!(cblk->down->status & FALLTHRU)) {
                   cblk->down->status |= FALLTHRU;
                   blkerrmsg(cblk->down, "no FT at rjmp", inform);
                }

                if (!fmatchlabel(jump->val, cblk->right->label))
                   blkerrmsg(cblk, "wrong label at rjmp", TRUE);
             }

             /* if the rtl is not a transfer of control */
             else if (!is_pc(jump->val)) {
                if (cblk->status & JUMP) {
                   cblk->status &= ~JUMP;
                   blkerrmsg(cblk, "JUMP at ft", inform);
                }

                if (cblk->down && !(cblk->down->status & FALLTHRU)) {
                   cblk->down->status |= FALLTHRU;
                   blkerrmsg(cblk->down, "no FT at ft", inform);
                }
             }

             /* else it must be a return */
             else {
                if (!(cblk->status & JUMP)) {
                   cblk->status |= JUMP;
                   blkerrmsg(cblk, "No JUMP at return", inform);
                }
             }
          }

          /* else if there is no rtl */
          else {
             if (cblk->status & JUMP) {
                cblk->status &= ~JUMP;
                blkerrmsg(cblk, "JUMP at ft", inform);
             }

             if (cblk->down && !(cblk->down->status & FALLTHRU)) {
                cblk->down->status |= FALLTHRU;
                blkerrmsg(cblk->down, "no FT at ft", inform);
             }
          }
       }

       /* if not a left successor and rtl was found in the block */
       else if (jump) {
          if (is_ucjmp(jump->val))
             blkerrmsg(cblk, "ucjmp without left pointer", TRUE);
          else if (is_rjmp(jump->val))
             blkerrmsg(cblk, "rjmp without left pointer", TRUE);
       }

       /* if a right successor */
       if (cblk->right) {

          /* check if target has this block as a predecessor */
          if (!inblist(cblk->right->preds, cblk))
             blkerrmsg(cblk, "right pred not found", TRUE);

          /* if an rtl was found in this block */
          if (jump)
             if (is_rjmp(jump->val)) {
                if (!fmatchlabel(jump->val, cblk->right->label))
                   blkerrmsg(cblk, "wrong label at rjmp", TRUE);
             }
          else if (!is_indjmp(jump->val))
             blkerrmsg(cblk, "no cond jump at block", TRUE);
       }

       /* if no right successor and a conditional branch */
       else if (jump && is_rjmp(jump->val))
          blkerrmsg(cblk, "rjmp without right pointer", TRUE);

       /* count the number of non-fallthru predecessors */
       i = 0;
       for (bptr = cblk->preds; bptr; bptr = bptr->next)
          if (!is_ft(bptr->ptr, cblk))
             i++;
    }

    /*
     * check if each rtl has its fields set correctly
     */
    for (cblk = top; cblk; cblk = cblk->down) {
       if (cblk->rtls && cblk->rtls->prev)
          blkerrmsg(cblk, "should be no prev before first rtl in blk", TRUE);
       for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
          if (ptr->daddy != cblk)
             blkerrmsg(cblk, "ptr in cblk does not have correct daddy", TRUE);
          if (ptr->next && ptr->next->prev != ptr){
	    fprintf(stderr, "RTL id: %d Block: %d\n", ptr->rtl_id, ptr->daddy);
	    blkerrmsg(cblk, "next does not match prev in rtl", TRUE);
	  }
       }
       if (cblk->rtlend && cblk->rtlend->next)
          blkerrmsg(cblk, "should be no next after last rtl in blk", TRUE);
    }

    /*
     *   I add one more check to test whether left & right & predecessor links
     *   exist in the given controlflow
     */
    for (cblk = top; cblk; cblk = cblk->down)
       cblk->status |= CHECKCF;

    for (cblk = top; cblk; cblk = cblk->down) {
       if (cblk->left)
          if (!cblk->left->label || strncmp(cblk->left->label, "RT", 2))
             if (!(cblk->left->status & CHECKCF))
                blkerrmsg(cblk, "left link does not exists in a controlflow",
                          TRUE);

       if (cblk->right)
          if (!cblk->right->label || strncmp(cblk->right->label, "RT", 2))
             if (!(cblk->right->status & CHECKCF))
                blkerrmsg(cblk, "right link does not exists in a controlflow",
                          TRUE);

       for (bptr = cblk->preds; bptr; bptr = bptr->next)
          if (!(bptr->ptr->status & CHECKCF))
             blkerrmsg(bptr->ptr, "pred does not exists in a a controlflow\n",
                       TRUE);

    }

    for (cblk=top; cblk; cblk=cblk->down)
          cblk->status &= ~CHECKCF;

    fflush(stderr);
}

/*
 * is_ft - checks if the predecessor falls into the current block
 */
int is_ft(pblk, cblk)
struct bblock * pblk, *cblk;
{
    /* check if predecessor precedes current block and falls into it */
    if (pblk == cblk->up && (cblk->status & FALLTHRU))
         return(TRUE);
    return(FALSE);
}

/*
 * blkerrmsg - prints an error message and identifies the basic block
 */
void blkerrmsg(struct bblock *cblk, char *msg, int inform)
{
    if (!inform)
         return;
    fprintf(stderr, "%s at %d", msg, cblk->bnum);
    if (cblk->label != 0)
       fprintf(stderr, " (%s)", cblk->label);
    else if (cblk->up && cblk->up->label)
       fprintf(stderr, " (below %s)", cblk->up->label);
    else if (cblk->down && cblk->down->label)
       fprintf(stderr, " (above %s)", cblk->down->label);
    fprintf(stderr, "\n");
    /*dump_function(1);
      exit(1);*/
    for(; ; )
      sleep(5);
}

/*
 * is_ucjmp - is RTL an unconditional jump
 */
int is_ucjmp(s)
register char *s;
{
   return(*s == 'P' && *(s + 1) == 'C' && *(s + 3) == 'L');
}

/*
 * is_indjmp - is RTL an indirect jump
 */
int is_indjmp(s)
register char *s;
{
   return (*s == 'P' && *(s + 1) == 'C' && *(s + 3) != 'L' &&
           *(s + 3) != 'R' && *(s + 3) != 'I' && *(s + 3) != 'F');
}

/*
 * fmatchlabel - check if label is found in an RTL
 */
int fmatchlabel(char *s, char *l)
{
   extern char *epos;

   if (fmatch(s,l))
      return *epos == ';' || *epos == '\0';
   return FALSE;
}

/*
 * inblist - determines if a block is in a blist
 */
int inblist(struct blist *first, struct bblock *cblk)
{
   struct blist *bptr;

   for (bptr = first; bptr; bptr = bptr->next) {
      if (bptr->ptr == cblk)
         return(TRUE);
   }
   return(FALSE);
}

/*
 * find_bb_in_next - find a basic block in the next list
 */
int find_bb_in_next(char *label)
{
   struct bblock *cblk;

   for (cblk = top; cblk; cblk = cblk->next){
      if (!compare_string(cblk ->label, label)){
         fprintf(stderr, "find the label in bb: %d\n", cblk);
         return 1;
      }
   }
   return 0;
}


/*
 * print_phases - list all the phases in the undo list
 */
void print_phases()
{
   int index = 1;
   struct transform_list *tptr;
   extern struct transform_list *transform_head;

   for(tptr=transform_head ; tptr ; tptr=tptr->next){
     if(tptr->mod_flag == BEGINPHASE){
       index = atoi(tptr->string);
       fprintf(stderr, "Phase begin %d: %s\n", index, phase_names[index]);
     }
   }
}

#ifdef DEBUG
/*functions defined just for debugging
 will be deleted afterwards*/


short regtovec_call(char *rtl)
{
   return vecmap[USP(rtl)];
}

unsigned short USP_call(char *s)
{
   return (*((unsigned short *) (s)) & (MAPREGS - 1));
}

#endif   /*endif DEBUG*/

/*
 * dump_trans - dump out the transformations to a file
 */
void dump_trans(struct freq_counts *initial_cnt)
{
  char str[MAXSTRING], filename[MAXSTRING];
  char transoutbuf[BUFSIZ];
  int count;
  struct freq_counts *freq_ptr;
  struct transform_list *tptr;
  extern char *fn;
  extern char pathname[], basename[];
  extern int afterinit_block_id, afterinit_rtl_id;
  FILE *f;

  /*open the file to dump transforms*/
  if (!fn)
    return;
  sprintf(filename, "%s%s.transout", pathname, basename);
  f = fopen(filename, "a");
  setbuf(f, transoutbuf);
  
  /* dump out the function name */
  fprintf(f, "f%s %d %d\n", fn, afterinit_block_id, afterinit_rtl_id);
  
  /*dump the initial frequency measures*/
  for(freq_ptr=initial_cnt,count=0 ; freq_ptr ; freq_ptr=freq_ptr->next, count++);
  fprintf(f, "F%d", count);
  
  for(freq_ptr=initial_cnt ; freq_ptr ; freq_ptr=freq_ptr->next)
    fprintf(f, " %d", freq_ptr->count);
  
  fprintf(f, "\n");

  /*dump out the transformations in the undo list*/
  for(tptr=transform_head ; tptr ; tptr=tptr->next){
    fprintf(f, "%c", tptr->mod_flag);
    fprintf(f, "%s", tptr->string);
  }
  
  /*--- for measurement phase ---*/
  
  /*if measures are requested before all initial transforms are applied
    then we mark the point upto which transformations have been applied and
    then copy over the remaining lines from the .trans file*/
  if(measure_phase && from_where==BEFORE_INIT_TRANS)
    {
      /*mark position in trans file*/
      fprintf(f, "%%\n");
      
      /*copy over the remaining lines from the trans file*/
      fgets(str, MAXSTRING-1, init_file);
      while((!feof(init_file)) && (strncmp(str, "f", 1)!=0))
	{
	  fprintf(f, "%s", str);
	  fgets(str, MAXSTRING-1, init_file);
	}
      fclose(init_file);
      from_where = AFTER_INIT_TRANS;
    }
  /*--- end measuremment phase changes ---*/
  
  fclose(f);
}

/*
 * nextstr - skip over the current string
 */
char *nextstr(char *s)
{
   char *p;

   for (p = s; *p && !isspace(*p); p++)
      ;
   for (; isspace(*p); p++)
      ;
   return p;
}

/*
 * remv_trailing_spaces - skip trailing spaces
 */
void remv_trailing_spaces(char *s)
{
  char *p;

  p = (char *)s;
  while(!isspace(*p))
    p++;
  *p = '\0';
}

/*
 * swaprtls - switch the location of two rtls in the same block
 */
void swaprtls(struct list *ptr1, struct list *ptr2)
{
   struct list *tptr;
   struct bblock *cblk;

   if (ptr1 == ptr2)
      return;
   if (ptr1->daddy != ptr2->daddy) {
      fprintf(stderr, "swaprtls (vpo) - rtls not in the same block\n");
      exit(1);
   }
   cblk = ptr1->daddy;
   tptr = ptr1->prev;
   if(ptr1->prev)
     ptr1->prev->next = ptr2;
   ptr1->prev = ptr2->prev;
   if(ptr2->prev)
     ptr2->prev->next = ptr1;
   ptr2->prev = tptr;

   tptr = ptr1->next;
   if(ptr1->next)
     ptr1->next->prev = ptr2;
   ptr1->next = ptr2->next;
   if(ptr2->next)
     ptr2->next->prev = ptr1;
   ptr2->next = tptr;

   if (cblk->rtls == ptr1)
      cblk->rtls = ptr2;
   else if (cblk->rtls == ptr2)
      cblk->rtls = ptr1;
   if (cblk->rtlend == ptr1)
      cblk->rtlend = ptr2;
   else if (cblk->rtlend == ptr2)
      cblk->rtlend = ptr1;
}


/*
 * dump_block_counts - passes basic block frequency counts to the viewer 
 *
 */

/* 
   This function reads the measurement file twice for each function. The first
   time, the offset of the function in the file is computed. The second time
   the block frequencies are dumped. This is necessary because the name of
   the function located after the block frequency counts
*/

void dump_block_counts(struct bblock *bblk, char *file_name, char *func_name, int func_offset) {
  int i;
  int junk;
  FILE *fp;
  int func_num = 0, found;
  struct list *ptr;

  if ((fp = fopen(file_name, "r")) == NULL)
    return;

  if (!func_name)
    write_record_id(BEGINFREQ);

  /* Skip over the not taken counts */

  for (i = 0; i < NUMJREL; i++) 
    fscanf(fp, "%d ", &junk);

  /* For each function */

  fscanf(fp, "%d ", &junk);
  while (!feof(fp)) {
    int equiv_classes;
    char name[100];
    char skip[100];


    fscanf(fp, "%d ", &equiv_classes);
    for (i = 0; i < equiv_classes; i++) {
      int freq_count;

      fscanf(fp, "%d ", &freq_count);
      if (!func_name && func_num == func_offset && bblk) {

	found = FALSE;
	/*check if there is a frequency count associated with this block */
	while(!found){
	  if(bblk){
	    for(ptr=bblk->rtls ; ptr ; ptr=ptr->next){
	      if(cnts(ptr))
		break;
	    }
	  }

	  /* if frequency measure exists */
	  if(ptr != NULL){
	    write_bb_pointer(bblk);
	    write_num(freq_count);
	    bblk = bblk->down;
	    found = TRUE;
	  }
	  else{
	    write_bb_pointer(bblk);
	    write_num(0);
	    bblk = bblk->down;
	  }
	}
      }
    }
    fscanf(fp, "%s ", name);
    if (func_name && !strcmp(func_name, name)) {
      fclose(fp);
      dump_block_counts(bblk, file_name, NULL, func_num);
      return;
    }

    do {
      fscanf(fp, "%s ", skip);
    } while (strcmp(skip, "-2"));

    do {
      fscanf(fp, "%s ", skip);
    } while (strcmp(skip, "-3"));

    do {
      fscanf(fp, "%s ", skip);
    } while (strcmp(skip, "-4"));

    do {
      fscanf(fp, "%s ", skip);
    } while (!feof(fp) && strcmp(skip, "-1"));
    func_num++;
  }
  fclose(fp);
  write_record_id(ENDFREQ);
}


/*get_ins_count - determine the instruction counts in each basic block*/

struct freq_counts *get_ins_count(struct bblock *bblk, char *file_name, char *func_name, int func_offset) {
  int i;
  int junk;
  FILE *fp;
  int func_num = 0, found;
  struct freq_counts *freq_head, *freq_ptr, *new_ptr;
  struct list *ptr;

  if ((fp = fopen(file_name, "r")) == NULL)
    return NULL;

  freq_head = NULL;
  /* Skip over the not taken counts */
  for (i = 0; i < NUMJREL; i++) 
    fscanf(fp, "%d ", &junk);

  /* For each function */
  fscanf(fp, "%d ", &junk);
  while (!feof(fp)) {
    int equiv_classes;
    char name[100];
    char skip[100];

    fscanf(fp, "%d ", &equiv_classes);
    for (i = 0; i < equiv_classes; i++) {
      int freq_count;

      fscanf(fp, "%d ", &freq_count);
      if (!func_name && func_num == func_offset && bblk) {

	found = FALSE;

	/*check if there is a frequency count associated with this block */
	while(!found && bblk){
	  if(bblk){
	    for(ptr=bblk->rtls ; ptr ; ptr=ptr->next){
	      if(cnts(ptr))
		break;
	    }
	  }

	  /* allocate new record for this block frequency*/
	  new_ptr = (struct freq_counts*)CALLOC(sizeof(struct freq_counts));
	  new_ptr->count = 0;
	  new_ptr->next = NULL;
	  if(freq_head == NULL)
	    freq_head = new_ptr;
	  else
	    freq_ptr->next = new_ptr;
	  freq_ptr = new_ptr;

	  /*if a frequency measure exists for this block */
	  if(ptr != NULL){
	    new_ptr->count = freq_count;
	    bblk->frequency = freq_count;
	    found = TRUE;
	  }
	  else /* no frequency measure for this block*/
	    bblk->frequency = 0;
	  
	  bblk = bblk->down;
	}
      }
    }
    if(!func_name && func_num == func_offset){
      while(bblk){
	fprintf(stderr, "no frequency measure, block: %d\n", bblk->block_id);
	bblk->frequency = 0;
	bblk = bblk->down;
      }
    }

    fscanf(fp, "%s ", name);
    if (func_name && !strcmp(func_name, name)) {
      fclose(fp);
      freq_head = (struct freq_counts*) get_ins_count(bblk, file_name, NULL, func_num);
      return freq_head;
    }

    do {
      fscanf(fp, "%s ", skip);
    } while (strcmp(skip, "-2"));

    do {
      fscanf(fp, "%s ", skip);
    } while (strcmp(skip, "-3"));

    do {
      fscanf(fp, "%s ", skip);
    } while (strcmp(skip, "-4"));

    do {
      fscanf(fp, "%s ", skip);
    } while (!feof(fp) && strcmp(skip, "-1"));
    func_num++;
  }
  fclose(fp);
  return freq_head;
}


/* determine_total_ins - calculate the number of instructions executed after
   this stage */
int determine_total_ins(struct bblock *bb_ptr)
{
  int sum;
  struct bblock *ptr;

  sum = 0;
  for(ptr=bb_ptr ; ptr ; ptr=ptr->down)
    sum += (getRTLCount(ptr) * ptr->frequency);

  return sum;
}

/*
 * count_RTL - count the total number of RTLs present in this function
 */
int count_RTL()
{
  struct bblock *cblk;
  int nrtl;
  extern struct bblock *top;

  nrtl = 0;
  for(cblk=top ; cblk ; cblk=cblk->down)
    nrtl += getRTLCount(cblk);

  return nrtl;
}

/*getRTLCount - Number of RTLs in this basic block*/
int getRTLCount(struct bblock *bb_ptr)
{
  struct list *RTL;
  int count;

  count = 0;
  for(RTL=bb_ptr->rtls ; RTL ; RTL=RTL->next)
    {
      if(cnts(RTL))
	count++;
    }

  return count;
}

#ifdef INTERACT
/*
 * apply_inittrans - apply the initial transformations that were stored
 *                   in a file
 */
int apply_inittrans(struct freq_counts **init_cnt, int num_changes, int source)
{
   unsigned long trans_checksum;
   char *a;
   int treg, tnouse, tatype, treps;
   int found, phase, blkid, blkid2, rtlid, rtlid2;
   unsigned int type, kind;
   unsigned long long s_measure, d_measure;
   rlist deads;
   struct list *ptr, *ptr2, *tptr, *rtlptr;
   struct bblock *cblk, *cblk2;
   struct freq_counts *phase_cnt;
   struct locuse *loc_ptr;
   struct transform_list *trans_ptr, *prev_ptr;
   char line[MAXSTRING], name[MAXSTRING], val[MAXSTRING], reg[REGSTRING];
   char val2[MAXSTRING];
   char regs[MAXSTRING], filename[MAXSTRING];
   char regname[MAXSTRING], tregin[MAXSTRING];
   extern char pathname[], basename[];
   extern char *fn;
   extern unsigned long cex_checksum;
   extern int block_id, rtl_id;
   extern unsigned int trans_done;
   extern int session_over;
   extern int phase_sanity_check, transform_sanity_check;
   extern int measure_phase, measure_init, tot_commands, measure_flag;
   extern int isused[MAXREGS], transnum;
   extern struct locuse *locs;
   extern struct bblock *top, *last;
   extern struct command commands[MAXCMDS];
   extern struct transform_list *transform_head;
   int is_proceed, initial_counts=-1, changes, cf_flag;
   int regm, index;
   extern int regmap[];
   extern char transbuf[];

   FILE *pkfile;
   int count=0;
   void print_block_rtlids(FILE *pkfile);
   cf_flag = FALSE;

   /*do this only if we are reading from file*/
   if(source == APPLY_FROM_FILE){

     /* open the trans file */
     sprintf(filename, "%s%s.trans", pathname, basename);
     init_file = fopen(filename, "r");
     if (!init_file) {
       if (!session_over && !measure_phase && !measure_init && !continuous_measures)
         write_record_id(ENDSEQ);
       return initial_counts;
     }
     setbuf(init_file, transbuf);

     /* verify that the trans file key matches the cex file key */
     fscanf(init_file, "k%u\n", &trans_checksum);
     
     if (cex_checksum != trans_checksum) {
       if (!session_over && !measure_phase && !measure_init && !continuous_measures)
         write_record_id(ENDSEQ);
       return initial_counts;
     }
     
     /* find the appropriate function in the trans file */
     found = FALSE;
     while (fgets(line, MAXSTRING, init_file)) {
       if (sscanf(line, "f%s", name) == 1 && strcmp(name, fn) == 0) {
         found = TRUE;
         break;
       }
     }
     if (!found) {
       fprintf(stderr,
	       "apply_inittrans (vpo) - could not find function %s\n", fn);
       exit(1);
     }
     
     /* get the frequency counts */
     if( (fscanf(init_file, "F%d", &initial_counts)) == 1)
       initial_counts = send_counts(init_file, initial_counts, init_cnt, 1);
     else
       initial_counts = send_counts(init_file, -1, init_cnt, 1);
     
     /* if no initial counts */
#ifdef MEASURE   
     if(continuous_measures && initial_counts == -1 &&
	(!(measure_init||measure_phase)))
       {
	 from_where = BEFORE_INIT_TRANS;
	 get_measure(init_cnt, -1);
	 read_perf_num(&s_measure, &d_measure);
	 if(viewing && !session_over)
	   send_measures(*init_cnt, s_measure, d_measure);
       }
#endif   
   }
   
   /* apply the transformations in the function */
   changes = 0;
   trans_ptr = NULL;
   prev_ptr = NULL;
   is_proceed = get_change(line, source, init_file, &trans_ptr);
   while (is_proceed)
     {
       /*break out if number of phases applied exceeds that specified*/
       if(num_changes!=-1 && changes>=num_changes && line[0]!=ENDPHASE &&
	  line[0]!=ENDTRANS && line[0]!=LOCAL_ALLOC && line[0]!=MAP_REGISTER)
	 break;

       switch (line[0])
	 {
	 case BEGINPHASE:
	   sscanf(&line[1], "%d", &phase);
	   trans_done |= (1 << phase);
	   begin_phase(phase);

	   /*initialization for register assignment*/
	   if(phase == REG_ASSIGNMENT_TRANS)
	     {
	       /*Set all basic blocks to UNDONE and mark all registers in use lines */
	       for (cblk = top; cblk; cblk = cblk->down) {
		 STATUS(cblk, UNDONE);
		 for (rtlptr = cblk->rtls; rtlptr; rtlptr = rtlptr->next)
		   if (rtlptr->type & USELINE)
		     isused[rtlptr->cost] = TRUE;
	       }
	     }
	   else if(phase == FILL_DELAY_SLOTS_TRANS){
	     /*need to do the check control flow before delay slots are
	       filled in */
	     check_control_flow(TRUE);
	     cf_flag = TRUE;
	   }

	   /*enter the command in the command array*/
	   if(!(measure_init || measure_phase))
 	     commands[tot_commands++].oper = phase;
	   break;
	   
	 case ENDPHASE:
	   sscanf(&line[1], "%d", &phase);
	   end_phase(phase);
	   
	   /* read the freqeuncy counts */
	   if(source == APPLY_FROM_FILE){
	     phase_cnt = NULL;
	     if( (fscanf(init_file, "F%d", &count)) == 1)
	       count = send_counts(init_file, count, &phase_cnt, 1);
	     else
	       count = send_counts(init_file, 0, &phase_cnt, 1);
#ifdef MEASURE	   
	     if(continuous_measures && count==-1 &&
		(!(measure_init||measure_phase)))
	       {
		 from_where = BEFORE_INIT_TRANS;
		 get_measure(&phase_cnt, -1);
		 read_perf_num(&s_measure, &d_measure);
		 if(viewing && !session_over)
		   send_measures(phase_cnt, s_measure, d_measure);
	       }
#endif
	     add_measures_transform_list(transform_tail, phase_cnt);
	   }
	   break;
	   
	 case BEGINTRANS:
	   if (NECESSARY_TRANS(1 << phase))
	     begin_trans(VET_REQUIRED_TRANS);
	   else
	     begin_trans(VET_OPT_TRANS);
	   sscanf(&line[1], "%d", &transnum);
	   break;
	   
	 case ENDTRANS:
	   end_trans();
	   break;
	   
	 case NEWBB:
	   sscanf(&line[1], "%d", &blkid);
	   block_id = blkid - 1;
	   new_basic_block((char *) NULL);
	   break;
	   
	 case UPDATEBB:
	   a = &line[1];
	   sscanf(a, "%d", &blkid);
	   cblk = find_block_id(blkid);
	   a = nextstr(a);
	   sscanf(a, "%d", &rtlid);
	   ptr2 = cblk->rtls;
	   while (rtlid != -1) {
	     ptr = find_rtl_id(rtlid, blkid);
	     swaprtls(ptr2, ptr);
	     a = nextstr(a);
	     sscanf(a, "%d", &rtlid);
	     ptr2 = ptr->next;
	   }
	   update_bb_rtls(cblk);
	   break;
	   
	 case LABELBB:
	   sscanf(&line[1], "%d %s", &blkid, name);
	   cblk = find_block_id(blkid);
	   label_bb(cblk, cblk->label, name);
	   if (cblk->label)
	     FREE(cblk->label);
	   cblk->label = copy_string(name);
	   break;
	   
         case MOD_LEFT:
	   sscanf(&line[1], "%d %d", &blkid, &blkid2);
	   cblk = find_block_id(blkid);
	   cblk2 = find_block_id(blkid2);
	   cblk->left = cblk2;
	   mod_bb_ptrs(MOD_LEFT, cblk, cblk2);
	   break;

         case MOD_RIGHT:
            sscanf(&line[1], "%d %d", &blkid, &blkid2);
            cblk = find_block_id(blkid);
            cblk2 = find_block_id(blkid2);
            cblk->right = cblk2;
            mod_bb_ptrs(MOD_RIGHT, cblk, cblk2);
            break;

         case MOD_DOWN:
            sscanf(&line[1], "%d %d", &blkid, &blkid2);
            cblk = find_block_id(blkid);
            cblk2 = find_block_id(blkid2);
            cblk->down = cblk2;
	    if(cblk2)
	      cblk2->up = cblk;
            mod_bb_ptrs(MOD_DOWN, cblk, cblk2);
            break;
            
         case DELRTL:
            sscanf(&line[1], "%d %d", &blkid, &rtlid);
            ptr = find_rtl_id(rtlid, blkid);
            remv_rtl(ptr);
            break;

         case DELBB:
            sscanf(&line[1], "%d", &blkid);
            cblk = find_block_id(blkid);
            for (cblk2 = top; cblk2->next != cblk; cblk2 = cblk2->next)
               ;
            cblk2->next = cblk->next;
	    if(!cblk->next)
	      last = cblk2;
            free_basic_block(cblk);
            break;

         case NEWRTL:
            line[strlen(line)-1] = '\0';
            sscanf(&line[1], "%d %d %d %u ", &blkid, &rtlid, &rtlid2, &type);
            cblk = find_block_id(blkid);
            ptr = find_rtl_id(rtlid, blkid);
            deads = rinit();
            a = nextstr(&line[1]);
            a = nextstr(a);
            a = nextstr(a);
            a = nextstr(a);
	    val[0] = '\0';
            if (type & RTLINE) {
               sscanf(a, "%s %s", val, regs);
               if (strcmp(regs, "none") != 0) {
                  a = regs;
                  reg[2] = '\0';
                  while (*a != '\0') {
                     reg[0] = *a;
                     reg[1] = *(a+1);
                     rinsert(reg, &deads);
                     a += REGLOCLENGTH;
                  }
               }
            }
            else
               strcpy(val, a);
	    remv_trailing_spaces(val);
	    if(strcmp(val, "none") == 0)
	      val[0] = '\0';
	    /*In some cases the current rtlid, which is the least rtl id
	      available may not match with the id to be given to this
	      RTL. This may be because some RTLs, esp. during Code Motion are
	      deleted immediately after creation, during which although the
	      rtlid is incremented, no messages are sent*/
	    if(rtl_id > rtlid2-1){
	      fprintf(stderr, "Warning: New RTL may be given same id as some previous RTL\n");
	    }
	    rtl_id = rtlid2-1;
            ptr = add_rtl_with_type(cblk, ptr, val, deads, type);
            break;
            
         case MOVERTLS:
            sscanf(&line[1], "%d %d %d %d", &blkid, &blkid2, &rtlid, &rtlid2);
            cblk = find_block_id(blkid);
            cblk2 = find_block_id(blkid2);
            ptr = find_rtl_id(rtlid, blkid);
            ptr2 = find_rtl_id(rtlid2, blkid);

	    if(cblk->rtls){
	      if(cblk->rtls->prev = cblk2->rtlend)
		cblk2->rtlend->next = cblk->rtls;
	      else
		cblk2->rtls = cblk->rtls;
	      cblk2->rtlend = cblk->rtlend;

	      for(tptr=cblk->rtls ; tptr ; tptr=tptr->next)
		tptr->daddy = cblk2;
	    }
	      /*cblk2->rtlend->next = ptr;
		ptr->prev = cblk2->rtlend;
		cblk2->rtlend = ptr2;
		for (tptr = ptr; tptr; tptr = tptr->next)
		tptr->daddy = cblk2;
		cblk->rtls = cblk->rtlend = (struct list *) NULL;*/

            /* send the messages to the viewer */
            move_rtls(cblk, cblk2, ptr, ptr2);
	    cblk->rtls = cblk->rtlend = (struct list *) NULL;
            break;

         case MOVERTL:
            sscanf(&line[1], "%d %d %d %d", &blkid, &rtlid, &blkid2, &rtlid2);
            cblk = find_block_id(blkid);
            ptr = find_rtl_id(rtlid, blkid);
            cblk2 = find_block_id(blkid2);
            ptr2 = find_rtl_id(rtlid2, blkid2);
      
            /* detach the RTL from the old block */
            if (ptr->next)
	      ptr->next->prev = ptr->prev;
            else
	      cblk->rtlend = ptr->prev;
	    if (ptr->prev)
	      ptr->prev->next = ptr->next;
	    else
	      cblk->rtls = ptr->next;
	    
            /* hook up the RTL to the new block */
            /*ptr->next = ptr2;*/
            if (ptr2) {
	      ptr->next = ptr2;
	      ptr->prev = ptr2->prev;
	      if(ptr2->prev)
		ptr2->prev->next = ptr;
	      else
		cblk2->rtls = ptr;
	      ptr2->prev = ptr;
            }
            else {
	      ptr->prev = cblk2->rtlend;
	      ptr->next = (struct list*) NULL;
	      if(cblk2->rtlend)
		cblk2->rtlend->next = ptr;
	      cblk2->rtlend = ptr;
	      if(!cblk2->rtls)
		cblk2->rtls = ptr;
            }
            ptr->daddy = cblk2;

            /* send the message to the viewer */
            move_rtl(ptr, cblk, ptr2, cblk2);
            break;

         case MOD_RTL:
            sscanf(&line[1], "%d %d %u", &blkid, &rtlid, &type);
            ptr = find_rtl_id(rtlid, blkid);
            a = nextstr(&line[1]);
            a = nextstr(a);
            a = nextstr(a);
            if (type & RTLINE)
               sscanf(a, "%s", val);
            else
               strcpy(val, a);
            strcpy(val2, ptr->val);

            if (ptr && ptr->val)
               FREE(ptr->val);
            ptr->val = copy_string(val);
            ptr->type = type;
            mod_rtl(ptr, val2, val);
            break;
       
         case MOD_SIDE:
            sscanf(&line[1], "%d %d %s", &blkid, &rtlid, val);
            ptr = find_rtl_id(rtlid, blkid);
            mod_side(ptr, ptr->se, val);
            if (ptr->se)
               FREE(ptr->se);
            if (strcmp(val, "none") == 0)
               ptr->se = (char *) NULL;
            else
               ptr->se = copy_string(val);
            break;

         case MOD_RES:
            sscanf(&line[1], "%d %d %s", &blkid, &rtlid, val);
            ptr = find_rtl_id(rtlid, blkid);
            mod_res(ptr, ptr->val, val);
            if (ptr->val)
               FREE(ptr->val);
            ptr->val = copy_string(val);
            break;

         case MOD_DEADS:
            sscanf(&line[1], "%d %d %s", &blkid, &rtlid, val);
            ptr = find_rtl_id(rtlid, blkid);
            deads = rinit();
	    ptr->deads = rinit();
            if (strcmp(val, "none") != 0) {
               reg[2] = '\0';
               a = val;
               while (*a != '\0') {
                  reg[0] = *a;
                  reg[1] = *(a+1);
                  rinsert_init(reg, &deads);
                  a += REGLOCLENGTH;
               }
            }
	    else
	      mod_deads(ptr, NULL, deads);
            rcpy(&ptr->deads, deads);
            rfree(deads);
            break;
    
         case MOD_TYPE:
            sscanf(&line[1], "%d %d %u", &blkid, &rtlid, &type);
            ptr = find_rtl_id(rtlid, blkid);
            ptr->old_type = ptr->type;
            ptr->type = type;
            mod_type(ptr);
            break;

         case MOD_KIND:
            sscanf(&line[1], "%d %d %u", &blkid, &rtlid, &kind);
            ptr = find_rtl_id(rtlid, blkid);
            ptr->old_kind = ptr->kind;
            ptr->kind = kind;
            mod_kind(ptr);
            break;

	 case LOCAL_ALLOC:
	   tregin[0] = '\0';
	   sscanf(&line[1], "%s %d %d %d %d %s", regname, &treg, &tnouse,
		  &treps, &tatype, tregin);
	   
	   /*update the variable information in struct locuse*/
	   for(loc_ptr=locs ; loc_ptr ; loc_ptr=loc_ptr->next)
	     {
	       if(strcmp(loc_ptr->item, regname) ==0)
		 {
		   loc_ptr->reg = treg;
		   loc_ptr->nouse = tnouse;
		   if(isreg(tregin))
		     strcpy(loc_ptr->regin, tregin);
		   loc_ptr->atype = tatype;
		   loc_ptr->reps = treps;
		   reorder_local(loc_ptr);
		   break;
		 }
	     }
	   if(loc_ptr == NULL)
	     fprintf(stderr, "apply_inittrans(LOCAL_ALLOC): variable %s not found\n", regname);
	   else
	     update_locuse(loc_ptr);
	   break;

	 case MAP_REGISTER:
	   sscanf(&line[1], "%d %d", &regm, &index);
	   regmap[regm] = index;
	   update_reg_mapping(regm, index);
	   break;
	   
         case MEASURE_POSITION:
	   if(measure_init)
	     return 0;
	   
         default:
            fprintf(stderr, "apply_inittrans (vpo) - bad msg type %c\n",
                    line[0]);
            exit(1);
	 }
       if(check_null_block() == 0)
	 {
	   printf("block list not null terminated\n");
	   /*exit(1);*/
	 }
       
       /*count the number of changes applied*/
       if(line[0]!=BEGINPHASE && line[0]!=BEGINTRANS && line[0]!=ENDPHASE
	  && line[0]!=ENDTRANS && line[0]!=LOCAL_ALLOC && line[0]!=MAP_REGISTER)
	 changes++;
       
       /*read the next line*/
       prev_ptr = trans_ptr;
       is_proceed = get_change(line, source, init_file, &trans_ptr);
     }

   /* close the trans file */
   if(source == APPLY_FROM_FILE)
     fclose(init_file);
   else{
     if(prev_ptr)
       prev_ptr->next = NULL;
     else
       transform_head = NULL;
     transform_tail = prev_ptr;
   }

   if (!session_over && !measure_phase && !measure_init)
      write_record_id(ENDSEQ);

   /* for a sanity check, we should check the control flow */
   if(!cf_flag)
     check_control_flow(TRUE);

   return initial_counts;
}

/*
 * get_next_change - get the next transform line
 */
int get_change(char line[], int from_where, FILE *init_file,
	       struct transform_list **ptr)
{
  extern struct transform_list *transform_head;
  
  if(from_where == APPLY_FROM_FILE){

    /*get line from file*/
    if((fgets(line, MAXSTRING, init_file) && line[0] != 'f'))
      return TRUE;
  }
  else{

    /*get next line from list*/
    if(*ptr)
      (*ptr) = (*ptr)->next;
    else
      (*ptr) = transform_head;
    
    if((*ptr)){
      sprintf(line, "%c%s", (*ptr)->mod_flag, (*ptr)->string);
      return TRUE;
    }
  }
  
  /*reached end of file / list*/
  return FALSE;
}
#endif /*INTERACT*/

/*calculate some required information and then check if the control flow
  is correct*/
void check_control_flow(int status)
{
  struct bblock *cblk, *cblk2;
  
  /* reset all of the up fields */
  top->up = (struct bblock *) NULL;
  for (cblk2 = top, cblk = top->down; cblk; cblk2 = cblk, cblk = cblk->down)
    cblk->up = cblk2;

  /* reset all of the predecessors */
  clear_all_preds();
  recalculate_all_preds();
  
  /* recalculate the status field as well */
  calc_status();
  
  /* for a sanity check, we should check the control flow */
  /*check_cf(status);*/
}
   
/*print the block and RTL ids to a file*/
void print_block_rtlids(FILE *pkfile)
{
  struct bblock *cblk;
  struct list *rtlptr;

  for (cblk = top; cblk; cblk = cblk->down)
    {
      fprintf(pkfile, "%d\n", cblk->block_id);
      for(rtlptr=cblk->rtls ; rtlptr ; rtlptr=rtlptr->next)
	fprintf(pkfile, "%d ", rtlptr->rtl_id);
      fprintf(pkfile, "\n");
    }
}  


/*check is block list is null terminated*/
int check_null_block()
{
  struct bblock *cblk;
  int count;

  for(cblk=top, count=0 ; cblk && count<100; cblk=cblk->down, count++);

  if(count == 100)
    return 0;
  else
    return 1;
}

#endif  /*endif INTERACT*/
