
/************************************************************************/
/*                                                                      */
/* Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   */
/*     The Rector and Visitors of the University of Virginia            */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from the University of Virginia.                                */
/* This code is distributed WITHOUT ANY WARRANTY.  No claims are        */
/*      made as to whether it serves any particular purpose or even     */
/*      works at all.                                                   */
/*                                                                      */
/************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/vectors.c,v $
 $Date: 2003/11/18 01:58:05 $
 $Revision: 1.24 $
 $Name:  $

 Log of changes:

 $Log: vectors.c,v $
 Revision 1.24  2003/11/18 01:58:05  clc5q
 Make number_basic_blocks() return the number of blocks in the function,
    and make calls to it reflect the fact that it now returns an int.

 Revision 1.23  2003/07/11 21:04:24  clc5q
 Add assertions in vinsert().

 Revision 1.22  2003/04/03 20:06:11  jg9h
 Merged latest in VISTA with latest in vpo.

 Revision 1.21  2003/03/05 15:45:58  clc5q
 Add assertions and diagnostic output.

 Revision 1.20  2003/02/06 04:46:29  jg9h
 Merged VISTA into mainline vpo under the INTERACT preprocessor flag

 Revision 1.19  2002/07/11 15:22:53  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.18  2002/06/05 20:08:54  jdh8d
 Changes to support the stache, including routines to build a callgraph.

 Some cosmetic changes.

 Minor bug fixes.

 Changed build scripts to build the mips compiler for SimpleScalar if
 SSAS is set to "-DSS_AS"

 Revision 1.17  2002/03/25 17:15:22  rv3s
 New routine rempty() added to check if a dead list is empty

 Revision 1.16  2002/02/12 16:08:34  jdh8d
 *** empty log message ***

 Revision 1.15  2001/09/24 20:35:36  jdh8d
 *** empty log message ***

 Revision 1.14  2001/09/24 20:33:21  jdh8d
 *** empty log message ***

 Revision 1.13  2001/09/12 17:13:04  jdh8d
 *** empty log message ***

 Revision 1.12  2001/07/23 18:35:00  rv3s
 *** empty log message ***

 Revision 1.11  2001/07/20 15:48:50  jdh8d
 *** empty log message ***

 Revision 1.10  2001/07/17 16:56:13  jdh8d
 *** empty log message ***

 Revision 1.9  2001/05/28 17:51:03  jdh8d
 *** empty log message ***

 Revision 1.8  2001/05/23 20:28:21  jdh8d


 Re-worked all of the backend to support variable
 sized registers.  IT IS  NO LONGER SAFE TO ASSUME
 THAT REGISTERS ARE TWO BYTES.  Mips currently has
 4-byte registers, others will follow (or be destroyed!)

 Revision 1.7  2000/10/18 02:46:54  jwd
 Added something to warn if reget was overflowing.

 Revision 1.6  1999/03/05 17:03:25  jwd
 Added a declaration to function rins()  for a temporary used by
 VET.
 Revision 1.5  1999/01/06 15:52:49  jwd
 Added VET directives.
 Revision 1.4  1998/06/26 21:00:22  jwd
 The type of the first parameter to the function ddelete was
 incorrect.  The function also had an extra left brace.
 Revision 1.3  1998/04/03 18:20:26  jwd
 First ANSI C version.

*************************************************************************/

/*------------------------------------------------
   History of Changes
   1/13/98: Modified code to produce more mudular and ansi-c acceptable
            function prototyping.

------------------------------------------------ */

#include <stdio.h>
#include <assert.h>

#include "vpo.h"
#include "vectors.h"
#include "assign.h"
#include "misc.h"

#ifdef INTERACT
#include "undo.h"
#endif

#ifdef VET
#include "view.h"
#endif

int veclen = 0;		/* length of a RTLINE number vector in integers */
int dveclen;		/* length of a dominator vector in integers     */
int rveclen;		/* length of a register vector in integers      */


/*
 * The following routines deal with vectors that contain a bit position for
 * each RTL in the function
 */

/*
 * vinsert - add an item to a vector
 */
void vinsert(vector * sptr, unsigned int r) {

   unsigned int *ptr, *end;

   assert(NULL != sptr);


   if (!*sptr) {
      assert(0 < veclen);
      (*sptr) = (vector) (ptr = (unsigned int*)VALLOC);
      for (end = ptr + veclen; ptr != end; *ptr++ = 0)
         ;
   }
   assert(NULL != *sptr);
   (*sptr)[r >> intscale] |= (((unsigned) 1) << (r & bitsperint));
}

/*
 * vin - indicate if an item line is in a vector
 */
int vin(vector ptr, unsigned int r) {

   return (ptr && ptr[r >> intscale] & (((unsigned) 1) << (r & bitsperint)));
}

/*
 * number_rtls - assign a bit position to each RTLs and set "veclen"
 */
void number_rtls() {

   register struct list *ptr;
   register int onum;
   register struct bblock *cblk;
   static int old_veclen=0, first=TRUE;
   static int allocated_veclen;

   onum = 1;
   for (cblk = top; cblk; cblk = cblk->down) {
      for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
	 if (!(ptr->type & ASMLINE))
	    ptr->onum = onum++;
      }
   }
   veclen = (onum >> intscale) + 1;

   /*because of the memory allocation used in VPO, vectors that are freed are
     saved on the 'freelist' for vectors, and new vectors are allocated
     memory from this list. Some optimizations like loop unrolling may
     increase the number of RTLs in the function, and which may result in the
     vector requiring more bytes that previously allocated. When this
     happens, and the memory is still allocated from the free list which has
     smaller length vectors we may run into potential memory problems. Check
     for this condition. A temporary fix to this problem would be restart the
     program with a larger vector size from the start. A permanant fix would
     require us to copy all the vectors to the larger size vectors and free
     the current 'freelist' for vectors*/
   if(old_veclen && old_veclen < veclen){
     
     fprintf(stderr, "number_rtls: New vector length exceeds the previous allocated vector length\n");
     /*exit(0);*/

     veclen = allocated_veclen = veclen + 3;
     realloc_vectors(old_veclen, veclen);
   }

   /*measurement code generally increases veclen later. So we give it more
     right upfront*/
   if(first){
     veclen = allocated_veclen = veclen + 3;
     first = FALSE;
   }
   else
     veclen = allocated_veclen;
   old_veclen = veclen;
}


/*
 * The following routines deal with vectors that contain a bit position for
 * each basic block in the function
 */

/*
 * dinsert - adds a block reference to a dvector list
 */
void dinsert(dvector * sptr, short unsigned int r) {

   register unsigned int *ptr, *end;

   if (!*sptr) {
      *sptr = (dvector) (ptr = DALLOC);
      end = ptr + dveclen;
      while (ptr != end)
	 *ptr++ = 0;
   }
   (*sptr)[r >> intscale] |= (((unsigned) 1) << (r & bitsperint));
}

/*
 * din - indicate if a basic block is in a dvector
 */
int din(dvector ptr, short unsigned int r) {

   return (ptr && ptr[r >> intscale] & (((unsigned) 1) << (r & bitsperint)));
}

/*
 * dinter - calculate the intersection of two dvector lists
 */
void dinter(dvector res, dvector ptr) {

   register unsigned int *ptr1, *ptr2, *end;

   if (res && ptr) {
      end = (ptr1 = (unsigned int *) res) + dveclen;
      ptr2 = (unsigned int *) ptr;
      while (ptr1 != end)
	 *ptr1++ &= *ptr2++;
   }
}

/*
 * ddelete - delete a block from a dvector list
 */
void ddelete(dvector *ptr, unsigned short r) {
   if (*ptr)
      (*ptr)[r >> intscale] &= ~(((unsigned) 1) << (r & bitsperint));
}

/*
 * dall - initialize a dvector list containing all basic blocks
 */
dvector dall() {

   register unsigned int *ptr, *end;

   end = (ptr = DALLOC) + dveclen;
   do {
      *(--end) = -1;
   } while (ptr != end);
   return ((dvector) ptr);
}

/*
 * dcpy - copy a dvector list to another one
 */
void dcpy(dvector * dst, dvector src) {

   register unsigned int *ptr1, *ptr2, *end;

   if (src) {
      if (!*dst)
	 *dst = (dvector) DALLOC;
      end = (ptr1 = (unsigned int *) *dst) + dveclen;
      ptr2 = (unsigned int *) src;
      while (ptr1 != end)
	 *ptr1++ = *ptr2++;
   }
   else if (*dst) {
      dfree(*dst);
      *dst = dinit();
   }
}

/*
 * dcmp - compare two dvector lists
 */
int dcmp(dvector a, dvector b) {

   register unsigned int *ptr1, *ptr2, *end;

   if (a && b) {
      end = (ptr1 = (unsigned int *) a) + dveclen;
      ptr2 = (unsigned int *) b;
      while (ptr1 != end)
	 if (*ptr1++ != *ptr2++)
	    return (FALSE);
      return (TRUE);
   }
   return (a == b);
}


/*
 * number_basic_blocks - number basic blocks and evaluate the dveclen value
 */
int number_basic_blocks() {

   struct bblock *cblk;
   int bnum;
   static int old_dveclen=0;

   /*
      Assign each basic block a bit vector position in a vector that can be
      used to contain each of the basic blocks in the function.
   */
   bnum = 1;
   for (cblk = top; cblk; cblk = cblk->next)
      cblk->bnum = (unsigned short) bnum++;

   /*
      Determine the length of such a vector in terms of the basic word size of
      the host machine.
   */
   dveclen = bnum >> intscale;
   if (bnum & bitsperint)
      dveclen++;

   /*look at comments for veclen in number_rtls*/
   if(old_dveclen && old_dveclen < dveclen){

     fprintf(stderr, "number_basic_blocks: New dveclen size exceeds previous size\n");
     /*exit(0);*/

     /*re-allocate all the dvectors to increase their size*/
     realloc_dvectors(old_dveclen, dveclen);
   }
   old_dveclen = dveclen;
   
   return bnum;
}
#ifdef VET
/*
 * rcmp - return TRUE if vectors have all items in common
 */
int rcmp(rlist res, rlist ptr)
{
    register unsigned int *ptr1, *ptr2, *end;

    if (res && ptr) {
      end = (ptr1 = (unsigned int *) res) + rveclen;
      ptr2 = (unsigned int *) ptr;
      while (ptr1 != end)
        if (*ptr1++ != *ptr2++)
          return(FALSE);
      return(TRUE);
      }
    return(res == ptr);
}

/*
 * rsave - save a register list
 */
void rsave(rlist *dst, rlist src)
{
    register unsigned int *ptr1, *ptr2, *end;

    *dst = rinit();
    if (ptr2 = (unsigned int *) src) 
    {
	if (!(ptr1 = (unsigned int *) *dst))
	{
	    *dst = (rlist) (ptr1 = RALLOC);
	}
	for (end = ptr1 + rveclen; ptr1 != end; *ptr1++ = *ptr2++);
    }
    else {
	rfree(*dst);
	*dst = rinit();
    }
}

/*
 * x_send_deads - finds dead list and sends message to VET
 */
void x_send_deads(rlist rold, rlist *sptr)
{
    struct bblock *cblk;
    struct list *ptr;

    /* This is an inefficient way to get to the appropriate rtl.
       But this method minimizes the changes to vpo. */
#ifdef INTERACT
    if ((started_trans || user_directed) && !rcmp(rold, *sptr)) {
#else
      if (started_trans && !rcmp(rold, *sptr)) {
#endif 	 
       /* if can't find rtl, then message is not sent */
       for (cblk = top; cblk; cblk = cblk->down)
          for (ptr = cblk->rtls; ptr; ptr = ptr->next)
             if (ptr->deads && ptr->deads == *sptr) {
#ifdef INTERACT
	       mod_deads(ptr, rold, ptr->deads);				 
#else 
	       mod_deads(ptr);
#endif
                return;
	     }
   }
}
#endif


/*
 * The following routines deal with vectors that contain a bit position for
 * each register and local used in the function
 */

/*
 * runion - computes the union of two rlists
 */
void runion(rlist * result, rlist vec) {

   register unsigned int *ptr1, *ptr2, *end;
#if defined(VET) || defined(INTERACT)
    rlist rold;
    void x_send_deads(), rsave();

#ifdef INTERACT
    if (swq)
#endif
      rsave(&rold, *result);
#endif

   if (ptr2 = (unsigned int *) vec)
      if (ptr1 = (unsigned int *) *result)
	 for (end = ptr1 + rveclen; ptr1 != end; *ptr1++ |= *ptr2++);
      else {
	 *result = (rlist) (ptr1 = RALLOC);
	 for (end = ptr1 + rveclen; ptr1 != end; *ptr1++ = *ptr2++);
      }
#if defined(VET) || defined(INTERACT)
    x_send_deads(rold, result);
#endif
}

/*
 * runiond - computes {a} = {a} + ({b} - {c}), returns TRUE if {a} changes
 */
int runiond(rlist * a, rlist b, rlist c) {

   register unsigned int *ptr1, *ptr2, *ptr3, *end, tmp, change;
#if defined(VET) || defined(INTERACT)
    rlist rold;
    void x_send_deads(), rsave();

#ifdef INTERACT
	 if (swq)
#endif
    rsave(&rold, *a);
#endif

   change = FALSE;
   if (b) {
      if (!*a) {
	 *a = (rlist) (ptr2 = ptr1 = RALLOC);
	 for (end = ptr2 + rveclen; ptr2 != end; *ptr2++ = 0);
      }
      else
	 end = (ptr1 = (unsigned int *) *a) + rveclen;
      ptr2 = (unsigned int *) b;
      if (ptr3 = (unsigned int *) c)
	 while (ptr1 != end) {
	    tmp = *ptr1;
	    if ((*ptr1++ |= (*ptr2++ & ~(*ptr3++))) != tmp)
	       change = TRUE;
	 }
      else
	 while (ptr1 != end) {
	    tmp = *ptr1;
	    if ((*ptr1++ |= *ptr2++) != tmp)
	       change = TRUE;
	 }
   }
#if defined(VET) || defined(INTERACT)
    x_send_deads(rold, a);
#endif
   return (change);
}

/*
 * rcommon - return TRUE if vectors have items in common
 */
int rcommon(rlist res, rlist ptr) {

   register unsigned int *ptr1, *ptr2, *end;

   if (res && ptr) {
      end = (ptr1 = (unsigned int *) res) + rveclen;
      ptr2 = (unsigned int *) ptr;
      while (ptr1 != end)
	 if (*ptr1++ & *ptr2++)
	    return (TRUE);
   }
   return (FALSE);
}

/*
 * rinsert - enters a register string into an rlist
 */
void rinsert(char *reg, rlist * sptr) {

   register unsigned int *ptr, *end;
   int r;
#if defined(VET) || defined(INTERACT)
    rlist rold;
    void x_send_deads(), rsave();

#ifdef INTERACT
	 rold=rinit();
	 if (swq)
#endif
    rsave(&rold, *sptr);
#endif

   if (!*sptr) {
      *sptr = (rlist) (ptr = RALLOC);
      for (end = ptr + rveclen; ptr != end; *ptr++ = 0);
   }
   r = regtovec(reg);
   if(r<0)
   {
      (void) fprintf(stderr,"unallocated item:");
      show(reg);
      if (NULL != fn)
         (void) fprintf(stderr, " in function: %s\n", fn);
      assert(r >= 0);
      exit(EXIT_FAILURE);
   }
   (*sptr)[r >> intscale] |= (((unsigned) 1) << (r & bitsperint));
#if defined(VET) || defined(INTERACT)
    x_send_deads(rold, sptr);
    rfree(rold);
#endif
}

/*
 * rinsert_init - enters a register string into an rlist
 */
void rinsert_init(char *reg, rlist * sptr) {

   register unsigned int *ptr, *end;
   int r;

   if (!*sptr) {
      *sptr = (rlist) (ptr = RALLOC);
      for (end = ptr + rveclen; ptr != end; *ptr++ = 0);
   }
   r = regtovec(reg);
   (*sptr)[r >> intscale] |= (((unsigned) 1) << (r & bitsperint));
}

/*
 * rins - enters a register number into an rlist
 */
void rins(unsigned int r, rlist * sptr) {

   register unsigned int *ptr, *end;
#if defined(VET) || defined(INTERACT)
   rlist rold;

#ifdef INTERACT
	if (swq)
#endif
    rsave(&rold, *sptr);
#endif

   if (!*sptr) {
      *sptr = (rlist) (ptr = RALLOC);
      for (end = ptr + rveclen; ptr != end; *ptr++ = 0);
   }
   (*sptr)[r >> intscale] |= (((unsigned) 1) << (r & bitsperint));
#if defined(VET) || defined(INTERACT)
    x_send_deads(rold, sptr);
#endif
}

/*
 * rdelete - deletes a register string from an rlist
 */
void rdelete(char *reg, rlist * ptr) {

   int r;
#if defined(VET) || defined(INTERACT)
    rlist rold;

#ifdef INTERACT
	 if (swq)
#endif
    rsave(&rold, *ptr);
#endif

   if (*ptr) {
      r = regtovec(reg);
      if(r>=0)
      	(*ptr)[r >> intscale] &= ~(((unsigned) 1) << (r & bitsperint));
   }
#if defined(VET) || defined(INTERACT)
    x_send_deads(rold, ptr);
#endif
}

/*
 * rdel - deletes a register number from an rlist
 */
void rdel(unsigned int r, rlist *ptr) {
#if defined(VET) || defined(INTERACT)
    rlist rold;

#ifdef INTERACT
	 if (swq)
#endif
    rsave(&rold, *ptr);
#endif
   if (*ptr)
      (*ptr)[r >> intscale] &= ~(((unsigned) 1) << (r & bitsperint));
#if defined(VET) || defined(INTERACT)
    x_send_deads(rold, ptr);
#endif
}

/*
 * rcpy - copy a register list
 */
void rcpy(rlist * dst, rlist src) {

   register unsigned int *ptr1, *ptr2, *end;
#if defined(VET) || defined(INTERACT)
    rlist rold=rinit();
    void x_send_deads(), rsave();

#ifdef INTERACT
    if (swq && !undoing_transformation)
#endif
    rsave(&rold, *dst);
#endif

   if (ptr2 = (unsigned int *) src) {
      if (!(ptr1 = (unsigned int *) *dst))
      {
	 rfree(*dst);
	 *dst = (rlist) (ptr1 = RALLOC);
      }
      for (end = ptr1 + rveclen; ptr1 != end; *ptr1++ = *ptr2++);
   }
   else {
      rfree(*dst);
      *dst = rinit();
   }
#if defined(VET) || defined(INTERACT)
#ifdef INTERACT
   if (swq && !undoing_transformation)
#endif
     x_send_deads(rold, dst);
   rfree(rold);
#endif
}

/*
 * reget - returns string containing registers indicated by vector
 */
void reget(rlist ptr, char *reg) {

   register unsigned int c;
   register int rnum, end;
   register unsigned int *rptr;
   char* initial_reg=reg;
   int copy=0;

   if (rptr = (unsigned int *) ptr) {
      end = rveclen;
      rnum = 0;
      while (end--)
	 if (c = *rptr++) {
	    do {
	       if (c & 1) {
                  copylocOr2(reg,sets[rnum].item);
                  *(skiplocOr2(reg)) = '\0';
                  *(skiplocOr2(reg) + 1) = '\0';
                  reg = skiplocOr2(reg)+1;

	       }
	       rnum++;
	    } while (c >>= (unsigned int) 1);
	    rnum = (rnum + bitsperint) & ~bitsperint;
	 }
	 else
	    rnum += (bitsperint + 1);
   }
   *reg = '\0';
   if (reg-initial_reg > MAXSTRING)
      fprintf(stderr,"Potential overrun of memory in reget\n");
}

/*
 * rexists - returns non-zero if register string is present in an rlist
 */
int rexists(char *reg, rlist ptr) {

   int r;

   r = regtovec(reg);
   if(r<0)
   {
	fprintf(stderr, "rexists called with non-register\n");
	 
   }
   return (ptr && ptr[r >> intscale] & (((unsigned) 1) << (r & bitsperint)));
}
/*
 * rin - returns non-zero if register number is present in an rlist
 */
int rin(unsigned int r, rlist ptr) {

   return (ptr && ptr[r >> intscale] & (((unsigned) 1) << (r & bitsperint)));
}

/*
 * rin - returns non-zero if register number is present in an rlist
 */
int raliasin(char* r, rlist ptr) 
{

	char* a;
	if(!isreg(r) || pseudoreg(r))
	{
		struct setlist* sptr=&sets[regtovec(r)];
		return rin(sptr->num,ptr);
	}
	for(a=aliases(r);*a;a=skiploc(a)+1)
	{
		struct setlist* sptr=&sets[regtovec(a)];
		if(rin(sptr->num,ptr))
			return TRUE;
	}
	return FALSE;
}

int rempty(rlist ptr) 
{
  char reg[MAXSTRING];
  reget(ptr,reg);
  if (reg[0] == 0)
    return TRUE;

  return FALSE;

}


/*
 * realloc_dvectors - reallocate all dvectors to new vector size, if the
 *                    dvector size ever increases from what has been
 *                    currently involved
 */
void realloc_dvectors(int old_vlen, int new_vlen)
{
  dvector dptr;
  struct bblock *cblk;
  struct locuse *cloc;
  extern struct bblock *top;
  extern struct locuse *locs;

  /*first we "free" all memory for dvectors from the freelist, so that when
    we allocate new dvectors they are not allocated the same old space from
    the old list !!. dvectors are never deallocated to the free list, so
    don't need this step*/

  /*create a new dvector for everything that has been allocated a dvector and
    copy it to the new storage*/
  for(cblk=top ; cblk ; cblk=cblk->next){

#if defined(MEASURE) || defined(INTERACT)
    if(dptr = cblk->dom){
      cblk->dom = DALLOC;
      memset(cblk->dom, 0, new_vlen);
      memcpy(cblk->dom, dptr, old_vlen);
    }
#endif

#ifdef MEASURE
    if(dptr = cblk->equivclass){
      cblk->equivclass = DALLOC;
      memset(cblk->equivclass, 0, new_vlen);
      memcpy(cblk->equivclass, dptr, old_vlen);
    }
#endif    
  }

  /*copy over the dvectors in the locuse structure*/
  for(cloc=locs ; cloc ; cloc=cloc->next){

    if(dptr = cloc->setin){
      cloc->setin = DALLOC;
      memset(cloc->setin, 0, new_vlen);
      memcpy(cloc->setin, dptr, old_vlen);
    }

    if(dptr = cloc->usedin){
      cloc->usedin = DALLOC;
      memset(cloc->usedin, 0, new_vlen);
      memcpy(cloc->usedin, dptr, old_vlen);
    }
  }
}


/*
 * realloc_vectors - reallocate all vectors to new vector size, if the
 *                   vector size ever increases from what has been
 *                   currently allocated
 */
void realloc_vectors(int old_vlen, int new_vlen)
{
  vector vec;
  struct bblock *cblk;
  struct list *rtl;
  struct locuse *loc;
  struct ignode *ig;
  struct loopl *loop;
  extern struct bblock *top;
  extern struct locuse *locs;
  extern struct ignode *igptr;
  extern struct loopl *loops;

  /*  fprintf(stderr, "Realloc vectors entered\n");
      fflush(stderr);*/
  
  /*set freelist to NULL, so that new blocks are not allocated from the
    freelist*/
  freeptrs[VECTORTYPE] = NULL;

  /*re-allocated structures in RTLs*/
  for(cblk=top ; cblk ; cblk=cblk->next){

    for(rtl=cblk->rtls ; rtl ; rtl=rtl->next){

#if SCHED || FILL
      if(vec = rtl->uses){
	rtl->uses = VALLOC;
	memset(rtl->uses, 0, new_vlen);
	memcpy(rtl->uses, vec, old_vlen);
      }

      if(vec = rtl->sets){
	rtl->sets = VALLOC;
	memset(rtl->sets, 0, new_vlen);
	memcpy(rtl->sets, vec, old_vlen);
      }
#endif
    }
  }

  /*re-allocate structures in locuse*/
  for(loc=locs ; loc ; loc=loc->next){

    if(vec = loc->olist){
      loc->olist = VALLOC;
      memset(loc->olist, 0, new_vlen);
      memcpy(loc->olist, vec, old_vlen);
    }
  }

  /*re-allocate structures in loopl*/
  for(loop=loops ; loop ; loop=loop->next){

    if(vec = loop->regs){
      loop->regs = VALLOC;
      memset(loop->regs, 0, new_vlen);
      memcpy(loop->regs, vec, old_vlen);
    }

    if(vec = loop->regs_in){
      loop->regs_in = VALLOC;
      memset(loop->regs_in, 0, new_vlen);
      memcpy(loop->regs_in, vec, old_vlen);
    }
  }

  /*re-allocate structures in ignode*/
  for(ig=igptr ; ig ; ig=ig->next){

    if(vec = ig->regs){
      ig->regs = VALLOC;
      memset(ig->regs, 0, new_vlen);
      memcpy(ig->regs, vec, old_vlen);
    }
  }

  /*  fprintf(stderr, "Realloc vectors exit\n");
      fflush(stderr);*/
}


/*
 * vec2char - copy the vector bits into character array
 */
void vec2char(vector vec, char *c_vec)
{
  char *c, regstr[MAXSTRING];
  int i, type, max, lregs_type[MAXCLASSES+1];

  for(i=0 ; i<=MAXCLASSES ; i++)
    lregs_type[i] = 0;

  reget(vec, regstr);
  for(c=regstr ; *c ; c=skiploc(c)+1){
    if(isRegister(c)){
      type = getRegType(c);
      lregs_type[type]++;
    }
  }

  /*construct a string containing the counts of live registers of each type*/
  c_vec[0] = '\0';
  for(i=0 ; i<=MAXCLASSES ; i++)
    sprintf(c_vec, "%s%d", c_vec, lregs_type[i]);
}
