/***********************************************************************
 * 
 * Copyright (c) 1995 Ann H. Bailey.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/view.c,v $
 * $Revision: 1.18 $
 ***********************************************************************/

/* FILE: view.c */
#ifdef VET

#define __VIEW_C__

#include <sys/select.h>
#include <sys/time.h>
#include <setjmp.h>

#include "vpo.h"
#include "view.h"
#include "assign.h"
#include "vpo2.h"
#include "undo.h"
#include "misc.h"
#include "server_view.h"

#ifdef INTERACT
#include "regs.h"
#include "view_socket.h"

#define ERR_QUIT(MSG) {\
 fprintf(stderr,MSG);\
 exit (-1);\
 }
#define ASSERT(a) {if (!a) { fprintf(stderr, "Assert failed\n"); foo();}}
void foo() {}

#endif

#define STRING     256          /* Length of work strings */
#ifdef DEBUG
FILE *errlog;			/* Log file. */
#endif

int viewing = FALSE;		/* Signal that VET is started. */
int initial_set_sent = FALSE;	/* Signal when the initial RTLS sent to VET. */
int send_messages = FALSE;	/* User signaled to view a function. */
int sanity_enabled = FALSE;	/* Sanity check enabled for a function. */
int delayed_viewing = FALSE;	/* Used to delay sending a transformation */
int DeadsareBitVectors;		/* Used to signal deads stored in sideffects */
int quit_vpo = FALSE;		/* User aborted but allow VPO to finish. */
char *current_func = NULL;	/* Function currently viewing. */
int started_phase = FALSE;	/* Signals begin phase occurred. */
int started_trans = FALSE;	/* Signals begin transformation occurred. */

int transnum;                   /* current transformation number   */
int opttransnum;                /* curr nonrequired trans number   */
char transbuf[BUFSIZ];


#ifdef INTERACT
/* The number of transformations information is stored as an array
   since this information will be retained between compilations of
   functions.  (Remember VPO frees all dynamically allocated
   information after the compilation of a function.) */

int m_trans_pending = FALSE;    /* Transformation pending to be sent. */
int m_phase_pending = FALSE;    /* Phase pending to be sent.          */
int type_trans = 0;             /* Type of the last transformation    */
#endif

/* The number of transformations information is stored as an array
   since this information will be retained between compilations of
   functions.  (Remember VPO frees all dynamically allocated
   information after the compilation of a function.) */

#define MAXFUNCS 2000

static struct {
   char basename[80];           
   char funcname[80];
   int numtrans;
   } trans[MAXFUNCS];
static int numfuncs;

/**********************************************************************
 * 
 * startup_viewer - Create an inter-process stream for communication 
 *                 between VPO and VET.
 *
 **********************************************************************/
void startup_viewer(char *host, int port)
{
  int process_id;	/* Process id used for creating log file */
  char fname[STRING];	/* Filename for log file */

#ifdef INTERACT  
  if (!swq || measure_flag || measure_phase)
    return;
#endif
  
#ifdef DEBUG
  /*
   * Obtain our current process identifier
   */
  process_id = (int) getpid();
  /* Create log file name based on pid */
  sprintf(fname, "vpo%d.log", process_id);
  
  if (! (errlog = fopen(fname, "w"))) {
    fprintf(stderr, "vpo (view): could not create %s\n", fname);
    exit(1);
  }
  setbuf(errlog, srrlogbuf);

  fprintf(errlog,"startup viewer\n");
#endif

  /* setup port/socket to send messages to vet */
  setup_client_socket(host, port);

  /* Enable flag which indicates VET is running */
  viewing = TRUE;

#ifdef INTERACT  
  /* Send message to user interface to indicate that the user will
       interact with the compiler. */
  if (user_directed)
    write_record_id(USER_INTERACT);
#endif
}


/***********************************************************************
 * 
 * send_func_name - Send the function name to VET.
 *
 ***********************************************************************/
void send_func_name(char *funcname)
{
    char command[MAXSTRING];
    char msgid;

#ifdef INTERACT    
    if (!swq || measure_flag || measure_phase)
      return;
#endif
    
    /* Save current function for error checking */
    current_func = funcname;

    /* 
     * Turn on flag send_messages so that function name
     * may be sent to VET.  This flag will be reset if
     * the user does not want to view this functions's rtls.
     */
    send_messages = TRUE;

    /* 
     * Reset sanity flag - this flag is changed from VET 
     * at the beginning of every new function.
     */
    sanity_enabled = FALSE;

    /* Send vet function name */
#ifdef INTERACT    
    if(!measure_init)
#endif      
      {
	write_record_id(BEGINFUNCTION);
	write_line(funcname);
      }

    /* 
     * Wait for vet to send reply to whether the initial set should be 
     * sent or next function.
     */

    /* Simulate old send_init_set */
    send_messages = TRUE;

#ifdef INTERACT
    if(!measure_init)
#endif
      {
      msgid = read_char();
      switch (msgid) {
      case SANITY_CHECK:
	msgid = read_char();
	sanity_enabled = (msgid == 'T') ? TRUE : FALSE;
	break;
      case ABORT_VPO:
	send_messages = FALSE;
	msgid = read_char();

	/*
	 * VET sent ABORT message - user may either have VPO continue
	 * and exit gracefully but not send messages to VET or kill VPO.
	 */
	if (msgid == 'T') {
	  end_viewer();
	  exit(0);
	}
	else 
	  quit_vpo = TRUE;
	break;
      default:
	fprintf(stderr, "vpo (send_func_name): wrong message id: %c\n",msgid);
	end_viewer();
	exit(1);
      } /* end switch */
    }
} /* end send_func_name */

/***********************************************************************
 * 
 * send_end_func - send VET and end of function message including the 
 *                 function name.
 *
 ***********************************************************************/
void send_end_func(char *funcname)
{
#ifdef INTERACT  
  if (!swq || measure_flag || measure_phase)
    return;
#endif  

  /* Check that we are ending the same function we started */
  if (strcmp(current_func, funcname)) {
    fprintf(stderr, "vpo (send_end_func): current func = %s | end func = %s\n", 
	    current_func, funcname);
    end_viewer();
    exit(1);
  }

#ifdef INTERACT  
  if (!user_directed) {
    dump_block_counts(top, "measure", fn, 0);
    write_record_id(ENDSEQ);
  }
#endif

  write_record_id(ENDFUNCTION);
  
  /* 
   * Reset flag so messages are not sent to VET until 
   * the next function's initial set is sent over.
   */
  initial_set_sent = FALSE;

  /*
   * Reset flag so messages are not sent until the
   * next function is started.
   */
  send_messages = FALSE;
} /* end send_end_func */

#ifdef INTERACT
void send_frequency_information(struct bblock *firstbb) {
  struct bblock *bblk;
  int i = 0;
  write_record_id(BEGINFREQ);

  for (bblk = firstbb; bblk; bblk = bblk->down) {
    write_bb_pointer(bblk);
    write_num(i++);
  }
  write_record_id(ENDFREQ);
}
#endif

/*********************************************************************
 *
 * send_init_set - Send the initial set of RTL's to VET.
 *
 *********************************************************************/
void send_init_set(char *funcname)
{
#ifdef INTERACT  
  if (!swq || measure_flag || measure_phase)
    return;
#endif
  
  /* 
   * Set global flag for viewing started.  Necessary because 
   * when vpo is reading in RTLs for a function it calls routines 
   * that have VET calls in it (i.e. NEWBB).  Therefore we don't 
   * want messages sent until the initial set is sent to VET.
   */
  initial_set_sent = TRUE;

  /*
   * Send the basename of the .c file
   */
#ifdef INTERACT
  write_line(basename);
#endif
  
  transnum = 0;
  opttransnum = 0;

  /*
   * Go through all of the basic blocks starting with the entry block and
   * continuing along in the output order.
   */
  send_basicblocks(top);

  write_record_id(ENDINITSET);
} /* end send_init_set */


/*********************************************************************
 *
 * prep_for_funstart - Reset the transformation counters to zero in 
 *                     preparation for a new function's optimization.
 *
 *********************************************************************/
void prep_for_funcstart()
{
  transnum = 0;
  opttransnum = 0;
} /* end prep_for_funcstart */



/***********************************************************************
 *
 * send_basicblocks - Send current state of RTL's to VET.
 *
 ***********************************************************************/
void send_basicblocks(struct bblock *firstbb)
{
  struct bblock *bblk;
#ifdef INTERACT
  extern char *ret;
  if (!swq || measure_flag || measure_phase)
    return;
#endif

  /*
   * Go through all of the basic blocks starting with the entry block and
   * continuing along in the output order.
   */
  for (bblk = firstbb; bblk; bblk = bblk->down) {
    
    /* Send the basic block label */
    write_record_id(NEWBB);
#ifdef INTERACT
    write_bb_pointer(bblk);
#else
    write_pointer(bblk);
#endif
    write_line(((bblk->label) ? bblk->label : "none"));

    /* Send the unconditional successor */
    if (bblk->left
#ifdef INTERACT
	&&!(bblk->left->label && strcmp(bblk->left->label, ret) == 0)
#endif
	) {
      write_record_id(BBLEFT);
#ifdef INTERACT
      write_bb_pointer(bblk->left);
#else
      write_pointer(bblk->left);         
#endif   
    }

    /* Send the conditional successor */
    if (bblk->right) {
      write_record_id(BBRIGHT);
#ifdef INTERACT         
      write_bb_pointer(bblk->right);
#else
      write_pointer(bblk->right);
#endif
    }

    /*
     * Go through all the rtls for the basic block
     */
	
    send_rtls(bblk->rtls);
  }
} /* end send_basicblocks */

/**********************************************************************
 *
 * send_rtls - send over all rtls in list
 *
 **********************************************************************/
void send_rtls(struct list *rtlptr)
{
  char *a, msg[MAXSTRING], reg[MAXSTRING];
  int i;  /* index */

#ifdef INTERACT  
  if (!swq || measure_flag || measure_phase)
    return;
#endif
  
  while(rtlptr) {
    
    switch (rtlptr->type & ~(ARGLINE | SIDEFFECT | COMBLINE |
			     FUNCNAME | CASELINE | DELAYLINE)) {
	    
    case ASMLINE:
    case COMMENT:
      write_record_id(VET_RTL);
      /* Send the rtl address */
#ifdef INTERACT
      write_rtl_pointer(rtlptr);
#else
      write_pointer(rtlptr);
#endif         
      write_num(rtlptr->type);
      write_line(rtlptr->val);
      break;
	    
    case RTLINE:
      write_record_id(VET_RTL);
      /* Send the rtl address */
#ifdef INTERACT
      write_rtl_pointer(rtlptr);
#else
      write_pointer(rtlptr);
#endif
      write_num(rtlptr->type & ~COMBLINE);
      write_line(convert_rtl(rtlptr->val));
#ifdef INTERACT
      /*write the assembly value to the socket
	added by cai, baosheng*/
      write_line(trans_to_assem(rtlptr));
#endif
	    
      /* Send the links */
      if (rtlptr->lnks[0]) {
	write_record_id(VET_RTLLINKS);
	for (i = 0; i < MAXLINKS && rtlptr->lnks[i]; i++) {
	  /* Send over the address of the rtl */
#ifdef INTERACT
	  write_rtl_pointer(rtlptr->lnks[i]);
#else
	  write_pointer(rtlptr->lnks[i]);
#endif
	}
	/* Signal end of links */
	write_num(0);
      }
	    
      /* 
       * Send the deads list
       * Before Vectorize_Deads is called the deads
       * are stored as a string in the SideEffects
       * list.  After Vectorize_Deads the deads lists
       * are bit vectors and are stored in the Deads
       * list.  
       */
	    
      if (DeadsareBitVectors) {
	/*
	 * If deads exist need to convert bit
	 * vectors to a string before sending to
	 * VET.
	 */
	if (rtlptr->deads) {
	  write_record_id(VET_DEADS);
	  reget(rtlptr->deads, reg);
	  msg[0] = '\0';
	  for (a = reg; *a; a = skiploc(a)+1) {
	    strcat(msg, convert_rtl(a));
	  }
	  write_line(msg);
		    
	}
	/* Check if a side effect exists */
	if (rtlptr->se) {
	  write_record_id(VET_SIDE);
	  write_line(convert_rtl(rtlptr->se));
	}
      }
      else {
	/*
	 * Deads are not bit vectors and if
	 * they exist they're stored in SideEffects
	 * as a string.
	 */
	if (rtlptr->se) {
	  write_record_id(VET_DEADS);
	  write_line(convert_rtl(rtlptr->se));
	}
      }
      break;
	    
    default:
      write_record_id(VET_RTL);
      /* Send the rtl address */
#ifdef INTERACT
      write_rtl_pointer(rtlptr);
#else
      write_pointer(rtlptr);
#endif
      write_num(rtlptr->type & ~COMBLINE);
      write_line(convert_rtl(rtlptr->val));
    } /* switch */
	
    rtlptr = rtlptr->next;

  } /* While rtlptr */
} /* end send_rtls */

#ifdef INTERACT

/**********************************************************************
 *
 * Functions for Interactive compiling 
 *
 **********************************************************************/

/*************************************************************
copy_string: copy one string from src to dst.
 ************************************************************* */
char *copy_string(char *src)
{
  char *dst=NULL;
  if (src)
    dst=strcpy((char *)CALLOC(strlen(src)+1), src);
  return dst;
}

/**********************************************************************
 *
 * print_undo_list - a debug fuction, prints the list of phases currently
 *                   applied to this function
 *
 **********************************************************************/
void print_undo_list()
{
  int i=0;
  int index;
  struct transform_list *tptr;
  extern struct transform_list *transform_head;

  fprintf(stderr, "Printing phases: \n");
  for(tptr=transform_head ; tptr ; tptr=tptr->next){
    if(tptr->mod_flag == BEGINPHASE){
      index = atoi(tptr->string);
      fprintf(stderr, "Phase %c: %s\n", tptr->mod_flag, phase_names[index]);
    }
  }
}


/**********************************************************************
 *
 * add_to_transform_list - add the string to the end of the transform list
 *
 **********************************************************************/
void add_to_transform_list(char msg_type, char buff[])
{
  struct transform_list *new;
  extern struct transform_list *transform_head, *transform_tail;

  /*create a new node*/
  new = (struct transform_list*)myalloc_trans(sizeof(struct transform_list));
  new->mod_flag = msg_type;
  if (buff){
    new->string = (char *)myalloc_trans(strlen(buff)+1);
    strcpy(new->string, buff);
  }
  else
    new->string = NULL;
  new->next = NULL;

  /*add node to the end of the list*/
  if(transform_head){
    transform_tail->next = new;
    transform_tail = new;
  }
  else
    transform_head = transform_tail = new;
  if(!(transform_tail->mod_flag) || !(new->mod_flag))
    fprintf(stderr, "addr of new: %u\n", new);
}   

/**********************************************************************
 *
 * add_measures_transform_list - append measures to endtrans list
 *
 **********************************************************************/
void add_measures_transform_list(struct transform_list *ptr, struct freq_counts *freq)
{
  char buff[MAXSTRING];
  int count, phase;
  struct freq_counts *fptr;

  /*verify that the last transform node is indeed ENDTRANS*/
  if(ptr->mod_flag != ENDPHASE){
    fprintf(stderr, "ENDTRANS node expected in add_measures_transform_list\nNode is: %c\n", ptr->mod_flag);
    return;
  }
  else{

    sscanf(ptr->string, "%d", &phase);
    /*get the number of counts*/
    for(fptr=freq, count=0 ; fptr ; fptr=fptr->next, count++);
    sprintf(buff, "%d\nF%d", phase, count);

    /*list the indivisual counts*/
    for(fptr=freq ; fptr ; fptr=fptr->next)
      sprintf(buff, "%s %d", buff, fptr->count);
    sprintf(buff, "%s\n", buff);

    /*update the transform node*/
    ptr->string = (char *)myalloc_trans(strlen(buff)+1);
    strcpy(ptr->string, buff);
  }
}
#endif /*INTERACT */


/**********************************************************************
 *
 * begin_trans - send a start transformation message string to VET 
 *
 **********************************************************************/
void begin_trans(int trans_type)
{
#ifdef INTERACT  
  char buff[MAXSTRING];

  if (!swq || measure_flag)
    return;

  m_trans_pending=TRUE;
#endif
  /*
   * Check if viewer started.  If we're not supposed to be
   * sending messages, then the write functions will check.
   * But don't want to check here so that we catch errors
   * in VPO's VET code.
   */
#ifdef INTERACT
  if (viewing && !measure_phase)
#else
    if (viewing)
#endif
      {
    /*
     * Instruction Selection phase is called on the last procedure
     * which is not a function and therefore no initial set is sent
     * to VET.
     */
    if (initial_set_sent) {
      /* 
       * Make sure we are in a phase - otherwise we have an error
       * and don't want to send the transformation.
       */
      if (started_phase) {
	/* 
	 * Check that we're not already in a transformation
	 * otherwise we missed an end transformation.
	 */
	if (!started_trans) {
#ifdef INTERACT
	  type_trans = trans_type;
#endif
	  write_record_id(BEGINTRANS);
	  started_trans = TRUE;
	}
	else {
	  fprintf(stderr, 
		  "vpo (begin_trans): already in a transformation\n");
	  end_viewer();
	  exit(1);
	}
      }
      else {
	fprintf(stderr, "vpo (begin_trans): not in phase\n");
	end_viewer();
	exit(1);
      }
    } /* if initial_set_sent */
  } /* if viewing */

  /* Check the special case of zero transformations requested for this */
  /* function when we are using vpoiso. In this case, we don't want any */
  /* (optional) transformations from taking place and we want to exit out */
  /* of this optimization phase now. Only the else part of the if executes */
  /* right now since the maxtrans = 0 case checked in vpo.c-- it's here */
  /* just in case. */

  if (swtm && fn && maxtrans == 0 && !(started_phase & REQUIRED_PHASE)) {
    fprintf(stderr, "vpo: stopped optimization of %s:%s() after zero (%d) improving transformations\n", basename, fn, opttransnum); 
    moreopts = FALSE;
    longjmp(my_env, 1);
  } else {
    /* update our transformation counters for vpoiso */
    transnum += 1;
    if (trans_type == VET_OPT_TRANS) 
      opttransnum += 1;
  }
} /* end begin_trans */

/***********************************************************************
 *
 * end_trans - send a end transformation message string to VET 
 *
 ***********************************************************************/
void end_trans(void)
{
#ifdef INTERACT  
  char buff[MAXSTRING];

  if (!swq || measure_flag)
    return;

  if(save_trans && !m_trans_pending){
    sprintf(buff, "\n");
    
    /*store it in the list of transformations*/
    add_to_transform_list(ENDTRANS, buff);
  }
#endif
  
  /*
   * Check if viewer started.  If we're not supposed to be
   * sending messages, then the write functions check.
   * But don't want to check here so that we catch errors
   * in VPO.
   */
#ifdef INTERACT  
  if (viewing && !measure_phase) 
#else
  if (viewing)
#endif
    {
    /*
     * Instruction Selection phase is called on the last procedure
     * which is not a function and therefore no initial set is sent
     * to VET.
     */
    if (initial_set_sent) {
      /* Verify we have started a transformation previously. */
      if (started_trans) {
	write_record_id(ENDTRANS);
	started_trans = FALSE;
      }
      else {
	fprintf(stderr, "vpo (end_trans): ending a trans. when not in one\n");
	end_viewer();
	exit(1);
      }
    } /* if initial_set_sent */

  } /* if viewing */

  /* if we reached the desired transformation number, quit */
  if (swtm && fn && maxtrans == opttransnum && !(started_phase & REQUIRED_PHASE)) {
    fprintf(stderr, "vpo: stopped optimization of %s:%s() after %d improving (%d total) transformations\n", basename, fn, opttransnum, transnum); 
    moreopts = FALSE;
    longjmp(my_env, 1);
  }

#ifdef INTERACT
  /*check id sanity check after each transformation enabled*/
  if(transform_sanity_check && viewing && !measure_phase && !m_trans_pending)
    {
      sanity_check(top);
      write_record_id_to_socket(ENDTRANS);
      printf("Sanity check performed after trans %d\n", type_trans);
    }

  m_trans_pending = FALSE;
#endif
  
} /* end end_trans */

/***********************************************************************
 * 
 * begin_phase - send a begin phase message string to VET
 *                
 ***********************************************************************/
void begin_phase(int phase)
{
#ifdef INTERACT
  if (!swq || measure_flag) 
    return;

  m_phase_pending=TRUE;
#endif

  /* Keep track of current phase for purposes of error checking */
  started_phase = phase;
  
  /*
   * Check if viewer started.  If we're not supposed to be
   * sending messages, then the write functions check.
   * But don't want to check here so that we catch errors
   * in VPO.
   */
#ifdef INTERACT  
  if (viewing && !measure_phase) 
#else
  if(viewing)
#endif
    {
    /*
     * Instruction Selection phase is called on the last procedure
     * which is not a function and therefore no initial set is sent
     * to VET.
     */
    if (initial_set_sent) {
      
      write_record_id(BEGINPHASE);
#ifndef INTERACT
      write_num(phase);
#endif
    }
  }
} /* end begin_phase */

/***********************************************************************
 * 
 * end_phase - send a end phase message string to VET
 *                
 ***********************************************************************/
void end_phase(int phase)
{
#ifdef INTERACT  
  char buff[MAXSTRING];
  extern int curr_oper;

  if (!swq || measure_flag)
    return;

  /*check is sanity check after each phase enabled*/
  if(phase_sanity_check && viewing && !measure_phase && !m_phase_pending && phase<=NUM_TRANS_TYPES)
    {
      sanity_check(top);
      write_record_id_to_socket(ENDTRANS);
      printf("Sanity check performed after phase: %d\n", phase);
    }
  
  /*make a new transform node*/
  if(save_trans && !m_phase_pending){
    if(curr_oper>0 && viewing && !measure_phase)
      sprintf(buff, "%d\n", curr_oper);
    else
      sprintf(buff, "%d\n", phase);
    add_to_transform_list(ENDPHASE, buff);
  }
  m_phase_pending = FALSE;
#endif
  
  /*
   * Check if viewer started.  If we're not supposed to be
   * sending messages, then the write functions check.
   * But don't want to check here so that we catch errors
   * in VPO.
   */
#ifdef INTERACT
  if (viewing && !measure_phase) 
#else
  if(viewing) 
#endif
    {
    /*
     * Instruction Selection phase is called on the last procedure
     * which is not a function and therefore no initial set is sent
     * to VET.
     */
    if (initial_set_sent) {
      /* Check that we're ending the same phase we're currently in */
      if (started_phase != phase) {
	fprintf(stderr, "vpo (end_phase): current phase = %d | end phase = %d\n",
		started_phase, phase);
	end_viewer();
	exit(1);
      }
	    
      write_record_id(ENDPHASE);
	
#ifndef INTERACT
      if (sanity_enabled) {
	sanity_check(top);
      }
#endif

    }
  } /* if viewing */

  /* Clear out phase flag */
  started_phase = FALSE;
} /* end end_phase */

/***********************************************************************
 * 
 * new_bb - send a new basic block message to VET
 *           
 ***********************************************************************/
void new_bb(struct bblock *cblk)
{
#ifdef INTERACT  
  char buff[MAXSTRING];

  if (!swq || measure_flag)
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();
    sprintf(buff, "%d\n", cblk->block_id);
    add_to_transform_list(NEWBB, buff);
  }
#endif
  
  /*
   * new_bb is called to create the initial set, therefore we do
   * not want to send over new_bb records until after VET has the
   * initial set.
   */
#ifdef INTERACT  
  if (initial_set_sent && viewing && !measure_phase) 
#else
    if(initial_set_sent) 
#endif      
    {
    /* Verify transformation was started. */
    if (!started_trans) {
      fprintf(stderr, "vpo (new_bb): not in transformation\n");
      end_viewer();
      exit(1);
    }
    /* assume down, left, and right fields are nil */
    write_record_id(NEWBB);
#ifdef INTERACT
    write_bb_pointer(cblk);
#else
    write_pointer(cblk);
#endif
	
    /* Block label, if any */
    write_line(((cblk->label) ? cblk->label : "none"));
  }
} /* end new_bb */

/**********************************************************************
 *
 * update_bb_rtls - send over all rtls for this basic block
 *
 **********************************************************************/
void update_bb_rtls(struct bblock *cblk)
{
  struct list *rtlptr;

#ifdef INTERACT
  char buff[NEW_BIGSTRING];
  struct list *tptr;
  
  if (!swq || measure_flag)
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    /*store the block id*/
    sprintf(buff, "%d", cblk->block_id);

    /*store the new order of RTLs in this block*/
    for(rtlptr=cblk->rtls ; rtlptr ; rtlptr=rtlptr->next)
      sprintf(buff, "%s %d", buff, rtlptr->rtl_id);
    sprintf(buff, "%s -1\n", buff);
    
    add_to_transform_list(UPDATEBB, buff);

    /* save the updated rtl fields to old fields. */
    if (cblk->rtls)
      init_old_rtls(cblk);    
  }
#endif
  
  /* Don't want to send to VET until initial set is created and sent. */
#ifdef INTERACT  
  if (initial_set_sent && viewing && !measure_phase) 
#else
  if(initial_set_sent)
#endif      
    {
    /* Verify transformation was started. */
    if (!started_trans) {
      fprintf(stderr, "vpo (update_bb_rtls): not in transformation\n");
      end_viewer();
      exit(1);
    }
    write_record_id(UPDATEBB);
#ifdef INTERACT
    write_bb_pointer(cblk);
#else
    write_pointer(cblk);
#endif

    /* send over addresses of rtls for this basic block */
    rtlptr = cblk->rtls;
    while (rtlptr) {
      if (rtlptr->type & ~(ARGLINE | SIDEFFECT | COMBLINE |
			   FUNCNAME | CASELINE | DELAYLINE)) {

	write_record_id(VET_RTL);
#ifdef INTERACT
	write_rtl_pointer(rtlptr);
#else
	write_pointer(rtlptr);
#endif
      }
      rtlptr = rtlptr->next;
    }
    write_record_id(ENDUPDATE);
  } /* end initial_set_sent */
} /* end update_bb_rtls */

/*********************************************************************
 *
 * label_bb - send a label basic block message for VET
 *
 *********************************************************************/
#ifdef INTERACT
void label_bb(struct bblock *cblk, char *oldlabel, char *newlabel)
#else
void label_bb(struct bblock *cblk, char *label)
#endif
{
#ifdef INTERACT  
  char buff[MAXSTRING];

  if (!swq || measure_flag)
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    /*store the new label in the list*/
    if(!newlabel){
      fprintf(stderr, "New label expected in label_bb, block_id: %d\n",
	      cblk->block_id);
      return;
    }
    sprintf(buff, "%d %s\n", cblk->block_id, newlabel);
    add_to_transform_list(LABELBB, buff);
  }
#endif
  
  /* Don't want to send to VET until initial set is created and sent. */
#ifdef INTERACT
  if (initial_set_sent && viewing && !measure_phase) 
#else
    if(initial_set_sent)
#endif
    {
    /* Verify transformation was started. */
    if (!started_trans) {
      fprintf(stderr, "vpo (label_bb): not in transformation\n");
      end_viewer();
      exit(1);
    }
    write_record_id(LABELBB);
#ifdef INTERACT
    write_bb_pointer(cblk);
    write_line((newlabel ? newlabel : "none"));
#else
    write_pointer(cblk);
    write_line((label ? label : "none"));
#endif
  } /* if initial_set_sent */
} /* end label_bb */

/*********************************************************************
 *
 * del_bb - send a delete basic block message for VET
 *
 *********************************************************************/
void del_bb(struct bblock *cblk)
{
#ifdef INTERACT  
  char buff[MAXSTRING];

  if (!swq || measure_flag)
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    /*save the id of the block to be deleted*/
    sprintf(buff, "%d\n", cblk->block_id);
    add_to_transform_list(DELBB, buff);
  }
#endif
  
  /* Don't want to send to VET until initial set is created and sent. */
#ifdef INTERACT
  if (initial_set_sent && viewing && !measure_phase) 
#else
  if(initial_set_sent)
#endif      
    {
    /* Verify transformation was started. */
    if (!started_trans) {
      fprintf(stderr, "vpo (del_bb): not in transformation\n");
      end_viewer();
      exit(1);
    }
    write_record_id(DELBB);
#ifdef INTERACT
    write_bb_pointer(cblk);
#else
    write_pointer(cblk);
#endif
  } /* if initial_set_sent */
} /* end del_bb */

/***********************************************************************
 * 
 * mod_bb_ptrs - send a modify message to VET if a pointer has changed 
 *               for a basic block.
 *               
  ***********************************************************************/
void mod_bb_ptrs(char msgid, struct bblock *cblk, struct bblock *newptr)
{
#ifdef INTERACT
  char buff[MAXSTRING];
  extern char *ret;
  extern int real_change;
  
  if(msgid == MOD_DOWN && !delayed_viewing)
    real_change++;
  if (!swq || measure_flag)
    return;
  if(!cblk)
    return;
  
  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    /*save the current block id and id of the its changed left/right block*/
    if(newptr)
      sprintf(buff, "%d %d\n", cblk->block_id, newptr->block_id);
    else
      sprintf(buff, "%d 0\n", cblk->block_id);
    add_to_transform_list(msgid, buff);
  }
#endif
  
  /* Don't want to send to VET until initial set is created and sent. */
#ifdef INTERACT
  if (initial_set_sent && viewing && !measure_phase) 
#else
  if(initial_set_sent)
#endif      
    {
    /* Verify transformation was started. */
    if (!started_trans) {
      fprintf(stderr, "vpo (mod_bb_ptrs): not in transformation\n");
      end_viewer();
      exit(1);
    }
    write_record_id(msgid);
#ifdef INTERACT
    write_bb_pointer(cblk);
    if (msgid == MOD_LEFT && newptr && newptr->label &&
	strcmp(newptr->label, ret) == 0)
      write_bb_pointer(NULL);
    else
      write_bb_pointer(newptr);
#else
    write_pointer(cblk);
    write_pointer(newptr);
#endif
  }
}

/***********************************************************************
 * 
 * insert_rtl - send an new rtl message string to VET
 *             
 ***********************************************************************/
void insert_rtl(struct list *rtlptr)
{
#ifdef INTERACT
  char *a, regs[MAXSTRING], buff[MAXSTRING];
  int next_rtlid;
  extern int real_change;

  if(!delayed_viewing)
    real_change++;
  if (!swq || measure_flag)
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    /*save information realted to the new RTL*/
    if(rtlptr->next)
      next_rtlid = rtlptr->next->rtl_id;
    else
      next_rtlid = 0;
    
    if(!rtlptr->val)
      sprintf(buff, "%d %d %d %u  ", rtlptr->daddy->block_id,
	      next_rtlid, rtlptr->rtl_id, rtlptr->type);
    else
      sprintf(buff, "%d %d %d %u %s ", rtlptr->daddy->block_id,
	      next_rtlid, rtlptr->rtl_id, rtlptr->type, rtlptr->val);

    if(rtlptr->type & RTLINE){
      reget(rtlptr->deads, regs);
      if(!regs[0])
	sprintf(buff, "%snone", buff);
      else
	for(a=regs ; *a && *a!='\n' ; a++)
	  sprintf(buff, "%s%c", buff, *a);
      /*	for(a=regs ; *a ; a+=3)
		sprintf(buff, "%s%c%c", buff, *a, *(a+1));*/
    }
    sprintf(buff, "%s\n", buff);
    add_to_transform_list(NEWRTL, buff);

    /*save old information regarding RTL type and kind*/
    init_old_single_rtl(rtlptr);    
  }
#endif

#ifdef INTERACT  
  if (!delayed_viewing && viewing && !measure_phase) 
#else
  if(!delayed_viewing)
#endif      
    {
    if (ok_to_send("insert_rtl")) {
      write_record_id(NEWRTL);
#ifdef INTERACT
      write_rtl_pointer(rtlptr);
#else
      write_pointer(rtlptr);
#endif
      write_num(rtlptr->type & ~COMBLINE);
      if (rtlptr->type & (ASMLINE | COMMENT)) {
	write_line(rtlptr->val);
      }
      else {
	write_line(convert_rtl(rtlptr->val));
#ifdef INTERACT
	/*write the assembly value to the socket */
	if (rtlptr->type & RTLINE) {
	  write_line(trans_to_assem(rtlptr));
	}
#endif
      }
#ifdef INTERACT
      write_bb_pointer(rtlptr->daddy);
      write_rtl_pointer(rtlptr->next);
#else
      write_pointer(rtlptr->daddy);
      write_pointer(rtlptr->next);
#endif
    } /* if ok_to_send */
  } /* if delayed_viewing */
}
    
/**********************************************************************
 *
 * del_rtl - send a delete rtl message string for VET
 *
 **********************************************************************/
void del_rtl(struct list *rtlptr)
{
#ifdef INTERACT  
  char buff[MAXSTRING];
  extern int real_change;

  if(!delayed_viewing && cnts(rtlptr))
    real_change++;
  if (!swq || measure_flag)
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    /*save the current block id and id of rtl to be deleted*/
    sprintf(buff, "%d %d\n", rtlptr->daddy->block_id, rtlptr->rtl_id);
    add_to_transform_list(DELRTL, buff);
  }
#endif

#ifdef INTERACT  
  if (!delayed_viewing && viewing && !measure_phase) 
#else
  if(!delayed_viewing)
#endif
    {
    /* Don't want to send to VET until initial set is created and sent. */
    if (initial_set_sent) {
      /* Verify transformation was started. */
      if (!started_trans) {
	fprintf(stderr, "vpo (del_rtl): not in transformation\n");
	end_viewer();
	exit(1);
      }
      write_record_id(DELRTL);
#ifdef INTERACT
      write_rtl_pointer(rtlptr);
      write_bb_pointer(rtlptr->daddy);
#else
      write_pointer(rtlptr);
      write_pointer(rtlptr->daddy);
#endif
    }
  }
}

#ifdef INTERACT

/**********************************************************************
 *
 * move_rtls - =merge_bb - send a move rtls ptr message for VET
 *
 **********************************************************************/
void move_rtls(struct bblock *from, struct bblock *to,
	       struct list *start_rtl, struct list *end_rtl)
{
  char buff[MAXSTRING];
  int ifrom, ito, start, end;

  if (!swq || measure_flag)
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    /*save the start and end pointers of the rtls to be moved*/
    ifrom = (from ? from->block_id : 0);
    ito = (to ? to->block_id : 0);
    start = (start_rtl ? start_rtl->rtl_id : 0);
    end = (end_rtl ? end_rtl->rtl_id : 0);
    
    sprintf(buff, "%d %d %d %d\n", ifrom, ito, start, end);
    add_to_transform_list(MOVERTLS, buff);
  }
  
  /* send messages to the viewer, no way to test its correctness now */
  if (!delayed_viewing && viewing && !measure_phase) {
    /* Don't want to send to VET until initial set is created and sent.*/
    if (initial_set_sent) {
      /* Verify transformation was started. */
      if (!started_trans) {
	fprintf(stderr, "vpo (move_rtl): not in transformation\n");
	end_viewer();
	exit(1);
      }
      write_record_id(MOVERTLS);
      write_bb_pointer(from);
      write_bb_pointer(to);
    } /* end if initial_set_sent */
  } /* end if delayed_viewing */
}
#endif
/**********************************************************************
 *
 * move_rtl - send a move rtl ptr message for VET
 *
 **********************************************************************/
void move_rtl(struct list *rtlptr, struct bblock *oldblk, 
	      struct list *before, struct bblock *newblk)
{
#ifdef INTERACT  
  int rtl_pos;
  char buff[MAXSTRING];
  extern int real_change;

  if(!delayed_viewing)
    real_change++;
  if (!swq || measure_flag)
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    /*save the positions of the moved RTL*/
    rtl_pos = (before ? before->rtl_id : 0);
    sprintf(buff, "%d %d %d %d\n", oldblk->block_id, rtlptr->rtl_id,
	    newblk->block_id, rtl_pos);
    add_to_transform_list(MOVERTL, buff);
  }
#endif
  
#ifdef INTERACT
  if (!delayed_viewing && viewing && !measure_phase) 
#else
  if(!delayed_viewing)
#endif      
    {
    /* Don't want to send to VET until initial set is created and sent. */
    if (initial_set_sent) {
      /* Verify transformation was started. */
      if (!started_trans) {
	fprintf(stderr, "vpo (move_rtl): not in transformation\n");
	end_viewer();
	exit(1);
      }
      write_record_id(MOVERTL);
#ifdef INTERACT
      write_rtl_pointer(rtlptr);
      write_bb_pointer(oldblk);
      write_rtl_pointer(before);
      write_bb_pointer(newblk);
#else
      write_pointer(rtlptr);
      write_pointer(oldblk);
      write_pointer(before);
      write_pointer(newblk);
#endif
    } /* end if initial_set_sent */
  } /* end if delayed_viewing */
} /* end move_rtl */

/**********************************************************************
 *
 * mod_type - send a modify type  message string for VET
 *
 **********************************************************************/
void mod_type(struct list *rtlptr)
{
#ifdef INTERACT
  char buff[MAXSTRING];

  if (!swq || measure_flag)
    return;

  if (BASETYPE(rtlptr->type) == BASETYPE(rtlptr->old_type))
     return;
  
  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    /*save the new type of the rtl*/
    sprintf(buff, "%d %d %d\n", rtlptr->daddy->block_id, rtlptr->rtl_id,
	    rtlptr->type);
    add_to_transform_list(MOD_TYPE, buff);

    /*store the updated value to the old type field*/
    rtlptr->old_type = rtlptr->type;    
  }
#endif
  
  /* Don't want to send to VET until initial set is created and sent. */
#ifdef INTERACT
  if (initial_set_sent && viewing && !measure_phase)
#else
  if(!delayed_viewing)
#endif
    {

    /* Verify transformation was started. */
    if (!started_trans) {
      fprintf(stderr, "vpo (mod_type): not in transformation\n");
      end_viewer();
      exit(1);
    }
    write_record_id(MOD_TYPE);
#ifdef INTERACT
    write_rtl_pointer(rtlptr);
#else
    write_pointer(rtlptr);      
#endif
    write_num((rtlptr->type & ~COMBLINE));
#ifdef INTERACT
    write_bb_pointer(rtlptr->daddy);
#else
    write_pointer(rtlptr->daddy);      
#endif
  }
}
/**********************************************************************
 *
 * mod_kind - add a mod kind node to the undo list
 *
 **********************************************************************/
void mod_kind(struct list *rtlptr)
{
#ifdef INTERACT
  char buff[MAXSTRING];

  if (!swq || measure_flag)
    return;

  if (BASEKIND(rtlptr->kind) == BASEKIND(rtlptr->old_kind))
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    /*save the rtl id and the new kind*/
    sprintf(buff, "%d %d %d\n", rtlptr->daddy->block_id, rtlptr->rtl_id,
	    rtlptr->kind);
    add_to_transform_list(MOD_KIND, buff);

    /* copy the updated value to old field */
    rtlptr->old_kind = rtlptr->kind;
  }
#endif
  
  /* Don't want to send to VET until initial set is created and sent. */
#ifdef INTERACT
  if (initial_set_sent && viewing && !measure_phase)
#else
  if(!delayed_viewing)
#endif
    {

      /* Verify transformation was started. */
    if (!started_trans) {
      fprintf(stderr, "vpo (mod_kind): not in transformation\n");
      end_viewer();
      exit(1);
    }
    write_record_id(MOD_KIND);
#ifdef INTERACT
    write_rtl_pointer(rtlptr);
#else
    write_pointer(rtlptr);      
#endif
    write_num(rtlptr->kind);
#ifdef INTERACT
    write_bb_pointer(rtlptr->daddy);
#else
    write_pointer(rtlptr->daddy);      
#endif
  }
}


/**********************************************************************
 *
 * mod_rtl - send a modify rtl message string for VET
 *
 **********************************************************************/
#ifdef INTERACT
void mod_rtl(struct list *rtlptr, char *oldval, char *newval)
#else
void mod_rtl(struct list *rtlptr)
#endif   
{
#ifdef INTERACT  
  char buff[MAXSTRING];
  extern int real_change;

  if(!delayed_viewing)
    real_change++;
  if (!swq || measure_flag)
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    /*save the information of the modified RTL*/
    sprintf(buff, "%d %d %u %s\n", rtlptr->daddy->block_id, rtlptr->rtl_id,
	    BASETYPE(rtlptr->type), newval);
    add_to_transform_list(MOD_RTL, buff);

    /*copy the updated value to old afield*/
    rtlptr->old_type = rtlptr->type;
  }
#endif
  
  /* Don't want to send to VET until initial set is created and sent. */
#ifdef INTERACT
  if (initial_set_sent && viewing && !measure_phase) 
#else
  if(initial_set_sent)
#endif
    {
    /* Verify transformation was started. */
    if (!started_trans) {
      fprintf(stderr, "vpo (mod_rtl): not in transformation\n");
      end_viewer();
      exit(1);
    }
    write_record_id(MOD_RTL);
#ifdef INTERACT
    write_rtl_pointer(rtlptr);
#else
    write_pointer(rtlptr);
#endif
    write_num(rtlptr->type & ~COMBLINE);
    if (rtlptr->type & (ASMLINE | COMMENT)) {
#ifdef INTERACT
      write_line(newval);
#else
      write_line(rtlptr->val);
#endif      
    }
    else {
#ifdef INTERACT
      write_line(convert_rtl(newval));
      /*write the assembly value to the socket */
      if (rtlptr->type & RTLINE) {
	write_line(trans_to_assem(rtlptr));
      }
#else
      write_line(convert_rtl(rtlptr->val));
#endif
    }
#ifdef INTERACT
    write_bb_pointer(rtlptr->daddy);
#else
    write_pointer(rtlptr->daddy);
#endif
  }
}

/*********************************************************************
 *
 * mod_links - send a new list of links for this rtl to VET
 *
 *********************************************************************/
void mod_links(struct list *rtlptr) 
{

  int i;	/* index to links */

#ifdef INTERACT
  /* currently this function has been disabled*/
  return;
  
  if (!swq || measure_flag)
    return;

  /*make a new undo node */
  if (save_trans && !delayed_viewing) {
    extern int links_modified;

    write_pending_phase();

    /* no need to save old or new links since this information
       can be recalculated */
    links_modified = 1;
  }
#endif

  /* Don't want to send to VET until initial set is created and sent. */
#ifdef INTERACT
  if (viewing && initial_set_sent && !delayed_viewing && 0==1) 
#else
    if (initial_set_sent) 
#endif
    {
    /* Verify transformation was started. */
    if (!started_trans) {
      fprintf(stderr, "vpo (mod_links): not in transformation\n");
      end_viewer();
      exit(1);
    }
    write_record_id(MOD_RTLLINKS);

    /* Send the rtl address */
#ifdef INTERACT
    write_rtl_pointer(rtlptr);
#else
    write_pointer(rtlptr);
#endif
    write_num(rtlptr->type & ~COMBLINE);
    /* Send the links */
    if (rtlptr->lnks[0]) {
      for (i = 0; i < MAXLINKS && rtlptr->lnks[i]; i++) {
	/* Send over the address of the rtl */
#ifdef INTERACT
	write_rtl_pointer(rtlptr->lnks[i]);
#else
	write_pointer(rtlptr->lnks[i]);
#endif
      }
    } /* end if links */
    /* Signal end of links */
    write_num(0);
#ifdef INTERACT
    write_bb_pointer(rtlptr->daddy);
#else
    write_pointer(rtlptr->daddy);
#endif
  } /* end if sent initial_set */
}

/*********************************************************************
 *
 * mod_side - modifies a side effect of an rtl
 *
 *********************************************************************/
#ifdef INTERACT
void mod_side(struct list *rtlptr, char *oldse, char *newse)
#else
void mod_side(struct list *rtlptr)
#endif
{
  char *a;
  char reg[MAXSTRING];

#ifdef INTERACT
  char buff[MAXSTRING];
  if (!swq || measure_flag)
    return;

  /* get rid of unnecessary messages */
  if (!oldse && !newse)
    return;
  if (oldse && newse && strcmp(oldse, newse) == 0)
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    /*save the new side effects information*/
    if(newse)
      sprintf(buff, "%d %d %s\n", rtlptr->daddy->block_id, rtlptr->rtl_id,
	      newse);
    else
      sprintf(buff, "%d %d none\n", rtlptr->daddy->block_id, rtlptr->rtl_id);
    add_to_transform_list(MOD_SIDE, buff);
  }
#endif
  
  /* Don't want to send to VET until initial set is created and sent. */
#ifdef INTERACT
  if (initial_set_sent && viewing && !measure_phase) 
#else
  if(initial_set_sent)
#endif
    {
    /* Verify transformation was started. */
    if (!started_trans) {
      fprintf(stderr, "vpo (mod_side): not in transformation\n");
      end_viewer();
      exit(1);
    }
    write_record_id(MOD_SIDE);
#ifdef INTERACT
    write_rtl_pointer(rtlptr);
#else
    write_pointer(rtlptr);
#endif
    write_num(rtlptr->type & ~COMBLINE);
#ifdef INTERACT
    write_line(convert_rtl(newse));
    write_bb_pointer(rtlptr->daddy);
#else
    write_line(convert_rtl(rtlptr->se));
    write_pointer(rtlptr->daddy);
#endif
  }
}

/*********************************************************************
 *
 * mod_res - modifies a reserve line
 *
 *********************************************************************/
#ifdef INTERACT  
void mod_res(struct list *rtlptr, char *oldval, char *newval)
#else
void mod_res(struct list *rtlptr)
#endif
{
#ifdef INTERACT  
  char buff[MAXSTRING];

  if (!swq || measure_flag)
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    /*save the information of the new reserve line*/
    sprintf(buff, "%d %d %s\n", rtlptr->daddy->block_id, rtlptr->rtl_id, newval);
    add_to_transform_list(MOD_RES, buff);
  }
#endif
  
  /* Don't want to send to VET until initial set is created and sent. */
#ifdef INTERACT
  if (initial_set_sent && viewing && !measure_phase) 
#else
  if(initial_set_sent) 
#endif      
    {
    /* Verify transformation was started. */
    if (!started_trans) {
      fprintf(stderr, "vpo (mod_res): not in transformation\n");
      end_viewer();
      exit(1);
    }
    write_record_id(MOD_RES);
#ifdef INTERACT
    write_rtl_pointer(rtlptr);
#else
    write_pointer(rtlptr);
#endif
    write_num(rtlptr->type & ~COMBLINE);
#ifdef INTERACT
    write_line(convert_rtl(newval));
    write_bb_pointer(rtlptr->daddy);
#else
    write_line(convert_rtl(rtlptr->val));
    write_pointer(rtlptr->daddy);
#endif
  }
}

/*********************************************************************
 *
 * mod_deads - send a modify dead list message for VET
 *
 *********************************************************************/
#ifdef INTERACT
void mod_deads(struct list *rtlptr, rlist olddeads, rlist newdeads)
#else
void mod_deads(struct list *rtlptr)
#endif
{
  char *a; 
  char msg[MAXSTRING], reg[MAXSTRING];

#ifdef INTERACT
  char buff[MAXSTRING];
  extern int dead_list_change;

  if(!delayed_viewing)
    dead_list_change = TRUE;
  
  if (!swq || measure_flag)
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    /*save the new dead register information*/
    sprintf(buff, "%d %d ", rtlptr->daddy->block_id, rtlptr->rtl_id);
    reget(newdeads, reg);
    if(!reg[0])
      sprintf(buff, "%snone", buff);
    else
      for(a=reg ; *a && *a!='\n' ; a++)
	sprintf(buff, "%s%c", buff, *a);
      /*for(a=reg ; *a ; a+=3)
	sprintf(buff, "%s%c%c", buff, *a, *(a+1));*/
    sprintf(buff, "%s\n", buff);
    add_to_transform_list(MOD_DEADS, buff);
  }
#endif
  
  /* Don't want to send to VET until initial set is created and sent. */
#ifdef INTERACT
  if (initial_set_sent && viewing && !measure_phase) 
#else
  if(initial_set_sent)
#endif      
    {
    /* Verify transformation was started. */
    if (!started_trans) {
      fprintf(stderr, "vpo (mod_deads): not in transformation\n");
      end_viewer();
      exit(1);
    }
    write_record_id(MOD_DEADS);
#ifdef INTERACT
    write_rtl_pointer(rtlptr);
#else
    write_pointer(rtlptr);
#endif
    write_num(rtlptr->type & ~COMBLINE);
      
    /* Before vectorize_deads is called the dead register list resides 
     * in the side effect as a string of registers.  After 
     * vectorize_deads is called the dead register list resides as bit 
     * vectors.  All function calls for VET expect the dead 
     * register list to be bit vectors and is converted to a string 
     * before this point.
     /* concatenate each register in dead list to the message */
    if (DeadsareBitVectors) {
#ifdef INTERACT
      reget(newdeads, reg);
#else
      reget(rtlptr->deads,reg);
#endif
      msg[0] = '\0';
      for (a = reg; *a; a = skiploc(a)+1) {
	strcat(msg, convert_rtl(a));
      }
      write_line(msg);
    }
    else {
      write_line(convert_rtl(rtlptr->se));
    }
#ifdef INTERACT
    write_bb_pointer(rtlptr->daddy);
#else
    write_pointer(rtlptr->daddy);
#endif
  } /* end if initial_set_sent */
}

#ifdef INTERACT
/*********************************************************************
 *
 * update_locuse - indiactes that variable has been assigned 
 *
 *********************************************************************/
void update_locuse(struct locuse *loc)
{
  char buff[MAXSTRING];
  extern int locals_change;
  
  if(!delayed_viewing)
    locals_change = TRUE;
  
  if (!swq || measure_flag)
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    sprintf(buff, "%s %d %d %d %d %s\n", loc->item, loc->reg, loc->nouse,
	    loc->reps, loc->atype, loc->regin);
    add_to_transform_list(LOCAL_ALLOC, buff);
  }
  /*don't send this information to the viewer*/
}  

/*********************************************************************
 *
 * update_reg_mapping - change register mapping information
 *
 *********************************************************************/
void update_reg_mapping(int base, int index)
{
  char buff[MAXSTRING];
  
  if (!swq || measure_flag)
    return;

  /*make a new transform node*/
  if(save_trans && !delayed_viewing){
    write_pending_phase();

    sprintf(buff, "%d %d\n", base, index);
    add_to_transform_list(MAP_REGISTER, buff);
  }
  /*don't send this information to the viewer*/
}


/*********************************************************************
 *
 * send_phase_msgs - send dummy begin and end phase messages to the viewer to
 *                   indicate that this phase has been performed. Only needed
 *                   for the two required phases
 *
 *********************************************************************/
void send_phase_msgs(int phase)
{
  char buff[MAXSTRING];

  if (!swq || measure_flag)
    return;

  /*make new transform nodes*/
  if(save_trans && !delayed_viewing){
    sprintf(buff, "%d\n", phase);
    add_to_transform_list(BEGINPHASE, buff);
  
    add_to_transform_list(ENDPHASE, buff);
  }
  
  /* send the BEGIN PHASE message*/
  if (!delayed_viewing && viewing && !measure_phase){

    if(initial_set_sent){
      write_record_id_to_socket(BEGINPHASE);
      write_num(phase);

      write_record_id_to_socket(ENDPHASE);
    }
  }
}

#endif /*INTERACT*/

/*********************************************************************
 *
 * convert_rtl - display an encoded RTL value in a readable format
 *
 *********************************************************************/
char *convert_rtl(register char *rtl)
{
    register char c;
    static char newrtl[BIGSTRING];
    static char rtlptr[BIGSTRING];


    /* Check for NULL value -- want to send empty string to VET */
    if (!rtl) {
	return("");
    }
    rtlptr[0] = '\0';
    while (c = *rtl) {
	if (isrg(rtl)) {
	    if (regtovec(rtl) < 0 || pseudoreg(rtl))
		(void) sprintf(newrtl, "%c[%d]", regprt[regtype(rtl)], regnum(rtl));
	    else
		(void) sprintf(newrtl, "%c[%d]", regprt[regtype(rtl)], map(rtl));
	    rtl = skiploc(rtl);
        }
	else if (islocal(rtl)) {
	    (void) sprintf(newrtl, "%s", localname(rtl));
	    rtl = skiploc(rtl);
        }
	else if (isglob(rtl)) {
	    (void) sprintf(newrtl, "%s", globalname(rtl));
	    rtl = skiploc(rtl);
        }
	else {
	    (void) sprintf(newrtl, "%c", c);
	    rtl++;
	}
	strcat(rtlptr, newrtl);
    }
    
    return(rtlptr);
} /* end convert_rtl */

/**********************************************************************
 *
 * map_registers - determines what RTLs and deads were affected by
 *                 mapping a register and informs xvpodb
 *
 **********************************************************************/
void map_registers(char *regold)
{
   struct bblock *cblk;
   struct list *ptr;
   char *s;

#ifdef INTERACT
   if (!swq || measure_flag)
     return;
#endif

   for (cblk = top; cblk; cblk = cblk->down) {
       for (ptr = cblk->rtls; ptr; ptr = ptr->next) {
         if (s = ptr->val) {
	     while (s = gtrg(s))
		 if  (cmploc(unskiploc(s), regold)==0){
#ifdef INTERACT
	     mod_rtl(ptr, ptr->val, ptr->val);					 
#else
		     mod_rtl(ptr);
#endif
		     break;
		 }
	     if (s = ptr->se)
		 while (s = gtrg(s))
		     if  (cmploc(unskiploc(s), regold)==0){
#ifdef INTERACT
	       mod_side(ptr, ptr->se, ptr->se);				  
#else
			 mod_side(ptr);
#endif

			 break;
                     }
	     if (ptr->deads && rexists(regold, ptr->deads))
#ifdef INTERACT
	   mod_deads(ptr, ptr->deads, ptr->deads);			  
#else
		 mod_deads(ptr);
#endif
	}
     }
   }
} /* end map_registers */

/***********************************************************************
 *
 * ok_to_send - Checks to see if it is okay to send the transformation.
 *
 **********************************************************************/
int ok_to_send(char *transformation_type) 
{
    
    /* Don't want to send to VET until initial set is created and sent. */
    if (initial_set_sent 
#ifdef INTERACT
	  &&  !measure_phase
#endif
	    ) {
	/* Verify transformation was started. */
	if (!started_trans) {
	    fprintf(stderr, 
		    "vpo (%s): not in transformation\n", transformation_type);
	    end_viewer();
	    exit(1);
	}
	return(TRUE);
    }
    return(FALSE);

} /* end ok_to_send */

/*********************************************************************
 *
 * end_viewer - send end of vpo message to vet and close it
 *
 *********************************************************************/
void end_viewer(void)
{
#ifdef INTERACT
  char *p;
  int num, len;
  char msg[MAXSTRING];
  extern int glo_sockfd;
  if (!swq || measure_flag)
    return;
#endif

#ifdef INTERACT
  if (viewing && !measure_phase) 
#else
  if(viewing)
#endif
    {
    /* 
     * Reset flag so that we send message to VET that
     * VPO is done.
       */
    send_messages = TRUE;
      
    if (!quit_vpo)
      write_record_id(ENDVIEW);

#ifdef INTERACT
/*
    if (user_directed) {
*/
    len = read(glo_sockfd, msg, MAXSTRING);
    if (len ==0) {
      fprintf(stderr, "end_viewer (vpo) : No options received\n");
      exit(0);
    }
    if (len == -1) {
      fprintf(stderr, "end_viewer (vpo) : Read error\n");
      exit(0);
    }
    p = msg;
    get_num(&p, &num);
    if (num != STOP_TRANS) {
      fprintf(stderr,
	      "end_viewer (vpo) - should receive STOP_TRANS msg\n");
      exit(1);
    }
    system("sleep 2");
/*
    }
*/
#endif

    /* Close socket to VET */
    close();

    /* 
       * Reset flag since we disconnected VET do
       * not want to send anymore messages.
       */
    send_messages = TRUE;
  }
}

#ifdef INTERACT
/*
 * write_record_id - writes a record id to the socket
 */
void write_record_id(char id)
{
  static int trans_pending = FALSE;
  static int phase_pending = FALSE;
  static int changes = FALSE;
  static int transactions = FALSE;
  extern int curr_oper;

  if (!viewing)
    return;

/*
  write_record_id_to_socket(id);
  return;
*/

  switch (id) {
  case BEGINPHASE:
    phase_pending = TRUE;
    return;

  case BEGINTRANS:
    trans_pending = TRUE;
    return;
 
  case ENDTRANS:
    if (!changes) {
      transnum -= 1;
      if (type_trans == VET_OPT_TRANS) 
	opttransnum -= 1;
      trans_pending = FALSE;
      return;
    }
    changes = FALSE;
    break;
 
  case ENDPHASE:
    if (!transactions) {
      phase_pending = FALSE;
      return;
    }
    transactions = FALSE;
    break;
 
  case NEWBB:
  case UPDATEBB:
  case LABELBB:
  case DELBB:
  case MOD_LEFT:
  case MOD_RIGHT:
  case MOD_DOWN:
  case NEWRTL:
  case DELRTL:
  case MOD_RTL:
  case MOD_SIDE:
  case MOD_RES:
  case MOD_DEADS:
  case MOD_TYPE:
  case MOD_KIND:
  case MOVERTL:
  case MOVERTLS:
    if (phase_pending) {
      if(!measure_phase){
	write_record_id_to_socket(BEGINPHASE);
	/* To keep the phase numbers consistent during the analysis as well
	   as the optimization phases we send the actual optimization phase
	   number as opposed to the phase number sent by the analysis
	   phase. But during init transforms, curr_oper=0, we send the actual
	   phase number read from the trans file*/
	if(curr_oper>0 && viewing && !measure_phase)
	  write_num(curr_oper);
	else
	  write_num(started_phase);

	/* to record that this phase has caused some changes*/
	is_change = TRUE;
      }
      phase_pending = FALSE;
    }
    if (trans_pending) {
      if(!measure_phase){
	write_record_id_to_socket(BEGINTRANS);
	/*write_num(transnum);*/
      }
      trans_pending = FALSE;
    }
    if (started_trans) {
      changes = TRUE;
      transactions = TRUE;
    }
    break;
 
  case BEGINFUNCTION:
  case ENDFUNCTION:
  case SEND_INIT_SET:
  case ENDINITSET:
  case ENDVIEW:
  case ABORT_VPO:
  case OK_TO_END:
  case BBLEFT:
  case BBRIGHT:
  case VET_RTL:
  case VET_RTLLINKS:
  case VET_SIDE:
  case VET_DEADS:
  case NEWRES_MSG:
  case NEWSET_MSG:
  case ENDSEQ:
  case ENDUPDATE:
  case BEGINLOOPS:
  case ENDLOOPS:
  case BEGINLOOP:
  case ENDLOOP:
  case BEGIN_LIVEREG:
  case VALIDATE_RESULT:
  case BLOCK_INFO:
  case USER_INTERACT:
  case BEGINFREQ:
  case ENDFREQ:
  case CHANGE_TEST_CONFIG_ID:
  case MOD_RTLLINKS:
  case SANITY_CHECK:
  case ENDSANITY:
    break;
 
  default:
    fprintf(stderr, "write_record_id (vpo) - unexpected record id %c\n",
	    id);
    exit(1);
  }

  if(!measure_phase)
    write_record_id_to_socket(id);
}

/*
 * write_pending_phase - writes a phase id to the socket
 */
void write_pending_phase()
{
  char buff[MAXSTRING];
  extern int curr_oper;
  
   if(m_phase_pending && save_trans){

     /*make a new transform node*/
     /* To keep the phase numbers consistent during the analysis as well
	as the optimization phases we send the actual optimization phase
	number as opposed to the phase number sent by the analysis
	phase. But during init transforms, cirr_oper=0, we send the actual
	phase number read from the trans file*/
     if(curr_oper>0 && viewing && !measure_phase)
       sprintf(buff, "%d\n", curr_oper);
     else
       sprintf(buff, "%d\n", started_phase);
     add_to_transform_list(BEGINPHASE, buff);
     m_phase_pending = FALSE;
   }
     
   if(m_trans_pending && save_trans){

     /*make a new transform node*/
     sprintf(buff, " %d\n", transnum);
     
     /*store it in the list of transformations*/
     add_to_transform_list(BEGINTRANS, buff);
     m_trans_pending = FALSE;
   }

}
#endif


/*********************************************************************
 *
 * sanity_check - sends over basic blocks and rtls to compare vpo
 *		  against what vet has to verify the transformations
 *		  are occurring correctly.
 *
 *********************************************************************/
void sanity_check(struct bblock *firstbb)
{
    struct bblock *bblk;
    struct list *rtlptr;	/* pointer to rtl value */

    if(viewing){
      write_record_id(SANITY_CHECK);
      
      /* Check debugging same function. */
      write_record_id(BEGINFUNCTION);
      write_line(current_func);
      
      send_basicblocks(firstbb);
      
      write_record_id(ENDSANITY);
    }

} /* end sanity_check */


/*********************************************************************
 *
 * initnumtranslist - read in number of transformations for each function
 *
 *********************************************************************/
void initnumtranslist(char *cexfilename, char *filename)
{
   FILE *f;
   char *d, *s, *p;
   int status, numtrans;
   char tmpfilename[80], funcname[80];
   void addtonumtranslist();

#ifdef INTERACT   
   if (!swq || measure_flag)
     return;
#endif
   
   for (s = cexfilename; *s; s++)
     if (*s == '.')
       p = s;
   if (strcmp(p, ".cex")) {
     fprintf(stderr,
	     "initnumtranslist (vpo) .cex expected in %s, not %s\n", cexfilename, p);
     exit(1);
   }
   for (s = cexfilename, d = basename; *s && s != p; )
     *d++ = *s++;
   *d = '\0';
   f = fopen(filename, "r");
   if (f) {
#ifdef DISPOSE
     setbuf(f, transbuf);
#endif
     status = fscanf(f, "%s : %s : %d\n", tmpfilename, funcname, &numtrans);
     numfuncs = 0;
     while (status != EOF) {
       addtonumtranslist(tmpfilename, funcname, numtrans);
       status = fscanf(f, "%s : %s : %d\n", tmpfilename, funcname, &numtrans);
     }
     fclose(f);
   }
   else if (swtm) {
     fprintf(stderr, "initnumtranslist (vpo) %s file not there\n", filename);
     exit(1);
   }
}

/*********************************************************************
 *
 * addtonumtranslist - update number of transformations for a function
 *
 *********************************************************************/
void addtonumtranslist(char *filename, char *funcname, int numtrans)
{
   int i;

   for (i = 0; i < numfuncs; i++)
      if (strcmp(trans[i].funcname, funcname) == 0) {
         trans[i].numtrans = numtrans;
         return;
         }
   if (numfuncs >= MAXFUNCS) {
      fprintf(stderr, "addtonumtranslist (vpo) too many functions\n");
      exit(1);
      }
   strcpy(trans[numfuncs].basename, filename);
   strcpy(trans[numfuncs].funcname, funcname);
   trans[numfuncs].numtrans = numtrans;
   numfuncs++;

}

/*********************************************************************
 *
 * maxtransforfunc - returns maximum number of transformations allowed
 *                   for a function (-1 if not found)
 *
 *********************************************************************/
int maxtransforfunc(char *funcname)
{
   int i;

   for (i = 0; i < numfuncs; i++)
      if (strcmp(trans[i].funcname, funcname) == 0)
         return trans[i].numtrans;
   return -1;
}

/*********************************************************************
 *
 * printnumtranslist - print functions and number of transformations
 *                     to a file
 *
 *********************************************************************/
void printnumtranslist(char *filename)
{
   int i;
   FILE *f;

   f = fopen(filename, "w");
#ifdef DISPOSE
   setbuf(f, transbuf);
#endif
   for (i = 0; i < numfuncs; i++)
      fprintf(f, "%s : %s : %d\n", trans[i].basename, trans[i].funcname,
              trans[i].numtrans);
   fclose(f);
}

#ifdef INTERACT
/*
 * write_bb_pointer - write the id associated with the block to the UI
 */
void write_bb_pointer(struct bblock *cblk)
{
   char command[MAXSTRING];
   int block_id;

   if (!swq || !viewing || measure_flag)
     return;
   
   if (cblk)
     block_id = cblk->block_id;
   else
     block_id = 0;

   sprintf(command, "%d", block_id);
   write_to_socket(command);
} /* end write_bb_pointer */

/*
 * write_rtl_pointer - write the id associated with the rtl to the UI
 */
void write_rtl_pointer(struct list *rtlptr)
{
  char command[MAXSTRING];
  int loc_rtl_id;
   
  if (!swq || !viewing || measure_flag)
    return;

  if (rtlptr)
    loc_rtl_id = rtlptr->rtl_id;
  else
    loc_rtl_id = 0;
    
  sprintf(command, "%d", loc_rtl_id);
  write_to_socket(command);
} /* end write_pointer */

/*
 * write_loops - write the information about the loops to the UI
 */
void write_loops()
{
  struct loopl *lptr;
  extern struct loopl *loops;
   
  if (!swq || !viewing || measure_flag)
    return;

  write_record_id(BEGINLOOPS);

    /* for each loop do */
  for (lptr = loops; lptr; lptr = lptr->next) {
    struct blist *bptr;
    
    write_record_id(BEGINLOOP);

    /* first, write the nesting level of this loop */
    write_num(lptr->ptr->freq);    
    for (bptr = lptr->blocks; bptr; bptr = bptr ->next){

      /* if the nesting level of the bb is the same as that
	 in the current loop */
      if (bptr->ptr->freq == lptr->ptr->freq)
	write_bb_pointer(bptr->ptr);

      /*else if bptr is the header of another loop*/
      else if (bptr->ptr->freq == (lptr->ptr->freq+1) && bptr->ptr->loop)
	write_signed_num(-bptr->ptr->block_id);
    }
    write_record_id(ENDLOOP);   
  }
  write_record_id(ENDLOOPS);  
}

/*************************************************************************
Functions to support queries for live register information
***************************************************************************/

/*vec2string: convert a vector into a string*/
void vec2string(char *dst_string, vector live_regs)
{
   char reg[REGSTRING], regs[MAXSTRING];
   int items, index, scratch, nonscratch;

   regs[0] = '\0';
   items = availreg(RTREG);
   for (index = firstreg(RTREG); items--; index += skipreg(RTREG)) {
      if (vin(live_regs, index)) {
         makereg(reg, RTREG, indextoreg(RTREG, index));
         strcat(regs, reg);
      }
   }

   items = availreg(FTREG);
   for (index = firstreg(FTREG); items--; index += skipreg(FTREG)) {
      if (vin(live_regs, index)) {
         makereg(reg, FTREG, indextoreg(FTREG, index));
         strcat(regs, reg);
      }
   }

   strcpy(dst_string, convert_rtl(regs));
}

char *get_record_name(char);

enum {
  SendingInitialSet,
  ReceivingBlocksToOptimize,
  ReceivingTransformations,
} OptState = SendingInitialSet;

char *OptStateNames[] = {
  "SendingInitialSet",
  "ReceivingBlocksToOptimize",
  "ReceivingTransformations",
};

char *get_viewer_request_name(int oper);

#define BLOCKID 0
void validate_incoming_message(int oper) {
  int error = 0;
  int previousState = OptState;

  /*fprintf(stderr, "Got %s\n", get_viewer_request_name(oper));*/

  return;
  if (oper >= 0)
    oper = BLOCKID;

  switch(OptState) {
  default:
    fprintf(stderr, "Unknown reading protocol state: %d\n", OptState);
    exit(1);
    break;

  case SendingInitialSet:
    switch(oper) {
    default:
      error = 1;
      break;
    case BLOCKID:
      OptState = ReceivingBlocksToOptimize;
      break;
    case HAND_TRANS:
      break;
    }	
    break;

  case ReceivingBlocksToOptimize:
    switch(oper) {
    default:
      error = 1;
      break;
    case BLOCKID:
      OptState = ReceivingTransformations;
      break;
    }
    break;

  case ReceivingTransformations:
    switch(oper) {
    default:
      error = 1;
      break;
    case BLOCKID:
      break;
    case QUIT_TRANS:
      break;
    }
    break;
  }
   

  if (previousState != OptState)
      fprintf(stderr, "\nMoving to optimizer state %s\n", OptStateNames[OptState]);
    
  if (error) {
    fprintf(stderr, "\n\nProtocol error -- vpo in state %s, when viewer sent %s\n", OptStateNames[OptState], get_viewer_request_name(oper));
/*    end_viewer();*/
  }

}


enum {
  WaitViewerMode,
  WaitInitialSet,
  ReceivingFunction,
  ReceivingInitialBlock,
  ReceivingInitialPhase,
  ReceivingInitialTransformation,
  UserInteraction,
  ReceivingBlock,
  ReceivingPhase,
  ReceivingTransformation,
  ReceivingLoop,
  Closed,
} ViewerState = WaitViewerMode;

char *ViewerStateNames[] = {
  "WaitViewerMode",
  "WaitInitialSet",
  "ReceivingFunction",
  "ReceivingInitialBlock",
  "ReceivingInitialPhase",
  "ReceivingInitialTransformation",
  "UserInteraction", 
  "ReceivingBlock",
  "ReceivingPhase",
  "ReceivingTransformation",
  "ReceivingLoop",
  "Closed",
};

void validate_outgoing_message(char id) {
  int error = 0;
  int previousState = ViewerState;

  return;
  fprintf(stderr, "\nGot %s\n", get_record_name(id));
  switch(ViewerState) {
  default:
    fprintf(stderr, "Unknown protocol state: %d\n", ViewerState);
    exit(1);
    break;

  case WaitViewerMode:
    switch(id) {
    default:
      error = 1;
      break;
    case USER_INTERACT:
      ViewerState = WaitInitialSet;
      break;
    case BEGINFUNCTION:	
      ViewerState = ReceivingFunction;
      break;
    }
    break;
    
  case WaitInitialSet:
    switch(id) {
    default:
      error = 1;
      break;
    case BEGINFUNCTION:	
      ViewerState = ReceivingFunction;
      break;
    }
    break;

  case ReceivingFunction:
    switch(id) {
    default:
      error = 1;
      break;
    case NEWBB:	
      ViewerState = ReceivingInitialBlock;
      break;
    case ENDFUNCTION:	
      ViewerState = WaitInitialSet;
      break;
    case BEGINPHASE:	
      ViewerState = ReceivingInitialPhase;
      break;
    case ENDSEQ:
      ViewerState = UserInteraction;
      break;
    }
    break;

  case ReceivingInitialBlock:
    switch(id) {
    default:
      error = 1;
      break;
    case BBLEFT:	
    case BBRIGHT:
    case VET_RTL:
    case VET_DEADS:
    case NEWBB:
      break;
    case ENDINITSET:
      ViewerState = ReceivingFunction;
      break;
    }
    break;

  case ReceivingInitialPhase:
    switch(id) {
    default:
      error = 1;
      break;
    case BEGINTRANS:
      ViewerState = ReceivingInitialTransformation;
      break;
    case ENDPHASE:
      ViewerState = ReceivingFunction;
      break;
    }
    break;

  case ReceivingInitialTransformation:
    switch(id) {
    default:
      error = 1;
      break;
    case MOD_LEFT:
    case MOD_RIGHT:
    case MOD_DOWN:
    case NEWRTL:
    case DELRTL:
    case MOD_RTLLINKS:


    case NEWBB:
    case UPDATEBB:
    case LABELBB:
    case DELBB:
    case MOD_RTL:
    case MOD_SIDE:
    case MOD_RES:
    case MOD_DEADS:
    case MOD_TYPE:
    case MOD_KIND:
    case MOVERTL:
    case MOVERTLS:


	break;
    case ENDTRANS:
      ViewerState = ReceivingInitialPhase;
      break;
    }
    break;

  case UserInteraction:
    switch(id) {
    default:
      error = 1;
      break;
    case ENDFUNCTION:
      ViewerState = WaitInitialSet;
      break;
    case BEGINPHASE:	
      ViewerState = ReceivingPhase;
      break;
    case BEGINLOOPS:
      ViewerState = ReceivingLoop;
      break;
    }
    break;

  case ReceivingBlock:
    switch(id) {
    default:
      error = 1;
      break;
    case BBLEFT:	
    case BBRIGHT:
    case VET_RTL:
    case VET_DEADS:
    case NEWBB:
      break;
    case ENDINITSET:
      ViewerState = ReceivingFunction;
      break;
    }
    break;

  case ReceivingPhase:
    switch(id) {
    default:
      error = 1;
      break;
    case BEGINTRANS:
      ViewerState = ReceivingTransformation;
      break;
    case ENDPHASE:
      ViewerState = UserInteraction;
      break;
    }
    break;

  case ReceivingTransformation:
    switch(id) {
    default:
      error = 1;
      break;

    case NEWBB:
    case UPDATEBB:
    case LABELBB:
    case DELBB:
    case MOD_LEFT:
    case MOD_RIGHT:
    case MOD_DOWN:
    case NEWRTL:
    case DELRTL:
    case MOD_RTL:
    case MOD_RTLLINKS:
    case MOD_SIDE:
    case MOD_RES:
    case MOD_DEADS:
    case MOD_TYPE:
    case MOD_KIND:
    case MOVERTL:
    case MOVERTLS:
	break;

    case ENDTRANS:
      ViewerState = ReceivingPhase;
      break;
    }
    break;


  case ReceivingLoop:
    switch(id) {
    default:
      error = 1;
      break;
    case BEGINLOOP:
    case ENDLOOP:
      break;
    case ENDLOOPS:
      ViewerState = UserInteraction;
      break;
    }
    break;

  }


  if (id == ENDVIEW) {
    error = 0;
    ViewerState = Closed;
  }

/*
  if (previousState != ViewerState) {
    fprintf(stderr, "\nGot %s\n", get_record_name(id));
      fprintf(stderr, "\nMoving to viewer state %s\n", ViewerStateNames[ViewerState]);
  }
*/
  if (error) {
    fprintf(stderr, "\n\nProtocol error -- viewer in state %s, when vpo sent %s\n", ViewerStateNames[ViewerState], get_record_name(id));
/*
    end_viewer();
    exit(1);
*/
  }
}

char *get_viewer_request_name(int oper) {
  switch (oper) {
  case FREQUENCY_REQ:
    return "FREQUENCY_MEASUREMENT_REQ";
  case SANITY_CHECK_REQ:
    return "SANITY_CHECK_REQ";
  case LIVE_REGS_QUERY:
    return "LIVE_REGS_QUERY";
  case LOOPS_QUERY:
    return "LOOPS_QUERY";
  case UNDO_CHANGE:
    return "UNDO_CHANGE";
  case HAND_TRANS:
    return "HAND_TRANS";
  case INSERT_UNDO_STOP:
    return "INSERT_UNDO_STOP";
  case CMP_FILES_TRANS:
    return "CMP_FILES_TRANS";
  case WRITE_RTLS_TRANS:
    return "WRITE_RTLS_TRANS";
  case UNDO_TRANS:
    return "UNDO_TRANS";
  case IF_FALSE_GOTO_TRANS:
    return "IF_FALSE_GOTO_TRANS";
  case IF_LOOP_TRUE_GOTO_TRANS:
    return "IF_LOOP_TRUE_GOTO_TRANS";
  case QUIT_TRANS:
    return "QUIT_TRANS";
  case STOP_TRANS:
    return "STOP_TRANS";
  case BEGIN_LOOP:
    return "BEGIN_LOOP";
  case IF_LOOP_FALSE_GOTO_TRANS:
    return "IF_LOOP_FALSE_GOTO_TRANS";
  case END_LOOP:
    return "END_LOOP";
  case DE_OPTIMIZE_FUNC:
    return "De-optimize function";
  default:
    if (oper >= 0)
      return "BLOCK_ID";
    return "unknown request";
  }
}    


/* send the frequency counts to the viewer */
int send_counts(FILE *fp, int initial_counts, struct freq_counts **init_cnt,
		int flag)
{
  struct bblock *bb_ptr;
  struct freq_counts *freq_ptr ,*new_ptr;
  int i, count;

  bb_ptr = top;
  if(continuous_measures && !measure_phase && !measure_init && viewing && initial_counts>0)
    write_record_id(BEGINFREQ);

  /*dump all block counts*/
  for(i=0 ; i<initial_counts ; i++)
    {
      fscanf(fp, "%d", &count);
      if(flag) {
	new_ptr = (struct freq_counts*)CALLOC(sizeof(struct freq_counts));
	new_ptr->count = count;
	new_ptr->next = NULL;
	if(*init_cnt == NULL)
	  *init_cnt = new_ptr;
	else
	  freq_ptr->next = new_ptr;
	freq_ptr = new_ptr;
      }
      if(continuous_measures && !measure_phase && !measure_init && viewing){
	write_bb_pointer(bb_ptr);
	write_num(count);
      }
      bb_ptr = bb_ptr->down;
    }

  fscanf(fp, " \n");

  if(continuous_measures && !measure_phase && !measure_init && viewing && initial_counts>0)
    write_record_id(ENDFREQ);
  if(initial_counts)
    return 1;
  return -1;
}


/*
 * send_measures - read the freq_counts record and send over the measures to
 *                 the viewer
 */
void send_measures(struct freq_counts *freq_ptr, unsigned long long s_measure,
		   unsigned long long d_measure)
{
  struct freq_counts *cur_ptr;
  struct bblock *bb_ptr;

  cur_ptr = freq_ptr;
  bb_ptr = top;

  if(viewing){
    
    /*send the start frequency measures record*/
    write_record_id(BEGINFREQ);
    
    write_long(s_measure);
    write_long(d_measure);
    
    while(cur_ptr) {
      /* send the basic block id and the corresponding frequency measure*/
      write_bb_pointer(bb_ptr);
      write_num(cur_ptr->count);
      bb_ptr = bb_ptr->down;
      cur_ptr = cur_ptr->next;
    }
    
    /*send end frequency measure record*/
    write_record_id(ENDFREQ);
  }
}


/*
 * sel_comb_initial_msg - send messages to the viewer at the start of the
 *                        Sel..COmb process
 */
void sel_comb_initial_msg(int noComb, int weight, int length,
			  unsigned long long s_measure,
			  unsigned long long d_measure)
{
  extern int swq_noviewer;

  if(!swq_noviewer){
    
    /*send message to viewer to accept SEL...COMB messages*/
    write_to_socket("SEL_COMB");
   
    /*send the number of combinations to the viewer, 0 if indeterminate*/
    write_num(noComb);
    
    /*send the string length to the viewer*/
    write_num(length);
    
    /*send the initial counts to the viewer*/
    write_long(s_measure);
    write_long(d_measure);
    write_num(weight);
    write_char(ENDSEQ);
  }
}


/*
 * sel_comb_msg - send performance update to the user, of the execution of the
 *                current function during the iterative search algorithm
 */
void sel_comb_msg(struct comb_struct *comb, int seq[],
		  unsigned long long s_measure,
		  unsigned long long d_measure)
{
  char string[MAXSTRING];
  extern int swq_noviewer;

  if(!swq_noviewer){
    
    /*send the the current sequence number and number of valid comb*/
    write_num(comb->done_comb);
    write_num(comb->valid_comb);

    /*send the elements in the current sequence*/
    get_opt_sequence(string, comb->base_opt, seq, comb->strlength);
    write_to_socket(string);
    
    /*send the current measures to the viewer*/
    write_long(s_measure);
    write_long(d_measure);
    
    write_char(ENDSEQ);
  }
}


/*
 * htable_msg - send an update to the viewer during a genetic algoritm search
 *              when the sequence hits in the hashtable
 */
void htable_msg(struct comb_struct *comb, int seq[],
		unsigned long long s_measure,
		unsigned long long d_measure)
{
  char string[MAXSTRING];
  extern struct chromosome popu[MAX_CHROMOSOMES];
  extern FILE *comb_log;
  extern int swq_noviewer;

  /*assign the measures from the hash table to the chromosome*/
  popu[comb->popu_count-1].s_measure = s_measure;
  popu[comb->popu_count-1].d_measure = d_measure;
  
  fprintf(comb_log, "codesize: %llu, inscount: %llu\n", s_measure, d_measure);

  if(!swq_noviewer){
    
    /*send the the current sequence number and number of valid comb*/
    write_num(comb->done_comb);
    write_num(comb->valid_comb);
    
    /*send the elements in the current sequence*/
    get_opt_sequence(string, comb->base_opt, seq, comb->strlength);
    write_to_socket(string);
    
    /*send the current measures to the viewer*/
    write_long(s_measure);
    write_long(d_measure);
    
    write_char(ENDSEQ);
  }
}

/*
 * succ_htable_msg - send an update to the viewer during a genetic algoritm search
 *                   when the sequence hits in the hashtable of only succ phases
 */
void succ_htable_msg(struct comb_struct *comb, int seq[],
		     unsigned long long s_measure,
		     unsigned long long d_measure)
{
  char string[MAXSTRING];
  extern FILE *comb_log;
  extern int swq_noviewer;

  fprintf(comb_log, "codesize: %llu, inscount: %llu\n", s_measure, d_measure);

  if(!swq_noviewer){
    
    /*send the the current sequence number and number of valid comb*/
    write_num(comb->done_comb);
    write_num(comb->valid_comb);

    /*send the elements in the current sequence*/
    get_opt_sequence(string, comb->base_opt, seq, comb->strlength);
    write_to_socket(string);
    
    /*send the current measures to the viewer*/
    write_long(s_measure);
    write_long(d_measure);
    
    write_char(ENDSEQ);
  }
}


/*
 * get_opt_sequence - get the string represtation of the numerical
 *                    optimization sequence
 */
void get_opt_sequence(char string[], int base[], int seq[], int length)
{
  int i;

  for(i=0 ; i<length ; i++){
    switch(base[seq[i]]){
    case BRANCH_CHAINING_TRANS:
      string[i] = 'b';
      break;
    case ELIM_EMPTY_BLOCKS_TRANS:
      string[i] = 'e';
      break;
    case USELESS_JUMP_ELIM_TRANS:
      string[i] = 'u';
      break;
    case DEAD_CODE_ELIM_TRANS:
      string[i] = 'd';
      break;
    case REVERSE_BRANCHES_TRANS:
      string[i] = 'r';
      break;
    case BLOCK_REORDERING_TRANS:
      string[i] = 'i';
      break;
    case MERGE_BLOCKS_TRANS:
      string[i] = 'm';
      break;
    case INST_SELECT_TRANS:
      string[i] = 's';
      break;
    case FIX_CONTROL_FLOW_TRANS:
      string[i] = 'f';
      break;
    case EVAL_ORDER_DETER_TRANS:
      string[i] = 'o';
      break;
    case GLOBAL_INST_SELECT_TRANS:
      string[i] = 'g';
      break;
    case REG_ASSIGNMENT_TRANS:
      string[i] = 'a';
      break;
    case MINIMIZE_LOOP_JUMPS_TRANS:
      string[i] = 'j';
      break;
    case DEAD_ASG_ELIM_TRANS:
      string[i] = 'h';
      break;
    case REG_ALLOCATION_TRANS:
      string[i] = 'k';
      break;
    case COMMON_SUBEXPR_ELIM_TRANS:
      string[i] = 'c';
      break;
    case CODE_MOTION_TRANS:
      string[i] = 'n';
      break;
    case LOOP_STRENGTH_REDUCT_TRANS:
      string[i] = 'l';
      break;
    case RECURRENCE_ELIM_TRANS:
      string[i] = 'p';
      break;
    case INDUCT_VAR_ELIM_TRANS:
      string[i] = 'v';
      break;
    case STRENGTH_REDUCT_TRANS:
      string[i] = 'q';
      break;
    case FIX_ENTRY_EXIT_TRANS:
      string[i] = 'w';
      break;
    case INST_SCHED_TRANS:
      string[i] = 't';
      break;
    case FILL_DELAY_SLOTS_TRANS:
      string[i] = 'u';
      break;
    case ALL_LOOP_TRANS:
      string[i] = 'a';
      break;
    default:
      string[i] = 'x';
    }
  }
  string[length] = '\0';
}


/*
 * is_interrupt - check if the user has sent the interrupt message
 */
int is_interrupt()
{
  /*the set contains the descriptors for reading*/
  fd_set rset;
  struct timeval timeout;
  extern int glo_sockfd;
  char buff[MAXSTRING];
  int len;
  extern int swq_noviewer;

  if(!swq_noviewer){
    
    /*set the descriptor*/
    FD_ZERO(&rset);
    FD_SET(glo_sockfd, &rset);

    /*initialize the timeval structure to return immediately*/
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;

    /*check if there is an interrupt from the viewer*/
    if(select(glo_sockfd+1, &rset, NULL, NULL, &timeout)){
      /*read the data*/
      len = read(glo_sockfd, buff, MAXSTRING);
      buff[strlen(buff)-1] = '\0';
      
      if(len == 0){
	fprintf(stderr, "No data from socket\n");
	exit(0);
      }
      if (len == -1) {
	fprintf(stderr, "getcmds: Apparently the user interface was killed\n");
	exit(0);
      }
      
      /*check if data is as expected*/
      if(buff[0] == 'i')
	return 1;
      else
	return 0;
    }
  }

  return 0;
}

/*
 * change_test_config - create / change the test configuration script
 */
void change_test_config()
{
  int status;
  char file[MAXSTRING], cmd[MAXSTRING], line[MAXSTRING];
  FILE *fp;
  extern pathname[];

  /*open the configuration file if it is present*/
  sprintf(file, "%stest.config", pathname);
  if( (fp = fopen(file, "r")) != NULL){

    /*send the current values in the config file to the viewer*/
    if( (get_next_line(line, MAXSTRING, fp)) &&
	(sscanf(line, "execution type: %s\n", cmd))==1)
      write_line(cmd);
    else
      write_line("N/A");
    
    if( (get_next_line(line, MAXSTRING, fp)) &&
	(sscanf(line, "link command: %[^\n]\n", cmd))==1)
      write_line(cmd);
    else
      write_line("N/A");

    if( (get_next_line(line, MAXSTRING, fp)) &&
	(sscanf(line, "desired output file: %s\n", cmd))==1)
      write_line(cmd);
    else
      write_line("N/A");

    if( (get_next_line(line, MAXSTRING, fp)) &&
	(sscanf(line, "actual output file: %s\n", cmd))==1)
      write_line(cmd);
    else
      write_line("N/A");

    if( (get_next_line(line, MAXSTRING, fp)) &&
	(sscanf(line, "execute command: %[^\n]\n", cmd))==1)
      write_line(cmd);
    else
      write_line("N/A");

    if( (get_next_line(line, MAXSTRING, fp)) &&
	(sscanf(line, "maximum time: %s\n", cmd))==1)
      write_line(cmd);
    else
      write_line("N/A");

    /*if( (get_next_line(line, MAXSTRING, fp)) &&
	(sscanf(line, "count file name: %s\n", cmd))==1)
      write_line(cmd);
    else
      write_line("N/A");

    if( (get_next_line(line, MAXSTRING, fp)) &&
	(sscanf(line, "genreport path: %s\n", cmd))==1)
      write_line(cmd);
    else
    write_line("N/A");*/
    fclose(fp);
  }

  write_char(ENDSEQ);

  /*check if the user wants to update the configuration file*/
  read_line_from_socket(line);
  if((sscanf(line, "%d", &status)) == 1 && status==CHANGE_TEST_CONFIG){

    /*read remaining lines and create new config file*/
    if((fp = fopen(file, "w")) == NULL){
      fprintf(stderr, "Unable to create file: %s\n", file);
      status = 0;
    }

    read_line_from_socket(line);
    if(status)
      fprintf(fp, "%s", line);
    read_line_from_socket(line);
    if(status)
      fprintf(fp, "%s", line);
    read_line_from_socket(line);
    if(status)
      fprintf(fp, "%s", line);
    read_line_from_socket(line);
    if(status)
      fprintf(fp, "%s", line);
    read_line_from_socket(line);
    if(status)
      fprintf(fp, "%s", line);
    read_line_from_socket(line);
    if(status)
      fprintf(fp, "%s", line);
    fprintf(fp, "EASE file: a.cnt\n");
    /*read_line_from_socket(line);
    if(status)
      fprintf(fp, "%s", line);
    read_line_from_socket(line);
    if(status)
    fprintf(fp, "%s", line);*/

    /*read end of record marker*/
    read_line_from_socket(line);
    if((sscanf(line, "%d", &status)) == 1 && status==QUIT_TRANS){
      fclose(fp);
      fprintf(stderr, "File %s sucessfully updated\n", file);
    }
  }
}

int get_next_line(char line[], int length, FILE *fp){

  /*read next line in the file*/
  if(fgets(line, length, fp)!= NULL)
    return 1;
  else
    return 0;
}

/* set the function name in variable current_func. Required during the
   measurement phase */
void set_function_name(char *fn)
{
  current_func = fn;
}

#endif /* interact */

#endif /*VET*/
