/***********************************************************************
 * 
 * Copyright (c) 1994 Ann H. Bailey.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/view.h,v $
 * $Revision: 1.15 $
 ***********************************************************************/

#include <stdio.h>
#include <sys/types.h>
#ifdef WIN32
#include <winsock.h>
#else /* !WIN32 */
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif /* WIN32 */

/*
 * These are message identifiers from the compiler to the user interface.
 */
#define ENDPROPERPHASE  'W'
#define BEGINFUNCTION   'f'
#define ENDFUNCTION	'*'
#define BEGINPHASE	'P'
#define ENDPHASE	'='
#define BEGINTRANS	'T'
#define ENDTRANS	'@'
#define SEND_INIT_SET	'I'
#define ENDINITSET	'E'
#define ENDVIEW	        '!'
#define BEGINFREQ	'F'
#define ENDFREQ		'z'
#define ABORT_VPO	'Q'
#define OK_TO_END	'O'
#define SANITY_CHECK	'C'
#define ENDSANITY	'$'
#define NEWBB	        'B'
#define UPDATEBB        'U'
#define ENDUPDATE       '.'
#define LABELBB	        'L'
#define DELBB		'X'
#define BBLEFT	        '<'
#define BBRIGHT	        '>'
#define MOD_LEFT	'['
#define MOD_RIGHT	']'
#define MOD_DOWN	'v'
#define NEWRTL	        '+'
#define DELRTL          '-'
#define RTLNUM	        '#'
#define VET_RTL		'R'
#define VET_RTLLINKS	'l'
#define VET_SIDE	's'
#define VET_DEADS	'D'
#define NEWRES_MSG	'r'
#define NEWSET_MSG	'S'
#define MOD_RTL         't'
#define MOD_RTLLINKS	'~'
#define MOD_SIDE        'i'
#define MOD_RES         'e'
#define MOD_DEADS       'd'
#define MOVERTL         'm'
#define MOVERTLS        'M'
#define LOCAL_ALLOC     '1'
#define MAP_REGISTER    '2'
#define CHANGE_TEST_CONFIG_ID 'H'

#ifdef INTERACT
#define MOD_TYPE        'a'       /* modify the type of an RTL              */
#define MOD_KIND        'k'       /* modify the kind of an RTL              */

#define ENDSEQ          'q'       /* indicates the end of a seq of msgs from
                                     the compiler to the user interface     */
#define USER_INTERACT   'u'       /* user interactive mode                  */
#define ERR_HAND        '%'       /* start of error msg for hand change     */

#define BEGINLOOPS      '^'       /* begin loops to respond to a loop query */
#define ENDLOOPS        '&'       /* end loops to respond to a loop query   */
#define BEGINLOOP       '('       /* begin info for a loop                  */
#define ENDLOOP         ')'       /* end info for a loop                    */
#define BEGIN_LIVEREG   '|'       /* begin info for a live register query   */
#define END_LIVEREG     '\\'      /* end info for a live register query     */
#define VALIDATE_RESULT 'V'       /* validation result message              */
#define CMP_FILES       'c'       /* compare files result message           */
#define BLOCK_INFO      'b'       /* basic block info message               */

/*
 * These are message identifiers from the user interface to the compiler
 */
#define MOD_ASSEM       'A'       /* modify assem inst from user interface  */
#define NEW_ASSEM       'Z'       /* insert new assem inst from UI          */
#define END_ASSEM       ';'       /* end of mod_assem or new_assem message  */
#endif

/*
 * Phase names
 */
#define BRANCHCHAIN_PHASE     (1 << 0)
#define USELESSJUMPS_PHASE    (1 << 1)
#define DEADCODEELIM_PHASE    (1 << 2)
#define REORDERBLOCKS_PHASE   (1 << 3)
#define MERGEBLOCKS_PHASE     (1 << 4)
#define INSTSELECT_PHASE      (1 << 5)
#define EVALORDERDET_PHASE    (1 << 6)
#define REGASSIGN_PHASE       (1 << 7)
#define JUMPS_PHASE           (1 << 8)
#define DEADVARELIM_PHASE     (1 << 9)
#define OPTLOOPS_PHASE        (1 << 10)
#define REGALLOC_PHASE        (1 << 11)
#define COMMONSUBEXPR_PHASE   (1 << 12)
#define CHEAPERINST_PHASE     (1 << 13)
#define CODEMOTION_PHASE      (1 << 14)
#define RECURRENCES_PHASE     (1 << 15)
#define STRENGTHRED_PHASE     (1 << 16)
#define INDUCTVARELIM_PHASE   (1 << 17)
#define FIXENTRYEXIT_PHASE    (1 << 18)
#define INSTSCHED_PHASE       (1 << 19)
#define FILLSLOTS_PHASE       (1 << 20)
#define VECTORIZEDEADS_PHASE  (1 << 21)
#define SETLOCALLINKS_PHASE   (1 << 22)
#define SETGLOBALLINKS_PHASE  (1 << 23)
#define PREDICATION_PHASE     (1 << 24)
#define USELESSCONDS_PHASE    (1 << 25)
#define CROSSJUMPING_PHASE    (1 << 26)
#define PLACEMENT_PHASE       (1 << 27)

#ifdef INTERACT
#define REQUIRED_PHASE  (REG_ASSIGNMENT_BIT|STRENGTH_REDUCT_BIT|FIX_ENTRY_EXIT_BIT|BLOCK_REORDERING_BIT)
#else
#define REQUIRED_PHASE  (REGASSIGN_PHASE|FIXENTRYEXIT_PHASE)
#endif

/* Global Data Exported from view.c */
extern int started_phase;
extern int started_trans;
extern int DeadsareBitVectors;
extern int sanity_enabled;
extern int opttransnum;
extern int delayed_viewing;
extern char basename[];
extern char *fn;                
extern int swtm;
extern int maxtrans;            
extern jmp_buf my_env;
extern int measure_init;

/*
 * ANSI FUNCTION PROTOTYPES
 */
void prep_for_funcstart();
void startup_viewer(char *, int);
void send_func_name(char *);
void send_end_func(char *);
void send_init_set(char *);
void send_basicblocks(struct bblock *);
void send_rtls(struct list *);

void begin_trans(int trans_type);
#ifdef INTERACT
void get_opt_sequence(char string[], int base[], int seq[], int length);
void sel_comb_initial_msg(int noComb, int weight, int length,
			  unsigned long long s_measure,
			  unsigned long long d_measure);
void sel_comb_msg(struct comb_struct *comb, int seq[],
		  unsigned long long s_measure,
		  unsigned long long d_measure);
void succ_htable_msg(struct comb_struct *comb, int seq[],
		     unsigned long long s_measure, unsigned long long d_measure);
void set_function_name(char *fn);
void change_test_config();
void read_perf_num(unsigned long long *s_measure, unsigned long long *d_measure);
void send_measures(struct freq_counts *freq_ptr, unsigned long long s_measure,
		   unsigned long long d_measure);
#endif

/*
 * possible values for trans_type parameter:
 * VET_OPT_TRANS says this caller is inside an improving optimization phase.
 * VET_REQUIRED_TRANS says the caller is doing a required operation, which
 *                    may or may not result in a transformation change.
 */
#define VET_OPT_TRANS 0
#define VET_REQUIRED_TRANS 1

void end_trans(void);
void begin_phase(int);
void end_phase(int);
void new_bb(struct bblock *);
void update_bb_rtls(struct bblock *);
#ifdef INTERACT
void label_bb(struct bblock *, char *, char *);
void mod_rtl(struct list *, char *, char *);
void mod_side(struct list *, char *, char *);
void mod_res(struct list *, char *, char *);
void mod_deads(struct list *, rlist, rlist);
#else
void label_bb(struct bblock *, char *);
void mod_rtl(struct list *);
void mod_side(struct list *);
void mod_res(struct list *);
void mod_deads(struct list *);
#endif

void del_bb(struct bblock *);
void mod_bb_ptrs(char, struct bblock *, struct bblock *);
void insert_rtl(struct list *);
void del_rtl(struct list *);
void move_rtl(struct list *, struct bblock *, struct list *, struct bblock *);
void mod_links(struct list *);
char *convert_rtl(register char *);
void map_registers(char *);
int ok_to_send(char *);
void end_viewer(void);
void sanity_check(struct bblock *);
#ifdef INTERACT
void move_rtls(struct bblock *from, struct bblock *to,
	      struct list *start_rtl, struct list *end_rtl);
void send_phase_msgs(int phase);
#endif

/*-----External Data references-------------*/
#ifdef DEBUGVET
extern FILE *errlog;
#endif
extern int send_messages;
extern int initial_set_sent;
extern int viewing;
extern int quit_vpo;
extern int phase_sanity_check, transform_sanity_check;


#ifdef INTERACT
extern char transbuf[BUFSIZ];
extern int swq, user_directed, save_trans, measure_phase, measure_flag;
extern int continuous_measures;
extern int is_change;
extern void assert_syn(struct list*);

/*
 * Added some functions for INTERACT/UNDO use
 */

void update_locuse(struct locuse *loc);

#ifdef DEBUG
void print_undo_list();
#endif

/*
 * The following functions were defined to support displaying of
 * compiler-related information.
 */
void mod_type(struct list *);
void mod_kind(struct list *);


/*
 * utility used for debugging
 */
int find_bb_in_next(char *);

/*
 * writes the new id pointer for blocks and rtls
 */
void write_bb_pointer(struct bblock *);
void write_rtl_pointer(struct list *);
void write_pending_phase();
void write_loops();
#endif
