/***********************************************************************
 * 
 * Copyright (c) 1996 Ann H. Bailey.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/view_socket.c,v $
 * $Revision: 1.9 $
 ***********************************************************************/
#include <sys/types.h>
#ifdef WIN32
#include <winsock.h>
#else /* !WIN32 */
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/param.h>
#include <netdb.h>
#endif /* WIN32 */
#include <stdio.h>
#include "vpo.h"
#include "view.h"
#include "view_socket.h"

#ifdef DEBUG
extern FILE *errlog;
#endif

static int sockfd;
#ifdef INTERACT
int glo_sockfd;                   /* global socket file descriptor */
#endif

#ifdef INTERACT
int open_bound_tcp_socket (int *port)
{
   int                   s;
   struct sockaddr_in   addr;
   int                 len = sizeof(addr);

   if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0){
      perror(": can't get socket");
      _exit(1);
   }

   addr.sin_addr.s_addr = INADDR_ANY;
   addr.sin_family      = AF_INET;
   addr.sin_port        = htons((short)*port);
   
   if (bind(s, (struct sockaddr *) &addr, len) < 0){
      perror(": bind");
      close(s);
      _exit(1);
   }

   if (getsockname(s, (struct sockaddr *)&addr, &len) < 0){
      perror(": can't get name");
      close(s);
      _exit(1);
   }

   *port = addr.sin_port;
   return s;
}
#endif

void setup_client_socket(char *hostname, int portnum)
{

    /* Set global variable to socket file descriptor. */
#ifdef INTERACT
    sockfd = glo_sockfd;
#else
    sockfd = client_socket(hostname, portnum);
#endif
}

int client_socket(char *hostname, int port)
{
    int sock;
    struct sockaddr_in server;
    struct sockaddr *server_ptr = (struct sockaddr *) &server;
    struct hostent *hp;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
	perror("vpo (client_socket): opening stream socket");
	exit(1);
    }
    server.sin_family = AF_INET;
    hp = gethostbyname(hostname);
    if (hp == 0) {
	fprintf(stderr, "vpo (client_socket): %s: unknown host\n", hostname);
	exit(1);
    }
    memcpy(&server.sin_addr, hp->h_addr, hp->h_length);
    server.sin_port = htons(port);
    
    if (connect(sock, server_ptr, sizeof(server)) < 0) {
	perror("vpo (client_socket): connecting stream socket");
	exit(1);
    }
    return sock;
}

char *get_record_name(char);
int msg_is_id = 0;
void write_to_socket(char *command)
{
  FILE *f;
  
    /* Give VET an empty string for NULL value */
    if (!command) {
	command = "";
    }
    /*
     * Check that viewer is running and that user wants to see
     * messages from VPO.
     */
    /*if (viewing)  {*/
      if (write(sockfd, command, (strlen(command) + 1)) < 0) {
	perror("vpo (write_to_socket): ");
	exit(1);
	/*}*/
    
#if DEBUG
      fprintf(errlog, "view_socket(write_to_socket): %s\n", command);
#endif
	 }

} /* end write_to_socket */

#ifdef INTERACT
char *get_record_name(char c) {
  switch (c) {
  case 'f':
    return "BEGINFUNCTION";
  case '*':
    return "ENDFUNCTION";
  case 'P':
    return "BEGINPHASE";
  case '=':
    return "ENDPHASE";
  case 'T':
    return "BEGINTRANS";
  case '@':
    return "ENDTRANS";
  case 'I':
    return "SEND_INIT_SET";
  case 'E':
    return "ENDINITSET";
  case '!':
    return "ENDVIEW";
  case 'Q':
    return "ABORT_VPO";
  case 'O':
    return "OK_TO_END";
  case 'C':
    return "SANITY_CHECK";
  case '$':
    return "ENDSANITY";
  case 'B':
    return "NEWBB";
  case 'U':
    return "UPDATEBB";
  case '.':
    return "ENDUPDATE";
  case 'L':
    return "LABELBB";
  case 'X':
    return "DELBB";
  case '<':
    return "BBLEFT";
  case '>':
    return "BBRIGHT";
  case '[':
    return "MOD_LEFT";
  case ']':
    return "MOD_RIGHT";
  case 'v':
    return "MOD_DOWN";
  case '+':
    return "NEWRTL";
  case '-':
    return "DELRTL";
  case '#':
    return "RTLNUM";
  case 'R':
    return "VET_RTL";
  case 'l':
    return "VET_RTLLINKS";
  case 's':
    return "VET_SIDE";
  case 'D':
    return "VET_DEADS";
  case 'r':
    return "NEWRES_MSG";
  case 'S':
    return "NEWSET_MSG";
  case 't':
    return "MOD_RTL";
  case '~':
    return "MOD_RTLLINKS";
  case 'i':
    return "MOD_SIDE";
  case 'e':
    return "MOD_RES";
  case 'd':
    return "MOD_DEADS";
  case 'm':
    return "MOVERTL";
  case 'M':
    return "MOVERTLS";
  case 'a':
    return "MOD_TYPE";
  case 'k':
    return "MOD_KIND";
  case 'A':
    return "MOD_ASSEM";
  case 'Z':
    return "NEW_ASSEM";
  case ';':
    return "END_ASSEM";
  case 'u':
    return "UNDO_STOP";
  case 'q':
    return "ENDSEQ";
  case '%':
    return "ERR_HAND";
  case '^':
    return "BEGINLOOPS";
  case '&':
    return "ENDLOOPS";
  case '(':
    return "BEGINLOOP";
  case ')':
    return "ENDLOOP";
  case '|':
    return "BEGIN_LIVEREG";
  case '\\':
    return "END_LIVEREG";
  default:
    return "unknown record id";
  }
}
#endif

#ifdef INTERACT
void write_record_id_to_socket(char id)
#else
void write_record_id(char id)
#endif
{
    char command[MAXSTRING];

#ifdef INTERACT
void validate_outgoing_message(char);
validate_outgoing_message(id);
#endif
    sprintf(command, "%c", id);
    msg_is_id = 1;
    write_to_socket(command);
    msg_is_id = 0;

} /* end write_record_id */
#ifdef INTERACT
void write_pointer(int num)
{
    char command[MAXSTRING];
    
    sprintf(command, "%d", num);
    write_to_socket(command);

} /* end write_pointer */
#else
void write_pointer(void *numo)
{
    char command[MAXSTRING];
    int num; 
    
    num= (int) numo;

    sprintf(command, "%u", num);
    write_to_socket(command);

} /* end write_pointer */
#endif

void write_num(int num)
{
    char command[MAXSTRING];

    sprintf(command, "%u", num);
    write_to_socket(command);

} /* end write_num */

#ifdef INTERACT
void write_long(unsigned long long num)
{
    char command[MAXSTRING];

    sprintf(command, "%llu", num);
    write_to_socket(command);

} /* end write_long */

void write_float(float num)
{
    char command[MAXSTRING];

    sprintf(command, "%f", num);
    write_to_socket(command);

} /* end write_float */

/*added a function called write_char
 by cai, baosheng*/
void write_char(char c)
{
    char command[2];

    sprintf(command, "%c", c);
    write_to_socket(command);
} /* end write_char*/

void write_signed_num(int num)
{
    char command[MAXSTRING];

    sprintf(command, "%d", num);
    write_to_socket(command);

} /* end write_num */

#endif

void write_line(char *command)
{
    write_to_socket(command);

}

char *read_from_socket(void)
{
    static char command[MAXSTRING];
    char *cmd_ptr = command;
    int error_result;

    /* Initialize first character to EOS in case nothing is read in. */
    *cmd_ptr = '\0';

    /*
     * Read one character at a time from the socket until reach end 
     * of string.
     */
    do {
	error_result = read(sockfd, cmd_ptr, 1);
	/* Read will return -1 */
	if (error_result < 0) {
	    perror("vpo (read_from_socket): ");
	    exit(1);
	}
	/* Read will return 0 if EOF */
	else if (error_result == 0) {
	    fprintf(stderr, "vpo (read_from_socket): EOF reached\n");
	    exit(1);
	}
    }
    /* Check for end of string. */
    while (*cmd_ptr++);

#if DEBUG
    /* Print out what was read to log file */
    fprintf(errlog, "view_socket(read_from_socket): %s\n", command);
#endif
    /* 
     * Since "command" is declared as static, then the memory continues
     * to be allocated outside this function.  Therefore the value of
     * "command" can by copied to a new memory location before this
     * function is called again.
     */
    return command;

} /* read_from_socket */

char read_char(void)
{
    char *line;		/* String read from socket. */
    char id;		/* Character from string. */

    /* Read string from socket. */
    line = read_from_socket();

    /* Convert string to character. */
    sscanf(line, "%c", &id);

    /* Return character. */
    return id;

} /* end read_char */


int read_num(void)
{
    char *line;		/* String read from socket. */
    int num;		/* Number converted from string. */

    /* Read in string from socket. */
    line = read_from_socket();

    /* Converted string to number. */
    sscanf(line, "%d", &num);
    
    /* Return number. */
    return num;

} /* end read_num */

void read_line_from_socket(char line[])
{
    char *cmd_ptr = line;
    int error_result;

    /* Initialize first character to EOS in case nothing is read in. */
    *cmd_ptr = '\0';

    /*
     * Read one character at a time from the socket until reach end 
     * of string.
     */
    do {
	error_result = read(sockfd, cmd_ptr, 1);
	/* Read will return -1 */
	if (error_result < 0) {
	    perror("vpo (read_from_socket): ");
	    exit(1);
	}
	/* Read will return 0 if EOF */
	else if (error_result == 0) {
	    fprintf(stderr, "vpo (read_from_socket): EOF reached\n");
	    exit(1);
	}
	else if (*cmd_ptr == '\n'){
	  cmd_ptr++;
	  *cmd_ptr = '\0';
	  break;
	}
    }
    /* Check for end of string. */
    while (*cmd_ptr++);
} /* read_line_from_socket */

