/* ***********************************************************
 * A server framework of the vpo. It calls vpomain, which is the
 original main function of vpo.
 **********************************************************/
#ifdef INTERACT

#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <string.h>
#include <netdb.h>
#include "server_view.h"
#include "view_socket.h"

#define MAX_CONN 3
#define TRUE   1
#define FALSE  0

/* maximum connection number                      */ 
/* can be any number allowed by OS.               */

#define COMMAND_LEN 100000
#define VPO_PORT 10633


#define CHECK(CALL)\
 if ((CALL) < 0) {\
    fprintf (stderr, #CALL);\
    _exit (-1);\
 }
#define ERR_QUIT(MSG) {\
 fprintf(stderr,MSG);\
 exit (-1);\
 }

#define SA struct sockaddr
/*no use after debugging*/
int debugging_mode = 0;
/*code added to support non-interactive version concurrently*/
/*interactive optimization flag*/
extern int swq, swq_noviewer, user_directed, save_trans;
extern int glo_sockfd;
extern int phase_sanity_check, transform_sanity_check;
extern int continuous_measures;
int vpo_main(int art_num, char **arg);
/*ended non-interactive version*/

void cmd_srv(int connfd, char *options, int argc, char *filename, char *rtlfile, char *asmfile);

/*right now, the vpo accepts 2 arguments,
 1. number of clients. if it is 1, then the vpo will run in
 single user mode, which means that it can be debugged using gdb
 or ddd. Otherwise, it can accepts multiple clients, but can not be
 debugged
 2. port number*/
int main(int argc, char **argv)
{
   int listenfd, connfd;
   pid_t childpid;
   int clilen;
   struct sockaddr_in cliaddr, servaddr;
   int port;
   char options[100];
   char input_file[100];

      if (argc < 2){
         ERR_QUIT("main(): Usage: vpo.sparc -q input_file  portnumber [RTL file] [ASM file] (check this message)\n");
      }

   /* check if the compiler is to be set in a mode where it performs a sanity
      check after each phase or transformation*/
   if( (strncmp(argv[1], "-s", 2)==0) ){
     phase_sanity_check = TRUE;
     sprintf(options, "-%s", &argv[1][2]);
     strcpy(argv[1], options);
   }
   else if( (strncmp(argv[1], "-S", 2)==0) ){
     transform_sanity_check = TRUE;
     sprintf(options, "-%s", &argv[1][2]);
     strcpy(argv[1], options);
   }

   /* check if user wants to send the execution frequency measurements after
      eash phase*/
   if( (strncmp(argv[1], "-m", 2)==0) ){
     continuous_measures = TRUE;
     sprintf(options, "-%s", &argv[1][2]);
     strcpy(argv[1], options);
   }

   /* -q    : open viewer, allow interaction, save changes
      -q... : open viewer
      -Q    : open viewer, save changes
      -Q... : open viewer, save changes */
   
   /* First parse the arguments
      If either of -q or -Q is specified then open up the viewer*/
   if ( (strncmp(argv[1], "-q", 2)==0) || (strncmp(argv[1], "-Q", 2)==0) ) {
      swq = TRUE;

      /* if -Q option is specified then always save changes to a file*/
      if( strncmp(argv[1], "-Q", 2) == 0)
	save_trans = TRUE;
      
      /* if only -q is specified then allow the user to specify optimisations
	 from the user interface. Also enable the option of saving all
	 changes to a file*/
      if (strcmp(argv[1], "-q") == 0) {
         user_directed = TRUE;
	 save_trans = TRUE;
         strcpy(options, "-a");
      }
      /* if "q1" option is specified then this is an option for de-optimizing
	 the routine and we do not want to perform register assignment. We still
	 want to open the viewer in interactive mode and save changes*/
      else if (strcmp(argv[1], "-q1") == 0) {
         user_directed = TRUE;
	 save_trans = TRUE;
         strcpy(options, "-a1");
      }
      else if (strcmp(argv[1], "-qd")==0 || strcmp(argv[1], "-qD")==0){
	fprintf(stderr, "VISTA in debugging mode. No viewer is needed and no messages are sent to through the socket\n");
	user_directed = TRUE;
	save_trans = TRUE;
	strcpy(options, "-a");
	swq_noviewer = TRUE;
      }
      else
         sprintf(options, "-%s", &argv[1][2]);
      
      if (argc < 4){
         ERR_QUIT("main(): Usage: vpo.sparc -q input_file  portnumber [RTL file] [ASM file]\n");
      }
      else
         port = atoi(argv[3]);     
        
      /*no use after debugging-below*/
      if (port >7410)
         debugging_mode = 1;
      /*end debugging*/
      
      /*bind to the port*/
      listenfd = open_bound_tcp_socket(&port);
      
      CHECK (listen (listenfd, 5));
      /*Revised to fit this program*/
#ifndef MULTI_USER_MODE
      for (;;){
	 clilen = sizeof(cliaddr);
	 if(!swq_noviewer){
	   CHECK(connfd = accept(listenfd, (SA *) &cliaddr, (int *) &clilen));
	 }
	 else
	   connfd = 0;

	 if(argc == 4)
	   cmd_srv(connfd, options, argc, argv[2], NULL, NULL);
	 else if(argc == 5)
	   cmd_srv(connfd, options, argc, argv[2], argv[4], NULL);
	 else
	   cmd_srv(connfd, options, argc, argv[2], argv[4], argv[5]);

	 if(!swq_noviewer)
	   CHECK (close (connfd)); /* parent closes connected socket */
      }
#endif
      
      /*multiple user mode*/
#ifdef MULTI_USER_MODE
      
      for (;;){
	 clilen = sizeof(cliaddr);
	 CHECK(connfd = accept(listenfd, (SA *) &cliaddr, (int *) &clilen));
	 
	 if ((childpid=fork())==0){      /*in child process*/
	    CHECK (close (listenfd));
	    cmd_srv(connfd, options, argv[2]); /* process the request */
	    CHECK (close (connfd)); /* parent closes connected socket */
	    exit(0);
	 }
	 /*in parent process*/
	 CHECK (close (connfd)); /* parent closes connected socket */
      }/*end for*/          
#endif
   }/*end if interactive*/
   /*non-interactive mode*/
   else{
      swq = FALSE;
      vpo_main(argc, argv);
   }
   
   return(0);
}

/*testing cmd_srv function*/
void cmd_srv(int connfd, char *options, int argc, char *filename, char *rtlfile, char *asmfile)
{
   char *argv[5];
   
   /*read something from the socket, probably "init"*/
   /*in_len = read(connfd, command, COMMAND_LEN);
   fprintf(stderr, "The received char is: %s\n", command);
   */
   /*specify the arguments*/
   /*the -th and -tp has nothing to do with the actual connection*/
   /*   sprintf(argvs[1], "-th=xi");
   sprintf(argvs[2], "-tp=10633");
   */
   
   argv[1] = options;
   argv[2] = filename;
   argv[3] = rtlfile;
   argv[4] = asmfile;
   /*call the actual vpo*/
   glo_sockfd = connfd;
   vpo_main(argc-1, argv);

   return;
}

#endif
