
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <strings.h>
#include "build_data_structure.h"
#include "misc.h"


static cache_item_t* create_cache_item()
{
	cache_item_t* new_item;
	new_item=calloc(sizeof(cache_item_t),1);	
	assert(new_item);
	return new_item;
}


static part_item_t* create_part_item()
{
	part_item_t* new_item;
	new_item=calloc(sizeof(part_item_t),1);	
	assert(new_item);
	return new_item;
}


cache_item_t* create_cache_item_offchip()
{
	cache_item_t* new_item=create_cache_item();
	new_item->type=CIT_offchip;
	return new_item;
}

cache_item_t* create_cache_item_ports(int i)
{
	cache_item_t* new_item=create_cache_item();
	new_item->type=CIT_ports;
	new_item->t.rw_ports=i;
	return new_item;
}

part_item_t* create_part_item_offchip()
{
        part_item_t* new_item=create_part_item();
        new_item->type=PIT_offchip;
        return new_item;
}

part_item_t* create_part_item_ports(int i)
{
        part_item_t* new_item=create_part_item();
        new_item->type=PIT_ports;
        new_item->t.rw_ports=i;
        return new_item;
}


cache_item_t* create_cache_item_blocksize(int i)
{
	cache_item_t* new_item=create_cache_item();
	new_item->type=CIT_blocksize;
	new_item->t.blocksize=i;
	return new_item;

}
cache_item_t* create_cache_item_sets                (int i)
{
	cache_item_t* new_item=create_cache_item();
	new_item->type=CIT_sets;
	new_item->t.sets=i;
	return new_item;

}
cache_item_t* create_cache_item_replacement         (char* r)
{
	cache_item_t* new_item=create_cache_item();
	new_item->type=CIT_replacement;
	new_item->t.replacement=r;
	return new_item;

}
cache_item_t* create_cache_item_hit_latency         (int i)
{
	cache_item_t* new_item=create_cache_item();
	new_item->type=CIT_hit_latency;
	new_item->t.hit_latency=i;
	return new_item;

}
cache_item_t* create_cache_item_extra_miss_latency  (int i)
{
	cache_item_t* new_item=create_cache_item();
	new_item->type=CIT_extra_miss_latency;
	new_item->t.extra_miss_latency=i;
	return new_item;

}
cache_item_t* create_cache_item_assoc               (int i)
{
	cache_item_t* new_item=create_cache_item();
	new_item->type=CIT_assoc;
	new_item->t.assoc=i;
	return new_item;

}
cache_item_t* create_cache_item_parts_cached        (ident_list_t* p)
{
	cache_item_t* new_item=create_cache_item();
	new_item->type=CIT_parts_cached;
	new_item->t.parts_cached=p;
	return new_item;

}
cache_item_t* create_cache_item_misses_to           (char* m)
{
	cache_item_t* new_item=create_cache_item();
	new_item->type=CIT_misses_to;
	new_item->t.misses_to=m;
	return new_item;

}

cache_item_t* add_cache_item                       (cache_item_t* a, cache_item_t* b)
{
	a->next=b;
	return a;
}


part_item_t* create_part_item_default() 
{
	part_item_t* new_item=create_part_item();
	new_item->type=PIT_default;
	return new_item;

}
part_item_t* create_part_item_size                (int i) 
{
	part_item_t* new_item=create_part_item();
	new_item->type=PIT_size;
	new_item->t.size=i;
	return new_item;

}

part_item_t* create_part_item_latency             (int i)
{
	part_item_t* new_item=create_part_item();
	new_item->type=PIT_latency;
	new_item->t.latency=i;
	return new_item;

}


part_item_t* add_part_item                        (part_item_t* a, part_item_t* b)
{
	a->next=b;
	return a;
}

cache_list_t* add_cache(cache_list_t* a, cache_list_t* b)

{
	a->next=b;
	return a;
}



part_list_t* add_memory_partition(part_list_t* a,part_list_t* b)
{
	a->next=b;
	return a;
}




ident_list_t* add_identifier_list    (char* new_ident, ident_list_t* list)
{
	ident_list_t* new_node=calloc(sizeof(ident_list_t),1);	
	assert(new_node);
	new_node->next=list;
	new_node->name=(char*)strdup(new_ident);
	assert(new_node->name);

	return new_node;
}


part_list_t* create_partition(char* name, part_item_t* items)
{
	part_list_t* new_part=calloc(sizeof(part_list_t),1);
	part_item_t* first;
	int found;
	assert(new_part);
	
	new_part->name=(char*)strdup(name);
	assert(new_part->name);


	if(strcmp(name,"instructions")==0)
		new_part->instructions=TRUE;
	else
		new_part->instructions=FALSE;

        new_part->default_item=FALSE;
        for(first=items,found=FALSE;first;first=first->next)
        {
                if(first->type==PIT_default && !found)
                {
                        new_part->default_item=TRUE;
			found=TRUE;
                }
                else if(first->type==PIT_latency && found)
                {
                        fatal("Multiple default markers for part %s", name);
                }
        }
	if(new_part->instructions && new_part->default_item)
		fatal("instruction partition is marked default\n");


	new_part->size=0;
	for(first=items,found=FALSE;first;first=first->next)
	{
		if(first->type==PIT_size && !found)
		{
#if 0
/* we allow a maximum size, only for power calculations */
			if(new_part->default_item)
				fatal("default part cannot have a maximum size");
#endif

			new_part->size=first->t.size;
			found=TRUE;
		}
		else if(first->type==PIT_size && found)
		{
			fatal("Multiple size specifications for part %s", name); 
		}
	}
	if(!found && !(new_part->default_item||new_part->instructions))
		fatal("No size specification for part %s", name); 


        new_part->latency=0;
        for(first=items,found=FALSE;first;first=first->next)
        {
                if(first->type==PIT_latency && !found)
                {
                        new_part->latency=first->t.latency;
			found=TRUE;
                }
                else if(first->type==PIT_latency && found)
                {
                        fatal("Multiple latency specifications for part %s", name);
                }
        }
        if(!found)
                fatal("No latency specification for part %s", name);


        new_part->rw_ports=-1;
	found=FALSE;
        for(first=items,found=FALSE;first;first=first->next)
        {
                if(first->type==PIT_ports && !found)
                {
                        new_part->rw_ports=first->t.rw_ports;
                        found=TRUE;
                }
                else if(first->type==PIT_ports && found)
                        fatal("Multiple ports specifications for cache %s", name);
        }
        if(!found)
                fatal("No ports specification for cache %s", name);

        new_part->offchip=FALSE;
        for(first=items,found=FALSE;first;first=first->next)
        {
                if(first->type==PIT_offchip && !found)
                {
                        new_part->offchip=TRUE;
                }
                else if(first->type==PIT_offchip && found)
                        fatal("Multiple offchip specifications for cache %s", name);
        }


	

	return new_part;
}

cache_list_t* create_cache(char* name, cache_item_t* items)
{
	cache_item_t* first;
	int found;
	cache_list_t* new_cache=calloc(sizeof(cache_list_t),1);
	assert(new_cache);

	new_cache->name=(char*)strdup(name);
	assert(new_cache->name);

        new_cache->sets=0;
        for(first=items,found=FALSE;first;first=first->next)
        {
                if(first->type==CIT_sets && !found)
                {
                        new_cache->sets=first->t.sets;
                        found=TRUE;
                }
                else if(first->type==CIT_sets && found)
                        fatal("Multiple sets specifications for cache %s", name);
        }
        if(!found)
                fatal("No sets specification for cache %s", name);

        new_cache->blocksize=0;
        for(first=items,found=FALSE;first;first=first->next)
        {
                if(first->type==CIT_blocksize && !found)
                {
                        new_cache->blocksize=first->t.blocksize;
                        found=TRUE;
                }
                else if(first->type==CIT_blocksize && found)
                        fatal("Multiple blocksize specifications for cache %s", name);
        }
        if(!found)
                fatal("No blocksize specification for cache %s", name);

        new_cache->assoc=0;
        for(first=items,found=FALSE;first;first=first->next)
        {
                if(first->type==CIT_assoc && !found)
                {
                        new_cache->assoc=first->t.assoc;
                        found=TRUE;
                }
                else if(first->type==CIT_assoc && found)
                        fatal("Multiple assoc specifications for cache %s", name);
        }
        if(!found)
                fatal("No assoc specification for cache %s", name);

        new_cache->replacement=NULL;
        for(first=items,found=FALSE;first;first=first->next)
        {
                if(first->type==CIT_replacement && !found)
                {
                        new_cache->replacement=first->t.replacement;
                        found=TRUE;
                }
                else if(first->type==CIT_replacement && found)
                        fatal("Multiple replacement specifications for cache %s", name);
        }
        if(!found)
                fatal("No replacement specification for cache %s", name);

        new_cache->hit_latency=0;
        for(first=items,found=FALSE;first;first=first->next)
        {
                if(first->type==CIT_hit_latency && !found)
                {
                        new_cache->hit_latency=first->t.hit_latency;
                        found=TRUE;
                }
                else if(first->type==CIT_hit_latency && found)
                        fatal("Multiple hit_latency specifications for cache %s", name);
        }
        if(!found)
                fatal("No hit_latency specification for cache %s", name);

        new_cache->extra_miss_latency=0;
        for(first=items,found=FALSE;first;first=first->next)
        {
                if(first->type==CIT_extra_miss_latency && !found)
                {
                        new_cache->extra_miss_latency=first->t.extra_miss_latency;
                        found=TRUE;
                }
                else if(first->type==CIT_extra_miss_latency && found)
                        fatal("Multiple extra_miss_latency specifications for cache %s", name);
        }
        if(!found)
                fatal("No extra_miss_latency specification for cache %s", name);


        new_cache->parts_cached=NULL;
        for(first=items,found=FALSE;first;first=first->next)
        {
                if(first->type==CIT_parts_cached && !found)
                {
                        new_cache->parts_cached=first->t.parts_cached;
                        found=TRUE;
                }
                else if(first->type==CIT_parts_cached && found)
                        fatal("Multiple parts_cached specifications for cache %s", name);
        }
        if(!found)
                fatal("No parts_cached specification for cache %s", name);

        new_cache->misses_to=NULL;
        for(first=items,found=FALSE;first;first=first->next)
        {
                if(first->type==CIT_misses_to && !found)
                {
                        new_cache->misses_to=first->t.misses_to;
                        found=TRUE;
                }
                else if(first->type==CIT_misses_to && found)
                        fatal("Multiple misses_serviced_by specifications for cache %s", name);
        }

        new_cache->rw_ports=-1;
	found=FALSE;
        for(first=items,found=FALSE;first;first=first->next)
        {
                if(first->type==CIT_ports && !found)
                {
                        new_cache->rw_ports=first->t.rw_ports;
                        found=TRUE;
                }
                else if(first->type==CIT_ports && found)
                        fatal("Multiple ports specifications for cache %s", name);
        }
        if(!found)
                fatal("No ports specification for cache %s", name);

	new_cache->offchip=FALSE;
        for(first=items,found=FALSE;first;first=first->next)
        {
                if(first->type==CIT_offchip && !found)
                {
                        new_cache->offchip=TRUE;
                        found=TRUE;
                }
                else if(first->type==CIT_offchip && found)
                        fatal("Multiple offchip specifications for cache %s", name);
        }

	return new_cache;
}



partition_description_t pd;
void final_check(part_list_t* parts, cache_list_t* caches)
{
	cache_list_t *all_caches, *caches2;

	/* make each cache have an array of patitions that're cached */
	for(all_caches=caches; all_caches; all_caches=all_caches->next)
	{
		ident_list_t *idents;
		int count=0;
		for(idents=all_caches->parts_cached; idents; idents=idents->next)
			count++;
		all_caches->the_parts_cached=calloc(sizeof(part_list_t*),count);
		all_caches->num_parts_cached=count;
		count=0;
		for(idents=all_caches->parts_cached; idents; idents=idents->next)
		{
			part_list_t* all_parts;;
			for(all_parts=parts; all_parts; all_parts=all_parts->next)
			{
				if(strcmp(all_parts->name,idents->name)==0)
				{
					if(all_caches->the_parts_cached[count]!=NULL)			
						fatal("Part %s is cached multiple times by cache %s\n",
							idents->name, all_caches->name);
					all_caches->the_parts_cached[count]=all_parts;
				}
			}
			count++;
		}
		
	
	}
	for(all_caches=caches; all_caches; all_caches=all_caches->next)
	{
		if(all_caches->misses_to==NULL)
			continue;
		for(caches2=caches; caches2; caches2=caches2->next)
		{
			if(strcmp(all_caches->misses_to,caches2->name)==0)
			{
				if(all_caches->misses_to_cache)
					fatal("Two (or more) caches named %s", all_caches->misses_to_cache);
				all_caches->misses_to_cache=caches2;
				
			}
		}
	}
	

	/* set the global partition description before yyparse exists */
	pd.parts=parts;
	pd.caches=caches;



}
