#ifndef BUILD_DS_H
#define BUILD_DS_H

typedef struct ident_list ident_list_t;
typedef enum cache_item_type cache_item_type_t;
typedef enum part_item_type part_item_type_t;
typedef struct cache_item cache_item_t;
typedef struct part_item part_item_t;
typedef struct cache_list cache_list_t;
typedef struct part_list part_list_t;

struct ident_list
{
	char* name;
	ident_list_t* next;
};



enum cache_item_type { 
	CIT_blocksize, 
	CIT_sets, 
	CIT_replacement, 
	CIT_hit_latency, 
	CIT_extra_miss_latency, 
	CIT_assoc, 
	CIT_parts_cached, 
	CIT_misses_to,
	CIT_ports,
	CIT_offchip
	 };

enum part_item_type { 
	PIT_default, 
	PIT_latency,
	PIT_size, 
	PIT_ports,
	PIT_offchip
};

struct cache_item
{
	cache_item_type_t type;

	union
	{
		int sets;
		int blocksize;
		int assoc;
		char* replacement;
		int hit_latency;
		int extra_miss_latency;
		ident_list_t* parts_cached;
		char*	misses_to;
		int rw_ports;
		
	} t;

	cache_item_t* next;
};

struct part_item
{
	part_item_type_t type;
	union
	{
		int size;
		int latency;
		int rw_ports;
	} t;
	part_item_t* next;

};


struct cache_list
{
	char* name;
        int sets;
        int blocksize;
        int assoc;
        char* replacement;
        int hit_latency;
        int extra_miss_latency;
        ident_list_t* parts_cached;

	/* malloced array of parts cached */
	int num_parts_cached;
	part_list_t** the_parts_cached;

        char*   misses_to;
	cache_list_t* misses_to_cache;

	int rw_ports;	/* how many rread/write ports to the cache */
	int offchip;	/* is this cache on the chip, and therefore counts as a power consumer */

	void* user_data;
	cache_list_t* next;
};

struct part_list
{
	char* name;
        int size;
        int latency;
	int default_item;	/* true if this part is the default */
	int instructions;	/* boolean, true if this is the instructions partition */

	int rw_ports;	/* how many rread/write ports to the cache */
	int offchip;	/* is this cache on the chip, and therefore counts as a power consumer */

	void* user_data;
        part_list_t* next;
};

typedef struct partition_description partition_description_t;
struct partition_description
{
        part_list_t* parts;
        cache_list_t* caches;
};

extern partition_description_t pd;


cache_item_t* create_cache_item_blocksize(int i);
cache_item_t* create_cache_item_sets                (int i);
cache_item_t* create_cache_item_replacement         (char* r);
cache_item_t* create_cache_item_hit_latency         (int i);
cache_item_t* create_cache_item_extra_miss_latency  (int i);
cache_item_t* create_cache_item_assoc               (int i);
cache_item_t* create_cache_item_parts_cached        (ident_list_t* p);
cache_item_t* create_cache_item_misses_to           (char* m);
cache_item_t* create_cache_item_offchip             ();
cache_item_t* create_cache_item_ports               (int i);
cache_item_t* add_cache_item                        (cache_item_t* a, cache_item_t* b);
part_item_t* create_part_item_default		    () ;
part_item_t* create_part_item_size                  (int i);
part_item_t* create_part_item_latency               (int i);
part_item_t* create_part_item_offchip             	();
part_item_t* create_part_item_ports               	(int i);

part_item_t* add_part_item(part_item_t* a, part_item_t* b);
cache_list_t* add_cache(cache_list_t* a, cache_list_t* b);
cache_list_t* create_cache(char* name, cache_item_t* items);
part_list_t* create_partition(char* name, part_item_t* items);
part_list_t* add_memory_partition(part_list_t* a,part_list_t* b);
ident_list_t* add_identifier_list    (char* new_ident, ident_list_t* list);

void final_check(part_list_t* parts, cache_list_t* caches);

#endif
