

%{
#include <stdio.h>
#include <strings.h>

#include "build_data_structure.h"

extern char *pdtooltext;
%}

%union {
	int constant;
	char* char_string;
	cache_list_t* cl;
	cache_item_t* ci;
	part_list_t* pl;
	part_item_t* pi;
	ident_list_t* il;
}

%token <constant> PD_PARTITION PD_CACHE PD_END PD_ACCESS_LATENCY PD_DEFAULT PD_SIZE 
%token <constant> PD_SETS PD_ASSOC PD_REPLACEMENT PD_HIT_LATENCY PD_EXTRA_MISS_LATENCY 
%token <constant> PD_PARTS_CACHED PD_MISSES_SERVICED_BY PD_BLOCKSIZE  PD_OFFCHIP PD_PORTS

%token <char_string>	PD_IDENTIFIER PD_CON

%type <constant> con

%type <pl>  parts_list part
%type <pi>  part_items part_item
%type <cl>  cache_list cache
%type <ci>  cache_items cache_item
%type <il>  ident_list 
%type <char_string>  ident


/*
 * we'll have a global structure.  it'll contain a list of partitions, and
 * a list of caches.
 * each partition will contain an access latency, size, and a bit to indicate if its default
 * each cache will contain set count, blocksize, assoc, replacement, hit latency,
 * miss detect latency, a list of partitions cached, and where misses should be serviced.
 */


%%

all: 	parts_list cache_list		{ final_check($1, $2);	}


/* 
 * note that the part list must at least 
 * contain a default partition, but that 
 * cache list may be empty
 */
parts_list:	part parts_list		{ $$=add_memory_partition($1,$2); }
	  |  	part			{ $$=add_memory_partition($1,NULL); }
	  ;

cache_list:	cache cache_list	{ $$=add_cache($1,$2);  	}
/*	  |  	cache			{ $$=add_cache($1,NULL); 	}*/
 	  |				{ $$=NULL;			}
	  ;

part:	  	PD_PARTITION ident part_items PD_END PD_PARTITION { $$=create_partition($2,$3); }
    ;

part_items:	part_item part_items	{ $$=add_part_item($1,$2); 	}
	  |	part_item		{ $$=add_part_item($1,NULL); 	}
	  ;	



part_item: 	PD_SIZE '=' con			{ $$=create_part_item_size($3); 	}
         |	PD_ACCESS_LATENCY '=' con	{ $$=create_part_item_latency($3); 	}
	 | 	PD_DEFAULT			{ $$=create_part_item_default(); 	}
	 |	PD_OFFCHIP				{ $$=create_part_item_offchip(); }
         |	PD_PORTS '=' con	 		{ $$=create_part_item_ports($3); }
	 ;


cache	:	  	PD_CACHE ident cache_items PD_END PD_CACHE	{ $$=create_cache($2,$3); }
    	;

cache_items:	cache_item cache_items	{ $$=add_cache_item($1,$2); 	}
	  |	cache_item		{ $$=add_cache_item($1,NULL); 	}
	  ;	

cache_item:	PD_SETS '=' con				{ $$=create_cache_item_sets($3); }
	 |	PD_BLOCKSIZE '=' con			{ $$=create_cache_item_blocksize($3); }
         |	PD_ASSOC '=' con    			{ $$=create_cache_item_assoc($3); }
         |	PD_REPLACEMENT '=' ident		{ $$=create_cache_item_replacement($3); }
         |	PD_HIT_LATENCY '=' con			{ $$=create_cache_item_hit_latency($3); }
         |	PD_EXTRA_MISS_LATENCY '=' con		{ $$=create_cache_item_extra_miss_latency($3); }
         |	PD_PARTS_CACHED '=' ident_list  	{ $$=create_cache_item_parts_cached($3); }
         |	PD_MISSES_SERVICED_BY '=' ident	 	{ $$=create_cache_item_misses_to($3); }
	 |	PD_OFFCHIP				{ $$=create_cache_item_offchip(); }
         |	PD_PORTS '=' con	 		{ $$=create_cache_item_ports($3); }
	 ; 

ident_list:	ident ',' ident_list	{ $$=add_identifier_list($1,$3); 	}
	  |	ident			{ $$=add_identifier_list($1,NULL); 	}
	  ;

ident	:	PD_IDENTIFIER 	{ $$=strdup(pdtooltext);}
	;

con     :       PD_CON    	{ $$=atoi(pdtooltext); } 
        ;


	

