
/************************************************************************/
/*                                                                      */
/* Copyright (c) 1990, 1991, 1992 -                                     */
/*     The Rector and Visitors of the University of Virginia            */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from the University of Virginia.                                */
/* This code is distributed WITHOUT ANY WARRANTY.  No claims are        */
/*      made as to whether it serves any particular purpose or even     */
/*      works at all.                                                   */
/*                                                                      */
/************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/regtool/regtool.c,v $
 $Date: 2003/10/08 14:58:05 $
 $Revision: 1.6 $
 $Name:  $

 Log of changes:

 $Log: regtool.c,v $
 Revision 1.6  2003/10/08 14:58:05  clc5q
 Fix nested comment problem in log message text.

 Revision 1.5  2003/10/05 16:44:43  clc5q
 At end of regs.h output, make the final line be "#endif // _REGS_H // "
    so that the vlink Makefiles can use the tag to grep out what
    is needed from regs.h for vlink.

 Revision 1.4  2002/06/28 14:47:02  jwd
 Added parentheses around the body of the REGCOM macro so that there would
 not be precedence problems if the user of the macro forgot to put parentheses
 around the macro call.

 Revision 1.3  2001/05/16 18:53:49  jdh8d
 Changes to support "soft floating point"

 Revision 1.2  2000/10/03 23:00:48  jwd
 Jason Hiser modified regtool to permit the specification of the encoding size
 of the locals, registers, and globals.  He also modified it so that the memory
 types and register types are decoupled.  Previously, they were linked.

 Revision 1.4  2000/07/18 18:36:00  jdh8d
 no

 Revision 1.3  1998/08/07 16:30:08  jwd
 Changed function fits to take an int as a first parameter.
 Visual C++ didn't like the short int parameter type.
 
 Revision 1.2  1998/04/08 17:26:36  jwd
 First ANSI C version.

*************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include "y.tab2.h"
#include "scan.h"
#include "regtool.h"
#include "regtool2.h"

/*
 * This file contains the main processing function of regtool
 */

extern struct classinfo *classes;
extern struct typeinfo *types;
extern int tnum;
extern int code_num;
extern char codes[];
extern char memprt[];
extern char regprt[];
extern char extraMems[];
void pruneExtraMems(char* memprt);

/* default encodings of locals, globals, and registers */
int localsize = 2, globalsize = 2, registersize = 2;
void emitRegisterOctal(unsigned int type, unsigned int number);


/*
 * main - entry routine for regtool
 */
int main(int argc, char **argv) {

   register int i;
   int direct, bitsperint;
   int cgen = FALSE;
   unsigned int ui;
   char *line;

   /*
      Determine size of an "int" on host machine
   */
   ui = bitsperint = 1;
   while (ui <<= 1)
      bitsperint++;

   /*
      Parse input parameters
   */
   direct = 0;
   while (--argc > 0) {
      if (*(*++argv) == '-') {
	 for (line = *argv + 1; *line; line++) {
	    switch (*line) {

	       case 'c':
		  cgen = TRUE;
		  break;

	       case 'w':
		  bitsperint = atoi(line + 1);
		  for (line++; *line; line++);
		  line--;
		  break;

	       default:
		  fprintf(stderr, "regtool: illegal option '%c'\n", *line);
		  exit(1);
	    }
	 }
      }
      else if (direct) {
	 if (!freopen(*argv, "w", stdout)) {
	    fprintf(stderr, "regtool: cannot open %s for output\n", *argv);
	    exit(1);
	 }
      }
      else {
	 if (!freopen(*argv, "r", stdin)) {
	    fprintf(stderr, "regtool: cannot open %s for input\n", *argv);
	    exit(1);
	 }
	 direct = 1;
      }
   }

   /*
      Initialize codition code information
   */
   *codes = '\0';
   code_num = 0;

   /*
      Initialize memory and register character definitions
   */
   for (i = 0; i < MAXTYPES; i++)
      memprt[i] = regprt[i] = ' ';

   /*
      Parse the input file
   */
   (void) yyparse();

   /*
      Perform semantic checks
   */
   checks();

   /*
      Emit .h file
   */
   hfile(cgen, bitsperint);

   /*
      Return with no errors
   */
   return (0);
}

/*
 * checks - perform semantic checks that could not be performed earlier
 */
void checks(void) {

   register short *vptr, *fptr;
   register struct classinfo *cptr;
   register struct typeinfo *tptr, *gptr;


   /*
      Ensure that no items that were marked reserved show up in the order list
   */
   for (cptr = classes; cptr; cptr = cptr->next) {
      if (cptr->order && cptr->reserve) {
	 for (fptr = cptr->order->num; *fptr != -1; fptr++) {
	    for (vptr = cptr->reserve->num; *vptr != -1; vptr++) {
	       if (*fptr == *vptr) {
		  fprintf(stderr, "regtool: attempt to order a reserved ");
		  fprintf(stderr, "register in class \"%s\"\n", cptr->name);
		  exit(1);
	       }
	    }
	 }
      }
   }

   /*
      Ensure that no scratch items were marked reserved
   */
   for (cptr = classes; cptr; cptr = cptr->next) {
      if (cptr->scratch && cptr->reserve) {
	 for (fptr = cptr->scratch->num; *fptr != -1; fptr++) {
	    for (vptr = cptr->reserve->num; *vptr != -1; vptr++) {
	       if (*fptr == *vptr) {
		  fprintf(stderr, "regtool: attempt to reserve a scratch ");
		  fprintf(stderr, "register in class \"%s\"\n", cptr->name);
		  exit(1);
	       }
	    }
	 }
      }
   }

   /*
      Ensure that no item in the scratch vector exceed the stated limit
   */
   for (cptr = classes; cptr; cptr = cptr->next) {
      if (cptr->scratch) {
	 for (vptr = cptr->scratch->num; *vptr != -1; vptr++) {
	    if (*vptr >= cptr->number) {
	       fprintf(stderr, "regtool: register in scratch list is above the ");
	       fprintf(stderr, "declared limit in class \"%s\"\n", cptr->name);
	       exit(1);
	    }
	 }
      }
   }

   /*
      Ensure that no item in the reserve vector exceed the stated limit
   */
   for (cptr = classes; cptr; cptr = cptr->next) {
      if (cptr->reserve) {
	 for (vptr = cptr->reserve->num; *vptr != -1; vptr++) {
	    if (*vptr >= cptr->number) {
	       fprintf(stderr, "regtool: register in reserve list is above the ");
	       fprintf(stderr, "declared limit in class \"%s\"\n", cptr->name);
	       exit(1);
	    }
	 }
      }
   }

   /*
      Ensure that no item in the order vector exceed the stated limit
   */
   for (cptr = classes; cptr; cptr = cptr->next) {
      if (cptr->order) {
	 for (vptr = cptr->order->num; *vptr != -1; vptr++) {
	    if (*vptr >= cptr->number) {
	       fprintf(stderr, "regtool: register in order list is above the ");
	       fprintf(stderr, "declared limit in class \"%s\"\n", cptr->name);
	       exit(1);
	    }
	 }
      }
   }

   /*
      Ensure that no items in the invariant vector exceed the stated limit
   */
   for (tptr = types; tptr; tptr = tptr->next) {
      if (tptr->invariant) {
	 for (vptr = tptr->invariant->num; *vptr != -1; vptr++) {
	    if (*vptr >= tptr->class->number) {
	       fprintf(stderr, "regtool: register in invariant list is above ");
	       fprintf(stderr, "the declared limit in type \"%s\"\n", tptr->name);
	       exit(1);
	    }
	 }
      }
   }

   /*
      Ensure that no limits are outside the stated limit
   */
   for (tptr = types; tptr; tptr = tptr->next) {
      if (tptr->ulimit >= 0) {
	 if (tptr->ulimit >= tptr->class->number) {
	    fprintf(stderr, "regtool: limit is outside declared range ");
	    fprintf(stderr, "in type \"%s\"\n", tptr->name);
	    exit(1);
	 }
      }
   }

   /*
      Ensure that all groups belong to the same class and that only the head
      group item contains an alignment or a limit specifier.
   */
   for (tptr = types; tptr; tptr = tptr->next) {
      if (tptr->group && tptr->head) {
	 tptr->htype = tptr;
	 for (gptr = tptr->group; gptr; gptr = gptr->group) {
	    if (gptr->class != tptr->class) {
	       fprintf(stderr, "regtool: group beginning with type ");
	       fprintf(stderr, "\"%s\" contains type ", tptr->name);
	       fprintf(stderr, "\"%s\" which is in another class\n", gptr->name);
	       exit(1);
	    }
	    if (gptr->alignment) {
	       fprintf(stderr, "regtool: type \"%s\" is in a group ", gptr->name);
	       fprintf(stderr, "and cannot have an alignment specification\n");
	       exit(1);
	    }
	    if (gptr->llimit >= 0 || gptr->ulimit >= 0) {
	       fprintf(stderr, "regtool: type \"%s\" is in a group ", gptr->name);
	       fprintf(stderr, "and cannot have a limit specification\n");
	       exit(1);
	    }
	    gptr->htype = tptr;
	 }
      }
   }
}

/*
 * hfile - use the collected data to emit an appropriate regs.h file
 */
void hfile(int cgen, int bitsperint) {

   register struct classinfo *cptr;
   register struct typeinfo *tptr;
   register short *vptr, *fptr, *sptr;
   register int index, maxregs, maxtemp, maxnum, limit, check, count, first;
   struct classinfo *lptr;
   struct typeinfo *xptr;
   int hi, lo, maxclasses;
   char *a;

   printf("#ifndef _REGS_H\n");
   printf("#define _REGS_H 1\n");

   /*
      Ouput MAXTEMP and MAXREGS definitions Assign "first" item to each class,
      based on how many locations each class requires.  Initialize the "layout"
      vector for each class, defining the location of each register item in
      order to ensure appropriate selection.
   */
   maxregs = maxtemp = maxnum = maxclasses = 0;
   lptr = (struct classinfo *) NULL;
   for (cptr = classes; cptr; cptr = cptr->next) {
      if (cptr->flags & SPILLC) {
	 if (cptr->number > 0)
	    maxtemp = cptr->number;
      }
      else {
	 cptr->num = maxclasses++;
	 cptr->first = maxregs;
	 if (cptr->number > 0 && cptr->types) {
	    maxregs += cptr->number;
	    if (cptr->number > maxnum)
	       maxnum = cptr->number;
	    cptr->ralloc = 0;
	    cptr->layout = (struct vector *) salloc(sizeof(struct vector));
	    vptr = cptr->layout->num;

	    /*
	       The "order" vector determines which items are placed at the
	       front of the "layout" vector.  These would then be selected by
	       the register allocator before any others.
	    */
	    if (cptr->order) {
	       for (fptr = cptr->order->num; *fptr != -1; fptr++) {
		  cptr->ralloc++;
		  *vptr++ = *fptr;
	       }
	    }

	    /*
	       After placing the items indicated in the "order" vector, we then
	       give preference to those items in the "scratch" vector, since
	       these should, by definition, be the cheapest to use.
	    */
	    if (cptr->scratch) {
	       for (fptr = cptr->scratch->num; *fptr != -1; fptr++) {
		  for (sptr = cptr->layout->num; sptr != vptr; sptr++)
		     if (*sptr == *fptr)
			break;
		  if (sptr == vptr) {
		     cptr->ralloc++;
		     *vptr++ = *fptr;
		  }
	       }
	    }

	    /*
	       Then we must add in the items which have not already been placed
	       and are not reserved.
	    */
	    for (index = 0; index < cptr->number; index++) {
	       if (cptr->reserve) {
		  for (fptr = cptr->reserve->num; *fptr != -1; fptr++)
		     if (*fptr == index)
			break;
		  if (*fptr == index)
		     continue;
	       }
	       for (sptr = cptr->layout->num; sptr != vptr; sptr++)
		  if (*sptr == index)
		     break;
	       if (sptr == vptr) {
		  cptr->ralloc++;
		  *vptr++ = index;
	       }
	    }

	    /*
	       Finally, the reserved items are placed at the end, to ensure
	       that they are never selected for use by the register assigner.
	    */
	    if (cptr->reserve) {
	       for (fptr = cptr->reserve->num; *fptr != -1; fptr++)
		  *vptr++ = *fptr;
	    }

	    *vptr = -1;
	    lptr = cptr;
	 }
	 else {
	    fprintf(stderr,
		"regtool: Warning: class '%s' lacks registers and/or types\n",
		    cptr->name);
	 }
      }
   }
   if (!maxtemp) {
      fprintf(stderr, "regtool: Warning: no spill class defined\n");
      maxtemp = 1;
   }
   if (!cgen) {
      printf("#define MAXTEMP %d\n", maxtemp);
      printf("#define MAXREGS %d\n", maxregs);
      printf("#define MAXCLASSES %d\n", (maxclasses) ? maxclasses : 1);
      printf("\n");
   }

   /*
      Output basic register type definitions
   */
   for (tptr = types; tptr; tptr = tptr->next)
      printf("#define %s %d\n", tptr->name, tptr->number);
   printf("\n");

   if (cgen) {

      printf("#define FIRST_PSEUDO_REG %d\n", maxnum);

      /*
         Output REGPRT definition
      */
      regprt[MAXTYPES] = '\0';
      printf("#define REGPRT \"%s\"\n", regprt);
      printf("\n");

      /*
         Emit macros that help form compressed register codes
      */
      for (tptr = types; tptr; tptr = tptr->next)
	 printf("#define %cR(n) TREG(%s, n)\n",
		toupper(regprt[tptr->number]), tptr->name);
      printf("\n");

      return;
   }

   /*
      Output macros to help identify register types
   */
   for (tptr = types; tptr; tptr = tptr->next)
   {
          printf("#define IS%s(c) (((c)&(char)0xfc)=="
			"(char)(0xc0|(%s<<2)))\n",
	     		tptr->name, tptr->name);

   }
   printf("\n");

   /*
      Output MEMPRT definition
   */
   memprt[tnum] = '\0';
   pruneExtraMems(memprt);
   printf("#define MEMPRT \"%s%s\"\n", memprt, extraMems);
   printf("\n");

   /*
      Emit regtoindex layouts for each different class Output regtoindex()
      macro definition
   */
   if (lptr) {
      printf("#ifdef ASSIGN\n");
      printf("short _r2i[%d][%d] = {\n", tnum, maxnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 if (!(cptr = tptr->class)) {
	    fprintf(stderr, "regtool: no definition for type \"%s\"\n",
		    tptr->name);
	    exit(1);
	 }
	 for (index = 0; index < cptr->number; index++) {
	    for (vptr = cptr->layout->num; *vptr != index; vptr++);
	    if (!index)
	       printf("  { %d", (vptr - cptr->layout->num) + cptr->first);
	    else
	       printf(", %d", (vptr - cptr->layout->num) + cptr->first);
	 }
	 while (index < maxnum)
	    if (!index++)
	       printf("  { 0");	/* } */
	    else
	       printf(", 0");
	 printf(" }");
	 if (!tptr->next)
	    printf("\n");
	 else
	    printf(",\n");
      }
      printf("  };\n");
      printf("#else\n");
      printf("extern short _r2i[%d][%d];\n", tnum, maxnum);
      printf("#endif\n");
      printf("\n");
      printf("#define regtoindex(t, n) _r2i[t][n]\n");
   }
   else
      printf("#define regtoindex(t, n) 0\n");
   printf("\n");

   /*
      Emit indextoreg layouts for each different class. Output indextoreg()
      macro definition.
   */
   if (lptr) {
      printf("#ifdef ASSIGN\n");
      printf("short _i2r[MAXREGS] = { ");
      for (cptr = classes; cptr; cptr = cptr->next) {
	 if (!(cptr->flags & SPILLC) && cptr->number > 0 && cptr->types) {
	    vptr = cptr->layout->num;
	    printf(" %d", *vptr++);
	    while (*vptr != -1)
	       printf(", %d", *vptr++);
	    if (cptr == lptr)
	       printf(" };\n");
	    else {
	       printf(",\n");
	       printf("                        ");
	    }
	 }
      }
      printf("#else\n");
      printf("extern short _i2r[MAXREGS];\n");
      printf("#endif\n");
      printf("\n");
      printf("#define indextoreg(t, n) _i2r[n]\n");
   }
   else
      printf("#define indextoreg(t, n) 0\n");
   printf("\n");

   /*
      Assign "first" item to each type, taking into consideration the type's
      limit (if any) and the type's alignment (if any). Output firstreg() macro
      definitions.
   */
   if (tnum) {
      printf("#ifdef ASSIGN\n");
      printf("short _firstreg[%d] = { ", tnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 cptr = tptr->class;
	 tptr->first = 0;
	 vptr = cptr->layout->num;
	 if (tptr->llimit > -1) {
	    while (*vptr != tptr->llimit) {
	       if (*vptr == tptr->ulimit) {
		  fprintf(stderr, "regtool: limit conflict at type \"%s\"\n",
			  tptr->name);
		  exit(1);
	       }
	       vptr++;
	    }
	    tptr->first = (vptr - cptr->layout->num);
	 }
	 while (!fits(*vptr, tptr->alignment)) {
	    vptr++;
	    tptr->first++;
	    if (*vptr == -1) {
	       fprintf(stderr, "regtool: cannot assign a first to type \"%s\"\n",
		       tptr->name);
	       exit(1);
	    }
	 }
	 if (tptr->number)
	    printf(", %d", tptr->first + tptr->class->first);
	 else
	    printf("%d", tptr->first + tptr->class->first);
      }
      printf(" };\n");
      printf("#else\n");
      printf("extern short _firstreg[%d];\n", tnum);
      printf("#endif\n");
      printf("\n");
      printf("#define firstreg(t) _firstreg[t]\n");
   }
   else
      printf("#define firstreg(t) 0\n");
   printf("\n");

   /*
      Assign "tfirst" item to each type, taking into consideration the type's
      spill aligment (if any). Output firsttemp() macro definitions.
   */
   if (tnum) {
      printf("#ifdef ASSIGN\n");
      printf("short _firsttemp[%d] = { ", tnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 tptr->tfirst = 0;
	 while (!fits(tptr->tfirst, tptr->talignment)) {
	    tptr->tfirst++;
	    if (tptr->tfirst == maxtemp) {
	       fprintf(stderr, "regtool: cannot assign a spill to type \"%s\"\n",
		       tptr->name);
	       exit(1);
	    }
	 }
	 if (tptr->number)
	    printf(", %d", tptr->tfirst);
	 else
	    printf("%d", tptr->tfirst);
      }
      printf(" };\n");
      printf("#else\n");
      printf("extern short _firsttemp[%d];\n", tnum);
      printf("#endif\n");
      printf("\n");
      printf("#define firsttemp(t) _firsttemp[t]\n");
   }
   else
      printf("#define firsttemp(t) 0\n");
   printf("\n");

   /*
      Output the definition of availreg() by determining the number of
      registers that are available from the firstreg().
   */
   if (tnum) {
      printf("#ifdef ASSIGN\n");
      printf("short _availreg[%d] = { ", tnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 struct typeinfo *gptr;
	 int size;

	 cptr = tptr->class;

	 /*
	    Set "limit" to the number of registers between the firstreg of the
	    type and the last register that can be used (determined by the
	    upper limit and where the reserved registers start).
	 */
	 limit = tptr->first;
	 vptr = cptr->layout->num + tptr->first;
	 while (limit < cptr->ralloc) {
	    limit++;
	    if (*vptr == tptr->ulimit)
	       break;
	    vptr++;
	 }

	 /*
	    Attempt to identify an appropriate "skip" between each acceptable
	    register (according to the stated alignment).  If no alignment is
	    specified, then the "skip" is one.  There might be no appropriate
	    "skip" value, in which case an error will be issued.  Note that we
	    know from the checks performed to calculate firstreg() that the
	    first register value satisfies the alignment requirements.
	 */
	 tptr->skip = 1;
	 if (tptr->size < 0)
	    tptr->size = 1;
	 if (tptr->alignment) {
	    vptr = cptr->layout->num + tptr->first;
	    for (index = tptr->first; index + tptr->size <= limit; tptr->skip++) {
	       index++;
	       *vptr++;
	       if (fits(*vptr, tptr->alignment))
		  break;
	    }
	    while (index + tptr->size <= limit) {
	       index += tptr->skip;
	       vptr += tptr->skip;
	       if (index < limit && !fits(*vptr, tptr->alignment)) {
		  fprintf(stderr,
			  "regtool: alignment difficulties in type \"%s\"\n",
			  tptr->name);
		  exit(1);
	       }
	    }
	 }

	 /*
	    The size that we need to use to obtain the "avail" value has to
	    take into consideration the length of a group.
	 */
	 size = tptr->size;
	 if (tptr->group && tptr->head) {
	    for (gptr = tptr->group; gptr; gptr = gptr->group)
	       size += gptr->size;
	 }

	 /*
	    Determine the "avail" value for this type (that is, the actual
	    number of registers that are available for this type).  All that we
	    need to do is use the "first," "limit," and "skip" values already
	    determined, along with the "size" of the type to count the number
	    of registers available.  Since it is assumed that an item whose
	    size is greater than one occupies a certain number of contiguous
	    registers, contiguity is checked.
	 */
	 tptr->avail = 0;
	 index = tptr->first;
	 vptr = cptr->layout->num + tptr->first;
	{
 	 int type_limit=limit;
/*
	 if(tptr->ulimit>=0)
		type_limit=tptr->ulimit;
	 if(tptr->llimit>=0)
		index=tptr->llimit;
*/
	 while (index + size <= type_limit) 
         {
	    sptr = fptr = vptr;
	    for (check = 1; check < size; check++) {
	       fptr++;
	       if (*fptr - *sptr != 1) {
		  fprintf(stderr,
			  "regtool: Warning! non-contiguous registers in type \"%s\"\n",
			  tptr->name);
/*		  exit(1); */
	       }
	       sptr++;
	    }
	    index += tptr->skip;
	    vptr += tptr->skip;
	    tptr->avail++;
	 }
	}

	 if (tptr == types)
	    printf(" %d", tptr->avail);
	 else
	    printf(", %d", tptr->avail);
      }
      printf(" };\n");
      printf("#else\n");
      printf("extern short _availreg[%d];\n", tnum);
      printf("#endif\n");
      printf("\n");
      printf("#define availreg(t) _availreg[t]\n");
   }
   else
      printf("#define availreg(t) 0\n");
   printf("\n");

   /*
      Output the definition of availtemp() by determining the number of valid
      spill locations available for each register type.
   */
   if (tnum) {
      printf("#ifdef ASSIGN\n");
      printf("short _availtemp[%d] = { ", tnum);
      for (tptr = types; tptr; tptr = tptr->next) {

	 /*
	    Attempt to identify an appropriate "skip" between each spill
	    location (according to the stated alignment).  If no alignment is
	    specified, then the "skip" is one.
	 */
	 tptr->tskip = 1;
	 if (tptr->tsize < 0)
	    tptr->tsize = 1;
	 if (tptr->talignment) {
	    for (index = tptr->tfirst;
		 index + tptr->tsize <= maxtemp;
		 tptr->tskip++)
	       if (fits(++index, tptr->talignment))
		  break;
	 }

	 /*
	    Determine the "tavail" value for this type (taht is, the actual
	    number of spill locations that are available for this type.  All
	    that we need to do is use the "tfirst", "maxtemp" and "tskip"
	    values already determined, along with the "tsize" of the type to
	    count the number of spill locations available.
	 */
	 tptr->tavail = 0;
	 index = tptr->tfirst;
	 while (index + tptr->tsize <= maxtemp) {
	    index += tptr->tskip;
	    tptr->tavail++;
	 }

	 if (tptr == types)
	    printf(" %d", tptr->tavail);
	 else
	    printf(", %d", tptr->tavail);
      }
      printf(" };\n");
      printf("#else\n");
      printf("extern short _availtemp[%d];\n", tnum);
      printf("#endif\n");
      printf("\n");
      printf("#define availtemp(t) _availtemp[t]\n");
   }
   else
      printf("#define availtemp(t) 0\n");
   printf("\n");

   /*
      Output the skipreg() macro.  This is the "skip" information that was
      calculated for availreg().
   */
   if (tnum) {
      printf("#ifdef ASSIGN\n");
      printf("short _skipreg[%d] = { ", tnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 if (tptr == types)
	    printf(" %d", tptr->skip);
	 else
	    printf(", %d", tptr->skip);
      }
      printf(" };\n");
      printf("#else\n");
      printf("extern short _skipreg[%d];\n", tnum);
      printf("#endif\n");
      printf("\n");
      printf("#define skipreg(t) _skipreg[t]\n");
   }
   else
      printf("#define skipreg(t) 1\n");
   printf("\n");

   /*
      Output the skiptemp() macro.  This is the "tskip" information that was
      calculated for availtemp().
   */
   if (tnum) {
      printf("#ifdef ASSIGN\n");
      printf("short _skiptemp[%d] = { ", tnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 if (tptr == types)
	    printf(" %d", tptr->tskip);
	 else
	    printf(", %d", tptr->tskip);
      }
      printf(" };\n");
      printf("#else\n");
      printf("extern short _skiptemp[%d];\n", tnum);
      printf("#endif\n");
      printf("\n");
      printf("#define skiptemp(t) _skiptemp[t]\n");
   }
   else
      printf("#define skiptemp(t) 1\n");
   printf("\n");

   /*
      Output the allocreg() macro.  This macro must take into consideration the
      sizes of all of the types that make up a group when giving the
      information for the head type in a group.  Except for this, this macro is
      identical to the contreg() macro declaration.
   */
   if (tnum) {
      printf("#ifdef ASSIGN\n");
      printf("short _allocreg[%d] = { ", tnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 struct typeinfo *gptr;
	 int size;

	 size = tptr->size;
	 if (tptr->group && tptr->head) {
	    for (gptr = tptr->group; gptr; gptr = gptr->group)
	       size += gptr->size;
	 }

	 if (tptr == types)
	    printf(" %d", size);
	 else
	    printf(", %d", size);
      }
      printf(" };\n");
      printf("#else\n");
      printf("extern short _allocreg[%d];\n", tnum);
      printf("#endif\n");
      printf("\n");
      printf("#define allocreg(t) _allocreg[t]\n");
   }
   else
      printf("#define allocreg(t) 1\n");
   printf("\n");

   /*
      Output the contreg() macro definition.  This will return the specified
      "size" for each given type.
   */
   if (tnum) {
      printf("#ifdef ASSIGN\n");
      printf("short _contreg[%d] = { ", tnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 if (tptr == types)
	    printf(" %d", tptr->size);
	 else
	    printf(", %d", tptr->size);
      }
      printf(" };\n");
      printf("#else\n");
      printf("extern short _contreg[%d];\n", tnum);
      printf("#endif\n");
      printf("\n");
      printf("#define contreg(t) _contreg[t]\n");
   }
   else
      printf("#define contreg(t) 1\n");
   printf("\n");

   /*
      Output the conttemp() macro.
   */
   if (tnum) {
      printf("#ifdef ASSIGN\n");
      printf("short _conttemp[%d] = { ", tnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 if (tptr == types)
	    printf(" %d", tptr->tsize);
	 else
	    printf(", %d", tptr->tsize);
      }
      printf(" };\n");
      printf("#else\n");
      printf("extern short _conttemp[%d];\n", tnum);
      printf("#endif\n");
      printf("\n");
      printf("#define conttemp(t) _conttemp[t]\n");
   }
   else
      printf("#define conttemp(t) 1\n");
   printf("\n");

   /*
      Output the classreg() macro definition.  This will return the specified
      "class" for each given type.
   */
   printf("#ifdef CLASSREG\n");
   if (tnum) {
      printf("#ifdef ASSIGN\n");
      printf("short _classreg[%d] = { ", tnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 if (tptr == types)
	    printf(" %d", tptr->class->num);
	 else
	    printf(", %d", tptr->class->num);
      }
      printf(" };\n");
      printf("#else\n");
      printf("extern short _classreg[%d];\n", tnum);
      printf("#endif\n");
      printf("\n");
      printf("#define classreg(t) _classreg[t]\n");
   }
   else
      printf("#define classreg(t) 0\n");
   printf("#endif\n");
   printf("\n");

   /*
      Output the localtype() macro
   */
   if (tnum) {
      printf("#ifdef ASSIGN\n");
      printf("short _localtype[%d] = { ", tnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 int ltype;

	 if (tptr->noalloc || tptr->class->flags & (FIFOC | STACKC))
	    ltype = -1;
	 else if (tptr->localtype)
	    ltype = tptr->localtype->number;
	 else
	    ltype = tptr->number;
	 if (tptr->number)
	    printf(", %d", ltype);
	 else
	    printf("%d", ltype);
      }
      printf(" };\n");
      printf("#else\n");
      printf("extern short _localtype[%d];\n", tnum);
      printf("#endif\n");
      printf("\n");
      printf("#define localtype(t) _localtype[t]\n");
   }
   else
      printf("#define localtype(t) 0\n");
   printf("\n");

   /*
      Output the _checktype array
   */
   printf("#ifdef ASSIGN\n");
   if (tnum) {
      printf("short _checktype[%d][%d] = { ", tnum, tnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 int ltype;

	 printf("{%2d", tptr->compatible[0]);
	 for (ltype = 1; ltype < tnum; ltype++)
	    printf(",%2d", tptr->compatible[ltype]);
	 if (tptr->next) {	/* { */
	    printf("},\n");
	    printf("                           ");
	 }
	 else
	    printf("} };\n");
      }
   }
   else
      printf("short _checktype[1, 1] = { { 0 } };\n");
   printf("#endif\n");
   printf("\n");

   /*
      Output the nextgroup() macro declaration
   */
   printf("#ifdef ASSIGN\n");
   if (tnum) {
      printf("short _nextgroup[%d] = { ", tnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 if (tptr->group) {
	    if (tptr == types)
	       printf(" %d", tptr->group->number);
	    else
	       printf(", %d", tptr->group->number);
	 }
	 else if (tptr == types)
	    printf(" -1");
	 else
	    printf(", -1");
      }
      printf(" };\n");
      printf("\n");
      printf("#define nextgroup(t) _nextgroup[t]\n");
   }
   else
      printf("#define nextgroup(t) -1\n");
   printf("\n");

   /*
      Output the headgroup() macro declaration
   */
   if (tnum) {
      printf("short _headgroup[%d] = { ", tnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 if (tptr->htype) {
	    if (tptr == types)
	       printf(" %d", tptr->htype->number);
	    else
	       printf(", %d", tptr->htype->number);
	 }
	 else if (tptr == types)
	    printf(" -1");
	 else
	    printf(", -1");
      }
      printf(" };\n");
      printf("\n");
      printf("#define headgroup(t) _headgroup[t]\n");
   }
   else
      printf("#define headgroup(t) -1\n");
   printf("#endif\n");
   printf("\n");

   /*
      Generate MARKINVARIANT macro
   */
   printf("#define MARKINVARIANT ");
   first = TRUE;
   for (tptr = types; tptr; tptr = tptr->next) {
      if (tptr->invariant) {
	 count = 0;
	 for (vptr = tptr->invariant->num;
	      count < VECLEN && *vptr != -1;
	      vptr++) {
	    if (!first)
	       printf("  \\\n                      ");
	    first = FALSE;
          /* 
	    hi = 0xc0 | (tptr->number << 2) | (*vptr >> 7);
	    lo = 0x80 | (*vptr & 0x7f);
	  */
	    printf("sptr = s_add(\"");
	    emitRegisterOctal(tptr->number, *vptr);
	    printf("\"); \\\n");
	    printf("                      sptr->ldead = top->rtls;");
	 }
      }
   }
   printf("\n");
   printf("\n");

   /*
      Calculate aliases by determining which type items that are in the same
      class overlap.  Output an array of character pointers containing an alias
      string for each possible register.
   */
   printf("#ifdef ASSIGN\n");
   if (tnum) {
      printf("static char *_aliases[%d][%d] = {\n", tnum, maxnum);
      for (tptr = types; tptr; tptr = tptr->next) {
	 cptr = tptr->class;
	 for (index = 0; index < maxnum; index++) {
	    if (index)
	       printf(",\n  ");
	    else
	       printf("{ ");
	    if (index + tptr->size > cptr->number ||
		!fits(index, tptr->alignment))
	       printf("\"\"");
	    else {
	       printf("\"");
	       for (xptr = types; xptr; xptr = xptr->next) {
		  if (xptr->class != cptr)
		     continue;
		  check = index - xptr->size + 1;
		  if (check < 0)
		     check = 0;
		  while (check < index + tptr->size) {
		     if (check + xptr->size > cptr->number)
			break;
		     if (fits(check, xptr->alignment)) {
			/*hi = 0xc0 | (xptr->number << 2) | (check >> 7);
			lo = 0x80 | (check & 0x7f);*/
	    		emitRegisterOctal(xptr->number, check);
			printf("\\0");
		     }
		     check++;
		  }
	       }
	       printf("\"");
	    }
	 }
	 printf(" }");
	 if (!tptr->next)
	    printf("\n");
	 else
	    printf(",\n");
      }
      printf("};\n");
      printf("\n");
   }

   /*
      Output information needed to initialize the hardware registers
   */
   limit = code_num;
   for (tptr = types; tptr; tptr = tptr->next) {
      cptr = tptr->class;
      for (index = 0; index + tptr->size <= cptr->number; index++)
	 if (fits(index, tptr->alignment))
	    limit++;
   }
   if (limit) {
      printf("#define HREGINFO %d\n", limit);
      printf("\n");
      printf("struct hreginfo _hrinfo[HREGINFO] = {\n");
      check = 0;
      for (a = codes; *a; a += 3) {
	 printf("  { HARDWARE, \"%s\" }", a);
	 check++;
	 if (check == limit)
	    printf("\n");
	 else
	    printf(",\n");
      }
      for (tptr = types; tptr; tptr = tptr->next) {
	 cptr = tptr->class;
	 for (index = 0; index + tptr->size <= cptr->number; index++) {
	    if (!fits(index, tptr->alignment))
	       continue;
	    printf("  { HARDWARE"); /* } */
	    if (cptr->scratch) {
	       vptr = cptr->scratch->num;
	       while (*vptr != -1) {
		  if (*(vptr++) == index) {
		     printf(" | SCRATCH");
		     break;
		  }
	       }
	    }
	    if (cptr->flags & FIFOC)
	       printf(" | FIFOREG");
	    if (cptr->flags & STACKC)
	       printf(" | STACKREG");
	    /* hi = 0xc0 | (tptr->number << 2) | (index >> 7);
	    lo = 0x80 | (index & 0x7f);*/
	    printf(", \"");
	    emitRegisterOctal(tptr->number, index);
	    /*{*/ printf("\" }"); 
	    check++;
	    if (check == limit)
	       printf("\n");
	    else
	       printf(",\n");
	 }
      }
      printf("};\n");
      printf("\n");
   }
   printf("#endif\n");
   printf("\n");

   /*
      Output REGINC macro
   */
   printf("#define REGINC(x, y) *x |= *y");
   index = 1;
   for (maxregs -= bitsperint; maxregs > 0; maxregs -= bitsperint) {
      printf("; \\\n");
      printf("                     *(x + %d) |= *(y + %d)", index, index);
      index++;
   }
   printf("\n");
   printf("\n");

   /*
      Output REGCOM macro
   */
   printf("#define REGCOM(x, y) (*x & *y");
   for (maxregs = 1; maxregs < index; maxregs++) {
      printf(" || \\\n");
      printf("                     *(x + %d) & *(y + %d)", maxregs, maxregs);
   }
   printf(")\n");
   printf("\n");

   /*
      Emit the TRASH_CC declaration
   */
   printf("#define TRASH_CC \"");
   for (a = codes; *a; a += 3)
      printf("%s\\0", a);
   printf("\"\n");
   printf("\n");

   /*
      Emit the IS_CC macro
   */
   printf("#define IS_CC(s) (" /*)*/ );
   if (!code_num)
      printf("FALSE");
   else {
      first = TRUE;
      for (a = codes; *a; a += 3) {
	 if (!first) {
	    printf(" || \\\n");
	    printf("                  ");
	 }
	 first = FALSE;
	 printf("(*(s) == '%c' && *((s) + 1) == '%c')", *a, *(a + 1));
      }
   }
   printf( /*(*/ ")\n");
   printf("\n");

   /*
      Emit the SET_CC macro
   */
   printf("#define SET_CC(s) (");
   if (!code_num)
      printf("FALSE");
   else {
      printf("(" /*)*/ );
      first = TRUE;
      for (a = codes; *a; a += 3) {
	 if (!first) {
	    printf(" || \\\n");
	    printf("                    ");
	 }
	 first = FALSE;
	 printf("(*(s) == '%c' && *((s) + 1) == '%c')", *a, *(a + 1));
      }
      printf(") && *((s) + 2) == '='");
   }
   printf( /*(*/ ")\n");
   printf("\n");

   /*
      Output the required typedefs for the reginfo and regarray items
   */
   printf("typedef struct {\n");
   printf("  struct reglist reg[MAXREGS];\n");
   printf("  struct reglist temp[MAXTEMP];\n");
   printf("  } reginfo;\n");
   printf("\n");
   printf("typedef reginfo *regarray;\n");

   printf("#define MAXLOCLENGTH (%d)\n\n",
	 (registersize>globalsize 
		? (registersize>localsize ? registersize : localsize) 
		: (globalsize  >localsize ? globalsize   : localsize)));

   printf("#define REGLOCLENGTH    (%d)\n\n", registersize);
   printf("#define GLOBALLOCLENGTH (%d)\n\n", globalsize);
   printf("#define LOCALLOCLENGTH  (%d)\n\n", localsize);

   printf("#endif /* _REGS_H */ \n");
}

/*
 * fits - examines a number with respect to an alignment
 */
int fits(int num, struct align_expr * align) {

   if (align) {
      num -= align->offset;
      if (num < 0)
	 return (FALSE);
      if (num % align->align)
	 return (FALSE);
   }
   return (TRUE);
}


/*
	Remove any elements from extraMems that
	are already defined in memprt to avoid duplicates
*/

#include <string.h>

void pruneExtraMems(char* memprt) {
   char* mem_ptr=extraMems;

   /* walk through the extra memories */
   while(*mem_ptr!=0) {
      /* if the current character is already in memprt */
      /* we need to prune it out			 */
      /* strchr returns 0 if it is *not* in the string */
      if (strchr(memprt, *mem_ptr) != 0) {
         /* move the last character to the current location */
         *mem_ptr = extraMems[strlen(extraMems)-1];
         /* and null terminate the last position */
			extraMems[strlen(extraMems)-1]=0;

			/* note this still works OK if mem_ptr is the
		   	 *	last thing on the list.  It copies
			 *	from the the last item, to the last item
			 *	then eliminates itself
			 *	from the end of the list 
			 */

			/* we don't advance to the next element here because
			 * we moved an element overtop of the current location.
			 * thus, we have to re-check the new element at this
			 * location
			 */	
		}
		else
		{
			/* advance to the next element */
			mem_ptr++;
		}
	}

}

void setGlobalSize(int size) {
	globalsize=size;
}

void setLocalSize(int size) {
        localsize=size;
}

void setRegisterSize(int size) {
        registersize=size;
}

void emitRegisterOctal(unsigned int type, unsigned int number) {
	int i;
	unsigned char reg[10];
	if(registersize>10)
	{
		fprintf(stderr, ("Registersize>10, bailing out ...\n"));
		exit(1);
	}	
	for(i=0;i<registersize; i++)
		reg[i]=0x80; 	/* set high bit */

	reg[0]|=(1<<6); /* set bit to indicate register */

	reg[0]|=(type<<2); /* set bits to indicate type */

	for(i=registersize-1;i>=0; i--)
	{
		/* or in the lower 7 bits of the number */
                reg[i]|=(number&((1<<7)-1));

		/* get rid of those bits from number */
                number=number>>7;
	}

	for(i=0;i<registersize; i++)
	{
		printf("\\%03o", (int)reg[i]);
	}
	
	
}
