/*
 * This file contains the regtool grammar
 */

%{

/************************************************************************/
/*                                                                      */
/* Copyright (c) 1990, 1991, 1992 -                                     */
/*     The Rector and Visitors of the University of Virginia            */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from the University of Virginia.                                */
/* This code is distributed WITHOUT ANY WARRANTY.  No claims are        */
/*      made as to whether it serves any particular purpose or even     */
/*      works at all.                                                   */
/*                                                                      */
/************************************************************************/

#include "regtool.h"

struct vector *list();
struct align_expr *expr();
%}

%union {
   int val;
   char chr;
   char *str;
   struct vector *numlist;
   struct align_expr *aexpr;
   };

%token <str> IDN
%token <val> CON
%token <chr> CHAR
%token TYPE
%token SIZE
%token FIFO
%token TYPES
%token MEMTYPES
%token CODES
%token CLASS
%token ORDER
%token GROUP
%token LIMIT
%token STACK
%token NUMBER
%token SCRATCH
%token RESERVE
%token MEMCHAR
%token REGCHAR
%token NOALLOC
%token ENDTYPE
%token ENDCLASS
%token ALIGNMENT
%token INVARIANT
%token LOCALTYPE
%token COMPATIBLE
%token REGISTER, LOCAL, GLOBAL 

%type <numlist> n_list
%type <aexpr> expr
%%
prog	:	types mem_types sizes codes classes
	;


sizes 	: 	size sizes
	| 
	;

/* sizes of the encoding for locals, globals, and registers */
size    :	localsize
	|	globalsize
	|	registersize
 	;

localsize :	LOCAL    SIZE CON  { setLocalSize($3);    } ;
globalsize :	GLOBAL   SIZE CON  { setGlobalSize($3);   } ;
registersize :	REGISTER SIZE CON  { setRegisterSize($3); } ;

types	:	TYPES t_list
	;

t_list	:	t_list ',' IDN			{ deftype($3); }
	|	IDN				{ deftype($1); }
	;

mem_types	:	MEMTYPES m_list
	| 	/* empty */
	;

m_list  :       m_list ',' IDN                  { defmem($3); }
        |       IDN                             { defmem($1); }
        ;


codes	:	CODES c_list
	|
	;

c_list	:	c_list ',' IDN			{ defcode($3); }
	|	IDN				{ defcode($1); }
	;

classes	:	classes class
	|
	;

class	:	CLASS '=' IDN			{ defclass($3); }
		options
		ENDCLASS
	;

options	:	options option
	|
	;

option	:	NUMBER '=' CON			{ number($3); }
	|	SCRATCH '=' n_list		{ scratch($3); }
	|	RESERVE '=' n_list		{ reserve($3); }
	|	ORDER '=' n_list		{ order($3); }
	|	FIFO				{ fifo(); }
	|	STACK				{ stack(); }
	|	TYPE '=' ty_list
		topts
		ENDTYPE				{ endtype(); }
	;

ty_list	:	ty_list ',' IDN			{ addtype($3); }
	|	IDN				{ addtype($1); }
	;

topts	:	topts topt
	|
	;

topt	:	ALIGNMENT '=' expr		{ alignment($3); }
	|	SIZE '=' CON			{ size($3); }
	|	LIMIT '=' CON range CON		{ limit($3, $5); }
	|	GROUP '=' IDN			{ group($3); }
	|	INVARIANT '=' n_list		{ invariant($3); }
	|	LOCALTYPE '=' IDN		{ localtype($3); }
	|	NOALLOC				{ noalloc(); }
	|	COMPATIBLE '=' IDN ',' IDN	{ compatible($3, $5); }
	|	MEMCHAR '=' CHAR		{ memchar($3); }
	|	REGCHAR '=' CHAR		{ regchar($3); }
	;

n_list	:					{ newlist(); }
		n_opts				{ $$ = list(); }
	;

n_opts	:	n_opts ',' n_opt
	|	n_opt
	;

n_opt	:	CON				{ con($1); }
	|	CON range CON			{ range($1, $3); }
	;

expr	:	CON '+' CON			{ $$ = expr($1, $3); }
	|	CON '-' CON			{ $$ = expr($1, -$3); }
	|	CON				{ $$ = expr($1, 0); }
	;

range	:	'.' '.'
	|	'-'
	;
%%
#include <stdio.h>
