
/************************************************************************/
/*                                                                      */
/* Copyright (c) 1990, 1991, 1992 -                                     */
/*     The Rector and Visitors of the University of Virginia            */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from the University of Virginia.                                */
/* This code is distributed WITHOUT ANY WARRANTY.  No claims are        */
/*      made as to whether it serves any particular purpose or even     */
/*      works at all.                                                   */
/*                                                                      */
/************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/regtool/scan.c,v $
 $Date: 2000/10/03 23:02:49 $
 $Revision: 1.2 $
 $Name:  $

 Log of changes:

 $Log: scan.c,v $
 Revision 1.2  2000/10/03 23:02:49  jwd
 Jason Hiser modified regtool to permit the specification of the encoding size
 of the locals, registers, and globals.  He also modified it so that the memory
 types and register types are decoupled.  Previously, they were linked.

 Revision 1.5  2000/07/18 18:36:01  jdh8d
 no

 Revision 1.3  1998/08/07 16:31:51  jwd
 Changed y.tab.h to ytab.h because mks yacc on NT
 produces this.
 Revision 1.2  1998/04/08 18:09:31  jwd
 First ANSI C version.

*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "regtool.h"
#include "y.tab.h"
#include "scan.h"
/*
 * This file contains the lexical scanner

  -------------------------------------------------------------
    History of Modifications:
    1/29/98: Function prototyping made to be ansi-C
       

  -------------------------------------------------------------

 */

int line = 1;		/* Current line in the input file */

/*
 * yylex - return the next token from the input file
 */
int yylex(void) {

   register int c, val;
   register char *a;
   char id[MAXID];

   for (;;) {
      switch (c = getc(stdin)) {

	 case EOF:

	    /*
	       End of file (return terminator)
	    */
	    return (0);

	 case (int) ',':
	 case (int) '=':
	 case (int) '.':
	 case (int) '+':
	 case (int) '-':

	    /*
	       Operators and special characters (return)
	    */
	    return (c);

	 case (int) '\t':
	 case (int) ' ':

	    /*
	       Separator (ignore)
	    */
	    break;

	 case (int) '#':

	    /*
	       Comment (skip to end of line and fall into newline code)
	    */
	    while ((c = getc(stdin)) != EOF && c != ((int) '\n'));

	 case (int) '\n':

	    /*
	       Newline (increment line counter)
	    */
	    line++;
	    break;

	 case (int) '0':
	 case (int) '1':
	 case (int) '2':
	 case (int) '3':
	 case (int) '4':
	 case (int) '5':
	 case (int) '6':
	 case (int) '7':
	 case (int) '8':
	 case (int) '9':

	    /*
	       Integer constant
	    */
	    val = 0;
	    do {
	       val = val * 10 + c - (int) '0';
	    } while ((c = getc(stdin)) >= ((int) '0') && c <= ((int) '9'));
	    ungetc(c, stdin);
	    yylval.val = val;
	    return (CON);

	 case (int) '\'':

	    /*
	       Character constant
	    */
	    c = getc(stdin);
	    if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z'))
	       yylval.chr = (char) c;
	    else {
	       fprintf(stderr,
		   "regtool (yylex): invalid character constant at line %d\n",
		       line);
	       exit(1);
	    }
	    if (getc(stdin) != (int) '\'') {
	       fprintf(stderr,
		   "regtool (yylex): invalid character constant at line %d\n",
		       line);
	       exit(1);
	    }
	    return (CHAR);

	 case (int) 'A':
	 case (int) 'B':
	 case (int) 'C':
	 case (int) 'D':
	 case (int) 'E':
	 case (int) 'F':
	 case (int) 'G':
	 case (int) 'H':
	 case (int) 'I':
	 case (int) 'J':
	 case (int) 'K':
	 case (int) 'L':
	 case (int) 'M':
	 case (int) 'N':
	 case (int) 'O':
	 case (int) 'P':
	 case (int) 'Q':
	 case (int) 'R':
	 case (int) 'S':
	 case (int) 'T':
	 case (int) 'U':
	 case (int) 'V':
	 case (int) 'W':
	 case (int) 'X':
	 case (int) 'Y':
	 case (int) 'Z':
	 case (int) 'a':
	 case (int) 'b':
	 case (int) 'c':
	 case (int) 'd':
	 case (int) 'e':
	 case (int) 'f':
	 case (int) 'g':
	 case (int) 'h':
	 case (int) 'i':
	 case (int) 'j':
	 case (int) 'k':
	 case (int) 'l':
	 case (int) 'm':
	 case (int) 'n':
	 case (int) 'o':
	 case (int) 'p':
	 case (int) 'q':
	 case (int) 'r':
	 case (int) 's':
	 case (int) 't':
	 case (int) 'u':
	 case (int) 'v':
	 case (int) 'w':
	 case (int) 'x':
	 case (int) 'y':
	 case (int) 'z':
	 case (int) '_':

	    /*
	       Identifier
	    */
	    a = id;
	    val = 1;
	    do {
	       if (++val > MAXID) {
		  fprintf(stderr,
			  "regtool (yylex): identifier too long at line %d\n",
			  line);
		  exit(1);
	       }
	       *a++ = (char) c;
	    } while (((c = getc(stdin)) >= ((int) 'A') && c <= ((int) 'Z')) ||
		     (c >= ((int) 'a') && c <= ((int) 'z')) ||
		     (c >= ((int) '0') && c <= ((int) '9')) ||
		     c == ((int) '_'));
	    ungetc(c, stdin);
	    *a = '\0';

	    /*
	       Check for keywords
	    */
	    if (!strcmp(id, "type"))
	       return (TYPE);
	    else if (!strcmp(id, "size"))
	       return (SIZE);
	    else if (!strcmp(id, "fifo"))
	       return (FIFO);
	    else if (!strcmp(id, "types"))
	       return (TYPES);
	    else if (!strcmp(id, "memtypes"))
	       return (MEMTYPES);
	    else if (!strcmp(id, "codes"))
	       return (CODES);
	    else if (!strcmp(id, "class"))
	       return (CLASS);
	    else if (!strcmp(id, "order"))
	       return (ORDER);
	    else if (!strcmp(id, "group"))
	       return (GROUP);
	    else if (!strcmp(id, "limit"))
	       return (LIMIT);
	    else if (!strcmp(id, "stack"))
	       return (STACK);
	    else if (!strcmp(id, "number"))
	       return (NUMBER);
	    else if (!strcmp(id, "scratch"))
	       return (SCRATCH);
	    else if (!strcmp(id, "reserve"))
	       return (RESERVE);
	    else if (!strcmp(id, "memchar"))
	       return (MEMCHAR);
	    else if (!strcmp(id, "regchar"))
	       return (REGCHAR);
	    else if (!strcmp(id, "noalloc"))
	       return (NOALLOC);
	    else if (!strcmp(id, "endtype"))
	       return (ENDTYPE);
	    else if (!strcmp(id, "endclass"))
	       return (ENDCLASS);
	    else if (!strcmp(id, "alignment"))
	       return (ALIGNMENT);
	    else if (!strcmp(id, "invariant"))
	       return (INVARIANT);
	    else if (!strcmp(id, "localtype"))
	       return (LOCALTYPE);
	    else if (!strcmp(id, "compatible"))
	       return (COMPATIBLE);
	    else if (!strcmp(id, "local"))
	       return (LOCAL);
	    else if (!strcmp(id, "global"))
	       return (GLOBAL);
	    else if (!strcmp(id, "register"))
	       return (REGISTER);
	    else if (!strcmp(id, "size"))
	       return (SIZE);

	    yylval.str = strcpy(salloc(val), id);
	    return (IDN);

	 default:
	    fprintf(stderr,
		    "regtool (yylex): invalid character '%c' at line %d\n",
		    c, line);
	    exit(1);
      }
   }
}

/*
 * yyerror - called if the parser encounters an error
 */
void yyerror(char *message) {


   fprintf(stderr,
	   "regtool (yyerror): %s at line %d\n",
	   message, line);
   exit(1);
}

/*
 * salloc - allocate a block of memory
 */
char *salloc(long unsigned int size) {

   register char *ptr;

   ptr = malloc(size);
   if (!ptr) {
      fprintf(stderr, "regtool (salloc): out of memory\n");
      exit(1);
   }
   return (ptr);
}
