
/************************************************************************/
/*                                                                      */
/* Copyright (c) 1990, 1991, 1992 -                                     */
/*     The Rector and Visitors of the University of Virginia            */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from the University of Virginia.                                */
/* This code is distributed WITHOUT ANY WARRANTY.  No claims are        */
/*      made as to whether it serves any particular purpose or even     */
/*      works at all.                                                   */
/*                                                                      */
/************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/regtool/semact.c,v $
 $Date: 2000/08/22 21:10:33 $
 $Revision: 1.1.1.1 $
 $Name:  $

 Log of changes:

 $Log: semact.c,v $
 Revision 1.1.1.1  2000/08/22 21:10:33  zephyr
 Created directory structure

 Revision 1.4  2000/07/18 18:36:01  jdh8d
 no

 Revision 1.3  1998/08/07 16:32:38  jwd
 Added include of stdlib.h.
 
 Revision 1.2  1998/04/08 18:10:23  jwd
 First ANSI C version.

*************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "regtool.h"
#include "scan.h"
#include "semact.h"

/*
 * This file contains the semantic actions required to support the regtool
 * grammar.
 --------------------------------------------------------------------
        History of Changes
        ------------------
        1/29/98: Modified to make function prottyping ansi C.        
	


 --------------------------------------------------------------------
 */

struct typeinfo *types = NULL;	/* Pointer to first item in types list   */
struct typeinfo *ctypes = NULL;	/* Pointer to the current types list     */
struct classinfo *classes = NULL;	/* Pointer to first item in classes
					   list */
struct classinfo *current = NULL;	/* Pointer to the current class          */
struct vector *clist = NULL;	/* Pointer to current vector list        */
int tnum = 0;		/* Number of types declared              */
int code_num;		/* Number of condition codes declared    */
char codes[MAXCODES * 3 + 1];	/* Condition code definitions            */
char memprt[MAXTYPES + 1];	/* Memory character definitions          */
char regprt[MAXTYPES + 1];	/* Register character definitions        */
char extraMems[MAXTYPES+1]; 	/* Extra memory characters */

extern int line;

/*
 * deftype - called when a new type is defined
 */
void deftype(char *id) {

   register struct typeinfo *tptr, *pptr;
   register int i;


   /*
      Search through types list to ensure that the new identifier is not a
      duplicate.
   */
   pptr = (struct typeinfo *) NULL;
   for (tptr = types; tptr; tptr = tptr->next) {
      if (!strcmp(id, tptr->name)) {
	 fprintf(stderr,
		 "regtool (deftype): mutiply defined type '%s' in line %d\n",
		 id, line);
	 exit(1);
      }
      pptr = tptr;
   }

   /*
      Ensure that the maximum number of types is not exceeded
   */
   if (tnum >= MAXTYPES) {
      fprintf(stderr, "regtool (endtypes): too many types defined\n");
      exit(1);
   }

   /*
      If the new type is unique, allocate a typeinfo structure for it and link
      it on to the end of the types list.
   */
   tptr = (struct typeinfo *) salloc(sizeof(struct typeinfo));
   tptr->next = (struct typeinfo *) NULL;
   tptr->name = id;
   tptr->number = tnum++;
   tptr->class = (struct classinfo *) NULL;
   tptr->spill = (struct classinfo *) NULL;
   tptr->alias = (struct typeinfo *) NULL;
   tptr->talias = (struct typeinfo *) NULL;
   tptr->alignment = (struct align_expr *) NULL;
   tptr->talignment = (struct align_expr *) NULL;
   tptr->size = -1;
   tptr->tsize = -1;
   tptr->llimit = -1;
   tptr->ulimit = -1;
   tptr->head = TRUE;
   tptr->group = (struct typeinfo *) NULL;
   tptr->htype = (struct typeinfo *) NULL;
   tptr->invariant = (struct vector *) NULL;
   tptr->localtype = (struct typeinfo *) NULL;
   tptr->noalloc = FALSE;
   for (i = 0; i < MAXTYPES; i++)
      tptr->compatible[i] = (i == tptr->number) ? tptr->number : -1;
   tptr->memchar = '\0';
   tptr->regchar = '\0';
   if (islower(*id))
      memprt[tptr->number] = toupper(*id);
   else
      memprt[tptr->number] = *id;
   if (isupper(*id))
      regprt[tptr->number] = tolower(*id);
   else
      regprt[tptr->number] = *id;
   if (pptr)
      pptr->next = tptr;
   else
      types = tptr;
}

/*
 * Defines a memory type that is not "linked" to any register types 
 */
 
void defmem(char* memName)
{
	char temp[2];

	/* snarf the first character of the name */
	temp[0]=memName[0];	
	temp[1]=0;

	/* and add it to the extra memories */
	strcat(extraMems, temp);
}

/*
 * defcode - called when a new condition code token is defined
 */
void defcode(char *id) {

   register char *a;


   if (code_num >= MAXCODES) {
      fprintf(stderr, "regtool (defcode): too many condition codes defined\n");
      exit(1);
   }

   if (strlen(id) != 2 || *id < 'A' || *id > 'Z' ||
       *(id + 1) < 'A' || *(id + 1) > 'Z') {
      fprintf(stderr, "regtool (defcode): illegal codition code token, ");
      fprintf(stderr, "\"%s\" at line %d\n", id, line);
      exit(1);
   }

   for (a = codes; *a; a += 3) {
      if (!strcmp(a, id)) {
	 fprintf(stderr, "regtool (defcode): mutiply defined condition ");
	 fprintf(stderr, "code \"%s\" at line %d\n", id, line);
	 exit(1);
      }
   }

   code_num++;

   *a++ = *id;
   *a++ = *(id + 1);
   *a++ = '\0';
}

/*
 * defclass - called when a new class is defined
 */
void defclass(char *id) {

   register struct classinfo *cptr, *pptr;

   /*
      Search through classes list to ensure that the new identifier is not a
      duplicate.
   */
   pptr = (struct classinfo *) NULL;
   for (cptr = classes; cptr; cptr = cptr->next) {
      if (!strcmp(id, cptr->name)) {
	 fprintf(stderr,
		 "regtool (defclass): mutiply defined class '%s' in line %d\n",
		 id, line);
	 exit(1);
      }
      pptr = cptr;
   }

   /*
      If the new class is unique, allocate a classinfo structure for it and
      link it on to the end of the class list.
   */
   cptr = (struct classinfo *) salloc(sizeof(struct classinfo));
   cptr->next = (struct classinfo *) NULL;
   cptr->name = id;
   cptr->flags = 0;
   cptr->number = -1;
   cptr->scratch = (struct vector *) NULL;
   cptr->reserve = (struct vector *) NULL;
   cptr->order = (struct vector *) NULL;
   cptr->types = (struct typeinfo *) NULL;
   if (pptr)
      pptr->next = cptr;
   else
      classes = cptr;

   /*
      Set the current class pointer to this new class
   */
   current = cptr;

   /*
      Check to see if the class is the special spill class
   */
   if (!strcmp(cptr->name, "SPILL"))
      cptr->flags |= SPILLC;
}

/*
 * number - number of registers in a class
 */
void number(int con) {

   if (current->number != -1) {
      fprintf(stderr,
	      "regtool (number): duplicate number field at line %d\n",
	      line);
      exit(1);
   }
   current->number = con;
}

/*
 * scratch - list of scratch registers in a class
 */
void scratch(struct vector * vec) {

   if (current->flags & SPILLC) {
      fprintf(stderr,
	      "regtool (scratch): scratch field in SPILL class at line %d\n",
	      line);
      exit(1);
   }

   if (current->scratch) {
      fprintf(stderr,
	      "regtool (scratch): duplicate scratch field at line %d\n",
	      line);
      exit(1);
   }

   current->scratch = vec;
}

/*
 * reserve - list of reserved registers in a class
 */
void reserve(struct vector * vec) {

   if (current->flags & SPILLC) {
      fprintf(stderr,
	      "regtool (reserve): reserve field in SPILL class at line %d\n",
	      line);
      exit(1);
   }

   if (current->reserve) {
      fprintf(stderr,
	      "regtool (reserve): duplicate reserve field at line %d\n",
	      line);
      exit(1);
   }

   current->reserve = vec;
}

/*
 * order - list of order preference in a class
 */
void order(struct vector * vec) {

   if (current->flags & SPILLC) {
      fprintf(stderr,
	      "regtool (order): order field in SPILL class at line %d\n",
	      line);
      exit(1);
   }

   if (current->order) {
      fprintf(stderr,
	      "regtool (order): duplicate order field at line %d\n",
	      line);
      exit(1);
   }

   current->order = vec;
}

/*
 * fifo - set fifo type for a class
 */
void fifo(void) {

   if (current->flags & SPILLC) {
      fprintf(stderr,
	      "regtool (fifo): fifo field in SPILL class at line %d\n",
	      line);
      exit(1);
   }

   if (current->flags & FIFOC) {
      fprintf(stderr,
	      "regtool (fifo): duplicate fifo field at line %d\n",
	      line);
      exit(1);
   }
   else if (current->flags & STACKC) {
      fprintf(stderr,
	      "regtool (fifo): cannot mix stack and fifo at line %d\n",
	      line);
      exit(1);
   }
   else
      current->flags |= FIFOC;
}

/*
 * stack - set stack type for a class
 */
void stack(void) {

   if (current->flags & SPILLC) {
      fprintf(stderr,
	      "regtool (stack): stack field in SPILL class at line %d\n",
	      line);
      exit(1);
   }

   if (current->flags & STACKC) {
      fprintf(stderr,
	      "regtool (stack): duplicate stack field at line %d\n",
	      line);
      exit(1);
   }
   else if (current->flags & FIFOC) {
      fprintf(stderr,
	      "regtool (stack): cannot mix stack and fifo at line %d\n",
	      line);
      exit(1);
   }
   else
      current->flags |= STACKC;
}

/*
 * addtype - add a type to the list for this class
 */
void addtype(char *id) {

   register struct typeinfo *tptr, *fptr, *pptr;

   /*
      Search through types list to ensure that the type identifier exists
   */
   for (tptr = types; tptr; tptr = tptr->next) {
      if (!strcmp(id, tptr->name)) {

	 /*
	    Check to see if this is the special SPILL class
	 */
	 if (current->flags & SPILLC) {
	    if (tptr->spill) {
	       fprintf(stderr,
		  "regtool (addtype): type '%s' already defined at line %d\n",
		       id, line);
	       exit(1);
	    }

	    /*
	       Link this type into the list of types for the current class
	    */
	    if (!ctypes)
	       ctypes = tptr;
	    pptr = (struct typeinfo *) NULL;
	    for (fptr = current->types; fptr; fptr = fptr->talias)
	       pptr = fptr;
	    if (pptr)
	       pptr->talias = tptr;
	    else
	       current->types = tptr;
	    tptr->spill = current;
	 }
	 else {
	    if (tptr->class) {
	       fprintf(stderr,
		  "regtool (addtype): type '%s' already defined at line %d\n",
		       id, line);
	       exit(1);
	    }

	    /*
	       Link this type into the list of types for the current class
	    */
	    if (!ctypes)
	       ctypes = tptr;
	    pptr = (struct typeinfo *) NULL;
	    for (fptr = current->types; fptr; fptr = fptr->alias)
	       pptr = fptr;
	    if (pptr)
	       pptr->alias = tptr;
	    else
	       current->types = tptr;
	    tptr->class = current;
	 }

	 return;
      }
   }

   fprintf(stderr,
	   "regtool (addtype): undefined type '%s' in line %d\n",
	   id, line);
   exit(1);
}

/*
 * endtype - end of the current type list for this class
 */
void endtype(void) {


   /*
      Set the current type list starting pointer to nothing
   */
   ctypes = (struct typeinfo *) NULL;
}

/*
 * alignment - alignment specification for a type in a class
 */
void alignment(struct align_expr * expr) {

   register struct typeinfo *tptr;

   /*
      See if this is the special SPILL class
   */
   if (current->flags & SPILLC) {
      if (ctypes->talignment) {
	 fprintf(stderr,
		 "regtool (alignment): duplicate alignment field at line %d\n",
		 line);
	 exit(1);
      }

      for (tptr = ctypes; tptr; tptr = tptr->talias)
	 tptr->talignment = expr;
   }
   else {
      if (ctypes->alignment) {
	 fprintf(stderr,
		 "regtool (alignment): duplicate alignment field at line %d\n",
		 line);
	 exit(1);
      }

      for (tptr = ctypes; tptr; tptr = tptr->alias)
	 tptr->alignment = expr;
   }
}

/*
 * size - size specification for a type in a class
 */
void size(int con) {

   register struct typeinfo *tptr;

   /*
      See if this is the special SPILL class
   */
   if (current->flags & SPILLC) {
      if (ctypes->tsize != -1) {
	 fprintf(stderr,
		 "regtool (size): duplicate size field at line %d\n",
		 line);
	 exit(1);
      }

      for (tptr = ctypes; tptr; tptr = tptr->talias)
	 tptr->tsize = con;
   }
   else {
      if (ctypes->size != -1) {
	 fprintf(stderr,
		 "regtool (size): duplicate size field at line %d\n",
		 line);
	 exit(1);
      }

      for (tptr = ctypes; tptr; tptr = tptr->alias)
	 tptr->size = con;
   }
}

/*
 * limit - limit specification for a type in a class
 */
void limit(int lcon, int ucon) {

   register struct typeinfo *tptr;

   if (current->flags & SPILLC) {
      fprintf(stderr,
	      "regtool (limit): limit field in SPILL class at line %d\n",
	      line);
      exit(1);
   }

   if (ctypes->llimit != -1) {
      fprintf(stderr,
	      "regtool (limit): duplicate limit field at line %d\n",
	      line);
      exit(1);
   }

   for (tptr = ctypes; tptr; tptr = tptr->alias) {
      tptr->llimit = lcon;
      tptr->ulimit = ucon;
   }
}

/*
 * group - group specification for a type in a class
 */
void group(char *id) {

   register struct typeinfo *type, *tptr, *fptr;

   if (current->flags & SPILLC) {
      fprintf(stderr,
	      "regtool (group): group field in SPILL class at line %d\n",
	      line);
      exit(1);
   }

   if (ctypes->group) {
      fprintf(stderr,
	      "regtool (group): duplicate group field at line %d\n",
	      line);
      exit(1);
   }

   /*
      Look for the type identifier
   */
   for (type = types; type; type = type->next) {
      if (!strcmp(id, type->name)) {
	 for (tptr = ctypes; tptr; tptr = tptr->alias) {
	    for (fptr = type; fptr; fptr = fptr->group)
	       if (fptr == tptr) {
		  fprintf(stderr,
		    "regtool (group): circular group definition at line %d\n",
			  line);
		  exit(1);
	       }
	    tptr->group = type;
	 }
	 type->head = FALSE;
	 return;
      }
   }

   fprintf(stderr,
	   "regtool (group): undefined type '%s' in line %d\n",
	   id, line);
   exit(1);
}

/*
 * invariant - invariant specification for a type in a class
 */
void invariant(struct vector * vec) {

   register struct typeinfo *tptr;

   if (current->flags & SPILLC) {
      fprintf(stderr,
	   "regtool (invariant): invariant field in SPILL class at line %d\n",
	      line);
      exit(1);
   }

   if (ctypes->invariant) {
      fprintf(stderr,
	      "regtool (invariant): duplicate invariant field at line %d\n",
	      line);
      exit(1);
   }

   for (tptr = ctypes; tptr; tptr = tptr->alias)
      tptr->invariant = vec;
}

/*
 * localtype - define the type that locals and arguments should be assigned to
 */
void localtype(char *id) {

   register struct typeinfo *type, *tptr;

   if (current->flags & SPILLC) {
      fprintf(stderr,
	   "regtool (localtype): localtype field in SPILL class at line %d\n",
	      line);
      exit(1);
   }

   if (ctypes->localtype) {
      fprintf(stderr,
	      "regtool (localtype): duplicate localtype field at line %d\n",
	      line);
      exit(1);
   }

   if (ctypes->noalloc) {
      fprintf(stderr,
	      "regtool (localtype): localtype field with noalloc at line %d\n",
	      line);
      exit(1);
   }

   /*
      Look for the type identifier
   */
   for (type = types; type; type = type->next) {
      if (!strcmp(id, type->name)) {
	 for (tptr = ctypes; tptr; tptr = tptr->alias)
	    tptr->localtype = type;
	 return;
      }
   }

   fprintf(stderr,
	   "regtool (localtype): undefined type '%s' in line %d\n",
	   id, line);
   exit(1);
}

/*
 * noalloc - declare the current types exempt from global allocation
 */
void noalloc(void) {

   register struct typeinfo *tptr;

   if (current->flags & SPILLC) {
      fprintf(stderr,
	      "regtool (noalloc): noalloc field in SPILL class at line %d\n",
	      line);
      exit(1);
   }

   if (ctypes->localtype) {
      fprintf(stderr,
	      "regtool (noalloc): noalloc field with localtype at line %d\n",
	      line);
      exit(1);
   }

   if (ctypes->noalloc) {
      fprintf(stderr,
	      "regtool (noalloc): duplicate noalloc field at line %d\n",
	      line);
      exit(1);
   }

   /*
      Mark all of the current types for no global allocation
   */
   for (tptr = ctypes; tptr; tptr = tptr->alias)
      tptr->noalloc = TRUE;
}

/*
 * compatible - declare global allocation type compatibility
 */
void compatible(char *id1, char *id2) {

   register struct typeinfo *type1, *type2, *tptr;

   if (current->flags & SPILLC) {
      fprintf(stderr,
	 "regtool (compatible): compatible field in SPILL class at line %d\n",
	      line);
      exit(1);
   }

   /*
      Look for the first type identifier
   */
   for (type1 = types; type1; type1 = type1->next) {
      if (!strcmp(id1, type1->name)) {
	 for (type2 = types; type2; type2 = type2->next) {
	    if (!strcmp(id2, type2->name)) {
	       for (tptr = ctypes; tptr; tptr = tptr->alias) {
		  if (tptr->compatible[type1->number] != -1) {
		     fprintf(stderr,
			     "regtool (compatible): duplicate compatible definition in line %d\n",
			     line);
		     exit(1);
		  }
		  tptr->compatible[type1->number] = type2->number;
	       }
	       return;
	    }
	 }

	 fprintf(stderr,
		 "regtool (compatible): undefined type '%s' in line %d\n",
		 id2, line);
	 exit(1);
      }
   }

   fprintf(stderr,
	   "regtool (compatible): undefined type '%s' in line %d\n",
	   id1, line);
   exit(1);
}

/*
 * memchar - define a memory character for the current types
 */
void memchar(char c) {

   struct typeinfo *tptr;

   if (current->flags & SPILLC) {
      fprintf(stderr,
	      "regtool (memchar): memchar field in SPILL class at line %d\n",
	      line);
      exit(1);
   }

   if (c < 'A' || c > 'Z') {
      fprintf(stderr,
	      "regtool (memchar): illegal memory character '%c' at line %d\n",
	      c, line);
   }

   for (tptr = ctypes; tptr; tptr = tptr->alias) {
      if (tptr->memchar) {
	 fprintf(stderr,
	       "regtool (memchar): duplicate memchar definition in line %d\n",
		 line);
	 exit(1);
      }
      tptr->memchar = c;
      memprt[tptr->number] = c;
   }
}

/*
 * regchar - define a register character for the current types
 */
void regchar(char c) {

   struct typeinfo *tptr;

   if (current->flags & SPILLC) {
      fprintf(stderr,
	      "regtool (regchar): regchar field in SPILL class at line %d\n",
	      line);
      exit(1);
   }

   if (c < 'a' || c > 'z') {
      fprintf(stderr,
	    "regtool (regchar): illegal register character '%c' at line %d\n",
	      c, line);
   }

   for (tptr = ctypes; tptr; tptr = tptr->alias) {
      if (tptr->regchar) {
	 fprintf(stderr,
	       "regtool (regchar): duplicate regchar definition in line %d\n",
		 line);
	 exit(1);
      }
      tptr->regchar = c;
      regprt[tptr->number] = c;
   }
}

/*
 * newlist - create a new vector list
 */
void newlist(void) {

   char *salloc();

   clist = (struct vector *) salloc(sizeof(struct vector));
   clist->num[0] = -1;
}

/*
 * con - add a constant to a vector list
 */
void con(int item) {

   register short *vptr;

   if (item < 0 || item >= VECLEN) {
      fprintf(stderr,
	      "regtool (con): constant '%d' is too large at line %d\n",
	      item, line);
      exit(1);
   };

   for (vptr = clist->num; *vptr != -1; vptr++)
      if (*vptr == item) {
	 fprintf(stderr,
		 "regtool (con): constant '%d' already in list at line %d\n",
		 item, line);
	 exit(1);
      }
   *vptr++ = item;
   *vptr = -1;
}

/*
 * range - add a range of constants to a vector list
 */
void range(int con1, int con2) {

   register int loop;
   void con(int item);

   if (con1 <= con2) {
      for (loop = con1; loop <= con2; loop++)
	 con(loop);
   }
   else {
      for (loop = con2; loop >= con1; loop--)
	 con(loop);
   }
}

/*
 * list - return a pointer to the current vector list
 */
struct vector *list(void) {

   register struct vector *nlist;

   nlist = clist;
   clist = (struct vector *) NULL;
   return (nlist);
}

/*
 * expr - return a pointer to a new alignment expression structure
 */
struct align_expr *expr(int align, int offset) {

   register struct align_expr *aptr;


   aptr = (struct align_expr *) salloc(sizeof(struct align_expr));
   aptr->align = align;
   aptr->offset = offset;
   return (aptr);
}







