/******************************************************************************
 *
 * Copyright (c) 1993 by Mark W. Bailey.
 *
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made as to
 * whether it serves any particular purpose or even works at all.
 *
 *****************************************************************************/

/******************************************************************************
 *
 * File: gencode.c
 *
 * Author: Mark W. Bailey
 *
 * Created: 12/18/93
 *
 * Description:
 *
 *      This file contains the code generators for both vpo and ntro 
 *      versions of ART. 
 *
 * RCS Header:
 * 	$Header: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/rtlprep/gencode.c,v 1.1.1.1 2000/08/22 21:10:33 zephyr Exp $
 *
 * Change Log:
 *	$Log: gencode.c,v $
 *	Revision 1.1.1.1  2000/08/22 21:10:33  zephyr
 *	Created directory structure
 *	
 *	Revision 1.3  2000/06/10 18:19:32  jwd
 *	Added some necessary includes.
 *	
 *	Revision 1.2  1998/08/07 16:35:51  jwd
 *	Added ARGBUILDAREA format. This is to support the new
 *	RTL interface.
 *	Revision 1.1  1998/01/29 20:53:54  jch8f
 *	Initial revision
 *	Revision 1.6  1997/07/11 12:17:57  mwb5y
 *	Fixed code generation of % operator.
 *
 *	Revision 1.5  1997/07/01 13:29:47  mwb5y
 *	Added macros to the dead register list.
 *
 *	Revision 1.4  1995/08/26 14:12:47  mwb5y
 *	Added semicolons to VPO generation and some error checking.
 *
 *	Revision 1.3  1995/08/21 16:07:06  mwb5y
 *	Fixed double quote problem with code expander version.
 *
 *	Revision 1.2  1995/08/21 15:48:02  mwb5y
 *	Updated to work with vpo and code expander.
 *
 * Revision 1.1  1994/01/13  16:40:03  mwb5y
 * Initial revision
 *
 *
 *****************************************************************************/

#define __GENCODE_C__
#include <stdio.h>
#include <ctype.h>
#include "gencode.h"
#include "parse-rtl-y.h"
#include "expr.h"

extern int swn;
extern int swv;
extern int swc;

char *ConstStack[10];		/* Constant stack */
int ConstSp = 0;		/* Constant stack pointer */

static int FirstEffect;		/* Flag indicates first effect in list */

static void VPO_GenerateFormatString(Expr *expr, char *strname);
static void VPO_GeneratePrintArguments(Expr *expr);


/****
 *
 *  VPO_Generate - Generate vpo code for the specified tree. The 
 *  		   function works in two passes. First the tree is 
 *  		   examined to generate the printf format string. 
 *  		   On the second examination of the tree, the printf 
 *  		   arguments are generated. 
 * 
 ****/

void VPO_Generate(Expr *expr, char *strname)
{
    FirstEffect = 1;		      /* Mark that no effects have been seen */
    VPO_GenerateFormatString(expr,strname);/* Generate the printf format string */
    VPO_GeneratePrintArguments(expr); /* Generate the printf arguments */
    ConstSp = 0;		      /* Reset the stack pointer. */
}

/****
 *
 *  VPO_GenerateFormatString - Recursively generate the printf format 
 *  			       string corresponding to the given tree. 
 * 
 ****/

static void VPO_GenerateFormatString(Expr *expr, char *strname)
{
    if(!expr) 
	return;
    
    switch(expr->type) {
      case TransferLineType: {
	  TransferExpr *node = &expr->value.transfer;
	  
	  if (swv) {
	      if (!*strname)
		  error_and_exit("`<string> := rtl' format required for vpo "
				 "ART statements");
	      printf("(void) sprintf(%s, \"",strname);
	  }
	  else {
	      if (*strname)
		  error_and_exit("`<string> := rtl' format is only used for "
				 "vpo ART statements");

	      printf("(void) printf(\"");
	  }
	  VPO_GenerateFormatString(node->transfers, strname);
	  if(node->deads) {
	      printf("\\t");
	      VPO_GenerateFormatString(node->deads, strname);
	  }
	  if (swc)
	      printf("\\n");
	  printf("\"");
	  break;
      }
	
      case ExprListType: {
	  ExprListExpr *node = &expr->value.exprlist;
	  
	  VPO_GenerateFormatString(node->expr, strname);
	  if (node->list) {
	      putchar(';');
	      VPO_GenerateFormatString(node->list, strname);
	  }
	  break;
      }
	
      case EffectType: {
	  EffectExpr *node = &expr->value.effect;
	  
	  if (node->optional)
	      break;

	  /* 
	   * If this is the first effect, place the RTL type on output 
	   * and unset FirstEffect for susequent effects.
	   */

	  if (FirstEffect) {
	      if (swc)
		  putchar(*node->type);
	      FirstEffect = 0;
	  }
	  VPO_GenerateFormatString(node->destination, strname);
	  putchar('=');
	  VPO_GenerateFormatString(node->source, strname);
	  if (swv)
	      putchar(';');
	  break;
      }
	
      case UseLineType: {
	  UseLineExpr *node = &expr->value.useline;
	  
	  printf("(void) printf(\"u");
	  VPO_GenerateFormatString(node->reg, strname);
	  if (swc)
	      printf("\\n");
	  printf("\"");
	  break;
      }
	
      case AsmLineType: {
	  AsmLineExpr *node = &expr->value.asmline;
	  
	  printf("(void) printf(\"-");
	  VPO_GenerateFormatString(node->expr, strname);
	  if (swc)
	      printf("\\n");
	  printf("\"");
	  break;
      }

      case ArgBuildAreaLineType: {
	  ArgBuildAreaExpr *node = &expr->value.argbuildarealine;
	  
	  printf("(void) printf(\"A");
	  VPO_GenerateFormatString(node->asize, strname);
	  if (swc)
	      printf("\\n");
	  printf("\"");
	  break;
      }
	
      case LabelLineType: {
	  LabelLineExpr *node = &expr->value.labelline;
	  
	  printf("(void) printf(\"");
	  VPO_GenerateFormatString(node->label, strname);
	  if (swc)
	      printf("\\n");
	  printf("\"");
	  break;
      }
	
      case DeadListType: {
	  DeadListExpr *node = &expr->value.deadlist;
	  
	  VPO_GenerateFormatString(node->reg, strname);
	  VPO_GenerateFormatString(node->list, strname);
	  break;
      }
	
      case MemoryType: {
	  MemoryExpr *node = &expr->value.memory;
	  
	  if (node->type->type != MemoryTypeType) 
	      printf("%%c");
	  else
	      VPO_GenerateFormatString(node->type, strname);
	  putchar('[');
	  VPO_GenerateFormatString(node->address, strname);
	  putchar(']');
	  break;
      }
	
      case MemoryTypeType: {
	  MemoryTypeExpr *node = &expr->value.memorytype;
	  
	  putchar(node->type);
	  break;
      }
	
      case RegisterType: {
	  RegisterExpr *node = &expr->value.reg;
	  
	  printf("%%c%%c");
	  break;
      }
	
      case RegisterTypeType:
	break;
	
      case GroupType: {
	  GroupExpr *node = &expr->value.group;
	  
	  putchar('(');
	  VPO_GenerateFormatString(node->expr, strname);
	  putchar(')');
	  break;
      }
	
      case BinaryType: {
	  BinaryExpr *node = &expr->value.binary;
	  
	  
	  VPO_GenerateFormatString(node->left, strname);
	  if (*node->op == '%')

	      /* Quote the '%' */

	      putchar('%');

	  putchar(*node->op);
	  VPO_GenerateFormatString(node->right, strname);
	  break;
      }
	
      case UnaryType: {
	  UnaryExpr *node = &expr->value.unary;
	  
	  putchar(*node->op);
	  VPO_GenerateFormatString(node->operand, strname);
	  break;
      }
	
      case MacroType: {
	  MacroExpr *node = &expr->value.macro;
	  
	  printf("%s", node->name);
	  if (node->operand) {
	      putchar('[');
	      VPO_GenerateFormatString(node->operand, strname);
	      putchar(']');
	  }
	  break;
      }
	
      case VariableType: {
	  VariableExpr *node = &expr->value.variable;
	  
	  printf("%%d");
	  break;
      }
	
      case LabelType: {
	  LabelExpr *node = &expr->value.label;
	  
	  printf("%s%%ld", node->prefix);
	  break;
      }
	
      case CType: {
	  CExpr *node = &expr->value.cexpr;
	  
	  printf("%%ld");
	  break;
      }
	
      case InternalCType:
	break;

      case EscExprType: {
	  EscExprExpr *node = &expr->value.escexpr;
	  
	  printf("%%s");
	  break;
      }
	
      case TypedConstType: {
	  TypedConstExpr *node = &expr->value.typedconstant;
	  
	  VPO_GenerateFormatString(node->value, strname);
	  break;
      }
	
      case ConstTypeType: 
	break;
	
      case InternalConstType:
	break;

      case IConstType: {
	  IConstExpr *node = &expr->value.iconst;
	  
	  printf("%s", node->value);
	  break;
      }
	
      case FConstType: {
	  FConstExpr *node = &expr->value.fconst;
	  
	  printf("%s", node->value);
	  break;
      }
	
      default:
	fprintf(stderr, "Unknown type in VPO_GenerateFormatString\n");
    }
}


/****
 *
 *  VPO_GeneratePrintArguments - Recursively generate the printf 
 *  				 arguments for the given tree. 
 * 
 ****/

static void VPO_GeneratePrintArguments(Expr *expr)
{
    if(!expr) 
	return;
    
    switch(expr->type) {
      case TransferLineType: {
	  TransferExpr *node = &expr->value.transfer;
	  
	  VPO_GeneratePrintArguments(node->transfers);
	  VPO_GeneratePrintArguments(node->deads);
	  printf(");\n");
	  break;
      }
	
      case ExprListType: {
	  ExprListExpr *node = &expr->value.exprlist;
	  
	  VPO_GeneratePrintArguments(node->expr);
	  if (node->list)
	      VPO_GeneratePrintArguments(node->list);
	  break;
      }
	
      case EffectType: {
	  EffectExpr *node = &expr->value.effect;
	  
	  if (node->optional)	/* Ignore optional effects */
	      break;
	  
	  VPO_GeneratePrintArguments(node->destination);
	  VPO_GeneratePrintArguments(node->source);
	  break;
      }
	
      case UseLineType: {
	  UseLineExpr *node = &expr->value.useline;
	  
	  VPO_GeneratePrintArguments(node->reg);
	  printf(");\n");
	  break;
      }
	
      case AsmLineType: {
	  AsmLineExpr *node = &expr->value.asmline;
	  
	  VPO_GeneratePrintArguments(node->expr);
	  printf(");\n");
	  break;
      }
	
      case ArgBuildAreaLineType: {
	  ArgBuildAreaExpr *node = &expr->value.argbuildarealine;
	  
	  VPO_GeneratePrintArguments(node->asize);
	  printf(");\n");
	  break;
      }

      case LabelLineType: {
	  LabelLineExpr *node = &expr->value.labelline;
	  
	  VPO_GeneratePrintArguments(node->label);
	  printf(");\n");
	  break;
      }
	
      case DeadListType: {
	  DeadListExpr *node = &expr->value.deadlist;
	  
	  VPO_GeneratePrintArguments(node->reg);
	  VPO_GeneratePrintArguments(node->list);
	  break;
      }
	
      case MemoryType: {
	  MemoryExpr *node = &expr->value.memory;
	  
	  if (node->type->type != MemoryTypeType) {
	      printf(", MEMCHR(");
	      VPO_GeneratePrintArguments(node->type);
	      printf(")");
	  }
	  else
	      VPO_GeneratePrintArguments(node->type);
	  VPO_GeneratePrintArguments(node->address);
	  break;
      }
	
      case MemoryTypeType:
	break;
	
      case RegisterType: {
	  RegisterExpr *node = &expr->value.reg;
	  
	  if (node->type->type == InternalCType) {
	      printf(", REG(");
	      VPO_GeneratePrintArguments(node->type);
	      printf(", ");
	  }
	  else {
	      printf(", ");
	      VPO_GeneratePrintArguments(node->type);
	      printf("R(");
	  }
	  VPO_GeneratePrintArguments(node->number);
	  putchar(')');
	  break;
      }
	
      case RegisterTypeType: {
	  RegisterTypeExpr *node = &expr->value.registertype;
	  
          putchar(toupper(node->type));
	  break;
      }
	
      case GroupType: {
	  GroupExpr *node = &expr->value.group;
	  
	  VPO_GeneratePrintArguments(node->expr);
	  break;
      }
	
      case BinaryType: {
	  BinaryExpr *node = &expr->value.binary;
	  
	  VPO_GeneratePrintArguments(node->left);
	  VPO_GeneratePrintArguments(node->right);
	  break;
      }
	
      case UnaryType: {
	  UnaryExpr *node = &expr->value.unary;
	  
	  VPO_GeneratePrintArguments(node->operand);
	  break;
      }
	
      case MacroType: {
	  MacroExpr *node = &expr->value.macro;
	  
	  if (node->operand)
	      VPO_GeneratePrintArguments(node->operand);
	  break;
      }
	
      case VariableType: {
	  VariableExpr *node = &expr->value.variable;
	  
	  printf(", %s(", node->type);
	  VPO_GeneratePrintArguments(node->name);
	  printf(")");
	  break;
      }
	
      case LabelType: {
	  LabelExpr *node = &expr->value.label;
	  
	  printf(", (long) (");
	  VPO_GeneratePrintArguments(node->name);
	  printf(")");
	  break;
      }
	
      case CType: {
	  CExpr *node = &expr->value.cexpr;
	  
	  printf(", (long) (%s)", node->expr);
	  break;
      }
	
      case InternalCType: {
	  CExpr *node = &expr->value.cexpr;

	  printf("%s", node->expr);
	  break;
      }

      case EscExprType: {
	  EscExprExpr *node = &expr->value.escexpr;
	  
	  printf(", %s", node->expr);
	  break;
      }
	
      case TypedConstType: {
	  TypedConstExpr *node = &expr->value.typedconstant;
	  
	  VPO_GeneratePrintArguments(node->value);
	  break;
      }
	
      case ConstTypeType:
	break;
	
      case InternalConstType: {
	  IConstExpr *node = &expr->value.iconst;

	  printf("%s", node->value);
	  break;
      }

      case IConstType:
	break;
	
      case FConstType:
	break;
	
      default:
	fprintf(stderr, "Unknown type\n");
    }
}

static void NTRO_GenerateCode(Expr *expr);


/****
 *
 *  NTRO_Generate - Generate ntro code from the given tree.
 * 
 ****/

void NTRO_Generate(Expr *expr)
{
    FirstEffect = 1;		/* Mark that no effects have been seen */
    NTRO_GenerateCode(expr);	/* Generate the ntro tree code */
    ConstSp = 0;		/* Reset the stack pointer */
}

static void NTRO_GenerateCode(Expr *expr)
{
    if(!expr) {
	printf("<NULL>");
	return;
    }
    
    switch(expr->type) {
      case TransferLineType: {
	  TransferExpr *node = &expr->value.transfer;
	  
	  printf("{");
	  NTRO_GenerateCode(node->transfers);
	  printf("}\n");
	  break;
      }
	
      case ExprListType: {
	  ExprListExpr *node = &expr->value.exprlist;
	  
	  NTRO_GenerateCode(node->expr);
	  NTRO_GenerateCode(node->list);
	  break;
      }
	
      case EffectType: {
	  EffectExpr *node = &expr->value.effect;
	  
	  /* 
	   * If any constant expressions hav been pushed on the stack, 
	   * emit code to declare the character strings and print the 
	   * value into them. 
	   */

	  if (ConstSp) {
	      int idcount = ConstSp;
	      
	      /* Emit the character string declarations. */

	      while (idcount--)
		  if (ConstStack[idcount])
		      printf("char __IntString%d[20];", idcount);

	      /* Emit the sprintf statements */

	      while (ConstSp--)
		  if (ConstStack[ConstSp])
		      printf("sprintf(__IntString%d, \"%%ld\", (long) (%s));",
			     ConstSp, ConstStack[ConstSp]);
	  }
	  ConstSp = 0;		/* Reset the stack pointer */

	  printf("emit_rtl(");
	  switch (*node->type) {

	    case 's':
	      printf("TREE_SIDE_EFFECT, ");
	      break;

	    case 'C':
	      printf("TREE_CASE, ");
	      break;

	    case 'r':
	      printf("TREE_RESERVE, ");
	      break;

	    case '+':
	      if (FirstEffect) {
		  printf("TREE_HEAD, ");	/* Emit first effect flag */
		  FirstEffect = 0;
	      }
	      else
		  if (node->optional)
		      printf("TREE_OPTION, ");  /* Emit optional flag */
		  else
		      printf("TREE_RTL, ");	/* Emit additional effect flag */
	      break;

	  }

	  if (node->destination)
	      NTRO_GenerateCode(node->destination);
	  else
	      printf("unknown()");
	  printf(", ");
	  if (node->source)
	      NTRO_GenerateCode(node->source);
	  else
	      printf("unknown()");
	  printf(");");
	  break;
      }
	
      case UseLineType: {
	  UseLineExpr *node = &expr->value.useline;
	  
	  printf("\n");
	  break;
      }
	
      case AsmLineType: {
	  AsmLineExpr *node = &expr->value.asmline;
	  
	  printf("emit_line(TREE_ASM, ");
	  NTRO_GenerateCode(node->expr);
	  printf(");\n");
	  break;
      }
	
      case LabelLineType: {
	  LabelLineExpr *node = &expr->value.labelline;
	  
	  printf("emit_");
	  NTRO_GenerateCode(node->label);
	  printf(";\n");
	  break;
      }
	
      case DeadListType: {
	  DeadListExpr *node = &expr->value.deadlist;
	  
	  break;
      }
	
      case MemoryType: {
	  MemoryExpr *node = &expr->value.memory;
	  
	  printf("memory(");
	  if (node->type->type != MemoryTypeType) {
	      printf("MEMCHR(");
	      NTRO_GenerateCode(node->type);
	      printf(")");
	  }
	  else
	      NTRO_GenerateCode(node->type);

	  printf(", ");
	  NTRO_GenerateCode(node->address);
	  printf(")");
	  break;
      }
	
      case MemoryTypeType: {
	  MemoryTypeExpr *node = &expr->value.memorytype;
	  
	  printf("%cTREG", node->type);
	  break;
      }
	
      case RegisterType: {
	  RegisterExpr *node = &expr->value.reg;
	  
	  if (node->type->type == InternalCType) {
	      printf("REG(");
	      VPO_GeneratePrintArguments(node->type);
	      printf(", ");
	  }
	  else {
	      VPO_GeneratePrintArguments(node->type);
	      printf("R(");
	  }
	  VPO_GeneratePrintArguments(node->number);
	  putchar(')');
	  break;
      }
	
      case RegisterTypeType: {
	  RegisterTypeExpr *node = &expr->value.registertype;
	  
	  putchar(toupper(node->type));
	  break;
      }
	
      case GroupType: {
	  GroupExpr *node = &expr->value.group;
	  
	  NTRO_GenerateCode(node->expr);
	  break;
      }
	
      case BinaryType: {
	  BinaryExpr *node = &expr->value.binary;
	  
	  printf("binary_op(%s, ", node->op);
	  NTRO_GenerateCode(node->left);
	  printf(", ");
	  NTRO_GenerateCode(node->right);
	  printf(")");
	  break;
      }
	
      case UnaryType: {
	  UnaryExpr *node = &expr->value.unary;
	  
	  printf("unary_op(%s, ", node->op);
	  NTRO_GenerateCode(node->operand);
	  printf(")");
	  break;
      }
	
      case MacroType: {
	  MacroExpr *node = &expr->value.macro;
	  
	  printf("mac(\"%s\", ", node->name);
	  if (node->operand)
	      NTRO_GenerateCode(node->operand);
	  else
	      printf("NONE");
	  printf(")");
	  break;
      }
	
      case VariableType: {
	  VariableExpr *node = &expr->value.variable;
	  
	  printf("%s(", node->type);
	  NTRO_GenerateCode(node->name);
	  printf(")");
	  break;
      }
	
      case LabelType: {
	  LabelExpr *node = &expr->value.label;
	  
	  printf("lbl(");
	  printf("%s", node->type);
	  printf(", (long) (");

	  NTRO_GenerateCode(node->name);
	  printf("))");
	  break;
      }
	
      case CType: {
	  CExpr *node = &expr->value.cexpr;
	  
	  /* 
	   * Substitute the converted string for the actual ctype 
	   * expression 
	   */

	  printf("__IntString%d", node->stackid);
	  break;
      }
	
      case InternalCType: {
	  CExpr *node = &expr->value.cexpr;

	  printf("%s", node->expr);
	  break;
      }

      case EscExprType: {
	  EscExprExpr *node = &expr->value.escexpr;
	  
	  printf("%s", node->expr);
	  break;
      }
	
      case TypedConstType: {
	  TypedConstExpr *node = &expr->value.typedconstant;
	  
	  printf("constant(");
	  if (node->type->type == InternalCType) {
	      printf("regdatatype(REGCHR(");
	      NTRO_GenerateCode(node->type);
	      printf("))");
	  }
	  else
	      NTRO_GenerateCode(node->type);
	  printf(", ");
	  NTRO_GenerateCode(node->value);
	  printf(")");
	  break;
      }
	
      case ConstTypeType: {
	  ConstTypeExpr *node = &expr->value.consttype;
	  
	  printf("DT_%s", node->type);
	  break;
      }
	
      case InternalConstType: {
	  IConstExpr *node = &expr->value.iconst;

	  printf("%s", node->value);
	  break;
      }

      case IConstType: {
	  IConstExpr *node = &expr->value.iconst;
	  
	  printf("\"%s\"", node->value);
	  break;
      }
	
      case FConstType: {
	  FConstExpr *node = &expr->value.fconst;
	  
	  printf("\"%s\"", node->value);
	  break;
      }
	
      default:
	fprintf(stderr, "Unknown type\n");
    }
}

