%{
/*****************************************************************************
 *
 * lex-rtl.l -
 *
 *    This is the lexical analyser for the ARTS language.
 *
 * Author: Mark Bailey
 * 11/22/93
  
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/rtlprep/lex-rtl.l,v $
 $Date: 2001/06/06 22:30:10 $
 $Revision: 1.2 $
 $Name:  $

 Log of changes:

 $Log: lex-rtl.l,v $
 Revision 1.2  2001/06/06 22:30:10  rv3s
 Support for WIN32

 Revision 1.1.1.1  2000/08/22 21:10:33  zephyr
 Created directory structure

 Revision 1.3  1998/09/29 15:46:10  jwd
 Changed name of included header file y.tab.h to parse-rtl.h.
 This name caused a problem with mks yacc.

 *
 *****************************************************************************/

#include <string.h>
#include "expr.h"
#include "op.h"
#include "parse-rtl.h"
int lineno = 1;

#ifdef WIN32                                                                  
int  yywrap() {return 1;}                                                     
#endif /* WIN32 */   

#ifndef YYCHAR
#define YYCHAR char
#endif

YYCHAR *stripquote(YYCHAR *s, YYCHAR q);
YYCHAR *newstr(YYCHAR *s);
int OpCheck(OpIndex op);
%}
%p 9000
AN	[A-Za-z0-9_]
NAN	[^A-Za-z0-9_]
NQ	([^'\n]|\\')
NDQ	([^"\n]|\\\")
NBQ	([^`\n]|\\`)
NP	[^#\n]
IDENT	{AN}{AN}*
TRUPP	[A-Z][A-Z][A-Z]
%%
^(#{NP}|{NP})*  ;
^(#{NP}|{NP})*\n { printf("%s", yytext); lineno++;}
##[ \t]*       	;
[ \t]*\n	{ return(EOL); }
\t+		{ return(DEADSEP); }
"%"		{ return(OpCheck(yylval.op = Chr_Percent)); }
"&"		{ return(OpCheck(yylval.op = Chr_Ampersand)); }
"*"		{ return(OpCheck(yylval.op = Chr_Asterix)); }
"+"		{ return(OpCheck(yylval.op = Chr_Plus)); }
","		{ return(OpCheck(yylval.op = Chr_Comma)); }
"-"		{ return(OpCheck(yylval.op = Chr_Minus)); }
"/"		{ return(OpCheck(yylval.op = Chr_Slash)); }
">"		{ return(OpCheck(yylval.op = Chr_Greater)); }
"<"		{ return(OpCheck(yylval.op = Chr_Less)); }
"?"		{ return(OpCheck(yylval.op = Chr_Question)); }
"^"		{ return(OpCheck(yylval.op = Chr_Caret)); }
"~"		{ return(OpCheck(yylval.op = Chr_Tilde)); }
"|"		{ return(OpCheck(yylval.op = Chr_VBar)); }
"!="		{ return(OpCheck(yylval.op = Chr_Exclam_Equal)); }
"<="		{ return(OpCheck(yylval.op = Chr_Less_Equal)); }
"->"		{ return(OpCheck(yylval.op = Chr_Minus_Equal)); }
"=="		{ return(OpCheck(yylval.op = Chr_Equal_Equal)); }
">="		{ return(OpCheck(yylval.op = Chr_Greater_Equal)); }
"<<"		{ return(OpCheck(yylval.op = Chr_Less_Less)); }
">>"		{ return(OpCheck(yylval.op = Chr_Greater_Greater)); }
":="		{ return(STRASSIGN); }
\`{NBQ}+\`	{ yylval.string = newstr(stripquote(yytext, '`'));	return(OPEXPR); }
GLOBAL/"["	{ return(GLOBAL); }
LOCAL/"["	{ return(LOCAL); }
LABEL_FE/"["	{ return(LABEL_FE); }
LABEL_BE/"["	{ return(LABEL_BE); }
LABEL_CE/"["	{ return(LABEL_CE); }
@C[ \t]+	{ return(CASE); }
@R[ \t]+	{ return(RESERVE); }
@U[ \t]+	{ return(USE); }
@A[ \t]+	{ return(ASM); }
@AB[ \t]+	{ return(ARGBUILDAREA); }
\'{NQ}+\'	{ yylval.string = newstr(stripquote(yytext, '\''));return(CEXPR); }
\"{NDQ}+\"	{ yylval.string = newstr(stripquote(yytext, '"'));return(SEXPR); }
-?[0-9]+	{ yylval.string = newstr(yytext); return(ICONST); }
-?[0-9]+\.[0-9]* { yylval.string = newstr(yytext); return(FCONST); }
[\(\)[\]\{\};=:] { return(yytext[0]); }
[A-Z]/"["	{ yylval.chr = *yytext; return(MTYPE); }
[a-z]/"["	{ yylval.chr = *yytext; return(RTYPE); }
[A-Z][A-Z]/{NAN} { yylval.string = newstr(yytext); return(MNAME); }
{IDENT}		{ yylval.string = newstr(yytext); return(IDENT); }
[A-Z]+/"{"	{ yylval.string = newstr(yytext); return(TYPEID); }
"MEM"/:		{ yylval.string = yytext; return(MEM); }
"REG"/:		{ yylval.string = yytext; return(REG); }
" "		;
.		{ fprintf(stderr, "line %d: character '%c' is not valid.\n", lineno, *yytext); }
%%

int OpCheck(OpIndex op)
{
    return Operators[op].binary.ntro ?
    	(Operators[op].unary.ntro ? UNBINOP : BINOP) : UNOP;
}


YYCHAR *newstr(YYCHAR *s)
{
    YYCHAR *t;

    t = (YYCHAR *) malloc(strlen(s) + 1);
    strcpy(t, s);
    return(t);
}

YYCHAR *stripquote(YYCHAR *s, YYCHAR q)
{
    static YYCHAR qstring[200];
    YYCHAR *c, *p = qstring;

    for(c = s + 1; *c; c++) {
	if (*c == '\\' && *(c + 1) == q)
	    c++;
	*p++ = *c;
    }
    *(p - 1) = '\0';
    return(qstring);
}

