/******************************************************************************
 *
 * Copyright (c) 1993 by Mark W. Bailey.
 *
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made as to
 * whether it serves any particular purpose or even works at all.
 *
 *****************************************************************************/

/******************************************************************************
 *
 * File: op.h
 *
 * Author: Mark W. Bailey
 *
 * Created: 12/1/93
 *
 * Description:
 *
 *      This file defines the data types used to define the operator 
 *      table for generating code for both the ntro and vpo versions.
 *
 * RCS Header:
 * 	$Header: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/rtlprep/op.h,v 1.1.1.1 2000/08/22 21:10:33 zephyr Exp $
 *
 * Change Log:
 *	$Log: op.h,v $
 *	Revision 1.1.1.1  2000/08/22 21:10:33  zephyr
 *	Created directory structure
 *	
 *	Revision 1.1  1998/01/29 20:54:04  jch8f
 *	Initial revision
 *	
 *	Revision 1.3  1997/07/09 14:29:14  mwb5y
 *	Removed trailing comma in OpIndex to shut up -pedantic for gcc.
 *
 *	Revision 1.2  1995/08/21 15:46:58  mwb5y
 *	Updated UTAB/BTAB macros to work with new swc/swv switches.
 *
 *	Revision 1.1  1994/01/13 16:41:36  mwb5y
 *	Initial revision
 *
 *
 *****************************************************************************/

#ifndef __OP_H__
#define __OP_H__

typedef struct {
    char *ntro;
    char *vpo;
} Ops;

typedef struct {
    Ops binary;
    Ops unary;
} OpInfo;

typedef enum {
    Chr_Percent		= 0,	/* Modulo		*/
    Chr_Ampersand 	= 1,	/* Bitwise AND		*/
    Chr_Asterix		= 2,	/* Multiplication	*/
    Chr_Plus		= 3,	/* Addition		*/
    Chr_Comma		= 4,	/* List			*/
    Chr_Minus		= 5,	/* Subtraction		*/
    Chr_Slash		= 6,	/* Division		*/
    Chr_Less		= 7,	/* Less Than		*/
    Chr_Greater		= 8,	/* Greater Than		*/
    Chr_Question 	= 9,	/* Comparison		*/
    Chr_Caret		= 10,	/* Bitwise XOR		*/
    Chr_VBar		= 11,	/* Bitwise OR		*/
    Chr_Tilde		= 12,	/* Logical NOT		*/
    Chr_Exclam_Equal 	= 13,	/* NOT Equal		*/
    Chr_Less_Equal 	= 14,	/* Less Than or Equal 	*/
    Chr_Minus_Equal 	= 15,	/* Then			*/
    Chr_Equal_Equal 	= 16,	/* Equal		*/
    Chr_Greater_Equal 	= 17,	/* Greater Than or Equal*/
    Chr_Less_Less	= 18,	/* Shift Left		*/
    Chr_Greater_Greater = 19	/* Shift Right		*/
} OpIndex;

#ifndef __OP_C__
extern OpInfo Operators[];
#endif

/* 
 * UTAB and BTAB macros are used to select the unary and binary 
 * operator information from the operator table.
 */

#define UTAB(INDEX) (swn ? Operators[(INDEX)].unary.ntro : Operators[(INDEX)].unary.vpo)
#define BTAB(INDEX) (swn ? Operators[(INDEX)].binary.ntro : Operators[(INDEX)].binary.vpo )

#endif /* __OP_H__ */
