/******************************************************************************
 *
 * Copyright (c) 1993 by Mark W. Bailey.
 *
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made as to
 * whether it serves any particular purpose or even works at all.
 *
 *****************************************************************************/

/******************************************************************************
 *
 * File: parse-rtl.y
 *
 * Author: Mark W. Bailey
 *
 * Created: 11/22/93
 *
 * Description:
 *
 *      The parser for the ART language. We build a parse tree that drives
 *      the generation of code to build rtl-trees (for ntro) or printf 
 *      statements (for vpo).
 *
 * RCS Header:
 * 	$Header: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/rtlprep/parse-rtl.y,v 1.1.1.1 2000/08/22 21:10:33 zephyr Exp $
 *
 * Change Log:
 *	$Log: parse-rtl.y,v $
 *	Revision 1.1.1.1  2000/08/22 21:10:33  zephyr
 *	Created directory structure
 *	
 *	Revision 1.3  2000/06/10 18:20:33  jwd
 *	Added some necessary includes.  Also deleted some NT specific
 *	code.
 *	
 *	Revision 1.2  1998/08/07 16:37:25  jwd
 *	Added support for ARGBUILDAREA interface. This is to
 *	support the new RTL creation interface. Also fixed some
 *	minor problems with the grammar. These were discovered
 *	by MKS yacc.
 *	Revision 1.1  1998/03/30 19:23:53  jwd
 *	Initial revision
 *	Revision 1.4  1997/07/01 13:29:29  mwb5y
 *	Added macros to the dead register list.
 *
 *	Revision 1.3  1995/08/26 14:13:38  mwb5y
 *	Added some error checking, fixed line number reporting
 *
 *	Revision 1.2  1995/08/21 15:46:14  mwb5y
 *	Changed code expander switch (default) to -c. Also added support for vpo.
 *
 * Revision 1.1  1994/01/13  16:42:49  mwb5y
 * Initial revision
 *
 *
 *****************************************************************************/

%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "parse-rtl-y.h"
#include "expr.h"
#include "gencode.h"
#include "op.h"
#define YYERROR_VERBOSE


extern int lineno;
int DoneFlag = 0;		/* Set when last statement is processed */
int swn = 0;			/* Set if ntro code is to be generated */
int swv = 0;			/* Set if vpo code is to be generated */
int swc = 0;			/* Set if vpo cgen code is to be generated */
char *infile = NULL;		/* Name of input file */
char strname[BUFSIZ];		/* Name of string to use for sprintf */

void yyerror(char *s);
int yylex();

%}

%union {
    char chr;
    char *string;
    Expr *node;
    OpIndex op;
}

%token GLOBAL LOCAL LABEL_BE LABEL_FE LABEL_CE RESERVE ASM ARGBUILDAREA
%token CASE EOL DEADSEP USE
%token <string> ICONST FCONST MNAME CEXPR SEXPR TYPEID MEM REG OPEXPR IDENT
%token STRASSIGN
%token <chr> MTYPE RTYPE
%token <op> UNBINOP UNOP BINOP

%type <chr> mtype rtype
%type <string> mname
%type <node> stmt rtl trans dlist rlist others expr mem reg mac constexpr
%type <node> global effect effects seffect sexpr fconst rmember
%type <node> local label_fe label_be label_ce label typeid 
%type <node> operand deficonst defcexpr deffconst
%type <node> iexpr iconst cexpr internconst internc interni
%nonassoc '(' ')'
%left UNBINOP BINOP UNOP OPEXPR

%%

    /* 
     * At this point, we have the complete tree. Generate the code for 
     * either ntro or vpo.
     */

start	: stmt EOL		{ if (swn)
				      NTRO_Generate($1);
				  else {
				      VPO_Generate($1, strname);
				      fflush(stdout);
				  }
				  lineno++;
				  YYACCEPT;
			      }
	| error EOL		{ YYACCEPT; }
	| 			{ DoneFlag = 1; YYACCEPT; }
	;

stmt	: IDENT { strcpy(strname, $1); } STRASSIGN rtl	{ $$ = $4; }
	| rtl			{ *strname = '\0'; $$ = $1; }
	| others
	;

rtl	: trans dlist		{ $$ = MakeTransfer($1, $2); }
	;

trans	: effects
	| RESERVE expr '=' expr { $$ = MakeEffect("r", $2, $4, 0); }
	| CASE expr '=' expr	{ $$ = MakeEffect("C", $2, $4, 0); }
	| 			{ $$ = NULL; }
	;

effects	: effect ';' effects	{ $$ = MakeExprList($1, $3); }
	| effect ';'
	| effect
	;

effect	: expr '=' expr		{ $$ = MakeEffect("+", $1, $3, 0); }
	| '[' expr '=' expr ']'	{ $$ = MakeEffect("+", $2, $4, 1); }
	| seffect
	;

seffect : expr '='		{ $$ = MakeEffect("s", $1, NULL, 0); }
	| '=' expr		{ $$ = MakeEffect("s", NULL, $2, 0); }
	;

dlist	: DEADSEP rlist 	{ $$ = $2; }
	| 			{ $$ = NULL; }
	;

rlist	: rmember rlist		{ $$ = MakeDeadList($1, $2); }
	| rmember
	;
	
rmember	: reg
	| mname			{ $$ = MakeMacro($1, NULL); }
	;

others	: USE reg		{ $$ = MakeUseLine($2); }
	| ASM sexpr		{ $$ = MakeAsmLine($2); }
	| ARGBUILDAREA expr	{ $$ = MakeArgBuildAreaLine($2); }
	| label			{ $$ = MakeLabelLine($1); }
	;

expr	: operand
	| OPEXPR expr 		{ $$ = MakeUnary($1, $2); }
	| expr OPEXPR expr	{ $$ = MakeBinary($2, $1, $3); }
	| UNOP expr 		{ $$ = MakeUnary(UTAB($1), $2); }
	| UNBINOP expr 		{ $$ = MakeUnary(UTAB($1), $2); }
	| expr BINOP expr	{ $$ = MakeBinary(BTAB($2), $1, $3); }
	| expr UNBINOP expr	{ $$ = MakeBinary(BTAB($2), $1, $3); }
	;

operand	: mem
	| reg
	| mac
	| deffconst
	| global
	| local
	| iexpr
	| label
	| '(' expr ')'			{ $$ = MakeGroup($2); }
	| sexpr
	| internc '{' constexpr '}' 	{ $$ = MakeTypedConst($1, $3); }
	| sexpr '{' constexpr '}' 	{ $$ = MakeTypedConst($1, $3); }
	| typeid '{' constexpr '}' 	{ $$ = MakeTypedConst($1, $3); }
	;

label	: label_fe
	| label_ce
	| label_be
	;

mem	: mtype '[' expr ']'		{ $$ = MakeMemory(MakeMemoryType($1), $3); }
	| MEM ':' internc '[' expr ']'	{ $$ = MakeMemory($3, $5); }
	;

reg	: rtype '[' internconst ']'	{ $$ = MakeRegister(MakeRegisterType($1), $3); }
	| REG ':' internc '[' internconst ']' { $$ = MakeRegister($3, $5); }
	| internc '[' internconst ']'	{ $$ = MakeRegister($1, $3); }
	;

mac	: mname '[' expr ']'		{ $$ = MakeMacro($1, $3); }
	| mname				{ $$ = MakeMacro($1, NULL); }
	;

global	: GLOBAL '[' internconst ']'	{ $$ = MakeVariable("gbl", $3); }
	;

local	: LOCAL '[' internconst ']'	{ $$ = MakeVariable("loc", $3); }
	;

label_ce: LABEL_CE '[' internconst ']' 	{ $$ = MakeLabel("LABEL_CE", "L0", $3); }
	;

label_fe: LABEL_FE '[' internconst ']' 	{ $$ = MakeLabel("LABEL_FE", "L", $3); }
	;

label_be: LABEL_BE '[' internconst ']' 	{ $$ = MakeLabel("LABEL_BE", "L00", $3); }
	;

internconst : internc
	| interni
	;

/* 
 * These two non-terminals (interni and internc) change the type of
 * their arguments to internal versions. These are used internally to
 * build larger expressions. For example, a register number is itself
 * not an expression, but, instead is part of a larger expression.
 */

interni : iconst	{ $$ = I2InternalConst($1); }
	;

internc : cexpr		{ $$ = C2InternalC($1); }
	;

iexpr	: deficonst
	| defcexpr
	;

constexpr : iconst
	| fconst
	| sexpr
	| cexpr
	;

/* 
 * defcexpr, deficonst and deffconst are expressions that have default 
 * type because the expressions they are derived from (cexpr, iconst 
 * and fconst) have no type explicitly defined.
 */

defcexpr : cexpr		{ $$ = MakeTypedConst(MakeConstType("WORD"), $1); }
	;

deficonst	: iconst	{ $$ = MakeTypedConst(MakeConstType("WORD"), $1); }
	;

deffconst	: fconst	{ $$ = MakeTypedConst(MakeConstType("DOBL"), $1); }
	;

cexpr	: CEXPR			{ $$ = MakeC($1); }
	| IDENT			{ $$ = MakeC($1); }
	;

iconst	: ICONST		{ $$ = MakeIConst($1); }
	;

fconst	: FCONST		{ $$ = MakeFConst($1); }
	;

sexpr	: SEXPR			{ $$ = MakeEscExpr($1); }
	;

mtype	: MTYPE
	;

rtype	: RTYPE
	;

mname	: MNAME
	;

typeid	: TYPEID		{ $$ = MakeConstType($1); }
	;
%%

#include <stdio.h>

main(int argc, char *argv[])
{
    int c;
    int arg;
    int errflag = 0;
    char *outfile = NULL;

    for (arg = 1; arg < argc; arg++) {
	if (*argv[arg] != '-')
	    if(infile)
		outfile = argv[arg];
	    else
		infile = argv[arg];
    }

    if (infile)
	freopen(infile, "r", stdin);
    else
	infile = "stdin";

    if (outfile)
	freopen(outfile, "w", stdout);

    printf("#line 1 \"%s\"\n", infile);
    /* comment this out until we have getopt on NT
    while ((c = getopt(argc, argv, "nvc")) != -1) {
	switch(c) {
	  case 'n':
	    swn = 1;
	    break;

	  case 'v':
	    swv = 1;
	    break;

	  case 'c':
	    swc = 1;
	    break;

	  case '?':
	    errflag++;
	    break;
	}
    }
    */
    if (!swn && !swv && !swc)
	swc = 1;
    if(errflag)
	exit(1);
    while(!DoneFlag)
	yyparse();
    exit(0);
    return 0;
}

void yyerror(char *s)
{
    fprintf(stderr, "\"%s\", line %d: %s\n", infile, lineno, s);
}

void error_and_exit(char *s)
{
    fprintf(stderr, "\"%s\", line %d: %s\n", infile, lineno, s);
    exit(1);
}
/* needed the line below for NT */
/* #include "c:\\bin\libmks\\yywrap.c" */
