/***********************************************************************
 * 
 * Copyright (c) 1996 Ann H. Bailey.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/vet/basicblock.c,v $
 * $Revision: 1.1.1.1 $
 ***********************************************************************/

#include <stdio.h>
#include <malloc.h>
#include "vet_io.h"
#include "basicblock.h"
#include "process_state.h"
#include "hash_functions.h"

hash_table bb_ht;
Bblockptr create_bb(void);		/* Allocates a new bb node */
void free_bb(Bblockptr);	   /* deallocates a bb node */

/**********************************************************************
 *
 * process_bb - this function initializes new basic block
 *              structure (vet's linked list structure).
 *              Returns pointer to the basic block.
 *
 **********************************************************************/
Bblockptr process_bb(Bblockptr *bb) {

    Bblockptr	current_bb;		/* bb node */
    int bbaddr;				/* bb address */
    static int number_of_bbs = 0;	/* number bb for output */

    /* Read new basicblock address */
    bbaddr = read_num();

    /* 
     * If the first basicblock has not been read, then
     * initialize the basicblock list.
     */
    if (!*bb) {
	*bb = create_bb();
	insert_hash_table(bb_ht, bbaddr, *bb);
 	(*bb)->address = bbaddr;
	(*bb)->label = read_line();
	(*bb)->number = ++number_of_bbs;
	return(*bb);
    }
    /* 
     * Since the first basic block has been read, then append this 
     * basic block to the end of the basic block list.
     */
    current_bb = *bb;
    while (current_bb->next) {
	current_bb = current_bb->next;
    }
    current_bb->down = bbaddr;
    current_bb->next = create_bb();
    insert_hash_table(bb_ht, bbaddr, current_bb->next);
    current_bb->next->address = bbaddr;
    current_bb->next->label = read_line();
    current_bb->next->prev = current_bb;
    current_bb->next->number = ++number_of_bbs;
    return(current_bb->next);

} /* end process_bb */

/***********************************************************************
 *
 * create_bb - this function allocates a new basic block node 
 *             Return pointer to the new basic block structure.
 *
***********************************************************************/
Bblockptr create_bb(void) {
    Bblockptr newnode;
    struct bblock *newvpo;

    /* Initialize basic block node */
    newnode = malloc(sizeof(Bblocknode));
    newnode->address = 0;
    newnode->next = NULL;
    newnode->prev = NULL;
    newnode->number = 0;
    newnode->left = 0;
    newnode->right = 0;
    newnode->down = 0;
    newnode->down_displayed = 0;
    newnode->label = NULL;
    newnode->rtls = NULL;
#ifdef XWINDOWS
    newnode->bb_widget = NULL;
    newnode->label_widget = NULL;
    newnode->down_widget = NULL;
    newnode->nofall_widget = NULL;
    newnode->left_widget = NULL;
    newnode->right_widget = NULL;
    newnode->rtl_widget = NULL;
    newnode->success_widget = NULL;
    newnode->rtl_scroll = NULL;
#endif
    return(newnode);

} /* end create_bb */

/*********************************************************************
 *
 * free_bb - Free basic block node.
 *
 *********************************************************************/
void free_bb(Bblockptr bb) {

    /* Free bb label */
    if (bb->label) {
	free(bb->label);
    }

    /* Note: Rtl node is freed during delete_rtl_hash_table call */
    bb->rtls = NULL;

    free(bb);
    bb = NULL;

} /* end free_bb */

void free_bb_wrapper(void *bb) {
    free_bb((Bblockptr)bb);
}

/******************************************************************
 * 
 * delete_bb_node - removes a bb node from vet bblock linked 
 *                  list, but not from hash table since it may
 *		    be used later for undoing transformation.
 *
 ******************************************************************/
void delete_bb_node(Bblockptr bb) {

    /* Set previous bb's next ptr to current bb's next ptr */
    if (bb->prev) {
	bb->prev->next = bb->next;
    }
    else {
	current_function->first_bb = bb->next;
    }
    /* Set next bb's previous ptr to current bb's prev ptr */
    if (bb->next) {
	bb->next->prev = bb->prev;
	bb->next = NULL;
	/*
	 * After setting the next node to point to bb's previous,
	 * reset bb's previous ptr.
	 */
	bb->prev = NULL;
    }

    /* Clear bb entry so if we do undo it's a clean entry */
    if (bb->label) {
	free(bb->label);
	bb->label = NULL;
    }
    if (bb->rtls){
	free_rtl(bb->rtls);
	bb->rtls = NULL;
    }
    
} /* end delete_bb_node */

/******************************************************************
 *
 * find_bb - finds the basicblock with the same address in the 
 *           vet bblock linked list.
 *           Returns a pointer to the basicblock.
 *
 ******************************************************************/
Bblockptr find_bb(int bbaddr) {
	
    Bblockptr bb;

    if (bb = (Bblockptr)search_hash_table(bb_ht, bbaddr)) {
	return(bb);
    } 
    else {
	fprintf(stderr, "ERROR vet (find_bb): missing basic block %d\n", bbaddr);
        exit(1);
    }

} /* end find_bb */

/******************************************************************
 *
 * find_bb_num - finds the basicblock with the same address in the 
 *           	 vet bblock linked list and returns the entry
 *		 doesn't exit if not found.
 *
 ******************************************************************/
Bblockptr find_bb_num(int bbaddr) {

    /* Searches hash table and returns pointer to node if bb exists */
    /* Returns NULL if not found */

    return(search_hash_table(bb_ht, bbaddr));

} /* end find_bb */

/*********************************************************************
 *
 * delete_bb_hash_table - deallocates hash table and clears entries.
 *
 *********************************************************************/
void delete_bb_hash_table(void) {

    delete_hash_table(bb_ht, free_bb_wrapper);

} /* end delete_bb_hash_table */

/*********************************************************************
 *
 * new_bb_hash_table - sets hash table entries to NULL.
 *
 *********************************************************************/
void new_bb_hash_table(void) {
    
    bb_ht = new_hash_table();

} /* end new_bb_hash_table */

Bblockptr insert_newbb_operation(int bbaddr) {

    Bblockptr bb;

    if (bb = (Bblockptr)search_hash_table(bb_ht, bbaddr))
	return(bb);
    
    bb = create_bb();
    insert_hash_table(bb_ht, bbaddr, bb);
    return(bb);
}
