/***********************************************************************
 * 
 * Copyright (c) 1998 Ann H. Bailey.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/vet/display_input.c,v $
 * $Revision: 1.1.1.1 $
 * 
 ***********************************************************************/
/* Need because compiler complains about empty file */
#include <stdio.h>		

#ifdef XWINDOWS
#include "vet.h"
#include "vpo.h"
#include "../vpo/lib/view.h"
#include "process_state.h"
#include "display_input.h"
#include "basicblock.h"
#include "vet_xinit.h"
#include <Xm/ArrowB.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h>
#include <Xm/Separator.h>
#include <Xm/List.h>
#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <Xm/PushB.h>
#include <Xm/ScrolledW.h>

void create_arrow(Widget, Bblockptr);
Widget allblocks = NULL;
void highlight_widget(Widget);
void outline_widget(Widget);
void reset_outline(Widget);
void highlight_rtlstring(Widget, Rtlptr);
void create_rtl_widget(Bblockptr, Widget);
void widget_visible(Widget, XmString, char *);

Widget area;

/*********************************************************************
 *
 * display_initial_set - this function displays the basicblocks and
 *			 rtls for each.
 *
 *********************************************************************/
void display_initial_set(void) {

    int index;
    Bblockptr bb;

    /* Create widget that holds the basic block widgets only */
    allblocks = XtVaCreateWidget("BBDisplay", 
				 xmRowColumnWidgetClass,
				 scrollwindow,
				 XmNisAligned, True,
				 XmNentryAlignment, XmALIGNMENT_CENTER,
				 NULL);


    bb = current_function->first_bb;
    current_function->first_displayed_bb = bb;
    while (bb) {
	create_basicblock_widget(bb);
	if (bb->down) {
	    /* Store down bb that's displayed */
	    bb->down_displayed = bb->down;
	    bb = find_bb(bb->down); 
	}
	else {
	    break;
	}
    }
    XtManageChild(allblocks);

} /* end display_initial_set */

/*********************************************************************
 *
 * create_basicblock_widget - creates the widget for each basicblock.
 *
 *********************************************************************/
void create_basicblock_widget(Bblockptr bb) {

    Widget pred_widget, success_widget, right_menu, separator;
    XmString labelstr, rightstr;
    char bb_display[BIGSTRING];

	/* Create a widget for displaying basic block */
	bb->bb_widget = XtVaCreateManagedWidget("BBlock", 
						xmFormWidgetClass,
						allblocks,
						NULL);
	
	/* Create the basic block label widget */
	bb->label_widget = XtVaCreateManagedWidget("Label", 
						   xmLabelWidgetClass, 
						   bb->bb_widget,
						   XmNalignment, XmALIGNMENT_BEGINNING,
						   XmNtopAttachment, XmATTACH_WIDGET,
						   XmNtopWidget, bb->bb_widget,
						   XmNleftAttachment, XmATTACH_FORM,
						   XmNrightAttachment, XmATTACH_FORM,
						   NULL);
	
    sprintf(bb_display, "%d %s", bb->number, bb->label);
    labelstr = XmStringCreateLocalized(bb_display);
    XtVaSetValues(bb->label_widget,
		  XmNlabelString, labelstr,
		  NULL);
    XmStringFree(labelstr);
	
    create_rtl_widget(bb, bb->label_widget);

    /* Create the basic block cond./uncond. arrows */
    bb->success_widget = XtVaCreateManagedWidget("Success", 
					     xmFormWidgetClass,
					     bb->bb_widget, 
					     XmNtopAttachment, XmATTACH_WIDGET,
					     XmNtopWidget, bb->rtl_widget,
					     XmNleftAttachment, XmATTACH_FORM,
					     XmNrightAttachment, XmATTACH_FORM,
					     NULL);
	
    if (bb->left) {
	/* 
	 * Left (conditional) is fall through to next bb.
	 * Display down pointer and no left ptr.
	 */
	if (bb->left == bb->down) {
	    bb->down_widget =
		XtVaCreateManagedWidget("Fallthrough", 
					xmArrowButtonWidgetClass, 
					allblocks,
					XmNshadowThickness, 0,
					XmNarrowDirection, XmARROW_DOWN,
					XmNhighlightThickness,0,
					XmNsensitive, False,
					NULL);
	}
	else {
	    bb->left_widget =
		XtVaCreateManagedWidget("Uncondjump", 
					xmArrowButtonWidgetClass, 
					bb->success_widget,
					XmNshadowThickness, 0,
					XmNarrowDirection, XmARROW_LEFT,
					XmNtopAttachment, XmATTACH_FORM,
					XmNbottomAttachment, XmATTACH_FORM,
					XmNleftAttachment, XmATTACH_FORM,
					NULL);
	    XtAddCallback(bb->left_widget, XmNactivateCallback, display_left, bb);
	
	    if (bb->down) {
		bb->nofall_widget = 
		    XtVaCreateManagedWidget("Nofallthrough",
					    xmLabelWidgetClass, 
					    allblocks,
					    XmNalignment, XmALIGNMENT_CENTER,
					    NULL);
	    }	    
	}
    } 
    if (bb->right) {
	bb->right_widget = 
	    XtVaCreateManagedWidget("Condjump", 
				    xmArrowButtonWidgetClass, 
				    bb->success_widget,
				    XmNshadowThickness, 0,
				    XmNarrowDirection, XmARROW_RIGHT,
				    XmNtopAttachment, XmATTACH_FORM,
				    XmNbottomAttachment, XmATTACH_FORM,
				    XmNrightAttachment, XmATTACH_FORM,
				    NULL);
	
	XtAddCallback(bb->right_widget, XmNactivateCallback, display_right, bb);
    }	
    XtManageChild(bb->bb_widget);

} /* end create_basicblock_widget */

/*********************************************************************
 *
 * create_rtl_widget - this function creates the strings to display
 *		       each RTL.
 *
 *********************************************************************/
void create_rtl_widget(Bblockptr bb, Widget attach_widget) {

    Widget rtls, rtl_scroll;
    XmStringTable	rtl_strings;
    int		num_visible, index, length;
    Rtlptr	rtl;

    rtl_scroll = XtVaCreateManagedWidget("RtlScroll",
					 xmScrolledWindowWidgetClass, 
					 bb->bb_widget,
					 XmNtopAttachment, XmATTACH_WIDGET,
					 XmNtopWidget, attach_widget,
					 XmNleftAttachment, XmATTACH_FORM,
					 XmNrightAttachment, XmATTACH_FORM,
					 NULL);

    rtl_strings = (XmStringTable) XtMalloc(sizeof(XmString *) * 1000);

    num_visible = 0;
    rtl = bb->rtls;
    for (rtl = bb->rtls; rtl; rtl = rtl->next) {
	if (rtl->type & rtls_to_display) {
	    /* Create rtl string for displaying and save it */
	    rtl->rtlstr = create_rtl_string(rtl);
	    if (rtl->rtlstr) {
		/* Add rtl line to widget array */
		rtl_strings[num_visible++] = rtl->rtlstr;
	    }
	}
    }

    if (num_visible == 0) {
	XtFree((char *)rtl_strings);
	rtl_strings = NULL;
    }
    bb->rtl_widget = 
	XtVaCreateManagedWidget("Rtl",
				xmListWidgetClass,	rtl_scroll,
				XmNlistSizePolicy, 	XmCONSTANT,
				XmNitems, 		rtl_strings,
				XmNitemCount, 		num_visible,
				XmNselectionPolicy,	XmEXTENDED_SELECT,
				NULL);

    XtFree((char *)rtl_strings);

} /* end create_rtl_widget */

void create_new_rtl_list(Bblockptr bb, Widget rtl_widget) {

    XmStringTable rtl_strings;
    int		num_visible;
    Rtlptr	rtl;
    char rtl_display[BIGSTRING];

    rtl_strings = (XmStringTable) XtMalloc(sizeof(XmString) * 1000);

    num_visible = 0;
    rtl = bb->rtls;
    for (rtl = bb->rtls; rtl; rtl = rtl->next) {
	if (rtl->type & rtls_to_display) {
	    /* Create rtl string for displaying and save it */
	    rtl->rtlstr = create_rtl_string(rtl);
	    if (rtl->rtlstr) {
		/* Add rtl line to widget array */
		rtl_strings[num_visible++] = rtl->rtlstr;
	    }
	}
    }
    if (num_visible == 0) {
	XtFree((char *)rtl_strings);
	rtl_strings = NULL;
    }
    /* Append new list to end of current list --
       if list is empty just inserts */
    XmListAddItems(rtl_widget, rtl_strings, num_visible, 0);

    XtFree((char *)rtl_strings);

} /* end create_new_rtl_list */

/*********************************************************************
 *
 * highlight_trans - cycles through all operations for transformation
 *		     and highlights rtls or outlines bb widgets to
 *		     be changed. Creates list of operations to be
 *		     performed.
 *
 *********************************************************************/
void highlight_trans(state_enum trans_state, char breakpoint_state) {
    Bblockptr bb;
    Rtlptr rtl;
    Operptr operation;
    char operation_text[BIGSTRING];
    XmString operation_str, newrtl_operation_str;
    char OperationVisible = FALSE;

    /* If we're skipping ahead to a breakpoint then don't highlight */
    if (breakpoint_state) {
	return;
    }

    for (operation = current_transformation->operation; 
	 operation; 
	 operation = operation->next) {
	    switch(operation->msgid) {

	      case LABELBB:

		/* highlight label */
		if (trans_state == intransbefore_state) {

 		    /* bb could be NULL if NEWBB being added */
		    if (bb = find_bb_num(operation->bbaddr)) {

			highlight_widget(bb->label_widget);
			widget_visible(bb->label_widget, NULL, &OperationVisible);

			/* Add transaction to scroll list */
			sprintf(operation_text, "Label BB# %d", bb->number);
		    }
		    else {
			/* Add transaction to be performed to scroll list */
			sprintf(operation_text, "Label BB");
		    }

		    operation_str = XmStringCreateLocalized(operation_text);

		    /* Append operation to end of list in status window */
		    XmListAddItemUnselected(operwidget, operation_str, 0);
		    XmStringFree(operation_str);
		}

		break;

	      case DELBB:
		if (trans_state == intransbefore_state) {

		    bb = find_bb(operation->bbaddr);
		    outline_widget(bb->bb_widget);

		    widget_visible(bb->bb_widget, NULL, &OperationVisible);

		    /* Add transaction to scroll list */
		    sprintf(operation_text, "Delete BB# %d", bb->number);
		    operation_str = XmStringCreateLocalized(operation_text);

		    /* Append operation to end of list in status window */
		    XmListAddItemUnselected(operwidget, operation_str, 0);
		    XmStringFree(operation_str);
		}

		break;

	      case NEWBB:
		if (trans_state == intransafter_state) {
		    bb = find_bb(operation->bbaddr);
		    outline_widget(bb->bb_widget);
		}
		else {
		    /* Add transaction to scroll list */
		    sprintf(operation_text, "Add BB");
		    operation_str = XmStringCreateLocalized(operation_text);

		    /* Append operation to end of list in status window */
		    XmListAddItemUnselected(operwidget, operation_str, 0);
		    XmStringFree(operation_str);
		}

		break;

	      case UPDATEBB:
		bb = find_bb(operation->bbaddr);
		outline_widget(bb->bb_widget);

		if (trans_state == intransbefore_state) {
		    widget_visible(bb->bb_widget, NULL, &OperationVisible);

		    /* Add transaction to scroll list */
		    sprintf(operation_text, "Update RTLs in BB# %d", bb->number);
		    operation_str = XmStringCreateLocalized(operation_text);

		    /* Append operation to end of list in status window */
		    XmListAddItemUnselected(operwidget, operation_str, 0);
		    XmStringFree(operation_str);
		}

		break;

	      case MOD_DOWN:
		bb = find_bb_num(operation->bbaddr);

		/* Check down widget not deleted */
		if (bb && bb->down_widget) {
		    highlight_widget(bb->down_widget);
		}

		if (trans_state == intransbefore_state) {

		    if (bb) {
			/* Display widget */
			widget_visible(bb->down_widget, NULL, &OperationVisible);
			/* Add transaction to scroll list */
			sprintf(operation_text, "Modify BB# %d down ptr", bb->number);
		    }
		    else {
			/* Add transaction to scroll list */
			sprintf(operation_text, "Modify new BB down ptr");
		    }

		    operation_str = XmStringCreateLocalized(operation_text);

		    /* Append operation to end of list in status window */
		    XmListAddItemUnselected(operwidget, operation_str, 0);
		    XmStringFree(operation_str);
		}

		break;

	      case MOD_LEFT:
		bb = find_bb_num(operation->bbaddr);

		/* Check left widget not deleted */
		if (bb && bb->left_widget) {
		    highlight_widget(bb->left_widget);
		}

		if (trans_state == intransbefore_state) {

		    if (bb) {
			/* Display widget */
			widget_visible(bb->left_widget, NULL, &OperationVisible);
			/* Add transaction to scroll list */
			sprintf(operation_text, "Modify BB# %d left ptr", bb->number);
		    }
		    else {
			/* Add transaction to scroll list */
			sprintf(operation_text, "Modify new BB left ptr");
		    }

		    operation_str = XmStringCreateLocalized(operation_text);

		    /* Append operation to end of list in status window */
		    XmListAddItemUnselected(operwidget, operation_str, 0);
		    XmStringFree(operation_str);
		}

		break;

	      case MOD_RIGHT:
		bb = find_bb_num(operation->bbaddr);

		/* Check right widget not deleted */
		if (bb && bb->right_widget) {
		    highlight_widget(bb->right_widget);
		}

		if (trans_state == intransbefore_state) {

		    if (bb) {
			/* Display widget */
			widget_visible(bb->right_widget, NULL, &OperationVisible);
			/* Add transaction to scroll list */
			sprintf(operation_text, "Modify BB# %d right ptr", bb->number);
		    }
		    else {
			/* Add transaction to scroll list */
			sprintf(operation_text, "Modify new BB right ptr");
		    }

		    operation_str = XmStringCreateLocalized(operation_text);

		    /* Append operation to end of list in status window */
		    XmListAddItemUnselected(operwidget, operation_str, 0);
		    XmStringFree(operation_str);
		}

		break;

	      case NEWRTL:
		bb = find_bb(operation->bbaddr);
		
		if (trans_state == intransafter_state) {

		    rtl = find_rtl(operation->rtladdr);

		    if (! (operation->rtltype & rtls_to_display)) {
			highlight_rtlstring(bb->rtl_widget, rtl);
		    }

		    /* Change string in scroll list to include number */
		    /* Find previous string */
		    sprintf(operation_text, "New line in BB# %d", bb->number);
		    operation_str = XmStringCreateLocalized(operation_text);

		    /* Create new string */
		    sprintf(operation_text, "New line #%d in BB# %d", 
		            rtl->number, bb->number);
		    newrtl_operation_str = XmStringCreateLocalized(operation_text);

		    XmListReplaceItems(operwidget, &operation_str, 1, 
		                       &newrtl_operation_str);
		    XmStringFree(operation_str);
		    XmStringFree(newrtl_operation_str);
		}
		else {

		    /* Add transaction to scroll list */
		    sprintf(operation_text, "New line in BB# %d", bb->number);
		    operation_str = XmStringCreateLocalized(operation_text);

		    /* Append operation to end of list in status window */
		    XmListAddItemUnselected(operwidget, operation_str, 0);
		    XmStringFree(operation_str);
		}

		break;

	      case DELRTL:
		if (trans_state == intransbefore_state) {
		    bb = find_bb(operation->bbaddr);
		    rtl = find_rtl(operation->rtladdr);

		    if (rtl->type & rtls_to_display) {
			
			widget_visible(bb->rtl_widget, rtl->rtlstr, &OperationVisible);
			highlight_rtlstring(bb->rtl_widget, rtl);

		    } /* if displayed rtl */
		    
		    /* Add transaction to scroll list */
		    sprintf(operation_text, "Delete line #%d", rtl->number);
		    operation_str = XmStringCreateLocalized(operation_text);

		    /* Append operation to end of list in status window */
		    XmListAddItemUnselected(operwidget, operation_str, 0);
		    XmStringFree(operation_str);
		} /* if before state */

		break;

	      case MOVERTL:
	      case MOD_RTL:
	      case MOD_RES:
	      case MOD_DEADS:
	      case MOD_SIDE:
	      case MOD_RTLLINKS:
		bb = find_bb(operation->bbaddr);

		rtl = find_rtl_num(operation->rtladdr);

		/* rtl might be NULL if modifies a newly added rtl */
		/* check if rtl displayed */
		if (rtl && (rtl->type & rtls_to_display)) {
		    highlight_rtlstring(bb->rtl_widget, rtl);
		}

		if (trans_state == intransbefore_state) {

		    /* Check if rtl displayed */
		    if (rtl && (rtl->type & rtls_to_display)) {
			widget_visible(bb->rtl_widget, rtl->rtlstr, &OperationVisible);
		    } /* end if displayed */

		    /* Add transaction to scroll list */
		    switch(operation->msgid) {
		      case MOVERTL: 
			if (rtl)
			    sprintf(operation_text, "Move line #%d", rtl->number);
			else
			    sprintf(operation_text, "Move line");

			break;

		      case MOD_RTL: 
			if (rtl)
			    sprintf(operation_text, "Mod line #%d", rtl->number);
			else
			    sprintf(operation_text, "Mod line");

			break;

		      case MOD_RES: 
			if (rtl)
			sprintf(operation_text, "Mod Res for line #%d", 
					    rtl->number);
			else
			    sprintf(operation_text, "Mod Res");

			break;

		      case MOD_DEADS: 
			if (rtl)
			sprintf(operation_text, "Mod Deads for line #%d", 
					      rtl->number);
			else
			    sprintf(operation_text, "Mod Deads");

			break;

		      case MOD_SIDE: 
			if (rtl)
			    sprintf(operation_text, "Mod Sides for line #%d", 
					     rtl->number);
			else
			    sprintf(operation_text, "Mod Sides");

			break;

		      case MOD_RTLLINKS: 
			if (rtl)
			    sprintf(operation_text, "Mod links for line #%d",
				    rtl->number);
			else
			    sprintf(operation_text, "Mod links");

			break;

		    } /* end switch */

		    operation_str = XmStringCreateLocalized(operation_text);

		    /* Append operation to end of list in status window */
		    XmListAddItemUnselected(operwidget, operation_str, 0);
		    XmStringFree(operation_str);
		}		    

		break;

	      default:
#if DEBUG >= DEBUG_LO
		/* missing operation id from case - ignore */
		fprintf(errlog, 
			"vet (highlight_trans): missing operation type: %c\n",operation->msgid);
#endif

		break;

	    } /* end switch */

	} /* end for */

} /* end highlight_trans */

/*********************************************************************
 *
 * clear_highlights - cycle through operations for transformation and
 *		      unhighlight any rtls or bbs.
 *
 ********************************************************************/
void clear_highlights(Transptr trans, state_enum trans_state, 
		      char breakpoint_state) {
    Bblockptr bb;
    Rtlptr rtl;
    Operptr operation;

    /* Verify transformation exists -- in running through breakpoint it
       could be empty OR if running breakpoint then do not show highlights.
     */
    if (!trans || breakpoint_state) {
	return;
    }

    for (operation = trans->operation; operation; operation = operation->next) {

	switch(operation->msgid) {
	  case LABELBB:
	    /* highlight label */
	    bb = find_bb(operation->bbaddr);

	    /* reverse back so doesn't appear highlighted */
	    highlight_widget(bb->label_widget);

	    break;

	  case DELBB:
	    if (trans_state == intransbefore_state) {
		bb = find_bb(operation->bbaddr);
		reset_outline(bb->bb_widget);
	    }

	    break;

	  case NEWBB:
	    if (trans_state == intransafter_state) {
		bb = find_bb(operation->bbaddr);
		reset_outline(bb->bb_widget);
	    }

	    break;

	  case UPDATEBB:
	    bb = find_bb(operation->bbaddr);
	    reset_outline(bb->rtl_widget);
	    
	    break;
	    
	  case MOD_DOWN:
	    /* Check for bb might have been deleted */
	    bb = find_bb_num(operation->bbaddr);
		
	    /* Check down widget not deleted */
	    if (bb && bb->down_widget) {
		highlight_widget(bb->down_widget);
	    }

	    break;

	  case MOD_LEFT:
	    /* Check for bb might have been deleted */
	    bb = find_bb_num(operation->bbaddr);

	    /* Check left widget not deleted */
	    if (bb && bb->left_widget) {
		highlight_widget(bb->left_widget);
	    }
	    
	    break;
	  
	  case MOD_RIGHT:
	    /* Check for bb might have been deleted */
	    bb = find_bb_num(operation->bbaddr);

	    /* Check right widget not deleted */
	    if (bb && bb->right_widget) {
		highlight_widget(bb->right_widget);
	    }

	    break;
	  
	  case DELRTL:
	    if (trans_state == intransbefore_state) {
		bb = find_bb(operation->bbaddr);
		rtl = find_rtl(operation->rtladdr);

		/* Check if rtl being displayed */
		if (rtl->type & rtls_to_display) {
		    XmListDeselectItem(bb->rtl_widget, rtl->rtlstr);
		    rtl->selected = FALSE;
		}
	    }

	    break;
	  
	  case NEWRTL:
	    /* Check if rtl being displayed */
	    if (operation->rtltype & rtls_to_display) {

		if (trans_state == intransafter_state) {
		    bb = find_bb(operation->bbaddr);
		    rtl = find_rtl(operation->rtladdr);
		    XmListDeselectItem(bb->rtl_widget, rtl->rtlstr);
		    rtl->selected = FALSE;
		}
	    }

	    break;

	  case MOVERTL:
	  case MOD_RTL:
	  case MOD_SIDE:
	  case MOD_RES:
	  case MOD_DEADS:
	  case MOD_RTLLINKS:
	    bb = find_bb(operation->bbaddr);
		
	    rtl = find_rtl_num(operation->rtladdr);

	    /* Check rtl not deleted and rtl being displayed */
	    if (rtl && (rtl->type & rtls_to_display)) {
		XmListDeselectItem(bb->rtl_widget, rtl->rtlstr);
		rtl->selected = FALSE;
	    }

	    break;

	  default:
#if DEBUG >= DEBUG_LO
	    /* missing operation id from case - ignore */
	    fprintf(errlog, 
		    "vet (clear_highlights): missing operation type: %c\n",operation->msgid);
#endif
	    break;
	
	} /* end switch */

    } /* end for */

} /* end clear_highlights */

/*********************************************************************
 *
 * highlight_widget - this function highlights widget by reverse colors
 *
 *********************************************************************/
void highlight_widget(Widget w) {

    Arg	wargs[3];
    Pixel fg,bg;

    if (!w) {
#if DEBUG >= DEBUG_LO
	fprintf(errlog, 
		"vet (highlight_widget): missing widget\n");
#endif
	return;
    }	
    /* Get the widget's current colors  */
    XtVaGetValues(w,
		  XmNforeground, &fg,
		  XmNbackground, &bg,
		  NULL);

    /* Reverse them and set the new colors  */
    XtVaSetValues(w,
		XmNforeground, bg,
		XmNbackground, fg,
		NULL);

} /* end highlight_widget */

/*********************************************************************
 *
 * outline_widget - this function outlines a widget by adding a border
 *
 *********************************************************************/
void outline_widget(Widget w) {

    if (!w) {
#if DEBUG >= DEBUG_LO
	fprintf(errlog, 
		"vet (outline_widget): missing widget\n");
#endif
	return;
    }	
    /* Increase the widget's border width */
    XtVaSetValues(w,
		  XmNborderWidth, 2,
		  NULL);

} /* end outline_widget */


/*********************************************************************
 *
 * reset_outline - this function resets a highlighted widget by 
 *		   changing the border back to 0.
 *
 *********************************************************************/
void reset_outline(Widget w) {

    if (!w) {
#if DEBUG >= DEBUG_LO
	fprintf(errlog, 
		"vet (reset_outline): missing widget\n");
#endif
	return;
    }	
    /* Decrease the widget's border width */
    XtVaSetValues(w,
		  XmNborderWidth, 1,
		  NULL);

} /* end reset_outline */

/*********************************************************************
 *
 * highlight_rtlstring - this function highlights an rtl by selecting
 *			 that line in the list widget.
 *
 *********************************************************************/
void highlight_rtlstring(Widget w, Rtlptr rtl) {

    if (rtl->selected) {
	return;
    }
    rtl->selected = TRUE;
    XtVaSetValues(w,
		  XmNselectionPolicy, XmMULTIPLE_SELECT,
		  NULL);

    XmListSelectItem(w, rtl->rtlstr, False);

    XtVaSetValues(w,
		  XmNselectionPolicy, XmEXTENDED_SELECT,
		  NULL);

} /* end highlight_rtlstring */

/*********************************************************************
 *
 * widget_visible - this function scrolls the window so that the
 *		    first operation is visible.
 *
 *********************************************************************/
void widget_visible(Widget operation_widget, XmString rtlstring, char *visible) {

    if (!*visible && operation_widget) {
	XmScrollVisible(scrollwindow, operation_widget, 0, 0);
	*visible = TRUE;
	if (rtlstring) {
	    XmListSetItem(operation_widget, rtlstring);
	}
    }	       

} /* end widget_visible */

/*********************************************************************
 *
 * reorder_down_bb - this function recalculates the down, left and
 *		     right pointers for a basicblock.
 *
 *********************************************************************/
void reorder_down_bb(Bblockptr bb) {
    
    int posnum = 0;
    Bblockptr displayedbb, newbb;

    displayedbb = current_function->first_displayed_bb;
    while(bb->address != displayedbb->address) {
	if (displayedbb->down) {
	    displayedbb = find_bb(displayedbb->down);
	    /* Increment position past down pointer and next bb */
	    posnum += 2;
	}
    }
    if (bb->left == bb->down) {
	if (bb->down) {
	    if (bb->down_widget) {
		XtManageChild(bb->down_widget);
	    }
	    else {
		bb->down_widget =  XtVaCreateManagedWidget("Fallthrough", 
							   xmArrowButtonWidgetClass, 
							   allblocks,
							   XmNshadowThickness, 0,
							   XmNarrowDirection, XmARROW_DOWN,
							   XmNhighlightThickness,0,
							   XmNpositionIndex, ++posnum,
							   NULL);
	    }
	}
	/* Don't display left widget */
	if (bb->left_widget) {
	    XtUnmanageChild(bb->left_widget);
	}
	/* Make sure nofall through widget isn't being displayed */
	if (bb->nofall_widget) {
	    XtUnmanageChild(bb->nofall_widget);
	}
    }
    else {
	/* Left is a JUMP to target and down bb is a no fall through */
	if (bb->left) {
	    if (bb->left_widget) {
		XtManageChild(bb->left_widget);
	    }
	    else {
		bb->left_widget =
		    XtVaCreateManagedWidget("Uncondjump", 
					    xmArrowButtonWidgetClass, 
					    bb->success_widget,
					    XmNshadowThickness, 0,
					    XmNarrowDirection, XmARROW_LEFT,
					    XmNtopAttachment, XmATTACH_FORM,
					    XmNbottomAttachment, XmATTACH_FORM,
					    XmNleftAttachment, XmATTACH_FORM,
					    NULL);
		XtAddCallback(bb->left_widget, XmNactivateCallback, display_left, bb);
	    }
	}
	if (bb->down) {
	    if (bb->nofall_widget) {
		XtManageChild(bb->nofall_widget);
	    }
	    else {
		bb->nofall_widget = XtVaCreateManagedWidget("Nofallthrough",
							    xmLabelWidgetClass, 
							    allblocks,
							    XmNalignment, XmALIGNMENT_CENTER,
							    XmNpositionIndex, ++posnum,
							    NULL);
	    }
	}
	if (bb->down_widget) {
	    XtUnmanageChild(bb->down_widget);
	}
    }
    /* VPO's down pointer is to next displayed basic block.
       Need to move basic blocks around so that the next
       basic block is what down is now pointing to.  This includes
       case where a new basic block was added at the end of the
       list and is now being correctly ordered. */
    newbb = find_bb(bb->down);
    /* Remove from previous location */
    if (newbb->bb_widget) {
	XtUnmanageChild(newbb->bb_widget);
	XtVaSetValues(newbb->bb_widget,
		      XmNpositionIndex, ++posnum,
		      NULL);
	/* Move to down location */
	XtManageChild(newbb->bb_widget);
    }

} /* end reorder_down_bb */

/*********************************************************************
 *
 * update_left_pointer - this function figures out whether a left
 *			 pointer is displayed based on conditional
 * 			 or unconditional jump.
 *
 *********************************************************************/
void update_left_pointer(Bblockptr bb) {


	/* 
	 * Left (conditional) is fall through to next bb.
	 * Display down pointer and no left ptr.
	 */

    	if (bb->left == bb->down) {
	    /* Do not display left arrow */
	    if (bb->left_widget) {
		XtUnmanageChild(bb->left_widget);
	    }
	    /* Modify down pointer transformation will fix arrows */
	    /* 
	     * If already created down widget then remanage
	     * otherwise create a new widget.
	     */
	    if (bb->down) {
		if (bb->down_widget) {
		    XtManageChild(bb->down_widget);
		}
		else {
		    bb->down_widget =
			XtVaCreateManagedWidget("Fallthrough", 
						xmArrowButtonWidgetClass, 
						allblocks,
						XmNshadowThickness, 0,
						XmNarrowDirection, XmARROW_DOWN,
						XmNhighlightThickness,0,
						XmNsensitive, False,
						NULL);
		}
	    }
	    /* Make sure no fallthrough widget not displayed */
	    if (bb->nofall_widget) {
		XtUnmanageChild(bb->nofall_widget);
	    }
	}
	else {
	    /* Left is a JUMP to target and down bb is a no fall through */
	    if (bb->left) {
		if (bb->left_widget) {
		    XtManageChild(bb->left_widget);
		}
		else {
		    bb->left_widget =
			XtVaCreateManagedWidget("Uncondjump", 
						xmArrowButtonWidgetClass, 
						bb->success_widget,
						XmNshadowThickness, 0,
						XmNarrowDirection, XmARROW_LEFT,
						XmNtopAttachment, XmATTACH_FORM,
						XmNbottomAttachment, XmATTACH_FORM,
						XmNleftAttachment, XmATTACH_FORM,
						NULL);
		    XtAddCallback(bb->left_widget, XmNactivateCallback, display_left, bb);
		}
	    } /* if bb->left */
	    /* No fall through case so unmanage down widget if displayed */
	    if (bb->down_widget) {
		XtUnmanageChild(bb->down_widget);
	    }
	    /* If down pointer exists then display no fall through widget */
	    if (bb->down) {
		if (bb->nofall_widget) {
		    XtManageChild(bb->nofall_widget);
		}
		else {
		    bb->nofall_widget = 
			XtVaCreateManagedWidget("Nofallthrough",
						xmLabelWidgetClass, 
						allblocks,
						XmNalignment, XmALIGNMENT_CENTER,
						NULL);
		}
	    }
	    /* 
	     * If down pointer does not exist then unmanage no fall through
	     * widget if still displayed. This could happen with the last
	     * bb because there is a bug in VPO in which the left pointer
	     * points to a bb that does not exist.
	     */
	    else {
		if (bb->nofall_widget) {
		    XtUnmanageChild(bb->nofall_widget);
		}
	    }
	} /* end if left == down */
	
} /* end update_left_pointer */

void create_right_pointer(Bblockptr bb) {

    bb->right_widget =
	XtVaCreateManagedWidget("Condjump", 
				xmArrowButtonWidgetClass, 
				bb->success_widget,
				XmNshadowThickness, 0,
				XmNarrowDirection, XmARROW_RIGHT,
				XmNtopAttachment, XmATTACH_FORM,
				XmNbottomAttachment, XmATTACH_FORM,
				XmNrightAttachment, XmATTACH_FORM,
				NULL);
	
    XtAddCallback(bb->right_widget, XmNactivateCallback, display_right, bb);
    
} /* end create_right_pointer */
#endif


