/***********************************************************************
 * 
 * Copyright (c) 1998 Ann H. Bailey.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/vet/hash_functions.c,v $
 * $Revision: 1.1.1.1 $
 ***********************************************************************/

#include <stdio.h>
#include <malloc.h>
#include "hash_functions.h"

/* Hash Table declarations */
#define HASH_MASK 0x3ff

int hash_table_index(int);		/* Hashes key - returns index */

void print_hash_table(hash_table);

hash_table new_hash_table() {

   int i;

   hash_table ht = malloc(sizeof(hash_entry *) * HASH_TABLE_SIZE);
   
   for (i = 0; i < HASH_TABLE_SIZE; i++) {
       ht[i] = NULL;
   }

    return(ht);

}


/*********************************************************************
 *
 * hash_table_index - applies hash mask and returns the index into the
 *              hash table.
 *
 *********************************************************************/
int hash_table_index(int key) {
    
    /* 
     * double word aligned (last 3 bits)
     * bblock struct is 112 bytes (last 7 bits)
     */
    return((key >> 7) & HASH_MASK);

} /* end hash_table_index */

hash_info search_hash_table(hash_table ht, int key) {

    hash_entry *table_entry;

    /*
     * Search through hashtable at this index and return a pointer
     * if matching key found; if not, return NULL.
     */
    for (table_entry = ht[hash_table_index(key)];
	 table_entry; 
	 table_entry = table_entry->next) {
	/* Check to see if the entry has been initialized */
	if (table_entry->key == key) {
	    return(table_entry->info);
	}
    }
    return NULL;

} /* end search_hash_table */

/*********************************************************************
 *
 * hash_insert
 *
 *********************************************************************/
void insert_hash_table(hash_table ht, int key, hash_info info) {

    int index;
    hash_info table_info;
    hash_entry *new_entry;

    /* Verify this rtl address does not already exist in table */
    if (! (table_info = search_hash_table(ht, key))) {

	/* Allocate node and initialize */
	new_entry = malloc(sizeof(hash_entry));
	new_entry->key = key;
	new_entry->info = info;

	index = hash_table_index(key);
	new_entry->next = ht[index];
	ht[index] = new_entry;
    }

} /* end insert_hash_table */

void print_hash_table(hash_table ht) {

    int i;
    hash_entry *table_entry;


    for (i = 0; i < HASH_TABLE_SIZE; i++) {
	printf("index %d\n", i);
	/*
	 * Search through hashtable at this index and return a pointer
	 * if matching key found; if not, return NULL.
	 */
	for (table_entry = ht[i];
	     table_entry; 
	     table_entry = table_entry->next) {
	    /* Check to see if the entry has been initialized */
	    printf("key = %d\n", table_entry->key);
	}
    }
}

/*********************************************************************
 *
 * delete_hash_info - free all node and all it's successors -
 *			effectively clearing all nodes at the same
 *			index in the hash table.
 *
 *********************************************************************/
hash_info delete_hash_info(hash_table ht, int key) {
    
    hash_info info;
    hash_entry *table_entry;
    hash_entry **prev_entry;
    int index;

    /*
     * Search through hashtable at this index and return a pointer
     * if matching key found; if not, return NULL.
     */
    index = hash_table_index(key);
    prev_entry = ht + index;
    for (table_entry = ht[index];
	 table_entry; 
	 table_entry = table_entry->next) {
	if (table_entry->key == key) {
	    *prev_entry = table_entry->next;
	    info = table_entry->info;
	    free(table_entry);
	    return(info);
	}
	prev_entry = &table_entry->next;
    }
    return(NULL);

} /* end destroy_hash_entry */

/*********************************************************************
 *
 *
 *********************************************************************/
void delete_hash_table(hash_table ht, void (*free_info)(hash_info)) {

    int i;
    hash_entry *table_entry;

    /* Clear all nodes at each index */
    for (i = 0; i < HASH_TABLE_SIZE; i++) {
	while(ht[i]) {
	    table_entry = ht[i]->next;
	    free_info(ht[i]->info);
	    free(ht[i]);
	    ht[i] = table_entry;
	}
    }
} /* end delete_hash_entry */

