/***********************************************************************
 * 
 * Copyright (c) 1998 Ann H. Bailey.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/vet/process_input.c,v $
 * $Revision: 1.1.1.1 $
 ***********************************************************************/

#define __PROCESS_INPUT_C__

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "vet.h"
#include "vpo.h"
#include "../vpo/lib/view.h"
#include "process_input.h"
#include "vet_io.h"
#include "process_state.h"
#ifdef XWINDOWS
#include "vet_xinit.h"
#include "display_input.h"
#include <Xm/MessageB.h>
#include <Xm/List.h>
#include <Xm/ScrolledW.h>
#endif

Funcptr create_function(void);
char *read_vpo_msg(void);
Phaseptr create_new_phase(void);
Transptr create_new_transformation(void);
Operptr create_new_operation(void);
void initialize_operations(Transptr);
void perform_operation(char, Operptr);
void update_rtls(Operptr);
void free_phase(Phaseptr);
void free_transformation(Transptr);
void free_operation(Operptr);
void sanity_check_num(int, int, char *);
void sanity_check_string(char *, char *, char *);
char *string_copy(char *);
void string_swap(char **, char **);
void num_swap(int *, int *);
int num_trans = 0;


Phaseptr current_phase;
Phaseptr first_phase;

char *phasenames[] = {
    "branchchain",
    "uselessjumps",
    "deadcodeelim",
    "reorderblocks",
    "mergeblocks",
    "instselect",
    "evalorderdet",
    "regassign",
    "jumps",
    "deadvarelim",
    "optloops",
    "regalloc",
    "commonsubexpr",
    "cheaperinst",
    "codemotion",
    "recurrences",
    "strengthred",
    "inductvarelim",
    "fixentryexit",
    "instsched",
    "fillslots",
    "vectorizedeads",
    "setlocallinks",
    "setgloballinks"
    };

const int NUMPHASES=(sizeof(phasenames) / sizeof(char *));

#ifdef XWINDOWS
    XmStringTable err_list;
    int err_num;
#endif

/*********************************************************************
 *
 * process_initial_set - process the rtl information passed from VPO
 *			 to create the initial set of BBs & RTLs.
 *
 *********************************************************************/
void process_initial_set(void) {

    int msgid, rtladdr;
    Bblockptr current_bb;
    Rtlptr current_rtl, searchrtl;
    int index;

    /* 
     * read a character from the socket to determine what
     * type of record follows.
     */
    while ((msgid = read_char()) != ENDINITSET) {

	switch (msgid) {

	  case LABELBB:
	  case NEWBB:
	    /* new basic block */
	    current_bb = process_bb(&current_function->first_bb);
	    break;

	  case BBLEFT:
	    /* link to unconditional successor */
	    if (current_bb) {
		current_bb->left = read_num();
	    }
#if DEBUG >= DEBUG_LO
	    else {
		fprintf(errlog,"ERROR vet (process_initial_set): BBLEFT missing bb for ptr %d\n",read_num());
	    }
#endif
	    break;

	  case BBRIGHT:
	    /* link to conditional successor */
	    if (current_bb) {
		current_bb->right = read_num();
	    }
#if DEBUG >= DEBUG_LO
	    else {
		fprintf(errlog,"ERROR vet (process_initial_set): BBRIGHT missing bb for ptr %d\n",read_num());
	    }
#endif
	    break;

	  case VET_RTL:  
	    /* new rtl - read in address*/
	    if (current_bb) {
		current_rtl = process_rtl(&current_bb->rtls);
	    }
#if DEBUG >= DEBUG_LO
	    else {
		fprintf(errlog,"ERROR vet (process_initial_set): VET_RTL missing bb for rtl %d\n",read_num());
	    }
#endif
	    break;

	  case VET_RTLLINKS:
	    /* process current rtl links */
	    index = 0;
	    while (rtladdr = read_num()) {
		if (current_rtl) {
		    current_rtl->links[index++] = rtladdr;
		}
#if DEBUG >= DEBUG_LO
                else {
		    fprintf(errlog,"ERROR vet (process_initial_set): VET_RTLLINKS missing rtl");
		}
#endif
	    } /* end while rtladdr */
	    /*
             * If we did not fill in entire array then set last
	     * index to 0 mark end of links.
	     */
	    if (current_rtl && index < MAXLINKS) {
		current_rtl->links[index] = 0;
	    }
	    break;

	  case VET_SIDE:
	    /* process current rtl side effects */
	    if (current_rtl) {
		current_rtl->side = read_line();
	    }
#if DEBUG >= DEBUG_LO
            else {
		fprintf(errlog,"ERROR vet (process_initial_set): VET_SIDE missing rtl");
	    }
#endif
	    break;

	  case VET_DEADS:
	    /* process current rtl deads list */
	    if (current_rtl) {
		current_rtl->deads = read_line();
	    }
#if DEBUG >= DEBUG_LO
            else {
		fprintf(errlog,"ERROR vet (process_initial_set): VET_DEADS missing rtl");
	    }
#endif
	    break;

	  default:
	    /* for testing purposes ignore everything else */
	    free(read_line());
#if DEBUG >= DEBUG_LO
	    fprintf(errlog,"ERROR vet (process_initial_set): unknown msgid = %c\n",msgid);
#endif
	} /* end switch */
    } /* end while */
} /* End process_initial_set */

/*********************************************************************
 *
 * initialize_transformation - process the transformations
 *                             passed from VPO.
 *
 *********************************************************************/
Transptr initialize_transformation(Phaseptr current_phase) {

    Transptr curr_trans;

    if (!current_phase->first_trans) {
	current_phase->first_trans = create_new_transformation();
	curr_trans = current_phase->first_trans;
    } 
    else {
	for (curr_trans = current_phase->first_trans;
	     curr_trans->next;
	     curr_trans = curr_trans->next)
	    ;
	curr_trans->next = create_new_transformation();
	curr_trans->next->prev = curr_trans;
	curr_trans = curr_trans->next;
    }


    curr_trans->num = ++num_trans;

    initialize_operations(curr_trans);

    return curr_trans;

} /* end initialize_transformation */

/*********************************************************************
 *
 * initialize_operations - this function reads in and stores all
 *			   the operations for a transformation.
 *
 *********************************************************************/
void initialize_operations (Transptr curr_trans) {

    char msgid;
    Operptr curr_operation = NULL;
    int linknum, index;
    
    /* 
     * read a character from the stream to determine what
     * type of record follows.
     */
    while ((msgid = read_char()) != ENDTRANS) {

	if (!curr_operation) {
	    curr_trans->operation = create_new_operation();
	    curr_operation = curr_trans->operation;
	}
	else {
	    curr_operation->next = create_new_operation();
	    curr_operation->next->prev = curr_operation;
	    curr_operation = curr_operation->next;
	}
	curr_operation->msgid = msgid;

	switch (msgid) {
	  case LABELBB:
	    /* label basic block */
	  case NEWBB:
	    /* new basic block */
	    curr_operation->bbaddr = read_num();
	    curr_operation->label = read_line();
	    break;
	  case UPDATEBB:
	    /* update rtls for basic block */
	    curr_operation->bbaddr = read_num();
	    update_rtls(curr_operation);
	    break;
	  case DELBB:
	    /* delete a basic block */
	    curr_operation->bbaddr = read_num();
	    break;
	  case MOD_LEFT:
	  case MOD_RIGHT:
	  case MOD_DOWN:
	    /* change a basic block's pointer */
	    curr_operation->bbaddr = read_num();
	    curr_operation->ptr = read_num();
	    break;
	  case NEWRTL:
	    curr_operation->rtladdr = read_num();
	    curr_operation->rtltype = read_num();
	    /* rtl value */
	    curr_operation->newline = read_line();
	    /* rtl's daddy */
	    curr_operation->bbaddr = read_num();
	    /* rtl to insert before */
	    curr_operation->ptr = read_num();
	    break;
	  case DELRTL:
	    curr_operation->rtladdr = read_num();
	    curr_operation->bbaddr = read_num();
	    break;
	  case MOVERTL:
	    curr_operation->rtladdr = read_num();
	    curr_operation->bbaddr = read_num();
	    curr_operation->ptr = read_num();
	    curr_operation->bbptr = read_num();
	    break;
	  case MOD_RTL:
	    curr_operation->rtladdr = read_num();
	    curr_operation->rtltype = read_num();
	    curr_operation->newline = read_line();
	    curr_operation->bbaddr = read_num();
	    break;
	  case MOD_SIDE:
	  case MOD_RES:
	  case MOD_DEADS:
	    curr_operation->rtladdr = read_num();
	    curr_operation->rtltype = read_num();
	    curr_operation->newline = read_line();
	    curr_operation->bbaddr = read_num();
	    break;
	  case MOD_RTLLINKS:
	    curr_operation->rtladdr = read_num();
	    curr_operation->rtltype = read_num();
	    index = 0;
	    while (linknum = read_num()) {
#if 0
		curr_operation->newlinks[index++] = (find_rtl(linknum))->number;
#endif
		curr_operation->newlinks[index++] = linknum;
	    }
	    /*
             * If we did not fill in entire array then set last
	     * index to 0 mark end of links.
	     */
	    if (index < MAXLINKS) {
		curr_operation->newlinks[index] = 0;
	    }
	    curr_operation->bbaddr = read_num();
	    break;
	  default:
	    fprintf(stderr, 
		    "ERROR vet (initialize_operations): invalid msgid = %c\n", msgid);
	    exit(1);
	} /* end switch */

    } /* end while */

} /* End initialize_operations */

/*********************************************************************
 *
 * update_rtls - special operation to change all rtls for a basic
 *		 block (special optimization).
 *
 *********************************************************************/
void update_rtls(Operptr curr_operation) {

    char msgid;
    int rtladdr, index;
    Rtlptr current_rtl = NULL;

    /* 
     * read a character from the stream to determine what
     * type of record follows.
     */
    while ((msgid = read_char()) != ENDUPDATE) {

	switch (msgid) {
	  case VET_RTL:  
	    rtladdr = read_num();
	    add_rtl_address(&curr_operation->rtls, rtladdr);

	    break;

	  default:
#if DEBUG >= DEBUG_LO
	    fprintf(errlog, "ERROR vet (update_rtls): unknown msgid %c\n", msgid);
#endif
	} /* end switch */
    } /* end while */

} /* end update_rtls */

/***********************************************************************
 * 
 * initialize_function - this function reads in a function name and 
 *                    initializes the data structure.
 *                    Return pointer to the function.
 *                    
 ***********************************************************************/
Funcptr initialize_function(void) {

    Funcptr current_func;

    current_func = create_function();
    current_func->name = read_line();
    return current_func;
    
} /* end initialize_function */

/***********************************************************************
 * 
 * create_function - this function creates a new node to represent a 
 *               function in the program.
 *               Return pointer to the new function structure.
 *
 ***********************************************************************/
Funcptr create_function(void) {

    Funcptr newnode;
    
    newnode = malloc(sizeof(Funcnode));
    newnode->first_bb = NULL;
    newnode->first_phase = NULL;
    newnode->first_displayed_bb = NULL;
    return newnode;

} /* end create_function */

/***********************************************************************
 * 
 * initialize_phase - this function reads in a phase name and 
 *                 initializes the data structure.
 *                 Return pointer to the phase structure.             
 *       
 ***********************************************************************/
Phaseptr initialize_phase(void) {

    Phaseptr curr_phase;
    int	index = 0;
    int phasenumber;
    /* 
     * If the first phase has not been read, then initialize
     * the list (setting the first_phase variable) and exit.
     */
    if (!current_function->first_phase) {
	current_function->first_phase = create_new_phase();
	phasenumber = read_num();
	while(!((1<<index++) & phasenumber));
	if (--index <= NUMPHASES) {
	    current_function->first_phase->name = phasenames[index];
	}
	else {
	    fprintf(stderr,"ERROR vet (initialize_phase): Unknown phase number %d\n",phasenumber);
	    current_function->first_phase->name = "Unknown Phase Number";
	}	    
	return current_function->first_phase;
    }
    /* 
     * Since the first phase has been read, then append this 
     * phase to the end of the list.
     */
    curr_phase = current_function->first_phase;
    while (curr_phase->next) {
	curr_phase = curr_phase->next;
    }
    curr_phase->next = create_new_phase();
    curr_phase->next->prev = curr_phase;
    phasenumber = read_num();
    while(!((1<<index++) & phasenumber));
    if (index <= NUMPHASES) {
	curr_phase->next->name = phasenames[--index];
    }
    else {
	fprintf(stderr,"ERROR vet (initialize_phase): Unknown phase number %d\n",phasenumber);
	curr_phase->next->name = "Unknown Phase Number";
    }	    
    return curr_phase->next;
    
} /* end initialize_phase */

/***********************************************************************
 * 
 * create_new_phase - this function creates a new node to represent a 
 *                    phase.
 *                    Return a pointer to the new structure.
 *
 ***********************************************************************/
Phaseptr create_new_phase(void) {

    Phaseptr newnode;
    
    newnode = malloc(sizeof(Phasenode));
    newnode->next = NULL;
    newnode->prev = NULL;
    newnode->first_trans = NULL;
    newnode->phase_read = FALSE;
    return newnode;

} /* end create_new_phase */

/***********************************************************************
 * 
 * create_new_transformation - allocates a new node to hold the 
 *			       transformations for a phase.
 *
 ***********************************************************************/
Transptr create_new_transformation(void) {

    Transptr newnode;

    newnode = malloc(sizeof(Transnode));
    newnode->next = NULL;
    newnode->prev = NULL;
    newnode->operation = NULL;
    return newnode;

} /* end create_new_transformation */

/***********************************************************************
 * 
 * create_new_operation - allocates a new node to hold the operationrmation
 *                        for an individual operation for transformation.
 *
 ***********************************************************************/
Operptr create_new_operation(void) {

    Operptr newnode;

    newnode = malloc(sizeof(Opernode));
    newnode->bbaddr = 0;
    newnode->rtladdr = 0;
    newnode->label = NULL;
    newnode->msgid = 0;
    newnode->rtltype = 0;
    newnode->bbptr = 0;
    newnode->ptr = 0;
    newnode->newlinks[0] = 0;
    newnode->number = 0;
    newnode->newline = NULL;
    newnode->rtls = NULL;
    newnode->oldrtls = NULL;
    newnode->next = NULL;
    newnode->prev = NULL;
    return newnode;

} /* end create_new_operation */

/***********************************************************************
 * 
 * free_operation - recursive function that traverses the individual 
 *                  operations for a transformation and deallocates.
 *
 ***********************************************************************/
void free_operation(Operptr curr_operation) {

    if (curr_operation->next) {
	free_operation(curr_operation->next);
    }
    if (curr_operation->label) {
	free(curr_operation->label);
    }
    if (curr_operation->newline) {
	free(curr_operation->newline);
    }

    if (curr_operation->rtls) {
	free_rtl_addresses(curr_operation->rtls);
    }

    if (curr_operation->oldrtls) {
	free_rtl_addresses(curr_operation->oldrtls);
    }
	
    free(curr_operation);

} /* end free_operation */

/***********************************************************************
 * 
 * free_transformation - recursive function that traverses the 
 *			 transformation structure and deallocates nodes.
 *
 ***********************************************************************/
void free_transformation(Transptr curr_trans) {

    if (curr_trans->next) {
	free_transformation(curr_trans->next);
    }
    if (curr_trans->operation) {
	free_operation(curr_trans->operation);
    }

    free(curr_trans);

} /* end free_transformation */

/***********************************************************************
 * 
 * free_phase - recursive function that traverses the phase structures 
 *              and deallocates each node.
 *
 ***********************************************************************/
void free_phase(Phaseptr curr_phase)
{
    if (curr_phase->next) {
	free_phase(curr_phase->next);
    }

    if (curr_phase->first_trans) {
	free_transformation(curr_phase->first_trans);
    }

    free(curr_phase);

} /* end free_phase */

/*********************************************************************
 * 
 * free_function - this function deallocates memory for the current
 *                 function structure.
 *
 *********************************************************************/
void free_function(void) {

    /* 
     * Free basic blocks when hash table is freed.
     * Necessary since all basic blocks every allocated
     * are in the hash table but some may not be linked
     * into this structure.
     */
    if (current_function->first_phase) {
	free_phase(current_function->first_phase);
    }
    if (current_function->name) {
	free(current_function->name);
    }
    free(current_function);

} /* end free_function */

/*********************************************************************
 *
 * apply_all_transformations - apply each transformation's operations.
 *
 *********************************************************************/
void apply_all_transformations(Transptr curr_trans) {

    while (curr_trans) {
	apply_single_transformation(curr_trans);
	curr_trans = curr_trans->next;
    }
} /* end apply_all_transformations */

/*********************************************************************
 *
 * apply_single_transformation - apply transformation's operations.
 *
 *********************************************************************/
void apply_single_transformation(Transptr curr_trans) {

    Operptr operation;

    /* Might have empty transformation if in phase with no trans. */
    if (!curr_trans) {
      return;
    }

    for (operation = curr_trans->operation; operation; operation = operation->next) {
	perform_operation(operation->msgid, operation);
    } /* end for */
} /* end apply_single_transformation */

/*********************************************************************
 *
 * undo_all_transformations - undo each transformation's operations.
 *
 *********************************************************************/
void undo_all_transformations(Transptr curr_trans) {

    Operptr operation;
    Transptr transformation;

    /* Undo all transformations from current transformation */
    while (curr_trans) {
	/* Could be empty transformation */
	if (curr_trans->operation) {
	    undo_single_transformation(curr_trans);
	}
	curr_trans = curr_trans->prev;
    } /* end while */
    
} /* end undo_all_transformations */

/*********************************************************************
 *
 * undo_single_transformation - undo transformation's operations.
 *
 *********************************************************************/
void undo_single_transformation(Transptr curr_trans) {

    Operptr operation;

    if (!curr_trans) {
	return;
    }
    /* Find last operation */
    for (operation = curr_trans->operation;
	 operation && operation->next;
	 operation = operation->next) 
	;
    /* Check for empty transformation */
    while (operation) {
	switch(operation->msgid) {
	  case NEWBB:
	    perform_operation(DELBB, operation);
	    break;
	  case DELBB:
	    perform_operation(NEWBB, operation);
	    break;
	  case NEWRTL:
	    perform_operation(DELRTL, operation);
	    break;
	  case DELRTL:
	    perform_operation(NEWRTL, operation);
	    break;
	  default:
	    perform_operation(operation->msgid, operation);
	    break;
	} /* end switch */
	operation = operation->prev;
    } /* end while */
    
} /* end undo_single_transformation */

/*********************************************************************
 *
 * perform_operation - applies transformation to bbs & rtls.
 *
 *********************************************************************/
void perform_operation(char msgid, Operptr operation) {

    Bblockptr bb;
    Bblockptr delete_bb;
    Rtlptr rtl, newrtl, rtlafter, searchrtl;
    Addrptr rtl_list, temp;
    int index, position;
    int rtlnext = 0;    /* rtl's next number saving in temp register */
    int prevrtl = 0; 	/* previous rtl's address */
    int length;

#ifdef XWINDOWS
    XmString newrtlstr;
#endif

    switch (msgid) {

      case LABELBB:
	/* label basic block */
	bb = find_bb(operation->bbaddr);
	string_swap(&bb->label, &operation->label);

#ifdef XWINDOWS
	/* Update BB label widget */
	newrtlstr = XmStringCreateLocalized(bb->label);
	XtVaSetValues(bb->label_widget,
		      XmNlabelString, newrtlstr,
		      NULL);
	XmStringFree(newrtlstr);
#endif

	break;

      case NEWBB:
	/* new basic block */
	/* 
         * Find last bb to add after -- the modify basic block
	 * pointer functions will order the block correctly. 
         */
	for (bb = current_function->first_bb;
	     bb && bb->next;
	     bb = bb->next)
	    ;
	bb->next = insert_newbb_operation(operation->bbaddr);
	bb->next->address = operation->bbaddr;
	bb->next->label = string_copy(operation->label);
	bb->next->prev = bb;

#ifdef XWINDOWS
	/* Check to see if widget already created */
	/* If so then probably in an undo transf. state */
	if (bb->next->bb_widget) {
	    XtManageChild(bb->next->bb_widget);
	    if (bb->next->right_widget) {
		XtManageChild(bb->next->right_widget);
	    }
	    update_left_pointer(bb->next);
	}
	else
	    create_basicblock_widget(bb->next);
#endif

	break;

      case UPDATEBB:
	/* update all rtls for this basic block */
	bb = find_bb(operation->bbaddr);

	/* Save the current rtl list if this is not an undo */
	if (!operation->oldrtls) {
	    for (rtl = bb->rtls; rtl; rtl = rtl->next) {
		add_rtl_address(&operation->oldrtls, rtl->address);
	    }
	}
	/* Point basic block to first rtl */
	bb->rtls = (find_rtl(operation->rtls->address));
	for (rtl_list = operation->rtls; rtl_list; rtl_list = rtl_list->next) {
	    Addrptr nextrtl = rtl_list->next;
	    rtl = (find_rtl(rtl_list->address));
	    rtl->next = (nextrtl) ? (find_rtl(nextrtl->address)) : NULL;
	    rtl->prev = (prevrtl) ? (find_rtl(prevrtl)) : NULL;
	    prevrtl = rtl_list->address;
	}
	/* Swap the rtl lists */
	temp = operation->oldrtls;
	operation->oldrtls = operation->rtls;
	operation->rtls = temp;

#ifdef XWINDOWS
	XmListDeleteAllItems(bb->rtl_widget);
	create_new_rtl_list(bb, bb->rtl_widget);
#endif

	break;

      case DELBB:
	/* delete a basic block */
	bb = find_bb(operation->bbaddr);
	operation->label = string_copy(bb->label);

#ifdef XWINDOWS
	if (bb->bb_widget) {
	    XtUnmanageChild(bb->bb_widget);
	}
	/*
	 * Unmanage down arrow if exists since not
	 * part of bb widget.
	 */
	if (bb->down_widget) {
	    XtUnmanageChild(bb->down_widget);
	}
	if (bb->nofall_widget) {
	    XtUnmanageChild(bb->nofall_widget);
	}
#endif

	delete_bb_node(bb);

	break;

      case MOD_LEFT:
	/* modify basic block's left pointer */
	bb = find_bb(operation->bbaddr);
	num_swap(&bb->left, &operation->ptr);

#ifdef XWINDOWS
	update_left_pointer(bb);
#endif

	break;

      case MOD_RIGHT:
	/* modify basic block's right pointer */
	bb = find_bb(operation->bbaddr);
	num_swap(&bb->right, &operation->ptr);

#ifdef XWINDOWS
	/* If right widget exists and new pointer
	   is NULL then unmanage widget */
	if (bb->right_widget) {
	    if (bb->right) {
		XtManageChild(bb->right_widget);
	    }
	    else {
		XtUnmanageChild(bb->right_widget);
	    }
	}
	else {
	    /* If right widget does not exist but
	       new pointer is not NULL then create. */
	    if (bb->right) {
		create_right_pointer(bb);
	    }
	}
#endif

	break;

      case MOD_DOWN:
	/* modify basic block's down pointer */
	bb = find_bb(operation->bbaddr);
	num_swap(&bb->down, &operation->ptr);

#ifdef XWINDOWS
	update_left_pointer(bb);
#endif

	break;

      case NEWRTL:
	/* insert new rtl */
	bb = find_bb(operation->bbaddr);
	/*
	 * if ptr is NULL then insert rtl at end of list
	 * otherwise insert before ptr.
	 */
	if (operation->ptr) {
	    rtl = find_rtl(operation->ptr);
	    newrtl = insert_newrtl_operation(operation->rtladdr);
	    newrtl->address = operation->rtladdr;
	    newrtl->line = string_copy(operation->newline);
	    newrtl->type = operation->rtltype;
	    newrtl->next = rtl;

	    if (rtl->prev) {
		newrtl->prev = rtl->prev;
		rtl->prev->next = newrtl;
	    }
	    else {
		/* rtl is first in list -- verify */
		if (bb->rtls == rtl) {
		    bb->rtls = newrtl;
		} 
		else {
		    fprintf(stderr, "ERROR vet (perform_operation): newrtl not first\n");
		    exit(1);
		} /* if first rtl */
		newrtl->prev = NULL;
	    } /* if rtl->prev */
	    rtl->prev = newrtl;

#ifdef XWINDOWS
	    if (newrtl->type & rtls_to_display) {
		/* Get position to insert before - checking for displayed rtls */
		while (!rtl->rtlstr) {
		    rtl = rtl->next;
		}
		position = XmListItemPos(bb->rtl_widget, rtl->rtlstr);

		/* Create string for list widget if not already created */
		/* may be an undo operation */
		if (!newrtl->rtlstr) {
		    newrtl->rtlstr = create_rtl_string(newrtl);
		}

		if (newrtl->rtlstr) {
		    /* Add new rtl before position */
		    XmListAddItemUnselected(bb->rtl_widget, newrtl->rtlstr, position);
		}
		else {
		    fprintf(stderr, "WARNING: vet (perform_operation): empty rtl xstring\n");
		}

	    } /* if displayed rtl */
#endif	    
	}
	else {
	    /* if no rtl to add before then add at the end */
	    if (bb->rtls) {
		for (rtl = bb->rtls; 
		     rtl && rtl->next; 
		     rtl = rtl->next)
		    ;
		rtl->next = insert_newrtl_operation(operation->rtladdr);
		rtl->next->address = operation->rtladdr;
		rtl->next->line = string_copy(operation->newline);
		rtl->next->type = operation->rtltype;
		rtl->next->prev = rtl;

#ifdef XWINDOWS
		if (rtl->next->type & rtls_to_display) {
		    /* Create string for list widget if string d.n.e. */
		    if (!rtl->next->rtlstr) {
			rtl->next->rtlstr = create_rtl_string(rtl->next);
		    }

		    if (rtl->next->rtlstr) {
			/* Add new rtl at end of list */
			XmListAddItemUnselected(bb->rtl_widget, rtl->next->rtlstr, 0);
		    }
		    else {
			fprintf(stderr, "WARNING: vet (perform_operation): empty rtl xstring\n");
		    }

		} /* if displayed rtl */
#endif	    
	    }
	    else {
		/* if no rtl then first rtl for bb */
		bb->rtls = insert_newrtl_operation(operation->rtladdr);
		bb->rtls->address = operation->rtladdr;
		bb->rtls->line = string_copy(operation->newline);
		bb->rtls->type = operation->rtltype;

#ifdef XWINDOWS
		if (bb->rtls->type & rtls_to_display) {
		    /* Create string for list widget */
		    if (!bb->rtls->rtlstr) {
			bb->rtls->rtlstr = create_rtl_string(bb->rtls);
		    }
		    if (bb->rtls->rtlstr) {
			/* Add new rtl at end of list, but list is empty so will be first*/
			XmListAddItemUnselected(bb->rtl_widget, bb->rtls->rtlstr, 0);
		    }
		    else {
			fprintf(stderr, "ERROR vet (perform_operation): empty rtl xstring\n");
			exit(1);
		    }
		} /* if displayed rtl */
#endif	    
	    }
	}

	break;

      case DELRTL:
	/* Find basic block first */
	bb = find_bb(operation->bbaddr);
	/* Search rtls for rtl to delete */
	rtl = find_rtl(operation->rtladdr);

	/* Save information for undo */
	operation->rtltype = rtl->type;
	operation->newline = string_copy(rtl->line);
	operation->ptr = (rtl->next) ? rtl->next->address : 0;
	operation->number = rtl->number;

#ifdef XWINDOWS
	if (rtl->type & rtls_to_display) {
	    XmListDeleteItem(bb->rtl_widget, rtl->rtlstr);
	}
#endif

	/* Remove node from list */
	delete_rtl_node(bb, rtl);

	break;

      case MOVERTL:
	/* Find basic block first */
	bb = find_bb(operation->bbaddr);
	/* Search rtls for rtl to move */
	rtl = find_rtl(operation->rtladdr);

#ifdef XWINDOWS
	if (rtl->type & rtls_to_display) {
	    /* Delete rtl at current position */
	    XmListDeleteItem(bb->rtl_widget, rtl->rtlstr);
	}
#endif

	/* Set previous rtl's next ptr to current rtl's next ptr */
	if (rtl->prev) {
	    rtl->prev->next = rtl->next;
	}
	else {
	    /* 
	     * If no previous rtl then it may be the first
	     * rtl in the basicblock, therefore we need to
	     * reset the bb's rtl pointer to the next rtl
	     * before removing this rtl.
	     */
	    if (bb->rtls->address == operation->rtladdr) {
		bb->rtls = rtl->next;
	    } 
	    else {
		fprintf(stderr, "ERROR vet (perform_operation): movertl's prev rtl not first rtl\n");
		exit(1);
	    }
	}
	
	/* Set next rtl's previous ptr to current rtl's prev ptr */
	if (rtl->next) {
	    rtl->next->prev = rtl->prev;
	}
	
	/* Save rtl's next pointer for undo */
	rtlnext = (rtl->next) ? rtl->next->address : 0;

	/* insert rtl in other basic block list */
	bb = find_bb(operation->bbptr);
	/*
	 * if ptr is NULL then insert rtl at end of list
	 * otherwise insert before ptr.
	 */
	if (operation->ptr) {
	    rtlafter = find_rtl(operation->ptr);
	    
	    /* if rtlafter's prev is NULL then first in list */
	    if (rtlafter->prev) {
		rtlafter->prev->next = rtl;
		rtl->prev = rtlafter->prev;
	    } 
	    else {
		/* verify first in list */
		if (bb->rtls == rtlafter) {
		    bb->rtls = rtl;
		    rtl->prev = NULL;
		}
		else {
		    fprintf(stderr, "ERROR vet (perform_operation): movertl rtl to move before is not first in list\n");
		    exit(1);
		}
	    }
	    rtl->next = rtlafter;
	    rtlafter->prev = rtl;
#ifdef XWINDOWS
	    /* Check if this rtl is to be displayed */
	    if (rtl->type & rtls_to_display) {

		/* Need to search for next displayed rtl to insert before */
		for (searchrtl = rtlafter; searchrtl; searchrtl = searchrtl->next) {

		    if ((searchrtl->type & rtls_to_display) && searchrtl->line) {
			position = XmListItemPos(bb->rtl_widget, searchrtl->rtlstr);
			XmListAddItemUnselected(bb->rtl_widget, rtl->rtlstr, position);

			break;
		    }
		}
		/* If no more displayed rtls then insert at end of list */
		if (!searchrtl) {
		    XmListAddItemUnselected(bb->rtl_widget,rtl->rtlstr, 0);
		}
	    } /* if displayed rtl */
#endif
	}
	/* Insert rtl at end of list */
	else {
	    for (rtlafter = bb->rtls; 
		 rtlafter && rtlafter->next;
		 rtlafter = rtlafter->next)
		;
	    if (rtlafter) {
		rtlafter->next = rtl;
		rtl->prev = rtlafter;
	    }
	    else {
		/* First rtl for bb */
		bb->rtls = rtl;
		rtl->prev = NULL;
	    } /* end if rtlafter */
	    rtl->next = NULL;
#ifdef XWINDOWS
	    if (rtl->type & rtls_to_display) {
		/* Add rtl to end of list */
		XmListAddItemUnselected(bb->rtl_widget,rtl->rtlstr, 0);
	    }
#endif
	} /* end if operation->ptr */

	/* Save information for undo */
	/* Swap basic blocks */
	num_swap(&operation->bbaddr, &operation->bbptr);
	/* Change the before pointer to what was rtl's next pointer */
	operation->ptr = rtlnext;

	break;

      case MOD_RTL:
      case MOD_SIDE:
      case MOD_RES:
      case MOD_DEADS:
      case MOD_RTLLINKS:
	bb = find_bb(operation->bbaddr);
	rtl = find_rtl(operation->rtladdr);

	rtl->type = operation->rtltype;

	switch (operation->msgid) {
	  case MOD_RTL:
	  case MOD_RES:
	    string_swap(&rtl->line, &operation->newline);

#ifdef XWINDOWS
	    /* Check if rtl to be displayed */
	    if (rtl->type & rtls_to_display) {
		newrtlstr = create_rtl_string(rtl);
		XmListReplaceItems(bb->rtl_widget, &rtl->rtlstr, 1, &newrtlstr);
		rtl->rtlstr = newrtlstr;
	    }
#endif
	    break;

	  case MOD_SIDE:
	    string_swap(&rtl->side, &operation->newline);

#ifdef XWINDOWS
	    /* Check if rtl to be displayed */
	    if (rtl->type & rtls_to_display) {
		newrtlstr = create_rtl_string(rtl);
		XmListReplaceItems(bb->rtl_widget, &rtl->rtlstr, 1, &newrtlstr);
		rtl->rtlstr = newrtlstr;
	    }
#endif
	    break;

	  case MOD_DEADS:
	    string_swap(&rtl->deads, &operation->newline);

#ifdef XWINDOWS
	    /* Check if rtl to be displayed */
	    if (rtl->type & rtls_to_display) {
		newrtlstr = create_rtl_string(rtl);
		XmListReplaceItems(bb->rtl_widget, &rtl->rtlstr, 1, &newrtlstr);
		rtl->rtlstr = newrtlstr;
	    }
#endif
	    break;

	  case MOD_RTLLINKS:
	    for (index = 0; 
		 index < MAXLINKS;
		 index++) {
		num_swap(&rtl->links[index], &operation->newlinks[index]);
	    }

#ifdef XWINDOWS
	    /* Check if rtl to be displayed */
	    if (rtl->type & rtls_to_display) {
		newrtlstr = create_rtl_string(rtl);
		XmListReplaceItems(bb->rtl_widget, &rtl->rtlstr, 1, &newrtlstr);
		rtl->rtlstr = newrtlstr;
	    }
#endif

	    break;

	  default:
	    break;

	} /* end switch */

	break; 

      default:
	/* missing operation id from case - ignore */
#if DEBUG >= DEBUG_LO
	fprintf(errlog, "ERROR vet (perform_operation): missing operation type\n");
#endif
	break;

    } /* end switch */

} /* end perform_operation */

/*********************************************************************
 *
 * process_sanity - special debug function to process sanity check
 *		    messages from VPO to determine if we're still in
 *		    sync with our basic block structure.
 *
 *********************************************************************/
void process_sanity(void) {

    char msgid;
    Bblockptr current_bb = NULL;
    int num, index;
    char *label, *rtl;
    char msg[BIGSTRING];
    Rtlptr current_rtl = NULL, temp_rtl = NULL;


    while ((msgid = read_char()) != ENDSANITY) {
	switch(msgid) {
	  case BEGINFUNCTION:
#ifdef XWINDOWS
	    err_list = (XmStringTable) XtMalloc(sizeof(XmString *) * 1000);
	    err_num = 0;
#endif
	    label = read_line();
	    sanity_check_string(label, current_function->name, "func name");
	    free(label);
	    break;

	  case NEWBB:
	    /* If already checking basic block verify that we reached
	       the end of the rtls before going to next basic block. */
	    if (current_rtl) {
		if (current_rtl->next) {
#if DEBUG >= DEBUG_LO
#ifdef XWINDOWS
		    sprintf(msg, "BB #%d: Vet contains more rtls: rtl #%d",
			    current_bb->address, current_rtl->next->address);
		    err_list[err_num++] = XmStringCreateLocalized(msg);
#else
		    fprintf(errlog, "ERROR vet (sanity_check): newbb: rtl ptr not null: rtl #%d, bb #%d\n",current_rtl->next->address, current_bb->address);
#endif
#endif
		    current_rtl = current_rtl->next;
		}
	    }
	    /* Go to next BB if this isn't the first BB for the func. */
	    if (current_bb) {
	       current_bb = find_bb_num(current_bb->down);
	    }
	    else {
	       current_bb = current_function->first_bb;
	    }
	    if (current_bb) {
		/* Clear previous BB's current rtl pointer */
		current_rtl = NULL;
		num = read_num();
		sanity_check_num(num, current_bb->address, "bbnum");
		label = read_line();
		sanity_check_string(label, current_bb->label, "label");
		free(label);
	    }
	    else {
		/* Clear input buffer for this rtl */
		read_num();
		free(read_line());
#if DEBUG >= DEBUG_LO
#ifdef XWINDOWS
		sprintf(msg, "Vet couldn't find next bb");
		    err_list[err_num++] = XmStringCreateLocalized(msg);
#else
		fprintf(errlog, "ERROR vet (sanity_check): newbb: no current bb\n");
#endif
#endif
	    } /* if not current_bb */
	    break;

	  case BBLEFT:
	    num = read_num();
	    if (current_bb) {
		sanity_check_num(num, current_bb->left, "bbleft");
	    }
#if DEBUG >= DEBUG_LO
	    else {
#ifdef XWINDOWS
		sprintf(msg, "Sanity check bbleft: no current bb");
		    err_list[err_num++] = XmStringCreateLocalized(msg);
#else
		fprintf(errlog, "ERROR vet (sanity_check): bbleft: no current bb\n");
#endif
	    }
#endif
	    break;

	  case BBRIGHT:
	    num = read_num();
	    if (current_bb) {
		sanity_check_num(num, current_bb->right, "bbright");
	    }
#if DEBUG >= DEBUG_LO
	    else {
#ifdef XWINDOWS
		sprintf(msg, "Sanity check bbright: no current bb");
		    err_list[err_num++] = XmStringCreateLocalized(msg);
#else
		fprintf(errlog, "ERROR vet (sanity_check): bbright: no current bb\n");
#endif
	    }
#endif
	    break;

	  case VET_RTL:
	    /* process rtls */
	    if (current_rtl) {
		current_rtl = current_rtl->next;
	    }
	    else {
	       current_rtl = current_bb->rtls;
	    }
	    if (current_rtl) {
		num = read_num();
		sanity_check_num(num, current_rtl->address, "rtl address");
		num = read_num();
		sanity_check_num(num, current_rtl->type, "rtl type");
		rtl = read_line();
		sanity_check_string(rtl, current_rtl->line, "rtl value");
		free(rtl);
	    }
	    else {
#if DEBUG >= DEBUG_LO
#ifdef XWINDOWS
		sprintf(msg, "Sanity check rtl: no next rtl pointer");
		    err_list[err_num++] = XmStringCreateLocalized(msg);
#else
		fprintf(errlog, "ERROR vet (sanity_check): rtl: No next rtl pointer\n");
#endif
#endif
		/* Clear input buffer for this rtl */
		read_num();
		read_num();
		free(read_line());
	    }
	    break;
	  case VET_RTLLINKS:
	    /* process current rtl links */
	    if (current_rtl) {
		index = 0;
		while (num = read_num()) {
		    if (current_rtl->links[index]) {
#if 0
			if (temp_rtl = find_rtl_num(num)) {
#endif
			    sanity_check_num(num, current_rtl->links[index++], "rtl links");
#if 0
			}
			else {
			    sanity_check_num(0, current_rtl->links[index++], "rtl links");
			}
#endif
		    }
#if DEBUG >= DEBUG_LO
                    else {
#ifdef XWINDOWS
			sprintf(msg, "Sanity check links: VET = no link at index %d VPO = %d",
				index, num);
		    err_list[err_num++] = XmStringCreateLocalized(msg);
#else
			fprintf(errlog, "ERROR vet (sanity_check): links: VET =  no link at index %d VPO = %d\n", index, num);
#endif
		    }
#endif
		}
		/* Verify the rtl does not have any more links */
		while (index < MAXLINKS && current_rtl->links[index]) {
#if DEBUG >= DEBUG_LO
#ifdef XWINDOWS
		    sprintf(msg, "Sanity check number links: VPO contains more links than VET: %d",
			    current_rtl->links[index]);
		    err_list[err_num++] = XmStringCreateLocalized(msg);
#else
		    fprintf(errlog, "ERROR vet (sanity_check): links: VPO contains more links than VET: %d\n", current_rtl->links[index]);
#endif
#endif
		    index++;
		}
	    }
#if DEBUG >= DEBUG_LO
            else {
#ifdef XWINDOWS
		sprintf(msg, "Sanity check links: no current rtl");
		    err_list[err_num++] = XmStringCreateLocalized(msg);
#else
		fprintf(errlog, "ERROR vet (sanity_check): links: no current rtl\n");
#endif
	    }
#endif
	    break;
	  case VET_SIDE:
	    /* process current rtl side effects */
	    rtl = read_line();
	    if (current_rtl) {
		if (current_rtl->side) {
		    sanity_check_string(rtl, current_rtl->side, "rtl side effects");
		}
#if DEBUG >= DEBUG_LO
		else {
#ifdef XWINDOWS
		    sprintf(msg, "Sanity check sides: VPO = %s VET = NULL", rtl);
		    err_list[err_num++] = XmStringCreateLocalized(msg);
#else
		    fprintf(errlog, "ERROR vet (sanity_check): sides: VPO = %s VET = NULL\n", rtl);
#endif
		}
#endif
	    }
#if DEBUG >= DEBUG_LO
	    else {
#ifdef XWINDOWS
		sprintf(msg, "Sanity check sides: no current rtl");
		    err_list[err_num++] = XmStringCreateLocalized(msg);
#else
		fprintf(errlog, "ERROR vet (sanity_check): sides: no current rtl\n");
#endif
	    }
#endif
	    free(rtl);
	    break;
	  case VET_DEADS:
	    /* process current rtl deads list */
	    rtl = read_line();
	    if (current_rtl) {
		if (current_rtl->deads) {
		    sanity_check_string(rtl, current_rtl->deads, "rtl deads list");
		}
#if DEBUG >= DEBUG_LO
		else {
#ifdef XWINDOWS
		    sprintf(msg, "Sanity check deads: VPO = %s VET = NULL", rtl);
		    err_list[err_num++] = XmStringCreateLocalized(msg);
#else
		    fprintf(errlog, "ERROR vet (sanity_check): deads: VPO = %s VET = NULL\n", rtl);
#endif
		}
#endif
	    } 
#if DEBUG >= DEBUG_LO
	    else {
#ifdef XWINDOWS
		sprintf(msg, "Sanity check deads: no current rtl");
		    err_list[err_num++] = XmStringCreateLocalized(msg);
#else
		fprintf(errlog, "ERROR vet (sanity_check): deads: no current rtl\n");
#endif
	    }
#endif
	    free(rtl);
	    break;
	  default:
	    fprintf(stderr, "ERROR vet (process_sanity): wrong msgid = %c\n", msgid);
	    exit(1);

	} /* end switch */
    } /* end while */

} /* end process_sanity */

/*********************************************************************
 *
 * sanity_check_num - sanity check function to compare numbers and
 *		      log errors.
 *
 *********************************************************************/
void sanity_check_num(int vponum, int vetnum, char *check) {

#ifdef XWINDOWS
    char display[BIGSTRING];
#endif

   if (vponum != vetnum) {
#if DEBUG >= DEBUG_LO
#ifdef XWINDOWS
    sprintf(display, "Sanity check number %s VPO = %d VET = %d",
	    check, vponum, vetnum);
    err_list[err_num++] = XmStringCreateLocalized(display);
#else
      fprintf(errlog, "ERROR vet (sanity_check_num): %s VPO = %d VET = %d\n",
	      check, vponum, vetnum);
#endif
#endif
   }

} /* end sanity_check_num */

/*********************************************************************
 *
 * sanity_check_string - special sanity function to compare strings
 *			 and log errors.
 *
 *********************************************************************/
void sanity_check_string(char *vpostring, char *vetstring, char *check) {

#ifdef XWINDOWS
    char display[BIGSTRING];
#endif

   if (strcmp(vpostring, vetstring)) {
#if DEBUG >= DEBUG_LO
#ifdef XWINDOWS
    sprintf(display, "Sanity check string %s VPO = %s VET = %s",
	    check, vpostring, vetstring);
    err_list[err_num++] = XmStringCreateLocalized(display);
#else
      fprintf(errlog, "ERROR vet (sanity_check_string): %s VPO = %s VET = %s\n",
	      check, vpostring, vetstring);
#endif
#endif
   }

} /* end sanity_check_string */

/*********************************************************************
 *
 * string_copy - copies a string allocating new memory for storage.
 *
 *********************************************************************/
char *string_copy(char *oldstring) {

    /* Allocate space for new string including EOS marker */
    char *newstring = malloc(strlen(oldstring) + 1);
    
    return strcpy(newstring,oldstring);

} /* end string_copy */

/*********************************************************************
 *
 * string_swap - swap two strings.
 *
 *********************************************************************/
void string_swap(char **string1, char **string2) {

    char *temp;

    temp = *string1;
    *string1 = *string2;
    *string2 = temp;

} /* end string_swap */

/*********************************************************************
 *
 * num_swap - swap two integers.
 *
 *********************************************************************/
void num_swap(int *num1, int *num2) {

   int temp;

   temp = *num1;
   *num1 = *num2;
   *num2 = temp;

} /* end num_swap */

#ifdef XWINDOWS
/********************************************************************
 *
 * create_rtl_string - creates string for displaying rtl
 *
 ********************************************************************/
XmString create_rtl_string(Rtlptr rtl) {

    char rtl_display[BIGSTRING];
    XmString str;
    int index, length, rtl_num;

    /* Check if user wants to display anything other than rtline */
    if (rtl->type & rtls_to_display) {
	/* Check if empty rtl line */
	if ((!(rtl->line) || !*(rtl->line)) && !display_sides && !display_deads) {
	    return(NULL);
	}
	/* Add rtl number to rtl line */
	sprintf(rtl_display, "%4d    ", rtl->number);
	/* Append rtl links */
	for (index = 0; rtl->links[index] != 0; index++) {
	    /* Replace all rtl addresses with VET rtl numbers */
	    rtl_num = (find_rtl(rtl->links[index]))->number;
	    sprintf(rtl_display + strlen(rtl_display), " %3d", rtl_num);
	}
	if (index == 1) {
	    strcat(rtl_display, "    ");
	}
	if (index == 0) {
	    strcat(rtl_display, "        ");
	}
	    /* Append rtl value */
	if (rtl->line && *(rtl->line)) {
	    strcat(rtl_display, " ");
	    strcat(rtl_display, rtl->line);
	}
	    /* Check to see if user want to see side effects */
	    if (display_sides && rtl->side && *(rtl->side)) {
		strcat(rtl_display, " Side: ");
		strcat(rtl_display, rtl->side);
	    }
	    /* Check to see if user wants to see deads */
	    if (display_deads && rtl->deads && *(rtl->deads)) {
		strcat(rtl_display, " Deads: ");
		strcat(rtl_display,rtl->deads);
	    }
	str = XmStringCreateLocalized(rtl_display);
	return(str);
    }
    return(NULL);

} /* end create_rtl_string */

void sanity_dialog() {

    char msg[BIGSTRING];
    Widget dialog;
    Widget error_scroll;
    Arg args[2];
    XmString message = XmStringCreateLocalized("Sanity Check Errors:");
    int index;

    if (!err_num) {
	return;
    }

    XtSetArg(args[0], XmNmessageString, message);
    XtSetArg(args[1], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);
    dialog = XmCreateMessageDialog(scrollwindow, "SanityError", args, 2);
    XmStringFree(message);

    /* No help is available */
    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
    /* Don't need the cancel button */
    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_CANCEL_BUTTON));
    
    error_scroll = XtVaCreateManagedWidget("ErrorScroll",
					   xmScrolledWindowWidgetClass, 
					   dialog,
					   NULL);
    
    XtVaCreateManagedWidget("Errors",
			    xmListWidgetClass,	error_scroll,
			    XmNlistSizePolicy, 	XmRESIZE_IF_POSSIBLE,
			    XmNitems, 		err_list,
			    XmNitemCount,	err_num,
			    NULL);

    XtManageChild(dialog);
    
    for (index = 0; index < err_num; index++) {
	XmStringFree(err_list[index]);
    }
    XtFree((void *)err_list);

} /* end sanity_dialog */
#endif
