/***********************************************************************
 * 
 * Copyright (c) 1998 Ann H. Bailey.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/vet/process_state.c,v $
 * $Revision: 1.1.1.1 $
 * 
 ***********************************************************************/
#define __PROCESS_STATE_C__

#include <stdio.h>
#include <stdlib.h>

#include "vet.h"
#include "vpo.h"
#include "../vpo/lib/view.h"
#include "process_input.h"
#include "process_state.h"
#include "basicblock.h"
#include "rtl.h"

#ifdef XWINDOWS
#include "vet_callbacks.h"
#endif

Funcptr current_function = NULL;
Phaseptr current_phase = NULL;
Transptr current_transformation = NULL;

/*********************************************************************
 *
 * process_function - start state; reads until hits endview from vpo
 *		      which signals vpo is finished.
 *		      NOTE: not used in XWINDOWS version
 *
 *********************************************************************/
void process_function(void) {

   char		msgid;	/* Record id from stream */
   
   /* ENDVIEW signals that VPO has finished */
   while ((msgid = read_char()) != ENDVIEW) {
       
       switch (msgid) {
	   
	 case BEGINFUNCTION:
	   /* 
	    * BEGINFUNCTION signals VPO has started optimizing
	    * a new function.
	    */
#if DEBUG >= DEBUG_MED
	 fprintf(errlog, "BEGINFUNCTION\n");
#endif

	 current_function = initialize_function();

#if DEBUG >= DEBUG_MED
	 fprintf(errlog, "Function: %s\n", current_function->name);
#endif

	 write_char(SANITY_CHECK);
	 write_char('T');

	 process_initial_set();
	 process_phases();

	 break;

       default: 
	 /* 
	  * Should never have any other input.  Exit due to this error
	  * since the stream is corrupt (not in the correct order).
	  */
	 fprintf(stderr, "ERROR vet (process_function): wrong msgid %c\n", msgid);
	 exit(1);

      } /* end switch read_char */
   } /* end while not ENDVIEW */

#if DEBUG >= DEBUG_MED
	 fprintf(errlog, "ENDVIEW\n");
#endif

   /* ENDVIEW means VPO closed it's socket now VET has to */
   close();

   exit(0);

} /* end process_function */

void process_phases(void) {
   char		msgid;	/* Record id from stream */
   Phaseptr	phase;

   while ((msgid = read_char()) != ENDFUNCTION) {

      switch (msgid) {

       case BEGINPHASE:
	 /* 
	  * BEGINPHASE signals VPO has started optimizing
	  * a new phase.
	  */
#if DEBUG >= DEBUG_MED
   fprintf(errlog, "BEGINPHASE ");
#endif

	 phase = initialize_phase();

#if DEBUG >= DEBUG_MED
   fprintf(errlog, "%s\n", phase->name);
#endif
	 process_transformations(phase);

	 break;

	case SANITY_CHECK:
#if DEBUG >= DEBUG_MED
   fprintf(errlog, "SANITY_CHECK\n");
#endif
	  process_sanity();

	  break;

       default: 
	 /* 
	  * Should never have any other input.  Exit due to this error
	  * since the stream is corrupt (not in the correct order).
	  */
	 fprintf(stderr, "ERROR vet (process_phases): wrong msgid %c\n", msgid);
	 exit(1);

      } /* end switch read_char */

   } /* end while not ENDFUNCTION */

#ifdef XWINDOWS
   if (sanity_enabled) {
       sanity_dialog();
   }
#else
   /* END_FUNCTION reached - free up function structure. */
   free_function();

   /* clear hash_tables for next function */
   delete_rtl_hash_table();
   delete_bb_hash_table();
#endif

} /* end process_phases */

void process_transformations(Phaseptr phase) {

   char		msgid;	/* Record id from stream */

   while ((msgid = read_char()) != ENDPHASE) {

      switch (msgid) {

       case BEGINTRANS:
	 /* 
	  * BEGINTRANS signals VPO has started
	  * a new transformation.
	  */
#if DEBUG >= DEBUG_MED
	  fprintf(errlog, "BEGINTRANS ");
#endif
	
	  current_transformation = initialize_transformation(phase);

#if DEBUG >= DEBUG_MED
	  fprintf(errlog, "#%d\n", current_transformation->num);
#endif

	 apply_single_transformation(current_transformation);

	 break;

       default: 
	 /* 
	  * Should never have any other input.  Exit due to this error
	  * since the stream is corrupt (not in the correct order).
	  */
	 fprintf(stderr, "ERROR vet (process_transformations): wrong msgid %c\n", msgid);
	 exit(1);

      } /* end switch read_char */

   } /* end while not ENDPHASE */
   
} /* end process_transformations */

