/***********************************************************************
 * 
 * Copyright (c) 1998 Ann H. Bailey.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/vet/process_xstate.c,v $
 * $Revision: 1.1.1.1 $
 * 
 ***********************************************************************/
#define __PROCESS_XSTATE_C__

#include <stdio.h>

#ifdef XWINDOWS
#include <stdlib.h>
#include "vet.h"
#include "vpo.h"
#include "../vpo/lib/view.h"
#include "process_input.h"
#include "process_xstate.h"
#include "process_state.h"
#include "vet_xinit.h"

process_enum read_function_name(void) {

   char		msgid;	/* Record id from stream */

   msgid = read_char();

   if (msgid == BEGINFUNCTION) {
       /* 
	* BEGINFUNCTION signals VPO has started optimizing
	* a new function.
	*/
#if DEBUG >= DEBUG_MED
       fprintf(errlog, "BEGINFUNCTION\n");
#endif

       current_function = initialize_function();
       
#if DEBUG >= DEBUG_MED
       fprintf(errlog, "Function: %s\n", current_function->name);
#endif

       return newfunction;
   }

   if (msgid == ENDVIEW) {

#if DEBUG >= DEBUG_MED
       fprintf(errlog, "ENDVIEW\n");
#endif

       return lastfunction;
   }

   /* 
    * Should never have any other input.  Exit due to this error
    * since the stream is corrupt (not in the correct order).
    */
   fprintf(stderr, "ERROR vet (read_function_name): wrong msgid %c\n", msgid);
   exit(1);

} /* end read_function_name */

process_enum read_phase_name(void) {

   char		msgid;	/* Record id from stream */

   while (msgid = read_char()) {
       
       switch(msgid) {
	   case BEGINPHASE:
	   /* 
	    * BEGINPHASE signals VPO has started optimizing
	    * a new phase.
	    */
#if DEBUG >= DEBUG_MED
	   fprintf(errlog, "BEGINPHASE\n");
#endif
	   current_phase = initialize_phase();
       
	   return newphase;
	   break;

	 case ENDFUNCTION:
#if DEBUG >= DEBUG_MED
	   fprintf(errlog, "ENDFUNCTION\n");
#endif

	   return lastphase;
	   break;

	 case SANITY_CHECK:

#if DEBUG >= DEBUG_MED
	   fprintf(errlog, "SANITY CHECK\n");
#endif
	   process_sanity();
	   sanity_dialog();

	   /* read in next message for next beginning of phase */
	   continue;

	   break;
	 default:
	   /* 
	    * Should never have any other input.  Exit due to this error
	    * since the stream is corrupt (not in the correct order).
	    */
	   fprintf(stderr, "ERROR vet (read_phase_name): wrong msgid %c\n", msgid);
	   exit(1);

	   break;
       }
   }
} /* end read_phase_name */
    
void read_transformations(Phaseptr phase) {

   char		msgid;	/* Record id from stream */

   while ((msgid = read_char()) != ENDPHASE) {

      switch (msgid) {

       case BEGINTRANS:
	 /* 
	  * BEGINTRANS signals VPO has started
	  * a new transformation.
	  */
#if DEBUG >= DEBUG_MED
	  fprintf(errlog, "BEGINTRANS\n");
#endif
	 current_transformation = initialize_transformation(phase);

	 break;

       default: 
	 /* 
	  * Should never have any other input.  Exit due to this error
	  * since the stream is corrupt (not in the correct order).
	  */
	 fprintf(stderr, "ERROR vet (read_transformations): wrong msgid %c\n", msgid);
	 exit(1);

      } /* end switch read_char */

   } /* end while not ENDPHASE */
   
} /* end read_transformations */
#endif
