/***********************************************************************
 * 
 * Copyright (c) 1998 Ann H. Bailey.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/vet/rtl.c,v $
 * $Revision: 1.1.1.1 $
 ***********************************************************************/

#define __RTL_C__

#include <stdio.h>
#include <malloc.h>
#include "rtl.h"
#include "vet_io.h"
#include "hash_functions.h"

hash_table rtl_ht;
void number_rtl(Rtlptr);
void free_rtl(Rtlptr);

/*********************************************************************
 *
 * process_rtl - this function initializes new rtl structure.
 * 		 Return pointer to the new rtl.
 *
 *********************************************************************/
Rtlptr process_rtl(Rtlptr *rtl) {

    Rtlptr current_rtl;			/* rtl node */
    int rtladdr;      			/* rtl address */

    /* Read new rtl address */
    rtladdr = read_num();

    /* 
     * If the first rtl has not been read, then
     * initialize the rtl list.
     */
    if (!*rtl) {
	*rtl = create_rtl();
	insert_hash_table(rtl_ht, rtladdr, *rtl);
	(*rtl)->address = rtladdr;
	(*rtl)->type = read_num();
	(*rtl)->line = read_line();
	number_rtl(*rtl);
	return(*rtl);
    }
    /* 
     * Since the first rtl has been read, then append this 
     * rtl to the end of the rtl list.
     */
    current_rtl = *rtl;
    while (current_rtl->next) {
	current_rtl = current_rtl->next;
    }
    current_rtl->next = create_rtl();
    insert_hash_table(rtl_ht, rtladdr, current_rtl->next);
    current_rtl->next->address = rtladdr;
    current_rtl->next->type = read_num();
    current_rtl->next->line = read_line();
    current_rtl->next->prev = current_rtl;
    number_rtl(current_rtl->next);
    return(current_rtl->next);

} /* end process_rtl */

void number_rtl(Rtlptr rtl) {
    
    static int number_of_rtls = 0;

    rtl->number = ++number_of_rtls;

} /* end number_rtl */

/***********************************************************************
 *
 * create_rtl - this function allocates a new rtl node 
 *                  Return pointer to the new structure.
 *
***********************************************************************/
Rtlptr create_rtl(void) {

    Rtlptr newnode;
    
    newnode = malloc(sizeof(Rtlnode));
    newnode->type = 0;
    newnode->line = NULL;
#ifdef XWINDOWS
    newnode->rtlstr = NULL;
    newnode->selected = FALSE;
#endif
    newnode->address = 0;
    newnode->number = 0;
    newnode->links[0] = 0;
    newnode->side = NULL;
    newnode->deads = NULL;
    newnode->next = NULL;
    newnode->prev = NULL;
    return(newnode);

} /* end create_rtl */

/*********************************************************************
 *
 * add_rtl_address - this function is for updatebb (update all rtls in
 *		     basicblock. Creates a linked list of the new
 *		     ordered list of rtls.
 *
 *********************************************************************/
 
void add_rtl_address(Addrptr *rtl_list, int new_address) {

    Addrptr current_rtl;

    if (!*rtl_list) {
	*rtl_list = malloc(sizeof(Addrnode));
	(*rtl_list)->address = new_address;
	(*rtl_list)->next = NULL;
	return;
    }

    current_rtl = *rtl_list;
    while (current_rtl->next) {
	current_rtl = current_rtl->next;
    }
    current_rtl->next = malloc(sizeof(Addrnode));
    current_rtl->next->address = new_address;
    current_rtl->next->next = NULL;
    return;

} /* end add_rtl_address */

/*********************************************************************
 *
 * find_rtl - find rtl in basic block.
 *
 *********************************************************************/
Rtlptr find_rtl(int rtladdr) {

    Rtlptr rtl;

    if (rtl = (Rtlptr)search_hash_table(rtl_ht, rtladdr)) {
	return(rtl);
    }
    else {
	fprintf(stderr, "ERROR vet (find_rtl): cannot find rtl %d\n", rtladdr);
	exit(1);
    }
} /* end find_rtl */


/*********************************************************************
 *
 * find_rtl_num - find rtl number - if does not exist returns NULL
 *
 *********************************************************************/
Rtlptr find_rtl_num(int rtladdr) {

    return((Rtlptr)search_hash_table(rtl_ht, rtladdr));

} /* end find_rtl_num */

/******************************************************************
 * 
 * delete_rtl_node - removes a rtl node from vet linked list.
 *
 ******************************************************************/
void delete_rtl_node(Bblockptr bb, Rtlptr rtl) {

    /* Set previous rtl's next ptr to current rtl's next ptr */
    if (rtl->prev) {
	rtl->prev->next = rtl->next;
    } 
    else {
	/* If rtl->prev is NULL then it's the first rtl */
	/* so need to reset bb's ptr to first rtl.      */
	bb->rtls = rtl->next;
    }
    /* Set next rtl's previous ptr to current rtl's prev ptr */
    if (rtl->next) {
	rtl->next->prev = rtl->prev;
	rtl->next = NULL;
	/*
	 * After setting the next node to point to rtl's previous,
	 * reset rtl's previous ptr.
	 */
	rtl->prev = NULL;
    }

    if (rtl->line) {
	free(rtl->line);
	rtl->line = NULL;
    }

#ifdef XWINDOWS
    if (rtl->rtlstr) {
	XmStringFree(rtl->rtlstr);
	rtl->rtlstr = NULL;
    }
#endif

    /* Free side effects list */
    if (rtl->side) {
	free(rtl->side);
	rtl->side = NULL;
    }

    /* Free deads list */
    if (rtl->deads) {
	free(rtl->deads);
	rtl->deads = NULL;
    }

} /* end delete_rtl_node */

/*********************************************************************
 *
 * free_rtl - recursive function that traverses the list of rtls
 *            and deallocates each rtl.
 *
 *********************************************************************/
void free_rtl(Rtlptr rtl) {

    /* Free rtl line */
    if (rtl->line) {
	free(rtl->line);
	rtl->line = NULL;
    }

#ifdef XWINDOWS
    if (rtl->rtlstr) {
	XmStringFree(rtl->rtlstr);
	rtl->rtlstr = NULL;
    }
#endif

    /* Free side effects list */
    if (rtl->side) {
	free(rtl->side);
	rtl->side = NULL;
    }

    /* Free deads list */
    if (rtl->deads) {
	free(rtl->deads);
	rtl->deads = NULL;
    }
    
    rtl->links[0] = 0;
    free(rtl);
    rtl = NULL;

} /* end free_rtl */

/*********************************************************************
 *
 * free_rtl_wrapper - this function is passed to the hash functions
 * 		      to free and rtl node.
 *
 *********************************************************************/
void free_rtl_wrapper(void *rtl) {

    free_rtl((Rtlptr)rtl);

} /* end free_rtl_wrapper */

/*********************************************************************
 *
 * free_rtl_addresses - this function frees the linked list for 
 *			order rtls for updating the rtls in a basic
 * 			block.
 *
 *********************************************************************/
void free_rtl_addresses(Addrptr rtl_list) {

    if (rtl_list->next) {
	free_rtl_addresses(rtl_list->next);
	rtl_list->next = NULL;
    }

    free(rtl_list);

} /* end free_rtl_addresses */

/*********************************************************************
 *
 * delete_rtl_hash_table - deallocates hash table and clears entries.
 *
 *********************************************************************/
void delete_rtl_hash_table(void) {

    delete_hash_table(rtl_ht, free_rtl_wrapper);

} /* end delete_rtl_hash_table */

/*********************************************************************
 *
 * new_rtl_hash_table - sets hash table entries to NULL.
 *
 *********************************************************************/
void new_rtl_hash_table(void) {
    
    rtl_ht = new_hash_table();

} /* end new_rtl_hash_table */

/*********************************************************************
 *
 * insert_newrtl_operation - checks to see if rtl already exists in
 *			     hash table. If not creates a new rtl node,
 *			     inserts in hash table, and numbers the
 *			     the transformation. Special function used
 *			     for insert_rtl transformation since 
 *			     rtl may already exist in hash table if this
 *			     is an undo transformation.
 * 
 *********************************************************************/
Rtlptr insert_newrtl_operation(int rtladdr) {

    Rtlptr rtl;

    if (rtl = (Rtlptr)search_hash_table(rtl_ht, rtladdr))
	return(rtl);
    
    rtl = create_rtl();
    insert_hash_table(rtl_ht, rtladdr, rtl);
    number_rtl(rtl);
    return(rtl);

} /* end insert_newrtl_operation */

/*********************************************************************
 *
 * print_rtls - debug tool to print all rtls in a basic block.
 *
 *********************************************************************/
void print_rtls(Bblockptr bb) {

    Rtlptr rtl;

    for (rtl = bb->rtls; rtl; rtl = rtl->next) {
	printf("Rtl address = %d, #%d", rtl->address, rtl->number);
	if (rtl->prev) {
	    printf(" prev #%d", rtl->prev->number);
	}
	printf("\n");
    }
} /* end print_rtls */

/*********************************************************************
 *
 * print_rtl_hash_table - debug tool to print all rtls in hash table.
 *
 *********************************************************************/
void print_rtl_hash_table(hash_table ht) {

    int i;
    hash_entry *table_entry;


    for (i = 0; i < HASH_TABLE_SIZE; i++) {
	printf("index %d\n", i);
	/*
	 * Search through hashtable at this index and return a pointer
	 * if matching key found; if not, return NULL.
	 */
	for (table_entry = ht[i];
	     table_entry; 
	     table_entry = table_entry->next) {
	    /* Check to see if the entry has been initialized */
	    printf("key = %d", table_entry->key);
	    if (table_entry->info) {
		printf(" rtl #%d",((Rtlptr)table_entry->info)->number);
	    }
	    if (((Rtlptr)table_entry->info)->prev) {
		printf("prev #%d ",((Rtlptr)table_entry->info)->prev->number);
	    }
	    if (((Rtlptr)table_entry->info)->next) {
		printf("next #%d",((Rtlptr)table_entry->info)->next->number);
	    }
	    printf("\n");
	}
    }
} /* end print_rtl_hash_table */

