/***********************************************************************
 * 
 * Copyright (c) 1998 Ann H. Bailey.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/vet/vet_callbacks.c,v $
 * $Revision: 1.1.1.1 $
 * 
 ***********************************************************************/
#include <stdio.h>

#ifdef XWINDOWS

#include "vpo.h"
#include "../vpo/lib/view.h"
#include "vet_callbacks.h"
#include "vet_xinit.h"
#include "display_input.h"
#include "process_xstate.h"
#include "process_state.h"

#include <Xm/SelectioB.h>
#include <Xm/MessageB.h>

state_enum viewer_state = initial_state;
process_enum process_state;

void transbreak_entered(Widget, XtPointer, XtPointer);
void destroy_widget(Widget, XtPointer, XtPointer);

void enable_button(Widget);
void disable_button(Widget);
char display_deads = FALSE;
char display_sides = FALSE;
char sanity_enabled = FALSE;
unsigned int rtls_to_display = (1 << 0);    /* Default so only RTLINEs set */
Phaseptr prev_phase = NULL;
char breakpoint_enabled = FALSE;
int transnum_break = 0;

/*********************************************************************
 *
 * next_func_pressed - user pressed Next Func button and next
 *		       function's initial set with first
 *		       transformations operation's highlighted.
 *
 *********************************************************************/
void next_func_pressed(Widget parent, XtPointer client_data,
		       XtPointer call_data) {

    switch (viewer_state) {

      case initial_state:
	/* Set scroll window for transformations to sensitive */
	enable_button(operwidget);

	/* Set breakpoint for phases sensitive */
	enable_button(phasemenu);

	/* Set breakpoint for transformations sensitive */
	enable_button(transbreak_menu);

	break;

      case intransbefore_state:
	apply_single_transformation(current_transformation);

	/* Fall Through */

      case intransafter_state:
	clear_highlights(current_transformation, viewer_state, breakpoint_enabled);

	/* Apply all remaining transformations */
	apply_all_transformations(current_transformation->next);

        /* Fall Through */

      case inphase_state:
	disable_button(prevtransbutton);
	disable_button(undotransbutton);
	disable_button(prevphasebutton);

	/* Clear Scroll List */
	XmListDeleteAllItems(operwidget);

	/* Read in remaining phases */
	process_phases();

	/* Fall Through */

      case lastphase_state:
	/* Clear status */
	XmTextSetString(statusmessage, "");

	/* Remove all basic blocks from screen */
	XtDestroyWidget(allblocks);

	/* END_FUNCTION reached - free up function structure. */
	free_function();

	/* clear hash_tables for next function */
	delete_rtl_hash_table();
	delete_bb_hash_table();

	break;

      default:
	break;
    } /* end switch */

    /* Clear current phase since we're starting a new function */
    current_phase = NULL;
    prev_phase = NULL;

    /* Read in function name */
    process_state = read_function_name();

    /* Last function for input file */
    if (process_state == lastfunction) {
	/* Display status */
	XmTextSetString(statusmessage, "No more functions");
	
	/* Clear function name */
	XmTextSetString(funcnameinput, "");
	/* Clear trans num */
	XmTextSetString(transnuminput, "");
	/* Clear trans state */
	XmTextSetString(transstate, "");
	/* Clear phase name */
	XmTextSetString(phasenameinput, "");

	/* Turn off buttons */
	disable_button(nextfuncbutton);
	disable_button(nextphasebutton);
	disable_button(prevphasebutton);
	disable_button(nexttransbutton);
	disable_button(prevtransbutton);
	disable_button(applytransbutton);
	disable_button(undotransbutton);
	disable_button(breakbutton);
	disable_button(phasemenu);
	disable_button(run);
	disable_button(options);

	viewer_state = final_state;
    }
    else {
	/* Display function name */
	XmTextSetString(funcnameinput, current_function->name);

	/* Set SANITY_CHECK flag based on users input */
	write_char(SANITY_CHECK);
	if (sanity_enabled)
	    write_char('T');
	else
	    write_char('F');

	/* Read in the initial set of RTLs */
	process_initial_set();

	/* Display initial set of basic blocks & rtls */
	display_initial_set();

	viewer_state = infunction_state;

	enable_button(nextfuncbutton);
	enable_button(breakbutton);

	/* Simulate moving to next state */
	next_phase_pressed(NULL, NULL, NULL);
    }

} /* end next_func_pressed */

/*********************************************************************
 *
 * next_phase_pressed - user has pressed the next_phase button.
 *                      Apply all transformations (but don't display)
 *                      and read in transformations for next phase.
 *                      Display first transformation.
 *
 *********************************************************************/
void next_phase_pressed(Widget parent, XtPointer client_data,
		      XtPointer call_data) {

     static char lastphase_read = FALSE;

     /* Clear trans state */
     XmTextSetString(transstate, "");
     enable_button(prevphasebutton); 
     disable_button(prevtransbutton);
     disable_button(applytransbutton);

     /* Clear status */
     XmTextSetString(statusmessage, "");

     switch (viewer_state) {
        case intransbefore_state: 
 	clear_highlights(current_transformation, viewer_state, breakpoint_enabled);

 	/* Apply current and all remaining transformations */
 	apply_all_transformations(current_transformation);

	/* Clear trans num */
	XmTextSetString(transnuminput, "");

	break;

      case intransafter_state:
	clear_highlights(current_transformation, viewer_state, breakpoint_enabled);

	/* Apply all remaining transformations */
	apply_all_transformations(current_transformation->next);

	break;

      case infunction_state:
	/* reset lastphase read flag since we're in a new function */
	lastphase_read = FALSE;
 
	break;

      case inphase_state:
      default:
	break;

    } /* end switch */

    /* If we've already viewed this phase don't read again */
    if (current_phase && current_phase->next) {
	/* Save previous phase in case this is last phase */
	/* since it's reset in read_phase_name */
	prev_phase = current_phase;
	current_phase = current_phase->next;
    }
    else {
        if (!lastphase_read) {
	    /* Read in phase name */
	    process_state = read_phase_name();
        }

	if (process_state == lastphase) {
	    /* Set flag for end of function */
            lastphase_read = TRUE;

	    /* Clear previous phase name */
	    XmTextSetString(phasenameinput, "");
	    /* Display status */
	    XmTextSetString(statusmessage, "No more phases");

	    disable_button(nextphasebutton);
	    disable_button(nexttransbutton);
	    
	    viewer_state = lastphase_state;
	    return;
	}	

	/* Read in remaining transformations for current phase*/
	read_transformations(current_phase);

    }
    /* Update phase name */
    XmTextSetString(phasenameinput, current_phase->name);
    
    enable_button(nextphasebutton);

    viewer_state = inphase_state;

    next_trans_pressed(NULL, NULL, NULL);

} /* end next_phase_pressed */

/*********************************************************************
 *
 * prev_phase_pressed - undo current transformations and previous
 *			phase transformations. Display first
 *                      transformation of previous phase.
 *
 *********************************************************************/
void prev_phase_pressed(Widget parent, XtPointer client_data,
		      XtPointer call_data) {
    Transptr trans;

    /* Need to remove the highlights on current transformations */
    clear_highlights(current_transformation, viewer_state, breakpoint_enabled);

    switch (viewer_state) {
      case intransafter_state:

	/* Undo transformations for current phase */
	undo_all_transformations(current_transformation);

	break;

      case intransbefore_state:

	/* Undo transformations for current phase but not
	   current transformation since never applied. */
	undo_all_transformations(current_transformation->prev);

	break;

      default:
	break;
    } /* end switch */

    /* Check if lastphase read since that means current_phase was
       not updated but phase is no longer displayed either. */
    if (viewer_state != lastphase_state) {
      /* Go back to previous phase */
      current_phase = current_phase->prev;
    }    

    /* Find last transformation and undo from there */
    for (trans = current_phase->first_trans;
	 trans && trans->next;
	 trans = trans->next)
	;
    undo_all_transformations(trans);

    /* Update phase name */
    XmTextSetString(phasenameinput, current_phase->name);

    if (!current_phase->prev) {
	disable_button(prevphasebutton);
	/* Display status */
	XmTextSetString(statusmessage, "First phase");
    }

    /* Clear state */
    XmTextSetString(transstate, "");

    /* Clear trans number */
    XmTextSetString(transnuminput, "");

    /* Clear status */
    XmTextSetString(statusmessage, "");

    enable_button(nextphasebutton);
    enable_button(nexttransbutton);
    disable_button(prevtransbutton);

    /* Clear current_transformation since starting new phase */
    current_transformation = NULL;

    viewer_state = inphase_state;

    next_trans_pressed(NULL, NULL, NULL);

} /* end prev_phase_pressed */

/*********************************************************************
 *
 * next_trans_pressed - apply current transformation and display 
 *                      next transformation.
 *
 *********************************************************************/
void next_trans_pressed(Widget parent, XtPointer client_data,
			XtPointer call_data) {

    char numberstr[10];

    /* Clear Scroll List */
    XmListDeleteAllItems(operwidget);

    switch (viewer_state) {
      case inphase_state:
	current_transformation = current_phase->first_trans;
	
	if (!current_transformation) {
	    /* Display status */
	    XmTextSetString(statusmessage, "No transformations for phase");

	    disable_button(nexttransbutton);
	    disable_button(applytransbutton);

	    return;
	}	
	break;

      case intransbefore_state:

	/* Clear highlights from previous transformation */
	clear_highlights(current_transformation, viewer_state, breakpoint_enabled);
	apply_single_transformation(current_transformation);

	/* Get next transformation */
	current_transformation = current_transformation->next;

	enable_button(prevtransbutton);

	/* Clear status */
	XmTextSetString(statusmessage, "");

	break;

      case intransafter_state:

	/* Clear highlights from previous transformation */
	clear_highlights(current_transformation, viewer_state, breakpoint_enabled);

	/* Get next transformation */
	current_transformation = current_transformation->next;

	enable_button(prevtransbutton);
	disable_button(undotransbutton);


	/* Clear status */
	XmTextSetString(statusmessage, "");

	break;

      default:
	break;

    } /* end switch */
	   
    /* Reset state */
    viewer_state = intransbefore_state;

    /* If using breakpoints could have empty transformation. The user
       interface doesn't allow the next transformation button to be pressed.
       Therefore exit this function to run pressed function.
     */
    if (!current_transformation) {
	return;
    }

    /* Display transformation number */
    sprintf(numberstr, "%d", current_transformation->num);
    XmTextSetString(transnuminput, numberstr);

    /* Check if empty transformation */
    if (!current_transformation->operation) {
	disable_button(applytransbutton);

	/* Display status */
	XmTextSetString(transstate, "Empty transformation");
    } 
    else {
	/* Display state */
	XmTextSetString(transstate, "BEFORE transformation applied");

	/* Highlight the transformations to be made */
	highlight_trans(viewer_state, breakpoint_enabled);
	enable_button(applytransbutton);
    }

    /* No more transformations */
    if (!current_transformation->next) {
	disable_button(nexttransbutton);

	/* Display status */
	XmTextSetString(statusmessage, "Last transformation");
    }
    else {
	enable_button(nexttransbutton);
    }

} /* end next_trans_pressed */

/*********************************************************************
 *
 * apply_trans_pressed - apply current transformation and highlight
 *                       newly inserted blocks or rtls.
 *
 *********************************************************************/
void apply_trans_pressed(Widget parent, XtPointer client_data,
			XtPointer call_data) {

    /* Clear highlights before doing new ones */
    clear_highlights(current_transformation, viewer_state, breakpoint_enabled);

    apply_single_transformation(current_transformation);

    /* Display state */
    XmTextSetString(transstate, "AFTER transformation applied");

    viewer_state = intransafter_state;

    /* Highlight the transformations made */
    highlight_trans(viewer_state, breakpoint_enabled);

    /* Clear status only if not last transformation -
       want user to still realize there are no more transformations */
    if (current_transformation->next) {

	/* Clear status */
	XmTextSetString(statusmessage, "");
    }

    disable_button(applytransbutton);
    enable_button(undotransbutton);

} /* end apply_trans_pressed */

/*********************************************************************
 *
 * prev_trans_pressed - undo current transformation and previous
 *                      transformation. Display previous transformation.
 *
 *********************************************************************/
void prev_trans_pressed(Widget parent, XtPointer client_data,
			XtPointer call_data) {

    char numberstr[10];

    /* Clear Scroll List */
    XmListDeleteAllItems(operwidget);

    /* Clear current transformation's highlighted rtl's */
    clear_highlights(current_transformation, viewer_state, breakpoint_enabled);

    switch (viewer_state) {
      case intransafter_state:
	undo_single_transformation(current_transformation);

	/* Turn off undo button */
	disable_button(undotransbutton);
	break;

      default:
	break;
    } /* end switch */

    /* Clear previous status */
    XmTextSetString(statusmessage, "");

    /* Go back to previous transformation */
    current_transformation = current_transformation->prev;

    /* Undo transformation for previous transformation */
    undo_single_transformation(current_transformation);

    if (!current_transformation->prev) {
	disable_button(prevtransbutton);

	/* Display status */
	XmTextSetString(statusmessage, "First transformation");
    }

    /* Display transformation number */
    sprintf(numberstr, "%d", current_transformation->num);
    XmTextSetString(transnuminput, numberstr);

    /* Check if empty transformation */
    if (!current_transformation->operation) {
	disable_button(applytransbutton);

	/* Display status */
	XmTextSetString(transstate, "Empty transformation");
    } 
    else {
	/* Display state */
	XmTextSetString(transstate, "BEFORE transformation applied");

	viewer_state = intransbefore_state;

	/* Highlight the transformations to be made */
	highlight_trans(viewer_state, breakpoint_enabled);
	enable_button(applytransbutton);
    }
    enable_button(nexttransbutton);

} /* end */

/*********************************************************************
 *
 * undo_trans_pressed - undo current transformation. Change highlights
 *			to rtls to be modified or blocks and rtls to
 * 			be deleted.
 *
 *********************************************************************/
void undo_trans_pressed(Widget parent, XtPointer client_data,
			XtPointer call_data) {

    /* Clear Scroll List */
    XmListDeleteAllItems(operwidget);

    /* Clear current highlights */
    clear_highlights(current_transformation, viewer_state, breakpoint_enabled);

    /* Undo current transformation */
    undo_single_transformation(current_transformation);

    viewer_state = intransbefore_state;

    highlight_trans(viewer_state, breakpoint_enabled);

    XmTextSetString(transstate, "BEFORE transformation applied");

    /* Turn off undo button */
    disable_button(undotransbutton);

    /* Turn on apply button */
    enable_button(applytransbutton);

} /* end undo_trans_pressed */

/*********************************************************************
 *
 * exit_pressed - send message to VPO and close socket.
 *
 *********************************************************************/
void exit_pressed(Widget parent, XtPointer client_data,
		  XtPointer call_data) {

    /* Let VPO know VET exiting */
    write_char(ABORT_VPO);

    /* Close socket to VPO */
    close();

    exit(0);

} /* end exit_pressed */

/*********************************************************************
 *
 * break_pressed - pop up dialog box for user to enter transformation
 *                 number to break on.
 *
 *********************************************************************/
void break_pressed(Widget parent, XtPointer client_data,
		  XtPointer call_data) {

    Widget dialog;
    XmString prompt = XmStringCreateLocalized("Enter transformation number:");
    Arg args[2];

    XtSetArg(args[0], XmNselectionLabelString, prompt);
    XtSetArg(args[1], XmNautoUnmanage, False);
    dialog = XmCreatePromptDialog(parent, "TransBreak", args, 2);
    
    /* When user enters the transformation number call transbreak_entered */
    XtAddCallback(dialog, XmNokCallback, transbreak_entered, parent);
    
    /* Destroy widget if user presses Cancel */
    XtAddCallback(dialog, XmNcancelCallback, destroy_widget, parent);

    /* No help available */
    XtUnmanageChild(XmSelectionBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
    
    XtManageChild(dialog);

} /* end break_pressed */

/*********************************************************************
 *
 * transbreak_entered - grab the transformation number to break on.
 * 			Apply next_trans until transformation is
 *			reached or apply next_phase. Displays
 *			transformation or if number invalid next
 *			valid transformation.
 *********************************************************************/
void transbreak_entered(Widget parent, XtPointer client_data,
			XtPointer transnum) {
    char *transnum_str;
    int transnum_break;
    char numberstr[10];

    /* Convert XmString to C string */
    XmStringGetLtoR(((XmSelectionBoxCallbackStruct *)transnum)->value, 
		    "", &transnum_str);
    /* Convert string to number */
    transnum_break = atoi(transnum_str);

    XtDestroyWidget(parent);

    /* Clear existing highlights */
    clear_highlights(current_transformation, viewer_state, breakpoint_enabled);

    breakpoint_enabled = TRUE;

    /* Go to next transformation if current number is less than breakpoint */
    while (current_transformation && (current_transformation->num < transnum_break)) {

	next_trans_pressed(NULL, NULL, NULL);

	/* If not transformations for current phase then go to next phase */
	while (!current_transformation) {

	    next_phase_pressed(NULL, NULL, NULL);

	    if (process_state == lastphase) {
		next_func_pressed(NULL, NULL, NULL);

		if (process_state == lastfunction) {
		    /* Display status */
		    XmTextSetString(statusmessage, "No more functions");
	
		    /* Clear function name */
		    XmTextSetString(funcnameinput, "");

		    /* Clear trans num */
		    XmTextSetString(transnuminput, "");

		    /* Clear trans state */
		    XmTextSetString(transstate, "");

		    /* Clear phase name */
		    XmTextSetString(phasenameinput, "");

		    /* Turn off buttons */
		    disable_button(nextfuncbutton);
		    disable_button(nextphasebutton);
		    disable_button(nexttransbutton);
		    disable_button(applytransbutton);
		    disable_button(breakbutton);
		    
		    viewer_state = final_state;

		    return;

		} /* if lastfunction */
	    } /* if lastphase */
	} /* if not transformation */
    } /* if not breakpoint number */

    /* Either found the matching transformation number or went past it */
    /* Reset breakpoint flag */
    breakpoint_enabled = FALSE;

    /* If transformation is not empty then highlight the operations */
    if (current_transformation && current_transformation->operation) {
	highlight_trans(viewer_state, breakpoint_enabled);
    }

} /* end run_pressed */

/*********************************************************************
 *
 * destroy_widget - removes widget displayed on screen.
 *
 *********************************************************************/
void destroy_widget(Widget parent, XtPointer client_data,
		    XtPointer transnum) {

    XtDestroyWidget(parent);

} /* end destroy_widget */

/*********************************************************************
 *
 * enable_button - enables button for user (such as apply_trans).
 *
 *********************************************************************/
void enable_button(Widget w) {

    XtVaSetValues(w,
		  XmNsensitive, True,
		  NULL);

} /* end enable_button */

/*********************************************************************
 *
 * disable_button - disables button so user cannot press.
 *
 *********************************************************************/
void disable_button(Widget w) {

    XtVaSetValues(w,
		  XmNsensitive, False,
		  NULL);

} /* end disable_button */

/*********************************************************************
 *
 * The following functions read the state of the toggle buttons
 * in the file menu display options. If the toggle button is on then
 * the appropriate flag is set to true.
 *
 *********************************************************************/
void sanity_pressed(Widget w, XtPointer client_data, 
		    XtPointer state) {

    sanity_enabled = (((XmToggleButtonCallbackStruct *)state)->set) ? TRUE : FALSE;
    
}

void asm_pressed(Widget w, XtPointer client_data, 
		   XtPointer state) {

    if (((XmToggleButtonCallbackStruct *)state)->set) {
        rtls_to_display |= ASMLINE;
    }
    else {
	rtls_to_display &= ~ASMLINE;
    }
    
}

void trash_pressed(Widget w, XtPointer client_data, 
		   XtPointer state) {

    if (((XmToggleButtonCallbackStruct *)state)->set) {
        rtls_to_display |= TRASHLINE;
    }
    else {
	rtls_to_display &= ~TRASHLINE;
    }
    
}

void use_pressed(Widget w, XtPointer client_data, 
		    XtPointer state) {

    if (((XmToggleButtonCallbackStruct *)state)->set) {
        rtls_to_display |= USELINE;
    }
    else {
	rtls_to_display &= ~USELINE;
    }
    
}

void reserve_pressed(Widget w, XtPointer client_data, 
		    XtPointer state) {

    if (((XmToggleButtonCallbackStruct *)state)->set) {
        rtls_to_display |= RESLINE;
    }
    else {
	rtls_to_display &= ~RESLINE;
    }
    
}

void comment_pressed(Widget w, XtPointer client_data, 
		    XtPointer state) {

    if (((XmToggleButtonCallbackStruct *)state)->set) {
        rtls_to_display |= COMMENT;
    }
    else {
	rtls_to_display &= ~COMMENT;
    }
    
}

void set_pressed(Widget w, XtPointer client_data, 
		    XtPointer state) {

    if (((XmToggleButtonCallbackStruct *)state)->set) {
        rtls_to_display |= SETLINE;
    }
    else {
	rtls_to_display &= ~SETLINE;
    }
    
}

void csline_pressed(Widget w, XtPointer client_data, 
		    XtPointer state) {

    if (((XmToggleButtonCallbackStruct *)state)->set) {
        rtls_to_display |= CSLINE;
    }
    else {
	rtls_to_display &= ~CSLINE;
    }
    
}

void sides_pressed(Widget w, XtPointer client_data, 
		    XtPointer state) {

    display_sides = (((XmToggleButtonCallbackStruct *)state)->set) ? TRUE : FALSE;
    
}

void deads_pressed(Widget w, XtPointer client_data, 
		    XtPointer state) {

    display_deads = (((XmToggleButtonCallbackStruct *)state)->set) ? TRUE : FALSE;
    
}

/*********************************************************************
 *
 * display_right - when user clicks on right arrow at bottom of
 *		   basicblock pop up a window showing the target
 *		   basicblock number.
 *
 *********************************************************************/
void display_right(Widget w, XtPointer basicblock,
		   XtPointer state) {

    Widget dialog;
    char display[MAXSTRING];
    Arg arg[1];
    XmString label;
    Bblockptr bb;

    bb = find_bb_num(((Bblockptr)basicblock)->right);

    if (bb && bb->number) {
	sprintf(display, "Target is basicblock #%d", bb->number);
    }
    else {
	sprintf(display, "VPO ERROR: target basicblock address #%d does not exist",
		((Bblockptr)basicblock)->right);
    }

    label = XmStringCreateLocalized(display);
    XtSetArg(arg[0], XmNmessageString, label);
    dialog = XmCreateMessageDialog(((Bblockptr)basicblock)->bb_widget, 
				   "RightArrow", arg, 1);
    XmStringFree(label);

    /* No help is available */
    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));

    /* Don't need the cancel button */
    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_CANCEL_BUTTON));

    XtManageChild(dialog);

} /* end display_right */


/*********************************************************************
 *
 * display_left - when user clicks on left arrow at bottom of
 *                basicblock pop up a window showing the target
 *		  basicblock number.
 *
 *********************************************************************/
void display_left(Widget w, XtPointer basicblock,
		   XtPointer state) {

    Widget dialog;
    char display[MAXSTRING];
    Arg arg[1];
    XmString label;
    Bblockptr bb;

    bb = find_bb_num(((Bblockptr)basicblock)->left);

    if (bb && bb->number) {
	sprintf(display, "Target is basicblock #%d", bb->number);
    }
    else {
	sprintf(display, "VPO ERROR: target basicblock address #%d does not exist",
		((Bblockptr)basicblock)->left); 
    }

    label = XmStringCreateLocalized(display);
    XtSetArg(arg[0], XmNmessageString, label);
    dialog = XmCreateMessageDialog(((Bblockptr)basicblock)->bb_widget, 
				   "LeftArrow", arg, 1);
    XmStringFree(label);

    /* No help is available */
    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));

    /* Don't need the cancel button */
    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_CANCEL_BUTTON));

    XtManageChild(dialog);

} /* end display_left */

/*********************************************************************
 *
 * phase_break_callback - this function is called when user clicks on
 * 			  the breakpoint menu for phase breakpoint. 
 *			  Skips ahead to phase; applying transformations.
 *			  If phase not reached by end of function then
 *			  stop with message no more phases.
 *
 *********************************************************************/
void phase_break_callback(Widget w, XtPointer phase_num,
			  XtPointer state) {

    /* Clear existing highlights */
    clear_highlights(current_transformation, viewer_state, breakpoint_enabled);

    /* Clear trans num */
    XmTextSetString(transnuminput, "");

    /* Clear trans state */
    XmTextSetString(transstate, "");

    /* Clear status */
    XmTextSetString(statusmessage, "");

    switch (viewer_state) {
      case intransbefore_state:
	/* Apply current and all remaining transformations */
	apply_all_transformations(current_transformation);

	enable_button(prevphasebutton);
	disable_button(prevtransbutton);
	disable_button(applytransbutton);

	break;

      case intransafter_state:
	/* Apply all remaining transformations */
	apply_all_transformations(current_transformation->next);

	enable_button(prevphasebutton);
	disable_button(prevtransbutton);
	disable_button(undotransbutton);

	break;

      default:
	break;

    } /* end switch */

    /* If we've already viewed this phase don't read again */
    if (current_phase && current_phase->next) {
	/* Save previous phase in case this is last phase */
	/* since it's reset in read_phase_name */
	prev_phase = current_phase;
	current_phase = current_phase->next;
    }
    else {
	/* Read in phase name */
	process_state = read_phase_name();
	    
	if (process_state == lastphase) {
	    /* Clear previous phase name */
	    XmTextSetString(phasenameinput, "");

	    /* Display status */
	    XmTextSetString(statusmessage, "No more phases");
		
	    disable_button(nextphasebutton);
	    disable_button(nexttransbutton);

	    return;
	}	

	/* Read in transformations for current phase*/
	read_transformations(current_phase);

	}

    while (strcmp(current_phase->name, phasenames[(int)phase_num])) {
	apply_all_transformations(current_phase->first_trans);

	/* If we've already viewed this phase don't read again */
	if (current_phase && current_phase->next) {
	    /* Save previous phase in case this is last phase */
	    /* since it's reset in read_phase_name */
	    prev_phase = current_phase;
	    current_phase = current_phase->next;
	}
	else {
	    /* Read in phase name */
	    process_state = read_phase_name();
	    
	    if (process_state == lastphase) {
		/* Clear previous phase name */
		XmTextSetString(phasenameinput, "");

		/* Display status */
		XmTextSetString(statusmessage, "No more phases");
		
		disable_button(nextphasebutton);
		disable_button(nexttransbutton);

		return;
	    }	

	    /* Read in transformations for current phase*/
	    read_transformations(current_phase);

	} /* end if */
    } /* end while */

    /* Update phase name */
    XmTextSetString(phasenameinput, current_phase->name);
    
    enable_button(nextphasebutton);

    viewer_state = inphase_state;

    next_trans_pressed(NULL, NULL, NULL);

} /* end phase_break_callback */
#endif
