/***********************************************************************
 * 
 * Copyright (c) 1996 Ann H. Bailey.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/vet/vet_init.c,v $
 * $Revision: 1.1.1.1 $
 * 
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "vet_init.h"

/* From vpcc.h */
#define FALSE        0     /* Logical false */
#define TRUE	     1     /* Logical true  */

char remotehostname[MAXHOSTNAMELEN + 1];  /* host running VPO */
char localhostname[MAXHOSTNAMELEN + 1];   /* host running VET */

/***********************************************************************
 *
 * parse_arguments - parse the arguments given in the command line.
 *
 ***********************************************************************/
char *parse_arguments(int argc, char **argv) {

    char *line;			/* Argument line */
    int v_arg_present = FALSE;	/* Command argument present flag */
    int h_arg_present = FALSE;  /* remote hostname arg present */
    static char command[LISTLEN];      /* rsh command to run VPO */
    char vpo_path[LISTLEN];     /* VPO path */
    char *chome;		/* CHOME environment variable */

    /* Initialize command with remote shell command */
    /* Path for SunOS and Solaris */
    strcpy(command, "rsh ");

    /* Skip first argument which is the process name */
    while (--argc) {
	line = *++argv;
	/*
	 * Look at the first letter of the argument.  Options start with 
	 * a '-'.
	 */
	if (*line == '-') {
	    /*
	     * Look at the second letter to determine whether it is a
	     * VET switch (starts with "T" or "t").
	     */
	    line++;
	    if ((*line == 'T') || (*line == 't')) {
		/*
		 * Look at the third letter of the argument to narrow down
		 * the possible things that this argument might be.
		 */
		switch (*++line) {
		
		    /*
		     * Check for v argument.  Need to get the command to 
		     * run VPO.
		     */
		  case 'V':
		  case 'v':
		    /* Check to see if argument for v is given. */
		    if (*++line == '=') {
			/* increment to VPO command */
			line++;
			strcpy(vpo_path, line);
		    }
		    else {
			fprintf(stderr, "vet: v option expects argument\n Usage: vet -Tv=path",
				*line);
			exit(1);
		    }
		    
		    /* Set flag true that -v argument processed */
		    v_arg_present = TRUE;
		    
		    break;
		    
		    /* 
		     * Check for h argument.  Need to get hostname for 
		     * remote machine to run VPO on.  If no-h argument is 
		     * given then default to local machine.
		     */
		  case 'H':
		  case 'h':
		    /* Check to see if argument for h is given. */
		    if (*++line == '=') {
			/* increment to hostname */
			line++;
			strcpy(remotehostname, line);
		    }
		    else {
			fprintf(stderr, "vet: h option expects argument\nUsage: vet -Th=vpohostname\n",
				*line);
			exit(1);
		    }

		    /* Set flag true that -h argument processed */
		    h_arg_present = TRUE;

		    break;

		  default:
		    break;
		} /* end switch (*++line) */
	    } /* end if 't' */
	} /* end if '-' */
    } /* end while argc */
    
    /* 
     * If no -v arg was given then check for an environment 
     * variable (CHOME -- used in VPCC).  Otherwise exit with
     * error.
     */
    if (!v_arg_present) {
	/* Get VPO path from CHOME environment variable */
       if (chome = getenv("CHOME")) {
	  strcpy(vpo_path, chome);
	  /* 
	   * If path obtained from CHOME env. var. then append
	   * default executable name for sparc platform.
	   */
	  strcat(vpo_path, "/vpo.sparc");
       }
       else {
	  fprintf(stderr, "vet: -v argument expected\nUsage: vet -Tv=vpopath\n");
	  exit(1);
       }
    }

    /* Find name of local machine */
    if (gethostname(localhostname, sizeof(localhostname)) < 0) {
	perror("vet_init: gethostname ");
	exit(1);
    }
    
    /* Special fix for running at home */
    if ((!strcmp(localhostname, "twiddler")) ||
        (!strcmp(localhostname,  "smokin"))) {
        strcpy(localhostname, "localhost");
    }
    /* 
     * If no -h arg was given then assume running on local
     * machine and therefore remotehost = localhost.
     */
    if (!h_arg_present) {
	strcpy(remotehostname, localhostname);
    }

    /* Set rsh remote machine name */
    strcat(command, remotehostname);

    /* Add vpo command */
    strcat(command, " ");
    strcat(command, vpo_path);

    /* Add vpo hostname argument to VPO command */
    strcat(command, " -th=");
    strcat(command, localhostname);

    return command;

} /* end parse_arguments */

