/***********************************************************************
 * 
 * Copyright (c) 1996 Ann H. Bailey.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/vet/vet_socket.c,v $
 * $Revision: 1.1.1.1 $
 * 
 ***********************************************************************/
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include "vet_init.h"
#include "vet_socket.h"
#include "vet_io.h"
#include "vet.h"

static int server_socket(int);  /* Accepts VPO connection on socket. */
static int server_port(int *sock_ptr);  /* Creates socket and returns port */
static int sockfd;    /* socket file descriptor needed for writes to socket */

/*********************************************************************
 *
 * setup_sockets - setup VET socket and execute rsh command for VPO.
 *
 *********************************************************************/
void setup_sockets(char *vpo_command, char **argv) {

        int serv_sock;                  /* VET's socket */
        int port;                       /* port number of VET's socket */
        int input_file_present = FALSE; /* flag for input file arg */
        char portstring[LISTLEN];       /* port for VPO to connect to */

#if DEBUG >= DEBUG_LO
        char logfile[LISTLEN];          /* logfile for debugging */

        /* assume that output is going to stdout */
        /* errlog will be reset if an output file is given */
        errlog = stdout;
#endif

        /* Set up VET socket and obtain port for VPO to connect to. */
        port = server_port(&serv_sock);

        /* Add port argument to rsh vpo command */
        sprintf(portstring, " -tp=%d ", port);
        strcat(vpo_command, portstring);

        /* Build VPO command from argv. Redirect file names correctly. */
        /* Skip process name in argv. */
        *++argv;
        while (*argv) {

            /* Skip arguments that start with a "-" -- processed earlier */
            if (**argv != '-') {
                /* If encountered input file already, this parameter
                   is the output file. */
                if (input_file_present) {
                    /* Add redirection to rsh vpo command. */
                    strcat(vpo_command, ">");
#ifdef DEBUG
                    /* 
                     * If an output filename is given then
                     * append .log to it to create the log file.
                     */
                    strcpy(logfile, *argv);
                    strcat(logfile, ".log");
                    if (!(errlog = fopen(logfile, "w"))) {
                        fprintf(stderr, "vet: cannot create %s\n",
                                logfile);
                        exit(1);
                    }
#endif
                }
                else {
                    /* 
                     * First filename is the input file.  Add redirection
                     * to the rsh vpo command.
                     */
                    strcat(vpo_command, "<");
                    input_file_present = TRUE;
                } /* end if input_file_present */
            } 
            else {
                /* 
                 * Don't add on the -t arguments - they've already 
                 * been processed in parse_arguments(). 
                 */
                if (*(*argv + 1) == 't' || *(*argv + 1) == 'T') {
                    *argv++;
                    continue;
                }
            } /* end if '-' */
            /* 
             * Add any remaining arguments to command -- this will allow
             * any special VPCC/VPO command to pass through.
             */  
            strcat(vpo_command, *argv++);
            strcat(vpo_command, " ");
        } /* end while */

        /* Illegal since VET needs to pass an input file to VPO. */
        if (!input_file_present) {
            fprintf(stderr, "vet: input file not specified\n");
            exit(1);
        }
        /* Execute rsh VPO command in the background. */
        strcat(vpo_command, "&");
        /* Execute rsh VPO command */
        system(vpo_command);
            
#if DEBUG == DEBUG_HI
        fprintf(errlog,"setup_sockets: vpo command = %s\n", vpo_command);
#endif

        /* Wait for VPO to establish a connection */
        /* Set global variable sockfd to socket file descriptor. */
        sockfd = server_socket(serv_sock);


} /* end setup_sockets */

/*********************************************************************
 *
 * server_port - creates a new stream socket and binds it to an
 *               address in the internet domain.
 *********************************************************************/
int server_port(int *sock_ptr) {

    int sock;
    int length;
    struct sockaddr_in server;
    struct sockaddr *server_ptr = (struct sockaddr *) &server;
    int port;


    /* Create an internet stream socket. */
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
        perror("ERROR vet (server_port): opening stream socket");
        exit(1);
    }
    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    server.sin_port = 0;
    /* Assign a name to this socket. */
    if (bind(sock, server_ptr, sizeof(server))) {
        perror("ERROR vet (server_port): binding stream socket");
        exit(1);
    }
    length= sizeof(server);
    if (getsockname(sock, server_ptr, &length)) {
        perror("ERROR vet (server_port): getting socket name");
        exit(1);
    }
    port = ntohs(server.sin_port);
    
    /* Listen for connections on this socket. */
    listen(sock, 5);
    *sock_ptr = sock;
    return port;
}

/*********************************************************************
 *
 * server_socket - this function waits until VPO establishes a 
 *                 connection to "sock" and creates a new socket
 *                 with the properties of "sock" and allocates a new
 *                 file descriptor for this new socket.
 *
 *********************************************************************/
int server_socket(int sock) {

        int msgsock;    /* File descriptor of new socket. */

        /* Blocks if no pending connections from VPO. */
        msgsock = accept(sock, 0, 0);
        if (msgsock == -1) {
                perror("ERROR vet (server_socket): accept");
                exit(1);
        }
        return msgsock;
}

/*********************************************************************
 * 
 * write_on_socket - writes string to socket.
 *
 *********************************************************************/
void write_on_socket(char *vet_string) {

    /* Write returns -1 if there's an error. */
    if (write(sockfd, vet_string, (strlen(vet_string) + 1)) < 0) {
        perror("ERROR vet (write_on_socket): ");
        exit(1);
    }

#if DEBUG == DEBUG_HI
    /* Print out what was written to the socket into log file. */
    fprintf(errlog, "write_on_socket: %s\n", vet_string);
#endif

} /* write_on_socket */

/*********************************************************************
 *
 * read_from_socket - read characters from socket until reach EOS.
 *
 *********************************************************************/
char *read_from_socket(void) {

    static char vpo_string[BIGSTRING]; /* Array to hold characters. */
    static FILE *sf = NULL;
    char *string_ptr = vpo_string;     /* Ptr to array. */
    int position = 0;           /* Position in array */

    /* If sf is null, open a stream from sockfd. */
    if ((sf == NULL) && ((sf = fdopen(sockfd, "r+")) == NULL)) {
        perror("ERROR vet (read_from_socket): ");
        exit(1);
    }
        
    /* Initialize first character to EOS in case nothing is read in. */
    *string_ptr = '\0';

    /*
     * Read one character at a time from the socket until reach end 
     * of string.
     */
    do {
        *string_ptr = getc(sf);


        /* Count characters so far*/
        position++;
        if (*string_ptr == '\t') {
            *string_ptr = ' ';
            /* If characters are not at tab stop add more spaces */
            while (position % 8 != 0) {
                *++string_ptr = ' ';
                position++;
            }

        }

        /* Successful read will return -1 */
        if (ferror(sf)) {
            perror("ERROR vet (read_from_socket): ");
            exit(1);
        }
        /* Read will return 0 if EOF */
        else if (feof(sf)) {
            fprintf(stderr, "ERROR vet (read_from_socket): EOF reached\n");
            exit(1);
        }
    }
    /* Check for end of string. */
    while (*string_ptr++);

#if DEBUG == DEBUG_HI
    /* Print out what was read to log file */
    fprintf(errlog, "read_from_socket: %s\n", vpo_string);
#endif
    

    /* 
     * Since "vpo_string" is declared as static, then the memory continues
     * to be allocated outside this function.  Therefore the value of
     * "vpo_string" needs to be copied to a new memory location before this
     * function is called again.
     */
    return vpo_string;

} /* end read_from_socket */
