/***********************************************************************
 * 
 * Copyright (c) 1997 Ann H. Bailey.
 * 
 * This code is distributed WITHOUT ANY WARRANTY. No claims are made 
 * as to whether it serves any particular purpose or even works at all.
 *
 * $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/vet/vet_xinit.c,v $
 * $Revision: 1.1.1.1 $
 * 
 ***********************************************************************/
#include <stdio.h>	/* Move here so can compile without warnings 
			   when compiling with no XWINDOWS. */

#ifdef XWINDOWS

#include "vet.h"
#include "vet_xinit.h"
#include "vet_callbacks.h"
#include "process_input.h"
#include <Xm/CascadeB.h>
#include <Xm/DialogS.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/List.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrolledW.h>
#include <Xm/Separator.h>
#include <Xm/Text.h>
#include <Xm/ToggleB.h>

Widget initialize_menu(Widget);
Widget initialize_buttons(Widget);
Widget initialize_status(Widget);
Widget CreateLabeledTextForm(Widget, char *, char *);

Widget scrollwindow;
Widget funcnameinput;
Widget phasenameinput;
Widget transnuminput;
Widget transtotalinput;

Widget exitbutton, nextfuncbutton, applytransbutton, undotransbutton;
Widget nextphasebutton, nexttransbutton, prevphasebutton, prevtransbutton;
Widget breakbutton, transbreak_menu;
Widget transstate, statusmessage;
Widget operwidget, phasemenu;
Widget run, options;

void optionsdialog(Widget, XtPointer, XtPointer);

/*********************************************************************
 *
 * initialize_xwindow - this functions creates the basic shell and
 *			menu for VET.
 *
 *********************************************************************/
XtAppContext initialize_xwindow(int argc, char **argv) {

    XtAppContext app_context;
    Widget vet, mainwindow, menubar, buttonbar, mainform;
    Widget statuswindow;

    vet = XtVaAppInitialize(&app_context, "Vet", NULL,
			    0, &argc, argv, NULL, NULL);

    /* Create the main window shell to hold the main widgets */
    mainwindow = XtVaCreateWidget("MainWindow", 
				  xmFormWidgetClass,
				  vet,
				  NULL);


    /* Create the menu */
    menubar = initialize_menu(mainwindow);

    /* Create the status window */
    statuswindow = initialize_status(mainwindow);

    /* Attach the status window under the menu */
    XtVaSetValues(statuswindow, 
		  XmNtopAttachment, XmATTACH_WIDGET,
		  XmNtopWidget, menubar, 
		  NULL);

    /* Create the buttons for traversing */
    buttonbar = initialize_buttons(mainwindow);

    /* Create the scroll window to hold the basic block widgets. */
    scrollwindow = XtVaCreateWidget("ScrollWindow", 
				    xmScrolledWindowWidgetClass, 
				    mainwindow,
				    XmNscrollingPolicy, XmAUTOMATIC,
				    XmNscrollBarDisplayPolicy, XmAS_NEEDED,
				    XmNvisualPolicy, XmVARIABLE,
				    XmNheight, 400,
				    XmNwidth, 400,
				    XmNtopAttachment, XmATTACH_WIDGET,
				    XmNtopWidget, statuswindow,
				    XmNbottomAttachment, XmATTACH_WIDGET,
				    XmNbottomWidget, buttonbar,
				    XmNleftAttachment, XmATTACH_FORM,
				    XmNrightAttachment, XmATTACH_FORM,
				    NULL);

    XtManageChild(scrollwindow);
    XtManageChild(mainwindow);

    XtRealizeWidget(vet);
    
    return app_context;

} /* end initialize_xwindow */

Widget initialize_menu(Widget parent) {

    Widget menubar, filemenu, optionsmenu;
    Widget breakmenu, phasebreak, phasewidget;
    Widget exit;
    int index;

    menubar = XmCreateMenuBar(parent, "menubar", NULL, 0);    
    XtVaSetValues(menubar,
		  XmNtopAttachment, XmATTACH_FORM,
		  XmNleftAttachment, XmATTACH_FORM,
		  XmNrightAttachment, XmATTACH_FORM,
		  NULL);

    /* Create File Menu */
    filemenu = XmCreatePulldownMenu(menubar, "filemenu", NULL, 0);

    /* Create the "File" button */
    XtVaCreateManagedWidget("File",
			    xmCascadeButtonWidgetClass, menubar,
			    XmNsubMenuId, filemenu,
			    NULL);

    run = XtVaCreateManagedWidget("Run",
				  xmPushButtonWidgetClass,
				  filemenu, NULL);
				  
    options = XtVaCreateManagedWidget("Options",
				       xmPushButtonWidgetClass,
				       filemenu, NULL);

    XtVaCreateManagedWidget("Separator",
			    xmSeparatorWidgetClass,
			    filemenu, NULL);

    exit = XtVaCreateManagedWidget("Exit", 
				   xmPushButtonWidgetClass,
				   filemenu, NULL);


    XtAddCallback(run, XmNactivateCallback, startvpo, NULL);
    XtAddCallback(options, XmNactivateCallback, optionsdialog, NULL);
    XtAddCallback(exit, XmNactivateCallback, exit_pressed, NULL);


    /* Create Break Menu */
    breakmenu = XmCreatePulldownMenu(menubar, "breakmenu", NULL, 0);

    /* Create the "Break" button */
    XtVaCreateManagedWidget("Break",
			    xmCascadeButtonWidgetClass, menubar,
			    XmNsubMenuId, breakmenu,
			    NULL);

    phasebreak = XmCreatePulldownMenu(breakmenu, "Phases", NULL, 0);

    phasemenu = XtVaCreateManagedWidget("PhaseBreak",
					xmCascadeButtonWidgetClass, breakmenu,
					XmNsubMenuId, phasebreak,
					XmNsensitive, False,
					NULL);
				  
    for (index = 0; index < NUMPHASES; index++) {
	phasewidget = XtVaCreateManagedWidget(phasenames[index],
					      xmPushButtonWidgetClass,
					      phasebreak,
					      NULL);
	XtAddCallback(phasewidget, XmNactivateCallback, phase_break_callback, (XtPointer)index);
    }

    transbreak_menu = XtVaCreateManagedWidget("TransBreak",
					      xmPushButtonWidgetClass,
					      breakmenu, 
					      XmNsensitive, False,
					      NULL);

    XtAddCallback(transbreak_menu, XmNactivateCallback, break_pressed, NULL);

    XtManageChild(menubar);
    return menubar;

} /* end initialize_menu */

void optionsdialog(Widget parent, XtPointer client_data,
		    XtPointer call_data) {

    Widget displaydialog, options;
    Arg args[1];
    XmString prompt;
    Widget sanity, assembly, trash, use, reserve, comment;
    Widget set, csline, sides, deads;


    prompt = XmStringCreateLocalized("Select RTLs to Display");
    XtSetArg(args[0], XmNmessageString, prompt);
    displaydialog = XmCreateMessageDialog(scrollwindow, "Display", args, 1);
    XmStringFree(prompt);

    options = XtVaCreateWidget("Options", xmRowColumnWidgetClass,
			       displaydialog, 
			       XmNradioBehavior, False,
			       NULL);

    sanity = XtVaCreateManagedWidget("Sanity",
				     xmToggleButtonWidgetClass,
				     options, 
				     XmNsensitive, True,
				     NULL);

    assembly = XtVaCreateManagedWidget("Asm",
				     xmToggleButtonWidgetClass,
				     options, 
				     NULL);

    trash = XtVaCreateManagedWidget("Trash",
				     xmToggleButtonWidgetClass,
				     options, 
				     NULL);

    use = XtVaCreateManagedWidget("Use",
				     xmToggleButtonWidgetClass,
				     options, 
				     NULL);

    reserve = XtVaCreateManagedWidget("Reserve",
				     xmToggleButtonWidgetClass,
				     options, 
				     NULL);

    comment = XtVaCreateManagedWidget("Comment",
				     xmToggleButtonWidgetClass,
				     options, 
				     NULL);

    set = XtVaCreateManagedWidget("Set",
				     xmToggleButtonWidgetClass,
				     options, 
				     NULL);

    csline = XtVaCreateManagedWidget("Csline",
				     xmToggleButtonWidgetClass,
				     options, 
				     NULL);

    sides = XtVaCreateManagedWidget("Sides",
				     xmToggleButtonWidgetClass,
				     options, 
				     NULL);

    deads = XtVaCreateManagedWidget("Deads",
				     xmToggleButtonWidgetClass,
				     options, 
				     NULL);

    XtAddCallback(assembly, XmNvalueChangedCallback, asm_pressed, NULL);
    XtAddCallback(trash, XmNvalueChangedCallback, trash_pressed, NULL);
    XtAddCallback(use, XmNvalueChangedCallback, use_pressed, NULL);
    XtAddCallback(reserve, XmNvalueChangedCallback, reserve_pressed, NULL);
    XtAddCallback(comment, XmNvalueChangedCallback, comment_pressed, NULL);
    XtAddCallback(set, XmNvalueChangedCallback, set_pressed, NULL);
    XtAddCallback(csline, XmNvalueChangedCallback, csline_pressed, NULL);
    XtAddCallback(sides, XmNvalueChangedCallback, sides_pressed, NULL);
    XtAddCallback(sanity, XmNvalueChangedCallback, sanity_pressed, NULL);
    XtAddCallback(deads, XmNvalueChangedCallback, deads_pressed, NULL);

    XtManageChild(options);

    XmToggleButtonSetState(assembly, (rtls_to_display & ASMLINE)?True:False, False);
    XmToggleButtonSetState(trash, (rtls_to_display & TRASHLINE)?True:False, False);
    XmToggleButtonSetState(use, (rtls_to_display & USELINE)?True:False, False);
    XmToggleButtonSetState(reserve, (rtls_to_display & RESLINE)?True:False, False);
    XmToggleButtonSetState(comment, (rtls_to_display & COMMENT)?True:False, False);
    XmToggleButtonSetState(set, (rtls_to_display & SETLINE)?True:False, False);
    XmToggleButtonSetState(csline, (rtls_to_display & CSLINE)?True:False, False);
    XmToggleButtonSetState(sides, display_sides, False);
    XmToggleButtonSetState(sanity, sanity_enabled, False);
    XmToggleButtonSetState(deads, display_deads, False);

    /* No help available */
    XtUnmanageChild(XmMessageBoxGetChild(displaydialog, XmDIALOG_HELP_BUTTON));

    
    XtManageChild(displaydialog);

} /* end optionsdialog */

void phasebreak_menu(Widget parent, XtPointer client_data,
		    XtPointer call_data) {



} /* end phasebreak_menu */

Widget initialize_status(Widget parent) {

    Widget status, operscroll;
    Widget statusright, statusleft;

    status = XtVaCreateWidget("Status",
			      xmRowColumnWidgetClass,
			      parent,
			      XmNpacking, XmPACK_COLUMN,
			      XmNnumColumns, 2,
			      XmNorientation, XmVERTICAL,
			      XmNleftAttachment, XmATTACH_FORM,
			      XmNrightAttachment, XmATTACH_FORM,
			      NULL);

    statusleft = XtVaCreateWidget("StatusLeft",
				  xmRowColumnWidgetClass,
				  status,
				  XmNpacking, XmPACK_COLUMN,
				  XmNnumColumns, 2,
				  XmNorientation, XmVERTICAL,
				  XmNentryAlignment, XmALIGNMENT_BEGINNING,
				  NULL);

    funcnameinput = CreateLabeledTextForm(statusleft, "FuncName", "FuncNameInput");
    phasenameinput = CreateLabeledTextForm(statusleft, "PhaseName", "PhaseNameInput");
    transnuminput = CreateLabeledTextForm(statusleft, "TransNum", "TransNumInput");
    transstate = XtVaCreateManagedWidget("TransState",
					 xmTextWidgetClass, statusleft,
					 XmNeditable, False,
					 XmNtraversalOn, False,
					 XmNcursorPositionVisible, False,
					 XmNshadowThickness, 0,
					 NULL);

    statusmessage = XtVaCreateManagedWidget("StatusMessage",
					    xmTextWidgetClass, statusleft,
					    XmNeditable, False,
					    XmNtraversalOn, False,
					    XmNcursorPositionVisible, False,
					    XmNshadowThickness, 0,
					    NULL);
    XtManageChild(statusleft);
    
    statusright = XtVaCreateWidget("StatusRight",
			      xmRowColumnWidgetClass,
			      status,
			      XmNpacking, XmPACK_TIGHT,
			      XmNnumColumns, 2,
			      XmNorientation, XmVERTICAL,
			      NULL);
    XtVaCreateManagedWidget("Operations",
			    xmLabelWidgetClass, statusright,
			    XmNalignment, XmALIGNMENT_BEGINNING,
			    NULL);
    
    operscroll = XtVaCreateManagedWidget("OperScroll",
					 xmScrolledWindowWidgetClass, 
					 statusright,
					 NULL);
    operwidget = XtVaCreateManagedWidget("OperWidget",
					 xmListWidgetClass, operscroll,
					 XmNlistSizePolicy, XmCONSTANT,
					 XmNitems,0,
					 XmNitemCount, 0,
					 XmNselectionPolicy, XmSINGLE_SELECT,
					 XmNvisibleItemCount, 4,
					 XmNsensitive, False,
					 NULL);

    XtManageChild(statusright);
    XtManageChild(status);

    return status;

} /* end initialize_status */

Widget initialize_buttons(Widget parent) {

   Widget buttonbar;

   buttonbar = XtVaCreateManagedWidget("buttonbar", 
				       xmRowColumnWidgetClass,
				       parent, 
				       XmNpacking, XmPACK_COLUMN,
				       XmNorientation, XmHORIZONTAL,
				       XmNnumColumns, 2,
				       XmNisAligned, True, 
				       XmNentryAlignment, XmALIGNMENT_CENTER,
				       XmNbottomAttachment, XmATTACH_FORM,
				       XmNleftAttachment, XmATTACH_FORM,
				       XmNrightAttachment, XmATTACH_FORM,
				       NULL);

    nextfuncbutton = XtVaCreateManagedWidget("NextFunc",
					     xmPushButtonWidgetClass,
					     buttonbar, 
					     XmNsensitive, False,
					     NULL);

    nextphasebutton = XtVaCreateManagedWidget("NextPhase",
					      xmPushButtonWidgetClass,
					      buttonbar, 
					      XmNsensitive, False,
					      NULL);

    nexttransbutton = XtVaCreateManagedWidget("NextTrans",
					      xmPushButtonWidgetClass,
					      buttonbar, 
					      XmNsensitive, False,
					      NULL);

    applytransbutton = XtVaCreateManagedWidget("ApplyTrans",
					       xmPushButtonWidgetClass,
					       buttonbar, 
					       XmNsensitive, False,
					       NULL);

    breakbutton = XtVaCreateManagedWidget("Break",
					  xmPushButtonWidgetClass,
					  buttonbar,
					  XmNsensitive, False,
					  NULL);

    exitbutton = XtVaCreateManagedWidget("Exit",
					 xmPushButtonWidgetClass,
					 buttonbar, NULL);

    prevphasebutton = XtVaCreateManagedWidget("PrevPhase",
					      xmPushButtonWidgetClass,
					      buttonbar, 
					      XmNsensitive, False,
					      NULL);

    prevtransbutton = XtVaCreateManagedWidget("PrevTrans",
					      xmPushButtonWidgetClass,
					      buttonbar, 
					      XmNsensitive, False,
					      NULL);

    undotransbutton = XtVaCreateManagedWidget("UndoTrans",
					      xmPushButtonWidgetClass,
					      buttonbar, 
					      XmNsensitive, False,
					      NULL);

    XtAddCallback(exitbutton, XmNactivateCallback, exit_pressed, 0);
    XtAddCallback(nextfuncbutton, XmNactivateCallback, next_func_pressed, 0);
    XtAddCallback(nextphasebutton, XmNactivateCallback, next_phase_pressed, 0);
    XtAddCallback(prevphasebutton, XmNactivateCallback, prev_phase_pressed, 0);
    XtAddCallback(nexttransbutton, XmNactivateCallback, next_trans_pressed, 0);
    XtAddCallback(applytransbutton, XmNactivateCallback, apply_trans_pressed, 0);
    XtAddCallback(prevtransbutton, XmNactivateCallback, prev_trans_pressed, 0);
    XtAddCallback(undotransbutton, XmNactivateCallback, undo_trans_pressed, 0);
    XtAddCallback(breakbutton, XmNactivateCallback, break_pressed, 0);

    return buttonbar;
} /* end initialize_buttons */

Widget CreateLabeledTextForm(Widget parent, char *label_name, 
			     char *text_name) {

    Widget form, label, ret;
    form = XtVaCreateWidget(NULL, 
			    xmFormWidgetClass, parent,
			    XmNorientation, XmHORIZONTAL,
			    NULL);
    label = XtVaCreateManagedWidget(label_name, 
				    xmLabelWidgetClass, form,
				    XmNalignment, XmALIGNMENT_BEGINNING,
				    XmNleftAttachment, XmATTACH_FORM,
				    XmNtopAttachment, XmATTACH_FORM,
				    XmNbottomAttachment, XmATTACH_FORM,
				    NULL);
    ret = XtVaCreateManagedWidget(text_name,
				  xmTextWidgetClass, form,
				  XmNeditable, False,
				  XmNcursorPositionVisible, False,
				  XmNtraversalOn, False,
				  XmNleftAttachment, XmATTACH_WIDGET,
				  XmNleftWidget, label,
				  XmNtopAttachment, XmATTACH_FORM,
				  XmNrightAttachment, XmATTACH_FORM,
				  XmNbottomAttachment, XmATTACH_FORM,
				  XmNshadowThickness, 0,
				  NULL);
    XtManageChild(form);
    
    return ret;

} /* end CreateLabeledTextForm */
#endif

