#ifndef _BASEOBJ_H
#define _BASEOBJ_H

//
// commonly used macros
//
#include <Macros.h>


//
// It's inefficent to include other class definitions in a .h file.  we're just
// make a forward declaration that this class exists 
//
class OptionsDatabase;


//
//  Class BaseObject - all objects of are this type.  It allows us to name each object, 
//   for debugging purposes, and also allows the program options to be passed around cleanly.
//
class BaseObject
{
	public:

		//
		// common constructers
		//
		BaseObject();
		BaseObject(OptionsDatabase* opt);

		// copy constructure
		BaseObject(const BaseObject& obj);
		~BaseObject();	


		// how we set/read the name of this object
		void setName(char* name);
		char* getName() const;

	protected:

		OptionsDatabase* getOptions() const;

	private:


		// we want to force sub-classes to overload getOptions
		// so that no one else can use the 'options' field.
		OptionsDatabase* getExactOptions() const;
		OptionsDatabase* options;

		// likewise with the name
		char* name;

};

#endif


