#include <BaseRtlFile.h>
#include <LinkedRtlFile.h>
#include <RtlFile.h>
#include <OptionsDatabase.h>

#include "my_stl.h"

using namespace std;

LinkedRtlFile::LinkedRtlFile(OptionsDatabase* opt) : BaseRtlFile(opt) 
{
	maxGlobalsSeen = 0;
	maxLabelsSeen = 0;
	maxFilesSeen = 0;
}


void LinkedRtlFile::mergeNewFile(RtlFile* rtlFile)
{
	LabelNumType maxGlobalsSeen2 = 0;

	int wasEmpty = rtlFile->isEmptyFile();

	assert(rtlFile);

	printf("\tmerging...\t");

	fflush(stdout);


	// We have to do an initial renumber to make sure we don't
	// remove a global declare for a static variable. Also renames
   // locals and parameter declarations.
	maxGlobalsSeen2 = rtlFile->renumberGlobalsAndLocals(maxGlobalsSeen,
                                                       maxFilesSeen);
	maxLabelsSeen = rtlFile->renumberLabels(maxLabelsSeen);
   ++maxFilesSeen;

	// Now we can remove any duplicate definitions since all statics
	//	have been renamed to something original.
	rtlFile->removeDuplicates(&alreadyDeclaredGlobals);

	// Finally, we renumber *again* just to make sure we don't use any
	//		extra global spaces.
	// However, this step isn't necessary... so, we skip it for efficiency.
	//maxGlobalsSeen2 = rtlFile->renumberGlobals(maxGlobalsSeen);

	maxGlobalsSeen=maxGlobalsSeen2;
	
	list<Function*> *rtlFuncs=rtlFile->getFunctions();
	list<Function*>::iterator it=rtlFuncs->begin();

   for (; it != rtlFuncs->end(); ++it) {
      functions.push_back(*it);
   }

	/* push an empty function here */
	/* so we get a "file terminator" between files */
	functions.push_back(new Function());	 


	/* clear the functions from the old RtlFile */
	rtlFuncs->erase(rtlFuncs->begin(), rtlFuncs->end());

	if (wasEmpty)
		printf("Empty input file\n");	
	else
		printf("Done\n");
}


void LinkedRtlFile::optimize()
{
	printf("Optimizing merged files\n");
}

void LinkedRtlFile::output()
{
	OptionsDatabase* options=getOptions();
	char* fn=options->getOutputFilename();
	FILE* fout;

        printf("Total of %d globals in this link\n", (int)maxGlobalsSeen);

 	if(fn==0)
	{
		fout=stdout;
		printf("Outputing linked file to stdout\n");
	}	
	else
	{
		printf("Outputing linked file to %s\n", fn);
		fout=fopen(fn, "w");
		if(!fout)
		{
			char temp[100];
			sprintf(temp, "Cannot open output file %s", fn);
			fatal(temp);
		}
	}

	list<Function*>::iterator it=functions.begin();

        for(; it != functions.end(); ++it)
        {
		(*it)->output(fout);
	}

}









