#ifndef _LINKEDRTLFILE_H
#define _LINKEDRTLFILE_H

#include <Macros.h>
#include <BaseRtlFile.h>
#include <string>

#include "my_stl.h"

class RtlFile;

//
// This class represents an RTL file that is capable of merging in new RTL
// files.  Note that the merge is destructive to the original RTL file
// (e.g.: the functions get stolen).
//
class LinkedRtlFile : public BaseRtlFile
{

	public:

		// We need an options database for proper output, thus
		// we prohibit the empty constructor. 
                LinkedRtlFile() NOT_IMPLEMENTED;
		// Avoid shallow copies.
                LinkedRtlFile(const LinkedRtlFile&) NOT_IMPLEMENTED;

		// Use this constructor exclusively.
                LinkedRtlFile(OptionsDatabase* opt);

		// Optimize currently does nothing, but is a stub for any whole program 
		// optimizations one might want to try.
		void optimize();

		// Output the linked RTL file to where the option says it should go.
		void output();

		// merge in and destroy rtlFile 
		void mergeNewFile(RtlFile* rtlFile);

	private:

	protected:

		// As we merge in new files, we'll need to renumber any labels and
      // static global variables, and rename locals (including function
      // parameters) by prefixing a file number to their name.
		// We use these to keep track of the "highest" label/global/file we've
      // seen so far so we can safely renumber and/or rename them.
		LabelNumType maxGlobalsSeen;
		LabelNumType maxLabelsSeen;
		int maxFilesSeen;

		// This map is used to map the global identifier code with its name.
		// It's important to not redeclare "printf" a bunch of times because we
		// may run out of global identifier space too soon if we allow it to be
      // declared multiple times.
		// Also, we don't want to "trick" vpo into thinking two globals are
		// different if they are really the same name.
  		map<string,string,less<string> > alreadyDeclaredGlobals;

			
};
#endif
