#include <iostream>
#include <string>
#include <map>

#include <stdlib.h>

#include "BaseRtlFile.h"
#include "RtlFile.h"
#include "rtltypes.h"
#include "machInterface.h"

extern "C" {
#include "../../../lib/rtlspec.h"
};

using namespace std;

char* getData(string &a);
static char* replaceLabel(map<string,string,less<string> > labelMap, char* rtl);
static char* replaceDeclaration(map<string,string,less<string> > labelMap, char* rtl);
/*
 * gtline - return remainder of input line
 */
static char *gtline(FILE* fin) 
{
	static int line=0;
	line++;
   	register int len;
   	register char *a;
   	char str[1000];
	
	
   	a = str;
   	for (len = 1; (*a++ = getc(fin)) != '\n'; len++)
   	{
		if(len>1000)
		{
			*a=0;
			printf("line=%d, str=%s\n", line, str);
			fatal("Input line too long");
		}
   	}
   	*(a - 1) = '\0';
   	return (strdup(str));
}

void RtlFile::readFile(char* filename)
{
	printf("Reading file %s...", filename);
	fflush(stdout);
	
	FILE* fin=fopen(filename, "r");	

	if(!fin)
	{
		printf("Cannot open %s, continuing to next file\n", filename);
		return;
	}
	
	Function* func=new Function();
	assert(func);


	int done=FALSE;
	while(!done)
	{
		/* get the rtl type */
		int type=getc(fin);

		switch(type)
		{
			/* end of file is found, break out */
			case EOF:
			{
				done=TRUE;
				/* add the last function */
				functions.push_back(func);	
				break;


			}
			/* the 'basename' of this file */
			case BASENAMERTL:
			{
				setName(gtline(fin));
				break;
			
			}
			/* end of function */
			case ENDOFFUNCRTL:
			{
				functions.push_back(func);	
			        func=new Function();
				gtline(fin);
       				assert(func);
				break;
			}
			case ARGBUILDRTL     :
			case ASMRTL          :
			case DELAYRTL        :
			case TRASHRTL        :
			case RESERVEold      :
			case RESERVEnew      :
			case PARAMRTL        :
			case SETRTL          :
			case CALLERSAVERTL   :
			case COMMENTRTL      :
			case REGMAPRTL       :
			case GLOBALREFRTL    :
			case SIDEEFFECTRTL   :
			case VARARGRTL       :
			case USERTL          :
			case DECLARATIONRTL  :
			case CASELINERTL1     :
			case CASELINERTL2     :
			case ENDFLOWRTL      :
			case LABELRTL        :
			case INSTRRTL        :
			case METADATARTL        :
			{
				char* newRtl=gtline(fin);
				func->addRtl(type,newRtl);
				break;

			}
			case FUNCNAMERTL:
			{
                                char* newRtl=gtline(fin);
                                func->addRtl(type, newRtl);
				func->setName(newRtl);
                                break;
			}
			default:
			{
				fprintf(stderr, "Bad rtl! Type=%c, line=%s\n",
					type, gtline(fin));
				break;
			}
		}
		
	}
	fclose(fin);
}



int isGlobalUse(char* id2)
{
	return isGlobal(id2);
}
int isLocalUse(char* id2)
{
	return isLocal(id2);
}

int isRegisterUse(char* id2)
{
	return isRegister(id2);
}


static int isGlobalDeclaration(Rtl theRtl)
{
	char* a=theRtl.getValue();

	while(*a++!='\t');// skip the name we don't care what it is 
	
	// identifier of this variable 
	a = skiploc(a);

	assert(*a == '\t');
   ++a; // skip the tab 

	char* startType = a;

	// the first and only digit of this field is 0
	// return true, otherwise return false
	return (*a == '0' && *(a+1) == '\t');	

}

static char* getGlobalName(char* a);
static char* getGlobalName(Rtl theRtl)
{
	return getGlobalName(theRtl.getValue());
}
static char* getGlobalName(char* a)
{
	char b[1000];
	char* c=b;
	int len=0;

	// copy the name until we get a tab
        while((*c++=*a++)!='\t')
	{
		len++;
		assert(len<1000);
	};

	*(c-1)=0;	// null terminate c in place of the tab 

	return strdup(b);
}


static void replaceID(Rtl *theRtl, char* newId)
{
        char* a = (char *) strdup(theRtl->getValue());
        char* b = a;

        while(*a++ != '\t');	// skip the name we don't care what it is

	copyloc(a, newId);
	a = skiploc(a);
	newId = skiploc(newId);
   assert(*a == '\t');
   ++a; 	// skip the tab

   // sanity check that this is a global
	// if so, the first and only digit of this field is 0
   assert (*a == '0' && *(a+1) == '\t');

	theRtl->setValue((char *) b);
}


static char* findID(Rtl theRtl)
{
        char* a=theRtl.getValue();      

        while(*a++!='\t');// skip the name we don't care what it is 
        
        char id1[100];
	copyloc(id1,a);
	*skiploc(id1)=0;

	return (char*) strdup((char*)id1);

}



/* see page 5 of "Register Transfer Standard" */
static int globalIDtoNum(char* id)
{
	return getGlobalNum(id);
}

static char* numToGlobalID(int num)
{
	char* id=getGlobal(num);

	assert(globalIDtoNum(id)==num);	

	return (char*)strdup((char*)id);
	
}

static int replaceRef(char* oldref, char* newref)
{
	copyloc(oldref,newref);
}


// Renumber the global ID for all global variables, using startNum
// as a base, and incrementing it with each renumbering.  Rename
// locals (including parameters) by adding a prefix such as ".p##"
// for parameters, or ".l##" for locals, where the "##" field is the file
// number.  The front end will keep all local and parameter names distinct
// within a file, but not across the files that we are merging.
// Return the update value of startNum.
LabelNumType RtlFile::renumberGlobalsAndLocals(LabelNumType startNum,
                                               int fileNum)
{
   map<int, int, less<int> > globalMap;
   list<Function*>::iterator func_it=functions.begin();
   int temp = fileNum;
   int digits = 1;  // How many digits in fileNum.

   assert(0 <= temp);
   while (temp > 9) {
      temp /= 10;
      ++digits;
   }
   
   for (; func_it != functions.end(); ++func_it) {
      Function* func = *func_it;
      list<Rtl*>::iterator rtl_it = func->getRtls()->begin(); 
      for (; rtl_it != func->getRtls()->end(); ++rtl_it) {
         Rtl *theRtl = *rtl_it;
         switch (theRtl->getType()) {
            // If it's a global declare, renumber it.
            // If it's a local declare, rename it.
         case DECLARATIONRTL:
            {
               if (!isGlobalDeclaration(*theRtl)) { // Local or parameter.
                  char *val = theRtl->getValue();
                  char newVal[1000];
                  char *a = val;
                  char *b = newVal;
                  char *temp;
                  assert(NULL != val);
                  while (*a && (*a != '.')) {
                     *b++ = *a++;
                  }
                  // Should be stopped at the beginning of a local, temp, or
                  // parameter, which will look like ".l3.4_foo" for
                  // a local, ".T359" for a temp, or ".p8.0_lambda"
                  // for a parameter.
                  assert(*a == '.');
                  if ((*(a + 1) != 'p') && (*(a + 1) != 'l')
                      && (*(a + 1) != 'T')) {
                     printf("Unknown declaration: %s\n", val);
                     break;
                  }
                  assert((*(a + 1) == 'p') || (*(a + 1) == 'l')
                         || (*(a + 1) == 'T'));
                  // Now, prefix an extra ".l", "T" or ".p" followed by the
                  // file number.
                  *b++ = *a;        // Get the dot.
                  *b++ = *(a + 1);  // Get the "p" or "l" or "T".
                  temp == lltostr((long long) fileNum,
                                  b + digits); // Int to ASCII, into b.
#if 0
                  printf("temp: %x b: %x\n", temp, b);
                  fflush(stdout);
                  assert(temp == b);  // temp is start, b + digits - 1 is end.
#endif
                  b += digits;
                  // Copy the rest of the RTL value after the new prefix.
                  while (*a != '\0')
                     *b++ = *a++;
                  *b = '\0'; // Null-terminate b, a.k.a. newVal.
                  theRtl->setValue(strdup(newVal)); // Store new RTL.
                  printf("Renaming %s to %s\n", val, newVal);
               }
               else { // Global.
                  char* id = findID(*theRtl);
                  int globalNum = globalIDtoNum(id); 
                  char* newId = numToGlobalID(startNum);
                  replaceID(theRtl, newId);
                  GetRidOfArray(newId);
               
                  globalMap[globalNum] = startNum;
                  startNum++;
                  //printf("Mapping %d to %d\n", globalNum,startNum);
               }
               break;
            }

         case INSTRRTL: 
            {
               /* otherwise search through the RTL */
               /* looking for a global        */
               char* a = strdup(theRtl->getValue());
               char* b = a;
               while (*a != 0 && *(a+1) != 0) {
                  if (isGlobalUse(a)) {
                     int globalNum = globalIDtoNum(a);
                     if (globalMap.find(globalNum) != globalMap.end()) {
                        char* newId = numToGlobalID(globalMap[globalNum]);
                        replaceRef(a, newId);
                     }
                  }
                  a = skiploc(a);
               }
               theRtl->setValue(b);
            }
         } // end switch (theRtl->getType())
      } // end for (; rtl_it != ...)
   } // end for (; func_it != ... )

   return startNum;
} // end renumberGlobalsAndLocals()


void print(string i) 
{
    cout << " '" << i << "'";
}

LabelNumType RtlFile::renumberLabels(LabelNumType startNum)
{
   list<string> globalNames;
   list<string> staticNames;
   list<string> externNames;
   list<string> labelNames;
   list<string> notExterns;
   list<string> noDeclare;
	
   LabelNumType maxSoFar=0;


   // make one pass over the file and snarf up and global 
   // names.
   list<Function*>::iterator func_it=functions.begin();
   for(; func_it != functions.end(); ++func_it)
   {
      Function* func=*func_it;
      list<Rtl*>::iterator rtl_it=func->getRtls()->begin();
      for(; rtl_it != func->getRtls()->end(); ++rtl_it)
      {
         Rtl *theRtl=*rtl_it;
         switch(theRtl->getType())
         {
		case DECLARATIONRTL:
               	{
			if(!isGlobalDeclaration(*theRtl))
				break;
			char* globName=getGlobalName(*theRtl);
			string blah(globName);
			globalNames.remove(blah);
			globalNames.push_back(blah);
               	}
         }
      }
   }

   noDeclare=globalNames;

   func_it=functions.begin();
   for(; func_it != functions.end(); ++func_it)
   {
      Function* func=*func_it;
      list<Rtl*>::iterator rtl_it=func->getRtls()->begin();
      for(; rtl_it != func->getRtls()->end(); ++rtl_it)
      {
	 Rtl *theRtl=*rtl_it;
 	 switch(theRtl->getType())
	 {
		case ASMRTL:
		case DELAYRTL:
		{
			char* name= getGlobalDeclareName(theRtl->getValue());

			/* did we find a external symbol */
			if(name!=0)
			{
				string blah(name);

				/* avoid duplicates */
				externNames.remove(blah); 	
				externNames.push_back(blah);	
				globalNames.remove(blah);
				break;	
			}

 			name=getStaticDeclareName(theRtl->getValue());

                        /* did we find a external symbol */
                        if(name!=0)
                        {
                                string blah(name);
                                staticNames.remove(blah);/* avoid duplicates */
                                staticNames.push_back(blah);  
				globalNames.remove(blah);
				break;	
                        }

                        name=findDeclare(theRtl->getValue());
                        /* did we find a external symbol */
                        if(name!=0)
                        {
                                string blah(name);
				noDeclare.remove(blah);
                                break;
                        }

			break;
		}
		case LABELRTL:
		{
			string blah(theRtl->getValue());
			labelNames.remove(blah);
			labelNames.push_back(blah);
		}
	 }
      }
   }

   int ual=unknownsAreLocal();
   list<string>::iterator it=globalNames.begin();
   for(; it!=globalNames.end(); ++it)
   {
   	if(ual) 
 	{
		staticNames.push_back(*it);
   	}
	else
	{
		externNames.push_back(*it);
	}
   }

   it=noDeclare.begin();
   for(; it!=noDeclare.end(); ++it)
	staticNames.remove(*it);


   map<string,string,less<string> > staticMapping;
   map<string,string,less<string> > labelMapping;
   map<string,string,less<string> > allMapping;

   it=staticNames.begin();
   for(; it!=staticNames.end(); ++it)
   {
	char temp[10];
	sprintf(temp, "L%lld", (unsigned long long)startNum++);
	staticMapping[*it]=string(temp);
	allMapping[*it]=string(temp);
//cout<<"Renaming static " << (*it) << " to " << temp<<endl;
   }

   it=labelNames.begin();
   for(; it!=labelNames.end(); ++it)
   {
        char temp[10], temp2[10];
        sprintf(temp, "L%lld", (unsigned long long)startNum++);
	string blah=*it;
	*temp2=0;
        strcat(temp2, "L");
	strcat(temp2+1, getData(blah));
	
        labelMapping[string(temp2)]=string(temp);
        allMapping[string(temp2)]=string(temp);
//printf("Renaming label %s to %s\n", temp2, temp);
   }

   func_it=functions.begin();
   for(; func_it != functions.end(); ++func_it)
   {
      Function* func=*func_it;
      list<Rtl*>::iterator rtl_it=func->getRtls()->begin();
      for(; rtl_it != func->getRtls()->end(); ++rtl_it)
      {
         Rtl *theRtl=*rtl_it;
	 char* newRtl;
	 newRtl=0;
         switch(theRtl->getType())
         {
                case ASMRTL:
                case DELAYRTL:
		{
			newRtl=replaceReferences(allMapping,theRtl->getValue());
			break;
		}
                case LABELRTL:
	        {
			newRtl=replaceLabel(labelMapping, theRtl->getValue());
			break;
		}
		case CASELINERTL1:
		case CASELINERTL2:
		{
			if(labelMapping.find(string(theRtl->getValue()))!=
				labelMapping.end())
			{
				theRtl->setValue(getData(labelMapping[
					string(theRtl->getValue())]));
			}
			break;
		}
                case DECLARATIONRTL:
	        {
			newRtl=replaceDeclaration(staticMapping, 
				theRtl->getValue());
			break;
			
		}
		case INSTRRTL:
		{
                        newRtl=fixRtl(allMapping,theRtl->getValue());
                        break;

		}
		case FUNCNAMERTL:
		{
			if(staticMapping.find(string(theRtl->getValue()))
				!= staticMapping.end())
			{
				char temp[1000];
				strcpy(temp, getData(
				  staticMapping[string(theRtl->getValue())]));
				newRtl=temp;
			}
			break;
		}
	 }
	 if(newRtl!=0)
	 {
		theRtl->setValue(newRtl);
	 }
      }
    }
    return startNum;
}

static char* replaceLabel(map<string,string,less<string> > labelMap, char* rtl)
{
	char temp[100];
	char temp2[100];
	sprintf(temp, "L%s", rtl);
	*temp2='L';
	if(labelMap.find(string(temp)) != labelMap.end())
	{
		strcpy(temp2, getData(labelMap[string(temp)])); 
		return strdup(temp2+1);
	}	
}

static char* replaceDeclaration(map<string,string,less<string> > labelMap, char* rtl)
{
	char* temp=getGlobalName(rtl);

	if(labelMap.find(string(temp))!= labelMap.end())
	{
		char temp2[1000];
		char* a=temp2;
		char* b=rtl;
		const char* c=getData(labelMap[string(temp)]);
		int len=strlen(c);
		int newlen=0;
		while(newlen++<len)
			*a++=*c++;
		a++;
		*a=0;
		a--; // back up over null terminator 
		while(*b++!='\t');
		b--; // back up to catch the tab
		while((*a++=*b++)!=0);

		return strdup(temp2);
	}
	return 0;
}






void RtlFile::removeDuplicates(map<string,string,less<string> > *alreadyDefined)
{
   map<string,string,less<string> > duplicates;

   // make one pass over the file and snarf up and global
   // names.
   list<Function*>::iterator func_it=functions.begin();
   for(; func_it != functions.end(); ++func_it)
   {
      Function* func=*func_it;
      list<Rtl*>::iterator rtl_it=func->getRtls()->begin();
      for(; rtl_it != func->getRtls()->end(); )
      {
         Rtl *theRtl=*rtl_it;
         switch(theRtl->getType())
         {
                case DECLARATIONRTL:
		{
			if(isGlobalDeclaration(*theRtl))
			{
				char* name=getGlobalName(*theRtl);
				char *id=findID(*theRtl);
//printf("Found global '%s'..", name);
				if((*alreadyDefined).find(string(name))!=
					alreadyDefined->end())	
				{
//printf("as a duplicate\n");
					//duplicate is found!
	
					// consquently we dont need this
					// global define 
					// erase bumps the iterator by one 
					list<Rtl*>::iterator temp_it=rtl_it;
					++rtl_it;
					func->getRtls()->erase(temp_it);

					//and we add a mapping in the duplicates
					// map
					char* prevId=getData(
					  (*alreadyDefined)[string(name)]);
//char* a=theRtl->getValue();
//while(*a++!='\t');
//printf("replacing %s (%c%c%c%c) with %c%c%c%c\n", theRtl->getValue(),
//a[0], a[1], a[2], a[3], prevId[0], prevId[1], prevId[2], prevId[3]);
					duplicates[string(id)]=
					  (*alreadyDefined)[string(name)];
				}
				else
				{
//printf("as a new symbol (id=%s)\n",id);
					// a new symbol
					// add it to alreadyDefined so
					// other we never define it again
					(*alreadyDefined)[string(name)]=
						string(id);
					++rtl_it;
				}
			}
			else
			{
				++rtl_it;
//if (theRtl->getValue()) printf("Found not-global '%s'\n", theRtl->getValue());
//else printf("Found empty declaration?\n");
			}
			break;
		}
		case INSTRRTL:
		{
			char temp[1000];
			strcpy(temp, theRtl->getValue());
			char* a=temp;
			while(*a!=0)
			{
				if(isRegisterUse(a) || isLocalUse(a))
				{
					a=skiploc(a);
				}
				else if(isGlobalUse(a))
				{
					char temp2[10];
				        copyloc(temp2,a);
					*skiploc(temp2)=0;
//printf("Found use of (%o %o %o %o)..",temp2[0],temp2[1], temp2[2], temp2[3]); 

					// is this something we're replacing
					// in this file?
					if(duplicates.find(string(temp2))!=
						duplicates.end())
					{
					  const char* b=duplicates[
						string(temp2)].data();
					  // duplicate found, replace entry
					  copyloc(a,(char*)b);
//printf("replacing with (%o %o %o %o)..\n",b[0],b[1], b[2], b[3]); 
					}
//else
//printf("not replacing.\n");

					a=skiploc(a);
				}
				else
				{
					a++;
				}
			}
			theRtl->setValue(temp);	
			++rtl_it;
			break;
		}
		default:
			++rtl_it;
		
	 }
      }
   }
}


char* getData(string &a)
{
	char temp[1000];
	
	strncpy(temp, a.data(), a.length());
	temp[a.length()]=0;

	return strdup(temp);
}

void RtlFile::output(char* blah)
{
	/* open the file is specified */
        FILE* fout=stderr;
	if(blah!=NULL)
		fout=fopen(blah,"w");

	if(!fout)
		fout=stderr;


        list<Function*>::iterator it=functions.begin();

        for(; it != functions.end(); ++it)
        {
                (*it)->output(fout);
        }

}

