
//**************************************************************//
// File:   	vlink.c						//
//								//
// Purpose: 	Main driver for the vlink program		//
//								//
// Author: 	Jason D. Hiser					//
//								//
// Created:  	7-12-2000					//
//								//
//**************************************************************//

//**************************************************************//
//								//
//Function: main						//
//								//
//Purpose:  the top level functionality.  			//
//								//
//Input:    (from command line) a list of options		//
//		followed by a list of files to link together	//
//								//
//Output:   assembly/rtl version of the multiple files linked 	//
//		together					//
//								//
//**************************************************************//



//
// header file that includes information about all options 
// 
#include <OptionsDatabase.h>

//
// include the information about a single rtl file
//
#include <RtlFile.h>

//
// Include information about linking an RtlFile together
//
#include <LinkedRtlFile.h>

//
// Useful macros
//
#include <Macros.h>

using namespace std;

int main(int argc, char** argv)
{

	//
	// read in and parse the options and save in a database
	//
	OptionsDatabase *options=new OptionsDatabase(argv, argc);
	assert(options);
	options->setName("Options Database");

	LinkedRtlFile* linked=new LinkedRtlFile(options);
	linked->setName("Linked");

	//
	// Loop through the input files and merge each one into the linked
	// File
	//
	list<char *>* filenames=options->getInputFilenames();
	assert(filenames);

	list<char*>::iterator it=filenames->begin();

	for(; it != filenames->end(); ++it)
	{
		char* filename=*it;
		RtlFile* rtlFile=new RtlFile(options);
		assert(rtlFile); 
		rtlFile->setName(filename);
		rtlFile->readFile(filename);
		if(!rtlFile->isEmptyFile())
		{
			linked->mergeNewFile(rtlFile);
		}
		GetRidOfObject(rtlFile);
	}


	// 
	// Optimize the merged input file
	//
	linked->optimize();

	//
	// Output it, to stdout or a file as dictated by the Options 
	// 
	linked->output();


	//
	// Cleanup memory
	//
	GetRidOfObject(linked);
	GetRidOfObject(options);

}
