/*
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/vpoiso/vpoiso.c,v $
 $Date: 2003/01/03 17:31:37 $
 $Revision: 1.11 $
 $Name:  $

 Log of changes:

 $Log: vpoiso.c,v $
 Revision 1.11  2003/01/03 17:31:37  jg9h
 Generate the rtl files once at the very end (instead of on each pass)

 Revision 1.10  2002/12/13 19:54:51  jdh8d
 added verbose option to vpoiso

 Revision 1.9  2002/11/18 19:34:01  jdh8d
 debug fixes, arm pre/post modification fixes

 Revision 1.8  2002/10/31 19:53:00  clc5q
 Fix the time limit feature by changing the executed shell command from
 "limit cputime" to "ulimit -t".

 Revision 1.7  2002/10/10 12:02:39  jg9h
 Now vpoiso creates the correct and incorrect versions in rtl format also (<filename>.incorr.rtl and <filename>.corr.rtl)

 Revision 1.6  2001/09/14 12:48:51  jdh8d
 no message

 Revision 1.5  2001/08/14 13:50:31  jdh8d
 *** empty log message ***

 Revision 1.4  2001/06/08 15:07:33  jdh8d
 *** empty log message ***

 Revision 1.3  2001/04/04 19:42:15  jdh8d

 increased the max number of funcs that can be in one file.

 Revision 1.2  2001/02/20 14:33:25  jdh8d

 Changed to emit a .s instead of a .o;  assembling is
 different on different machines.

 Revision 1.1.1.1  2000/08/22 21:10:34  zephyr
 Created directory structure

 Revision 1.1  2000/07/20 16:39:19  rv3s
 *** empty log message ***

 Revision 1.1  2000/06/10 19:32:05  jwd
 Initial revision

*/

#include <stdio.h>
#include <stdlib.h>

#define MAXFUNCS 1000
#define MAXFILES 25
#define MAXNAME 600
#define MAXCMD 1600
#define MAXDISREGARD 10
#define TRUE -1
#define FALSE 0
#define RECOMPILE 1

struct {
   char basename[MAXNAME];
   char funcname[MAXNAME];
   int numtrans;
   int starttrans;
   int endtrans;
   int compiledwithopts;
   } funcinfo[MAXFUNCS];

struct {
   char basename[MAXNAME];
   int firstfuncnum;
   int lastfuncnum;
   } fileinfo[MAXFILES];

char linkcmd[MAXCMD];
char executecmd[MAXCMD];
char rmcmd[MAXCMD];
char desiredoutfile[MAXNAME];
char actualoutfile[MAXNAME];
char compileflags[MAXNAME];
char compileflags_rtl[MAXNAME]; /* same as compileflags but without the A (so as to get rtls)*/
char disregard[MAXDISREGARD][MAXNAME];

int numfiles, numfuncs, numtrans, numdisregard;

unsigned long run_count = 0;

int verbose=FALSE;

int verbose_system(char* line)
{
	if(line && verbose)
	{
		printf("Executing : %s\n", line);
	}
	return system(line);
}


void error(char *msg)
{
   fprintf(stderr, "%s\n", msg);
   exit(1);
}

/*
 * get_env - return an environment value
 */
char *get_env(char **envp, char *name)
{
    register char *a, *b;

    if (envp) {
      while (*envp) {
        a = *envp;
        b = name;
        while (*a == *b) {
          a++;
          b++;
          }
        if (*a == '=' && !*b)
          return(a + 1);
        *envp++;
        }
      }
    return((char *) NULL);
}

int check_output()
{
   int i;
   char line[MAXCMD];
   char a[10];
   static int ask_output = TRUE;

   for (i = 0; i < numdisregard; i++) {
      sprintf(line, "grep -v '%s' %s > disregard.txt", disregard[i],
              actualoutfile);  
      verbose_system(line);
      sprintf(line, "mv disregard.txt %s", actualoutfile);
      verbose_system(line);
      }

   /* prompt user on how to proceed. Be flexible, allow user to force */
   /* fail, force pass, go ahead and compare output file against desired */
   /* or do the comparison and never prompt again. */
   if (ask_output) {
      fprintf(stderr, "---> execution result ([p]ass/[f]ail/[c]ompare/[d]on't ask again)? ");
      a[0]=getchar(); getchar();
      if (a[0]=='d' || a[0] == 'D') 
         ask_output = FALSE;
      else if (a[0] != 'c' && a[0] != 'C') 
         return (a[0]=='f' || a[0] == 'F');
   }
   sprintf(line, "cmp -s desired.txt %s", actualoutfile);
   return verbose_system(line);
}

void prepare_maxtrans_file(int upto);
main(int argc, char **argv, char **envp)
{
	int wrong_trans;
	int right_trans;
   int i, j, badfunc, cutnum, lastresultbad;
   int lastmax, lastmin, needstocompile, lastwithin, within;
   char line[MAXCMD];
   FILE *f;
   FILE *fopen();
   int verbose_system();
   char *spec_file;
   char *vpo, *chome;
   char *get_env();
   void error();
   char* verbose_p;


   if (argc > 1) {
      spec_file = argv[1];
   } else {
      spec_file = "isolate.txt";
   }

   /* get values of environmental variables */
   if (!(vpo = get_env(envp, "VPO")))
      error("VPO environment variable is not set");
   if (!(chome = get_env(envp, "CHOME")))
      error("CHOME environment variable is not set");


   /* get/set the verbose level for vpoiso */
   verbose_p=get_env(envp,"VERBOSE_VPOISO");
   if(verbose_p)
	verbose=atoi(verbose_p);
   if(verbose)
	fprintf(stderr,"Verbose level set to %d\n",verbose);

   /* first remove the two data files we'll be using */
   verbose_system("rm -f numtrans.txt maxtrans.txt");

   verbose_system("date '+START TIME: %H:%M:%S'");
   fprintf(stderr, "\n");


   /* read in setup information */
   numfiles = 0;
   f = fopen(spec_file, "r");
   if (!f) {
		sprintf(line,"could not open specification file: %s", spec_file);   
      error(line);
	}
   if (fscanf(f, "cexfiles: %s", fileinfo[numfiles].basename) != 1)
      error("no cexfiles found");
   while (fscanf(f, "%s", fileinfo[++numfiles].basename) != EOF)
      if (fileinfo[numfiles].basename[0] == '#')
         break;
      else {
         fileinfo[numfiles].firstfuncnum = MAXFUNCS+1;
         fileinfo[numfiles].lastfuncnum = -1;
         }
   fscanf(f, "\n");
   fgets(line, MAXCMD, f);
   if (strncmp(line, "link command: ", 14))
      error("no link command");
   strcpy(linkcmd, line+14);
   strcpy(executecmd, "echo \"ulimit -t ");
   fgets(line, MAXCMD, f);
   if (strncmp(line, "maximum time: ", 14))
      error("no maximum time");
   strcat(executecmd, line+14);
   executecmd[strlen(executecmd)-1] = '\0';
   strcat(executecmd, "; ");
   fgets(line, MAXCMD, f);
   if (strncmp(line, "execute command: ", 17))
      error("no execute command");
   strcat(executecmd, line+17);
   executecmd[strlen(executecmd)-1] = '\0';
   strcat(executecmd, "\" | sh -s");
   if (fscanf(f, "desired output file: %s\n", desiredoutfile) != 1)
      error("no desired output file");
   if (fscanf(f, "actual output file: %s\n", actualoutfile) != 1)
      error("no actual output file");
   
   /* $$$ note, must always at least supply *some* compilation flag */
   /* because parsing the command line switches in vpo assumes there */
   /* are flags. Need to make this clear in the isolate.txt user doc. */
   fscanf(f, "compilation flags: "); 
   fgets(compileflags, sizeof(compileflags), f);	
   compileflags[strlen(compileflags)-1]=0;
	strcpy(compileflags_rtl,compileflags);

	{
		/* The compileflags_rtl is a copy of compileflags, but with the A flag (for generate assembly) removed. */
		char *a=(char*)strchr(compileflags_rtl,'A');
		char *b=a+1;
		if (a) {
			while (*a++=*b++);
		}
	}
	
   fgets(line, MAXCMD, f);
   if (strncmp(line, "disregard strings:", 18) != 0)
      error("no disregard strings line");
   numdisregard = 0;
   for (i = 19; line[i] && line[i] != '\n'; ) {
      for (j = 0; line[i] && line[i] != '&' && line[i] != '\n'; j++, i++)
         disregard[numdisregard][j] = line[i];
      disregard[numdisregard][j] = '\0';
      if (line[i] == '&')
         i++;
      if (++numdisregard > MAXDISREGARD)
         error("too many disregard strings");
      }
   fclose(f);
   sprintf(line, "cp %s desired.txt", desiredoutfile);
   verbose_system(line);
   for (i = 0; i < numdisregard; i++) {
      sprintf(line, "grep -v '%s' desired.txt > disregard.txt", disregard[i],
              desiredoutfile);  
      verbose_system(line);
      sprintf(line, "mv disregard.txt desired.txt", desiredoutfile);
      verbose_system(line);
      }
   sprintf(rmcmd, "rm %s\n", actualoutfile);

   /* compile files to dump out number of transformations as a
      side effect */
   fprintf(stderr, "\n\n==================== Pass %lu\n", run_count++);
   fprintf(stderr, "compiling program to check if works with all improving transformations\n");
   for (i = 0; i < numfiles; i++) {
      fprintf(stderr, "compiling %s.cex\n", fileinfo[i].basename);
      sprintf(line, "%s/%s %s.cex -%sTD > %s.s ",
              chome, vpo, fileinfo[i].basename, compileflags,
              fileinfo[i].basename, fileinfo[i].basename);
	fprintf(stderr,"Executing: %s\n", line);
      verbose_system(line);
      }

   /* link and execute the program, print transformation info and then */
   /* check the program's output if the execution was correct. */
   fprintf(stderr, "linking program\n");
   verbose_system(linkcmd);
   verbose_system(rmcmd);
   fprintf(stderr, "executing program\n");
   verbose_system(executecmd);

   /* read in the number of transformations for each function */
   if (f = fopen("numtrans.txt", "r")) {
      numfuncs = 0;
      numtrans = 1;
      while (fscanf(f, "%s : %s : %d\n", funcinfo[numfuncs].basename,
                    funcinfo[numfuncs].funcname,
                    &funcinfo[numfuncs].numtrans) != EOF) {
         for (i = 0; i < numfiles; i++)
            if (strcmp(fileinfo[i].basename,
                       funcinfo[numfuncs].basename) == 0) {
               if (fileinfo[i].firstfuncnum > numfuncs)
                  fileinfo[i].firstfuncnum = numfuncs;
               if (fileinfo[i].lastfuncnum < numfuncs)
                  fileinfo[i].lastfuncnum = numfuncs;
               break;
               }
         funcinfo[numfuncs].starttrans = numtrans;
         numtrans += funcinfo[numfuncs].numtrans;
         funcinfo[numfuncs].endtrans = numtrans - 1;
         numfuncs++;
         }
      fclose(f);
      }
   else
      error("could not open numtrans.txt file");

   fprintf(stderr, "\n%d total improving transformations for program:\n",
           numtrans-1);

   for (i = 0; i < numfuncs; i++)
      fprintf(stderr, "   %s: %s: %d-%d\n", funcinfo[i].basename,
              funcinfo[i].funcname, funcinfo[i].starttrans,
              funcinfo[i].endtrans);
   fprintf(stderr, "\n");

   /* check output from last execution */
   if (check_output() == 0)
      error("nothing wrong with completely optimized program");
   else
      fprintf(stderr, "the output was NOT correct\n");

   /* set maximum transformations for each function to 0 */
   f = fopen("maxtrans.txt", "w");
   for (i = 0; i < numfuncs; i++) {
      fprintf(f, "%s : %s : 0\n", funcinfo[i].basename,
              funcinfo[i].funcname);
      funcinfo[i].compiledwithopts = FALSE;
      }
   fclose(f);
   
   /* compile files to check if program works with no optimizations */
   fprintf(stderr, "\n\n==================== Pass %lu\n", run_count++);
   fprintf(stderr,
    "compiling program to check if works with no improving transformations\n");

   for (i = 0; i < numfiles; i++) {
      fprintf(stderr, "compiling %s.cex\n", fileinfo[i].basename);
      /* $$$ modified the following line for vpo.out's path */ 
      sprintf(line, "%s/%s %s.cex -ATM > %s.s; ",
              chome, vpo, fileinfo[i].basename,
              fileinfo[i].basename, fileinfo[i].basename);
      verbose_system(line);
      }

   /* link and execute to see if output is correct */
   fprintf(stderr, "linking program\n");
   verbose_system(linkcmd);
   verbose_system(rmcmd);
   fprintf(stderr, "executing program\n");
   verbose_system(executecmd);
   if (check_output() != 0)
      error("output wrong even with no optimizations");
   else
      fprintf(stderr, "the output was correct\n");

   /* do binary search to isolate transformation causing an optimization
      error */
   lastmax = funcinfo[numfuncs-1].endtrans;
   fprintf(stderr,
      "starting binary search to isolate error within %d transformations\n",
           lastmax);
   lastmin = 1;
   within = -1;
   badfunc = -1;
   lastresultbad = TRUE;
   while (lastmax - lastmin > 0) {
      cutnum = (lastmin + lastmax)/2;
      if (badfunc == -1) {
         for (i = 0; i < numfuncs; i++)
            if (funcinfo[i].endtrans >= lastmax) {
               if (funcinfo[i].starttrans <= lastmin) {
                  fprintf(stderr,
                     "incorrect transformation isolated to function %s:%s()\n",
                          funcinfo[i].basename, funcinfo[i].funcname);
                  badfunc = i;
                  }
               break;
               }
         }

      /* set maximum number of transformations for all functions to 0
         except the bad function which is set to the cut number */
      f = fopen("maxtrans.txt", "w");
      lastwithin = within;
      for (i = 0; i < numfuncs; i++)
         if (funcinfo[i].starttrans > cutnum) /* $$$ || funcinfo[i].endtrans < lastmin */
            fprintf(f, "%s : %s : 0\n", funcinfo[i].basename,
                    funcinfo[i].funcname);
         else if (funcinfo[i].endtrans < cutnum)
            fprintf(f, "%s : %s : %d\n", funcinfo[i].basename,
                    funcinfo[i].funcname, funcinfo[i].numtrans);
         else {
            fprintf(f, "%s : %s : %d\n", funcinfo[i].basename,
                    funcinfo[i].funcname, cutnum - funcinfo[i].starttrans + 1);
            within = i;
            }
      fclose(f);

      if (badfunc == -1) {
         for (i = 0; i < numfuncs; i++)
            if (funcinfo[i].endtrans >= lastmin) {
               fprintf(stderr, "error within %s:%s()", funcinfo[i].basename, funcinfo[i].funcname);
               break;
               }
         for (i += 1; i < numfuncs; i++)
            if (funcinfo[i].starttrans > lastmax)
               break;
         fprintf(stderr, " to %s:%s() (transformation %d to %d)\n",
                 funcinfo[i-1].basename, funcinfo[i-1].funcname, lastmin, lastmax);
         }
      else
         fprintf(stderr, "error within %s (transformation %d to %d)\n",
                 funcinfo[badfunc].funcname, lastmin, lastmax);

      /* compile files to check if program works (only recompile
         a file if it is necessary) */
      fprintf(stderr, "\n\n==================== Pass %lu\n", run_count++);
      fprintf(stderr, "compiling program: applying transformations %d to %d\n"
,
              lastmin, cutnum);
      for (i = 0; i < numfiles; i++) {
         needstocompile = FALSE;
         for (j = fileinfo[i].firstfuncnum;
              j <= fileinfo[i].lastfuncnum; j++) {
            if (j == within || j == lastwithin || RECOMPILE ||
                (funcinfo[j].compiledwithopts &&
                 funcinfo[j].endtrans > cutnum) ||
                (!funcinfo[j].compiledwithopts &&
                 funcinfo[j].endtrans >= lastmin &&
                 funcinfo[j].starttrans < cutnum))
               needstocompile = TRUE;
            funcinfo[j].compiledwithopts =
               !(funcinfo[j].starttrans > cutnum ||
                 funcinfo[j].endtrans < lastmin);
            }
         if (needstocompile) {
            /* $$$ modified the following line for vpo.out's path */ 
            fprintf(stderr, "compiling %s.cex\n", fileinfo[i].basename);
            sprintf(line,
                    "%s/%s %s.cex -%sTM > %s.s; ",
                    chome, vpo, fileinfo[i].basename, compileflags, 
                    fileinfo[i].basename, fileinfo[i].basename);
            verbose_system(line);
            }
         }

      /* link and execute to see if output is correct */
      fprintf(stderr, "linking program\n");
      verbose_system(linkcmd);
      verbose_system(rmcmd);
      fprintf(stderr, "executing program\n");
      verbose_system(executecmd);
      if (check_output() == 0) {
         fprintf(stderr, "execution was correct\n");
         lastresultbad = FALSE;
         lastmin = cutnum+1;
         for (i = 0; i < numfiles; i++) {
            sprintf(line, "cp %s.s %s.corr.s", fileinfo[0].basename,
                    fileinfo[0].basename);
            verbose_system(line);
            }
         }
      else {
         fprintf(stderr, "execution was NOT correct\n");
         lastresultbad = TRUE;
         lastmax = cutnum;
         for (i = 0; i < numfiles; i++) {
            sprintf(line, "cp %s.s %s.incorr.s", fileinfo[0].basename,
                    fileinfo[0].basename);
            verbose_system(line);
            }
         }
      }


	{
		/* Prepare the rtl files of the correct and incorrect versions */
		int badfile=0;
		
		if (lastresultbad) 
			wrong_trans=cutnum;
		else 
			wrong_trans=cutnum+1;

		right_trans=wrong_trans-1;


		for (i = 0; i < numfiles; i++) {
			if (badfunc>=fileinfo[i].firstfuncnum && badfunc<=fileinfo[i].lastfuncnum) {
				badfile=i;
				break;
			}
		}
		prepare_maxtrans_file(wrong_trans);
		fprintf(stderr, "compiling %s.cex\n", fileinfo[badfile].basename);
		sprintf(line,
				"%s/%s %s.cex -%sTM -dno_rtl_line_nums > %s.incorr.rtl; ",
				chome, vpo, fileinfo[badfile].basename, compileflags_rtl, 
				fileinfo[badfile].basename, fileinfo[badfile].basename);
		verbose_system(line);

		prepare_maxtrans_file(right_trans);
		fprintf(stderr, "compiling %s.cex\n", fileinfo[badfile].basename);
		sprintf(line,
				"%s/%s %s.cex -%sTM -dno_rtl_line_nums > %s.corr.rtl; ",
				chome, vpo, fileinfo[badfile].basename, compileflags_rtl, 
				fileinfo[badfile].basename, fileinfo[badfile].basename);
		verbose_system(line);
	}

   /* determine the bad transformation */
   if (lastresultbad)
      fprintf(stderr,
      "\nIncorrect transformation isolated to optimization %d in function %s:%s()\n",
              cutnum - funcinfo[badfunc].starttrans,
              funcinfo[badfunc].basename,
              funcinfo[badfunc].funcname);
   else
      fprintf(stderr,
      "\nincorrect transformation isolated to optimization %d in function %s:%s()\n\n",
              cutnum - funcinfo[badfunc].starttrans + 1,
              funcinfo[badfunc].basename,
              funcinfo[badfunc].funcname);
   verbose_system("date '+STOP TIME: %H:%M:%S'");
}


void prepare_maxtrans_file(int upto)
{
	FILE *f;
	int i;
	f = fopen("maxtrans.txt", "w");
	if (!f) {
		fprintf(stderr,"Error writing into maxtrans.txt\n");
		exit(1);
	}
	for (i = 0; i < numfuncs; i++) {
		if (funcinfo[i].starttrans > upto)
			fprintf(f, "%s : %s : 0\n", funcinfo[i].basename,
					funcinfo[i].funcname);
		else if (funcinfo[i].endtrans < upto)
			fprintf(f, "%s : %s : %d\n", funcinfo[i].basename,
					funcinfo[i].funcname, funcinfo[i].numtrans);
		else {
			fprintf(f, "%s : %s : %d\n", funcinfo[i].basename,
					funcinfo[i].funcname, upto - funcinfo[i].starttrans+1 );
		}
	}
	fclose(f);
}
