/************************************************************************/
/*                                                                      */
/* Copyright (c) 1990, 1991, 1992 -                                     */
/*     The Rector and Visitors of the University of Virginia            */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from the University of Virginia.                                */
/* This code is distributed WITHOUT ANY WARRANTY.  No claims are        */
/*      made as to whether it serves any particular purpose or even     */
/*      works at all.                                                   */
/*                                                                      */
/************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/tools/yyfast/yyfast.c,v $
 $Date: 2003/03/12 15:35:09 $
 $Revision: 1.4 $
 $Name:  $

 Log of changes:

 $Log: yyfast.c,v $
 Revision 1.4  2003/03/12 15:35:09  jdh8d
 *** empty log message ***

 Revision 1.3  2003/02/06 03:21:38  jg9h
 Added ERROR token (for VISTA)

 Revision 1.2  2002/06/04 16:06:31  jg9h
 The YYSTYPE bug, where yyfast was checking for 3 spaces

 Revision 1.1.1.1  2000/08/22 21:10:35  zephyr
 Created directory structure

 Revision 1.2  1998/04/08 18:40:00  jwd
 First ANSI C version.


*************************************************************************/

/*
 * yyfast - this utility reads a yacc parser and builds a faster version
 *          which performs the same function.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "yyfast.h"
#include "yyfast2.h"

array yyexca;		/* Contents of the exception array           */
array yyact;		/* Contents of the action lookup table array */
array yypact;		/* Contents of the action pointer array      */
array yypgo;		/* Contents of the goto pointer array        */
array yyr1;		/* Contents of the goto lookup array         */
array yyr2;		/* Contents of the production size array     */
array yychk;		/* Contents of the action validate array     */
array yydef;		/* Contents of the default action array      */
action_pointer semantic_actions;	/* Pointer to semantic action data base      */
int number_of_productions;	/* Number of productions in the grammar      */
int number_of_actions;	/* Number of semantic actions actions parsed */
int number_of_states;	/* Number of states in the parser            */
int number_of_gotos;	/* Number of goto items in the parser        */
int table_size;		/* Size of the parser lookup table           */
int highest_token;	/* Maximum token value defined               */
int verbose;		/* Display parser information switch         */
string_pointer yyparse_name;	/* Identifier to use for the parser function */
string_pointer myname;	/* Name that we where invoked with           */
state_info info[MAXSTATES];	/* Information concerning each parse state   */
token_info tokens[MAXTOKEN];	/* Token information database                */
reduce_info reduce[MAXGOTO];	/* Information concerning each goto state    */

/*
 * main - parse the command line arguments and drive the processing
 */
int main(int argc, char **argv) {

   register string_pointer ptr;
   int infile;

   /*
      Remember the name that we are invoked with
   */
   myname = *argv++;

   /*
      Initialize default options and tokens data base
   */
   verbose = FALSE;
   yyparse_name = "yyparse";
   initialize_tokens();

   /*
      Parse the command line
   */
   infile = 0;
   while (--argc > 0) {
      ptr = *argv++;

      /*
         Look for options
      */
      if (*ptr == '-' && *(ptr + 1)) {
	 while (*++ptr) {
	    switch (*ptr) {

	       case 'f':
		  yyparse_name = ptr + 1;
		  while (*(ptr + 1))
		     ptr++;
		  break;

	       case 'v':
		  verbose = TRUE;
		  break;

	       default:
		  (void) fprintf(stderr, "%s: invalid option '%c'\n", myname, *ptr);
		  exit(1);
	    }
	 }
      }
      else {
	 if (infile == 3) {
	    (void) fprintf(stderr, "%s: too many file specifications\n", myname);
	    exit(1);
	 }
	 if (*ptr != '-') {
	    if (!infile) {
	       if (!freopen(ptr, "r", stdin)) {
		  (void) fprintf(stderr, "%s: cannot open '%s' for input\n",
				 myname, ptr);
		  exit(1);
	       }
	    }
	    else if (infile == 1) {
	       if (!freopen(ptr, "w", stdout)) {
		  (void) fprintf(stderr, "%s: cannot open '%s' for output\n",
				 myname, ptr);
		  exit(1);
	       }
	    }
	    else if (!freopen(ptr, "w", stderr)) {
	       (void) fprintf(stderr, "%s: cannot open '%s' for output\n",
			      myname, ptr);
	       exit(1);
	    }
	 }
	 infile++;
      }
   }

   /*
      Perform the main processing task of this tool
   */
   read_yacc_parser();
   build_state_info();
   write_yacc_parser();

   /*
      Finish indicating that no errors where found
   */
   exit(0);
}

/*
 * read_yacc_parser - scan the input parser to obtain the table information
 */
void read_yacc_parser(void) {

   string line;
  
   /*
      Read all of the lines from the input file and transfer them to the output
      file until we encounter the line that marks the start of the token
      definitions.
   */
   do {
      if (!read_input_line(line)) {
	 (void) fprintf(stderr, "%s: token definitions not found\n");
	 exit(1);
      }
      write_output_line(line);

      /* loop until we find the end of the YYSTYPE declaration */
   } /* { */ while (strstr(line, "YYSTYPE;\n")==NULL);


   /* Previously, we looped on this condition:
    *   while (strstr(line, "   } YYSTYPE;\n"));
    * However, this depends on the user putting exactly 3 spaces before the } in the md.y file.
    * This turned out to be true because everyone copied exactly one md.y file and used it.
    * Joshua pulled his hair out for 3 hours because he changed the indentation.  I've glued
    * a little hair back on, but he's rather bald still - jdh.
    */

   /*
      Examine each token definition to determine the value of each special
      token value that can be returned by the scanner.
   */
   highest_token = 255;
   for (;;) {
      if (!read_input_line(line)) {
	 (void) fprintf(stderr, "%s: end of token definitions not found\n");
	 exit(1);
      }
      write_output_line(line);
      if (!strcmp(line, "#ifndef YYMAXDEPTH\n"))
	 break;
      token_value(line);
   }

   /*
      Continue moving all of the lines from the input file to the output file
      until we encounter the declaration for the "yyexca" array.
   */
   for (;;) {
      if (!read_input_line(line)) {
	 (void) fprintf(stderr, "%s: yyexca array not found\n");
	 exit(1);
      }
      if (!strcmp(line, "short yyexca[] ={\n"))	/* } */
	 break;
      write_output_line(line);
   }

   /*
      Read and store the contents of the "yyexca" array
   */
   (void) read_array(&yyexca);

   /*
      Read the definition for the YYNPROD constant
   */
   number_of_productions = read_define("YYNPROD");

   /*
      Read the definition for the YYLAST constant
   */
   (void) read_define("YYLAST");

   /*
      Read and store the contents of the "yyact" array
   */
   for (;;) {
      if (!read_input_line(line)) {
	 (void) fprintf(stderr, "%s: yyact array not found\n");
	 exit(1);
      }
      if (!strcmp(line, "short yyact[]={\n"))	/* } */
	 break;
   }
   table_size = read_array(&yyact);

   /*
      Read and store the contents of the "yypact" array
   */
   for (;;) {
      if (!read_input_line(line)) {
	 (void) fprintf(stderr, "%s: yypact array not found\n");
	 exit(1);
      }
      if (!strcmp(line, "short yypact[]={\n"))	/* } */
	 break;
   }
   number_of_states = read_array(&yypact);
   if (number_of_states > MAXSTATES) {
      (void) fprintf(stderr, "%s: %d states is above the limit of %d states\n",
		     myname, number_of_states, MAXSTATES);
      exit(1);
   }

   /*
      Read and store the contents of the "yypgo" array
   */
   for (;;) {
      if (!read_input_line(line)) {
	 (void) fprintf(stderr, "%s: yypgo array not found\n");
	 exit(1);
      }
      if (!strcmp(line, "short yypgo[]={\n"))	/* } */
	 break;
   }
   number_of_gotos = read_array(&yypgo);
   if (number_of_gotos > MAXGOTO) {
      (void) fprintf(stderr, "%s: %d gotos is above the limit of %d gotos\n",
		     myname, number_of_gotos, MAXGOTO);
      exit(1);
   }

   /*
      Read and store the contents of the "yyr1" array
   */
   for (;;) {
      if (!read_input_line(line)) {
	 (void) fprintf(stderr, "%s: yyr1 array not found\n");
	 exit(1);
      }
      if (!strcmp(line, "short yyr1[]={\n"))	/* } */
	 break;
   }
   (void) read_array(&yyr1);

   /*
      Read and store the contents of the "yyr2" array
   */
   for (;;) {
      if (!read_input_line(line)) {
	 (void) fprintf(stderr, "%s: yyr2 array not found\n");
	 exit(1);
      }
      if (!strcmp(line, "short yyr2[]={\n"))	/* } */
	 break;
   }
   (void) read_array(&yyr2);

   /*
      Read and store the contents of the "yychk" array
   */
   for (;;) {
      if (!read_input_line(line)) {
	 (void) fprintf(stderr, "%s: yychk array not found\n");
	 exit(1);
      }
      if (!strcmp(line, "short yychk[]={\n"))	/* } */
	 break;
   }
   (void) read_array(&yychk);

   /*
      Read and store the contents of the "yydef" array
   */
   for (;;) {
      if (!read_input_line(line)) {
	 (void) fprintf(stderr, "%s: yydef array not found\n");
	 exit(1);
      }
      if (!strcmp(line, "short yydef[]={\n"))	/* } */
	 break;
   }
   (void) read_array(&yydef);

   /*
      Read each semantic action in the file
   */
   number_of_actions = read_semantic_actions();

   /*
      Show the user the size of the parser if he wants to know
   */
   if (verbose) {
      (void) fprintf(stderr, "%d semantic actions\n", number_of_actions);
      (void) fprintf(stderr, "%d productions\n", number_of_productions);
      (void) fprintf(stderr, "%d states\n", number_of_states);
      (void) fprintf(stderr, "%d gotos\n", number_of_gotos);
      (void) fprintf(stderr, "highest token number is %d\n", highest_token);
   }
}

/*
 * build_state_info - determine specific information about each parser state
 */
void build_state_info(void) {

   register int state;
   register short token, lookup, token_state, state_lookup, default_lookup;

   /*
      Initialize the information item for each state
   */
   for (state = 0; state < number_of_states; state++) {
      info[state].needs_shift = FALSE;
      info[state].needs_state = FALSE;
      info[state].uses_exception = FALSE;
      info[state].complex_token = FALSE;
      info[state].complex_check = FALSE;
      info[state].visited = FALSE;
      info[state].might_shift = FALSE;
      info[state].might_reduce = FALSE;
      info[state].save_yylval = FALSE;
      info[state].token_checks = 0;
   }

   for (state = 0; state < number_of_gotos; state++) {
      reduce[state].uses = 0;
      reduce[state].state_checks = 0;
      reduce[state].label_emitted = FALSE;
      reduce[state].reduce_label_num=state;
   }

   /*
      Determine which states are transfered to from a shift transition
   */
   for (state = 0; state < number_of_states; state++) {

      /*
         The first step when entering a new state is to look up the value in
         the "yypact" array for this state.
      */
      lookup = get_array(yypact, state);

      /*
         If the value is less than -1000, then there are no shift states in
         this item.
      */
      if (lookup > -1000) {
	 info[state].might_shift = TRUE;
	 for (token = 0; token <= highest_token; token++) {
	    if (lookup + token >= 0 && lookup + token < table_size) {
	       token_state = get_array(yyact, (int) (lookup + token));
	       if (get_array(yychk, (int) token_state) == token) {
		  info[state].token_checks++;
		  if (complex_token(token))
		     info[state].complex_token = TRUE;
		  info[token_state].needs_shift = TRUE;
		  if (uses_yylval(token))
		     info[token_state].save_yylval = TRUE;
	       }
	    }
	 }
      }

      /*
         In the case that none of the shifts are taken, then we have to decide
         what to do in the default case.  The possibilities are that we have
         reached an accepting state, that we can reduce a production, or that
         we reach an error state.
      */
      lookup = get_array(yydef, state);

      /*
         A -2 value in the default table for this item means that we have to
         consult the exception table to determine what action we must take
         next.
      */
      if (lookup == -2) {
	 info[state].uses_exception = TRUE;
	 for (default_lookup = 0;
	      get_array(yyexca, (int) default_lookup) != (short) -1 ||
	      get_array(yyexca, (int) default_lookup + 1) != (short) state;
	      default_lookup += 2);
	 do {
	    default_lookup += 2;
	    token_state = get_array(yyexca, (int) default_lookup);
	    state_lookup = get_array(yyexca, (int) default_lookup + 1);
	    if (state_lookup < 0) {
	       if (token_state >= 0) {
		  if (complex_token(token_state))
		     info[state].complex_token = TRUE;
	       }
	    }
	    else if (token_state >= 0) {
	       if (complex_token(token_state))
		  info[state].complex_token = TRUE;
	       build_reduce(state_lookup, state);
	    }
	    else
	       build_reduce(state_lookup, state);
	 } while (token_state >= 0);
      }

      /*
         Otherwise, a non-zero value indicates that the state must now reduce a
         production.
      */
      else if (lookup)
	 build_reduce(lookup, state);
   }

   /*
      Now we perform another pass through the states to determine which shift
      actions will only require a simple token check to be performed.
   */
   for (state = 0; state < number_of_states; state++)
      (void) complex_check(state);
}

/*
 * build_reduce - determine which states might be reached when reducing
 */
void build_reduce(short int production, int state) {

   register short token_state, goto_lookup, state_lookup, new_state;

   /*
      Indicate that this state might reduce a production
   */
   info[state].might_reduce = TRUE;

   token_state = get_array(yyr1, (int) production);
   goto_lookup = get_array(yypgo, (int) token_state);

   /*
      Determine if this goto table is shared with any other reductions in the
      parser.
   */

   /* memorize the first "state" that this reduction was used in */
   if (!reduce[token_state].uses++)
   {
	/* 
	 * however, there's a bug here.. more than 1 reduction may be used first
	 * in a given state, ie:

state 136
        src1 :  dst_    (99)
        src :  dst_    (102)

        ;  reduce 102
        ?  reduce 102
        u  reduce 102
        .  reduce 99


		Here, reduction 102 and reduction 99 are first seen in state 136.
		This leads to reduce136: being defined twice.

		Actually, it's worse than that.  reduce136 is defined 4 times
		(3 times for reduction 102, and once for reduction 99)

		Thus, I'm fixing this.  Instead of reduce136, we'll be
		changing the label to reduce102/reduce99.  I'm also adding
		a bit to indicate whether the label has been emitted or not.
		

	 */
      	reduce[token_state].state = state;
   }

   /*
      Check for those goto transitions that have to be taken depending on the
      current state on the top of the state stack.
   */
   for (state_lookup = 0; state_lookup < number_of_states; state_lookup++) {
      if (goto_lookup + state_lookup + 1 < table_size) {
	 new_state = get_array(yyact, (int) (goto_lookup + state_lookup + 1));
	 if (get_array(yychk, (int) new_state) == -token_state) {
	    info[new_state].needs_state = TRUE;
	    if (reduce[token_state].state == state)
	       reduce[token_state].state_checks++;
	 }
      }
   }

   /*
      Goto the default state
   */
   new_state = get_array(yyact, (int) goto_lookup);
   info[new_state].needs_state = TRUE;
}

/*
 * complex_check - determine if the state can perform a simple token lookup
 */
int complex_check(int state) {

   register short lookup, default_lookup, token_state, state_lookup;

   /*
      If this state has already been checked, then there is no need to do so
      again.
   */
   if (info[state].complex_check)
      return (info[state].complex_token);

   /*
      If this state has no reduce transitions, or, if it has complex token
      checks in any of its shift transitions, then we do not need to move any
      further to determine if it can perform a simple token lookup.
   */
   if (info[state].complex_token || !info[state].might_reduce) {
      info[state].complex_check = TRUE;
      return (info[state].complex_token);
   }

   /*
      If we have been around the block and not found anything yet, then there
      is no need to keep looking further.
   */
   if (info[state].visited)
      return (FALSE);

   /*
      Mark this state as visited so that we do not get into infinite loops
   */
   info[state].visited = TRUE;

   /*
      Now we need to to look at all of the states that this state can transfer
      control to after a production is reduced.  If all such states will check
      for only simple tokens, then we can go ahead and perform only a simple
      token lookup in this state.
   */
   if (info[state].uses_exception) {
      lookup = -2;
      for (default_lookup = 0;
	   get_array(yyexca, (int) default_lookup) != (short) -1 ||
	   get_array(yyexca, (int) default_lookup + 1) != (short) state;
	   default_lookup += 2);
      do {
	 default_lookup += 2;
	 token_state = get_array(yyexca, (int) default_lookup);
	 state_lookup = get_array(yyexca, (int) default_lookup + 1);
	 if (state_lookup > 0) {
	    if (token_state >= 0) {
	       if (reduce_complex_check(state_lookup)) {
		  info[state].complex_token = TRUE;
		  info[state].complex_check = TRUE;
		  return (TRUE);
	       }
	    }
	    else
	       lookup = state_lookup;
	 }
      } while (token_state >= 0);
   }
   else
      lookup = get_array(yydef, state);

   /*
      If any of the states transfered to after this production is reduced check
      for a complex token, then a shift from this state must scan for a complex
      token.
   */
   if (lookup > 0 && reduce_complex_check(lookup))
      info[state].complex_token = TRUE;

   info[state].complex_check = TRUE;
   return (info[state].complex_token);
}

/*
 * reduce_complex_check - determine which states might be reached when reducing
 */
int reduce_complex_check(short int production) {

   register short token_state, goto_lookup, state_lookup, new_state;

   token_state = get_array(yyr1, (int) production);
   goto_lookup = get_array(yypgo, (int) token_state);

   /*
      Check for those goto transitions that have to be taken depending on the
      current state on the top of the state stack.
   */
   for (state_lookup = 0; state_lookup < number_of_states; state_lookup++) {
      if (goto_lookup + state_lookup + 1 < table_size) {
	 new_state = get_array(yyact, (int) (goto_lookup + state_lookup + 1));
	 if (get_array(yychk, (int) new_state) == -token_state) {
	    if (complex_check(new_state))
	       return (TRUE);
	 }
      }
   }

   /*
      Goto the default state
   */
   new_state = get_array(yyact, (int) goto_lookup);
   return (complex_check(new_state));
}

/*
 * write_yacc_parser - emit the optimized parser
 */
void write_yacc_parser(void) {

   register int state;
   register short token, lookup, token_state, state_lookup, default_lookup;

   /*
      Emit the starting declaration for the parse function
   */
   (void) printf("\n");
   (void) printf("int %s()\n", yyparse_name);
   (void) printf("{\n");/* } */
   (void) printf("    register short *state_stack;\n");
   (void) printf("    register YYSTYPE *value_stack;\n");
   (void) printf("    register int token;\n");
   (void) printf("    YYSTYPE value_stack_ary[YYMAXDEPTH];\n");
   (void) printf("    short state_stack_ary[YYMAXDEPTH];\n");
   (void) printf("    extern char *linptr;\n");
   (void) printf("    int yylex();\n");
   (void) printf("\n");
   (void) printf("    state_stack = state_stack_ary;\n");
   (void) printf("    value_stack = value_stack_ary;\n");
   if (!info[0].needs_shift) {
      if (info[0].complex_token)
	 (void) printf("    token = yylex();\n");
      else
	 (void) printf("    token = (int) *linptr++;\n");
   }

   /*
      Emit the code that is required to perform the parsing actions for each
      individual state.
   */
   for (state = 0; state < number_of_states; state++) {

      if (state)
	 (void) printf("\n");

      /*
         If any of the other states transfer to this state through a shift
         action, then we need to consume the current token value and go on to
         the next token value.  We also need a label so that we can get to this
         location.
      */
      if (info[state].needs_shift) {
	 (void) printf("shift%d:\n", state);

	 /*
	    If all of the token values that could have caused us to enter this
	    state where of a type that do not return additional information in
	    the "yylval" global variable, then there is no need to save any
	    such information on the value stack.  We do, however, still need to
	    maintain a proper stack position for this token.
	 */
	 if (info[state].save_yylval &&
	     (info[state].might_shift || info[state].complex_token))
	    (void) printf("    *++value_stack = yylval;\n");
	 else
	    (void) printf("    value_stack++;\n");

	 /*
	    If we have determined that all of the possible tokens that are
	    valid for the next input token are single characters, then we can
	    get the next character directly from the input string rather than
	    going through the additional overhead of calling the scanner.
	 */
	 if (info[state].complex_token)
	    (void) printf("    token = yylex();\n");
	 else
	    (void) printf("    token = (int) *linptr++;\n");
      }

      /*
         If some state will transfer to this point after reducing a production,
         then we will need to emit a label here.
      */
      if (info[state].needs_state)
	 (void) printf("state%d:\n", state);

      /*
         If there are any shift actions associated with this state, then we
         emit the code required to perform the appropriate shift based on the
         value of the current token.  Note that if there are no possible shifts
         out of this state, that we do not bother to push the current state
         value onto the parse stack.  This is because we know that it will not
         be looked at.
      */
      if (info[state].might_shift) {

	 /*
	    Note that for state zero, if we do not shift or reduce into this
	    state, then there is no need to increment the state stack before
	    placing our state value on the stack.
	 */
	 if (!state && !info[0].needs_shift && !info[0].needs_state)
	    (void) printf("    *state_stack = 0;\n");
	 else
	    (void) printf("    *++state_stack = %d;\n", state);

	 /*
	    Now we will generate the code required to determine which state we
	    need to shift to depending on the input token value.  If we must
	    check many items, then we will use a switch statement.
	 */
	 if (info[state].token_checks > 4)
	    (void) printf("    switch (token) {\n");	/* } */
	 lookup = get_array(yypact, state);
	 for (token = 0; token <= highest_token; token++) {
	    if (((lookup + token) >= 0) && (lookup + token) < table_size) {
	       token_state = get_array(yyact, (int) (lookup + token));
	       if (get_array(yychk, (int) token_state) == token)
		  if (info[state].token_checks > 4)
		     (void) printf("      case %s: goto shift%d;\n",
				   token_identifier(token), token_state);
		  else
		     (void) printf("    if (token == %s) goto shift%d;\n",
				   token_identifier(token), token_state);
	    }
	 }
	 if (info[state].token_checks > 4)	/* { */
	    (void) printf("      }\n");
      }

      /*
         If we have to look at the exception table to determine what default
         action to take, then do so at this point.
      */
      if (info[state].uses_exception) {
	 lookup = -2;
	 for (default_lookup = 0;
	      get_array(yyexca, (int) default_lookup) != (short) -1 ||
	      get_array(yyexca, (int) default_lookup + 1) != (short) state;
	      default_lookup += 2);
	 do {
	    default_lookup += 2;
	    token_state = get_array(yyexca, (int) default_lookup);
	    state_lookup = get_array(yyexca, (int) default_lookup + 1);
	    if (state_lookup < 0) {
	       if (token_state >= 0)
		  (void) printf("    if (token == %s) return(0);\n",
				token_identifier(token_state));
	       else
		  (void) printf("    return(0);\n");
	    }
	    else if (token_state >= 0) {
	       (void) printf("    if (token == %s) {\n",
			     token_identifier(token_state));
	       reduce_production("      ", state_lookup, state);
	       (void) printf("      }\n");
	    }
	    else
	       lookup = state_lookup;
	 } while (token_state >= 0);
      }

      /*
         Otherwise we consult the default table directly to determine if we
         reduce a production or reach an error state.
      */
      else
	 lookup = get_array(yydef, state);

      /*
         If we have a valid reduction item, then generate the code to reduce
         it.
      */
      if (lookup >= 0)
	 reduce_production("    ", lookup, state);
   }

   /* { */
   (void) printf("\n}\n");
}

/*
 * reduce_action - emit the code to reduce a production
 */
void reduce_production(string_pointer tab, short int production, int state) {

   register short token_state, goto_lookup, state_lookup, new_state;

   /*
      If there is not production to reduce, then we have reached an error
      state.
   */
   if (!production) {
      (void) printf("%sreturn(1);\n", tab);
      return;
   }

   (void) printf("%s/* Reduce production %d */\n", tab, production);

   /*
      Determine how many items we have to take off of the parse stack. If this
      state has no shift actions, then we did not push our current state on the
      stack, so do not attempt to take it off.
   */
   token_state = get_array(yyr2, (int) production);
   if (info[state].might_shift) {
      if (token_state == 1)
	 (void) printf("%sstate_stack--;\n", tab);
      else
	 (void) printf("%sstate_stack -= %d;\n", tab, token_state);
   }
   else if (token_state > 1) {
      if (token_state == 2)
	 (void) printf("%sstate_stack--;\n", tab);
      else
	 (void) printf("%sstate_stack -= %d;\n", tab, token_state - 1);
   }

   /*
      Emit the semantic action required with the reduction of this production
   */
   process_semantic_action(tab, production, token_state, state);

   token_state = get_array(yyr1, (int) production);
   goto_lookup = get_array(yypgo, (int) token_state);

   /*
      If this goto setup is duplicated in other places, then if this is the
      first instance, place a label at this point so that the other instances
      can transfer control to it.  If an appropriate sequence has already been
      generated, transfer control to it directly.
   */

   if (reduce[token_state].state_checks > 1 && reduce[token_state].uses > 1) {
      if (reduce[token_state].state == state && !reduce[token_state].label_emitted)
	{
	 (void) printf("reduce%d:\n", reduce[token_state].reduce_label_num);
	 reduce[token_state].label_emitted=TRUE;
	}
      else {
	 (void) printf("%sgoto reduce%d;\n", tab, reduce[token_state].reduce_label_num);
	 return;
      }
   }

   /*
      If we have determine that there are many state checks in this item, then
      we will use a switch statement.
   */
   if (reduce[token_state].state_checks > 3)
      (void) printf("%sswitch (*state_stack) {\n", tab);	/* } */

   /*
      Check for those goto transitions that have to be taken depending on the
      current state on the top of the state stack.
   */
   for (state_lookup = 0; state_lookup < number_of_states; state_lookup++) {
      if (goto_lookup + state_lookup + 1 < table_size) {
	 new_state = get_array(yyact, (int) (goto_lookup + state_lookup + 1));
	 if (get_array(yychk, (int) new_state) == -token_state) {
	    if (reduce[token_state].state_checks > 3)
	       (void) printf("%s  case %d: goto state%d;\n",
			     tab, state_lookup, new_state);
	    else
	       (void) printf("%sif (*state_stack == %d) goto state%d;\n",
			     tab, state_lookup, new_state);
	 }
      }
   }

   /*
      Goto the default state
   */
   new_state = get_array(yyact, (int) goto_lookup);
   if (reduce[token_state].state_checks > 3) {
      (void) printf("%s  default: goto state%d;\n", tab, new_state);	/* { */
      (void) printf("%s  }\n", tab);
   }
   else
      (void) printf("%sgoto state%d;\n", tab, new_state);
}

/*
 * process_semantic_action - emit the most efficient semantic action
 */
void process_semantic_action(string_pointer tab, short int production, short int items, int state) {

   register string_pointer ptr, ptr2, ptr3;
   string line, num;
   int index;
   action_pointer action_ptr;

   /*
      Adjust the value stack if we need to
   */
   if (items > 1) {
      if (items == 2)
	 (void) printf("%svalue_stack--;\n", tab);
      else
	 (void) printf("%svalue_stack -= %d;\n", tab, items - 1);
   }

   /*
      Determine if there is any semantic action associated with this
      production.
   */
   for (action_ptr = semantic_actions;
	action_ptr && action_ptr->number != production;
	action_ptr = action_ptr->next);

   /*
      If there is a semantic action, then we must parse it and modify it to use
      the appropriate value stack items.
   */
   if (action_ptr) {

      ptr2 = line;

      /*
         Take care of the left side of the assignment to the value stack, if
         there is any.
      */
      ptr = action_ptr->body;
      if (*ptr == 'y' && *(ptr + 1) == 'y' && *(ptr + 2) == 'v' &&
	  *(ptr + 3) == 'a' && *(ptr + 4) == 'l' && *(ptr + 5) == '.') {
	 *ptr2++ = 'v';
	 *ptr2++ = 'a';
	 *ptr2++ = 'l';
	 *ptr2++ = 'u';
	 *ptr2++ = 'e';
	 *ptr2++ = '_';
	 *ptr2++ = 's';
	 *ptr2++ = 't';
	 *ptr2++ = 'a';
	 *ptr2++ = 'c';
	 *ptr2++ = 'k';
	 *ptr2++ = '-';
	 *ptr2++ = '>';
	 ptr += 6;
	 while ((*ptr >= 'A' && *ptr <= 'Z') || *ptr == '_' ||
		(*ptr >= 'a' && *ptr <= 'z') || (*ptr >= '0' && *ptr <= '9'))
	    *ptr2++ = *ptr++;
	 *ptr2++ = ' ';
	 while (*ptr++ != '=');
	 *ptr2++ = '=';
	 while (*ptr == ' ')
	    ptr++;
	 *ptr2++ = ' ';
      }

      /*
         Now handle the expression
      */
      while (*ptr) {
	 if (*ptr == 'y' && *(ptr + 1) == 'y' && *(ptr + 2) == 'p' &&
	     *(ptr + 3) == 'v' && *(ptr + 4) == 't' && *(ptr + 5) == '[') {
	    ptr += 6;
	    index = atoi(ptr);
	    if (*ptr == '-')
	       ptr++;
	    while (*ptr++ != ']');
	    if (!(items + index - 1)) {
	       if (info[state].needs_shift &&
		   !info[state].complex_token &&
		   items == 1) {
		  *ptr2++ = 'y';
		  *ptr2++ = 'y';
		  *ptr2++ = 'l';
		  *ptr2++ = 'v';
		  *ptr2++ = 'a';
		  *ptr2++ = 'l';
		  if (*ptr == '.') {
		     *ptr2++ = *ptr++;
		     while ((*ptr >= 'A' && *ptr <= 'Z') || *ptr == '_' ||
			    (*ptr >= 'a' && *ptr <= 'z') ||
			    (*ptr >= '0' && *ptr <= '9'))
			*ptr2++ = *ptr++;
		  }
	       }
	       else {
		  *ptr2++ = 'v';
		  *ptr2++ = 'a';
		  *ptr2++ = 'l';
		  *ptr2++ = 'u';
		  *ptr2++ = 'e';
		  *ptr2++ = '_';
		  *ptr2++ = 's';
		  *ptr2++ = 't';
		  *ptr2++ = 'a';
		  *ptr2++ = 'c';
		  *ptr2++ = 'k';
		  if (*ptr == '.') {
		     ptr++;
		     *ptr2++ = '-';
		     *ptr2++ = '>';
		     while ((*ptr >= 'A' && *ptr <= 'Z') || *ptr == '_' ||
			    (*ptr >= 'a' && *ptr <= 'z') ||
			    (*ptr >= '0' && *ptr <= '9'))
			*ptr2++ = *ptr++;
		  }
	       }
	    }
	    else {
	       *ptr2++ = 'v';
	       *ptr2++ = 'a';
	       *ptr2++ = 'l';
	       *ptr2++ = 'u';
	       *ptr2++ = 'e';
	       *ptr2++ = '_';
	       *ptr2++ = 's';
	       *ptr2++ = 't';
	       *ptr2++ = 'a';
	       *ptr2++ = 'c';
	       *ptr2++ = 'k';
	       *ptr2++ = '[';
	       (void) sprintf(num, "%d", items + index - 1);
	       ptr3 = num;
	       while (*ptr3)
		  *ptr2++ = *ptr3++;
	       *ptr2++ = ']';
	       if (*ptr == '.') {
		  *ptr2++ = '.';
		  ptr++;
		  while ((*ptr >= 'A' && *ptr <= 'Z') || *ptr == '_' ||
			 (*ptr >= 'a' && *ptr <= 'z') ||
			 (*ptr >= '0' && *ptr <= '9'))
		     *ptr2++ = *ptr++;
	       }
	    }
	 }
	 else
	    *ptr2++ = *ptr++;
      }
      *ptr2 = '\0';

      (void) printf("%s", action_ptr->reference);
      (void) printf("%s%s\n", tab, line);
   }

   /*
      If there is no semantic action, then we may have to generate code to
      perform the default action.
   */
   else if (info[state].needs_shift && info[state].save_yylval &&
	    !info[state].might_shift && !info[state].complex_token)
      (void) printf("%s*value_stack = yylval;\n", tab, items - 1);
}

/*
 * get_array - get the value of an array element from an array
 */
short get_array(register array array_item, register int index) {

   while (array_item) {
      if (index < ARRAYSIZE)
	 return (array_item->value[index]);
      array_item = array_item->next;
      index -= ARRAYSIZE;
   }

   (void) fprintf(stderr, "%s: array index does not exist\n", myname);
   exit(1);
}

/*
 * read_semantic_actions - scan the input file for the semantic actions
 */
int read_semantic_actions(void) {

   register action_pointer new_action, last;
   register string_pointer ptr;
   string line;
   int items;

   items = 0;
   last = (action_pointer) NULL;
   semantic_actions = (action_pointer) NULL;
   for (;;) {
      if (!read_input_line(line))
	 return (items);
      if (*line == 'c' && *(line + 1) == 'a' && *(line + 2) == 's' &&
	  *(line + 3) == 'e' && *(line + 4) == ' ') {
	 new_action = (action_pointer) malloc(sizeof(action));
	 new_action->number = (short) atoi(line + 5);
	 new_action->next = (action_pointer) NULL;
	 if (last)
	    last->next = new_action;
	 else
	    semantic_actions = new_action;
	 last = new_action;
	 if (!read_input_line(line)) {
	    (void) fprintf(stderr, "%s: semantic action expected\n", myname);
	    exit(1);
	 }
	 new_action->reference =
	  (string_pointer) strcpy((char *) malloc(strlen(line) + 1), line);
	 if (!read_input_line(line)) {
	    (void) fprintf(stderr, "%s: semantic action expected\n", myname);
	    exit(1);
	 }
	 for (ptr = line; *ptr; ptr++);
	 if (*(ptr - 1) == '\n' && *(ptr - 2) == ';' && *(ptr - 3) == 'k' &&
	     *(ptr - 4) == 'a' && *(ptr - 5) == 'e' && *(ptr - 6) == 'r' &&
	     *(ptr - 7) == 'b' && *(ptr - 8) == ' ') {
	    ptr -= 9;
	    while (*ptr == ' ')
	       ptr--;	/* { */
	    if (*ptr == '}')
	       ptr--;
	    while (*ptr == ' ')
	       ptr--;
	    *(ptr + 1) = '\0';
	    ptr = line;
	    while (*ptr == ' ')
	       ptr++;
	    if (*ptr == '{')	/* } */
	       ptr++;
	    while (*ptr == ' ')
	       ptr++;
	    new_action->body =
	     (string_pointer) strcpy((char *) malloc(strlen(ptr) + 1), ptr);
	 }
	 else {
	    (void) fprintf(stderr, "%s: semantic action expected\n", myname);
	    exit(1);
	 }
	 items++;
      }
   }
}

/*
 * read_array - scan the input file while reading the contents of an array
 */
int read_array(array * array_item) {

   register char c;
   register string_pointer ptr;
   register int length;
   string line;
   array item, last;

   length = 0;
   last = (array) NULL;
   *array_item = (array) NULL;
   for (;;) {
      if (!read_input_line(line)) {
	 (void) fprintf(stderr, "%s: unexpected end of file\n", myname);
	 exit(1);
      }

      ptr = line;
      while (c = *ptr) {/* { */
	 if (c == '}')
	    return (length);
	 if (c == '-' || (c <= '9' && c >= '0')) {
	    if (!(length & (ARRAYSIZE - 1))) {
	       item = (array) malloc(sizeof(array_element));
	       item->next = (array) NULL;
	       if (last)
		  last->next = item;
	       else
		  *array_item = item;
	       last = item;
	    }
	    item->value[length & (ARRAYSIZE - 1)] = (short) atoi(ptr);
	    do {
	       ptr++;
	    } while (*ptr <= '9' && *ptr >= '0');
	    length++;
	 }
	 else
	    ptr++;
      }
   }
}

/*
 * read_define - read a preprocessor constant definition value
 */
int read_define(char *ident) {

   register string_pointer ptr;
   string line;

   if (!read_input_line(line)) {
      (void) fprintf(stderr, "%s: constant definition expected\n", myname);
      exit(1);
   }

   ptr = line;
   if (*ptr++ == '#' && *ptr++ == ' ' && *ptr++ == 'd' &&
       *ptr++ == 'e' && *ptr++ == 'f' && *ptr++ == 'i' &&
       *ptr++ == 'n' && *ptr++ == 'e' && *ptr++ == ' ') {
      while (*ident && *ptr++ == *ident++);
      if (*ptr++ == ' ' && !*ident)
	 return (atoi(ptr));
   }

   (void) fprintf(stderr, "%s: constant definition expected\n", myname);
   exit(1);
}

/*
 * read_input_line - read the next line from the input file
 */
int read_input_line(string_pointer line) {

   register int c, length;

   length = 0;
   do {
      c = getchar();
      if (c == EOF) {
	 *line = '\0';
	 return (length);
      }
      if (++length == STRINGSIZE) {
	 (void) fprintf(stderr, "%s: input line too long\n", myname);
	 exit(1);
      }
      *line++ = (char) c;
   } while (c != (int) '\n');
   *line = '\0';
   return (length);
}

/*
 * write_output_line - write a line to the output file
 */
void write_output_line(register string_pointer line) {

   while (*line)
      putchar(*line++);
}

/*
 * intialize_tokens - initialize the tokens data base
 */
void initialize_tokens(void) {

   register int token;
   string line;
 
   /*
      Only the normal character tokens are initially defined
   */
   highest_token = 126;

   /*
      Initialize control characters
   */
   for (token = 0; token < 32; token++) {
      (void) sprintf(line, "%d", token);
      tokens[token].ident = strcpy((char*) malloc(4), line);
      tokens[token].complex_token = FALSE;
      tokens[token].uses_yylval = FALSE;
   }

   /*
      Initialize the printable characters
   */
   for (token = 32; token < 127; token++) {
      if (token == (int) '\'')
	 tokens[token].ident = "((int) '\\'')";
      else if (token == (int) '\\')
	 tokens[token].ident = "((int) '\\\\')";
      else {
	 (void) sprintf(line, "((int) '%c')", token);
	 tokens[token].ident = strcpy((char *) malloc(12), line);
      }
      tokens[token].complex_token = FALSE;
      tokens[token].uses_yylval = FALSE;
   }

   /*
      Indicate that the rest of the tokens are undefined
   */
   for (token = 127; token < MAXTOKEN; token++)
      tokens[token].ident = (string_pointer) NULL;
}

/*
 * token_value - parse a define line to obtain the value of a token
 */
void token_value(register string_pointer ptr) {

   register string_pointer ptr2;
   string ident;
   int token;
 
   /*
      Ensure that this is a define line
   */
   if (*ptr++ == '#' && *ptr++ == ' ' && *ptr++ == 'd' &&
       *ptr++ == 'e' && *ptr++ == 'f' && *ptr++ == 'i' &&
       *ptr++ == 'n' && *ptr++ == 'e' && *ptr++ == ' ') {

      /*
         Read the name of the identifier defined
      */
      ptr2 = ident;
      while (*ptr != ' ' && *ptr != '\n' && *ptr)
	 *ptr2++ = *ptr++;
      *ptr2 = '\0';

      /*
         Determine the value that the identifier is defined to
      */
      if (*ident && *ptr++ == ' ') {
	 token = atoi(ptr);

	 /*
	    Make sure that it is a valid token number
	 */
	 if (token >= MAXTOKEN) {
	    (void) fprintf(stderr,
			   "%s: token value %d is above the limit of %d\n",
			   myname, token, MAXTOKEN - 1);
	    exit(1);
	 }

	 /*
	    Check to be sure that the identifier is a valid scanner token
	 */
	 if (!strcmp(ident, "CON") || !strcmp(ident, "LBL")) {
	    tokens[token].ident = strcpy((char *) malloc(4), ident);
	    tokens[token].complex_token = TRUE;
	    tokens[token].uses_yylval = TRUE;
	 }
	 else if (!strcmp(ident, "FCON")) {
	    tokens[token].ident = strcpy((char *) malloc(5), ident);
	    tokens[token].complex_token = TRUE;
	    tokens[token].uses_yylval = TRUE;
	 }
	 else if (!strcmp(ident, "LOCAL") || !strcmp(ident,"ERROR")) {
	    tokens[token].ident = strcpy((char *) malloc(6), ident);
	    tokens[token].complex_token = TRUE;
	    tokens[token].uses_yylval = TRUE;
	 }
	 else if (!strcmp(ident, "GLOBAL")) {
	    tokens[token].ident = strcpy((char *) malloc(7), ident);
	    tokens[token].complex_token = TRUE;
	    tokens[token].uses_yylval = TRUE;
	 }
	 else if (*ident == 'R' && *(ident + 1) == 'E' && *(ident + 2) == 'G' &&
		  ((*(ident + 3) >= '0' && *(ident + 3) <= '9') ||
		   (*(ident + 3) >= 'A' && *(ident + 3) <= 'F')) &&
		  !*(ident + 4)) {
	    tokens[token].ident = strcpy((char *) malloc(5), ident);
	    tokens[token].complex_token = TRUE;
	    tokens[token].uses_yylval = TRUE;
	 }
	 else if (*ident >= 'A' && *ident <= 'Z' && *(ident + 1) == 'M' &&
		*(ident + 2) == 'E' && *(ident + 3) == 'M' && !*(ident + 4)) {
	    tokens[token].ident = strcpy((char *) malloc(5), ident);
	    tokens[token].complex_token = TRUE;
	    tokens[token].uses_yylval = FALSE;
	 }
	 else if (*ident >= 'A' && *ident <= 'Z' && *(ident + 1) >= 'A' &&
		  *(ident + 1) <= 'Z' && !*(ident + 2)) {
	    tokens[token].ident = strcpy((char *) malloc(3), ident);
	    tokens[token].complex_token = TRUE;
	    tokens[token].uses_yylval = FALSE;
	 }
	 else {
	    (void) fprintf(stderr, "%s: invalid token identifier %s\n",
			   myname, ident);
	    exit(1);
	 }

	 /*
	    Keep track of the highest token number
	 */
	 if (token > highest_token)
	    highest_token = token;
      }
   }
}

/*
 * token_identifier - return the string used to identify this token
 */
string_pointer token_identifier(int token) {

   if (token <= highest_token && tokens[token].ident)
      return (tokens[token].ident);

   (void) fprintf(stderr, "%s: undefined token value %d\n", myname, token);
   exit(1);
}

/*
 * complex_token - indicate if the token can only be obtained by calling yylex()
 */
int complex_token(int token) {

   if (token <= highest_token && tokens[token].ident)
      return (tokens[token].complex_token);

   (void) fprintf(stderr, "%s: undefined token value %d\n", myname, token);
   exit(1);
}

/*
 * uses_yylval - indicate which tokens return a value in the "yylval" variable
 */
int uses_yylval(int token) {

   if (token <= highest_token && tokens[token].ident)
      return (tokens[token].uses_yylval);

   (void) fprintf(stderr, "%s: undefined token value %d\n", myname, token);
   exit(1);
}








