
/************************************************************************
 *                                                                      *
 * Copyright (c) 1987, 1988, 1989, 1990, 1991, 1992 -                   *
 *     The Rector and Visitors of the University of Virginia            *
 *                                                                      *
 * This code may not be distributed further without permission          *
 *      from the University of Virginia.                                *
 * This code is distributed WITHOUT ANY WARRANTY.  No claims are        *
 *      made as to whether it serves any particular purpose or even     *
 *      works at all.                                                   *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 
 Revision Control information

 $Source: /home/zephyr/zephyr-cvs/zephyr/backend/vpo/lib/locals.c,v $
 $Date: 2004/01/29 01:21:17 $
 $Revision: 1.37 $
 $Name:  $

 Log of changes:

 $Log: locals.c,v $
 Revision 1.37  2004/01/29 01:21:17  clc5q
 Maintain list of loops containing each CFG node throughout all global
    data profiling phases.
 Correct macro name used to control global profiling of stores.
 Set a threshold of minimum improvement percentage in reduction of cache
    interference by local refinement of our global data placement. If the
    improvement is less than this percentage, consider it to be zero.

 Revision 1.36  2004/01/21 18:39:55  jdh8d
 Latest Vista additions/changes.  Don't yell at me if it's broken/bad.. prasad wanted these changes.  I've reviewed them, and tested,  minimally.. things appear OK (for the batch compiler)

 Revision 1.35  2003/12/27 02:35:21  clc5q
 Tinker with repetition limits in the local refinement stage of data placement.
 Experiment with path profiling thresholds; use the L1 data cache size rather
    than the L2 cache size when imposing a size limit on a path, and cut
    the node weight cutoff point down to 1.2 to avoid terminating a path within
    a loop, as profiling is relatively cheap.
 Account for programs translated from Fortran by f2c, which have MAIN__()
    instead of main() as their starting function.
 Special cases in BypassCFG() for left back edges, making it easier to bypass
    in the presence of left self edges and left back edges to bypassed nodes.
 Maintain the bypassed field in CFG nodes during BypassCFG() processing.
 Properly account for globals that are bigger than the entire L2 cache. They
    are not placed in the CacheMap[], but must be part of the GlobalsSize.
    Hence CacheMapSize is no longer always equal to GlobalsSize.

 Revision 1.34  2003/12/17 16:11:17  clc5q
 Bracket data placement functions with ifdef CALLGRAPH for targets that
    are built without callgraph code.

 Revision 1.33  2003/12/16 22:12:18  clc5q
 Massive update to global data placement implementation.

 Revision 1.32  2003/11/18 02:01:58  clc5q
 Major implementation portion of global data profiling and placement.

 Revision 1.31  2003/11/03 00:36:25  clc5q
 Begin implementing data placement profiling of globals.
 Code functions addGlobalSym(), findGlobalFromName(), and emitAsmForGlobals().
 Call these functions from declaration() as needed.

 Revision 1.30  2003/10/21 17:28:45  clc5q
 Ran indent tool and made other cosmetic changes by hand.

 Revision 1.29  2003/07/11 21:00:14  clc5q
 Add assertions and clean up code in locals_used().

 Revision 1.28  2003/05/04 02:01:56  clc5q
 In reorder_local(), maintain "loce", which points to the end of the
     locals list, as we sort the locals.
 Re-indent reorder_local() to make it readable.
 Correct cut-and-paste initialization error in declaration().

 Revision 1.27  2003/04/09 20:54:55  clc5q
 Link each ignode and basic block to the loops that contain it.

 Revision 1.26  2003/04/03 20:06:10  jg9h
 Merged latest in VISTA with latest in vpo.

 Revision 1.25  2003/03/18 18:14:59  jdh8d
 arg, i'll get it right now... its tested for sure

 Revision 1.24  2003/03/18 17:18:34  jdh8d
 *** empty log message ***

 Revision 1.22  2003/03/06 16:14:18  jdh8d
 *** empty log message ***

 Revision 1.21  2003/02/06 13:13:33  jdh8d
 *** empty log message ***

 Revision 1.20  2003/02/06 04:46:29  jg9h
 Merged VISTA into mainline vpo under the INTERACT preprocessor flag

 Revision 1.19  2003/02/02 03:37:26  clc5q
 Initialize the locuse->indref field to TRUE in order to be safe during
    the optimizations (such as instruction selection) that occur before
    the locals are processed by live_variable_analysis() or locals_used().
 
 Revision 1.18  2003/01/07 22:43:48  clc5q
 Guard against NULL pointer dereferencing in is_pc() by not calling it if
    cblk->rtlend->val is NULL.

 Revision 1.17  2002/11/27 16:28:08  jwd
 Sampoorani updated this code so that variables that heavily used
 in loops can be allocated even if no variable is available.  This
 code was originally written by Raja, but it was broken from the get
 go.

 Revision 1.16  2002/10/24 21:36:14  clc5q
 Ensure that all fields in a new struct locuse are initialized in
   declaration(), including new fields added for stack frame optimizations
   and load coalescing.

 Revision 1.15  2002/10/03 15:49:37  clc5q
 Replace "extern ..." garbage in *.c files with extern declarations in
 header files.

 Revision 1.14  2002/08/08 19:00:12  clc5q
 Add parameter "sort" to locals_used(). Use it as a guard for whether to
 call reorder_locals().

 Revision 1.13  2002/07/11 15:22:31  clc5q
 Clean up #include chains. Use #include instead of extern to get access
 to global variables from other modules.

 Revision 1.12  2002/06/05 20:08:53  jdh8d
 Changes to support the stache, including routines to build a callgraph.

 Some cosmetic changes.

 Minor bug fixes.

 Changed build scripts to build the mips compiler for SimpleScalar if
 SSAS is set to "-DSS_AS"

 Revision 1.11  2001/11/27 20:40:55  rv3s
 color_global_allocation now uses the lv_freq field to adjust access counts

 Revision 1.10  2001/11/16 17:23:11  jdh8d
 *** empty log message ***

 Revision 1.9  2001/06/13 15:50:30  rv3s
 In color_global_allocation, if a variable is not accessed frequently enough
 in a function, attempt allocation over the life-time of a loop.

 Revision 1.8  2001/05/23 20:28:21  jdh8d


 Re-worked all of the backend to support variable
 sized registers.  IT IS  NO LONGER SAFE TO ASSUME
 THAT REGISTERS ARE TWO BYTES.  Mips currently has
 4-byte registers, others will follow (or be destroyed!)

 Revision 1.7  2001/04/20 19:08:54  jdh8d
 *** empty log message ***

 Revision 1.6  2001/03/23 21:58:18  jdh8d

 initialized memChar in locals.c
 fixed combiner bug, see comments in combine.c

 Revision 1.5  2000/06/09 19:37:50  jwd
 Changed calls to VET to indicate which phases are mandatory and
 which are optional.

 Revision 1.4  1999/01/06 15:36:59  jwd
 Added VET directives.
 Revision 1.3  1998/04/03 18:19:37  jwd
 First ANSI C version.

*************************************************************************/

/*-------------History of Changes -------------------
   1/12/98:  Modified for ANSI-C function prototyping
   ---------------------------------------------------*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "vpo.h"
#include "assign.h"
#include "cdmotion.h"
#include "controlflow.h"
#include "dataflow.h"
#include "links.h"
#include "misc.h"
#include "vectors.h"
#include "vpo2.h"
#include "regs.h"
#include "locals.h"

#include "regs2.h" /* Machine Dependent */
#include "rtl.h"   /* Machine Dependent */
#include "inst.h"  /* Machine Dependent */

#ifdef VET
#include "view.h"
#endif

#ifdef INTERACT
#include "encode_rtl.h"
#endif

#ifdef CALLGRAPH
#include "callgraph.h"
#include "placement.h" /* Machine Dependent */
#endif

/*
 * This file contains functions to perform global register allocation and
 * associated optimizations on procedures.
 */

struct locinfo {
   short st;                    /* local has been stored flag    */
   short ld;                    /* local has been loaded flag    */
   struct locuse *ptr;          /* pointer to local referenced   */
};

/* Minimum estimate field value for a CFG node to be considered
 * important enough to include in global path profiling.
 * First constant is for intra-node profiling only, second (larger)
 * constant is for path profiling.  Determined by trial and error.
 */
#define GLOBAL_PROFILING_THRESHOLD1 0.01
#define GLOBAL_PROFILING_THRESHOLD2 (DEFAULT_LOOP_ITERATIONS * 8.2)

/* When starting a path at minimum estimate field value == X, at what
 * estimate field value (fraction * X) should we terminate loops?
 * NOTE: We will never allow the cutoff to fall as low as 1.0, else
 * we will cross from one outer loop to another.
 */
#define PATH_CUTOFF_FRACTION  0.0

/* UpdateGlobalVariableAffinity() is recursive, and the thresholds
 * above and the size limits might not be enough to keep it from
 * recursing so deeply that it overflows the run-time stack.
 * Limit the recursion depth without preventing good profiling.
 */
#define MAX_PROFILING_RECURSION_DEPTH  3000

/* Do we use the ProfilingStack to follow function calls as we profile? */
#define INTERPROCEDURAL_PROFILING FALSE

/* How many path heads should we try to find, i.e. we
 * should start path profiling at the busiest how many
 * CFG nodes, based on their estimate field values.
 */
#define MAX_PROFILING_PATHS 1200

/* How many CFG nodes can a leaf function have before we stop
 * trying to inline it because of time and virtual memory
 * considerations?
 */
#define MAX_SIZE_TO_INLINE 10000

/* An alternative approach: Inline until the total nodes
 * will exceed a fixed budget. Small programs can inline
 * to a greater degree than programs with numerous functions.
 */
#define CFG_NODE_BUDGET  600000
static int CFGNodeBudget = CFG_NODE_BUDGET;
#define ALLOW_NODE_BUDGET_GROWTH  FALSE

/* Set a time limit on inlining, in seconds, in case we
 * start paging out virtual memory and the other limitations
 * have not done the job yet.
 */
#define INLINING_TIME_LIMIT  480.0

/* How many iterations of the final loop of InlineFunctionsIntoMain()
 * will we perform before we stop (to save time and memory) ?
 */
#define MAX_INLINING_ITERATIONS 200

/* What fraction of the original cache interference is considered to
 * signal essentially no improvement in our local refinement stage in
 * global data placement?
 */
#define LOCAL_REFINEMENT_THRESHOLD 0.99

/* Output debug data about CFG inlining. */
#define DEBUG_INLINING   TRUE

/* Output debug data about global data profiling. */
#define DEBUG_PROFILING   TRUE
#define DEBUG_PROFILING2 FALSE /* Verbose. */
#define DEBUG_PROFILING3 FALSE /* Verbose. */

/* Output debug data about global data placement. */
#define DEBUG_PLACEMENT   TRUE

/* Set all functions' topEstimate field to a constant 1.0 ? */
#define CONSTANT_CALLS  TRUE
/* Skip inlining functions? */
#define SKIP_INLINING FALSE
/* Profile all globals, no matter how big? */
#define NO_PROFILE_SIZE_LIMIT  TRUE

/* Profile loads and stores to globals, or just loads? */
#if L1_DATA_CACHE_NO_WRITE_ALLOCATE
#define IGNORE_STORES  TRUE
#else
#define IGNORE_STORES FALSE
#endif

/* Compile code for global data placement? */
#define GLOBAL_DATA_PLACEMENT FALSE

/* The Chaco graph partitioning library has no header files to include, so
 * we will prototype its interface here.
 */
extern
int interface(int, int *, int *, int *, float *, float *, float *, float *,
              char *, char *, short *, int, int, int *, double *, int, int,
              int, int, int, double, long);

/* Macro to compute the number of cache lines occupied by a chain of globals. */
#define CACHE_LINES(gptr) \
 (((GlobalChainSize(gptr) - 1) / CACHE_LINE_PACKING_QUANTUM) + 1)

/* Macro to compute the number of cache lines occupied by a size in bytes. */
#define CACHE_LINES_FROM_SIZE(size) \
 (((size - 1) / CACHE_LINE_PACKING_QUANTUM) + 1)

struct hashedCFGNode {
   struct GlobalPlacementCFGNode *node;
   int index;
   struct hashedCFGNode *next;
};

struct globlist {
   struct globalsym *current;
   struct globlist *prev;
   struct globlist *next;
   unsigned short ArrIndex;   /* Into malloc()-ed affinity array. */
};

/* Mapping of globals to cache lines. */
static struct globalsym **CacheMap = NULL;
static int CacheMapSize; /* Number of bytes of data in the CacheMap. */
static int CacheMapLines; /* Number of (L2) cache lines in the CacheMap. */
static int *PartitionStartingLine = NULL;
  /* First cache line # per partition #. */

static int *PartitionGlobalsCount = NULL;
  /*  # used globals per partition #. */
static int *PartitionShadowCount = NULL;
  /*  # coalesced, non-leftmost globals per partition #. */

static int InlineSizeLimit = MAX_SIZE_TO_INLINE;
static int InlineLimitDoubled = FALSE;

/* We keep a stack of "return addresses" if we are forced to profile
 * without being able to inline all functions until main() has no call
 * sites.  We traverse functions when we come across function calls,
 * leaving on the top of the stack a struct containing a pointer to the
 * CFG node from which we came, as well as a pointer to the global var
 * entry in that node that had the function call.
 */
struct ReturnAddress {
   struct GlobalPlacementCFGNode *node;
   struct varl *call;
};

static struct ReturnAddress ProfilingStack[MAX_PROFILING_RECURSION_DEPTH];
static int ProfilingStackPointer = 0; /* Points to entry above the top. */

/* For each function in the program, we need a data structure mapping
 * loops within that function to an ID that is unique across the whole
 * program.
 */
struct RTLLoopData {
   struct loopl *loop; /* PTR value from RTL for current function. */
   int ID;             /* From CFGLoopID when loop is processed. */
   struct RTLLoopData *left;
   struct RTLLoopData *right;
};

static struct RTLLoopData *RTLLoopTree = NULL;
static int CFGLoopID = 0;

/* Pad arrays that are big enough to fill the L1 cache way size so that
 * no two such arrays have the same starting address modulo the cache
 * size, to avoid thrashing if they are accessed within the same loop.
 */
#define PAD_LARGE_ARRAYS  TRUE

/* What is the maximum edge weight we want to send to Chaco (to prevent
 * it from having an overflow problem in its computations)?
 */
#define CHACO_AFFINITY_VALUE_LIMIT  10000.0

/* Control the time consumed by 2-opt refinement of our data placement. */
#define MAX_CONSECUTIVE_BAD_SWAPS   80
#define MAX_SWAPS_PER_PARTITION     40
#define MAX_SWAPS_PER_REFINEMENT 10000
#define TWO_OPT_ITERATIONS          10

static FILE *DebugFile = NULL;  /* Holds debugging output. */

struct globalsym *globSymStart = NULL; /* Global variable list start. */
struct globalsym *globSymEnd = NULL;   /* Global variable list end. */
int CollectingGlobals = FALSE; /* Is data placement gathering global lines? */

struct locuse *locs = NULL;     /* local variable and argument info */
struct locuse *loce = NULL;     /* last item in local information   */
struct ignode *igptr = NULL;    /* pointer to coloring graph        */
struct ignode *iglast = NULL;   /* pointer to end of coloring graph */
int   csnum[MAXREGS];           /* callersave points for each reg   */

static int GlobalsCount = 0;
static struct globalsym **Globals; /* Array of globalsym pointers, which
                                    * can be sorted, directly indexed, etc.
                                    */
static float *AffinityArray = NULL;

#ifdef CALLGRAPH
/* Because the current function has not finished optimizations and has
 * not yet been added to the callgraph, we do not have a call_graph_node
 * when we perform our global profiling.  Therefore, we must save the
 * control flow graph we build until we can transfer the pointer into
 * the topCFG field within the call_graph_node.
 */
static struct GlobalPlacementCFGNode *savedCFGNode = NULL;
static int IrreducibleFlag;  /* Save status of savedCFGNode. */
static int NumBasicBlocks = 0;
static int CFGLoopIDCounter = 0; /* Unique ID # across all loops in program. */

void  GroupGlobalChain(struct globalsym *);
void  SortCoalescedGlobals();
static int GetUniformRandom(int, int);
void  InsertGlobal(struct globalsym *, struct globalsym *);
void  ComputeStaticCallEstimates();
void  CopyCFGLoopList(struct GlobalPlacementCFGNode *,
                      struct GlobalPlacementCFGNode *);
#endif  /* CALLGRAPH */

#define DEBUG_VECTOR_FREE_LIST FALSE

#if DEBUG_VECTOR_FREE_LIST
void validate_vector_free_list()
{
   char **ptr = freeptrs[12];
   int   count = 0;

   while (NULL != ptr) {
      (void) fprintf(stderr, "vector free list address: %x\n", ptr);
      assert(0 != (0x00ffffff & (unsigned int) ptr));
      ptr = (char **) *ptr;
      ++count;
   }

   (void) fprintf(stderr, "Vector free list length: %d \n", count);
}
#endif

/*
 * OpenGlobalDebugFile - open a disk file to hold debugging displays.
 */
FILE *OpenGlobalDebugFile() {
   char *filename;

   filename =
      (char *) SALLOC((int) (strlen("GlobalDebug") + strlen(".debug") + 2));
   if (NULL == filename) {
      (void) fprintf(stderr, "Out of virtual memory: "
                             "locals.c::OpenGlobalDebugFile() \n");
      exit(EXIT_FAILURE);
   }

   (void) strcpy(filename, "GlobalDebug");
   (void) strcat(filename, ".debug");

   DebugFile = fopen(filename, "w");
   if (NULL == DebugFile) {
      (void) fprintf(stderr, "ERROR: Cannot open file %s for writing.\n",
                     filename);
   }
   return DebugFile;

} /* end of OpenGlobalDebugFile() */

void CloseGlobalDebugFile() {
   assert(NULL != DebugFile);
   (void) fflush(DebugFile);
   (void) fclose(DebugFile);
   (void) fprintf(stderr, "Closed the DebugFile.\n");
   DebugFile = NULL;
}

/*
 * PrintLocalTime - Print the time (in local time zone) in a readable format.
 */
void PrintLocalTime() {
   time_t tempTime;
   struct tm *timeStruct;
   char *ASCIITime;

   tempTime = time(&tempTime);
   timeStruct = localtime(&tempTime);
   assert(NULL != timeStruct);
   ASCIITime = asctime(timeStruct);
   assert(NULL != ASCIITime);
   (void) fprintf(stderr, "%s \n", ASCIITime);

   return;
} /* end of PrintLocalTime() */

#ifdef CALLGRAPH
/*
 * FindNodeByNumber - Given a node number, find it in the call graph.
 * Used in the debugger.
 */
struct GlobalPlacementCFGNode *FindNodeByNumber(struct call_graph_node *cgnode,
                                                int number) {
   struct GlobalPlacementCFGNode *cnode;

   assert(NULL != cgnode);
   assert(0 <= number);

   cnode = cgnode->topCFG;
   do {
      if (number == cnode->NodeNum)
         return cnode;
      cnode = cnode->next;
   } while (NULL != cnode);
   return NULL;
} /* end of FindNodeByNumber() */

/*
 * FindGlobalReferences - Given a global symbol, find all occurrences of it in
 * the call graph. Used in the debugger.
 */
void FindGlobalReferences(struct call_graph_node *cgnode, char *global) {
   struct GlobalPlacementCFGNode *cnode;
   struct varl *tempVar;

   assert(NULL != cgnode);
   assert(NULL != global);

   cnode = cgnode->topCFG;
   do {
      if (NULL != cnode->prof->FirstGlobal) {
         tempVar = cnode->prof->FirstGlobal;
         do {
            if (0 == strcmp(tempVar->name, global))
               (void) fprintf(stderr, "Found in node %d\n", cnode->NodeNum);
            tempVar = tempVar->next;
         } while (NULL != tempVar);
      }
      cnode = cnode->next;
   } while (NULL != cnode);
   return;
} /* end of FindGlobalReferences() */

/* FindCFGNode - Return TRUE if we can find a given CFG node number by
 * starting at the given node and recursing only through left and right
 * children, terminating each path if we fall below the cutoff for node
 * weight.
 */
int FindCFGNode(struct GlobalPlacementCFGNode *cnode, float cutoff,
                int number) {
   assert(NULL != cnode);

   if (number == cnode->NodeNum)
      return TRUE;
   else if (cutoff > cnode->estimate)
      return FALSE;
   else if ((NULL != cnode->left)
            && (FindCFGNode(cnode->left, cutoff, number)))
      return TRUE;
   else
      return ((NULL != cnode->right)
              && (FindCFGNode(cnode->right, cutoff, number)));

} /* end of FindCFGNode() */

/*
 * FindCFGCycle - Return TRUE if we can find a cycle that begins and ends
 * at the given node without passing through nodes with estimate fields
 * below the cutoff threshold.
 */
int FindCFGCycle(struct GlobalPlacementCFGNode *cnode, float cutoff) {
   int number;

   assert(NULL != cnode);

   number = cnode->NodeNum;
   if ((NULL != cnode->left) && (FindCFGNode(cnode->left, number, cutoff)))
      return TRUE;
   else
      return ((NULL != cnode->right)
              && (FindCFGNode(cnode->right, number, cutoff)));

} /* end of FindCFGCycle() */
#endif  /* CALLGRAPH */

/*
 * ShowGlobalsList - debug function to display globalsym list.
 */
void ShowGlobalsList() {
   struct globalsym *gptr;

   if ((NULL == globSymStart) || (NULL == DebugFile))
      return;

   for (gptr = globSymStart; NULL != gptr; gptr = gptr->next) {
      if (NULL == gptr->CoalescedMSNeighbor)
         (void) fprintf(DebugFile, "\n");
      (void) fprintf(DebugFile, "%s  %d ", globals + gptr->gallocIndex,
                     gptr->AffinityIndex);
   }
   (void) fprintf(DebugFile, "\n");
   return;
} /* end of ShowGlobalsList() */

/*
 * ShowGlobalsArray - debug function to display Globals[] array.
 */
void ShowGlobalsArray() {
   int index;

   if ((NULL == Globals) || (NULL == DebugFile))
      return;

   for (index = 0; index < GlobalsCount; ++index) {
      if (NULL == Globals[index]->CoalescedMSNeighbor)
         (void) fprintf(DebugFile, "\n");
      (void) fprintf(DebugFile, "%s  %d ",
                     globals + Globals[index]->gallocIndex,
                     Globals[index]->AffinityIndex);
   }
   (void) fprintf(DebugFile, "\n");
   (void) fprintf(DebugFile, "\n");
   ShowGlobalsList();
   return;
} /* end of ShowGlobalsArray() */

/*
 * ShowAffinityArray - debug function to display Affinity[] array.
 */
static void ShowAffinityArray() {
   int row;
   int col;
   int *Relevant;
   int AllZeros;

   if ((NULL == AffinityArray) || (NULL == DebugFile))
      return;

   Relevant = (int *) SALLOC_PERM(sizeof(int) * GlobalsCount);
   if (NULL == Relevant) {
      (void) fprintf(stderr, "Out of memory: ShowAffinityArray().\n");
      exit(EXIT_FAILURE);
   }
   for (row = 0; row < GlobalsCount; ++row) {
      Relevant[Globals[row]->AffinityIndex] 
        = (NULL == Globals[row]->CoalescedMSNeighbor);
   }
   for (row = 0; row < GlobalsCount; ++row) {
      if (!Relevant[row])
         continue;
      AllZeros = TRUE;
      for (col = 0; col < GlobalsCount; ++col) {
         if (0.001 <  *(AffinityArray + row * GlobalsCount + col)) {
            AllZeros = FALSE;
            break;
         }
      }
      if (AllZeros)
         Relevant[row] = FALSE;
   }

   /* Print header showing which columns are relevant. */
   (void) fprintf(DebugFile, "Column:    ");
   for (row = 0; row < GlobalsCount; ++row) {
      if (!Relevant[row])
         continue;
      (void) fprintf(DebugFile, "%10d", row);
   }
   (void) fprintf(DebugFile, "\n");

   for (row = 0; row < GlobalsCount; ++row) {
      /* Don't print a row of all zeros. */
      if (!Relevant[row])
         continue;

      (void) fprintf(DebugFile, "Row: %5d ", row);
      for (col = 0; col < GlobalsCount; ++col) {
         if (!Relevant[col])
            continue;
         (void) fprintf(DebugFile, "%10.1f",
                        *(AffinityArray + row * GlobalsCount + col));
      }
      (void) fprintf(DebugFile, "\n");
   }
   (void) fprintf(DebugFile, "\n");

   FREE(Relevant);
   return;
}  /* end of ShowAffinityArray() */

#ifdef CALLGRAPH
/*
 * ShowCacheMap - debug function to display CacheMap[] array.
 */
static void ShowCacheMap(int CacheLevel) {
   int index;
   int col;
   int CacheColumns, CacheWayStride, CacheAssociativity;
   int CacheLine;

   if ((NULL == CacheMap) || (NULL == DebugFile) || (1 > CacheLevel))
      return;

   if (1 == CacheLevel) {
      CacheColumns = 1 + ((CacheMapSize - 1) / L1_DATA_CACHE_WAY_SIZE);
      CacheWayStride = L1_DATA_CACHE_WAY_SIZE / CACHE_LINE_PACKING_QUANTUM;
      CacheAssociativity = L1_DATA_CACHE_ASSOCIATIVITY;
   }
   else { /* Must be level 2 cache. */
      CacheColumns = 1 + ((CacheMapSize - 1) / L2_DATA_CACHE_WAY_SIZE);
      CacheWayStride = L2_DATA_CACHE_WAY_SIZE / CACHE_LINE_PACKING_QUANTUM;
      CacheAssociativity = L2_DATA_CACHE_ASSOCIATIVITY;
   }

   for (index = 0; index < CacheWayStride; ++index) {
      for (col = 0; col < CacheColumns; ++col) {
         CacheLine = (col * CacheWayStride) + index;
         if (CacheLine >= CacheMapLines) /* Last column can be partial. */
            break;
         (void) fprintf(DebugFile, "%12s", 
                        globals + CacheMap[CacheLine]->gallocIndex);
      }
      (void) fprintf(DebugFile, "\n");
   }
   (void) fprintf(DebugFile, "\n");
   return;
} /* end of ShowCacheMap() */

/*
 * ShowCFG - debug display of a CFG.
 */
void ShowCFG(struct call_graph_node *cgnode) {
   struct GlobalPlacementCFGNode *cnode;

   if (NULL == cgnode)
      return;

   cnode = cgnode->topCFG;

   (void) fprintf(stderr, "%12s %12s %12s %12s %12s %10s\n", cgnode->name,
                  "node", "left", "right", "next", "estimate");
   while (NULL != cnode) {
      (void) fprintf(stderr, "%12d %12x", cnode->NodeNum, cnode);
      if (NULL == cnode->prof->FirstGlobal)
         (void) fprintf(stderr, " %12x", cnode->left);
      else /* Print an asterisk next to the node address. */
         (void) fprintf(stderr, "%1s %11x", "*", cnode->left);

      if (!cnode->prof->LeftBackEdgeFlag)
         (void) fprintf(stderr, " %12x", cnode->right);
      else /* Print an asterisk next to the node address. */
         (void) fprintf(stderr, "%1s %11x", "*", cnode->right);

      if (!cnode->prof->RightBackEdgeFlag)
         (void) fprintf(stderr, " %12x", cnode->next);
      else /* Print an asterisk next to the node address. */
         (void) fprintf(stderr, "%1s %11x", "*", cnode->next);

      (void) fprintf(stderr, " %10.4f", cnode->estimate);

      if (TRUE == cnode->done) /* TRUE, not random garbage value. */
         (void) fprintf(stderr, "%1s\n", "*");
      else
         (void) fprintf(stderr, "\n");
      cnode = cnode->next;
   }
   return;
} /* end of ShowCFG() */

/*
 * DumpCFG - debug dump of a CFG to DebugFile.
 */
void DumpCFG(struct call_graph_node *cgnode) {
   struct GlobalPlacementCFGNode *cnode;

   if (NULL == cgnode)
      return;

   if (NULL == DebugFile)
      DebugFile = OpenGlobalDebugFile();

   assert(NULL != DebugFile);

   cnode = cgnode->topCFG;

   (void) fprintf(DebugFile, "%12s %12s %12s %12s %12s %10s\n", cgnode->name,
                  "node", "left", "right", "next", "estimate");
   while (NULL != cnode) {
      (void) fprintf(DebugFile, "%12d %12x", cnode->NodeNum, cnode);
      if (NULL == cnode->prof->FirstGlobal)
         (void) fprintf(DebugFile, " %12x", cnode->left);
      else /* Print an asterisk next to the node address. */
         (void) fprintf(DebugFile, "%1s %11x", "*", cnode->left);

      if (!cnode->prof->LeftBackEdgeFlag)
         (void) fprintf(DebugFile, " %12x", cnode->right);
      else /* Print an asterisk next to the node address. */
         (void) fprintf(DebugFile, "%1s %11x", "*", cnode->right);

      if (!cnode->prof->RightBackEdgeFlag)
         (void) fprintf(DebugFile, " %12x", cnode->next);
      else /* Print an asterisk next to the node address. */
         (void) fprintf(DebugFile, "%1s %11x", "*", cnode->next);

      (void) fprintf(DebugFile, " %10.4f", cnode->estimate);

      if (TRUE == cnode->done) /* TRUE, not random garbage value. */
         (void) fprintf(DebugFile, "%1s\n", "*");
      else
         (void) fprintf(DebugFile, "\n");
      cnode = cnode->next;
   }
   return;
} /* end of DumpCFG() */

/*
 * ShowCallees - debugging function to go through a CFG and display
 * the function names of callees.  Useful for figuring out what work remains
 * to be done to get a function (especially main()) to become a leaf node.
 */
void ShowCallees(struct GlobalPlacementCFGNode *cnode) {
   struct varl *tempVar;

   if (NULL == cnode)
      return;

   while (NULL != cnode) {
      assert(NULL != cnode->prof);
      tempVar = cnode->prof->FirstGlobal;
      while (NULL != tempVar) {
         if (0 > tempVar->index)
            (void) fprintf(stderr, "%s  %d\n", tempVar->name, cnode->NodeNum);
         tempVar = tempVar->next;
      }
      cnode = cnode->next;
   }
   return;
} /* end of ShowCallees() */

/* GetUniformRandom - return an integer uniformly distributed from
 * the lower bound to the upper bound.
 */
static int GetUniformRandom(int lower, int upper) {
   unsigned int RandomInt = rand();
   float Rfloat = ((float) RandomInt) / ((float) RAND_MAX);

   assert(lower < upper);
   assert(0 <= lower);
   assert(0.0 <= Rfloat);
   assert(1.0 >= Rfloat);

   if (1.0 == Rfloat)
      Rfloat = 0.9999;

   return ((int) (floor(Rfloat * (upper - lower + 1))));
}
#endif /* CALLGRAPH */

/*
 * simple_global_allocation - perform simple global register allocation
 */
int simple_global_allocation(int invocation, int *csalloc)
{
   register struct bblock *cblk;
   register struct locuse *ptr;
   register struct list *rtlptr;
   int   change, first, ifcheap, ifcs;
   char  reg[REGSTRING];

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_simple_global_allocation", invocation))
      dump_function_i("before simple_global_allocation()", invocation,
                      SHOW_SE);

   first = (invocation == 1);
#ifdef VET
   if (!moreopts)
      return;
#endif

#ifdef VET
#ifdef INTERACT
   begin_phase(REG_ALLOCATION_TRANS);
#else
   begin_phase(REGALLOC_PHASE);
#endif
#endif

   /*
    *  If this is not the first invocation of this allocator, then forget 
    *  all of the old (stale) local variable information.
    */
   if (!first)
      clean_locals();

   /*
    * Determine where locals are used in the function.  These data are
    * essential for determining the payoff of allocating registers to locals.
    */
   locals_used(TRUE);

   change = FALSE;
   for (ptr = locs; ptr; ptr = ptr->next) {

      /*
       * If this local has been marked to be ignored, then ignore it
       */
      if (ptr->nouse)
         continue;

      /*
       * Emit warning messages about locals that are declared but not used and
       * perform a cost-benefit analysis on each local.
       */
      ifcheap = FALSE;
      if (!ptr->reps) {
         if (first && !sww) {
            if (ptr->arg)
               (void) fprintf(stderr, "vpo: argument '%s'", ptr->item);
            else
               (void) fprintf(stderr, "vpo: local '%s'", ptr->item);
            if (fn)
               (void) fprintf(stderr, " not used in %s()\n", fn);
            else
               (void) fprintf(stderr, " not used\n");
         }
         ptr->nouse = TRUE;
#ifdef INTERACT
         begin_trans(VET_OPT_TRANS);
         update_locuse(ptr);
         end_trans();
#endif
         continue;
      }
      else if (!ptr->arg && ptr->reps == 1) {
         if (first && !sww) {
            (void) fprintf(stderr, "vpo: local '%s'", ptr->item);
            if (fn)
               (void) fprintf(stderr, " referenced only once in %s()\n", fn);
            else
               (void) fprintf(stderr, " referenced only once\n");
         }
         ptr->nouse = TRUE;
#ifdef INTERACT
         begin_trans(VET_OPT_TRANS);
         update_locuse(ptr);
         end_trans();
#endif
         continue;
      }
      else if (ptr->vol)
         continue;
      else if (!ptr->arg && ptr->reps < minlocrefs) {
         ptr->nouse = TRUE;
         ifcheap = TRUE;
#ifdef INTERACT
         begin_trans(VET_OPT_TRANS);
         update_locuse(ptr);
         end_trans();
#endif
      }
      else if (ptr->arg && ptr->reps < minargrefs) {
         ptr->nouse = TRUE;
         ifcheap = TRUE;
#ifdef INTERACT
         begin_trans(VET_OPT_TRANS);
         update_locuse(ptr);
         end_trans();
#endif
      }

      /*
       * Determine the maximum number of saves and restores that can be
       * tolerated if a callersaves register is allocated to this local.
       */
      ifcs = ptr->reps - minargrefs;
      if (ifcs < 0)
         ifcs = 0;

      /*
       *  If there is no reason why this local should not be allocated to a
       * register, then attempt to allocate a register of the appropriate type
       * to this local.  If a register is allocated, modify the function to
       * change all of the references of the local to references to the
       * register.
       */
      if (canreplace(ptr->atype, top->regstate, reg, ifcheap, ifcs)) {

#ifdef VET
         begin_trans(VET_OPT_TRANS);
#endif
         /*
          * Indicate that this item has no more references, this may change if
          * the item is an argument or if it has been allocated to a
          * callersaves register.
          */
         ptr->reps = 0;

         /*
          * If the local is an argument, then we need to place an initial load
          * of the register at the entry point of the function unless the
          * argument enters the function via a register, in which case it will
          * be the task of fixentry() to ensure that the entry register and the
          * use register match up.
          */
         if (ptr->arg) {
            if (regarg(ptr))
               (void) strcpy(ptr->regin, reg);
            else {
               load(ptr->atype, TRUE, ptr->code, top, top->rtls, reg, TRUE);
               ptr->reps++;
            }
         }

         /*
          * Modify the function to change all of the references of the local to
          * references to the allocated register.
          */
         for (cblk = top; cblk; cblk = cblk->down) {
            for (rtlptr = cblk->rtls; rtlptr; rtlptr = rtlptr->next) {
               if (cnts(rtlptr) && vin(ptr->olist, rtlptr->onum))
                  loctoreg(ptr, rtlptr, ptr->atype, reg);
            }
         }

         /*
          * Indicate that the local has been allocated to a register and that
          * some modifications have been made to the function.  If the register
          * that the local was allocated to needs to be saved and restored
          * around function calls, then we place the value of the register in
          * the "csreg" field so that move_caller_save() knows that this item
          * needs to be looked at.
          */
         ptr->reg = ptr->nouse = change = TRUE;
#ifdef INTERACT
         update_locuse(ptr);
#endif
         if (hascs && callersave(reg)) {
            *csalloc = TRUE;
            (void) strcpy(ptr->csreg, reg);
         }
#ifdef VET
         end_trans();
#endif
      }
   }
#ifdef VET
#ifdef INTERACT
   end_phase(REG_ALLOCATION_TRANS);
#else
   end_phase(REGALLOC_PHASE);
#endif
#endif

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_simple_global_allocation", invocation))
      dump_function_i("after simple_global_allocation()", invocation,
                      SHOW_SE);

   /*
    * Return indicating if any changes have been made to the function
    */
   return (change);
}

/*
 * color_global_allocation - allocates using a coloring algorithm
 */
int color_global_allocation(int invocation)
{
   struct ignode *iptr, *sptr;
   struct ignode *eptr;
   struct list *rtlptr;
   struct locuse *ptr;
   struct conflict *cptr;
   vector allocate;
   int   freq, type = -1, change, first, ifcheap, entry;
   char  reg[REGSTRING];

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_color_global_allocation", invocation))
      dump_function_i("before color_global_allocation()", invocation,
                      SHOW_LABELS | SHOW_USE | SHOW_NONRTLS | SHOW_SE);

   first = (invocation == 1);
   change = FALSE;

   /*
    * Complain about locals that are declared and not used and items that are
    * only written to (if the "no warning messages" flag is not set).
    */
#ifdef VET
#ifdef INTERACT
   begin_phase(REG_ALLOCATION_TRANS);
#else
   begin_phase(REGALLOC_PHASE);
#endif
#endif

   for (ptr = locs; ptr; ptr = ptr->next) {
      if (ptr->nouse)
         continue;
      if (!ptr->reps) {
         if (first && !sww) {
            if (ptr->arg)
               (void) fprintf(stderr, "vpo: argument '%s'", ptr->item);
            else
               (void) fprintf(stderr, "vpo: local '%s'", ptr->item);
            if (fn)
               (void) fprintf(stderr, " not used in %s()\n", fn);
            else
               (void) fprintf(stderr, " not used\n");
         }
         ptr->nouse = TRUE;
#ifdef INTERACT
         begin_trans(VET_OPT_TRANS);
         update_locuse(ptr);
         end_trans();
#endif
      }
      else if (!ptr->arg && ptr->reps == 1) {
	if(first && !sww){
	  (void) fprintf(stderr, "vpo: local '%s'", ptr->item);
	  if (fn)
	    (void) fprintf(stderr, " referenced only once in %s()\n", fn);
	  else
	    (void) fprintf(stderr, " referenced only once\n");
	}
	ptr->nouse = TRUE;
#ifdef INTERACT
	begin_trans(VET_OPT_TRANS);
	update_locuse(ptr);
	end_trans();
#endif	 	 
      }
      else if (ptr->arg && ptr->reps == 1 && !regarg(ptr)){
	ptr->nouse = TRUE;
#ifdef INTERACT
	 begin_trans(VET_OPT_TRANS);
	 update_locuse(ptr);
	 end_trans();
#endif	 	 
      }		 
   }

   /*
    * Generate an interference graph for the current function.  Sort the nodes
    * in the graph so that the most frequently used items are considered for
    * allocation first.
    */
   build_interference_graph();
   sort_interference_graph();
   /*new_sort_ig();*/

   /*
    * Perform global register mapping using a map-coloring style algorithm
    * which is driven by access frequency estimation information.
    */
   allocate = regnew();
   iptr = igptr;
   while (iptr) {

      /*
       * Our first attempt at coloring is to consider the whole life of a
       * specific item together at once.  To do so, we consider all of the
       * "blobs" of life that are related together at once.  If we cannot do
       * so, then we mark such "blobs" as CONSIDERED and move on to other
       * possible strategies.
       */
      if (!(iptr->status & (COLORED | CONSIDERED))) {

         /*
          * If the local has been identified as either one that should not be
          * allocated to a register (because the code takes the address of it)
          * or if it was declared as a volatile item (in ANSI C), then skip it.
          */
         ptr = iptr->item->local;
         if (ptr->nouse || ptr->vol) {
            iptr = iptr->next;
            continue;
         }
#ifdef INTERACT
         if (limited_scope && swq) {
            for (sptr = iptr; sptr->nsib != iptr; sptr = sptr->nsib)
               if (sptr->ptr->inscope)
                  break;
            if (!sptr->ptr->inscope) {
               iptr = iptr->next;
               continue;
            }
         }
#endif
         /*
          * Look through all of the "blobs" of life that comprise this life of
          * a local to determine which registers it conflicts with, what type
          * it is referenced as, how often it is referenced, and if it extends
          * to the entry point of the function.
          */
         regcopy(allocate, iptr->regs);
         type = iptr->type;
         freq = iptr->lv_freq;
         for (sptr = iptr->nsib; sptr != iptr; sptr = sptr->nsib) {

            /*
             * This macro is created by regtool in order to include in the
             * "allocate" vector all of the items that conflict with this
             * "blob" of life (as given by the "sptr->regs" item).  On most
             * machines, this macro will emit the following code:
             *
             * *allocate |= *sptr->regs;
             *
             * but on machines where there are more allocable registers than
             * there are bits in an "int", there will be additional lines to
             * move the whole register vector as quickly as possible.
             */
            REGINC(allocate, sptr->regs);

            /*
             * Here we ensure that the item was referenced consistently with
             * respect to its type.  If the same local is referenced as an
             * "int" in one place and as a "double" in another, then there is
             * some obvious confusion as to which type of register should be
             * allocated for it, so we mark it as a local that should not be
             * allocated.
             */
            if (sptr->type != NOTYPE) {
               if ((type = checktype(type, sptr->type)) == NOTYPE) {
                  ptr->nouse = TRUE;
#ifdef INTERACT
                  begin_trans(VET_OPT_TRANS);
                  update_locuse(ptr);
                  end_trans();
#endif
                  break;
               }
            }

            /*
             * Add the number of times that the item is referenced in this
             * "blob" of life to its overall reference count.
             */
            freq += sptr->lv_freq;
         }

         /*
          * Determine if the local is used often enough to justify allocating
          * it to a register.  If it is used infrequently, it might be okay to
          * allocate it to a scratch register, if one is available, since these
          * registers are "cheap" to use relative to non-scratch ones.
          */
         if (ptr->nouse || !freq) {
            iptr->status |= CONSIDERED;
            for (sptr = iptr->nsib; sptr != iptr; sptr = sptr->nsib)
               sptr->status |= CONSIDERED;
            iptr = iptr->next;
            continue;
         }
         if (!ptr->arg && freq < minlocrefs)
            ifcheap = TRUE;
         else if (ptr->arg && freq == 1 && !regarg(ptr)) {
            iptr->status |= CONSIDERED;
            for (sptr = iptr->nsib; sptr != iptr; sptr = sptr->nsib)
               sptr->status |= CONSIDERED;
            iptr = iptr->next;
            continue;
         }
         else if (ptr->arg && freq < minargrefs)
            ifcheap = TRUE;
         else
            ifcheap = FALSE;

         /*
          * Attempt to allocate a register for this life of the local.  If one
          * is found, then update the RTLs of the function to change the local
          * variable reference to appropriate register references.
          */
         if (cancolor(type, allocate, reg, ifcheap)) {
#ifdef VET
            begin_trans(VET_OPT_TRANS);
#endif
            sptr = iptr;
            entry = !ptr->arg;
            do {

               /*
                * If this item of life extends to the entry point of the
                * function, then set "entry" to TRUE and place the register
                * that was allocated to this local in the "regin" field of the
                * local variable information structure.
                */
               if (!entry && sptr->ptr == top) {
                  for (eptr = top->conin; eptr; eptr = eptr->conin) {
                     if (eptr == sptr) {
                        entry = TRUE;
                        if (regarg(ptr)) {

                           (void) strcpy(ptr->regin, reg);
                           ptr->atype = type;
                           sptr->status &= ~MUSTLOAD;
#ifdef INTERACT
                           update_locuse(ptr);
#endif
                        }
                        break;
                     }
                  }
               }

               /*
                * Mark this local item as "colored" and the register that is
                * allocated to it as unavailable for allocation to any of the
                * other locals whose life intersect this one.
                */
               sptr->status |= COLORED;
               for (cptr = sptr->conflicts; cptr; cptr = cptr->next)
                  touchreg(cptr->ptr->regs, reg);

               /*
                * Change the local variable references in the function to refer
                * to the appropriate register instead.  Note that in some
                * instances we must load and spill the value of the local when
                * indicated by the interference graph.
                */
               if (rtlptr = sptr->start) {
                  for (;;) {
                     if (cnts(rtlptr) && vin(ptr->olist, rtlptr->onum))
                        loctoreg(ptr, rtlptr, type, reg);
                     if (rtlptr == sptr->end)
                        break;
                     if (!(rtlptr = rtlptr->next))
                        break;
                  }
                  if (sptr->status & MUSTLOAD)
                     load(type, ptr->arg, ptr->code, sptr->ptr, sptr->start,
                          reg, TRUE);
               }
               else if (sptr->status & MUSTLOAD) {
                  rtlptr = sptr->ptr->rtlend;
                  if (rtlptr && rtlptr->val && is_rjmp(rtlptr->val)) {
                     if (rtlptr->prev)
                        rtlptr = rtlptr->prev;
                  }
                  else if (!rtlptr || !rtlptr->val || !is_pc(rtlptr->val))
                     rtlptr = (struct list *) NULL;
                  load(type, ptr->arg, ptr->code, sptr->ptr, rtlptr, reg,
                       TRUE);
               }

               /*
                * Continue on to the next "blob" that makes up part of this
                * local's life.
                */
               sptr = sptr->nsib;
            } while (sptr != iptr);

            process_work_list();


            /*
             * Note that changes have been made to the function.  Also update
             * the local variable usage count and set the "reg" field when no
             * references remain to the local in order for fixentry() to know
             * that it should not allocate any space to hold the local.
             */
            change = TRUE;
            ptr->reps -= freq;
            if (!ptr->reps)
               ptr->reg = ptr->nouse = TRUE;

#ifdef INTERACT
            update_locuse(ptr);
#endif


#ifdef VET
            end_trans();
#endif
            /*
             * Since we are done with this "blob" of life, move on to the next
             */
            iptr = iptr->next;
            continue;
         }

         /*
          * Mark all "blobs" that make up this life as considered, since there
          * is no use in trying to consider them again as a whole. They will
          * and should, however, be considered in other groupings.
          */
         iptr->status |= CONSIDERED;
         for (sptr = iptr->nsib; sptr != iptr; sptr = sptr->nsib)
            sptr->status |= CONSIDERED;
      }
      /*
       * It is  worthwhile to attempt allocation over the life of a loop.  
       * Some locals tend to be used heavily inside just one single loop.  
       * This would allow some items which cannot be allocated over their 
       * whole life, but are used heavily within a single loop to be 
       * allocated at such loops.
       */
      /* Try and allocate registers at the loop level */
      if (!(iptr->status & COLORED)) {

         struct loopl *liteml;
         int   must_save = FALSE, must_load = FALSE;
         struct blist *blklist, *bptr;
         struct bblock *cblk;
         ptr = iptr->item->local;

         /* Go through all loops in the function. Control flow optimisation,
          * called before this phase orders the loops starting from
          * the biggest
          */
         for (liteml = loops; liteml; liteml = liteml->next) {

	   /* Jason found an error caused by misplaced STORES in the
             * switch table. This was found in a loop that spanned a few 
             * switch table entries and some of their targets. So on 
             * promoting a variable to a register in such a loop, a STORE
             * was placed in the loop exit - which happened to be the middle
             * of a jump table.
             * Check if any of the exit blocks of the loop end with
             * an assembly line (i.e.) do not end with a branch. If so
             * we have an exit at a switch table. For now, we will avoid 
             * promoting locals over such loops because the loop exits include
             * all switch cases which are not part of the loop - unlike what
             * is reflected in the control flow structure.
             */
            if (!check_loop_exits(liteml))
               continue;

            /* Determine which loop a sibling block is part of. Then attempt to
             * color for that loop. If not possible, move on to the next loop
             */
            sptr = iptr;
            do {

               /* Check if it is part of any loop */
               if (block_in_loop(liteml, sptr->ptr)) {
                  type = sptr->type;
                  freq = 0;
                  regcopy(allocate, sptr->regs);
                  sptr->loop = liteml;
                  break;
               }
               else {
                  sptr = sptr->nsib;
               }
            }
            while (sptr != iptr);
            /* If not part of this loop, continue for  next loop */
            if (sptr == iptr || (sptr->status & COLORED))
               continue;
            /* liteml is the loop over which we will try to color */
            /*
             * Look through all of the "blobs" of life that comprise this life
             * of a local to determine which registers it conflicts with, what
             * type it is referenced as, how often it is referenced, and if it
             * extends to the entry point of the function.
             */
            do {

               /*-
                * This macro is created by regtool in order to include in the
                * "allocate" vector all of the items that conflict with this
                * "blob" of life (as given by the "sptr->regs" item).  On most
                * machines, this macro will emit the following code:
                * 
                * *allocate |= *sptr->regs;
                * 
                * but on machines where there are more allocable registers than
                * there are bits in an "int", there will be additional lines to
                * move the whole register vector as quickly as possible.
                */
               if ((sptr->loop == liteml)
                   || block_in_loop(liteml, sptr->ptr)) {
                  sptr->loop = liteml;
                  REGINC(allocate, sptr->regs);
                  /* Its possible that this block is colored by a previous
                   * invocation.  So move on to next loop.
                   */
                  if (sptr->status & COLORED) {
                     freq = 0;
                     break;
                  }
                  /*
                   * Here we ensure that the item was referenced consistently
                   * with respect to its type.  If the same local is referenced
                   * as an "int" in one place and as a "double" in another,
                   * then there is some obvious confusion as to which type of
                   * register should be allocated for it, so we mark it as a
                   * local that should not be allocated.
                   */
                  if (sptr->type != NOTYPE) {
                     if ((type = checktype(type, sptr->type)) == NOTYPE) {
                        ptr->nouse = TRUE;
                        break;
                     }
                  }

                  /*
                   * Add the number of times that the item is referenced in
                   * this "blob" of life to its overall reference count.
                   */

                  freq += sptr->lv_freq;
                  if (sptr->status & ISSET)
                     must_save = TRUE;
                  if (sptr->status & ISUSED)
                     must_load = TRUE;
               }
               sptr = sptr->nsib;
            }
            while (sptr != iptr);

            if (ptr->nouse || !freq) {
               continue;
            }

            /*
             * Determine if the local is used often enough to justify
             * allocating it to a register.  If it is used infrequently, it
             * might be okay to allocate it to a scratch register, if one is
             * available, since these registers are "cheap" to use relative
             * to non-scratch ones.
             */
            if (!ptr->arg && freq < minlocrefs)
               ifcheap = TRUE;
            else if (ptr->arg && freq == 1 && !regarg(ptr)) {
               continue;
            }
            else if (ptr->arg && freq < minargrefs)
               ifcheap = TRUE;
            else
               ifcheap = FALSE;

            /*
             * Attempt to allocate a register for this life of the local.  If
             * one is found, then update the RTLs of the function to change the
             * local variable reference to appropriate register references.
             */
            if (cancolor(type, allocate, reg, ifcheap)) {
#ifdef VET
               begin_trans(VET_OPT_TRANS);
#endif
               sptr = iptr;
               entry = !ptr->arg;
               /* A local has to be loaded in the loop preheader, whether
                * it has been read or written to in the loop. This is because, 
                * if the local is written to only in parts of the loop, correct
                * value should be written to it on exit. This should be the
                * value it had prior to loop entry, if it was not written
                * inside the loop.
                */
               if (must_load || must_save) {
                  cblk = preheader(liteml->ptr);
                  if (cnts(cblk->rtlend) && is_pc(cblk->rtlend->val))
                     load(type, ptr->arg, ptr->code, cblk, cblk->rtlend, reg,
                          FALSE);
                  else  /* add it as the last RTL */
                     load(type, ptr->arg, ptr->code, cblk, NULL, reg, FALSE);
               }


               do {
                  if (sptr->loop != liteml) {
                     sptr = sptr->nsib;
                     continue;
                  }


                  /*
                   * Mark this local item as "colored" and the register that is
                   * allocated to it as unavailable for allocation to any of
                   * the other locals whose life intersect this one.
                   */
                  sptr->status |= COLORED;
                  for (cptr = sptr->conflicts; cptr; cptr = cptr->next)
                     touchreg(cptr->ptr->regs, reg);

                  /*
                   * Change the local variable references in the function to
                   * refer to the appropriate register instead.  Note that in
                   * some instances we must load and spill the value of the
                   * local when indicated by the interference graph.
                   */
                  /* Checking for stores and loads as done in next case */
                  if (rtlptr = sptr->start) {
                     for (;;) {
                        if (cnts(rtlptr) && vin(ptr->olist, rtlptr->onum))
                           loctoreg(ptr, rtlptr, type, reg);
                        if (rtlptr == sptr->end)
                           break;
                        if (!(rtlptr = rtlptr->next))
                           break;
                     }
                     if (sptr->status & MUSTLOAD)
                        load(type, ptr->arg, ptr->code, liteml->preheader,
                             sptr->start, reg, TRUE);
                  }
                  else if (sptr->status & MUSTLOAD) {
                     rtlptr = sptr->ptr->rtlend;
                     if (rtlptr && rtlptr->val && is_rjmp(rtlptr->val)) {
                        if (rtlptr->prev)
                           rtlptr = rtlptr->prev;
                     }
                     else if (!rtlptr || !rtlptr->val || !is_pc(rtlptr->val))
                        rtlptr = (struct list *) NULL;
                     load(type, ptr->arg, ptr->code, liteml->preheader,
                          rtlptr, reg, TRUE);
                  }

                  /*
                   * Continue on to the next "blob" that makes up part of this
                   * local's life.
                   */
                  sptr = sptr->nsib;
               } while (sptr != iptr);

               if (must_save) {

                  /* If there is only one predecessor to this exit block, we 
                   * know its a loop block. Just insert the save at the 
                   * beginning of this exit block.
                   */

                  for (blklist = liteml->exits; blklist;
                       blklist = blklist->next) {

                     /* check if this  exit block has more than one pred */
                     if (!blklist->ptr->preds->next) {
                        store(type, ptr->arg, ptr->code, blklist->ptr,
                              blklist->ptr->rtls, reg, FALSE);
                     }
                     else {
                        /* Now more than one predecessor to this exit block.
                         * If all predecessors are from the loop, do the same
                         * as above.
                         */
                        for (bptr = blklist->ptr->preds;
                             bptr && block_in_loop(liteml, bptr->ptr);
                             bptr = bptr->next) ;

                        if (!bptr) {
                           store(type, ptr->arg, ptr->code, blklist->ptr,
                                 blklist->ptr->rtls, reg, FALSE);
                        }
                        else {
                           /* We have to create a new exit block. */
                           cblk = loop_exit(liteml, blklist->ptr);
                           store(type, ptr->arg, ptr->code, cblk, NULL, reg,
                                 FALSE);
                        }
                     }
                  }
               }

               process_work_list();
#ifdef VET
               end_trans();
#endif

               /*
                * Note that changes have been made to the function.  Also
                * update the local variable usage count and set the "reg" field
                * when no references remain to the local in order for
                * fixentry() to know that it should not allocate any space to
                * hold the local.
                */
               change = TRUE;
               ptr->reps -= freq;
               if (!ptr->reps)
                  ptr->reg = ptr->nouse = TRUE;
               break;

            }
         }         /* end of iterating through all loops */

         if (!(iptr->status & COLORED)) {
            /*
             * Mark all "blobs" that make up this life as considered, since
             * there is no use in trying to consider them again as a whole.
             * They will and should, however, be considered in other groupings.
             */
            iptr->status |= CONSIDERED;
            for (sptr = iptr->nsib; sptr != iptr; sptr = sptr->nsib)
               sptr->status |= CONSIDERED;
         }
         else {
            iptr = iptr->next;
            continue;
         }
      }

      /*
       * If no register could be allocated for the whole life of a local, then
       * consider individual life "blobs" to see if it is worthwhile to
       * allocate the local to a register for the duration of a "blob". Note
       * that we will have to consider the fact that a register load and spill
       * might have to take place at the borders of the life.
       */
      if (!(iptr->status & COLORED)) {
         /*
          * Estimate the releative advantage of allocating this "blob" of life
          * to a register.  If the item needs to be loaded or spilled at its
          * life boundaries, then these factors count against it.
          */
         freq = iptr->lv_freq;

         /* If this is a loop we may be able to hoist the load/store, so
          * don't penalize so much.
          */
         if (iptr->status & ISSET)
            freq -= (iptr->ptr->freq << 4) + 1;
         if (iptr->status & ISUSED)
            freq -= (iptr->ptr->freq << 4) + 1;

         /*
          * If the item would cost most to load and spill that it would save on
          * references or it is only used in a single RTL, then there is no
          * advantage in allocating it to a register.
          */

         if (freq <= 0 || iptr->start == iptr->end) {
            iptr = iptr->next;
            continue;
         }

         /*
          * Ensure that the item is used enough to warrant its allocation.
          * Also, determine if the register used must be cheap to make the
          * allocation worthwhile.
          */
         ptr = iptr->item->local;
         if (!ptr->arg && freq < minlocrefs)
            ifcheap = TRUE;
         else if (ptr->arg && freq == 1 && !regarg(ptr)) {
            iptr = iptr->next;
            continue;
         }
         else if (ptr->arg && freq < minargrefs)
            ifcheap = TRUE;
         else
            ifcheap = FALSE;

         /*
          * Attempt to locate a suitable register to contain this "blob" of
          * life.
          */
         if (cancolor(iptr->type, iptr->regs, reg, ifcheap)) {
#ifdef VET
            begin_trans(VET_OPT_TRANS);
#endif

            /*
             * Determine if this life "blob" reaches the entry point of the
             * function.  If it does, we set the "regin" field to indicate
             * which register contains the argument at the entry point of the
             * function.  Also, if the local is passed to this function in a
             * register, we abstain from placing a load of it at the entry
             * point of the function, since fixentry() will ensure that the
             * appropriate value winds up in the register given by the "regin"
             * field for this local.
             */
            if (iptr->ptr == top) {
               for (eptr = top->conin; eptr; eptr = eptr->conin) {
                  if (eptr == iptr) {
                     if (ptr->arg && regarg(ptr)) {
                        (void) strcpy(ptr->regin, reg);
                        ptr->atype = iptr->type;
#ifdef INTERACT
                        update_locuse(ptr);
#endif
                     }
                     if (!ptr->arg || regarg(ptr)) {
                        iptr->status &= ~ISUSED;
                        if (ptr->arg)
                           iptr->status |= ISSET;
                     }
                     break;
                  }
               }
            }

            /*
             * Indicate that this "blob" has been colored and make the register
             * used conflict with any "blobs" that are live simultaneously with
             * this one.
             */
            iptr->status |= COLORED;
            for (cptr = iptr->conflicts; cptr; cptr = cptr->next)
               touchreg(cptr->ptr->regs, reg);

            /*
             * If the local is used (referenced) in this "blob" of life, then
             * we must first load it before referencing it.
             */
            rtlptr = iptr->start;
            if (iptr->status & ISUSED)
               load(iptr->type, ptr->arg, ptr->code, iptr->ptr, rtlptr, reg,
                    FALSE);

            /*
             * Change all of the RTLs that reference this local in the "blob"
             * of life to appropriate register references.
             */
            for (;;) {
               if (cnts(rtlptr) && vin(ptr->olist, rtlptr->onum))
                  loctoreg(ptr, rtlptr, iptr->type, reg);
               if (rtlptr == iptr->end)
                  break;
               rtlptr = rtlptr->next;
            }

            /*
             * If the local is set (modified) within this "blob" of life, then
             * we must update its value before exiting the "blob". We must be
             * sure not to place the store in a location that destroys a
             * condition code value.
             */
            if (cnts(rtlptr) && !is_pc(rtlptr->val))
               rtlptr = rtlptr->next;
            if (iptr->status & ISSET) {
               if (rtlptr && rtlptr->val && is_rjmp(rtlptr->val)
                   && sets_only_cc(rtlptr->prev))
                  rtlptr = rtlptr->prev;
               store(iptr->type, ptr->arg, ptr->code, iptr->ptr, rtlptr, reg,
                     FALSE);
            }
            process_work_list();


            /*
             * Note the changes that have taken place because of this
             * allocation.
             */
            change = TRUE;
            ptr->reps -= freq;
            if (!ptr->reps)
               ptr->reg = ptr->nouse = TRUE;

#ifdef INTERACT
            update_locuse(ptr);
#endif

#ifdef VET
            end_trans();
#endif
         }

         /*
          * Having considered this "blob" singly, move on to the next one
          */
         iptr = iptr->next;
      }

      /*
       * This item has already been colored, so we skip on to the next one
       */
      else
         iptr = iptr->next;
   }
   mrfree(allocate);

   /*
    * Release the space allocated to hold the interference graph
    */
   clear_interference_graph();

#ifdef VET
#ifdef INTERACT
   end_phase(REG_ALLOCATION_TRANS);
#else
   end_phase(REGALLOC_PHASE);
#endif
#endif
   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_after_color_global_allocation", invocation))
      dump_function_i("after color_global_allocation()", invocation,
                      SHOW_LABELS | SHOW_USE | SHOW_NONRTLS | SHOW_SE);

   /*
    * Return indicating if any changes have been made to the function
    */
   return change;
}

/*
 * block_in_list - returns TRUE if a given basic block is in a list of
 *                 basic blocks
 */

int block_in_list(struct bblock *cblk, struct blist *blklist)
{
   for (; blklist; blklist = blklist->next) {
      if (blklist->ptr == cblk)
         return TRUE;
   }
   return FALSE;
}


/*
 * loctoreg - replace local reference with a register
 */
void loctoreg(struct locuse *ptr, struct list *rtl, int type, char *reg)
{
   register struct list *lptr;

   if (!(rtl->type & ARGLINE) && (lptr = locsub(ptr, rtl, type, reg))) {
      combmark(lptr);
      lptr->cost = 0;
      lptr->kind = 0;
      lptr->type &= ~SIDEFFECT;
      if (lptr->se) {
         FREE(lptr->se);
         lptr->se = (char *) NULL;
      }
#ifdef INTERACT
      mod_kind(lptr);
      mod_type(lptr);
#endif
#ifdef VET
      /* Type has changed and possibly the side effect was cleared. */
#ifndef INTERACT
      mod_side(lptr);
#endif
#endif
   }
}

/*
 * locals_used - look for local variable references in the function
 */
void locals_used(int sort)
{
   struct locuse *ptr;
   int   i;
   struct list *rtl;
   struct bblock *cblk;
   int   type, ind, set;
   char *c, code[REGSTRING];
#if DEBUG_VECTOR_FREE_LIST
   /* For debugging a particular function's compilation. */
   int   debugflag = (fn && (0 == strcmp(fn, "gl_flush_pb")));
#endif

   /*
    * Look through every basic block and every RTL in the function for local
    * register references.  Update the local information structure as each
    * reference is found to indicate where each local is used.
    */
   for (cblk = top; cblk; cblk = cblk->down) {
      for (rtl = cblk->rtls; rtl; rtl = rtl->next) {
         if (c = cnts(rtl)) {
            while (locget(&c, code, &type, &ind, &set, rtl)) {
#if DEBUG_VECTOR_FREE_LIST
               if (debugflag)
                  validate_vector_free_list();
#endif
               i = vecmap[USP(code)];
               if (i < 0) {
                  (void) fprintf(stderr,
                            "vpo (locals_used): undeclared local reference\n");
                  exit(EXIT_FAILURE);
               }
               ptr = sets[i].local;
               assert(NULL != ptr);
               vinsert(&ptr->olist, rtl->onum);
               if (!ptr->nouse) {
                  if (set)
                     dinsert(&ptr->setin, cblk->bnum);
                  else
                     dinsert(&ptr->usedin, cblk->bnum);
               }
               if (ind)
                  ptr->nouse = ptr->indref = TRUE;
               else if ((ptr->atype = checktype(ptr->atype, type)) == NOTYPE)
                  ptr->nouse = TRUE;
               ptr->reps += (cblk->freq << 4) + 1;
               if (sort)
                  reorder_local(ptr);
#ifdef INTERACT
               update_locuse(ptr);
#endif
            }
         }
      }
   }
}

/*
 * reorder_local - place local in order according to its "reps" field
 */
void reorder_local(struct locuse *ptr)
{
   struct locuse *prev;

   /*
    * Order this local in the coloring graph list according to the estimate of
    * the number of times that it is referenced.
    */
   while ((prev = ptr->prev) && prev->reps < ptr->reps) {
      if (ptr->next)
         ptr->next->prev = prev;
      else {       /* ptr pointed to last local. */
         assert(loce == ptr);
         loce = prev;
      }
      prev->next = ptr->next;
      if (prev->prev)
         prev->prev->next = ptr;
      else {       /* prev pointed to the first local. */
         assert(locs == prev);
         locs = ptr;
      }
      ptr->next = prev;
      ptr->prev = prev->prev;
      prev->prev = ptr;
   }
}


/*
 * swap_ignode - check for conditions when we need to swap the nodes of the
 *               interference graph
 */
int swap_ignode(struct igsort *sptr, struct igsort *nptr)
{
  /*if a node has no references, then it should be percolated down the tree*/
  /*if(sptr->ptr->start == NULL && nptr->ptr->start != NULL)
    return TRUE;*/
  
  if(sptr->freq < nptr->freq)
    return TRUE;

  if(sptr->freq == nptr->freq){
    if(strcmp(sptr->ptr->item->local->item, nptr->ptr->item->local->item) < 0)
      return TRUE;
  }

  return FALSE;
}

void remove_node(struct ignode *ptr)
{
  if(ptr->prev)
    ptr->prev->next = ptr->next;
  else
    igptr=ptr->next;
  if(ptr->next)
    ptr->next->prev = ptr->prev;
}

void insert_node(struct ignode *ptr, struct ignode *after)
{
  struct ignode *next;

  if(after){
    next = after->next;
    after->next = ptr;
    ptr->prev = after;
    if(next)
      next->prev = ptr;
    ptr->next = next;
  }
  else{
    ptr->next = igptr;
    igptr->prev = ptr;
    igptr = ptr;
  }
}


/*uses a simple insertion sort algorithm*/
void new_sort_ig()
{
  struct ignode *ptr, *ptr2, *prev, *prev2, *next, *nptr, *large;

  if(!igptr)
    return;
  
  for(ptr=igptr ; ptr->next ; ptr=nptr){
    nptr = ptr->next;
    large = ptr;

    /*find the largest node*/
    for(ptr2=ptr->next ; ptr2 ; ptr2=ptr2->next){

      if(large->start==NULL && ptr2->start!=NULL)
	large = ptr2;
      else if(large->freq < ptr2->freq)
	large = ptr2;
      else if(large->freq == ptr2->freq && 
	      strcmp(large->item->local->item, ptr2->item->local->item) < 0)
	large = ptr2;
    }

    /*swap the two nodes*/
    if(large != ptr){
      prev = ptr->prev;
      prev2 = large->prev;

      remove_node(ptr);
      remove_node(large);

      insert_node(large, prev);
      insert_node(ptr, prev2);
    }
  }
}


/*
 * sort_interference_graph - sort the interference graph nodes by "freq"
 */
void sort_interference_graph()
{
   register struct igsort *sptr, *nptr, *rptr;
   register int find;
   struct igsort (*igptrs)[];
   struct igsort tmp;
   struct ignode *ptr;
   int   heap;

   /*
    * Sort the interference graph according to the value of the "freq" field.
    * This function utilizes a heap sort because its fast and I happen to like
    * heap sorts.
    */

   /*
    * Check to be sure that there is something to sort
    */
   if (!igptr)
      return;

   /*
    * Allocate a heap of the appropriate size
    */
   igptrs = (struct igsort(*)[]) SALLOC((int) ignums * sizeof(struct igsort));

   heap = 0;
   for (ptr = igptr; ptr; ptr = ptr->next) {

      /*
       * Insert the item into the next slot in the heap
       */
      sptr = &(*igptrs)[++heap];
      sptr->freq = ptr->freq;
      sptr->ptr = ptr;

      /*
       * Percolate the item up to the point where its parent has a value at
       * least as large as the item does.
       */
      for (find = heap >> 1; find; find >>= 1) {
         nptr = &(*igptrs)[find];
	 /*if(sptr->freq > nptr->freq){*/
	 if(!swap_ignode(sptr, nptr)){
	   tmp = *nptr;
	   *nptr = *sptr;
	   *sptr = tmp;
	   sptr = nptr;
         }
         else
	   break;
      }
   }

   /*
    * The top item in the heap is has the largest value
    */
   sptr = &(*igptrs)[1];
   igptr = iglast = sptr->ptr;

   /*
    * Move the last item in the heap to the top and percolate it down to allow
    * the next highest item to move up to the top.  This is what the heap sort
    * is all about.
    */
   while (heap > 1) {
      *sptr = (*igptrs)[heap--];
      find = 2;
      while (find <= heap) {
         nptr = &(*igptrs)[find];
         if (find == heap) {
	   /*if (sptr->freq < nptr->freq) {*/
	   if(swap_ignode(sptr, nptr)){
	       tmp = *nptr;
               *nptr = *sptr;
               *sptr = tmp;
               sptr = nptr;
            }
            break;
         }
         else {
            rptr = &(*igptrs)[find + 1];
            /*if (rptr->freq > nptr->freq) {*/
	      /*if (sptr->freq < rptr->freq) {*/
	    if(!swap_ignode(rptr, nptr)){
	      if(swap_ignode(sptr, rptr)){
                  tmp = *rptr;
                  *rptr = *sptr;
                  *sptr = tmp;
                  sptr = rptr;
                  find++;
               }
               else
                  break;
            }
            else {
	      /*if (sptr->freq < nptr->freq) {*/
	      if(swap_ignode(sptr, nptr)){
                  tmp = *nptr;
                  *nptr = *sptr;
                  *sptr = tmp;
                  sptr = nptr;
               }
               else
                  break;
            }
         }
         find <<= 1;
      }

      /*
       * The new top is now the next item to go on the list.
       */
      sptr = &(*igptrs)[1];
      iglast->next = sptr->ptr;
      iglast = sptr->ptr;
   }

   /*
    * Mark the end of the list
    */
   iglast->next = (struct ignode *) NULL;

   /*
    * Return the heap
    */
   FREE(igptrs);
}

/*
 * sets_only_cc - check RTL to ensure that is only sets condition codes
 */
int sets_only_cc(struct list *ptr)
{
   register char *rtl, c;

   if (rtl = cnts(ptr)) {
      while (c = *rtl++) {
         if (c == '=' && !is_cc(rtl - 3))
            return FALSE;
      }
      return TRUE;
   }
   return FALSE;
}

/*
 * count_caller_save - count the number of callersave points in the function
 */
int count_caller_save()
{
   register struct list *rtl;
   register char *c;
   register struct bblock *cblk;
   register int index;
   int   rcont, rt, calls;

   /*
    * Clear out the existing caller save items found for each register
    */
   for (index = 0; index < MAXREGS; csnum[index++] = 0) ;

   /*
    * Search through the function for caller save points and note which
    * registers are affected.
    */
   calls = FALSE;
   for (cblk = top; cblk; cblk = cblk->down) {
      for (rtl = cblk->rtls; rtl; rtl = rtl->next) {
         if (rtl->type & CSLINE) {
            for (c = rtl->val; *c; c = skiploc(c)) {
               rt = regtype(c);
               index = regtoindex(rt, regnum(c));
               for (rcont = contreg(rt); rcont--; index++)
                  csnum[index] += (cblk->freq << 5) + 2;
            }
            calls = TRUE;
         }
      }
   }

   /*
    * Indicate if any caller save lines are present in the function
    */
   return calls;
}

/*
 * callersave - return the number of times register must be saved and restored
 */
int callersave(char *reg)
{

   register int index, max, rcont;
   int   rt;

   max = 0;
   rt = regtype(reg);
   index = regtoindex(rt, regnum(reg));
   for (rcont = contreg(rt); rcont--; index++) {
      if (csnum[index] > max)
         max = csnum[index];
   }
   return max;
}

/*
 * move_caller_save - locate the best place to save locals around calls
 */
void move_caller_save(int invocation)
{
   register struct list *rtl;
   register int fptr, iptr;
   struct bblock *cblk, *hblk;
   struct list *ptr;
   struct locinfo item[MAXREGS];
   struct blist *eptr;
   struct locuse *lptr;
   struct loopl *loop;

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_move_caller_save", invocation))
      dump_function_i("before move_caller_save()", invocation,
                      SHOW_LABELS | SHOW_SUCCS | SHOW_PREDS | SHOW_LOOP |
                      SHOW_NONRTLS);

   /*
    * The purpose of this routine is to locate the best place to save and
    * restore registers that need to be saved by the caller around function
    * calls.  We only deal with registers that are allocated by the simple
    * global register allocator in this function.  First, we start out by
    * locating the callersave points in the function.
    */
   for (cblk = top; cblk; cblk = cblk->down) {
      for (rtl = cblk->rtls; rtl; rtl = rtl->next) {
         if (rtl->type & CSLINE) {

            /*
             * Save a pointer to the local information record of every local
             * that we must insert save and restore information for at this
             * call site. If there are no locals that need to be saved and
             * restored here, proceed to the next site.
             */
            iptr = 0;
            for (lptr = locs; lptr; lptr = lptr->next) {
               if (*lptr->csreg && fmatch(rtl->val, lptr->csreg)) {
                  item[iptr].ptr = lptr;
                  item[iptr].st = FALSE;
                  item[iptr++].ld = FALSE;
               }
            }
            if (!iptr)
               continue;

            /*
             * Now examine the control-flow information to determine the outer-
             * most loop in which this call site is in.  We will then move on
             * into inner loops unless we manage to save and restore all
             * caller- save items.  The reason we want to do this is because it
             * pays off to move the save and restores outside of loops if
             * possible, since these points will usually be executed less
             * often.  We can only do this, however, if the item we have to
             * save and restore is not set or referenced inside the loop.
             */
            for (loop = loope; iptr && loop; loop = loop->prev) {
               if (block_in_loop(loop, cblk)) {

                  /*
                   * Now we examine all of the items that still need to be
                   * stored. If the item is not set inside the current loop,
                   * then we can store it just once at the preheader of the
                   * loop.
                   */
                  for (fptr = 0; fptr < iptr; fptr++) {
                     if (!
                         (item[fptr].st
                          || any_block_in_loop(loop,
                                               item[fptr].ptr->setin))) {

                        /*
                         * Make a loop pre-header, if one doesn't already exist
                         */
                        if (!(hblk = loop->preheader))
                           hblk = loop->preheader = preheader(loop->ptr);

                        /*
                         * Find a suitable place to place a store operation at
                         * the end of the pre-header block.
                         */
                        if (!
                            ((ptr = hblk->rtlend) && cnts(ptr)
                             && is_pc(ptr->val)))
                           ptr = (struct list *) NULL;

                        /*
                         * Insert the store of the local and update the
                         * information to reflect the fact that the store has
                         * been placed.  Note that we check to see if the item
                         * has already been stored in the pre-header of this
                         * loop and, if it has, we avoid placing another store
                         * at this point.
                         */
                        lptr = item[fptr].ptr;
                        if (!rexists(lptr->code, loop->stored)) {
                           store(lptr->atype, lptr->arg, lptr->code, hblk,
                                 ptr, lptr->csreg, TRUE);
                           rinsert(lptr->code, &loop->stored);
                        }
                        item[fptr].st = TRUE;

                        /*
                         * If the item is now both saved and restored, we can
                         * remove it from the list of items still needing
                         * attention.
                         */
                        if (item[fptr].ld) {
                           item[fptr].ptr = item[--iptr].ptr;
                           item[fptr].st = item[iptr].st;
                           item[fptr--].ld = item[iptr].ld;
                        }
                     }

                     /*
                      * Now we examine all of the items that we still need to
                      * load. If there are no sets or uses of the item in this
                      * loop, then we can place the load at the exit points of
                      * the loop.
                      */
                     if (!
                         (item[fptr].ld
                          || any_block_in_loop(loop, item[fptr].ptr->setin)
                          || any_block_in_loop(loop,
                                               item[fptr].ptr->usedin))) {

                        /*
                         * Place a load at each exit point of the loop.  If the
                         * exit points are not adequate for placing stores
                         * (because more than just the exit of this loop flows
                         * through it), then make a suitable exit points.
                         */
                        if (!rexists(lptr->code, loop->loaded)) {
                           for (eptr = loop->exits; eptr; eptr = eptr->next) {
                              hblk = loop_exit(loop, eptr->ptr);
                              lptr = item[fptr].ptr;
                              load(lptr->type, lptr->arg, lptr->code, hblk,
                                   hblk->rtls, lptr->csreg, TRUE);
                           }
                           rinsert(lptr->code, &loop->loaded);
                        }

                        /*
                         * Note the fact that this item has been loaded.  If
                         * the item has been both loaded and stored, then we
                         * can remove it from the list of items that still
                         * require consideration.
                         */
                        item[fptr].ld = TRUE;
                        if (item[fptr].st) {
                           item[fptr].ptr = item[--iptr].ptr;
                           item[fptr].st = item[iptr].st;
                           item[fptr--].ld = item[iptr].ld;
                        }
                     }
                  }
               }
            }

            /*
             * If there are any items left that we were not able to locate a
             * suitable place to save and restore for, we must place the save
             * and restore instructions immediately before and after this
             * callersave site.
             */
            if (iptr) {
               for (ptr = rtl->prev; !(ptr->type & RTLINE); ptr = ptr->prev) ;
               for (fptr = 0; fptr < iptr; fptr++) {
                  if (!item[fptr].st) {
                     lptr = item[fptr].ptr;
                     store(lptr->atype, lptr->arg, lptr->code, cblk, ptr,
                           lptr->csreg, TRUE);
                  }
                  if (!item[fptr].ld) {
                     lptr = item[fptr].ptr;
                     load(lptr->atype, lptr->arg, lptr->code, cblk, rtl,
                          lptr->csreg, TRUE);
                  }
               }
            }
         }
      }
   }

   /*
    * Now mark all items that were taken care of by this fuction so that we
    * don't attempt to save and restore them in a future pass through this
    * function.
    */
   for (lptr = locs; lptr; lptr = lptr->next)
      *lptr->csreg = '\0';

   /*
    * Check to see if the user would like to see the state of the function at
    * this point in time.
    */
   if (swd && debug_opt_i("dump_before_move_caller_save", invocation))
      dump_function_i("before move_caller_save()", invocation,
                      SHOW_LABELS | SHOW_SUCCS | SHOW_PREDS | SHOW_LOOP |
                      SHOW_NONRTLS);
}

/*
 * clean_locals - selectively forget some things about locals
 */
void clean_locals()
{

   register struct locuse *ptr;

#ifdef VET
   begin_trans(VET_REQUIRED_TRANS);
#endif
   for (ptr = locs; ptr; ptr = ptr->next) {
      ptr->reps = 0;
      ptr->indref = FALSE;
      if (!(ptr->nouse = ptr->reg))
         ptr->atype = NOTYPE;
      vfree(ptr->olist);
      ptr->olist = vinit();
      dfree(ptr->setin);
      ptr->setin = dinit();
      dfree(ptr->usedin);
      ptr->usedin = dinit();
      ptr->addr_mode = 0;
   }
#ifdef VET
   end_trans();
#endif
}

/*
 * clear_locals - return local variable information structures to the heap
 */
void clear_locals()
{

#ifndef DISPOSE
   register struct locuse *ptr, *dptr;

   ptr = locs;
   while (ptr) {
      dptr = ptr;
      ptr = ptr->next;
      free(dptr->item);
      vfree(dptr->olist);
      dfree(dptr->setin);
      dfree(dptr->usedin);
      free((char *) dptr);
   }
#endif

   locs = (struct locuse *) NULL;
}

/*
 * CompactGlobalsList - Remove functions from the globals list, leaving only
 * data variables.  Return the number of such variables that are less than
 * a given size.
 */
int CompactGlobalsList(int MaxSize) {
   int GlobalsIndex = 0;
   int BigGlobalsIndex = 0;
   struct globalsym *globptr = globSymStart;
   struct globalsym *saveptr;
   int codeFlag;

   assert(0 < MaxSize);

   while (NULL != globptr) {
      codeFlag = FALSE;
      if (NULL == globptr->sectionBuffer) {
         /* Section was ".text" and was not recorded.*/
         codeFlag = TRUE;
         if (NULL != globptr->prev)
            globptr->prev->next = globptr->next;
         else
            globSymStart = globptr->next;
         if (NULL != globptr->next)
            globptr->next->prev = globptr->prev;
         else
            globSymEnd = globptr->prev;
         saveptr = globptr->next;
         FREE(globptr);
         globptr = saveptr;
      }
      else {
         ++GlobalsIndex;
         if ((globptr->size + globptr->PadBytes) >= MaxSize) {
            ++BigGlobalsIndex;
#if DEBUG_PLACEMENT
            (void) fprintf(stderr, "Big global found: %s size: %d\n",
                           (globals + globptr->gallocIndex), 
                           (globptr->size + globptr->PadBytes));
#endif
         }
      }
      if (!codeFlag)
         globptr = globptr->next;
   }

   return GlobalsIndex;
} /* end of CompactGlobalsList() */


/*
 * BuildGlobalsArray - build the Globals[] array of globalsym pointers from
 * the linked list of global symbols.
 * Assumes that Globals[] has already been allocated to the needed size.
 * Returns the number of globals processed.
 */
int BuildGlobalsArray(int MaxSize) {
   int GlobalsIndex = 0;
   struct globalsym *globptr;

   assert(0 < MaxSize);  /* NOTE: MaxSize is deprecated. */
   assert(NULL != Globals);

   for (globptr = globSymStart; globptr; globptr = globptr->next) {
      Globals[GlobalsIndex++] = globptr;
   }
   return GlobalsIndex;
}

/*
 * GetGlobalsIndex - Return the subscript into the Globals[] array that
 * matches a given globalsym pointer.
 */
int GetGlobalsIndex(struct globalsym *gptr, struct globalsym *Globals[], int n)
{
   int GlobalsIndex;

   assert(NULL != Globals);

   for (GlobalsIndex = 0; GlobalsIndex < n; ++GlobalsIndex)
      if (gptr == Globals[GlobalsIndex])
         return GlobalsIndex;

   return (-1); /* Failure */
}

/* addGlobalSym - add a new entry to the global symbol table with the
 * given code and index into the globals name array.
 */
struct globalsym *addGlobalSym(char *code, int nameIndex) {
   assert(NULL != code);
   assert(0 <= nameIndex);

   if (NULL == globSymStart) { /* First global. */
      globSymStart = (struct globalsym *)
         SALLOC_PERM(sizeof(struct globalsym));
      if (NULL == globSymStart) {
         (void) fprintf(stderr, "vpo::addGlobalSym: out of virtual memory.\n");
         exit(EXIT_FAILURE);
      }
      globSymEnd = globSymStart;
      (void) strcpy(globSymStart->code, code);
      globSymStart->gallocIndex = nameIndex;
      globSymStart->size = 0;
      globSymStart->PadBytes = 0;
      globSymStart->SectionNumber = -1;
      globSymStart->PlacementPartition = 0;
      globSymStart->next = NULL;
      globSymStart->prev = NULL;
      globSymStart->CoalescedMSNeighbor = NULL;
      globSymStart->CoalescedLSNeighbor = NULL;
      globSymStart->FirstLine = NULL;
      globSymStart->LastLine = NULL;
      globSymStart->sectionBuffer = NULL;
   }
   else { /* Not first global; add at the end of the list. */
      assert(NULL != globSymEnd);
      assert(NULL == globSymEnd->next);
      globSymEnd->next = (struct globalsym *)
         SALLOC_PERM(sizeof(struct globalsym));
      if (NULL == globSymEnd->next) {
         (void) fprintf(stderr, "vpo::addGlobalSym: out of virtual memory.\n");
         exit(EXIT_FAILURE);
      }
      globSymEnd->next->prev = globSymEnd;
      globSymEnd = globSymEnd->next;
      (void) strcpy(globSymEnd->code, code);
      globSymEnd->gallocIndex = nameIndex;
      globSymEnd->size = 0;
      globSymEnd->PadBytes = 0;
      globSymEnd->SectionNumber = -1;
      globSymEnd->PlacementPartition = 0;
      globSymEnd->next = NULL;
      globSymEnd->CoalescedMSNeighbor = NULL;
      globSymEnd->CoalescedLSNeighbor = NULL;
      globSymEnd->FirstLine = NULL;
      globSymEnd->LastLine = NULL;
      globSymEnd->sectionBuffer = NULL;
   }
   return globSymEnd;
} /* end of addGlobalSym() */

/* findGlobalFromAffinityIndex - Given an affinity index number, find the
 * corresponding entry in the global symbol table. Useful in debugger.
 */
struct globalsym *findGlobalFromAffinityIndex(int index) {
   struct globalsym *globalPtr = globSymStart;
   char *globName;

   assert(0 <= index);
   while (NULL != globalPtr) {
      if (index == globalPtr->AffinityIndex)
         return globalPtr;
      else
         globalPtr = globalPtr->next;
   }
   return NULL;
} /* end of findGlobalFromAffinityIndex() */

/* findGlobalFromName - Given a name, find the corresponding entry in the
 * global symbol table.
 */
struct globalsym *findGlobalFromName(char *name) {
   struct globalsym *globalPtr = globSymStart;
   char *globName;

   assert(NULL != name);
   while (NULL != globalPtr) {
      globName = globals + globalPtr->gallocIndex;
      if (0 == strcmp(globName, name))
         return globalPtr;
      else
         globalPtr = globalPtr->next;
   }
   return NULL;
} /* end of findGlobalFromName() */

/*
 * findGlobalFromCode - Given a code, find the corresponding entry in the
 * global symbol table.
 */
struct globalsym *findGlobalFromCode(char *code) {
   char name[MAXSTRING];
   int hash;
   int vecIndex;

   assert(NULL != code);

   hash = USP(code);
   vecIndex = vecmap[hash];
   assert(0 <= vecIndex);
   return findGlobalFromName(globals + vecIndex);
} /* end of findGlobalFromCode() */

/* emitAsmForGlobals - go through the global symbol table linked list and
 * emit all the ASMLINEs stored for each global.
 */
void emitAsmForGlobals(int alignment) {
   struct globalsym *CurrentGlobal = globSymStart;
   struct globalAsmLines *CurrentLine;
   char *printable;
   char oldSection[MAXSTRING] = {'\0', };
   int PaddingDone = FALSE;

   while (NULL != CurrentGlobal) {
      CurrentLine = CurrentGlobal->FirstLine;
      if ((NULL != CurrentGlobal->sectionBuffer)
          && (0 != strcmp(CurrentGlobal->sectionBuffer, oldSection))) {
         /* Entered a new section type. */
         (void) strcpy(oldSection, CurrentGlobal->sectionBuffer);
         /* NOTE: .align could be machine-dependent.!!!!! */
         (void) printf("%s\n.align %d\n", oldSection, alignment);
#define SPARC_PAD_TEST FALSE
#if SPARC_PAD_TEST
         (void) printf(".skip %d\n", 12288);
#endif
      }
      else if ((NULL != CurrentLine)
               && (NULL == CurrentGlobal->CoalescedMSNeighbor))
         /* NOTE: .align could be machine-dependent.!!!!! */
         (void) printf(".align %d\n", alignment);

      if (GlobalIsCommonData(CurrentGlobal))
         mdEmitCommonGlobal(CurrentGlobal);
      else {
         while (NULL != CurrentLine) {
            printable = CurrentLine->line;
            assert(NULL != printable);
            if (('-' == *printable) || ('_' == *printable))
               ++printable;
            (void) printf("%s\n", printable);
            CurrentLine = CurrentLine->next;
         }
      }

      if (0 < CurrentGlobal->PadBytes) {
         /* Call a machine-dependent ASM emitting function
          * to emit a certain number of bytes of dead space, such
          * as the .space or .skip directives on some machines.
          */
         mdEmitPadBytes(CurrentGlobal);
      }
      CurrentGlobal = CurrentGlobal->next;
   }
} /* end of emitAsmForGlobals() */

/*
 * declaration - handle local and global variable declarations
 */
int declaration(char *val)
{
   register struct locuse *ptr;
   struct globalsym *globPtr = NULL;
   int   blev, type;
   char *a, *start, id[MAXSTRING], code[REGSTRING];
   int   hash = 0;

   /*
    * The first item in a declare line is the identifier name
    */
   start = val;
   a = id;
   while ((*a++ = *val++) != '\t') ;
   *(a - 1) = *(val - 1) = '\0';

   /*
    * The second item is the two-character code that will be used in RTLs
    */
   copyloc(code, val);
   val = skiploc(val);
   *skiploc(code) = 0;
   val++;
#ifdef CALLGRAPH
   /* Need to memorize globals for the callgraph building routines. */
   if (isGlobal(code)) {
      if (maxGlobalNum < getGlobalNum(code))
         maxGlobalNum = getGlobalNum(code);
   }
#endif

   /*
    * The third item is the identifier level:
    *
    * 0  -- global 1  -- argument 2  -- normal local >2 -- locals within blocks
    */
   blev = atoi(val);
   val = is_const(val);
   if (0 == blev) { /* Global variable. */
      if (vecmap[hash = USP(code)] < 0) {
#ifdef INTERACT
         struct glouse *glos_ptr;
         if (swq) {
	   /*glos_ptr = (struct glouse *) SALLOC(sizeof(struct glouse));*/
	   glos_ptr = (struct glouse *) myalloc_perm(sizeof(struct glouse));
            if (NULL == glos_ptr) {
               (void) fprintf(stderr,
                              "vpo (declaration): out of virtual storage\n");
               exit(EXIT_FAILURE);
            }
            strcpy(glos_ptr->code, code);
            glos_ptr->global_name_ptr = globals + gallocptr;
            add_glouse_node(glos_ptr);
         }
#endif
         vecmap[hash] = gallocptr;
         if (CollectingGlobals) {
            globPtr = addGlobalSym(code, gallocptr);
            assert(NULL != globPtr);
         }
         for (a = globals + gallocptr; *a++ = *start++; gallocptr++) {
            if (gallocptr == (GLOBALS - 1)) {
               (void) fprintf(stderr,
                              "vpo (declaration): out of global storage\n");
               exit(EXIT_FAILURE);
            }
         }
         gallocptr++;
      }
      else {
         (void) fprintf(stderr,
                        "vpo (declaration): global code multiply declared\n");
         fprintf(stderr, "Code=%d, old name=%s new name=%s, map to %d",
                 getGlobalNum(code), globalname(code), start + 1, hash);
#ifdef INTERACT
	 /*this seems to happen sometime on the ARM, when undoing, or reading
	   the function back in*/
	 return blev;
#else	 
         exit(EXIT_FAILURE);
#endif	 
      }
   }
   else { /* Argument or local; basically the same for our purposes. */
      if (vecmap[hash = USP(code)] < 0) {
         if (regallocptr == MAXRVEC) {
            (void) fprintf(stderr, "vpo (declaration): too many items\n");
            exit(EXIT_FAILURE);
         }
         (void) strcpy(sets[regallocptr].item, code);
         sets[regallocptr].num = regallocptr;
         vecmap[regname(regallocptr)] = regallocptr;
         ptr = (struct locuse *) SALLOC(sizeof(struct locuse));
         if (!locs) {
            ptr->prev = (struct locuse *) NULL;
            locs = loce = ptr;
         }
         else {
            loce->next = ptr;
            ptr->prev = loce;
            loce = ptr;
         }
         ptr->next = (struct locuse *) NULL;
         sets[regallocptr].local = ptr;
         regallocptr++;
         ptr->item = start;
         ptr->level = blev;
         *(skiploc(code)) = '\0';
         (void) strcpy(ptr->code, code);
         ptr->setin = dinit();
         ptr->usedin = dinit();
         ptr->size = 0;
         ptr->offset = 0;
         ptr->type = 0;
         ptr->vol = FALSE;
         ptr->reps = (swr) ? 0 : 3;
#if 0
         ptr->indref = (swr || swo) ? FALSE : TRUE;
#else
         /* It is safer to start with indref set to TRUE, then have
          * live_variable_analysis() or locals_used() call clean_locals()
          * to set it to FALSE.  Then, they can analyse the whole function
          * at once and set indref to TRUE whenever they see an indirect
          * reference.  This way, during all the optimization that precedes
          * the first call to live_variable_analysis(), indref has the safe
          * value of TRUE.  In particular, instruction_selection() is called
          * several times before the locals are analyzed, and it calls
          * maybe_same(), which needs the indref field to be safe.
          */
         ptr->indref = TRUE;
#endif
         ptr->nouse = ptr->reg = ptr->structure = FALSE;
         ptr->arg = (blev == 1) ? TRUE : FALSE;
         ptr->atype = NOTYPE;
         *ptr->regin = '\0';
         *ptr->csreg = '\0';
         ptr->olist = vinit();
         ptr->conptr = (struct ignode *) NULL;
         ptr->memChar = 0;
         ptr->addr_mode = 0;
         ptr->PlacementPartition = 0;
         ptr->PadBytes = 0;
         ptr->CoalescingStatus = 0;
         ptr->CoalescedMSNeighbor = (struct locuse *) NULL;
         ptr->CoalescedLSNeighbor = (struct locuse *) NULL;

#ifdef STACHE
         ptr->stache_data = NULL;
#endif
      }
      else {
         (void) fprintf(stderr,
                        "vpo (declaration): local code multiply declared\n");
         fprintf(stderr, "Code=%d, old name=%s new name=%s, map to %d",
                 getLocalNum(code), localname(code), start + 1, hash);
         exit(EXIT_FAILURE);
      }
   }

   /*
    * The fourth item is the register type that corresponds to the type that
    * the item was declared as.  This may or may not be the register type that
    * will be used to actually hold this local if it is allocated to a
    * register, but it does provide a method to differentiate between locals
    * that should be allocated to address registers versus data registers on
    * machines like the Motorola 68020.  The base type is limited to the eight
    * least-significant bits of the type.  The ninth bit, if set, indicates
    * that the item is a structure.  This field is not used on all machines,
    * only on those that need to know that the item is a structure for some
    * particular reason.
    */
   type = atoi(++val);
   val = is_const(val);
   if (0 < blev) { /* Not for globals. */
      ptr->type = type & 255;
      if (type & 256)
         ptr->structure = TRUE;
   }

   /*
    * The fifth (optional) item is the size of the item
    */
   if (*val) {
      if (0 < blev)
         ptr->size = atoi(++val);
      else if (CollectingGlobals) { /* Global. */
         assert(NULL != globPtr);
         globPtr->size = atoi(++val);
      }
      val = is_const(val);

      /*
       * The sixth (optional) item is the offset of the item
       */
      if ((0 < blev) && *val) {
         ptr->offset = atoi(++val);
         val = is_const(val);

         /*
          * The seventh (optional) item is non-zero if the item is volatile
          */
         if (*val)
            ptr->vol = atoi(++val);
      }
   }

   return blev;
} /* end of declaration() */

/*
 * makeglobal - sets up a new global identifier and returns its code
 */
void makeglobal(char *id, char *code)
{
   int i;
   char *a;

   /*
    * Locate a free slot for the global
    */
   for (i = 0; i < GLOBALS; i++) {
      copyloc(code, getGlobal(i));
      if (vecmap[USP(code)] == -1)
         break;
   }
   if (i < 0) {
      (void) fprintf(stderr, "vpo (makeglobal): out of global storage\n");
      exit(EXIT_FAILURE);
   }

   /*
    * Insert the identifier in the appropriate slot
    */
   vecmap[USP(code)] = gallocptr;
   for (a = globals + gallocptr; *a++ = *id++; gallocptr++) {
      if (gallocptr == (unsigned short) (GLOBALS - 1)) {
         (void) fprintf(stderr, "vpo (makeglobal): out of global storage\n");
         exit(EXIT_FAILURE);
      }
   }
   gallocptr++;
   *(skiploc(code)) = '\0';
} /* end of makeglobal() */

#ifdef CALLGRAPH

/*
 * Return TRUE if the global is leftmost in a chain; FALSE if it is not
 * in a chain, or it is not leftmost in its chain.
 */
int LeftmostInGlobalChain(struct globalsym *global)
{
   assert(NULL != global);
   return ((NULL == global->CoalescedMSNeighbor)
           && (NULL != global->CoalescedLSNeighbor));
}

/*
 * Return TRUE if the global is rightmost in a chain; FALSE if it is not
 * in a chain, or it is not rightmost in its chain.
 */
int RightmostInGlobalChain(struct globalsym *global)
{
   assert(NULL != global);
   return ((NULL == global->CoalescedLSNeighbor)
           && (NULL != global->CoalescedMSNeighbor));
}

/*
 * Follow the chain leftward and return the leftmost in the chain of
 * coalesced globals.
 */
struct globalsym *GetLeftmostInGlobalChain(struct globalsym *global)
{
   assert(NULL != global);
   /* If it is coalesced, but has no MS neighbor, it is leftmost. */
   while (NULL != global->CoalescedMSNeighbor) {
      global = global->CoalescedMSNeighbor;
      assert(NULL != global->CoalescedLSNeighbor);
   }

   return global;
}

/*
 * Follow the chain rightward and return the rightmost in the chain of
 * coalesced globals.
 */
struct globalsym *GetRightmostInGlobalChain(struct globalsym *global)
{
   assert(NULL != global);
   /* If it is coalesced, but has no LS neighbor, it is rightmost. */
   while (NULL != global->CoalescedLSNeighbor) {
      global = global->CoalescedLSNeighbor;
   }

   return global;
}

/* DeallocateGlobalDataProfile - free the memory in a struct GlobalDataProfile
 * contained in the given CFG node.
 */
void DeallocateGlobalDataProfile(struct GlobalPlacementCFGNode *cnode) {
   struct varl *temp, *temp2;

   assert(NULL != cnode);
   assert(NULL != cnode->prof);

   temp = cnode->prof->FirstGlobal;

   while (NULL != temp) {
      temp2 = temp->next;
      FREE(temp);
      temp = temp2;
   }
   cnode->prof->FirstGlobal = NULL;
   cnode->prof->LastGlobal = NULL;
   FREE(cnode->prof);
} /* end of DeallocateGlobalDataProfile() */

/* DeallocateCFGLoopList - free the memory in a struct CFGloopl
 * contained in the given CFG node.
 */
void DeallocateCFGLoopList(struct GlobalPlacementCFGNode *cnode) {
   struct CFGloopl *temp, *temp2;

   assert(NULL != cnode);
   assert(NULL != cnode->LoopList);

   temp = cnode->LoopList;

   while (NULL != temp) {
      temp2 = temp->next;
      FREE(temp);
      temp = temp2;
   }
   return;
} /* end of DeallocateCFGLoopList() */

/*
 * FreeCFGNodes - helper function for FreeCFG().
 */
void FreeCFGNodes(struct GlobalPlacementCFGNode *CFGnode1) {
   struct GlobalPlacementCFGNode *CFGnode2;

   while (NULL != CFGnode1) {
      CFGnode2 = CFGnode1->next;
      assert(NULL != CFGnode1->prof);
      DeallocateGlobalDataProfile(CFGnode1);
      if (NULL != CFGnode1->LoopList)
         DeallocateCFGLoopList(CFGnode1);
      FREE(CFGnode1);
      CFGnode1 = CFGnode2;
   }
}

/*
 * FreeCFG - Free all the memory for the CFG of a given
 * call graph node.  This was allocated via SALLOC_PERM, so it is
 * important to free it ASAP, whether VPO was build with DISPOSE turned
 * on or off, so that we don't run out of virtual memory on large
 * benchmarks.
 */
void FreeCFG(struct call_graph_node *cnode) {
   struct GlobalPlacementCFGNode *CFGnode1;

   assert(NULL != cnode);

   CFGnode1 = cnode->topCFG;
   FreeCFGNodes(CFGnode1);
   return;
}

/* CacheInterferenceAffinity - add up the affinity of cache lines that
 * conflict with each other at a given level of cache.
 */
static float CacheInterferenceAffinity(int CacheLevel) {
   float Interference = 0.0;
   float proportion1, proportion2;
   int CacheLines1, CacheLines2;
   int CacheColumns; /* How many times do we wrap around the cache way size? */
   int CacheWayStride;
   int CacheAssociativity;
   int CacheLine;
   int CacheMapIndex;
   int row, col, index2;
   int CacheWayIndex;
   struct globalsym *gptr1, *gptr2;

   assert((1 <= CacheLevel) && (2 >= CacheLevel));
   assert(NULL != CacheMap);
   assert(0 < CacheMapSize);
   assert(0 < CacheMapLines);
   assert(0 < GlobalsCount);
   assert(NULL != AffinityArray);

   if (1 == CacheLevel) {
      CacheColumns = 1 + ((CacheMapSize - 1) / L1_DATA_CACHE_WAY_SIZE);
      CacheWayStride = L1_DATA_CACHE_WAY_SIZE / CACHE_LINE_PACKING_QUANTUM;
      CacheAssociativity = L1_DATA_CACHE_ASSOCIATIVITY;
   }
   else { /* Must be level 2 cache. */
      CacheColumns = 1 + ((CacheMapSize - 1) / L2_DATA_CACHE_WAY_SIZE);
      CacheWayStride = L2_DATA_CACHE_WAY_SIZE / CACHE_LINE_PACKING_QUANTUM;
      CacheAssociativity = L2_DATA_CACHE_ASSOCIATIVITY;
   }

   /* If we don't wrap around the cache more times than the associativity,
    * then there cannot be any conflict.
    */
   if (CacheColumns <= CacheAssociativity)
      return Interference; /* i.e. return 0.0 */

   /* Otherwise, go through the first way of the cache map and compute the
    * total affinity among all data items from later ways of the cache that
    * map into the same cache line number.  Repeat for all cache ways.
    */
   for (CacheLine = 0; CacheLine < CacheWayStride; ++CacheLine) {
      for (CacheWayIndex = 0; CacheWayIndex < CacheColumns; ++CacheWayIndex) {
         CacheMapIndex = (CacheWayIndex * CacheWayStride) + CacheLine;
         if (CacheMapIndex >= CacheMapLines) {
            /* We can only be out of the CacheMap bounds on the final way. */
            assert(CacheWayIndex == (CacheColumns - 1));
            continue;  /* Avoid going out of CacheMap bounds. */
         }
         gptr1 = CacheMap[CacheMapIndex];
         assert(NULL != gptr1);
         row = gptr1->AffinityIndex;
         assert(0 <= row);
         CacheLines1 = CACHE_LINES(gptr1);
         assert(0 < CacheLines1);
         proportion1 = 1.0 / CacheLines1;

         for (index2 = CacheWayIndex + 1; index2 < CacheColumns; ++index2) {
            CacheMapIndex = (index2 * CacheWayStride) + CacheLine;
            if (CacheMapIndex >= CacheMapLines) {
               /* We can only be out of bounds on the final way. */
               assert(index2 == (CacheColumns - 1));
               continue;  /* Avoid going out of CacheMap bounds. */
            }
            gptr2 = CacheMap[CacheMapIndex];
            assert(NULL != gptr2);
            col = gptr2->AffinityIndex;
            assert(0 <= col);
            CacheLines2 = CACHE_LINES(gptr2);
            assert(0 < CacheLines2);
            proportion2 = 1.0 / CacheLines2;
            /*-
             * We can finally add affinity values. A note on the
             * use of proportion1 and proportion2: Picture a loop
             * nest such as the following:
             *
             * for (i = 0; i < 100; ++i) {
             *    for (j = 0; j < 10; ++j)
             *       A[j] = j;
             *
             *    for (k = 0; k < 3; ++k)
             *       B[k] = k;
             * }
             *
             * What conflict affinity shall be recorded here?
             * For simplicity, assume each array element in A[]
             * and B[] occupies a full cache line, and that
             * A[] has 10 elements and B[] has 3.  B[] can evict
             * at most 3 of the 10 cache lines belonging to A[]
             * on each outer loop iteration.  Thus, B[] can cause
             * 300 misses in A[], and A[] can cause 300 misses
             * in B[].  Now, examine what the "estimate" fields
             * will be in each basic block above.  The assignment
             * into A[] will have an estimate field of 1000.0,
             * while the assignment into B[] will have an estimate
             * field of 300.0.  Path profiling will cause a value
             * of 300.0 to be added to the (A, B) entry in the
             * AffinityArray[], because A comes first.  If another
             * loop nest were identical to the one above, but
             * had the order of the inner loops switched, then
             * a value of 1000.0 would be added to the (B, A)
             * entry.  However, only 300 cache misses could be
             * caused in each array in the second loop nest,
             * just as in the loop nest above.  In order to
             * use the AffinityArray[] values properly, we would
             * want each B[] cache line to be responsible for
             * one third of the 300.0 value in the (A, B) entry
             * in AffinityArray[], and each cache line in A[]
             * to be responsible for one tenth of the 1000.0
             * value in the (B, A) entry.  This is accomplished
             * by the proportioning below.
             */
            Interference += 
             (proportion2 * (*(AffinityArray + (row * GlobalsCount) + col)));
            Interference +=
             (proportion1 * (*(AffinityArray + (col * GlobalsCount) + row)));
         }
      } /* end for (CacheWayIndex = 0; ...) */
   } /* end for (CacheLine = 0; ...) */

   assert(0.0 <= Interference);
   return Interference;

} /* end of CacheInterferenceAffinity() */


/* BuildCacheMap - go through globals list and map globals into the cache
 * lines they occupy.
 */
static void BuildCacheMap(int BuildPartitionStartingLine) {
   struct globalsym *gptr;
   int size;
   int CacheLines, CacheLineLimit;
   int CacheLineIndex = 0;
   int CurrentPartition = -1;
   int CurrentUsedGlobals = 0;
   int CurrentShadowGlobals = 0;  /* Non-leftmost coalesced globals. */

   assert(NULL != CacheMap);
   assert(0 < CacheMapLines);
   assert(0 < CacheMapSize);

   if (BuildPartitionStartingLine) {
      assert(NULL != PartitionGlobalsCount);
      assert(NULL != PartitionShadowCount);
      assert(NULL != PartitionStartingLine);
   }

   for (gptr = globSymStart; gptr; gptr = gptr->next) {
      size = GlobalChainSize(gptr);
      if (0 >= size)
         continue; /* Too small or junk; who knows; could assert here. */
      if (NULL != gptr->CoalescedMSNeighbor) {
         assert(!BuildPartitionStartingLine
                || (gptr->PlacementPartition == CurrentPartition));
         ++CurrentShadowGlobals;
         continue; /* Don't care about middle of chain. */
      }
      if (size >= L2_DATA_CACHE_SIZE)
         continue;  /* Too big; conflicts with everything anyway. */
      if ((gptr->PlacementPartition != CurrentPartition)
          && BuildPartitionStartingLine
          && (gptr->PlacementPartition >= 0)) {
         if (CurrentPartition >= 0) {
            PartitionGlobalsCount[CurrentPartition] = CurrentUsedGlobals;
            PartitionShadowCount[CurrentPartition] = CurrentShadowGlobals;
         }
         CurrentPartition = gptr->PlacementPartition;
         PartitionStartingLine[CurrentPartition] = CacheLineIndex;
         CurrentUsedGlobals = 0;
         CurrentShadowGlobals = 0;
      }
      CacheLines = CACHE_LINES(gptr);
      CacheLineLimit = CacheLineIndex + CacheLines;
      assert(CacheLineLimit <= CacheMapLines);
      while (CacheLineIndex < CacheLineLimit) {
         CacheMap[CacheLineIndex++] = gptr;
      }
      ++CurrentUsedGlobals;
   } /* end for (gptr = globSymStart; ...) */

   /* Record # globals for final partition. */
   if (BuildPartitionStartingLine && (CurrentUsedGlobals > 0)) {
      assert(0 <= CurrentPartition);
      PartitionGlobalsCount[CurrentPartition] = CurrentUsedGlobals;
      PartitionShadowCount[CurrentPartition] = CurrentShadowGlobals;
   }

   /* Due to mysterious rounding of sizes, etc., we have a hard time
    * sizing the CacheMap exactly, so we allocated a few extra lines.
    * Now that we know exactly how many lines there are, adjust
    * the value of CacheMapLines accordingly.
    */
   assert(CacheLineIndex == CacheLineLimit);
   if (CacheLineIndex < CacheMapLines)
      CacheMapLines = CacheLineIndex;
} /* end of BuildCacheMap() */

/* SwapEntriesInCache - given a pair of pointers to globals, swap them in the
 * given cache map and realign entries between them to leave no gaps. The
 * pair of globals MUST be in the same placement partition.
 */
static void 
SwapEntriesInCache(struct globalsym *gptr1, struct globalsym *gptr2) {
   int CacheLines1, CacheLines2;
   int StartingLine1 = -1, StartingLine2 = -1;
   int TempLine1, TempLine2;
   int partition;
   int index;
   int temp; /* for swapping */
   struct globalsym *gptr3; /* For swapping. */

   assert(NULL != gptr1);
   assert(NULL != gptr2);
   assert(gptr1 != gptr2);
   assert(NULL != CacheMap);
   assert(0 < CacheMapSize);
   assert(0 < CacheMapLines);
   assert(gptr1->PlacementPartition == gptr2->PlacementPartition);
   assert(gptr1->SectionNumber == gptr2->SectionNumber);
   assert(NULL != PartitionStartingLine);

   partition = gptr1->PlacementPartition;
   CacheLines1 = CACHE_LINES(gptr1);
   CacheLines2 = CACHE_LINES(gptr2);

   index = PartitionStartingLine[partition];
   do {
      if (CacheMap[index] == gptr1) { /* Found gptr1. */
         StartingLine1 = index;
         if (StartingLine2 >= 0) break; /* Must have found gptr2 earlier. */
         index += CacheLines1; /* Skip over duplicate gptr1 entries in map. */
      }
      else if (CacheMap[index] == gptr2) { /* Found gptr2. */
         StartingLine2 = index;
         if (StartingLine1 >= 0) break; /* Must have found gptr1 earlier. */
         index += CacheLines2; /* Skip over duplicate gptr2 entries in map. */
      }
      else
         index += CACHE_LINES(CacheMap[index]);
      assert(index < CacheMapLines); /* Else we didn't find 1 of the 2. */
   } while (CacheMap[index]->PlacementPartition == partition);

   /* To reduce the case by case analysis during swapping, make gptr1 and its
    * associated variables refer to the global found earlier in the cache.
    */
   if (StartingLine1 > StartingLine2) { /* Need to swap pointers, etc. */
      gptr3 = gptr1;
      gptr1 = gptr2;
      gptr2 = gptr3;
      temp = StartingLine1;
      StartingLine1 = StartingLine2;
      StartingLine2 = temp;
      temp = CacheLines1;
      CacheLines1 = CacheLines2;
      CacheLines2 = temp;
   }

   /* Now, swap the gptr1 and gptr2 entries. If they are not the same size,
    * move the globals between them up or down.
    */
   if (CacheLines1 <= CacheLines2) {
      /* Move in-between entries down by (CacheLines2 - CacheLines1) lines.
       * Note that this might be zero, meaning gptr1 and gptr2 are contiguous,
       * which is fine.
       */
      for (TempLine2 = StartingLine2 - 1;
           TempLine2 >= (StartingLine1 + CacheLines1);
           --TempLine2) {
         CacheMap[TempLine2 + (CacheLines2 - CacheLines1)] 
            = CacheMap[TempLine2]; 
      }
   }
   else {
      /* Move in-between entries up by (CacheLines1 - CacheLines2) lines.
       * Note that this might be zero, meaning gptr1 and gptr2 are contiguous,
       * which is fine.
       */
      for (TempLine1 = StartingLine1 + CacheLines2;
           TempLine1 < (StartingLine2 + CacheLines2 - CacheLines1);
           ++TempLine1) {
         CacheMap[TempLine1] 
            = CacheMap[TempLine1 + (CacheLines1 - CacheLines2)];
      }
   }
   /* Write gptr1 entries onto the final CacheLines1 entries for gptr2. */
   for (TempLine2 = StartingLine2 + CacheLines2 - 1;
        TempLine2 >= StartingLine2 + CacheLines2 - CacheLines1;
        --TempLine2) {
      CacheMap[TempLine2] = gptr1;
   }
   /* Now, write gptr2 into the (CacheLines2) entries starting where gptr1
    * used to be.
    */
   for (TempLine1 = StartingLine1;
        TempLine1 < (StartingLine1 + CacheLines2);
        ++TempLine1) {
      CacheMap[TempLine1] = gptr2;
   }
} /* end of SwapEntriesInCache() */

/* SwapGlobalsInList - swap two globals in the globSymStart list.
 */
static void 
SwapGlobalsInList(struct globalsym *gptr1, struct globalsym *gptr2) {
   struct globalsym *gptr3;
   int adjacent12;
   int adjacent21;

   assert(NULL != gptr1);
   assert(NULL != gptr2);
   assert(gptr1 != gptr2);
   assert(gptr1->PlacementPartition == gptr2->PlacementPartition);
   assert(gptr1->SectionNumber == gptr2->SectionNumber);

#ifdef VET
   begin_trans(VET_OPT_TRANS);
#endif
   /* Special case: one is already the successor of the other. */
   adjacent12 = (gptr2 == gptr1->next);
   adjacent21 = (gptr1 == gptr2->next);
   if (adjacent12) {
      /* gptr1 comes before gptr2.  Just insert gptr2 before gptr1. */
      InsertGlobal(gptr1, gptr2);
      return;
   }
   if (adjacent21) {
      /* gptr2 comes before gptr1.  Just insert gptr1 before gptr2. */
      InsertGlobal(gptr2, gptr1);
      return;
   }

   /* First, insert gptr2 to the left of gptr1.  Keep track of gptr2's
    * old successor in the list.
    */
   gptr3 = gptr2->next;
   if (NULL == gptr3) { /* Swap for simplicity below. */
      gptr3 = gptr1;
      gptr1 = gptr2;
      gptr2 = gptr3;
      gptr3 = gptr2->next;
      assert(NULL != gptr3);
   }
   InsertGlobal(gptr1, gptr2);
   /* Now, insert gptr1 where gptr2 used to be. */
   InsertGlobal(gptr3, gptr1);
   /* If either global was part of a chain, bring the rest of the
    * chain along with it.
    */
   if ((NULL != gptr1->CoalescedMSNeighbor)
       || (NULL != gptr1->CoalescedLSNeighbor)) {
      assert(LeftmostInGlobalChain(gptr1));
      GroupGlobalChain(gptr1);
   }
   if ((NULL != gptr2->CoalescedMSNeighbor)
       || (NULL != gptr2->CoalescedLSNeighbor)) {
      assert(LeftmostInGlobalChain(gptr2));
      GroupGlobalChain(gptr2);
   }
#ifdef VET
   end_trans();
#endif
} /* end of SwapGlobalsInList() */

/* TwoOptRefinement - performs swapping of globals in the CacheMap (and in the
 * globals list) as long as a reduction in the cache interference is
 * measured, or until some number of swaps has been performed (to limit
 * the time cost of the optimization).  The starting interference and the
 * highest partition number (the one used by the variables bigger than the
 * L1_DATA_CACHE_WAY_SIZE, not actually partitioned by Chaco) are passed in.
 */
static float TwoOptRefinement(float OldInterference, int MinPartition, 
                              int HighestPartition) {
   float interference;
   int partition;
   int StartingLine;
   int TotalSwaps = 0; /* Total across all partitions. */
   int CurrentSwaps;   /* Just within this partition. */
   int BadSwaps;       /* Consecutive swaps that pessimized cache layout. */
   int RandomGlobal1, RandomGlobal2;
   int CacheLines1;
   int IndexIntoGlobals = 0;
   struct globalsym *gptr1;
   struct globalsym *gptr2;
   int temp;
   int count;
   int GlobalsIndex;

   assert(0 <= MinPartition);
   assert(0 <= HighestPartition);

   if (0.0 >= OldInterference)
      return OldInterference;

   assert(NULL != Globals);

   for (partition = MinPartition; partition <= HighestPartition; ++partition) {
      if (0 < partition) {
         IndexIntoGlobals += PartitionGlobalsCount[partition - 1];
         IndexIntoGlobals += PartitionShadowCount[partition - 1];
      }
      StartingLine = PartitionStartingLine[partition];
      count = PartitionGlobalsCount[partition]
              + PartitionShadowCount[partition];
      if ((0 > StartingLine) || (0 == count))
         /* Partition with just Chaco dummies. */
         continue;

      assert(0 < count);
      if (1 == PartitionGlobalsCount[partition]) /* Cannot do any swapping. */
         continue;

      CurrentSwaps = 0;
      BadSwaps = 0;
      do {
         /* Get two global variable numbers from 0..(count - 1). */
         RandomGlobal1 = GetUniformRandom(0, count - 1);
         do {
            RandomGlobal2 = GetUniformRandom(0, count - 1);
         } while (RandomGlobal1 == RandomGlobal2);

         /* Get globalsym pointers to these two globals.*/
         if (RandomGlobal1 > RandomGlobal2) {
            temp = RandomGlobal1;
            RandomGlobal1 = RandomGlobal2;
            RandomGlobal2 = temp;
         }
         gptr1 = Globals[IndexIntoGlobals + RandomGlobal1];
         assert(NULL != gptr1);
         gptr1 = GetLeftmostInGlobalChain(gptr1);
         CacheLines1 = CACHE_LINES(gptr1);
         gptr2 = Globals[IndexIntoGlobals + RandomGlobal2];
         assert(NULL != gptr2);
         gptr2 = GetLeftmostInGlobalChain(gptr2);
         if (gptr1 == gptr2)  /* Accidentally hit two in same chain. */
            continue;

         /* Swap these two globals and recompute the cache interference. */
         SwapEntriesInCache(gptr1, gptr2);
         interference = CacheInterferenceAffinity(1);

         /* See if cache interference decreased or not. */
         if (interference >= OldInterference) { /* No improvement. */
            ++BadSwaps;
            SwapEntriesInCache(gptr2, gptr1); /* Undo the swap. */
            if (BadSwaps >= MAX_CONSECUTIVE_BAD_SWAPS) {
#if DEBUG_PLACEMENT
               (void) fprintf(stderr,
                              "Bad swaps termination on partition %d\n",
                              partition);
#endif
               break; /* Go to next partition. */
            }
         }
         else { /* Improvement. */
#if DEBUG_PLACEMENT
            (void) fprintf(stderr,
                           "Good swap in partition %d Delta: %12.6f Bad: %d\n",
                           partition, (interference - OldInterference),
                           BadSwaps);
#endif
            BadSwaps = 0;
            ++CurrentSwaps;
            OldInterference = interference;
            /* Swap gptr1 and gptr2 in the globalsym list and in Globals[]. */
            SwapGlobalsInList(gptr1, gptr2);
            GlobalsIndex = BuildGlobalsArray(L2_DATA_CACHE_SIZE);
         }
         ++TotalSwaps;
         if (TotalSwaps >= MAX_SWAPS_PER_REFINEMENT) {
#if DEBUG_PLACEMENT
            (void) fprintf(stderr,
                           "Total swaps termination on partition %d\n",
                           partition);
#endif
            return OldInterference;
         }
         if (2 == PartitionGlobalsCount[partition])
            /* We have done the only possible swap. */
            break;
      } while (CurrentSwaps < MAX_SWAPS_PER_PARTITION);

#if DEBUG_PLACEMENT
      if (CurrentSwaps >= MAX_SWAPS_PER_PARTITION) {
            (void) fprintf(stderr,
                           "Good swaps termination on partition %d\n",
                           partition);
      }
#endif

   } /* end for (partition = MinPartition; ... ) */

   return OldInterference;
} /* end of TwoOptRefinement() */

/* Insert global2 before global1 in the globalsym list.  If global1 is NULL,
 * we are inserting global2 at the end of the list.
 */
void InsertGlobal(struct globalsym *global1, struct globalsym *global2) {
   assert(NULL != global2);
   assert(global1 != global2);

   /* Boundary condition: if global2 already lies just before global1,
    * just return.
    */
   if (global1 == global2->next)  /* Even if they are NULL here. */
      return;

   /* Make global2->prev bypass global2 with its "next" pointer. */
   if (NULL != global2->prev)
      global2->prev->next = global2->next;
   else {     /* global2 was the first global. */
      assert(globSymStart == global2);
      globSymStart = global2->next;  /* Update beginning pointer. */
   }

   /* Make global2->next bypass global2 with its "prev" pointer. */
   if (NULL != global2->next)
      global2->next->prev = global2->prev;
   else { /* global2 was the last global. */
      assert(globSymEnd == global2);
      globSymEnd = global2->prev;
   }

   /* Make global2 point backward to global1->prev and forward to global1. */   
   global2->next = global1;
   if (NULL != global1)
      global2->prev = global1->prev;
   else { /* global2 has moved to end of globals list. */
      global2->prev = globSymEnd;
      globSymEnd->next = global2;
      globSymEnd = global2;
   }

   /* Make global1->prev point forward to global2. */
   if (globSymStart == global1) {
      assert(NULL == global1->prev);
      globSymStart = global2;
   }
   else {
      if (global1) {
         assert(NULL != global1->prev);
         global1->prev->next = global2;
      }
   }

   /* Make global1 point backward to global2. */
   if (NULL != global1)
      global1->prev = global2;
} /* end of InsertGlobal() */


/* Collect together the globals in a coalesced chain, starting with global1,
 * so that there are no other globals in their midst in the globals list.
 */
void GroupGlobalChain(struct globalsym *global1)
{
   struct globalsym *global2;
   struct globalsym *temp;

   assert(NULL != global1);
   assert(LeftmostInGlobalChain(global1));

   /* We have the leftmost member of a coalesced pair or chain. */
   global2 = global1->CoalescedLSNeighbor;
   assert(NULL != global2);
   assert(global2->CoalescedMSNeighbor == global1);
   temp = global1->next;
   while (global2) {       /* Drop out of loop after end of chain. */
      if (temp != global2) {
         /* Insert global2 to the left of temp, which is to the
          * right of global2->CoalescedMSNeighbor.
          */
         InsertGlobal(temp, global2);
         assert(global2->prev == global2->CoalescedMSNeighbor);
      }
      else {
         temp = temp->next;
      }
      /* Move global2 down the chain. */
      global2 = global2->CoalescedLSNeighbor;
   }
} /* end of GroupGlobalChain() */

/*
 * Sort coalesced variables so that they are paired together and properly
 * aligned.  Return TRUE if there are no problems.  Return FALSE otherwise,
 * which indicates that coalescing cannot proceed and we cannot alter RTLs
 * safely to coalesce loads.
 */
void SortCoalescedGlobals()
{
   struct globalsym *global1;
   struct globalsym *global2;
   struct globalsym *temp;
#ifndef NDEBUG          /* Lengthy assertion logic follows for debugging. */
   int   count = 0;
   int   count2 = 0;

   for (global1 = globSymStart; global1; global1 = global1->next) {
      if (NULL == global1->next)
         assert(global1 == globSymEnd);

      ++count;
   }
#endif

#ifdef VET
   begin_trans(VET_REQUIRED_TRANS);
#endif
   /* Traverse the globalsym list.  Each time we encounter a global that
    * is the absolute leftmost (MS = most significant) member of a coalesced
    * pair or chain, leave it in its current position but traverse rightward
    * (LS direction) and successively insert the other members of the chain
    * after it, moving them from their current positions in the list.
    */
   for (global1 = globSymStart; global1; global1 = global1->next) {
      if (LeftmostInGlobalChain(global1)) {
         /* We have the leftmost member of a coalesced pair or chain. */
         GroupGlobalChain(global1);
      }
   }
#ifdef VET
   end_trans();
#endif

#ifndef NDEBUG          /* Confirm globSymStart list sanity before leaving. */
   for (global1 = globSymStart; global1; global1 = global1->next) {
      ++count2;
      if (NULL != global1->CoalescedMSNeighbor) {
         assert(global1->CoalescedMSNeighbor == global1->prev);
      }
      else if (NULL != global1->CoalescedLSNeighbor) {
         assert(global1->CoalescedLSNeighbor == global1->next);
      }
   }
   assert(count == count2);
#endif
} /* end of SortCoalescedGlobals() */

/*
 * ScaleAffinityArray - Scale the floating point values in the affinity array
 * up or down to get the minimum value to be 1.0 while not allowing the
 * maximum value to exceed a #defined limit.  If these two goals conflict,
 * toss out the values less than 1.0 (intra-block affinity from rarely
 * executed blocks) and reexamine the minimum and maximum values left.
 * When computing how much we need to scale, ignore affinity array values
 * that are not within a single section that is big enough to be
 * processed by Chaco, as indicated by argument RelevantSection[].
 * Return the scale factor used to scale up or down.
 */
static float ScaleAffinityArray(float *AffinityArray, int GlobalsCount,
                               int *RelevantSection) {
   double MinValue = CHACO_AFFINITY_VALUE_LIMIT;
   double MaxValue = 0.0;
   double TempValue;
   double ScaleFactor = 1.0;
   int row, col;
   int index;
   int section;
   int ScaleDown;
   int *RelevantGlobal;
   int *AffinitySection;

   assert(NULL != AffinityArray);
   assert(0 < GlobalsCount);

   RelevantGlobal = (int *) SALLOC(GlobalsCount * sizeof(int));
   AffinitySection = (int *) SALLOC(GlobalsCount * sizeof(int));
   if ((NULL == RelevantGlobal) || (NULL == AffinitySection)) {
      (void) fprintf(stderr, "Out of memory: ScaleAffinityArray()\n");
      exit(EXIT_FAILURE);
   }
   bzero(RelevantGlobal, GlobalsCount * sizeof(int));
   bzero(AffinitySection, GlobalsCount * sizeof(int));

   for (index = 0; index < GlobalsCount; ++index) {
      row = Globals[index]->AffinityIndex;
      section = Globals[index]->SectionNumber;
      assert(0 <= section);
      RelevantGlobal[row] = RelevantSection[section];
      AffinitySection[row] = section;
   }

#ifdef VET
   begin_trans(VET_REQUIRED_TRANS);
#endif
   /* Calculate the minimum and maximum values in the array.  Because
    * the array is asymmetric, we get the affinity value for variables
    * A and B by adding the (A, B) and (B, A) values from the array.
    */
   for (row = 0; row < GlobalsCount; ++row) {
      if (!RelevantGlobal[row])
         continue;
      for (col = row + 1; col < GlobalsCount; ++col) {
         if (!RelevantGlobal[col])
            continue;
         if (AffinitySection[col] != AffinitySection[row])
            continue;  /* Relevant, but different, sections. */
         TempValue = *(AffinityArray + (row * GlobalsCount) + col)
            + *(AffinityArray + (col * GlobalsCount) + row);
         if ((TempValue > 0.0) && (TempValue < MinValue))
            MinValue = TempValue;
         if (TempValue > MaxValue)
            MaxValue = TempValue;
      }
   }

   if (MinValue > MaxValue) {
      (void) fprintf(stderr, "WARNING: No affinity to be scaled!\n");
      goto cleanup;
   }

   /* Determine if we are going to scale down or up. */
   ScaleDown = ((MaxValue > CHACO_AFFINITY_VALUE_LIMIT)
                || (MinValue > 1.0));

   /* Compute the factor by which we will multiply all array values. */
   if (!ScaleDown) {
      /* There are two limits on the scaling up factor.  We don't need
       * to scale up by more than the reciprocal of MinValue, because that
       * is enough to get our minimum up to 1.0.  However, that might cause
       * MaxValue to exceed its Chaco limit, which is more serious.  We
       * would rather toss some small values than exceed the Chaco limit.
       */
      ScaleFactor = 1.0 / MinValue;
      if ((CHACO_AFFINITY_VALUE_LIMIT / MaxValue) < ScaleFactor)
         ScaleFactor = (CHACO_AFFINITY_VALUE_LIMIT / MaxValue);
   }
   else { /* ScaleDown == TRUE */
      /* We scale down by the lesser of two factors, each of which lies
       * between 0.0 and 1.0: the factor we need to bring MaxValue down
       * to the Chaco limit, and the factor we need to bring MinValue
       * down to 1.0 when MinValue is greater than 1.0.
       */
      if (CHACO_AFFINITY_VALUE_LIMIT < MaxValue)
         ScaleFactor = (CHACO_AFFINITY_VALUE_LIMIT / MaxValue);
      if ((1.0 < MinValue) && ((1.0 / MinValue) < ScaleFactor))
         ScaleFactor = (1.0 / MinValue);
   }

#if DEBUG_PLACEMENT
   (void) fprintf(stderr, "AffinityArray scaling factor: %12.7f \n", 
                  ScaleFactor);
#endif

   /* Scale all values in the array by the ScaleFactor.  If any two
    * variables have a mutual affinity that is still less than 0.5 after
    * scaling, set their two affinity entries in the array to 0.0.
    */
   for (row = 0; row < GlobalsCount; ++row) {
      for (col = row + 1; col < GlobalsCount; ++col) {
         TempValue = *(AffinityArray + (row * GlobalsCount) + col)
            + *(AffinityArray + (col * GlobalsCount) + row);
         TempValue *= ScaleFactor;
         if (0.5 > TempValue) {  /* Too small; reset to zero. */
            *(AffinityArray + (row * GlobalsCount) + col) = 0.0;
            *(AffinityArray + (col * GlobalsCount) + row) = 0.0;
         }
         else { /* OK to scale it. */
            *(AffinityArray + (row * GlobalsCount) + col) *= ScaleFactor;
            *(AffinityArray + (col * GlobalsCount) + row) *= ScaleFactor;
         }
      }
   }

#ifdef VET
   end_trans();
#endif

 cleanup:
   FREE(RelevantGlobal);
   FREE(AffinitySection);

   /* NOTE: Diagonal elements are not scaled, but they are currently unused. */
   return (float) ScaleFactor;

} /* end of ScaleAffinityArray() */

#if GLOBAL_DATA_PLACEMENT

/* Write out a *.graph file in Chaco format so that we can run the graph
 * through the standalone version of Chaco and debug it.
 */
static void 
WriteChacoGraphFile(int ChacoVertices, int ChacoEdges, int *ChacoStart, 
                    int *ChacoAdjacency, int *ChacoVertexWeights,
                    float *ChacoEdgeWeights) {
   FILE *GraphFile;
   char *filename;
   int vertex;
   int index, index_max;

   filename =
      (char *) SALLOC((int) (strlen("GlobalDebug  ") + strlen(".graph")));
   if (NULL == filename) {
      (void) fprintf(stderr, "Out of virtual memory: "
                             "locals.c::WriteChacoGraphFile() \n");
      exit(EXIT_FAILURE);
   }

   (void) strcpy(filename, "GlobalDebug");
   (void) strcat(filename, ".graph");

   GraphFile = fopen(filename, "w");
   if (NULL == GraphFile) {
      (void) fprintf(stderr, "ERROR: Cannot open file %s for writing.\n",
                     filename);
      return;
   }

   /* ChacoEdges is a double count, as each edge is listed twice. */
   assert(0 == (ChacoEdges % 2));
   ChacoEdges /= 2;

   /*-
    * The first line of a Chaco *.graph file looks like:
    *
    * num-vertices num-edges  control-number
    *
    * The control number has up to three decimal digits.  The hundreds
    * digit, if present and non-zero, means that we will use vertex numbers
    * explicitly.  We never do this.  The tens digit means we will use edge
    * weights, and the ones digit means we will use vertex weights.  We set
    * each of these to 1, making our control number eleven.
    */
   (void) fprintf(GraphFile, "%d  %d  %d\n", ChacoVertices, ChacoEdges, 11);

   /*-
    * The remaining lines give the vertices and edges with all of their
    * weights.  Each line looks like:
    *
    * vertex-weight neighbor1 edgeweight1 neighbor2 edgeweight2 ....
    *
    * All the neighbors of a vertex appear on its line.  You could precede
    * the vertex-weight by its vertex number (if you set the 100's digit
    * of the control number; see above comment) but we let Chaco implicitly
    * number the vertices starting from 1 for the first vertex line.
    */
   assert(0 == ChacoStart[0]);
   for (vertex = 1; vertex <= ChacoVertices; ++vertex) {
      (void) fprintf(GraphFile, "%d ", ChacoVertexWeights[vertex - 1]);
      for (index = ChacoStart[vertex - 1], index_max = ChacoStart[vertex];
           index < index_max;
           ++index) {
         (void) fprintf(GraphFile, "%d %10.2f ", ChacoAdjacency[index],
                        ChacoEdgeWeights[index]);
      }
      (void) fprintf(GraphFile, "\n");
   }

   if (fclose(GraphFile))
      (void) fprintf(stderr, "Failed to close file %s \n", filename);

   FREE(filename);

   return;
} /* end of WriteChacoGraphFile() */

/* MarkPlacementPartition - 
 * Mark a global, including any other members of its coalesced chain, with
 * a certain PlacementPartition number.
 */
static void MarkPlacementPartition(struct globalsym *gptr, int Partition) {
   struct globalsym *gptr2;

   assert(NULL != gptr);
   gptr2 = GetLeftmostInGlobalChain(gptr);
   while (NULL != gptr2) {
      gptr2->PlacementPartition = Partition;
      gptr2 = gptr2->CoalescedLSNeighbor;
   }
}

/* IncreasePlacementPartition -
 * Increase the partition numbers of all globals with current sections >=
 * a certain SectionNumber.
 */
static void IncreasePlacementPartition(int MinSection, int increase) {
   struct globalsym *gptr2 = globSymStart;

   assert(NULL != gptr2);
   while (NULL != gptr2) {
      if (gptr2->SectionNumber >= MinSection)
         gptr2->PlacementPartition += increase;
      gptr2 = gptr2->next;
   }
}

/* 
 * SortGlobalsByPartition - Sort globals in the given section by the partition
 * number assigned by Chaco.
 */
void SortGlobalsByPartition(int section) {
   /* As our data structure is a linked list, rather than an array or tree
    * or heap, we will perform a simple bubblesort here.  Should this
    * prove to be a bottleneck, we should re-code.
    */
   int change;
   struct globalsym *current;
   struct globalsym *next;
   assert(NULL != globSymStart);

#ifdef VET
   begin_trans(VET_REQUIRED_TRANS);
#endif

   do {
      change = FALSE;
      current = globSymStart;
      next = GetRightmostInGlobalChain(current);
      next = next->next; /* Move past coalesced chain of current. */

      while (NULL != next) {
         if ((current->PlacementPartition > next->PlacementPartition)
             && (current->SectionNumber == section)
             && (next->SectionNumber == section)) {
            change = TRUE;
            InsertGlobal(current, next); /* Insert next before current. */
            if (NULL != next->CoalescedLSNeighbor)
               GroupGlobalChain(next); /* Bring its chain with it. */
            next = GetRightmostInGlobalChain(current);
            next = next->next;
         }
         else {
            current = next;
            next = GetRightmostInGlobalChain(next);
            next = next->next;
            if (current->SectionNumber > section)
               break;
         }
      } /* end while (NULL != next) */
   } while (change);

#ifdef VET
   end_trans();
#endif

   return;
} /* end of SortGlobalsByPartition() */

/*
 * SortGlobalsBySection - Sort globals by the SectionNumber field.
 */
void SortGlobalsBySection() {
   /* As our data structure is a linked list, rather than an array or tree
    * or heap, we will perform a simple bubblesort here.  Should this
    * prove to be a bottleneck, we should re-code.
    */
   int change;
   struct globalsym *current;
   struct globalsym *next;

   assert(NULL != globSymStart);

#ifdef VET
   begin_trans(VET_REQUIRED_TRANS);
#endif

   do {
      change = FALSE;
      current = globSymStart;
      next = GetRightmostInGlobalChain(current);
      next = next->next; /* Move past coalesced chain of current. */

      while (NULL != next) {
         if (current->SectionNumber > next->SectionNumber) {
            change = TRUE;
            InsertGlobal(current, next); /* Insert next before current. */
            if (NULL != next->CoalescedLSNeighbor)
               GroupGlobalChain(next); /* Bring its chain with it. */
            next = GetRightmostInGlobalChain(current);
            next = next->next;
         }
         else {
            current = next;
            next = GetRightmostInGlobalChain(next);
            next = next->next;
         }
      }
   } while (change);

#ifdef VET
   end_trans();
#endif

   return;
} /* end of SortGlobalsBySection() */

/*
 * RecordAssemblerSections - We must keep data from different sections
 * (e.g. uninitialized common blocks, read-only data, initialized read-
 * write data) separate from each other.  They will be placed in contiguous
 * sections together with other data items of the same section by the
 * assembler, so we do not have the freedom to interleave different data
 * types when we place data. We need to call a machine-dependent function
 * for each global to analyze its sectionBuffer field and return a section
 * number, which we place in the SectionNumber field of the symbol table
 * table entry for each global. Return one plus the highest section number
 * seen.
 */
int RecordAssemblerSections() {
   int HighestSection = -2;
   int section;
   struct globalsym *global = globSymStart;

   assert(NULL != global);
   assert(NULL == global->prev);

   while (NULL != global) {
      section = GetGlobalPlacementSection(global);
      global->SectionNumber = section;
      if (section > HighestSection)
         HighestSection = section;
      global = global->next;
   }

   /* We number data sections from zero, so using 0..3 is using 4 sections.
    * Return 1 more than HighestSection.
    */
   return (1 + HighestSection);
} /* end of RecordAssemblerSections() */

/*
 * PackCacheLines - Perform a pseudo-coalescing over all globals smaller than
 * the given CacheLineSize in order to pack as many cache lines as possible.
 * Use the AffinityArray[] to group efficiently.  Pad out remaining partial
 * lines by extending the size of the final item in the line or by adding
 * a dummy variable if alignment considerations require it.  Also pad out
 * variables larger than 1 line to be an integral multiple of the line size.
 * Merge values in rows and columns of the AffinityArray[] to make all affinity
 * refer to the leftmost item in the cache line chain.
 * Limit all operations to the given section number.
 * Return the number of padding bytes added.
 */
static int PackCacheLines(int CacheLineSize, int section) {
   int row, col;
   float MaxDensity;
   int MaxRow, MaxCol;
   int change;
   int size1, size2;  /* Coalesced sizes of row, col variables. */
   int AlignSize1, AlignSize2; /* Alignment required for row, col vars. */
   int Padding12, Padding21;
      /* Pad bytes if size1 precedes size2, vice versa. */
   int AffIndex1, AffIndex2;  /* Indices into the AffinityArray[]. */

   int SavePadding;
   int Var1First;
   int TotalPadBytes = 0;
   float Density;
   float TempValue;
   float Value12, Value21;  /* Affinity values; (row, col), (col, row). */
   struct globalsym *gptr;

   assert(NULL != AffinityArray);
   assert(NULL != Globals);
   assert(0 < GlobalsCount);
   assert(0 <= section);

   /* NOTE: This is really a graph partitioning problem.  We cannot trust
    * Chaco to partition into EXACTLY CacheLineSize partitions, and we are
    * not sure what we would do with irregular-sized partitions returned
    * from Chaco other than apply some greedy method to regroup items.
    * So, we will use a greedy bottom-up clustering here, in which the
    * mutual affinity between a pair of variables is translated into a
    * density of affinity by dividing by the sum of their sizes.  Thus,
    * an 8-byte variable and a 40-byte variable might have a certain affinity,
    * but they use up almost a whole 64-byte cache line between them. A pair
    * of 8-byte variables with slightly less mutual affinity might be a
    * better choice, as they leave room in the cache line for a lot more
    * affinity in future steps of the clustering.
    */

   do {
      change = FALSE;
      MaxDensity = -1.0; /* To pair everything, even zero affinity. */
      MaxRow = MaxCol = -1;
      for (row = 0; row < GlobalsCount; ++row) {
         if (NULL != Globals[row]->CoalescedMSNeighbor)
            continue;  /* Only want leftmost items in chains of vars. */
         if (section != Globals[row]->SectionNumber)
            continue;  /* Invalid to pack across sections. */
         size1 = GlobalChainSize(Globals[row]);
         if (size1 >= CacheLineSize) {
            /* Too big to pack.  Round size up to CacheLineSize multiple. */
            if (0 != (size1 % CacheLineSize)) {
#ifdef VET
               begin_trans(VET_OPT_TRANS);
#endif
               assert(0 == Globals[row]->PadBytes);
               Globals[row]->PadBytes
                  += (CacheLineSize - (size1 % CacheLineSize));
               TotalPadBytes += Globals[row]->PadBytes;
#ifdef VET
               end_trans();
#endif
            }
            continue; /* Cannot be packed into a line with anything else. */
         }
         assert(0 < size1);
         if (size1 <= 2)
            AlignSize1 = size1;
         else if (size1 >= 8)
            AlignSize1 = 8;
         else
            AlignSize1 = 4;

         for (col = row + 1; col < GlobalsCount; ++col) {
            if (NULL != Globals[col]->CoalescedMSNeighbor)
               continue;  /* Only want leftmost items in chains of vars. */
            if (section != Globals[col]->SectionNumber)
               continue;  /* Invalid to pack across sections. */
            size2 = GlobalChainSize(Globals[col]);
            if ((size1 + size2) > CacheLineSize)
               continue;  /* Too big to coalesce together. */
            assert(0 < size2);
            if (size2 <= 2)
               AlignSize2 = size2;
            else if (size2 >= 8)
               AlignSize2 = 8;
            else
               AlignSize2 = 4;
            Padding12 = AlignSize2 - (size1 % AlignSize2);
            assert((0 <= Padding12) && (AlignSize2 >= Padding12));
            if (Padding12 == AlignSize2)
               Padding12 = 0;
            Padding21 = AlignSize1 - (size2 % AlignSize1);
            assert((0 <= Padding21) && (AlignSize1 >= Padding21));
            if (Padding21 == AlignSize1)
               Padding21 = 0;

            if (((size1 + size2 + Padding12) > CacheLineSize)
                && ((size1 + size2 + Padding21) > CacheLineSize))
               continue;  /* Padding will make it too big for cache line. */

            Value12 = *(AffinityArray + row * GlobalsCount + col);
            Value21 = *(AffinityArray + col * GlobalsCount + row);
            TempValue = Value12 + Value21;
            Density = TempValue * ((float) CacheLineSize / (size1 + size2));
            if (Density > MaxDensity) {
               change = TRUE;
               MaxDensity = Density;
               MaxRow = row;
               MaxCol = col;
               /* Does the MaxRow variable come first in the pair, either
                * because of alignment constraints or because it is desired
                * to have the first-accessed variable come first, or desired
                * to minimize padding?
                */
               Var1First = (((size1 + size2 + Padding12) <= CacheLineSize)
                            /* It is legal to have MaxRow var first. */
                            && (((size1 + size2 + Padding21) > CacheLineSize)
                                /* It is not legal to have MaxCol first... */
                                || (Value12 >= Value21)
                                /* ... or we want to order accesses ... */
                                || (Padding21 > Padding12)));
                                /* ... or we want to reduce padding. */
               if (Var1First)
                  SavePadding = Padding12;
               else
                  SavePadding = Padding21;
            }
         }
      }

      if (0 <= MaxRow) { /* Found one to coalesce. */
#ifdef VET
         begin_trans(VET_OPT_TRANS);
#endif
         AffIndex1 = Globals[MaxRow]->AffinityIndex;
         AffIndex2 = Globals[MaxCol]->AffinityIndex;
         /* Link them in the globals list & coalesce affinity entries. */
         if (Var1First) { /* MaxRow is leftmost, MaxCol is rightmost. */
            gptr = GetRightmostInGlobalChain(Globals[MaxRow]);
            gptr->PadBytes += SavePadding;
            TotalPadBytes += SavePadding;
            gptr->CoalescedLSNeighbor = Globals[MaxCol];
            Globals[MaxCol]->CoalescedMSNeighbor = gptr;
            /* Merge col for MaxCol into col for MaxRow in AffinityArray[]. */
            for (row = 0; row < GlobalsCount; ++row) {
               *(AffinityArray + row * GlobalsCount + AffIndex1)
                  += *(AffinityArray + row * GlobalsCount + AffIndex2);
               *(AffinityArray + row * GlobalsCount + AffIndex2) = 0.0;
            }
            /* Merge row for MaxCol into row for MaxRow in AffinityArray[]. */
            for (col = 0; col < GlobalsCount; ++col) {
               *(AffinityArray + AffIndex1 * GlobalsCount + col)
                  += *(AffinityArray + AffIndex2 * GlobalsCount + col);
               *(AffinityArray + AffIndex2 * GlobalsCount + col) = 0.0;
            }
         }
         else { /* MaxCol is leftmost, MaxRow is rightmost. */
            gptr = GetRightmostInGlobalChain(Globals[MaxCol]);
            gptr->PadBytes += SavePadding;
            TotalPadBytes += SavePadding;
            gptr->CoalescedLSNeighbor = Globals[MaxRow];
            Globals[MaxRow]->CoalescedMSNeighbor = gptr;
            /* Merge col for MaxRow into col for MaxCol in AffinityArray[]. */
            for (row = 0; row < GlobalsCount; ++row) {
               *(AffinityArray + row * GlobalsCount + AffIndex2)
                  += *(AffinityArray + row * GlobalsCount + AffIndex1);
               *(AffinityArray + row * GlobalsCount + AffIndex1) = 0.0;
            }
            /* Merge row == MaxRow into row == MaxCol in AffinityArray[]. */
            for (col = 0; col < GlobalsCount; ++col) {
               *(AffinityArray + AffIndex2 * GlobalsCount + col)
                  += *(AffinityArray + AffIndex1 * GlobalsCount + col);
               *(AffinityArray + AffIndex1 * GlobalsCount + col) = 0.0;
            }
         }
#ifdef VET
         end_trans();
#endif
         
      } /* end if (0 <= MaxRow) */
   } while(change);

   /* Because we set MaxDensity to -1.0 on each iteration above, we should
    * have paired everything that could be paired, right down to unpaired
    * variables that had no mutual affinity at all but fit together in
    * a cache line.  However, there could still be dead space at the end
    * of some cache lines that we need to pad out.  Find it and pad it.
    */
   for (row = 0; row < GlobalsCount; ++row) {
      if (NULL != Globals[row]->CoalescedMSNeighbor)
         continue;  /* Only want leftmost items in chains of vars. */
      if (section != Globals[row]->SectionNumber)
         continue;
      size1 = GlobalChainSize(Globals[row]);
      if (size1 >= CacheLineSize)
         continue; /* No padding needed. */
      assert(0 < size1);
#ifdef VET
   begin_trans(VET_OPT_TRANS);
#endif
      gptr = GetRightmostInGlobalChain(Globals[row]);
      assert(NULL != gptr);
      gptr->PadBytes += (CacheLineSize - size1);
      TotalPadBytes += (CacheLineSize - size1);
#ifdef VET
   end_trans();
#endif
   }
   return TotalPadBytes;
} /* end of PackCacheLines() */

/*
 * PadLargeArrays - Examine all arrays of size greater than or equal to MinSize
 * (not counting their PadBytes) and add PadBytes to prevent any array from
 * having the same starting address, modulo MinSize, as any other array.
 * This will reduce the chances of pathological thrashing.
 * Limit all operations to the given section number.
 * Return the number of cache lines of the total size, after padding, of all
 * these arrays, modulo MinSize.
 */

static int PadLargeArrays(int MinSize, int section) {
   int CacheLinesPerWay = CACHE_LINES_FROM_SIZE(MinSize);
   struct globlist *ArrayList = NULL;
   struct globlist *CurrentArray = NULL;
   struct globalsym *gptr;
   struct globalsym **Arrays;
   int CacheLines;
   int CacheLineIndex;
   int ArrayCount = 0;
   int index;
   int PadLineIndex;
   int TotalPadLines = 0;
   int TotalCacheLines = 0;
   char *StartingLines;  /* Boolean array. */

   assert(0 < MinSize);
   assert(0 == (MinSize % CACHE_LINE_PACKING_QUANTUM));

#if PAD_LARGE_ARRAYS

   for (gptr = globSymStart; NULL != gptr; gptr = gptr->next) {
      if (section < gptr->SectionNumber)
         break;
      if ((section > gptr->SectionNumber) || (MinSize > gptr->size))
         continue;
      TotalCacheLines += CACHE_LINES(gptr);
      if (NULL == CurrentArray) { /* First one. */
         assert(0 == ArrayCount);
         ArrayList = (struct globlist *) SALLOC(sizeof(struct globlist));
         if (NULL == ArrayList) {
            (void) fprintf(stderr, "Out of virtual memory: PadLargeArrays\n");
            exit(EXIT_FAILURE);
         }
         CurrentArray = ArrayList;
         CurrentArray->prev = NULL;
      }
      else {
         CurrentArray->next = (struct globlist *)
            SALLOC(sizeof(struct globlist));
         if (NULL == CurrentArray->next) {
            (void) fprintf(stderr, "Out of virtual memory: PadLargeArrays\n");
            exit(EXIT_FAILURE);
         }
         CurrentArray->next->prev = CurrentArray;
         CurrentArray = CurrentArray->next;
      }
      CurrentArray->next = NULL;
      CurrentArray->current = gptr;
      CurrentArray->ArrIndex = ArrayCount++;
   } /* end for (gptr = globSymStart; ...) */

   if (0 == ArrayCount)
      return 0;
   if (1 == ArrayCount)
      goto cleanup;

   /* Allocate an array of globalsym pointers so we don't have to traverse
    * linked list for the rest of the function.
    */
   Arrays = (struct globalsym **)
      SALLOC(sizeof(struct globalsym *) * ArrayCount);
   if (NULL == Arrays) {
      (void) fprintf(stderr, "Out of virtual memory: PadLargeArrays\n");
      exit(EXIT_FAILURE);
   }
   CurrentArray = ArrayList;
   for (index = 0; index < ArrayCount; ++index) {
      assert(NULL != CurrentArray);
      Arrays[index] = CurrentArray->current;
      CurrentArray = CurrentArray->next;
   }

   /* Allocate a boolean map of the cache line numbers that have an
    * array starting in that line.
    */
   StartingLines = SALLOC(CacheLinesPerWay * sizeof(char));
   if (NULL == StartingLines) {
      (void) fprintf(stderr, "Out of virtual memory: PadLargeArrays\n");
      exit(EXIT_FAILURE);
   }
   bzero(StartingLines, CacheLinesPerWay * sizeof(char));

   /* Go through the arrays and pad if conflicts arise among
    * starting addresses.
    */
   CacheLineIndex = 0;
   for (index = 0; index < ArrayCount; ++index) {
      if (!StartingLines[CacheLineIndex]) { /* No conflict. */
         StartingLines[CacheLineIndex] = (char) TRUE;
         CacheLines = CACHE_LINES(Arrays[index]);
         assert(CacheLinesPerWay <= CacheLines);
         CacheLineIndex += CacheLines;
         CacheLineIndex %= CacheLinesPerWay;
      }
      else { /* Conflict. Must pad until no more conflict, if possible. */
         assert(0 < index);
         for (PadLineIndex = 1;
              PadLineIndex < CacheLinesPerWay;
              ++PadLineIndex) {
            if (!StartingLines[(CacheLineIndex + PadLineIndex)
                               % CacheLinesPerWay]) {
#ifdef VET
               begin_trans(VET_OPT_TRANS);
#endif
               CacheLineIndex += PadLineIndex;
               CacheLineIndex %= CacheLinesPerWay;
               TotalPadLines += PadLineIndex;
               Arrays[index - 1]->PadBytes 
                  += (PadLineIndex * CACHE_LINE_PACKING_QUANTUM);
#ifdef VET
               end_trans();
#endif
               break;
            }
         }
         /* We could fall out of the for-loop without finding a
          * way to pad this array, in which case every cache line
          * has an array that starts there.  Might as well exit.
          */
         if (PadLineIndex == CacheLinesPerWay) {
            assert(ArrayCount == CacheLinesPerWay);
            goto cleanup;
         }
      } /* end if (.. no conflict ..) ... else ... */
   } /* end for (index = 0; ...) */

 cleanup:
   if (NULL != Arrays) FREE(Arrays);
   if (NULL != StartingLines) FREE(StartingLines);
   CurrentArray = ArrayList;
   while (NULL != CurrentArray) {
      CurrentArray = CurrentArray->next;
      FREE(ArrayList);
      ArrayList = CurrentArray;
   }

#if DEBUG_PLACEMENT
   (void) fprintf(stderr, "Array padding bytes added for globals : %d \n",
                  TotalPadLines * CACHE_LINE_PACKING_QUANTUM);
#endif

#endif  /* PAD_LARGE_ARRAYS */

   /* Return the offset in the CacheMap that the arrays will create. */
   return ((TotalCacheLines + TotalPadLines) % CacheLinesPerWay);
} /* end of PadLargeArrays() */

/* Combine the optimization of displacement mode addressing with the placement
 * of global data.  Return FALSE if we cannot place globals, TRUE otherwise.
 */
int GlobalDataPlacement()
{
   struct globalsym *gptr;
   int row, col;
   int GlobalsSize = 0;
   int MonsterGlobalsSize = 0;
   int *SmallGlobals = NULL;
   int *SmallGlobalsTotalSize;
   int TempSize;
   int PadBytesTotalSize = 0;
   int ArrayOffsetLines = 0;
   int HasEdges;
   int *PartitionSize = NULL;
   int MaxChacoPartition = 0; /* Highest partition number worked on by Chaco */
   int HighestPartition = 0;  /* Partition # assigned to too-big variables. */
   int HighestPartitionUsed = FALSE; /* Are there any too-big globals? */
   int TotalSections;       /* # Assembler sections in which data is placed. */
   int *SectionSize;        /* # bytes in each section. */
   int *ScaleSection;       /* Boolean: section should affect scaling? */
   int ScalingNeeded = FALSE; /* Any TRUE entries in ScaleSection[] ? */
   float ScaleFactor;
   int index;
   int AffIndex1, AffIndex2;  /* Indices into AffinityArray[]. */
   int section;
   int GlobalsIndex;
   int NextMarkedPartition = 0;
   int MaxPartitions;
   int NoBenefit = FALSE;  /* From local refinement. */
   float interference;  /* Sum of affinities of all conflicting data items. */
   float MinInterference;

   /* Variables to be passed to Chaco. */
   int ChacoEdges = 0;
   int SaveChacoEdges;
   int ChacoVertices = 0;
   int ChacoGlobals = 0;
   int *ChacoGlobalsIndex = NULL; /* Maps ChacoGlobals[] back to Globals[].*/
   int *InverseGlobalsIndex = NULL; /* From Globals[] to ChacoGlobals[]. */
   int *ChacoStart = NULL;
   int *ChacoAdjacency = NULL;
   int *ChacoVertexWeights = NULL;
   float *ChacoEdgeWeights = NULL;
   float ChacoTempEdgeSum;
   short *ChacoAssignment = NULL;
   int ChacoMeshDimensions[3] = {0, 0, 0};
   double *ChacoGoal = NULL;
   int ChacoVMax;
   int ChacoNDimensions;
   int ChacoTemp;
   int ChacoNSection; /* 1 : bisection, 2 : quadrisection, 3 : octasection. */
   double ChacoEigenTolerance = 0.001;
   long ChacoRandomSeed = 0L;
   int ChacoError;
   int ChacoDummiesNeeded;

   if ((0 >= GlobalsCount) || (NULL == Globals))
      return FALSE;

   srand(12345); /* Set random number seed. */

   TotalSections = RecordAssemblerSections();
   assert(0 < TotalSections);
   SortGlobalsBySection();

   for (section = 0; section < TotalSections; ++section) {
      PadBytesTotalSize += PackCacheLines(CACHE_LINE_PACKING_QUANTUM, section);
      ArrayOffsetLines += PadLargeArrays(L1_DATA_CACHE_WAY_SIZE, section);
   }

   SortCoalescedGlobals();
   GlobalsIndex = BuildGlobalsArray(L2_DATA_CACHE_SIZE);

#if DEBUG_PLACEMENT
   (void) fprintf(stderr, "%d pad bytes added for small globals \n",
                  PadBytesTotalSize);
#endif

   /* Produce a boolean array marking just the globals that are smaller than
    * the way size of the cache.  Use only the leftmost in coalesced chains
    * of globals.
    */
   SmallGlobals = (int *) SALLOC((int) sizeof(int) * GlobalsCount);

   /* Need to move this out to global extern space for fixentry() use!!!!! */
   SectionSize = (int *) SALLOC_PERM(sizeof(int) * TotalSections);
   ScaleSection = (int *) SALLOC_PERM(sizeof(int) * TotalSections);
   SmallGlobalsTotalSize = (int *) SALLOC_PERM(sizeof(int) * TotalSections);
   if ((NULL == SectionSize) || (NULL == SmallGlobalsTotalSize)
       || (NULL == ScaleSection) || (NULL == SmallGlobals)) {
      (void) fprintf(stderr, "Out of virtual memory:"
                     " locals.c::PlaceStackData()\n");
      exit(EXIT_FAILURE);
   }
   bzero(SectionSize, TotalSections * sizeof(int));
   bzero(SmallGlobalsTotalSize, TotalSections * sizeof(int));
   bzero(SmallGlobals, GlobalsCount * sizeof(int));

   for (index = 0; index < GlobalsCount; ++index) {
      if (NULL != Globals[index]->CoalescedMSNeighbor)
         continue;
      section = Globals[index]->SectionNumber;
      assert(0 <= section);
      TempSize = GlobalChainSize(Globals[index]);
      assert(0 < TempSize);
      if (TempSize < L1_DATA_CACHE_SIZE) {
         SmallGlobals[index] = TRUE;
         SmallGlobalsTotalSize[section] += TempSize;
      }
      else if (TempSize >= L2_DATA_CACHE_SIZE) {
         MonsterGlobalsSize += TempSize;
      }
      SectionSize[section] += TempSize;
   }

   for (section = 0; section < TotalSections; ++section) {
      GlobalsSize += SectionSize[section];
      ScaleSection[section]
         = (L1_DATA_CACHE_SIZE < SmallGlobalsTotalSize[section]);
      ScalingNeeded |= ScaleSection[section];
#if DEBUG_PLACEMENT
      (void) fprintf(stderr, "Section %d total size: %d\n", section,
                     SectionSize[section]);
      (void) fprintf(stderr, "Section %d small globals size: %d \n", section,
                     SmallGlobalsTotalSize[section]);
#endif
   }
#if DEBUG_PLACEMENT
   (void) fprintf(stderr, "Globals total size: %d\n", GlobalsSize);
   (void) fprintf(stderr, "Mega-globals size: %d\n", MonsterGlobalsSize);
#endif

   /* Don't try to put monsters in the CacheMap. */
   GlobalsSize -= MonsterGlobalsSize;

   CacheMapSize = GlobalsSize;
   CacheMapLines = CACHE_LINES_FROM_SIZE(CacheMapSize) + 3;

   MaxPartitions 
      = ((GlobalsSize / L1_DATA_CACHE_WAY_SIZE) * 2) + TotalSections;

   PartitionGlobalsCount = (int *) SALLOC(sizeof(int) * MaxPartitions);
   PartitionShadowCount = (int *) SALLOC(sizeof(int) * MaxPartitions);
   PartitionStartingLine = (int *) SALLOC(sizeof(int) * MaxPartitions);
   if ((NULL == PartitionGlobalsCount)
       || (NULL == PartitionShadowCount)
       || (NULL == PartitionStartingLine)) {
      (void) fprintf(stderr, "Out of virtual memory:"
                     " locals.c::GlobalDataPlacement()\n");
      exit(EXIT_FAILURE);
   }
   bzero(PartitionGlobalsCount, sizeof(int) * MaxPartitions);
   bzero(PartitionShadowCount, sizeof(int) * MaxPartitions);
   bzero(PartitionStartingLine, sizeof(int) * MaxPartitions);

   ChacoGlobalsIndex = (int *) SALLOC((int) sizeof(int) * GlobalsCount);
   InverseGlobalsIndex = (int *) SALLOC((int) sizeof(int) * GlobalsCount);
   if ((NULL == ChacoGlobalsIndex) || (NULL == InverseGlobalsIndex)) {
      (void) fprintf(stderr, "Out of virtual memory:"
                     " locals.c::PlaceStackData()\n");
      exit(EXIT_FAILURE);
   }

   if (ScalingNeeded)
      ScaleFactor = ScaleAffinityArray(AffinityArray, GlobalsCount,
                                       ScaleSection);

   /* The big loop: Use Chaco to partition within each section. */
   for (section = 0; section < TotalSections; ++section) {
      if (L1_DATA_CACHE_WAY_SIZE >= SmallGlobalsTotalSize[section])
         goto BypassChaco;
      ChacoEdges = 0;
      ChacoVertices = 0;
      ChacoGlobals = 0;
      bzero(ChacoGlobalsIndex, GlobalsCount * sizeof(int));
      bzero(InverseGlobalsIndex, GlobalsCount * sizeof(int));

      for (gptr = globSymStart; gptr; gptr = gptr->next) {
         if (section > gptr->SectionNumber)
            continue;
         if (section < gptr->SectionNumber)
            break;  /* Sorted by section number, so we are done. */

         row = GetGlobalsIndex(gptr, Globals, GlobalsCount);
         assert(row >= 0);
         assert(row < GlobalsCount);
         if (NULL == gptr->CoalescedMSNeighbor) {
            TempSize = GlobalChainSize(gptr);
            assert(0 < TempSize);
            if (TempSize < L1_DATA_CACHE_SIZE) {
               InverseGlobalsIndex[row] = ChacoGlobals;
               ChacoGlobalsIndex[ChacoGlobals++] = row;
            }
         }
      } /* end for (gptr = globSymStart; ...) */

      if (1 >= ChacoGlobals) 
         goto BypassChaco;

      /* Using only those globals with TRUE entries in SmallGlobals[], 
       * create the input data structures for Chaco.  First, determine
       * the edge count.
       */
      for (row = 0; row < GlobalsCount; ++row) {
         if (!SmallGlobals[row] || (section != Globals[row]->SectionNumber))
            continue;
         AffIndex1 = Globals[row]->AffinityIndex;
         for (col = row + 1; col < GlobalsCount; ++col) {
            if (!SmallGlobals[col])
               continue;
            if (section != Globals[col]->SectionNumber)
               continue;
            AffIndex2 = Globals[col]->AffinityIndex;
            ChacoTempEdgeSum =
              *(AffinityArray + (AffIndex1 * GlobalsCount) + AffIndex2)
               + *(AffinityArray + (AffIndex2 * GlobalsCount) + AffIndex1);
            if (0.5 <= ChacoTempEdgeSum) {
               ++ChacoEdges;
            }
         }
      }

      if (1 > ChacoEdges)
         goto BypassChaco;

      SaveChacoEdges = 2 * ChacoEdges; /* Chaco wants A->B and B->A edges. */
      ChacoNDimensions = 
          (SmallGlobalsTotalSize[section] / L1_DATA_CACHE_WAY_SIZE);
      assert(0 <= ChacoNDimensions);
      if (0 >= ChacoNDimensions) {
         goto BypassChaco;  /* Partition by page size as enhancement, only if
                             * SmallGlobalsTotalSize > 4KB.
                             * !!!!!
                             */
      }

      /* Convert to power of two. i.e. compute the number of hypercube
       * dimensions for Chaco by rounding down to nearest power of two.
       */
      ChacoTemp = 1;
      while (1 < ChacoNDimensions) {
         ChacoNDimensions >>= 1;
         ++ChacoTemp;
      }
      ChacoNDimensions = ChacoTemp;
      ChacoNSection = ChacoNDimensions;
      if (3 < ChacoNSection)
         ChacoNSection = 3;
      ChacoDummiesNeeded =
          (((1 << ChacoNDimensions) * L1_DATA_CACHE_WAY_SIZE)
             - SmallGlobalsTotalSize[section]) / CACHE_LINE_PACKING_QUANTUM;

      if (NULL != PartitionSize)
         FREE(PartitionSize);
      PartitionSize =
         (int *) SALLOC(sizeof(int) * (2 + (1 << ChacoNDimensions)));
      if (NULL == PartitionSize) {
         (void) fprintf(stderr, "Out of virtual memory:"
                        " locals.c::PlaceStackData()\n");
         exit(EXIT_FAILURE);
      }
      bzero(PartitionSize, (2 + (1 << ChacoNDimensions)) * sizeof(int));

      ChacoTemp = ChacoGlobals + ChacoDummiesNeeded;
      if (NULL != ChacoStart) FREE(ChacoStart);
      ChacoStart = (int *) SALLOC((int) sizeof(int) * (ChacoTemp + 1));
      if (NULL != ChacoAdjacency) FREE(ChacoAdjacency);
      ChacoAdjacency = (int *) SALLOC((int)sizeof(int) * (2 * ChacoEdges + 1));
      if (NULL != ChacoVertexWeights) FREE(ChacoVertexWeights);
      ChacoVertexWeights = (int *) SALLOC((int) sizeof(int) * ChacoTemp);
      if (NULL != ChacoEdgeWeights) FREE(ChacoEdgeWeights);
      ChacoEdgeWeights = (float *) 
         SALLOC((int) sizeof(float) * (2 * ChacoEdges + 1));
      if (NULL != ChacoAssignment) FREE(ChacoAssignment);
      ChacoAssignment = (short *) SALLOC((int) sizeof(short) * ChacoTemp);
      if ((NULL == ChacoStart) || (NULL == ChacoAdjacency)
          || (NULL == ChacoVertexWeights) || (NULL == ChacoEdgeWeights)
          || (NULL == ChacoAssignment)) {
         (void) fprintf(stderr, "Out of virtual memory:"
                        " locals.c::PlaceStackData()\n");
         exit(EXIT_FAILURE);
      }
      (void) memset(ChacoStart, 0, sizeof(int) * (ChacoTemp + 1));
      (void) memset(ChacoAdjacency, 0, sizeof(int) * (2 * ChacoEdges + 1));
      (void) memset(ChacoVertexWeights, 0, sizeof(int) * ChacoTemp);
      (void) memset(ChacoEdgeWeights, 0, sizeof(float) * (2 * ChacoEdges + 1));

      /* Now, go through the AffinityArray[][] and create Chaco vertices
       * and edges, with edge weights and vertex weights.
       */
      ChacoEdges = 0;
      for (row = 0; row < GlobalsCount; ++row) {
         if (!SmallGlobals[row] || (section != Globals[row]->SectionNumber))
            continue;
         /* We have a vertex for Chaco. */
         AffIndex1 = Globals[row]->AffinityIndex;
         ChacoVertexWeights[ChacoVertices] = GlobalChainSize(Globals[row]);
         ChacoStart[ChacoVertices++] = ChacoEdges;
         for (col = 0; col < GlobalsCount; ++col) {
            if ((col == row) || (!SmallGlobals[col])
                || (section != Globals[col]->SectionNumber))
               continue;
            AffIndex2 = Globals[col]->AffinityIndex;
            ChacoTempEdgeSum =
              *(AffinityArray + (AffIndex1 * GlobalsCount) + AffIndex2)
               + *(AffinityArray + (AffIndex2 * GlobalsCount) + AffIndex1);
            if (0.5 <= ChacoTempEdgeSum) {
               /* We have an edge for Chaco.  Collapse the two directed
                * graph affinities from AffinityArray[row][col] and
                * AffinityArray[col][row] into a single edge weight.
                * For the adjacency list, remember that Chaco starts
                * numbering vertices at 1, not 0.
                */
               ChacoTempEdgeSum += 0.5;
               ChacoTempEdgeSum = (float) floor((double) ChacoTempEdgeSum);
               ChacoAdjacency[ChacoEdges] = 1 + InverseGlobalsIndex[col];
               ChacoEdgeWeights[ChacoEdges] = ChacoTempEdgeSum;
               ++ChacoEdges;
               assert(ChacoEdges <= SaveChacoEdges);
            }
         }
      }
      ChacoStart[ChacoVertices] = ChacoEdges;
      assert(ChacoEdges == SaveChacoEdges);

      /* We need to create dummy variables to fill up our Chaco globals list
       * to a power-of-two multiple of the partition size.  So, if we are
       * creating 16KB partitions and have 17KB of globals for Chaco, we need
       * 15KB of dummy globals.
       */
      for (index = 0; index < ChacoDummiesNeeded; ++index) {
         ChacoVertexWeights[ChacoVertices] = CACHE_LINE_PACKING_QUANTUM;
         ChacoStart[ChacoVertices++] = ChacoEdges;
      }
      ChacoStart[ChacoVertices] = ChacoEdges;
      /* NOTE: We are not incrementing ChacoGlobals, which will not count the
       * dummy globals added.  We don't care what partition Chaco assigns to
       * dummies.
       */

      /* We are finally ready to call Chaco's interface function to partition
       * the graph.
       */
      ChacoVMax = ChacoVertices / 8;
      if (4 > ChacoVMax) {
         if (0 >= ChacoVertices) goto BypassChaco;
         ChacoVMax = 4;
      }
      /* Chaco has some lower limit on VMax that I will enforce here to keep
       * warning messages from being emitted.  Shows up in check_input.c code
       * in Chaco.  Does not say why in the code, so I will just imitate it.
       */
      if ((1 << (ChacoNDimensions + 1)) > ChacoVMax)
         ChacoVMax = (1 << (ChacoNDimensions + 1));

      if (0 < ChacoVMax) {
#if DEBUG_PLACEMENT
         if (NULL == CacheMap)
            CacheMap = (struct globalsym **)
               SALLOC(sizeof(struct globalsym *) * CacheMapLines);
         if (NULL == CacheMap) {
            (void) fprintf(stderr, "Out of virtual memory:"
                           " locals.c::GlobalDataPlacement()\n");
            exit(EXIT_FAILURE);
         }
         BuildCacheMap(FALSE);
         interference = CacheInterferenceAffinity(1);
         (void) fprintf(stderr, "Global interference before Chaco: %12.3f \n", 
                        interference);
#endif
#define DEBUG_CHACO  TRUE
#if DEBUG_CHACO
         WriteChacoGraphFile(ChacoVertices, ChacoEdges, ChacoStart,
                             ChacoAdjacency, ChacoVertexWeights,
                             ChacoEdgeWeights);
#endif

#ifdef VET
         begin_trans(VET_OPT_TRANS);
#endif
         ChacoError = interface(ChacoVertices, ChacoStart, ChacoAdjacency,
                                ChacoVertexWeights, ChacoEdgeWeights, NULL,
                                NULL, NULL, NULL, NULL, ChacoAssignment,
                                0, ChacoNDimensions, ChacoMeshDimensions,
                                ChacoGoal, 1, 1, 0, ChacoVMax, ChacoNSection,
                                ChacoEigenTolerance, ChacoRandomSeed);
         assert(!ChacoError);
         (void) fprintf(stderr, "Chaco succeeded!\n");
#if !DEBUG_PLACEMENT
         rewind(stdout);
#endif
         /*-
          * Now that Chaco has partitioned the smaller globals into groups,
          * we need to do the following:  !!!!! Update comment
          * 1) Mark the PlacementPartition field for each global, with 0 up to
          *    N-1 for the N partitions that Chaco created, and N for globals
          *    too large to be part of the Chaco partitioning.
          * 2) Sort the globals list by partition number.
          * 3) Place partitions 0..N to minimize conflict misses.
          * 4) Perform immediate mode addressing optimization.
          */
         HighestPartition = (1 << ChacoNDimensions);
         for (gptr = globSymStart; NULL != gptr; gptr = gptr->next) {
            if (section > gptr->SectionNumber)
               continue;
            if (section < gptr->SectionNumber)
               break;
            if (L1_DATA_CACHE_WAY_SIZE <= gptr->size) {
               gptr->PlacementPartition 
                 = HighestPartition + NextMarkedPartition;
               HighestPartitionUsed = TRUE;
            }
         }

         /* Mark the Chaco variables with their newly assigned partition
          * numbers, to which we add the base NextMarkedPartition.
          */
         for (index = 0; index < ChacoGlobals; ++index) {
            row = ChacoGlobalsIndex[index];
            assert(0 <= row);
            assert(row < GlobalsCount);
            MarkPlacementPartition(Globals[row],
               NextMarkedPartition + (int) ChacoAssignment[index]);
         }

         /* Bump up all other sections' partition numbers above us. This
          * avoids problems with the globals list moving from partition 0
          * to 1 to 2 to 0 again for some section we have not reached yet.
          * This pattern causes problems in BuildCacheMap(TRUE).
          */
         if (HighestPartitionUsed)
            IncreasePlacementPartition(section + 1, HighestPartition + 1);
         else
            IncreasePlacementPartition(section + 1, HighestPartition);

         for (gptr = globSymStart; NULL != gptr; gptr = gptr->next) {
            if (section < gptr->SectionNumber)
               break;
            if (section == gptr->SectionNumber) {
               if (NULL == gptr->CoalescedMSNeighbor)
                  PartitionSize[gptr->PlacementPartition]
                     += GlobalChainSize(gptr);
            }
         }
         MaxChacoPartition = 0;
         for (index = 0; index < (1 << ChacoNDimensions); ++index) {
            assert(0 <= PartitionSize[index]);
            /* Either the partition has only dummies, or else it must have a
             * positive size.  If it has only dummies, bypass this iteration.
             */
            if (0 == PartitionSize[index]) {
               continue;
            }
            if (MaxChacoPartition < index)
               MaxChacoPartition = index;
         }
#ifdef VET
         end_trans();
#endif
         SortGlobalsByPartition(section); 
         GlobalsIndex = BuildGlobalsArray(L2_DATA_CACHE_SIZE);

         NextMarkedPartition += HighestPartition;
         if (HighestPartitionUsed)
            ++NextMarkedPartition;

#if DEBUG_PLACEMENT
         BuildCacheMap(FALSE);
         interference = CacheInterferenceAffinity(1);
         (void) fprintf(stderr, "Global interference after Chaco: %12.3f \n", 
                        interference);
#endif
         continue;  /* in for (section = 0; ...) */
      } /* end if (0 < ChacoVMax) */

   BypassChaco:

      /* Mark all globals in this section with NextMarkedPartition. If
       * there are any bigger than a Chaco partition (L1_DATA_CACHE_WAY_SIZE),
       * mark them with NextMarkedPartition + 1.
       */
#ifdef VET
      begin_trans(VET_OPT_TRANS);
#endif
      HighestPartitionUsed = FALSE;
      for (gptr = globSymStart; NULL != gptr; gptr = gptr->next) {
         if (section < gptr->SectionNumber)
            break;
         if ((section > gptr->SectionNumber)
             || (NULL != gptr->CoalescedMSNeighbor))
            continue;
         if (GlobalChainSize(gptr) >= L1_DATA_CACHE_WAY_SIZE) {
            HighestPartitionUsed = TRUE;
            MarkPlacementPartition(gptr, NextMarkedPartition + 1);
         }
         else
            MarkPlacementPartition(gptr, NextMarkedPartition);
      }
#ifdef VET
      end_trans();
#endif

      /* Bump up all other section's partition numbers above us. This
       * avoids problems with the globals list moving from partition 0
       * to 1 to 2 to 0 again for some section we have not reached yet.
       * This pattern causes problems in BuildCacheMap(TRUE).
       */
      if (HighestPartitionUsed) {
         NextMarkedPartition += 2;
         SortGlobalsByPartition(section); 
#ifdef VET
         begin_trans(VET_REQUIRED_TRANS);
#endif
         IncreasePlacementPartition(section + 1, 2);
#ifdef VET
         end_trans();
#endif
      }
      else {
         ++NextMarkedPartition;
#ifdef VET
         begin_trans(VET_REQUIRED_TRANS);
#endif
         IncreasePlacementPartition(section + 1, 1);
#ifdef VET
         end_trans();
#endif
      }

      (void) fprintf(stderr, "Section %d needs non-Chaco placement.\n",
                     section);
   } /* end for (section = 0; ...) */

   /* Build the CacheMap and determine the affinity conflict score. */
   if (NULL == CacheMap)
      CacheMap = (struct globalsym **)
         SALLOC(sizeof(struct globalsym *) * CacheMapLines);
   if (NULL == CacheMap) {
      (void) fprintf(stderr, "Out of virtual memory:"
                     " locals.c::GlobalDataPlacement()\n");
      exit(EXIT_FAILURE);
   }

   GlobalsIndex = BuildGlobalsArray(L2_DATA_CACHE_SIZE);
   BuildCacheMap(TRUE);
   interference = CacheInterferenceAffinity(1);

#if DEBUG_PLACEMENT
   (void) fprintf(stderr, "Interference before 2-opt: %12.3f \n",
                  interference);
#endif
   /* Do a local refinement to reduce the interference. */
   for (index = 0; index < TWO_OPT_ITERATIONS; ++index) {
      if ((interference / ScaleFactor) < 10.0) 
         break;   /* Any improvement would be trivial. */

      MinInterference = TwoOptRefinement(interference, 0, 
                                         NextMarkedPartition - 1);
      if (MinInterference >= (LOCAL_REFINEMENT_THRESHOLD * interference)) {
         if (NoBenefit) /* Two in a row with no benefit. */
            break;

         NoBenefit = TRUE;
#if DEBUG_PLACEMENT
         (void) fprintf(stderr, "Negligible benefit from 2-opt.\n");
#endif
      }
      else {
         interference = MinInterference;
         NoBenefit = FALSE;
#if DEBUG_PLACEMENT
         (void) fprintf(stderr, "Interference after 2-opt: %12.3f \n",
                        interference);
#endif
      }
   }

#if DEBUG_PLACEMENT
   (void) fprintf(stderr, "Interference after TwoOpt: %12.3f \n", 
                  interference);
#endif

   /* !!!!! Any further placement refinement goes here !!!!! */

cleanup:

#if 0 /* Keep these around for use in fixentry() later. */
   if (NULL != AffinityArray) FREE(AffinityArray);
   if (NULL != Globals) FREE(Globals);
   GlobalsCount = 0;
#endif
   if (NULL != PartitionStartingLine) FREE(PartitionStartingLine);
   if (NULL != PartitionGlobalsCount) FREE(PartitionGlobalsCount);
   if (NULL != PartitionShadowCount) FREE(PartitionShadowCount);

   return TRUE;

} /* end of GlobalDataPlacement() */

#endif  /* GLOBAL_DATA_PLACEMENT */

/*
 * GetBlockIndex - Find the index into the cblkArray that corresponds to
 * the given basic block pointer.  To save time, do a wrap-around circular
 * search starting at "start", which is the best guess at where to start
 * looking, based on whether it was a back edge or forward edge that led
 * to this block.
 */
int GetBlockIndex(struct bblock *cblk, struct bblock *cblkArray[],
                  int NumBasicBlocks, int start) {
   int index;

   for (index = start; index < NumBasicBlocks; ++index)
      if (cblk == cblkArray[index])
         return index;
   for (index = 0; index < start; ++index)
      if (cblk == cblkArray[index])
         return index;
   (void) fprintf(stderr, "Corrupt basic block array.\n");
   assert(FALSE);
   exit(EXIT_FAILURE);
} /* end of GetBlockIndex() */

/*
 * GetNodeIndex - Find the index into the CFGArray that corresponds to
 * the given CFG node pointer. The inlined CFGs are getting to be a million
 * nodes for some benchmarks!! This demands a hash table lookup rather than
 * something simpler.
 */
int GetNodeIndex(struct GlobalPlacementCFGNode *cnode,
                 struct hashedCFGNode **HashTable) {
   int hash;
   struct hashedCFGNode *tempNode;

   assert(NULL != cnode);
   assert(NULL != HashTable);

   hash = (((unsigned int) cnode) >> 5) & 4095;
   assert((0 <= hash) && (4096 > hash));

   tempNode = HashTable[hash];
   assert(NULL != tempNode);
   while (cnode != tempNode->node) {
      tempNode = tempNode->next;
      assert(NULL != tempNode);
   }

   return tempNode->index;
} /* end of GetNodeIndex() */

/*
 * TransferProfStructToCFG - Transfer the DataProfile info from a basic
 * block into the GlobalDataProfile struct for a CFG node.
 */
void TransferProfStructToCFG(struct GlobalDataProfile *CFGprof,
                             struct DataProfile *CBLKprof) {
   struct varl *tempVar = NULL;  /* From CFG.  */
   struct varl *tempVar2;        /* From cblk. */

   assert(NULL != CFGprof);
   assert(NULL != CBLKprof);

   CFGprof->LeftBackEdgeFlag = CBLKprof->LeftBackEdgeFlag;
   CFGprof->RightBackEdgeFlag = CBLKprof->RightBackEdgeFlag;

   CFGprof->FirstGlobal = CFGprof->LastGlobal = NULL;
   tempVar2 = CBLKprof->FirstLocal;
   while (NULL != tempVar2) {
      if (NULL == tempVar) {  /* First global in list. */
         CFGprof->FirstGlobal = (struct varl *)
            SALLOC_PERM(sizeof(struct varl));
         if (NULL == CFGprof->FirstGlobal) {
            (void) fprintf(stderr,
                          "Out of virtual memory: TransferProfStructToCFG\n");
            exit(EXIT_FAILURE);
         }
         tempVar = CFGprof->FirstGlobal;
      }
      else {
         tempVar->next = (struct varl *) SALLOC_PERM(sizeof(struct varl));
         if (NULL == tempVar->next) {
            (void) fprintf(stderr,
                          "Out of virtual memory: TransferProfStructToCFG\n");
            exit(EXIT_FAILURE);
         }
         tempVar = tempVar->next;
      }
      tempVar->index = tempVar2->index;
      tempVar->name = CALLOC_PERM(strlen(tempVar2->name));
      if (NULL == tempVar->name) {
         (void) fprintf(stderr,
                        "Out of virtual memory: TransferProfStructToCFG\n");
         exit(EXIT_FAILURE);
      }
      strcpy(tempVar->name, tempVar2->name);
      tempVar->cgnode = tempVar2->cgnode;
      tempVar->next = NULL;
      tempVar2 = tempVar2->next;
   } /* end while (NULL != tempVar2) */
   CFGprof->LastGlobal = tempVar;
   return;
} /* end of TransferProfStructToCFG() */

/*
 * GetCFGLoopID - Given a pointer to a loop (in the RTL form of the code),
 * traverse the RTLLoopTree, find it, and return the unique CFG loop ID
 * for that loop.
 */
int GetCFGLoopID(struct loopl *loop, struct RTLLoopData *TreeNode) {
   assert(NULL != TreeNode);

   if (loop == TreeNode->loop)
      return TreeNode->ID;
   else if (loop < TreeNode->loop)
      return GetCFGLoopID(loop, TreeNode->left);
   else {
      assert(loop > TreeNode->loop);
      return GetCFGLoopID(loop, TreeNode->right);
   }

} /* end of GetCFGLoopID() */


/*
 * BuildCFGLoopList - Build a list of loops containing the given CFG node,
 *  using the given looplist from its corresponding basic block, and attach
 *  the list to the LoopList field in the CFG node.
 */
void BuildCFGLoopList(struct GlobalPlacementCFGNode *CFGnode,
                      struct listl *looplist) {
   struct CFGloopl *loop1;
   struct CFGloopl *loop2 = NULL;
   struct CFGloopl *loop3;
   struct listl *blkloopl;

   assert(NULL != CFGnode);
   assert(NULL != looplist);

   /* Note: We want the looplist sorted by number of blocks, from most to
    * fewest blocks, which will put outer loops before inner loops the
    * list. We keep them sorted in that way in the CFG loop list.
    */
   for (blkloopl = looplist; blkloopl; blkloopl = blkloopl->next) {
      loop1 = (struct CFGloopl *) SALLOC_PERM(sizeof(struct CFGloopl));
      if (NULL == loop1) {
         (void) fprintf(stderr, "Out of virtual memory: BuildCFGLoopList()\n");
         exit(EXIT_FAILURE);
      }
      loop1->loopID = GetCFGLoopID(blkloopl->ptr, RTLLoopTree);
      loop1->blockCount = blkloopl->ptr->blockCount;
      if (blkloopl->ptr->valid && ('\0' != blkloopl->ptr->iterations[0])
          && isdigit(blkloopl->ptr->iterations[0]))
         loop1->iterations = atoi(blkloopl->ptr->iterations);
      else
         loop1->iterations = (int) DEFAULT_LOOP_ITERATIONS;
      loop1->next = NULL;
      /* Attach loop1 to the list, sorted by blockCount. */
      if (NULL == CFGnode->LoopList) /* First item. */
         CFGnode->LoopList = loop1;
      else {
         /* Insert by blockCount field. */
         for (loop2 = CFGnode->LoopList; loop2; loop2 = loop2->next) {
            if (loop1->blockCount > loop2->blockCount) {
               /* Insert at beginning. */
               loop1->next = loop2;
               CFGnode->LoopList = loop1;
               break;
            }
            else if (NULL == loop2->next) {
               /* Insert at end. */
               loop2->next = loop1;
               break;
            }
            else if (loop1->blockCount > loop2->next->blockCount) {
               /* Insert after loop2 and before loop2->next. */
               loop1->next = loop2->next;
               loop2->next = loop1;
               break;
            }
         } /* end for (loop2 = ... ) */
      } /* end if (NULL == CFGnode->LoopList) */
   } /* end for (blkloopl = ... ) */
   return;
} /* end of BuildCFGLoopList() */

/*
 * TransferProfileToCFG - Transfer the struct DataProfile info from all
 * basic blocks into corresponding CFG (control flow graph) nodes.
 */
void TransferProfileToCFG() {
   struct GlobalPlacementCFGNode *CFGnode;
   struct GlobalPlacementCFGNode **CFGArray;
   int index;
   int childIndex;
   struct bblock *cblk;
   struct bblock **cblkArray;

   /* Number the basic blocks so we can hold pointers to them in an array. */
   NumBasicBlocks = number_basic_blocks();
   assert(0 < NumBasicBlocks);

   /* Make an array of pointers to all basic blocks in the function,
    * indexed by their basic block numbers.  Adjust basic block numbers to
    * be zero-based, rather than 1-based, as we process them.
    */
   --NumBasicBlocks;
   cblkArray = (struct bblock **)
      SALLOC(NumBasicBlocks * sizeof(struct bblock *));
   if (NULL == cblkArray) {
      (void) fprintf(stderr, "Out of virtual memory: TransferProfileToCFG\n");
      exit(EXIT_FAILURE);
   }
   for (cblk = top; NULL != cblk; cblk = cblk->next) {
      assert(0 < cblk->bnum);  /* Currently 1-based. */
      --cblk->bnum; /* Make it zero-based. */
      cblkArray[cblk->bnum] = cblk;
   }

   /* Make a CFG node corresponding to each basic block. */
   CFGArray = (struct GlobalPlacementCFGNode **)
      SALLOC(NumBasicBlocks * sizeof(struct GlobalPlacementCFGNode *));
   if (NULL == CFGArray) {
      (void) fprintf(stderr, "Out of virtual memory: TransferProfileToCFG\n");
      exit(EXIT_FAILURE);
   }
   for (index = 0; index < NumBasicBlocks; ++index) {
      assert(NULL != cblkArray[index]);
      CFGArray[index] = (struct GlobalPlacementCFGNode *)
         SALLOC_PERM(sizeof(struct GlobalPlacementCFGNode));
      if (NULL == CFGArray[index]) {
         (void) fprintf(stderr,
                        "Out of virtual memory: TransferProfileToCFG\n");
         exit(EXIT_FAILURE);
      }
      CFGArray[index]->estimate = cblkArray[index]->estimate;
      CFGArray[index]->left = NULL;
      CFGArray[index]->right = NULL;
      CFGArray[index]->prof = NULL;
      CFGArray[index]->LoopList = NULL;
      CFGArray[index]->done = FALSE;
      if (NULL != cblkArray[index]->prof) {
         CFGArray[index]->prof = (struct GlobalDataProfile *)
            SALLOC_PERM(sizeof(struct GlobalDataProfile));
         if (NULL == CFGArray[index]->prof) {
            (void) fprintf(stderr,
                           "Out of virtual memory: TransferProfileToCFG\n");
            exit(EXIT_FAILURE);
         }
         TransferProfStructToCFG(CFGArray[index]->prof,
                                 cblkArray[index]->prof);
      }
      if (NULL != cblkArray[index]->looplist)
         BuildCFGLoopList(CFGArray[index], cblkArray[index]->looplist);
   }

   /* Go back and set the "next" fields for the CFG nodes. */
   for (index = 0; index < (NumBasicBlocks - 1); ++index)
      CFGArray[index]->next = CFGArray[index + 1];
   CFGArray[NumBasicBlocks - 1]->next = NULL;

   /*
    * Now we need to set the left and right child links in each CFG node
    * to correspond to the left and right child links in the basic blocks.
    * This could be sped up by setting up a hash table that maps basic
    * block pointers to array indices (i.e. cblk->bnum values), should
    * this computation prove to be a bottleneck.  We will try to speed
    * things up by paying attention to whether the link we are following
    * is a back edge or not.
    */
   for (index = 0; index < NumBasicBlocks; ++index) {
      cblk = cblkArray[index];
      if (NULL != cblk->left) {
         if ((NULL != cblk->prof) && (!cblk->prof->LeftBackEdgeFlag))
            childIndex = GetBlockIndex(cblk->left, cblkArray, NumBasicBlocks,
                                       index);
         else
            childIndex = GetBlockIndex(cblk->left, cblkArray, NumBasicBlocks,
                                       0);
         CFGArray[index]->left = CFGArray[childIndex];
      }
      if (NULL != cblk->right) {
         if ((NULL != cblk->prof) && (!cblk->prof->RightBackEdgeFlag)) {
            childIndex = GetBlockIndex(cblk->right, cblkArray, NumBasicBlocks,
                                       index);
            CFGArray[index]->right = CFGArray[childIndex];
         }
         else {
            /* Get rid of right back edges. We will never follow one
             * in any of our processing.
             */
            CFGArray[index]->right = NULL;
            CFGArray[index]->prof->RightBackEdgeFlag = FALSE;
         }
      }
   }

   /* Make savedCFGNode point at the node for the top block. */
   savedCFGNode = CFGArray[top->bnum];

   /* Now we no longer need our arrays. */
   FREE(cblkArray);
   FREE(CFGArray);

   return;
} /* end of TransferProfileToCFG() */

/*
 * FreeRTLLoopTree - Recursively free the RTLLoopTree data structure.
 */
void FreeRTLLoopTree(struct RTLLoopData *TreeNode) {
   assert(NULL != TreeNode);

   if (NULL != TreeNode->left)
      FreeRTLLoopTree(TreeNode->left);
   if (NULL != TreeNode->right)
      FreeRTLLoopTree(TreeNode->right);
   FREE(TreeNode);
   return;
} /* end of FreeRTLLoopTree() */

/*
 * InsertRTLLoopInTree - Insert the given loop ptr into the binary tree
 * pointed to by the global static variable RTLLoopTree.  It is assumed that
 * there are not enough loops per function to make it necessary to maintain
 * good tree balance with advanced data structures such as an AVL tree.  If
 * the facts prove otherwise, this could be enhanced.
 */
void InsertRTLLoopInTree(struct loopl *loop, struct RTLLoopData **TreeNode) {
   assert(NULL != TreeNode);

   if (NULL == *TreeNode) {
      *TreeNode =
          (struct RTLLoopData *) SALLOC_PERM(sizeof(struct RTLLoopData));
      if (NULL == *TreeNode) {
         (void) fprintf(stderr, "Out of memory: InsertRTLLoopInTree().\n");
         exit(EXIT_FAILURE);
      }
      (*TreeNode)->loop = loop;
      (*TreeNode)->ID = CFGLoopID++;
      (*TreeNode)->left = NULL;
      (*TreeNode)->right = NULL;
   }
   else if (loop > (*TreeNode)->loop)
      InsertRTLLoopInTree(loop, &((*TreeNode)->right));
   else {
      assert(loop < (*TreeNode)->loop);
      InsertRTLLoopInTree(loop, &((*TreeNode)->left));
   }
   return;
} /* end of InsertRTLLoopInTree() */

/* CreateCFGLoopData -
 * Create the data structures containing info about each
 * loop, so that we can track loop nesting, loop iterations, etc., for
 * each CFG node.
 */
void CreateCFGLoopData() {
   int numLoops = 0;
   struct loopl *loopPtr;

   /* Deallocate old data. */
   if (NULL != RTLLoopTree) {
      FreeRTLLoopTree(RTLLoopTree);
      RTLLoopTree = NULL;
   }

   /* Insert the loops from this function into the tree. */
   for (loopPtr = loops; loopPtr; loopPtr = loopPtr->next) {
      InsertRTLLoopInTree(loopPtr, &RTLLoopTree);
   }

   return;
} /* end of CreateCFGLoopData() */

/*
 * BuildBlockGlobalProfile() - build the DataProfile structure in each block
 * to list the first through last globals accessed, and function calls made,
 * within each block, with edge weights and node weights taken
 * from the estimate field within each basic block.
 * Return count of global variables that are accessed from memory.
 */
void BuildBlockGlobalProfile(int MaxSize) {
   struct globalsym *globptr;
   struct bblock *cblk;
   struct bblock *sblk; /* successors to cblk */
   struct list *rtl;
   struct varl *cur_var;
   char *c;
   char *name;
   int successors;
   char expandedRTL[BIGSTRING];
   int loadFlag;
   int callFlag;

   /* If we are not inside a function, just return. */
   if (NULL == fn)
      return;

   savedCFGNode = NULL;

   /* We really cannot do anything with a function that has an irreducible
    * flow graph.  estimate_execution_frequency() does not complete its
    * filling in of the estimate fields for these functions.  Our traversal
    * of the function's blocks below is an imitation of that code and will
    * have the same problems.
    */
   IrreducibleFlag = estimate_execution_frequency(1);
   if (IrreducibleFlag) {
      savedCFGNode = (struct GlobalPlacementCFGNode *)
         SALLOC_PERM(sizeof(struct GlobalPlacementCFGNode));
      if (NULL == savedCFGNode) {
         (void) fprintf(stderr,
               "Out of virtual memory: locals.c::BuildBlockGlobalProfile()\n");
         exit(EXIT_FAILURE);
      }
      savedCFGNode->left = NULL;
      savedCFGNode->right = NULL;
      savedCFGNode->estimate = 1.0;
      return;
   }

   /* Create the hash table of data structures containing info about each
    * loop, so that we can track loop nesting, loop iterations, etc., for
    * each CFG node.
    */
   CreateCFGLoopData();

   /* Allocate the DataProfile struct, and reset the status, for each block. */
   for (cblk = top; cblk; cblk = cblk->next) {
      /* This function can be called more than once, to create profiles with
       * different MaxSize thresholds for different coalescing and placement
       * purposes.  If there is profiling data from a previous execution,
       * deallocate the data structures.
       */
      if (NULL != cblk->prof) {
         DeallocateDataProfile(cblk);
      }
      cblk->prof =
        (struct DataProfile *) SALLOC((int) sizeof(struct DataProfile));
      if (NULL == cblk->prof) {
         (void) fprintf(stderr,
               "Out of virtual memory: locals.c::BuildBlockGlobalProfile()\n");
         exit(EXIT_FAILURE);
      }
      cblk->prof->FirstLocal = NULL;
      cblk->prof->LastLocal = NULL;
      cblk->status &= ~DONE;
   }
   /* Imitating the traversal logic of 
    * controlflow.c::estimate_execution_frequency(), initialize the fields
    * within the DataProfile struct for each block.
    */
   for (cblk = top; cblk; cblk = cblk->next) {
      if (DONE & cblk->status) continue;

      cblk->prof->CaseFlag = (cblk->rtlend && (cblk->rtlend->type & CASELINE));

      if (cblk->left) {
         cblk->prof->LeftBackEdgeFlag = dominates(cblk->left, cblk);
         if (cblk->prof->LeftBackEdgeFlag) {
            cblk->prof->LeftEdgeWeight = 
              (float) inner_loop_iterations(cblk) - 1;
            /* We loop back N-1 times for an N-iteration loop. */
         }
      }
      else {
         cblk->prof->LeftEdgeWeight = 0.0;
         cblk->prof->LeftBackEdgeFlag = FALSE;
      }

      if (cblk->right) {
         cblk->prof->RightBackEdgeFlag = dominates(cblk->right, cblk);
         if (cblk->prof->RightBackEdgeFlag) {
            cblk->prof->RightEdgeWeight = 
               (float) inner_loop_iterations(cblk) - 1;
            /* We loop back N-1 times for an N-iteration loop. */
         }
      }
      else {
         cblk->prof->RightEdgeWeight = 0.0;
         cblk->prof->RightBackEdgeFlag = FALSE;
      }

      /*
       * Now we determine how many children this basic block has.  We
       * must make a special case for switch lookup tables.  Note that
       * these lookup tables are made up of a chain of basic blocks,
       * but that for the purposes of execution estimation we will
       * treat such a sequence as if it were one single basic block.
       */
      successors = 0;
      if (cblk->prof->CaseFlag) {
         if (!cblk->prof->RightBackEdgeFlag)
            ++successors;
         for (sblk = cblk->left;
              sblk->rtlend && (sblk->rtlend->type & CASELINE);
              sblk = sblk->left) {
            if (!dominates(sblk->right, cblk))
               ++successors;
         }
      }
      else {
         if (cblk->left && !cblk->prof->LeftBackEdgeFlag)
            ++successors;
         if (cblk->right && !cblk->prof->RightBackEdgeFlag)
            ++successors;
      }

      /*
       * Now we take the reciprocal of the number of successors of this
       * block, not counting back edges, and assign it to each forward edge
       * as the edge weight.
       */
      if (cblk->prof->CaseFlag) {
         cblk->prof->LeftEdgeWeight = 1.0;
         if (!cblk->prof->RightBackEdgeFlag)
            cblk->prof->RightEdgeWeight = (1.0 / successors);
         for (sblk = cblk->left;
              sblk->rtlend && (sblk->rtlend->type & CASELINE);
              sblk = sblk->left) {
            sblk->status |= DONE;
            sblk->prof->CaseFlag = TRUE;
            sblk->prof->LeftBackEdgeFlag = FALSE;
            sblk->prof->RightBackEdgeFlag = dominates(sblk->right, cblk);
            sblk->prof->LeftEdgeWeight = 1.0;
            if (!sblk->prof->RightBackEdgeFlag)
               sblk->prof->RightEdgeWeight = (1.0 / successors);
         }
      }
      else {
         if (cblk->left && !cblk->prof->LeftBackEdgeFlag)
            cblk->prof->LeftEdgeWeight = (1.0 / successors);
         if (cblk->right && !cblk->prof->RightBackEdgeFlag)
            cblk->prof->RightEdgeWeight = (1.0 / successors);
      }

      /* Now we go through all global references in this block, recording
       * all globals in the chain from cblk->prof->FirstLocal to ...->LastLocal.
       * NOTE: we are using the "FirstLocal" and "LastLocal" names that will
       * later be used for profiling local variables.
       * The first and last globals may be the same; there may be only one
       * global, or none.
       */
      cur_var = cblk->prof->FirstLocal;
      for (rtl = cblk->rtls; rtl; rtl = rtl->next) {
         if (NULL != (c = cnts(rtl))) {
            callFlag = is_call(c);
            if (!callFlag && (NULL == gtmem(c)))
               continue;  /* No calls, loads or stores in this RTL. */
            loadFlag = is_a_load(c);
#if IGNORE_STORES
            if (!loadFlag && !callFlag)
               continue;
#endif
            (void) strcpy(expandedRTL, rtl->val);
            expand(cblk->loop, rtl, expandedRTL);
            c = expandedRTL;
            while (NULL != (c = gtglobal(c))) {
#if IGNORE_STORES
               if (*(c + 1) == '=')  /* Store, not load. */
                  continue;
#endif
               globptr = findGlobalFromCode(c - GLOBALLOCLENGTH);
               assert(NULL != globptr);
               /* We have found a reference to a global variable.
                * gtglobal() returned with "c" pointing just past
                * the global in the RTL. "globptr" points to the globalsym
                * record for the global.
                */
#if NO_PROFILE_SIZE_LIMIT
               /* Experimenting with deprecating the size limit feature. */
               if (globptr->size >= MaxSize)
                  continue;  /* Not eligible for this profile; too big. */
#endif

               /* Eliminate duplication in consecutive accesses. */
               name = (globals + globptr->gallocIndex);
               if (cur_var && (0 == strcmp(name, cur_var->name)))
                  continue;

               if (NULL == cur_var) { /* Need to allocate first global. */
                  assert(NULL == cblk->prof->FirstLocal);
                  cblk->prof->FirstLocal =
                    (struct varl *) SALLOC((int) sizeof(struct varl));
                  if (NULL == (cur_var = cblk->prof->FirstLocal)) {
                     (void) fprintf(stderr, "Out of virtual memory:"
                                    " locals.c::BuildBlockGlobalProfile()\n");
                     exit(EXIT_FAILURE);
                  }
               }
               else {
                  cur_var->next = 
                    (struct varl *) SALLOC((int) sizeof(struct varl));
                  if (NULL == cur_var->next) {
                     (void) fprintf(stderr, "Out of virtual memory:"
                                    " locals.c::BuildBlockGlobalProfile()\n");
                     exit(EXIT_FAILURE);
                  }
                  cur_var = cur_var->next;
               }
               cur_var->index = 0;  /* Will be filled in at EOF. */
               cur_var->name  = (globals + globptr->gallocIndex);
               cur_var->next  = NULL;
               cur_var->cgnode  = NULL;
            } /* end while (c = gtglobal(c)) */
         } /* end if (c = cnts(rtl)) */
      } /* end for (rtl = cblk->rtls; ...) */

      cblk->prof->LastLocal = cur_var;
      cblk->status |= DONE;
   } /* end for (cblk = top; cblk; cblk = cblk->next) */

   for (cblk = top; cblk; cblk = cblk->next) {
      cblk->status &= ~DONE;
   }

   /* Transfer the profiling info into the control flow graph nodes
    * for this function within the call graph.
    */
   TransferProfileToCFG();

   return;
} /* end of BuildBlockGlobalProfile() */

/*
 * BypassCFG - See if CFG nodes that access no globals can be bypassed
 * by their predecessors.  Mark as done which ever nodes are reached
 * and leave unmarked those that are bypassed. Recurse to successors.
 * Given node is predecessor; we start looking with its successors.
 */
void BypassCFG(struct GlobalPlacementCFGNode *cnode) {
   struct GlobalPlacementCFGNode *tempNode;
   int LeftFlag;
   int RightFlag;
   int LoopFlag;   /* Is cnode in a loop? */
   int RightBackEdgeToBypassedNode;
   int LeftBackEdgeToBypassedNode;
   int LeftBackEdge;
   int LeftSelfEdge;
   int SameRightChild;
   int SameLeftChild;
   int change = FALSE;

   assert(NULL != cnode);
   assert(NULL != cnode->prof);

   LeftFlag = (NULL != cnode->left);
   RightFlag = (NULL != cnode->right);
   LeftBackEdge = LeftFlag && cnode->prof->LeftBackEdgeFlag;
   LeftSelfEdge = LeftBackEdge && (cnode == cnode->left);
   SameRightChild = RightFlag && LeftFlag 
                     && (cnode->right == cnode->left->right);
   SameLeftChild = RightFlag && LeftFlag 
                     && (cnode->left == cnode->right->left);
   cnode->done = TRUE;

   if (!LeftFlag && !RightFlag)
      return;

   LoopFlag = (1.0 < cnode->estimate);

   do {
      if (change) { /* Not the first iteration. */
         LeftFlag = (NULL != cnode->left);
         RightFlag = (NULL != cnode->right);
         LeftBackEdge = LeftFlag && cnode->prof->LeftBackEdgeFlag;
         LeftSelfEdge = LeftBackEdge && (cnode == cnode->left);
         if (LeftFlag && RightFlag && (cnode->left == cnode->right)) {
            /* Bypassing has made our left child be the same as our right. */
            RightFlag = FALSE;
            cnode->right = NULL;
            cnode->prof->RightBackEdgeFlag = FALSE;
         }
         /* If the node has a left back edge to itself and has a right child,
          * we can eliminate the left self edge and make the right child
          * be the left child instead.
          */
         if (LeftSelfEdge && RightFlag) {
            cnode->left = cnode->right;
            cnode->prof->LeftBackEdgeFlag = cnode->prof->RightBackEdgeFlag;
            RightFlag = FALSE;
            cnode->right = NULL;
            cnode->prof->RightBackEdgeFlag = FALSE;
            LeftFlag = TRUE;
            LeftBackEdge = cnode->prof->LeftBackEdgeFlag;
            LeftSelfEdge = LeftBackEdge && (cnode == cnode->left);
         }
         SameRightChild = RightFlag && LeftFlag 
                          && (cnode->right == cnode->left->right);
         SameLeftChild = RightFlag && LeftFlag 
                          && (cnode->left == cnode->right->left);
      }
      change = FALSE;

      /* We can be fooled into thinking that a successor has a significant
       * right child of its own, when in fact it is just a right back edge
       * back to some node we have already decided to bypass.
       * NOTE: Now that we have removed right back edges, this will come
       * out FALSE and do no harm, but we leave it in case we change our
       * design decisions in the future about removing back edges.
       */
      RightBackEdgeToBypassedNode 
         = (RightFlag && cnode->right->prof->RightBackEdgeFlag
            && (NULL != cnode->right->right)
            && (cnode->right->right->bypassed));
      /* Ditto for left child chains that follow back edges. We will not
       * eliminate left back edges, so these can potentially cause an
       * infinite loop if the entire chain has no globals and we keep
       * bypassing nodes while following left back edges in a circle.
       */
      LeftBackEdgeToBypassedNode 
         = (LeftFlag && cnode->left->prof->LeftBackEdgeFlag
            && (NULL != cnode->left->left)
            && (cnode->left->left->bypassed));

      if (LeftFlag && RightFlag) {
         /*-
          * Both successors exist. One can be bypassed only if it has a
          * single unconditional successor, and it accesses no globals,
          * and bypassing it does not cause us to merge two loops that
          * this non-loop node was separating (which would mess up our
          * computation of conflict affinities in profiling by making it
          * look like the globals in the two loops can thrash each other).
          * The "single unconditional successor" criterion can be relaxed
          * if our left child has two children, but its right child is the
          * same as our right child:
          *
          *        O
          *       / \
          *      /   \
          *     O---->O
          *    /
          *   /
          *  O
          *
          * Here, the path that goes through the left child to the right
          * child will never contribute anything new to global profiling if
          * the left child has no globals.  The mirror image case is
          * detected when SameLeftChild is TRUE, i.e. the right child
          * has the left child as its own left child.
          */
         if (((NULL == cnode->left->prof)
              || (NULL == cnode->left->prof->FirstGlobal))
             && ((NULL == cnode->left->right) || SameRightChild)
             && (NULL != cnode->left->left)
             && (!LeftBackEdgeToBypassedNode)
             && (!(LoopFlag && (1.0 >= cnode->left->estimate)
                   && (1.0 < cnode->left->left->estimate)))) {
            /* We can bypass this node. */
            tempNode = cnode->left;
            cnode->left = tempNode->left;  /* Bypass useless node. */
            cnode->prof->LeftBackEdgeFlag |= tempNode->prof->LeftBackEdgeFlag;
            change = TRUE;
            tempNode->bypassed = TRUE;
         }
         if (((NULL == cnode->right->prof)
              || (NULL == cnode->right->prof->FirstGlobal))
             && ((NULL == cnode->right->right) || RightBackEdgeToBypassedNode
                 || SameLeftChild)
             && (NULL != cnode->right->left)
             && (!(LoopFlag && (1.0 >= cnode->right->estimate)
                   && (1.0 < cnode->right->left->estimate)))) {
            /* We can bypass this node. */
            tempNode = cnode->right;
            if (SameLeftChild) {
               cnode->right = tempNode->right;
               cnode->prof->RightBackEdgeFlag
                  |= tempNode->prof->RightBackEdgeFlag;
            }
            else {
               cnode->right = tempNode->left;
               cnode->prof->RightBackEdgeFlag 
                  |= tempNode->prof->LeftBackEdgeFlag;
            }

            /* Don't create right back edges. */
            if (cnode->prof->RightBackEdgeFlag) {
               cnode->right = NULL;
               cnode->prof->RightBackEdgeFlag = FALSE;
            }
            change = TRUE;
            tempNode->bypassed = TRUE;
         }
      }
      else if (LeftFlag && (NULL == cnode->right)) {
         /* Only the left successor exists. It can be bypassed if it has no
          * globals, even if it has two successors. Avoid bypassing the 
          * bottom node, as it is useful.  If the node separates two loops,
          * the top one being cnode (i.e. LoopFlag == TRUE) and the bottom
          * being EITHER the left or right child of cnode->left, then do not
          * attempt to bypass, as this merges loops (see comment above).
          * Do not attempt to bypass a node that has a left back edge to
          * itself, as we will go into an infinite loop, and we would not
          * be "bypassing" node A to get to node B if A == B, anyway.
          */
         if (((NULL == cnode->left->prof)
              || (NULL == cnode->left->prof->FirstGlobal)) /* No globals. */
             && (NULL != cnode->left->left) /* Not bottom. */
             && (!LeftBackEdgeToBypassedNode)
             && (cnode->left != cnode->left->left) /* Not back to itself. */
             && (!(LoopFlag && (1.0 >= cnode->left->estimate)
                   && ((1.0 < cnode->left->left->estimate)
                       || ((NULL != cnode->left->right)
                           && (1.0 < cnode->left->right->estimate)))))) {
            /* We can bypass this node. */
            tempNode = cnode->left;
            cnode->left = tempNode->left;
            /* If we are bypassing a node that dominates us, i.e. our left
             * edge was a back edge, then we don't want to create a right
             * back edge.  In that case, we leave our right child set to NULL.
             */
            if (!LeftBackEdge && (NULL != tempNode->right)) {
               cnode->right = tempNode->right;
               cnode->prof->RightBackEdgeFlag = 
                   tempNode->prof->RightBackEdgeFlag;
               assert(!cnode->prof->RightBackEdgeFlag);
            }
            else {
               cnode->prof->RightBackEdgeFlag = FALSE;
            }
            cnode->prof->LeftBackEdgeFlag |= tempNode->prof->LeftBackEdgeFlag;
            change = TRUE;
            tempNode->bypassed = TRUE;
         }
      } /* end if (LeftFlag && RightFlag) ... else if ... */
   } while (change);

   /* Recurse into successors. */
   LeftFlag = ((NULL != cnode->left) && (!cnode->left->done)
#if 0
               && (!cnode->prof->LeftBackEdgeFlag)
#endif
              );
   if (LeftFlag)
      BypassCFG(cnode->left);

   RightFlag = ((NULL != cnode->right) && (!cnode->right->done));
   if (RightFlag)
      BypassCFG(cnode->right);

} /* end of BypassCFG() */

/*
 * PruneCFG - Use BypassCFG() to mark the nodes that must remain in the CFG,
 * because they are reachable from the top node, have globals in them, or
 * are the bottom node or a combination join/split node (more than one
 * predecessor, more than one successor; hard to bypass).  Then delete
 * the unmarked nodes to make the CFG smaller before we start inlining
 * CFGs into each other.
 * To simplify, because we never follow back edges, just delete any right
 * edge that is a back edge.
 * Input parameter CFGnode should be the top node of a function's CFG.
 */
void PruneCFG(struct call_graph_node *cgnode) {
   struct GlobalPlacementCFGNode *CFGnode;
   struct GlobalPlacementCFGNode *cnode;
   struct GlobalPlacementCFGNode *TempNode;
   int NodesDeleted = 0;

   assert(NULL != cgnode);
   CFGnode = cgnode->topCFG;
   cnode = CFGnode;
   assert(NULL != cnode);

   /* Prepare for BypassCFG() by marking all nodes as NOT done. */
   while (NULL != cnode) {
      cnode->done = FALSE;
      /* Delete right back edges. */
      if ((NULL != cnode->right) && cnode->prof->RightBackEdgeFlag) {
         cnode->right = NULL;
         cnode->prof->RightBackEdgeFlag = FALSE;
      }
      cnode = cnode->next;
   }

   /* Bypass unneeded or unreachable nodes; mark the others as done. */
   BypassCFG(CFGnode);

   /* Audit the results.  No left or right child pointers should be dangling
    * or aimed at a node that is not marked as done, unless it is a child
    * pointer from a node that is also not marked as done.
    */
   cnode = CFGnode;
   while (NULL != cnode) {
      if (cnode->done) {
         if (NULL != cnode->left) {
            assert(cnode->left->done);
         }
         if (NULL != cnode->right) {
            assert(cnode->right->done);
            assert(!cnode->prof->RightBackEdgeFlag);
         }
      }
      cnode = cnode->next;
   }

   /* Eliminate the unmarked nodes. */
   cnode = CFGnode;
   while (NULL != cnode->next) { /* Looking ahead one node. */
      /* If the left and right successors of cnode have become the same,
       * we can get rid of the right pointer to cut down unnecessary
       * recursion in the future.
       */
      if ((NULL != cnode->right) && (cnode->right == cnode->left)) {
         cnode->right = NULL;
         cnode->prof->RightBackEdgeFlag = FALSE;
      }

      if (!cnode->next->done) { /* Bypassed and unmarked. */
         if (NULL != cnode->next->prof->FirstGlobal) {
            (void) fflush(stderr);
            DumpCFG(cgnode);
         }
         TempNode = cnode->next;
         cnode->next = TempNode->next; /* Bypass with next link. */
         assert(NULL == TempNode->prof->FirstGlobal);
         if (NULL != TempNode->LoopList)
            DeallocateCFGLoopList(TempNode);
         FREE(TempNode->prof); /* NOTE: no global list to free here. */
         FREE(TempNode);
         ++NodesDeleted;
         if (NULL == cnode->next) /* Bypassed last node (in next list). */
            break;
      }
      else 
         cnode = cnode->next;
   }
#if DEBUG_PROFILING
   (void) fprintf(stderr, "%d CFG nodes deleted.\n", NodesDeleted);
#endif

   return;
} /* end of PruneCFG() */

/*
 * CountGlobalsInFunction - Recursively count the global variables (not
 * including function calls) in a CFG.  Return as soon as we find one;
 * the purpose is just to determine if the function has become irrelevant
 * after removing library calls, etc.
 */
int CountGlobalsInFunction(struct GlobalPlacementCFGNode *CFGnode) {
   int count = 0;
   struct varl *tempVar;
 
   assert(NULL != CFGnode);

   while (NULL != CFGnode) {
      if ((NULL != CFGnode->prof) && (NULL != CFGnode->prof->FirstGlobal)) {
         for (tempVar = CFGnode->prof->FirstGlobal;
              NULL != tempVar;
              tempVar = tempVar->next) {
            if (0 <= tempVar->index) {
               ++count;
               return count;
            }
         }
      }
      CFGnode = CFGnode->next;
   }

   return count;
} /* end of CountGlobalsInFunction() */

/*
 * ResetNodeFlag - reset the "done" and "bypassed" flags to FALSE for the
 * given CFG node and recurse through its children.
 */
void ResetNodeFlag(struct GlobalPlacementCFGNode *CFGnode) {
   assert(NULL != CFGnode);
   assert(NULL != CFGnode->prof);

   while (NULL != CFGnode) {
      CFGnode->done = FALSE;
      CFGnode->bypassed = FALSE;
      CFGnode = CFGnode->next;
   }

   return;
} /* end of ResetNodeFlag() */

/*
 * RemoveLeafCall - Given a top CFG node and a leaf function, remove any calls
 * to that leaf function in the entire CFG.
 * Return a count of calls removed.
 */
int RemoveLeafCall(struct GlobalPlacementCFGNode *CFGnode,
                   struct call_graph_node *cnode) {
   int callsRemoved = 0;
   struct varl *tempVar;
   struct varl *tempVar2;

   assert(NULL != cnode);
   assert(NULL != CFGnode);
   assert(NULL != CFGnode->prof);

   while (NULL != CFGnode) {
      if (NULL != CFGnode->prof) {
         tempVar = CFGnode->prof->FirstGlobal;
         tempVar2 = NULL;
         while (NULL != tempVar) {
            if (tempVar->cgnode == cnode) { /* Found a call to remove. */
               ++callsRemoved;
               if (NULL == tempVar2) { /* Call is first in list. */
                  tempVar2 = tempVar;
                  tempVar = tempVar->next;
                  CFGnode->prof->FirstGlobal = tempVar;
                  if (NULL == tempVar) /* Call was only item in list. */
                     CFGnode->prof->LastGlobal = NULL;
                  FREE(tempVar2);
                  tempVar2 = NULL;
               }
               else { /* Call is NOT first in list. */
                  tempVar2->next = tempVar->next;
                  if (NULL == tempVar->next) /* Call was last in list. */
                     CFGnode->prof->LastGlobal = tempVar2;
                  FREE(tempVar);
                  tempVar = tempVar2->next;
               }
            }
            else { /* Did not find call; move on in list. */
               tempVar2 = tempVar;
               tempVar = tempVar->next;
            }
         } /* end while (NULL != tempVar) */
      } /* end if (NULL != CFGnode->prof) */
      CFGnode = CFGnode->next;
   } /* end while (NULL != CFGnode) */

   return callsRemoved;
} /* end of RemoveLeafCall() */

/*
 * RemoveCallsToLeafWithNoGlobals - Given a leaf node with no globals and
 * no call sites, remove all calls to it from other functions, as it has
 * no effect on global profiling or global data placement.
 */
void RemoveCallsToLeafWithNoGlobals(struct call_graph_node *cnode) {
   struct call_graph_node *tempNode = the_call_graph;
   int callsRemoved;

   assert(NULL != tempNode);
   assert(NULL != cnode);
   assert(0 == cnode->CallSiteCount);

   while (NULL != tempNode) {
      if (tempNode->library_func)
         break;
      if (!tempNode->IrreducibleCFG && (tempNode->has_data)
          && (0 < tempNode->CallSiteCount) && (tempNode != cnode)) {
         callsRemoved = RemoveLeafCall(tempNode->topCFG, cnode);
         tempNode->CallSiteCount -= callsRemoved;
         if ((0 == tempNode->CallSiteCount)
             && (0 == CountGlobalsInFunction(tempNode->topCFG))) {
            /* We turned one of the callers into a function with
             * no calls and no globals.  Recursively process it.
             */
            RemoveCallsToLeafWithNoGlobals(tempNode);
         }
      }
      tempNode = tempNode->next_func;
   }

} /* end of RemoveCallsToLeafWithNoGlobals() */

/*
 * FixCFGIndices - For the given top CFG node, fill in its global indices
 * and process the entire CFG likewise.  Remove calls
 * to library functions, and give all other functions an index of -1.
 * Return a count of function calls seen.
 */
int FixCFGIndices(struct GlobalPlacementCFGNode *CFGnode, int *GlobalsSeen) {
   struct varl *tempVar;
   struct varl *prevVar = NULL;
   struct globalsym *tempGlobal;
   struct call_graph_node *cnode;
   char *id;
   int RemovedLibraryFunc;
   int FunctionFlag;
   int CallSiteCount = 0;

#if 0
   assert(NULL != CFGnode);
#else
   assert((NULL != CFGnode) && (NULL != CFGnode->prof));
#endif

   while (NULL != CFGnode) {
      /* Process the list of globals for this node. */
      if (NULL != CFGnode->prof)
         tempVar = CFGnode->prof->FirstGlobal;
      else
         tempVar = NULL;

      while (NULL != tempVar) {
         FunctionFlag = TRUE;
         RemovedLibraryFunc = FALSE;
         tempGlobal = findGlobalFromName(tempVar->name);
         tempVar->index = -1;
         if (NULL != tempGlobal) {
            FunctionFlag = GlobalIsFunction(tempGlobal);
         }
         if (!FunctionFlag) {
            tempVar->index = GetGlobalsIndex(tempGlobal, Globals, GlobalsCount);
            tempVar->cgnode = NULL;  /* Not a function call. */
            ++(*GlobalsSeen);
         }
         else { /* Must have been a function name, probably a library func. */

            /* If it is a call to a library function, remove it. */
            assert(NULL != tempVar->name);
            id = function_name_to_id(tempVar->name);
            if (NULL != id)
               cnode = find_call_graph_node(id);
            else
               cnode = NULL;
            tempVar->cgnode = cnode; /* Save to avoid lots of lookups. */

            if ((NULL == cnode) || cnode->library_func
                || cnode->IrreducibleCFG || !cnode->has_data) {
               /* Unknown or library function, or function with an
                * irreducible control flow graph, which we will never
                * be able to analyze or inline properly. Need to remove.
                */
               RemovedLibraryFunc = TRUE;
               if (NULL != prevVar) {
                  prevVar->next = tempVar->next;
                  if (NULL == prevVar->next)
                     CFGnode->prof->LastGlobal = prevVar;
                  FREE(tempVar);
                  tempVar = prevVar->next;
               }
               else { /* tempVar was FirstGlobal. */
                  CFGnode->prof->FirstGlobal = tempVar->next;
                  if (NULL == tempVar->next) {
                     CFGnode->prof->LastGlobal = NULL;
                     (void) fprintf(stderr, "CFG node emptied.\n");
                  }
                  FREE(tempVar);
                  tempVar = CFGnode->prof->FirstGlobal;
               }
            }
            else { /* Not a library call. */
               ++CallSiteCount;
            }
         } /* end if (!FunctionFlag) ... else ... */

         if (!RemovedLibraryFunc) {
            prevVar = tempVar;
            tempVar = tempVar->next;
         }
      } /* end while (NULL != tempVar) */
      CFGnode = CFGnode->next;
      prevVar = NULL;
   } /* end while (NULL != CFGnode) */

   return CallSiteCount;
} /* end of FixCFGIndices() */

/*
 * FillInGlobalIndices - Go through all CFG nodes for the whole file and
 * fill in the "index" field for each global from prof->FirstGlobal through
 * prof->LastGlobal, now that the Globals[] array has been built and the
 * indices into it can be determined.  As we go, we will prune the CFG of
 * many nodes that have no global accesses.
 */
void FillInGlobalIndices(struct call_graph_node *callGraph) {
   struct GlobalPlacementCFGNode *CFGnode;
   int GlobalsSeen;

   assert(NULL != callGraph);

   while (NULL != callGraph) {
      GlobalsSeen = 0;
      if (callGraph->library_func)
         break;
      if (!callGraph->IrreducibleCFG && callGraph->has_data) {
         /* Don't care about standard C library; it does not
          * use our global variables without us seeing it.
          * We cannot do anything with irreducible control flow graphs,
          * either, so we don't waste time fixing them up.
          */
         CFGnode = callGraph->topCFG;
         callGraph->InlinedIntoCallers = FALSE;
         if (NULL != CFGnode) {
            callGraph->CallSiteCount = FixCFGIndices(CFGnode, &GlobalsSeen);
            if ((0 == GlobalsSeen) && (0 == callGraph->CallSiteCount)) {
               /* No globals; save future time by bypassing this function. */
               callGraph->InlinedIntoCallers = TRUE;
               FreeCFG(callGraph);
            }
            else
               PruneCFG(callGraph);
         }
      }
      else {
         if (NULL != callGraph->topCFG)
            FreeCFG(callGraph);
         callGraph->CallSiteCount = 0;
         callGraph->InlinedIntoCallers = TRUE;
      }
      callGraph = callGraph->next_func;
   }

   /* FixCFGIndices() has already removed calls to library functions,
    * unknown functions, and functions with irreducible CFGs. There are
    * still two other categories of calls that need to be removed: calls to
    * functions that were marked "InlinedIntoCallers" above because they
    * had no call sites and used no globals, and calls to functions (probably
    * library functions that escaped our input list of library functions) that
    * do not appear in our input file (hence their "has_data" field is FALSE)
    * but they are not marked as library functions. Remove such calls here.
    */
   for (callGraph = the_call_graph; 
        NULL != callGraph;
        callGraph = callGraph->next_func) {
      if (callGraph->library_func)
         break;
      if (!callGraph->IrreducibleCFG && (0 == callGraph->CallSiteCount)
          && (callGraph->InlinedIntoCallers || !callGraph->has_data))
         RemoveCallsToLeafWithNoGlobals(callGraph);
   }

   return;
} /* end of FillInGlobalIndices() */

/*
 * CopyCFGNode - Copy the contents of node2 into node1.
 * If node2 has no children, make node1 flow into successor.
 * Scale the inlined node2's estimate field by multiplying it by the
 * passed in estimate (from the original calling CFG node).
 * If this is a direct recursion, eliminate calls to Callee as we inline.
 */
void CopyCFGNode(float estimate,
                 struct call_graph_node *RootNode,
                 struct GlobalPlacementCFGNode *node1,
                 struct GlobalPlacementCFGNode *node2,
                 struct GlobalPlacementCFGNode *successor,
                 struct call_graph_node *Callee,
                 int *RecursionsRemoved) {
   struct varl *tempVar;
   struct varl *tempVar2;
   struct GlobalPlacementCFGNode *tempCFGNode1;
   struct GlobalPlacementCFGNode *tempCFGNode2;
   int NewRecursion; /* True if we are introducing direct recursion. */
   int RecursiveFlag = (Callee == RootNode);
   int LeftFlag, RightFlag;

   assert(NULL != node1);
   assert(NULL != node2);
   assert(NULL != RecursionsRemoved);
   assert(NULL != RootNode);

   node1->estimate = node2->estimate * estimate;
   node1->LoopList = NULL;
   if (NULL != node2->LoopList)
      CopyCFGLoopList(node1, node2);

   /* We should be able to get away with aliasing the globals list
    * instead of copying it, because the function being inlined into
    * all of its callers will become dead after it is inlined and will
    * not be subject to further processing, meaning that its CFG should
    * remain untouched.  This is only true if we are not introducing a
    * direct recursive call in the particular CFG node.  In that CFG node,
    * if RecursiveFlag is TRUE, we would want to do a copy, deleting
    * the new recursive call as we copy.
    */
   if (NULL != node2->prof) {
      NewRecursion = FALSE;
      for (tempVar2 = node2->prof->FirstGlobal;
           NULL != tempVar2;
           tempVar2 = tempVar2->next) {
         NewRecursion |= (tempVar2->cgnode == RootNode);
      }
#if 0
      /* Prof can be an alias, and it is allocated via SALLOC_PERM, meaning
       * it is really malloc'ed.  If PruneCFG() is called later, it could
       * delete a CFG node that has no globals. Then a dangling pointer alias
       * exists, the same address is now available for another malloc(),
       * and you have big problems. Better to copy.  Might get away with
       * aliasing if we have globals, but the no globals case will definitely
       * be pruned later.
       */
      if (!(RecursiveFlag && NewRecursion)) {
         node1->prof = node2->prof;
      }
      else {  /* Have to copy for safety. */
#endif
         node1->prof = (struct GlobalDataProfile *)
            SALLOC_PERM(sizeof(struct GlobalDataProfile));
         if (NULL == node1->prof) {
            (void) fprintf(stderr, "Out of virtual memory: CopyCFGnode\n");
            exit(EXIT_FAILURE);
         }
         node1->prof->LeftBackEdgeFlag = node2->prof->LeftBackEdgeFlag;
         node1->prof->RightBackEdgeFlag = node2->prof->RightBackEdgeFlag;
         node1->prof->FirstGlobal = 
            (struct varl *) SALLOC_PERM(sizeof(struct varl));
         if (NULL == node1->prof->FirstGlobal) {
            (void) fprintf(stderr, "Out of virtual memory: CopyCFGnode\n");
            exit(EXIT_FAILURE);
         }
         tempVar = node1->prof->FirstGlobal;
         tempVar2 = node2->prof->FirstGlobal;
         /* Set a marker in case globals list becomes empty because the
          * only entries were directly recursive calls to Callee.
          */
         tempVar->cgnode = Callee;
         while (NULL != tempVar2) {
            if (RecursiveFlag && (tempVar2->cgnode == Callee)) {
               /* Skip over recursive call. */
               ++(*RecursionsRemoved);
               tempVar2 = tempVar2->next;
            }
            else {
               tempVar->index = tempVar2->index;
               tempVar->name = tempVar2->name;
               tempVar->cgnode = tempVar2->cgnode;
               tempVar2 = tempVar2->next;
               /* Allocate tempVar->next only if we are going to need it. */
               if ((NULL != tempVar2)
                   && (!(Callee == tempVar2->cgnode && NULL == tempVar2->next))) {
                  tempVar->next = 
                     (struct varl *) SALLOC_PERM(sizeof(struct varl));
                  if (NULL == tempVar->next) {
                     (void) fprintf(stderr,
                                    "Out of virtual memory: CopyCFGnode\n");
                     exit(EXIT_FAILURE);
                  }
                  tempVar = tempVar->next;
               }
               else
                  tempVar->next = NULL;
            }
         }
         node1->prof->LastGlobal = tempVar;
         /* See if we made the list empty by eliminating recursion. */
         if (tempVar->cgnode == Callee) { /* Found the marker. */
            node1->prof->LastGlobal = NULL;
            FREE(node1->prof->FirstGlobal);
         }
#if 0
      }
#endif
   } /* end if (NULL != node2->prof) */

   /* If node2 has no children, we need to make successor be
    * the left child of node1 and return.
    */
   node1->left = NULL;
   node1->right = NULL;
   LeftFlag = (NULL != node2->left);
   RightFlag = (NULL != node2->right);
   if (!LeftFlag && !RightFlag) {
      node1->left = successor;  /* successor could be NULL. */
      node1->prof->LeftBackEdgeFlag = FALSE;
      node1->prof->RightBackEdgeFlag = FALSE;
   }

   return;
} /* end of CopyCFGNode() */

/*
 * NumberCFGNodes - Run through the CFG nodes and number them, starting
 * with 0 for the top node. Place the number in the NodeNum field.
 * Return the number of nodes.
 */
int NumberCFGNodes(struct GlobalPlacementCFGNode *top) {
   struct GlobalPlacementCFGNode *temp = top;
   int num = 0;

   assert(NULL != top);

   while (NULL != temp) {
      temp->NodeNum = num++;
      temp = temp->next;
   }
   return num;
} /* end of NumberCFGNodes() */

/*
 * FreeCFGHashTable - Walk the hash table and free all of its
 * entries in memory.
 */
void FreeCFGHashTable(struct hashedCFGNode **table) {
   struct hashedCFGNode *tempNode;
   struct hashedCFGNode *currentNode;
   int index;

   assert(NULL != table);

   for (index = 0; index < 4096; ++index) {
      currentNode = table[index];
      while (NULL != currentNode) {
         tempNode = currentNode->next;
         FREE(currentNode);
         currentNode = tempNode;
      }
   }
   return;
} /* end of FreeCFGHashTable() */

/*
 * InsertInCFGHashTable - insert a CFG node pointer value, and its index
 * value, into the CFG node hash table at subscript "hash".
 */
void InsertInCFGHashTable(struct hashedCFGNode **table,
                          struct GlobalPlacementCFGNode *node,
                          int index, int hash) {
   struct hashedCFGNode *newNode;

   assert(NULL != table);
   assert(NULL != node);
   assert(0 <= index);
   assert((0 <= hash) && (4096 > hash));

   /* Allocate a new hash table entry. */
   newNode =
        (struct hashedCFGNode *) SALLOC_PERM(sizeof(struct hashedCFGNode));
   if (NULL == newNode) {
      (void) fprintf(stderr, "Out of memory: InsertInCFGHashTable.\n");
      exit(EXIT_FAILURE);
   }

   /* Insert newNode at front of hash list at table[hash]. */   
   newNode->node = node;
   newNode->index = index;
   newNode->next = table[hash];  /* Could be NULL. */
   table[hash] = newNode;

   return;
} /* end of InsertInCFGHashTable() */

/* SetNodeArrayChildren -
 * Set the left and right child links in each NodeArray
 * node to correspond to the left and right child links in the
 * StartNode (source) array. This uses a hash table
 * that maps node pointers to array indices.
 */
void SetNodeArrayChildren(int NumNodes, 
                          struct GlobalPlacementCFGNode **StartNode,
                          struct GlobalPlacementCFGNode **NodeArray) {
   int index;
   int childIndex;
   int hash;
   static struct GlobalPlacementCFGNode **savedStartNode = NULL;
   static int savedNumNodes = 0;
   static struct hashedCFGNode *CFGHashTable[4096];

   if (-1 == NumNodes) { /* Signal to free memory. */
      FreeCFGHashTable(CFGHashTable);
      savedStartNode = NULL;
      savedNumNodes = 0;
      return;
   }

   assert(NULL != StartNode);
   assert(NULL != NodeArray);
   assert(0 < NumNodes);

   /* If we have no existing hash table, zero out the table. */
   if (NULL == savedStartNode)
      bzero(CFGHashTable, 4096 * sizeof(struct hashedCFGNode *));

   /* Build a new hash table if we need to. */
   if ((savedStartNode != StartNode) || (savedNumNodes != NumNodes)) {
      if (NULL != savedStartNode)
         FreeCFGHashTable(CFGHashTable);
      savedStartNode = StartNode;
      savedNumNodes = NumNodes;
      for (index = 0; index < NumNodes; ++index) {
         assert(NULL != StartNode[index]);
         hash = (((unsigned int) StartNode[index]) >> 5) & 4095;
         assert((0 <= hash) && (4095 >= hash));
         InsertInCFGHashTable(CFGHashTable, StartNode[index], index, hash);
      }
   }

   for (index = 0; index < NumNodes; ++index) { 
      assert(NULL != StartNode[index]);
      assert(NULL != NodeArray[index]);
      if (NULL != StartNode[index]->left) {
         childIndex = GetNodeIndex(StartNode[index]->left,
                                   CFGHashTable);
         NodeArray[index]->left = NodeArray[childIndex];
      }
      if (NULL != StartNode[index]->right) {
         childIndex = GetNodeIndex(StartNode[index]->right,
                                   CFGHashTable);
         NodeArray[index]->right = NodeArray[childIndex];
      }
   }
} /* end of SetNodeArrayChildren() */

/*
 * CopyNodeArray - Fill the given NodeArray with pointers to newly
 * allocated CFG nodes, and copy the contents of the array of CFG
 * nodes that starts at StartNode into the newly allocated nodes,
 * following the next pointers and determining the proper left and
 * right pointers.  If there is a loop list to prepend, do so for
 * each node.
 */
void CopyNodeArray(int NumNodes, float estimate,
                   struct GlobalPlacementCFGNode **StartNode,
                   struct GlobalPlacementCFGNode **NodeArray,
                   struct call_graph_node *RootNode,
                   struct call_graph_node *Callee,
                   struct GlobalPlacementCFGNode *NewNode,
                   struct CFGloopl *PrependLoops,
                   int *RecursionsRemoved) {
   int index;
   struct GlobalPlacementCFGNode *CFGnode = StartNode[0];
   struct CFGloopl *TempLoop;
   struct CFGloopl *NewLoop;
   struct CFGloopl *SaveLoop;
   struct CFGloopl *FirstLoop;

   assert(0 < NumNodes);

   for (index = 0; index < NumNodes; ++index) {
      assert(NULL != CFGnode);
      NodeArray[index] = (struct GlobalPlacementCFGNode *)
         SALLOC_PERM(sizeof(struct GlobalPlacementCFGNode));
      if (NULL == NodeArray[index]) {
         (void) fprintf(stderr,
                        "Out of virtual memory: CopyNodeArray\n");
         exit(EXIT_FAILURE);
      }
      CopyCFGNode(estimate, RootNode, NodeArray[index],
                  CFGnode, NewNode, Callee, RecursionsRemoved);
      NodeArray[index]->done = TRUE;
      NodeArray[index]->NodeNum = 0; /* Temporary. */
      if (NULL != PrependLoops) {
         /* Copy the PrependLoops list into this node's loop list. */
         TempLoop = PrependLoops;
         SaveLoop = NULL;
         FirstLoop = NULL;
         do {
            NewLoop = (struct CFGloopl *) SALLOC_PERM(sizeof(struct CFGloopl));
            if (NULL == NewLoop) {
               (void) fprintf(stderr,
                              "Out of virtual memory: CopyNodeArray\n");
               exit(EXIT_FAILURE);
            }
            NewLoop->loopID = TempLoop->loopID;
            NewLoop->blockCount = TempLoop->blockCount;
            NewLoop->iterations = TempLoop->iterations;
            if (NULL != SaveLoop)
               SaveLoop->next = NewLoop;
            else
               FirstLoop = NewLoop;
            SaveLoop = NewLoop;
            TempLoop = TempLoop->next;
         } while (NULL != TempLoop);
         assert(NULL != SaveLoop);
         assert(NULL != FirstLoop);
         SaveLoop->next = NodeArray[index]->LoopList;
         NodeArray[index]->LoopList = FirstLoop;
      } /* end if (NULL != PrependLoops) */
      CFGnode = CFGnode->next;
   } /* end for (index = 0; index < NumNodes; ...) */

   for (index = 0; index < (NumNodes - 1); ++index) 
      NodeArray[index]->next = NodeArray[index + 1];
   NodeArray[NumNodes - 1]->next = NewNode;

   /* Set the left/right children pointers in NodeArray
    * to mimic the structure of StartNode.
    */
   SetNodeArrayChildren(NumNodes, StartNode, NodeArray);
   return;
} /* end of CopyNodeArray() */

/*
 * CopyCFGLoopList - Allocate a new loop list for the first CFG node, and
 * copy the contents of the loop list of the second CFG node into it.
 */
void CopyCFGLoopList(struct GlobalPlacementCFGNode *DestNode,
                     struct GlobalPlacementCFGNode *SourceNode) {
   struct CFGloopl *SourceLoop;
   struct CFGloopl *DestLoop = NULL;

   assert(NULL != SourceNode);
   assert(NULL != DestNode);

   SourceLoop = SourceNode->LoopList;
   DestNode->LoopList = NULL;
   while (NULL != SourceLoop) {
      if (NULL == DestLoop) { /* First loop in list. */
         DestNode->LoopList =
             (struct CFGloopl *) SALLOC_PERM(sizeof(struct CFGloopl));
         DestLoop = DestNode->LoopList;
      }
      else {
         DestLoop->next =
             (struct CFGloopl *) SALLOC_PERM(sizeof(struct CFGloopl));
         DestLoop = DestLoop->next;
      }
      if (NULL == DestLoop) {
         (void) fprintf(stderr, "Out of memory: CopyCFGLoopList().\n");
         exit(EXIT_FAILURE);
      }
      DestLoop->loopID = SourceLoop->loopID;
      DestLoop->blockCount = SourceLoop->blockCount;
      DestLoop->iterations = SourceLoop->iterations;
      DestLoop->next = NULL;
      SourceLoop = SourceLoop->next;
   }
   return;
} /* end of CopyCFGLoopList */

/*
 * IncreaseLoopBlockCounts - Go through the CFG for the given call graph node
 * and find all occurrences of the loops in the given loop list.  Increase the
 * block counts of these loops by the given increment, which is the number of
 * blocks in the callee that was inlined into some node that is contained in
 * all loops in that loop list.
 */
void IncreaseLoopBlockCounts(struct call_graph_node *RootNode,
                             struct CFGloopl *Loops, int increment) {
   struct GlobalPlacementCFGNode *CFGnode;
   struct CFGloopl *SourceLoop;
   struct CFGloopl *CFGLoop;

   assert(NULL != RootNode);
   assert(0 < increment);

   if (NULL == Loops)
      return;

   CFGnode = RootNode->topCFG;

   /* For each loop list in each CFG node in the function, we can find
    * matches with the Loops argument using the fact that loopID fields
    * are unique over the whole source program, and the loop lists are kept
    * sorted by blockCount from largest to smallest.
    */
   while (NULL != CFGnode) {
      CFGLoop = CFGnode->LoopList;
      SourceLoop = Loops;
      while ((NULL != CFGLoop) && (NULL != SourceLoop)) {
         if (CFGLoop->loopID == SourceLoop->loopID) {
            CFGLoop->blockCount += increment;
            SourceLoop = SourceLoop->next;
            CFGLoop = CFGLoop->next;
         }
         else if (SourceLoop->blockCount > CFGLoop->blockCount)
            SourceLoop = SourceLoop->next;
         else
            CFGLoop = CFGLoop->next;
      }

      CFGnode = CFGnode->next;
   } /* end while (NULL != CFGnode) */
   return;
} /* end of IncreaseLoopBlockCounts() */

/*
 * InlineNodeInCFGNode - Given a candidate CFG node and a particular
 * callee node to look for, find any calls to the callee node, inline the
 * callee node's CFG in the given CFG.
 * Return TRUE if we inline, FALSE otherwise.
 * NOTE: Passing NULL for RootNode is a signal to deallocate static data
 * and return, because an inlining phase has been completed.
 */
int InlineNodeInCFGNode(struct GlobalPlacementCFGNode *CFGnode,
                        struct call_graph_node *Callee,
                        struct call_graph_node *RootNode) {
   struct varl *tempVar;
   struct GlobalPlacementCFGNode *NewNode;
   struct GlobalPlacementCFGNode *tempNode;
   struct varl *tempVar2;
   struct CFGloopl *TempLoop;
   int change = FALSE;
   int index;
   int childIndex;
   int RecursionsRemoved = 0;
   int RecursiveFlag = (Callee == RootNode);

   static int NumNodes = 0;             /* Node count from Callee. */
   static int CallSites = 0;            /* Use if RecursiveFlag */
   static int DirectRecursionsRemoved = 0;
   static struct GlobalPlacementCFGNode **SourceArray = NULL;
   static struct GlobalPlacementCFGNode **RecursiveSourceArray = NULL;
   static struct GlobalPlacementCFGNode **DestArray = NULL;
   static char *SourceName = NULL;

   if (NULL == RootNode) { /* Signal to deallocate static data. */
      if (NULL != SourceArray) {
         /* SourceArray has data from a previously processed Callee. */
         FREE(SourceArray);
         SourceArray = NULL;
         NumNodes = 0;
         if (NULL != RecursiveSourceArray) {
            FreeCFGNodes(RecursiveSourceArray[0]);
            FREE(RecursiveSourceArray);
            DirectRecursionsRemoved = 0;
            CallSites = 0;
         }
      }
      SetNodeArrayChildren(-1, NULL, NULL); /* Signal to deallocate. */
      return FALSE;
   }

   assert(NULL != CFGnode);
   assert(NULL != Callee);
   assert(NULL != Callee->topCFG);

   if (CFGnode->done)
      return FALSE;

   if (NULL != CFGnode->prof) {
      CFGnode->done = TRUE;
      tempVar = CFGnode->prof->FirstGlobal;
      while (NULL != tempVar) {
         if (tempVar->cgnode == Callee) {
            /* Found it. Inline the whole CFG of Callee right here. */
            change = TRUE;
            /* In order to inline a copy of the CFG of Callee and keep
             * all of the "next" and "left" and "right" fields in order
             * in the inlined copy, we need to have indexed arrays of CFG
             * nodes for Callee and the copy of Callee, analogous to what we
             * did in TransferProfileToCFG().
             */
            if ((NULL != SourceArray) && (SourceName != Callee->name)) {
               /* SourceArray has data from a previously processed Callee. */
               FREE(SourceArray);
               SourceArray = NULL;
               if (NULL != RecursiveSourceArray) {
                  FreeCFGNodes(RecursiveSourceArray[0]);
                  FREE(RecursiveSourceArray);
               }
               SetNodeArrayChildren(-1, NULL, NULL); /* Signal: deallocate. */
            }
            if (NULL == SourceArray) { /* Not yet allocated. */
               NumNodes = NumberCFGNodes(Callee->topCFG);
               assert(0 < NumNodes);
               SourceArray = (struct GlobalPlacementCFGNode **)
                  SALLOC_PERM(NumNodes
                              * sizeof(struct GlobalPlacementCFGNode *));
               if (RecursiveFlag) {
                  RecursiveSourceArray = (struct GlobalPlacementCFGNode **)
                        SALLOC_PERM(NumNodes
                                    * sizeof(struct GlobalPlacementCFGNode *));
                  CallSites = Callee->CallSiteCount; 
                   /* Preserve original count. */
               }
               if ((NULL == SourceArray)
                   || (RecursiveFlag && (NULL == RecursiveSourceArray))) {
                  (void) fprintf(stderr,
                               "Out of virtual memory: InlineNodeInCFGNode\n");
                  exit(EXIT_FAILURE);
               }
               SourceName = Callee->name;
               tempNode = Callee->topCFG;
               while (NULL != tempNode) {
                  SourceArray[tempNode->NodeNum] = tempNode;
                  tempNode = tempNode->next;
               }
               if (RecursiveFlag) {
                  /* Allocate and copy nodes into RecursiveSourceArray.
                   * This gives us a single copy of the original CFG node
                   * array prior to inlining.  We don't want to inline 100
                   * nodes at the site of the first direct recursion, 200
                   * nodes at the second site, etc.
                   */
                  DirectRecursionsRemoved = 0;
                  CopyNodeArray(NumNodes, 1.0, SourceArray,
                                RecursiveSourceArray, RootNode, Callee, NULL,
                                NULL, &DirectRecursionsRemoved);
                  CallSites -= DirectRecursionsRemoved;
               }
            }
            --RootNode->CallSiteCount;
            if (!RecursiveFlag) {
               RootNode->CallSiteCount += Callee->CallSiteCount;
            }
            else
               RootNode->CallSiteCount += CallSites;

            /* We are going to split CFGnode.  The globals list prior to
             * tempVar will be put in the upper node after the split, while
             * the rest of the globals list after tempVar will be put in
             * the lower node.  The upper node will unconditionally flow
             * into the CFG for the Callee, which will be copied in at
             * this point.  The Callee CFG will unconditionally flow
             * into the lower split node.  The lower split node will inherit
             * the entire list of successors
             * from the original unsplit CFGnode.  CFGnode will then be made
             * to point to the lower split node so that we can continue.
             */
            NewNode = (struct GlobalPlacementCFGNode *)
               SALLOC_PERM(sizeof(struct GlobalPlacementCFGNode));
            if (NULL == NewNode) {
               (void) fprintf(stderr,
                              "Out of virtual memory: InlineNodeInCFGNode\n");
               exit(EXIT_FAILURE);
            }
            NewNode->estimate = CFGnode->estimate;
            NewNode->prof = (struct GlobalDataProfile *)
               SALLOC_PERM(sizeof(struct GlobalDataProfile));
            if (NULL == NewNode->prof) {
               (void) fprintf(stderr,
                              "Out of virtual memory: InlineNodeInCFGNode\n");
               exit(EXIT_FAILURE);
            }
            NewNode->left = CFGnode->left;
            NewNode->right = CFGnode->right;
            NewNode->next = CFGnode->next;
            NewNode->done = FALSE;
            CopyCFGLoopList(NewNode, CFGnode);
            NewNode->prof->LeftBackEdgeFlag = CFGnode->prof->LeftBackEdgeFlag;
            NewNode->prof->RightBackEdgeFlag = CFGnode->prof->RightBackEdgeFlag;

            /* NewNode will be the lower node after the split. We need to
             * divide the globals list of tempVar between the upper and lower
             * nodes, CFGnode and NewNode.
             */
            tempVar2 = CFGnode->prof->FirstGlobal;
            if (tempVar2 == tempVar) { /* Leaf call was first item. */
               NewNode->prof->FirstGlobal = tempVar->next;
               if (NULL != tempVar2->next)  /* Leaf call is not last item. */
                  NewNode->prof->LastGlobal = CFGnode->prof->LastGlobal;
               else
                  NewNode->prof->LastGlobal = NULL;
               CFGnode->prof->FirstGlobal = NULL;
               CFGnode->prof->LastGlobal = NULL;
            }
            else {  /* Leaf call was NOT the first item. */
               while (tempVar2 != tempVar) {
                  /* These globals stay in CFGNode. */
                  if (tempVar2->next == tempVar) {
                     /* Terminate CFGnode's list with tempVar2. */
                     CFGnode->prof->LastGlobal = tempVar2;
                     tempVar2->next = NULL;
                     break;
                  }
                  tempVar2 = tempVar2->next;
               }
               /* Now, make NewNode's globals list start after the call
                * to the leaf node being inlined, which is tempVar.
                */
               NewNode->prof->FirstGlobal = tempVar->next;
               NewNode->prof->LastGlobal = NULL;
               for (tempVar2 = tempVar->next;
                    NULL != tempVar2;
                    tempVar2 = tempVar2->next) {
                  NewNode->prof->LastGlobal = tempVar2;
               }
            } /* end if (tempVar2 == tempVar) ... else ... */

            /* Now we copy the CFGnodes of the inlined function into the
             * calling function. Use a temporary node, so that we don't
             * infinitely extend the function with new nodes.
             */
            if (NULL != DestArray)
               FREE(DestArray);
            DestArray = (struct GlobalPlacementCFGNode **)
               SALLOC_PERM(NumNodes * sizeof(struct GlobalPlacementCFGNode *));
            if (NULL == DestArray) {
               (void) fprintf(stderr,
                              "Out of virtual memory: InlineNodeInCFGNode\n");
               exit(EXIT_FAILURE);
            }

            if (RecursiveFlag) /* Use RecursiveSourceArray as our source. */
               CopyNodeArray(NumNodes, CFGnode->estimate, RecursiveSourceArray,
                             DestArray, RootNode, Callee, NewNode,
                             CFGnode->LoopList, &RecursionsRemoved);
            else
               CopyNodeArray(NumNodes, CFGnode->estimate, SourceArray,
                             DestArray, RootNode, Callee, NewNode,
                             CFGnode->LoopList, &RecursionsRemoved);

            CFGnode->left = DestArray[0]; /* Safe to extend the CFG now. */
            CFGnode->next = DestArray[0];
            CFGnode->prof->LeftBackEdgeFlag = FALSE;
            CFGnode->prof->RightBackEdgeFlag = FALSE;
            CFGnode = NewNode; /* Resume after the inlined CFG. */
            tempVar = NewNode->prof->FirstGlobal;
            CFGnode->done = TRUE;
            if (!RecursiveFlag)
               RootNode->CallSiteCount -= RecursionsRemoved;
            RecursionsRemoved = 0;
            RootNode->blockCount += Callee->blockCount;
            /* Now we need to increase the block counts of the loops that
             * already existed in the caller that have increased in size
             * because Callee was inlined into their loop bodies.  This
             * set of loops is exactly CFGnode->LoopList.
             */
            IncreaseLoopBlockCounts(RootNode, CFGnode->LoopList,
                                    Callee->blockCount);
#if 0
            /* If we inlined a directly recursive call, then the number
             * of nodes in the function will have increased.  The
             * SourceArray[] is no longer valid.
             */
            if (RecursiveFlag) {
               FREE(SourceArray);
               SourceArray = NULL;
            }
#endif

         } /* end if (tempVar->cgnode == cnode) */
         else
            tempVar = tempVar->next;
      } /* end while (NULL != tempVar) */
   } /* end if (NULL != CFGnode->prof) */

   CFGnode->done = TRUE;

   return change;
} /* end of InlineNodeInCFGNode() */

/*
 * InlineLeafNode - Given a leaf node (no call sites in it), inline it
 * into all of its callers.  Because it is a leaf node, this will not
 * create any new direct or indirect recursions.  Decrement the call site
 * total of each caller by one for each inlining.  Prune the CFG of each
 * caller after inlining.
 * Return TRUE if we do any inlining, FALSE otherwise.
 */
int InlineLeafNode(struct call_graph_node *cnode) {
   struct GlobalPlacementCFGNode *CFGnode;
   struct call_graph_node *cgnode = the_call_graph;
   int change = FALSE;
   int NumNodes;
   int newchange;

   assert(NULL != cnode);
   assert(NULL != cgnode);
   assert(0 == cnode->CallSiteCount);

   do {
      if (cgnode->library_func)
         break;
      if (!cgnode->InlinedIntoCallers
          && (NULL != cgnode->topCFG) && (0 < cgnode->CallSiteCount)
          && cgnode->has_data && !cgnode->IrreducibleCFG) {
         assert(cnode != cgnode); /* Leaf cannot call itself. */
         CFGnode = cgnode->topCFG;
         ResetNodeFlag(CFGnode);
         /* Recurse through CFG looking for calls to this function. */
         newchange = FALSE;
         while (NULL != CFGnode) {
            newchange |= InlineNodeInCFGNode(CFGnode, cnode, cgnode);
            change |= newchange;
            CFGnode = CFGnode->next;
         }
         if (newchange) {
            PruneCFG(cgnode);
            NumNodes = NumberCFGNodes(cgnode->topCFG);
            if ((0 != strcmp(cgnode->name, "main"))
                && (0 != strcmp(cgnode->name, "MAIN__")))
               cgnode->InlineEligible = (NumNodes <= InlineSizeLimit);
         }
      }
      cgnode = cgnode->next_func;
   } while (NULL != cgnode);

   cnode->InlinedIntoCallers = TRUE;
   cnode->InlineEligible = FALSE;
   FreeCFG(cnode);

   return change;
} /* end of InlineLeafNode() */

/*
 * EliminateDirectRecursion - Search a callgraph node for direct calls to
 * itself. Inline a copy of its CFG at the call site, but eliminate self
 * calls during the copy to cut the recursion depth at 1. Prune the CFG
 * if we did any inlinings.
 */
int EliminateDirectRecursion(struct call_graph_node *cnode) {
   int change = FALSE;
   int newChange;
   int NumNodes;
   struct GlobalPlacementCFGNode *CFGnode;

   assert(NULL != cnode);
   assert(0 < cnode->CallSiteCount);

   CFGnode = cnode->topCFG;
   assert(NULL != CFGnode);

   /* Recurse through CFG looking for calls to this function. */
   ResetNodeFlag(CFGnode);
   while (NULL != CFGnode) {
      newChange = InlineNodeInCFGNode(CFGnode, cnode, cnode);
      change |= newChange;
      if (newChange)
         cnode->topEstimate *= (1.0 + CFGnode->estimate);
      CFGnode = CFGnode->next;
   }
   if (change) {
      PruneCFG(cnode);
      NumNodes = NumberCFGNodes(cnode->topCFG);
      if ((0 != strcmp(cnode->name, "main"))
          && (0 != strcmp(cnode->name, "MAIN__")))
         cnode->InlineEligible = (NumNodes <= InlineSizeLimit);
   }

   return change;
} /* end of EliminateDirectRecursion() */

/*
 * InlineRecursiveNode - Inline a non-leaf node and break any direct recursions
 * that result.  Prune the CFG of any callers we inlined into.
 */
int InlineRecursiveNode(struct call_graph_node *cnode, time_t StartTime) {
   struct GlobalPlacementCFGNode *CFGnode;
   struct call_graph_node *cgnode = the_call_graph;
   int change = FALSE;
   int NumNodes;
   int newChange;
   int newChange2 = FALSE;
   time_t CurrentTime;
   double ElapsedSeconds;

   assert(NULL != cnode);
   assert(NULL != cgnode);
   assert(0 <= cnode->CallSiteCount);

   do {
      if (cgnode->library_func)
         break;
      if (!cgnode->InlinedIntoCallers
          && (NULL != cgnode->topCFG) && (0 < cgnode->CallSiteCount)
          && cgnode->has_data && !cgnode->IrreducibleCFG && (cnode != cgnode)) {
         CFGnode = cgnode->topCFG;

         /* Recurse through CFG looking for calls to the cnode function. */
         ResetNodeFlag(CFGnode);
         newChange = FALSE;
         while (NULL != CFGnode) {
            newChange |= InlineNodeInCFGNode(CFGnode, cnode, cgnode);
            if (newChange) {
               (void) time(&CurrentTime);
               ElapsedSeconds = difftime(CurrentTime, StartTime);
               if (INLINING_TIME_LIMIT < ElapsedSeconds) {
                  (void) fprintf(stderr,
                           "Time limit reached: InlineFunctionsIntoMain()\n");
                  return TRUE;
               }
            }
            change |= newChange;
            CFGnode = CFGnode->next;
         }

         if (newChange) {
            /* Eliminate direct recursions we created in our caller, unless
             * our caller is so big as to not be eligible for inlining.
             */
            ResetNodeFlag(cgnode->topCFG);
            if ((0 < cgnode->CallSiteCount) && (cgnode->InlineEligible)
                 && !InlineLimitDoubled)
               newChange2 = EliminateDirectRecursion(cgnode);
            else
               newChange2 = FALSE;
            change |= newChange2;
            if (newChange2) {
               ResetNodeFlag(cgnode->topCFG);
               /* Deallocate static data in InlineNodeInCFGNode() */
               (void) InlineNodeInCFGNode(NULL, NULL, NULL);
            }
         }

         if (newChange || newChange2) {
            PruneCFG(cgnode);
            NumNodes = NumberCFGNodes(cgnode->topCFG);
            if ((0 != strcmp(cgnode->name, "main"))
                && (0 != strcmp(cgnode->name, "MAIN__")))
               cgnode->InlineEligible = (NumNodes <= InlineSizeLimit);
         }
      }
      cgnode = cgnode->next_func;
   } while (NULL != cgnode);

   cnode->InlinedIntoCallers = TRUE;
   cnode->InlineEligible = FALSE;
   FreeCFG(cnode);

   return change;
} /* end of InlineRecursiveNode() */

/* Recompute our CFG node budget limitations per function, mark
 * functions as InlineEligible or not, and return the number of
 * inline-eligible functions remaining to be processed.
 */
int ScanRemainingFunctions(struct call_graph_node *mainNode,
                           int *MinCallSiteCount) {
   int RemainingFuncs = 0;
   int NonInlinedFuncs = 0;
   struct call_graph_node *cnode;
   int NumNodes;

   /* Find out the maximum number of functions left to process. */
   cnode = the_call_graph;
   *MinCallSiteCount = 50000;
   do {
      if (cnode->library_func)
         break;
      else if ((cnode != mainNode) && !cnode->InlinedIntoCallers
               && (!cnode->IrreducibleCFG) && (cnode->has_data)) {
         ++NonInlinedFuncs;
         assert(0 <= cnode->CallSiteCount);
         if (*MinCallSiteCount > cnode->CallSiteCount)
            *MinCallSiteCount = cnode->CallSiteCount;
      }
      cnode = cnode->next_func;
   } while (NULL != cnode);

   if (0 == NonInlinedFuncs)
      return 0;

   /* Apply the node budget limitation. */
   InlineSizeLimit = CFGNodeBudget / NonInlinedFuncs;
   cnode = the_call_graph;
   *MinCallSiteCount = 50000;
   RemainingFuncs = 0;
   do {
      if (cnode->library_func)
         break;
      else if (!cnode->InlinedIntoCallers && (cnode->has_data)
               && (!cnode->IrreducibleCFG) && (cnode != mainNode)) {
         NumNodes = NumberCFGNodes(cnode->topCFG);
         cnode->InlineEligible = (NumNodes <= InlineSizeLimit);
         if (cnode->InlineEligible) {
            ++RemainingFuncs;
            assert(0 <= cnode->CallSiteCount);
            if (*MinCallSiteCount > cnode->CallSiteCount)
               *MinCallSiteCount = cnode->CallSiteCount;
         }
      }
      else
         cnode->InlineEligible = FALSE;
      cnode = cnode->next_func;
   } while (NULL != cnode);

   return RemainingFuncs;
} /* end of ScanRemainingFunctions() */


/*
 * InlineFunctionsIntoMain - inline the CFGs of all functions other than main()
 * until we have a huge CFG for main() that represents the entire program, or
 * until we exhaust our CFGNodeBudget.
 * Library function calls, unknown function calls, indirect function calls
 * through pointers, and calls to functions with irreducible CFGs will be
 * treated as no-ops and no inlining will occur at these call sites.
 */
struct call_graph_node *
InlineFunctionsIntoMain(struct call_graph_node *callGraph) {
   int change;
   int outerChange;
   struct call_graph_node *cnode = callGraph;
   struct call_graph_node *mainNode = NULL;
   int CallSiteTarget;
   int LoopIterations;
   int RemainingFuncs;
   int NumNodes;
   int MinCallSiteCount;
   time_t StartTime, CurrentTime;
   double ElapsedSeconds;

   assert(NULL != callGraph);

   while (NULL != cnode) {
      /* Fortran programs translated by f2c will have MAIN__ instead
       * of "main".
       */
      if ((0 == strcmp(cnode->name, "main"))
          || (0 == strcmp(cnode->name, "MAIN__"))) {
         mainNode = cnode;
         mainNode->InlineEligible = FALSE;
         break;
      }
      cnode = cnode->next_func;
   } /* end while (NULL != cnode) */

   if (NULL == mainNode) {
      (void) fprintf(stderr, "Cannot find main() function !!\n");
      (void) fprintf(stderr, "Aborting data placement optimization.\n");
      return NULL;
   }

   (void) time(&StartTime);

   outerChange = FALSE;
#if SKIP_INLINING
   /* skip inlining entirely. */
   return mainNode;
#endif

   /* Let's try to break any direct recursions by inlining the recursive
    * function ONCE into itself at each recursive call site.
    */
   cnode = callGraph;
   do {
      if (cnode->library_func)
         break;

      if (cnode->is_directly_recursive && !cnode->InlinedIntoCallers
          && (0 < cnode->CallSiteCount)) {
         /* We have a non-leaf function. Search it for direct recursions
          * and perform a single copy of its own CFG into itself at the
          * call site, removing the self-call in the copy.
          */
         ResetNodeFlag(cnode->topCFG);
         if (EliminateDirectRecursion(cnode)) {
#if DEBUG_INLINING
            (void) fprintf(stderr, "Direct recursion removed for %s\n",
                           cnode->name);
            (void) fflush(stderr);
#endif
            outerChange = TRUE;
            (void) time(&CurrentTime);
            ElapsedSeconds = difftime(CurrentTime, StartTime);
            if (INLINING_TIME_LIMIT < ElapsedSeconds) {
               (void) fprintf(stderr,
                           "Time limit reached: InlineFunctionsIntoMain()\n");
               break;
            }
         }
      }
      cnode = cnode->next_func;
   } while (NULL != cnode);

   if (outerChange) {
      /* Deallocate static data in InlineNodeInCFGNode() */
      (void) InlineNodeInCFGNode(NULL, NULL, NULL);
   }

   do {
      outerChange = FALSE;
      (void) fflush(stderr);
      RemainingFuncs = ScanRemainingFunctions(mainNode, &MinCallSiteCount);
      do {
         change = FALSE;
         cnode = callGraph;
         do {
            if (cnode->library_func)
               break;

            if (cnode->InlineEligible && !cnode->InlinedIntoCallers
                && (0 == cnode->CallSiteCount)) {
               /* We have a function that was either a leaf, or has become
                * a leaf node through inlining, and it has not yet been inlined
                * into the CFGs of all of its callers.  It is time to inline
                * it into the CFGs of all callers, IF it is small enough.
                */
               change |= InlineLeafNode(cnode);
               (void) time(&CurrentTime);
               ElapsedSeconds = difftime(CurrentTime, StartTime);
               if (INLINING_TIME_LIMIT < ElapsedSeconds) {
                  (void) fprintf(stderr,
                            "Time limit reached: InlineFunctionsIntoMain()\n");
                  break;
               }
            }
            cnode = cnode->next_func;
         } while (NULL != cnode);
         outerChange |= change;
      } while (change);
   } while (outerChange);

   (void) fflush(stderr);

   /* If the call site count for main() has gone to zero, we have avoided
    * indirect recursion loops and are finished.
    */
   if (0 == mainNode->CallSiteCount) {
      return mainNode;
   }

   /* Otherwise, we have to break recursion chains in order to complete as
    * much inlining as possible.  To do this, we inline a function from the
    * indirectly recursive chain into all its callers, and then remove
    * any directly recursive calls that are created by the inlining.
    * We try to work bottom up from functions that have
    * only one call site, only two call sites, etc.
    */
   LoopIterations = 0;
   while (0 < mainNode->CallSiteCount) {
      RemainingFuncs = ScanRemainingFunctions(mainNode, &MinCallSiteCount);
      /* If the node budget is about to cause us to terminate, even though
       * we are far short of our time limit, then we are obviously not
       * exhausting virtual memory yet, which was the point of these
       * budgets.  Double the node budget.
       */
      (void) time(&CurrentTime);
      ElapsedSeconds = difftime(CurrentTime, StartTime);
#if ALLOW_NODE_BUDGET_GROWTH
      if ((0 == RemainingFuncs) 
          && (ElapsedSeconds < (INLINING_TIME_LIMIT * 0.5))) {
         CFGNodeBudget *= 2;
         InlineLimitDoubled = TRUE;
#if DEBUG_INLINING
         (void) fprintf(stderr, "Doubling node budget to: %d\n",
                        CFGNodeBudget);
         (void) fprintf(stderr, "Elapsed seconds: %12.3f\n",
                        ElapsedSeconds);
#endif
         RemainingFuncs = ScanRemainingFunctions(mainNode, &MinCallSiteCount);
      }
#endif

#if DEBUG_INLINING
      (void) fprintf(stderr, "Funcs remaining to inline: %d\n",
                     RemainingFuncs);
      PrintLocalTime();
      (void) fflush(stderr);
#endif
      if (0 == RemainingFuncs) {
         break;
      }
      if (MAX_INLINING_ITERATIONS < LoopIterations++) {
         (void) fprintf(stderr,
                       "Iteration limit reached: InlineFunctionsIntoMain()\n");
         break;
      }
      if (INLINING_TIME_LIMIT < ElapsedSeconds) {
         (void) fprintf(stderr,
                        "Time limit reached: InlineFunctionsIntoMain()\n");
         break;
      }

      change = FALSE;
      cnode = callGraph;
      CallSiteTarget = MinCallSiteCount;
      do {
         if (cnode->library_func)
            break;
         if (cnode->InlineEligible && !cnode->InlinedIntoCallers) {
            assert(cnode != mainNode); /* main() is never eligible. */
            if (CallSiteTarget == cnode->CallSiteCount) {
               /* We have a function that has the right number of calls
                * remaining, and it has not yet been inlined
                * into the CFGs of all of its callers.  It is time to inline
                * it into the CFGs of all callers, IF it is small enough.
                */
               change |= InlineRecursiveNode(cnode, StartTime);
               (void) time(&CurrentTime);
               ElapsedSeconds = difftime(CurrentTime, StartTime);
               if (INLINING_TIME_LIMIT < ElapsedSeconds) {
                  (void) fprintf(stderr,
                            "Time limit reached: InlineFunctionsIntoMain()\n");
                  break;
               }
            }
         }
         cnode = cnode->next_func;
      } while (NULL != cnode);

      if (change) {
         /* See if we created any leaf nodes and inline them here. */
         change = FALSE;
         cnode = callGraph;
         do {
            if (cnode->library_func)
               break;
            if (cnode->InlineEligible && !cnode->InlinedIntoCallers
                && (0 == cnode->CallSiteCount)) {
               /* We have a function that was either a leaf, or has become
                * a leaf node through inlining, and it has not yet been inlined
                * into the CFGs of all of its callers.  It is time to inline
                * it into the CFGs of all callers, IF it is small enough.
                */
               assert(cnode != mainNode); /* main() is never eligible. */
               change |= InlineLeafNode(cnode);
               (void) time(&CurrentTime);
               ElapsedSeconds = difftime(CurrentTime, StartTime);
               if (INLINING_TIME_LIMIT < ElapsedSeconds) {
                  (void) fprintf(stderr,
                            "Time limit reached: InlineFunctionsIntoMain()\n");
                  break;
               }
            }
            cnode = cnode->next_func;
         } while (NULL != cnode);
      } /* end if (change) ... */
      (void) fflush(stderr);
   } /* end while (0 < mainNode->CallSiteCount) */

#if DEBUG_INLINING
   (void) time(&CurrentTime);
   ElapsedSeconds = difftime(CurrentTime, StartTime);
   (void) fprintf(stderr, "Inlining loop time: %12.3f\n", ElapsedSeconds);
   (void) fflush(stderr);
#endif

#if 0
   /* Might have functions only called by pointers. */
   if (0 < mainNode->CallSiteCount)
#endif
      ComputeStaticCallEstimates();


   return mainNode;
} /* end of InlineFunctionsIntoMain() */

/*
 * Compute the total size of this object with respect to coalescing.  If it is
 * not first in any coalesced pair, return its own size.  If it is first in a
 * pair, follow its least-significant neighbor chains until we reach the end,
 * adding the sizes of its neighbors as we go along.
 */
int GlobalChainSize(struct globalsym *global)
{
   int   Size;

   assert(NULL != global);
   Size = global->size + global->PadBytes;
   while (NULL != global->CoalescedLSNeighbor) {
      global = global->CoalescedLSNeighbor;
      Size += (global->size + global->PadBytes);
   }

   return Size;
}

/*
 * GetInnerLoopIterations - Return the iteration count of the innermost loop
 * in a loop nest.
 */
int GetInnerLoopIterations(struct CFGloopl *LoopList) {
   struct CFGloopl *TempLoop = LoopList;

   assert(NULL != LoopList);

   /* The LoopList is sorted from outer to inner, so just find the last
    * entry in the list.
    */
   while (NULL != TempLoop->next)
      TempLoop = TempLoop->next;

   return TempLoop->iterations;
} /* end of GetInnerLoopIterations() */

/*
 * ComputeLoopIterationOverlap - For a pair of loop lists, divide up the
 * iterations into the shared and unshared (second loop list only) iterations.
 */
void ComputeLoopIterationOverlap(struct CFGloopl *Loops1,
                                 struct CFGloopl *Loops2,
                                 int *SharedIterations,
                                 int *SecondOnlyIterations) {
   struct CFGloopl *TempLoop1 = Loops1;
   struct CFGloopl *TempLoop2 = Loops2;
   int SharedFlag = FALSE;
   int SecondOnlyFlag = FALSE;

   assert(NULL != Loops1);
   assert(NULL != Loops2);

   *SharedIterations = 1;
   *SecondOnlyIterations = 1;

   while (NULL != TempLoop2) {
      if (NULL == TempLoop1) {
         /* Loops2 has remaining items after Loops1 has been processed. */
         (*SecondOnlyIterations) *= TempLoop2->iterations;
         SecondOnlyFlag = TRUE;
         TempLoop2 = TempLoop2->next;
      }
      else if (TempLoop1->loopID == TempLoop2->loopID) {
         (*SharedIterations) *= TempLoop1->iterations;
         SharedFlag = TRUE;
         TempLoop1 = TempLoop1->next;
         TempLoop2 = TempLoop2->next;
      }
      else if (TempLoop1->blockCount > TempLoop2->blockCount) {
         TempLoop1 = TempLoop1->next;
      }
      else if (TempLoop1->blockCount == TempLoop2->blockCount) {
         /* An unshared pair of loops has the same number of blocks by
          * coincidence.  All we care about is the fact that the loops
          * are unshared.
          */
         (*SecondOnlyIterations) *= TempLoop2->iterations;
         SecondOnlyFlag = TRUE;
         TempLoop1 = TempLoop1->next;
         TempLoop2 = TempLoop2->next;
      }
      else {
         /* Must be TempLoop1->blockCount < TempLoop2->blockCount */
         (*SecondOnlyIterations) *= TempLoop2->iterations;
         SecondOnlyFlag = TRUE;
         TempLoop2 = TempLoop2->next;
      }
   } /* end while (NULL != TempLoop2) */

   if (!SharedFlag)
      *SharedIterations = 0;
   if (!SecondOnlyFlag)
      *SecondOnlyIterations = 0;

   return;
} /* end of ComputeLoopIterationOverlap() */

/*
 * IntraBlockGlobalAffinity -
 * Increment the affinity array entries for globals that are used within
 * a single block.  This happens at the beginning of each path that we
 * follow through the control flow graph.
 */
int IntraBlockGlobalAffinity(struct GlobalPlacementCFGNode *CFGnode,
                             float topEstimate, int NumGlobs, 
                             struct globalsym *Globals[], float *AffinityArray)
{
   struct varl *var1;
   struct varl *var2;
   struct varl *temp1;
   struct globalsym *global;
   int AccumSize = 0;  /* Size of all globals in this block. */
   int WindowSize;
   int index;
   int SaveIndex;

   /* These next two are static to save memory allocations and time. */
   static int *GlobalsUsed = NULL;
     /* Boolean array; TRUE -> found in this CFG node. */
   static int count = 0;

   assert(NULL != CFGnode);
   assert(NULL != CFGnode->prof);
   var1 = CFGnode->prof->FirstGlobal;
   assert(NULL != var1);

   if ((count < GlobalsCount) && (count > 0)) {
      /* Need to reallocate GlobalsUsed[] with a bigger size than some
       * previous function in this file needed.
       */
      FREE(GlobalsUsed);
   }

   if (NULL == GlobalsUsed) {
      GlobalsUsed = (int *) SALLOC((int) sizeof(int) * GlobalsCount);
      count = GlobalsCount;
   }
   if (NULL == GlobalsUsed) {
      (void) fprintf(stderr, "Out of virtual memory: "
                     "regs.c::IntraBlockGlobalAffinity() \n");
      exit(EXIT_FAILURE);
   }

   bzero(GlobalsUsed, GlobalsCount * sizeof(int));

   /* Now that we do not inline all functions into main(), but stop at
    * a size limit, we might still encounter function calls, indicated
    * by an index of -1.  Skip over these.  For all others, count the
    * total size we will encounter. If it is zero, return.
    */
   while (NULL != var1) {
      if (0 <= var1->index) {
         global = GetLeftmostInGlobalChain(Globals[var1->index]);
         index = GetGlobalsIndex(global, Globals, GlobalsCount);
         assert(index < GlobalsCount);
         if (!GlobalsUsed[index]) {
            AccumSize += GlobalChainSize(global);
            GlobalsUsed[index] = TRUE;
         }
      }
      var1 = var1->next;
   }

   if (0 == AccumSize)
      return AccumSize;

   /* Now, use GlobalsUsed[] to record which globals
    * have been used to start a path profile within
    * this block.
    */
   bzero(GlobalsUsed, GlobalsCount * sizeof(int));
   var1 = CFGnode->prof->FirstGlobal;
   while ((NULL != var1) && (0 > var1->index))
      var1 = var1->next;

   do {
      global = GetLeftmostInGlobalChain(Globals[var1->index]);
      index = GetGlobalsIndex(global, Globals, GlobalsCount);
      assert(index < GlobalsCount);
      if (GlobalsUsed[index]) {
         /* We have already used this global to start a path
          * within this basic block. Move on to the next global.
          */
         var1 = var1->next;
         while ((NULL != var1) && (0 > var1->index))
            var1 = var1->next;
         continue;
      }
      GlobalsUsed[index] = TRUE;
      var2 = var1->next;
      while ((NULL != var2) && (0 > var2->index))
         var2 = var2->next;  /* Skip over function calls. */

      SaveIndex = index;
      temp1 = var1;
      while (NULL != var2) {
         global = GetLeftmostInGlobalChain(Globals[var2->index]);
         index = GetGlobalsIndex(global, Globals, GlobalsCount);
         assert(index < GlobalsCount);
         if (index != SaveIndex) {
#if 1
            /* Experimenting with not requiring a repetition of globals
             * before we record affinity.  If global1 and global2 are accessed
             * once each in a basic block, and we loop around to that
             * block, they can thrash each other even if they only appear
             * once each in the block.  Using SaveIndex was keeping us
             * from recording any affinity.  With this experimental code,
             * the continued use of SaveIndex in the else case below will
             * give extra weight to intrablock affinity that does not depend
             * on looping back to this block.
             */
            *(AffinityArray + (NumGlobs * SaveIndex) + index) 
                += (CFGnode->estimate * topEstimate);
#endif
         }
         else  {  /* index == SaveIndex */
            /* Within our window, we have encountered the global "var1"
             * twice.  Everything in between the most recent prior
             * occurrence (pointed to by temp1) and the current occurrence
             * (var2) should be marked as having extra affinity (potential
             * cache conflict) with var1.
             */
            for (temp1 = temp1->next;
                 (NULL != temp1) && (temp1 != var2);
                 temp1 = temp1->next) {
               if (0 > temp1->index)
                  continue;  /* Skip function calls. */
               global = GetLeftmostInGlobalChain(Globals[temp1->index]);
               index = GetGlobalsIndex(global, Globals, GlobalsCount);
               assert(index < GlobalsCount);
               assert(index != SaveIndex);
               *(AffinityArray + (NumGlobs * SaveIndex) + index)
                   += (CFGnode->estimate * topEstimate);
            }
            assert(temp1 == var2);
         }

         var2 = var2->next;
         while ((NULL != var2) && (0 > var2->index))
            var2 = var2->next;  /* Skip over function calls. */
      } /* end while (NULL != var2) */
      var1 = var1->next;
      while ((NULL != var1) && (0 > var1->index))
         var1 = var1->next;
   } while (NULL != var1);

   return AccumSize;
} /* end of IntraBlockGlobalAffinity() */

/*
 * UpdateGlobalVariableAffinity - Starting with the current control flow
 * graph node (CFGnode) and the starting set of global variables (PrevGlobals)
 * and the accumulated weight of the path to CFGnode from
 * the starting block (IncomingWeight), increment the AffinityArray entries
 * that match the PrevGlobals to the globals seen in CFGnode.  Terminate if we
 * see a total size of globals in our path that exceeds MaxSize, or if
 * the accumulated path weight drops below MinWeight, or if we reach the end
 * of the function, or our recursion depth exceeds a defined limit.  Otherwise,
 * recursively follow the path.
 */
void UpdateGlobalVariableAffinity(struct GlobalPlacementCFGNode *CFGnode,
                                  struct varl *PrevGlobals, int AccumSize,
                                  float MinWeight, int MaxSize, int *depth,
                                  int NumGlobs, struct globalsym *Globals[],
                                  float topEstimate, float *AffinityArray) {
   struct varl *OldVar;
   struct varl *NewVar;
   struct globalsym *global1;
   struct globalsym *global2;
   int index1, index2;
   int WindowSize;
   float ScaleDown;
   int InnerLoopIterations, SharedIterations, OtherIterations;

   /* Keep track of the loop list at the start of the chain. */
   static struct CFGloopl *PrevLoops = NULL;

   assert(NULL != CFGnode);
   assert(0.0 < topEstimate);

#if DEBUG_PROFILING3
#if 0
   (void) fprintf(stderr, "%x  %d\n", CFGnode, CFGnode->NodeNum);
   (void) fflush(stderr);
#endif
#endif

   /* End our recursion when we exceed the size threshold, or fall below
    * the weight threshold, which means the weights are getting small enough
    * to be irrelevant to the optimization, or when we reach the recursion
    * depth limit (to avoid overflowing the stack).
    */
#if DEBUG_PROFILING2
   if (MaxSize >= L2_DATA_CACHE_SIZE) {
      if (AccumSize >= MaxSize)
         (void) fprintf(stderr, "Profiling size limit reached.\n");
      if ((CFGnode->estimate * topEstimate) < MinWeight)
         (void) fprintf(stderr, "Profiling threshold reached.\n");
   }
#endif
#if DEBUG_PROFILING
   if (*depth > MAX_PROFILING_RECURSION_DEPTH)
      (void) fprintf(stderr, "Profiling recursion limit reached.\n");
#endif

   if (CFGnode->done || (AccumSize >= MaxSize)
       || ((CFGnode->estimate * topEstimate) < MinWeight)
       || (*depth > MAX_PROFILING_RECURSION_DEPTH)) {
      --(*depth);
      return;
   }

   /* If we are processing this node, set its "done" flag to avoid
    * duplicate processing.  NOTE that these flags must be cleared to
    * FALSE over the entire function before any new path is started.
    * This is handled in PlaceGlobalData().
    */
   CFGnode->done = TRUE;

   /* If PrevGlobals is NULL, we have not yet gotten started on a chain.
    * Register allocation might have eliminated all the loads and stores
    * in this block.  Keep trying to start the chain.
    */
   if (NULL == PrevGlobals) {
      assert(0 == AccumSize);
      if ((NULL != CFGnode->prof->FirstGlobal)
          && (NULL != CFGnode->LoopList)) {
         /* We can start the list of globals with the current block. */
         PrevGlobals = CFGnode->prof->FirstGlobal;
         PrevLoops = CFGnode->LoopList;
         AccumSize = IntraBlockGlobalAffinity(CFGnode, topEstimate, NumGlobs,
                                              Globals, AffinityArray);
      }
#if 1
      /* With the for loop at the bottom of PlaceGlobalData(), there is
       * no reason to start at the current block if it has no globals.
       * CFGnode->left and CFGnode->right will get their turn to start chains
       * of globals later in that for loop, so recursing down into them now
       * will cause duplicate counting.
       */
      if (0 == AccumSize) {
         --(*depth);
         return;
      }
#endif
      if ((NULL != CFGnode->left) && (!CFGnode->prof->LeftBackEdgeFlag)) {
         ++(*depth);
         UpdateGlobalVariableAffinity(CFGnode->left, PrevGlobals, AccumSize,
                                      MinWeight, MaxSize, depth, NumGlobs,
                                      Globals, topEstimate, AffinityArray);
      }
      if ((NULL != CFGnode->right) && (!CFGnode->prof->RightBackEdgeFlag)) {
         ++(*depth);
         UpdateGlobalVariableAffinity(CFGnode->right, PrevGlobals, AccumSize,
                                      MinWeight, MaxSize, depth, NumGlobs,
                                      Globals, topEstimate, AffinityArray);
      }
   }
   else { /* We are in the middle of a chain of globals. */
      OldVar = PrevGlobals;
      while (NULL != OldVar) {
         assert(0 <= OldVar->index); /* No function calls in PrevGlobals. */
         OldVar = OldVar->next;
      }
      OldVar = PrevGlobals;
      InnerLoopIterations = GetInnerLoopIterations(CFGnode->LoopList);
      assert(0.0 < InnerLoopIterations);
      ComputeLoopIterationOverlap(PrevLoops, CFGnode->LoopList,
                                  &SharedIterations, &OtherIterations);
      do {
         /* We need to have each CFG node (starting and current nodes) be
          * inside loops to do meaningful profiling. If the current node has
          * no loop list, just get on to its children as quickly as possible.
          */
         if (NULL == CFGnode->LoopList)
            break;

         global1 = GetLeftmostInGlobalChain(Globals[OldVar->index]);
         index1 = global1->AffinityIndex;
         assert(index1 < GlobalsCount);
         assert(0 <= index1);
         WindowSize = AccumSize;
         NewVar = CFGnode->prof->FirstGlobal;
#if INTERPROCEDURAL_PROFILING  /* !!!! Design not finished. */
         if ((NULL != NewVar) && (0 > NewVar->index)) { /* Function call. */
            assert(0 <= ProfilingStackPointer);
            assert(NULL != NewVar->cgnode);
            assert(NULL != NewVar->cgnode->topCFG);
            ProfilingStack[ProfilingStackPointer]->node = CFGnode;
            ProfilingStack[ProfilingStackPointer]->call = NewVar;
            ProfilingStack[ProfilingStackPointer++]->depth = 1;
            ++(*depth);
            UpdateGlobalVariableAffinity(NewVar->cgnode->topCFG, PrevGlobals,
                                         AccumSize, MinWeight, MaxSize, depth,
                                         NumGlobs, Globals,
                                         topEstimate * CFGnode->estimate
                                            * NewVar->cgnode->topEstimate,
                                         AffinityArray);
         }
#else
         while ((NULL != NewVar) && (0 > NewVar->index))
            NewVar = NewVar->next;  /* Skip function calls. */
#endif
         while (NULL != NewVar) {
            global2 = GetLeftmostInGlobalChain(Globals[NewVar->index]);
            index2 = global2->AffinityIndex;
            assert(index2 < GlobalsCount);
            assert(0 <= index2);
            WindowSize += GlobalChainSize(global2);
            if (WindowSize > MaxSize)
               break;
            if (0 != OtherIterations) {
               /* If there are no shared iterations, then we have crossed
                * outer loop boundaries and do not want to record any
                * affinity.  We could consider asserting that this will
                * never happen, as our thresholds should have prevented
                * this.
                */
               if (0 == SharedIterations)
                  ScaleDown = 0.0;
               else {
                  /* We have both shared and unshared loops. We cannot
                   * thrash on each element within a line, as the two
                   * nodes do not share the same inner loop, so we need
                   * to scale down by an estimate of the elements per
                   * load unit (subblock; may be same as block size on
                   * a machine with no subblocking). We assume that the
                   * inner loop iterations for the later CFG node are
                   * the number of iterations needed to move through the
                   * entire data object; this is often the case, and we
                   * are not always able to analyze the exact number of
                   * iterations in a loop, so we don't want to make too
                   * much distinction between loops that were and were
                   * not analyzable.  All unshared iterations for the second
                   * CFG node that are not the innermost loop iterations
                   * need to be factored out.
                   */
                  ScaleDown = (INTS_PER_SUBBLOCK *
                     ((float) OtherIterations / (float) InnerLoopIterations));
                  ScaleDown = 1.0 / ScaleDown;
               }
            }
            else {
               /* As PrevLoops != NULL, there must be SharedIterations. */
               assert(0 < SharedIterations);
               /* They share the entire loop nest, down to the innermost loop
                * of CFGnode->LoopList, so there is no scaling down to be done.
                */
               ScaleDown = 1.0;
            }
            *(AffinityArray + (NumGlobs * index1) + index2) 
                += (CFGnode->estimate * ScaleDown * topEstimate);
            NewVar = NewVar->next;
            while ((NULL != NewVar) && (0 > NewVar->index))
               NewVar = NewVar->next;  /* Skip function calls. */
         }
         OldVar = OldVar->next;
         while ((NULL != OldVar) && (0 > OldVar->index))
            OldVar = OldVar->next;  /* Skip function calls. */
      } while (NULL != OldVar);
      AccumSize = WindowSize;
      /* Now we recurse down the control flow graph. */
      if ((NULL != CFGnode->left) && (!CFGnode->prof->LeftBackEdgeFlag)) {
         ++(*depth);
         UpdateGlobalVariableAffinity(CFGnode->left, PrevGlobals, AccumSize,
                                      MinWeight, MaxSize, depth, NumGlobs,
                                      Globals, topEstimate, AffinityArray);
      }
      if ((NULL != CFGnode->right) && (!CFGnode->prof->RightBackEdgeFlag)) {
         ++(*depth);
         UpdateGlobalVariableAffinity(CFGnode->right, PrevGlobals, AccumSize,
                                      MinWeight, MaxSize, depth, NumGlobs,
                                      Globals, topEstimate, AffinityArray);
      }
   } /* end if (NULL == PrevGlobals) ... else ... */

   --(*depth);
   return;
} /* end of UpdateGlobalVariableAffinity() */

/*
 * ComputeStaticCallEstimates - Increase the topEstimate field for each
 * call graph node in the call graph by the size of the estimate fields
 * within all CFG nodes that call that function.
 */
void ComputeStaticCallEstimates() {
   struct call_graph_node *cgnode = the_call_graph;
   struct GlobalPlacementCFGNode *CFGnode;
   struct varl *tempVar;
   int count;

   assert(NULL != cgnode);
   while (NULL != cgnode) {
      if (cgnode->library_func)
         break;

      if (cgnode->has_data && !cgnode->IrreducibleCFG
          && !cgnode->InlinedIntoCallers
#if !CONSTANT_CALLS
          && (0 < cgnode->CallSiteCount)
#endif
         ) {
#if CONSTANT_CALLS
         /* Experiment with setting all topEstimate fields to a
          * constant value.
          */
         cgnode->topEstimate = 1.0;
#else
         CFGnode = cgnode->topCFG;
         count = 0;
         while ((NULL != CFGnode) && (count < cgnode->CallSiteCount)) {
            assert(NULL != CFGnode->prof);
            tempVar = CFGnode->prof->FirstGlobal;
            while (NULL != tempVar) {
               if (NULL != tempVar->cgnode) {
                  assert(0 > tempVar->index);
                  tempVar->cgnode->topEstimate += CFGnode->estimate;
                  ++count;
               }
               tempVar = tempVar->next;
            }
            CFGnode = CFGnode->next;
         }
#endif
      }
      cgnode = cgnode->next_func;
   }

   return;
} /* end of ComputeStaticCallEstimates() */

/*
 * RecordAffinityIndex - Go through the Globals[] array and record the
 * array index in the AffinityIndex field of each global.  After we shuffle
 * the globals list repeatedly during placement, the AffinityIndex field
 * will have remained constant and can always be used to index into
 * the AffinityArray[], while the index into Globals[] will stop corresponding
 * to the indices into the AffinityArray[] as we reorder globals.
 */
void RecordAffinityIndex() {
   int index;

   assert(NULL != Globals);
   assert(0 < GlobalsCount);

   for (index = 0; index < GlobalsCount; ++index)
      Globals[index]->AffinityIndex = index;

   return;
} /* end of RecordAffinityIndex() */

/* 
 * RecordCurrentCallGraphNode - Now that the end of the function has been
 * reached and a new call_graph_node has been created, we can associate
 * our saved CFG with that call_graph_node.
 */
void RecordCurrentCallGraphNode(struct call_graph_node *cgnode) {
   assert(NULL != cgnode);

   cgnode->topCFG = savedCFGNode;
   cgnode->IrreducibleCFG = IrreducibleFlag;
   cgnode->blockCount = NumBasicBlocks;
   if (!IrreducibleFlag)
      PruneCFG(cgnode);
   else {
      FREE(cgnode->topCFG);
      cgnode->InlineEligible = FALSE;
#if DEBUG_INLINING || DEBUG_PROFILING || DEBUG_PLACEMENT
      (void) fprintf(stderr, "Irreducible CFG for %s\n", cgnode->name);
      (void) fflush(stderr);
#endif
   }
   return;
}

/* SortLibraryFuncsToEnd - 
 * Sort all of the library functions to the end of the call graph list
 * so that we can iterate through less of the list in all future
 * processing.
 */
void SortLibraryFuncsToEnd() {
   struct call_graph_node *FirstLibNode = the_call_graph;
   struct call_graph_node *cgnode;
   struct call_graph_node *tempNode;
   struct call_graph_node *tempNode2;
   int LibSeen = FALSE;

   assert(NULL != FirstLibNode);

   /* Find the first library function. */
   while (!FirstLibNode->library_func) {
      FirstLibNode = FirstLibNode->next_func;
      assert(NULL != FirstLibNode); /* Must have -lib_info. */
   }

   /* Get some non-library node in the first position in the list. */
   if (the_call_graph == FirstLibNode) {
      tempNode = the_call_graph;
      do {
         tempNode = tempNode->next_func;
         assert(NULL != tempNode); /* Must have at least a main(). */
      } while (tempNode->library_func);
      /* tempNode is now the first non-lib function. */

      cgnode = FirstLibNode;
      while (cgnode != tempNode) {
         /* Move all library functions at the beginning of the list
          * to the position after the first non-lib function.
          */
         tempNode2 = cgnode->next_func;
         cgnode->next_func = tempNode->next_func;
         tempNode->next_func = cgnode;
         cgnode = tempNode2;
      }
      the_call_graph = tempNode;
      FirstLibNode = tempNode->next_func;
   } /* end if (the_call_graph == FirstLibNode) */

   cgnode = the_call_graph->next_func;
   tempNode = the_call_graph;
   while (NULL != cgnode) {
      if (cgnode->library_func) {
         if (!LibSeen) {
            LibSeen = TRUE;
            assert(cgnode == FirstLibNode);
         }
         tempNode = cgnode;
         cgnode = cgnode->next_func;
      }
      else if (!LibSeen) {
         tempNode = cgnode;
         cgnode = cgnode->next_func;
      }
      else { /* Non-lib func after LibSeen has been set. */
         /* Move cgnode to second position in list. */
         tempNode->next_func = cgnode->next_func;
         cgnode->next_func = the_call_graph->next_func;
         the_call_graph->next_func = cgnode;
         cgnode = tempNode->next_func;
      }
   } /* end while (NULL != cgnode) */
   return;
} /* end of SortLibraryFuncsToEnd() */

#define NODE_COUNT_BUCKETS  181

/*
 * PlaceGlobalData - Finish profiling global data variables and place them
 * to minimize cache misses and addressing costs.
 */
void PlaceGlobalData(struct call_graph_node *callGraph) {
   int GlobalsIndex;
   int SmallNodeCount = 0;
   int MediumNodeCount = 0;
   int LargeNodeCount = 0;
   int NumNodes;
   int RecursionDepth = 0;
   int NodeCounts[NODE_COUNT_BUCKETS];
   int NodeIndex;
   int NodeTotal;
   int row, col;
   float threshold[3] = {0.0, GLOBAL_PROFILING_THRESHOLD1,
                         GLOBAL_PROFILING_THRESHOLD2};
   float cutoff = 1.2;
   struct call_graph_node *mainNode;
   struct call_graph_node *cgnode;
   struct GlobalPlacementCFGNode *CFGnode;

#if DEBUG_PROFILING || DEBUG_INLINING || DEBUG_PLACEMENT
   (void) fprintf(stderr, "Starting PlaceGlobalData: ");
   PrintLocalTime();
#if NO_PROFILE_SIZE_LIMIT
   (void) fprintf(stderr, "No size limit for global profiling.\n");
#endif
#if IGNORE_STORES
   (void) fprintf(stderr, "Using only loads in global profiling.\n");
#else
   (void) fprintf(stderr, "Using loads and stores in global profiling.\n");
#endif
#if SKIP_INLINING
   (void) fprintf(stderr, "Skipping all function inlining.\n");
#else
   (void) fprintf(stderr, "Node budget: %d\n", CFGNodeBudget);
   (void) fprintf(stderr, "Inlining time limit: %12.3f\n",
                  INLINING_TIME_LIMIT);
#endif
#if CONSTANT_CALLS
   (void) fprintf(stderr, "topEstimate fields set to 1.0.\n");
#endif
#if CONVERT_RODATA_TO_DATA
   (void) fprintf(stderr, "Merging read-only data into data section.\n");
#endif
#if CONVERT_COMMON_TO_DATA
   (void) fprintf(stderr, "Merging uninitialized data into data section.\n");
#endif
#endif

   GlobalsCount = CompactGlobalsList(L2_DATA_CACHE_SIZE);
   if (0 >= GlobalsCount)
      return;

   Globals = (struct globalsym **)
      SALLOC_PERM(GlobalsCount * sizeof(struct globalsym *));
   if (NULL == Globals) {
      (void) fprintf(stderr, "Out of virtual memory: PlaceGlobalData()\n");
      exit(EXIT_FAILURE);
   }

   GlobalsIndex = BuildGlobalsArray(L2_DATA_CACHE_SIZE);
   assert(GlobalsIndex == GlobalsCount);
   RecordAffinityIndex();

   /* Sort all of the library functions to the end of the call graph list
    * so that we can iterate through less of the list in all future
    * processing.
    */
   SortLibraryFuncsToEnd();
   callGraph = the_call_graph; /* Changed when sorted. */

   /* Go back through all of the CFG nodes and fill in the Globals[] indices
    * in the struct varl entries.  As many CFG nodes access no globals, we
    * will prune the CFG as we go.
    */
#if DEBUG_PROFILING || DEBUG_INLINING || DEBUG_PLACEMENT
   (void) fprintf(stderr, "Starting FillInGlobalIndices: ");
   PrintLocalTime();
#endif

   FillInGlobalIndices(callGraph);

   AffinityArray = 
      (float *) SALLOC_PERM(GlobalsCount * GlobalsCount * sizeof(float));
   if (NULL == AffinityArray) {
      (void) fprintf(stderr, "Out of virtual memory: PlaceGlobalData()\n");
      exit(EXIT_FAILURE);
   }
   for (row = 0; row < GlobalsCount; ++row)
      for (col = 0; col < GlobalsCount; ++col)
         *(AffinityArray + row * GlobalsCount + col) = 0.0;

#if DEBUG_PROFILING || DEBUG_INLINING || DEBUG_PLACEMENT
   (void) fprintf(stderr, "Starting function inlining: ");
   PrintLocalTime();
#endif

   mainNode = InlineFunctionsIntoMain(the_call_graph);
   assert(NULL != mainNode);

#if DEBUG_INLINING
   if (0 < mainNode->CallSiteCount) {
      (void) fprintf(stderr, "\nCallees for main:\n");
      ShowCallees(mainNode->topCFG);
   }
#endif

   /* Loop through the non-inlined CFGs and perform path profiling, starting
    * candidate paths at each CFG node that has globals in it and meets
    * certain execution frequency criteria.
    */
#if DEBUG_PROFILING || DEBUG_INLINING || DEBUG_PLACEMENT
   (void) fprintf(stderr, "Starting path profiling: ");
   PrintLocalTime();
#endif

   for (NodeIndex = 0; NodeIndex < NODE_COUNT_BUCKETS; ++NodeIndex)
      NodeCounts[NodeIndex] = 0;

   for (cgnode = the_call_graph; cgnode; cgnode = cgnode->next_func) {
      if (cgnode->library_func)
         break;
      if (!cgnode->has_data || cgnode->IrreducibleCFG
          || cgnode->InlinedIntoCallers)
         continue;
      NumNodes = NumberCFGNodes(cgnode->topCFG) - 1;
   
#if DEBUG_PROFILING
      (void) fprintf(stderr,
                     "Nodes for %s: %d Calls: %d Est: %10.2f Static: %d\n",
                     cgnode->name, NumNodes, cgnode->CallSiteCount,
                     cgnode->topEstimate, cgnode->static_calls);
#endif 
      /* Scale our profiling thresholds based on the number of CFG nodes
       * we have to deal with.  We want to start about MAX_PROFILING_PATHS
       * in our profiling loop below (i.e. we want that many LargeNodes).
       * Group the node counts by the range of their estimate fields, where
       * 0.0 <= estimate < 10.0 goes into bucket 0, 10.0 to 19.9999 into
       * bucket 1, etc., and all estimate fields >= 200.0 are counted in
       * bucket 20.
       */
      for (CFGnode = cgnode->topCFG; NULL != CFGnode; CFGnode = CFGnode->next) {
         if (NULL != CFGnode->prof->FirstGlobal) {
            NodeIndex = (int) (floor(CFGnode->estimate * cgnode->topEstimate)
                               / 10.0);
            if (NODE_COUNT_BUCKETS <= NodeIndex)
               NodeIndex = NODE_COUNT_BUCKETS - 1; 
               /* Biggest estimates are grouped together. */
            assert((0 <= NodeIndex) && (NODE_COUNT_BUCKETS > NodeIndex));
            ++NodeCounts[NodeIndex];
         }
      }
   } /* end for (cgnode = the_call_graph; ...) */

   /* Now, start with the last bucket and work downward until we determine how
    * far down we would need to go to get just more than MAX_PROFILING_PATHS
    * large nodes to start our paths.
    */
   NodeTotal = 0;
   NodeIndex = NODE_COUNT_BUCKETS - 1;
   while (NodeIndex > 0) {
      NodeTotal += NodeCounts[NodeIndex];
      if (MAX_PROFILING_PATHS < NodeTotal)
         break;
      --NodeIndex;
   }
   threshold[2] = ((float) NodeIndex) * 10.0;
   /* In order to prevent paths from flowing from one loop
    * to another loop, we should never permit the highest
    * threshold to fall as low as 1.0.  If two loops are
    * separated by a node of weight 1.0, the data in the
    * second loop might evict the data from the first loop
    * out of the cache, but there is obviously no outer loop
    * that carries us back to the first loop, else the node
    * that lies between the loops would have a weight equal
    * to at least the iteration count of that outer loop.
    * Thus, the two loops are independent and the eviction
    * of data was irrelevant.
    */
   if (1.2 > threshold[2])
      threshold[2] = 1.2;

   if (threshold[1] >= threshold[2])
      threshold[1] = 0.0;

   cutoff = PATH_CUTOFF_FRACTION * threshold[2];
   if (1.2 > cutoff)
      cutoff = 1.2;

#if DEBUG_PROFILING2
   for (NodeIndex = 0; NodeIndex < NODE_COUNT_BUCKETS; ++NodeIndex)
      (void) fprintf(stderr, "NodeCounts[%d] = %d \n", NodeIndex,
                     NodeCounts[NodeIndex]);
#endif
#if DEBUG_PROFILING
   (void) fprintf(stderr, "Path count target: %d\n", MAX_PROFILING_PATHS);
   (void) fprintf(stderr, "Path threshold: %8.2f \n", threshold[2]);
   (void) fprintf(stderr, "Path cutoff: %8.2f \n", cutoff);
   (void) fflush(stderr);
#endif

#if DEBUG_PROFILING3
   for (CFGnode = mainNode->topCFG; NULL != CFGnode; CFGnode = CFGnode->next) {
      if (NULL != CFGnode->prof->FirstGlobal) {
         float estimate = (CFGnode->estimate * mainNode->topEstimate);
         CFGnode->done = TRUE;
         if ((cutoff <= estimate)
             && FindCFGCycle(CFGnode, cutoff)) {
            (void) fprintf(stderr, "Cycle at: %x  %d\n", CFGnode,
                           CFGnode->NodeNum);
            (void) fflush(stderr);
         }
      }
   }
#endif

   for (cgnode = the_call_graph; cgnode; cgnode = cgnode->next_func) {
      if (cgnode->library_func)
         break;
      if (!cgnode->has_data || cgnode->IrreducibleCFG
          || cgnode->InlinedIntoCallers)
         continue;
      ResetNodeFlag(cgnode->topCFG);
      for (CFGnode = cgnode->topCFG; NULL != CFGnode; CFGnode = CFGnode->next) {
         assert(FALSE == CFGnode->done);
         if (NULL != CFGnode->prof->FirstGlobal) {
            float estimate = (CFGnode->estimate * cgnode->topEstimate);
            if (threshold[2] <= estimate) {
               ++LargeNodeCount;
               ++RecursionDepth;
#if DEBUG_PROFILING3
               (void) fprintf(stderr, "Starting path: %x\n", CFGnode);
               (void) fflush(stderr);
#endif
               UpdateGlobalVariableAffinity(CFGnode, NULL, 0, cutoff,
#if 0
                                            2 * L2_DATA_CACHE_SIZE,
#else
                                            2 * L1_DATA_CACHE_SIZE,
#endif
                                            &RecursionDepth, GlobalsCount,
                                            Globals, cgnode->topEstimate,
                                            AffinityArray);
               assert(0 == RecursionDepth);
               ResetNodeFlag(cgnode->topCFG);
            }
            else if (threshold[1] <= estimate) {
               ++MediumNodeCount;
               (void) IntraBlockGlobalAffinity(CFGnode, cgnode->topEstimate,
                                               GlobalsCount, Globals,
                                               AffinityArray);
            }
            else
               ++SmallNodeCount;
         }
      } /* end for (CFGnode = cgnode->topCFG; ...) */
      FreeCFG(cgnode);
   } /* end for (cgnode = the_call_graph; ...) */

#if DEBUG_PROFILING
   (void) fprintf(stderr, "Node count - Small: %d Medium: %d Large: %d\n",
                  SmallNodeCount, MediumNodeCount, LargeNodeCount);
#endif 

#if DEBUG_PROFILING || DEBUG_INLINING || DEBUG_PLACEMENT
   (void) fprintf(stderr, "Starting global data placement: ");
   PrintLocalTime();
   (void) fflush(stderr);
#endif

#if GLOBAL_DATA_PLACEMENT
#define DO_PLACEMENT  TRUE
#if DO_PLACEMENT
   if (0 < (MediumNodeCount + LargeNodeCount)) {
      if (!GlobalDataPlacement())
         (void) fprintf(stderr, "Global data placement did not succeed.\n");
   }
#if DEBUG_PROFILING || DEBUG_PLACEMENT
   else {
      (void) fprintf(stderr, "No paths profiled; skipping global placement\n");
   }
#endif  /* DEBUG_PROFILING || DEBUG_PLACEMENT */
#endif  /* DO_PLACEMENT */
#endif  /* GLOBAL_DATA_PLACEMENT */

#if DEBUG_PROFILING || DEBUG_INLINING || DEBUG_PLACEMENT
   (void) fprintf(stderr, "Finished global data placement: ");
   PrintLocalTime();
#endif

   (void) fflush(stderr);

   return;
} /* end of PlaceGlobalData() */

#endif  /* CALLGRAPH */
